unit OPJLoader;
{ Copyright (C) 2012 C M Western; see NLFit.pas for license}
{ Import Origin .OPJ files }
interface

uses Classes, ExpPlotLoader;

type

  TOPJFileLoader = class(TExperimentalPlotLoader)
  public
    function Load: Boolean; override;
  end;

implementation

uses SysUtils, Units, TextStrm, Math, MathSup;

function ReadWord(var p: PAnsiChar):Word;
begin
  Result :=
  {$IFDEF FPC_BIG_ENDIAN}
    SwapEndian(PWord(p)^);
  {$ELSE}
    PWord(p)^;
  {$ENDIF}
  Inc(p,2);
end;

function ReadInteger(var p: PAnsiChar):Integer;
begin
  Result :=
  {$IFDEF FPC_BIG_ENDIAN}
    SwapEndian(PInteger(p)^);
  {$ELSE}
    PInteger(p)^;
  {$ENDIF}
  Inc(p,4);
end;

function ReadSingle(var p: PAnsiChar):Double;
var
  v: Integer;
begin
  v := ReadInteger(p);
  Result := PSingle(@v)^;
end;

type
  Tm = record
    case Integer of
      1: (r8: Double);
      {$IFDEF FPC_BIG_ENDIAN}
      2: (hi, lo: Longint);
      {$ELSE}
      2: (lo, hi: Longint);
      {$ENDIF}
      3: (i8: Int64);
    end;

function ReadDouble(var p: PAnsiChar):Double;
{$IFDEF FPC_BIG_ENDIAN}
var
  v: Tm;
{$ENDIF}
begin
  {$IFDEF FPC_BIG_ENDIAN}
    v.i8 := SwapEndian(PInt64(p)^);
    Result := v.r8;
  {$ELSE}
    Result := PDouble(p)^;
  {$ENDIF}
  Inc(p,8);
end;

{ TOPJFileLoader }

function TOPJFileLoader.Load: Boolean;
var
  SheetStart, Eof: PAnsiChar;
  Version, LastnPoints: Integer;
  SheetNames: array of AnsiString;
  function ReadSheets(SheetName: AnsiString; FindRange: Boolean):Integer;
  var
    p, s, se: PAnsiChar;
    Size, DataType, DataTypeU, ValueSize, Signature, nPoints, i: Integer;
    Value: Double;
    Name, SpreadName, ColName: AnsiString;
    Found: Boolean;
  begin
    { One pass over available data }
    p := SheetStart;
    Size := ReadInteger(p);
    Inc(p);
    Result := -1;
    LastnPoints := 0;
    while (Size > 0) and (Size <= $8c) and (p+Size <= Eof) do begin
      s := p;
      Inc(p,$16);
      DataType := ReadWord(p);
      DataTypeU := Ord((s+$3f)^);
      ValueSize := Ord((s+$3d)^);
      p := s+$58;
      se := p + 24;
      while (se > p) and (se^ = #0) do
        Dec(se);
      SetString(Name, p, se-p+1);
      while (se > p) and (se^ <> '_') do
        Dec(se);
      if se^ = '_' then begin
        SetString(SpreadName,p,se-p);
        ColName := Copy(Name,se-p+2,Length(Name));
        Inc(s,Size+1);
        p := s;
        nPoints := ReadInteger(p);
        if (nPoints mod ValueSize <> 0) or (p+nPoints > Eof) then
          Connector.ReportError('Column size error',True);
        if nPoints > 0 then
          Inc(p);
        if SheetName = '' then begin
          { Find sheet names }
          Found := False;
          for i := 0 to High(SheetNames) do
            Found := Found or (SpreadName = SheetNames[i]);
          if not Found then begin
            SetLength(SheetNames, Length(SheetNames)+1);
            SheetNames[High(SheetNames)] := SpreadName;
          end;
          Inc(p, nPoints);
        end else if SheetName <> SpreadName then
          Inc(p, nPoints)
        else begin
          Inc(Result);
          if Result >= Length(ChannelDest) then
            SetnChannels(Result+1);
          LastnPoints := Max(LastnPoints, nPoints div ValueSize);
          for i := 0 to nPoints div ValueSize - 1 do begin
            Found := False;
            if ValueSize <= 8 then begin
              Value := ReadDouble(p);
              Found := True;
            end else if ((DataType and 256) = 256) and (p^ = #0) then begin
              Inc(p,2);
              Value := ReadDouble(p);
              Inc(p, ValueSize-10);
              Found := True;
              // -1.23456789e-300 is marker for unset values
              if IsNanOrInf(Value) or (Abs(Value+1.23456789e-300) < 1e-308) then
                Found := False;
            end else
              Inc(p, ValueSize); { Text - skip }
            if Found then
              if FindRange then
                AddChange(Result,Value)
              else
                SetValue(-1, Result, i, Value)
            else if not FindRange then
              SetValue(-1, Result, i, Nan);
          end;
          if not FindRange and (Connector is TExperimentalPlotLoaderConnector) then
            SetNameN(-1, Result, Name);
        end;
      end else begin
        { else matrix or function - skip }
        p := s + $58 + 25;
        Signature := ReadWord(p);
        p := s + Size + 1;
        Size := ReadInteger(p);
        Inc(p,Size+1);
        if (ValueSize <> 8) and (ValueSize <= 16) then
          Inc(p,2);
      end;
      Inc(p);
      Size := ReadInteger(p);
      Inc(p, Size + 1 + Ord(Size > 0));
      Size := ReadInteger(p);
      Inc(p);
    end;
  end;
var
  Position: Int64;
  Buffer: string[11];
  p: PAnsiChar;
  i, j, nCols, PlotOffset: Integer;
  Line: AnsiString;
begin
  Result := False;
  if (Ext = '') or (Ext = '.opj') then begin
    SetLength(buffer,11);
    if (Stream.Read(Buffer[1],11) = 11) and (Copy(buffer,1,4) = 'CPYA') and TryStrToInt(Trim(Copy(Buffer,8,4)), Version) then begin
      if buffer[6] = '3' then
        Version := 0;
    end else
      Exit; { Unrecognized }
  end else
    Exit; { Wrong extension }
  Stream.Position := Position;
  Line := '';
  repeat { Fill buffer with rest of file }
    i := Length(Line);
    SetLength(Line, i+100000);
    SetLength(Line, Stream.Read((PAnsiChar(Line)+i)^,100000)+i);
  until Length(Line) <> i+100000;
  p := PAnsiChar(Line);
  Eof := p + Length(Line);
  while (p < Eof) and (p^ <> #10) do
    Inc(p);
  Inc(p);
  i := ReadInteger(p);
  Inc(p, i+7);
  SheetStart := p;
  ReadSheets('',True); { Find sheet names }
  PlotOffset := 0;
  for i := 0 to High(SheetNames) do begin
    nCols := ReadSheets(SheetNames[i], True); { Find sheet names }
    if (nCols < 0) or (nValues[0] <= 1) then
      Continue; { Only one row of data }
    { Treat channel as valid if > 50% of the lines have data }
    for j := 0 to High(ChannelDest) do
      if (nValues[j] >= LastnPoints div 2) and (nValues[j] > 0) then
        nCols := j+1;
    SetLength(ChannelDest,nCols);
    SetnChannels(nCols);
    if nCols < 1 then
      Break;
    DoAutoDetectFrequency;
    CheckAvailableChannels(Length(ChannelDest), LastnPoints, AskAboutWeights);
    ReadSheets(SheetNames[i], False);
    Result := True;
    if i < High(SheetNames) then
      ResetConnector;
  end;
end;

initialization
  RegisterPlotLoader(TOPJFileLoader, 'Origin Files', '*.opj');
end.
