unit OpusLoader;
{ Copyright (C) 2016 C M Western; see NLFit.pas for license}
{ Import Bruker Opus spectrometer files; Initial implementaion based on inspection of files
  and web searches including Le Kuai (JPL) and Soil Spectroscopy and Chemometrics project
  (Andrew Sila and Tomislav Hengl )
}

interface

uses ExpPlotLoader, Classes;

type

  { TOpusLoader }

  TOpusLoader = class(TExperimentalPlotLoader)
  public
    function Load: Boolean; override;
  end;

implementation

uses SysUtils, MathSup;

{ TOpusLoader }

type

  TDirEnty = record
    Dtype: Integer;
    Dlength: Integer;
    Dstart: Integer;
  end;

function TOpusLoader.Load: Boolean;
var
  Buffer: string[4];
  ItemType, ItemSize: Integer;
  DirStart, DirSize, MaxDirSize: Integer;
  Dir: array of TDirEnty;
  nPoints: Integer;
  fXV, lXV: Double;
  Comment: string;
  function ValidExt(const Name:String): Boolean;
  var
    i: Integer;
  begin
    Result := False;
    if Name[1] <> '.' then
      Exit;
    for i := 2 to Length(Name) do
      if not (Name[i] in ['0'..'9']) then
        Exit;
    Result := True;
  end;
  function NextItem: AnsiString;
  begin
    SetLength(Buffer,3);
    Stream.Read(Buffer[1], 4); { Reads null as well }
    ItemType := NextInt16;
    ItemSize := NextInt16;
    if ItemSize > 0 then
      Result := Buffer
    else
      Result := 'END';
  end;
  procedure AddString;
  var
    s: AnsiString;
  begin
    SetLength(s, 2*ItemSize);
    Read(s[1], 2*ItemSize);
    while Pos(#0, s) > 0 do
      Delete(s, Pos(#0,s), 1);
    if s = '' then
      Exit;
    if Comment <> '' then
      Comment := Comment + ' ';
    Comment := Comment + s;;
  end;

  procedure SkipItem;
  begin
    Stream.Seek(ItemSize*2, soFromCurrent);
  end;
  function TrySpectrumRange(Entry, Offset: Integer):Boolean;
  var
    i: Integer;
  begin
    Result := (Entry+Offset < Length(Dir)) and (Entry+Offset >= 0);
    if not Result then
      Exit;
    Result := (Dir[Entry+Offset].Dtype shr 4 and $3f) = 1;
    if not Result then
      Exit;
    Result := ((Dir[Entry].Dtype shr 10) and $7f)
              = ((Dir[Entry+Offset].Dtype shr 10) and $7f);
    if not Result then
      Exit;
    nPoints := -1;
    SetNan(fXV);
    SetNan(lXV);
    Stream.Position := InitialPosition + Dir[Entry+Offset].Dstart;
    while NextItem <> 'END' do begin
      //WriteLn(Log^, IntToHex(Stream.Position, 8), ' ', Buffer);
      if Buffer = 'NPT' then
        nPoints := NextInt32
      else if Buffer = 'FXV' then
        fXV := NextDouble
      else if Buffer = 'LXV' then
        lXV := NextDouble
      else if (Buffer = 'DAT') or (Buffer = 'TIM') then
        AddString
      else
        SkipItem;
    end;
    Result := (nPoints > 0) and not IsNanOrInf(fXV) and not IsNanOrInf(lXV);
  end;
var
  i: Integer;
  Sample, Ratio: Integer; { Directory entries }
begin
  Result := False;
  SetLength(Buffer, 4);
  if (Ext = '') or ValidExt(Ext) then begin
    if (Stream.Read(Buffer[1],4) = 4) and (Buffer = #10#10#254#254) then begin
      NextDouble; { Version number }
      DirStart := NextInt32;
      MaxDirSize := NextInt32;
      DirSize := NextInt32;
      if (DirSize > MaxDirSize) or (DirSize > 50 { probably implausible }) then
        Exit;
    end else
      Exit; { Unrecognized }
  end else
    Exit; { Wrong extension }
  Stream.Position := InitialPosition + DirStart;
  SetLength(Dir, DirSize);
  Sample := -1;
  Ratio := -1;
  Comment := '';
  for i := 0 to DirSize-1 do
    with Dir[i] do begin
      Dtype := NextInt32;
      Dlength := NextInt32;
      Dstart := NextInt32;
    end;
  for i := 0 to DirSize-1 do begin
    if Verbose then
      WriteLn(Log^, IntToHex(Dir[i].Dtype, 8),  Dir[i].Dlength:12, ' ', IntToHex(Dir[i].Dstart,8), ' ',
                    IntToHex((Dir[i].Dtype shr 4) and $3f, 2), ' ', IntToHex((Dir[i].Dtype shr 10) and $7f, 3), ' ', IntToHex((Dir[i].Dtype shr 17), 8));
    case Dir[i].Dtype shr 4 and $3f of
      0: { Data }
        if ((Dir[i].Dtype shr 10) and $7f) <> 2 then
          case Dir[i].Dtype and $f of
            $7: if Sample >= 0 then begin
                  Connector.ReportError('Multiple Sample Spectra',False)
                end else
                  Sample := i;
            $f: if Ratio >= 0 then
                  Connector.ReportError('Multiple Ratioed Spectra',False)
                else
                  Ratio := i;
          end;
      1: ; { Spectrum range / number of points - handled elsewhere }
    else
      Stream.Position := InitialPosition + Dir[i].Dstart;
      while NextItem <> 'END' do begin
        if (Buffer = 'INS') or (Buffer = 'BMS') or (Buffer = 'EXP')  then
          AddString
        else
          SkipItem;
      end;
    end;
  end;
  { Prefer ratioed spectrum }
  if Ratio < 0 then
    Ratio := Sample;
  if Ratio > 0 then
    { Search for nearest directory entry with spectrum parameters }
    for i := 1 to DirSize - 1 do begin
      Result := TrySpectrumRange(Ratio,i) or TrySpectrumRange(Ratio,-i);
      if Result then
        Break;
    end;
  if Result then begin
    if Verbose then
      WriteLn(Log^, fXV, ' ', lXV, ' ', nPoints);
    CheckAvailableChannels(2, nPoints, False);
    Stream.Position := InitialPosition + Dir[Ratio].Dstart;
    Connector.ReadAndCopyArray(0,1,tySingle);
    Connector.SetCommentN(1, Comment);
    for i := 0 to nPoints-1 do
      Connector.SetValue(0,0,i,fXV+i*(lXV-fXV)/(nPoints-1));
  end;
end;

initialization
  RegisterPlotLoader(TOpusLoader, 'Bruker Opus Files', '*.0;*.1;*.2;*.3;*.4;*.5;*.6;*.7;*.8;*.9');
end.

