unit SSLoader;
{ Copyright (C) 2016 C M Western; see NLFit.pas for license}
interface

uses Classes, MathSup, ExpPlotLoader, ItemStrm;

type

  TSpreadSheet = class
  private
    FSheet: Integer;
    procedure SetSheet(const Value: Integer);
  protected
    procedure AddSharedIndex(Row, Col, Index: Integer);
  public
    nShared: Integer;
    Verbose: Boolean;
    SharedStrings: array of string;
    Log: PText;
    Rows, Cols: Integer;
    Sheets: array of TDoubleArrayArray;
    SheetNames: array of string;
    Strings: TStringList;
    constructor Create(var TheLog: Text; BeVerbose: Boolean = False);
    destructor Destroy; override;
    function PackIndex(Index: Integer):Double;
    function UnpackIndex(Value: Double): string;
    procedure AddShared(S: string);
    procedure SetDim(nRows, nCols: Integer);
    procedure MakeAvailable(Row, Col: Integer);
    procedure AddValue(Row, Col: Integer; Value: Double);
    procedure AddSharedString(Row, Col, Index: Integer);
    procedure AddString(Row, Col: Integer; const Value: string);
    procedure Trim;
    function StrRep(Row, Col: Integer): String;
    function nCols: Integer;
    function Column(Col: Integer): TDoubleArray;
    property Sheet: Integer read FSheet write SetSheet;
  end;

  TSpreadSheetFileLoader = class;

  TSSItemStream = class(TItemStream)
  private
    FLine, FColumn, FSheet: Integer;
    Source: TSpreadSheetFileLoader;
    CurVal: Double;
    FEOL, FEOF: Boolean;
  public
    constructor Create(TheSource: TSpreadSheetFileLoader);
    destructor Destroy; override;
    function Line:Integer; override;
    function Column: Integer; override;
    function GetFreeItem: Boolean; override;
    function Token: string; override;
    function IsIntegerToken(var Value: Integer):Boolean; override;
    function IsRealToken(var Value: Double):Boolean; override;
    function IsEmptyToken: Boolean; override;
    function NextLine: Boolean; override;
    function AtEndOfFile: Boolean; override;
    procedure SkipBlankLines; override;
    procedure Rewind; override;
  end;

  TSpreadSheetFileLoader = class(TExperimentalPlotLoader)
  public
    SS: TSpreadSheet;
    nWritten: Integer;
    procedure ClearSpreadSheet;
    function DetectChannels(Sheet: Integer):Boolean;
    function ProcessSpreadSheet:Boolean;
    function MakeItemStream(var TheLog: TextFile): TItemStream;
    destructor Destroy; override;
  end;

var
  EmptyCell: Double;

implementation

uses Math, SysUtils, UCompat;

function IsEmptyValue(Value: Double):Boolean;
var
  Index: Integer;
begin
  Result := IsNan(Value) and (PInteger(@Value)^ = 0);
end;

{ TSpreadSheet }

procedure TSpreadSheet.AddShared(S: string);
begin
  SetLength(SharedStrings, nShared+1);
  SharedStrings[nShared] := S;
  if Verbose then
    WriteLn(Log^, 'Shared ', nShared:10, ' ', S);
  Inc(nShared);
end;

procedure TSpreadSheet.AddSharedIndex(Row, Col, Index: Integer);
begin
  AddValue(Row, Col, PackIndex(Index));
end;

procedure TSpreadSheet.AddSharedString(Row, Col, Index: Integer);
begin
  AddSharedIndex(Row, Col, Index+1);
end;

procedure TSpreadSheet.AddString(Row, Col: Integer; const Value: string);
var
  Index: Integer;
begin
  if Value = '' then
    Exit;
  if Strings = nil then
    Strings := TStringList.Create;
  Index := Strings.IndexOf(Value);
  if Index < 0 then
    Index := Strings.Add(Value);
  AddSharedIndex(Row, Col, -(Index+1));
end;

procedure TSpreadSheet.AddValue(Row, Col: Integer; Value: Double);
begin
  MakeAvailable(Row, Col);
  Sheets[Sheet][Col][Row] := Value;
  if Verbose then
    WriteLn(Log^, Sheet:2, Row:10, Col:10, ' ', StrRep(Row,Col));
end;

function TSpreadSheet.Column(Col: Integer): TDoubleArray;
begin
  Result := Sheets[Sheet][Col];
end;

constructor TSpreadSheet.Create(var TheLog: Text; BeVerbose: Boolean);
begin
  EmptyCell := PackIndex(0);
  inherited Create;
  Log := @TheLog;
  Verbose := BeVerbose;
end;

destructor TSpreadSheet.Destroy;
begin
  Strings.Free;
  inherited;
end;

procedure TSpreadSheet.MakeAvailable(Row, Col: Integer);
var
  i, Old: Integer;
  C: TDoubleArray;
begin
  Rows := Max(Row+1, Rows);
  Cols := Max(Col+1, Cols);
  if Col > High(Sheets[Sheet]) then
    SetLength(Sheets[Sheet], Col+1);
  if Row > High(Sheets[Sheet][Col]) then begin
    Old := Length(Sheets[Sheet][Col]);
    SetLength(Sheets[Sheet][Col], Rows);
    C := Column(Col);
    for i := Old to Rows-1 do
      C[i] := EmptyCell;
  end;
end;

function TSpreadSheet.nCols: Integer;
begin
  Result := Length(Sheets[Sheet]);
end;

function TSpreadSheet.PackIndex(Index: Integer): Double;
begin
  SetNan(Result);
  PInteger(@Result)^ := Index;
end;

procedure TSpreadSheet.SetDim(nRows, nCols: Integer);
begin
  Rows := Max(nRows, Rows);
  Cols := Max(nCols, Cols)
end;

procedure TSpreadSheet.SetSheet(const Value: Integer);
var
  i, Old: Integer;
begin
  FSheet := Value;
  if Sheet >= Length(Sheets) then begin
    SetLength(Sheets, Sheet+1);
    Old := Length(Sheets);
    SetLength(SheetNames, Sheet+1);
    for i := Old to Sheet do
      SheetNames[i] := IntToStr(i+1);
  end;
  Cols := Length(Sheets[Sheet]);
  Rows := 0;
  for i := 0 to Cols-1 do
    Rows := Max(Rows, Length(Sheets[Sheet][i]));
end;

function TSpreadSheet.StrRep(Row, Col: Integer): String;
begin
  Result := UnpackIndex(Sheets[Sheet][Col][Row]);
end;

procedure TSpreadSheet.Trim;
var
  r, c, s, LastRow, LR, f, t: Integer;
  Col: TDoubleArray;
begin
  for s := High(Sheets) downto 0 do begin
    LastRow := -1;
    for c := 0 to High(Sheets[s]) do begin
      LR := -1;
      Col := Sheets[s][c];
      for r := High(Col) downto 0 do
        if not IsEmptyValue(Col[r]) then begin
          LR := r;
          Break;
        end;
      if LR < 0 then begin
        Sheets[s][c] := nil;
        if c = High(Sheets[s]) then
          SetLength(Sheets[s],c);
      end;
      LastRow := Max(LR, LastRow);
    end;
    if LastRow < 0 then begin
      Sheets[s] := nil;
      if S = High(Sheets) then
        SetLength(Sheets,S);
    end else begin
      { Trim leading columns }
      f := 0;
      c := 0;
      while (c < Length(Sheets[s])) and (Length(Sheets[s][c]) = 0) do
        Inc(c);
      if c > f then begin
        while c < Length(Sheets[s]) do begin
          Sheets[s][f] := Sheets[s][c];
          Inc(c);
          Inc(f);
        end;
        SetLength(Sheets[s],f);
      end;
      for c := 0 to High(Sheets[s]) do
        SetLength(Sheets[s][c], Min(LastRow+1,Length(Sheets[s][c])));
    end;
  end;
end;

function TSpreadSheet.UnpackIndex(Value: Double): string;
var
  Index: Integer;
begin
  if IsNan(Value) then begin
    Index := PInteger(@Value)^;
    if Index = 0 then
      Result := ''
    else if Index > 0 then
      Result := SharedStrings[Index-1]
    else
      Result := Strings[-Index-1];
  end else
    Result := FloatToStr(Value);
end;

{ TSpreadSheetFileLoader }

procedure TSpreadSheetFileLoader.ClearSpreadSheet;
begin
  SS.Free;
  SS := TSpreadSheet.Create(Log^, Verbose);
end;

destructor TSpreadSheetFileLoader.Destroy;
begin
  FreeAndNil(SS);
  inherited;
end;

function TSpreadSheetFileLoader.DetectChannels(Sheet: Integer):Boolean;
var
  i, j: Integer;
  Value: Double;
  Col: TDoubleArray;
  N, R, nValid, nSet, Row: Integer;
  Comment, C: string;
  nActive: array of Integer;
  HadNumber: Boolean;
begin
  SS.Sheet := Sheet;
  Result := False;
  if SS.nCols = 0 then
    Exit;
  SetnChannels(SS.nCols);
  ColumnNames := nil;
  SetLength(ColumnNames, SS.nCols);
  { Guess number of 'Active' Columns by looking for the most common number }
  SetLength(nActive, SS.nCols+1);
  Row := 0;
  for i := 0 to Min(SS.Rows-1, 20) do begin
    nSet := 0;
    for j := 0 to SS.nCols-1 do
      if i >= Length(SS.Sheets[Sheet][j]) then begin
        nSet := -1;
        Break;
      end else if not IsNanOrInf(SS.Sheets[Sheet][j][i]) then
        Inc(nSet);
    if nSet < 0 then
      Break
    else
      Inc(nActive[nSet]);
    if (nSet > 0) and (nActive[nSet] > 10) then
      Break;
    Inc(Row);
  end;
  for i := 0 to SS.nCols do
    if nActive[i] > nActive[nSet] then
      nSet := i;
  R := Row;
  { Guess column names }
  for i := 0 to SS.nCols-1 do begin
    Col := SS.Column(i);
    HadNumber := False;
    for j := Min(Row,High(SS.Sheets[Sheet][i])) downto 0 do
      if IsNanOrInf(SS.Sheets[Sheet][i][j]) then begin
        C := Trim(SS.StrRep(j,i));
        if C <> '' then begin
          ColumnNames[i] := C;
          if HadNumber then          
            R := Min(j, R);
          Break;
        end;
      end else if j = 0 then
        R := 0
      else
        HadNumber := True;
    if ColumnNames[i] = '' then
      ColumnNames[i] := IntToStr(i+1);
    if Verbose then
      WriteLn(Log^, i:3, ' ', ColumnNames[i]);
  end;
  { Guess comment }
  Comment := '';
  nSet := 0;
  for i := 0 to R-1 do begin
    if nSet > 0 then
      Comment := Comment + LineEnding;
    nSet := 0;
    for j := 0 to SS.nCols-1 do begin
      C := Trim(SS.StrRep(i,j));
      if C <> '' then begin
        if nSet > 0 then
          C := ' ' + C;
        Inc(nSet);
        Comment := Comment + C;
      end;
    end;
  end;
  for i := 0 to SS.nCols-1 do begin
    Col := SS.Column(i);
    for j := 0 to High(Col) do
        AddChange(i,Col[j]);
  end;
  N := 0;
  for i := 0 to SS.nCols-1 do
    N := Max(N, nValues[i]);
  if N <= 1 then
    Exit;
  { Treat channel as valid if > 50% of the lines have data, and at least 2 points }
  nValid := 0;
  for i := 0 to SS.nCols-1 do
    if nValues[i] >= Max(N div 2, 2) then
      Inc(nValid)
    else
      ChannelDest[i] := -1000; { Flag as not to be scanned }
  if nValid = 0 then
    Exit;
  DoAutoDetectFrequency;
  CheckAvailableChannels(SS.nCols, N, AskAboutWeights);
  { Count number of channels to be set }
  nValid := 0;
  for i := 0 to High(ChannelDest) do
    if ChannelDest[i] >= 0 then
      Inc(nValid);
  R := 0;
  for i := 0 to SS.Rows-1 do begin
    nSet := 0;
    for j := 0 to SS.Cols-1 do
      if i >= Length(SS.Sheets[SS.Sheet][j]) then
        Break
      else if ChannelDest[j] >= 0 then begin
        Value := SS.Sheets[SS.Sheet][j][i];
        if IsNanOrInf(Value) then
          Break;
        Inc(nSet);
        if Verbose then
          WriteLn(Log^, i:5, R:5, j:3, ' ', Value);
        SetValue(-1, j, R, Value);
      end;
    if nSet = nValid then
      Inc(R)
    else if Verbose then
      WriteLn(Log^, 'Skip ',i:5, ' ',nSet);
  end;
  if Verbose and (R < N) then
    WriteLn(Log^, 'Skipping from ',R,' to ',N-1);
  for i := R to N-1 do
    for j := 0 to SS.Cols-1 do
      if ChannelDest[j] >= 0 then
        SetValue(-1, j, i, Nan);
  Inc(nWritten);
  if Comment <> '' then
    Connector.Comment := Comment;
  Result := True;
end;

function TSpreadSheetFileLoader.MakeItemStream(var TheLog: TextFile): TItemStream;
begin
  Log := @TheLog;
  Result := nil;
  if Setup then
    if Load then
      Result := TSSItemStream.Create(Self);
  if Result = nil then
    Rewind;
end;

function TSpreadSheetFileLoader.ProcessSpreadSheet: Boolean;
var
  i: Integer;
begin
  if Connector = nil then begin
    Result := Length(SS.Sheets) > 0;
    Exit;
  end;
  Result := False;
  for i := 0 to High(SS.Sheets) do begin
    if i > 0 then
      ResetConnector;
    if Length(SS.Sheets) > 1 then
      Connector.SetName(SS.SheetNames[i]);
    Result := DetectChannels(i) or Result;
    SS.Sheets[i] := nil; { Free memory }
  end;
end;

{ TSSItemStream }

function TSSItemStream.AtEndOfFile: Boolean;
begin
  Result := FEOF;
end;

function TSSItemStream.Column: Integer;
begin
  Result := FColumn+1;
end;

constructor TSSItemStream.Create(TheSource: TSpreadSheetFileLoader);
begin
  inherited Create;
  Source := TheSource;
  Source.SS.Trim;
  Rewind;
end;

destructor TSSItemStream.Destroy;
begin
  FreeAndNil(Source);
  inherited;
end;

function TSSItemStream.GetFreeItem: Boolean;
begin
  Result := False;
  if FEOL or FEOF then
    Exit;
  if (FSheet > High(Source.SS.Sheets)) or (FColumn > High(Source.SS.Sheets[FSheet])) then
    FEOL := True
  else begin
    if FLine <= High(Source.SS.Sheets[FSheet][FColumn]) then
      CurVal := Source.SS.Sheets[FSheet][FColumn][FLine]
    else
      CurVal := EmptyCell;
    Inc(FColumn);
    Result := True;
  end;
end;

function TSSItemStream.IsEmptyToken: Boolean;
begin
  Result := IsEmptyValue(CurVal)
end;

function TSSItemStream.IsIntegerToken(var Value: Integer): Boolean;
begin
  Result := not IsNanOrInf(CurVal);
  if Result then
    Result := (Abs(CurVal) <= High(Integer)) and (CurVal = Round(CurVal));
  if Result then
    Value := Round(CurVal);
end;

function TSSItemStream.IsRealToken(var Value: Double): Boolean;
begin
  Result := not IsNanOrInf(CurVal);
  if Result then
    Value := CurVal;
end;

function TSSItemStream.Line: Integer;
begin
  Result := FLine+1 + FSheet*1000000;
end;

function TSSItemStream.NextLine: Boolean;
begin
  Result := True;
  Inc(FLine);
  FColumn := 0;
  if FLine >= Source.SS.Rows then begin
    Fline := 0;
    repeat
      Result := FSheet < High(Source.SS.Sheets);
      if Result then begin
        Inc(FSheet);
        Source.SS.Sheet := FSheet;
      end else
        FLine := Source.SS.Rows; { Stay at end of last sheet }
    until not Result or (Source.SS.Cols > 0);
    FEOF := not Result;
  end;
  FEOL := not Result;
end;

procedure TSSItemStream.Rewind;
begin
  FLine := 0;
  FSheet := 0;
  FColumn := 0;
  FEOF := Length(Source.SS.Sheets) = 0;
  FEOL := False;
end;

procedure TSSItemStream.SkipBlankLines;
begin
  repeat
    while GetFreeItem do
      if not IsEmptyToken then begin
        FColumn := 0;
        FEOL := False;
        Exit;
      end;
  until not NextLine;
end;

function TSSItemStream.Token: string;
begin
  Result := Source.SS.UnpackIndex(CurVal);
end;

end.
