unit TabbedLine;
{ Copyright (C) 2008,2016 C M Western; see NLFit.pas for license}
interface

uses MathSup;

type
  TSeparator = (tlSpace=0, tlComma=1, tlTab=2, tlAny, tlDetect);

function TryReadTabbedLine(Separator: TSeparator; s, Last: PAnsiChar; V: array of PDouble; out nFields: Integer; var SeparatorCount: TIntegerArray):PAnsiChar;
function TryReadIntegerTabbedLine(Separator: TSeparator; s, Last: PAnsiChar; V: array of PInteger; out nFields: Integer; var SeparatorCount: TIntegerArray):PAnsiChar;
function TryReadInt64TabbedLine(Separator: TSeparator; s, Last: PAnsiChar; V: array of PInt64; out nFields: Integer; var SeparatorCount: TIntegerArray):PAnsiChar;
function ReadArrayArray(S: AnsiString): TDoubleArrayArray;
function ReadArray(S: AnsiString): TDoubleArray;
function ReadIntegerArrayArray(S: AnsiString): TIntegerArrayArray;
function ReadIntegerArray(S: AnsiString): TIntegerArray;
function ReadInt64Array(S: AnsiString): TInt64Array;
function WriteArray(const A: TDoubleArrayArray):string; overload;
function WriteArray(const A: TDoubleArray):string; overload;
function WriteArray(const A: TIntegerArrayArray):string; overload;
function WriteArray(const A: TIntegerArray):string; overload;
function WriteArray(const A: TInt64Array):string; overload;
function SkipLineEnding(var P: PAnsiChar):Boolean;
function SkipString(var P: PAnsiChar; Last: PAnsiChar; Separator: TSeparator):UTF8String;
function SkipItem(var P: PAnsiChar; Last: PAnsiChar; Separator: TSeparator):UTF8String;
function SkipToLineEnd(var P: PAnsiChar; Last: PAnsiChar): Boolean;

var
  EmptyArray: TIntegerArray = nil;

implementation

uses UCompat, SysUtils;

function SkipLineEnding(var P: PAnsiChar):Boolean;
begin
  Result := True;
  case P^ of
    #10: if (P+1)^ = #13 then
           Inc(P,2)
         else
           Inc(P);
    #13: if (P+1)^ = #10 then
           Inc(P,2)
         else
           Inc(P);
  else
    Result := False;
  end;
end;

function SkipToLineEnd(var P: PAnsiChar; Last: PAnsiChar): Boolean;
begin
  while (P <> Last) and not (P^ in [#0, #10, #13]) do
    Inc(P);
  Result := (P <> Last) and (P^ <> #0);
end;

function SkipString(var P: PAnsiChar; Last: PAnsiChar; Separator: TSeparator):UTF8String;
var
  C: AnsiChar;
  S: PAnsiChar;
  OldLength: Integer;
begin
  S := P;
  C := P^;
  if C in ['''', '"'] then begin
    { Quoted string }
    Result := '';
    repeat
      Inc(P);
      S := P;
      while (P <> Last) and not (P^ in [C, #0, #10, #13]) do
        Inc(P);
      OldLength := Length(Result);
      if OldLength > 0 then
        Dec(S); { Include quote if not first time round }
      SetLength(Result,P-S+Length(Result));
      Move(S^,Result[OldLength+1],P-S);
      if P = Last then
        Break;
      Inc(P);
    until (P = Last) or (P^ <> C);
    if (P <> Last) and not (P^ in [#0, #10, #13]) then
      Inc(P); { Skip field separator, but not end of line }
    Exit;
  end else if Separator = tlComma then begin
    while (P <> Last) and not (P^ in [',', #0, #10, #13]) do
      Inc(P);
  end else if Separator = tlTab then begin
    while (P <> Last) and not (P^ in [#9, #0, #10, #13]) do
      Inc(P);
  end else begin
    while (P <> Last) and not (P^ in [' ', #9, #0, #10, #13]) do
      Inc(P);
  end;
  SetLength(Result,P-S);
  Move(S^,Result[1],P-S);
  if (P <> Last) and not (P^ in [#0, #10, #13]) then
    Inc(P); { Skip field separator, but not end of line }
end;

function SkipItem(var P: PAnsiChar; Last: PAnsiChar; Separator: TSeparator):UTF8String;
var
  C: AnsiChar;
  S: PAnsiChar;
  OldLength: Integer;
begin
  S := P;
  C := P^;
  if Separator = tlComma then begin
    while (P <> Last) and not (P^ in [',', #0, #10, #13]) do
      Inc(P);
  end else if Separator = tlTab then begin
    while (P <> Last) and not (P^ in [#9, #0, #10, #13]) do
      Inc(P);
  end else begin
    while (P <> Last) and not (P^ in [' ', #9, #0, #10, #13]) do
      Inc(P);
  end;
  SetLength(Result,P-S);
  Move(S^,Result[1],P-S);
  if (P <> Last) and not (P^ in [#0, #10, #13]) then
    Inc(P); { Skip field separator, but not end of line }
end;

function TryReadTabbedLine(Separator: TSeparator; s, Last: PAnsiChar; V: array of PDouble; out nFields: Integer; var SeparatorCount: TIntegerArray):PAnsiChar;
{ Read a tab (or comma) separated set of Double values; empty fields cause value to
  be left unchanged;
  result is nil if end of string reached, otherwise points to first char not read;
  If Newline found, routine exits with result pointing to new line;
  nFields is the number of values read (including null values)
  Last points to char beyond end of string; can be nil if this is readable and points to #0.
  SeparatorCount is only used if Separator = tlDetect
}
var
  P, B, Start: PAnsiChar;
  i: Integer;
  Buffer: ShortString;
  IsString: Boolean;
begin
  nFields := 0;
  Result := nil;
  if s = nil then
    Exit; { No values read }
  P := s;
  while (nFields < Length(V)) and (V[nFields] <> nil) do begin
    while (P <> Last) and (P^ = ' ') do
      Inc(p);
    if (P = Last) or (P^ = #0) then
      Exit
    else if P^ in [#10,#13] then
      Break
    else if (P^ = #9) then begin
      if Separator in [tlComma, tlSpace] then
        Continue { Treat as white space }
      else if Separator = tlDetect then
        Inc(SeparatorCount[Ord(tlTab)]);
      Inc(P) { Empty field }
    end else if P^ = ',' then begin
      if Separator in [tlSpace, tlTab] then
        Break { Start of string field }
      else if Separator = tlDetect then
        Inc(SeparatorCount[Ord(tlComma)]);
      Inc(P) { Empty field }
    end else begin
      Start := P;
      IsString := False;
      { Must make a copy, as correct length required for Val() }
      B := @Buffer[1];
      for i := 0 to 253 do begin
        if P^ in [#0, ' ', #9, ',', #10, #13] then
          Break;
        if i > 50 then
          IsString := True
        else if (i > 3) and not (P^ in ['0'..'9','+','-','.','e','E']) then
          IsString := True
        else if P^ in ['d', 'D'] then
          B^ := 'e'
        else
          B^ := P^;
        Inc(B);
        Inc(P);
        if P = Last then
          Break;
      end;
      if not IsString then begin
        // B^ := #0;
        SetLength(Buffer,B-@Buffer[1]);
        Val(Buffer,V[nFields]^,i);
        if i <> 0 then begin
          { Check for Nan, Inf }
          if Buffer[1] in ['+','-'] then
            Delete(Buffer,1,1);
          if (Length(Buffer) = 3) and ((CompareText(Buffer,'Nan') = 0) or (CompareText(Buffer,'Inf') = 0)) then
            SetNan(V[nFields]^)
          else
            IsString := True;
        end;
      end;
      B := P;
      while (P <> Last) and ((P^ = ' ') or ((Separator in [tlSpace, tlComma]) and (P^ = #9))) do
        Inc(P);
      if (P <> Last) and not IsString then
        if P^ = #9 then begin
          if Separator = tlDetect then
            Inc(SeparatorCount[Ord(tlTab)]);
          Inc(P); { Non-empty field }
        end else if P^ = ',' then begin
          if Separator in [tlSpace, tlTab] then
            Break { Start of string field }
          else if Separator = tlDetect then
            Inc(SeparatorCount[Ord(tlComma)]);
          Inc(P); { Non-empty field }
        end else if (Separator = tlDetect) and (B <> P) then
          Inc(SeparatorCount[Ord(tlSpace)]);
      if IsString then begin
        P := Start;
        Break;
      end;
    end;
    Inc(nFields);
  end;
  if (P = Last) or (P^ = #0) then
    Result := nil
  else
    Result := P;
end;

function TryReadIntegerTabbedLine(Separator: TSeparator; s, Last: PAnsiChar; V: array of PInteger; out nFields: Integer; var SeparatorCount: TIntegerArray):PAnsiChar;
{ Read a tab (or comma) separated set of Integer values; empty fields cause value to
  be left unchanged;
  result is nil if end of string reached, otherwise points to first char not read;
  If Newline found, routine exits with result pointing to new line;
  nFields is the number of values read (including null values)
  Last points to char beyond end of string; can be nil if this is readable and points to #0.
  SeparatorCount is only used if Separator = tlDetect
}
var
  P, B, Start: PAnsiChar;
  Value: Integer;
  IsString, Negative: Boolean;
begin
  nFields := 0;
  Result := nil;
  if s = nil then
    Exit; { No values read }
  P := s;
  while (nFields < Length(V)) and (V[nFields] <> nil) do begin
    while (P <> Last) and (P^ = ' ') do
      Inc(p);
    if (P = Last) or (P^ = #0) then
      Exit
    else if P^ in [#10,#13] then
      Break
    else if (P^ = #9) then begin
      if Separator in [tlComma, tlSpace] then
        Continue { Treat as white space }
      else if Separator = tlDetect then
        Inc(SeparatorCount[Ord(tlTab)]);
      Inc(P) { Empty field }
    end else if P^ = ',' then begin
      if Separator in [tlSpace, tlTab] then
        Break { Start of string field }
      else if Separator = tlDetect then
        Inc(SeparatorCount[Ord(tlComma)]);
      Inc(P) { Empty field }
    end else begin
      Start := P;
      IsString := False;
      Value := 0;
      Negative := P^ = '-';
      if P^ in ['+', '-'] then
        Inc(P);
      while not IsString and (P <> Last) and (P^ in ['0'..'9']) do begin
        if Value <= High(Integer) div 10 then
          Value := Value*10 + Ord(P^)-Ord('0')
        else
          IsString := True;
        Inc(P);
      end;
      while (P <> Last) and not (P^ in [#0, ' ', #9, ',', #10, #13]) do begin
        IsString := True;
        Inc(P);
      end;
      if not IsString then begin
        if Negative then
          Value := -Value;
        V[nFields]^ := Value;
      end;
      B := P;
      while (P <> Last) and ((P^ = ' ') or ((Separator in [tlSpace, tlComma]) and (P^ = #9))) do
        Inc(P);
      if (P <> Last) and not IsString then
        if P^ = #9 then begin
          if Separator = tlDetect then
            Inc(SeparatorCount[Ord(tlTab)]);
          Inc(P); { Non-empty field }
        end else if P^ = ',' then begin
          if Separator in [tlSpace, tlTab] then
            Break { Start of string field }
          else if Separator = tlDetect then
            Inc(SeparatorCount[Ord(tlComma)]);
          Inc(P); { Non-empty field }
        end else if (Separator = tlDetect) and (B <> P) then
          Inc(SeparatorCount[Ord(tlSpace)]);
      if IsString then begin
        P := Start;
        Break;
      end;
    end;
    Inc(nFields);
  end;
  if (P = Last) or (P^ = #0) then
    Result := nil
  else
    Result := P;
end;

function TryReadInt64TabbedLine(Separator: TSeparator; s, Last: PAnsiChar; V: array of PInt64; out nFields: Integer; var SeparatorCount: TIntegerArray):PAnsiChar;
{ Read a tab (or comma) separated set of Integer values; empty fields cause value to
  be left unchanged;
  result is nil if end of string reached, otherwise points to first char not read;
  If Newline found, routine exits with result pointing to new line;
  nFields is the number of values read (including null values)
  Last points to char beyond end of string; can be nil if this is readable and points to #0.
  SeparatorCount is only used if Separator = tlDetect
}
var
  P, B, Start: PAnsiChar;
  Value: Int64;
  IsString, Negative: Boolean;
begin
  nFields := 0;
  Result := nil;
  if s = nil then
    Exit; { No values read }
  P := s;
  while (nFields < Length(V)) and (V[nFields] <> nil) do begin
    while (P <> Last) and (P^ = ' ') do
      Inc(p);
    if (P = Last) or (P^ = #0) then
      Exit
    else if P^ in [#10,#13] then
      Break
    else if (P^ = #9) then begin
      if Separator in [tlComma, tlSpace] then
        Continue { Treat as white space }
      else if Separator = tlDetect then
        Inc(SeparatorCount[Ord(tlTab)]);
      Inc(P) { Empty field }
    end else if P^ = ',' then begin
      if Separator in [tlSpace, tlTab] then
        Break { Start of string field }
      else if Separator = tlDetect then
        Inc(SeparatorCount[Ord(tlComma)]);
      Inc(P) { Empty field }
    end else begin
      Start := P;
      IsString := False;
      Value := 0;
      Negative := P^ = '-';
      if P^ in ['+', '-'] then
        Inc(P);
      while not IsString and (P <> Last) and (P^ in ['0'..'9']) do begin
        if Value <= High(Int64) div 10 then
          Value := Value*10 + Ord(P^)-Ord('0')
        else
          IsString := True;
        Inc(P);
      end;
      while (P <> Last) and not (P^ in [#0, ' ', #9, ',', #10, #13]) do begin
        IsString := True;
        Inc(P);
      end;
      if not IsString then begin
        if Negative then
          Value := -Value;
        V[nFields]^ := Value;
      end;
      B := P;
      while (P <> Last) and ((P^ = ' ') or ((Separator in [tlSpace, tlComma]) and (P^ = #9))) do
        Inc(P);
      if (P <> Last) and not IsString then
        if P^ = #9 then begin
          if Separator = tlDetect then
            Inc(SeparatorCount[Ord(tlTab)]);
          Inc(P); { Non-empty field }
        end else if P^ = ',' then begin
          if Separator in [tlSpace, tlTab] then
            Break { Start of string field }
          else if Separator = tlDetect then
            Inc(SeparatorCount[Ord(tlComma)]);
          Inc(P); { Non-empty field }
        end else if (Separator = tlDetect) and (B <> P) then
          Inc(SeparatorCount[Ord(tlSpace)]);
      if IsString then begin
        P := Start;
        Break;
      end;
    end;
    Inc(nFields);
  end;
  if (P = Last) or (P^ = #0) then
    Result := nil
  else
    Result := P;
end;

function ReadArrayArray(S: AnsiString): TDoubleArrayArray;
var
  P: PAnsiChar;
  Buffer: TDoubleArray;
  PBuffer: array of PDouble;
  nFields, nF, nRows: Integer;
  procedure ExtendBuffer(nItems:Integer);
  var
    Old, i: Integer;
  begin
    Old := Length(Buffer);
    Inc(nItems, Old);
    SetLength(Buffer, nItems);
    SetLength(PBuffer, nItems);
    for i := 0 to nItems-1 do
      PBuffer[i] := @Buffer[i];
  end;
begin
  Buffer := nil;
  Result := nil;
  nRows := 0;
  ExtendBuffer(5);
  P := PAnsiChar(S);
  while P^ <> #0 do begin
    P := TryReadTabbedLine(tlAny, P, nil, PBuffer, nFields, EmptyArray);
    while (P <> nil) and not (P^ in [#10,#13]) do begin
      ExtendBuffer(1);
      P := TryReadTabbedLine(tlAny, P, nil, [@Buffer[High(Buffer)]], nF, EmptyArray);
      Inc(nFields, nF);
    end;
    if nFields > 0 then begin
      if nRows >= Length(Result) then
        if Length(Result) = 0 then
          SetLength(Result, 16)
        else
          SetLength(Result, Length(Result)*2);
      Result[nRows] := Copy(Buffer, 0, nFields);
      Inc(nRows);
    end;
    if P = nil then
      Break;
    SkipLineEnding(P);
  end;
  SetLength(Result, nRows);
end;

function ReadIntegerArrayArray(S: AnsiString): TIntegerArrayArray;
var
  P: PAnsiChar;
  Buffer: TIntegerArray;
  PBuffer: array of PInteger;
  nFields, nF, nRows: Integer;
  procedure ExtendBuffer(nItems:Integer);
  var
    Old, i: Integer;
  begin
    Old := Length(Buffer);
    Inc(nItems, Old);
    SetLength(Buffer, nItems);
    SetLength(PBuffer, nItems);
    for i := 0 to nItems-1 do
      PBuffer[i] := @Buffer[i];
  end;
begin
  Buffer := nil;
  Result := nil;
  nRows := 0;
  ExtendBuffer(5);
  P := PAnsiChar(S);
  while P^ <> #0 do begin
    P := TryReadIntegerTabbedLine(tlAny, P, nil, PBuffer, nFields, EmptyArray);
    while (P <> nil) and not (P^ in [#10,#13]) do begin
      ExtendBuffer(1);
      P := TryReadIntegerTabbedLine(tlAny, P, nil, [@Buffer[High(Buffer)]], nF, EmptyArray);
      Inc(nFields, nF);
    end;
    if nFields > 0 then begin
      if nRows >= Length(Result) then
        if Length(Result) = 0 then
          SetLength(Result, 16)
        else
          SetLength(Result, Length(Result)*2);
      Result[nRows] := Copy(Buffer, 0, nFields);
      Inc(nRows);
    end;
    if P = nil then
      Break;
    SkipLineEnding(P);
  end;
  SetLength(Result, nRows);
end;

function ReadArray(S: AnsiString): TDoubleArray;
var
  P: PAnsiChar;
  V: Double;
  nRows, nFields: Integer;
begin
  Result := nil;
  nRows := 0;
  P := PAnsiChar(S);
  while P^ <> #0 do begin
    P := TryReadTabbedLine(tlAny, P, nil, [@V], nFields, EmptyArray);
    if nFields > 0 then begin
      if nRows >= Length(Result) then
        if Length(Result) = 0 then
          SetLength(Result, 16)
        else
          SetLength(Result, Length(Result)*2);
      Result[nRows] := V;
      Inc(nRows);
    end;
    if P = nil then
      Break;
    while not (P^ in [#10, #13]) do
      Inc(P);
    SkipLineEnding(P);
  end;
  SetLength(Result, nRows);
end;

function ReadIntegerArray(S: AnsiString): TIntegerArray;
var
  P: PAnsiChar;
  V, nRows, nFields: Integer;
begin
  Result := nil;
  nRows := 0;
  P := PAnsiChar(S);
  while P^ <> #0 do begin
    P := TryReadIntegerTabbedLine(tlAny, P, nil, [@V], nFields, EmptyArray);
    if nFields > 0 then begin
      if nRows >= Length(Result) then
        if Length(Result) = 0 then
          SetLength(Result, 16)
        else
          SetLength(Result, Length(Result)*2);
      Result[nRows] := V;
      Inc(nRows);
    end;
    if P = nil then
      Break;
    while not (P^ in [#10, #13]) do
      Inc(P);
    SkipLineEnding(P);
  end;
  SetLength(Result, nRows);
end;

function ReadInt64Array(S: AnsiString): TInt64Array;
var
  P: PAnsiChar;
  V: Int64;
  nRows, nFields: Integer;
begin
  Result := nil;
  nRows := 0;
  P := PAnsiChar(S);
  while P^ <> #0 do begin
    P := TryReadInt64TabbedLine(tlAny, P, nil, [@V], nFields, EmptyArray);
    if nFields > 0 then begin
      if nRows >= Length(Result) then
        if Length(Result) = 0 then
          SetLength(Result, 16)
        else
          SetLength(Result, Length(Result)*2);
      Result[nRows] := V;
      Inc(nRows);
    end;
    if P = nil then
      Break;
    while not (P^ in [#10, #13]) do
      Inc(P);
    SkipLineEnding(P);
  end;
  SetLength(Result, nRows);
end;

function WriteArray(const A: TDoubleArrayArray):string;
var
  i, j: Integer;
  Line: string;
begin
  Result := '';
  for i := 0 to High(A) do begin
    Line := '';
    for j := 0 to High(A[i]) do begin
      if j > 0 then
        Line := Line + ' ';
      Line := Line + FloatToStr(A[i][j]);
    end;
    Result := Result + LineEnding + Line;
  end;
end;

function WriteArray(const A: TIntegerArrayArray):string;
var
  i, j: Integer;
  Line: string;
begin
  Result := '';
  for i := 0 to High(A) do begin
    Line := '';
    for j := 0 to High(A[i]) do begin
      if j > 0 then
        Line := Line + ' ';
      Line := Line + FloatToStr(A[i][j]);
    end;
    Result := Result + LineEnding + Line;
  end;
end;

function WriteArray(const A: TDoubleArray):string; overload;
var
  i: Integer;
begin
  Result := '';
  for i := 0 to High(A) do
    Result := Result + LineEnding + FloatToStr(A[i]);
end;

function WriteArray(const A: TIntegerArray):string; overload;
var
  i: Integer;
begin
  Result := '';
  for i := 0 to High(A) do
    Result := Result + LineEnding + IntToStr(A[i]);
end;

function WriteArray(const A: TInt64Array):string; overload;
var
  i: Integer;
begin
  Result := '';
  for i := 0 to High(A) do
    Result := Result + LineEnding + IntToStr(A[i]);
end;

end.
