﻿unit Units;
{ Copyright (C) 2005 C M Western; see NLFit.pas for license}
interface

type
  TResonanceUnits = (cm1=0, MHz, Kelvin, eV, sqrtcm1, nmVac, nmAir, angVac, angAir);
  TEnergyUnits = cm1..eV;
  TTransitionType = (ttDipole, ttQuadrupole);

function NameOfUnit(Units: TResonanceUnits): string; overload;
function TypeOfUnit(Units: TResonanceUnits): string;
function StrToUnits(s: string): TResonanceUnits;
function ConversionFactor(UFrom, UTo: TEnergyUnits):Extended;
function Convert(Value:Double; UFrom, UTo: TResonanceUnits):Extended;
function ConvertWidth(Value, Centre:Double; UFrom, UTo: TResonanceUnits):Extended;

{ Is the unit proportional to energy? }
function IsEnergyUnit(Units: TResonanceUnits): Boolean; inline;

{ Is a reciprocal involved in converting to any energy unit? }
function IsWavelengthUnit(Units: TResonanceUnits): Boolean; inline;

{ Is a square root involved in converting to energy ? }
function IsSqrtUnit(Units: TResonanceUnits): Boolean; inline;

type
  TIntensityUnits = (Arbitrary=0,
                     Normalized,
                     Squared,
                     HonlLondon,   { Dimensionless line strength factor summed over degenerate upper and lower states }
                     OscillatorStrength, { Dimensionless; summed over upper degenerate states only }
                     EinsteinASum, { s-1; summed over all degenerate states }
                     EinsteinA,    { s-1; summed over all lower degnerate states only, so reciprocal is radiative lifetime }
                     nm2MHzperMolecule {nm^2 MHz / molecule},
                     cm2WavenumberperMolecule { cm^2 cm-1 / molecule },
                     HzperMolecule {s-1, emission including Boltzmann factor },
                     PopDist { Population Distrubution },
                     NormPopDist { Normalised Population Distrubution } );

function NameOfUnit(Units: TIntensityUnits; Integrated: Boolean; FrequencyUnits:TResonanceUnits): string; overload;
function StrToIntensityUnits(s: string): TIntensityUnits;
function IntensityFactor(Frequency: Double; FrequencyUnits: TResonanceUnits; IntensityUnits: TIntensityUnits; TransitionType: TTransitionType): Double; overload;
function QuadrupoleConversion(Frequency: Double; Units: TResonanceUnits):Double;
function ConvertIntensity(Value, Frequency: Double; FFrom: TResonanceUnits; IFrom: TIntensityUnits; FTo: TResonanceUnits; ITo: TIntensityUnits; Integrated: Boolean = False):Double;

function IsArbitraryStyle(Units: TIntensityUnits): Boolean;
function HasUnits(Units: TIntensityUnits):Boolean;
function IsNormalized(Units: TIntensityUnits):Boolean;
function IsPopDist(Units: TIntensityUnits):Boolean;
function IncludesPopulation(Units: TIntensityUnits):Boolean;

procedure SetUpUnitFactors(Planck, ElectronCharge, Avogadro, BoltzmannConstant, BohrMagneton, BohrRadius: Double; const Source: string);

var
  FundamentalConstantsSummary: string;
  BoltzmannConstants: array [TEnergyUnits] of Double;
{ This is computed to ensure the values are identical, even when switching units.
  The previous values were 0.6950356 ( for cm-1 ) and 20836.644 (for MHz )
  but 20836.644 / c * 1e4 =0.69503563028260037 }
  MomentOfInertiaTimesRotationalConstant: array [TEnergyUnits] of Double;

var
  { Computed derived constants to avoid discrepancies }
  hCross: Double;
  eVTocm1: Double; { 8065.54400479571 8065.54429 8065.54465 }
  KelvinToMHz: Double; { 20836.6122251252 20836.6179288395 20836.644 }
  cm1toJ: Double; { 19.8644582417176E-24 19.8644568326930306e-24 = 100 h c }
  amuAngstoms2toSI: Double; { 16.6053904042716E-48 16.6053892103218986e-48 }
  SpeedOfLightTimesPlanck: Double;
  BohrMagnetonOverPlanck: Double;
  BohrMagneton_e21: Double;
  gammaMhalf: Double; { Conversion to dimensionless normal coordinates is Q = gamma^-1/2 q
                        where gamma^-1/2 = h^1/2 /(20 pi c^1/2) / sqrt(omega/cm-1)
                        so gammaMhalf = h/(20 pi c^1/2) = 23.6612549932097616e-24 }
  ReduceFrequency: Double; { 4 pi^2 10^-21 c / h / NA = 0.02966016141352712 }
  lambdaConversion: Double; { = 5.89182972041606619e-7; 1/Sqrt(LambdaConversion) = 1302.79147066436 }
  gammaConversion: Double;  { = 38.641005308071206; 1/(Sqrt(lambdaconv)*gammaconv) =  33.7152581895232 }
  aJConversion: Double;     { = 50341.1721332513141 = 1e-18/(100hc); 1 aJ in cm-1 }
  QuadrupoleFactor: Double; { = 9.93333451715843E-75 = 5000 pi^2 e^2 a0^4; See QuadrupoleFactor routine }

implementation

uses TypInfo, UCompat, AirVac, SysUtils, Math, CODATA, MathSup;

var { Actually const, but compiler can't cope with these as expressions just involving constants }
  IntensityFactors: array [TIntensityUnits] of Double;
  { Conversion factors excluding frequency }

function NameOfUnit(Units: TResonanceUnits): string;
begin
  case Units of
    cm1: Result := 'cm-1';
    MHz: Result := 'MHz';
    Kelvin: Result := 'Kelvin';
    eV:  Result := 'eV';
    sqrtcm1: Result := 'cm-1/2';
    nmVac, nmAir:  Result := 'nm';
    angVac, angAir: Result := {$IFNDEF FPC} #197 {$ELSE} 'Å' {$ENDIF};
  end;
end;

function StrToUnits(s: string): TResonanceUnits;
var
  I: Integer;
begin
  Result := cm1;
  I := GetEnumValue(TypeInfo(TResonanceUnits),s);
  if I >= 0 then
    Result := TResonanceUnits(I)
  else
    repeat
      if CompareText(s, NameOfUnit(Result)) = 0 then
        Exit;
      if Result = High(Result) then begin
        if (Length(s) >= 3) and (CompareText(s, Copy('Angstroms',1,Length(s))) = 0) then begin
          Result := angVac;
          Exit;
        end;
        raise EReportException.Create(s + ' is not a valid unit name');
      end;
      Inc(Result);
    until False;
end;

function TypeOfUnit(Units: TResonanceUnits): string;
begin
  case Units of
    cm1: Result := 'Wavenumber';
    MHz: Result := 'Frequency';
    eV:  Result := 'Energy';
    Kelvin: Result := 'Temperature';
    sqrtcm1: Result := 'Velocity';
    nmAir, angAir:  Result := 'Air Wavelength';
    nmVac, angVac:  Result := 'Vacuum Wavelength';
  end;
end;

function IsEnergyUnit(Units: TResonanceUnits): Boolean; inline;
begin
  Result := Units <= High(TEnergyUnits);
end;

function IsWavelengthUnit(Units: TResonanceUnits): Boolean; inline;
begin
  Result := Units >= nmVac;
end;

function IsSqrtUnit(Units: TResonanceUnits): Boolean; inline;
begin
  Result := Units = sqrtcm1;
end;

const
  SpeedOfLight4: Double = 29979.2458;

function ConversionFactor(UFrom, UTo: TEnergyUnits):Extended;
begin
  { Can't use array of constants, because expressions like 1/SpeedOfLight don't count as constant expressions }
  if UTo < UFrom then
    Result := 1/ConversionFactor(UTo, UFrom)
  else if UFrom = UTo then
    Result := 1
  else
    case UFrom of
      cm1:    case UTo of
                MHz:    Result := SpeedOfLight4;
                Kelvin: Result := SpeedOfLight4/KelvinToMHz;
                ev:     Result := 1/eVTocm1;
              end;
      MHz:    case UTo of
                Kelvin: Result := 1/KelvinToMHz;
                ev:     Result := 1/eVTocm1/SpeedOfLight4;
              end;
      Kelvin: case UTo of
                ev:     Result := KelvinToMHz/SpeedOfLight4/eVTocm1;
              end;
    end;
end;

function Convert(Value:Double; UFrom, UTo: TResonanceUnits):Extended;
begin
  Result := Value;
  if UFrom = UTo then
    Exit;
  case UFrom of
    sqrtcm1: begin UFrom := cm1; Result := Sqr(Result); end;
    nmVac:   begin UFrom := cm1; Result := nmcm1(Result); end;
    nmAir:   begin UFrom := cm1; Result := vac(Result); end;
    angVac:  begin UFrom := cm1; Result := nmcm1(Result/10); end;
    angAir:  begin UFrom := cm1; Result := vac(Result/10); end;
  end;
  if IsEnergyUnit(UTo) then
    Result := Result*ConversionFactor(UFrom, UTo)
  else begin
    Result := Result*ConversionFactor(UFrom, cm1);
    case UTo of
      sqrtcm1: if Result < 0 then
                 { Allow energy scale to go through zero }
                 Result := -Sqrt(Abs(Result))
               else
                 Result := Sqrt(Result);
      nmVac:   Result := nmcm1(Result);
      nmAir:   Result := air(Result);
      angVac:  Result := 10*nmcm1(Result);
      angAir:  Result := 10*air(Result);
    end
  end;
end;

{ Note that this function is not reversible, and as it depends on the centre
  frequency it can't be made usefully so. }
function ConvertWidth(Value, Centre:Double; UFrom, UTo: TResonanceUnits):Extended;
begin
  if UFrom = UTo then
    Result := Value
  else if IsSqrtUnit(UFrom) xor IsSqrtUnit(UTo) then
    Result := Abs(Convert(Centre,UFrom,UTo)-Convert(Centre+Value,UFrom,UTo))
  else if IsWavelengthUnit(UFrom) or IsWavelengthUnit(UTo) then
    { Taking inverse or air <-> nm conversion - scale according to Value/Centre }
    if Centre > 1/Max_nm_cm1 then
      Result := Value/Centre * Convert(Centre, UFrom, UTo)
    else
      Result := 0 { Invalid, but zero is probably harmless }
  else
    Result := Convert(Value, UFrom, UTo)
end;

function NameOfUnit(Units: TIntensityUnits; Integrated: Boolean; FrequencyUnits:TResonanceUnits): string;
begin
  case Units of
    Arbitrary:    Result := '';
    Squared:      Result := 'Intensity^2';
    Normalized:   Result := 'Normalized Intensity';
    HonlLondon:   Result := 'Honl-London Factor';
    OscillatorStrength: Result := 'Oscillator Strength';
    EinsteinASum: Result := 'Summed Einstein A coefficient / s-1';
    EinsteinA:    Result := 'Einstein A coefficient per state / s-1';
    nm2MHzperMolecule: Result := 'nm^2 MHz / molecule';
    cm2WavenumberperMolecule: Result := 'cm^2 cm-1 / molecule';
    HzperMolecule: Result := 's-1 / molecule';
    PopDist: Result := 'Population';
    NormPopDist: Result := 'Normalized Population';
  end;
  if not Integrated and (Units <> Arbitrary) then
    Result := Result + ' / ' + NameOfUnit(FrequencyUnits);
end;

function StrToIntensityUnits(s: string): TIntensityUnits;
var
  I: Integer;
begin
  I := GetEnumValue(TypeInfo(TIntensityUnits),s);
  if I >= 0 then
    Result := TIntensityUnits(I)
  else
    raise EReportException.Create('Unknown intensity unit:'+s);
end;

const
  FrequencyUnit: array [TIntensityUnits] of TResonanceUnits =
    (cm1 { Arbitrary: Actually none },
     cm1 { Squared: Actually none },
     cm1 { Normalized: Actually none },
     cm1 { HonlLondon: Actually none },
     MHz { OscillatorStrength: In principle none, but conversion factor given assuming frequency in MHz },
     cm1 { Hz },
     cm1 { Hz },
     MHz, { nm2MHzperMolecule }
     cm1, { cm2WavenumberperMolecule }
     cm1  { HzperMolecule },
     cm1 { Actually none },
     cm1 { Actually none }
     );

  FrequencyPower: array [TIntensityUnits] of Integer =
    { Power of frequency in relationship to dipole moment }
    (0 { Arbitrary },
     0 { Normalized },
     0 { Squared },
     0 { Honl-London },
     1 { OscillatorStrength},
     3 { EinsteinASum },
     3 { EinsteinA },
     1, { nm2MHzperMolecule }
     1, { cm2WavenumberperMolecule }
     3, { HzperMolecule }
     0, { PopDist }
     0  { NormPopDist } );

function QuadrupoleConversion(Frequency: Double; Units: TResonanceUnits):Double;
begin
  Result := Convert(Frequency, Units, cm1);
  { 3 * pi^2 / 10 (nu/c)^2 in SI units }
  { = 3000 * pi^2 * (nu/cm-1)^2 }
  { multiply by 5/3 to correct for wrong conversion to unpolarised light in IntensityFactor routine }
  { divide by (1e-21/c)^2 to cancel out the conversion to Debye }
  { multiply by e^2 a0^4 to convert quadrupole moment from atomic units }
  { Overall factor is 5e45 pi^2 e^2 a0^4 c^2 = 8.92763583942E-16 }
  Result := QuadrupoleFactor * Sqr(Result);
end;

function IntensityFactor(Frequency: Double; FrequencyUnits: TResonanceUnits; IntensityUnits: TIntensityUnits; TransitionType: TTransitionType): Double;
begin
  if HasUnits(IntensityUnits) or (IntensityUnits = OscillatorStrength) then begin
    if FrequencyUnits <> FrequencyUnit[IntensityUnits] then
      Frequency := Convert(Abs(Frequency),FrequencyUnits,FrequencyUnit[IntensityUnits])
    else
      Frequency := Abs(Frequency);
    Result := 3 { Conversion factor for unpolarised light }
              * IntensityFactors[IntensityUnits]
              * Math.IntPower(Frequency,FrequencyPower[IntensityUnits]);
    if TransitionType = ttQuadrupole then
      Result := Result * QuadrupoleConversion(Frequency,FrequencyUnit[IntensityUnits]);
  end else if IntensityUnits = HonlLondon then
    if TransitionType = ttQuadrupole then
      Result := 5
    else
      Result := 3
  else
    Result := 1;
end;

function ConvertIntensity(Value, Frequency: Double;
                          FFrom: TResonanceUnits; IFrom: TIntensityUnits;
                          FTo: TResonanceUnits; ITo: TIntensityUnits;
                          Integrated: Boolean):Double;
var
  s: Double;
begin
  { Attempt to convert intetsity units; some conversions are only approximate,
    and values without a specific unit (Arbitrary, Normalized, Squared) are left unchanged.
    Call with Integrated=True for integrated line intensities }
  { Note that EinsteinA and EinsteinASum are considered as equivalent in this routine } 
  { TODO: Do the conversions make sense for sqrtcm1? }
  Result := Value; { If all else fails, leave unchanged }
  case IFrom of
    Arbitrary, Normalized, PopDist, NormPopDist, HonlLondon, OscillatorStrength, EinsteinASum, EinsteinA: if ITo = Squared then Result := Sqr(Result);
    Squared: if IFrom <> Squared then Result := Sqrt(Abs(Result));
  else
    case ITo of
      Arbitrary: ;
      Normalized: ;
      Squared: Result := Sqr(Result);
      HonlLondon: ;
      OscillatorStrength: ;
      EinsteinASum: ;
      EinsteinA: ;
      PopDist: ;
      NormPopDist: ;
    else
      { below here both From and To have units }
      Result := Result * IntensityFactors[ITo]/IntensityFactors[IFrom];
      if (FrequencyUnit[IFrom] <> FrequencyUnit[ITo]) or (FrequencyPower[ITo] <> FrequencyPower[IFrom]) then
        if FrequencyPower[ITo] = FrequencyPower[IFrom] then
          Result := Result * Math.IntPower(ConversionFactor(FrequencyUnit[IFrom],FrequencyUnit[ITo]),FrequencyPower[IFrom])
        else begin
          Frequency := Convert(Frequency, FFrom, FrequencyUnit[IFrom]);
          if Frequency <= 0 then
            Result := 0
          else
            Result := Result * Math.IntPower(Convert(Frequency, FrequencyUnit[IFrom], FrequencyUnit[ITo]), FrequencyPower[ITo])
                             / Math.IntPower(Frequency, FrequencyPower[IFrom]);
        end;
    end;
  end;
  if not Integrated and (FFrom <> FTo) and not(IFrom in [Arbitrary,Normalized]) then begin
    { This conversion fails if the zero level is incorrect in experimental spectra }
    { TODO: Review this for experimental spectra } 
    s := Convert(Frequency, FFrom, FTo);
    if s > 0 then
      Result := Result * Frequency / s
    else
      Result := 0;
  end;
end;

function IsArbitraryStyle(Units: TIntensityUnits): Boolean;
begin
  Result := Units in [Arbitrary, Squared, PopDist];
end;

function HasUnits(Units: TIntensityUnits):Boolean;
begin
  Result := Units in [EinsteinASum, EinsteinA, nm2MHzperMolecule, cm2WavenumberperMolecule, HzperMolecule];
end;

function IsNormalized(Units: TIntensityUnits):Boolean;
begin
  Result := Units in [Normalized, nm2MHzperMolecule, cm2WavenumberperMolecule, HzperMolecule, NormPopDist];
end;

function IsPopDist(Units: TIntensityUnits):Boolean;
begin
  Result := Units in [PopDist, NormPopDist];
end;

function IncludesPopulation(Units: TIntensityUnits):Boolean;
begin
  Result := not (Units in [HonlLondon, OscillatorStrength, EinsteinASum, EinsteinA]);
end;

procedure SetUpUnitFactors(Planck, ElectronCharge, Avogadro, BoltzmannConstant, BohrMagneton, BohrRadius: Double; const Source: string);
var
  I: TEnergyUnits;
begin

  hCross := Planck / Two / Pi;
  eVTocm1 := ElectronCharge / Planck / SpeedOfLight / 100;
  KelvinToMHz := BoltzmannConstant / Planck / 1000000;
  SpeedOfLightTimesPlanck := SpeedOfLight * Planck;
  cm1toJ := SpeedOfLightTimesPlanck * 100;
  amuAngstoms2toSI := 1e-23 / Avogadro;
  BohrMagnetonOverPlanck := BohrMagneton/Planck;
  BohrMagneton_e21 := 1e21 * BohrMagneton;

  IntensityFactors[Arbitrary] := 1;
  IntensityFactors[Squared] := 1;
  IntensityFactors[Normalized] := 1;
  IntensityFactors[HonlLondon] := 1;
  IntensityFactors[OscillatorStrength] := 8*Pi*Pi*ElectronMass*1e-36/3/Planck/Sqr(ElectronCharge*SpeedOfLight);
  IntensityFactors[nm2MHzperMolecule] := 8*Pi*Pi*Pi/(3e31*SpeedOfLight*Planck);
  IntensityFactors[cm2WavenumberperMolecule] := IntensityFactors[nm2MHzperMolecule] / 1e14;
  IntensityFactors[HzperMolecule] := 64*Math.IntPower(pi,4)*1e-43/3/Planck;
  IntensityFactors[EinsteinASum] := IntensityFactors[HzperMolecule];
  IntensityFactors[EinsteinA] := IntensityFactors[HzperMolecule];
  IntensityFactors[PopDist] := 1;
  IntensityFactors[NormPopDist] := 1;
  BoltzmannConstants[cm1] := Convert(1,Kelvin,cm1);
  BoltzmannConstants[MHz] := Convert(1,Kelvin,MHz);
  BoltzmannConstants[Kelvin] := 1;
  BoltzmannConstants[eV] := Convert(1,Kelvin,eV);
  MomentOfInertiaTimesRotationalConstant[MHz] :=  Planck / (8000*Sqr(Pi)) * 1e20 * Avogadro;
  for I := Low(I) to High(I) do
    MomentOfInertiaTimesRotationalConstant[I] := Convert(MomentOfInertiaTimesRotationalConstant[MHz], MHz, I);
  ReduceFrequency := 4 * Pi * Pi * SpeedOfLight * 1e-21 / Planck / Avogadro;
  gammaMhalf := Sqrt(Planck/SpeedOfLight) / (Pi * 20);
  LambdaConversion := 2*pi*pi*SpeedOfLight*SpeedOfLight/5/Avogadro;
  gammaConversion := 2*Pi/1e19/Planck/Sqrt(1000*Avogadro);
  aJConversion := 1e-20/Planck/SpeedOfLight;
  QuadrupoleFactor := 5e45 * Sqr(pi * ElectronCharge * SpeedOfLight * Sqr(BohrRadius));

  FundamentalConstantsSummary := 'Fundamental Constants from '+Source+': h='+FloatToStr(Planck)+' k='+FloatToStr(BoltzmannConstant)+' NA='+FloatToStr(Avogadro)+' e='+FloatToStr(ElectronCharge)+' me='+FloatToStr(ElectronMass);
end;

initialization
  SetUpUnitFactors(CODATA.Planck, CODATA.ElectronCharge, CODATA.Avogadro, CODATA.BoltzmannConstant, CODATA.BohrMagneton, CODATA.BohrRadius, CODATA.FundamentalConstantsSource);
end.
