unit ZipLoader;
{ Copyright (C) 2016 C M Western; see NLFit.pas for license}
interface

uses {$IFDEF FPC} zipper, {$ENDIF}
     Classes, SSLoader;

type

{$IFNDEF FPC}
  { Copied directly from zipper in FPC FCL, as not available in all versions of Delphi }
  Central_File_Header_Type = Packed Record
    Signature            :  LongInt;
    MadeBy_Version       :  Word;
    Extract_Version_Reqd :  Word;
    Bit_Flag             :  Word;
    Compress_Method      :  Word;
    Last_Mod_Time        :  Word;
    Last_Mod_Date        :  Word;
    Crc32                :  LongWord;
    Compressed_Size      :  LongInt;
    Uncompressed_Size    :  LongInt;
    Filename_Length      :  Word;
    Extra_Field_Length   :  Word;
    File_Comment_Length  :  Word;
    Starting_Disk_Num    :  Word;
    Internal_Attributes  :  Word;
    External_Attributes  :  LongInt;
    Local_Header_Offset  :  LongInt;
  End;
{$ENDIF}

  TZipFileLoader = class(TSpreadSheetFileLoader)
  protected
    nEntries, CenDirPos : LongInt;
    CurrentEntry: LongInt;
    CentralHdr: Central_File_Header_Type;
    CurrentFileName: AnsiString;
    DStream: TStream;
    StreamSize: LongInt;
  public
    destructor Destroy; override;
    procedure FreeStream;
    function ReadHeader: Boolean;
    procedure FirstEntry;
    function ReadEntry: Boolean;
    function NextEntry: Boolean;
    function PrepareStream: Boolean;
    function Setup: Boolean; override;
    function FindFile(const Name: string):Boolean;
    function ReadXMLFile(const Name: string):Boolean;
  end;

implementation

uses {$IFDEF FPC}
       zstream,
     {$ELSE}
       {$IFDEF UNICODE}
         ZLib,
       {$ELSE}
         MZLib, { For earlier versions of Delphi, use MZLib from GraphicEX, as some access to internal state required }
         {$DEFINE REIMP}
       {$ENDIF}
     {$ENDIF}
     SysUtils, Math, pgXMLRead, Worker;

{$IFNDEF FPC}
{ Copied directly from zipper in FPC FCL, as not available in Delphi }

Const
  { Signatures }
{$ifdef FPC_BIG_ENDIAN}
  END_OF_CENTRAL_DIR_SIGNATURE  = $504B0506;
  LOCAL_FILE_HEADER_SIGNATURE   = $504B0304;
  CENTRAL_FILE_HEADER_SIGNATURE = $504B0102;
{$else FPC_BIG_ENDIAN}
  END_OF_CENTRAL_DIR_SIGNATURE  = $06054B50;
  LOCAL_FILE_HEADER_SIGNATURE   = $04034B50;
  CENTRAL_FILE_HEADER_SIGNATURE = $02014B50;
{$endif FPC_BIG_ENDIAN}

Type
   Local_File_Header_Type = Packed Record
     Signature              :  LongInt;
     Extract_Version_Reqd   :  Word;
     Bit_Flag               :  Word;
     Compress_Method        :  Word;
     Last_Mod_Time          :  Word;
     Last_Mod_Date          :  Word;
     Crc32                  :  LongWord;
     Compressed_Size        :  LongInt;
     Uncompressed_Size      :  LongInt;
     Filename_Length        :  Word;
     Extra_Field_Length     :  Word;
   end;

  End_of_Central_Dir_Type =  Packed Record
    Signature               :  LongInt;
    Disk_Number             :  Word;
    Central_Dir_Start_Disk  :  Word;
    Entries_This_Disk       :  Word;
    Total_Entries           :  Word;
    Central_Dir_Size        :  LongInt;
    Start_Disk_Offset       :  LongInt;
    ZipFile_Comment_Length  :  Word;
  end;

{$ENDIF}

{$IFDEF REIMP}

{ Partial re-implementation to allow header skip with MZlib }

type

  TMyDecompressionStream = class(TStream)
  private
    FStream: TStream;
    State: TZState;
    FFullSise: Integer;
    ExpectedCRC, ActualCRC: LongWord;
    Buffer: AnsiString;
  protected
    function GetSize:Int64; override;
  public
    constructor Create(Source: TStream; CompSize, FullSize: Integer; AExpectedCRC: LongWord);
    destructor Destroy; override;
    function Read(var Buffer; Count: Longint): Longint; override;
  end;

constructor TMyDecompressionStream.Create(Source: TStream; CompSize, FullSize: Integer; AExpectedCRC: LongWord);
begin
  inherited Create;
  FStream := Source;
  if Source is TMemoryStream then
    State.NextInput := PByte(PAnsiChar(TMemoryStream(Source).Memory) + Source.Position)
  else begin
    SetLength(Buffer, CompSize);
    Source.Read(Buffer[1], CompSize);
    State.NextInput := PByte(Buffer);
  end;
  State.AvailableInput := CompSize;
  InflateInit2(State, -MAX_WBITS); { Flag to skip header }
  FFullSise := FullSize;
  ExpectedCRC := AExpectedCRC;
  ActualCRC := 0;
end;

destructor TMyDecompressionStream.Destroy;
begin
  inflateEnd(State);
  inherited;
end;

function TMyDecompressionStream.GetSize: Int64;
begin
  Result := FFullSise;
end;

function TMyDecompressionStream.Read(var Buffer; Count: Integer): Longint;
var
  Initial: Cardinal;
  Status: Integer;
begin
  State.NextOutput := @Buffer;
  State.AvailableOutput := Count;
  Initial := State.TotalOutput;
  Status := inflate(State,0);
  Result := State.TotalOutput - Initial;
  if Status >= 0 then
    ActualCRC := CRC32(ActualCRC,@Buffer,Result);
  case Status of
    Z_OK: { No action };
    Z_STREAM_END: { Check CRC }
          if ActualCRC <> ExpectedCRC then
            raise Exception.Create('CRC error');
  else
    raise Exception.Create('Decompression Error');
  end;
end;

{$ELSE}

type

  TMyDecompressionStream = class(TDecompressionStream)
    constructor Create(ASource: TStream; CompSize, FullSize: Integer; AExpectedCRC: LongWord);
  end;

constructor TMyDecompressionStream.Create(ASource: TStream; CompSize, FullSize: Integer; AExpectedCRC: LongWord);
begin
  inherited Create(ASource, {$IFDEF FPC}True{$ELSE}-15{$ENDIF});
end;

{$ENDIF}

{ TZipFileLoader }

type
  PEnd_of_Central_Dir_Type = ^End_of_Central_Dir_Type;

destructor TZipFileLoader.Destroy;
begin
  FreeStream;
  FreeAndNil(XMLDocument);
  inherited;
end;

function TZipFileLoader.FindFile(const Name: string): Boolean;
begin
  FreeStream;
  Result := False;
  FirstEntry;
  while not Result and NextEntry do
    if CompareText(Name,CurrentFileName) = 0 then
      Result := PrepareStream;
end;

procedure TZipFileLoader.FirstEntry;
begin
  Stream.Seek(CenDirPos,soFrombeginning);
  CurrentEntry := -1;
end;

procedure TZipFileLoader.FreeStream;
begin
  if DStream = Stream then
    DStream := nil
  else
    FreeAndNil(DStream);
end;

function TZipFileLoader.NextEntry: Boolean;
begin
  Inc(CurrentEntry);
  Result := ReadEntry;
end;

function TZipFileLoader.PrepareStream: Boolean;
var
   LocalHdr: Local_File_Header_Type;
begin
  Result := False;
  FreeStream;
  Stream.Seek(CentralHdr.Local_Header_Offset,soFromBeginning);
  if Stream.Read(LocalHdr,SizeOf(LocalHdr)) <> SizeOf(LocalHdr) then
    Exit;
  if LocalHdr.Signature<>LOCAL_FILE_HEADER_SIGNATURE then
    Exit;
  Stream.Seek(LocalHdr.Filename_Length+CentralHdr.Extra_Field_Length,soCurrent);
  if CentralHdr.Compress_Method = 0 then begin
     DStream := Stream;
     StreamSize := CentralHdr.Compressed_Size;
  end else begin { Assuming Method is 8 }
     DStream := TMyDecompressionStream.Create(Stream, CentralHdr.Compressed_Size, CentralHdr.Uncompressed_Size, CentralHdr.Crc32);
     StreamSize := CentralHdr.Uncompressed_Size;
  end;
  Result := True;
end;

function TZipFileLoader.ReadEntry: Boolean;
begin
  Result := False;
  if CurrentEntry >= nEntries then
    Exit;
  if Stream.Read(CentralHdr, SizeOf(CentralHdr)) <> SizeOf(CentralHdr) then
    Exit;
  with CentralHdr do begin
    if Signature<>CENTRAL_FILE_HEADER_SIGNATURE then
      Exit;
    SetLength(CurrentFileName,Filename_Length);
    if Stream.Read(CurrentFileName[1],Filename_Length) <> Filename_Length then
      Exit;
    Stream.Seek(Extra_Field_Length+File_Comment_Length,soCurrent);
  end;
  Result := True;
end;

function TZipFileLoader.ReadHeader: Boolean;
var
  i: Integer;
  Buffer: array [0..4095] of AnsiChar;
begin
   Result := False;
   Stream.Seek(Max(Stream.Size-SizeOf(Buffer),0), soFromBeginning);
   i := Stream.Read(Buffer[0], SizeOf(Buffer));
   Dec(i, SizeOf(End_of_Central_Dir_Type));
   while i >= 0 do begin
     if PEnd_of_Central_Dir_Type(@Buffer[i])^.Signature = END_OF_CENTRAL_DIR_SIGNATURE then
       Break;
     Dec(i);
   end;
   if i < 0 then
     Exit;
   with PEnd_of_Central_Dir_Type(@Buffer[i])^ do begin
     CenDirPos := Start_Disk_Offset;
     nEntries := Entries_This_Disk;
   end;
   Result := True;
end;

function TZipFileLoader.ReadXMLFile(const Name: string): Boolean;
begin
  FreeAndNil(XMLDocument);
  Result := FindFile(Name);
  if Result then
    pgXMLRead.ReadXMLFile(XMLDocument, DStream, CurrentFileName, StreamSize);
end;

function TZipFileLoader.Setup: Boolean;
begin
  Result := inherited Setup;
  if not Result then
    Exit;
  Result := ReadHeader;
end;

end.
