unit AutoFit;
{$I utils.inc}{Copyright (C) 2016 C M Western; see commandline.pas for license}
interface

uses SysUtils, Classes, MathSup, MNode, Mixture, Fitter, ExpPlot, pgDOM, Worker;

type

  TStatusEvent = procedure (Sender: TObject; Trial: Int64) of object;

  TAutoFitStatus = (afOK, afOrder, afBlends, afParameter, afIntensity, afResidual, afSVD,
                    afCheckCalc);
  { TAutoFitter }

  TAutoFitter = class(TStreamableMNode)
  private
    FnBest: Integer;
    TA: TIntegerArray;  { Current assignment }
    FOnUpdateStatus: TStatusEvent;
    LineList: TExperimentalPlot;
    function NextAssignment(nSteps: Integer):Boolean;
    procedure SetnBest(AValue: Integer);
  protected
    procedure LoadStartingValues(Source: TAutoFitter);
    procedure MergeBest(Source: TAutoFitter);
  public
    { Input: }
    Obs0: TObservationArray; { All the observations }
    { Settings controlling the process }
    aW: Double; { The acceptance window }
    Mixture: TMixture; { The mixture to work on }
    { Working }
    f: TDoubleArray; { The linelist frequencies to assign }
    Intensity: TDoubleArray; { Intensities for the above }
    Weights: TDoubleArray; { Weights for above; may be empty }
    FP: TIntegerArray; { Lowest index of possible assignment within range }
    PU: TIntegerArray; { Closest assignment for lines not part of trial }
    Observations: TObservationArray; { The observations for the trial assignments, and those already assigned; standard fitting process is applied to these. }
    UndoPosition: Integer;
    OriginalParameters: TParameters;
    Fitters: array of TAutoFitter; { If parallel search active, this contains the sub fitter }
    nObs: Integer; { The number of observations to apply the standard fitting process to }
    { Output: }
    nFitted: Integer; { The number of trial assignments }
    nFail: Integer;
    nTrials: Int64;
    Trial: Int64; { Progress indicator }
    logNtrials: Double;
    nCheckOK: TIntegerArray; { Number of check lines that were accepted }
    TrialNo: TInt64Array; { Trial id; Can be set to nil to not track }
    FitIndices: TIntegerArrayArray; { Assignment for each trial; index is [Trial][Line] }
    Best: TDoubleArrayArray; { Indexes are [ trial ] [ Parameters ], with the first two "parameters" Sum(Obs-Calc) and Sum(Intensity) }
    StatusCount: array [TAutoFitStatus] of Integer; { Count of each possible failure types }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function ReadProperties(Node: TDOMNode; CheckUnread: Boolean = True):Integer; override;
    function WriteProperties(Parent: TDOMNode; AllParameters: Boolean = True; WithChildren: Boolean = True): TDOMElement; override;
    function HelpOnSetting(AIndex: Integer): string; override;
    procedure SetObservations(var TheObservations: TObservationArray);
    procedure SetLineList(Plot: TExperimentalPlot);
    function ComputeSearchRange: Integer; { Result is number of lines for which no assignment is possible }
    procedure Search(Offset, nSteps: Integer; var Problem: Exception; ShowStatus: TOnShowStatus = nil);
    procedure Closest;
    procedure ShowSVD(ShowDerivatives: Boolean);
    function SourceName: string;
    function AssignedLines(out Wts: TDoubleArray): TDoubleArray;
    function ConvertIndices(const Lines: array of Double):TIntegerArrayArray;
    function CountDifferences(Fit1, Fit2: Integer): Integer;
    procedure SetParametersFrom(Index: Integer; SetStdDev: Boolean);
    procedure WriteInfo(var FF: Text); { Details of fit }
    property nBest: Integer read FnBest write SetnBest; { The number of trials kept }
    property OnUpdateStatus: TStatusEvent read FOnUpdateStatus write FOnUpdateStatus;
  published
    property SearchWindow:  Double index  1 read GetParameter write SetParameter;
    property Irange:        Double index  2 read GetParameter write SetParameter;
    property Keep:       Integer index 1 read GetSetting write SetSetting;
    property MaxBlends:  Integer index 2 read GetSetting write SetSetting;
    property LimitRange: Boolean index 3 read GetFlag write SetFlag;
    property KeepOrder:  Boolean index 4 read GetFlag write SetFlag;
    property MaxCycles:  Integer index 5 read GetSetting write SetSetting;
    property Verbose:    Boolean index 6 read GetFlag write SetFlag;
    property FitDetails: Boolean index 7 read GetFlag write SetFlag;


  end;

const
  AutoFitStatusNames : array [TAutoFitStatus] of string = (
    'OK', 'Lines out of order', 'Too many blends', 'Parameter moved out of range',
    'Intensity out of range', 'Fit error out of range', 'Undetermined Paramteters in fit',
    'Calculation of check lines failed');

implementation

uses StrUtils, Math, UCompat, Progress, Profile, ShellSort, MatrixVec,
     PeakFinders, Units, TextAssign, TabbedLine, PGWork, Sysinfo;

var
  AutoFitTime: string = 'Auto fit search';
  FullAutoFitTime: string = 'Auto fit';
  FullFitTime: string = 'Full fits';
  FitAssignmentTime: string = 'Fit trial assignment';
  TestAssignmentTime: string = 'Test trial assignment';

type

  TTrialSorter = class(TShellSorter)
    AutoFitter: TAutoFitter;
    function InOrder(I, J: Integer): Boolean; override;
    class procedure SortBest(TheAutoFitter: TAutoFitter);
  end;

function TTrialSorter.InOrder(I, J: Integer): Boolean;
begin
  with AutoFitter do begin
    if nCheckOK[I] = nCheckOK[J] then
      Result := Best[I][0] <= Best[J][0]
    else
      Result := nCheckOK[I] > nCheckOK[J];
    if Result then
      Exit;
    SwapValues(nCheckOK[I], nCheckOK[J], SizeOf(Integer));
    if Length(TrialNo) > 0 then
      SwapValues(TrialNo[I], TrialNo[J], SizeOf(Int64));
    SwapValues(Best[I], Best[J], SizeOf(Best[I]));
    SwapValues(FitIndices[I], FitIndices[J], SizeOf(FitIndices[I]));
  end;
end;

class procedure TTrialSorter.SortBest(TheAutoFitter: TAutoFitter);
begin
  with TTrialSorter.Create do
    try
      AutoFitter := TheAutoFitter;
      Sort(AutoFitter.nBest);
    finally
      Free;
    end;
end;

{ TAutoFitter }

procedure TAutoFitter.MergeBest(Source: TAutoFitter);
var
  Cur: Integer;
  i: Integer;
  afS: TAutoFitStatus;
begin
  Cur := nBest;
  nBest := nBest + Source.nBest;
  if Length(Source.TrialNo) > 0 then
    SetLength(TrialNo, nBest);
  for i := Cur to nBest - 1 do begin
    nCheckOK[i] := Source.nCheckOK[i-Cur];
    if Length(Source.TrialNo) > 0 then
      TrialNo[i] := Source.TrialNo[i-Cur];
    Best[i] := Source.Best[i-Cur];
    FitIndices[i] := Source.FitIndices[i-Cur];
  end;
  TTrialSorter.SortBest(Self);
  nBest := Min(nBest, Keep);
  Inc(nFail, Source.nFail);
  for afS := Low(StatusCount) to High(StatusCount) do
    Inc(StatusCount[afS], Source.StatusCount[afS]);
end;

function TAutoFitter.NextAssignment(nSteps: Integer): Boolean;
var
  i: Integer;
begin
  Result := False;
  while (nSteps > 0) and not Result do begin
    { Bump to next possible assignment }
    for i := 0 to nObs-1 do begin
      if (TA[i] >= 0) and (Obs0[i].Weight <= 0) then begin
        Inc(TA[i]);
        if (TA[i] > High(f)) or (f[TA[i]] > Obs0[i].Frequency+SearchWindow) then begin
          TA[i] := FP[i];
        end else
          Break;
      end;
      Result := i = nObs-1;
    end;
    Dec(nSteps);
  end;
end;

procedure TAutoFitter.SetLineList(Plot: TExperimentalPlot);
var
  i: Integer;
begin
  LineList.Name := ExtractFileName(Plot.Name);
  SetLength(LineList.Frequency, Plot.EndPoint-Plot.StartPoint+1);
  SetLength(LineList.Intensity, Length(LineList.Frequency));
  if Length(Plot.Weights) > 0 then
    SetLength(LineList.Weights, Length(LineList.Frequency))
  else
    SetLength(LineList.Weights, 0);
  for i := 0 to High(LineList.Frequency) do begin
    LineList.Frequency[i] := Convert(Plot.LocalFrequency(i+Plot.StartPoint), Plot.Units, Mixture.Units);
    LineList.Intensity[i] := Plot.AdjustIntensity(Plot.Intensity[i+Plot.StartPoint]);
    if Length(LineList.Weights) > 0 then
      LineList.Weights[i] := ConvertWidth(Plot.Weights[i+Plot.StartPoint], Plot.LocalFrequency(i+Plot.StartPoint), Plot.Units, Mixture.Units);
  end;
  LineList.Sort;
  f := LineList.Frequency;
  Intensity := LineList.Intensity;
  Weights := LineList.Weights;
end;

procedure TAutoFitter.SetnBest(AValue: Integer);
var
  i: Integer;
begin
  SetLength(nCheckOK, AValue);
  SetLength(FitIndices, AValue);
  SetLength(Best, AValue, 2+Length(OriginalParameters));
  for i := FnBest to AValue-1 do
    nCheckOK[i] := -1;
  if Length(TrialNo) > 0 then
    SetLength(TrialNo, AValue);
  FnBest := AValue;
end;

procedure TAutoFitter.Search(Offset, nSteps: Integer; var Problem: Exception; ShowStatus: TOnShowStatus);
var
  i, j, LastTA, nBlends, Cycles, nA, nO, LP, P, nIaverage: Integer;
  CurWorst: Integer; { Index into best of current worst fit }
  ResetNeeded, Complete, CalcOK: Boolean;
  Calc: TColumnVector; { For working out calculated test values }
  Fitter: TAutoFitFitter;
  Status: TAutoFitStatus;
  StatusInfo: Integer;
  FitStatus, LogString: string;
  OldSD, LastSD, SumA, SumI, V, R, fMin, fMax, Iaverage: Double;
  RLabel: string;

  function CheckInRange:TAutoFitStatus;
  var
    i: Integer;
  begin
    for i := 0 to High(Mixture.FittingParameters) do
      if (Fitter.InitialParameters[i].StdDev > 0) and (Abs(Fitter.par[i+1] - Mixture.FittingParameters[i].Value) > Fitter.InitialParameters[i].StdDev) then begin
        Result := afParameter;
        StatusInfo := i;
        Exit;
      end;
    Result := afOK;
  end;
var
  ClonedMixtures: TMixtureArray;
  Worker: TPgopherWorker;
  nSplit: Integer;
begin
  nFail := 0;
  for Status := Low(StatusCount) to High(StatusCount) do
    StatusCount[Status] := 0;
  OriginalParameters := nil;
  if nFitted > 0 then begin
    Mixture.FillParameters;
    OriginalParameters := Copy(Mixture.FittingParameters);
    for i := 0 to High(OriginalParameters) do
      with OriginalParameters[i] do
        Name := MNode.FullParameterName(Index);
    if IsMainThread and (nWorkers > 0) and (nSteps = 1) then begin
      { Parallelize calculation }
      ClonedMixtures := nil;
      nBest := 0;
      nSplit := Min(nWorkers, nCPUs);
      TStart(FullAutoFitTime);
      try
        ClonedMixtures := Mixture.Clone(nSplit);
        SetLength(Fitters, nSplit);
        for i := 0 to High(ClonedMixtures) do begin
          Fitters[i] := ClonedMixtures[i].ComponentWithName(Name, TAutoFitter, nil) as TAutoFitter;
          Fitters[i].LoadStartingValues(Self);
          Worker := FindWorkerWait as TPgopherWorker;
          if Assigned(Worker) then
            Worker.AutoFit(Fitters[i], i, nSplit, Problem)
          else
            raise NotYet; { Internal error }
        end;
        WaitNoBusyWorkers(ShowStatus);
        for i := 0 to High(ClonedMixtures) do
          MergeBest(Fitters[i]);
      finally
        for i := 0 to High(ClonedMixtures) do
          ClonedMixtures[i].Free;
        SetLength(Fitters,0);
        TEnd(FullAutoFitTime);
      end;
      Exit;
    end;
  end;
  FnBest := 0; { Force clear }
  nBest := Keep;
  fMin := f[0];
  fMax := f[High(f)];
  SetLength(TrialNo, nBest); { Note - TrialNo can be set to nil to omit record }
  Trial := Offset-nSteps;
  CurWorst := -1;
  for i := 0 to nBest-1 do
    nCheckOK[i] := -1;

  Calc := TColumnVector.Create(Length(Obs0));
  Fitter := nil;
  UndoPosition := -1;
  try
    TStart(AutoFitTime);
    if nFitted > 0 then begin
      Fitter := TAutoFitFitter.Create(Mixture.FittingParameters, Observations, Mixture, 0);
      UndoPosition := Mixture.UndoPosition;
      Mixture.PrepareForFit(ftLine, nil);
    end;
    Complete := NextAssignment(Offset);
    while not Complete do begin
      Inc(Trial,nSteps);
      if (nSteps = 1) and Assigned(FOnUpdateStatus) and (Trial mod 1024 = 0) then
        FOnUpdateStatus(Self, Trial)
      else
        CheckForAbort;
      if Verbose then begin
        LogString := ''; { Accumulate single line in string to allow for multithreaded output }
        for i := 0 to nObs-1 do
          if (TA[i] >= 0) and (Obs0[i].Weight > 0) then
            LogString := LogString + AtLeast(f[TA[i]],14,Mixture.Precision);
        LogString := LogString + ' ';
      end;
      ResetNeeded := False;
      Status := afOK;
      StatusInfo := 0;
      if KeepOrder then begin
        LastTA := -1;
        i := 0;
        while i < nObs do begin
          j := TA[i];
          if j >= 0 then begin
            if j <= LastTA then begin
              Status := afOrder;
              Break;
            end;
            LastTA := j;
          end;
          Inc(i);
        end;
      end else if (MaxBlends >= 0) and (MaxBlends < nFitted) then begin
        nBlends := 0;
        for i := 1 to nObs-1 do
          if TA[i] >= 0 then begin
            for j := 0 to i-1 do
              if (TA[i] = TA[j]) and ((Obs0[i].Weight <= 0) or (Obs0[j].Weight <= 0)) then
                Inc(nBlends);
            if nBlends > MaxBlends then begin
              Status := afBlends;
              StatusInfo := TA[i];
              Break;
            end;
          end;
      end;
      Iaverage := 0; { Average intensity of fit transitions }
      nIaverage := 0; { Number of assigned fit transitions }
      if (Status = afOK) and (Irange > 0) then begin
        { Find average intensity of fit transitions }
        for i := 0 to nObs-1 do begin
          j := TA[i];
          if j >= 0 then begin
            Inc(nIaverage);
            Iaverage := Iaverage + Intensity[j];
          end;
        end;
        { Check all lines reasonably close to average intensity }
        if nIaverage > 0 then begin
          Iaverage := Iaverage/nIaverage;
          for i := 0 to nObs-1 do begin
            j := TA[i];
            if j >= 0 then begin
              R := Intensity[j]/Iaverage;
              if (R <= Irange) or (R>= 1/Irange) then begin
                Status := afIntensity;
                Break;
              end;
            end;
          end;
        end;
      end;
      if Status = afOK then begin
        TStart(FitAssignmentTime);
        for i := 0 to nObs-1 do begin
          j := TA[i];
          if j >= 0 then
            Observations[i].Frequency := f[j];
        end;
        Cycles := 0;
        OldSD := 0;
        if Assigned(Fitter) then
          Fitter.PrepareResiduals(Log, FitDetails, False);
        if (Length(Mixture.FittingParameters) > 0) and Assigned(Fitter) then begin
          Inc(Cycles);
          FitStatus := Fitter.DoLeastSQ(Mixture.FittingParameters, Mixture.Precision, False, not FitDetails);
          ResetNeeded := True;
          if Fitter.nDiscard > 0 then
            Status := afSVD
          else if Fitter.UnWtNewSD > aW then
            Status := afResidual
          else if LimitRange then
            Status := CheckInRange;
          if Status = afOK then begin
            Fitter.ApplyParameters(Mixture.FittingParameters, False);
            ResetNeeded := True;
            LastSD := Fitter.OldSD;
            Fitter.PrepareCalculated;
            Fitter.PrepareResiduals(Log, FitDetails, False);
            { Test here exits if first cycle was sufficient }
            if Fitter.OldSD > aW/10 then begin
              Inc(Cycles);
              { Do one cycle with old derivatives }
              FitStatus := Fitter.DoLeastSQ(Mixture.FittingParameters, Mixture.Precision, False, not FitDetails);
              if Fitter.nDiscard > 0 then
                Status := afSVD
              else begin
                Fitter.ApplyParameters(Mixture.FittingParameters, False);
                TStart(FullFitTime);
                while (Cycles < MaxCycles) and not((Fitter.OldSD < aW/10) or (Abs(LastSD - Fitter.OldSD) < aW/10)) and (Status = afOK) do begin
                  LastSD := Fitter.OldSD;
                  Fitter.PrepareCalculated;
                  Fitter.PrepareResiduals(Log, FitDetails, False);
                  if (Fitter.OldSD < aW/10) or (Abs(LastSD - Fitter.OldSD) < aW/10) then
                    Break; { Exit here implies the previous cycle is fine }
                  Fitter.CalculateDerivatives(Mixture.FittingParameters);
                  Inc(Cycles);
                  FitStatus := Fitter.DoLeastSQ(Mixture.FittingParameters, Mixture.Precision, False, not FitDetails);
                  if Fitter.nDiscard > 0 then
                    Status := afSVD
                  else
                    Fitter.ApplyParameters(Mixture.FittingParameters, False);
                end;
                TEnd(FullFitTime);
                if (Status = afOK) and LimitRange then
                  Status := CheckInRange;
              end;
            end;
          end;
        end else
          FitStatus := '';
        TEnd(FitAssignmentTime);
        nA := 0; { Number of accepted lines }
        nO := 0; { Number of observations (including previous assignments) }
        SumA := 0;
        SumI := 0;
        LP := -1;
        if Status = afOK then begin
          TStart(TestAssignmentTime);
          if Assigned(Fitter) then begin
            if not Fitter.CalculateAllObs(Calc, Obs0, nObs) then
              Status := afCheckCalc
          end else
            for i := nObs to High(Obs0) do
              if CalculateObs(Obs0[i], V, RLabel, False, obLine) then
                Calc[i+1] := V
              else begin
                Status := afCheckCalc;
                Break;
              end;
          if Status = afOK then
            for i := nObs to High(Obs0) do begin
              V := Calc[i+1];
              //CalcOK := CalculateObs(Obs0[i], V, RLabel, False, obLine);
              if Obs0[i].Weight > 0 then begin
                { TODO: Special case with nFitted = 0, nObs > 0 }
                { Already existing assignment }
                CalcOK := True;
                P := PU[i];
                if P < 0 then
                  SumA := SumA + Sqr(Obs0[i].Frequency-V);
                  { SumI not changed }
              end else begin
                PU[i] := -1;
                if (V >= fMin-aw) and (V <= fMax+aW) then begin
                  P := ClosestPoint(V, f, True);
                  CalcOK := Abs(f[P]-V) <= aW; { Within acceptance window }
                  if CalcOK then begin
                    if nIaverage > 0 then begin
                      R := Intensity[P]/Iaverage;
                      CalcOK := (R <= Irange) and (R >= 1/Irange);
                    end;
                    if CalcOK then
                      Inc(nA);
                  end;
                end else
                  CalcOK := False;
              end;
              if CalcOK then begin
                if P >= 0 then begin
                  if KeepOrder and (P <= LP) then begin
                    Status := afOrder;
                    Break;
                  end;
                  LP := P;
                  SumA := SumA + Sqr(f[P]-V);
                  SumI := SumI + Intensity[P];
                  PU[i] := P;
                end;
                Inc(nO);
              end;
            end;
          if nO > 0 then
            SumA := Sqrt(SumA/nO)
          else
            SumA := LastSD;
          // OK := OK and ((nA > 0) or (Length(Obs0) = nFit));
          if (Status = afOK) and ((CurWorst < 0) or (nA > nCheckOK[CurWorst])
                                  or ((nA = nCheckOK[CurWorst]) and (SumA < Best[CurWorst][0]))) then begin
            if CurWorst < 0 then begin
              { Initial phase - finding initial 10 }
              i := 0;
              while (i < nBest) and (nCheckOK[i] >= 0) do
                Inc(i);
              if i = nBest-1 then
                CurWorst := 0; { Flag just filled last empty slot }
            end else
              i := CurWorst;
            nCheckOK[i] := nA;
            if Length(TrialNo) > 0 then
              TrialNo[i] := Trial;
            Best[i][0] := SumA;
            Best[i][1] := SumI;
            for j := 0 to High(OriginalParameters) do
              Best[i][j+2] := Mixture.FittingParameters[j].Value;
            for j := 0 to nObs-1 do
              PU[j] := TA[j];
            FitIndices[i] := Copy(PU);
            if CurWorst >= 0 then begin
              { Find current worst }
              CurWorst := 0;
              for i := 1 to nBest-1 do
                if (nCheckOK[i] < nCheckOK[CurWorst])  then
                  CurWorst := i
                else if (nCheckOK[i] = nCheckOK[CurWorst]) and (Best[i][0] > Best[CurWorst][0]) then
                  CurWorst := i;
            end;
          end;
          TEnd(TestAssignmentTime);
        end else
          Inc(nFail);
        if Verbose then begin
          LogString := LogString + PadLeft(IntToStr(Cycles),4) + PadLeft(IntToStr(nA),5)
                       + DtosMA(SumA, 10, -1) + DtosMA(SumI, 10, -1);
          for i := 0 to High(Mixture.FittingParameters) do
            LogString := LogString + DtosMA(Mixture.FittingParameters[i].Value, 15, -1);
          if Status <> afOK then
            LogString := 'Fail: ' + AutoFitStatusNames[Status] + ' ' + LogString + ' ' + FitStatus
          else
            LogString := LogString + ' ' + FitStatus;
        end;
        if ResetNeeded then begin
          Mixture.RestartFitAt(UndoPosition);
          Fitter.Reset;
        end;
      end else begin
        Inc(nFail);
        if Verbose then
          LogString := 'Fail: ' + AutoFitStatusNames[Status] + ' ' + LogString;
      end;
      if Verbose then begin
        LogString := 'Trial'+PadLeft(IntToStr(Trial),21)+': '+LogString;
        WriteLn(Log, LogString);
      end;
      Inc(StatusCount[Status]);
      Complete := (nFitted = 0) or NextAssignment(nSteps);
    end;
  finally
    Calc.Free;
    Fitter.Free;
    if UndoPosition >= 0 then begin
      Mixture.LoadParameters(UndoPosition);
      Mixture.DiscardFits(Mixture.UndoPosition-UndoPosition-1);
    end;
    TEnd(AutoFitTime);
  end;
  TTrialSorter.SortBest(Self);
  while (nBest > 0) and (nCheckOK[nBest-1] < 0) do
    nBest := nBest - 1;
end;

procedure TAutoFitter.Closest;
var
  i, P: Integer;
begin
  nBest := 2;
  nCheckOK[0] := 0;
  nCheckOK[1] := 0;
  Best[0][0] := 0; { For obs-calc }
  Best[0][1] := 0; { For SumI }
  for i := 2 to High(Best[0]) do
    Best[0][i] := OriginalParameters[i-2].Value;
  Best[1] := Copy(Best[0]);
  SetLength(FitIndices, 2, nObs);
  for i := 0 to nObs-1 do
    if Obs0[i].Weight <= 0 then begin
      P := ClosestPoint(Obs0[i].Frequency, f);
      FitIndices[1][i] := P;
      Inc(nCheckOK[1]);
      Best[1][0] := Best[1][0] + Sqr(Obs0[i].Frequency-f[P]);
      Best[1][1] := Best[1][1] + Intensity[P];
      if Abs(f[P]-Obs0[i].Frequency) < SearchWindow then begin
        Inc(nCheckOK[0]);
        FitIndices[0][i] := P;
        Best[0][0] := Best[0][0] + Sqr(Obs0[i].Frequency-f[P]);
        Best[0][1] := Best[0][1] + Intensity[P];
      end else
        FitIndices[0][i] := -1;
    end;
  if nCheckOK[1] > 0 then begin
    Best[1][0] := Sqrt(Best[1][0]/nCheckOK[1]);
    if nCheckOK[0] > 0 then
      Best[0][0] := Sqrt(Best[0][0]/nCheckOK[0]);
  end;
  if nCheckOK[1] = nCheckOK[0] then
    nBest := 1; { Both rows identical }
end;

type

  TObsSorter = class(TShellSorter)
    AutoFitter: TAutoFitter;
    Offset: Integer;
    OnTrialFlag: Boolean;
    function InOrder(I, J: Integer): Boolean; override;
  end;

function TObsSorter.InOrder(I, J: Integer): Boolean;
begin
  Inc(I,Offset);
  Inc(J,Offset);
  if OnTrialFlag then
    if (ofTrial in AutoFitter.Obs0[I].Flags) xor (ofTrial in AutoFitter.Obs0[J].Flags) then
      Result := ofTrial in AutoFitter.Obs0[I].Flags
    else if (AutoFitter.Obs0[I].Weight > 0) xor (AutoFitter.Obs0[J].Weight > 0) then
      Result := AutoFitter.Obs0[I].Weight > 0
    else
      Result := True
  else
    Result := AutoFitter.Obs0[I].Frequency <= AutoFitter.Obs0[J].Frequency;
  if not Result then
    SwapValues(AutoFitter.Obs0[I], AutoFitter.Obs0[J], SizeOf(TObservation));
end;

procedure TAutoFitter.SetObservations(var TheObservations: TObservationArray);
begin
  Obs0 := TheObservations;
  with TObsSorter.Create do
    try
      AutoFitter := Self;
      Offset := 0;
      OnTrialFlag := True;
      Sort(Length(Obs0));
      nFitted := 0;
      while (nFitted < Length(Obs0)) and (ofTrial in Obs0[nFitted].Flags) do
        Inc(nFitted);
      nObs := nFitted;
      while (nObs < Length(Obs0)) and (Obs0[nObs].Weight > 0) do
        Inc(nObs);
      if KeepOrder then begin
        { Require assignments to be in order of original estimates }
        OnTrialFlag := False;
        Sort(nObs);
        Offset := nObs;
        Sort(Length(Obs0)-Offset);
      end;
      Observations := Copy(Obs0,0,nObs);
    finally
      Free;
    end;
end;

procedure TAutoFitter.SetParametersFrom(Index: Integer; SetStdDev: Boolean);
var
  i: Integer;
  viInfo: TValueInfo;
  S: string;
begin
  for i := 0 to High(OriginalParameters) do begin
    viInfo := Mixture.SubPropInfo(OriginalParameters[i].Name, S);
    if not Assigned(viInfo.MNode) or (viInfo.viType <> viParameter) then
      if Index > Length(Best) then begin
        { Forcing valid OriginalParameters, but not valid so simply empty }
        OriginalParameters := nil;
        Exit;
      end else
        raise EReportException.Create(OriginalParameters[i].Name +' not found');
    OriginalParameters[i].MNode := viInfo.MNode;
    OriginalParameters[i].Index := viInfo.Index;
    if Index < 0 then
      viInfo.MNode.Parameters[viInfo.Index] := OriginalParameters[i].Value
    else if Index < Length(Best) then
      viInfo.MNode.Parameters[viInfo.Index] := Best[Index][i+2];
    if SetStdDev then
      viInfo.MNode.StdDev[viInfo.Index] := OriginalParameters[i].StdDev;
  end;
end;

function TAutoFitter.ComputeSearchRange: Integer;
var
  i, P: Integer;
begin
  SetLength(TA, nObs);
  SetLength(FP, nObs);
  nTrials := 1;
  logNtrials := 0;
  Result := 0; { Result is number of lines for which no assignment is possible }
  for i := 0 to nObs-1 do begin
    if Obs0[i].Weight > 0 then begin
      { Don't search on assigned observations }
      TA[i] := ClosestPoint(Obs0[i].Frequency, f, True);
      if f[TA[i]] <> Obs0[i].Frequency then
        TA[i] := -1; { Not in linelist }
      FP[i] := TA[i];
      P := FP[i]+1;
    end else begin
      TA[i] := ClosestPoint(Obs0[i].Frequency-SearchWindow, f, True);
      if f[TA[i]] < Obs0[i].Frequency-SearchWindow then begin
        Inc(TA[i]);
        { No lines in range - take closest line }
        if TA[i] >= Length(f) then begin
          Dec(TA[i]);
          Inc(Result);
        end else
          if f[TA[i]] > Obs0[i].Frequency+SearchWindow then begin
            if f[TA[i]]-Obs0[i].Frequency > Obs0[i].Frequency - f[TA[i]-1] then
              Dec(TA[i]);
            Inc(Result);
          end;
      end;
      FP[i] := TA[i];
      P := FP[i]+1;
      while (P < Length(f)) and (f[P] <= Obs0[i].Frequency+SearchWindow) do
        Inc(P);
      Observations[i].Weight := 1;
    end;
    logNtrials := logNtrials + log10(P-FP[i]);
    if logNtrials < 18 then
      nTrials := nTrials * (P-FP[i]);
  end;
  SetLength(PU, Length(Obs0));
  { TODO: Loop below no longer required }
  for i := nObs to High(Obs0) do
    { Existing assignments not part of trial - check for lines currently in linelist }
    if Obs0[i].Weight > 0 then begin
      PU[i] := ClosestPoint(Obs0[i].Frequency, f, True);
      if f[PU[i]] <> Obs0[i].Frequency then
        PU[i] := -1; { Not in linelist }
    end;
end;

constructor TAutoFitter.Create(AOwner: TComponent);
begin
  SavedProperties := SavedProperties + ',SearchWindow=100,Keep=10,LimitRange=True,MaxCycles=20';
  // DimensionlessParameters := 'Concentration,' + DimensionlessParameters;
  inherited Create(AOwner);
  if AOwner <> nil then
    Mixture := RootNode as TMixture;
  LineList := TExperimentalPlot.Create(nil);
end;

destructor TAutoFitter.Destroy;
begin
  FreeAndNil(LineList);
  inherited;
end;

procedure TAutoFitter.LoadStartingValues(Source: TAutoFitter);
  procedure FixManifold(var Manifold: TManifold);
  var
    viInfo: TValueInfo;
    ErrMsg: string;
  begin
    if Manifold = nil then
      Exit;
    viInfo := Mixture.SubPropInfo(Manifold.FullHeirarchicName(True,True)+'.Colour',ErrMsg,False);
    if viInfo.MNode = nil then
      raise Exception.Create('Can''t find '+Manifold.FullHeirarchicName+':'+ErrMsg)
    else
      Manifold := viInfo.MNode as TManifold;
  end;
var
  i: Integer;
begin
  nFitted := Source.nFitted;
  nObs := Source.nObs;
  aW := Source.aW;
  f := Source.f;
  Intensity := Source.Intensity;
  Weights := Source.Weights;
  FP := Source.FP;
  PU := Copy(Source.PU);
  TA := Copy(Source.TA);
  Obs0 := Copy(Source.Obs0);
  Observations := Copy(Source.Observations);
  { Update Manifold pointers }
  for i := 0 to High(Obs0) do begin
    FixManifold(Obs0[i].ManifoldL);
    FixManifold(Obs0[i].ManifoldU);
  end;
  for i := 0 to High(Observations) do begin
    Observations[i].ManifoldL := Obs0[i].ManifoldL;
    Observations[i].ManifoldU := Obs0[i].ManifoldU;
  end;
end;

procedure TAutoFitter.ShowSVD(ShowDerivatives: Boolean);
var
  i: Integer;
begin
  with TAutoFitFitter.Create(Mixture.FittingParameters, Observations, Mixture, 0) do
    try
      MakeSVD;
      ShowSVD(Parameters);
    finally
      Free;
    end;
  if not ShowDerivatives then
    Exit;
  { Calculation with all observations }
  for i := 0 to High(Obs0) do
    if Obs0[i].Weight < 0 then
      Obs0[i].Weight := 1;
  with TAutoFitFitter.Create(Mixture.FittingParameters, Obs0, Mixture, 0) do
    try
      ShowDerivativeHeader(Log, 'Parameters:');
      ShowResidualHeader(Log, 'Derivatives of all observations', False);
      for i := 0 to nObs - 1 do begin
        ShowResidual(Log, i, 0, False);
        ShowDerivative(Log, i);
      end;
      WriteLn(Log, 'If all lines assigned:');
      MakeSVD;
      ShowSVD(Parameters);
    finally
      Free;
    end;
end;

function TAutoFitter.SourceName: string;
begin
  Result := LineList.Name;
end;

function TAutoFitter.AssignedLines(out Wts: TDoubleArray): TDoubleArray;
var
  n, i, j, p: Integer;
  V: Double;
begin
  { For saving, return a list of frequencies used in any assignment }
  Result := Copy(f);
  Wts := Copy(Weights);
  n := 0;
  for i := 0 to nBest - 1 do
    for j := 0 to High(FitIndices[i]) do
      if FitIndices[i][j] >= 0 then begin
        V := f[FitIndices[i][j]];
        if n > 0 then begin
          p := ClosestPoint(V, Result, False, 0, n-1);
          if Result[p] = V then
            Continue;
        end;
        Result[n] := V;
        if Length(Wts) > 0 then
          Wts[n] := Weights[FitIndices[i][j]];
        Inc(n);
      end;
  SetLength(Result, n);
  if Length(Wts) > 0 then begin
    { TODO: save intensity information also }
    SetLength(Wts, n);
    Sort2(Result, Wts);
  end else
    Sort(Result);
end;

function TAutoFitter.ConvertIndices(const Lines: array of Double):TIntegerArrayArray;
var
  i, j: Integer;
begin
  { For saving convert fit indices to index into array produced by AssignedLines(); }
  Result := Copy(FitIndices);
  for i := 0 to High(Result) do begin
    Result[i] := Copy(FitIndices[i]);
    for j := 0 to High(FitIndices[i]) do
      if Result[i][j] >= 0 then
        Result[i][j] := ClosestPoint(f[FitIndices[i][j]], Lines, True);
  end;
end;

function TAutoFitter.CountDifferences(Fit1, Fit2: Integer): Integer;
  function CD(F1, F2: Integer): Integer;
  var
    i, j: Integer;
    Found: Boolean;
  begin
    Result := 0;
    for i := 0 to High(FitIndices[F1]) do
      if FitIndices[F1][i] >= 0 then begin
        Found := False;
        for j := 0 to High(FitIndices[F2]) do begin
          Found := FitIndices[F1][i] = FitIndices[F2][j];
          if Found then
            Break;
        end;
        if not Found then
          Inc(Result);
      end;
  end;
begin
  Result := CD(Fit1,Fit2) + CD(Fit2,Fit1);
end;

function TAutoFitter.ReadProperties(Node: TDOMNode; CheckUnread: Boolean): Integer;
var
  i, j, nObs0: Integer;
  S, OriginalParametersString: AnsiString;
  P: PAnsiChar;
begin
  TryStrToInt(ReadValue(Node, 'nFit', '0'), nFitted);
  TryStrToInt(ReadValue(Node, 'nObs', '0'), nObs);
  nObs := Max(nFitted, nObs); { Allow for early files without nObs saved }
  TryStrToInt(ReadValue(Node, 'nFail', '0'), nFail);
  TryStrToInt64(ReadValue(Node, 'nTrials', '0'), nTrials);
  Best := ReadArrayArray(ReadValue(Node, 'Best', ''));
  LineList.Name := ReadValue(Node, 'Source', '');
  f := ReadArray(ReadValue(Node, 'Lines', ''));
  Intensity := nil;
  Weights := ReadArray(ReadValue(Node, 'Weights', ''));
  FitIndices := ReadIntegerArrayArray(ReadValue(Node, 'Assignments', ''));
  TrialNo := ReadInt64Array(ReadValue(Node, 'TrialNo', ''));
  OriginalParameters := nil;
  OriginalParametersString := ReadValue(Node, 'OriginalParameters', '');
  P := PAnsiChar(OriginalParametersString);
  while P^ <> #0 do begin
    S := SkipItem(P,nil,tlTab);
    if S <> '' then begin
      SetLength(OriginalParameters, Length(OriginalParameters)+1);
      with OriginalParameters[High(OriginalParameters)] do begin
        Name := S;
        S := SkipItem(P,nil,tlTab);
        TryStrToFloatEx(S, Value);
        if not (P^ in [#10, #13]) then begin
          S := SkipItem(P,nil,tlTab);
          TryStrToFloatEx(S, StdDev);
        end else
          StdDev := -1;
      end;
    end;
    if P = nil then
      Break;
    SkipLineEnding(P);
  end;
  nBest := Length(Best);
  nObs0 := -1;
  for i := 0 to High(FitIndices) do begin
    nCheckOK[i] := Length(FitIndices[i]) - nObs;
    for j := nFitted to High(FitIndices[i]) do
      if FitIndices[i][j] < 0 then
        Dec(nCheckOK[i])
      else
        nObs0 := Max(nObs0,FitIndices[i][j]);
  end;
  SetLength(Obs0, nObs0+1);
  S := ReadValue(Node, 'Observations', '');
  P := PAnsiChar(S);
  nObs0 := 0;
  if Length(Obs0) > 0 then  
    while P^ <> #0 do begin
      if nObs0 > High(Obs0) then
        SetLength(Obs0, nObs0+1); { Shouldn't happen, but prevent crash on loading }
      if TryStrToInt(SkipItem(P,nil,tlTab),Obs0[nObs0].LineNumber) then begin
        if Obs0[nObs0].LineNumber >= 0 then
          with Obs0[nObs0] do begin
            FileName := LineListFileName;
            Weight := -1;
          end;
        Inc(nObs0);
      end;
      if P = nil then
        Break;
      if SkipToLineEnd(P, nil) then
        SkipLineEnding(P);
    end;
  SetLength(Obs0,nObs0);
  Result := inherited ReadProperties(Node, CheckUnread);
end;

procedure TAutoFitter.WriteInfo(var FF: Text);
var
  afS: TAutoFitStatus;
  i, j: Integer;
begin
  WriteLn(FF, 'Search on ', nFitted, ' transitions, ', nObs-nFitted, ' already assigned, ', Length(Obs0)-nObs, ' check transitions');
  Write(FF, '  nOK Residual     SumI      Trial');
  for i := 0 to High(OriginalParameters) do
    Write(FF, ' ', OriginalParameters[i].Name);
  WriteLn(FF);
  for j := 0 to nBest-1 do begin
    Write(FF, nCheckOK[j]:5, ' ', DtosM(Best[j][0],8,-1,True), ' ', DtosM(Best[j][1],8,-1,True));
    if j < Length(TrialNo) then
      Write(FF, TrialNo[j]:11)
    else
      Write(FF, ' ':11);
    for i := 0 to High(OriginalParameters) do
      Write(FF, ' ', DtosM(Best[j][i+2],Length(OriginalParameters[i].Name),-1,True));
    WriteLn(FF);
  end;
  WriteLn(FF, 'Breakdown of ', nTrials, ' trials:');
  for afS := Low(StatusCount) to High(StatusCount) do
    WriteLn(FF, StatusCount[afS]:10, ' ', AutoFitStatusNames[afS]);
  WriteLn(FF);
end;

function TAutoFitter.WriteProperties(Parent: TDOMNode; AllParameters: Boolean; WithChildren: Boolean): TDOMElement;
var
  i: Integer;
  Lines, Wts: TDoubleArray;
  s: string;
begin
  Result := inherited WriteProperties(Parent, AllParameters, WithChildren);
  if nBest = 0 then
    Exit; { No current fit }
  WriteValue(Result, 'nFit', IntToStr(nFitted), '0');
  WriteValue(Result, 'nObs', IntToStr(nObs), '0');
  WriteValue(Result, 'nFail', IntToStr(nFail), '0');
  WriteValue(Result, 'nTrials', IntToStr(nTrials), '0');
  WriteValue(Result, 'Best', WriteArray(Best), '');
  WriteValue(Result, 'Source', SourceName, '');
  Lines := AssignedLines(Wts);
  WriteValue(Result, 'Lines', WriteArray(Lines), '');
  if Length(Wts) > 0 then
    WriteValue(Result, 'Weights', WriteArray(Wts), '');
  WriteValue(Result, 'Assignments', WriteArray(ConvertIndices(Lines)), '');
  WriteValue(Result, 'TrialNo', WriteArray(TrialNo), '');
  s := '';
  for i := 0 to High(OriginalParameters) do
    with OriginalParameters[i] do begin
      s := s + Name + #9 + FloatToStr(Value);
      if StdDev > 0 then
        s := s + #9 + FloatToStr(StdDev);
      s := s + LineEnding;
    end;
  WriteValue(Result, 'OriginalParameters', s, '');
  s := '';
  for i := 0 to High(Obs0) do
    with Obs0[i] do
      if (FileName = LineListFileName) and (Weight < 0) and (LineNumber >= 0) then
        s := s + IntToStr(LineNumber) + LineEnding
      else
        s := s + '-1' + LineEnding;
  WriteValue(Result, 'Observations', s, '');
end;

const
  AutoFitterSettingNames: array[1..7] of string = (
'Number of possible fits to keep',
'Maximum number of blends to allow',
'Limit range of fitted parameters to Std Dev',
'Force trial fits to be in same frequency order as initial calculated position',
'Maximum cycles on each trial fit',
'Display status of each possible assignment',
'Display details of each fit'
);

function TAutoFitter.HelpOnSetting(AIndex: Integer): string;
begin
  if (AIndex >= Low(AutoFitterSettingNames))
     and (AIndex <= High(AutoFitterSettingNames)) then
    Result := AutoFitterSettingNames[AIndex]
  else
    Result := inherited HelpOnSetting(AIndex);
end;

initialization
  TAutoFitter.RegisterClass(TMixture);
end.
