unit CheckImport;
{ Check an imported file yields the same .pgo file }

{$I utils.inc}{Copyright (C) 2009 C M Western; see commandline.pas for license}

interface

uses Classes, TextStrm, MNode, Mixture;

function CheckImported(IMix, Mixture: TMixture; var nTried: Integer; Tolerance: Double; var Failures: Integer; Script: TStrings):Double;
function CheckCalculatedParameters(Mixture: TMixture; S: TTextStream; var nTried: Integer; Tolerance: Double; var Failures: Integer;
                                   Flags: Integer { 1: Allow parameters to be equal in magnitude but opposite sign }
                                   ):Double;
procedure CompareNode(A, B: TMNode; var Failures: Integer; var MaxErr: Double; Tolerance: Double);
procedure ApplyScript(Mixture: TMixture; Script: TStrings);

implementation

uses SysUtils, Math, MathSup, UCompat, Worker, pgDOM, MNodeExpr, AutoFit;

function DiffNos(n1, n2, Tolerance: Double; var MaxErr: Double):Boolean;
var
  RelTol: Double;
begin
  if IsNanOrInf(n1) then begin
    Result := not IsNanOrInf(n2);
    Exit;
  end;
  if Abs(n1-n2) <= Tolerance then begin
    MaxErr := Max(MaxErr, Abs(n1-n2));
    Result := False;
    Exit;
  end;
  { Try again on fractional difference }
  RelTol := Max(Abs(n1),Abs(n2))*Tolerance;
  Result := Abs(n1-n2) >= RelTol;
  if RelTol > Tolerance then
    MaxErr := Max(MaxErr, Abs(n1-n2)/Max(Abs(n1),Abs(n2)))
  else
    MaxErr := Max(MaxErr, Abs(n1-n2));
end;

procedure CompareNode(A, B: TMNode; var Failures: Integer; var MaxErr: Double; Tolerance: Double);
  procedure Fail(Name: string);
  begin
    Inc(Failures);
    if Failures > 0 then begin
      WriteLn(Log, Name, ' in ', A.FullHeirarchicName, ':', A.ClassName);
    end;
  end;
  function CheckStr(Name, sA, sB: string):Boolean;
  var
    i: Integer;
  begin
    Result := sA = sB;
    if not Result then begin
      if (Pos(#13, sA) > 0) xor (Pos(#13, sB) > 0) then begin
        sA := StringReplace(sA, #13#10, #10, [rfReplaceAll]);
        sB := StringReplace(sB, #13#10, #10, [rfReplaceAll]);
        Result := sA = sB
      end;
      if not Result then
        if Length(sA)+Length(sB) > 100 then begin
          i := 1;
          while (i <= Min(Length(sA), Length(sb))) and (sA[i] = sB[i]) do
            Inc(i);
          Fail(Format('%s difference at %d/%d %d (%s %s):'+LineEnding+'%s'+LineEnding+'Other:'+LineEnding+'%s',
                      [Name, i, Length(sA), Length(sB), Copy(sA,i,10), Copy(sB,i,10), sA, sB]))
        end else
          Fail(Format('%s difference: "%s" "%s"', [Name, sA, sB]));
    end;
  end;
  function VarAndTol(const S: string; out Value, Tolerance: Double; var MaxVal: Double):Boolean;
  begin
    Result := True;
    if not TryStrToFloat(S,Value) then begin
      Tolerance := 0;
      if (CompareText(s,'Nan') = 0) or (CompareText(s,'Inf') = 0) then
        SetNan(Value)
      else
        Result := False;
      Exit;
    end;
    MaxVal := Max(MaxVal, Abs(Value));
    Tolerance := Precision(S)*1.6;
    { be more slack on very small values }
    Tolerance := Max(Tolerance, 1e-14*MaxVal);
  end;
  function CheckTokenStr(Name, sA, sB: DOMString):Boolean;
  var
    tA, tB: TTextStream;
    vA, eA, vB, eB, MaxVal: Double;
  begin
    Result := sA = sB;
    if Result then
      Exit;
    MaxVal := 0;
    tA := TTextStream.Create(TStringStream.Create(sA));
    tA.FlowToNextLine := True;
    tB := TTextStream.Create(TStringStream.Create(sB));
    tB.FlowToNextLine := True;
    try
      while not tA.AtEndOfFile do begin
        if ta.GetFreeItem xor tB.GetFreeItem then
          Fail('Missing token in '+Name+':'+ta.Token+'<>'+tb.Token);
        if ta.Token <> tB.Token then begin
          if ((Pos('.', ta.Token) > 0) or (Pos('.', tb.Token) > 0))
             and VarAndTol(ta.Token, vA, eA, MaxVal) and VarAndTol(tb.Token, vB, eB, MaxVal) then begin
            if (Abs(vA-vB) > Max(eA,eB)) and DiffNos(vA,vB,Tolerance,MaxErr) then
              Fail('Token difference in '+Name+':'+ta.Token+'<>'+tb.Token);
          end else
            Fail('Token mistmatch in '+Name+':'+ta.Token+'<>'+tb.Token);
        end;
      end;
      if not tB.AtEndOfFile then begin
        sB := Trim(tB.RestOfFile);
        if sB <> '' then
          Fail('Excess text in '+Name+':'+sB);
      end;
    finally
      tA.UnderlyingStream.Free;
      tB.UnderlyingStream.Free;
      tA.Free;
      tB.Free;
    end;
  end;

  function CheckInt(Name: string; iA, iB: Integer):Boolean;
  begin
    Result := iA = iB;
    if not Result then
      Fail(Format('%s difference: %d %d', [Name, iA, iB]));
  end;
  procedure CheckNo(Name: string; n1, n2, tol: Double);
  var
    Msg: string;
    Bad: Boolean;
  begin
    if DiffNos(n1, n2, tol, MaxErr) then
      Fail(Format('%s out of tolerance: %g %g %g (%g)', [Name, n1, n2, n1-n2, n1/n2]))
  end;
  procedure CheckArray(Name: string; A, B: TStringArrayArray);
  var
    i, j: Integer;
    n1, n2, Tol: Double;
  begin
    if not CheckInt(Name+' count', Length(A), Length(B)) then
      Exit;
    for i := 0 to High(A) do begin
      if not CheckInt(Name+' count '+IntToStr(i), Length(A[i]), Length(B[i])) then
        Exit;
      for j := 0 to High(A[i]) do
        if A[i][j] <> B[i][j] then begin
          if TryStrToFloat(A[i][j],n1) and TryStrToFloat(B[i][j],n2) then begin
            Tol := Max(Abs(n1)*Tolerance, Tolerance);
            if not DiffNos(n1,n2,Tol,MaxErr) then
              Exit;
          end;
          Fail(Format('%s[%d][%d] difference: %s %s', [Name, i, j, A[i][j], B[i][j]]));
        end;
    end;
  end;
  procedure CheckAttributes(xA, xB: TDOMElement);
  var
    aA, aB: TDOMAttr;
  begin
    repeat
      aA := xA.LastAttribute;
      aB := xB.LastAttribute;
      if aA = nil then begin
        if aB <> nil then
          Fail('Extra Attribute:'+aB.Name);
        Exit;
      end else if aB = nil then begin
        Fail('Missing Attribute:'+aA.Name);
        Exit;
      end else begin
        CheckStr('Attribute Name', aA.Name, aB.Name);
        CheckTokenStr(aA.Name, aA.Value, aB.Value);
        xA.RemoveAttributeNode(aA).Free;
        xB.RemoveAttributeNode(aB).Free;
      end
    until False;
  end;
  procedure CheckViaSave(A, B: TMNode);
  { Save MNodes as XML; may be better overall model }
  var
    XMLDocument: TXMLDocument;
    xA, xB: TDOMElement;
    cA, cB: TDOMNode;
  begin
    XMLDocument := TXMLDocument.Create;
    try
      xA := A.WriteProperties(XMLDocument, True, False);
      xB := B.WriteProperties(XMLDocument, True, False);
      cA := xA.FirstChild;
      cB := xB.FirstChild;
      while cA <> nil do begin
        if cB = nil then begin
          Fail('Child missing:'+cA.NodeName);
          Break;
        end;
        CheckStr('Child Name', cA.NodeName, cB.NodeName);
        CheckStr(ca.NodeName, cA.NodeValue, cB.NodeValue);
        if cA is TDOMElement then
          CheckAttributes(TDOMElement(cA), cB as TDOMElement);
        cA := cA.NextSibling;
        cB := cB.NextSibling;
      end;
      if cB <> nil then
          Fail('Extra Child:'+cB.NodeName);
      CheckAttributes(xA, xB);
    finally
      XMLDocument.Free;
    end;
  end;
var
  i: Integer;
  Tol: Double;
begin
  if not CheckStr('Class', A.ClassName, B.ClassName) then
    Exit;
  CheckStr('Name', A.Name, B.Name);
  CheckStr('Comment', A.Comment, B.Comment);
  if CheckInt('nSettings', A.nSettings, B.nSettings) then
    for i := 1 to A.nSettings do begin
      CheckInt(A.SettingNames[i], A.Settings[i], B.Settings[i]);
      CheckStr('Setting name', A.SettingNames[i], B.SettingNames[i]);
    end;
  if CheckInt('nParameters', A.nParameters, B.nParameters) then
    for i := 1 to A.nParameters do begin
      Tol := Max(Abs(A.Parameters[i])*Tolerance, Tolerance);
      CheckNo(A.ParameterNames[i], A.Parameters[i], B.Parameters[i], Tol);
      CheckStr('Parameter name', A.ParameterNames[i], B.ParameterNames[i]);
    end;
  if CheckInt('Component Count', A.ComponentCount, B.ComponentCount) then
    for i := 0 to A.ComponentCount - 1 do
      CompareNode(A.Components[i] as TMNode, B.Components[i] as TMNode, Failures, MaxErr, Tolerance);
  if A is TMNodeWithArray then begin
    CheckArray('A', TMNodeWithArray(A).A, TMNodeWithArray(B).A);
    CheckArray('AF', TMNodeWithArray(A).AF, TMNodeWithArray(B).AF);
  end;
  if A is TAutoFitter then begin
    CheckViaSave(A, B);
  end;
end;

function CheckImported(IMix, Mixture: TMixture;
                       var nTried: Integer; Tolerance: Double; var Failures: Integer; Script: TStrings):Double;
var
  E: Exception;
begin
  Result := 0;
  ApplyScript(IMix, Script);
  E := IMix.FixupProperties;
  if Assigned(E) then begin
    WriteLn(Log, 'Imxix:',E.Message);
    E.Free;
  end;
  E := Mixture.FixupProperties;
  if Assigned(E) then begin
    WriteLn(Log, 'Mixture:',E.Message);
    E.Free;
  end;
  { Copy some values across, as these will never be identical }
  IMix.Name := Mixture.Name;
  IMix.ShowEstUnc := Mixture.ShowEstUnc; { so .var files with correlations will compare correctly }
  CompareNode(Mixture, IMix, Failures, Result, Tolerance);
  IMix.Free;
end;

procedure ApplyScript(Mixture: TMixture; Script: TStrings);
var
  i: Integer;
  s: string;
begin
  if Script = nil then
    Exit;
  for i := 0 to Script.Count-1 do begin
    s := Trim(Script[i]);
    if S = '' then
      Continue;
    if s[1] in [';', '/', '#'] then
      Continue;
    try
      Mixture.HandleExpression(s);
    except
      on E: Exception do
        raise EReportException.Create('Error on line '+IntToStr(i+1)+' of script:'+E.Message);
    end;
  end;
end;

function CheckCalculatedParameters(Mixture: TMixture; S: TTextStream; var nTried: Integer; Tolerance: Double; var Failures: Integer; Flags: Integer):Double;
  procedure Fail(Name: string);
  begin
    Inc(Failures);
    if Failures > 0 then begin
      WriteLn(Log, 'Line ', S.Line, ':', Name);
    end;
  end;
var
  m, ErrMsg: string;
  Molecule: TMolecule;
  Manifold: TManifold;
  MNode: TMNode;
  State: TState;
  viInfo: TValueInfo;
  V, Vc: Double;
  Tol: Double;
begin
  Result := 0;
  while not S.AtEndOfFile do begin
    if S.GetFreeItem then
      if S.Token <> 'Parameter' then
        Fail(S.Token)
      else begin
        Inc(nTried);
        S.NextStringVar(m);
        Molecule := Mixture.FindMolecule(m);
        if not Assigned(Molecule) then
          Molecule := Mixture.Species[0].Isotopomer[0]
        else
          S.NextStringVar(m);
        Manifold := Molecule.ManifoldWithName(m);
        if Assigned(Manifold) then
          S.NextStringVar(m)
        else
          Manifold := Molecule.Manifold[0];
        State := Manifold.StateWithName(m);
        if Assigned(State) then begin
          S.NextStringVar(m);
          viInfo := State.SubPropInfo(m,ErrMsg);
        end else begin
          viInfo := Manifold.SubPropInfo(m,ErrMsg);
          if not viInfo.IsSingleValue then
            viInfo := Mixture.SubPropInfo(m,ErrMsg);
          if not viInfo.IsSingleValue and (Manifold.nComponentsWithName(m,MNode) = 1) then begin
            S.NextStringVar(m);
            if m <> '=' then
              viInfo := MNode.SubPropInfo(m,ErrMsg);
          end;
        end;
        if not viInfo.IsSingleValue then
          Fail(m+' '+ErrMsg)
        else begin
          S.NextStringVar(m);
          if m <> '=' then
            Fail('Expecting =, found '+m);
          S.NextStringVar(m);
          V := S.RealToken;
          Tol := Max(Abs(V)*Tolerance,Tolerance);
          Vc := viInfo.GetValue;
          if ((Flags and 1) = 1) and ((Vc < 0) xor (V < 0)) then
            Vc := -Vc; {Ignore sign differences }
          if DiffNos(V, VC, Tol, Result) then
            Fail(Format('%s out of tolerance %g %g diff=%g',[viInfo.FullName,V,Vc,V-Vc]));
        end;
      end;
    S.NextLine;
  end;
end;

end.
