unit CheckPgopher;
{ Read a pgopher output file, and compare to a calculation of the same thing }

{$I utils.inc}{Copyright (C) 2005-14 C M Western; see commandline.pas for license}

interface

uses
  Classes, SysUtils, Mixture, TextStrm;

function CheckPgopherFile(Generated: Boolean; Mixture: TMixture; S: TTextStream; var nTried: Integer;
                          Tolerance: Double; var Failures: Integer;
                          InputPgopherVersion: Integer; PseudoSym: Boolean; LineListMode: Integer):Double;

function CheckEigenvalueFile(Mixture: TMixture; S: TTextStream; var nTried: Integer; Tolerance: Double; var Failures: Integer; InputPgopherVersion: Integer; QnoMask: Integer): Double;

function CheckPlotStream(Mixture: TMixture; S: TTextStream; CM: TStream; var nTried: Integer; Tolerance: Double; var Failures: Integer): Double;

function CheckFitStream(Cur: TStream; OldName: string; nParO: Integer; var nTried: Integer; Tolerance: Double; var Failures: Integer; LineListMode: Integer): Double;

function CheckExprs(S: TTextStream; var nTried: Integer; Tolerance: Double; var Failures: Integer): Double;
function CheckAlg(S: TTextStream; var nTried: Integer; Tolerance: Double; var Failures: Integer): Double;

implementation

uses Math, MathSup, UCompat, MatrixVec, Worker, QuantumNumber, MNode, Obs, OAsymT, StrUtils,
     ExternalList, Abelian, ShellSort, Progress, NextTok, DRintensity, Units, Profile, Expr;

var
  CompareListsTime: string = 'Compare Line Lists';
  ReadLineTime: string = 'Read Line';
  CompareLineTime: string = 'Compare Line';

procedure VarAndTol(const S: string; out Value, Tolerance: Double; var MaxVal: Double);
begin
  if not TryStrToFloat(S,Value) then begin
    Tolerance := 0;
    if (CompareText(s,'Nan') = 0) or (CompareText(s,'Inf') = 0) then
      SetNan(Value)
    else
      raise Exception.Create('Bad number:'+S);
    Exit;
  end;
  MaxVal := Max(MaxVal, Abs(Value));
  Tolerance := Precision(S)*1.6;
  { be more slack on very small values }
  Tolerance := Max(Tolerance, 1e-14*MaxVal);
end;

procedure NextVarAndTol(S: TTextStream; out Value, Tolerance: Double; var MaxVal: Double);
begin
  if not S.GetFreeItem then
    raise Exception.Create('No Number');
  VarAndTol(S.Token, Value, Tolerance, MaxVal);
end;

{ For comparing linelists, prepare list of transitions,
  sorted by J', J", Sym', Sym" and frequency }

type

  TSLine = record
    JU, JL: TQuantumNumber;
    SymU, SymL: Integer;
    Frequency: Double;
    Intensity: Double;
    Strength: Double;
    ITol, STol: Double;
  end;

  TSLineList = array of TSLine;

  TSLineListSorter = class(TShellSorter)
    LineList: TSLineList;
    function InOrder(I,J:Integer):Boolean; override;
  end;

var
  FreqDiffThreshold: Double;

function CompareQnos(const A, B: TSLine):Integer;
begin
  Result := A.SymU - B.SymU;
  if Result = 0 then begin
    Result := A.SymL - B.SymL;
      if Result = 0 then begin
      Result := A.JU - B.JU;
      if Result = 0 then
        Result := A.JL - B.JL;
    end;
  end;
end;

function Compare(const A, B: TSLine; out Close: Boolean; AlternateTest: Boolean = False):Integer;
begin
  Close := False;
  Result := CompareQnos(A,B);
  if Result = 0 then begin
    Result := Sign(A.Frequency - B.Frequency);
    if Result = 0 then
      Close := True
    else begin
      Close := Abs(A.Frequency-B.Frequency) <= FreqDiffThreshold;
      if Close then { Actually want string representation to match }
        if AlternateTest then
          Close := Abs(A.Frequency-B.Frequency) <= FreqDiffThreshold/2
        else
          Close := RoundToIncrement(A.Frequency,FreqDiffThreshold) = RoundToIncrement(B.Frequency,FreqDiffThreshold);
    end;
  end;
end;

function TSLineListSorter.InOrder(I, J: Integer): Boolean;
var
  Close: Boolean;
  L: TSLine;
begin
  Result := Compare(LineList[J],LineList[I], Close) >= 0;
  if Result then
    Exit;
  SwapValues(LineList[I], LineList[J], SizeOf(LineList[I]));
end;

function MakeTSLineList(Simulation: TSimulation):TSLineList;
var
  Line: TLine;
  i, w: Integer;
  Problem: Exception;
  LineList: TLineList;
begin
  LineList := Simulation.Mixture.LineList;
  if not Assigned(LineList) then
    raise EReportException.Create('No lines');
  SetLength(Result,LineList.nLines);
  w := 0;
  for i := 0 to LineList.nLines - 1 do begin
    Line := LineList.Line[i];
    with Result[w] do begin
      if Assigned(Line.EigenValuesU) then begin
        JU := Line.EigenvaluesU.J;
        SymU := Line.EigenvaluesU.Sym;
      end else begin
        JU := -987;
        SymU := -89;
      end;
      if Assigned(Line.EigenvaluesL) then begin
        JL := Line.EigenvaluesL.J;
        SymL := Line.EigenvaluesL.Sym;
      end else begin
        JL := -978;
        SymL := -87;
      end;
      Frequency := Convert(Line.Frequency, Simulation.Mixture.Units, Simulation.PlotUnits);
      Strength := Line.Strength;
      Problem := nil;
      Intensity := Simulation.LineIntensity(Line, Simulation.Mixture.Units, Problem);
      if Assigned(Problem) then
        raise Problem;
      ITol := 0;
      STol := 0;
      if Intensity >= Simulation.OThreshold then
        Inc(w);
    end;
  end;
  SetLength(Result,w);
end;

procedure ShowSLine(var F: Text; const Line: TSLine; ShowStr: Boolean);
begin
  Write(Log,'Ju=',QuantumNumberToStr(Line.JU,qftJ):5, ' SymU=',Line.SymU:3,
            ' Jl=',QuantumNumberToStr(Line.JL,qftJ):5, ' SymL=',Line.SymL:3,
            ' Freq=',FloatToStr(Line.Frequency));
  if ShowStr then
    WriteLn(Log, ' ', Line.Strength, ' ', Line.STol:6, ' ', Line.Intensity, Line.ITol:6)
   else
     WriteLn(Log);
end;

function CompareLists(CalcList, ReadList: TSLineList; MaxLineStrength: Double; IgnoreWeak: Boolean):Integer;
  function SumLines(out SSum, ISum, STol, ITol:Double; p: Integer; List: TSLineList; var Template: TSLine; ALternateTest: Boolean; var StrMax, IMax: Double):Integer;
  var
    Close: Boolean;
  begin
    if Template.JU < 0 then begin
      SSum := List[p].Strength;
      ISum := List[p].Intensity;
      STol := List[p].STol;
      ITol := List[p].ITol;
      Result := p + 1;
      Template := List[p];
      if Template.Frequency > FreqDiffThreshold then begin
        StrMax := Max(StrMax, SSum);
        IMax := Max(IMax, ISum);
      end;
    end else begin
      Result := p;
      SSum := 0;
      ISum := 0;
      STol := 0;
      ITol := 0;
    end;
    while Result < Length(List) do begin
      Compare(Template,List[Result], Close, AlternateTest);
      if not Close then
        Break;
      if Template.Frequency > FreqDiffThreshold then begin
        StrMax := Max(StrMax, List[Result].Strength);
        IMax := Max(IMax, List[Result].Intensity);
      end;
      SSum := SSum + List[Result].Strength;
      ISum := ISum + List[Result].Intensity;
      STol := STol + List[Result].STol;
      ITol := ITol + List[Result].ITol;
      Inc(Result);
    end;
  end;
var
  Sorter: TSLineListSorter;
  CP, RP, d, nC, nR, i: Integer;
  SSumC, ISumC, SSumR, ISumR, STol, ITol, Junk, SIg, Iig, StrReadMax, IReadMax, StrCalcMax, ICalcMax: Double;
  Close, AlternateTest: Boolean;
  Current: TSLine;
begin
  Sorter := TSLineListSorter.Create;
  Sorter.LineList := CalcList;
  Sorter.Sort(Length(CalcList));
  Sorter.LineList := ReadList;
  Sorter.Sort(Length(ReadList));
  Sorter.Free;
  CP := 0;
  RP := 0;
  Result := 0;
  StrReadMax := 0;
  IReadMax := 0;
  StrCalcMax := 0;
  ICalcMax := 0;
  Sig := 0; { unused }
  Iig := 0; { unused }

  while (Cp < Length(CalcList)) and (Rp < Length(ReadList)) do begin
    AlternateTest := False;
    if Cp >= Length(CalcList) then
      d := 1
    else if Rp >= Length(ReadList) then
      d := -1
    else begin
      d := Compare(CalcList[Cp],ReadList[Rp],Close);
      if Close then
        d := 0;
      if (d <> 0) and (CompareQnos(CalcList[Cp],ReadList[Rp]) = 0)
         and (Abs(CalcList[Cp].Frequency-ReadList[Rp].Frequency) < FreqDiffThreshold / 2 ) then begin
        { Allow for shift to other side of rounding }
        AlternateTest := True;
        d := 0;
      end;
    end;
    if d = 0 then begin
      { Sum lines with the same frequency }
      Current.JU := -1;
      nR := SumLines(SSumR, ISumR, STol, ITol, Rp, ReadList, Current, False, Sig, Iig);
      nC := SumLines(SSumC, ISumC, Junk, Junk, Cp, CalcList, Current, AlternateTest, Sig, Iig); { These will have STol = ITol = 0 }
      if ((Abs(SSumC-SSumR)>STol) or (Abs(ISumC-ISumR)>ITol))
         and (CalcList[Cp].Frequency > FreqDiffThreshold) { ignore very low frequency lines, as these can be intermittently discarded }
       then begin
        Inc(Result);
        Write(Log, 'Line intensity mismatch ', SSumC, ' ', SSumR, ' diff/STol=', Abs(SSumC-SSumR)/STol:5:1, ' ', ISumC, ' ', ISumR, ' diff/ITol=', Abs(ISumC-ISumR)/ITol:5:1, ' ');
        ShowSLine(Log,ReadList[Rp],False);
        Write(Log, 'Calc Strength: ');
        for i := Cp to nC - 1 do
          Write(Log, CalcList[i].Strength);
        WriteLn(Log);
        Write(Log, 'Read Strength: ');
        for i := Rp to nR - 1 do
          Write(Log, ReadList[i].Strength, ' ', ReadList[i].STol:6);
        WriteLn(Log);
        Write(Log, 'Calc Intensity: ');
        for i := Cp to nC - 1 do
          Write(Log, CalcList[i].Intensity);
        WriteLn(Log);
        Write(Log, 'Read Intensity: ');
        for i := Rp to nR - 1 do
          Write(Log, ReadList[i].Intensity, ' ', ReadList[i].ITol:6);
        WriteLn(Log);
      end;
      Cp := nC;
      Rp := nR;
    end else if d < 0 then begin
      Current.JU := -1;
      nC := SumLines(SSumC, ISumC, STol, ITol, Cp, CalcList, Current, False, StrCalcMax, ICalcMax);
      if (not IgnoreWeak or (ISumC > MaxLineStrength*1e-15))
         and (CalcList[Cp].Frequency > FreqDiffThreshold) then begin
        Inc(Result);
        Write(Log, 'Discarding ', nC-Cp, ' calculated lines (', Cp, ' ', Rp, ') :', SSumC:10, ' ', ISumC:10, ' ');
        ShowSLine(Log, CalcList[Cp], True);
        WriteLn(Log,CalcList[Cp].frequency,' ',CalcList[Max(Cp-1,0)].Frequency);
      end;
      Cp := nC;
    end else begin
      Current.JU := -1;
      nR := SumLines(SSumR, ISumR, STol, ITol, Rp, ReadList, Current, False, StrReadMax, IReadMax);
      if (not IgnoreWeak or (SSumR > STol) or (ISumR > ITol)) { Ignore bonus low intensity lines }
         and (ReadList[Rp].Frequency > FreqDiffThreshold) then begin
        Inc(Result);
        Write(Log, 'Discarding ', nR-Rp, ' read lines (', Cp, ' ', Rp, ') :', SSumR:10, ' ', ISumR:10, ' ');
        ShowSLine(Log, ReadList[Rp], True);
        WriteLn(Log,CalcList[Cp].frequency,' ',ReadList[Rp].Frequency);
      end;
      Rp := nR;
    end;
  end;
  if Result > 0 then begin
    if StrCalcMax + ICalcMax > 0 then
      WriteLn(Log, 'Strongest discarded calculated line: S=', StrCalcMax, ', I=', ICalcMax);
    if StrReadMax + IReadMax > 0 then
      WriteLn(Log, 'Strongest discarded read line: S=', StrReadMax, ', I=', IReadMax);
  end;
end;

function CheckEigenvalueFile(Mixture: TMixture; S: TTextStream; var nTried: Integer; Tolerance: Double; var Failures: Integer; InputPgopherVersion: Integer; QnoMask: Integer): Double;
  { QnoMask allow selective removal of quantum numers from the comparison }
  function RemoveQnos(s: string):string;
  var
    Tok: string;
    Mask: Integer;
  begin
    s := AnsiReverseString(s);
    Result := '';
    Mask := QnoMask;
    repeat
      Tok := NextToken(s);
      if not Odd(Mask) then
        Result := Result + ' ' + Tok;
      Mask := Mask div 2;
    until s = '';
    Result := AnsiReverseString(Result);
  end;
var
  J, sym, n: Integer; { State Details }
  E, ETol, EMax: Double;
  Manifold: TManifold;
  Molecule: TMolecule;
  Eigenvalues: TEigenvalues;
  Simulation: TSimulation;
  procedure Fail(Name: string);
  begin
    Inc(Failures);
    if Failures > 0 then begin
      WriteLn(Log, 'Line ', S.Line, ':', Name);
      WriteLn(Log, J:5, sym:2, n:2, E:12:4);
    end;
  end;
  procedure CheckNo(n1, n2, tol: Double; Name: string);
  begin
    if isNanOrInf(n2) then
      Fail(Format('%s fails: %g Nan', [Name, n1]))
    else if Abs(n1-n2) > tol then
      Fail(Format('%s out of tolerance: Read=%g Calc=%g %g', [Name, n1, n2, n1-n2]));
    if Abs(n1-n2) > Result then
      Result := Abs(n1-n2);
  end;
  function FindSym(Manifold:TManifold; J: TQuantumNumber; SymName:String):Integer;
  begin
    if TryStrToInt(SymName, Result) then
      Result := Manifold.Molecule.FixSym(J, Result, InputPgopherVersion)
    else if not Manifold.Molecule.TryStrToAnySymName(J, SymName, Result) then begin
      Fail('Bad symmetry name:'+SymName);
      Result := -1;
    end;
  end;
var
 SymS: string;
 g: Integer;
 pop, popTol, popMax, popCalc: Double;
 Q, QTol, QMax, Qcalc, Qsum, Qscale: Double;
 Overflow: Exception;
 LabRd, LabEx: string;
 Emin: Double;
begin
  nTried := 0;
  Mixture.CheckMatrixElementsAvailable;
  Simulation := Mixture.Simulations[0];
  Molecule := Mixture.Species[0].Isotopomer[0];
  Manifold := Molecule.Manifold[0];
  Result := 0;
  EMax := 0;
  popMax := 0;
  QMax := 0;
  Overflow := nil;
  Qsum := 0;
  if IncludesPopulation(Simulation.IntensityUnits) then
    Qcalc := Manifold.Molecule.PartitionFunction(Simulation,Overflow,Emin)
  else
    Qcalc := 1;
  Qscale := Manifold.bf(Simulation.Temperature,Emin-Molecule.ActualAssumedOrigin);
  while not S.AtEndOfFile do begin
    { Skip lines not starting with an integer or manifold }
    if not S.GetFreeItem then begin
      S.Nextline;
      Continue;
    end;
    if S.Token = 'Pass' then
      Break; { TODO: Could check partition function calc also }
    Manifold := Molecule.ManifoldWithName(S.Token);
    if Assigned(Manifold) then
      S.GetFreeItem
    else
      Manifold := Molecule.Manifold[0];
    if not TryStrToQuantumNumber(S.Token,J,Mixture.QuantumNumberFormat) then begin
      S.Nextline;
      Continue;
    end;
    S.NextStringVar(SymS);  S.NextIntegerVar(n);
    NextVarAndTol(S, E, ETol, EMax);
    Sym := FindSym(Manifold,J,SymS);
    Inc(nTried);
    if Sym >= 0 then begin
      Eigenvalues := Manifold.Eigenvalues(J, sym, [meHrot, meWidth], Mixture.Simulations[0].ExField);
      if n > Eigenvalues.Size then
        Fail(Format('Eigenvalue (%d) > %d', [n, Eigenvalues.Size]))
      else begin
        CheckNo(E, Eigenvalues.E[n], ETol, 'E');
        if S.GetFreeItem and S.IsIntegerToken(g)  then begin
          CheckNo(g,Eigenvalues.Degeneracy(n), 0, 'g');
          if S.GetFreeItem then begin
            VarAndTol(S.Token, pop, popTol, popMax);
            popCalc := g*Manifold.Population(Eigenvalues,n,Simulation,Overflow,0);
            QSum := Qsum + popCalc/Qscale;
            CheckNo(pop, popCalc/Qcalc, popTol, 'Population');
            LabRd := '';
            if S.GetFreeItem then
              if TryStrToFloatEx(S.Token,Q) then begin
                VarAndTol(S.Token, Q, QTol, QMax);
                CheckNo(Q, Qsum, QTol, 'Q');
                LabRd := S.RestOfLine;
              end else
                LabRd := S.Token + S.RestOfLine;
            LabRd := Trim(LabRd);
            if LabRd <> '' then begin
              LabEx := Trim(Eigenvalues.DominantLabel(n,Mixture.QuantumNumberFormat));
              if LabRd <> LabEx then begin
                if QnoMask <> 0 then begin
                   LabRd := RemoveQnos(LabRd);
                   LabEx := RemoveQnos(LabEx);
                end;
                if LabRd <> LabEx then
                  Fail('State Label mismatch:'+LabRd+'<>'+LabEx);
              end;
            end;
          end;
        end;
      end;
     end;
    S.NextLine;
  end;
  Overflow.Free;
end;

function CheckPgopherFile(Generated: Boolean; Mixture: TMixture; S: TTextStream;
                          var nTried: Integer; Tolerance: Double; var Failures: Integer;
                          InputPgopherVersion: Integer; PseudoSym: Boolean; LineListMode: Integer):Double;
var
  MoleculeName: string;
  Manp, sympStr: string;
  Jp, np: Integer; { Upper State }
  symp: Shortint;
  Man, symStr: string;
  J,  n : Integer; { Lower State }
  sym: Shortint;
  f, cstr, w, Ep, E, Sum: Double;
  Wt: string;
  linestrength, linestrengthTol, linestrengthMax,
  Str, StrTol, StrMax,
  EstUnc, EstUncMax, EstUncTol,
  EinsteinA, EinsteinATol, EinsteinAMax: Double;
  ManifoldL, ManifoldU: TManifold;
  Molecule: TMolecule;
  TransitionMoments: TTransitionMoments;
  Ds: TTransitionMatrices;
  Line: TLine;
  Failed: Boolean;
  CurrentFailure: string;
  MaxDifference, CurrentDifference: Double;
  Simulation: TSimulation;
  procedure Fail(Name: string);
  begin
    Inc(Failures);
    if Failures > 0 then begin
      WriteLn(Log, 'Line ', S.Line, ':', Name);
      WriteLn(Log, MoleculeName, ' ', Manp, ' ', Jp:3, symp:2, np:2, ' ', Man, ' ', J:3, sym:2, n:2, f:12:4, str:8:4, w:8, Ep:12:4, E:12:4);
    end;
    Failed := True;
  end;
  function CheckNo(n1, n2, tol: Double; Name: string; Relative: Boolean = False; Defer: Boolean = False):Boolean;
  var
    Msg: string;
  begin
    Result := True; { Assume fail }
    if isNanOrInf(n2) then
      Msg := Format('%s fails: %g Nan', [Name, n1])
    else if (Relative and (Abs(n1-n2)/Abs(n1+n2) > tol)) or (not Relative and (Abs(n1-n2) > tol)) then
      Msg := Format('%s out of tolerance: %g %g %g (%g)', [Name, n1, n2, n1-n2, n1/n2])
    else
      Result := False; { OK }
    if Defer then begin
      if Abs(n1-n2) > CurrentDifference then
        CurrentDifference := Abs(n1-n2);
      if Result then begin
        if CurrentFailure <> '' then
          CurrentFailure := CurrentFailure + sLineBreak;
        CurrentFailure := CurrentFailure + Msg;
      end;
    end else begin
      if Abs(n1-n2) > MaxDifference then
        MaxDifference := Abs(n1-n2);
      if Result then
        Fail(Msg);
    end;
  end;
  function FindSym(Manifold:TManifold; J: TQuantumNumber; SymName:String):Integer;
  begin
    if Simulation.HasFields then
      Result := -1
    else if TryStrToInt(SymName, Result) then
      Result := Manifold.Molecule.FixSym(J, Result, InputPgopherVersion)
    else if not Manifold.Molecule.TryStrToAnySymName(J, SymName, Result) then
      if PseudoSym then
        Result := -1 { Symmetry and eigenvalue number must come from symmetry name }
      else
        Fail('Bad symmetry name:'+SymName);
    if PseudoSym and (Manifold.Molecule is TAsymmetricMolecule)
       and (TAsymmetricMolecule(Manifold.Molecule).PointGroup in [C2h, Ci]) then
      Result := -1; { For these point groups, Ag and Au are in both the high and low symmetry versions
                      of the point group, so force search mode }
  end;
  procedure NextQuantumVar(var Q: TQuantumNumber);
  var
    str: String;
  begin
    str := S.NextString('?');
    if not TryStrToQuantumNumber(str, Q, Mixture.QuantumNumberFormat) then
      if (str = '-') and (Molecule is TExternalMolecule) then
        Q := -1
      else
        Fail('Bad quantum number:'+str);
  end;
  procedure CheckDerivatives(Derivs: string);
    procedure CheckDerivative(const Derivative: TDerivative; Value: Double);
    var
      s, e: string;
      v, Tol, Mx: Double;
    begin
      s := NextToken(Derivs);
      e := Derivative.MNode.FullParameterName(Derivative.Index);
      if s <> e then
        CurrentFailure := 'Derivative Names: '+s+' <> '+e
      else begin
        s := NextToken(Derivs);
        VarAndTol(s, v, Tol, Mx);
        CheckNo(v, Value, Tol, e, False, True);
      end;
    end;
  var
    i: Integer;
  begin
    if Line.EigenvaluesU.Manifold = Line.EigenvaluesL.Manifold then
      for i := 0 to High(Line.EigenvaluesL.Derivatives) do
        CheckDerivative(Line.EigenvaluesU.Derivatives[i], Line.EigenvaluesU.Derivatives[i].Values[Line.IndexU]-Line.EigenvaluesL.Derivatives[i].Values[Line.IndexL])
    else begin
      for i := 0 to High(Line.EigenvaluesU.Derivatives) do
        CheckDerivative(Line.EigenvaluesU.Derivatives[i], Line.EigenvaluesU.Derivatives[i].Values[Line.IndexU]);
      for i := 0 to High(Line.EigenvaluesL.Derivatives) do
        CheckDerivative(Line.EigenvaluesL.Derivatives[i], -Line.EigenvaluesL.Derivatives[i].Values[Line.IndexL]);
    end;
  end;
  function ReadManifold(var Name: string): TManifold;
  begin
    { Allow for spaces in file names - see TObservationsReader.ReadOptionalManifold }
    S.NextStringVar(Name);
    Result := Molecule.ManifoldWithName(Name);
    if not Assigned(Result) and not (Name[1] in ['.', '+', '-', '0'..'9']) then begin
      { Might well be a Manifold name, as it is not a number;
        Allow for Manifold name to have spaces in it }
      while not Assigned(Result) and (Molecule.FindMNodeStartingWith(Name,TManifold) <> nil) do begin
        if not S.GetFreeItem then
          Break;
        Name := Name + ' ' + S.Token;
        Result := Molecule.ManifoldWithName(Name);
      end;
    end;
  end;
  function ReadMolecule(var Name: string): TMolecule;
  begin
    { Allow for spaces in file names - see TObservationsReader.ReadOptionalMolecule }
    Result := Mixture.FindMolecule(Name);
    if not Assigned(Result) and not (Name[1] in ['.', '+', '-', '0'..'9']) then begin
      { Might well be a Molecule name, as it is not a number;
        Allow for Molecule name to have spaces in it }
      while not Assigned(Result) and (Mixture.FindMNodeStartingWith(Name,TMolecule,True) <> nil) do begin
        if not S.GetFreeItem then
          Break;
        Name := Name + ' ' + S.Token;
        Result := Mixture.FindMolecule(Name);
      end;
    end;
  end;
var
  Problem: Exception;
  nCounted, nZero, nNotCounted, nIntensityFailures: Integer;
  TxDetails, TxOnly, ReadTx, CalcTx, AdjTx, Buffer, Derivs: string;
  StepIndex, OriginalL, OriginalU: Integer;
  i, P, Q, R: Integer;
  MaxIntensity: Double;
  HasA, HasEstUnc, HasWt: Boolean;
  CalcList, ReadList: TSLineList;
  nReadList: Integer;
  LR: TSLine;
  FreqTol, FreqMax, EpTol, EpMax, ETol, EMax, wTol, wMax, ITotal: Double;
begin
  { LineListMode is a mask to turn off various checks:
    1: TxDetails
    2: Line Count
    4: Compare Linelist
    8: Line Strength/Intensity/A/Width
   16: Ignore linelist differences involving very weak lines
   32: Skip linelist generation completely
   64: Adjust A values - not currently used
  128: Looser intensity tolerance checks
  }

  nTried := 0;
  nZero := 0;
  MaxDifference := 0;
  Problem := nil;
  FreqDiffThreshold := 0;
  FreqMax := 0; { Not used }
  EpMax := 0;
  EMax := 0;
  StrMax := 0;
  linestrengthMax := 0;
  EinsteinAMax := 0;
  wMax := 0;
  nIntensityFailures := 0;
  S.AllowDefaults := False;
  S.CommentChars := S.CommentChars - ['/'];
  if Mixture = nil then
    raise EReportException.Create('No Mixcture');
  Simulation := Mixture.Simulations[0];
  CalcList := nil;
  ReadList := nil;
  nReadList := 0;
  if Mixture.PrintLevel < plEigenVectors then
    Mixture.PrintLevel := plNone;
  if (LineListMode and 32) = 0 then begin
    Simulation.MakeLineList(Problem);
    if Assigned(Problem) then
      raise Problem;
    if not Generated and ((LineListMode and 4) = 0) then begin
      CalcList := MakeTSLineList(Simulation);
      SetLength(ReadList, Length(CalcList));
    end;
  end;
  i := -1;
  { Look for line list heading }
  while not S.AtEndOfFile do begin
    Buffer := S.RestOfLine;
    i := Max(Pos('Frequency', Buffer),Pos('Position', Buffer));
    if i > 0 then
      Break;
    S.NextLine;
  end;
  if i < 0 then
    Fail('Line list heading not found');
  { Look at column headings to decide on file type }
  Delete(Buffer,1,i);
  { Check for A as column heading }
  HasA := Pos(' A ', Buffer) > 0;
  HasEstUnc := Pos(' Est Unc ', Buffer) > 0;
  HasWt := Pos(' Wt ', Buffer) > 0;
  S.NextLine;
  i := -1;
  Sum := 0;
  while not S.AtEndOfFile do begin
    if (S.Line mod 100) = 0 then
      SetProgress(S.Position);
    Failed := False;
    MoleculeName := S.NextString('');
    if CompareText(MoleculeName, 'stop') = 0 then
      Exit;
    if (MoleculeName = '') then begin
      S.NextLine;
      Continue;
    end;
    Molecule := ReadMolecule(MoleculeName);
    if not Assigned(Molecule) then
      if TryStrToInt(MoleculeName,i) then
        Break
      else
        raise EReportException.Create('Molecule '+MoleculeName+' not found');
    TStart(ReadLineTime);
    if not IsPopDist(Simulation.IntensityUnits) then begin
      ManifoldU := ReadManifold(Manp); NextQuantumVar(Jp); S.NextStringVar(sympStr); S.NextIntegerVar(np);
    end else begin
      ManifoldU := nil;
      Jp := -987;
      symp := -89;
    end;
    ManifoldL := ReadManifold(Man);  NextQuantumVar(J);  S.NextStringVar(symStr);  S.NextIntegerVar(n);
    NextVarAndTol(S,f,FreqTol,FreqMax);
    FreqDiffThreshold := Max(FreqDiffThreshold,FreqTol);
    if HasEstUnc then
      NextVarAndTol(S,EstUnc,EstUncTol,EstUncMax);
    if Generated then begin
      S.GetFreeItem; { Actually weight }
      Wt := S.Token;
      NextVarAndTol(S,Str,StrTol,StrMax);
      if (Wt <> '-') and (StrToFloat(Wt) <= 0) then begin
        S.NextLine;
        TEnd(ReadLineTime);
        Continue; { Skip }
      end;
    end else begin
      NextVarAndTol(S,Str,StrTol,StrMax);
      if (LineListMode and 128) <> 0 then
        StrTol := StrTol*6;
      if not IsPopDist(Simulation.IntensityUnits) then
        NextVarAndTol(S,Ep,EpTol,EpMax);
      NextVarAndTol(S,E,ETol,EMax);
      NextVarAndTol(S,linestrength,linestrengthTol,linestrengthMax);
      if (LineListMode and 128) <> 0 then
        linestrengthTol := linestrengthTol*6;
      if HasA then begin
        NextVarAndTol(S,EinsteinA,EinsteinATol,EinsteinAMax);
        if (LineListMode and 128) <> 0 then
          EinsteinATol := EinsteinATol*6;
      end;
      NextVarAndTol(S,w,wTol,wMax);
      TxDetails := Trim(S.RestOfLine);
      if (Length(TxDetails) >0) and (TxDetails[1] = ':') then
        TxDetails := Trim(Copy(TxDetails,2,Length(TxDetails)));
      { May need to do more trimming if format changes in future }
    end;
    Inc(nTried);
    TEnd(ReadLineTime);
    TStart(CompareLineTime);
    if not Assigned(Molecule) then Fail('Can''t find Molecule:'+MoleculeName);
    if not Failed then begin
      Line.Width := 0;
      if not Assigned(ManifoldU) then begin
        if not IsPopDist(Simulation.IntensityUnits) then
          Fail('Can''t find Upper Manifold:'+Manp);
        Line.EigenvaluesU := nil;
        Line.IndexU := -27;
      end else begin
        symp := FindSym(ManifoldU, Jp, sympStr);
        if symp < 0 then
          Line.IndexU := FindEigenvaluesAndIndex(Line.EigenvaluesU, ManifoldU, Jp, symp, -1, TxDetails, False, S.Line, '', 0, Mixture.QuantumNumberFormat, Simulation.ExField)
        else begin
          Line.EigenvaluesU := ManifoldU.Eigenvalues(Jp, symp, [meHrot, meWidth], Simulation.ExField, True, plNone, nil, True, 0);
          Line.IndexU := np;
        end;
        if (Line.IndexU <= 0) or (Line.IndexU > Line.EigenvaluesU.Size) then
          Fail(Format('Upper eigenvalue (%d) > number of eigenvalues (%d)', [np, Line.EigenvaluesU.Size]))
        else begin
          if not Generated then
            CheckNo(Ep, Line.EigenvaluesU.Energy[Line.IndexU], EpTol, 'E''');
          if Assigned(Line.EigenValuesU.Widths) then
            Line.Width := Line.Width + Line.EigenValuesU.Widths[np];
        end;
      end;
      if not Assigned(ManifoldL) then
        Fail('Can''t find Lower Manifold:'+Man)
      else begin
        sym := FindSym(ManifoldL, J, symStr);
        if sym < 0 then
          Line.IndexL := FindEigenvaluesAndIndex(Line.EigenvaluesL, ManifoldL, J, sym, -1, TxDetails, True, S.Line, '', 0, Mixture.QuantumNumberFormat, Simulation.ExField)
        else begin
          Line.EigenvaluesL := ManifoldL.Eigenvalues(J, sym, [meHrot, meWidth], Simulation.ExField, True, plNone, nil, True, 0);
          Line.IndexL := n;
        end;
        if (Line.IndexL <= 0) or (Line.IndexL > Line.EigenvaluesL.Size) then
          Fail(Format('Lower eigenvalue (%d) > number of eigenvalues (%d)', [n, Line.EigenvaluesL.Size]))
        else begin
          if Assigned(Line.EigenValuesL.Widths) then
            Line.Width := Line.Width + Line.EigenValuesL.Widths[n];
          if not Generated then
            CheckNo(E, Line.EigenvaluesL.Energy[Line.IndexL], ETol, 'E"');
        end;
      end;
      if (LineListMode and 64) <> 0 then begin
        Str := Str / 2.5706;
        linestrength := linestrength * 2 / 3;
        EinsteinA := EinsteinA * 2 / 3;
        if (symp >= 4) and (sym < 4) then
          EinsteinA := EinsteinA / 2
        else if (symp >= 4) and (sym >= 4) then
            EinsteinA := EinsteinA / 2;
      end;
      if not failed then begin
        if IsPopDist(Simulation.IntensityUnits) then
          Line.Frequency := Line.EigenvaluesL.Energy[Line.IndexL]
        else
          Line.Frequency := Line.EigenvaluesU.Energy[Line.IndexU]-Line.EigenvaluesL.Energy[Line.IndexL];
        if not (Molecule is TExternalMolecule) then begin
          Line.Frequency := Convert(Line.Frequency, Mixture.Units, Simulation.PlotUnits);
          CheckNo(f, Line.Frequency, FreqTol, 'Frequency');
        end;
        if HasEstUnc then
          CheckNo(EstUnc, EstErr(Line.EigenvaluesU, Line.EigenvaluesL, Line.IndexL, Line.IndexU), EstUncTol, 'Est Unc');
        TransitionMoments := Molecule.FindTransitionMoments(ManifoldU,ManifoldL);
        if (LineListMode and 8) <> 0 then
          { Skip intensity checks }
          Sum := Sum + Str
        else if IsPopDist(Simulation.IntensityUnits) then begin
          Line.CachedName := '';
          CheckNo(f, Line.Frequency, FreqTol, 'Frequency');
          Problem := nil;
          Line.Strength := Line.EigenvaluesL.StatisticalWeight(Line.IndexL); { TODO: Incomplete }
          cstr := Simulation.LineIntensity(Line, Mixture.Units, Problem);
          if Assigned(Problem) then begin
            Fail(Problem.Message);
            Problem.Free;
          end;
          Sum := Sum + cstr;
          if not CheckNo(linestrength, Line.Strength, linestrengthTol, 'Line Strength') then
            CheckNo(w, Line.Width, wTol, 'Line Width');
          if not CheckNo(Str, cstr, StrTol, 'Intensity') then
            if (MakeTxName(Line) <> TxDetails) and ((LineListMode and 1) = 0) then
              Fail('TxDetails mismatch|'+Trim(Line.CachedName)+'|'+TxDetails+'| ');
        end else if not Assigned(TransitionMoments) then
          Fail('Can''t find transition moment')
        else if Molecule is THITRANMolecule then begin
          { Simple check of line strength }
          { First find strength }
          if THITRANMolecule(Molecule).FindLine(Line,TxDetails) < 0 then
            Fail('Line not found');
          CheckNo(f, Line.Frequency, FreqTol, 'Frequency');
          if not Generated then begin
            if not CheckNo(linestrength, Line.Strength, linestrengthTol, 'Line Strength') then begin
              if HasA then
                CheckNo(EinsteinA, Molecule.EinsteinA(Line, Mixture.Units), EinsteinATol, 'A');
              CheckNo(w, Line.Width, wTol, 'Line Width');
            end;
          end;
          Problem := nil;
          cstr := Simulation.LineIntensity(Line, Mixture.Units, Problem);
          if Assigned(Problem) then begin
            Fail(Problem.Message);
            Problem.Free;
          end;
          if not CheckNo(Str, cstr, StrTol, 'Intensity') then
            if (Line.CachedName <> TxDetails) and ((LineListMode and 1) = 0) then
              Fail('TxDetails mismatch|'+Trim(Line.CachedName)+'|'+TxDetails+'| ');
        end else begin
          if Line.Frequency = 0 then
            Inc(nZero); { Zero frequency lines are discarded in more recent versions }
          Ds := TransitionMoments.TransitionMatrix(Line.EigenValuesL, Line.EigenValuesU, True, True, nil, Simulation);
          { Accept eigenvalue number changes, if the other eigenvalue has the
            same, or very similar energy. }
          OriginalL := Line.IndexL;
          OriginalU := Line.IndexU;
          cstr := 0;
          { Cycle through all pairs of eigenvalues; the +1 ensures the original is used to give the error message }
          for StepIndex := 0 to Line.EigenvaluesL.Size * Line.EigenvaluesU.Size + 1 do begin
            Line.IndexL := (OriginalL + StepIndex - 1) mod Line.EigenvaluesL.Size + 1;
            if not SameValue(Line.EigenvaluesL.E[OriginalL],Line.EigenvaluesL.E[Line.IndexL]) then
              Continue;
            Line.IndexU := (OriginalU + StepIndex div Line.EigenvaluesL.Size - 1) mod Line.EigenvaluesU.Size + 1;
            if not SameValue(Line.EigenvaluesU.E[OriginalU],Line.EigenvaluesU.E[Line.IndexU]) then
              Continue;
//            if StepIndex <> 0 then
//              WriteLn(Log, StepIndex:3, OriginalL:3, Line.IndexL:3, OriginalU:3, Line.IndexU:3,
//                           ' ', Line.EigenvaluesL.E[OriginalL]-Line.EigenvaluesL.E[Line.IndexL],
//                           ' ', Line.EigenvaluesU.E[OriginalU]-Line.EigenvaluesU.E[Line.IndexU],
//                           CurrentFailure);
            CurrentFailure := '';
            CurrentDifference := 0;
            if TransitionMoments.MakeLine(Line,Ds,Line.EigenvaluesL,Line.EigenvaluesU,Line.IndexL,Line.IndexU,RandomExcitation,True) then
              if f <> 0 then
                Fail('States out of order');
            if not Generated then begin
              if CheckNo(linestrength, Line.Strength, linestrengthTol, 'Line Strength', False, True) then
                Continue;
              if HasA then
                CheckNo(EinsteinA, Molecule.EinsteinA(Line, Mixture.Units), EinsteinATol, 'A', False, True);
            end;
            Problem := nil;
            cstr := Simulation.LineIntensity(Line, Mixture.Units, Problem);
            if Assigned(Problem) then begin
              Fail(Problem.Message);
              Problem.Free;
            end;
            if CheckNo(Str, cstr, StrTol, 'Intensity', False, True) then
              Continue;
            if not Generated then begin
              if CheckNo(w, Line.Width, wTol, 'Width', False, True) then
                Continue;
              Line.CachedName := '';
              if (TxDetails <> '') and (Trim(MakeTxName(Line)) <> TxDetails) then begin
                { Format Adjust }
                P := Pos(':',Line.CachedName);
                Q := Pos(':',TxDetails);
                if (P > 0) and (Q > 0) then begin
                  { Compare branch details }
                  ReadTx := Trim(Copy(TxDetails,1,Q-1));
                  CalcTx := Trim(Copy(Line.CachedName,1,P-1));
                  if ReadTx <> CalcTx then begin
                    { Recheck, removing spaces }
                    AdjTx := ReadTx;
                    R := 2;
                    while R < Length(AdjTx) do begin
                      if AdjTx[R] = ' ' then
                        if (AdjTx[R-1] in ['0'..'9']) and (AdjTx[R+1] in ['0'..'9']) then
                          AdjTx[R] := ','
                        else begin
                          Delete(AdjTx,R,1);
                          Dec(R);
                        end;
                      Inc(R);
                    end;
                    if AdjTx <> CalcTx then
                      CurrentFailure := 'Branch Details mismatch|'+AdjTx+'|'+CalcTx+'| ';
                  end;
                end;
                if P > 0 then
                  Line.CachedName := Trim(Copy(Line.CachedName, P+1, Length(Line.CachedName)));
                Derivs := '';
                if Q > 0 then begin
                  TxOnly := Trim(Copy(TxDetails, Q+1, Length(TxDetails)));
                  Q := Pos(':',TxOnly);
                  if Q > 0 then begin
                    Derivs := Trim(Copy(TxOnly,Q+1,Length(TxOnly)));
                    TxOnly := Trim(Copy(TxOnly,1,Q-1));
                  end;
                end else
                  TxOnly := TxDetails;
                P := Pos('- ',Line.CachedName);
                if (P > 0) and (Pos('- ',TxOnly) <= 0) then
                  Delete(Line.CachedName, P, 2);
                if ((LineListMode and 1) = 0) and not CompareStateLabels(Line.CachedName,TxOnly) then begin
                  CurrentFailure := 'TxDetails mismatch|'+Trim(Line.CachedName)+'|'+TxOnly+'| ';
                end;
                if (CurrentFailure = '') and (Derivs <> '') then
                  CheckDerivatives(Derivs);
              end;
            end;
            if CurrentFailure = '' then
              Break; { Success }
          end;
          Sum := Sum + cstr;
          if CurrentDifference > MaxDifference then
            MaxDifference := CurrentDifference;
          if CurrentFailure <> '' then begin
            Inc(nIntensityFailures);
            Fail(CurrentFailure);
          end;
          // WriteLn(Log, Line.Strength:8:4, LineIntensity(Line, Mixture.Temperature, Mixture.Tvib)/5:8:4, str:8:4);
        end;
      end;
    end;
    if Length(CalcList) > 0 then begin
      if Length(ReadList) <= nReadList then
        SetLength(ReadList, nReadList+10000);
      LR.Frequency := f;
      LR.JU := Jp;
      LR.JL := J;
      LR.SymU := symp;
      LR.SymL := sym;
      LR.Strength := linestrength;
      LR.Intensity := Str;
      LR.STol := linestrengthTol;
      LR.ITol := StrTol;
      ReadList[nReadList] := LR;
      Inc(nReadList);
    end;
    S.NextLine;
    TEnd(CompareLineTime);
  end;
  if (i <> -1) and ((LineListMode and 32) = 0) then begin
    { Calculate expected values }
    with Simulation.Mixture.LineList do begin
      ITotal := 0;
      nCounted := 0;
      MaxIntensity := 0;
      for j := 0 to nLines-1 do begin
        Problem := nil;
        Str := Simulation.LineIntensity(Line[j], Simulation.Mixture.Units, Problem);
        if Assigned(Problem) then
          raise Problem;
        ITotal := ITotal + Str;
        if Str >= Simulation.OThreshold then
          Inc(nCounted);
      end;
      nNotCounted := nLines - nCounted;
    end;
    CheckNo(i, nTried, 0, 'Line Count');
    if ((LineListMode and 2) = 0) and (Abs(i-nZero-nCounted) > 20) then
      CheckNo(i-nZero, nCounted, 0, 'Simulated Line Count');
    while not S.AtEndOfFile and (S.NextString('') <> '=') do
      if S.Token = '' then Fail('Format Error');
    Str := S.NextDouble(Sum);
    if nIntensityFailures = 0 then
      CheckNo(Str, Sum, 1e-4, 'Line intensity sum(a)'); { Note that this will happen if there are any intensity errors }
    CheckNo(Str, ITotal, 1e-4, 'Line intensity sum(b)');
  end;
  if (Length(CalcList) > 0) and (Failures < 50) then begin
    SetLength(ReadList,nReadList);
    TStart(CompareListsTime);
    Inc(Failures, CompareLists(CalcList, ReadList, linestrengthMax, (LineListMode and 16) <> 0));
    TEnd(CompareListsTime);
  end;
  Result := MaxDifference;
end;

function CheckPlotStream(Mixture: TMixture; S: TTextStream; CM: TStream; var nTried: Integer; Tolerance: Double; var Failures: Integer): Double;
  procedure Fail(S: TTextStream; Name: string);
  begin
    Inc(Failures);
    if Failures > 0 then begin
      WriteLn(Log, 'Line ', S.Line, ':', Name);
    end;
  end;
var
  C: TTextStream;
  nCols, n: Integer;
  V: array of array of Double;
  Maxes: array of Double;
  procedure ReadBlock(S: TTextStream);
  var
    r, c: Integer;
    T: Double;
  begin
    r := 0;
    nCols := 0;
    n := 0;
    while not S.GetFreeItem and not S.AtEndOfFile do
      S.NextLine;
    if S.AtEndOfFile then
      Exit;
    repeat
      c := 0;
      repeat
        if S.IsRealToken(T) then begin
          if c >= nCols then begin
            nCols := c+1;
            if nCols > Length(V) then begin
              SetLength(V, nCols);
              SetLength(Maxes, nCols);
            end;
            Maxes[c] := 0;
          end;
          if r >= n then
            n := r+1;
          if r >= Length(V[c]) then
            SetLength(V[c],r+1000);
          V[c][r] := T;
          Maxes[c] := Max(Maxes[c], Abs(T));
          Inc(nTried);
          Inc(c);
        end else
          Fail(S, 'Bad token:'+S.Token);
      until not S.GetFreeItem;
      while c < nCols do begin
        SetNan(V[c][r]);
        Inc(c);
      end;
      S.NextLine;
      Inc(r);
    until not S.GetFreeItem;
  end;
  procedure CheckBlock(S: TTextStream);
  var
    r, c: Integer;
    T: Double;
    ColFail: Boolean;
  begin
    r := 0;
    while not S.GetFreeItem and not S.AtEndOfFile do
      S.NextLine;
    if S.AtEndOfFile then begin
      if n <> 0 then
        Fail(S,'Extra Read Block');
      Exit;
    end;
    if n = 0 then begin
      Fail(S,'Extra calculated Block staring with '+S.Token);
      Exit;
    end;
    ColFail := False;
    repeat
      c := 0;
      repeat
        if S.IsRealToken(T) then begin
          if c >= nCols then
            Fail(S, 'Extra Calculated Columns:'+S.Token)
          else if r >= n then
            Fail(S, 'Extra Calculated Values:'+S.Token)
          else if IsNanOrInf(V[c][r]) xor IsNanOrInf(T) then
            Fail(S,Format('Non Blank Value %d %d %g %g',[c, r, T, Maxes[c]]))
          else begin
            if Abs(V[c][r]-T) >= Tolerance*Maxes[c] then
              Fail(S,Format('col=%d col0=%g %g %g %g',[c, V[0][r], V[c][r], T, Maxes[c]]));
            T := Abs(V[c][r]-T);
            if Maxes[c] > 0 then
              T := T / Maxes[c];
            Result := Max(T, Result);
          end;
          Inc(c);
        end else
          Fail(S, 'Bad token:'+S.Token);
      until not S.GetFreeItem;
      if (c < nCols) and not ColFail then begin
        Fail(S, 'Not Enough Calculated Columns');
        ColFail := True; { Prevent multiple messages }
      end;
      S.NextLine;
      Inc(r);
    until not S.GetFreeItem or S.AtEndOfFile;
    if r <> n then
      Fail(S, Format('Different number of values: %d %d', [r, n]));
  end;
begin
  CM.Position := 0;
  C := TTextStream.Create(CM,False);
  C.Comments := False;
  C.FlowToNextLine := False;
  S.Comments := False;
  S.FlowToNextLine := False;
  Result := 0;
  try
    repeat
      ReadBlock(S);
      CheckBlock(C);
    until S.AtEndOfFile;
    while not C.AtEndOfFile do begin
      if C.GetFreeItem then begin
        Fail(C, 'Extra Items:'+C.Token);
        Break;
      end;
      C.NextLine;
    end;
  finally
    C.Free;
  end;
end;

type
  TPar = record
    n: Integer;
    OldV, OldTol, OldMx, NewV, NewTol, NewMx,
    Sigma, SigmaTol, SigmaMx, Change, ChangeTol, ChangeMx,
    Sens, SenTol, SensMx, Scale, ScaleTol, ScaleMx: Double;
    Summary, Name: string
  end;

function CheckFitStream(Cur: TStream; OldName: string; nParO: Integer; var nTried: Integer; Tolerance: Double; var Failures: Integer; LineListMode: Integer): Double;
  procedure Fail(S: TTextStream; Name: string);
  begin
    Inc(Failures);
    if Failures > 0 then begin
      WriteLn(Log, 'Line ', S.Line, ' of ' + S.StreamName + ':', Name);
    end;
  end;
  procedure LookFor(S: TTextStream; Target: string);
  var
    buf: string;
  begin
    while not S.AtEndOfFile do begin
      buf := S.NextString('');
      if buf = Target then
        Exit; { Found it }
      S.NextLine;
    end;
    Fail(S, Target + ' not found');
  end;
  function ReadPar(S: TTextStream; var Par: TPar; HasChange, HasScale: Boolean): Boolean;
  begin
    Result := False;
    if not S.GetFreeItem then
      Exit;
    if not S.IsIntegerToken(Par.n) then
      Exit;
    if Par.n <= 0 then Exit;
    NextVarAndTol(S, Par.OldV, Par.OldTol, Par.OldMx);
    NextVarAndTol(S, Par.NewV, Par.NewTol, Par.NewMx);
    NextVarAndTol(S, Par.Sigma, Par.SigmaTol, Par.SigmaMx);
    Par.NewTol := Par.NewTol*4e6; { Large slack here; significant difference picked up by summary test }
    if HasChange then begin
      NextVarAndTol(S, Par.Change, Par.ChangeTol, Par.ChangeMx);
      if Abs(Par.Change) > 10000 then
        Par.ChangeTol := Par.ChangeTol*2; { More slack on large changes }
    end;
    NextVarAndTol(S, Par.Sens, Par.SenTol, Par.SensMx);
    if HasScale then
      NextVarAndTol(S, Par.Scale, Par.ScaleTol, Par.ScaleMx);
    S.NextStringVar(Par.Summary);
    Par.Name := Trim(S.RestOfLine);
    Result := True;
  end;
  procedure WrP(var Par: TPar);
  begin
    WriteLn(Log, DtosMA(Par.OldV,20,-1), DtosMA(Par.NewV,20,-1), DtosMA(Par.Sigma,10,-1), DtosMA(Par.Change,10,4), DtosMA(Par.Sens,8,-1), ' ', Par.Summary:20, ' ',Par.Name);
  end;

var
  C, O: TTextStream;
  ParC, ParO: TPar;
  procedure FailP(Name: string);
  var
    l: Integer;
  begin
    Fail(C, Name);
    if Failures <= 0 then
      Exit;
    l := Max(Length(C.StreamName), Length(O.StreamName));
    Write(Log, C.StreamName:l, ' ');
    WrP(ParC);
    Write(Log, O.StreamName:l, ' ');
    WrP(ParO);
  end;

  procedure CheckNo(n1, n2, tol: Double; Name: string);
  begin
    Inc(nTried);
    if IsNanOrInf(n1) xor IsNanOrInf(n2) then
      FailP(Name + ' different')
    else if Abs(n1-n2) > tol then
      FailP(Format('%s out of tolerance(%s): Read=%g Calc=%g %g', [Name, DtosM(tol,7,-1), n2, n1, n1-n2]));
    if Abs(n1-n2) > Result then
      Result := Abs(n1-n2);
  end;

  procedure SplitSummary(S: TTextStream; b: string; out Val, Tol, Err: Double);
  var
    Ps, Pe: Integer;
    V, E: String;
    Mx: Double;
  begin
    Ps := Pos('(', b);
    Pe := Pos(')', b);
    if (Ps <= 0) or (Pe <= 0) then
      Fail(S, 'Bad ()');
    V := Copy(b,1,Ps-1)+Copy(b,Pe+1,Length(b));
    Mx := 0;
    VarAndTol(V, Val, Tol, Mx);
    if IsNanOrInf(Val) then
      SetNan(Err)
    else begin
      E := Copy(b,Ps+1, Pe-Ps-1);
      if not TryStrToFloat(E,Err) then
        Fail(S, 'Bad err:'+E);
      if Precision(E)/Err < 1e-5 then
        Tol := Tol*2; { Allow some slack if too many figures in error }
      Err := Err * Precision(V);
    end;
  end;

var
  nPar, i, j, n: Integer;
  OK, HasChange, CHasScale, OHasScale: Boolean;
  CV, Ctol, Cerr, OV, Otol, Oerr, CMx: Double;
  OName, OHeader: string;
begin
  Result := 0;
  nPar := 0;
  FillChar(ParO, Sizeof(ParO), 0);
  FillChar(ParC, Sizeof(ParC), 0);
  C := nil;
  O := nil;
  try
    Cur.Position := 0;
    C := TTextStream.Create(Cur, False);
    C.StreamName := 'New';
    OName := ChangeFileExt(OldName, '.out');
    if not FileExists(OName) then
      OName := OldName + '.out';
    O := TTextStream.Create(OName, True);
    O.StreamName := OName;
    if nParO = 0 then begin
      LookFor(C, 'Average');
      LookFor(O, 'Average');
      Exit; { TODO: Check error }
    end else begin
      LookFor(C, 'Initial');
      LookFor(O, 'Initial');
    end;
    if Failures > 0 then
      Exit;
    LookFor(C, 'Parameters:');
    C.NextLine;
    OHeader := C.RestOfLine;
    CHasScale := Pos('Scale',OHeader) > 0;
    LookFor(O, 'Parameters:');
    O.NextLine;
    OHeader := O.RestOfLine;
    HasChange := Pos('Change',OHeader) > 0;
    OHasScale := Pos('Scale',OHeader) > 0;
    repeat
      C.NextLine;
      O.NextLine;
      if ReadPar(C, ParC, True, CHasScale) then begin
        if not ReadPar(O, ParO, HasChange, OHasScale) then begin
          Fail(C, 'Extra parameter '+ParC.Name);
          Break;
        end;
      end else begin
        if ReadPar(O, ParO, HasChange, OHasScale) then
          Fail(O, 'Extra parameter '+ParO.Name);
        Break;
      end;
      Inc(nPar);
      if (ParC.n <> nPar) or (ParC.n <> ParO.n) then
        FailP('Parameters out of order');
      if ParC.Name <> ParO.Name then
        FailP('Different Names');
      CheckNo(ParC.OldV, ParO.OldV, ParC.OldTol+ParO.OldTol, 'Old');
      CheckNo(ParC.NewV, ParO.NewV, (ParC.NewTol+ParO.NewTol), 'New');
      if (LineListMode and 1) = 0 then begin
        CheckNo(ParC.Sigma, ParO.Sigma, ParC.SigmaTol+ParO.SigmaTol, 'Sigma');
        if HasChange then
          CheckNo(ParC.Change, ParO.Change, ParC.ChangeTol+ParO.ChangeTol, 'Change');
        CheckNo(ParC.Sens, ParO.Sens, ParC.SenTol+ParO.SenTol, 'Sens');
        if OHasScale and CHasScale then
          CheckNo(ParC.Scale, ParO.Scale, ParC.ScaleTol+ParO.ScaleTol, 'Scale');
      end;
      if ParC.Summary <> ParO.Summary then begin
        SplitSummary(C, ParC.Summary, CV, Ctol, Cerr);
        SplitSummary(O, ParO.Summary, OV, Otol, Oerr);
        CheckNo(CV, OV, Ctol+Otol, 'Summary value');
        CheckNo(Cerr, Oerr, Ctol+Otol, 'Summary error');
      end;
    until C.AtEndOfFile;
    if nPar = 1 then
      Exit; { No Correlation matrix }
    LookFor(C, 'Correlation'); C.NextLine;
    LookFor(O, 'Correlation'); O.NextLine;
    if C.NextString('') = 'Determinant' then
      C.NextLine;
    if O.NextString('') = 'Determinant' then
      O.NextLine;
    for i := 1 to nPar do begin
      C.NextLine;
      O.NextLine;
      if C.NextInteger(-1) <> i then
        Fail(C, 'Expected '+IntToStr(i));
      if O.NextInteger(-1) <> i then
        Fail(O, 'Expected '+IntToStr(i));
      CMx := 1;
      for j := 1 to i do begin
        NextVarAndTol(C, CV, Ctol, CMx);
        NextVarAndTol(O, OV, Otol, CMx);
        CheckNo(CV, OV, Ctol+Otol, Format('%d,%d', [i, j]));
      end;
    end;
  finally
    C.Free;
    O.Free;
  end;
end;

{ Expression evaluation checks }
function CheckExprs(S: TTextStream; var nTried: Integer; Tolerance: Double; var Failures: Integer): Double;
var
  Line, Expected: string;
  v, vEx: Double;
  Variables: TVariableList;
  LastResult: TVariable;
  procedure Fail(Name: string);
  begin
    Inc(Failures);
    if Failures > 0 then begin
      WriteLn(Log, 'Line ', S.Line, ' of ' + S.StreamName + ':' + Line + ':', Name);
    end;
  end;
  procedure CheckNo(n1, n2, tol: Double);
  begin
    if IsNanOrInf(n1) xor IsNanOrInf(n2) then
      Fail('Different')
    else if Abs(n1-n2) > tol then
      Fail(Format('Out of tolerance(%s): Read=%g Calc=%g %g', [DtosM(tol,7,-1), n2, n1, n1-n2]));
    if Abs(n1-n2) > Result then
      Result := Abs(n1-n2);
  end;
begin
  Result := 0;
  Variables := TVariableList.Create;
  LastResult := TVariable.Create('!');
  Variables.Add(LastResult);
  while not S.AtEndOfFile do begin
    Line := S.RestOfLine;
    S.NextLine;
    if Trim(Line) = '' then
      Continue;
    if Line[1] <> '>' then
      Fail('Bad line');
    Line := Trim(Copy(Line,2,Length(Line)));
    Expected := S.RestOfLine;
    S.NextLine;
    Inc(nTried);
    vEx := StrToFloatEx(Expected);
    v := EvaluateExpression(Line, Variables, [xImplicitArrays]);
    LastResult.Value := v;
    CheckNo(vEx, v, Tolerance);
  end;
  Variables.Free;
end;

{ Algebraic simplification checks }
function CheckAlg(S: TTextStream; var nTried: Integer; Tolerance: Double; var Failures: Integer): Double;
var
  Line, Expected, Actual: string;
  Lex: TLex;
  Expr: TExpr;
  VarList: TVariableList;
  ExprNode: TExprNode;
  Actions: TSimplifyActions;
  procedure Fail(Name: string);
  begin
    Inc(Failures);
    if Failures > 0 then begin
      WriteLn(Log, 'Line ', S.Line, ' of ' + S.StreamName + ':' + Line + ':', Name);
    end;
  end;
  procedure CheckNo(n1, n2, tol: Double);
  begin
    if IsNanOrInf(n1) xor IsNanOrInf(n2) then
      Fail('Different')
    else if Abs(n1-n2) > tol then
      Fail(Format('Out of tolerance(%s): Before=%g After=%g %g', [DtosM(tol,7,-1), n2, n1, n1-n2]));
    if Abs(n1-n2) > Result then
      Result := Abs(n1-n2);
  end;
var
  j, BaseC: Integer;
  vn: string;
  V, VA: Double;
begin
  BaseC := ExprNodesCreated - ExprNodesDestroyed;
  Actions := [siMinusOneInteger];
  while not S.AtEndOfFile do begin
    Line := S.RestOfLine;
    S.NextLine;
    if Trim(Line) = '' then
      Continue;
    if Trim(Line) = 'Expand' then begin
      Actions := [siMinusOneInteger,siExpand];
      Continue;
    end;
    if Line[1] <> '>' then
      Fail('Bad line');
    Line := Trim(Copy(Line,2,Length(Line)));
    Expected := Trim(S.RestOfLine);
    S.NextLine;
    Inc(nTried);
    Lex := TLex.Create;
    Lex.Options := Lex.Options - [xFundamentalConstants];
    Expr := TExpr.Create(Lex, True);
    VarList := Expr.MakeVariables(Line);
    for j := 0 to VarList.Count - 1 do begin
      { Essentially random test values }
      vn := LowerCase(VarList[j]);
      if vn = 'j' then V := 24
      else if vn = 'f' then V := 24.5
      else if vn = 'm' then V := -19
      else if vn = 'j1' then V := 28
      else if vn = 'j2' then V := 34
      else if vn = 'j3' then V := 32
      else if vn = 'n' then V := 23
      else if vn = 's' then V := 1
      else if vn = 'k' then V := -13
      else if vn = 'h' then V := 15.5
      else
        V := sin(27-j);
      VarList.Objects[j].Value := V;
    end;
    Expr.Text := Line;
    V := Expr.Value;
    ExprNode := Expr.ExprTree.Simplify(Actions);
    Actual := ExprNode.Text(True);
    if Actual <> Expected then
      Fail(Actual+ ' but expected ' + Expected);
    with ExprNode.Compile do begin
      VA := Value;
      Free;
    end;
    CheckNo(VA, V, Tolerance);
    ExprNode.Free;
    if ExprNodesCreated-BaseC <> ExprNodesDestroyed then
      Fail(IntToStr(ExprNodesCreated-BaseC-ExprNodesDestroyed)+' nodes leaked');
    Expr.Free;
  end;
end;

end.

