unit CheckSpecView;
{ Read a specview output file, and compare to a calculation of the same thing }

{$I utils.inc}{Copyright (C) 2007 C M Western; see commandline.pas for license}

interface

uses Mixture, TextStrm;

function CheckSpecViewFile(Mixture: TMixture; S: TTextStream;
                          var nTried: Integer; Tolerance: Double; var Failures: Integer;
                          IsGHz: Boolean):Double;

implementation

uses SysUtils, QuantumNumber, Worker, MathSup, MatrixVec, OAsymT, MNode, DRintensity;

function CheckSpecViewFile(Mixture: TMixture; S: TTextStream;
                          var nTried: Integer; Tolerance: Double; var Failures: Integer;
                          IsGHz: Boolean):Double;
var
  Failed: Boolean;
  Jp, J: TQuantumNumber;
  ManifoldL, ManifoldU: TManifold;
  Molecule: TMolecule;
  np: Smallint; { Upper State }
  symp: Shortint;
  StrU, StrL: string;
  n: SmallInt; { Lower State }
  sym: Shortint;
  Frequency, Intensity, cstr: Double;
  MaxDifference: Double;
  Simulation: TSimulation;
  Line: TLine;
  Problem: Exception;
  procedure Fail(Name: string);
  begin
    Inc(Failures);
    if Failures > 0 then begin
      WriteLn(Log, 'Line ', S.Line, ':', Name);
      WriteLn(Log, Jp:3, J:5, Frequency:12:4, Intensity:8:4, ' ', StrU, ' ', StrL);
      Mixture.LineList.ShowLine(Line, 0, Mixture.Units, Mixture.Units, False, Problem);
    end;
    Failed := True;
  end;
  function CheckNo(n1, n2, tol: Double; Name: string; Defer: Boolean = False):Boolean;
  var
    Msg: string;
  begin
    Result := True; { Assume fail }
    if isNanOrInf(n2) then
      Msg := Format('%s fails: %g Nan', [Name, n1])
    else if Abs(n1-n2) > tol then begin
      Msg := Format('%s out of tolerance: %g %g %g (%g)', [Name, n1, n2, n1-n2, n1/n2])
    end else
      Result := False; { OK }
    if Abs(n1-n2) > MaxDifference then
      MaxDifference := Abs(n1-n2);
    if Result then
      Fail(Msg);
  end;
  procedure NextQuantumVar(var Q: TQuantumNumber);
  var
    str: String;
  begin
    str := S.NextString('?');
    if (str <> '') and (str[Length(str)] = '>') then
      Delete(str,Length(str),1);
    if (str <> '') and (str[Length(str)] = ',') then
      Delete(str,Length(str),1);
    if not TryStrToQuantumNumber(str, Q, qftJ) then
      Fail('Bad quantum number:'+str);
  end;
  procedure CheckAsymRot;
  var
    Kap, Kcp, Ka, Kc: TQuantumNumber;
    AlternateSym: ShortInt;
  begin
    NextQuantumVar(Kap); NextQuantumVar(Kcp);
    if (S.NextString('') <> '-->') or (S.NextString('') <> '|') then begin
      Fail('Bad line');
      S.Newline;
      Exit;
    end;
    NextQuantumVar(J); NextQuantumVar(Ka); NextQuantumVar(Kc);
    S.NextDoubleVar(Frequency); S.NextDoubleVar(Intensity);
    StrU := ManifoldU.FindIndex(Line.EigenvaluesU, np, symp, AlternateSym, Jp, [Kap, Kcp], qftJ, Simulation.ExField);
    if np < 0 then
      Fail(StrU)
    else begin
      StrL := ManifoldL.FindIndex(Line.EigenvaluesL, n, sym, AlternateSym, J, [Ka, Kc], qftJ, Simulation.ExField);
      if n < 0 then
        Fail(StrL);
    end;
  end;
  procedure CheckSpinRot;
  var
    QNp, Kap, Pp, QN, Ka, P: TQuantumNumber;
    AlternateSym: ShortInt;
  begin
    NextQuantumVar(QNp); NextQuantumVar(Kap); NextQuantumVar(Pp);
    if (S.NextString('') <> '-->') or (S.NextString('') <> '|') then begin
      Fail('Bad line');
      S.Newline;
      Exit;
    end;
    NextQuantumVar(J); NextQuantumVar(QN); NextQuantumVar(Ka); NextQuantumVar(P);
    S.NextDoubleVar(Frequency); S.NextDoubleVar(Intensity);
    StrU := ManifoldU.FindIndex(Line.EigenvaluesU, np, symp, AlternateSym, Jp, [Abs(Kap)], qftJ, Simulation.ExField, Pp div 2, -1, QNp);
    if np < 0 then
      Fail(StrU)
    else begin
      StrL := ManifoldL.FindIndex(Line.EigenvaluesL, n, sym, AlternateSym, J, [Abs(Ka)], qftJ, Simulation.ExField, P div 2, -1, QN);
      if n < 0 then
        Fail(StrL);
    end;
  end;
  procedure CheckTriplet;
  var
    QNu, QN: TQuantumNumber;
    AlternateSym: ShortInt;
  begin
    NextQuantumVar(QNu);
    if (S.NextString('') <> '-->') or (S.NextString('') <> '|') then begin
      Fail('Bad line');
      S.Newline;
      Exit;
    end;
    NextQuantumVar(J); NextQuantumVar(QN);
    S.NextDoubleVar(Frequency); S.NextDoubleVar(Intensity);
    StrU := ManifoldU.FindIndex(Line.EigenvaluesU, np, symp, AlternateSym, Jp, [], qftJ, Simulation.ExField, -1, -1, QNu);
    if np < 0 then
      Fail(StrU)
    else begin
      StrL := ManifoldL.FindIndex(Line.EigenvaluesL, n, sym, AlternateSym, J, [], qftJ, Simulation.ExField, -1, -1, QN);
      if n < 0 then
        Fail(StrL);
    end;
  end;
  procedure CheckDoublet;
  var
    Pu, P: Integer;
    Omegau, Omega: TQuantumNumber; { Or N }
    function FindState(J, Omega: TQuantumNumber; Parity: Integer;
                       Manifold: TManifold; var Eigenvalues: TEigenvalues;
                       var n: Smallint; var Sym: Shortint): string;
    var
      AlternateSym: ShortInt;
    begin
      Parity := (2-Parity) div 4;
      if Odd(Omega) then
        Result := Manifold.FindIndex(Eigenvalues, n, sym, AlternateSym, J, [Omega], qftJ, Simulation.ExField, Parity)
      else
        Result := Manifold.FindIndex(Eigenvalues, n, sym, AlternateSym, J, [], qftJ, Simulation.ExField, Parity, -1, Omega)
    end;
  begin
    NextQuantumVar(Pu); NextQuantumVar(Omegau);
    if (S.NextString('') <> '-->') or (S.NextString('') <> '|') then begin
      Fail('Bad line');
      S.Newline;
      Exit;
    end;
    NextQuantumVar(J); NextQuantumVar(P); NextQuantumVar(Omega);
    S.NextDoubleVar(Frequency); S.NextDoubleVar(Intensity);
    StrU := FindState(Jp, Omegau, Pu, ManifoldU, Line.EigenvaluesU, np, symp);
    if np < 0 then
      Fail(StrU)
    else begin
      StrL := FindState(J, Omega, P, ManifoldL, Line.EigenvaluesL, n, sym);
      if n < 0 then
        Fail(StrL);
    end;
  end;
var
  s1, s2: string;
  TransitionMoments: TTransitionMoments;
begin
  S.Comments := False;
  MaxDifference := 0;
  Simulation := Mixture.Simulations[0];
  Problem := nil;
  Simulation.MakeLineList(Problem);
  if Assigned(Problem) then
    raise Problem;
  Molecule := Mixture.Species[0].Isotopomer[0];
  ManifoldL := Molecule.Manifold[0];
  if Molecule.nManifolds > 1 then
    ManifoldU := Molecule.Manifold[1]
  else
    ManifoldU := ManifoldL;
  while not S.AtEndOfFile do begin
    { Line format is | 9, 9, 1,> --> | 10, 10, 0,>  22391.2317  0.59 # ... }
    Failed := False;
    s1 := S.NextString(''); s2 := S.NextString('');
    if (s2 <> '') and (s2[Length(s2)] = ',') then
      Delete(s2,Length(s2),1);
    if (s1 <> '|') or not TryStrToQuantumNumber(s2,Jp,qftJ) then begin
      S.NextLine;
      Continue;
    end;
    Failed := False;
    Inc(nTried);
    if ManifoldU is TAsymmetricManifold then
      if ManifoldL.State[0].S > 0 then
        CheckSpinRot
      else
        CheckAsymRot
    else if ManifoldL.State[0].S = 2 then
      CheckTriplet
    else
      CheckDoublet;
    if not Failed then begin
      Line.Width := 0;
      Line.IndexU := np;
      Line.IndexL := n;
      Line.Frequency := Line.EigenvaluesU.E[Line.IndexU]-Line.EigenvaluesL.E[Line.IndexL];
      Line.CachedName := '';
      TransitionMoments := Molecule.FindTransitionMoments(ManifoldU,ManifoldL);
      if Assigned(TransitionMoments) then
        { Do this here so line printout has correct intensity }
        TransitionMoments.MakeLine(Line,Line.EigenvaluesL,Line.EigenvaluesU,Line.IndexL,Line.IndexU,Simulation,True);
      if IsGhz then
        CheckNo(Frequency, Line.Frequency/1000, Tolerance, 'Frequency')
      else
        CheckNo(Frequency, Line.Frequency, Tolerance, 'Frequency');
      if not Assigned(TransitionMoments) then
        Fail('Can''t find transition moment')
      else begin
        Problem := nil;
        cstr := Simulation.LineIntensity(Line, Mixture.Units, Problem);
        if Assigned(Problem) then begin
            Fail(Problem.Message);
            Problem.Free;
          end;
        { Intensity is only printed to 0.01 }
        CheckNo(Intensity, cstr, 0.01, 'Intensity', True);
      end;
    end;
    S.NextLine;
  end;
  Result := MaxDifference;
end;

end.
