unit CommandLine;
{$I utils.inc}
{
PGOPHER - a program for simulating rotational structure
Copyright (C) 2005-2015 C M Western
School of Chemistry, University of Bristol, Bristol, UK
For further information please see http://pgopher.chm.bris.ac.uk/
or contact help-pgopher@bristol.ac.uk

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
}
interface

uses Classes, SysUtils;

function HandleCommandLine: Boolean;
procedure DisconnectLogs;

implementation

{.$DEFINE SKELETON}

uses Profile,
     {$IFNDEF CONSOLE}
       {$IFDEF MSWINDOWS}
         Windows,
       {$ENDIF}
       {$IFDEF LCL_} LCLType, {$ENDIF}
       Main, LogUnit, Forms, OvrlyFrm, SyncObjs,
     {$ENDIF}
     {$IFDEF VCL_}
       {$IFNDEF CONSOLE }
         StackDump,
       {$ENDIF}
     {$ENDIF}
     {$IFDEF SKELETON}
     Skeleton,
     {$ENDIF}
     Global, Mixture, D3hImport, PgopherImport, CheckPgopher, CheckOverlay,
     TextStrm, Overlay, MNode, JPLCat, MatrixElements, Obs, Fitter, Expr,
     Worker, TextAssign, Sysinfo, PGWork, Math, CheckSpecView, CheckJB95,
     CheckImport, Hitran, DataSource, UCompat, Plotter, ExportG, ImportLmatrix,
     GraphM, LevelPlotter, ExpCan, Progress, Units, NextTok, Clock, MathSup,
     ImportCALPGM, InternalCoords, ExoMol, OverlaySumFitter, Exporter, Importer,
     AutoFitHelper;

{$IFNDEF CONSOLE}

{$IFNDEF MSWINDOWS}
var
  LastProcessMessages: TDateTime = 0;
{$ENDIF}

function WaitGui(var Event: TEvent; Timeout: Cardinal; ReturnOnAbort: Boolean = True; ShowStatus: TOnShowStatus = nil): TWaitResult;
var
  Handle: {$IFDEF FPC}TEventHandle{$ELSE}THandle{$ENDIF};
  EventCopy: TEvent;
  procedure UpdateStatus;
  begin
    if Assigned(ShowStatus) then
      ShowStatus(ReturnOnAbort)
    else
      SetProgress(CurrentProgress, -1, ReturnOnAbort);
    Application.ProcessMessages;
  end;
begin
  if GetCurrentThreadID <> MainThreadID then
    Result := SimpleWait(Event, Timeout)
  else begin
    Result := wrAbandoned;
    {$IFDEF MSWINDOWS}
    // Yield;
    repeat
      if (Aborted and ReturnOnAbort) or Application.Terminated then
        Exit;
      EventCopy := Event; { Copy as Event can be set to nil after entry }
      if not Assigned(EventCopy) then
        Exit; { Event no longer set up }
      Handle := EventCopy.Handle;
      case MsgWaitForMultipleObjects(1, Handle, False, Timeout, QS_ALLINPUT) of
        WAIT_OBJECT_0: Result := wrSignaled;
        WAIT_ABANDONED_0: Result := wrAbandoned;
        WAIT_OBJECT_0+1: begin;
                           UpdateStatus;
                           Continue;
                         end;
        WAIT_FAILED: RaiseLastOSError;
        WAIT_TIMEOUT:
          begin
            UpdateStatus;
            Result := wrTimeout;
          end;
      end;
      Exit;
    until False;
    {$ELSE}
    // Yield;
    repeat
      if (Aborted and ReturnOnAbort) or Application.Terminated then
        Exit;
      if not Assigned(Event) then
        Exit; { Event no longer set up }
      if Now-LastProcessMessages > 200/MSecsPerDay then begin
        UpdateStatus;
        LastProcessMessages := Now;
      end;
      EventCopy := Event; { Copy as Event can be set to nil after entry }
      if not Assigned(EventCopy) then
        Exit;
      Result := EventCopy.WaitFor(Timeout); { Ideally should wait on message available as well }
      case Result of
        wrError: RaiseLastOSError;
//        wrTimeout: Continue;
      end;
    until True;
    {$ENDIF}
  end;
end;
{$ENDIF}

var
  LogConnected: Boolean = False;

procedure ConnectLog(OutFile: string);
{$IFNDEF CONSOLE}
var
  FileName: string;
{$ENDIF}
begin
  {$IFNDEF CONSOLE}
  if not Assigned(LogForm) then
    Application.CreateForm(TLogForm, LogForm);
  if not Assigned(OverlayForm) then
    Application.CreateForm(TOverlayForm, OverlayForm);
  with Application do begin
    FileName := HelpFile;
    if not FileExists(FileName) then begin
      FileName := ExtractFilePath(Application.ExeName) + FileName;
      if not FileExists(FileName) then
        FileName := ExtractFilePath(Application.ExeName) + 'Help' + PathDelim + ExtractFileName(FileName);
    end;
    HelpFile := ExpandFileName(FileName);
  end;
  if OutFile = '' then begin
    { Output connection is just to catch any Write()s to Output,
      though they should be all to Log }
    //AssignStrings(Output, LogForm.Log.Lines);
    //Rewrite(Output);
    AssignStrings(Log, LogForm.Log.Lines);
    Rewrite(Log);
  end else
  {$ENDIF}
  begin
    if OutFile <> '' then
      AssignFile(Output, OutFile);
    {$IFDEF DCC} {Delphi bug? stdout not forced to be crlf - OK for console, but divert to file weird }
    SetLineBreakStyle(Output, DefaultTextLineBreakStyle);
    {$ENDIF}
    Rewrite(Output);
    AssignTextFile(Log,Output);
    Rewrite(Log);
  end;
  LogConnected := True;
end;

procedure DisconnectLogs;
begin
  { Clean shutdown }
  StartWorkers(0,nil);
  CloseFile(Log);
  {$IFDEF LINUX}
  AssignFile(Log,'');
  Append(Log);
  {$ENDIF}
end;

type
  TNewPgoOpt = (npgNone, npgOver, npgNew);

var
  nTried: Integer;
  nFailures: Integer;
  Tolerance: Double;
  LineListMode: Integer;
  CheckFile: string;
  PGOoutput: string;

{$IFDEF CONSOLE}
procedure WriteNewPgoFile(NewPgoOpt: TNewPgoOpt);
var
  i: Integer;
  OrgName, OtherFileName: string;
begin
  { Write out new parameters }
  if PGOoutput <> '' then begin
    NewPgoOpt := npgNew;
    OtherFileName := PGOoutput;
  end else if NewPgoOpt = npgNone then
    Exit
  else begin
    { Find unused "other" file name }
    OrgName := GlobalMixture.DirectoryName + PathDelim + GlobalMixture.Name;
    i := 0;
    repeat
      Inc(i);
      OtherFileName := ChangeFileExt(OrgName, '.'+IntToStr(i)+'.pgo');
    until not FileExists(OtherFileName);
  end;
  { TODO: Consider saving overlay parameters also }
  if NewPgoOpt = npgOver then
    { Overwrite input pgo file after backing it up }
    if not RenameFile(OrgName, OtherFileName) then
      raise Exception.Create('Can''t rename '+OrgName+' to '+OtherFileName)
    else begin
      WriteLn('Writing new parameters to ', OrgName, ' and saving old parameters to ', OtherFileName);
      GlobalMixture.SaveToFile(OrgName);
    end
  else begin
    { Write to new file }
    WriteLn('Writing new parameters to ', OtherFileName);
    GlobalMixture.SaveToFile(OtherFileName);
  end;
end;

function DivertCheckLog(ObsFile: string): string;
var
  OutName, OutDir: string;
begin
  OutName := StringReplace(ObsFile,'/','__',[rfReplaceAll]);
  OutName := StringReplace(OutName,'\','__',[rfReplaceAll]);
  {$IFDEF FPC}
    OutDir := GetTempDir+'/pgopher-'+{$INCLUDE %FPCTARGETCPU%}+'-'+{$INCLUDE %FPCTARGETOS%}+'/';
  {$ELSE}
    OutDir := GetEnvironmentVariable('TEMP')+'\pgopher';
    {$IFDEF CPU32}
      OutDir := OutDir + '-32';
    {$ENDIF}
    {$IFDEF CPU64}
      OutDir := OutDir + '-64';
    {$ENDIF}
    {$IFDEF UNICODE}
      OutDir := OutDir + 'U';
    {$ENDIF}
    OutDir := OutDir + '\';
  {$ENDIF}
  ForceDirectories(OutDir);
  Result := OutDir+OutName+'.out.new';
  AssignFile(Log,Result);
  Rewrite(Log);
end;

procedure CancelDivert;
begin
  CloseFile(Log);
  AssignFile(Log,''); { Reconnect to standard output }
  {$IFNDEF FPC}
    Rewrite(Log); { Delphi Bug - Append tries to open STDIN, not STDOUT }
  {$ELSE}
    Append(Log);
  {$ENDIF}
end;

function DoFit(ObsFile, ObsFile1: string; ObsType:TObsType; Check, OutFile: string; NewPgoOpt: TNewPgoOpt; CheckDerivatives: Double; TakeWeight: TTakeWeight = tkObs):string;
var
  Prec, i: Integer;
  function CheckResult(Expected, Actual: string):Boolean;
  var
    E, A: string;
    EV, AV, Tol: Double;
    First: Boolean;
  begin
    { Compare tokens in the string, allowing for small differences in the numbers }
    Result := True;
    First := True;
    repeat
      E := NextToken(Expected);
      A := NextToken(Actual);
      if E <> A then begin
        Result := False;
        if (Length(E) > 1)  and (E[1] in ['0'..'9','+','-','.']) then begin
          if E[Length(E)] = ')' then
            Delete(E, Length(E), 1);
          if A[Length(A)] = ')' then
            Delete(A, Length(A), 1);
          if CompareText(A,'Nan') = 0 then begin
            Result := False;
            Exit;
          end;
          EV := StrToFloat(E);
          Tol := Precision(E)*1.6;
          AV := StrToFloat(A);
          Result := SameValue(EV,AV,Tol);
          First := False;
        end;
        if not Result then
          Exit;
      end;
    until (Expected = '') and (Actual = '');
  end;
  procedure WriteParams;
  var
    i: Integer;
    Prm: string;
  begin
    for i := 1 to Paramcount do begin
      Prm := ParamStr(i);
      if (Pos(' ', Prm) > 0) or (Pos('"', Prm) > 0) then
        Prm := AnsiQuotedStr(Prm, '"');
      Write(Log, ' ', Prm);
    end;
  end;
var
  ObsStream: TStream;
  ObsFileToRead: string;
  FitType: TFitType;
  StartTime: Int64;
  CheckOut, Prm: string;
  LargestError: Double;
  NewS: TMemoryStream;
  OverallOK: Boolean;
  OldSD, NewSD: Double;
begin
  StartTime := Clocks();
  GlobalMixture.CheckProperties;
  try
    if (Check <> '') and (OutFile = '') then
      CheckOut := DivertCheckLog(ObsFile)
    else
      CheckOut := '';
    WriteLn(Log, Global.VersionInfo);
    WriteParams;
    WriteLn(Log);
    case ObsType of
      obContour:    FitType := ftContour;
      obExpression: FitType := ftOverlaySum;
    else
      FitType := ftLine;
    end;
    ObsFileToRead := '';
    if FitType <> ftLine then begin
      ObsType := obContour;
      if not Assigned(GlobalOverlays) then
        raise EReportException.Create('Overlays required');
      if GlobalOverlays.ContourFile then
          if ObsFile1 = '' then
            raise EReportException.Create('ContourFile flag set but no external file provided')
          else
            ObsFileToRead := ObsFile1
      else
        ClearObservations;
    end else
      ObsFileToRead := ObsFile;
    if ObsFileToRead <> '' then begin
      ObsStream := TFileStream.Create(ObsFileToRead, fmOpenRead or DefaultShareMode);
      try
        LoadObservations(GlobalMixture, ObsStream, ObsFileToRead, ObsType, TakeWeight);
      finally
        ObsStream.Free;
      end;
    end;
    for i := 1 to GlobalMixture.FitCycles do begin
      case FitType of
        ftContour:
          begin
            GlobalMixture.PrepareForFit(ftContour,GlobalOverlays);
            Result := FitContour(GlobalMixture, GlobalOverlays, nil, GlobalMixture.FittingParameters, False, CheckDerivatives);
          end;
        ftOverlaySum:
          begin
            GlobalMixture.PrepareForFit(ftOverlaySum,GlobalOverlays);
            Result := FitOverlaySum(GlobalMixture, GlobalOverlays, nil, GlobalMixture.FittingParameters, False, CheckDerivatives);
          end;
      else
        if ObsType = obIntensity then
          Prec := -1
        else
          Prec := GlobalMixture.Precision;
        GlobalMixture.PrepareForFit(ftLine,nil);
        Result := FitLines(GlobalMixture.FittingParameters, Observations, Prec, GlobalMixture, CheckDerivatives, OldSD, NewSD);
      end;
    end;
    if Length(GlobalMixture.FittingParameters) > 0 then
      WriteNewPgoFile(NewPgoOpt);
  except
    on E: Exception do
      if Check <> '' then
        Result := E.Message
      else
        raise;
  end;
  if Check <> '' then begin
    CancelDivert;
    NewS := TMemoryStream.Create;
    try
      UnlockedLoadFromFile(NewS, CheckOut);
      if CheckFile = '' then
        CheckFile := ObsFileToRead;
      if CheckFile = '' then
        CheckFile := ObsFile; { Probably .ovr file }
      LargestError := CheckFitStream(NewS, CheckFile, Length(GlobalMixture.FittingParameters), nTried, Tolerance, nFailures, LineListMode);
    finally
      NewS.Free;
    end;
    if nFailures <> 0 then
      Write(Log, nFailures, ' failure(s) out of ', nTried);
    OverallOK := CheckResult(Check,Result);
    if (nFailures = 0) and OverallOK then
      Write(Log, 'OK (', nTried:5, ')');
    if OverallOK then
      Write(Log, ' Largest Error = ', LargestError:20, (Clocks()-StartTime)/ClockFrequency:10:3, ' s  : (', Result, ')')
    else
      Write(Log, '**** failure: Fit mismatch :', Result, '<>');
    WriteParams;
    WriteLn(Log);
  end else
    WriteLn(Log, Result);
end;
{$ENDIF}

var
  LoadFileTime: string = 'Load Comparision File';

function HandleCommandLine: Boolean;
var
  Stream: TTextStream;
  CMStream: TMemoryStream;
  i, d, np, nParams, CODATAVersion: Integer;
  Params, MixtureChanges, OverlayChanges: array of string;
  Script, ScriptFile: TStringList;
  LargestError, Offset, CheckDerivatives: Double;
  SortBoth, Interpolate, CvtToPert, SkipELowerCheck, DoubleABlevels, NoCalc: Boolean;
  NewPgoOpt: TNewPgoOpt;
  InputPgopherVersion, PgopherVersion, QNOmask: Integer;
  Check, OutFile: string;
  Templates, TemplateFiles: array of string;
  procedure ApplyChanges;
  var
    i: Integer;
  begin
    if (Length(MixtureChanges) > 0) or (Length(Templates) > 0) then begin
      if GlobalMixture = nil then
        {$IFDEF CONSOLE}
          GlobalMixture :=
        {$ELSE}
          MainForm.Mixture :=
        {$ENDIF}
          TMixture.Create(nil);
      for i := 0 to High(MixtureChanges) do
        GlobalMixture.HandleAssignment(MixtureChanges[i]);
      SetLength(MixtureChanges, 0); { Allow routine to be called multiple times }
      for i := 0 to High(Templates) do
        ImportMNodes(GlobalMixture, Templates[i], TemplateFiles[i], GlobalMixture.PrintLevel >= plDetail);
      SetLength(TemplateFiles,0);
    end;
    if GlobalOverlays <> nil then begin
      for i := 0 to High(OverlayChanges) do
        GlobalOverlays.HandleAssignment(OverlayChanges[i]);
      SetLength(OverlayChanges, 0);
    end;
  end;
  procedure FitHelper(ObsType: TObsType);
  var
    i: Integer;
    o: string;
    {$IFNDEF CONSOLE}
    Stream: TMemoryStream;
    {$ENDIF}
  begin
    {$IFNDEF CONSOLE}
      case ObsType of
        obCDLower:    i := 6;
        obCDUpper:    i := 7;
        obExpression: i := 8;
      else
        i := Ord(ObsType);
      end;
      LogForm.FitType.ItemIndex := i;
      LogForm.DoFit(LogForm.ObsFileName.FileName, False, tkObs, CheckDerivatives, -1);
      if Check <> '' then
        try
          Stream := TMemoryStream.Create;
          LogForm.Log.Lines.SaveToStream(Stream);
          if CheckFile = '' then
            CheckFile := LogForm.ObsFileName.FileName;
          CheckFitStream(Stream, CheckFile, Length(GlobalMixture.FittingParameters), nTried, Tolerance, nFailures, LineListMode);
        finally
          Stream.Free;
        end;
    {$ELSE}
      if High(Params) > 2 then
        o := Params[3]
      else
        o := '';
      DoFit(Params[2],o,ObsType,Check,OutFile,NewPgoOpt,CheckDerivatives);
    {$ENDIF}
  end;
var
  Problem: Exception;
  LoadedMixture: TMixture;
  CDir, FName: string;
  StartTime: Int64;
  Ext: string;
  Reader: TMixtureReader;
  IsOverlay: array of Boolean;
begin
  SetLength(MixtureChanges,0);
  SetLength(OverlayChanges,0);
  Stream := nil;
  Script := nil;
  {$IFDEF CONSOLE}
    Global.VersionInfo := {$I commandlinever.inc} ;
    {$IFDEF FPC}
      Global.VersionInfo := Global.VersionInfo + ' ' + {$I arch.inc} + ' ' + {$INCLUDE %DATE%};
    {$ENDIF}
    ShowTimingAtEnd := False;
  {$ELSE}
    DoWait := {$IFDEF FPC}@{$ENDIF}WaitGui;
    with MainForm.Project do
      Global.VersionInfo := 'Pgopher ' + Version + '.' + IntToStr(Build) + ' ' + BuildDate;
    {$IFDEF FPC}
      Global.VersionInfo := Global.VersionInfo + ' ' + {$I arch.inc};
    {$ENDIF}
  {$ENDIF}
  {$IFDEF CPU32}
    Global.VersionInfo := Global.VersionInfo + ' 32 bit';
  {$ENDIF}
  {$IFDEF CPU64}
    Global.VersionInfo := Global.VersionInfo + ' 64 bit';
  {$ENDIF}
  {$IFDEF UNICODE}
    Global.VersionInfo := Global.VersionInfo + ' Unicode';
  {$ENDIF}
  {$IFDEF FPC}
    Global.VersionInfo := Global.VersionInfo + ' (fpc ' + {$INCLUDE %FPCVERSION%}+' '+{$INCLUDE %FPCTARGETCPU%}+'-'+{$INCLUDE %FPCTARGETOS%}+')';
  {$ELSE}
    Global.VersionInfo := Global.VersionInfo + ' (Delphi ' + FloatToStr(CompilerVersion) + '/' + FloatToStr(RTLVersion) + ')';
  {$ENDIF}
  PgopherVersion := DecodePgopherVersion(Global.VersionInfo);
  AddExpectedException(EReportException);
  AddExpectedException(EExprFormatError);
  AddExpectedException(EFitError);
  UCompat.VersionInfo := Global.VersionInfo;
  { Ignore system setting - there are internal strings that are not localised
    and data files should not be localised. Text input could be localised, though
    the langauge of the program is English }
  {$IFDEF UNICODE}FormatSettings.{$ENDIF}DecimalSeparator := '.';
  // SetTrackMemory(True);
  np := -1; { Default is to auto calculate the number of worker threads }
  try
    Result := False;
    nTried := 0;
    Tolerance := 1e-4;
    nFailures := 0;
    nExprFailures := 0;
    SortBoth := False;
    Interpolate := False;
    CvtToPert := False;
    SkipELowerCheck := False;
    DoubleABlevels := False;
    QNOmask := 0;
    CODATAVersion := 0;
    Check := '';
    OutFile := '';
    PGOoutput := '';
    CheckFile := '';
    Offset := 0;
    CheckDerivatives := 0;
    NewPgoOpt := npgNone;
    LineListMode := 0;
    InputPgopherVersion := PgopherVersion;
    nParams := ParamCount;
    NoCalc := False;
    {$IFDEF DARWIN}{$IFNDEF CONSOLE}
    nParams := 0; { Parameters to gui files come via drop files, not command line }
    {$ENDIF}{$ENDIF}
    SetLength(Params, nParams);
    d := 0;
    i := 1;
    while i <= nParams do begin
      Params[d] := ParamStr(i);
      if Copy(Paramstr(i), 1, 8) = 'Mixture.' then begin
        SetLength(MixtureChanges, Length(MixtureChanges)+1);
        MixtureChanges[High(MixtureChanges)] := Copy(Paramstr(i), 9, Length(Paramstr(i)));
      end else if Copy(Paramstr(i), 1, 9) = 'Overlays.' then begin
        SetLength(OverlayChanges, Length(OverlayChanges)+1);
        OverlayChanges[High(OverlayChanges)] := Copy(Paramstr(i), 10, Length(Paramstr(i)));
      end else if ParamStr(i) = '--tolerance' then begin
        Tolerance := StrToFloatEx(ParamStr(i+1));
        Inc(i);
      end else if ParamStr(i) = '--fail' then begin
        nFailures := -StrToInt(ParamStr(i+1));
        Inc(i);
      end else if ParamStr(i) = '--linelistmode' then begin
        LineListMode := StrToInt(ParamStr(i+1));
        Inc(i);
      end else if ParamStr(i) = '--sort' then
        SortBoth := True
      else if ParamStr(i) = '--offset' then begin
        Offset := StrToFloatEx(ParamStr(i+1));
        Inc(i);
      end else if ParamStr(i) = '--checkderivatives' then begin
        CheckDerivatives := StrToFloatEx(ParamStr(i+1));
        Inc(i);
      end else if (ParamStr(i) = '-o') or (ParamStr(i) = '--o') then begin
        OutFile := ParamStr(i+1);
        Inc(i);
      end else if ParamStr(i) = '--pgo' then begin
        PGOoutput := ParamStr(i+1);
        Inc(i);
      end else if ParamStr(i) = '--interpolate' then
        Interpolate := True
      else if ParamStr(i) = '--np' then begin
        np := StrToInt(ParamStr(i+1));
        Inc(i);
      end else if ParamStr(i) = '--check' then begin
        Check := ParamStr(i+1);
        Inc(i)
      end else if ParamStr(i) = '--checkfile' then begin
        CheckFile := ParamStr(i+1);
        Inc(i)
      end else if ParamStr(i) = '--pgopherversion' then begin
        InputPgopherVersion := DecodePgopherVersion(ParamStr(i+1));
        Inc(i);
      end else if ParamStr(i) = '--cvttopert' then
        CvtToPert := True
      else if ParamStr(i) = '--skipelowercheck' then
        SkipELowerCheck := True
      else if ParamStr(i) = '--doubleablevels' then
        DoubleABlevels := True
      else if ParamStr(i) = '--QNOmask' then begin
        QNOmask := StrToInt(ParamStr(i+1));
        Inc(i);
      end else if ParamStr(i) = '--exprcheck' then
        ExprChecking := True
      else if ParamStr(i) = '--overwrite' then
        NewPgoOpt := npgOver
      else if ParamStr(i) = '--newpgo' then
        NewPgoOpt := npgNew
      else if ParamStr(i) = '--AllCALPGMpert' then
        AllCALPGMpert := True
      else if ParamStr(i) = '--closed' then
        DefaultClosed := True
      else if ParamStr(i) = '--CO06' then
        CODATAVersion := 2006 { Need to scale temperature to account for difference in fundamental constants }
      else if ParamStr(i) = '--CO10' then
        CODATAVersion := 2010 { Need to scale temperature to account for difference in fundamental constants }
      else if ParamStr(i) = '--overlay' then begin
        SetLength(IsOverlay, d+1);
        Params[d] := ParamStr(i+1);
        IsOverlay[d] := True;
        Inc(i);
        Inc(d);
      end else if ParamStr(i) = '--trackmemory' then begin
        SetTrackMemory(True);
        {$IFDEF CONSOLE}
          ShowTimingAtEnd := True;
        {$ENDIF}
      end else if ParamStr(i) = '--times' then
        {$IFDEF CONSOLE}
          ShowTimingAtEnd := True
        {$ENDIF}
      else if ParamStr(i) = '--script' then begin
        if not Assigned(Script) then
          Script := TStringList.Create;
        ScriptFile := TStringList.Create;
        try
          ScriptFile.LoadFromFile(ParamStr(i+1));
          Script.AddStrings(ScriptFile);
        finally
          ScriptFile.Free;
        end;
        Inc(i);
      end else if ParamStr(i) = '--exec' then begin
        if not Assigned(Script) then
          Script := TStringList.Create;
        Inc(i);
        Script.Add(ParamStr(i));
      end else if ParamStr(i) = '--template' then begin
        SetLength(Templates, Length(Templates)+1);
        SetLength(TemplateFiles, Length(Templates));
        Templates[High(Templates)] := ParamStr(i+1);
        FName := ParamStr(i+2);
        if (FName <> '') and FileExists(FName) then
          TemplateFiles[High(Templates)] := ExpandFileName(FName)
        else begin
          WriteLn(ErrOutput, 'Template file ' + FName + ' not found');
          SetLength(Templates, Length(Templates)-1);
        end;
        Inc(i,2);
      end else if ParamStr(i) = '--nocalc' then
        NoCalc := True
      else
        Inc(d);
      Inc(i);
    end;
    nParams := d;
    SetLength(Params, nParams); { These are only filenames and actions required }
    SetLength(IsOverlay, nParams+1);
    ConnectLog(OutFile);
    CDir := GetCurrentDir;
    { Partial fix for change in fundamental constants }
    case CODATAVersion of
      2006: begin
              // Was GlobalMixture.Simulations[0].Temperature := GlobalMixture.Simulations[0].Temperature * 20836.644 / KelvinToMHz;
              SetUpUnitFactors(6.62606896e-34, 1.602176487e-19, 6.02214179e23, 1.3806504e-23, 927.400915e-26, 0.52917720859e-10, 'CODATA 2006');
              BoltzmannConstants[cm1] := 0.6950356302826; //0.6950356; 0.69503563008641966
              // BoltzmannConstants[MHz] := 20836.644; //20836.644
            end;
      2010: SetUpUnitFactors(6.62606957e-34, 1.602176565e-19, 6.02214129e23, 1.3806488e-23, 927.400968e-26, 0.52917721092e-10, 'CODATA 2010');
    end;

    if (nParams > 1) and (Copy(Params[0],1,2) = '--') then begin
      { Automated testing or specific command line actions }
      StartWorkers(np,TPgopherWorker);
      if Params[0] <> '--import' then begin
        {$IFNDEF CONSOLE}
          {$IFNDEF FPC}
            LogForm.Show;
          {$ELSE}
            { Must defer showing the log until the main form is visible }
            MainForm.ShowLogForm := True;
          {$ENDIF}
          MainForm.EmptyMixtureItemClick(nil);
          if not FileExists(Params[1]) then
            raise EReportException.Create(Params[1]+' does not exisit');
          LoadGeneralFile(Params[1],IsOverlay[1],True,cm1);
          if nParams > 2 then begin
            DoDirSeparators(Params[nParams-1]);
            LogForm.ObsFileName.Text := Params[nParams-1];
          end;
          SetCurrentDir(CDir); { As Mixture load can change current directory }
        {$ELSE}
          case FileType(Params[1],True, Reader) of
            fiParameters: if not Assigned(Reader) then begin
                            GlobalMixture := TMixture.Create(nil);
                            GlobalMixture.LoadFromFile(Params[1]);
                          end else
                            GlobalMixture := ImportMixture(Reader, Params[1]);
            fiOverlay:    AddToOverlay(GlobalOverlays, Params[1], True, cm1, False, Arbitrary);
            fiObservations: { ??? }
          end;
        {$ENDIF}
      end;
      ApplyChanges;
      if GlobalMixture <> nil then begin
        if CvtToPert then
          GlobalMixture.ConvertToPerturbations;
        Problem :=  GlobalMixture.FixupProperties;
        if Assigned(Problem) then begin
          WriteLn(Log, 'Load Exception: ', Problem.Message);
          Problem.Free; { Could well have same problem later }
        end;
      end;
      try
        if Copy(Params[0],1,4) = '--H(' then
          ExportHamiltonian(Copy(Params[0],4,Length(Params[0])))
        else if Copy(Params[0],1,4) = '--T(' then
          ExportTransitionMatrix(Copy(Params[0],4,Length(Params[0])))
        else if Params[0] = '--fit' then
          FitHelper(obLine)
        else if Params[0] = '--intensityfit' then
          FitHelper(obIntensity)
        else if Params[0] = '--fitlower' then
          FitHelper(obELower)
        else if Params[0] = '--fitupper' then
          FitHelper(obEUpper)
        else if Params[0] = '--contourfit' then begin
          AddOverlay(Params[2], True, cm1);
          FitHelper(obContour);
        end else if Params[0] = '--fitcdlower' then
          FitHelper(obCDLower)
        else if Params[0] = '--fitcdupper' then
          FitHelper(obCDUpper)
        else if Params[0] = '--overlaysumfit' then begin
          AddOverlay(Params[2], True, cm1);
          FitHelper(obExpression) { obExpression is just used to flag overlay sum fit }
        end else if Params[0] = '--autofit' then begin
          AddOverlay(Params[2], True, cm1);
          DoAutoFit(GlobalMixture);
          {$IFDEF CONSOLE}
            WriteNewPgoFile(NewPgoOpt);
          {$ENDIF}
        end else if Params[0] = '--plot' then begin
          for i := 2 to nParams-2 do
            AddOverlay(Params[i], True, cm1);
          MakePlotFile(Params[d-1],False);
        end else if Params[0] = '--eplot' then
          MakePlotFile(Params[2],True)
        else if Params[0] = '--import' then begin
          Stream := TFreeFormatStream.Create(Params[2], True);
          if LowerCase(Params[1]) = 'lmatrix' then
            LoadedMixture := ImportLmatrixFile(Stream as TFreeFormatStream)
          else if LowerCase(Params[1]) = 'hitran' then
            LoadedMixture := ImportHITRANFile(Stream as TFreeFormatStream)
          else if LowerCase(Params[1]) = 'exomol' then begin
            LoadedMixture := ImportExoMolFile(Stream as TFreeFormatStream);
            for i := 3 to High(Params) do
              (LoadedMixture.Species[0].Isotopomer[0] as TExoMolMolecule).LoadTransFile(Params[i]);
          end else
            raise EReportException.Create('Unknown import:'+Params[1]);
          ApplyChanges;
          ApplyScript(LoadedMixture, Script);
          {$IFDEF CONSOLE}
          GlobalMixture :=
          {$ELSE}
          MainForm.Mixture :=
          {$ENDIF}
            LoadedMixture;
        end
        {$IFDEF SKELETON}
        else if Params[0] = '--misc' then
          SkeletonFunction(Params[2])
        {$ENDIF}
        else begin
          DoDirSeparators(Params[2]);
          TStart(LoadFileTime);
          Stream := TTextStream.Create(Params[2], True);
          TEnd(LoadFileTime);
          Stream.Comments := True;
          StartTime := Clocks();
          if ExtractFilePath(Params[2]) <> '' then
            SetCurrentDir(ExtractFilePath(Params[2]));
          PushProgress('Checking',Stream.Size);
          try
            if Params[0] = '--testd3h' then
              LargestError := CheckD3hFile(GlobalMixture, Stream, nTried, Tolerance, nFailures)
            else if Params[0] = '--testold' then
              LargestError := CheckOldPgopherFile(GlobalMixture, Stream, nTried, Tolerance, nFailures, False, False, False)
            else if Params[0] = '--testoldnoom' then
              LargestError := CheckOldPgopherFile(GlobalMixture, Stream, nTried, Tolerance, nFailures, True, False, False)
            else if Params[0] = '--testoldnof1' then
              LargestError := CheckOldPgopherFile(GlobalMixture, Stream, nTried, Tolerance, nFailures, False, True, False)
            else if Params[0] = '--testoldnoomega' then
              LargestError := CheckOldPgopherFile(GlobalMixture, Stream, nTried, Tolerance, nFailures, False, False, True)
            else if Params[0] = '--testpgo' then
              LargestError := CheckPgopherFile(False, GlobalMixture, Stream, nTried, Tolerance, nFailures, InputPgopherVersion, False, LineListMode)
            else if (Params[0] = '--testplot') or (Params[0] = '--testeplot') then begin
              SetCurrentDir(CDir);
              for i := 3 to nParams-1 do
                AddOverlay(Params[i], True, cm1);
              ApplyChanges;
              try
                CMStream := TMemoryStream.Create;
                CMStream.Size := Stream.Size; { Pre-allocate memory }
                MakePlotStream(Params[0] = '--testeplot',CMStream,TNullCanvas);
                CMStream.Size := CMStream.Position;
                LargestError := CheckPlotStream(GlobalMixture, Stream, CMStream, nTried, Tolerance, nFailures)
              finally
                CMStream.Free;
              end;
            end else if Params[0] = '--testpseudo' then
              LargestError := CheckPgopherFile(False, GlobalMixture, Stream, nTried, Tolerance, nFailures, InputPgopherVersion, True, LineListMode)
            else if Params[0] = '--testgen' then
              LargestError := CheckPgopherFile(True, GlobalMixture, Stream, nTried, Tolerance, nFailures, InputPgopherVersion, False, LineListMode)
            else if Params[0] = '--testeig' then
              LargestError := CheckEigenvalueFile(GlobalMixture, Stream, nTried, Tolerance, nFailures, InputPgopherVersion, LineListMode)
            else if Params[0] = '--testovr' then
              LargestError := CheckOverlays(GlobalOverlays, Stream, nTried, Tolerance, nFailures, SortBoth, Interpolate, ExtractFileExt(Params[2]))
            else if Params[0] = '--testexpr' then
              LargestError := CheckExprs(Stream, nTried, Tolerance, nFailures)
            else if Params[0] = '--testalg' then
              LargestError := CheckAlg(Stream, nTried, Tolerance, nFailures)
            else if Params[0] = '--testjplcat' then
              LargestError := CheckJPLCatFile(GlobalMixture, Stream, nTried, Tolerance, nFailures, Offset, SkipELowerCheck, DoubleABlevels, QNOmask)
            else if Params[0] = '--testspecview' then
              LargestError := CheckSpecViewFile(GlobalMixture, Stream, nTried, Tolerance, nFailures, False)
            else if Params[0] = '--testspecviewghz' then
              LargestError := CheckSpecViewFile(GlobalMixture, Stream, nTried, Tolerance, nFailures, True)
            else if Params[0] = '--testjb95' then
              LargestError := CheckJB95ViewFile(GlobalMixture, Stream, nTried, Tolerance, nFailures)
            else if Params[0] = '--testlimport' then
              LargestError := CheckImported(ImportLmatrixFile(Stream), GlobalMixture, nTried, Tolerance, nFailures, Script)
            else if Params[0] = '--testhitranimport' then
              LargestError := CheckImported(ImportHITRANFile(Stream), GlobalMixture, nTried, Tolerance, nFailures, Script)
            else if Params[0] = '--testparimport' then
              LargestError := CheckImported(ImportCALPGMFile(Stream,True,{$IFDEF CONSOLE}True{$ELSE}False{$ENDIF}), GlobalMixture, nTried, Tolerance, nFailures, Script)
            else if Params[0] = '--testparimportclosed' then
              LargestError := CheckImported(ImportCALPGMFile(Stream,False,{$IFDEF CONSOLE}True{$ELSE}False{$ENDIF}), GlobalMixture, nTried, Tolerance, nFailures, Script)
            else if Params[0] = '--testpgoimport' then begin
              LoadedMixture := TMixture.Create(nil);
              LoadedMixture.LoadFromStream(Stream,Params[2]);
              LargestError := CheckImported(LoadedMixture, GlobalMixture, nTried, Tolerance, nFailures, Script)
            end else if Params[0] = '--checkcalcpar' then begin
              ApplyScript(GlobalMixture, Script);
              LargestError := CheckCalculatedParameters(GlobalMixture, Stream, nTried, Tolerance, nFailures, LineListMode)
            end else if Params[0] = '--testautofit' then begin
              SetCurrentDir(CDir);
              AddOverlay(Params[2], True, cm1);
              LoadedMixture := TMixture.Create(nil);
              LoadedMixture.LoadFromFile(Params[3]);
              {$IFDEF CONSOLE}
                if OutFile = '' then
                  DivertCheckLog(Params[1]);
                { TODO: Check this output also }
              {$ENDIF}
              DoAutoFit(GlobalMixture);
              {$IFDEF CONSOLE}
                if OutFile = '' then
                  CancelDivert;
              {$ENDIF}
              LargestError := CheckImported(LoadedMixture, GlobalMixture, nTried, Tolerance, nFailures, Script)
            end else
              raise Exception.Create('Bad test:'+Params[0]);
          finally
            PopProgress;
          end;
          Inc(nFailures, nExprFailures);
          if nFailures <> 0 then begin
            Write(Log, nFailures, ' failure(s) out of ', nTried);
            if nExprFailures > 0 then
              Write(Log, ' of which ', nExprFailures, ' expression failures ');
            ExitCode := 1;
          end else
            Write(Log, 'OK (', nTried:5, ')');
          Write(Log, ' Largest Error = ', LargestError:20, (Clocks()-StartTime)/ClockFrequency:10:3, ' s  :');
          for i := 1 to ParamCount do
            Write(Log, ' ', ParamStr(i));
          WriteLn(Log);
        end;
      finally
        FreeAndNil(Stream);
      end;
      {$IFDEF CONSOLE}
      FreeAndNil(GlobalMixture);
      FreeAndNil(GlobalOverlays);
      DisconnectLogs;
      {$ENDIF}
    end else begin
      { Simple Calculation }
      StartWorkers(np,TPgopherWorker);
      {$IFNDEF CONSOLE}
        if GlobalMixture = nil then
          MainForm.EmptyMixtureItemClick(nil);
        for i := 0 to High(Params) do
          try
            LoadGeneralFile(Params[i],i>0,IsOverlay[i],MainForm.Simulation.PlotUnits);
            SetCurrentDir(CDir); { As Mixture load can change current directory }
          except
            on E:Exception do begin
              Application.MessageBox(PChar('Error on loading '+Params[i]+':'+sLineBreak+E.Message), 'Error', MB_OK+MB_ICONERROR);
              Break;
            end;
          end;
        ApplyChanges;
        if GlobalMixture <> nil then
          ApplyScript(GlobalMixture, Script);
        MainForm.ConsiderReplot(GlobalMixture.nSpecies = 0);
      {$ELSE}
        if d = 0 then begin
          WriteLn(Log, Global.VersionInfo);
          WriteLn(Log, DetermineCpu(False));
        end else if (d = 1) and (FileType(Params[0],False,Reader) = fiParameters) then begin
          if {$IFNDEF FPC}@{$ENDIF}Reader = nil then begin
            GlobalMixture := TMixture.Create(nil);
            GlobalMixture.LoadFromFile(Params[0]);
          end else
            GlobalMixture := ImportMixture(Reader, Params[0]);
        end else
          raise EReportException.Create('Arguments not recognized');
        if GlobalMixture <> nil then begin
          ApplyChanges;
          ApplyScript(GlobalMixture, Script);
          if not LineListWanted(GlobalMixture.PrintLevel) and not NoCalc then
            GlobalMixture.PrintLevel := plLineList;
          if GlobalMixture.PrintLevel >= plDetail then
            GlobalMixture.PrettyPrint(Log, GlobalMixture.PrintLevel);
          Problem := nil;
          if not NoCalc then
            GlobalMixture.Simulations[0].MakeLineList(Problem);
          if Assigned(Problem) then begin
            WriteLn(Log, Problem.Message);
            Problem.Free;
          end;
          if PGOoutput <> '' then
            GlobalMixture.SaveToFile(PGOoutput);
        end;
        Flush(Log);
        FreeAndNil(GlobalMixture);
        DisconnectLogs;
      {$ENDIF}
    end;
  except
    on E:Exception do begin
      if not LogConnected then
        ConnectLog('');
      ConsiderException(Log, E);
      WriteLn(Log, '**** failure Exception:', E.message, ':', CmdLine);
      Flush(Log);
      {$IFNDEF CONSOLE}
        Application.MessageBox(PChar('Error while processing parameters: '+E.Message), 'Error', MB_OK+MB_ICONERROR);
      {$ENDIF}
    end;
  end;
  Script.Free;
end;

initialization
finalization
  Global.VersionInfo := '';
end.
