unit Constrain;
{$I utils.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses Classes, Mixture, Obs, Expr, MNode, TypInfo, pgDOM, Overlay, MNodeExpr;

type

  { TMNodeVariable }

  TMNodeVariable = class (TFunction)
  public
    viInfo: TValueInfo;
    constructor Create(Text: string; AMNode: TMNode; Index: Integer); overload;
    constructor Create(Text: string; AviInfo: TValueInfo); overload;
    constructor Create(Text: string; const SearchPath: array of TMNode); overload;
    function Value(var a:array of Double):Double; override;
    procedure SetValue(AValue: Double);
    function SameVariable(AMNode: TMNode; AIndex: Integer): Boolean;
  end;

  { TExpression }

  TExpression = class (TMNodeExpr)
  protected
    FText: string;
    SearchPath: array [0..8] of TMNode;
    procedure SetSearchPath(Reader: TObservationsReader; Mixture: TMixture; Overlays: TOverlays);
  public
    procedure SetText(Value: string; Reader: TObservationsReader; Mixture: TMixture; Overlays: TOverlays);
    property Text: string read FText;
  end;

  { TConstraint }

  TConstraint = class (TExpression)
  private
    LHS: TMNodeVariable;
    FullText: string;
  protected
    function MakeLex: TLex; override;
  public
    procedure Apply;
    procedure SetText(Value: string; Reader: TObservationsReader; Mixture: TMixture; Overlays: TOverlays);
    procedure Write(var F: Text);
    property Text: string read FullText;
  end;

procedure ApplyConstraints;
procedure CheckConstraints(var TheParameters: TParameters);
procedure ClearConstraintsAndExpressions;
function AddConstraint(Text: string; Reader: TObservationsReader; Mixture: TMixture; Overlays: TOverlays): TConstraint;
procedure WriteConstraints(var F: Text);

function AddExpression(Text: string; Reader: TObservationsReader; Mixture: TMixture; Overlays: TOverlays): TExpression;

function ReadArbitraryParameters(MNode: TMNode; Node: TDOMNode): Integer;

implementation

uses SysUtils, Contnrs, PropHelp, UCompat;

{ Constraints on parameters; Global Variable for now,
  and only applies to Mixture it was set up for; could
  make it a variable in TMixture }

var
  Constraints, Expressions: TComponentList;

procedure ClearConstraintsAndExpressions;
begin
  FreeAndNil(Constraints);
  FreeAndNil(Expressions);
end;

procedure ApplyConstraints;
var
  I: Integer;
begin
  if Assigned(Constraints) then
    for I := 0 to Constraints.Count - 1 do
      (Constraints[i] as TConstraint).Apply;
end;

procedure CheckConstraints(var TheParameters: TParameters);
var
  I, J, K: Integer;
begin
  if not Assigned(Constraints) then
    Exit;
  { If a constraint applied to a parameter, remove from list of floated variables }
  for I := 0 to Constraints.Count - 1 do
    with (Constraints[i] as TConstraint) do
      for J := High(TheParameters) downto 0 do
        if LHS.SameVariable(TheParameters[J].MNode, TheParameters[J].Index) then begin
          for K := J to High(TheParameters)-1 do
            TheParameters[K] := TheParameters[K+1];
          SetLength(TheParameters, Length(TheParameters)-1);
        end;
end;

function AddConstraint(Text: string; Reader: TObservationsReader; Mixture: TMixture; Overlays: TOverlays): TConstraint;
begin
  Result := TConstraint.Create(nil);
  try
    Result.SetText(Text,Reader,Mixture,Overlays);
    Result.Apply; { Test it out }
    if not Assigned(Constraints) then
      Constraints := TComponentList.Create; { Default is owns objects }
    Constraints.Add(Result);
  except
    Result.Free;
    raise;
  end;
end;

procedure WriteConstraints(var F: Text);
var
  I: Integer;
begin
  if not Assigned(Constraints) or (Constraints.Count = 0) then
    Exit;
  WriteLn(F,'Constraints:');
  for I := 0 to Constraints.Count - 1 do
    { TODO: (Constraints[I] as TConstraint).Write(F) }
    WriteLn(F, (Constraints[I] as TConstraint).Text);
end;

{ TExpression }

procedure TExpression.SetSearchPath(Reader: TObservationsReader; Mixture: TMixture; Overlays: TOverlays);
begin
  SearchPath[0] := Overlays;
  SearchPath[1] := Mixture;
  SearchPath[2] := Mixture.Simulations[0];
  if Mixture.nSpecies = 1 then
    SearchPath[3] := Mixture.Species[0]
  else
    SearchPath[3] := nil;
  SearchPath[4] := Reader.DefaultMolecule;
  SearchPath[5] := Reader.DefaultManifoldU;
  SearchPath[6] := Reader.DefaultManifoldL;
  SearchPath[7] := Reader.DefaultStateU;
  SearchPath[8] := Reader.DefaultStateL;
end;

procedure TExpression.SetText(Value: string; Reader: TObservationsReader; Mixture: TMixture; Overlays: TOverlays);
var
  E: Exception;
begin
  FText := Value;
  SetSearchPath(Reader, Mixture, Overlays);
  E := ParseString(Value, FExpr, SearchPath, False);
  if Assigned(E) then
    raise E;
end;

function AddExpression(Text: string; Reader: TObservationsReader; Mixture: TMixture; Overlays: TOverlays): TExpression;
begin
  Result := TExpression.Create(nil);
  try
    Result.SetText(Text,Reader,Mixture,Overlays);
    Result.Expr.Value; { Test it out }
    if not Assigned(Expressions) then
      Expressions := TComponentList.Create; { Default is owns objects }
    Expressions.Add(Result);
  except
    Result.Free;
    raise;
  end;
end;

{ TConstraint }

function TConstraint.MakeLex: TLex;
begin
  Result := inherited MakeLex;
  { Note that constraints have different identifier rules - may need spaces to force parsing of comparison operators }
  Result.IdentifierChars := Result.IdentifierChars + ['.','=','>','<','|'];
  Result.AddFunction(LHS,True);
end;

procedure TConstraint.Apply;
var
  Value: Double;
begin
  Value := Expr.Value;
  LHS.SetValue(Value);
end;

procedure TConstraint.SetText(Value: string; Reader: TObservationsReader; Mixture: TMixture; Overlays: TOverlays);
var
  eq: Integer;
  LHSname: string;
begin
  FullText := Value;
  ClearExpr;
  eq := Pos(':=', Value);
  { = on it's own is no good as v=0 is a default name! }
  if eq <= 1 then
    raise EReportException.Create('Constraint must contain :=');
  LHSname := Copy(Value,1,eq-1);
  LHSname := Trim(LHSname);
  SetSearchPath(Reader, Mixture, Overlays);
  LHS := TMNodeVariable.Create(LHSname, SearchPath);
  Delete(Value,1,eq+2);
  inherited SetText(Value, Reader, Mixture, Overlays);
end;

procedure TConstraint.Write(var F: Text);
var
  ExprText: string;
  ExprNode: TExprNode;
begin
  ExprText := FullText;
  ExprNode := nil;
  try
    try
      ExprNode := FExpr.ExprTree;
      ExprText := LHS.Name + ' := ' + ExprNode.Text(True);
    finally
      ExprNode.Free;
    end;
  except
    { Fall back to input text }
  end;
  Writeln(F, ExprText);
end;

{ TMNodeVariable }

constructor TMNodeVariable.Create(Text: string; AMNode: TMNode; Index: Integer);
begin
  inherited Create(Text, 0);
  if Index < 0 then
    viInfo.viType := viCalculatedParameter
  else
    viInfo.viType := viParameter;
  viInfo.MNode := AMNode;
  viInfo.Index := Abs(Index);
end;

constructor TMNodeVariable.Create(Text: string; AviInfo: TValueInfo);
begin
  inherited Create(Text, 0);
  viInfo := AviInfo;
end;

constructor TMNodeVariable.Create(Text: string; const SearchPath: array of TMNode);
begin
  inherited Create(Text, 0);
  viInfo := FindSubProp(Text, SearchPath, True);
  if not Assigned(viInfo.MNode) and (Length(Text) > 2) and (Text[1] in ['''','"'])
     and (Pos(Text[1], Copy(Text,2,Length(Text))) = Length(Text)-1) then
    viInfo := FindSubProp(Copy(Text, 2, Length(Text)-2), SearchPath, True);
  if not Assigned(viInfo.MNode) then
    raise EReportException.Create('Can''t find '+Text);
end;

function TMNodeVariable.SameVariable(AMNode: TMNode; AIndex: Integer): Boolean;
begin
  if AMNode = viInfo.MNode then
    Result := (viInfo.viType = viParameter)
              and (viInfo.Index = AIndex)
  else
    Result := False;
end;

procedure TMNodeVariable.SetValue(AValue: Double);
begin
  viInfo.SetValue(AValue);
end;

function TMNodeVariable.Value(var a: array of Double): Double;
begin
  Result := viInfo.GetValue;
end;

function ReadArbitraryParameters(MNode: TMNode; Node: TDOMNode): Integer;
var
  Child, NextChild: TDOMNode;
  Attr: TDOMAttr;
begin
  { All unread parameters are read as dynamic parameters }
  Result := 0;
  Child := Node.FirstChild;
  { Read (and create) dynamic parameters containing variables }
  while Assigned(Child) do begin
    NextChild := Child.NextSibling;
    if (Child is TDOMElement) and (CompareText(Child.NodeName, 'Parameter') = 0) then begin
      Attr := (Child as TDOMElement).GetAttributeNode('Name');
      Result := Result + MNode.ReadFloatableDOMNode(MNode.AddParameter(Attr.NodeValue),'0',Child as TDOMElement);
      Node.RemoveChild(Child);
    end;
    Child := NextChild;
  end;
end;

initialization
finalization
  ClearConstraintsAndExpressions;
end.

constructor TMNodeVariable.Create(Text: string; Reader: TObservationsReader; Mixture: TMixture; Overlays: TOverlays);
var
  Variables: TVariables;
  Ignored: string;
begin
  inherited Create(Text, 0);
  viInfo.MNode := nil;
  if Assigned(Overlays) then begin
    Variables := TVariables(Overlays.ChildWithClass(TVariables));
    if Assigned(Variables) then
      viInfo := Variables.SubPropInfo(Text, Ignored);
  end;
  if not viInfo.IsSingleValue then begin
    Variables := TVariables(Mixture.ChildWithClass(TVariables));
    if Assigned(Variables) then
      viInfo := Variables.SubPropInfo(Text, Ignored);
  end;
  if not viInfo.IsSingleValue then
    viInfo := Mixture.SubPropInfo(Text, Ignored);
  if not viInfo.IsSingleValue and (Mixture.nSpecies = 1) then
    viInfo := Mixture.Species[0].SubPropInfo(Text, Ignored);
  if not viInfo.IsSingleValue and Assigned(Reader.DefaultMolecule) then
    viInfo := Reader.DefaultMolecule.SubPropInfo(Text, Ignored);
  if not viInfo.IsSingleValue and Assigned(Reader.DefaultManifoldU) then
    viInfo := Reader.DefaultManifoldU.SubPropInfo(Text, Ignored);
  if not viInfo.IsSingleValue and Assigned(Reader.DefaultManifoldL) then
    viInfo := Reader.DefaultManifoldL.SubPropInfo(Text, Ignored);
  if not viInfo.IsSingleValue and (Reader.DefaultStateU <> nil) then
    viInfo := Reader.DefaultStateU.SubPropInfo(Text, Ignored);
  if not viInfo.IsSingleValue and (Reader.DefaultStateL <> nil) then
    viInfo := Reader.DefaultStateL.SubPropInfo(Text, Ignored);
  if not viInfo.IsSingleValue and Assigned(Overlays) then
    viInfo := Overlays.SubPropInfo(Text, Ignored);
  if not viInfo.IsSingleValue then
    raise Exception.Create('Can''t find '+Text);
end;

procedure TConstraint.SetText(Value: string; Reader: TObservationsReader; Mixture: TMixture; Overlays: TOverlays);
var
  Variables: TVariableList;
  i, eq: Integer;
  LHSname: string;
begin
  FuText := Value;
  ClearExpr;
  eq := Pos(':=', Value);
  { = on it's own is no good as v=0 is a default name! }
  if eq <= 1 then
    raise EReportException.Create('Constraint must contain :=');
  LHSname := Copy(Value,1,eq-1);
  LHSname := Trim(LHSname);
  LHS := TMNodeVariable.Create(LHSname, Reader, Mixture, Overlays);
  Delete(Value,1,eq+2);
  try
    FLex := TLex.Create;
    FExpr := TExpr.Create(FLex, True);
    FLex.IdentifierChars := FLex.IdentifierChars + ['.','=','>','<','|'];
    FLex.AddFunction(ConstantPi);
    FLex.AddFunction(LHS);
    Variables := FExpr.MakeVariables(Value);
    try
      for i := 0 to Variables.Count - 1 do begin
        FLex.AddFunction(TMNodeVariable.Create(Variables[i],Reader,Mixture,Overlays));
        FLex.RemoveVariable(Variables.Objects[i]);
      end;
    finally
      Variables.Free;
    end;
    FLex.ReturnUnknown := False;
    FExpr.Text := Value;
    with FLex.Functions do
      for i := 0 to Count - 1 do
        if Objects[i] is TMNodeVariable then
          with TMNodeVariable(Objects[i]) do
            Name := viInfo.MNode.FullHeirarchicName;
  except
    ClearExpr;
    raise;
  end;
end;


