unit Decomp;
{$I utils.inc}{Copyright (C) 2008 C M Western; see commandline.pas for license}
interface

uses MatrixVec;

function FindRotations(Matrix: TMatrix):TMatrix;
{ Given a Matrix transform to an equivalent set of rotations by least
  squares fitting }

implementation

uses SysUtils, NLFit, Worker, Math;

procedure Rotate(M: TMatrix; Angle: Double; r1, r2: Integer);
{ Postmultiply by matrix:
   (.... .........)
r1 (....c...s....)
   (.... .........)
r2 (....-s...c....)
   (.... .........)
  with r1 < r2, s = sin(Angle/Degrees)
}
var
  cs, sn: Extended;
  i: Integer;
  t: Double;
begin
  SinCos(Angle,sn,cs);
  for i := 1 to M.RowCount do begin
    { Only two elements of any given row change }
    t := M[i,r1]*cs - M[i,r2]*sn;
    M[i,r2] := M[i,r1]*sn + M[i,r2]*cs;
    M[i,r1] := t;
  end;
end;

function MakeFromRotations(Rotations: TMatrix):TMatrix;
var
  n, i, j: Integer;
begin
  n := Rotations.RowCount;
  Result := TMatrix.Create(n,n);
  Result.Zero;
  Result.AddDiagonal(1);
  for i := 1 to n do
    for j := 1 to i - 1 do
      Rotate(Result,Rotations[i,j]*Pi/180,i,j);
end;

function FindRotations(Matrix: TMatrix):TMatrix;
var
  a, ada, SVD: TMatrix;
  nPar, nObs, n, i, j, np, iteration: Integer;
  obs, par, calc, b, SVDw: TColumnVector;
  k: Integer;
  OldSD,NewSD,UnWtOldSD,UnWtNewSD: Double;
begin
  n := Matrix.RowCount;
  Result := TMatrix.Create(n,n);
  Result.Zero;
  if n <= 1 then
    Exit;
  nPar := (n*n - n) div 2; { One parameter for each element below the diagonal }
  nObs := n*n;
  par := TColumnVector.Create(nPar);
  b := TColumnVector.Create(nPar);
  SVDw := TColumnVector.Create(nPar);
  obs := nil;
  a := nil;
  ada := nil;
  SVD := nil;
  try
    { These can cause memory overflow }
    a := TMatrix.Create(nObs,nPar);
    ada := TMatrix.Create(nPar,nPar);
    SVD := TMatrix.Create(nPar,nPar);
    for iteration := 0 to n*n do begin
      WriteLn(Log,'Iteration '+IntToStr(iteration));
      obs.Free;
      obs := MakeFromRotations(Result);
      { For each parameter }
      np := 1;
      for i := 1 to n do
        for j := 1 to i - 1 do begin
          par[np] := Result[i,j];
          { Calculate derivative }
          if par[np] = 0 then
            Result[i,j] := 1
          else
            Result[i,j] := par[np]*1.0001;
          calc := MakeFromRotations(Result);
          // calc.Write(Log,'calc');
          for k := 1 to nObs do
            a[k,np] := (obs[k]-calc[k])/(par[np]-Result[i,j]);
          Result[i,j] := par[np];
          calc.Free;
          Inc(np);
        end;
      obs.Scale(-1);
      obs.Add(Matrix);
      // a.Write(Log,'a');
      obs.Write(Log,'Residuals');
      // par.Write(Log,'par');
      LeastSQ(a,obs,par,nil,OldSD,NewSD,UnWtOldSD,UnWtNewSD,b,ada,SVDw,SVD,0,False,1);
      WriteLn(Log, 'Sums: Old ', FloatToStr(OldSD), ', New ',FloatToStr(NewSD)); 
      // b.Write(Log,'b');
      // ada.Write(Log,'ada');
      { Copy back new parameters }
      np := 1;
      for i := 1 to n do
        for j := 1 to i - 1 do begin
          Result[i,j] := par[np];
          Inc(np);
        end;
      Result.Write(Log,'Decomp');
      if Abs(OldSD-NewSD) < 1e-10 then
        Exit;
    end;
  finally
    a.Free;
    obs.Free;
    par.Free;
    ada.Free;
    SVD.Free;
    SVDw.Free;
  end;
end;

end.
