unit Exporter;
{ Utility class for exporting values for use with other programs}

{$I utils.inc}{Copyright (C) 2015 C M Western; see commandline.pas for license}
interface

uses Classes, MathSup, MatrixVec, ComplexMatrix, MNode;

type
  PText = ^Text;

  TShower = record
    F: PText;
    Separator: Char;
    bbuf: string[20];
    FirstField: Boolean;
    procedure ShowField(const Field: AnsiString);
    procedure ShowEOL;
    procedure ShowBuffer;
    procedure ShowD(V: Double);
    procedure ShowComplex(V: Complex);
    procedure ShowInt(V: Integer);
    procedure Setup(var PF: Text; PrintLevel: TPrintLevel);
    procedure ShowTransitionMatrix(Title: string; Eigenvalues, Eigenvalues1: TComponent { Must be TEigenvalues } ; M: TMatrix; Offset: Integer; ShowBasis: Boolean);
  end;

procedure ExportHamiltonian(s: string);
procedure ExportTransitionMatrix(s: string);

implementation

uses SysUtils, UCompat, Worker, QuantumNumber, Mixture, Global;

{ TShower }

procedure TShower.ShowField(const Field: AnsiString);
begin
  if Separator = #0 then begin
    Write(F^,AnsiChar(Length(Field)));
    Write(F^,Field);
  end else begin
    if FirstField then
      FirstField := False
    else
      Write(F^,Separator);
    if Separator = ',' then
      if (Pos('"', Field) <> 0) or (Pos(',', Field) > 0) then
        Write(F^,AnsiQuotedStr(Field,'"'))
      else
        Write(F^,Field)
    else
      Write(F^,Field);
  end;
end;

procedure TShower.ShowEOL;
begin
  if Separator = #0 then
    ShowField(#10) { Must write end of line to avoid text file mis-handling }
  else
    WriteLn(F^);
  FirstField := True;
end;

procedure TShower.Setup(var PF: Text; PrintLevel: TPrintLevel);
begin
  F := @PF;
  case PrintLevel of
    plCSV: Separator := ',';
    plText: Separator := #9;
    plBinary: Separator := #0;
  else
    Separator := ' ';
  end;
  FirstField := True;
end;

procedure TShower.ShowBuffer;
begin
  SetLength(bbuf, (Ord(bbuf[1])-200) mod 16 + 1);
  Write(F^,bbuf);
  SetLength(bbuf,20); { Avoids range check errors }
end;

procedure TShower.ShowComplex(V: Complex);
var
  s: string;
begin
  if Separator = #0 then begin
    bbuf[1] := AnsiChar(SizeOf(Complex)+216);
    PComplex(@bbuf[2])^ := V;
    SetLength(bbuf,1+Sizeof(Complex)); { Avoids range check errors }
    Write(F^,bbuf);
  end else begin
    if V.re <> 0 then
      s := FloatToStr(V.re)
    else
      s := '';
    if V.im <> 0 then begin
      if (V.im > 0) and (s <> '') then
        s := s + '+';
      s := s + FloatToStr(V.im)+'i';
    end;
    if s = '' then
      s := '0';
    ShowField(s);
  end;
end;

procedure TShower.ShowD(V: Double);
begin
  if Separator = #0 then begin
    bbuf[1] := AnsiChar(SizeOf(Double)+216);
    PDouble(@bbuf[2])^ := V;
    ShowBuffer;
  end else
    ShowField(FloatToStr(V));
end;

procedure TShower.ShowInt(V: Integer);
begin
  if Separator = #0 then begin
    bbuf[1] := AnsiChar(Sizeof(Integer)+200);
    PInteger(@bbuf[2])^ := V;
    ShowBuffer;
  end else
    ShowField(IntToStr(V));
end;

procedure TShower.ShowTransitionMatrix(Title: string; Eigenvalues, Eigenvalues1: TComponent; M: TMatrix; Offset: Integer; ShowBasis: Boolean);
  function SLabel(i: Integer; Ev: TEigenValues): string;
  begin
    if ShowBasis then
      Result := Ev.Basis[i+Offset].BasisLabel
    else
      Result := Ev.DominantBasis[i+Offset].StateLabel(Ev.Manifold.Mixture.QuantumNumberFormat);
  end;
var
  Ev, Ev1: TEigenvalues;
  i, j, n: Integer;
begin
  Ev := Eigenvalues as TEigenvalues;
  Ev1 := Eigenvalues1 as TEigenvalues;
  ShowField(Title);
  if not Assigned(M) then begin
    ShowEOL;
    Exit;
  end;
  if IsComplexMatrix(M) then
    n := TComplexMatrix(M).ColCount
  else
    n := M.ColCount;
  for i := 1 to n do
    ShowField(SLabel(i,Ev1));
  ShowEOL;
  for i := 1 to M.RowCount do begin
    ShowField(SLabel(i,Ev));
    for j := 1 to n do
      if IsComplexMatrix(M) then
        ShowComplex(TComplexMatrix(M)[i,j])
      else
        ShowD(M[i,j]);
    ShowEOL;
  end;
end;

function FirstArg(var s: string): string;
var
  P: Integer;
begin
  P := Pos(',', s);
  if P <= 0 then begin
    Result := Trim(s);
    s := '';
  end else begin
    Result := Trim(Copy(s,1,P-1));
    s := Trim(Copy(s,P+1,Length(s)));
  end
end;

function FindEv(var s, Title: string; ExtraOK: Boolean): TEigenValues;
var
  Arg, ErrMsg: string;
  vi: TValueInfo;
  Manifold: TManifold;
  J: TQuantumNumber;
  Sym: Integer;
begin
  Result := nil;
  if GlobalMixture = nil then
    raise EReportException.Create('No Mixture');
  Arg := FirstArg(s);
  vi := GlobalMixture.SubPropInfo(Arg+'.Initial', ErrMsg);
  if Assigned(vi.MNode) and (vi.MNode is TManifold) then begin
    Manifold := vi.MNode as TManifold;
    Arg := FirstArg(s);
  end else if (GlobalMixture.nMolecules = 1) and (GlobalMixture.nSpecies = 1)
              and (GlobalMixture.Species[0].Isotopomer[0].nManifolds =1)
              and TryStrToQuantumNumber(Arg,J,GlobalMixture.QuantumNumberFormat)
              then
    Manifold := GlobalMixture.Species[0].Isotopomer[0].Manifold[0]
  else
    raise EReportException.Create('Manifold not recognized:'+Arg+' ('+ErrMsg+')');
  if not TryStrToQuantumNumber(Arg,J,GlobalMixture.QuantumNumberFormat) then
    raise EReportException.Create('Bad J:'+Arg);
  Sym := 0; { Default }
  Arg := FirstArg(s);
  if Arg <> '' then begin
    if not Manifold.Molecule.TryStrToAnySymName(J,Arg,Sym) then
      if ExtraOK then
        s := Arg+','+s { put back as Arg is Manifold or J of other state }
      else
        raise EReportException.Create('Bad Symmetry:'+Arg)
  end else if (s <> '') and not ExtraOK then
    raise EReportException.Create('Too many arguments, or arguments not recognized');
  Result := TEigenvalues.Create(Manifold, J, Sym, GlobalMixture.Simulations[0].ExField);
  Result.MakeBasis(Manifold.Mixture.BasisOrder);
  Title := Manifold.FullHeirarchicName+','+GlobalMixture.QuantumNumberToStr(J)+','+Manifold.Molecule.SymName(Sym)+')';
end;

procedure ExportHamiltonian(s: string);
var
  SS: TShower;
  Eigenvalues: TEigenvalues;
  Matrices: TComplexMatrices;
  i, Base: Integer;
  Title: string;
begin
  if (Length(s) < 3) or (s[1] <> '(') or (s[Length(s)] <> ')') then
    raise EReportException.Create('Argument not recognized:'+s);
  s := Copy(s,2,Length(s)-2); { Strip brackets }
  { --H(Manifold,J,Sym)
    Mainfold and Sym optional if otherwise unambiguous }
  Matrices := nil;
  EigenValues := FindEv(s, Title, False);
  SS.Setup(Log, GlobalMixture.PrintLevel);
  if SS.Separator = ' ' then
    SS.Separator := #9;
  Title := 'H('+Title;
  try
    EigenValues.MakeHamiltonian(Matrices, EigenValues.Manifold.Molecule.BlockMatrix);
    case Length(Matrices) of
      0: SS.ShowField(Title); { Empty matrix }
      1: SS.ShowTransitionMatrix(Title, Eigenvalues, Eigenvalues, Matrices[0], 0, True);
    else
      Base := 0;
      for i := 0 to High(Matrices) do begin
        SS.ShowTransitionMatrix(Title+'#'+IntToStr(i+1), Eigenvalues, Eigenvalues, Matrices[i], Base, True);
        Inc(Base, Matrices[i].RowCount);
      end;
    end;
  finally
    EigenValues.Free;
    for i := 0 to High(Matrices) do
      Matrices[i].Free;
  end;
end;

procedure ExportTransitionMatrix(s: string);
var
  Ev1, Ev2: TEigenvalues;
  Title, t1: string;
  TM: TTransitionMoments;
  M: TComplexMatrix;
  SS: TShower;
begin
  if (Length(s) < 5) or (s[1] <> '(') or (s[Length(s)] <> ')') then
    raise EReportException.Create('Argument not recognized:'+s);
  s := Copy(s,2,Length(s)-2); { Strip brackets }
  Ev1 := FindEv(s, Title, True);
  Ev2 := nil;
  try
    Ev2 := FindEv(s, t1, False);
    TM := Ev1.Manifold.Molecule.FindTransitionMoments(Ev1.Manifold, Ev2.Manifold);
    if (TM = nil) or (TM.nRanks = 0) then
      raise EReportException.Create('No Transition Moments');
    if TM.nRanks <> 1 then
      raise EReportException.Create('Multiple ranks not implemented');
    M := TM.OriginalTransitionMatrix(Ev1, Ev2, TM.TransitionMoment[0].Rank, nil, True, True, True, GlobalMixture.Simulations[0]);
    SS.Setup(Log, GlobalMixture.PrintLevel);
    if SS.Separator = ' ' then
      SS.Separator := #9;
    Title := 'T(' + Title + ',' + t1 + ')';
    SS.ShowTransitionMatrix(Title, Ev1, Ev2, M, 0, True);
    M.Free;
  finally
    Ev1.Free;
    Ev2.Free;
  end;
end;

end.
