unit ExternalList;

{ Use externally generated line list for plots }

interface
{$I utils.inc}{Copyright (C) 2008 C M Western; see commandline.pas for license}

uses {$IFNDEF CONSOLE} Menus, {$ENDIF}
     SysUtils, Classes, MNode, QuantumNumber, Mixture, DRintensity, pgDOM,
     QuantumMath, MathSup, StaticField;

type

  { TExternalEigenvalues }

  TExternalEigenvalues = class(TEigenvalues)
  private
    FEnergies: TDoubleArray;
    FStatisticalWeight: Integer; { TODO: Should probably be array for generality }
  protected
    Fg: array of Integer;
    FQuanta: array of string;
    function GetEnergy(AIndex: Integer): Double; override;
  public
    constructor Create(AManifold: TManifold; AJ: TQuantumNumber);
    procedure FixMem;
    function Degeneracy(AIndex: Integer): Integer; override;
    function Quanta(AIndex: Integer):string;
    function Qno(AIndex: Integer; QnoType: TQnoType):TQuantumNumber; override;
    function AddLevel(AEnergy: Double; g: Integer; const AQuanta: string; EDiff: Double = 0):Integer;

    function DominantLabel(AIndex: Integer; QuantumNumberFormat: TQuantumNumberFormat):string; override;
    function StateLabelKey(AIndex: Integer): string; override;
    function TrackIndex(AIndex: Integer):Integer; override;
    function JDeAdjustedSym(AIndex: Integer):Integer; override;
    function TxName(AIndex: Integer; OtherEv: TEigenvalues; OtherIndex: Integer):string; override;

    procedure SetStatisticalWeight(AValue: Integer);
    function StatisticalWeight(Index: Integer):Integer; override;
    function SpinSpecies(Index: Integer):Integer; override;
  end;

  TExternalManifold = class(TManifold)
  protected
    FDummyEigenvalues: array of TExternalEigenvalues;
    FHalfSpin: Integer;
    FnDummyEigenvalues: Integer;
  public
    destructor Destroy; override;
    procedure ClearEigenvalues;
    function FixupProperties:Exception; override;
    function AddLevel(Energy: Double; g: Integer; J: TQuantumNumber; const Quanta: string):Integer;
//    function Eigenvalues(J: TQuantumNumber; Sym: Integer; MatElemTypes: TMatElemTypes;
//                         Wait: Boolean = True;
//                         PrintLevel: Integer = -10; SelState: TState = nil; IncludePerturbations: Boolean = True;
//                         RemoveSpins: Byte = 0): TEigenvalues;
    function nSym: Integer; override; { Number of different symmetries }
    function Population(AEigenvalues: TEigenvalues; i: Integer; Simulation: TSimulation; var Overflow: Exception; IncludeTspin: Integer): Double; override;
    function HalfSpin(RemovedSpins: Byte = 0): TQuantumNumber; override;
    function FindJmin(Default: Integer; RemovedSpins: Byte = 0):Integer; override;
    function Eigenvalues(J: TQuantumNumber): TExternalEigenvalues; overload;
    function Eigenvalues(J: TQuantumNumber; Sym: Integer; MatElemTypes: TMatElemTypes;
                         const AExField: TStaticFields;
                         Wait: Boolean = True;
                         PrintLevel: TPrintLevel = plDefault; SelState: TState = nil; IncludePerturbations: Boolean = True;
                         RemoveSpins: Byte = 0; CreateIfNotFound: Boolean = True): TEigenvalues; overload; override;
  end;

  { TExternalMolecule }

  TExternalMolecule = class(TMolecule)
  public
    function LastSym: Integer; override;
    function JAdjustedSym(J: TQuantumNumber; Sym: Integer): Integer; override;
    function QnoInfo(QnoType: TQnoType): TQnoInfo; override;
  end;

  THITRANManifold = class(TExternalManifold)
  public
    procedure AddAllLevels(AState: TState); override;
    procedure ClearCache(ClearChildren: Boolean); override;
    function AddLevel(Energy: Double; g: Integer; J: TQuantumNumber; const Quanta: string; EDiff: Double = 0):Integer;
    function Eigenvalues(g: Integer; J: TQuantumNumber): TExternalEigenvalues; overload;
  end;

  TExternalEntry = record
    LineNumber: Integer; { Line number in file }
    Frequency: Double; { Transition frequency }
    Intensity: Double;
    EinsteinA: Double; { If available; stored as Spol }
    Width: Double;
    ELower:    Double; { Lower State energy (if available) }
    gL, gU: Integer;        { Upper and lower state degeneracy, if available }
    JL, JU: TQuantumNumber; { Upper and lower state J, if available }
    ManifoldL: THITRANManifold;
    ManifoldU: THITRANManifold;
    UpperQuanta, LowerQuanta: string;
    IndexU, IndexL: Integer; { Computed }
  end;

  THITRANMolecule = class(TExternalMolecule)
  private
    FnEntries: Integer;
    FDummyManifold: THITRANManifold;
  public
    Entries: array of TExternalEntry;
    constructor Create(AOwner: TComponent); override;
    function FixupProperties:Exception; override;
    function ReadProperties(Node: TDOMNode; CheckUnread: Boolean = True):Integer; override;
    function WriteProperties(Parent: TDOMNode; AllParameters: Boolean = True; WithChildren: Boolean = True): TDOMElement; override;
    function DummyManifold: THITRANManifold;

    function AddToSpectrum(Spectrum: TSpectrum; PrintLevel: TPrintLevel; const Excitation: TExcitation; var Problem: Exception;
                           Simulation: TSimulation;
                           J1Sel: Integer = Low(Integer); Sym1Sel: Integer = -1; J2Sel: Integer = Low(Integer); Sym2Sel: Integer = -1;
                           Manifold1Sel: TManifold = nil; Manifold2Sel: TManifold = nil;
                           DeltaJ: TQuantumNumber = -10000;
                           TransitionMomentsSel: TTransitionMoments = nil; TransitionMomentSel: TTransitionMoment = nil;
                           State1Sel: TState = nil; State2Sel: TState = nil;
                           IncludePerturbations: Boolean = True; IncludeZero: Boolean = False; RemoveSpins: Byte = 0;
                           IgnoreInitial: Boolean = False):Boolean; override;
    function HelpOnParameter(AIndex: Integer): string; override;

    {$IFNDEF CONSOLE}
    procedure AddHandlers; override;
    procedure LoadFromFileClick(Sender: TObject);
    {$ENDIF}

    procedure AddEntry(const Entry: TExternalEntry);
    function FindLine(var ALine: TLine; const TxDetails: string): Integer;
    property nEntries: Integer read FnEntries;
  published
    property WScale: Double index 3 read GetParameter write SetParameter;
    property EDiff:  Double index 4 read GetParameter write SetParameter;
    property IScale: Double index 5 read GetParameter write SetParameter;
    property gDiv:  Integer index LMI read GetSetting write SetSetting;
  end;

implementation

uses {$IFNDEF CONSOLE} Forms, Dialogs, IPlotLoader, {$ENDIF}
      Units, Overlay, Worker, Hitran, ExpPlotLoader, PartFunc,
      Graphics, Math, TabbedLine, UCompat;

var
  EmptyEigenvalues: TEigenvalues;

{ TExternalMolecule }

function TExternalMolecule.JAdjustedSym(J: TQuantumNumber; Sym: Integer): Integer;
begin
  Result := Sym;
end;

function TExternalMolecule.LastSym: Integer;
begin
  Result := 0;
end;

function TExternalMolecule.QnoInfo(QnoType: TQnoType): TQnoInfo;
begin
  Result := inherited QnoInfo(QnoType);
  case QnoType of
    qtA: Result.Name := '#';
  end;
end;

{ TExternalEigenvalues }

function TExternalEigenvalues.Qno(AIndex: Integer; QnoType: TQnoType): TQuantumNumber;
begin
  case QnoType of
    qtA: Result := AIndex;
    qtB: Result := 0;
  else
    Result := J;
  end;
end;

function TExternalEigenvalues.Quanta(AIndex: Integer): string;
begin
  Result := FQuanta[AIndex-1];
end;

procedure TExternalEigenvalues.SetStatisticalWeight(AValue: Integer);
begin
  FStatisticalWeight := AValue;
end;

function TExternalEigenvalues.StateLabelKey(AIndex: Integer): string;
begin
  Result := '';
end;

function TExternalEigenvalues.StatisticalWeight(Index: Integer): Integer;
begin
  Result := FStatisticalWeight;
end;

function TExternalEigenvalues.TrackIndex(AIndex: Integer): Integer;
begin
  Result := 0;
end;

function TExternalEigenvalues.TxName(AIndex: Integer; OtherEv: TEigenvalues; OtherIndex: Integer): string;
begin
  Result := '';
end;

function TExternalEigenvalues.SpinSpecies(Index: Integer): Integer;
begin
  Result := 1;
end;

function TExternalEigenvalues.AddLevel(AEnergy: Double; g: Integer; const AQuanta: string; EDiff: Double): Integer;
var
  i: Integer;
begin
  if EDiff > 0 then
    for i := 0 to FSize - 1 do
      if (AQuanta = FQuanta[i]) or (AQuanta = '') then
        if (Abs(FEnergies[i]-AEnergy)<EDiff) and (Fg[i] = g) then begin
          Result := i+1;
          Exit;
        end else if AQuanta <> '' then
          WriteLn(Log, 'Warning: Quanta equal, but energy or degeneracy mismatch in state ', Manifold.UniqueName, ' #', i:3, ' Quanta:', AQuanta, ' ', FEnergies[i]:15:6, ' ', AEnergy:15:6, ' ', Fg[i]:3, ' ', g:3, ' Difference:', Abs(FEnergies[i]-AEnergy):15:6);
  Inc(FSize);
  if FSize > Length(FEnergies) then begin
    if FSize < 10 then
      i := 10
    else
      i := 100;
    SetLength(FEnergies, FSize+i);
    SetLength(Fg, Length(FEnergies));
    SetLength(FQuanta, Length(FEnergies));
  end;
  Result := FSize;
  FEnergies[Result-1] := AEnergy;
  Fg[Result-1] := g;
  FQuanta[Result-1] := AQuanta;
end;

constructor TExternalEigenvalues.Create(AManifold: TManifold; AJ: TQuantumNumber);
begin
  inherited Create(AManifold, AJ, 0, ZeroField);
  FSize := 0;
end;

function TExternalEigenvalues.Degeneracy(AIndex: Integer): Integer;
begin
  Result := Fg[AIndex-1];
end;

function TExternalEigenvalues.DominantLabel(AIndex: Integer; QuantumNumberFormat: TQuantumNumberFormat): string;
begin
  Result := Manifold.UniqueName + ' ' + FQuanta[AIndex-1] + ' ' + QuantumNumberToStr(J, QuantumNumberFormat);
end;

procedure TExternalEigenvalues.FixMem;
begin
  SetLength(FEnergies, FSize);
  SetLength(Fg, Length(FEnergies));
  SetLength(FQuanta, Length(FEnergies));
end;

function TExternalEigenvalues.GetEnergy(AIndex: Integer): Double;
begin
  Result := FEnergies[AIndex-1];
end;

function TExternalEigenvalues.JDeAdjustedSym(AIndex: Integer): Integer;
begin
  Result := 0;
end;

{ TExternalManifold }

function TExternalManifold.AddLevel(Energy: Double; g: Integer; J:TQuantumNumber; const Quanta: string): Integer;
begin
  Result := Eigenvalues(J).AddLevel(Energy, g, Quanta);
end;

procedure TExternalManifold.ClearEigenvalues;
var
  i: Integer;
begin
  for i := 0 to High(FDummyEigenvalues) do
    FreeAndNil(FDummyEigenvalues[i]);
  FnDummyEigenvalues := 0;
end;

destructor TExternalManifold.Destroy;
begin
  ClearEigenvalues;
  inherited;
end;

function TExternalManifold.Eigenvalues(J: TQuantumNumber): TExternalEigenvalues;
var
  Index: Integer;
begin
  Index := ((J or 1)+1) div 2;
  if Index > High(FDummyEigenvalues) then
    SetLength(FDummyEigenvalues,Index+20);
  if not Assigned(FDummyEigenvalues[Index]) then begin
    if FnDummyEigenvalues = 0 then
      SetCachedInformationPresent;
    FDummyEigenvalues[Index] := TExternalEigenvalues.Create(Self, J);
    Inc(FnDummyEigenvalues);
  end;
  Result := FDummyEigenvalues[Index];
end;

function TExternalManifold.Eigenvalues(J: TQuantumNumber; Sym: Integer;
  MatElemTypes: TMatElemTypes; const AExField: TStaticFields;
  Wait: Boolean; PrintLevel: TPrintLevel;
  SelState: TState; IncludePerturbations: Boolean;
  RemoveSpins: Byte; CreateIfNotFound: Boolean): TEigenvalues;
var
  Index: Integer;
begin
  Index := ((J or 1)+1) div 2;
  Result := nil;
  if Index <= High(FDummyEigenvalues) then
    Result := FDummyEigenvalues[Index];
  if not Assigned(Result) and CreateIfNotFound then
    Result := EmptyEigenvalues;
end;

function TExternalManifold.FindJmin(Default: Integer; RemovedSpins: Byte): Integer;
begin
  if (Length(FDummyEigenvalues) = 0) or Assigned(FDummyEigenvalues[0]) then
    Result := -1
  else
    Result := 0;
end;

function TExternalManifold.FixupProperties: Exception;
var
  i: Integer;
begin
  for i := 0 to High(FDummyEigenvalues) do
    if Assigned(FDummyEigenvalues[i]) then
      FDummyEigenvalues[i].FixMem;
  Result := inherited FixupProperties;
end;

function TExternalManifold.nSym: Integer;
begin
  Result := 1;
end;

function TExternalManifold.HalfSpin(RemovedSpins: Byte): TQuantumNumber;
begin
  Result := FHalfSpin;
end;

function TExternalManifold.Population(AEigenvalues: TEigenvalues; i: Integer;
  Simulation: TSimulation; var Overflow: Exception; IncludeTspin: Integer): Double;
  function bfp(T, E: Double): Double;
  begin
    Result := bf(T,E);
    if Result < 0 then begin
      if not Assigned(Overflow) then
        Overflow := EReportException.CreateFmt('Overflow in Bolzmann for line %d: E=%g',[i, E]);
      Result := 0;
    end;
  end;
var
  PIndex: Integer;
begin
  if not (UsePopParams or (Simulation.Temperature <= 0)) then
    { Normal Boltzman calculation }
    Result := bfp(Simulation.Temperature, AEigenvalues.Energy[i])
  else begin
    { Custom population }
    PIndex := PopulationIndex(0,0,i);
    if PIndex > 0 then
      Result := Parameters[PIndex]
    else
      Result := 0;
  end;
  { TODO: May be able to handle Tspin > 0 }
//    if i >= 0 then begin
//      Result := 1;
//      if i <= High((Molecule as TExternalMolecule).Entries) then
//        with (Molecule as TExternalMolecule).Entries[i] do
//          if gL > 0 then { TODO: Handle -ve T => External population }
//            Result := bfp(Simulation.Temperature,ELower);
//    end else begin
//      Result := 0;
//      if -i-1 <= High((Molecule as TExternalMolecule).Entries) then
//        with (Molecule as TExternalMolecule).Entries[-i-1] do
//          if gU > 0 then { TODO: Handle -ve T => External population }
//            Result := bfp(Simulation.Temperature,Frequency+ELower);
//    end
//  else if i >= 0 then begin
//    { For custom populations only use lower state }
//    PIndex := PopulationIndex(0,0,i);
//    if PIndex > 0 then
//      Result := Parameters[PIndex]
//    else
//      Result := 0;
//  end else
//    Result := 0;
end;

{ THITRANMolecule }

procedure THITRANMolecule.AddEntry(const Entry: TExternalEntry);
var
  tm: TTransitionMoments;
begin
  if High(Entries) < nEntries then
    SetLength(Entries,nEntries+10000);
  Entries[nEntries] := Entry;
  with Entries[nEntries] do begin
    if not Assigned(ManifoldL) then
      ManifoldL := DummyManifold;
    if not Assigned(ManifoldU) then
      ManifoldU := DummyManifold;
    tm := FindTransitionMoments(ManifoldU, ManifoldL);
    if not Assigned(tm) then
      TTransitionMoments.Create(Self).SetBraKetName(ManifoldU.Name,ManifoldL.Name);
  end;
  Inc(FnEntries);
  if not Modified then
    Modified := True; { Don't call every time - much too slow }
end;

function THITRANMolecule.AddToSpectrum(Spectrum: TSpectrum;
  PrintLevel: TPrintLevel; const Excitation: TExcitation;
  var Problem: Exception; Simulation: TSimulation; J1Sel, Sym1Sel, J2Sel,
  Sym2Sel: Integer; Manifold1Sel, Manifold2Sel: TManifold;
  DeltaJ: TQuantumNumber; TransitionMomentsSel: TTransitionMoments;
  TransitionMomentSel: TTransitionMoment; State1Sel, State2Sel: TState;
  IncludePerturbations, IncludeZero: Boolean; RemoveSpins: Byte;
  IgnoreInitial: Boolean): Boolean;
var
  i: Integer;
  Line: TLine;
  wt: Double;
begin
  if RemoveSpins <> 0 then
    raise NotYet;
  with Line do begin
    Width := 0;
  end;
  Line.ColourAndType := Colour;
  wt := Abundance*Species.Concentration*IScale;
  for i := 0 to nEntries-1 do
    with Entries[i] do begin
      if Assigned(TransitionMomentsSel) then
        if not( ((TransitionMomentsSel.Bra = ManifoldU)
                  and (TransitionMomentsSel.Ket = ManifoldL))
               or ((TransitionMomentsSel.Bra = ManifoldL)
                  and (TransitionMomentsSel.Ket = ManifoldU))
                )
         then
        Continue;
      if Assigned(Manifold1Sel) and (ManifoldL <> Manifold1Sel) then
        Continue;
      if Assigned(Manifold2Sel) and (ManifoldU <> Manifold2Sel) then
        Continue;
      Line.Frequency := Frequency;
      if (UpperQuanta <> '') and (LowerQuanta <> '') then
        Line.CachedName := UpperQuanta + ' - ' + LowerQuanta
      else
        Line.CachedName := UpperQuanta + LowerQuanta;
      { Use negative index values here to flag upper state }
      // Line.IndexU := -1-i; { Flag upper state index }
      // Line.IndexL := i;
      Line.IndexL := IndexL;
      Line.IndexU := IndexU;
      Line.EigenvaluesL := ManifoldL.Eigenvalues(gL,JL);
      Line.EigenvaluesU := ManifoldU.Eigenvalues(gU,JU);
      if Line.EigenvaluesU.Manifold.Colour <> clNone then
        Line.ColourAndType := Line.EigenvaluesU.Manifold.Colour
      else if Line.EigenvaluesL.Manifold.Colour <> clNone then
        Line.ColourAndType := Line.EigenvaluesL.Manifold.Colour
      else
        Line.ColourAndType := Colour;
      Line.Width := Width*WScale;
      if EinsteinA > 0 then
        Line.Strength := EinsteinA
      else
        Line.Strength := Intensity;
      Line.Strength := Line.Strength*wt;
      if PrintLevel >= plLineList then
        Spectrum.ShowLine(Line, 0, Mixture.Units, Mixture.Simulations[0].PlotUnits, False, Problem);
      Spectrum.AddLine(Line, Mixture.Simulations[0], cm1, Problem, False);
    end;
  Result := True;
end;

constructor THITRANMolecule.Create(AOwner: TComponent);
begin
  SavedProperties := SavedProperties + ',EDiff=1e-3,IScale=1';
  inherited;
  ModifyParameter(IndexOfParameter('EDiff'),1,False,False);
end;

function THITRANMolecule.DummyManifold: THITRANManifold;
begin
  if not Assigned(FDummyManifold) then  begin
    FDummyManifold := THITRANManifold.Create(Self);
    FDummyManifold.Name := 'Dummy';
    FDummyManifold.Initial := True;
  end;
  Result := FDummyManifold;
end;

function THITRANMolecule.FindLine(var ALine: TLine; const TxDetails: string): Integer;
var
  i: Integer;
begin
  { This is intended for testing purposes only; given eigenvalue numbers find corresponding
    line list entry }
  for i := 0 to High(Entries) do
    with Entries[i] do
      if (ALine.IndexU = IndexU) and (ALine.IndexL = IndexL)
         and (ALine.EigenvaluesU = ManifoldU.Eigenvalues(gU,JU))
         and (ALine.EigenvaluesL = ManifoldL.Eigenvalues(gL,JL)) then begin
        if (UpperQuanta <> '') and (LowerQuanta <> '') then
          ALine.CachedName := UpperQuanta + ' - ' + LowerQuanta
        else
          ALine.CachedName := UpperQuanta + LowerQuanta;
        Result := i;
        if EinsteinA > 0 then
          ALine.Strength := EinsteinA
        else
          ALine.Strength := Intensity;
        ALine.Strength := ALine.Strength*IScale;
        ALine.Frequency := Frequency;
        ALine.Width := Width*WScale;
        ALine.ColourAndType := Colour;
        if (TxDetails = '') or (TxDetails = ALine.CachedName) then
          Exit;
      end;
  Result := -1;
end;

function THITRANMolecule.FixupProperties: Exception;
var
  i: Integer;
begin
  SetLength(Entries, nEntries); { Discard unused memory }
//  if ActivePartitionFunction.ClassType = TDefaultPartitionFunction then
//    with TInterpolatedPartitionFunction.Create(Self) do begin
//      Name := 'Partition Function';
//      AutoZero := False;
//    end;
  Mixture.LargestIndex := Max(Mixture.LargestIndex,nEntries*10);
  for i := 0 to nEntries-1 do
    with Entries[i] do
      if (UpperQuanta <> '') and (LowerQuanta <> '') then begin
        IndexL := ManifoldL.AddLevel(ELower, gL, JL, LowerQuanta, EDiff);
        IndexU := ManifoldU.AddLevel(ELower+Frequency, gU, JU, UpperQuanta, EDiff);
      end else begin
        IndexL := ManifoldL.AddLevel(ELower, gL, JL, '', EDiff);
        IndexU := ManifoldU.AddLevel(ELower+Frequency, gU, JU, '', EDiff);
      end;
  Result := inherited FixupProperties;
end;

function THITRANMolecule.HelpOnParameter(AIndex: Integer): string;
begin
  case AIndex of
    3: Result := 'Scale factor for widths';
    4: Result := 'Energy levels closer than this considered identical';
  else
    Result := inherited HelpOnParameter(AIndex);
  end;
end;

{$IFNDEF CONSOLE}
procedure THITRANMolecule.AddHandlers;
begin
  inherited;
  // TODO: AddHandler('Load...', 'Load from file', {$IFDEF FPC}@{$ENDIF}LoadFromFileClick);
end;

procedure THITRANMolecule.LoadFromFileClick(Sender: TObject);
//var
//  Stream: TMemoryStream;
begin
{ TODO: Restore }
//  with TOpenDialog.Create(Application) do
//    try
//      Filter := 'Hitran file (*.par)|*.par|All Files|'+AllFilesMask;
//      Title := 'Load External Line List';
//      if Execute then begin
//        Stream := TMemoryStream.Create;
//        try
//          Stream.LoadFromFile(FileName);
//          ReadHitranFile(Stream, FileName);
//        finally
//          Stream.Free;
//        end;
//      end;
//    finally
//      Free;
//    end;
end;
{$ENDIF}

function THITRANMolecule.ReadProperties(Node: TDOMNode; CheckUnread: Boolean): Integer;
var
  Child: TDOMNode;
  b: AnsiString;
  P, PS, PP: PAnsiChar;
  nFields: Integer;
  LineNo, gUt, gLt, JUt, JLt: Double;
  Entry: TExternalEntry;
  Version: Integer;
  function NextField(Text:Boolean):AnsiString;
  var
    S: PAnsiChar;
  begin
    Inc(P);
    S := P;
    while not ( (S^ in [#0, #10, #13])
                or (not Text and (S^ = ',')) { Was S^ in [',', '/'] }
                or (Text and (S^ = #9))
                 ) do
      Inc(S);
    SetLength(Result, S-P);
    if S > P then
      StrMove(@Result[1], P, S-P);
    Result := Trim(Result);
    P := S;
  end;
begin
  Result := inherited ReadProperties(Node, False);
  Version := ReadVersion;
  Child := Node.FirstChild;
  if Assigned(Child) and (Child is TDOMCDATASection) then begin
    b := TDOMCDATASection(Child).Data;
    P := @b[1];
    FnEntries := 0;
    while P^ <> #0 do begin
      with Entry do begin
        Width := 0;
        EinsteinA := 0;
        SetNan(ELower);
        LineNo := -1; { Transfer variables - must be double }
        gUt := 0;
        gLt := 0;
        JUt := -0.5;
        JLt := -0.5;
        if Version < 7001485 then
          P := TryReadTabbedLine(tlAny, P, nil, [@LineNo,@Frequency,@Intensity,@Width,@ELower,@gUt,@gLt,@JUt,@JLt], nFields, EmptyArray)
        else
          P := TryReadTabbedLine(tlAny, P, nil, [@LineNo,@Frequency,@Intensity,@EinsteinA,@Width,@ELower,@gUt,@gLt,@JUt,@JLt], nFields, EmptyArray);
        if P = nil then
          P := @b[Length(b)]
        else
          Dec(P);
        if nFields > 0 then begin
          LineNumber := Round(LineNo);
          gU := Round(gUt);
          gL := Round(gLt);
          JU := Round(JUt*2);
          JL := Round(JLt*2);
          if P^ = ',' then begin
            if Version < 8000193  then begin
              { Adjust to allow for "/" in manifold name }
              { Assume last "/" on line is quantum number delimiter }
              PS := nil;
              PP := P;
              while not (PP^ in [#0, #10, #13]) do begin
                if PP^ = '/' then
                  PS := PP;
                Inc(PP);
              end;
              if PS <> nil then
                PS^ := #9;
            end;
            ManifoldU := FindOrCreateNode(NextField(Version >= 8000193), THITRANManifold) as THITRANManifold;
          end else
            ManifoldU := nil;
          if ((Version < 8000193) and (P^ = ',')) or ((Version >= 8000193) and (P^ = #9)) then
            ManifoldL := FindOrCreateNode(NextField(True), THITRANManifold) as THITRANManifold
          else
            ManifoldL := nil;
          if not (P^ in [#0, #10, #13]) then
            UpperQuanta := NextField(True)
          else
            UpperQuanta := '';
          if not (P^ in [#0, #10, #13]) then
            LowerQuanta := NextField(True)
          else
            LowerQuanta := '';
          AddEntry(Entry);
        end else
          Inc(P);
      end;
      SkipLineEnding(P);
    end;
    Node.RemoveChild(Child);
  end;
  if CheckUnread then
    Result := Result + nUnreadProperties(Node);
end;

function THITRANMolecule.WriteProperties(Parent: TDOMNode; AllParameters, WithChildren: Boolean): TDOMElement;
var
  Strings: TStringList;
  i: Integer;
  s: string;
begin
  Result := inherited WriteProperties(Parent, AllParameters, WithChildren);
  Strings := TStringList.Create;
  try
    Strings.Capacity := Length(Entries)+1;
    Strings.Add('');
    for i := 0 to High(Entries) do begin
      with Entries[i] do begin
        s := IntToStr(LineNumber)
             +#9+FloatToStr(Frequency)
             +#9+FloatToStr(Intensity)
             +#9+FloatToStr(EinsteinA)
             +#9;
        if Width > 0 then
          s := s+FloatToStr(Width);
        s := s+#9;
        if not IsNan(ELower) then
          s := s+FloatToStr(ELower);
        s := s+#9;
        if gU > 0 then
          s := s+FloatToStr(gU);
        s := s+#9;
        if gL > 0 then
          s := s+FloatToStr(gL);
        s := s+#9;
        if JU >= 0 then
          s := s+QuantumNumber.QuantumNumberToStr(JU,qftJ);
        s := s+#9;
        if JL >= 0 then
          s := s+QuantumNumber.QuantumNumberToStr(JL,qftJ);
        s := s+',';
        if ManifoldU <> FDummyManifold then
          s := s+ManifoldU.Name;
        s := s+#9; { Was ',' in versions before 8.0.193 }
        if ManifoldL <> FDummyManifold then
          s := s+ManifoldL.Name;
        s := s+#9+UpperQuanta;  { Was '/' in versions before 8.0.193 }
        if LowerQuanta <> '' then
          s := s+#9+LowerQuanta;
      end;
      Strings.Add(s);
    end;
    Result.AppendChild(Parent.OwnerDocument.CreateCDATASection(Strings.Text));
  finally
    Strings.Free;
  end;
end;

{ THITRANManifold }

procedure THITRANManifold.AddAllLevels(AState: TState);
var
  i, Ind: Integer;
  pop: Double;
  Overflow: Exception;
  Simulation: TSimulation;
  Mol: THITRANMolecule;
begin
  Simulation := Mixture.Simulations[0];
  Overflow := nil;
  Mol := Molecule as THITRANMolecule;
  for i := 0 to Mol.nEntries - 1 do
    if Mol.Entries[i].ManifoldL = Self then begin
      pop := Population(Eigenvalues(Mol.Entries[i].gL,Mol.Entries[i].JL), i, Simulation, Overflow, 0);
      Ind := PopulationIndex(0,0,i,True);
      Parameters[Ind] := pop;
    end;
end;

function THITRANManifold.AddLevel(Energy: Double; g: Integer; J: TQuantumNumber;
  const Quanta: string; EDiff: Double): Integer;
begin
  Result := Eigenvalues(g,J).AddLevel(Energy, g, Quanta, EDiff);
end;

procedure THITRANManifold.ClearCache(ClearChildren: Boolean);
begin
  inherited ClearCache(ClearChildren);
  FHalfSpin := 0;
  ClearEigenvalues;
end;

function THITRANManifold.Eigenvalues(g: Integer; J: TQuantumNumber): TExternalEigenvalues;
begin
  if (J < 0) and (g > 0) and ((Molecule as THITRANMolecule).gDiv > 0) then begin
    J := Round(g/(Molecule as THITRANMolecule).gDiv-1);
    FHalfSpin := SmallestProjection(J);
  end;
  Result := Eigenvalues(J);
  { TODO: Fails for equivalent nuclei }
  Result.SetStatisticalWeight((Molecule as THITRANMolecule).gDiv);
end;

initialization
  THITRANMolecule.RegisterClass(TSpecies);
  THITRANManifold.RegisterClass(THITRANMolecule);
  EmptyEigenvalues := TEigenvalues.Create(nil,0,0,ZeroField);
finalization
  FreeAndNil(EmptyEigenvalues);
end.
