unit Fitter;
{$I utils.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses Graphics, MatrixVec, MNode, Mixture, Overlay, Obs, MathSup;

type
  OnShowContour = procedure (Spectra: TOverlays; Simulation: TColumnVector; ObsMinusCalc: Boolean) of object;

function FitLines(var Parameters: TParameters; Obs: TObservationArray; Precision: Integer; Mixture: TMixture; CheckDerivatives: Double; out OldSD, NewSD: Double; Quiet: Boolean = False):string;
function FitContour(Mixture: TMixture; Spectra: TOverlays; ShowContour: OnShowContour; var Parameters: TParameters; JustResiduals: Boolean; CheckDerivatives: Double):string;
function CalculateObs(var Observation: TObservation; out Value: Double; out RegeneratedLabel: string; RegenerateLabel:Boolean; AObsType: TObsType; Wait: Boolean = True): Boolean;
function CalculateObsEx(var Observation: TObservation; out Value: Double; out RegeneratedLabel: string; RegenerateLabel:Boolean;
                        AObsType: TObsType; out EigU, EigL: TEigenvalues; out Line: TLine; Wait: Boolean = True): Boolean;
function FindObservationColour(const Obs: TObservation; Upper: Boolean):TColor;

type

  { TFitter }

  TFitter = class
  public
    nObs, nPar: Integer;
    a: TMatrix;   { d(obs)/d(par) }
    omc: TColumnVector; { normally obs - calc }
    par: TColumnVector; { Parameters - changed by this routine }
    weights: TColumnVector; { if WeightedFit then the relative standard deviation of each value in obs }
    OldSD, NewSD: Double; { Previous and estimated next standard deviations }
    UnWtOldSD, UnWtNewSD: Double; { Previous and estimated next standard deviations ignoring weights }
    b: TColumnVector;  { Standard Deviations of the parameters}
    ada: TMatrix; { Correlation Matrix }
    SVD: TMatrix; SVDw: TColumnVector; { Details of Singular Value Decomposition }
    Correlations: TDoubleArrayArray; { Correlation of observation with following observations }
    ParameterScales: TColumnVector; { Scaling factors for parameters; can be nil }
    CheckDerivatives: Double;
    AlternateFitter: TFitter;
    Mixture: TMixture;
    nDiscard: Integer; { Number of paramters not floated because of SVD threshold }
    SubtractDerivs: Boolean;
    WeightedFit: Boolean;
    constructor Create(AnPar, AnObs: Integer; AMixture: TMixture; ACheckDerivatives: Double);
    destructor Destroy; override;
    function DoLeastSQ(var Parameters: TParameters; Precision: Integer; Apply: Boolean = True; Quiet: Boolean = False):string;
    procedure CalculateDerivatives(var Parameters: TParameters);
    procedure ApplyParameters(var Parameters: TParameters; SetCorrelations: Boolean = True);
    function CalculateAllObs(v: TColumnVector): Boolean; virtual; abstract;
    procedure AddAlternateFitter;
    procedure ShowSVD(var Parameters: TParameters);
    procedure SetupMatrices;
  end;

  { TLineFitter }

  TLineFitter = class(TFitter)
  protected
    { For determining printout format }
    LargestMolecule, MolName, LargestLower, LargestUpper, LargestName: string;  { For Determining field lengths }
    SymWidth: Integer;
    MoleculeChanged, LowerChanged, UpperChanged: Boolean; { For Determining if field needs printing }
    JHeader, Header: string;
    HasEField, HasBField, HasJ, HasF: Boolean;
  public
    Parameters: TParameters;
    Obs: TObservationArray;
    PreCalcPartitionFunction: Boolean;
    MaxCorrelatedObservations: Integer;
    Precision: Integer;
    function CalculateAllObs(v: TColumnVector): Boolean; overload; override;
    function CalculateAllObs(v: TColumnVector; Observations: TObservationArray; Start: Integer = 0): Boolean; overload;
    constructor Create(var TheParameters: TParameters; var TheObservations: TObservationArray; TheMixture: TMixture; ACheckDerivatives: Double);
    procedure ShowResidualHeader(var F: Text; Title: string; ShowCalc: Boolean);
    procedure ShowResidual(var F: Text; Index: Integer; Calc: Double; ShowCalc: Boolean);
    procedure ShowDerivativeHeader(var F: Text; Title: string);
    procedure ShowDerivative(var F: Text; Index: Integer);

    function PrepareCalculated:Boolean;
    procedure PrepareResiduals(var F: Text; Show: Boolean; ShowProgress: Boolean);
  end;

  { TAutoFitFitter }

  TAutoFitFitter = class(TLineFitter)
  public
    IntialCalculated: TColumnVector;
    InitialDerivatives: TMatrix;
    InitialParameters: TParameters;
    constructor Create(var TheParameters: TParameters; var TheObservations: TObservationArray; TheMixture: TMixture; ACheckDerivatives: Double);
    destructor Destroy; override;
    procedure Reset;
    function MakeSVD:Double;
  end;


implementation

uses Progress, SysUtils, NLFit, ComplexMatrix, Vib, StaticField,
     SpectrumPlot, ExpPlot, Units, DRintensity, StrUtils, UCompat,
     Worker, PGWork, Math, QuantumNumber, Constrain, Profile, MatrixOps;

var
  ResidualOutput: string = 'Residual Output';
  LeastSquaresTime: string = 'Least Squares';
  LineFitTime: string = 'Line Fit';
  ContourFitTime: string = 'Contour Fit';
  PrepareCalculatedTime: string = 'Prepare Calculated';
  CalculateDerivativesTime: string = 'Derivatives of Observations';
  ApplyParametersTime: string = 'Apply Parameters';

{ TFitter }

procedure TFitter.AddAlternateFitter;
begin
  AlternateFitter := TFitter.Create(nPar, nObs, Mixture, 0);
  AlternateFitter.nPar := nPar;
  AlternateFitter.nObs := nObs;
  AlternateFitter.a.Assign(a);
  AlternateFitter.omc.Assign(omc);
  AlternateFitter.weights.Assign(weights);
  AlternateFitter.Correlations := Copy(Correlations);
end;

procedure TFitter.ShowSVD(var Parameters: TParameters);
var
  i, j, l, p: Integer;
  Sorter: TEigenvectorSorter;
begin
  if SVD = nil then
    Exit;
  WriteLn(Log);
  WriteLn(Log, 'Singular Value Decomposition');
  if (Mixture.SVDlimit <= 0) or (Mixture.SVDlimit > nPar) then begin
    Write(Log,' ':17);
    for i := 0 to nPar-1 do
      Write(Log, ' ', Parameters[i].Name:13);
    WriteLn(Log);
    for i := 1 to nPar do begin
      Write(Log, i:3, DTosMA(SVDW[i],14,-1));
      for j := 1 to nPar do
        Write(Log, DTosMA(SVD[i,j],Max(14,Length(Parameters[j-1].Name)+1),-1));
      WriteLn(Log);
    end;
  end else begin
    l := 0;
    for i := 0 to nPar-1 do
      l := Max(l, Length(Parameters[i].Name));
    Sorter := TEigenvectorSorter.Create(SVD, False);
    for i := 1 to nPar do begin
      Write(Log, i:3, DTosMA(SVDW[i],14,-1));
      Sorter.Process(i, Mixture.SVDlimit);
      for j := 1 to Mixture.SVDlimit do begin
        p := Sorter.Indices[j-1];
        Write(Log, DTosMA(SVD[i,p],Mixture.Precision+6,-1), ' ', Parameters[p-1].Name, ' ':l-Length(Parameters[p-1].Name)+1);
        if j < Mixture.SVDlimit then
          Write(Log, '+');
      end;
      WriteLn(Log);
    end;
    Sorter.Free;
  end;
end;

procedure TFitter.SetupMatrices;
begin
  if Mixture.SVDThresh >= 0 then
    if SVD = nil then begin
      SVD := TMatrix.Create(nPar, nPar);
      SVDw := TColumnVector.Create(nPar);
    end else begin
      SVD.ColCount := nPar;
      SVD.RowCount := nPar;
      SVDw.Size := nPar;
    end
  else begin
    FreeAndNil(SVD);
    FreeAndNil(SVDw);
  end;
  if Mixture.ScaledParameters then begin
    if ParameterScales = nil then
      ParameterScales := TColumnVector.Create
  end else
    FreeAndNil(ParameterScales);
end;

constructor TFitter.Create(AnPar, AnObs: Integer; AMixture: TMixture; ACheckDerivatives: Double);
begin
  AMixture.CheckProperties;
  nObs := AnObs;
  nPar := AnPar;
  if (nObs <= 0) and (nPar > 0) then
    raise EReportException.Create('No observations');
  if nObs < nPar then
    raise EReportException.CreateFmt('More parameters (%d) than observations (%d)', [nPar, nObs]);
  a := TMatrix.Create(nObs, nPar);
  omc := TColumnVector.Create(nObs);
  weights := TColumnVector.Create(nObs); { Also used as workspace }
  Mixture := AMixture;
  CheckDerivatives := ACheckDerivatives;
  SubtractDerivs := True;
  WeightedFit := True;
end;

destructor TFitter.Destroy;
begin
  FreeAndNil(AlternateFitter);
  a.Free;
  omc.Free;
  par.Free;
  b.Free;
  ada.Free;
  weights.Free;
  SVD.Free;
  SVDw.Free;
  ParameterScales.Free;
  inherited;
end;

function TFitter.DoLeastSQ(var Parameters: TParameters; Precision: Integer; Apply: Boolean; Quiet: Boolean):string;
var
  i, j, p, l, nErrDig, nDerivs, iMx, jMx: Integer;
  r1: string;
  Sum, C, v, x, Cmax: Double;
  MaxDerivative: array of Double;
  ObsMax: array of Integer;
  ShowScale: Boolean;
  WtOrNil: TColumnVector;
begin
  TStart(LeastSquaresTime);
  if (Mixture.FitDetails or Assigned(AlternateFitter)) and (nPar > 0) then begin
    if Assigned(AlternateFitter) then
      nDerivs := 2
    else
      nDerivs := 1;

    if (Mixture.FitDetails) and not Quiet then begin
      WriteLn(Log);
      Write(Log,' ':4);
      if nDerivs > 1 then
        WriteLn(Log, 'Derivative Matrix; alternate derivatives listed second')
      else
        WriteLn(Log, 'Derivative Matrix');
      for i := 0 to nPar-1 do
        Write(Log, ' ', Parameters[i].Name:13*nDerivs);
      WriteLn(Log);
      for i := 1 to nObs do begin
        Write(Log, i:6);
        for j := 1 to nPar do begin
          Write(Log, DTosMA(a[i,j],Max(14,Length(Parameters[j-1].Name)+1),-1));
          if nDerivs = 2 then begin
            Write(Log, DTosMA(AlternateFitter.a[i,j],Max(14,Length(Parameters[j-1].Name)+1),-1));
          end;
        end;
        WriteLn(Log);
      end;
    end;

    if Assigned(AlternateFitter) then begin
      SetLength(MaxDerivative, nPar);
      SetLength(ObsMax, nPar);
      for j := 1 to nPar do begin
        C := 0;
        { First pass: find biggest derivative }
        for i := 1 to nObs do
          C := Max(C, Max(Abs(a[i,j]), Abs(AlternateFitter.a[i,j])));
        { Second pass: find biggest difference in derivatives }
        x := 0;
        ObsMax[j-1] := 1;
        if C > 0 then
          for i := 1 to nObs do begin
            v := Abs(a[i,j]-AlternateFitter.a[i,j])/C;
            if v > x then begin
              x := v;
              ObsMax[j-1] := i;
            end;
          end;
        MaxDerivative[j-1] := x;
      end;

      WriteLn(Log);
      WriteLn(Log, 'Alternate fit with scaled step size for derivatives');
      AlternateFitter.omc.Assign(omc);
      if Assigned(weights) then
        AlternateFitter.weights.Assign(weights)
      else
        FreeAndNil(AlternateFitter.weights);
      AlternateFitter.DoLeastSQ(Parameters, Precision, False);
      WriteLn(Log);
    end;
  end;

  { Fill in parameter info }
  if par = nil then begin
    par := TColumnVector.Create(nPar);
    b := TColumnVector.Create(nPar);
    ada := TMatrix.Create(nPar, nPar);
  end else begin
    par.Size := nPar;
    b.Size := nPar;
    ada.RowCount := nPar;
    ada.ColCount := nPar;
  end;
  for i := 0 to nPar-1 do
    par[i+1] := Parameters[i].Value;
  SetupMatrices;
  if WeightedFit then
    WtOrNil := weights
  else
    WtOrNil := nil;
  nDiscard := LeastSQ(a, omc, par, WtOrNil, OldSD, NewSD, UnWtOldSD, UnWtNewSD, b, ada, SVDw, SVD,
                      Mixture.SVDThresh, Mixture.FitDetails, Mixture.ScaleChanges, Correlations, ParameterScales);
  if not Quiet then begin
    WriteLn(Log);
    if nPar > 0 then
      if SVD <> nil then
        Write(Log, 'SVD fit: ')
      else
        Write(Log, 'QR fit: ');
    Write(Log, nObs, ' Observations,  ', nPar, ' Parameters');
    if ParameterScales <> nil then
      WriteLn(Log, ' (scaled)')
    else
      WriteLn(Log);
  end;
  if nPar > 0 then
    Result := 'Initial '
  else
    Result := '';
  Result := Result + 'Average Error: ' + FloatToStr(OldSD);
  if OldSD <> UnWtOldSD then
    Result := Result + '  (Unweighted: ' +  FloatToStr(UnWtOldSD) + ')';
  if not Quiet then
    WriteLn(Log, Result);
  if nPar > 0 then begin
    r1 := 'Predicted New Error:   ' + FloatToStr(NewSD);
    if NewSD <> UnWtNewSD then
      r1 := r1 + '  (Unweighted: ' +  FloatToStr(UnWtNewSD) + ')';
    if not Quiet then
      WriteLn(Log, r1);
    Result := Result + ' - ' + r1;
    if (nDiscard > 0) then begin
      Result := '**** '+IntToStr(nDiscard) + ' combinations of parameters not floated ' + Result;
      if not Quiet then
        WriteLn(Log, '**** ', nDiscard, ' combinations of parameters not floated **** (SVD threshold = ', FloatToStr(Mixture.SVDThresh),')');
    end;

    { For constants (such as the observed spectrum scale factor) where the residual is
      inversly proportional to the constant, the estimated change is particularly poor if
      the change is large. Adjust the change by converting to a change in the inverse,
      and then inverting. [The ratio of derivatives of a parameter p and it's inverse is -1/p^2
      so a change d in p corresponds to a change -d/p^2 in 1/p so the new p is
      1/(1/p-d/p^2) = p/(1-d/p) }
    for i := 0 to nPar - 1 do
      if Parameters[i].Inverse then begin
        if not Quiet then
          WriteLn(Log, 'Linear New Value for ',Parameters[i].Name,': ',par[i+1]);
        C := Parameters[i].Value/(1-(par[i+1]-Parameters[i].Value)/Parameters[i].Value);
        { Check that parameter does not change sign; fall back to original method if it does }
        if C * Parameters[i].Value > 0 then
          par[i+1] := C;
      end;

    if not Quiet then begin
      WriteLn(Log, 'Parameters:');
      Write(Log, '  # Old                 New                 Std Dev   Change/Std Sens    ');
      if Assigned(AlternateFitter) then
        Write(Log, 'Deriv Diff  Obs# ');
      ShowScale := (Mixture.FitDetails or (nDiscard > 0)) and (ParameterScales <> nil);
      if ShowScale then
        Write(Log, 'Scale   ');
      WriteLn(Log, 'Summary              Name');
    end;
    for i := 0 to nPar-1 do begin
      { Calculate parameter sensitivity; defined in JKG Watson, J Mol Spec 66 500 (1977)
        See also RJ LeRoy J Mol Spec 191 223 (1998) for additional ways of dealing with this issue.
        Choose number of significant figures such that using the rounded values allows the
        calculated standard deviation to rise by no more then 10% }
      Sum := 0;
      for j := 1 to nObs do
        { Note - don't need Sqr(a[j,i+1]/weights[j]) as a has been adjusted to give observations with standard deviation 1 }
        Sum := Sum + Sqr(a[j,i+1]);
      { and use it to determine the number of uncertainty digits to show }
      Sum := 0.1*NewSD/nPar/Sqrt(Sum/nObs);
      if not Quiet then begin
        nErrDig := Max(Trunc(Log10(b[i+1]/Sum))+1,2);
        Write(Log, Format('%3d%s%s%s%s %s', [i+1, DtosMA(Parameters[i].Value,20,-1),
                   DtosMA(par[i+1],20,-1), DtosMA(b[i+1],10,-1), DtosMA((par[i+1]-Parameters[i].Value)/b[i+1],10,4),
                   DtosMA(Sum,8,-1)]));
        if Assigned(AlternateFitter) then
          Write(Log, DtosMA(MaxDerivative[i],11,-1), ' ', ObsMax[i]:5);
        if ShowScale then
          Write(Log, DtosMA(ParameterScales[i+1],8,-1));
        WriteLn(Log, Format(' %-20s %s', [DtosErr(par[i+1],b[i+1],nErrDig), Parameters[i].Name]));
      end;
      if Apply then begin
        Parameters[i].Value := par[i+1];
        Parameters[i].StdDev := b[i+1];
      end;
    end;
    if Apply and not Quiet then
      WriteConstraints(Log);
    if (nPar > 1) and not Quiet then begin
      WriteLn(Log);
      WriteLn(Log, 'Correlation Matrix');
      if nPar > 1 then begin
        Write(Log, 'Determinant = ', FloatToStrF(Determinant(ada), ffFixed, 15, Mixture.Precision-1));
        if nPar > 2 then begin
          Cmax := 0;
          for i := 1 to nPar do
            for j := 1 to i-1 do
              if Abs(ada[i,j]) > Cmax then begin
                Cmax := Abs(ada[i,j]);
                iMx := i;
                jMx := j;
              end;
          if Cmax > 0 then begin
            Write(Log, ', Largest off-diagonal element = ', FloatToStrF(ada[iMx,jMx], ffFixed, 15, Mixture.Precision-1), ' at ', iMx, ',', jMx);
            Write(Log, ' = ', Parameters[iMx-1].Name, ', ', Parameters[jMx-1].Name);
          end;
        end;
        WriteLn(Log);
      end;
      for i := 1 to nPar do
        Write(Log, i:Mixture.Precision+4);
      WriteLn(Log);
      for i := 1 to nPar do begin
        Write(Log, i:3);
        for j := 1 to i do
          Write(Log, FloatToStrF(ada[i,j], ffFixed, 15, Mixture.Precision-1):Mixture.Precision+4);
        WriteLn(Log);
      end;

      if (Mixture.FitDetails or (nDiscard > 0)) and (SVD <> nil) then
        ShowSVD(Parameters);
    end;
    if Apply then
      ApplyParameters(Parameters);
  end;
  TEnd(LeastSquaresTime);
end;

procedure TFitter.ApplyParameters(var Parameters: TParameters; SetCorrelations: Boolean = True);
var
  i, j: Integer;
  OldParameters: TDoubleArray;
begin
  TStart(ApplyParametersTime);
  if SetCorrelations then begin
    if Length(Mixture.Correlations.FitDetails.Parameters) <> Length(Parameters) then
      Mixture.Correlations.FitDetails.Parameters := Copy(Parameters)
    else if nPar > 0 then
      Move(Parameters[0], Mixture.Correlations.FitDetails.Parameters[0], nPar*SizeOf(Parameters[0]));
    with Mixture.Correlations do begin
      if Length(FitDetails.Correlation) <> ada.RowCount then
        SetLength(FitDetails.Correlation, ada.RowCount);
      for i := 0 to ada.RowCount - 1 do begin
        if Length(FitDetails.Correlation[i]) <> i then
          SetLength(FitDetails.Correlation[i],i);
        for j := 0 to i - 1 do
          FitDetails.Correlation[i][j] := ada[i+1,j+1];
      end;
    end;
  end else
    Mixture.Correlations.FitDetails.Parameters := nil; { Simply clear }
  SetLength(OldParameters, nPar);
  try
    for i := 0 to nPar-1 do
      with Parameters[i] do begin
        OldParameters[i] := MNode.Parameters[Index];
        Parameters[i].Value := par[i+1];
        Parameters[i].StdDev := b[i+1];
        MNode.SetParameter(Index, par[i+1]);
        MNode.StdDev[Index] := b[i+1];
      end;
    ApplyConstraints;
    Mixture.CheckProperties(True);
  except
    on E:Exception do begin
      WriteLn(Log,'**** Failure at end of fit:'+E.Message+' - reverting parameters');
        for i := 0 to nPar-1 do
          with Parameters[i] do
            MNode.SetParameter(Index, OldParameters[i]);
      ApplyConstraints;
    end;
  end;
  TEnd(ApplyParametersTime);
end;

procedure TFitter.CalculateDerivatives(var Parameters: TParameters);
var
  i, j, deriv, nDerivs: Integer;
  vo, vn, dv: Double;
  Msg: string;
  AA: TMatrix;
begin
  if CheckDerivatives <> 0 then begin
    AddAlternateFitter;
    nDerivs := 2;
  end else
    nDerivs := 1;
  {  Calculate Derivatives wrt each parameter }
  TStart(CalculateDerivativesTime);
  try
    PushProgress('Derivatives', High(Parameters));
    for i := 0 to High(Parameters) do begin
      SetProgress(i);
      vo := Parameters[i].Value;
      try
        if Parameters[i].Increment <= 0 then begin
          vn := vo * 1.0001;
          if vn = 0 then
            vn := 0.001; { Random Guess }
        end else
          vn := vo + Parameters[i].Increment;
        for deriv := 1 to nDerivs do begin
          if deriv = 2 then
            vn := (vn-vo)*CheckDerivatives + vo;
          try
            Parameters[i].MNode.SetParameter(Parameters[i].Index, vn);
            ApplyConstraints;
            Mixture.CheckProperties(True); { TODO: Was Parameters[i].MNode.CheckProperties; inefficient as is }
            if not CalculateAllObs(weights) then
              if not CheckForAbort then
                raise EReportException.Create('Transitions not found - see log for details');
          except
            on EAbort do
              Abort;
            on E:Exception do begin
              Msg := 'Error while stepping '+Parameters[i].MNode.FullParameterName(Parameters[i].Index)+' to '+FloatToStr(vn)
                           +': '+E.Message;
              WriteLn(Log, '**** ',Msg);
              raise;
            end;
          end;
          if deriv = 1 then
            AA := a
          else
            AA := AlternateFitter.a;
          for j := 1 to nObs do begin
            if SubtractDerivs then
              dv := omc[j]
            else
              dv := 0;
            AA[j,i+1] := (dv - weights[j])/(vn-vo);
          end
        end;
      finally
        Parameters[i].MNode.SetParameter(Parameters[i].Index, vo);
        ApplyConstraints;
        { TODO: See above Parameters[i].MNode.CheckProperties; }
      end;
    end;
  finally
    PopProgress;
    TEnd(CalculateDerivativesTime);
  end;
end;

type

  { TContourFitter }

  TContourFitter = class(TFitter)
    OriginalCalcI, OriginalObsI: array of Double;
    Original: Boolean;
    Spectra: TOverlays;
    Simulation: TFitPlot;
    constructor Create(AnPar, AnObs: Integer; AMixture: TMixture; ACheckDerivatives: Double);
    destructor Destroy; override;
    function CalculateAllObs(v: TColumnVector): Boolean; override;
  end;

constructor TContourFitter.Create(AnPar, AnObs: Integer; AMixture: TMixture; ACheckDerivatives: Double);
begin
  inherited Create(AnPar, AnObs, AMixture, ACheckDerivatives);
  SubtractDerivs := False;
end;

destructor TContourFitter.Destroy;
begin
  FreeAndNil(Simulation);
  inherited Destroy;
end;

function TContourFitter.CalculateAllObs(v: TColumnVector): Boolean;
var
  i, j, l: Integer;
  Problem: Exception;
  CalcI, ObsI, df: Double;
  ConditionSimulation: TSimulation;
  procedure PlaceObsCalc;
  begin
    if IsNanOrInf(CalcI) then
      raise EReportException.CreateFmt('Calculated value of observation %d is %g',[l+1, CalcI]);
    if Original then begin
      OriginalCalcI[l] := CalcI;
      OriginalObsI[l] := ObsI;
    end else begin
      { Slightly odd logic here as "observed" values can actually change if baseline/scaling/correction changed }
      { Order of calculation here to minimize rounding errors }
      CalcI := CalcI - OriginalCalcI[l];
      ObsI := ObsI - OriginalObsI[l];
    end;
    v[l+1] := ObsI - CalcI;
    Inc(l);
  end;
begin
  Result := True;
  ConditionSimulation := Mixture.Simulations[0];
  { Copy in required frequencies }
  l := 0;
  for i := 0 to Spectra.nPlots-1 do
   if Spectra[i] is TExperimentalPlot then
    with TExperimentalPlot(Spectra[i]) do
      if not Hide then begin
        if Count > 0 then
          CheckProperties(True); { Calculate corrected intensity }
        if ConvolutionNeeded then begin
          PrepareConvolution(ConditionSimulation, ConditionSimulation.IntensityUnits, 0, True);
          with ConvolutedSpectrum do begin
            df := (FMax-FMin)/(nDF-1);
            for j := 0 to nDF-1 do
              Simulation.Frequency[l+j] := FMin+j*df;
            Inc(l, nDF);
          end;
        end else
          for j := StartPoint to EndPoint do
            if not IsNanOrInf(Intensity[j]) then begin
              Simulation.Frequency[l] := Convert(AdjustFrequency(Frequency[j]),Units,ConditionSimulation.PlotUnits);
              Inc(l);
            end;
      end;
  Simulation.MakeFromSimulation(ConditionSimulation, not Mixture.NoLineList, Problem);
  if Assigned(Problem) then
    if Problem is EWarningException then
      Problem.Free
    else
      raise Problem;
  { On first cycle, populate Original Calclated and "observed" inteities }
  if Original then begin
    SetLength(OriginalCalcI, l);
    SetLength(OriginalObsI, l);
  end;
  l := 0; { Copy simulation, doing vertical scaling as we go }
  for i := 0 to Spectra.nPlots-1 do
   if Spectra[i] is TExperimentalPlot then
    with TExperimentalPlot(Spectra[i]) do
      if not Hide then begin
        if ConvolutionNeeded then
          with ConvolutedSpectrum do
            for j := 0 to nDF-1 do begin
              CalcI := Simulation.Plot[l];
              ObsI := Plot[j];
              PlaceObsCalc;
            end
        else
          for j := StartPoint to EndPoint do
            if not IsNanOrInf(Intensity[j]) then begin
              CalcI := ConvertIntensity(Simulation.Plot[l],Simulation.Frequency[l],ConditionSimulation.PlotUnits,ConditionSimulation.IntensityUnits,Units,IntensityUnits);
              ObsI := AdjustIntensity(Intensity[j]);
              PlaceObsCalc;
            end;
      end;
end;

function FitContour(Mixture: TMixture; Spectra: TOverlays; ShowContour: OnShowContour;
                    var Parameters: TParameters; JustResiduals: Boolean; CheckDerivatives: Double):string;
var
  f: TContourFitter;
  nObs: Integer;
  i: Integer;
  vo, vn: Double;
  ConditionSimulation: TSimulation;
begin
  TStart(ContourFitTime);
  ConditionSimulation := Mixture.Simulations[0];
  WriteLn(Log);
  { Find total number of observed points }
  nObs := 0;
  for i := 0 to Spectra.nPlots-1 do
    if Spectra[i] is TExperimentalPlot then
      with TExperimentalPlot(Spectra[i]) do
        if not Hide then begin
          Limits(Frequency, vo, vn, StartPoint, EndPoint);
          if ConvolutionNeeded then begin
            Inc(nObs, ConditionSimulation.nDF);
            Writeln(log, 'Contour fit to ', FloatToStr(AdjustFrequency(vo)), ' - ', FloatToStr(AdjustFrequency(vn)), ' ', NameOfUnit(Units),
                         ' convoluted to ', ConditionSimulation.nDF: 7, ' points from ', Name);
          end else begin
            Inc(nObs, EndPoint-StartPoint+1);
            Writeln(log, 'Contour fit to ', FloatToStr(AdjustFrequency(vo)), ' - ', FloatToStr(AdjustFrequency(vn)), ' ', NameOfUnit(Units),
                         ' (Points ', StartPoint:7, ' - ', EndPoint:7, ') of ', Name);
          end;
        end;
  ApplyConstraints;
  CheckConstraints(Parameters);
  f := TContourFitter.Create(Length(Parameters), nObs, Mixture, CheckDerivatives);
  f.Spectra := Spectra;
  try
    f.Simulation := TFitPlot.Create(ConditionSimulation.PlotUnits,IsPopDist(ConditionSimulation.IntensityUnits));
    f.Simulation.ndf := nObs;
    f.Original := True;
    f.WeightedFit := False;
    f.CalculateAllObs(f.omc);
    if Mixture.FitDetails then begin
      WriteLn(Log, 'Observations');
      WriteLn(Log, '#':6, ' Frequency           Observed       Calculated');
      for i := 0 to nObs-1 do
        WriteLn(Log, i:6, DtosMA(f.Simulation.Frequency[i],20,-1), DtosMA(f.OriginalObsI[i],15,-1), DtosMA(f.OriginalCalcI[i],15,-1));
    end;
    if Assigned(ShowContour) then
      ShowContour(Spectra, f.omc, True);
    if JustResiduals then Exit;
    {  Calculate Derivatives wrt each parameter }
    { TODO: Optimize for simple derivatives }
    f.Original := False;
    f.CalculateDerivatives(Parameters);
    FreeAndNil(f.Weights);
    Spectra.ClearStdDev;
    Result := f.DoLeastSQ(Parameters, Mixture.Precision);
  finally
    f.Free;
    TEnd(ContourFitTime);
  end;
end;

function CalculateObsEx(var Observation: TObservation; out Value: Double; out RegeneratedLabel: string; RegenerateLabel:Boolean;
                        AObsType: TObsType; out EigU, EigL: TEigenvalues; out Line: TLine; Wait: Boolean): Boolean;
var
  LineList: TLineList;
  i: Integer;
  Mixture: TMixture;
  Molecule: TMolecule;
  TransitionMoments: TTransitionMoments;
  Intensity: Double;
  Problem: Exception;
  CalcLine: TLine;
  ConditionSimulation: TSimulation;
  viInfo: TValueInfo;
begin
  Result := False;
  if RegenerateLabel or not Wait then
    RegeneratedLabel := 'l';
  with Observation do begin
    if (AObsType = obIntensity) and not Wait then
      Exit
    else if AObsType = obParameter then begin
      viInfo.MNode := MNode;
      viInfo.viType := TValueInfoType(IndexL);
      viInfo.Index := IndexU;
      Value := viInfo.GetValue;
    end else if AObsType = obExpression then
      Value := (MNode as TExpression).Expr.Value
    else if AObsType <> obEUpper then begin
      AIndexL := FindEigenvaluesAndIndex(EigL, ManifoldL, JL, SymL, IndexL, Name, True, LineNumber, FileName, RemoveSpins, QuantumNumberFormat, ExField, True, Wait);
      if not Assigned(EigL) or (AIndexL <= 0) then
        Exit; { not waiting or error }
      Value := EigL.E[AIndexL];
      if AObsType <> obELower then
        Value := -Value;
      if RegenerateLabel then
        RegeneratedLabel := EigL.DominantBasis[AIndexL].StateLabel(ManifoldL.Mixture.QuantumNumberFormat);
    end else
      Value := 0;
    Result := True;
    if not (AObsType in [obELower, obParameter, obExpression]) then begin
      AIndexU := FindEigenvaluesAndIndex(EigU, ManifoldU, JU, SymU, IndexU, Name, False, LineNumber, FileName, RemoveSpins, QuantumNumberFormat, ExField, True, Wait);
      Result := Assigned(EigU) and (AIndexU > 0);
      if not Result then begin
        { not waiting or error }
        if not Wait then
          RegeneratedLabel := 'l';
        Exit;
      end;
      Value := Value + EigU.E[AIndexU];
      if RegenerateLabel then begin
        if AObsType <> obEUpper then
          RegeneratedLabel := ' - ' + RegeneratedLabel;
        RegeneratedLabel := EigU.DominantBasis[AIndexU].StateLabel(ManifoldU.Mixture.QuantumNumberFormat) + RegeneratedLabel;
        if AObsType <> obEUpper then
          RegeneratedLabel := ' : ' + EigU.DominantBasis[AIndexU].TxName(EigU.DominantBasis[AIndexU],EigL.DominantBasis[AIndexL]) + ' : ' + RegeneratedLabel;
      end;
    end;
    if RegenerateLabel and (AObsType = obLine) and (ManifoldU.Initial or ManifoldL.Initial) then begin
      { Consider regenerating intensity calculation }
      Molecule := ManifoldU.Molecule;
      TransitionMoments := Molecule.FindTransitionMoments(ManifoldU, ManifoldL);
      Line.CachedName := Copy(RegeneratedLabel,4,Length(RegeneratedLabel));
      Line.Frequency := Value;
      if Assigned(TransitionMoments) and TransitionMoments.Allowed(JU, SymU, JL, SymL) then begin
        TransitionMoments.MakeLine(CalcLine, EigL, EigU, AIndexL, AIndexU, ManifoldL.Mixture.Simulations[0]);
        Problem := nil;
        Mixture := ManifoldL.Mixture;
        Intensity := Mixture.Simulations[0].LineIntensity(CalcLine, Mixture.Units, Problem);
        if not Assigned(Problem) then begin
          RegeneratedLabel := FloatToStrF(Intensity, ffExponent, 5, 2) + RegeneratedLabel;
          Line.Strength := Intensity;
        end else
          Problem.Free;
      end;
    end;
    if AObsType = obIntensity then begin
      { Slightly inefficient - calculate all lines between these pairs of Manifold, J and Sym.
        Could be optimized by sorting observations by Manifold, J and sym }
      Mixture := ManifoldL.Mixture;
      ConditionSimulation := Mixture.Simulations[0];
      if ConditionSimulation.IntensityUnits <> Observation.IntensityUnits then begin
        ConditionSimulation := TSimulation.Create(nil);
        ConditionSimulation.UpdateBaseNode(Mixture.Simulations[0]);
        ConditionSimulation.IntensityUnits := Observation.IntensityUnits;
      end;
      LineList := TLineList.Create(ConditionSimulation.PlotUnits,IsPopDist(ConditionSimulation.IntensityUnits));
      { TODO: IsPopDist(Mixture.IntensityUnits) may need treating as a special case }
      try
        Problem := nil;
        ManifoldU.Molecule.AddToSpectrum(LineList, Mixture.PrintLevel, RandomExcitation, Problem, ConditionSimulation,
                                         JL, ManifoldL.Molecule.JDeAdjustedSym(JL,SymL),
                                         JU, ManifoldU.Molecule.JDeAdjustedSym(JU,SymU), ManifoldL, ManifoldU,
                                         -10000, nil, nil, nil, nil, true, true, Observation.RemoveSpins);
        for i := 0 to LineList.nLines-1 do
          with LineList.Line[i] do
            if ((EigenvaluesU.Manifold = ManifoldU) and (IndexU = AindexU) and (IndexL = AindexL) and (EigenvaluesU.J = JU) and (EigenvaluesU.Sym = SymU) ) { TODO: -cField: Sym test fails with Field }
               or ((EigenvaluesU.Manifold = ManifoldL) and (IndexL = AindexU) and (IndexU = AindexL) and (EigenvaluesL.J = JU) and (EigenvaluesL.Sym = SymU)) then begin
              Value := ConditionSimulation.LineIntensity(LineList.Line[i], LineList.Units, Problem, Observation.RemoveSpins);
              if Assigned(Problem) then begin
                Result := False;
                WriteLn(Log, Problem.Message);
                Problem.Free;
              end else if IsNanOrInf(Value) then begin
                LineList.ShowLine(i, 0, LineList.Units, LineList.Units, True, Problem);
                Problem.Free;
                Result := False;
              end;
              Line := LineList.Line[i];
              Exit;
            end;
        if Mixture.AllowZeroI then
          Value := 0
        else begin
          WriteLn(Log, 'Transition not found during fitting - line number ', LineNumber, ' of ', FileName, ':', Name);
          for i := 0 to LineList.nLines-1 do
            LineList.ShowLine(i, 0, ConditionSimulation.PlotUnits, ConditionSimulation.PlotUnits, False, Problem);
          Problem.Free;
          Result := False;
        end;
      finally
        LineList.Free;
        if ConditionSimulation <> Mixture.Simulations[0] then
          ConditionSimulation.Free;
      end;
    end;
  end;
end;

function CalculateObs(var Observation: TObservation; out Value: Double; out RegeneratedLabel: string; RegenerateLabel:Boolean; AObsType: TObsType; Wait: Boolean): Boolean;
var
  EigU, EigL: TEigenvalues;
  Line: TLine;
begin
  Result := CalculateObsEx(Observation, Value, RegeneratedLabel, RegenerateLabel, AObsType, EigU, EigL, Line, Wait)
end;

function FindObservationColour(const Obs: TObservation; Upper: Boolean):TColor;
  procedure FindColour(Manifold: TManifold; J: TQuantumNumber; Sym: Shortint; Ind: Integer; IsUpper: Boolean);
  var
    Ev: TEigenvalues;
  begin
    Ev := nil;
    with Obs do begin
      if not Assigned(Manifold) or (ObsType in [obParameter,obExpression]) or (Ind <= 0) then
        Exit;
      if (Ind = FindEigenvaluesAndIndex(Ev, Manifold, J, Sym, Ind, Name,
                                       not IsUpper, LineNumber, FileName, RemoveSpins, QuantumNumberFormat, ExField))
         and Assigned(Ev) then
      Result := Ev.DominantBasis[Ind].State.FindColour;
    end;
  end;
begin
  Result := Obs.Colour;
  if Result <> clNone then
    Exit;
  with Obs do begin
    if Upper or (ObsType = obEUpper) then
      FindColour(ManifoldU, JU, SymU, AIndexU, True)
    else
      FindColour(ManifoldL, JL, SymL, AIndexL, False);
    if (Result = clNone) and (ObsType in [obLine, obIntensity, obCDLower, obCDUpper]) then
      if not Upper then { Try other state for colour }
        FindColour(ManifoldU, JU, SymU, AIndexU, True)
      else
        FindColour(ManifoldL, JL, SymL, AIndexL, False);
  end;
  if Result = clNone then
    Result := clBlack;
end;

function TLineFitter.CalculateAllObs(v: TColumnVector; Observations: TObservationArray; Start: Integer): Boolean;
var
  i, n, Block, BlockSize, iWorker: Integer;
  val: Double;
  Worker: TPgopherWorker;
  RLab: string;
  Problem: Exception;
begin
  Result := True;
  n := Length(Observations)-Start;
  Problem := nil;
  if PreCalcPartitionFunction then
    Mixture.Simulations[0].CalculatePartitionFunctions(Problem);
  if Assigned(Problem) then begin
    Result := False;
    WriteLn(Log, Problem.Message);
    Problem.Free;
    Exit;
  end;
  PushProgress('Calculating Observations',n);
  try
    if (nWorkers > 0) and (n > 1) and IsMainThread then begin
      { Parallel algorithm }
      { Heuristic to maximise independent calculations:
        make large steps through observations array }
      {$IFDEF ALG1}
      for Block := 0 to BlockSize-1 do begin
        SetProgress(Block, BlockSize-1);
        i := Block;
        while (i < nObs) and Result do begin
          Worker := FindWorkerWait as TPgopherWorker;
          if Assigned(Worker) and Result then
            Worker.CalculateObs(@Obs[i], v.Item[i+1], @Result)
          else begin
            { Something has failed }
            Result := False;
            if Assigned(Worker) then
              Worker.Skip;
          end;
          Inc(i, BlockSize);
        end;
        if not Result then
          Break;
      end;
      {$ENDIF}
      {$IFDEF ALG2}
      DivMod(nObs, nWorkers, BlockSize, i);
      //Slower on test
      if i > 0 then
        Inc(BlockSize);
      Block := 0;
      for i := 1 to nWorkers do begin
        Worker := FindWorkerWait as TPgopherWorker;
        if Assigned(Worker) and Result then
          Worker.CalculateObsBlock(Block, BlockSize, v, Obs, @Result)
        else begin
          { Something has failed }
          Result := False;
          if Assigned(Worker) then
            Worker.Skip;
        end;
        Inc(Block, BlockSize);
      end;
      {$ENDIF}
      { Not entirely clear why, but for CN test case, this gives the fastest results on test system
        May well depend on specific machine, though }
      BlockSize := n div nWorkers;
      if BlockSize = 0 then
        BlockSize := n; { Less than one observation per worker }
      for Block := 0 to BlockSize-1 do begin
        i := Block + (Start div BlockSize) * BlockSize;
        if i < Start then
          Inc(i, BlockSize);
        if i < Length(Observations) then begin
          Worker := FindWorkerWait as TPgopherWorker;
          if Assigned(Worker) and Result then
            Worker.CalculateObsBlock(i, -BlockSize, v, Observations, @Result)
          else begin
            { Something has failed }
            Result := False;
            if Assigned(Worker) then
              Worker.Skip;
          end
        end;
      end;
      WaitNoBusyWorkers;
      if not Result then
        for Block := 0 to BlockSize-1 do
          for iWorker := 0 to nWorkers do begin
            i := Block + iWorker*BlockSize;
            if (i >= Start) and (i < n) and IsNanOrInf(v[i+1]) then
              raise EReportException.CreateFmt('Calculated value of observation %d is %g',[i+1, v[i+1]]);
          end;
    end else { Serial algorithm }
      for i := Start to High(Observations) do begin
        SetProgress(i-Start);
        Result := Result and CalculateObs(Observations[i], val, RLab, False, Observations[i].ObsType);
        if IsNanOrInf(val) then
          raise EReportException.CreateFmt('Calculated value of observation %d is %g',[i+1, val]);
        v[i+1] := val;
      end;
  finally
    PopProgress;
  end;
end;

function TLineFitter.CalculateAllObs(v: TColumnVector): Boolean;
begin
  Result := CalculateAllObs(v, Obs, 0);
end;

constructor TLineFitter.Create(var TheParameters: TParameters;
  var TheObservations: TObservationArray; TheMixture: TMixture;
  ACheckDerivatives: Double);
var
  i: Integer;
  viInfo: TValueInfo;
  FullLabel: string;
  Manifold: TManifold;
begin
  ApplyConstraints;
  CheckConstraints(Parameters);
  Parameters := TheParameters;
  Obs := TheObservations;
  inherited Create(Length(Parameters), Length(Obs), TheMixture, ACheckDerivatives);
  Precision := Mixture.Precision;
  MaxCorrelatedObservations := 0;
  { Work out widths of fields for printout }
  MoleculeChanged := False;
  LowerChanged := False;
  UpperChanged := False;
  LargestLower := '';
  LargestUpper := '';
  LargestMolecule := '';
  LargestName := '';
  SymWidth := 1;
  HasEField := False;
  HasBField := False;
  HasJ := False;
  HasF := False;
  PreCalcPartitionFunction := False;
  for i := 0 to nObs-1 do
    with Obs[i] do begin
      if ObsType = obParameter then begin
        viInfo.MNode := MNode;
        viInfo.viType := TValueInfoType(IndexL);
        viInfo.Index := IndexU;
        FullLabel := viInfo.MNode.UniqueName+'.'+viInfo.ValueName;
        if Length(FullLabel) > Length(LargestName) then
          LargestName := FullLabel;
        Continue;
      end;
      if ObsType = obExpression then begin
        FullLabel := (MNode as TExpression).Text;
        if Length(FullLabel) > Length(LargestName) then
          LargestName := FullLabel;
        Continue;
      end;
      if ObsType <> obEUpper then
        Manifold := ManifoldL
      else
        Manifold := ManifoldU;
      if (ObsType = obIntensity)  then
        PreCalcPartitionFunction := PreCalcPartitionFunction or IsNormalized(Mixture.Simulations[0].IntensityUnits);
      MolName := Manifold.Molecule.Name;
      SymWidth := Max(SymWidth, Manifold.Molecule.SymLength);
      if LargestMolecule <> MolName then begin
        if Length(LargestMolecule) > 0 then
          MoleculeChanged := True;
        if Length(LargestMolecule) < Length(MolName) then
          LargestMolecule := MolName;
      end;
      if (ObsType <> obEUpper) and (LargestLower <> ManifoldL.Name) then begin
        if Length(LargestLower) > 0 then
          LowerChanged := True;
        if Length(LargestLower) < Length(ManifoldL.Name) then
          LargestLower := ManifoldL.Name
      end;
      if (ObsType <> obELower) and (LargestUpper <> ManifoldU.Name) then begin
        if Length(LargestUpper) > 0 then
          UpperChanged := True;
        if Length(LargestUpper) < Length(ManifoldU.Name) then
          LargestUpper := ManifoldU.Name
      end;
      if IsNonZero(ExField) then begin
        HasEField := HasEField or (ExField.E <> 0);
        HasBField := HasBField or (ExField.B <> 0);
      end else if not (Manifold is TVibrationalManifold) then
        if Length(Manifold.Molecule.HyperfineBasisStates)-RemoveSpins > 1 then
          HasF := True
        else
          HasJ := True;
      MaxCorrelatedObservations := Max(MaxCorrelatedObservations, Length(Correlation));
    end;
  if MaxCorrelatedObservations > 0 then
    SetLength(Correlations,nObs-1);
end;

procedure TLineFitter.ShowResidualHeader(var F: Text; Title: string; ShowCalc: Boolean);
var
  i: Integer;
begin
  WriteLn(F);
  Write(F, Title);
  if MoleculeChanged then begin
    if Length(LargestMolecule) < 3 then
      LargestMolecule := 'Mol';
  end else begin
    if LargestMolecule <> '' then
      Write(F, ' for ', LargestMolecule);
    LargestMolecule := '';
  end;
  if not (LowerChanged or UpperChanged) then
    if (LargestLower = '') or (LargestUpper = LargestLower) then begin
      if LargestUpper <> '' then
        Write(F, ' for the ', LargestUpper, ' manifold')
    end else if LargestUpper = '' then
      Write(F, ' for the ', LargestLower, ' manifold')
    else
      Write(F, ' for the ', LargestUpper, ' - ', LargestLower, ' Transition');
  WriteLn(F);
  Header := '';
  if Length(LargestMolecule) > 0 then
    Header := Copy('Molecule', 1, Length(LargestMolecule)) + ' ';
  if UpperChanged then
    Header := Header + StringOfChar(' ', Length(LargestUpper)+1);
  JHeader := '';
  if HasJ then
    JHeader := JHeader + 'J';
  if HasF then
    JHeader := JHeader + 'F';
  if HasBField or HasEField then
    JHeader := JHeader + 'M';
  if JHeader = '' then
    JHeader := '-'; { Vibrational only }
  if Mixture.QuantumNumberFormat = qftJ then
    JHeader := '  '+JHeader;
  Header := Header + '  ' + JHeader + '''' + 'S''' + StringOfChar(' ', SymWidth+1) + '#''';
  if LowerChanged then
    Header := Header + StringOfChar(' ', Length(LargestLower)+1);
  Header := Header + '  ' + JHeader + '"' + 'S"' + StringOfChar(' ', SymWidth+1) + '#"';
  if HasEField then
    Header := Header + ' E        ';
  if HasBField then
    Header := Header + '  B       ';
  if Length(LargestName)+2 > Length(Header) then
    Header := Header + StringOfChar(' ',Length(LargestName) - Length(Header) + 2);
  Write(F, Header, '     Observed');
  if ShowCalc then
    Write(F, '    Calculated','      Obs-Calc');
  Write(F, ' StdDev ');
  for i := 1 to MaxCorrelatedObservations do
    Write(F, ' Correl');
  WriteLn(F);
end;

procedure TLineFitter.ShowDerivative(var F: Text; Index: Integer);
var
  j: Integer;
begin
  for j := 1 to nPar do
    Write(F, DTosMA(a[Index+1,j],Max(14,Length(Parameters[j-1].Name)+1),-1));
  WriteLn(F);
  if Assigned(AlternateFitter) then
    (AlternateFitter as TLineFitter).ShowDerivative(F, Index);
end;

procedure TLineFitter.ShowDerivativeHeader(var F: Text; Title: string);
var
  i: Integer;
begin
  Write(F, Title);
  if Assigned(AlternateFitter) then
    WriteLn(F, '; alternate derivatives listed second')
  else
    WriteLn(F);
  for i := 0 to nPar-1 do
    Write(F, ' ', Parameters[i].Name:13);
  WriteLn(F);
end;

procedure TLineFitter.ShowResidual(var F: Text; Index: Integer; Calc: Double; ShowCalc: Boolean);
var
  viInfo: TValueInfo;
  FullLabel, RLab: string;
  j, wl: Integer;
  vo: Double;
begin
  with Obs[Index] do begin
    if ObsType = obParameter then begin
      if Length(LargestMolecule) > 0 then
        Write(F, ' ':Length(LargestMolecule)+1);
      viInfo.MNode := MNode;
      viInfo.viType := TValueInfoType(IndexL);
      viInfo.Index := IndexU;
      FullLabel := viInfo.MNode.UniqueName+'.'+viInfo.ValueName;
      Write(F, FullLabel, ' ':Length(Header)-Length(FullLabel)-1);
    end else if ObsType = obExpression then begin
      if Length(LargestMolecule) > 0 then
        Write(F, ' ':Length(LargestMolecule)+1);
      FullLabel := (MNode as TExpression).Text;
      Write(F, FullLabel, ' ':Length(Header)-Length(FullLabel)-1);
    end else begin
      if ObsType <> obELower then begin
        if Length(LargestMolecule) > 0 then
          Write(F, ManifoldU.Molecule.Name:Length(LargestMolecule), ' ');
        if UpperChanged then
          Write(F, ManifoldU.Name:Length(LargestUpper), ' ');
        Write(F, ManifoldU.Mixture.QuantumNumberToStr(JU):2+Length(JHeader),
                   ' ', PadRight(ManifoldU.Molecule.JAdjustedSymName(JU,SymU),SymWidth),
                   ' ', AIndexU:3, ' ')
      end else begin
        if Length(LargestMolecule) > 0 then
          Write(F, ManifoldL.Molecule.Name:Length(LargestMolecule), ' ');
        wl := 8+SymWidth+Length(JHeader);
        if UpperChanged then
          wl := wl + Length(LargestUpper)+1;
        Write(F, ' ':wl);
      end;
      if ObsType <> obEUpper then begin
        if LowerChanged then
            Write(F, ManifoldL.Name:Length(LargestLower), ' ');
        Write(F, ManifoldL.Mixture.QuantumNumberToStr(JL):2+Length(JHeader),
                   ' ', PadRight(ManifoldL.Molecule.JAdjustedSymName(JL,SymL),SymWidth),
                   ' ', AIndexL:3)
      end else begin
        wl := 7+SymWidth+Length(JHeader);
        if LowerChanged then
          wl := wl + Length(LargestLower)+1;
        Write(F, ' ':wl)
      end;
    end;
    if HasEField then
      Write(F, DtosMA(ExField.E,10,-1));
    if HasBField then
      Write(F, DtosMA(ExField.B,10,5));
    if ObsType = obIntensity then begin
      Write(F, Frequency:14);
      if ShowCalc then
        Write(F, Calc:14, Frequency-Calc:14);
    end else begin
      Write(F, AtLeast(Frequency,14,Precision));
      if ShowCalc then
        Write(F, AtLeast(Calc,14,Precision), AtLeast(Frequency-Calc,14,Precision));
    end;
    Write(F, DtosMA(Weight,8,-1));
    for j := 1 to MaxCorrelatedObservations do
      if (Index < nObs-1) and (j <= Length(Correlations[Index])) and (Correlations[Index][j-1] <> 0) then
        Write(F, DtosMA(Correlations[Index][j-1],7,4))
      else
        Write(F, ' ':7);
    if ofRegenerateLabel in Flags then begin
      CalculateObs(Obs[Index], vo, RLab, True, ObsType);
      Write(F, ' ', RLab, ' :':(LongestObsLabel+2-Length(Rlab)));
    end else begin
      Write(F, ' ', Name, ' :':(LongestObsLabel+2-Length(Name)));
    end;
    Writeln(F, LineNumber:4, ':', FileName);
 end;
end;

function TLineFitter.PrepareCalculated:Boolean;
var
  i: Integer;
begin
  TStart(PrepareCalculatedTime);
  Result := Length(Correlations) <> 0;
  if not CalculateAllObs(omc) then
    if not CheckForAbort then
      raise EReportException.Create('Transitions not found - see log for details');
  for i := 0 to nObs-1 do
    with Obs[i] do begin
      Calculated := omc[i+1];
      Flags := Flags + [ofCalculatedSet];
      Result := Result or (Weight <> 1);
      if Length(Correlation) > 0 then
        Correlations[i] := Correlation;
    end;
  TEnd(PrepareCalculatedTime);
end;

procedure TLineFitter.PrepareResiduals(var F: Text; Show: Boolean; ShowProgress: Boolean);
var
  i: Integer;
begin
  UnWtOldSD := 0;
  OldSD := 0;
  if ShowProgress then
    PushProgress('Residual Output');
  if Show then begin
    TStart(ResidualOutput);
    ShowResidualHeader(F, 'Residuals before fit', True);
  end;
  { Work out obs-calc, weights }
  for i := 0 to nObs-1 do begin
    if (ShowProgress) and (i mod 128 = 0) then
      SetProgress(i, nObs);
    if Show then
      ShowResidual(F, i, omc[i+1], True);
    omc[i+1] := omc[i+1] - Obs[i].Frequency;
    UnWtOldSD := UnWtOldSD + Sqr(omc[i+1]);
    if WeightedFit then begin
      weights[i+1] := Obs[i].Weight;
      OldSD := OldSD + Sqr(omc[i+1]/weights[i+1]);
    end;
  end;
  if Show then
    TEnd(ResidualOutput);
  if ShowProgress then
    PopProgress;
  if nObs > nPar then
    i := nObs - nPar
  else
    i := 1;
  UnWtOldSD := Sqrt(UnWtOldSD/i);
  if WeightedFit then
    if Length(Correlations) > 0 then
      OldSD := MaxDouble { Needs more work to calculate }
    else
      OldSD := Sqrt(OldSD/i)
  else
    OldSD := UnWtOldSD;
end;

{ TAutoFitFitter }

constructor TAutoFitFitter.Create(var TheParameters: TParameters;
  var TheObservations: TObservationArray; TheMixture: TMixture;
  ACheckDerivatives: Double);
begin
  inherited Create(TheParameters, TheObservations, TheMixture, 0);
  WeightedFit := PrepareCalculated;
  IntialCalculated := TColumnVector.Create(omc);
  CalculateDerivatives(Parameters);
  InitialDerivatives := TMatrix.Create(a);
  InitialParameters := Copy(Parameters);
  SetupMatrices;
end;

destructor TAutoFitFitter.Destroy;
begin
  FreeAndNil(InitialDerivatives);
  FreeAndNil(IntialCalculated);
  inherited;
end;

procedure TAutoFitFitter.Reset;
var
  i: Integer;
begin
  omc.Assign(IntialCalculated);
  a.Assign(InitialDerivatives);
  for i := 0 to nPar-1 do begin
    Parameters[i].Value := InitialParameters[i].Value;
    Parameters[i].Increment := InitialParameters[i].Increment;
  end;
end;

function TAutoFitFitter.MakeSVD:Double;
var
  obsb: TColumnVector;
  U: TMatrix;
begin
  if Length(Correlations) > 0 then
    raise NotYet;
  DecorrelateObservations(obsb, U, a, nil, nil, nil, False);
  ScaleParameters(ParameterScales, U);
  NLFit.MakeSVD(U, SVDw, SVD, False);
  if SVDw.Size > 0 then
    Result := SVDW[SVDW.Size]/SVDw[1]
  else
    Result := 0;
  U.Free;
end;

function FitLines(var Parameters: TParameters; Obs: TObservationArray; Precision: Integer;
                  Mixture: TMixture; CheckDerivatives: Double; out OldSD, NewSD: Double; Quiet: Boolean):string;
var
  f: TLineFitter;
begin
  TStart(LineFitTime);
  f := TLineFitter.Create(Parameters, Observations, Mixture, CheckDerivatives);
  f.Precision := Precision;
  try
    f.WeightedFit := f.PrepareCalculated;
    f.CalculateDerivatives(Parameters);
    f.PrepareResiduals(Log, not Quiet, True);
    Result := f.DoLeastSQ(Parameters, Precision, True, Quiet);
    OldSD := f.OldSD;
    NewSD := f.NewSD;
  finally
    f.Free;
    TEnd(LineFitTime);
  end;
end;

end.
