unit FormUtils;
{$I utils.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
{ Various routines common to pgopher forms }
interface

uses Classes, NGrid, Mixture, MNode, StdCtrls, Graphs, ComCtrls, NewSpin, MatrixVec, Global, ComplexMatrix;

procedure UpdateEditControl(Edit: TCustomEdit; Value: Double);
function ShowEigenvaluesInGrid(Eigenvalues: TEigenvalues; Grid:TNumberGrid; BasisLabel: TCustomLabel; nBasisStates: Integer = 0):string;
procedure ShowComplexMatrixInGrid(Matrix: TMatrix; Grid:TNumberGrid; LeadingCols:Integer; GrowOnly, Square, Transpose: Boolean; LeadingRows: Integer; Factor: Double = 1);
function ShowGraphInfo(G: TGraphD; const PlottedPoints: TPlottedPoints; OnGetPointInfo: TOnGetPointInfo): string;
procedure EditObservation(Obs: TObservation);

implementation

uses Math, Obs, SysUtils, MathSup, LineFrm, SlaveEdit, Profile, Controls;

var
 FindLargestCoefs: string = 'FindLargestCoefs';
 DisplayEigenvectors: string = 'DisplayEigenvectors';

procedure UpdateEditControl(Edit: TCustomEdit; Value: Double);
var
  Update: Boolean;
begin
  Update := (Edit.Owner is TControl) and not TControl(Edit.Owner).Visible;
  Update := Update or not Edit.HandleAllocated or not Edit.Focused;
  if Update or not Edit.Modified then begin
    Edit.Text := FloatToStr(Value);
    { TODO: Trim to displayable number of figures? }
    Edit.Modified := False;
  end;
end;

function ShowEigenvaluesInGrid(Eigenvalues: TEigenvalues; Grid:TNumberGrid; BasisLabel: TCustomLabel; nBasisStates: Integer):string;
var
  Woff, i, ii, nCof: Integer;
  Sorter: TEigenvectorSorter;
  z: Complex;
  IsComplex: Boolean;
begin
  TStart(DisplayEigenvectors);
  with EigenValues do begin
    if Assigned(Widths) then
      Woff := 1
    else
      Woff := 0;
    Grid.RowCount := Size + 1;
    nBasisStates := Min(nBasisStates,Size);
    IsComplex := IsComplexMatrix(Eigenvalues.C);
    if nBasisStates = 0 then begin
      Grid.ColCount := Size + 3 + Woff;
      Sorter := nil;
    end else begin
      Sorter := TEigenvectorSorter.Create(Eigenvalues.C);
      nCof := 2 + Ord(IsComplex);
      Grid.ColCount := nCof*nBasisStates + 3 + Woff;
    end;
    Grid.Title[2] := 'E';
    if Woff > 0 then
      Grid.Title[3] := 'Width';
    for i := 1 to Size do begin
      Grid.Lbl[i] := DominantBasis[i].StateLabel(Eigenvalues.Manifold.Mixture.QuantumNumberFormat);
      Grid.Value[i,2] := E[i];
      if Woff > 0 then
        Grid.Value[i, 3] := Widths[i];
      if nBasisStates = 0 then begin
        Grid.Title[i+Woff+2] := Basis[i].BasisLabel;
      end else begin
        TStart(FindLargestCoefs);
        Sorter.Process(i, nBasisStates);
        TEnd(FindLargestCoefs);
        for ii := 0 to nBasisStates-1 do begin
          if IsComplex then begin
            z := C[Sorter.Indices[ii],i];
            Grid.Value[i,ii*nCof+3+Woff] := z.re;
            Grid.Value[i,ii*nCof+4+Woff] := z.im;
          end else
            Grid.Value[i,ii*nCof+3+Woff] := C.Re[Sorter.Indices[ii],i];
          Grid.Value[i,ii*nCof+2+nCof+Woff] := Sorter.Indices[ii];
        end;
      end;
    end;
    if nBasisStates > 0 then begin
      for i := 1 to nBasisStates do begin
        Grid.Title[nCof*(i-1)+3+Woff] := IntToStr(i);
        if IsComplex then begin
          Grid.Title[nCof*(i-1)+4+Woff] := 'i';
          Grid.Title[nCof*(i-1)+5+Woff] := ' ';
        end else
          Grid.Title[nCof*(i-1)+4+Woff] := ' ';
      end;
      for i := 1 to Grid.ColCount do
        if Grid.ColWidths[i] < 30 then
          Grid.ColWidths[i] := 30;
    end else
      ShowComplexMatrixInGrid(C, Grid, 2+Woff, Grid.Width < Grid.Parent.Width, False, True, 0);
    Result := 'Eigenvalues';
    if Eigenvalues.Sym >= 0 then
      Result := Result + ' for Sym = '+Manifold.Molecule.BothSymNames(Eigenvalues.J,Sym);
    Grid.Title[1] := Basis[1].StateLabelKey;
    Grid.AutoAdjustColumn(1,True,True);
    Grid.AutoAdjustColumn(0);
    BasisLabel.Caption := '| ' + Basis[1].BasisLabelKey + ' >';
  end;
  Sorter.Free;
  Grid.AutoDefaultRowHeight;
  TEnd(DisplayEigenvectors);
end;

function ShowGraphInfo(G: TGraphD; const PlottedPoints: TPlottedPoints; OnGetPointInfo: TOnGetPointInfo):string;
var
  i: Integer;
begin
  Result := Concat(RoundedStr(G.LastMouseX, (G.Xmax-G.Xmin)/G.Width), '  ',
                   RoundedStr(G.LastMouseY, (G.Ymax-G.Ymin)/G.Height));
  if G.AreaSelected then
    with G.SelectedMathArea do
      Result := Result + ', [' + RoundedStr(Right-Left, (G.Xmax-G.Xmin)/G.Width) + ', '
                               + RoundedStr(Bottom-Top, (G.Ymax-G.Ymin)/G.Height) + ']';
  if Assigned(OnGetPointInfo) then
    for i := 0 to High(PlottedPoints) do
      with PlottedPoints[i] do
        if G.CloseToX(G.LastMouseX, X) and G.CloseToY(G.LastMouseY, Y) then
          Result := Result + OnGetPointInfo(Index,Extra);
end;

procedure EditObservation(Obs: TObservation);
begin
  if Length(Obs.FileName) = 0 then
    Exit; { Filename not given }
  if Obs.FileName[1] = '<' then
    if Obs.FileName = LineListFileName then
      LineForm.HighlightRow(Obs.LineNumber)
    else
      Exit { Clipboard (or other source) }
  else
    LaunchEditor(Obs.FileName, Obs.LineNumber);
end;

procedure ShowMatrixInGrid(Matrix: TMatrix; Grid:TNumberGrid; LeadingCols:Integer; GrowOnly, Square, Transpose: Boolean; LeadingRows: Integer; Factor: Double = 1);
var
  r, c: Integer;
  v: Double;
begin
  if not Assigned(Matrix) then
    Grid.Clear
  else if Square then
    for c := 1 to Matrix.Colcount do begin
      for r := 1 to Matrix.Rowcount do
        if Transpose then
          Grid[r+LeadingRows,c+LeadingCols] := Sqr(Matrix[c,r]*Factor)
        else
          Grid[r+LeadingRows,c+LeadingCols] := Sqr(Matrix[r,c]*Factor);
      Grid.AutoAdjustColumn(c+LeadingCols,GrowOnly);
    end
  else
    for c := 1 to Matrix.Colcount do begin
      for r := 1 to Matrix.Rowcount do begin
        if Transpose then
          v := Matrix[c,r]
        else
          v := Matrix[r,c];
        Grid[r+LeadingRows,c+LeadingCols] := v*Factor;
      end;
      Grid.AutoAdjustColumn(c+LeadingCols,GrowOnly);
    end;
end;

procedure ShowComplexMatrixInGrid(Matrix: TMatrix; Grid:TNumberGrid; LeadingCols:Integer; GrowOnly, Square, Transpose: Boolean; LeadingRows: Integer; Factor: Double = 1);
var
  r, c, cb: Integer;
  Z: Complex;
  SumIm, v: Double;
  M: TComplexMatrix;
begin
  if not (Matrix is TComplexMatrix) or (TComplexMatrix(Matrix).Mode = cmReal) then
    ShowMatrixInGrid(Matrix, Grid, LeadingCols, GrowOnly, Square, Transpose, LeadingRows, Factor)
  else if not Assigned(Matrix) then
    Grid.Clear
  else begin
    M := TComplexMatrix(Matrix);
    if Square then
      for c := 1 to M.Colcount do begin
        for r := 1 to M.Rowcount do
          if Transpose then
            Grid[r+LeadingRows,c+LeadingCols] := M.SqrMod[c,r]*Sqr(Factor)
          else
            Grid[r+LeadingRows,c+LeadingCols] := M.SqrMod[r,c]*Sqr(Factor);
        Grid.AutoAdjustColumn(c+LeadingCols,GrowOnly);
      end
    else begin
      Grid.ColCount := Max(Grid.ColCount, 2*M.ColCount+LeadingCols+1);
      for c := M.ColCount downto 1 do begin
        SumIm := 0;
        cb := 2*c+LeadingCols-1;
        for r := 1 to M.Rowcount do begin
          if Transpose then
            Z := M[c,r]
          else
            Z := M[r,c];
          Grid[r+LeadingRows,cb] := Z.re*Factor;
          Grid[r+LeadingRows,cb+1] := Z.im*Factor;
          SumIm := SumIm + Abs(Z.im);
        end;
        Grid.Title[cb] := Grid.Title[c+LeadingCols];
        Grid.AutoAdjustColumn(cb,GrowOnly);
        Grid.Title[cb+1] := 'i ';
        if SumIm > 0 then
          Grid.ColWidths[cb+1] := Grid.ColWidths[cb]
        else
          Grid.AutoAdjustColumn(cb+1,False,True); { Empty imaginary part }
      end
    end;
  end;
  Grid.AutoDefaultRowHeight;
end;

end.
