unit Hitran;
{$I utils.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses Mixture, ExpPlotLoader, Classes, Obs, TextStrm, ExternalList, MathSup;

type

  THITRANReader = class(TObservationsReader)
    ForceGlobalClass: Integer;
    DiscardL: Boolean;
    constructor Create; override;
    function DoReadOneLine: Boolean; override;
    function HandleDirective(var Handled: Boolean):Boolean; override;
  end;

  TBaseHITRANLoader = class(TExperimentalPlotLoader)
  private
    EolLength: Integer;
    EolChar: AnsiChar;
  protected
    Buffer: AnsiString;
    LineLength: Integer;
    nPoints: Integer;
    P: Integer;
    procedure Setup; virtual; abstract;
    procedure AddLine; virtual; abstract;
    procedure Done; virtual; abstract;
    function GF(Width: Integer): Double;
  public
    function Load: Boolean; override;
  end;

  THITRANLoader = class(TBaseHITRANLoader)
  protected
    procedure Setup; override;
    procedure AddLine; override;
    procedure Done; override;
  end;

  TExternalHITRANLoader = class(TBaseHITRANLoader)
  protected
    Mixture: TMixture;
    ExternalMolecule: THITRANMolecule;
    FullImport: Boolean;
//    Parsums: TParsums;
    procedure Setup; override;
    procedure AddLine; override;
    procedure Done; override;
  end;

function IsHITRANfile(FileName: string):Boolean;

function ImportHITRANFile(S: TTextStream): TMixture;
//function ImportHITRANFile(S: TTextStream; AParsums: TParsums ): TMixture; overload;
//function ImportHITRANFile(AFileName, ParsumName: string): TMixture; overload;
//function LoadHITRANParsumFile(S: TFreeFormatStream):TParsums;

implementation

uses {$IFNDEF CONSOLE} IPlotLoader, {$ENDIF}
     MNode, HundA, SymT, OAsymT, QuantumNumber, StrUtils, SysUtils, Math, Graphics,
     Units, Worker, PartFunc, QuantumMath, StaticField, Profile, SimpleQnos, UCompat;

var
  HitranImportTime:string = 'Import HITRAN file';
  HITRANFindOrCreateNode: string = 'Find or create HITRAN state';
  HITRANAddEntry: string = 'Add HITRAN line';

function IsHitranBuffer(const buffer: string):Boolean;
var
  s, e: Integer;
  LineLength, nOK: Integer;
begin
  LineLength:= -1;
  nOK := 0;
  { Count number of 100 or 160 character lines }
  s := 1;
  while s <= Length(buffer) do begin
    e := s;
    while (e <= Length(buffer)) and not (buffer[e] in [#10,#13]) do
      Inc(e);
    if (e-s = 100) or (e-s = 160) then begin
      if LineLength < 0 then
        LineLength := e-s;
      if e-s = LineLength then
        Inc(nOK);
    end;
    while (e <= Length(buffer)) and (buffer[e] in [#10,#13]) do
      Inc(e);
    s := e;
  end;
  Result := nOK > 5;
end;

function IsHitranStream(Stream: TStream):Boolean;
var
  buffer: AnsiString;
  Position: Int64;
  nb: Integer;
begin
  Result := False;
  SetLength(Buffer,1024);
  Position := Stream.Position;
  nb := Stream.Read(buffer[1], Length(Buffer));
  Stream.Position := Position;
  if nb > 0 then begin
    SetLength(buffer,nb);
    Result := IsHitranBuffer(buffer);
  end;
end;

function IsHITRANfile(FileName: string):Boolean;
var
  Stream: TStream;
begin
  Result := False;
  { Quick test to see if the file might be a HITRAN .par file }
  try
    Stream := TFileStream.Create(FileName,fmOpenRead or DefaultShareMode);
    try
      Result := IsHitranStream(Stream);
    finally
      Stream.Free;
    end;
  except
  end;
end;

{ TBaseHITRANLoader }

function TBaseHITRANLoader.GF(Width: Integer): Double;
begin
  if not TryStrToFloatEx(Copy(Buffer,P,Width),Result) then
    FileErr('Bad Floating point value:'+Copy(Buffer,P,Width));
  Inc(P,Width);
end;

function TBaseHITRANLoader.Load: Boolean;
var
  nl, cr, e, nb: Integer;
  s: string;
  HasNumber: Boolean;
begin
  Result := False;
  if ((Ext = '') and not IsHitranStream(Stream)) or (Ext <> '.par') then
    Exit;
  LineLength := -1;
  LineNumber := 1;
  nPoints := 0;
  { Determine line length, end of line style }
  SetLength(Buffer, 162);
  nb := Stream.Read(Buffer[1],162);
  SetLength(Buffer, nb);
  if nb < 100 then begin
    FileErr('File too short');
    Exit;
  end;
  nl := Pos(#10,Buffer);
  cr := Pos(#13,Buffer);
  EolLength := 1;
  if nl > 0 then
    if cr > 0 then begin
      EolLength := 2;
      EolChar := Buffer[Min(nl,cr)];
    end else
      EolChar := #10
  else if cr > 0 then
    EolChar := #13
  else
    FileErr('Line too long');
  repeat
    { Parse line }
    { Is first non-blank character a number? }
    e := 1;
    while (e < nb) and (Buffer[e] = ' ') do
      Inc(e);
    HasNumber := (e < nb) and (Buffer[e] in ['0'..'9']);
    e := Pos(EolChar, Buffer)-1;
    if e < 0 then
      e := nb;  { No eol }
    if HasNumber or (LineLength < 0) and ((e=100) or (e=160)) then begin
      LineLength := e;
      Stream.Position := Stream.Position - nb + LineLength + EolLength;
      SetLength(Buffer, LineLength + EolLength);
      Setup;
    end;
    if not HasNumber or (e <> LineLength) or (e < 0) then begin
      s := LowerCase(Trim(Copy(Buffer,1,e)));
      if (s = 'end') or (s = 'stop') then
        Break;
      if (e > 0) and HasNumber then
        FileErr('Bad length to line');
      { Skip This line }
      if Pos(EolChar, Buffer) > 0 then
        Stream.Position := Stream.Position - (Int64(nb)-e) + EolLength
      else begin
        while (Stream.Read(Buffer[1],1) = 1) and (Buffer[1] <> EolChar) do
          ;
        if EolLength = 2 then
          Stream.Position := Stream.Position + 1;
      end;
    end else begin
      P := 1;
      AddLine;
      Inc(nPoints);
    end;
    Inc(LineNumber);
    nb := Stream.Read(Buffer[1],160+EolLength);
  until nb = 0;
  Result := nPoints > 0;
  Done;
end;

{ THITRANLoader }

procedure THITRANLoader.AddLine;
begin
  Inc(P,3);
  Connector.SetValue(0,0,nPoints,GF(12));
  Connector.SetValue(0,1,nPoints,GF(10));
end;

procedure THITRANLoader.Done;
begin
  if nPoints > 0 then begin
    Connector.SetLineList(0,True);
    Connector.SetLength(0,nPoints,False,True,True);
    Connector.SetIntensityUnits (cm2WavenumberperMolecule);
    Connector.SetScale(1);
  end;
end;

procedure THITRANLoader.Setup;
begin
  CheckAvailableChannels(2, Stream.Size div (Int64(LineLength) + EolLength), False);
end;

{ THITRANReader }

constructor THITRANReader.Create;
begin
  inherited;
  Units := cm1;
  ForceGlobalClass := 0;
end;

function THITRANReader.DoReadOneLine: Boolean;
var
  p, nErrorsBefore: Integer;
  Buffer, Branch, LName, UName: string;
  Molecule: TMolecule;
  Fp, Fpp: TQuantumNumber;
  GlobalClass: Integer;
  ExtraU, ExtraL: string;
  AStateU, AStateL: TState;
  ss: string;
  OtherObs: TObservation; { For case where, for example, both parities are assigned to one line }
  UseOtherObs: Integer;
  Str, ELow: Double;
  function StrToFloatEx1(const s: string):Double;
  begin
    if not TryStrToFloatEx(s, Result) then
      ReportError(' Bad number: '+s);
  end;
  function GetField(Width: Integer): string;
  begin
    Result := Copy(buffer, p, Width);
    Result := Trim(Result);
    Inc(p, Width);
  end;
  function GetQno(Width: Integer): TQuantumNumber;
  begin
    Result := Round(StrToFloatEx1(GetField(Width))*2);
  end;
  function GetOptionalQno(Width: Integer): TQuantumNumber;
  var
    s: string;
  begin
    s := Trim(GetField(Width));
    if s = '' then
      Result := Low(TQuantumNumber)
    else
      Result := Round(StrToFloatEx1(s)*2);
  end;
  function DetermineGlobalClass(Field: string): Integer;
  begin
    Result := ForceGlobalClass;
    if Result > 0 then
      Exit;
    if Molecule.Manifold[0] is TSymmetricManifold then begin
      case Length(Field) of
        9,10:  Result := 8;
        11,12: Result := 10;
      else
        Result := 11; { Free format name }
      end;
      Exit;
    end;
    if Molecule.Manifold[0] is TAsymmetricManifold then begin
      case Length(Field) of
        5,6:   Result := 5;
        11,12: Result := 9;
      else
        Result := 11; { Free format name }
      end;
      Exit;
    end;
    Field := Trim(Field);
    case Length(Field) of
      1, 2: if Molecule.Manifold[0] is TLinearManifold then
              Result := 1
            else
              Result := 10;
      3:    Result := 2; { Note 2004 documentation seems to be in error, at least for 2008 db - O2 is A1I7, so 8 used below }
      5,6:  Result := 6;
      7:    if Molecule.Manifold[0].State[0] is TLinear then
              Result := 4
            else
              Result := 8;
      8:    if (Molecule.Manifold[0].State[0] is TLinear)
               and (TLinear(Molecule.Manifold[0].State[0]).S > 0) then
              if TLinear(Molecule.Manifold[0].State[0]).S > 1 then
                Result := 2 { O2 }
              else
                Result := 3
            else begin
              {Could be 4 or 5 - only difference is position of possible spaces }
              Result := Pos(' ', Field);
              if (Result > 0) and not Odd(Length(Field)-Result) then
                Result := 5
              else
                Result := 4;
            end;
      9:  { if Molecule.Manifold[0] is TSymmetricManifold then
            Result := 8
           else }
            Result := 5;
      10:   Result := 8;
      13:   Result := 11;
      14,15:Result := 7;
      11,12: if Molecule.Manifold[0] is TLinearManifold then
               Result := 7
             else
               Result := 9; { or 10, but do not differentiate }
    else
      ReportError('Bad global quantum numbers format:'+Field);
    end;
  end;
  function ManifoldName(Field: string; out Extra: string): string;
    procedure ReadnCh(nCh, n:Integer);
    var
      i: string;
    begin
      repeat
        i := Trim(Copy(Field, Length(Field)-nCh+1, nCh));
        if Length(Field) > nCh then
          Delete(Field, Length(Field)-nCh+1, nCh)
        else
          Field := '';
        if Result = '' then
          Result := i
        else
          Result := i + '_' + Result;
        Dec(n);
      until (n <= 0) or (Length(Field) = 0);
    end;
  begin
    Result := '';
    Extra := '';
    case GlobalClass of
      1: ReadnCh(2,1);
      2: begin ReadnCh(2,1); Field := Trim(Field); ReadnCh(1,1); end;
      3: begin
           ReadnCh(2,1);
           Delete(Field, Length(Field)-1, 2);
           Extra := Copy(Field, Length(Field)-2, 3);
           Delete(Field, Length(Field)-2, 3);
           ReadnCh(1,1);
         end;
      4: ReadnCh(2,4);
      5: ReadnCh(2,5);
      6: ReadnCh(2,3);
      7: begin ReadnCh(1,3); ReadnCh(2,6); end;
      8: ReadnCh(2,5);
      9: ReadnCh(2,6);
     10: Result := Trim(Field);
     11: Result := ReplaceStr(Trim(Field),' ','_'); { Newer format NH3 }
    else
      ReportError('Bad global class:' + IntToStr(GlobalClass));
    end;
  end;
  { Local quantum number subroutines }
  procedure Local2_5_6(Multiplicity: Integer);
  var
    NL, NU: TQuantumNumber;
    NBranch, MagFlag: string;
    SwitchParity: Integer;
    function FindState(SName: string; Manifold: TManifold; State: TState; out IndexF:Smallint; out SymF: Shortint; FW: TQuantumNumber; SymW: Shortint; JW, NW: TQuantumNumber):string;
    var
      AlternateSym: ShortInt;
    begin
      Result := Manifold.FindStateLabel(IndexF, SymF, AlternateSym, FW, [], obs.QuantumNumberFormat, ExField, SymW, JW, NW, State, obs.RemoveSpins);
      if (SymW < 0) and (IndexF = -3) then begin
        { Parity not specified, so add both parities if there is one match of each }
        UseOtherObs := AlternateSym;
        Result := Manifold.FindStateLabel(IndexF, SymF, AlternateSym, FW, [], obs.QuantumNumberFormat, ExField, -SymF-1, JW, NW, State, obs.RemoveSpins);
        if IndexF <= 0 then
          UseOtherObs := -1;
      end;
      if IndexF < 0 then
        ReportError(SName+Result)
      else
        IndexF := -IndexF;
    end;
    function DetermineN(Manifold: TManifold; J: TQuantumNumber; const Extra: string): TQuantumNumber;
    var
      Linear: TLinear;
      Ninc: TQuantumNumber;
    begin
      Linear := Manifold.State[0] as TLinear;
      if Linear.A >= 0 then
        Ninc := -1
      else
        Ninc := +1;
      if Extra[1] = '1' then
        Result := J + Ninc
      else
        Result := J - Ninc;
      if Result < Linear.Lambda then
        Result := Linear.Lambda;
    end;
    procedure DetermineParity(const s: string; var Sym: Shortint; J: TQuantumNumber);
    begin
      case s[1] of
        '+': Sym := 0;
        '-': Sym := 1;
        'e', 'E': Sym := (J div 2) and 1;
        'f', 'F': Sym := 1 - ((J div 2) and 1);
        { TODO: What does g mean? }
        'd', 'q', 'g': begin Sym := -1; SwitchParity := 0; end;
        ' ': Sym := (J div 2) and 1;
      else
        ReportError('Bad symmetry:'+s);
        Exit;
      end;
    end;
    var
      sp: Integer;
      s, su: string;
  begin { Normal linear }
    { Group 2 Local Quantum numbers }
    MagFlag := GetField(10);
    Fp := GetOptionalQno(5);
    NL := -1;
    NU := -1;
    case Multiplicity of
      1: begin
           GetField(5);
           Branch := GetField(1);
           obs.JL := GetQno(3);
         end;
      2: begin
           { Format change 2004/2008, so make flexible }
           sp := p;
           { Skip at most 3 blanks }
           while (p < sp+3) and (Buffer[p] = ' ') do
             Inc(p);
           Branch := GetField(1);
           if (Buffer[p] in ['M'..'U','m'..'u']) and (p < sp+4) then
             Branch := Branch+GetField(1);
           obs.JL := GetQno(5);
           NL := DetermineN(obs.ManifoldL, obs.JL, ExtraL);
           if (Branch = '') or not IsDeltaJName(Branch[Length(Branch)], obs.JU) then begin
             ReportError('Bad branch:'+Branch);
             Exit;
           end;
           obs.JU := obs.JL + obs.JU;
           NU := DetermineN(obs.ManifoldU, obs.JU, ExtraU);
         end;
      3: begin
           GetField(1);
           NBranch := GetField(1);
           if (NBranch = '') or not IsDeltaJName(NBranch[1], NU) then begin
             ReportError('Bad delta J name:'+NBranch);
             Exit;
           end;
           NL := GetQno(3);
           NU := NL+NU;
           Branch := GetField(1);
           obs.JL := GetQno(3);
         end;
    end;
    if (Branch = '') or not IsDeltaJName(Branch[Length(Branch)], obs.JU) then begin
      ReportError('Bad branch:'+Branch);
      Exit;
    end;
    obs.JU := obs.JL + Obs.JU;
    su := ' ';
    if Multiplicity = 3 then begin
      Fpp := GetOptionalQno(5);
      s := GetField(1)+' '; { Sym }
    end else begin
      s := GetField(1)+' '; { Sym }
      if Buffer[p] in ['e','f'] then begin
        su := s;
        s := GetField(1)+ ' ';
      end;
      Fpp := GetOptionalQno(5);
    end;
    if MagFlag = 'm' then
      SwitchParity := 0
    else
      SwitchParity := 1;
    DetermineParity(s, Obs.SymL, Obs.JL);
    if su <> ' ' then
      DetermineParity(su, Obs.SymU, Obs.JU)
    else
      Obs.SymU := -1;
    if Fp >= 0 then begin
      if Molecule.nNuclei = 0 then begin
        ReportError('Hyperfine quantum numbers in file but not in molecule');
        Exit;
      end;
      if Obs.RemoveSpins = 0 then
        Obs.RemoveSpins := Obs.RemoveSpins + Max(Molecule.nNuclei-1,0); { Single hyperfine provided, so discard any others }
    end else begin
      if Obs.RemoveSpins = 0 then
        Obs.RemoveSpins := Obs.RemoveSpins + Molecule.nNuclei; { No hyperfine provided, so discard }
      Fp := Obs.JU;
      Fpp := Obs.JL;
    end;
    LName := FindState('Lower:', obs.ManifoldL, AStateL, obs.IndexL, obs.SymL, Fpp, obs.SymL, obs.JL, NL);
    if UseOtherObs >= 0 then begin
      OtherObs := Obs;
      OtherObs.Name := FindState('Other Lower:', obs.ManifoldL, AStateU, OtherObs.IndexL, OtherObs.SymL, Fpp, UseOtherObs, obs.JL, NL)
                       + ':' + obs.Name;
      OtherObs.JL := Fpp;
    end;
    obs.JL := Fpp;
    if Obs.SymU < 0 then    
      obs.SymU := SwitchParity xor obs.SymL;
    UName := FindState('Upper:', obs.ManifoldU, AStateU, obs.IndexU, obs.SymU, Fp, obs.SymU, obs.JU, NU);
    if UseOtherObs >= 0 then begin
      OtherObs.Name := FindState('Other Upper:', obs.ManifoldU, AStateU, OtherObs.IndexU, OtherObs.SymU, Fp, SwitchParity xor obs.SymU, obs.JU, NU)
                       + ' - ' + OtherObs.Name;
      OtherObs.JU := Fp;
    end;
    obs.JU := Fp;
    obs.Name := UName + ' - ' + LName + ':' + obs.Name;
  end;
  procedure Local4; { Symmetric Tops }
  {$DEFINE USESIMPLE}
  {$IFDEF USESIMPLE}
  var
    NQN, DefSym: Integer;
    Q: TSimpleQnos;
    function DoState(Lab: string; var Q: array of Integer):Boolean;
    var
      el: TQuantumNumber;
      SymS: string;
      SymW: Integer;
      FullSymmetryGiven, NewerFormat, ReadSpin: Boolean;
      v: Double;
    begin
      Result := True; { OK }
      ReadSpin := True;
      Q[0] := GetOptionalQno(3); { J }
      if Q[0] = Low(TQuantumNumber) then begin
        Q[0] := GetQno(3);
        ReadSpin := False;
      end;
      Q[1] := GetQno(3); { K }
      SymS := GetField(2); { l (older) or s/a (newer) }
      el := 0;
      NewerFormat := False;
      if SymS <> '' then
        if TryStrToFloatEx(SymS, v) then begin
          el := round(v*2);
          if el < 0 then
            el := 2
          else if el > 0 then
            el := 1;
          if (Q[1] = 0) or DiscardL then
            el := 0; { Discard any l given }
        end else
          NewerFormat := True;
      SymS := Trim(GetField(3));
      if NewerFormat then
        SymS := Trim(GetField(4));
      FullSymmetryGiven := False;
      SymW := -1;
      if Molecule.TryStrToAnySymName(Q[0], SymS, SymW) then
        FullSymmetryGiven := True
      else if (Molecule as TSymmetricMolecule).IsPartialSymmetry(SymS) then
        { For now, simply discard symmetry name }
      else if SymS <> '' then begin
        ReportError(Lab+'Unknown Symmetry:'+SymS);
        Result := False;
      end;
      if (SymW >= 0) or (el <> 0) then
        Q[1] := (Q[1]+100000)*1600 + el*100 + SymW + 1;
      Q[2] := Low(TQuantumNumber);
      if not NewerFormat then
        if ReadSpin then
          Q[2] := GetOptionalQno(4)
        else
          GetField(1); { Skip - special case for molecule 46 = SO3 }
      if (Q[2] < 0) or (Molecule.nNuclei-Obs.RemoveSpins <= 0) then
        Obs.RemoveSpins := Max(Obs.RemoveSpins,Molecule.nNuclei)
      else begin
        Obs.RemoveSpins := Max(Molecule.nNuclei-1,0); { Single hyperfine provided, so discard any others }
        NQN := 3;
      end;
    end;
  begin
    { Map to simple quantum number format }
    NQN := 2;
    if not (DoState('Upper:', Q.QNP) and DoState('Lower:', Q.QN)) then
      Exit;
    if ConvertQuantumNumbers(Obs.ManifoldU, AStateU, Obs.JU, Obs.SymU, Obs.IndexU,
                             Obs.ManifoldL, AStateL, Obs.JL, Obs.SymL, Obs.IndexL,
                             Obs.Name, Molecule, 2, NQN, Q, ZeroField, StateMap, nil, nil,
                             qftJ, Obs.RemoveSpins, 2, DefaultSymmetry, True) then
      ReportError(Obs.Name);
  {$ELSE}
  var
    LastSymW: Integer;
    US: string;
    function FindState(Lab: string; Manifold: TManifold; State: TState; out F: TQuantumNumber; out Sym: Shortint; out Index: Smallint): string;
    var
      Extra: array [0..1] of TQuantumNumber;
      SymS: string;
      J: TQuantumNumber;
      SymW, SymA: Integer;
      FullSymmetryGiven: Boolean;
      AlternateSym: ShortInt;
    begin
      Index := -1;
      J := GetQno(3);
      Extra[0] := GetQno(3); { K }
      Extra[1] := GetOptionalQno(2); { l }
      SymS := Trim(GetField(3));
      FullSymmetryGiven := False;
      if Molecule.TryStrToAnySymName(J, SymS, SymW) then
        FullSymmetryGiven := True
      else if SymS = 'E' then
        SymW := 2
      else if SymS = 'A+' then
        SymW := 0
      else if SymS = 'A-' then
        SymW := 1
      else if SymS = 'A' then
        if (LastSymW=0) or (LastSymW=1) then
          SymW := LastSymW { Guess that symmetry does not change - could be refined }
        else
          ReportError(Lab+'Ambiguous symmetry:A - specify A+ or A-')
      else if (SymS = '') or (SymS = 's') or (SymS = 'a') then
        if LastSymW < 0 then
          SymW := -1 { Accept any symmetry }
        else begin
          { Look for dipole allowed transition }
          if Odd(BranchRank div 2) then begin
            for SymA := 0 to Molecule.LastSym do
              if (Molecule as TSymmetricMolecule).DipoleAllowed(SymA, LastSymW) then begin
                SymW := SymA;
                Break;
              end;
          end else
            SymW := LastSymW;
        end
      else begin
        ReportError(Lab+'Unknown sym:'+SymS);
        Sym := -1;
        Exit;
      end;
      if not FullSymmetryGiven and (SymW in [0,1]) and Odd(J div 2) and (Length(SymS) = 2) then
        SymW := 1 - SymW; { May need further adjustment }
      F := GetOptionalQno(4);
      if (F < 0) or (Molecule.nNuclei-Obs.RemoveSpins <= 0) then begin
        F := J;
        Obs.RemoveSpins := Max(Obs.RemoveSpins,Molecule.nNuclei);
      end else
        Obs.RemoveSpins := Max(Molecule.nNuclei-1,0); { Single hyperfine provided, so discard any others }
      Result := Manifold.FindStateLabel(Index, Sym, AlternateSym, F, Extra, obs.QuantumNumberFormat, ExField, SymW, J, -1, State, obs.RemoveSpins);
      if (Index = -3) and (SymW = -1) then begin
        { Two matches - take first of presumably degenerate pair }
        if ObsType = obIntensity then
          { Only half the intensity in one symmetry }
          { TODO: Should use statistical weights for this }
          Obs.Frequency := Obs.Frequency / 2;
        SymW := Min(-Sym-1, AlternateSym);
        Result := Manifold.FindStateLabel(Index, Sym, AlternateSym, F, Extra, obs.QuantumNumberFormat, ExField, SymW, J, -1, State, obs.RemoveSpins);
        LastSymW := SymW;
      end else if (Index < 0) and (Extra[1] <> 0) and (Extra[1] <> Low(TQuantumNumber)) then begin
        { Try with other sign of l }
        Extra[1] := -Extra[1];
        Manifold.FindStateLabel(Index, Sym, AlternateSym, F, Extra, obs.QuantumNumberFormat, ExField, SymW, J, -1, State, obs.RemoveSpins);
      end;
      LastSymW := SymW;
      if Index < 0 then begin
        ReportError(Lab+Result);
        Result := '';
      end else
        Index := 0;
    end;
  begin
    LastSymW := -1;
    US := FindState('Upper:',obs.ManifoldU,AStateU,obs.JU,obs.SymU,obs.IndexU);
    if obs.IndexU >= 0 then
      obs.Name := US + ' - ' + FindState('Lower:',obs.ManifoldL,AStateL,obs.JL,obs.SymL,obs.IndexL) + ':' + obs.Name;
  {$ENDIF}
  end;
  procedure Local1; { Asymmetric Tops }
  var
    NU, NL, JU, JL: TQuantumNumber;
    Nam: string;
    ExtraU, ExtraL: array [0..1] of TQuantumNumber;
    AlternateSym: ShortInt;
    procedure HandleSym(out N: TQuantumNumber; var J, F: TQuantumNumber);
    var
      Sym: string;
    begin
      Sym := GetField(1);
      if Sym = '+' then begin
        N := J;
        J := N + 1;
      end else if Sym = '-' then begin
        N := J;
        J := N - 1;
      end else
        N := -1;
      if (Obs.RemoveSpins > Molecule.nNuclei) and (N >= 0) then
        J := N;
      if (F < 0) or (Molecule.nNuclei-Obs.RemoveSpins <= 0) then begin
        F := J;
        Obs.RemoveSpins := Max(Obs.RemoveSpins,Molecule.nNuclei);
      end else
        Obs.RemoveSpins := Max(Molecule.nNuclei-1,0); { Single hyperfine provided, so discard any others }
    end;
  begin
    { Asymmetric Tops }
    JU := GetQno(3);
    ExtraU[0] := GetQno(3);
    ExtraU[1] := GetQno(3);
    obs.JU := GetOptionalQno(5);
    HandleSym(NU, JU, obs.JU);
    JL := GetQno(3);
    ExtraL[0] := GetQno(3);
    ExtraL[1] := GetQno(3);
    obs.JL := GetOptionalQno(5);
    HandleSym(NL, JL, obs.JL);
    if (NU < 0) xor (NL < 0) then begin
      ReportError('N only specified for one state');
      Exit;
    end;
    Nam := obs.ManifoldL.FindStateLabel(obs.IndexL, obs.SymL, AlternateSym, obs.JL, ExtraL, obs.QuantumNumberFormat, ExField, -1, JL, NL, AStateL, obs.RemoveSpins);
    if Obs.IndexL < 0 then
        ReportError('Lower:'+Nam)
    else begin
      Obs.IndexL := -Obs.IndexL;
      Obs.Name := Nam + ':' + Obs.Name;
    end;
    Nam := obs.ManifoldU.FindStateLabel(obs.IndexU, obs.SymU, AlternateSym, obs.JU, ExtraU, obs.QuantumNumberFormat, ExField, -1, JU, NU, AStateU, obs.RemoveSpins);
    if Obs.IndexU < 0 then
        ReportError('Upper:'+Nam)
    else begin
      Obs.IndexU := -Obs.IndexU;
      Obs.Name := Nam + ' - ' + Obs.Name;
    end;
  end;
  function FindManifold(Name: string; DefaultManifold: TManifold; out State: TState):TManifold;
  var
    i: Integer;
  begin
    State := nil;
    Result := Molecule.ManifoldWithName(Name);
    if not Assigned(Result) and Assigned(DefaultManifold) then begin
      State := DefaultManifold.StateWithName(Name);
      if Assigned(State) then
        Result := DefaultManifold;
    end;
    if not Assigned(Result) then begin
      for i := 0 to Molecule.nManifolds - 1 do begin
        State := Molecule.Manifold[i].StateWithName(Name);
        if Assigned(State) then begin
          Result := Molecule.Manifold[i];
          Break;
        end;
      end;
    end;
    if not Assigned(Result) then
      Result := DefaultManifold;
  end;
var
  EinA: Double;
begin
  Result := False;
  Buffer := StringOfChar(' ',Max(f.Column-Length(f.Token),0)) + f.Token;
  Buffer := Buffer + f.RestOfLine;
  UseOtherObs := -1;
  if not ((Length(Buffer)=100) or (Length(Buffer)=160)) then
    ReportError('Bad line length')
  else begin
    p := 1;
    while (p < Length(Buffer)) and (Buffer[p] = ' ') do
      Inc(p);
    if (p >= Length(Buffer)) or not(Buffer[p] in ['0'..'9']) then
      Exit; { First non-blank character not number }
    p := 1;
    s := GetField(3);
    if Mixture.nMolecules = 1 then
      Molecule := DefaultMolecule
    else
      Molecule := Mixture.FindMolecule(s);
    if not Assigned(Molecule) then
      ReportError('Molecule not found:'+s)
    else begin
      Obs.Name := Buffer; { Overriden if hyperfine present }
      Obs.Frequency := StrToFloatEx1(GetField(12));
      Str := StrToFloatEx1(GetField(10));
      EinA := StrToFloatEx1(GetField(10));
      Inc(p,5+5);
      ELow := StrToFloatEx1(GetField(10));
      Inc(p,4+8);
      case ObsType of
        obLine, obCDLower, obCDUpper: ;
        obIntensity: if Mixture.Simulations[0].IntensityUnits = EinsteinA then
                       Obs.Frequency := EinA
                     else
                       Obs.Frequency := Str;
        obELower: Obs.Frequency := ELow;
        obEUpper: Obs.Frequency := Obs.Frequency + ELow;
      end;
      if Length(Buffer) = 100 then begin
        raise NotYet;
      end else begin
        { Determine Class - see Table 3 of HITRAN 2004 paper }
        s := GetField(15);
        GlobalClass := DetermineGlobalClass(s);
        ss := ManifoldName(s,ExtraU);
        obs.ManifoldU := FindManifold(ss, DefaultManifoldU, AStateU);
        if not Assigned(AStateU) then
          AStateU := StateU;
        if not Assigned(obs.ManifoldU) then begin
          ReportError('Upper Manifold/State not found:'+ss);
          Exit;
        end;
        s := GetField(15);
        ss := ManifoldName(s,ExtraL);
        obs.ManifoldL := FindManifold(ss, DefaultManifoldL, AStateL);
        if not Assigned(AStateL) then
          AStateL := StateL;
        if not Assigned(obs.ManifoldL) then begin
          ReportError('Lower Manifold/State not found:'+ss);
          Exit;
        end;
        nErrorsBefore := nErrors;
        if Molecule.Manifold[0] is TSymmetricManifold then
          Local4
        else if Molecule.Manifold[0] is TAsymmetricManifold then
          Local1
        else
          case GlobalClass of
            2: Local2_5_6(3);
            3: Local2_5_6(2);
            1,4,5,7: Local2_5_6(1);
          else
            ReportError('Unexpected Global Class '+IntToStr(GlobalClass));
          end;
        if nErrorsBefore <> nErrors then
          Exit;
        s := GetField(1)+' ';
        if ObsType = obLine then
          case s[1] of
            '1': obs.Weight := 1;
            '2': obs.Weight := 0.1;
            '3': obs.Weight := 0.01;
            '4': obs.Weight := 0.001;
            '5': obs.Weight := 0.0001;
            '6': obs.Weight := 0.00001;
          else
            obs.Weight := 10;
          end
        else
          obs.Weight := 1; { TODO: Use intensity accuracy information }
        if UseOtherObs >= 0 then
          AddObservation(OtherObs);
        Result := True;
      end;
    end;
  end;
end;

function THITRANReader.HandleDirective(var Handled: Boolean): Boolean;
begin
  Result := inherited HandleDirective(Handled);
  if not Handled and (LowerCase(s) = 'globalclass') then begin
    if f.GetFreeItem then
      ForceGlobalClass := f.IntegerToken
    else
      ForceGlobalClass := 1;
    Handled := True;
  end;
  if not Handled and (LowerCase(s) = 'discardl') then begin
    if f.GetFreeItem then
      DiscardL := f.BooleanToken
    else
      DiscardL := True;
    Handled := True;
  end;
end;

procedure TExternalHITRANLoader.AddLine;
var
  Entry: TExternalEntry;
  Lower, Upper: String;
  function F(Width:Integer):AnsiString;
  begin
    Result := Copy(Buffer,P,Width);
    Inc(P,Width);
  end;
  function FirstQuantum(s: string): TQuantumNumber;
  var
    i, L: Integer;
  begin
    { Determine quantum number format from first non-space character }
    i := 1;
    while (i < 5) and (s[i] = ' ') do
      Inc(i);
    case i of
      5: L := 1; // ____n.... -> Group 3, J up to 9
      4: L := 2; // ___nn.... -> Group 3, J 10 - 99
      3: if s[4] = ' ' then
           L := 1  // __n__n... => Group 1, 4, J up to 9
         else
           L := 3; // __nnn.... -> Group 3, J 99+
      2: L := 2; // _nn_?n... => Group 1, 4, J 10 - 99
      1: L := 3; // nnn??n... => Group 1, 4, J 99+
    end;
    if not TryStrToQuantumNumber(Copy(S,i,L),Result,qftJ) then
      Result := -1;
  end;
var
  MoleculeNumber: string;
  Isotopologue: string;
  Species: TSpecies;
  Molecule: THITRANMolecule;
  nGlobal, nLocal, nSkip: Integer;
  s: string;
  DeltaJ: TQuantumNumber;
  JStart, Jlength: Integer;
begin
  with Entry do begin
    MoleculeNumber := Trim(F(2));
    Isotopologue := F(1);
    if FullImport then begin
      Species := Mixture.FindOrCreateNode(MoleculeNumber, TSpecies) as TSpecies;
      Molecule := Species.FindOrCreateNode(MoleculeNumber+Isotopologue, THITRANMolecule) as THITRANMolecule;
    end else begin
      Molecule := ExternalMolecule;
      // QuantumNumbers := MoleculeNumber+Isotopologue;
    end;
    Frequency := GF(12);
    Intensity := GF(10);
    if FullImport and (LineLength > 100) then
      EinsteinA := GF(10) { Read A }
    else begin
      Inc(P,10); { Ignore transition moment }
      EinsteinA := 0;
    end;
    Width := GF(5); { Air broadened width, gamma(air) }
    Inc(P,5); { Skip Self broadening, gamma(self) }
    ELower := GF(10);
    Intensity := Intensity/IntensityFactor(Frequency, cm1, cm2WavenumberperMolecule, ttDipole)
                          /(Exp(-ELower/BoltzmannConstants[cm1]/296)-Exp(-(ELower+Frequency)/BoltzmannConstants[cm1]/296));
    Inc(P,4+8); { Skip T dep gamma(air), air induced shift }
    if LineLength = 100 then begin
      nGlobal := 3;
      nLocal := 9;
      nSkip := 3+6;
    end else begin
      nGlobal := 15;
      nLocal := 15;
      nSkip := 6+12+1;
    end;
    Upper := AnsiReplaceStr(Trim(F(nGlobal)), '  ', ' '); { Double spaces not preserved in MNode Names }
    Lower := AnsiReplaceStr(Trim(F(nGlobal)), '  ', ' ');
    if FullImport then begin
      TStart(HITRANFindOrCreateNode);
      ManifoldU := Molecule.FindOrCreateNode(Upper, THITRANManifold) as THITRANManifold;
      ManifoldL := Molecule.FindOrCreateNode(Lower, THITRANManifold) as THITRANManifold;
      TEnd(HITRANFindOrCreateNode);
      ManifoldU.Initial := True;
      ManifoldL.Initial := True;
      UpperQuanta := '';
      LowerQuanta := '';
    end else begin
      UpperQuanta := Upper + ' ';
      LowerQuanta := Lower + ' ';
      ManifoldL := nil;
      ManifoldU := nil;
    end;
    Upper := F(nLocal);
    Lower := F(nLocal);
    UpperQuanta := UpperQuanta + Trim(Upper);
    LowerQuanta := LowerQuanta + Trim(Lower);
    gL := 0; { Flag unset }
    gU := 0;
    JL := -1;
    JU := -1;
    Inc(P, nSkip);
    if LineLength > 100 then begin
      gU := Round(GF(7));
      gL := Round(GF(7));
      EinsteinA :=  EinsteinA * gU / IntensityFactor(Frequency, cm1, HzperMolecule, ttDipole); { Convert to Spol }
      s := Trim(Lower);
      DeltaJ := Low(TQuantumNumber);
      JLength := 3;
      if (Length(s) > 1) and IsDeltaJName(s[1],DeltaJ) then begin
        { Diatomic, so lower state contains branch }
        JStart := Pos(s[1],Lower);
        if JStart = 6 then
          JStart := 7 { Local quanta style 2 }
        else if (JStart = 2) and not IsDeltaJName(s[2],DeltaJ) then
          Inc(JStart,5) { Local quanta style 5 }
        else begin
          { Local quanta style 6 }
          Inc(JStart);
          if IsDeltaJName(s[2],DeltaJ) then
            Inc(JStart);
          JLength := 5;
        end;
        if TryStrToQuantumNumber(Copy(Lower,JStart,Jlength),JL,qftJ) then
          JU := JL + DeltaJ;
      end else begin
        JL := FirstQuantum(Lower);
        JU := FirstQuantum(Upper);
      end;
    end;
  end;
  Entry.LineNumber := LineNumber;
  TStart(HITRANAddEntry);
  Molecule.AddEntry(Entry);
  TEnd(HITRANAddEntry);
end;

procedure TExternalHITRANLoader.Done;
//var
//  i, j, k, mn: Integer;
//  Q: TInterpolatedPartitionFunction;
//  E: Exception;
begin
  Mixture.FixupProperties.Free; { Discard errors at this point }
//  for i := 0 to Mixture.nSpecies - 1 do
//    for j := 0 to Mixture.Species[i].Count - 1 do
//      with Mixture.Species[i].Isotopomer[j] do begin
//        Q := TInterpolatedPartitionFunction.Create(Mixture.Species[i].Isotopomer[j]);
//        Q.AutoZero := False;
//        mn := StrToInt(Name);
//        for k := 0 to High(Parsums) do
//          if Parsums[k].MoleculeNumber*10 + Parsums[k].Isotopologue = mn then begin
//            Q.Comment := Parsums[k].Name;
//            Q.Plot.Frequency := Parsums[k].T;
//            Q.Plot.Intensity := Parsums[k].Q;
//            Break;
//          end;
//      end;
end;

procedure TExternalHITRANLoader.Setup;
begin
  Mixture.Simulations[0].IntensityUnits := cm2WavenumberperMolecule;
end;

function ImportHITRANFile(S: TTextStream): TMixture;
var
  Dummy: TExperimentalPlotLoaderConnector;
begin
  Result := TMixture.Create(nil);
  Result.Simulations[0].Temperature := 296; { HITRAN default }
  Dummy := {$IFNDEF CONSOLE}TInteractivePlotLoaderConnector{$ELSE}TExperimentalPlotLoaderConnector{$ENDIF}.Create(nil,'');
  TStart(HitranImportTime);
  with TExternalHITRANLoader.Create(Dummy,S,'.par',cm1,cm2WavenumberperMolecule) do
    try
      Mixture := Result;
      FullImport := True;
//      Parsums := AParsums;
      DoLoad(Worker.Log, False);
    finally
      Free;
      TEnd(HitranImportTime);
    end;
end;

//function ImportHITRANFile(S: TTextStream): TMixture;
//begin
//  Result := ImportHITRANFile(S, nil);
//end;
//
//  TParsumEntry = record
//    Name: string;
//    MoleculeNumber: Integer;
//    Isotopologue: Integer;
//    T: TDoubleArray;
//    Q: TDoubleArray;
//  end;
//
//  TParsums = array of TParsumEntry;

//function LoadHITRANParsumFile(S: TFreeFormatStream):TParsums;
//var
//  nCol, n: Integer;
//  Value: Double;
//  T: TDoubleArray;
//  i: Integer;
//  LastMolecule, ThisMolecule: string;
//  nIso, nMol: Integer;
//begin
//  { Load HITRAN pasrsum.dat from global data }
//  { Read header line }
//  nCol := 0;
//  LastMolecule := '';
//  nMol := 0;
//  S.FlowToNextLine := False;
//  while not S.GetFreeItem do
//    S.NextLine; { Skip Temperature and blank lines }
//  while not S.AtEndOfLine do
//    if S.GetFreeItem then begin
//      SetLength(Result, Length(Result)+1);
//      Result[nCol].Name := s.Token;
//      i := Pos('_',s.Token);
//      if i > 0 then
//        ThisMolecule := Copy(s.Token, 1{$IFDEF MSWINDOWS}  {$ENDIF} {$IFDEF LINUX}  {$ENDIF}, i-1)
//      else
//        raise Exception.Create('Unexpected molecule name :'+s.Token);
//      if ThisMolecule <> LastMolecule then begin
//        nIso := 1;
//        Inc(nMol);
//        LastMolecule := ThisMolecule;
//      end else
//        Inc(nIso);
//      Result[nCol].MoleculeNumber := nMol;
//      Result[nCol].Isotopologue := nIso;
//      Inc(nCol);
//    end;
//  S.NextLine;
//  n := 0;
//  while not S.AtEndOfFile do begin
//    while not S.GetFreeItem and not S.AtEndOfFile do
//      S.NextLine; { Skip blank lines }
//    if not S.IsRealToken(Value) then
//      Break;
//    if Length(T) <= n  then
//      SetLength(T, n+3000);
//    T[n] := Value;
//    nCol := 0;
//    while not S.AtEndOfLine do
//      if S.GetFreeItem then begin
//        if not S.IsRealToken(Value) then
//          Break;
//        if Length(Result[nCol].Q) <= n  then
//          SetLength(Result[nCol].Q, n+3000);
//        Result[nCol].Q[n] := Value;
//        Inc(nCol);
//      end;
//    if nCol > 0 then
//      Inc(n)
//    else
//      Dec(n);
//    S.NextLine;
//  end;
//  SetLength(T,n);
//  for i := 0 to High(Result) do begin
//    Result[i].T := T;
//    SetLength(Result[i].Q,n);
//  end;
//end;

//function ImportHITRANFile(AFileName, ParsumName: string): TMixture;
//var
//  MStream: TMemoryStream;
//  Parsums: TParsums;
//  Stream: TFreeFormatStream;
//begin
//  Stream := nil;
//  MStream := TMemoryStream.Create;
//  try
//    MStream.LoadFromFile(ParsumName);
//    Stream := TFreeFormatStream.Create(MStream);
//    Parsums := LoadHITRANParsumFile(Stream as TFreeFormatStream);
//    FreeAndNil(Stream);
//    MStream.LoadFromFile(AFileName);
//    MStream.Position := 0;
//    Stream := TFreeFormatStream.Create(MStream);
//    Stream.Comments := True;
//    Result := ImportHITRANFile(Stream as TFreeFormatStream,Parsums);
//  finally
//    MStream.Free;
//    Stream.Free;
//  end;
//end;

initialization
  RegisterPlotLoader(THITRANLoader, 'HITRAN Line lists', '*.par');
end.

