unit ImportCALPGM;
{$I utils.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}

{ Import data file from Herb Pickett's CALPGM program suite }
interface

uses TextStrm, Mixture;

function ImportCALPGMFile(S: TTextStream): TMixture; overload;
function ImportCALPGMFile(S: TTextStream; MaybeOpenShell, Quiet: Boolean): TMixture; overload;

var
  AllCALPGMpert: Boolean = False; { Mainly for testing }
  DefaultClosed: Boolean = False;

implementation

uses {$IFNDEF CONSOLE} Dialogs, {$ENDIF}
     Math, Classes, SysUtils, HundA, OAsymT, SymT, Units, MNode, MathSup, Expr,
     QuantumNumber, Worker, Abelian, MatrixVec, pgDOM, UCompat, Progress, Profile,
     CODATA;

function ShowScale(Num, Den:Integer):string;
begin
  with EN([Abs(Num), '/', Den, 'sqrt', '*', Sign(Num)]) do begin
    Result := Text(True);
    Free;
  end;
end;

function MakeScale(Num, Den:Integer):Double;
begin
  Result := Sqrt(Abs(Num)/Den)*Sign(Num);
end;

type

  TCalOptionInfo = record
    CHR: Char;
    NVIB, KNMIN, KNMAX, IXX, IAX, WTPL, WTMN, EWT, DIAG, XOPT: Integer;
    SPIND, VSYM: Int64;
    Name: string;
  end;

  TCALPGMReader = class
  public
    Title, Comment: string;
    OI: TCalOptionInfo;
    OIs: array of TCalOptionInfo;
    { Parameter details }
    IDPAR: Int64;
    FF, I2, I1, NS, TYP, KSQ, NSQ, V2, V1: Integer;
    PAR, ERRPAR: Double;
    ID: Integer; { IDPAR less vibrational number }
    nPar: Integer; { Maximum number of parameters left to read }
    S: TTextStream;
    Mixture: TMixture;
    Quiet, MaybeOpenShell: Boolean;
    constructor Create(Stream: TTextStream; AQuiet: Boolean);
    function ReadOptions(FirstLine: Boolean): TCalOptionInfo;
    function ReadParameterLine(out LineOK:Boolean): Boolean;
  end;

  TMoleculeReader = class;

  { TIntReader }

  TIntReader = class
    R: TMoleculeReader;
    Molecule: TMolecule;
    Title, Comment: string;
    S: TTextStream;
    IRFLAG, TAG, FBGN, FEND, MAXV: Integer;
    QROT, STR0, STR1, FQLIM, TEMP: Double;
    IDIP: Int64;
    SYM, V1, V2, I1, TYP: Integer;
    DIPOLE: Double;
    ScaleNum, ScaleDen: Integer;
    { Parameters for current object }
    ManifoldU, ManifoldL: TManifold;
    StateU, StateL: TState;
    TransitionMoments: TTransitionMoments;
    TransitionMoment: TRootedMNodePair;
    constructor Create(Stream: TTextStream; AReader: TMoleculeReader);
    function ReadParameterLine(out LineOK:Boolean): Boolean;
    function FindState(const V: Integer; var State: TState; var Manifold: TManifold):Boolean;
    function MakeTM(TransitionMomentClass: TTransitionMomentClass): TTransitionMoment;
    function FindTM(TransitionMomentClass: TTransitionMomentClass; Component: TQuantumNumber): TTransitionMoment;
    procedure CompleteTM(TM: TRootedMNodePair);
    procedure ShowProblem;
    procedure HandleParameter;
  end;

  { TMoleculeReader }

  TMoleculeReader = class
  public
    R: TCALPGMReader;
    RI: TIntReader;
    nProblems: Integer;
    ElectronSpin: TQuantumNumber;
    ElectronSpinIndex, SpinDigits: Integer;
    ElectronSpinIndexKnown: Boolean;
    NuclearSpins: array of TQuantumNumber;
    nItot: Integer;
    SpinIndex: array of Integer;
    StateMap: array of TState;
    StateIndexMap: array of Integer;
    Deleted: array of Boolean;
    Parameters: TParameters;
    nParameters: Integer;
    ParameterDeleted: Boolean;
    CurrentTemplate: Integer;
    Isotopomers: array of TMolecule;
    TemplateOIs: array of TCalOptionInfo;
    Species: TSpecies;
    Verbose: Boolean;
    constructor Create(AReader: TCALPGMReader);
    destructor Destroy; override;
    procedure Read; virtual;
    procedure ReadInt(IntStream: TTextStream);
    procedure ShowProblem;
    procedure SetValues(MNode: TMNode; Index: Integer; Scale: Double; OtherPar: string = '');
    procedure CheckForDuplicatePerturbation(var MNode: TMNode);
    function AddState(Manifold: TManifold; Index: Integer; FlipSym: Boolean; OI: TCalOptionInfo):TState; virtual; abstract;
    function ConvertParameter(State: TState; out Scale: Double; out MNode: TMNode):Integer; virtual; abstract;
    function SParameter(IDI: Integer; State: TState; var Scale: Double; var MNode: TMNode):Integer; virtual; abstract;
    function ISParameter(INuc: Integer; IDI: Integer; State: TState; var Scale: Double; var MNode: TMNode):Integer; virtual; abstract;
    function IParameter(INuc: Integer; IDI: Integer; State: TState; var Scale: Double; var MNode: TMNode):Integer; virtual; abstract;
    function IIParameter(INuc1, INuc2: Integer; IDI: Integer; State: TState; var Scale: Double; var MNode: TMNode):Integer; virtual; abstract;
    function ConvertPerturbation(State1, State2: TState; out Scale: Double; out MNode: TMNode):Integer; virtual; abstract;
    procedure MakeTransitionMoment(Magnetic: Boolean); virtual; abstract;
    procedure AddSpins(const OI: TCalOptionInfo; Molecule: TMolecule);
    function SetTemplate(const OI: TCalOptionInfo; out Manifold: TManifold):Boolean; { True if flipped symmetries }
    function SwitchIsotopomer(var Molecule: TMolecule; OI: TCalOptionInfo; Flip: Boolean):TManifold; virtual; abstract;
    function ConvertParametersWithSpin(State: TState; out Scale: Double; out MNode: TMNode):Integer;
  end;

  { TAsymmetricMoleculeReader }

  TAsymmetricMoleculeReader = class(TMoleculeReader)
  public
    Molecule: TAsymmetricMolecule;
    ReductionKnown, SReduction: Boolean;
    procedure Read; override;
    function SetReduction(TheSReduction: Boolean):Boolean;
    function AddState(Manifold: TManifold; Index: Integer; FlipSym: Boolean; OI: TCalOptionInfo):TState; override;
    function ConvertParameter(State: TState; out Scale: Double; out MNode: TMNode):Integer; override;
    function SParameter(IDI: Integer; State: TState; var Scale: Double; var MNode: TMNode):Integer; override;
    function ISParameter(INuc: Integer; IDI: Integer; State: TState; var Scale: Double; var MNode: TMNode):Integer; override;
    function IParameter(INuc: Integer; IDI: Integer; State: TState; var Scale: Double; var MNode: TMNode):Integer; override;
    function IIParameter(INuc1, INuc2: Integer; IDI: Integer; State: TState; var Scale: Double; var MNode: TMNode):Integer; override;
    function ConvertPerturbation(State1, State2: TState; out Scale: Double; out MNode: TMNode):Integer; override;
    function ConvertPE(PE: OAsymT.TPertEquiv; State1, State2: TState; out Scale: Double; out MNode: TMNode):Integer;
    function MakePE:OAsymT.TPertEquiv;
    procedure MakeTransitionMoment(Magnetic: Boolean); override;
    function SwitchIsotopomer(var NewMolecule: TMolecule;  OI: TCalOptionInfo; Flip: Boolean):TManifold; override;
  end;

  { TSymmetricMoleculeReader }

  TSymmetricMoleculeReader = class(TMoleculeReader)
  public
    Molecule: TSymmetricMolecule;
    LastEState: Integer;
    ZetaEtaMsg: Boolean;
    procedure Read; override;
    function AddState(Manifold: TManifold; Index: Integer; FlipSym: Boolean; OI: TCalOptionInfo):TState; override;
    function ConvertParameter(State: TState; out Scale: Double; out MNode: TMNode):Integer; override;
    function SParameter(IDI: Integer; State: TState; var Scale: Double; var MNode: TMNode):Integer; override;
    function ISParameter(INuc: Integer; IDI: Integer; State: TState; var Scale: Double; var MNode: TMNode):Integer; override;
    function IParameter(INuc: Integer; IDI: Integer; State: TState; var Scale: Double; var MNode: TMNode):Integer; override;
    function IIParameter(INuc1, INuc2: Integer; IDI: Integer; State: TState; var Scale: Double; var MNode: TMNode):Integer; override;
    function ConvertPerturbation(State1, State2: TState; out Scale: Double; out MNode: TMNode):Integer; override;
    function ConvertPE(PE: SymT.TPertEquiv; State1, State2: TState; out Scale: Double; out MNode: TMNode):Integer;
    procedure MakeTransitionMoment(Magnetic: Boolean); override;
    function SwitchIsotopomer(var NewMolecule: TMolecule;  OI: TCalOptionInfo; Flip: Boolean):TManifold; override;
  end;

  { TLinearMoleculeReader }

  TLinearMoleculeReader = class(TMoleculeReader)
  public
    Molecule: TLinearMolecule;
    function AddState(Manifold: TManifold; Index: Integer; FlipSym: Boolean; OI: TCalOptionInfo):TState; override;
    function ConvertParameter(State: TState; out Scale: Double; out MNode: TMNode):Integer; override;
    function SParameter(IDI: Integer; State: TState; var Scale: Double; var MNode: TMNode):Integer; override;
    function ISParameter(INuc: Integer; IDI: Integer; State: TState; var Scale: Double; var MNode: TMNode):Integer; override;
    function IParameter(INuc: Integer; IDI: Integer; State: TState; var Scale: Double; var MNode: TMNode):Integer; override;
    function IIParameter(INuc1, INuc2: Integer; IDI: Integer; State: TState; var Scale: Double; var MNode: TMNode):Integer; override;
    function ConvertPerturbation(State1, State2: TState; out Scale: Double; out MNode: TMNode):Integer; override;
    function ConvertPE(PE: HundA.TPertEquiv; State1, State2: TState; out Scale: Double; out MNode: TMNode):Integer;
    function MakePE(State1, State2: TState):HundA.TPertEquiv;
    procedure MakeTransitionMoment(Magnetic: Boolean); override;
    function SwitchIsotopomer(var NewMolecule: TMolecule;  OI: TCalOptionInfo; Flip: Boolean):TManifold; override;
  end;

{ TIntReader }

constructor TIntReader.Create(Stream: TTextStream; AReader: TMoleculeReader);
begin
  S := Stream;
  R := AReader;
  Molecule := R.R.Mixture.Species[0].Isotopomer[0];
  Title := S.RestOfLine;
  S.NextLine;
  S.FlowToNextLine := False;
  IRFLAG := S.NextInteger(0);
  if (IRFLAG div 1000) mod 10 > 0 then
    Molecule.Mixture.Units := cm1;
  TAG := S.NextInteger(0);
  Molecule.Name := Format('C%d',[TAG]);
  QROT := S.NextDouble(1);
  FBGN := S.NextInteger(-1);
  if FBGN > 0 then
    Molecule.Jmin := 2*FBGN;
  FEND := S.NextInteger(999);
  Molecule.Jmax := 2*FEND;
  STR0 := S.NextDouble(0);
  if STR0 < 0 then
    Molecule.Mixture.Simulations[0].OThreshold := power(10,STR0);
  STR1 := S.NextDouble(0);
  FQLIM := S.NextDouble(1e3);
  Molecule.Mixture.Simulations[0].Fmax := Convert(FQLIM*1000, MHz, Molecule.Mixture.Simulations[0].PlotUnits);
  TEMP := S.NextDouble(300);
  Molecule.Mixture.Simulations[0].Temperature := TEMP;
  MAXV := S.NextInteger(999);
  S.NextLine;
end;

function TIntReader.ReadParameterLine(out LineOK: Boolean): Boolean;
var
  t: Int64;
  idstring: string;
  function ND(n: Integer = 10): Integer;
  begin
    Result := t mod n;
    t := t div n;
  end;
begin
  LineOK := True;
  Result := False;
  ScaleNum := 1;
  ScaleDen := 1;
  if not S.GetFreeItem or (S.Token = '') then
    Exit;
  idstring := S.Token;
  if idstring[Length(idstring)] = '.' then
    Delete(idstring, Length(idstring), 1);
  try
    IDIP := StrToInt64(idstring);
    t := Abs(IDIP);
    SYM := ND;
    if Abs(R.R.OI.NVIB) < 10 then begin
      V1 := ND;
      V2 := ND;
    end else if Abs(R.R.OI.NVIB) < 100 then begin
      V1 := ND(100);
      V2 := ND(100);
    end else begin
      V1 := ND(1000);
      V2 := ND(1000);
    end;
    I1 := ND;
    TYP := ND(100);
    S.NextDoubleVar(DIPOLE);
    Comment := S.RestOfLine;
  except
    on E: Exception do begin
      Writeln(Log, 'Format problem on line ', S.Line, ':', E.Message);
      LineOK := False;
    end;
  end;
  S.NextLine;
  Result := True; { Read a line }
end;

function TIntReader.FindState(const V: Integer; var State: TState; var Manifold: TManifold): Boolean;
var
  i, n: Integer;
begin
  if V < length(R.StateMap) then begin
    State := R.StateMap[V];
    Manifold := State.Manifold;
    Result := True;
  end else begin
    Write(Log, 'State ', V, ' not found - ');
    ShowProblem;
    Result := False;
  end;
end;

procedure TIntReader.ShowProblem;
begin
  Inc(R.nProblems);
  if not R.R.Quiet then
    WriteLn(Log, 'Problem with intensity parameter on line ', S.Line-1, ' of ', S.StreamName, ': ',
                  IDIP, ' ', DIPOLE, ' ', Comment);
end;

function TIntReader.MakeTM(TransitionMomentClass: TTransitionMomentClass): TTransitionMoment;
begin
  if TransitionMoments.Bra = ManifoldU then
    Result := TransitionMomentClass.Create(TransitionMoments, StateU.Name, StateL.Name)
  else
    Result := TransitionMomentClass.Create(TransitionMoments, StateL.Name, StateU.Name);
  TransitionMoment := Result;
end;

function TIntReader.FindTM(TransitionMomentClass: TTransitionMomentClass; Component: TQuantumNumber): TTransitionMoment;
var
  i: Integer;
  Found: Boolean;
begin
  for i := 0 to TransitionMoments.ComponentCount-1 do
    if TransitionMoments.TransitionMoment[i].InheritsFrom(TransitionMomentClass) then begin
      Result := TransitionMoments.TransitionMoment[i];
      if TransitionMoments.Bra = ManifoldU then
        Found := (Result.Bra = StateU) and (Result.Ket = StateL)
      else
        Found := (Result.Bra = StateL) and (Result.Ket = StateU);
      if Found and (Result is TSphericalTransitionMoment) then
        Found := Found and (TSphericalTransitionMoment(Result).Component = Component);
      if Found then
        Exit;
    end;
  Result := MakeTM(TransitionMomentClass);
  if Result is TSphericalTransitionMoment then
    TSphericalTransitionMoment(Result).Component := Component;
  if Result is TSingleTransitionMoment then
    TSingleTransitionMoment(Result).Strength := 0;
end;

procedure TIntReader.CompleteTM(TM: TRootedMNodePair);
var
  C: string;
begin
  C := IntToStr(IDIP);
  if ScaleNum <> ScaleDen then
    C := Format('%s * %s', [C, ShowScale(ScaleNum, ScaleDen)]);
  if Comment <> '' then
    C := Comment + ' ' + C;
  TM.Comment := C;
  if TM is TSingleTransitionMoment then
    TSingleTransitionMoment(TM).Strength := DIPOLE*MakeScale(ScaleNum, ScaleDen)
  else if TM is TPerturbation then
    TPerturbation(TM).Value := DIPOLE*MakeScale(ScaleNum, ScaleDen);

end;

procedure TIntReader.HandleParameter;
begin
  if FindState(V1,StateU,ManifoldU)
     and FindState(V2,StateL,ManifoldL) then begin
    { Find or create TransitionMoments object }
    TransitionMoments := Molecule.FindTransitionMoments(ManifoldU,ManifoldL);
    if not Assigned(TransitionMoments) then
      TransitionMoments := TTransitionMoments.Create(Molecule, ManifoldU.Name, ManifoldL.Name);
    Molecule.FixupProperties.Free;
    TransitionMoment := nil;
    if SYM > 3 then begin
      Write(Log, 'SYM =', SYM, ' - ');
      ShowProblem;
    end else begin
      R.MakeTransitionMoment((SYM = 0) or (I1 > 0));
      if TransitionMoment = nil then begin
        Write(Log, 'TYP =', TYP, ', SYM =', SYM, ' - ');
        ShowProblem;
      end else
        CompleteTM(TransitionMoment);
    end;
  end;
end;

{ TMoleculeReader }

constructor TMoleculeReader.Create(AReader: TCALPGMReader);
begin
  R := AReader;
  nProblems := 0;
  CurrentTemplate := -1;
  Species := R.Mixture.Species[0];
  ElectronSpinIndex := -1;
  //Verbose := True;
end;

destructor TMoleculeReader.Destroy;
begin
  inherited Destroy;
  RI.Free;
end;

procedure TMoleculeReader.ShowProblem;
begin
  Inc(nProblems);
  if not R.Quiet then
    WriteLn(Log, 'Problem with parameter on line ', R.S.Line-1, ': ',
                  R.IDPAR, ' ', R.PAR, ' ', R.ERRPAR, ' ', R.Comment);
end;

procedure TMoleculeReader.SetValues(MNode: TMNode; Index: Integer; Scale: Double; OtherPar: string);
begin
  if Index < 0 then begin
    ShowProblem;
    ParameterDeleted := True;
    Parameters[nParameters].MNode := nil;
  end else begin
    if (Abs(R.ERRPAR) > 1e-10) and (Abs(R.ERRPAR) < 1e17) then begin
      if (MNode.Parameters[Index] <> 0) then begin
        ShowProblem;
        if not R.Quiet then
          WriteLn(Log, 'Standard Deviation ignored as it provides multiple contribution to ', MNode.FullParameterName(Index));
      end;
      if OtherPar <> '' then begin
        ShowProblem;
        if not R.Quiet then
          WriteLn(Log, 'Standard Deviation ignored as it contributes to ', MNode.FullParameterName(Index), ' and ', OtherPar);
      end;
    end;
    MNode.Floating[Index] := Abs(R.ERRPAR) > 1e-16;
    MNode.Parameters[Index] := MNode.Parameters[Index] + Scale*R.PAR;
    if MNode.Floating[Index] then
      MNode.StdDev[Index] := Sqrt(Sqr(MNode.StdDev[Index]) + Sqr(Scale*R.ERRPAR));
    Parameters[nParameters].MNode := MNode;
    Parameters[nParameters].Value := Sign(Scale); { Needed to convert correlation matrix }
  end;
  Parameters[nParameters].Index := Index;
end;

procedure TMoleculeReader.CheckForDuplicatePerturbation(var MNode: TMNode);
var
  Manifold: TManifold;
  i: Integer;
  P: TPerturbation;
begin
  { Combine identical perturbations; value is handled elsewhere }
  if not (MNode is TPerturbation) then
    Exit;
  with MNode.Owner as TManifold do
    for i := 0 to nPerturbations - 1 do
      if (Perturbation[i] <> MNode) and Perturbation[i].SameSettingsAs(MNode) then begin
        P := Perturbation[i];
        P.Comment := P.Comment + '/' + MNode.Comment;
        MNode.Free;
        MNode := P;
        Exit;
      end;
end;

var
  NewManifoldTime: string = 'New Manifold';
  UpdateParametersTime: string = 'Update Parameters';
  MovePerturbationsTime: string = 'Move Perturbations';
  UpdateStateMapTime: string = 'Update State Map';
  FreeStateTime: string = 'Free State';


procedure TMoleculeReader.Read;
  procedure SeparateStates(Manifold: TManifold; Prefix: string);
  var
    i, j, k, n: Integer;
    XMLDocument: TXMLDocument;
    DOMElement: TDOMElement;
    NewManifold: TManifold;
    NewPerturbation: TPerturbation;
  begin
    { Separate out non-connected states into separate manifolds }
    if not Assigned(Manifold) then
      Exit;
    n := 0;
    PushProgress('Separating States '+Prefix,Manifold.nStates);
    try
      while  Manifold.nStates > 0 do begin
        { Peel off first remaining state each time round loop }
        i := 0;
        SetProgress(n);
        TStart(NewManifoldTime);
        XMLDocument := TXMLDocument.Create;
        DOMElement := Manifold.WriteProperties(XMLDocument,True,False);
        Manifold.State[i].WriteProperties(DOMElement);
        NewManifold := TMNodeClass(Manifold.ClassType).Create(Manifold.Owner) as TManifold;
        NewManifold.ReadProperties(XMLDocument.FirstChild);
        NewManifold.Name := Prefix+IntToStr(n);
        Inc(n);
        XMLDocument.Free;
        TEnd(NewManifoldTime);
        TStart(UpdateParametersTime);
        { Update parameters list to point to new object }
        for j := 0 to nParameters - 1 do
          if not Assigned(Parameters[j].MNode) then
            { Skip }
          else if Parameters[j].MNode = Manifold.State[i] then
            Parameters[j].MNode := NewManifold.State[0]
          else if Parameters[j].MNode.Owner = Manifold.State[i] then
            Parameters[j].MNode := NewManifold.State[0].Components[Parameters[j].MNode.ComponentIndex] as TMNode;
        TEnd(UpdateParametersTime);
        TStart(MovePerturbationsTime);
        { Now transfer diagonal perturbations }
        j := 0; { loop from start allowing for perturbations to go }
        while j < Manifold.nPerturbations do
          if Manifold.Perturbation[j].Bra = Manifold.State[i] then begin
            XMLDocument := TXMLDocument.Create;
            DOMElement := Manifold.Perturbation[j].WriteProperties(XMLDocument,True,False);
            NewPerturbation := TMNodeClass(Manifold.Perturbation[j].ClassType).Create(NewManifold) as TPerturbation;
            NewPerturbation.ReadProperties(XMLDocument.FirstChild);
            for k := 0 to nParameters - 1 do
              if Parameters[k].MNode = Manifold.Perturbation[j] then
                Parameters[k].MNode := NewPerturbation;
            Manifold.Perturbation[j].Free;
            XMLDocument.Free;
          end else
            Inc(j);
        TEnd(MovePerturbationsTime);
        TStart(UpdateStateMapTime);
        for j := 0 to High(StateMap) do
          if StateMap[j] = Manifold.State[i] then
            StateMap[j] := NewManifold.State[0];
        TEnd(UpdateStateMapTime);
        TStart(FreeStateTime);
        Manifold.State[i].Free;
        TEnd(FreeStateTime);
      end
    finally
      PopProgress;
    end;
    Manifold.Free;
  end;
var
  i, j, StateMin, StateMax, s, p: Integer;
  State: TState;
  LineOk, Flip: Boolean;
  MNode: TMNode;
  Index, nPerturbations, nCor, MaxState: Integer;
  Scale, v: Double;
  line, val: string;
  NewName: string;
  Cor, CorT: TMatrix;
  Molecule: TMolecule;
  Manifold: TManifold;
  OtherManifold: TManifold;
begin
  val := Trim(R.Title);
  j := 0;
  for i := 1 to Length(val) do { Guess for end of species name }
    if (val[i] in [',','!']) and (i > 2) then begin
      val := Trim(Copy(val,1,i-1));
      Break;
    end else if val[i] <> ' ' then
      j := 0
    else if j > 3 then begin
      val := Trim(Copy(val,1,i-1));
      Break;
    end else
      Inc(j);
  if val = '' then
    val := 'S';
  Species.Name := val;
  SetLength(StateMap, ABS(R.OI.NVIB));
  SetLength(StateIndexMap, ABS(R.OI.NVIB));
  SetLength(Deleted, ABS(R.OI.NVIB));
  for i := 0 to Abs(R.OI.NVIB) - 1 do
    StateIndexMap[i] := i;
  for i := 0 to Abs(R.OI.NVIB)-1 do begin
    Flip := SetTemplate(R.OIs[i],Manifold);
    State := AddState(Manifold,i,Flip,R.OIs[i]);
    NewName := R.OIs[i].Name;
    State.CleanName(NewName);
    if (NewName = '') or (Manifold.ComponentWithName(NewName,TState,nil) <> nil) then
      NewName := IntToStr(i);
    State.Name := NewName;
    State.Comment := R.OIs[i].Name;
    State.AdjustNuclei(True);
    for j := 0 to Manifold.Molecule.nNuclei-1 do
      State.NucleusWithSpin(j).Spin := NuclearSpins[j];
    for j := Manifold.Molecule.nNuclei-2 downto Manifold.Molecule.nNuclei-nItot do
      State.NucleusWithSpin(j).AsNext := True;
    StateMap[i] := State;
    if Verbose then begin
      WriteLn(Log, 'State ', i, ' = ', State.FullHeirarchicName);
      State.PrettyPrintSettings(Log, plDefault, '  ');
    end;
  end;
  { Delete unwanted states }
  i := Abs(R.OI.NVIB)-1; { Logical index }
  repeat
    Deleted[i] := StateIndexMap[i] <> i;
    if Deleted[i] then begin
      FreeAndNil(StateMap[i]);
      StateMap[i] := StateMap[StateIndexMap[i]];
      for j := i to Abs(R.OI.NVIB) - 1 do
        if StateIndexMap[j] >= i then
          Dec(StateIndexMap[j]);
    end;
    Dec(i);
  until i < 0;
  nParameters := 0;
  nPerturbations := 0;
  ParameterDeleted := False;
  while R.ReadParameterLine(LineOK) do
    if LineOK then begin
      if Verbose then
        WriteLn(Log, R.S.Line-1:6, ': ', R.IDPAR, ' ', R.PAR, ' ', R.ERRPAR, ' ', R.Comment);
      if R.S.Line mod 100 = 0 then
        SetProgress(R.S.Line, R.S.Line * R.S.Size div R.S.Position);
      if nParameters >= High(Parameters) then
        SetLength(Parameters, nParameters+20);
      StateMin := R.V1;
      StateMax := R.V2;
      if Abs(R.OI.NVIB) >= 100 then
        MaxState := 999
      else if Abs(R.OI.NVIB) >= 10 then
        MaxState := 99
      else
        MaxState := 9;
      if (R.V1 = MaxState) and (R.V2 = MaxState) then begin
        StateMin := 0;
        StateMax := Abs(R.OI.NVIB)-1;
      end else if R.V1 <> R.V2 then begin
        if Max(R.V1,R.V2) >= Abs(R.OI.NVIB) then begin
          if not R.Quiet then
            WriteLn(Log, 'Ignoring parameter on line ', R.S.Line-1, ': ', R.IDPAR, ' ', R.PAR, ' ', R.ERRPAR, ' ', R.Comment);
          ParameterDeleted := True;
          Continue; { Skip this parameter see c034001.par for example, though not imported correctly }
        end;
        Index := ConvertPerturbation(StateMap[R.V1], StateMap[R.V2], Scale, MNode);
        CheckForDuplicatePerturbation(MNode);
        SetValues(MNode, Index, Scale);
        if StateMap[R.V1] <> StateMap[R.V2] then
          Inc(nPerturbations);
        Continue;
      end;
      if StateMax >= Abs(R.OI.NVIB) then begin
        WriteLn(Log, 'State index out of bounds:', StateMax, '>', Abs(R.OI.NVIB)-1);
        ShowProblem;
      end else begin
        for i := StateMin to StateMax do
          if (StateMin = StateMax) or not Deleted[i] then begin
            Index := ConvertParametersWithSpin(StateMap[i], Scale, MNode);
            CheckForDuplicatePerturbation(MNode);
            SetValues(MNode, Index, Scale);
          end;
        Inc(nParameters);
      end;
    end else
      Inc(nProblems);
  SetLength(Parameters, nParameters);
  Cor := nil; { Only create matrix if needed - would be very large for some files }
  i := 1;
  j := 1;
  while not ParameterDeleted and not R.S.AtEndOfFile and (i <= nParameters) do begin
    { Read correlation coeficients }
    line := R.S.RestOfLine;
    s := 1;
    while s < Length(line) do begin
      p := s+1;
      while (p < Length(line)) and not (line[p] in [' ', '-']) do
        Inc(p);
      val := Trim(Copy(line,s,p-s));
      if val <> '' then
        if val[1] = '!' then
          Break
        else if TryStrToFloatEx(val, v) then begin
          if not Assigned(Cor) then begin
            Cor := TMatrix.Create(nParameters,nParameters);
            Cor.Zero;
          end;
          Cor[i,j] := v;
          Inc(j);
          if j > i then begin
            Inc(i);
            if i > nParameters then
              Break;
            j := 1;
          end;
        end else begin
          Inc(nProblems);
          WriteLn(Log, 'Problem with correlation matrix on line ', R.S.Line, ':', v, ' - is NPAR set correctly? (', R.nPar, ')');
        end;
      s := p;
    end;
    R.S.NextLine;
  end;
  for i := 0 to Species.ComponentCount-1 do
    if nPerturbations = 0 then begin
      Molecule := Species.Isotopomer[i];
      Manifold := Molecule.Manifold[0];
      if Molecule.nManifolds > 1 then
        OtherManifold := Molecule.Manifold[1]
      else
        OtherManifold := nil;
      { Doing it this way ensures state order is preserved }
      if Manifold.nStates > 1 then
        SeparateStates(Manifold, Manifold.Name);
      if Assigned(OtherManifold) and (OtherManifold.nStates > 1) then
        SeparateStates(OtherManifold, OtherManifold.Name);
    end;
  if Assigned(Cor) then begin
    CorT := Cor.Transpose;
    CorT.PostMultiply(Cor);
    R.Mixture.Correlations.FitDetails.Parameters := Parameters;
    R.Mixture.Correlations.AdjustSize;
    for i := 0 to nParameters - 1 do
      for j := 0 to i - 1 do
        R.Mixture.Correlations.FitDetails.Correlation[i][j] := CorT[i+1,j+1]*Parameters[i].Value*Parameters[j].Value;
    CorT.Free;
  end;
  Cor.Free;
  if R.Quiet then
    Exit;
end;

procedure TMoleculeReader.ReadInt(IntStream: TTextStream);
var
  LineOK: Boolean;
begin
  RI := TIntReader.Create(IntStream, Self);
  while RI.ReadParameterLine(LineOK) do
    if LineOK then
      RI.HandleParameter
    else
      Inc(nProblems);
end;

procedure TMoleculeReader.AddSpins(const OI: TCalOptionInfo; Molecule: TMolecule);
var
  SPIND, t: Int64;
  i: Integer;
  Index, SIndex: Integer;
  Buffer: string;
  HasSpins, NewMolecule: Boolean;
  function NextSpin: TQuantumNumber;
  begin
    Result := (t mod SpinDigits) - 1;
    t := t div SpinDigits;
  end;
  procedure StartSpin;
  begin
    t := SPIND;
    if t mod 10 = 0 then begin
      { Spin multiplicities are two digits }
      SpinDigits := 100;
      t := t div 10;
    end else
      SpinDigits := 10;
  end;
begin
  ElectronSpin := 0;
  SPIND := Abs(OI.SPIND);
  NewMolecule := Molecule.Name = '';
  if NewMolecule then begin
    Buffer := IntToStr(SPIND);
    while Species.ComponentWithName(Buffer, TMolecule, nil) <> nil do
      Buffer := Buffer + Format('_%d_%d', [OI.WTPL, OI.WTMN]);
    Molecule.Name := Buffer;
    Molecule.Comment := R.Title;
  end;
  //for i := 0 to High(R.OIs) do
  //  if ABS(R.OIs[i].SPIND) <> SPIND then begin
  //    //if not R.Quiet then
  //    //  WriteLn(Log, 'Inconsistent SPIND values: ', SPIND, ' <> ', ABS(R.OIs[i].SPIND), ' for state ', i+1,
  //    //               ' - Simulations with differing numbers of spins must be done in separate molecules in PGOPHER');
  //    SPIND := Max(SPIND, ABS(R.OIs[i].SPIND));
  //  end;
  HasSpins := False;
  StartSpin;
  while (t > 0) and not HasSpins do
    HasSpins := NextSpin > 0;
  if HasSpins then begin
    if not ElectronSpinIndexKnown then
      if R.MaybeOpenShell then begin
        Buffer := '1';
        {$IFNDEF CONSOLE}
        if not InputQuery('Select Electron Spin',
                          'Enter spin number for electron (Enter 0 for singlet states) (SPIND='+IntToStr(SPIND)+')', Buffer) then
          Abort;
        {$ENDIF}
        ElectronSpinIndex := StrToInt(Buffer);
        if ElectronSpinIndex = 0 then
          ElectronSpinIndex := -1;
      end else
        ElectronSpinIndex := -1;
    ElectronSpinIndexKnown := True;
    Index := 1;
    SIndex := 0;
    StartSpin;
    if not R.Quiet and NewMolecule then
      WriteLn(Log, 'For SPIND = ', SPIND, ' :');
    while t <> 0 do begin
      if Length(SpinIndex) <= Index then
        SetLength(SpinIndex, Index+1); { Ignore first entry }
      if Index = ElectronSpinIndex then begin
        ElectronSpin := NextSpin;
        if not R.Quiet and NewMolecule then
          WriteLn(Log, '  Assuming electron is spin number ', ElectronSpinIndex, ' (', QuantumNumberToStr(ElectronSpin,qftJ), ')');
        SpinIndex[Index] := -1;
      end else begin
        SetLength(NuclearSpins, Length(NuclearSpins)+1);
        NuclearSpins[High(NuclearSpins)] := NextSpin;
        SpinIndex[Index] := SIndex;
        Inc(SIndex);
        if not R.Quiet and NewMolecule then
          WriteLn(Log, '  Assuming spin number ', Index, ' is a nuclear spin of ', QuantumNumberToStr(NuclearSpins[High(NuclearSpins)],qftJ));
        Molecule.JAdjustSym := False;
      end;
      Inc(Index);
    end;
    if NewMolecule then begin
      Molecule.nNuclei := SIndex;
      for i := 0 to Molecule.nNuclei - 1 do
        Molecule.Nuclei[i].Name := 'Nucleus' + IntToStr(i+1);
    end
  end else begin
    SIndex := 0;
    Index := 0;
    StartSpin;
    while t > 0 do begin
      Inc(Index);
      if (Index <> ElectronSpinIndex) and (not (Index = 1) and not ElectronSpinIndexKnown) then
        Inc(SIndex);
      NextSpin;
    end;
  end;
  if Molecule.nNuclei <> SIndex then begin
    Inc(nProblems);
    WriteLn(Log, 'nNuclei changing for ',Molecule.FullHeirarchicName, ' ', Molecule.nNuclei, ' ', SIndex, ' ', SPIND);
  end;
end;

function TMoleculeReader.SetTemplate(const OI: TCalOptionInfo; out Manifold: TManifold):Boolean;
  function NuclearSPIND(i: Int64):Int64;
  var
    p: Int64;
    j: Integer;
  begin
    i := Abs(i);
    Result := i;
    if ElectronSpinIndex > 0 then begin
      p := SpinDigits;
      for j := 2 to ElectronSpinIndex do
        p := p * SpinDigits;
      Result := i div p * p;
      p := p div SpinDigits;
      if p > 0 then
        Result := Result + i mod p;
    end;
  end;
var
  i : Integer;
  NucSpins: Int64;
begin
  i := 0;
  Result := False; { Not flipped }
  NucSpins := NuclearSPIND(OI.SPIND);
  while i < Length(TemplateOIs) do begin
    with TemplateOIs[i] do
      if NuclearSPIND(SPIND) = NucSpins then
        if (OI.WTPL = WTPL) and (OI.WTMN = WTMN) then
          Break
        else if (OI.WTPL = WTMN) and (OI.WTMN = WTPL) then begin
          Result := True;
          Break;
        end;
    Inc(i);
  end;
  if i >= Length(TemplateOIs) then begin
    SetLength(TemplateOIs,i+1);
    SetLength(Isotopomers,i+1);
    TemplateOIs[i] := OI;
  end;
  if OI.IAX < 0 then begin
    if -OI.IAX < 6 then
      nItot := 2
    else
      case -OI.IAX of
        6:     nItot := 3;
        7,8:   nItot := 4;
        9:     nItot := 5;
        10,11: nItot := 6;
      else
        Inc(nProblems);
        WriteLn(Log, 'Itot coupling indicated; IAX = ', OI.IAX,' - need to set Nucleus.AsNext');
      end
  end else
    nItot := 0;
  Manifold := SwitchIsotopomer(Isotopomers[i], OI, Result);
end;

function ImportCALPGMFile(S: TTextStream; MaybeOpenShell, Quiet: Boolean): TMixture;
var
  Reader: TCALPGMReader;
  MoleculeReader: TMoleculeReader;
  IntStream: TTextStream;
  FileName: string;
  i: Integer;
begin
  S.ItemSeparators := S.ItemSeparators + [isComma];
  S.Comments := True;
  S.FlowToNextLine := True;
  S.FlowToNextRecord := False;
  S.AllowDefaults := True;
  IntStream := nil;
  PushProgress('Importing SPCAT/SPFIT',S.Size div 60);
  Reader := TCALPGMReader.Create(S, Quiet);
  try
    Reader.MaybeOpenShell := MaybeOpenShell;
    Result := TMixture.Create(nil);
    try
      Reader.Mixture := Result;
      Result.Name := 'Mixture';
      while Result.Units <> MHz do
        Result.CycleUnits;
      Result.Simulations[0].IntensityUnits := nm2MHzperMolecule;
      Result.Simulations[0].PlotUnits := MHz;
      TSpecies.Create(Result);
      if Reader.OI.KNMIN = Reader.OI.KNMAX then
        MoleculeReader := TLinearMoleculeReader.Create(Reader)
      else if Reader.OI.SPIND >= 0 then
        MoleculeReader := TAsymmetricMoleculeReader.Create(Reader)
      else
        MoleculeReader := TSymmetricMoleculeReader.Create(Reader);
      try
        MoleculeReader.Read;
        Result.FixupProperties.Free;
        if S.StreamName <> '' then begin
          FileName := ChangeFileExt(ExtractFileName(S.StreamName), '.int');
          if FileExists(FileName) then begin
            IntStream := TTextStream.Create(FileName,True);
            IntStream.ItemSeparators := IntStream.ItemSeparators + [isComma];
            MoleculeReader.ReadInt(IntStream);
          end;
        end;
        {$IFNDEF CONSOLE}
        for i := 0 to Abs(Reader.OI.NVIB)-1 do
          WriteLn(Log, 'MapState', i:5, ' ', MoleculeReader.StateMap[i].FullHeirarchicName(True, True));
        if MoleculeReader.nProblems > 0 then
          ShowMessage('Problems with import - see log window');
        {$ENDIF}
      finally
        MoleculeReader.Free;
      end;
      Result.FixupProperties.Free;
    except
      Result.Free;
      Raise;
    end;
  finally
    Reader.Free;
    IntStream.Free;
    PopProgress;
  end;
end;

function ImportCALPGMFile(S: TTextStream): TMixture;
begin
  Result := ImportCALPGMFile(S,not DefaultClosed,False);
end;

function TMoleculeReader.ConvertParametersWithSpin(State: TState; out Scale: Double; out MNode: TMNode):Integer;
var
  IDI: Integer;
begin
 Scale := 1;
 MNode := State;
 if (R.I1 = 0) and (R.I2 = 0) and (R.NS <> 0) and (R.FF = 0) then
   R.I1 := 1; { See FSO3 file }
 if (R.I1 <> 0) or (R.I2 <> 0) then begin
   IDI := (((R.FF*1000 + R.NS)*100 + R.TYP)*10 + R.KSQ)*10 + R.NSQ;
   if R.I1 = ElectronSpinIndex then
     if R.I2 = 0 then
       Result := SParameter(IDI, State, Scale, MNode)
     else if R.I2 = ElectronSpinIndex then
       Result := SParameter(IDI + 1100000, State, Scale, MNode)
     else
       Result := ISParameter(R.I2, IDI, State, Scale, MNode)
   else if R.I2 = ElectronSpinIndex then
     if R.I1 = 0 then { Documentation implies commutator with iN^2? }
       Result := SParameter(IDI+100000, State, Scale, MNode)
     else
       Result := ISParameter(R.I1, IDI, State, Scale, MNode)
   else if R.I2 = 0 then
     Result := IParameter(R.I1, IDI, State, Scale, MNode)
   else if R.I1 = 0 then
     Result := IParameter(R.I2, IDI, State, Scale, MNode)
   else if R.I1 = R.I2 then
     Result := IParameter(R.I2, IDI + 1100000, State, Scale, MNode)
   else
     Result := IIParameter(R.I1, R.I2, IDI, State, Scale, MNode);
 end else
   Result := ConvertParameter(State, Scale, MNode)
end;

{ TCALPGMReader }

constructor TCALPGMReader.Create(Stream: TTextStream; AQuiet: Boolean);
var
  OIt: TCalOptionInfo;
  i, VSYM, t: Integer;
begin
  MaybeOpenShell := True;
  Quiet := AQuiet;
  S := Stream;
  Title := S.RestOfLine;
  S.NextLine;
  nPar := S.NextInteger(999);
  S.NextLine; { Skip NLINE, NITR, NXPAR,  THRESH, ERRTST, FRAC, CAL }
  S.FlowToNextLine := False;
  { Line 3 }
  OI := ReadOptions(True);
  SetLength(OIs, Abs(OI.NVIB));
  for i := 0 to Abs(OI.NVIB) - 1 do
    OIs[i] := OI;
  OIt := OI;
  while OIt.VSYM < 0 do begin
    OIt := ReadOptions(False);
    OIs[Abs(OIt.NVIB)] := OIt;
  end;
  VSYM := OI.VSYM;
  i := 0;
  while (VSYM > 0) and (i < Length(OIs)) do begin
    if VSYM mod 10 > 0 then begin
      t := OIs[i].WTMN;
      OIs[i].WTMN := OIs[i].WTPL;
      OIs[i].WTPL := t;
    end;
    Inc(i);
    VSYM := VSYM div 10;
  end;
end;

function TCALPGMReader.ReadOptions(FirstLine: Boolean): TCalOptionInfo;
var
  c: string;
  p: Integer;
begin
  c := S.NextString('g');
  if (Length(c) = 1) and not (c[1] in ['0'..'9']) then begin
    Result.CHR := c[1];
    S.NextInt64Var(Result.SPIND);
  end else
    Result.SPIND := StrToInt64(c);
  Result.NVIB := 0;
  S.NextIntegerVar(Result.NVIB);
  if FirstLine and (Result.NVIB = 0) then begin
    Result.NVIB := 1;
    if not Quiet then
      WriteLn(Log, 'NVIB read as 0 - assuming 1');
  end;
  Result.KNMIN := S.NextInteger(0);
  Result.KNMAX := S.NextInteger(-1);
  Result.IXX := S.NextInteger(0);
  if FirstLine and (Result.IXX <> 0) and not Quiet then
    WriteLn(Log, 'IXX = ', Result.IXX, ' ignored - all off-diagonal interactions will be included');
  Result.IAX := S.NextInteger(1);
  Result.WTPL := S.NextInteger(1);
  Result.WTMN := S.NextInteger(1);
  Result.VSYM := S.NextInt64(0);
  Result.EWT := S.NextInteger(99);
  Result.DIAG := 0;
  Result.XOPT := 0;
  if S.GetFreeItem and S.IsIntegerToken(Result.DIAG) then begin
    if S.GetFreeItem and S.IsIntegerToken(Result.XOPT) then
      if Result.XOPT <> 0 then
        WriteLn(Log,'XOPT = ',Result.XOPT,'; Phase choice may not be consistent');
  end;
  S.NextInteger(0);
  { Allow text labels at the end of each line }
  c := S.RestOfLine;
  p := 1;
  while (p <= Length(c)) and (c[p] in [' ', ',']) do
    Inc(p);
  Result.Name := Copy(c,p,Length(c));
  S.NextLine;
end;

function TCALPGMReader.ReadParameterLine(out LineOK:Boolean): Boolean;
var
  t: Int64;
  idstring: string;
  function ND(n: Integer = 10): Integer;
  begin
    Result := t mod n;
    t := t div n;
  end;
begin
  LineOK := True;
  Result := False;
  if (nPar <= 0) or not S.GetFreeItem or (S.Token = '') then
    Exit;
  idstring := S.Token;
  if idstring[Length(idstring)] = '.' then
    Delete(idstring, Length(idstring), 1);
  try
    IDPAR := StrToInt64(idstring);
    if IDPAR = MaxInt then
      Exit;
    t := Abs(IDPAR);
    if Abs(OI.NVIB) < 10 then begin
      V1 := ND;
      V2 := ND;
    end else if Abs(OI.NVIB) < 100 then begin
      V1 := ND(100);
      V2 := ND(100);
    end else begin
      V1 := ND(1000);
      V2 := ND(1000);
    end;
    ID := t;
    NSQ := ND;
    KSQ := ND;
    TYP := ND(100);
    NS := ND;
    I1 := ND;
    I2 := ND;
    FF := ND;
    S.NextDoubleVar(PAR);
    S.NextDoubleVar(ERRPAR);
    Comment := S.RestOfLine;
  except
    on E: Exception do begin
      Writeln(Log, 'Format problem on line ', S.Line, ':', E.Message);
      LineOK := False;
    end;
  end;
  S.NextLine;
  Dec(nPar);
  Result := True; { Read a line }
end;

{ TAsymmetricMoleculeReader }

function TAsymmetricMoleculeReader.ConvertParameter(State: TState; out Scale: Double; out MNode: TMNode):Integer;
var
  ParName: string;
  PE: OAsymT.TPertEquiv;
  OneOfPair, OrgSReduction, MixedReduction: Boolean;
  AScale: Double;
begin
  MixedReduction := False;
  Scale := 1;
  ParName := '?';
  MNode := State;
  { Alternate method interpreting flags directly }
  PE := MakePE;
  if PE.Name <> '?' then begin
    OrgSReduction := Molecule.SReduction;
    PE.Name := OAsymT.FindParamEquiv(PE, OrgSReduction, OneOfPair, AScale);
    if Molecule.SReduction <> OrgSReduction then
      MixedReduction := SetReduction(OrgSReduction);
  end;
  { TODO - when all cases handled above, remove redundant stuff below }
  with State as TAsymmetricTop do
  case R.ID of
     0: ParName := 'Origin';
   100: ParName := 'A';
   200: ParName := 'B';
   300: ParName := 'C';

     1: begin
          ParName := 'BBar';
          SetValues(State, zIndex, 1, ParName);
        end;
    10: ParName := ParameterNames[zIndex];
   400: begin; ParName := 'BDelta'; Scale := 4; end;

     2: begin; ParName := 'DJ'; Scale := -1; end;
    11: begin; ParName := 'DJK'; Scale := -1; end;
    20: begin; ParName := 'DK'; Scale := -1; end;
   410: begin; ParName := 'deltaK'; SetReduction(False); Scale := -1; end;
   401: begin; ParName := 'deltaJ'; if not Molecule.SReduction then Scale := -1; end;
   500: begin; ParName := 'deltaK'; SetReduction(True); end;

     3: ParName := 'HJ';
    12: ParName := 'HJK';
    21: ParName := 'HKJ';
    30: ParName := 'HK';
   501: begin; ParName := 'phiJK'; SetReduction(True); end;
   600: begin; ParName := 'phiK'; SetReduction(True); end;
   402: ParName := 'phiJ';
   411: begin; ParName := 'phiJK'; SetReduction(False); end;
   420: begin; ParName := 'phiK'; SetReduction(False); end;

     4: ParName := 'LJ';
    13: ParName := 'LJJK';
    22: ParName := 'LJK';
    31: ParName := 'LKKJ';
    40: ParName := 'LK';
   403: ParName := 'llJ';
   412: begin; ParName := 'llJK'; SetReduction(False); end;
   421: begin; ParName := 'llKJ'; SetReduction(False); end;
   430: begin; ParName := 'llK'; SetReduction(False); end;
   502: begin; ParName := 'llJK'; SetReduction(True); end;
   601: begin; ParName := 'llKJ'; SetReduction(True); end;
   700: begin; ParName := 'llK'; SetReduction(True); end;
  end;
  if (PE.Name <> ParName) and (ParName <> '?') and (PE.Name <> '') then
    WriteLn(Log, 'Inconsistent ID=', R.ID, ' ', AScale, ' ', PE.Name, ' <> ', Scale, ' ', Parname, ' ', R.Comment, ' ',R.S.Line);
  if (ParName <> '?') and not MixedReduction and (not AllCALPGMpert or OneOfPair or (ParName = 'Origin')) then
    Result := State.IndexOfParameter(ParName)
  else if (PE.Name <> '?') and not OneOfPair then begin
    Scale := 1;
    Result := ConvertPE(PE, State, State, Scale, MNode)
  end else
    Result := ConvertPerturbation(State, State, Scale, MNode);
end;

function TAsymmetricMoleculeReader.MakePE: OAsymT.TPertEquiv;
var
  oP, OldScale4: Integer;
begin
  Result := OAsymT.EmptyPertEquiv;
  with Result do begin
    JPower := 2*R.NSQ;
    zPower := 2*R.KSQ;
    if ((R.I1 = ElectronSpinIndex) or ((R.I1 = 0) and (R.NS > 0) and (R.FF = 0))) and (R.I2 = 0) then begin
      if R.NS > 4 then begin
        Inc(zPower);
        SzPower := 1;
        Scale4 := 2;
        NSPower := R.NS-5;
      end else
        NSPower := R.NS;
      case R.TYP of
        0: { No operator };
        1: begin aPower := 1; SaPower := 1; Scale4 := 2; end;
        2: begin bPower := 1; SbPower := 1; Scale4 := 2; end;
        3: begin cPower := 1; ScPower := 1; Scale4 := 2; end;
        4..11:
           begin
             pPower:= 2*(R.TYP-3);
             if R.FF + R.NS = 0 then begin
                { If no spin operators specifed, add one }
               Dec(pPower);
               Inc(SpPower);
               Scale4 := 2;
             end;
           end;
       21: begin bPower := 1; ScPower := 1; Scale4 := 2; end;
       41: begin aPower := 1; ScPower := 1; Scale4 := 2; end;
       61: begin aPower := 1; SbPower := 1; Scale4 := 2; end;
      else
        Name := '?';
      end;
      if R.FF + R.NS + R.TYP = 0 then
        Inc(NSPower); { If no spin operators specifed, add one }
    end else if R.I1 + R.I2 > 0 then begin
      Name := '?'; { Other spin }
    end else begin
      { No spin operators }
      case R.TYP of
        0: { No operator };
        1: aPower := 2;
        2: bPower := 2;
        3: cPower := 2;
        4..11: pPower:= 2*(R.TYP-3);
       20: aPower := 1;
       21: begin bPower := 1; cPower := 1; end;
       40: bPower := 1;
       41: begin aPower := 1; cPower := 1; end;
       60: cPower := 1;
       61: begin
             aPower := 1;
             bPower := 1;
             { eg. 6110 TYP=61, KSQ=1 prolate should be 1/2[[Na,Nb]+,Na^2]+
               but the PGOPHER direct equivalent is [Na^3,Nb]+, which differs
               by a factor of two, and additional lower order term ~Na^2Nc }
             if zPower > 0 then begin
               Write(Log, 'Approximate conversion - ');
               ShowProblem;
               Scale4 := 8;
             end;
           end;
       { What is equivalent operator for 22, 42, 62? Does not seem to be simply N+^3 or Nc^3 }
      else
        Name := '?';
      end;
      oP := bPower + pPower;
      if R.OI.NVIB >= 0 then
        Inc(oP, cPower)
      else
        Inc(oP, aPower);
      if (zPower > 0) and (op > 0) then
        Scale4 := Scale4 div 2;
      OldScale4 := 4;
      if (zPower > 0) and
         ((bPower + pPower > 0)
          or ((aPower > 0) and (R.OI.NVIB <= 0))
          or ((cPower > 0) and (R.OI.NVIB > 0))
         ) then
        OldScale4 := 2;
      if OldScale4 <> Scale4 then
        WriteLn(Log, 'Diff scales ', R.TYP, ', IDI=', R.ID, ' ', OldScale4, ' ', Scale4, ' ', R.OI.NVIB, ' ', zPower, ' ', oP);
    end;
    if R.FF <> 0 then
      Name := '?';
  end;
  if Result.Name = '?' then
    WriteLn(Log, 'Unknown TYP=', R.TYP, ', IDI=', R.ID);
end;

function TAsymmetricMoleculeReader.ConvertPerturbation(State1, State2: TState; out Scale: Double; out MNode: TMNode):Integer;
var
  PE: OAsymT.TPertEquiv;
begin
  Result := -1;
  Scale := 1;
  PE := MakePE;
  if PE.Name <> '?' then
    Result := ConvertPE(PE, State1, State2, Scale, MNode);
end;

function TAsymmetricMoleculeReader.ConvertPE(PE: OAsymT.TPertEquiv; State1, State2: TState; out Scale: Double; out MNode: TMNode): Integer;
var
  Perturbation: TAsymmetricTopPerturbation;
begin
  Perturbation := TAsymmetricTopPerturbation.Create(State1.Manifold);
  with Perturbation do begin
    SetBraKetName(State1.Name, State2.Name);
    MNode := Perturbation;
    SetFromPertEquiv(PE);
    Comment := IntToStr(R.IDPAR);
    if PE.Scale4 <> 4 then
      Comment := Comment + '/' +FloatToStr(Four/PE.Scale4);
    Scale := PE.Scale4/Four;
    Result := MNode.IndexOfParameter('Value');
  end;
end;

procedure TAsymmetricMoleculeReader.MakeTransitionMoment(Magnetic: Boolean);
begin
  if Magnetic then
    Exit;
  case RI.TYP of
    0: (RI.MakeTM(TCartesianTransitionMoment) as TCartesianTransitionMoment).Axis := TabcAxis(RI.SYM-1);
  end;
end;

function TAsymmetricMoleculeReader.SwitchIsotopomer(var NewMolecule: TMolecule; OI: TCalOptionInfo; Flip: Boolean):TManifold;
begin
  if NewMolecule = nil then begin
    Molecule := TAsymmetricMolecule.Create(Species);
    NewMolecule := Molecule;
    if R.OI.NVIB > 0 then
      Molecule.Representation := Ir
    else
      Molecule.Representation := IIIl;
    { Heuristic for point group }
    if OI.WTPL = OI.WTMN then
      Molecule.PointGroup:= C1
    else
      Molecule.PointGroup := C2v;
    { Set C2zAxis anyway }
    case Abs(OI.IAX) of
      1: begin
          Molecule.C2zAxis := aAxis;
          Molecule.eeWt := OI.WTPL; Molecule.eoWt := OI.WTPL;
          Molecule.oeWt := OI.WTMN; Molecule.ooWt := OI.WTMN;
        end;
      2: begin
          Molecule.C2zAxis := bAxis;
          Molecule.eeWt := OI.WTPL; Molecule.eoWt := OI.WTMN;
          Molecule.oeWt := OI.WTMN; Molecule.ooWt := OI.WTPL;
        end;
      3: begin
          Molecule.C2zAxis := cAxis;
          Molecule.C2xAxis := aAxis; { Must not equal C2zAxis, and c is default }
          Molecule.eeWt := OI.WTPL; Molecule.eoWt := OI.WTMN;
          Molecule.oeWt := OI.WTPL; Molecule.ooWt := OI.WTMN;
        end;
      0:
    else
      Inc(nProblems);
      WriteLn(Log, 'Unexpected value for IAX:',R.OI.IAX);
    end;
    if ReductionKnown then
      Molecule.SReduction := SReduction;
  end else
    Molecule := NewMolecule as TAsymmetricMolecule;
  AddSpins(OI, Molecule);
  if Molecule.ComponentCount = 0 then begin
    Result := TAsymmetricManifold.Create(Molecule);
    Result.AdjustDefaults;
    Result.Initial := True;
    Result.Name := 'X';
  end else
    Result := Molecule.Manifold[0];
end;

procedure TAsymmetricMoleculeReader.Read;
begin
  ReductionKnown := True;
  case R.OI.CHR of
    'a', 'A': SReduction := False;
    's', 'S': SReduction := True;
  else
    ReductionKnown := False;
  end;
  inherited;
end;

function TAsymmetricMoleculeReader.SetReduction(TheSReduction: Boolean):Boolean;
var
  i, m: Integer;
  Mol: TAsymmetricMolecule;
begin
  Result := False;
  if ReductionKnown then begin
    if SReduction <> TheSReduction then begin
      Result := True;
      if not R.Quiet then
        Write(Log, 'Mixed reductions - '); { e.g. c038002.par; could fix by using diagonal perturbations for selected parameters }
      ShowProblem;
    end;
  end else begin
    SReduction := TheSReduction;
    for m := 0 to Species.ComponentCount-1 do begin
      Mol := Species.Isotopomer[m] as TAsymmetricMolecule;
      if SReduction <> Mol.SReduction then begin
        for i := 0 to Mol.Manifold[0].nStates do
          with Mol.Manifold[0].State[i] as TAsymmetricTop do
            deltaJ := - deltaJ;
        for i := 0 to nParameters - 1 do
          if (Parameters[i].MNode is TAsymmetricTop) and (Parameters[i].MNode.ParameterNames[Parameters[i].Index] = 'deltaJ') then
            Parameters[i].Value := -Parameters[i].Value;
        Mol.SReduction := SReduction;
      end;
    end;
  end;
  ReductionKnown := True;
end;

function TAsymmetricMoleculeReader.AddState(Manifold: TManifold; Index: Integer; FlipSym: Boolean; OI: TCalOptionInfo): TState;
var
  State: TAsymmetricTop;
begin
  State := TAsymmetricTop.Create(Manifold);
  State.AdjustDefaults;
  State.A := 0; { Must be set to zero, as subsequent updates only add to the values }
  State.B := 0;
  State.C := 0;
  State.S := ElectronSpin;
  Result := State;
  if FlipSym and (Molecule.PointGroup <> C1) then
    case Molecule.C2zAxis of
      aAxis: State.Symmetry := 2;
      bAxis: State.Symmetry := 3;
      cAxis: State.Symmetry := 1;
    end;
  State.Kmin := 2*OI.KNMIN;
  if OI.KNMAX >= 0 then
    State.Kmax := 2*OI.KNMAX;
end;

function TAsymmetricMoleculeReader.SParameter(IDI: Integer; State: TState; var Scale: Double; var MNode: TMNode):Integer;
var
  ParName: string;
  OneOfPair: Boolean;
  procedure OffDiag(x, y: Char);
  var
    Perturbation: TAsymmetricTopPerturbation;
  begin
    { Antisymmetric off-diagonal spin rotation components }
    { Brown + Sears JMS 75 111 (1979) implies combinations
      can be chosen such that these are not determinable. }
    { Seems to be phase problem in relative signs of matrix
      elements of e.g. ab+ba and ab-ba at the moment }
    Perturbation := TAsymmetricTopPerturbation.Create(State.Manifold);
    with Perturbation do begin
      SetBraKetName(State.Name, State.Name);
      MNode := Perturbation;
      SetSetting(IndexOfSetting(x+'Power'),1);
      SetSetting(IndexOfSetting('S'+y+'Power'),1);
      Comment := IntToStr(R.IDPAR);
    end;
    with TAsymmetricTopPerturbation.Create(State.Manifold) do begin
      SetBraKetName(State.Name, State.Name);
      MNode := Perturbation;
      SetSetting(IndexOfSetting(y+'Power'),1);
      SetSetting(IndexOfSetting('S'+x+'Power'),1);
      ScalePrev := -1;
      Comment := IntToStr(R.IDPAR);
    end;
    ParName := 'Value';
    Scale := 1/2;
    OneOfPair := True;
  end;
var
  PE: OAsymT.TPertEquiv;
  OrgSReduction: Boolean;
  Perturbation: TAsymmetricTopPerturbation;
  AScale: Double;
begin
  { Alternate method interpreting flags directly }
  PE := MakePE;
  if IDI >= 100000 then
    PE.Name := '?'; { Spin-spin or other oddity }
  OneOfPair := False;
  if PE.Name <> '?' then begin
    OrgSReduction := Molecule.SReduction;
    PE.Name := OAsymT.FindParamEquiv(PE, OrgSReduction, OneOfPair, AScale);
    if (PE.Name = '') and (Molecule.Representation = Ir) then begin
      Inc(PE.zPower,PE.aPower); PE.aPower := 0;
      Inc(PE.SzPower, PE.SaPower); PE.SaPower := 0;
      PE.Name := OAsymT.FindParamEquiv(PE, OrgSReduction, OneOfPair, AScale);
    end;
    if (PE.Name = '') and (Molecule.Representation = IIIl) then begin
      Inc(PE.zPower,PE.cPower); PE.cPower := 0;
      Inc(PE.SzPower, PE.ScPower); PE.ScPower := 0;
      PE.Name := OAsymT.FindParamEquiv(PE, OrgSReduction, OneOfPair, AScale);
    end;
    if Molecule.SReduction <> OrgSReduction then
      SetReduction(OrgSReduction);
  end;
  { TODO - when all cases handled above, remove redundant stuff below }
  Result := -1;
  ParName := '?';
  with State as TAsymmetricTop do
    case IDI of
    100: ParName := 'eaa';
    200: ParName := 'ebb';
    300: ParName := 'ecc';
   2100: ParName := 'ebcbar';
   4100: ParName := 'eacbar';
   6100: ParName := 'eabbar';
 102100: OffDiag('b','c');
 104100: OffDiag('a','c');
 106100: OffDiag('a','b');
      1: ParName := 'DsN';  { S Reduction }
     10: ParName := 'DsKN'; { S Reduction }
    101: ParName := 'DsNK'; { S Reduction }
    110: ParName := 'DsK';  { S Reduction }
    500: ParName := 'ds2';
1100100: begin; ParName := 'alpha'; Scale := 2/(3*S*(2*S-2)); end;

  10001: ParName := 'DsN';  { A Reduction }
  50001: ParName := 'DsNK'; { A Reduction }
  10010: ParName := 'DsKN'; { A Reduction }
  50010: ParName := 'DsK';  { A Reduction }
  10400: ParName := 'ds1';
  50400: ParName := 'ds2';  { A Reduction }
    end;
  if (PE.Name <> ParName) and (ParName <> '?') and (PE.Name <> '') and not OneOfPair then
    WriteLn(Log, 'Inconsistent IDI=', IDI, AScale, ' ', PE.Name, ' <> ', Scale, ' ', Parname, ' ', R.Comment, ' ', R.S.Line);
  if ParName <> '?' then
    Result := MNode.IndexOfParameter(ParName)
  else if (PE.Name <> '?') and not OneOfPair then
    Result := ConvertPE(PE, State, State, Scale, MNode);
end;

function TAsymmetricMoleculeReader.ISParameter(INuc: Integer; IDI: Integer;
  State: TState; var Scale: Double; var MNode: TMNode): Integer;
var
  ParName: string;
  procedure SetTqq(abcAxis: TabcAxis);
  var
    xyzAxis: TxyzAxis;
  begin
    xyzAxis := (State.Molecule as TAsymmetricMolecule).xyzAxis(abcAxis);
    with State.Nucleus(SpinIndex[INuc]) as TAsymmetricTopNucleus do
      case xyzAxis of
        xAxis: ParName := 'Txxmyy';
        yAxis: begin; ParName := 'Txxmyy'; Scale := -1; end;
        zAxis: begin ParName := 'Tzz'; Scale := 2/3; end ; { Note 2/3 scale factor has caused problems }
      end;
  end;
begin
  Result := -1;
  MNode := State.Nucleus(SpinIndex[INuc]);
  case IDI of
    0: ParName := 'aF';
  100: SetTqq(aAxis);
  200: SetTqq(bAxis);
  300: SetTqq(cAxis);
 2100: ParName := 'Tbc';
 4100: ParName := 'Tac';
 6100: ParName := 'Tab';
  400: begin; ParName := 'Txxmyy'; Scale := 4; end;
//  110: TKaa := TKaa*2/3;
//  101: TJaa := TKaa*2/3;
//  410: TKxxmyy := TKxxmyy + PAR*4;
  end;
  if ParName <> '' then
    Result := MNode.IndexOfParameter(ParName);
end;

function TAsymmetricMoleculeReader.IIParameter(INuc1, INuc2: Integer;
  IDI: Integer; State: TState; var Scale: Double; var MNode: TMNode): Integer;
begin
  Result := -1;
end;

function TAsymmetricMoleculeReader.IParameter(INuc: Integer; IDI: Integer;
  State: TState; var Scale: Double; var MNode: TMNode): Integer;
var
  ParName: string;
  procedure SetChiqq(abcAxis: TabcAxis);
  var
    xyzAxis: TxyzAxis;
  begin
    xyzAxis := (State.Molecule as TAsymmetricMolecule).xyzAxis(abcAxis);
    with State.Nucleus(SpinIndex[INuc]) as TAsymmetricTopNucleus do
      case xyzAxis of
        xAxis: begin ParName := 'CHIxxmyy'; SetValues(MNode,IndexOfParameter('CHIzz'), -1/3, ParName); end;
        yAxis: begin; ParName := 'CHIxxmyy'; Scale := -1; SetValues(MNode,IndexOfParameter('CHIzz'), -1/3, ParName); end;
        zAxis: begin ParName := 'CHIzz'; Scale := 2/3; end;
      end;
  end;
begin
  Result := -1;
  MNode := State.Nucleus(SpinIndex[INuc]);
  case IDI of
    100: ParName := 'Caa';
    200: ParName := 'Cbb';
    300: ParName := 'Ccc';
//  410: chiKxxmyy := chiKxxmyy + PAR*4; ? need spin scaling
1100100: SetChiqq(aAxis);
1100200: SetChiqq(bAxis);
1100300: SetChiqq(cAxis);
1100400: begin; ParName := 'CHIxxmyy'; Scale := 4; end;
1102100: ParName := 'Chibc';
1104100: ParName := 'Chiac';
1106100: ParName := 'Chiab';
  end;
  if ParName <> '' then
    Result := MNode.IndexOfParameter(ParName)
end;

{ TSymmetricMoleculeReader }

procedure TSymmetricMoleculeReader.Read;
begin
  LastEState := -999;
  Inc(nProblems);
  if not R.Quiet then
    WriteLn(Log, 'Point group, state symmetries and statistical weights may well need attention');
  inherited;
end;

function TSymmetricMoleculeReader.SwitchIsotopomer(var NewMolecule: TMolecule; OI: TCalOptionInfo; Flip: Boolean):TManifold;
begin
  if NewMolecule = nil then begin
    Molecule := TSymmetricMolecule.Create(Species);
    NewMolecule := Molecule;
    case Abs(OI.IAX) of
      6:     begin
               if OI.WTPL = OI.WTMN then begin
                 Molecule.Wt0 := OI.WTPL div 2;
                 Molecule.Wt1 := OI.EWT mod 100;
                 Molecule.PointGroup := C3v
               end else begin
                 Molecule.PointGroup := D3h;
                 Molecule.Wt0 := OI.WTPL;
                 Molecule.Wt1 := OI.WTMN;
                 Molecule.Wt2 := OI.EWT mod 100;
               end;
             end;
      7,8:   Molecule.PointGroup := C4v;
      9:     Molecule.PointGroup := C5v;
      10,11: Molecule.PointGroup := C6v;
    else
      { Need to scan all states }
    end;
  end else
    Molecule := NewMolecule as TSymmetricMolecule;
  AddSpins(OI, Molecule);
  if Molecule.ComponentCount = 0 then begin
    Result := TSymmetricManifold.Create(Molecule);
    Result.AdjustDefaults;
    Result.Initial := True;
    Result.Name := 'X';
  end else
    Result := Molecule.Manifold[0] as TSymmetricManifold;
end;

function TSymmetricMoleculeReader.AddState(Manifold: TManifold; Index: Integer; FlipSym: Boolean; OI: TCalOptionInfo): TState;
var
  State: TSymmetricTop;
  EWT, n: Integer;
begin
  { TODO: Handle FlipSym? }
  State := TSymmetricTop.Create(Manifold);
  State.AdjustDefaults;
  State.B := 0; { Must be set to zero, as subsequent updates only add to the values }
  State.C := 0;
  State.S := ElectronSpin;
  EWT := R.OIs[Index].EWT;
  if (EWT > 0) and (EWT mod 100 <> 99) then begin
    n := State.Molecule.CharacterTable.n;
    EWT := (EWT div 100) * 2;
    if EWT = 0 then begin
      { A1/A2 Symmetry }
      if (R.OIs[Index].WTPL <> R.OI.WTPL) and (Molecule.PointGroup = D3h) then
        State.Symmetry := 4; { Assume A2" }
    end else if EWT = n then
      { B1/B2 Symmetry }
    else if EWT < n then begin
      { E Symmetry }
      if LastEState > 0 then
        Writeln(Log, 'State order problem for ', Index)
      else if LastEState = -999 then begin
        LastEState := Index;
      end else begin
        StateIndexMap[-LastEState-1] := Index;
        LastEState := -999;
      end;
      State.Symmetry := 2; { TODO - will need adjustment }
    end else begin
      { Other half of E symmetry - discard }
      if LastEState >= 0 then begin
        StateIndexMap[Index] := LastEState;
        LastEState := -999;
      end else if LastEState = -999 then
         LastEState := -(Index+1)
      else
        WriteLn(Log, 'State ordering problem for ', Index);
      State.Symmetry := 2; { TODO - will need adjustment }
    end;
  end;
  Result := State;
end;

function TSymmetricMoleculeReader.ConvertParameter(State: TState; out Scale: Double; out MNode: TMNode):Integer;
  procedure CheckZetaEta;
  begin
    if not ZetaEtaMsg and not R.Quiet then
      WriteLn(Log, 'Zeta and eta values may need checking');
    ZetaEtaMsg := True;
  end;
var
  ParName, AParName: string;
  PE: SymT.TPertEquiv;
  OneOfPair: Boolean;
  AScale: Double;
begin
 Result := -1;
 Scale := 1;
 PE := SymT.EmptyPertEquiv;
 PE.JPower := 2*R.NSQ;
 PE.zPower := 2*R.KSQ;
 AParName := '';
 case R.TYP of
    0: { No operator };
    4,5,6,7,8,9,10,11:
       begin
         PE.pPower:= 2*(R.TYP-3);
         if PE.zPower > 0 then
           PE.Scale4 := 2;
       end;
    20: if R.OI.NVIB > 0 then begin
          Inc(PE.zPower); { Prolate }
          PE.lPower := 1;
        end else
          AParName := '?';
    60: if R.OI.NVIB < 0 then begin
          Inc(PE.zPower); { Oblate }
          PE.lPower := 1;
        end else
          AParName := '?';
 else
    AParName := '?';
 end;
 if AParName = '?' then
    WriteLn(Log, 'Unknown TYP=', R.TYP, ', IDI=', R.ID, ' ',R.Comment);
 if AParName <> '?' then
   AParName := FindParamEquiv(PE, OneOfPair, AScale);
 { TODO - when all cases handled above, remove redundant stuff below }
 MNode := State;
 with State as TSymmetricTop do
  case Self.R.ID of
     0: ParName := 'Origin';
     1: begin
          ParName := 'B';
          SetValues(State,IndexOfParameter('C'), 1, 'B');
        end;
    10: ParName := 'C';

     2: begin; ParName := 'DJ'; Scale := -1; end;
    11: begin; ParName := 'DJK'; Scale := -1; end;
    20: begin; ParName := 'DK'; Scale := -1; end;

     3: ParName := 'HJ';
    12: ParName := 'HJK';
    21: ParName := 'HKJ';
    30: ParName := 'HK';

     4: ParName := 'LJ';
    13: ParName := 'LJJK';
    22: ParName := 'LJK';
    31: ParName := 'LKKJ';
    40: ParName := 'LK';

  2000: begin ParName := 'zeta'; Scale := - 1/2/C; CheckZetaEta; end;
  2001: begin ParName := 'etaJ'; CheckZetaEta; end;
  2010: begin ParName := 'etaK'; CheckZetaEta; end;

  end;
  if (AParname <> ParName) and not OneOfPair then { OneOfPair only applies to B, which needs special handling anyway }
    WriteLn(Log, 'Inconsistent ID=', R.ID, ' ', AScale, ' ', AParname, ' <> ', Scale, ' ', Parname, ' ', R.Comment, ' ',R.S.Line);
  if ParName <> '' then
    Result := State.IndexOfParameter(ParName)
  else if (AParName <> '?') and not OneOfPair then
    Result := ConvertPE(PE, State, State, Scale, MNode)
  else
    Result := ConvertPerturbation(State, State, Scale, MNode);
end;

function TSymmetricMoleculeReader.SParameter(IDI: Integer; State: TState; var Scale: Double; var MNode: TMNode):Integer;
var
  ParName, AParName: string;
  PE: SymT.TPertEquiv;
  OneOfPair: Boolean;
  AScale: Double;
begin
  PE := SymT.EmptyPertEquiv;
  PE.JPower := 2*R.NSQ;
  PE.zPower := 2*R.KSQ;
  if R.NS > 4 then begin
    Inc(PE.zPower);
    PE.SzPower := 1;
    PE.Scale4 := 2;
    PE.NSPower := R.NS-5;
  end else
    PE.NSPower := R.NS;
  case R.TYP of
    0: { No operator };
    1: if R.OI.NVIB > 0 then begin
         Inc(PE.zPower); { Prolate }
         Inc(PE.SzPower); { Prolate }
       end else begin
         { Oblate }
         PE.pPower:= 1;
         PE.SpPower:= -1;
         PE.Scale4 := 1;
       end;
    4,5,6,7,8,9,10,11:
       begin
         PE.pPower:= 2*(R.TYP-3);
         if IDI < 1000 then begin
            { If no spin operators specifed, add one }
           Dec(PE.pPower);
           Inc(PE.SpPower);
           PE.Scale4 := 2;
         end;
       end;
  else
    WriteLn(Log, 'Unknown TYP=', R.TYP, ', IDI=', IDI);
    AParName := '?';
  end;
  if IDI < 100 then
    Inc(PE.NSPower); { If no spin operators specifed, add one }
  if IDI >= 100000 then
    AParName := '?'; { Spin-spin or other oddity }
  if AParName <> '?' then
    AParName := SymT.FindParamEquiv(PE, OneOfPair, AScale);
  Result := -1;
  with State as TSymmetricTop do
    case IDI of
    100: ParName := 'ecc';
      0: begin
           ParName := 'ebb'; { 0 is actually aa+bb+cc; PGOPHER uses xx = cc and xx = yy = bb }
           SetValues(State,IndexOfParameter('ecc'),1,'ebb');
           OneOfPair := True; { Special handling required for ParamEquiv }
         end;
  10001: ParName := 'DsN';
  50001: ParName := 'DsNK';
  10010: ParName := 'DsKN';
  50010: ParName := 'DsK';
//1100100: begin; ParName := 'alpha'; Scale := 2/(3*S*(2*S-2)); end;
    end;
  if (AParname <> ParName) and not OneOfPair then
    WriteLn(Log, 'Inconsistent ID=', R.ID, ' ', AScale, ' ', AParname, ' <> ', Scale, ' ', Parname, ' ', R.Comment, ' ',R.S.Line);
  if ParName <> '' then
    Result := MNode.IndexOfParameter(ParName)
  else if (AParName <> '?') and not OneOfPair then
    Result := ConvertPE(PE, State, State, Scale, MNode)
end;

function TSymmetricMoleculeReader.ISParameter(INuc: Integer; IDI: Integer; State: TState; var Scale: Double; var MNode: TMNode):Integer;
begin
  Result := -1;
end;

function TSymmetricMoleculeReader.IIParameter(INuc1, INuc2: Integer;
  IDI: Integer; State: TState; var Scale: Double; var MNode: TMNode): Integer;
begin
  Result := -1;
end;

function TSymmetricMoleculeReader.IParameter(INuc: Integer; IDI: Integer; State: TState; var Scale: Double; var MNode: TMNode):Integer;
var
  ParName: string;
begin
  Result := -1;
  MNode := State.Nucleus(SpinIndex[INuc]);
  case IDI of
1100100: begin; ParName := 'eQq'; Scale := 2/3; end;
      0: begin
           ParName := 'Cperp';
           SetValues(MNode, MNode.IndexOfParameter('Czz'), 1, 'Cperp');
         end;
    100: ParName := 'Czz';
  end;
  if ParName <> '' then
    Result := MNode.IndexOfParameter(ParName)
end;

function TSymmetricMoleculeReader.ConvertPerturbation(State1, State2: TState; out Scale: Double; out MNode: TMNode):Integer;
var
  Perturbation: TSymmetricTopPerturbation;
  procedure Considerl(lChange, lPower: Integer);
  begin
    if Molecule.IsDegenerate((State1 as TSymmetricTop).Symmetry)
       or Molecule.IsDegenerate((State2 as TSymmetricTop).Symmetry) then begin
      { TODO: These should be decided on symmetry grounds }
      Perturbation.lChange := lChange;
      Perturbation.lPower := lPower;
    end;
  end;
begin
  Result := -1;
  Scale := 1;
  if State1 = State2 then
    with State1 as TSymmetricTop do begin
      case Self.R.ID of
        { These are really diagonal, but expressed as matrix elements between
          degenerate states }
        400: begin Result := IndexOfParameter('qplus'); Scale := -2; end;
        401: begin Result := IndexOfParameter('DqJ'); Scale := -2; end;
        410: begin Result := IndexOfParameter('DqK'); Scale := -1; end;
      end;
      if Result > 0 then
        MNode := State1;
    end;
  if (Result < 0) and (R.FF = 0) and (R.I2 = 0) and (R.I1 = 0) and (R.NS = 0) then begin
    Perturbation := TSymmetricTopPerturbation.Create(State1.Manifold);
    with Perturbation do begin
      SetBraKetName(State1.Name, State2.Name);
      Result := IndexOfParameter('Value');
      MNode := Perturbation;
      JPower := R.NSQ*2;
      zPower := 2*R.KSQ;
      Comment := IntToStr(R.IDPAR);
      case R.TYP of
        0: ;
        // 1..3: NotYet;
        4..10: begin
                 pPower := 2*(R.TYP-3);
                 Considerl(2,0);
                 Scale := -Scale; // Seems to be sometimes needed; TODO - decide on symmetry grounds?
               end;
        20..39: if R.OI.NVIB > 0 then begin
                  { Prolate }
                  pPower := (R.TYP-19) div 2 * 2;
                  if not Odd(R.TYP) then
                    zPower := zPower + 1;
                  Considerl(0,1);
                end else begin
                  { Oblate }
                  pPower := (R.TYP-20) div 2 * 2 + 1;
                  if Odd(R.TYP) then
                    zPower := zPower + 1;
                  Considerl(-2,0);
                end;
        // 40..59: NotYet;
        60..79: if R.OI.NVIB > 0 then begin
                  { Prolate }
                  pPower := (R.TYP-59) div 2 * 2 + 1;
                  if Odd(R.TYP) then
                    zPower := zPower + 1;
                  Considerl(0,1);
                end else begin
                  { Oblate }
                  pPower := (R.TYP-60) div 2 * 2;
                  if not Odd(R.TYP) then
                    zPower := zPower + 1;
                  Considerl(0,1);
                end;
      else
        FreeAndNil(MNode);
        Result := -1;
      end;
      if (Result >= 0) and (zPower <> 0) and (pPower <> 0) then begin
        Scale := Scale/sqrt(Two);
        if zPower >= 2 then
          if Odd(zPower) then begin
            zPower := -zPower;
            Scale := Scale/2;
          end else
            Scale := Scale/sqrt(Two);
      end;
    end;
  end;
end;

function TSymmetricMoleculeReader.ConvertPE(PE: SymT.TPertEquiv; State1, State2: TState; out Scale: Double; out MNode: TMNode): Integer;
var
  Perturbation: TSymmetricTopPerturbation;
begin
  Perturbation := TSymmetricTopPerturbation.Create(State1.Manifold);
  with Perturbation do begin
    SetBraKetName(State1.Name, State2.Name);
    MNode := Perturbation;
    SetFromPertEquiv(PE);
    Comment := IntToStr(R.IDPAR);
    Scale := PE.Scale4/Four;
    Result := MNode.IndexOfParameter('Value');
  end;
end;

procedure TSymmetricMoleculeReader.MakeTransitionMoment(Magnetic: Boolean);
var
  Parallel: Boolean;
  Component: TQuantumNumber;
begin
  if Magnetic then
    Exit;
  Parallel := ((R.OI.NVIB > 0 { Prolate }) and (RI.SYM = 1 {a}))
              or ((R.OI.NVIB < 0 { Oblate }) and (RI.SYM = 3 {c}));
  if Parallel then
    Component := 0
  else
    Component := 2;
  case RI.TYP of
    0: begin
        (RI.MakeTM(TSphericalTransitionMoment) as TSphericalTransitionMoment).Component := Component;
        if Component = 2 then
          RI.ScaleDen := 2;
       end;
  end;
end;

{ TLinearMoleculeReader }

function TLinearMoleculeReader.SwitchIsotopomer(var NewMolecule: TMolecule; OI: TCalOptionInfo; Flip: Boolean):TManifold;
begin
  if NewMolecule = nil then begin
    Molecule := TLinearMolecule.Create(Species);
    NewMolecule := Molecule;
    Molecule.Symmetric := OI.WTMN <> OI.WTPL;
    if Flip then begin
      Molecule.SymWt := OI.WTMN;
      Molecule.AsymWt := OI.WTPL;
    end else begin
      Molecule.SymWt := OI.WTPL;
      Molecule.AsymWt := OI.WTMN;
    end;
  end else
    Molecule := NewMolecule as TLinearMolecule;
  AddSpins(OI, Molecule);
  if Molecule.ComponentCount = 0 then begin
    Result := TLinearManifold.Create(Molecule);
    Result.AdjustDefaults;
    Result.Initial := True;
    if Molecule.Symmetric then
      Result.Name := 'g'
    else
      Result.Name := 'X';
  end else
    Result := Molecule.Manifold[0];
  { gerade and ungerade states must be in different manifolds }
  if Flip then
    if Molecule.ComponentCount = 1 then begin
      Result := TLinearManifold.Create(Molecule);
      Result.AdjustDefaults;
      Result.Initial := True;
      Result.Name := 'u';
    end else
      Result := Molecule.Manifold[1];
end;

function TLinearMoleculeReader.AddState(Manifold: TManifold; Index: Integer; FlipSym: Boolean; OI: TCalOptionInfo): TState;
var
  State: TLinear;
begin
  State := TLinear.Create(Manifold);
  State.AdjustDefaults;
  State.B := 0; { Must be set to zero, as subsequent updates only add to the values }
  State.S := ElectronSpin;
  State.Lambda := OI.KNMIN*2;
  if FlipSym then
    State.gerade := False;
  Result := State;
end;

function TLinearMoleculeReader.ConvertParameter(State: TState; out Scale: Double; out MNode: TMNode):Integer;
var
  ParName: string;
  PE: HundA.TPertEquiv;
  AScale: Double;
begin
  Result := -1;
  Scale := 1;
  PE := MakePE(State, State);
  if PE.Name <> '?' then
    PE.Name := HundA.FindParamEquiv(PE, AScale, State);
  MNode := State;
  ParName := '?';
  with State as TLinear do
    case R.ID of
      0: ParName := 'Origin';
      1: ParName := 'B';
      2: begin; ParName := 'D'; Scale := -1; end;
      3: ParName := 'H';
      4: ParName := 'L';
      5: ParName := 'M';
      6: ParName := 'PP';

    400 {pi}, 500 {Delta}:
         begin; ParName := 'q'; Scale := 2; end;
    401: begin; ParName := 'qD'; Scale := 2; end;
    402: begin; ParName := 'qH'; Scale := 2; end;
    403: begin; ParName := 'qL'; Scale := 2; end;
    end;
  if ((PE.Name <> ParName) or (AScale <> Scale)) and (ParName <> '?') and (PE.Name <> '') then
    WriteLn(Log, 'Inconsistent Linear ID=', R.ID, ' ', AScale, ' ', PE.Name, ' <> ', Scale, ' ', Parname, ' ', R.Comment, ' ',R.S.Line);
  if ParName <> '?' then
    Result := State.IndexOfParameter(ParName)
  else if PE.Name <> '?'then
    if PE.Name <> '' then begin
      Scale := AScale;
      Result := State.IndexOfParameter(PE.Name)
    end else begin
      Scale := 1;
      Result := ConvertPE(PE, State, State, Scale, MNode)
    end
  else
    Result := ConvertPerturbation(State, State, Scale, MNode);
end;

function TLinearMoleculeReader.ConvertPerturbation(State1, State2: TState; out Scale: Double; out MNode: TMNode):Integer;
var
  PE: HundA.TPertEquiv;
begin
  Result := -1;
  Scale := 1;
  PE := MakePE(State1, State2);
  if PE.Name <> '?' then
    Result := ConvertPE(PE, State1, State2, Scale, MNode);
end;

function TLinearMoleculeReader.ConvertPE(PE: HundA.TPertEquiv; State1, State2: TState; out Scale: Double; out MNode: TMNode): Integer;
var
  Perturbation: TLinearPerturbation;
begin
  Result := -1;
  if State1.Manifold <> State2.Manifold then begin
    WriteLn(Log, 'Perturbation between manifolds:',State1.Name,'-',State2.Name,' - ');
    Exit;
  end;
  Perturbation := TLinearPerturbation.Create(State1.Manifold);
  with Perturbation do begin
    SetBraKetName(State1.Name, State2.Name);
    MNode := Perturbation;
    if PE.Den = 0 then begin
      WriteLn(Log, 'Scale factor = ', PE.Num, '/', PE.Den);
      ShowProblem;
      Comment := IntToStr(R.IDPAR) + '* ???';
    end else begin
      SetFromPertEquiv(PE);
      Scale := MakeScale(PE.Num, PE.Den);
      if PE.Num <> PE.Den then
        Comment := IntToStr(R.IDPAR) + '*' + ShowScale(PE.Num,PE.Den) + ' '
      else
        Comment := IntToStr(R.IDPAR);
    end;
    if PE.Num <> 0 then
      Result := MNode.IndexOfParameter('Value');
  end;
end;

function TLinearMoleculeReader.MakePE(State1, State2: TState): HundA.TPertEquiv;
var
  K1, K2: Integer;
begin
  Result := HundA.EmptyPertEquiv;
  with Result do begin
    NPower := 2*R.NSQ;
    if R.KSQ > 0 then begin
      { Actually Lambda^2, so a constant }
      K1 := Max((State1 as TLinear).Lambda,0) div 2;
      K2 := Max((State2 as TLinear).Lambda,0) div 2;
      Num := Num * Sqr(Round(MathSup.IntPower(K1,2*R.KSQ) + MathSup.IntPower(K2,2*R.KSQ)));
      Den := Den * 4;
    end;
    if R.NS > 4 then
      Name := '?'
    else if R.NS >0 then begin
      Op := NS;
      n := R.NS;
    end else begin
      Op := LNuncouple;
      n := 0;
    end;

    if (R.I1 = ElectronSpinIndex) and (R.I2 = ElectronSpinIndex) then begin
      { Electron spin spin }
      case R.TYP of
        1: begin { LambdaSS }
             Srank := 4;
             with State2 do
               Den := Den * (S*(2*S-2));
           end;
        4: begin { o }
             Srank := -1;
             Scomp := -4
           end
      else
        Name := '?';
      end;
    end else if ((R.I1 = ElectronSpinIndex) or ((R.I1 = 0) and (R.NS > 0) and (R.FF = 0))) and (R.I2 = 0) then begin
      if R.NS > 0 then
        Name := '?';
      case R.TYP of
        0: if R.NS in [0..4] then
             Name := ''; { No main operator }
        1: begin { A } Result.Num := State2.S*(State2.S+2)*(State2.S+1);
                       Result.Den := 4;
                       Op := LS;
                       n := 1;
           end;
        4..11: begin { p } Op := LNuncouple; n := 2*(R.TYP-3); Srank := -1; Scomp := -2; end;
      else
        Name := '?';
      end;
      if R.FF + R.NS + R.TYP = 0 then begin
        Op := LNuncouple; { If no spin operators specifed, add one }
        n := 1;
      end;
    end else if R.I1 + R.I2 > 0 then begin
      Name := '?'; { Other spin }
    end else begin
      { No spin operators }
      case R.TYP of
        0: { No operator };
        4..11: begin { q } Op := LNuncouple; n := 2*(R.TYP-3); { Den := Den * 4; } end;
      else
        Name := '?';
      end;
    end;
    if R.FF <> 0 then
      Name := '?';
  end;
  if Result.Name = '?' then
    WriteLn(Log, 'Unknown Linear TYP=', R.TYP, ', IDI=', R.ID);
end;

function TLinearMoleculeReader.SParameter(IDI: Integer; State: TState; var Scale: Double; var MNode: TMNode):Integer;
var
  ParName: string;
  PE: HundA.TPertEquiv;
  AScale, EScale: Double;
begin
  Result := -1;
  PE := MakePE(State, State);
  if PE.Name <> '?' then
    PE.Name := HundA.FindParamEquiv(PE, AScale, State);
  ParName := '?';
  with State as TLinear do
    case IDI of
      0: ParName := 'gamma';
      1: ParName := 'gammaD';
      2: ParName := 'gammaH';
      3: ParName := 'gammaL';

    400: begin; ParName := 'p'; Scale := -2; end;
    401: begin; ParName := 'pD'; Scale := -2; end;
    402: begin; ParName := 'pH'; Scale := -2; end;
    403: begin; ParName := 'pL'; Scale := -2; end;

    100: ParName := 'A';
    101: ParName := 'AD';

    1100100: begin; ParName := 'LambdaSS'; Scale := 1/(S*(2*S-2)); end;
    1100101: begin; ParName := 'LambdaD'; Scale := 1/(S*(2*S-2)); end;
    1100102: begin; ParName := 'LambdaH'; Scale := 1/(S*(2*S-2)); end;
    1100103, 1100104, 1100105, 1100106, 1100107, 1100108, 1100109:
             begin
               PE.Num := 8; { Copied from LambdaSS and friends }
               PE.Den := 3 * Sqr(S*(2*S-2));
             end;
    1100400: ParName := 'o';
    1100401: ParName := 'oD';
    end;
  if ((PE.Name <> ParName) or (AScale <> Scale)) and (ParName <> '?') and (PE.Name <> '') then
    WriteLn(Log, 'Inconsistent Linear IDI=', IDI, AScale, ' ', PE.Name, ' <> ', Scale, ' ', Parname, ' ', R.Comment, ' ', R.S.Line);
  if ParName <> '?' then
    Result := State.IndexOfParameter(ParName)
  else if PE.Name <> '?'then
    if PE.Name <> '' then begin
      Scale := AScale;
      Result := State.IndexOfParameter(PE.Name)
    end else begin
      Scale := 1;
      Result := ConvertPE(PE, State, State, Scale, MNode)
    end
  else
    Result := ConvertPerturbation(State, State, Scale, MNode);
end;

function TLinearMoleculeReader.ISParameter(INuc: Integer; IDI: Integer;
  State: TState; var Scale: Double; var MNode: TMNode): Integer;
var
  ParName: string;
  PE: HundA.TPertEquiv;
begin
  Result := -1;
  MNode := State.Nucleus(SpinIndex[INuc]);
  ParName := '';
  PE := HundA.EmptyPertEquiv;
  with MNode as TLinearNucleus do begin
    if AsNext and (Self.Molecule.Symmetric) then
      Scale := Scale * 0.5; { Only one of pair of nuclei in .par file }
    case IDI of
      0: ParName := 'b';
    100: begin
           ParName := 'c';
           SetValues(MNode, MNode.IndexOfParameter('b'), -Scale/3, 'c');
         end;
    400: begin; ParName := 'd'; Scale := -2*Scale; end;
    401: begin;
           PE.Name := 'dJ';
           PE.Num := -4;
           PE.NPower := 2*R.NSQ;
           PE.Op := bF;
           PE.n := 2;
         end; // Was: ParName := 'dJ'; Scale := -2; ShowProblem; WriteLn(Log, 'Conversion is approximate'); end;
    end;
    if ParName <> '' then
      Result := MNode.IndexOfParameter(ParName)
    else if PE.Name <> '' then begin
      if AsNext and (Self.Molecule.Symmetric) then
        PE.Den := PE.Den*4;
      Result := ConvertPE(PE, State, State, Scale, MNode);
      (MNode as TPerturbation).Nucleus := SpinIndex[INuc]+1;
    end;
  end;
end;

function TLinearMoleculeReader.IIParameter(INuc1, INuc2: Integer; IDI: Integer;
  State: TState; var Scale: Double; var MNode: TMNode): Integer;
var
  SpinSpin :TLinearSpinSpin;
  ParName: string;
begin
  Result := -1;
  MNode := State.FindSpinSpin(SpinIndex[INuc1], SpinIndex[INuc2]);
  with MNode as TLinearSpinSpin do
    case IDI of
        0:  ParName := 'J';
      100:  begin; ParName := 'S'; Scale := 1/3; end;
    end;
  if ParName <> '' then
    Result := MNode.IndexOfParameter(ParName)
end;

function TLinearMoleculeReader.IParameter(INuc: Integer; IDI: Integer;
  State: TState; var Scale: Double; var MNode: TMNode): Integer;
var
  ParName: string;
  PE: HundA.TPertEquiv;
begin
  Result := -1;
  MNode := State.Nucleus(SpinIndex[INuc]);
  PE := HundA.EmptyPertEquiv;
  with MNode as TLinearNucleus do begin
    if AsNext and (Self.Molecule.Symmetric) then
      Scale := Scale * 0.5; { Only one of pair of nuclei in .par file }
    case IDI of
            0: begin
                 ParName := 'cI';
                 { actually I.N = I.(J-S), not I.J, so extra -I.S }
                 if State.S > 0 then
                   SetValues(MNode, MNode.IndexOfParameter('b'), -Scale, 'cI');
               end;
          400: begin
                 // Probably OK: PE.Name := 'cI2';  PE.Op := IJ;
                 // Probably not OK below; More likely involves N+^2 and possibly L-^2
                 //PE.n := 2;
                 //if State.S > 0 then
                 //  SetValues(MNode, MNode.IndexOfParameter('d'), 2, 'cI2'); { See cI }
               end;
          100: ParName := 'a';
      1100100: begin; ParName := 'eQq0'; Scale := 2/3*Scale; end;
      1100400: begin; ParName := 'eQq2'; Scale := 4*Scale; end; { TODO: Sign may not be consistent }
    end;
    if ParName <> '' then
      Result := MNode.IndexOfParameter(ParName)
    else if PE.Name <> '' then begin
      if AsNext and (Self.Molecule.Symmetric) then
        PE.Den := PE.Den*4;
      Result := ConvertPE(PE, State, State, Scale, MNode);
      (MNode as TPerturbation).Nucleus := SpinIndex[INuc]+1;
    end;
  end;
end;

procedure TLinearMoleculeReader.MakeTransitionMoment(Magnetic: Boolean);
var
  Parallel: Boolean;
  Component: TQuantumNumber;
  TM: TTransitionMoment;
  IP: TLinearIntensityPerturbation;
  Bra, Ket: TLinear;
begin
  Bra := RI.StateU as TLinear;
  Ket := RI.StateL as TLinear;
  Parallel := RI.SYM <= 1;
  if Parallel then
    Component := 0
  else begin
    Component := 2;
    RI.ScaleDen := 2;
  end;
  if Molecule.Symmetric and (Bra.gerade = Ket.gerade) then
    Magnetic := True;
  if Magnetic then begin
    if RI.I1 = ElectronSpinIndex then begin
      case RI.TYP of
        0: begin
             if RI.SYM = 0 then
               { need both parallel and perpendicular components }
               RI.CompleteTM(RI.FindTM(TSpinTransitionMoment, 2));
             TM := RI.FindTM(TSpinTransitionMoment, Component);
           end;
      end;
    end else if RI.I1 > 0 then
      Exit { Nuclear dipole }
    else
      case RI.TYP of
        0: if RI.SYM = 0 then begin
             TM := RI.FindTM(TRotationalTransitionMoment, Component);
             WriteLn(Log, TM.FullHeirarchicName+' will also contribute ', RI.DIPOLE/2, ' to gS');
             RI.DIPOLE := RI.DIPOLE / MeOverMp;
           end else
             TM := RI.FindTM(TOrbitalTransitionMoment, Component);
      end;
  end else
    case RI.TYP of
      0: begin
           TM := RI.FindTM(TSphericalTransitionMoment, Component);
           { TODO: Check for duplicate transition moments }
         end;
      3, 5:
         begin
           TM := RI.FindTM(TSphericalTransitionMoment, Component);
           IP := TLinearIntensityPerturbation.Create(TM);
           RI.TransitionMoment := IP;
           IP.SetBraKetName(TM.Bra.Name, TM.Ket.Name);
           IP.JPower := 2;
           IP.AntiSym := RI.TYP = 5;
         end;
      9, 10:
         begin
           TM := RI.FindTM(TSphericalTransitionMoment, Component);
           IP := TLinearIntensityPerturbation.Create(TM);
           RI.TransitionMoment := IP;
           IP.SetBraKetName(TM.Bra.Name, TM.Ket.Name);
           IP.JPower := 4;
           IP.AntiSym := RI.TYP = 10;
           if RI.TYP = 10 then
             RI.ScaleDen := RI.ScaleDen * 4;
         end;
    end;
end;

end.
