unit ImportLmatrix;
{ Import normal modes from MOLPRO xml output file and other sources }

{$I utils.inc}{Copyright (C) 2009 C M Western; see commandline.pas for license}
interface

uses TextStrm, Mixture, Vib;

function ImportLmatrixFile(Stream: TTextStream): TMixture;
function ImportMoldenFile(Stream: TTextStream): TMixture;
function CheckLMatrix(State: TElectronicState; Threshold: Double; Verbose: Boolean = False):Double;

implementation

uses SysUtils, Math, MathSup, CODATA, pgDOM, pgXMLRead, MNode, InternalCoords,
     Global, Worker, UCompat, NuclearInfo, Units, NextTok;

function MakeBase(Mixture: TMixture):TVibratingMolecule;
var
  Species: TSpecies;
begin
  Mixture.Name := 'Mixture';
  Species := TSpecies.Create(Mixture);
  Species.Name := 'LmatrixImport';
  Result := TVibratingMolecule.Create(Species);
  Result.Name := 'LmatrixImport';
end;

function ExtractBase(Name: string; out Id: Integer):string;
var
  p: Integer;
begin
  { Given a name like H12, return H and set Id to 12 }
  Id := -1;
  for p := 1 to Length(Name) do
    if Name[p] in ['0'..'9'] then begin
      Result := Copy(Name,1,p-1);
      Id := StrToInt(Copy(Name,p,Length(Name)));
      Exit;
    end;
  Result := Name; { No number }
end;

procedure NumberNucleus(Molecule: TMolecule; Id: Integer; Base: string);
var
  I, NId, LastId: Integer;
begin
  { Set nucleus name; number identical nuclei }
  LastId := -1;
  with Molecule do begin
    for I := 0 to Id - 1 do
      if ExtractBase(Nuclei[i].Name,NId) = Base then begin
        if NId < 0 then begin
          NId := 1;
          Nuclei[i].Name := Base + IntToStr(Nid);
        end;
        LastId := NId;
      end;
    if LastId > 0 then
      Base := Base + IntToStr(LastId+1);
    Nuclei[Id].Name := Base;
  end;
end;

function AddState(nStates: Integer; Molecule: TMolecule):TElectronicState;
var
  Manifold: TVibrationalManifold;
begin
  Manifold := TVibrationalManifold.Create(Molecule);
  Manifold.Name := 'S'+IntToStr(nStates);
  Result := TElectronicState.Create(Manifold);
  Result.Name := Manifold.Name;
  Molecule.CheckProperties;
end;

function AddAtom(State: TElectronicState; var n: Integer; { Current atom count }
                  Element: string; x,y,z: Double; UseAverageMass, WriteNames: Boolean):Double;
var
  Nuc: TNuclearInfo;
  Molecule: TVibratingMolecule;
  ij: Integer;
  Nu: TMolecularNucleus;
begin
  Result := 0;
  Nuc := FindMostAbundantNucleus(Element, True);
  if Nuc.Mass < 0 then
    raise EReportException.Create('Unknown element:'+Element);
  if Nuc.Mass = 0 then
    Exit;
  Molecule := State.Molecule;
  if n >= Molecule.nNuclei then begin
    { New nucleus }
    Molecule.nNuclei := n+1;
    NumberNucleus(Molecule,n,Nuc.Symbol);
  end else if WriteNames then
    NumberNucleus(Molecule,n,Nuc.Symbol)
  else begin
    { Old nucleus - check element matches }
    if ExtractBase(State.Nucleus(n).Name,ij) <> Nuc.Symbol then
      raise EReportException.CreateFmt('Atoms out of order: %s %s',[State.Nucleus(n).Name, Nuc.Symbol]);
  end;
  Nu := State.Nucleus(n) as TMolecularNucleus;
  if UseAverageMass then begin
    Nu.Mass := Nuc.StandardWeight;
    Result := Nuc.Mass;
  end else begin
    Nu.Mass := Nuc.Mass;
    Result := Nuc.StandardWeight;
  end;
  Nu.x := x;
  Nu.y := y;
  Nu.z := z;
  Inc(n);
end;

procedure AddMode(State: TElectronicState; var n: Integer; Frequency: Double; c: array of Double);
var
  Molecule: TVibratingMolecule;
  o, i: Integer;
begin
  Molecule := State.Molecule;
  if Molecule.nNuclei = 0 then
    Molecule.nNuclei := Length(c) div 3
  else if 3*Molecule.nNuclei <> Length(c) then
    raise EReportException.CreateFmt('Nuclear count mismatch: %d nuclei, %d modes',[Molecule.nNuclei, Length(c)]);
  if Frequency < 1 then
    Exit; { Skip low frequency modes }
  if n > 3*Molecule.nNuclei-5 then
    raise EReportException.Create('Too many modes for Nuclei');
  if n >= Molecule.nModes then
    Molecule.nModes := n+1;
  State.Mode(n).Omega := Frequency;
  o := State.Mode(n).FirstDynamicParameter;
  for i := 0 to High(c) do
    State.Mode(n).SetParameter(o+i,c[i]);
  Inc(n);
end;

procedure AddFCF(State,FirstState: TElectronicState);
begin
  State.Molecule.CheckProperties;
  FirstState.Manifold.Initial := True;
  if State <> FirstState then
    TMultidimensionalFCF.Create(TTransitionMoments.Create(State.Molecule,State.Manifold.Name,FirstState.Manifold.Name),
                                State.Name, FirstState.Name).Calculate := True;
end;

function FindValue(Name: string; Node:TDOMNode; BlankOK: Boolean = False):string;
var
  Attr: TDOMNode;
begin
  Attr := Node.Attributes.GetNamedItem(Name);
  if not Assigned(Attr) then
    if BlankOK then
      Result := ''
    else
      raise EReportException.CreateFmt('Can''t find attribute %s in %s',[Name,Attr.NodeName])
  else
    Result := Trim(Attr.NodeValue);
end;

{ Import control file from ezSpectrum program - Vadim Mozhayskiy
  http://iopenshell.usc.edu/downloads/ }

function ImportezSpectrumXLMFile(XMLDocument: TXMLDocument): TMixture;
var
  Molecule: TVibratingMolecule;
  nStates: Integer;
  InitialState: TElectronicState;
  InitialZptE: Double;
  procedure ReadJobParameters(Node: TDOMNode);
  begin
    with Result.Simulations[0] do
      Temperature := StrToFloatDef(FindValue('temperature',Node),Temperature);
  end;
  procedure ReadState(Node: TDOMNode; Initial: Boolean);
  var
    State: TElectronicState;
    b, i, j, k, ij: Integer;
    line, s: string;
    linear, mass_weighted, ChangeMade: Boolean;
    v, ZptE, x, y, z: Double;
    OrgOrder: array of TVibrationalMode;
    NamesSet: Boolean;
  begin
    State := AddState(nStates,Molecule);
    Inc(nStates);
    ZPtE := 0;
    Node := Node.FirstChild;
    NamesSet := False;
    while Assigned(Node) do begin
      if Node.NodeName = 'ip' then
        State.Origin := Convert(StrToFloat((Node.FirstChild as TDOMCharacterData).Data),eV,cm1)
      else if Node.NodeName = 'geometry' then begin
        linear := StrToBoolDef(FindValue('linear',Node,True),False);
        i := StrToInt(FindValue('number_of_atoms',Node));
        if Molecule.nNuclei = 0 then begin
          Molecule.nNuclei := i;
          if linear then
            Molecule.nModes := 3*i-5
          else
            Molecule.nModes := 3*i-6;
        end else if i <> Molecule.nNuclei then
          raise EReportException.CreateFmt('Inconsistent number of atoms: %d %d',[i, Molecule.nNuclei]);
        line := FindValue('text',Node);
        i := 0;
        while line <> '' do begin
          s := NextToken(line);
          x := StrToFloat(NextToken(line));
          y := StrToFloat(NextToken(line));
          z := StrToFloat(NextToken(line));
          AddAtom(State, i, s, x, y, z, False, not NamesSet);
        end;
        NamesSet := True;
      end else if Node.NodeName = 'normal_modes' then begin
        line := FindValue('atoms',Node);
        mass_weighted := StrToBoolDef(FindValue('if_mass_weighted',Node,True),True);
        i := 0;
        while line <> '' do begin
          s := NextToken(line);
          if ExtractBase(Molecule.Nuclei[i].Name,ij) <> s then
            raise EReportException.CreateFmt('Atoms out of order in modes: %s %s',[Molecule.Nuclei[i].Name, s]);
          Inc(i);
        end;
        line := FindValue('text',Node);
        for b := 0 to (Molecule.nModes-1) div 3 do
          for j := 0 to Molecule.nNuclei - 1 do
            for i := b to Min(b+2,Molecule.nNuclei-1) do
              with State.Mode(i) do
                for k := 0 to 2 do begin
                  v := StrToFloat(NextToken(line));
                  if not mass_weighted then
                    v := v/Sqrt((State.Nucleus(j) as TMolecularNucleus).Mass);
                  Parameters[FirstDynamicParameter+3*j+k] := v;
                end;
      end else if Node.NodeName = 'frequencies' then begin
        line := FindValue('text',Node);
        i := 0;
        while line <> '' do begin
          v := StrToFloat(NextToken(line));
          ZptE := ZPtE + v/2;
          State.Mode(i).Omega := v;
          Inc(i);
        end;
      end else if Node.NodeName = 'manual_coordinates_transformation' then begin
        if StrToBoolDef(FindValue('if_manual',Node,True),False) then
          raise EReportException.Create(Node.NodeName+' not implemented');
      end else if Node.NodeName = 'manual_normal_modes_reordering' then begin
        if StrToBoolDef(FindValue('if_manual',Node,True),False) then begin
          { Save original order }
          SetLength(OrgOrder, Molecule.nModes);
          for i := 0 to Molecule.nModes - 1 do
            OrgOrder[i] := State.Mode(i);
          for i := 0 to Molecule.nModes do begin
            ChangeMade := False;
            line := FindValue('new_order',Node,True);
            for j := 0 to Molecule.nModes - 1 do begin
              k := StrToInt(NextToken(line));
              if OrgOrder[j].ComponentIndex <> k then begin
                OrgOrder[j].ComponentIndex := k;
                ChangeMade := True;
              end;
            end;
            if not ChangeMade then
              Break;
          end;
          if ChangeMade then
            raise EReportException.Create('Can''t reorder modes');
        end;
      end;
      Node := Node.NextSibling;
    end;
    State.Normalize(nil);
    State.DoCheckOrthogonal(2e-3);
    CheckLMatrix(State,0.0001);
    Molecule.CheckProperties;
    if not Initial then begin
      AddFCF(State,InitialState);
      State.Origin := State.Origin - ZptE + InitialZptE;
    end else begin
      InitialState := State;
      InitialZptE := ZptE;
    end;
  end;
var
  Node: TDOMNode;
begin
  nStates := 0;
  InitialState := nil;
  Result := TMixture.Create(nil);
  try
    Molecule := MakeBase(Result);
    Node := XMLDocument.FirstChild;
    if Node.NodeName <> 'root' then
      raise EReportException.Create('Expected root node');
    Node := Node.FirstChild;
    while Assigned(Node) do begin
      if Node.NodeName = 'job_parameters' then
        ReadJobParameters(Node)
      else if Node.NodeName = 'initial_state' then
        ReadState(Node,True)
      else if Node.NodeName = 'target_state' then
        ReadState(Node,False);
      Node := Node.NextSibling;
    end;
  except
    Result.Free;
    raise;
  end;
end;

function ImportMOLPROXLMFile(XMLDocument: TXMLDocument): TMixture;
var
  Molecule: TVibratingMolecule;
  nStates: Integer;
  AtomArray, VibNode: TDOMNode;
  FirstState: TElectronicState;
  procedure LoadAtoms(State: TElectronicState);
  var
    Atom: TDOMNode;
    i: Integer;
    s: string;
    x, y, z: Double;
  begin
    Atom := AtomArray.FirstChild;
    { Count number of atoms }
    i := 0;
    while Assigned(Atom) do begin
      Inc(i);
      Atom := Atom.NextSibling;
    end;
    if (Molecule.nNuclei <> 0) and (Molecule.nNuclei <> i) then
      raise EReportException.CreateFmt('Inconsistent number of atoms %d %d',[Molecule.nNuclei,i]);
    Atom := AtomArray.FirstChild;
    i := 0;
    while Assigned(Atom) do begin
      s := FindValue('elementType',Atom);
      x := StrToFloat(FindValue('x3',Atom));
      y := StrToFloat(FindValue('y3',Atom));
      z := StrToFloat(FindValue('z3',Atom));
      AddAtom(State,i, s, x, y, z, True, False);
      Atom := Atom.NextSibling;
    end;
  end;
  procedure LoadVibrations;
  var
    State: TElectronicState;
    Mode: TDOMNode;
    n, i: Integer;
    nu: Double;
    line, s: string;
    c: array of Double;
  begin
    if VibNode = nil then
      Exit;
    State := AddState(nStates,Molecule);
    LoadAtoms(State);
    Mode := VibNode.FirstChild;
    n := 0;
    while Assigned(Mode) do begin
      if Mode.NodeName = 'normalCoordinate' then begin
        Nu := StrToFloat(FindValue('wavenumber',Mode));
        if Mode.FirstChild is TDOMCharacterData then
          line := (Mode.FirstChild as TDOMCharacterData).Data
        else
          line := '';
        i := 0;
        while line <> '' do begin
          SetLength(c, i+1);
          s := NextToken(line);
          if (s <> '') and TryStrToFloatEx(s,c[i]) then
            Inc(i);
        end;
        SetLength(c,i);
        AddMode(State, n, Nu, c);
      end;
      Mode := Mode.NextSibling;
    end;
    State.FixupProperties.Free;
    State.DoCheckOrthogonal(2e-4);
    CheckLMatrix(State,0.1);
    if nStates = 0 then
      FirstState := State;
    AddFCF(State,FirstState);
    Inc(nStates);
    AtomArray := nil;
    VibNode := nil;
  end;
var
  Job, JobStep, Node, Attr: TDOMNode;
begin
  nStates := 0;
  FirstState := nil;
  VibNode := nil;
  Result := TMixture.Create(nil);
  try
    Molecule := MakeBase(Result);
    Job := XMLDocument.FirstChild;
    if Job.NodeName <> 'molpro' then
      raise EReportException.Create('Expected molpro node');
    Job := Job.FirstChild;
    while Assigned(Job) do begin
      if Job.NodeName = 'job' then begin
        JobStep := Job.FirstChild;
        while Assigned(JobStep) do begin
          if JobStep.NodeName = 'jobstep' then begin
            Attr := JobStep.Attributes.GetNamedItem('command');
            if Assigned(Attr) and (Attr.NodeValue = 'FREQ') then begin
              Node := JobStep.FirstChild;
              while Assigned(Node) do begin
                if Assigned(VibNode) and (Node.NodeName = 'jobstep') then begin
                  { Take next atom array, if present, as atom array in FREQ node uses different axis
                    system from the vibrations }
                  Node := Node.FirstChild;
                  Continue;
                end;
                if Node.NodeName = 'cml:atomArray' then
                  AtomArray := Node
                else if Node.NodeName = 'vibrations' then
                  VibNode := Node;
                Node := Node.NextSibling;
              end;
              LoadVibrations;
            end;
          end;
          JobStep := JobStep.NextSibling;
        end;
      end;
      Job := Job.NextSibling;
    end;
  except
    Result.Free;
    raise;
  end;
end;

function ImportGaussianFile(Stream: TTextStream): TMixture;
var
  State, FirstState: TElectronicState;
  Molecule: TVibratingMolecule;
  nStates: Integer;
  line: string;
  Atoms: array of string;
  function NumericLine: Boolean;
  begin
    Result := (Length(line) > 0) and (line [1] in ['0'..'9']);
  end;
  procedure ReadAtoms;
  var
    n: Integer;
  begin
    { Find atom co-ordinate lines - just save for now }
    repeat
      line := Trim(Stream.RestOfLine);
    until Stream.AtEndOfFile or NumericLine ;
    n := 0;
    while not Stream.AtEndOfFile and NumericLine do begin
      if n >= Length(Atoms) then
        Setlength(Atoms,n+1);
      Atoms[n] := line;
      Inc(n);
      line := Trim(Stream.RestOfLine);
    end;
    SetLength(Atoms,n);
  end;
  procedure ParseAtoms;
  var
    n, i: Integer;
    Nuc: TNuclearInfo;
    s: string;
    x, y, z: Double;
  begin
    n := 0;
    for i := 0 to High(Atoms) do begin
      s := Atoms[i];
      if i+1 <> StrToInt(NextToken(s)) then
        raise EReportException.Create('Atom out of order:'+Atoms[i]);
      Nuc := FindMostAbundantNucleus(StrToInt(NextToken(s)));
      if Nuc.Mass < 0 then
        raise EReportException.Create('Bad Atom:'+Atoms[i]);
      NextToken(s); { Skip atom type }
      x := StrToFloat(NextToken(s));
      y := StrToFloat(NextToken(s));
      z := StrToFloat(NextToken(s));
      AddAtom(State,n,Nuc.Symbol,x,y,z,False,False);
    end;
  end;
  function ReadFrequencies: Boolean;
  var
    s: string;
    Frequencies: array of Double;
    c: array of array of Double;
    n, i, j, k: Integer;
    HighPrecision: Boolean;
  begin
    State := AddState(nStates,Molecule);
    Inc(nStates);
    ParseAtoms;
    SetLength(c,3*Molecule.nNuclei);
    n := 0;
    repeat
      { Read Frequencies line }
      SetLength(Frequencies,0);
      NextToken(line); { Skip Frequencies label }
      NextToken(line); { Skip -- }
      while line <> '' do begin
        s := NextToken(line);
        SetLength(Frequencies,Length(Frequencies)+1);
        Frequencies[High(Frequencies)] := StrToFloat(s);
      end;
      { Skip until co-ordinates line }
      repeat
        line := Trim(Stream.RestOfLine);
      until Stream.AtEndOfFile or NumericLine ;
      SetLength(c,Length(Frequencies),3*Molecule.nNuclei);
      { if third token is an integer assume high precision mode used }
      s := line;
      NextToken(s);
      NextToken(s);
      HighPrecision := TryStrToInt(NextToken(s),i);
      if HighPrecision then
        { high precision mode }
        for i := 0 to Molecule.nNuclei - 1 do
          for k := 0 to 2 do begin
            if k+1 <> StrToInt(NextToken(line)) then
              raise EReportException.Create('Coord out of order:'+line);
            if i+1 <> StrToInt(NextToken(line)) then
              raise EReportException.Create('Atom out of order:'+line);
            NextToken(line); { Skip atomic number }
            for j := 0 to High(Frequencies) do
              c[j,3*i+k] := StrToFloat(NextToken(line));
            line := Trim(Stream.RestOfLine);
          end
      else
        { Default low precision mode }
        for i := 0 to Molecule.nNuclei - 1 do begin
          if i+1 <> StrToInt(NextToken(line)) then
            raise EReportException.Create('Atom out of order:'+line);
          NextToken(line); { Skip atomic number }
          for j := 0 to High(Frequencies) do
            for k := 0 to 2 do
              c[j,3*i+k] := StrToFloat(NextToken(line));
          line := Trim(Stream.RestOfLine);
        end;
      for j := 0 to High(Frequencies) do
        AddMode(State,n,Frequencies[j],c[j]);
      for i := 0 to 4 do begin
        line := Trim(Stream.RestOfLine);
        if Pos('Frequencies',line) = 1 then
          Break;
      end;
    until Stream.AtEndOfFile or (Pos('Frequencies',line) <> 1);
    State.FixupProperties.Free;
    State.Normalize(nil);
    if HighPrecision then begin
      State.DoCheckOrthogonal(2e-4);
      CheckLMatrix(State,0.0001);
    end else begin
      State.DoCheckOrthogonal(2e-2);
      CheckLMatrix(State,0.3);
    end;
    if not Assigned(FirstState) then
      FirstState := State;
    AddFCF(State,FirstState);
    Result := HighPrecision;
  end;
begin
  State := nil;
  FirstState := nil;
  nStates := 0;
  Result := TMixture.Create(nil);
  Stream.FlowToNextLine := True;
  try
    Molecule := MakeBase(Result);
    repeat
      line := Trim(Stream.RestOfLine);
      if (Pos('Standard orientation', line) = 1) or (Pos('Input orientation', line) = 1) then
        { relies on most recent co-ordinate set being the correct one for the co-ordinates given;
          typically "Standard orientation" follows "Input orientation" if nosymm is not set }
        ReadAtoms
      else if (Length(Atoms) > 0) and (Pos('Frequencies',line) = 1) then begin
        if ReadFrequencies then
          SetLength(Atoms,0); // High Precision set can have following low precision set following; this forces discard
      end;
    until Stream.AtEndOfFile;
    if not Assigned(FirstState) then
      raise EReportException.Create('No Vibrations found');
  except
    Result.Free;
    raise;
  end;
end;

function ImportLmatrixFile(Stream: TTextStream): TMixture;
var
  P: Int64;
  C: AnsiChar;
  XMLDocument: TXMLDocument;
begin
  Result := nil;
  { Find first non-blank character }
  P := Stream.Position;
  while Stream.Read(C,SizeOf(C)) = 1 do
    if C > ' ' then
      Break;
  Stream.Position := P;
  if C = '<' then begin
    ReadXMLFile(XMLDocument, Stream);
    try
      if XMLDocument.FirstChild.NodeName = 'molpro' then
        Result := ImportMOLPROXLMFile(XMLDocument)
      else if XMLDocument.FirstChild.NodeName = 'root' then
        Result := ImportezSpectrumXLMFile(XMLDocument)
      else
        raise EReportException.Create('Unknown XML file type');
    finally
      XMLDocument.Free;
    end;
  end else if C = '[' then
    Result := ImportMoldenFile(Stream)
  else
    while (Stream.Line <= 5) and not Assigned(Result) do
      if Pos('Gaussian', Stream.RestOfLine) > 0 then begin
        Result := ImportGaussianFile(Stream);
        Break;
      end else
        Stream.NextLine;
  if not Assigned(Result) then
    raise EReportException.Create('File format not recognized');
end;

function ImportMoldenFile(Stream: TTextStream): TMixture;
var
  State, FirstState: TElectronicState;
  Molecule: TVibratingMolecule;
  nStates: Integer;
  Frequencies, OtherMass: array of Double;
  UseAverageMass, MassStyleSet: Boolean;
  procedure ReadFrequencies;
  var
    V: Double;
  begin
    { Start of new section }
    State := AddState(nStates,Molecule);
    Stream.NextLine;
    SetLength(Frequencies,0);
    while Stream.GetFreeItem and Stream.IsRealToken(V) do begin
      SetLength(Frequencies,Length(Frequencies)+1);
      Frequencies[High(Frequencies)] := V;
      Stream.NextLine;
    end;
    if not Assigned(FirstState) then
      FirstState := State;
    AddFCF(State,FirstState);
  end;
  procedure ReadAtoms;
  var
    n: Integer;
    s: string;
    x, y, z, m: Double;
  begin
    n := 0;
    repeat
      Stream.NextLine;
      if not Stream.GetFreeItem or (Stream.Token[1] = '[') then
        Break;
      s := Stream.Token;
      if Stream.NextDoubleVar(x) and Stream.NextDoubleVar(y) and Stream.NextDoubleVar(z) then begin
        m := AddAtom(State,n,s,x*BohrRadius*1e10,y*BohrRadius*1e10,z*BohrRadius*1e10,UseAverageMass,False);
        if m > 0 then begin
          SetLength(OtherMass, n);
          OtherMass[n-1] := m;
        end;
      end else
        raise EReportException.Create('Bad atom:'+s);
    until False;
  end;
  procedure ReadModes;
  var
    n, i, j, k: Integer;
    c: array of Double;
    AvErr, ExactErr, MA, ME, V: Double;
  begin
    SetLength(c, 3*Molecule.nNuclei);
    n := 0;
    Stream.NextLine;
    AvErr := 0; { Calculate normalisation constant with both sorts of masses }
    ExactErr := 0;
    for k := 0 to High(Frequencies) do begin
      AvErr := AvErr - 1;
      ExactErr := ExactErr - 1;
      if CompareText(Stream.NextString(''),'vibration') <> 0 then
        raise EReportException.Create('Vibration expected, found '+Stream.Token);
      Stream.NextLine; { Ignore vibration number }
      for i := 0 to Molecule.nNuclei - 1 do begin
        MA := (State.Nucleus(i) as TMolecularNucleus).Mass;
        ME := OtherMass[i];
        for j := 0 to 2 do begin
          Stream.NextDoubleVar(V);
          AvErr := AvErr + Sqr(V)*MA;
          ExactErr := ExactErr + Sqr(V)*ME;
          c[3*i+j] := V;
        end;
        Stream.NextLine;
      end;
      AddMode(State,n,Frequencies[k],c);
    end;
    if not MassStyleSet and (Abs(AvErr) < 0.1) and (Abs(ExactErr) < 0.1) then begin
      if Abs(ExactErr) < Abs(AvErr)/2 then begin
        UseAverageMass := False;
        for i := 0 to Molecule.nNuclei - 1 do
          (State.Nucleus(i) as TMolecularNucleus).Mass := OtherMass[i];
        ShowInfo('Using exact rather than average mass for molden file import');
      end;
    end;
    MassStyleSet := True;
    State.FixupProperties.Free;
    State.Normalize(nil);
    State.DoCheckOrthogonal(1e-4);
    CheckLMatrix(State,0.0001);
  end;
begin
  UseAverageMass := True;
  MassStyleSet := False;
  Stream.FlowToNextLine := False;
  Stream.FlowToNextRecord := True;
  if CompareText(Stream.NextString(''),'[Molden') <> 0 then
    raise EReportException.Create('Missing [Molden Format] header');
  State := nil;
  FirstState := nil;
  nStates := 0;
  Result := TMixture.Create(nil);
  try
    Molecule := MakeBase(Result);
    repeat
      { Find section header }
      while not Stream.AtEndOfFile and ((Stream.Token = '') or(Stream.Token[1] <> '[')) do begin
        Stream.NextLine;
        Stream.GetFreeItem;
      end;
      if CompareText(Stream.Token, '[FREQ]') = 0 then
         ReadFrequencies
      else if CompareText(Stream.Token, '[FR-COORD]') = 0 then
        ReadAtoms
      else if CompareText(Stream.Token, '[FR-NORM-COORD]') = 0 then
        ReadModes
      else
        Stream.GetFreeItem;
    until Stream.AtEndOfFile;
  except
    Result.Free;
    raise;
  end;
end;

function CheckLMatrix(State: TElectronicState; Threshold: Double; Verbose: Boolean):Double;
var
  CD: TCartesianDisplacements;
  SC: TSymmetryCoords;
  i: Integer;
  InF: Double;
  Mixture: TMixture;
  ErrMsg: string;
begin
  { Use to check imported lmatrices by calculating the frequencies via the cartesian displacements }
  { May pick up issues with co-ordinate definitions }
  { Creates and destroys some objects, so should only be used if no object InternalCoords are present }
  Mixture := State.RootNode as TMixture;
  CD := TCartesianDisplacements.Create(State);
  SC := nil;
  ErrMsg := '';
  try
    CD.Prepare(False);
    CD.CalculateFromL(False,True);
    SC := TSymmetryCoords.Create(State);
    SC.Active := False;
    SC.PreserveS := True;
    //SC.MakeU(CD);
    SC.CalculateF(Verbose,True);
    SC.MakeLMatrix(State, Verbose, CD);
    ErrMsg := SC.LastError;
    if ErrMsg = '' then begin
      Result := 0;
      for i := 1 to SC.lambda.Size do begin
        InF := Convert(State.Mode(i-1).Omega, Mixture.Units, cm1);
        InF := Abs(InF - sqrt(SC.lambda[i]/LambdaConversion));
        if InF > Result then
          Result := InF;
      end;
      if Verbose then
        WriteLn(Log, 'Maximum Error ', Result, ' cm-1');
      if Result > Threshold then
        ErrMsg := 'Warning - '+State.UniqueName+' - frequencies inconsistent with l matrix; maximum error '+DtosM(Result,10,-1)+' cm-1';
    end else if Verbose then
      WriteLn(Log, ErrMsg);
  finally
    CD.Free;
    SC.Free;
  end;
  if not Verbose and (ErrMsg <> '') then begin
    ShowInfo(ErrMsg);
    CheckLMatrix(State, 0, True);
  end;
end;

end.
