unit ImportSTROTAB;
{$I utils.inc}{Copyright (C) 2008 C M Western; see commandline.pas for license}

{ Import constants from .sta files as used by the STROTAB program.
  John Kodet, Richard H. Judge Computer Physics Communications 176 601 (2007)
}
interface

uses TextStrm, Mixture, Obs;

function ImportSTROTABFile(Stream: TTextStream): TMixture;

implementation

uses {$IFNDEF CONSOLE} Dialogs, {$ENDIF}
     MathSup, SysUtils, OAsymT, Worker, MNode, Abelian;

const NameChanges: array [0..7] of array [0..1] of string =
 (('DELTAJK', 'deltaK'), ('d1', 'deltaJ'), ('d1', 'deltaK'),
  ('DSKJ','DsKN'), ('DSJK','DsNK'), ('DSJ','DsN'), ('DELTASJ','ds1'), ('DELTASK','ds22')
 );

function ImportSTROTABFile(Stream: TTextStream): TMixture;
var
  SectionName, Buffer: string;
  nErrors: Integer;
  Name, Value: string;
  procedure ReportError(Msg: string);
  begin
    Inc(nErrors);
    WriteLn(Log, 'Line ', Stream.Line, ' ', Msg);
  end;
  procedure Unknown;
  begin
    ReportError('Unknown or unimplemented setting in '+SectionName+': '+Buffer);
  end;
  function FindNextLine: Boolean;
  begin
    repeat
      Result := not Stream.AtEndOfFile;
      if not Result then
        Exit;
      Buffer := Trim(Stream.RestOfLine);
      Stream.NextLine;
    until Buffer <> '';
  end;
  function FindNextItem: Boolean;
  begin
    Result := FindNextLine and (Buffer[1] <> '}');
  end;
  function FindNextSection: Boolean;
  begin
    Result := True;
    while FindNextLine do begin
      if Buffer[1] = '[' then begin
        SectionName := UpperCase(Copy(Buffer, 2, Length(Buffer)-2));
        if FindNextLine and (Buffer[1] = '{') then
          Exit
        else
          ReportError('Bad start to '+SectionName);
      end;
    end;
    Result := False;
  end;
  function FindNextValue: Boolean;
  var
    p: Integer;
  begin
    repeat
      Result := FindNextItem;
      if Result then begin
        p := Pos('=',buffer);
        if p > 1 then begin
          Name := Trim(Copy(Buffer,1,p-1));
          Value := Trim(Copy(Buffer,p+1,Length(Buffer)));
          Exit;
        end else
          ReportError('Bad line:'+Buffer);
      end;
    until not Result;
  end;
  procedure ReadState(State: TAsymmetricTop);
  var
    i: Integer;
    v: Double;
    Found: Boolean;
  begin
    while FindNextValue do begin
      if not TryStrToFloatEx(Value, v) then
        ReportError('Bad value:'+Value);
      if Name = 'a' then
      else if Name = 'b' then
      else if Name = 'a0' then
      else begin
        for i := 0 to High(NameChanges) do
          if CompareText(NameChanges[i][0], Name) = 0 then begin
            Name := NameChanges[i][1];
            Break;
          end;
        Found := False;
        for i := 1 to State.nParameters do
          if CompareText(Name,State.ParameterNames[i]) = 0 then begin
            State.Parameters[i] := v;
            Found := True;
            Break;
          end;
        if not Found then
          Unknown;
      end;
    end;
  end;
var
  Species: TSpecies;
  Molecule: TAsymmetricMolecule;
  Ground, Excited: TAsymmetricTop;
  GroundM, ExcitedM: TAsymmetricManifold;
  TM: TTransitionMoments;
  procedure ReadCalcParameters;
  var
    v: Integer;
  begin
    while FindNextValue do begin
      Name := UpperCase(Name);
      if not TryStrToInt(Value,v) then
        ReportError('Bad Value:'+Value);
      if Name = 'JSMAX' then
        GroundM.Jmax := 2*v
      else if Name = 'JSMIN' then
        GroundM.Jmin := 2*v
      else if Name = 'JTMAX' then
        ExcitedM.Jmax := 2*v
      else if Name = 'JTMIN' then
        ExcitedM.Jmin := 2*v
      else
        Unknown;
    { TODO: 'PTMAX' 'KTMAX' 'PKTMAX' 'PSMAX' 'KSMAX' 'PKSMAX''PMAX' 'KMAX' 'PKMAX' }
    end;
  end;
  procedure ReadSymmetryParameters;
  begin
    while FindNextValue do begin
      Name := UpperCase(Name);
      Value := UpperCase(Value);
      if Name = 'POINT_GROUP' then
        Molecule.SetSettingFromStr('PointGroup',Value)
      else if Name = 'REPRESENTATION' then begin
        if Value[1] in ['1'..'3'] then
          Value := StringOfChar('I',Ord(Value[1])-Ord('0'))+Copy(Value,2,Length(Value));
        Molecule.SetSettingFromStr('Representation',Value)
      end else if Name = 'REDUCTION' then
        Molecule.SReduction := Value[1] = 'S'
      else if Name = 'TRANSITION_PRODUCT' then
        Excited.SetSettingFromStr('Symmetry',Value)
      else
        Unknown;
    end;
  end;
  procedure ReadIntensityParameters;
  begin
    while FindNextValue do begin
      Name := UpperCase(Name);
      if (Name = 'EE') or (Name = 'EO') or (Name = 'OE') or (Name = 'OO') then
        Molecule.SetSettingFromStr(Name+'wt',Value)
      else if (Name = 'X') or (Name = 'Y') or (Name = 'Z') then
        with TCartesianTransitionMoment.Create(TM,Excited.Name,Ground.Name) do begin
          SetSettingFromStr('Axis',Name);
          SetSettingFromStr('Value',Value);
        end
      else if Name = 'TEMPERATURE' then
        Result.Simulations[0].Temperature := StrToFloat(Value)
      else
        Unknown;
    end;
  end;
  procedure ReadContourParameters;
  begin
    while FindNextValue do begin
      Name := UpperCase(Name);
      if Name = 'BOXES' then
        Result.Simulations[0].nDF := StrToInt(Value)
      else if Name = 'RANGEMAX' then
        Result.Simulations[0].Fmax := StrToFloat(Value)
      else if Name = 'RANGEMIN' then
        Result.Simulations[0].Fmin := StrToFloat(Value)
      else if Name = 'FWHM' then
        Result.Simulations[0].Gaussian := StrToFloat(Value)
      else
        Unknown;
    end;
  end;
begin
  nErrors := 0;
  Result := TMixture.Create(nil);
  try
    Result.Name := 'Mixture';
    Species := TSpecies.Create(Result);
    Species.Name := 'Species';
    Molecule := TAsymmetricMolecule.Create(Species);
    with Molecule do begin
      Name := 'Asymmetric Top';
      PointGroup := C2v;
    end;
    GroundM := TAsymmetricManifold.Create(Molecule);
    with GroundM do begin
      Name := 'Ground';
      Initial := True;
    end;
    Ground := TAsymmetricTop.Create(GroundM);
    with Ground do begin
      Name := 'v=0';
    end;
    ExcitedM := TAsymmetricManifold.Create(Molecule);
    with ExcitedM do begin
      Name := 'Excited';
    end;
    Excited := TAsymmetricTop.Create(ExcitedM);
    with Excited do begin
      S := 2;
      Name := 'v=0';
    end;
    TM := TTransitionMoments.Create(Molecule);
    with TM do
      SetBraKetName(ExcitedM.Name,GroundM.Name);
    while FindNextSection do begin
      if SectionName = 'TITLE' then begin
        while FindNextItem do
          Molecule.Comment := Molecule.Comment + ' ' + Buffer;
      end else if SectionName = 'GROUND STATE' then
        ReadState(Ground)
      else if SectionName = 'EXCITED STATE' then
        ReadState(Excited)
      else if SectionName = 'CALCULATION PARAMETERS' then
        ReadCalcParameters
      else if SectionName = 'SYMMETRY PARAMETERS' then
        ReadSymmetryParameters
      else if SectionName = 'TRANSITIONS' then
        { Ignore }
      else if SectionName = 'LEAST SQUARES' then
        { Ignore }
      else if SectionName = 'INTENSITY PARAMETERS' then
        ReadIntensityParameters
      else if SectionName = 'OUTPUT' then
        { Ignore }
      else if SectionName = 'CONTOUR PARAMETERS' then
        ReadContourParameters
      else
        ReportError('Unknown Section Name:'+SectionName);
    end;
    {$IFNDEF CONSOLE}
    if nErrors > 0 then
      ShowMessage('Problems with import - see log window');
    {$ENDIF}
  except
    Result.Free;
    raise;
  end;
end;

end.
