unit ImportSpecView;
{$I utils.inc}{Copyright (C) 2007 C M Western; see commandline.pas for license}
{ Import data file from Ohio SpecView program }
interface

uses TextStrm, Mixture;

function ImportSpecViewFile(S: TTextStream): TMixture;

implementation

uses Classes, SysUtils, OAsymT, HundA, Worker, Units, Math, Abelian;

function ImportSpecViewFile(S: TTextStream): TMixture;
var
  Species: TSpecies;
  Molecule: TMolecule;
  ManifoldL, ManifoldU: TManifold;
  StateL, StateU: TState;
  TransitionMoments: TTransitionMoments;
  IsGHz: Boolean;
  IsMicrowave: Boolean;
  Float: array of Boolean;
  ParamStart: int64;
  Origin: Double;

  procedure MakeManifoldPair(out ManifoldL, ManifoldU: TManifold;
                             out TransitionMoments: TTransitionMoments);
  begin
    ManifoldL := Molecule.ChildClassWithType(TManifold).Create(Molecule) as TManifold;
    with ManifoldL do begin
      Name := 'Ground';
      Initial := True;
      StateL := ChildClassWithType(TState).Create(ManifoldL) as TState;
      StateL.Name := 'Lower';
    end;
    if IsMicrowave then begin
      ManifoldU := ManifoldL;
      StateU := StateL;
    end else begin
      ManifoldU := Molecule.ChildClassWithType(TManifold).Create(Molecule) as TManifold;
      with ManifoldU do begin
        Name := 'Excited';
        StateU := ChildClassWithType(TState).Create(ManifoldU) as TState;
        StateU.Name := 'Upper';
      end;
    end;
    TransitionMoments := TTransitionMoments.Create(Molecule);
    TransitionMoments.SetBraKetName(ManifoldU.Name, ManifoldL.Name);
  end;

  procedure MakeBaseMixture(MolculeClass: TMoleculeClass; MakeManifold: Boolean = True);
  begin
    Result := TMixture.Create(nil);
    Species := TSpecies.Create(Result);
    Species.Name := 'Species';
    Molecule := MolculeClass.Create(Species);
    Molecule.Name := Molecule.ClassName;
    if MakeManifold then
      MakeManifoldPair(ManifoldL, ManifoldU, TransitionMoments)
    else
      StateU := nil;
  end;

  function MakeTransitonMoment(TransitionMomentClass: TTransitionMomentClass):TTransitionMoment;
  begin
    Result := TransitionMomentClass.Create(TransitionMoments);
    Result.SetBraKetName(StateU.Name, StateL.Name);
  end;

  procedure Next(out V; Count: Integer);
  begin
    if S.Read(V,Count) <> Count then
      raise Exception.Create('Unexpected end of file');
  end;

  function NR: Double;
  begin
    Next(Result, SizeOf(Double));
    {$IFDEF FPC_BIG_ENDIAN}
      Result := Double(SwapEndian(QWord(Result)));
    {$ENDIF}
  end;

  function NP: Double;
  begin
    Result := NR;
    if IsGHz then
      Result := Result * 1000;
  end;

  function NW: Integer;
  begin
    Next(Result, SizeOf(Integer));
    {$IFDEF FPC_BIG_ENDIAN}
     Result := SwapEndian(Result);
    {$ENDIF}
  end;

  procedure ReadHeader;
  begin
    with Result.Simulations[0] do begin
      Temperature := NR;
      Gaussian := NR;
      Lorentzian := NR;
      NR; { Min Intensity }
      NR; { Space to skip }
      NR; { Plot Resolution }
      case Round(NR) of
        0: Mixture.Units := cm1;
        1: Mixture.Units := MHz;
        2: begin Mixture.Units := MHz; IsGHz := True; end;
      else
        raise Exception.Create('Bad units');
      end;
      if IsGHz then begin
        Gaussian := Gaussian * 1000;
        Lorentzian := Lorentzian *1000;
      end;
    end;
  end;

  procedure ReadTail;
  begin
    NR; { 0 ?? }
    NW; { 1 ?? }
    NW; { 3 ?? }
    Origin := NP;
    NW; { 121 ?? }
    NW; { 121 ?? }
    NW; { 10000 ?? }
    IsMicrowave := NW <> 0;
    if not IsMicrowave and Assigned(StateU) then
      StateU.Origin := Origin;
    { Some additional bytes ignored }
    S.Position := ParamStart;
  end;

  procedure ReadFitFlags(n: Integer);
  var
    i: Integer;
  begin
    ParamStart := S.Position;
    S.Position := S.Position + Int64(n) * SizeOf(Double);
    SetLength(Float, n+1);
    for i := 0 to n do
      Float[i] := NW <> 0;
  end;

  procedure ReadAsym(WithSpin: Boolean);
  var
    d: Double;
    procedure ReadManifold(Manifold: TAsymmetricManifold);
    var
      d: Double;
      function MakePert: TAsymmetricTopPerturbation;
      begin
        Result := TAsymmetricTopPerturbation.Create(Manifold);
        Result.SetBraKetName(Manifold.State[0].Name, Manifold.State[0].Name);
        Result.Value := d;
      end;
    begin
      with Manifold.State[0] as TAsymmetricTop do begin
        if WithSpin then begin
          BBar := NP;
          A := NP + BBar;
          BDelta := 2*NP;
        end else begin
          A := NP;
          B := NP;
          C := NP;
        end;
        d := NP;
        if d <> 0 then
          MakePert.aPower := 1;
        d := NP;
        if d <> 0 then
          MakePert.bPower := 1;
        d := NP;
        if d <> 0 then
          MakePert.aPower := 3;
        DK := NP;
        DJK := NP;
        DJ := NP;
        deltaK := NP;
        deltaJ := NP;
        if WithSpin then begin
          eaa := NP;
          ebb := NP;
          ecc := NP;
          eabbar := 2*NP;
        end;
      end;

    end;
  begin
    { Import of Asym.Rotor(&int.rotation) }
    MakeBaseMixture(TAsymmetricMolecule);
    NW; NW;
    if WithSpin then
      ReadFitFlags(30)
    else
      ReadFitFlags(22);
    ReadHeader;
    Species.Jmax := Round(NR)*2;
    NR; { Max delta K }
    d := NR;
    if d <> 0 then
      (MakeTransitonMoment(TCartesianTransitionMoment) as TCartesianTransitionMoment).Axis := aAxis;
    d := NR;
    if d <> 0 then
      (MakeTransitonMoment(TCartesianTransitionMoment) as TCartesianTransitionMoment).Axis := bAxis;
    d := NR;
    if d <> 0 then begin
      (MakeTransitonMoment(TCartesianTransitionMoment) as TCartesianTransitionMoment).Axis := cAxis;
      if WithSpin then
        (StateU as TAsymmetricTop).Symmetry := 1; { Must have symmetry change }
    end;
    if WithSpin then begin
      NR; { LoSt K }
      NR; { UpSt K }
      NR; { LoSt restrict }
      NR; { UpSt restrict }
    end;
    ReadTail;
    ReadManifold(ManifoldL as TAsymmetricManifold);
    if not IsMicrowave then
      ReadManifold(ManifoldU as TAsymmetricManifold);
    with Molecule as TAsymmetricMolecule do
      if WithSpin then begin
        PointGroup := Cs;
        C2zAxis := cAxis;
        StateU.S := 1;
        StateL.S := 1;
      end else
        PointGroup := C1;
  end;

  procedure ReadLinearDoublet;
  var
    Offset: Integer;
    procedure SetStateType(State: TLinear);
    begin
      Inc(Offset, 7);
      case NW of
        0:  begin State.Lambda := 0; Dec(Offset, 3); end;
        1,2: State.Lambda := 2;
        3,4: State.Lambda := 4;
      end;
      State.S := 1;
    end;
    procedure ReadState(Manifold: TManifold);
    begin
      with Manifold.State[0] as TLinear do begin
        B := NP;
        D := NP;
        H := NP;
        gamma := NP;
        if Lambda > 0 then begin
          A := NP;
          p := NP;
          q := NP;
        end;
      end;
    end;
  begin
    { Import of Linear Doublet }
    MakeBaseMixture(TLinearMolecule);
    Offset := 0;
    SetStateType(StateU as TLinear);
    SetStateType(StateL as TLinear);
    ReadFitFlags(Offset);
    ReadHeader;
    Species.Jmax := Round(NR)*2;
    ReadTail;
    ReadState(ManifoldL);
    if not IsMicrowave then
      ReadState(ManifoldU);
    (MakeTransitonMoment(TSphericalTransitionMoment) as TSphericalTransitionMoment).Component := -9999;
    (Molecule as TLinearMolecule).RSquaredH := True;
  end;

  procedure ReadLinearTriplet;
    procedure ReadState(Manifold: TManifold);
    begin
      with Manifold.State[0] as TLinear do begin
        B := NP;
        D := NP;
        LambdaSS := NP;
        LambdaD := NP;
        gamma := NP;
      end;
    end;
  begin
    { Import of Linear Doublet }
    MakeBaseMixture(TLinearMolecule);
    NW; NW;
    StateU.S := 2;
    StateL.S := 2;
    ReadFitFlags(10);
    ReadHeader;
    Species.Jmax := Round(NR)*2;
    ReadTail;
    ReadState(ManifoldL);
    if not IsMicrowave then
      ReadState(ManifoldU);
    (MakeTransitonMoment(TSphericalTransitionMoment) as TSphericalTransitionMoment).Component := -9999;
    (Molecule as TLinearMolecule).RSquaredH := True;
  end;

  procedure ReadSinglet;
  var
    Offset: Integer;
    procedure SetStateType(State: TLinear);
    var
      Lamg: Integer;
    begin
      Lamg := NW;
      Inc(Offset, 2);
      with State do
        case Lamg of
          0: { Sigmag + };
          1: Lambda := -1;
          2: gerade := False;
          3: begin gerade := False; Lambda := -1 end;
        else
          Lambda := (Lamg div 2) * 2 - 2;
          gerade := not Odd(Lamg);
         Inc(Offset);
        end;
    end;
    procedure ReadState(Manifold: TManifold);
    begin
      with Manifold.State[0] as TLinear do begin
        B := NP;
        D := NP;
        if Lambda > 0 then
          q := NP;
      end;
    end;
  begin
    { Import of Linear Doublet }
    MakeBaseMixture(TLinearMolecule);
    Offset := 0;
    SetStateType(StateU as TLinear);
    SetStateType(StateL as TLinear);
    ReadFitFlags(Offset);
    ReadHeader;
    Species.Jmax := Round(NR)*2;
    ReadTail;
    ReadState(ManifoldL);
    if not IsMicrowave then
      ReadState(ManifoldU);
    (MakeTransitonMoment(TSphericalTransitionMoment) as TSphericalTransitionMoment).Component := -9999;
    with Molecule as TLinearMolecule do begin
      RSquaredH := True;
      Symmetric := True;
      AsymWt := 0;
    end;
  end;

  procedure ReadMultipleSinglet;
  var
    VibMax: Integer;
    i: Integer;
    FC: array [0..19] of Double;
    ManifoldsU, ManifoldsL: array [0..19] of TManifold;
    TransitionMoments: array [0..19] of TTransitionMoments;
    Problem: Exception;
    procedure ReadState(Manifolds: array of TManifold; Base: Double);
    var
      v, Lamgu: Integer;
    begin
      for v := 0 to 19 do begin
        if v >= VibMax then begin
          NR; NR; NR; NR; NR; { Skip state }
        end else begin
          with Manifolds[v].State[0] as TLinear do begin
            Origin := NP+Base;
            Lamgu := Round(NR);
            case Lamgu of
              800: { Sigmag+ };
              801: Lambda := -1;
              802: gerade := False;
              803: begin gerade := False; Lambda := -1; end;
            else
              Lambda := (Lamgu div 10) * 2;
              gerade := not Odd(Lamgu);
            end;
            B := NP;
            D := NP;
            q := NP;
            Name := IntToStr(v)+':'+SettingToStr(IndexOfSetting('Lambda'));
            if gerade then
              Name := Name + 'g'
            else
              Name := Name + 'u';
            Name := Name + ':'+IntToStr(Round(Origin));
          end;
        end;
      end;
    end;
  begin
    { Import of Linear Doublet }
    MakeBaseMixture(TLinearMolecule, False);
    NW; NW;
    ReadFitFlags(200);
    ReadHeader;
    Species.Jmax := Round(NR)*2;
    VibMax := Min(Round(NR),20);
    for i := 0 to High(FC) do
      FC[i] := NR;
    NR; { Unity }
    ReadTail;
    for i := 0 to VibMax-1 do begin
      MakeManifoldPair(ManifoldsL[i], ManifoldsU[i], TransitionMoments[i]);
      ManifoldsL[i].Name := 'X'+IntToStr(i);
      ManifoldsU[i].Name := 'A'+IntToStr(i);
    end;
    ReadState(ManifoldsL, 0);
    if not IsMicrowave then
      ReadState(ManifoldsU, Origin);
    with Molecule as TLinearMolecule do begin
      RSquaredH := True;
      Symmetric := True;
      AsymWt := 0;
    end;
    for i := 0 to VibMax - 1 do
      with TSphericalTransitionMoment.Create(TransitionMoments[i]) do begin
        Problem := SetBraKetNameE(ManifoldsU[i].State[0].Name, ManifoldsL[i].State[0].Name);
        if Assigned(Problem) then
          WriteLn(Log, Problem.Message);
        Problem.Free;
        Strength := Sqrt(FC[i]);
      end;
  end;

var
  l: Byte;
  ModelName: string;
begin
  IsGHz := False;
  IsMicrowave := False;
  S.Position := 0;
  Next(l,1);
  SetLength(ModelName, l);
  Next(ModelName[1], l);
  if ModelName = 'Asym.Rotor(&int.rotation)' then
    ReadAsym(False)
  else if ModelName = 'Asym.Rotor(&spin-rot.) case B Watson' then
    ReadAsym(True)
  else if ModelName = 'Linear Doublet' then
    ReadLinearDoublet
  else if ModelName = 'Diatomic Triplet Sigma (heteronuclear)' then
    ReadLinearTriplet
  else if ModelName = 'Linear singlets bos NoLambda' then
    ReadSinglet
  else if ModelName = 'Linear singlets bos NoLambda multistate' then
    ReadMultipleSinglet
  else
    raise Exception.Create('Import of  '+ModelName+' not implemented');
  Result.Simulations[0].PlotUnits := Result.Units;
end;

end.
