unit ImportVIBCA;
{$I utils.inc}{Copyright (C) 2013 C M Western; see commandline.pas for license}
{ Import of data files from
  VIBCA - VIBRATIONAL CALCULATIONS PROGRAM
  Zbigniew KISIEL based on program from P W Fowler
  http://info.ifpan.edu.pl/~kisiel/prospe.htm }

interface

uses
  Mixture, TextStrm, Vib, InternalCoords;

function ImportVIBCAFile(S: TTextStream): TMixture;
procedure WriteVIBCAFile(var fi: Text; IC: TInternalCoords);

implementation

uses Classes, SysUtils, zMatrix;

function ImportVIBCAFile(S: TTextStream): TMixture;
var
  Species: TSpecies;
  Molecule: TVibratingMolecule;
  Manifold: TVibrationalManifold;
  State: TElectronicState;
  zMatrix: TzMatrix;
  IC: TInternalCoords;
  Sy: TSymmetryCoords;
  NAT, NFK, NLBE, IFVCT, IPRFUL: Integer;
  FSCALE, t: Double;
  AT: string;
  JJ, KK, LL: Integer;
  R, THETA, PHI, NAX, ALFA: string;
  ICC1, ICC2, INDEX: Integer;
  LIST: array [1..2] of array [1..4] of Integer;
  i, j, Mode1, Mode2: Integer;
  Line: string;
  FK: array of Double;
  procedure ReadLine;
  begin
    Line := S.RestOfLine;
    S.NextLine;
  end;
  function field(n: Integer):string;
  begin
    Result := Trim(Copy(Line, 1, n));
    Delete(Line,1,n);
  end;
  function ifield(n: Integer):Integer;
  begin
    Result := StrToInt(field(n));
  end;
  function izfield(n: Integer):Integer;
  var
    s: string;
  begin
    s := field(n);
    if s <> '' then
      Result := StrToInt(s)
    else
      Result := 0;
  end;
  function dfield(n: Integer):Double;
  begin
    Result := StrToFloat(field(n));
  end;
  function FindMode(ICC: Integer; LIST: array of Integer):Integer;
  var
    L: array [0..4] of string;
    i, j: Integer;
  begin
    for i := 1 to 4 do
      if LIST[i-1] = 0 then
        L[i] := ''
      else
        L[i] := zMatrix.Atoms[LIST[i-1]-1].Name;
    case ICC of
      1: L[0] := 'stretch';
      2: L[0] := 'bend';
      3: begin
           L[0] := 'bondplane';
           L[1] := zMatrix.Atoms[LIST[3]-1].Name; { Apex atom }
           L[2] := zMatrix.Atoms[LIST[0]-1].Name; { end atom }
           L[3] := zMatrix.Atoms[LIST[1]-1].Name; { anchor 1 }
           L[4] := zMatrix.Atoms[LIST[2]-1].Name; { anchor 2 }
         end;
      4: L[0] := 'bend';
      5: L[0] := 'torsion';
      6: L[0] := 'torsion';
    else
      L[0] := IntToStr(ICC);
    end;
    for i := 0 to High(IC.A) do
      for j := 0 to High(IC.A[i]) do
        if IC.A[i][j] <> L[j] then
          Break
        else if j = 4 then begin
          Result := i;
          Exit;
        end;
    Result := Length(IC.A);
    SetLength(IC.A, Result+1);
    SetLength(IC.A[Result],5);
    for i := 0 to 4 do
      IC.A[Result][i] := L[i];
  end;
begin
  Result := TMixture.Create(nil);
  try
    Result.Name := 'Mixture';
    Species := TSpecies.Create(Result);
    Species.Name := 'S';
    Molecule := TVibratingMolecule.Create(Species);
    Molecule.Name := 'M';
    Manifold := TVibrationalManifold.Create(Molecule);
    Manifold.Name := 'Ground';
    State := TElectronicState.Create(Manifold);
    State.Name := 'X';
    State.Comment := S.RestOfLine;
    S.NextIntegerVar(NAT);
    S.NextIntegerVar(NFK);
    S.NextIntegerVar(NLBE);
    S.NextIntegerVar(IFVCT);
    S.NextIntegerVar(IPRFUL);
    S.NextDoubleVar(FSCALE);
    S.NextLine;
    Molecule.nNuclei := Abs(NAT);
    zMatrix := TzMatrix.Create(State);
    zMatrix.Name := 'zMatrix';
    SetLength(zMatrix.A,ABS(NAT)+2*NLBE,7);
    IC := TInternalCoords.Create(State);
    IC.Name := 'Internal Coords';
    IC.Active := False; { For now }
    for i := 0 to High(zMatrix.A) do begin
      ReadLine;
      field(1);
      AT := field(4);
      JJ := ifield(5);
      KK := ifield(5);
      R  := field(10);
      THETA := field(10);
      PHI := field(10);
      LL := izfield(5);
      NAX := field(5);
      ALFA := field(10);
      if AT = '*' then
        AT := 'X'; { Dummy }
      zMatrix.A[i][0] := AT;
      zMatrix.A[i][2] := R;
      zMatrix.A[i][4] := THETA;
      zMatrix.A[i][6] := PHI;
      if JJ <> 0 then
        zMatrix.A[i][1] := IntToStr(JJ);
      { TODO: non-Cartesian co-ordinates not handled properly }
      if KK <> 0 then
        zMatrix.A[i][3] := IntToStr(JJ);
      if LL <> 0 then
        zMatrix.A[i][5] := Trim(IntToStr(LL)+' '+NAX+' '+ALFA);
    end;
    zMatrix.NumberAtoms;
    State.FixupForceFieldCalc(False);
    if NAT < 0 then begin
      ReadLine;
      ReadLine;
      ReadLine;
    end;
    SetLength(FK, NFK);
    NFK := 0;
    while NFK < Length(FK) do begin
      for i := 1 to 10 do begin
        S.NextDoubleVar(FK[NFK]);
        FK[NFK] := FK[NFK]*FSCALE;
        Inc(NFK);
        if NFK >= Length(FK) then
          Break;
      end;
      if NFK >= Length(FK) then
        Break;
      S.NextLine;
    end;
    repeat
      ReadLine;
      ICC1 := ifield(5);
      if ICC1 = 0 then
        Break;
      for i := 1 to 4 do
        LIST[1][i] := izfield(5);
      Mode1 := FindMode(ICC1,LIST[1]);
      INDEX := ifield(5);
      ICC2 := izfield(5);
      if ICC2 <> 0 then begin
        for i := 1 to 4 do
          LIST[2][i] := izfield(5);
        Mode2 := FindMode(ICC2, LIST[2]);
      end else
        Mode2 := Mode1;
      i := Length(IC.AF);
      SetLength(IC.AF,i+1);
      SetLength(IC.AF[i],3);
      IC.AF[i][0] := FloatToStr(FK[INDEX-1]);
      IC.AF[i][1] := IntToStr(Mode1+1);
      IC.AF[i][2] := IntToStr(Mode2+1);
    until False;
    IC.Active := Length(IC.AF) > 0;
    ReadLine;
    if Trim(Line) = 'U MATRIX' then begin
      Sy := TSymmetryCoords.Create(State);
      Sy.Name := 'Symmetry Coords';
      IC.MakeBMatrix(State,False);
      SetLength(Sy.A,3*Molecule.nNuclei-5,IC.nInternal);
      for i := 0 to High(Sy.A) do
        if S.GetFreeItem and S.IsRealToken(t) then begin
          Sy.A[i][0] := S.Token;
          for j := 1 to IC.nInternal-1 do
            S.NextStringVar(Sy.A[i][j]);
          end
        else if i = High(Sy.A) then
          SetLength(Sy.A,i);
      Sy.Active := True;
    end;
  except
    Result.Free;
    raise;
  end;
end;

procedure WriteVIBCAFile(var fi: Text; IC: TInternalCoords);
var
  NFK: Integer;
  ICC: array of Integer;
  State: TElectronicState;
  procedure DoInternal2(n1, n2: Integer);
    procedure DoInternal(n:Integer);
    var
      i: Integer;
    begin
      Write(fi,ICC[n]:5);
      if ICC[n] = 3 then
        Write(fi, IC.AtomIndex[n][2]+1:5, { end atom }
                  IC.AtomIndex[n][3]+1:5, { anchor 1 }
                  IC.AtomIndex[n][4]+1:5, { anchor 2 }
                  IC.AtomIndex[n][1]+1:5) { Apex atom }
      else
        for i := 1 to 4 do
          if IC.AtomIndex[n][i] >= 0 then
            Write(fi, IC.AtomIndex[n][i]+1:5)
          else
            Write(fi, ' ':5)
    end;
  begin
    { ICC1,LIST(1,1),...,LIST(1,4),INDEX,ICC2,LIST(1,2),...,LIST(4,2) }
    Inc(NFK);
    DoInternal(n1);
    Write(fi, NFK:5);
    if n1 <> n2 then
      DoInternal(n2)
    else
      Write(fi, 0:5);
    WriteLn(fi);
  end;
var
  NAT, NLBE, IFVCT, IPRFUL: Integer;
  FSCALE: Double;
  i, j, pass: Integer;
  s: string;
begin
  { Export to .VIB file for VIBCA
    Program by Zbigniew Kisiel/Patrick Fowler }
  State := IC.ElectronicState;
  IC.Apply(False);
  WriteLn(fi, State.UniqueName+' '+State.Comment); { HDG  - TITLE }
  NAT := 0;
  with IC.zMatrix do
    for i := 0 to High(Atoms) do
    if Atoms[i].Info.Mass > 0 then
      Inc(NAT);
  NFK := 0; { NUMBER OF FORCE CONSTANT VALUES TO BE SPECIFIED }
  if Assigned(IC.F) then
    for i := 1 to IC.F.RowCount do
      for j := 1 to i do
        if IC.F[i,j] <> 0 then
          Inc(NFK);
  NLBE := IC.nLinearBends; { NUMBER OF LINEAR BENDS IN THE MOLECULE }
  IFVCT := 1;
  IPRFUL := 1;
  FSCALE := 1;
  WriteLn(fi, NAT, ' ', NFK, ' ', NLBE, ' ', IFVCT, ' ', IPRFUL, ' ',  FSCALE);
  { Atom positions - Cartesian only output }
  with IC.zMatrix do
    for i := 0 to High(Atoms) do begin
      if Atoms[i].Info.Mass > 0 then
        s := IntToStr(Round(Atoms[i].Info.Mass))+UpperCase(Atoms[i].Info.Symbol)
      else
        s:= '*';
      WriteLn(fi, ' ', s:4, 0:5, 0:5, Atoms[i].xyz[1]:10:7, Atoms[i].xyz[2]:10:7, Atoms[i].xyz[3]:10:7) { AT JJ KK R THETA PHI LL NAX ALFA, 1X,A4,2I5,3F10.4,2I5,F10.4 }
    end;
  { Compute ICC values for each internal co-ordinate }
  j := 0;
  SetLength(ICC, IC.nInternal);
  with IC do begin
    for i := 0 to High(A) do
     if A[i][0] <> '' then begin
       if LowerCase(A[i][0]) = 'stretch' then
         ICC[j] := 1
       else if LowerCase(A[i][0]) = 'bend' then
         if AtomIndex[j][4] >= 0 then
           ICC[j] := 3 { Linear Bend }
         else
           ICC[j] := 2
       else if LowerCase(A[i][0]) = 'plane' then
         raise NotYet { TODO }
       else if LowerCase(A[i][0]) = 'bondplane' then
         ICC[j] := 3
       else if LowerCase(A[i][0]) = 'torsion' then
         if AtomIndex[j][4] >= 0 then
           ICC[j] := 6 { 4 atom torsion }
         else
           ICC[j] := 5 { Generalized torsion }
       else
         raise NotYet; { Error }
       Inc(j);
     end;
    { Force constants }
    if Assigned(F) then
      for pass := 1 to 2 do begin
        NFK := 0;
        for i := 1 to F.RowCount do
          if F[i,i] <> 0 then
            if pass = 1 then begin
              Write(fi, ' ',FloatToStr(F[i,i]));
              Inc(NFK);
              if NFK mod 10 = 0 then
                WriteLn(fi);
            end else
              DoInternal2(i-1,i-1);
        for i := 1 to F.RowCount do
          for j := 1 to i-1 do
            if F[i,j] <> 0 then
              if pass = 1 then begin
                Write(fi, ' ',FloatToStr(F[i,j]));
                Inc(NFK);
                if NFK mod 10 = 0 then
                  WriteLn(fi);
              end else
                DoInternal2(i-1,j-1);
        if (pass = 1) and (NFK mod 10 <> 0) then
          WriteLn(fi);
      end;
  end;
  WriteLn(fi, 0:5);
  WriteLn(fi, 'END');
end;

end.
