unit Importer;
{ Import multiple objects from a csv or tab delimited file }
{$I utils.inc}{Copyright (C) 2016 C M Western; see commandline.pas for license}
interface

uses
  Classes, SysUtils, MNode, Mixture, ItemStrm;

function ImportMNodes(Template: TMNode; FileName: string; Verbose: Boolean): Boolean; overload; { Intended for command line use }
function ImportMNodes(Mixture: TMixture; Template: string; FileName: string; Verbose: Boolean): Boolean; overload; { Intended for command line use }
function ImportMNodes(var F: Text; Template: TMNode; S: TItemStream; FileName: string; Update, Make, Verbose: Boolean): Boolean; overload;

implementation

uses {$IFDEF MSWINDOWS} Windows, {$ENDIF}
     UCompat, TextStrm, Huge, pgDOM, Worker;

function ImportMNodes(var F: Text; Template: TMNode; S: TItemStream; FileName: string; Update, Make, Verbose: Boolean): Boolean;
var
  Headers: array of String;
  BaseName, AllNames, NewName: string;
  BaseNameKet, NewNameBra, NewNameKet: string;
  i, nErrors, nNew, nOld, nUnread, nValues, StarPos, StarPosKet, InitialSkip: Integer;
  Indices, Scale: array of Integer;
  Parent, OldNode, NewNode, Dummy: TMNode;
  XMLDocument: TXMLDocument;
  TemplateXML: TDOMElement;
  TemplateClass: TMNodeClass;
  Pair: TMNodePair;
  p, NewVal: Double;
  IsPair: Boolean;
  Bra, Ket: TMNode;
  function MakeDummy: TMNode;
  begin
    if Dummy = nil then
      Dummy := TemplateClass.Create(nil);
    Result := Dummy;
  end;
  procedure Error(const Msg: string; OK: Boolean);
  begin
    WriteLn(F, 'Error on line ', S.Line, ' of ', FileName, ': ', Msg);
    Inc(nErrors);
    if OK then
      Exit;
    Update := False;
    Make := False;
  end;
  function ExtractBase(const s: string):string;
  begin
    Result := s;
    while (Result <> '') and (Result[Length(Result)] in ['0'..'9']) do
      Delete(Result,Length(Result),1);
  end;
  procedure FindBaseName(MNode: TMNode; out BaseName: string; out StarPos: Integer);
  begin
    StarPos := -1;
    if (S.Token <> 'v') and (S.Token <> 'v''') and (S.Token <> 'v"') then begin
      BaseName := S.Token;
      StarPos := Pos('*', BaseName);
      if StarPos <= 0 then
        BaseName := ExtractBase(BaseName);
    end else
      BaseName := ExtractBase(MNode.Name);
  end;
  function FindNewName(out NewName: string; const BaseName: string; StarPos: Integer):Boolean;
  var
    NewV: Integer;
  begin
    Result := S.GetFreeItem and not S.IsEmptyToken;
    if Result then begin
      if S.IsIntegerToken(NewV) then begin
        NewName := IntToStr(NewV);
        if StarPos <= 0 then
          NewName := BaseName + NewName
        else
          NewName := Copy(BaseName,1,StarPos-1) + NewName + Copy(BaseName,StarPos+1,Length(BaseName));
      end else
        NewName := S.Token;
    end;
  end;
begin
  Result := False;
  Dummy := nil;
  Parent := Template.Owner as TMNode;
  try
    TemplateClass := TMNodeClass(Template.ClassType);
    S.FlowToNextLine := False;
    S.SkipBlankLines;
    IsPair := Template is TMNodePair;
    InitialSkip := 0;
    while S.GetFreeItem and (S.Token = '') do
      Inc(InitialSkip);
    if IsPair then begin
      Pair := TMNodePair(Template);
      FindBaseName(Pair.Bra, BaseName, StarPos);
      if not S.GetFreeItem then
        Error('Second vibrational quantum number not found',False);
      FindBaseName(Pair.Ket, BaseNameKet, StarPosKet);
    end else
      FindBaseName(Template, BaseName, StarPos);
    nValues := 0;
    nErrors := 0;
    while S.GetFreeItem do begin
      SetLength(Headers, nValues+1);
      Headers[nValues] := S.Token;
      Inc(nValues);
    end;
    if nValues < 1 then
      Error('No value names found',False);
    SetLength(Indices, nValues);
    SetLength(Scale, nValues);
    for i := 0 to nValues-1 do
      if Headers[i] = '' then
        Indices[i] := 0
      else begin
        Indices[i] := Template.IndexOfParameter(Headers[i]);
        Scale[i] := 1;
        if Indices[i] <= 0 then begin
          Indices[i] := Template.IndexOfSetting(Headers[i]);
          if Indices[i] <= 0 then begin
            if Headers[i][1] = '-' then begin
              Indices[i] := Template.IndexOfParameter(Copy(Headers[i],2,Length(Headers[i])));
              Scale[i] := -1;
            end;
            if (Indices[i] <= 0) and (Pos('^2', Headers[i]) = Length(Headers[i]) -1) then begin
              Indices[i] := Template.IndexOfParameter(Copy(Headers[i],1,Length(Headers[i])-2));
              Scale[i] := 0;
            end;
            if Indices[i] <= 0 then begin
              Error(Headers[i]+' not found in '+Template.UniqueName,True);
              Indices[i] := Low(Integer);
            end;
          end else
            Indices[i] := -Indices[i]; { Flag setting, rather than parameter }
        end;
      end;
    S.NextLine;
    nNew := 0;
    nOld := 0;
    AllNames := #9; { Check no duplicate imports }
    while not S.AtEndOfFile do begin
      for i := 1 to InitialSkip do
        S.GetFreeItem;
      if FindNewName(NewNameBra, BaseName, StarPos)
         and (not IsPair or FindNewName(NewNameKet, BaseNameKet, StarPosKet)) then begin
        if IsPair then begin
          Bra := Pair.BraOwner.ComponentWithName(NewNameBra, Pair.BraKetClass, nil);
          if Bra = nil then
            Error('Can''t find Bra('+NewNameBra+') in '+Pair.BraOwner.UniqueName,False);
          Ket := Pair.KetOwner.ComponentWithName(NewNameKet, Pair.BraKetClass, nil);
          if Ket = nil then
            Error('Can''t find Ket('+NewNameKet+') in '+Pair.KetOwner.UniqueName,False);
          NewName := '<' + NewNameBra + '|' + Pair.PairName + '|' + NewNameKet + '>';
        end else
          NewName := NewNameBra;
        if Pos(#9+LowerCase(NewName)+#9,AllNames) > 0 then
          Error(NewName + ' imported twice',True);
        AllNames := AllNames + LowerCase(NewName) + #9;
        OldNode := Parent.ComponentWithName(NewName, TMNode, nil);
        if OldNode <> nil then begin
          Inc(nOld);
          if Verbose then begin
            if not Update then
              Write(F, '(Skip) ');
            WriteLn(F, 'Overwriting parameters in '+OldNode.UniqueName);
          end;
          if Update then
            NewNode := OldNode
          else begin
            NewNode := MakeDummy;
            Dummy.Name := OldNode.Name;
          end;
        end else begin
          Inc(nNew);
          OldNode := Template;
          if not IsPair then
            if ExtractBase(NewName) <> ExtractBase(Template.Name) then
              { look for first component in Parent with same base name }
              for i := 0 to Parent.ComponentCount-1 do
                if ExtractBase((Parent.Components[i] as TMNode).Name) = ExtractBase(NewName) then begin
                  OldNode := Parent.Components[i] as TMNode;
                  Break;
                end;
          if Make then
            NewNode := TemplateClass.Create(Parent)
          else
            NewNode := MakeDummy;
          if Verbose then begin
            if not Make then
              Write(F, '(Skip) ');
            WriteLn(F, 'Creating '+NewName+' and copying values from '+OldNode.UniqueName);
          end;
          XMLDocument := TXMLDocument.Create;
          try { Note: must create a new document each time, as read process removes items read }
            TemplateXML := Template.WriteProperties(XMLDocument);
            nUnread := NewNode.ReadProperties(TemplateXML);
          finally
            XMLDocument.Free;
          end;
          if nUnread > 0 then
            Error(IntToStr(nUnread)+' unread properties for '+NewNode.UniqueName,False);
          if IsPair then
            (NewNode as TMNodePair).SetBraKetName(NewNameBra, NewNameKet)
          else
            NewNode.Name := NewName;
        end;
        i := 0;
        while S.GetFreeItem do begin
          if i >= nValues then begin
            Error('Excess value for '+NewNode.UniqueName+':'+S.Token,False);
          end else
            if Indices[i] > 0 then
              if S.IsRealToken(P) then begin
                if Scale[i] = 0 then
                  NewVal := Sqrt(P)
                else
                  NewVal := P*Scale[i];
                NewNode.Parameters[Indices[i]] := NewVal;
                if Verbose then
                  WriteLn(F, ' ':10, NewNode.ParameterNames[Indices[i]]:10, ' := ', FloatToStr(NewVal))
              end else
                Error('Bad value for '+Headers[i]+' in '+NewNode.UniqueName+':'+S.Token,True)
            else if Indices[i] = Low(Integer) then begin
              if Verbose then
                WriteLn(F, 'Skipping ',Headers[i],' = ',S.Token)
            end else if Indices[i] < 0 then
              try
                NewNode.SetSettingFromStr(-Indices[i], S.Token);
              except
                on E: Exception do
                  Error('Bad value for '+Headers[i]+' in '+NewNode.UniqueName+' ('+E.Message+'):'+S.Token,True);
              end;
          Inc(i);
        end;
      end;
      S.NextLine;
    end;
    WriteLn(F, nOld, ' objects updated');
    WriteLn(F, nNew, ' objects created');
    WriteLn(F, nErrors, ' errors');
    Result := nErrors = 0;
  finally
    Dummy.Free;
  end;
end;

function ImportMNodes(Template: TMNode; FileName: string; Verbose: Boolean): Boolean;
var
  S: TItemStream;
begin
  Result := False;
  S := SetupImportStream(FileName);
  try
    Result := ImportMNodes(Log, Template, S, FileName, True, True, Verbose);
  finally
    S.Free;
  end;
end;

function ImportMNodes(Mixture: TMixture; Template: string; FileName: string; Verbose: Boolean): Boolean; overload; { Intended for command line use }
var
  ValueInfo: TValueInfo;
  MNode: TMNode;
  Msg: string;
begin
  MNode := nil;
  if not Mixture.nComponentsWithName(Template, MNode) = 1 then begin
    ValueInfo := Mixture.SubPropInfo(Template+'.Origin',Msg,False);
    if ValueInfo.MNode = nil then begin
      ValueInfo := Mixture.SubPropInfo(Template+'.Strength',Msg,False);
      if ValueInfo.MNode = nil then
        ValueInfo := Mixture.SubPropInfo(Template+'.Value',Msg,False);
    end;
    MNode := ValueInfo.MNode;
  end;
  if MNode = nil then
    raise EReportException.Create('Can''t find '+Template);
  Result := ImportMNodes(MNode, FileName, Verbose);
end;

end.

