unit Interlock;
{$I utils.inc}{ Copyright (C) 2005 C M Western; see UCompat.pas for license}
interface

{$IFDEF DELPHI}
  {$IF CompilerVersion >= 19.0}
  {$DEFINE NEWDELPHIEX}
  {$IFEND}
{$ENDIF}

uses {$IFDEF MSWINDOWS} Windows, {$ENDIF}
     {$IFDEF NEWDELPHIEX} StackTrace, {$ENDIF}
     Classes, SysUtils, Worker, SyncObjs;

type

  TInterlockedWorker = class;

  { Component that records state while being assembled, so that
    threads can co-operate on its use }

  { TInterlockedComponent }

  TInterlockedComponent = class(TComponent)
  private
    FErrorMessage: string;
    FCurrentWorker: TInterlockedWorker;
    FExceptionClass: ExceptClass;
    {$IFDEF NEWDELPHIEX}
    CallStack: TCallStack;
    {$ENDIF}
    function GetReady: Boolean;
  protected
    FInValid: Boolean;
    FDoneEvent: TEvent;
  public
    procedure BeforeDestruction; override;
    function Make(Wait:Boolean; RequiredAttributes: Pointer = nil):Boolean;
    procedure DoMake(RequiredAttributes: Pointer); virtual; abstract;
    procedure RunDoMake(RequiredAttributes: Pointer);
    procedure WaitFor(Timeout: Cardinal = INFINITE);
    procedure RaiseInvalid;
    function HasCorrectAttributes(Attributes: Pointer):Boolean; virtual; abstract;
    property InValid: Boolean read FInValid;
    property Ready: Boolean read GetReady;
    property ErrorMessage: string read FErrorMessage;
    property ExceptionClass: ExceptClass read FExceptionClass;
    property CurrentWorker: TInterlockedWorker read FCurrentWorker;
  end;

  TInterlockedWorker = class(TWorkerThread)
  private
    FRequiredAttributes: Pointer;
    FInterlockedComponent: TInterlockedComponent;
  public
    procedure DoWork; override;
  end;

implementation

uses UCompat, Contnrs, Profile;

threadvar
  ThreadDoneEvent: TEvent;

var
  DoneEvents: TObjectList; { Used to track allocations so that they can be cleaned up }

type
  EPropagatedException = class(Exception)
  protected
    {$IFDEF NEWDELPHIEX}
      procedure RaisingException(P: PExceptionRecord); override;
    {$ENDIF}
  end;

  EReportPropagatedException = class(EPropagatedException);

{ EPropagatedException }

{$IFDEF NEWDELPHIEX}
procedure EPropagatedException.RaisingException(P: PExceptionRecord);
begin
  SetInnerException;
  { Skip setting up stack info }
end;
{$ENDIF}

{ TInterlockedComponent }

function TInterlockedComponent.GetReady: Boolean;
begin
  Result := not Assigned(FDoneEvent);
end;

procedure TInterlockedComponent.BeforeDestruction;
begin
  inherited;
  WaitFor;
end;

var
  { One per object might be fastest, but might end up with a lot of unused objects }
  InterlockCriticalSection: {$IFDEF FPC}System.{$ENDIF}TRTLCriticalSection;

function TInterlockedComponent.Make(Wait: Boolean; RequiredAttributes: Pointer):Boolean;
{ Ensure component is being processed so it has the RequiredAttributes;
  Wait if required }
var
  StartWork: Boolean;
begin
  Result := True;
  if InValid then
    RaiseInvalid;
  {$IFDEF FPC}System.{$ENDIF}EnterCriticalSection(InterlockCriticalSection);
  try
    if Ready and HasCorrectAttributes(RequiredAttributes) then
      Exit; { Easy - done }
    StartWork := not Assigned(FDoneEvent);
    if StartWork then
      if nWorkers > 0 then begin
        if not Assigned(ThreadDoneEvent) then begin
          if not Assigned(DoneEvents) then
            DoneEvents := TObjectList.Create;
          ThreadDoneEvent := TEvent.Create(nil, True, False, '');
          DoneEvents.Add(ThreadDoneEvent);
        end;
        FDoneEvent := ThreadDoneEvent;
        FDoneEvent.ResetEvent;
      end else
        StartWork := Wait;
  finally
    {$IFDEF FPC}System.{$ENDIF}LeaveCriticalSection(InterlockCriticalSection);
  end;
  if StartWork then begin
    if Wait or (GetCurrentThreadID <> MainThreadID) then begin
      RunDoMake(RequiredAttributes); { This thread can do the work }
      if nWorkers > 0 then
        ThreadDoneEvent.SetEvent;
      if InValid then
        RaiseInvalid;
    end else begin
      Result := False; { Not completed in this call }
      try
        FCurrentWorker := FindWorkerWait as TInterlockedWorker;
      except
        if nWorkers > 0 then
          ThreadDoneEvent.SetEvent;
        FDoneEvent := nil;
        raise;
      end;
      if not Assigned(FCurrentWorker) then begin
        if nWorkers > 0 then
          ThreadDoneEvent.SetEvent;
        FDoneEvent := nil;
        AbortCalculation;
      end;
      FCurrentWorker.FRequiredAttributes := RequiredAttributes;
      FCurrentWorker.FInterlockedComponent := Self;
      {$IFDEF FPC}System.{$ENDIF}EnterCriticalSection(InterlockCriticalSection);
      { Transfer done event to new worker }
      FDoneEvent := FCurrentWorker.Done;
      ThreadDoneEvent.SetEvent;
      {$IFDEF FPC}System.{$ENDIF}LeaveCriticalsection(InterlockCriticalSection);
      FCurrentWorker.StartWork;
    end;
  end else if Wait then begin
    { another thread doing it - wait for it }
    WaitFor;
    {$IFDEF FPC}System.{$ENDIF}EnterCriticalSection(InterlockCriticalSection);
    try
      if Ready and HasCorrectAttributes(RequiredAttributes) then
        Exit;
    finally
      {$IFDEF FPC}System.{$ENDIF}LeaveCriticalsection(InterlockCriticalSection);
    end;
    Make(Wait, RequiredAttributes); { Need to restart calculation }
  end else
    Result := False; { Not ready yet, and not waiting }
end;

procedure TInterlockedComponent.RaiseInvalid;
var
  E: EPropagatedException;
  {$IFDEF NEWDELPHIEX}
  C: PCallStack;
  {$ENDIF}
begin
  if IsExpectedException(ExceptionClass) then
    E := EReportPropagatedException.Create('')
  else
    E := EPropagatedException.Create('');
  E.Message := 'Invalid Object: ' + ErrorMessage;
  {$IFDEF NEWDELPHIEX}
    New(C);
    C^ := CallStack;
    E.SetStackInfo(C);
  {$ENDIF}
  raise E;
end;

procedure TInterlockedComponent.RunDoMake(RequiredAttributes: Pointer);
begin
  try
    DoMake(RequiredAttributes);
  except
    on E:Exception do begin
      {$IFDEF CONSOLE}
        ConsiderException(Log, E);
      {$ENDIF}
      FErrorMessage := E.Message;
      FExceptionClass := ExceptClass(E.ClassType);
      {$IFDEF NEWDELPHIEX}
        if E.StackInfo <> nil then
          CallStack := PCallStack(E.StackInfo)^
        else
          FillChar(CallStack, SizeOf(CallStack), 0);
      {$ENDIF}
      FInValid := True;
    end;
  end;
  {$IFDEF FPC}System.{$ENDIF}EnterCriticalSection(InterlockCriticalSection);
  try
    FDoneEvent := nil;
  finally
    {$IFDEF FPC}System.{$ENDIF}LeaveCriticalSection(InterlockCriticalSection);
  end;
end;

procedure TInterlockedComponent.WaitFor(Timeout: Cardinal);
begin
  while Assigned(FDoneEvent) do
    { The calculation could finish before the wait starts }
    { Even if event signaled, may not be complete as DoneEvent might have been switched }
    if DoWait(FDoneEvent, 1000) = wrAbandoned then
      if Assigned(FDoneEvent) then      
        AbortCalculation;
end;

{ TInterlockedWorker }

procedure TInterlockedWorker.DoWork;
begin
  if Assigned(FInterlockedComponent) then
    FInterlockedComponent.RunDoMake(FRequiredAttributes);
  FInterlockedComponent := nil;
end;

initialization
  AddExpectedException(EReportPropagatedException);
  {$IFDEF FPC}System.{$ENDIF}InitCriticalSection(InterlockCriticalSection);
finalization
  {$IFDEF FPC}System.{$ENDIF}DoneCriticalSection(InterlockCriticalSection);
  DoneEvents.Free;
end.

