unit InternalCoords;
{Copyright (C) 2013 C M Western; see commandline.pas for license}
{$DEFINE NOAT}{$I utils.inc}

{ Internal co-ordinates for vibrational calculations }

interface

uses SysUtils, Classes, MathSup, pgDOM, MNode, Mixture, MNodeExpr, Vib, MatrixVec, zMatrix;

type

  TPotentialTerm = record
    Bra, Ket: TElectronicState;
    Value: Double;
    Modes: TIntegerArray;
  end;

  { TMNodeWithFMatrix }

  TMNodeWithFMatrix = class(TlMatrixMNode)
  public
    F: TMatrix;
    Flinear: TMatrix;
    ExtraTerms: array of TPotentialTerm;
    procedure ClearArrays; override;
    procedure MakeFMatrix(nModes: Integer; const Caption: string; Verbose: Boolean; ShowName: TShowName);
    procedure SortF(ByValue: Boolean);
    procedure LoadF(NewF: TMatrix);
    procedure CalculateF(RowLabel, SRowLabel: TShowName; B, U: TMatrix; InPrincipalAxis, Verbose: Boolean; DoApply: Boolean);
  end;

  { TInternalCoords }

  TModeType = (icStretch, icBend, icLinBend, icPlaneSym, icPlaneGen, icTorsion, icGenTorsion, icError);

  TInternalCoords = class(TMNodeWithFMatrix)
  public
    nInternal, nLinearBends: Integer;
    B: TMatrix;
    Brows: TIntegerArray; { Mapping from row in string array A to row in B }
    AtomIndex: array of T4Integers;
    ModeAtoms: array of TIntegerArray; { Index into Atoms[] array of atoms involved in each mode }
    ModeType: array of TModeType;
    function ReadProperties(Node: TDOMNode; CheckUnread: Boolean = True):Integer; override;
    procedure ShowInternalCoords(Sender: TObject);
    procedure AddHandlers; override;
    procedure MakeBMatrix(State: TElectronicState; Verbose: Boolean);
    procedure MakeExtraTerms(dint: TMatrix; Verbose: Boolean; Apply: Boolean; CreatePert: Boolean);
    procedure Prepare(Verbose: Boolean);
    procedure CalculateF(Verbose: Boolean; DoApply: Boolean);
    procedure ClearArrays; override;
    function Apply(Verbose: Boolean):string; override;
    procedure Makel(Verbose, ApplyOperators, CreateOperators: Boolean);
    function nameR(Mode: Integer):string;
    function HelpOnSetting(AIndex: Integer): string; override;
  published
    property InPrincipalAxis: Boolean index 2 read GetFlag write SetFlag;
    procedure CalculateFfromF(Verbose: Boolean; DoApply: Boolean);
  end;

  TCartesianDisplacements = class(TlMatrixMNode)
  { Allow B matrix to be specified independently;
    if TInternalCoords active, rows can be added on the end }
  private
    FInternalCoords: TInternalCoords;
    function GetInternalCoords: TInternalCoords;
  public
    B: TMatrix;
    Brows: array of Integer; { Mapping from row in string array A to row in B }
    RowActive: array of Boolean; { Active implies not derived from TInternalCoords }
    procedure ClearArrays; override;
    procedure ShowCartesianDisplacements(Sender: TObject);
    procedure AddHandlers; override;
    procedure Prepare(Verbose: Boolean);
    procedure MakeBMatrix(State: TElectronicState; Verbose: Boolean);
    procedure CheckBMatrix(Verbose: Boolean; RemoveRT: Boolean);
    procedure UpdateBMatrix(C: TMatrix; const Changed: array of Boolean; Verbose: Boolean);
    function Apply(Verbose: Boolean):string; override;
    procedure Makel(Verbose, ApplyOperators, CreateOperators: Boolean);
    function nameB(Mode: Integer):string;
    function HelpOnSetting(AIndex: Integer): string; override;
    procedure UpdateObjectChain; override;
    property InternalCoords: TInternalCoords read GetInternalCoords;
  published
    property InPrincipalAxis: Boolean index 2 read GetFlag write SetFlag;
    property Normalize: Boolean       index 3 read GetFlag write SetFlag;
    procedure CalculateFromL(Verbose: Boolean; DoApply: Boolean);
  end;

  { TSymmetryCoords }

  TForceElement = record
    V: Double;
    Coords: array [1..2] of Integer;
  end;

  TSymmetryCoords = class(TMNodeWithFMatrix)
  private
    FCartesianDisplacements: TCartesianDisplacements;
    function GetCartesianDisplacements: TCartesianDisplacements;
  public
    U, Bs, G, GA, bigL{, Lm1T, Gm1, AL}: TMatrix;
    ComputedF: array of TForceElement;
    Brows: array of Integer; { Mapping from row in string array A to row in B }
    procedure ShowSymmetryCoords(Sender: TObject);
    procedure AddHandlers; override;
    procedure Prepare(Verbose: Boolean);
    procedure MakeU(CartesianDisplacements: TCartesianDisplacements);
    procedure MakeLMatrix(State: TElectronicState; Verbose, ApplyOperators, CreateOperators, ForDisplay: Boolean); overload;
    procedure MakeLMatrix(State: TElectronicState; Verbose: Boolean; ACartesianDisplacements: TCartesianDisplacements); overload;
    function InvertU(Verbose: Boolean): TMatrix;
    procedure ClearArrays; override;
    function Apply(Verbose: Boolean):string; override;
    function nameS(Mode: Integer):string;
    function HelpOnSetting(AIndex: Integer): string; override;
    procedure UpdateObjectChain; override;
    property CartesianDisplacements: TCartesianDisplacements read GetCartesianDisplacements;
  published
    property Normalize: Boolean index 2 read GetFlag write SetFlag;
    property PreserveS: Boolean index 3 read GetFlag write SetFlag;
    procedure CalculateF(Verbose: Boolean; DoApply: Boolean);
  end;

implementation

uses Math, MatrixOps, Lapack, EigBlock, Units, ShellSort, UCompat,
     {$IFNDEF CONSOLE} InternalCoordsFrm, CartesianDisplacementsFrm, SymmetryCoordsFrm, {$ENDIF}
     Worker;

procedure EigenGF(var GF: TMatrix; out lambda: TColumnVector; Small: Double; Verbose: Boolean; BasisLabel: TShowName; Width: Integer);
var
  i, ii, j: Integer;
  t: Double;
begin
  lambda := TColumnVector.Create(GF.ColCount);
  Eigen(lambda, GF, True, Small, Verbose, BasisLabel, Width);
  { Reverse order of eigenvectors so highest frequency first
    and zero frequency modes at the end }
  for i := 1 to GF.ColCount div 2 do begin
    ii := GF.ColCount-i+1;
    t := lambda[i];
    lambda[i] := lambda[ii];
    lambda[ii] := t;
    for j := 1 to GF.RowCount do begin
      t := GF[j,i];
      GF[j,i] := GF[j,ii];
      GF[j,ii] := t;
    end;
  end;
end;

procedure ShowDimemsionless(l, B: TMatrix; lambda: TColumnVector; State: TElectronicState; ModeLabel: TShowName; Verbose, ApplyOperators, CreateOperators: Boolean);
{ Note that l is discarded on exit }
var
  dint: TMatrix;
  i, j: Integer;
  m: Double;
  IC: TInternalCoords;
begin
  for i := 1 to l.RowCount do begin
    m := 1/Sqrt((State.Nucleus((i-1) div 3) as TMolecularNucleus).Mass);
    for j := 1 to l.ColCount do
      l[i,j] := l[i,j]*m/Sqrt(gammaConversion*Sqrt(lambda[j]));
  end;
  dint := TMatrix.Create(B);
  dint.PostMultiply(l);
  if Verbose then begin
    WriteLn(Log, 'Transformation between dimensionless normal coordinates and Cartesian displacements');
    l.Write(Log, 'd', State.nameNucleusXYZ, State.nameMode, State.Precision+11, State.Precision+6);
    WriteLn(Log, 'Transformation between dimensionless normal coordinates and Internal coordinates');
    dint.Write(Log, 'dint', ModeLabel, State.nameMode, State.Precision+11, State.Precision+6);
  end;
  IC := State.ForceFieldNode(TInternalCoords) as TInternalCoords;
  if Assigned(IC) then begin
    if Assigned(IC.Flinear) and Verbose then begin
      WriteLn(Log, 'Linear terms in dimensionless normal coordinates');
      l.Free;
      l := TMatrix.Create(IC.Flinear);
      l.PostMultiply(dint);
      l.Write(log, 'Mode', nil, State.nameMode, State.Precision+6);
    end;
    if Length(IC.ExtraTerms) > 0 then begin
      if Verbose then
        WriteLn(Log, 'Extra Potential Terms');
      IC.MakeExtraTerms(dint, Verbose, ApplyOperators, CreateOperators);
    end;
  end;
  l.Free;
  dint.Free;
end;

function MakelMatrix(var lambda: TColumnVector; State: TElectronicState; B, f: TMatrix; Verbose, ApplyOperators, CreateOperators: Boolean; ModeLabel: TShowName):TMatrix;
var
  i, j: Integer;
  m: Double;
  d: TMatrix;
  IC: TInternalCoords;
begin
  { Work out l Matrix given B and f }
  Result := nil;
  FreeAndNil(lambda);
  if not (Assigned(f) and Assigned(B)) then
    Exit;
  { The l Matrix is defined such that
      rho = M^1/2 Dx = l Q
    where:
           M is the matrix with masses on the diagonal
           Dx are cartesian displacements
           Q are the normal co-ordinates
    Given this definition, the kinetic energy is diagonal:
      2T = rho'* rho' = Q'* l* l Q' = Q'*Q'
    where ' implies d/dt, * implies transpose, provided
      l* l = 1
    The f matrix gives the potential energy in terms of internal co-ordinates, R:
      2V = R* f R
    where the internal co-ordinates, R are expressed as:
      R = B Dx
    Substituting Dx = M^-1/2 rho in the potential energy expression:
      2V = rho* M^-1/2 B* f B M^-1/2 rho
    and also rho = l Q:
      2V = Q* l* M^-1/2 B* f B M^-1/2 l Q
    Consider the matrix made of the central terms:
      GF = M^-1/2 B* f B M^-1/2
    This can be shown to be symmetric:
      GF* = (M^-1/2 B f B* M^-1/2)* = GF
    as f is symmetric. If l taken as the eigenvectors of the GF matrix then
    both the potential and kinetic energy will be diagonal.
  }
  try
    Result := B.Transpose;
    Result.PostMultiply(f);
    Result.PostMultiply(B);
    if Verbose then begin
      WriteLn(Log, 'f Matrix in Cartesian co-ordinates');
      Result.Write(Log, 'f', State.nameNucleusXYZ, State.nameNucleusXYZ, State.Precision+6);
      IC := State.ForceFieldNode(TInternalCoords) as TInternalCoords;
      if Assigned(IC) and Assigned(IC.Flinear) then begin
        d := TMatrix.Create(IC.Flinear);
        d.PostMultiply(B);
        WriteLn(Log, 'Linear Terms in Cartesian co-ordinates');
        d.Write(Log, 'f', nil, State.nameNucleusXYZ, State.Precision+6);
        d.Free;
        WriteLn(Log);
      end;
    end;
    for i := 1 to Result.RowCount do begin
      m := 1/Sqrt((State.Nucleus((i-1) div 3) as TMolecularNucleus).Mass);
      for j := 1 to Result.ColCount do
        Result[i,j] := Result[i,j]*m/Sqrt((State.Nucleus((j-1) div 3) as TMolecularNucleus).Mass);
    end;
    if Verbose then
      Result.Write(Log, 'GF', State.nameNucleusXYZ, State.nameNucleusXYZ, State.Precision+6);
    EigenGF(Result, lambda, State.Molecule.SmallCoefficient, Verbose, State.nameNucleusXYZ, State.Precision+6);
    if Verbose then begin
      lambda.Write(Log, 'lambda', State.Precision+6);
      WriteLn(Log, 'l Matrix (R and T labels arbitrary)');
      Write(Log, 'cm-1');
      for i := 1 to lambda.Size do
        Write(Log, Sign(lambda[i])*sqrt(Abs(lambda[i])/LambdaConversion):State.Precision+11:State.Precision);
      WriteLn(Log);
      Result.Write(Log, 'l', State.nameNucleusXYZ, State.nameMode, State.Precision+11, State.Precision+6);
    end;
    if Verbose or CreateOperators then begin
      d := TMatrix.Create(Result);
      i := 2;
      while (i <= lambda.Size) and (lambda[i] >= State.Molecule.SmallCoefficient*lambda[1]) do
        Inc(i);
      Dec(i);
      d.ColCount := i; { Discard T and R }
      ShowDimemsionless(d, B, lambda, State, ModeLabel, Verbose, ApplyOperators, CreateOperators);
    end;
  except
    FreeAndNil(lambda);
    FreeAndNil(Result);
  end;
end;

{ TInternalCoords }

procedure TInternalCoords.AddHandlers;
begin
  inherited;
  {$IFNDEF CONSOLE}
  AddHandler('View...', 'View and edit internal co-ordintes', ShowInternalCoords);
  {$ENDIF}
end;

function TInternalCoords.Apply(Verbose: Boolean):string;
begin
  Makel(Verbose, True, False);
  if (LastError = '') and Assigned(l) then
    ElectronicState.ApplylMatrix(Self, InPrincipalAxis, Verbose);
  Result := LastError;
end;

procedure TInternalCoords.Makel(Verbose, ApplyOperators, CreateOperators: Boolean);
begin
  MakeBMatrix(ElectronicState, Verbose);
  if LastError = '' then
    MakeFMatrix(nInternal, 'f', Verbose, nameR);
  if LastError = '' then
    try
      l := MakelMatrix(lambda, ElectronicState, B, f, Verbose, ApplyOperators, CreateOperators, nameR);
    except
      on E: Exception do
        HandleError(-1,-1,E.Message);
    end;
end;

function TInternalCoords.nameR(Mode: Integer): string;
begin
  if Mode > nInternal then
    Result := 'B' + IntToStr(Mode)
  else
    Result := 'R' + IntToStr(Mode);
end;

procedure TInternalCoords.Prepare(Verbose: Boolean);
begin
  { Use for partial calculations from user interface;
    Make sure dependent nodes in order }
  if zMatrix.ApplyOK(False) then
    Apply(Verbose)
  else
    LastError := zMatrix.LastError;
end;

function TInternalCoords.ReadProperties(Node: TDOMNode; CheckUnread: Boolean): Integer;
var
  i: Integer;
begin
  Result := inherited ReadProperties(Node, CheckUnread);
  for i := 0 to High(A) do
    if (Length(A[i]) > 0) and (LowerCase(A[i][0]) = 'plane1') then
      A[i][0] := 'Plane';
end;

procedure TInternalCoords.CalculateF(Verbose, DoApply: Boolean);
begin
  inherited CalculateF(nameR, nil, B, nil, InPrincipalAxis, Verbose, DoApply);
end;

procedure TInternalCoords.CalculateFfromF(Verbose, DoApply: Boolean);
  function ActiveSymmetryCoords: TSymmetryCoords;
  var
    i: Integer;
  begin
    Result := nil;
    with ElectronicState do
      for i := 0 to ComponentCount-1 do
        if (Components[i] is TSymmetryCoords) and TSymmetryCoords(Components[i]).ReallyActive then
          if Assigned(Result) then
            raise EReportException.Create('More than one active Symmetry Coords')
          else
            Result := TSymmetryCoords(Components[i]);
  end;
var
  S: TSymmetryCoords;
  Fcalc: TMatrix;
begin
  if not Active then
    raise EReportException.Create('Internal Co-ords must be active for this transformation');
  { Calculate f (in internal co-ords) from F (in symmetry co-ords) }
  S := ActiveSymmetryCoords;
  if not Assigned(S) then
    Exit;
  S.Prepare(Verbose);
  S.MakeU(S.CartesianDisplacements);
  if Verbose then begin
    WriteLn(Log, 'Transformation between internal and symmetry co-ordinates');
    S.U.Write(Log, 'U   ', S.nameS, nameR, Precision, Precision-4);
  end;
  S.MakeFMatrix(S.U.RowCount, 'F', Verbose, S.nameS);
  if not Assigned(S.F) then
    raise EReportException.Create('No F matrix');
  { Equating potential energy formulae:
      R* f R = S* F S
    and using S = UR;
      S* F S = R* U* F U R
    so
      f = U* F U
  }
  Fcalc := nil;
  try
    Fcalc := S.U.Transpose;
    Fcalc.PostMultiply(S.F);
    Fcalc.PostMultiply(S.U);
    if Verbose then
      Fcalc.Write(Log, 'f', nameR, nameR, Precision);
    if DoApply then
      LoadF(Fcalc);
  finally
    Fcalc.Free;
  end;
end;

procedure TInternalCoords.ClearArrays;
begin
  FreeAndNil(B);
  inherited;
end;

function TInternalCoords.HelpOnSetting(AIndex: Integer): string;
begin
  if AIndex = 2 then
    Result := 'True to calculate Cartesian displacements in Principal Axis System.'
  else
    Result := inherited HelpOnSetting(AIndex);
end;

procedure TInternalCoords.MakeBMatrix(State: TElectronicState; Verbose: Boolean);
var
  nNuclei: Integer;
  Atoms: array [1..4] of Txyz;
  xyz: Integer;
  nBonds: array of Integer; { For each atom, how many bonds connected }
  Bonds: array of array of Integer; { For each atom, a list of atoms it is bonded to }
  function TryStrToIntEx(const s: string; out Value: Integer): Boolean;
  begin
    { Do not accept hex numbers }
    if (Length(s) > 0) and (s[1] in ['0'..'9']) then
      Result := TryStrToInt(s, Value)
    else
      Result := False;
  end;
  function FindAtoms(r, n: Integer; LastOptional: Boolean; LastTwoOptional: Boolean = False):Boolean;
  var
    i, j, na: Integer;
    s: string;
  begin
    Result := False;
    for i := 1 to 4 do
      AtomIndex[nInternal-1][i] := -2; { No atom }
    SetLength(ModeAtoms[nInternal-1],n);
    for i := 1 to n do begin
      if i <= High(A[r]) then
        s := Trim(A[r][i])
      else
        s := '';
      if s = '' then
        if (i = n) and LastOptional then begin
          n := i-1;
          Break
        end else if (i = n-1) and LastTwoOptional then begin
          n := i-1;
          Break;
        end else begin
          HandleError(r,i,'Missing Atom '+IntToStr(i));
          Exit;
        end
      else begin
        na := zMatrix.FindAtom(s);
        if na < 0 then begin
          HandleError(r,i,'Bad Atom name:'+s);
          Exit;
        end else begin
          if InPrincipalAxis then
            Atoms[i] := zMatrix.Atoms[na].pxyz
          else
            Atoms[i] := zMatrix.Atoms[na].xyz;
          ModeAtoms[nInternal-1][i-1] := na;
          na := zMatrix.Atoms[na].Index;
          if na < 0 then
            if not LastOptional or (i <> n) then begin
              HandleError(r,i,'Dummy Atom not allowed:'+s);
              Exit;
            end else
              na := nNuclei; { Flag for dummy }
          AtomIndex[nInternal-1][i] := na;
          for j := 1 to i-1 do
            if AtomIndex[nInternal-1][j] = na then begin
              HandleError(r,i,Format('Atoms %d and %d identical',[j,i]));
              Exit;
            end;
        end;
      end;
    end;
    SetLength(ModeAtoms[nInternal-1],n); { Trim optional atoms not used }
    for i := n+1 to 4 do
      if (i <= High(A[r])) and (Trim(A[r][i]) <> '') then
        HandleError(r,i,'Warning - extra atom:'+A[r][i]);
    CurrentAtoms := ModeAtoms[nInternal-1];
    Result := True;
  end;
  function CosToSin(x:Double):Double;
  begin
    Result := 1 - Sqr(x);
    if Result < 0 then
      Result := 0
    else
      Result := Sqrt(Result);
  end;

  procedure AddBond(n1, n2: Integer);
  var
    i: Integer;
  begin
    if n1 < n2 then
      AddBond(n2,n1); { Bonds work both ways }
    n1 := ModeAtoms[nInternal-1][n1-1];
    n2 := ModeAtoms[nInternal-1][n2-1];
    for i := 0 to nBonds[n1]-1 do
      if Bonds[n1][i] = n2 then
        Exit; { ALready in the list }
    if nBonds[n1] > High(Bonds[n1]) then
      SetLength(Bonds[n1],nBonds[n1]+3);
    Bonds[n1][nBonds[n1]] := n2;
    Inc(nBonds[n1]);
  end;

var
  mlk, mlm, mln, ma, mkm, mkn, mkl, mmn: Double;
  elk, elm, eln, ea, ekm, ekn, ekl, emn, elkc, elmc, enc: Txyz;
  cosphi, sinphi, costheta, sintheta, cosklm, sinklm, coslmn, sinlmn: Double;

  function AddTorsion(const AtomNos: array of Integer; ShowError:Boolean; ErrRow: Integer):Boolean;
  var
    i, j, xyz: Integer;
    AtomIndex: T4Integers;
  begin
    Result := False;
    for j := 1 to 4 do begin
      if InPrincipalAxis then
        Atoms[j] := zMatrix.Atoms[AtomNos[j-1]].pxyz
      else
        Atoms[j] := zMatrix.Atoms[AtomNos[j-1]].xyz;
      AtomIndex[j] := zMatrix.Atoms[AtomNos[j-1]].Index;
    end;
    mkl := NDiff(ekl,Atoms[2],Atoms[1]);
    if mkl < LinearTest then begin
      if ShowError then
        HandleError(ErrRow,-1,'Atoms %%1 and %%2 co-incident');
      Exit;
    end;
    mlm := NDiff(elm,Atoms[3],Atoms[2]);
    if mlm < LinearTest then begin
      if ShowError then
        HandleError(ErrRow,-1,'Atoms %%2 and %%3 co-incident');
      Exit;
    end;
    mmn := NDiff(emn, Atoms[4],Atoms[3]);
    if mmn < LinearTest then begin
      if ShowError then
        HandleError(ErrRow,-1,'Atoms %%3 and %%4 co-incident');
      Exit;
    end;
    { A2.5 }
    cosklm := -Dot(ekl,elm);
    coslmn := -Dot(elm,emn);
    sinklm := Cross(elkc,ekl,elm);
    if Abs(sinklm) < LinearTest then begin
      if ShowError then
        HandleError(ErrRow,-1,'Atoms %%1-%%2 and %%2-%%3 parallel');
      Exit;
    end;
    DivBy(elkc,-mkl*Sqr(sinklm));
    sinlmn := Cross(enc,elm,emn);
    if Abs(sinlmn) < LinearTest then begin
      if ShowError then
        HandleError(ErrRow,-1,'Atoms %%2-%%3 and %%3-%%4 parallel');
      Exit;
    end;
    DivBy(enc, mmn*Sqr(sinlmn));
    for xyz := 1 to 3 do begin
      i := AtomIndex[1]*3+xyz; B[nInternal, i] := B[nInternal, i] + elkc[xyz];
      i := AtomIndex[4]*3+xyz; B[nInternal, i] := B[nInternal, i] + enc[xyz];
      i := AtomIndex[2]*3+xyz; B[nInternal, i] := B[nInternal, i] -(mlm-mkl*cosklm)/mlm*elkc[xyz] - mmn/mlm*coslmn*enc[xyz];
      i := AtomIndex[3]*3+xyz; B[nInternal, i] := B[nInternal, i] -(mlm-mmn*coslmn)/mlm*enc[xyz] - mkl/mlm*cosklm*elkc[xyz];
    end;
    Result := True;
  end;

  function AddGeneralTorsion(const AtomNos: array of Integer; ErrRow: Integer): Integer;
  var
    i, j, nT: Integer;
    TAtoms: T4Integers;
    procedure AddAtom(nA: Integer);
    var
      k: Integer;
    begin
      for k := 2 to nT-1 do
        if ModeAtoms[nInternal-1][k] = nA then
          Exit; { Already in list }
      if nT > High(ModeAtoms[nInternal-1]) then
        SetLength(ModeAtoms[nInternal-1],2*nT);
      ModeAtoms[nInternal-1][nT] := nA;
      Inc(nT);
    end;
  begin
    { Generalized tosion after VIBCA }
    TAtoms[2] := AtomNos[0];
    TAtoms[3] := AtomNos[1];
    ModeAtoms[nInternal-1][0] := TAtoms[2];
    ModeAtoms[nInternal-1][1] := TAtoms[3];
    nT := 2;
    Result := 0;
    for i := 0 to nBonds[TAtoms[2]]-1 do begin
      TAtoms[1] := Bonds[TAtoms[2]][i];
      for j := 0 to nBonds[TAtoms[3]]-1 do begin
        TAtoms[4] := Bonds[TAtoms[3]][j];
        if AddTorsion(TAtoms, False, ErrRow) then begin
          AddAtom(-(TAtoms[1]+1));
          AddAtom(TAtoms[4]);
          Inc(Result);
        end;
      end;
    end;
    SetLength(ModeAtoms[nInternal-1],nT);
    if Result > 0 then
      { Use simple average of possible torsions. Note that
        VIBCA uses sum, but scales force constant by Result^2 }
      for i := 1 to B.ColCount do
        B[nInternal,i] := B[nInternal,i]/Result;
  end;

var
  OK: Boolean;
  i, j: Integer;
begin
  { See L Hedberg and I M Mills JMS 160 117 (1993) }
  ClearArrays;
  { TODO: Following should probably not call apply }
  LastError := zMatrix.Apply(Verbose,InPrincipalAxis);
  if LastError <> '' then
    Exit;
  nInternal := 0;
  nLinearBends := 0;
  for i := 0 to High(A) do
    if (Length(A[i]) > 0) and (A[i][0] <> '') then
      Inc(nInternal);
  SetLength(Brows, Length(A));
  SetLength(AtomIndex, nInternal);
  SetLength(ModeType, nInternal);
  SetLength(nBonds,Length(zMatrix.Atoms));
  SetLength(Bonds,Length(nBonds),3);
  SetLength(ModeAtoms, nInternal);
  nNuclei := (State.Owner.Owner as TVibratingMolecule).nNuclei;
  B := TMatrix.Create(nInternal, 3*nNuclei);
  B.Zero;
  nInternal := 0;
  for i := 0 to High(A) do
    if (Length(A[i]) > 0) and (A[i][0] <> '') then begin
      Inc(nInternal);
      Brows[i] := nInternal;
      if LowerCase(A[i][0]) = 'stretch' then begin
        if FindAtoms(i,2,False) then begin
          AddBond(1, 2);
          ModeType[nInternal-1] := icStretch;
          { A2.1 }
          mkl := NDiff(ekl,Atoms[2],Atoms[1]);
          if mkl < LinearTest then
            HandleError(1,-1,'Atoms %%1 and %%2 are co-incident');
          for xyz := 1 to 3 do begin
            B[nInternal, AtomIndex[nInternal-1][1]*3+xyz] := -ekl[xyz];
            B[nInternal, AtomIndex[nInternal-1][2]*3+xyz] := ekl[xyz];
          end;
        end;
      end else if LowerCase(A[i][0]) = 'bend' then begin
        if FindAtoms(i,4,True) then begin
          AddBond(1, 2);
          AddBond(2, 3);
          if AtomIndex[nInternal-1][4] < 0 then
            ModeType[nInternal-1] := icBend
          else
            ModeType[nInternal-1] := icLinBend;
          { A2.2; If 4 atoms given then ea is constrcted by replacing elm with eln }
          mlk := NDiff(elk, Atoms[1], Atoms[2]);
          if mlk < LinearTest then
            HandleError(i,-1,'Atoms %%1 and %%2 co-incident')
          else if NDiff(elm, Atoms[1], Atoms[3]) < LinearTest then
            HandleError(i,-1,'Atoms %%1 and %%3 co-incident')
          else begin
            mlm := NDiff(elm, Atoms[3], Atoms[2]);
            if mlm < LinearTest then
              HandleError(i,-1,'Atoms %%2 and %%3 co-incident')
            else begin
              OK := False;
              if AtomIndex[nInternal-1][4] < 0 then begin
                ma := Cross(ea, elm, elk);
                OK := ma >= LinearTest;
                if not OK then
                  HandleError(i,-1,'%%1-%%2 and %%2-%%3 parallel - specify plane with atom 4')
              end else begin
                { linear bend }
                Inc(nLinearBends);
                mln := NDiff(eln, Atoms[4], Atoms[2]);
                if mln < LinearTest then
                  HandleError(i,-1,'Atoms %%2 and %%4 co-incident')
                else begin
                  ma := Cross(ea, elk, eln);
                  OK := ma >= LinearTest;
                  if not OK then
                    HandleError(i,-1,'%%1-%%2 and %%2-%%4 parallel')
                end;
              end;
              if OK then begin
                DivBy(ea,ma);
                Cross(elkc,ea,elk);
                Cross(elmc,elm,ea);
                for xyz := 1 to 3 do begin
                  B[nInternal, AtomIndex[nInternal-1][1]*3+xyz] := elkc[xyz]/mlk;
                  B[nInternal, AtomIndex[nInternal-1][3]*3+xyz] := elmc[xyz]/mlm;
                  B[nInternal, AtomIndex[nInternal-1][2]*3+xyz] := -(elkc[xyz]/mlk+elmc[xyz]/mlm);
                end;
              end;
            end;
          end;
        end;
      end else if (LowerCase(A[i][0]) = 'plane') or (LowerCase(A[i][0]) = 'bondplane') then begin
        if FindAtoms(i,4,False) then begin
          AddBond(1, 2);
          AddBond(1, 3);
          AddBond(1, 4);
          if LowerCase(A[i][0]) = 'plane' then
            ModeType[nInternal-1] := icPlaneSym
          else
            ModeType[nInternal-1] := icPlaneGen;
          mkl := NDiff(ekl,Atoms[2],Atoms[1]);
          if mkl < LinearTest then
            HandleError(i,-1,'Atoms %%1 and %%2 co-incident')
          else begin
            mkm := NDiff(ekm,Atoms[3],Atoms[1]);
            if mkm < LinearTest then
              HandleError(i,-1,'Atoms %%1 and %%3 co-incident')
            else begin
              mkn := NDiff(ekn, Atoms[4],Atoms[1]);
              if mkn < LinearTest then
                HandleError(i,-1,'Atoms %%1 and %%4 co-incident')
              else begin
                { A2.3 }
                Cross(elkc,ekm,ekn);
                Cross(elmc,ekn,ekl);
                Cross(enc, ekl,ekm);
                if LowerCase(A[i][0]) = 'bondplane' then begin
                  { A2.4 }
                  cosphi := Dot(ekm,ekn);
                  sinphi := Cross(ea,ekm,ekn);
                  sintheta := Dot(ekl,ea)/sinphi;
                  costheta := CosToSin(sintheta);
                  for xyz := 1 to 3 do begin
                    elkc[xyz] := elkc[xyz]/(costheta*sinphi) - sintheta/costheta*ekl[xyz];
                    elmc[xyz] := elmc[xyz]/(costheta*sinphi) - sintheta/costheta/Sqr(sinphi)*(ekm[xyz]-cosphi*ekn[xyz]);
                    enc[xyz]  := enc[xyz] /(costheta*sinphi) - sintheta/costheta/Sqr(sinphi)*(ekn[xyz]-cosphi*ekm[xyz]);
                  end;
                end;
                for xyz := 1 to 3 do begin
                  B[nInternal, AtomIndex[nInternal-1][2]*3+xyz] := elkc[xyz]/mkl;
                  B[nInternal, AtomIndex[nInternal-1][3]*3+xyz] := elmc[xyz]/mkm;
                  B[nInternal, AtomIndex[nInternal-1][4]*3+xyz] := enc[xyz]/mkn;
                  B[nInternal, AtomIndex[nInternal-1][1]*3+xyz] := -(B[nInternal, AtomIndex[nInternal-1][2]*3+xyz]+B[nInternal, AtomIndex[nInternal-1][3]*3+xyz]+B[nInternal, AtomIndex[nInternal-1][4]*3+xyz]);
                end;
              end;
            end;
          end;
        end;
      end else if LowerCase(A[i][0]) = 'torsion' then begin
        if FindAtoms(i,4,False,True) then begin
          if AtomIndex[nInternal-1][3] >= 0 then begin
            AddBond(1, 2);
            AddBond(2, 3);
            AddBond(3, 4);
            ModeType[nInternal-1] := icTorsion;
            AddTorsion(ModeAtoms[nInternal-1],True,i)
          end else begin
            AddBond(1, 2);
            ModeType[nInternal-1] := icGenTorsion;
          end;
        end;
      end else begin
        ModeType[nInternal-1] := icError;
        HandleError(i,1,'Unknown mode type:'+A[i][0]);
      end
    end else
      Brows[i] := -1; { Blank Line }
  nInternal := 0;
  for i := 0 to High(A) do
    if A[i][0] <> '' then begin
      Inc(nInternal);
      if ModeType[nInternal-1] = icGenTorsion then
        if AddGeneralTorsion(ModeAtoms[nInternal-1],i) = 0 then
          HandleError(i,-1,'No chains for torsion');
    end;
  if HasDuplicateRows(A, i, j, True) then
    HandleError(i,-1,'Identical to row '+IntToStr(j+1));
  if Verbose and (nInternal > 0) then begin
    if LastError <> '' then
      WriteLn(Log, LastError);
    if InPrincipalAxis then
      WriteLn(Log, 'B matrix in Principal Axis System')
    else
      WriteLn(Log, 'B matrix in Original Axis System');
    B.Write(Log, 'B', nameR, State.nameNucleusXYZ, Precision, Precision-4);
  end;
end;

procedure TInternalCoords.MakeExtraTerms(dint: TMatrix; Verbose, Apply, CreatePert: Boolean);
var
  Handled: array of Boolean;
  i, j, ii, jj, nToHandle: Integer;
  Order: Integer;
  Bra, Ket: TElectronicState;
  Fe, dT: TMatrix;
  Mx: Double;
  hShown: Boolean;
  procedure AddPerturbation(Value: Double; mode1, mode2: Integer);
  var
    Pnam: string;
    i: Integer;
    Pert, P: TVibrationalPerturbation;
  begin
    { Perturbation name }
    Pnam := 'q'+IntToStr(mode1);
    if mode1 = mode2 then
      Pnam := Pnam + '^2'
    else if mode2 > 0 then
      Pnam := Pnam + 'q'+IntToStr(mode2);
    if Verbose then begin
      if not hShown then
        WriteLn(Log, 'Equivalent Perturbations');
      Write(Log, Pnam, ' ', FloatToStr(Value));
    end;
    Pert := nil;
    for i := 0 to Bra.Manifold.nPerturbations - 1 do begin
      P := Bra.Manifold.Perturbation[i] as TVibrationalPerturbation;
      if (((P.Bra = Bra) and (P.Ket = Ket)) or ((P.Ket = Bra) and (P.Bra = Ket)))
         and (P.PairName = Pnam) then
        if Pert = nil then
          Pert := P
        else begin
          LastError := 'Duplicate perturbations: '+Pert.FullHeirarchicName;
        end;
    end;
    if (Pert = nil) and CreatePert then begin
      Pert := TVibrationalPerturbation.Create(Bra.Manifold);
      Pert.SetBraKetName(Bra.Name, Ket.Name);
      Pert.Op := q;
      Pert.Value := Value;
      if mode1 = mode2 then
        Pert.Settings[Pert.FirstDynamicSetting+mode1-1] := 2
      else begin
        Pert.Settings[Pert.FirstDynamicSetting+mode1-1] := 1;
        if mode2 > 0 then
          Pert.Settings[Pert.FirstDynamicSetting+mode2-1] := 1;
      end;
    end;
    if Verbose then
      if Pert <> nil  then
        WriteLn(Log, ' currently ', FloatToStr(Pert.PreviousValue))
      else
        WriteLn(Log);
    if (Apply or CreatePert) and Assigned(Pert) then begin
      Pert.FromForceField := True;
      Pert.Value := Value;
    end;
    hShown := True;
  end;
begin
  if Length(ExtraTerms) = 0 then
    Exit;
  nToHandle := Length(ExtraTerms);
  SetLength(Handled, nToHandle);
  for i := 0 to High(ExtraTerms) do
    Handled[i] := False;
  Fe := nil;
  Order := 1;
  repeat
    Bra := nil;
    Ket := nil;
    FreeAndNil(Fe);
    for j := 0 to High(ExtraTerms) do
      if not Handled[j] and (Length(ExtraTerms[j].Modes) = Order) then begin
        if Bra = nil then begin
          Bra := ExtraTerms[j].Bra;
          Ket := ExtraTerms[j].Ket;
          if Verbose then
            WriteLn(Log, 'Potential Terms of order ', Order, ' connecting ', Bra.Name, ' and ', Ket.Name);
          case Order of
            1: Fe := TMatrix.Create(1,nInternal);
            2: Fe := TMatrix.Create(nInternal,nInternal);
          end;
          Fe.Zero;
        end else if (Bra <> ExtraTerms[j].Bra) or (Ket <> ExtraTerms[j].Ket) then
          Continue;
        Handled[j] := True;
        Dec(nToHandle);
        ii := ExtraTerms[j].Modes[0];
        case Order of
          1: Fe[1,ii] := Fe[1,ii] + ExtraTerms[j].Value;
          2: begin
               jj := ExtraTerms[j].Modes[1];
               Fe[jj,ii] := Fe[jj,ii] + ExtraTerms[j].Value;
               if ii <> jj then
                 Fe[ii,jj] := Fe[ii,jj] + ExtraTerms[j].Value;
             end;
        end;
      end;
    if Assigned(Fe) then begin
      case Order of
        1: begin
             if Verbose then begin
               WriteLn(Log, 'Potential matrix in internal coordinates');
               Fe.Write(Log, 'f', nil, nameR, Precision);
             end;
             Fe.PostMultiply(dint);
             Fe.Scale(aJConversion);
             if Verbose then begin
               WriteLn(Log, 'Potential matrix in dimensionless normal coordinates / cm-1');
               Fe.Write(Log, 'f', nil, ElectronicState.nameMode, Precision);
             end;
           end;
        2: begin
             if Verbose then begin
               WriteLn(Log, 'Potential matrix in internal coordinates');
               Fe.Write(Log, 'f', nameR, nameR, Precision);
             end;
             Fe.PostMultiply(dint);
             dT := dint.Transpose;
             dT.PostMultiply(Fe);
             Fe.Free;
             Fe := dT;
             Fe.Scale(aJConversion);
             if Verbose then begin
               WriteLn(Log, 'Potential matrix in dimensionless normal coordinates / cm-1');
               Fe.Write(Log, 'f', ElectronicState.nameMode, ElectronicState.nameMode, Precision);
             end;
           end;
      end;
      hShown := False;
      Mx := Fe.MaxAbs * Bra.Molecule.SmallCoefficient;
      if Order = 1 then begin
        for i := 1 to Fe.ColCount do
          if Abs(Fe[1,i]) > Mx then
            AddPerturbation(Fe[1,i], i, -1);
      end else if Order = 2 then begin
        for i := 1 to Fe.RowCount do
          for j := 1 to i do
            if Abs(Fe[i,j]) > Mx then
              if i = j then
                AddPerturbation(Fe[i,i]/2, i, i)
              else
                AddPerturbation(Fe[i,j], j, i);
      end;
    end else begin
      Inc(Order);
      if (nToHandle > 0) and (Order > 2) then
        raise NotYet;
    end;
  until nToHandle = 0;
end;

procedure TInternalCoords.ShowInternalCoords(Sender: TObject);
begin
  {$IFNDEF CONSOLE}
  ConnectInternalCoordsForm(Self);
  {$ENDIF}
end;

{ TCartesianDisplacements }

procedure TCartesianDisplacements.AddHandlers;
begin
  inherited;
  {$IFNDEF CONSOLE}
  AddHandler('View...', 'View and edit cartesian displacements', ShowCartesianDisplacements);
  {$ENDIF}
end;

procedure TCartesianDisplacements.Makel(Verbose, ApplyOperators, CreateOperators: Boolean);
begin
  MakeBMatrix(ElectronicState, Verbose);
  if LastError = '' then
    try
      l := MakelMatrix(lambda, ElectronicState, B, InternalCoords.F, Verbose, ApplyOperators, CreateOperators, nameB);
    except
      on E: Exception do
        HandleError(-1,-1,E.Message);
    end;
end;

function TCartesianDisplacements.Apply(Verbose: Boolean):string;
begin
  Makel(Verbose, True, False);
  if (LastError = '') and Assigned(l) then
    ElectronicState.ApplylMatrix(Self, InPrincipalAxis, Verbose);
  Result := LastError;
end;

procedure TCartesianDisplacements.CalculateFromL(Verbose, DoApply: Boolean);
var
  State: TElectronicState;
  C: TMatrix;
  i, j: Integer;
  m: Double;
begin
  { Back calculate displacements from l matrix, taking internal co-ordinates the same as normal co-ordinates }
  { TODO: Take U into account, if present }
  { TODO: Check effect of InPrincipalAxis flag }
  State := ElectronicState;
  if not Assigned(State.adm) then
    Exit;
  if Verbose or DoApply then
    Prepare(False);
  C := nil;
  try
    if InPrincipalAxis then
      C := State.lMatrix.Transpose
    else
      C := State.adm.Transpose;
    C.RowCount := (State.Owner.Owner as TVibratingMolecule).nModes; { Discard rotations and translations, if present }
    { Post Multiply by M^1/2 or M = diagonal matrix of Sqrt(masses) }
    for j := 1 to C.ColCount do begin
      m := (State.Nucleus((j-1) div 3) as TMolecularNucleus).Mass;
      if InPrincipalAxis then
        m := Sqrt(m); { Using lMatrix, so different scaling }
      for i := 1 to C.RowCount do
        C[i,j] := C[i,j]*m;
    end;
    if Verbose then
      C.Write(Log, 'B', InternalCoords.nameR, State.nameNucleusXYZ, Precision, Precision-4);
    if not DoApply then
      Exit;
    UpdateBMatrix(C, [True], False);
  finally
    C.Free;
  end;
end;

procedure TCartesianDisplacements.CheckBMatrix(Verbose: Boolean; RemoveRT: Boolean);
var
  Bfull: TMatrix;
  i, j, k, nCoordsChanged: Integer;
  Sum: Double;
  CoordsChanged: array of Boolean;
begin
  { Check if co-ordinates orthogonal to Rotation and Translation and each other }
  Bfull := nil;
  try
    Bfull := TMatrix.Create(B);
    zMatrix.MassScale(Bfull,-1);
    zMatrix.AddRT(Bfull, B.RowCount+1, InPrincipalAxis, True);
    if Verbose then begin
      WriteLn(Log, 'Mass scaled B matrix Bm = BM^-1/2 with rotations and translations added');
      Bfull.Write(Log, 'Bm', nameB, ElectronicState.nameNucleusXYZ, Precision, Precision-4);
      Bfull.CheckOrthogonal(Log, 'BmBm^T', False, ElectronicState.Molecule.SmallCoefficient, True, nameB, nameB);
    end;
    if not RemoveRT then
      Exit;
    SetLength(CoordsChanged, B.Colcount);
    nCoordsChanged := 0;
    for i := 1 to B.RowCount do begin { For each mode }
      CoordsChanged[i-1] := False;
      for k := B.RowCount+1 to Bfull.RowCount do begin { For Rx, Ry, Rz, Tx, Ty, Tx }
        Sum := 0;
        for j := 1 to B.Colcount do
          Sum := Sum + Bfull[i,j]*Bfull[k,j];
        if Abs(Sum) > ElectronicState.Molecule.SmallCoefficient*B.RowCount then begin
          CoordsChanged[i-1] := True;
          for j := 1 to B.Colcount do
            Bfull[i,j] := Bfull[i,j] - Sum*Bfull[k,j];
        end;
      end;
      if CoordsChanged[i-1] then
        Inc(nCoordsChanged);
    end;
    if nCoordsChanged > 0 then begin
      if Verbose then begin
        WriteLn(Log, nCoordsChanged, ' co-ordinates changed');
        WriteLn(Log, 'Updated Mass scaled B matrix Bm = BM^-1/2');
        Bfull.Write(Log, 'Bm', nameB, ElectronicState.nameNucleusXYZ, Precision, Precision-4);
        Bfull.CheckOrthogonal(Log, 'BmBm^T', False, ElectronicState.Molecule.SmallCoefficient, True, nameB, nameB);
      end;
      Bfull.RowCount := B.RowCount;
      zMatrix.MassScale(Bfull,1); { Undo mass scaling }
      UpdateBMatrix(Bfull, CoordsChanged, Verbose);
    end;
  finally
    Bfull.Free;
  end;
end;

procedure TCartesianDisplacements.ClearArrays;
begin
  FreeAndNil(B);
  inherited;
end;

function TCartesianDisplacements.GetInternalCoords: TInternalCoords;
begin
  if not Assigned(FInternalCoords) then
    FInternalCoords := ElectronicState.ForceFieldNode(TInternalCoords) as TInternalCoords;
  Result := FInternalCoords;
end;

function TCartesianDisplacements.HelpOnSetting(AIndex: Integer): string;
begin
  if AIndex = 2 then
    Result := 'True if displacements are in Principal Axis System.'
  else if AIndex = 3 then
    Result := 'If true, normalize input symmetry co-ordinate coefficients'
  else
    Result := inherited HelpOnSetting(AIndex);
end;

procedure TCartesianDisplacements.MakeBMatrix(State: TElectronicState; Verbose: Boolean);
var
  i, j, nManual: Integer;
  s: string;
  V, Sum: Double;
begin
  ClearArrays;
  try
    InternalCoords.MakeBMatrix(State, Verbose);
    if InternalCoords.LastError = '' then
      InternalCoords.MakeFMatrix(InternalCoords.nInternal, 'f', Verbose, InternalCoords.nameR);
    LastError := InternalCoords.LastError;
  except
    on E: Exception do
      LastError := E.Message;
  end;
  if LastError <> '' then
    Exit;
  if Assigned(InternalCoords.B) then
    B := TMatrix.Create(InternalCoords.B)
  else
    B := TMatrix.Create(0,3*(State.Owner.Owner as TVibratingMolecule).nNuclei);
  SetLength(Brows, Max(B.RowCount, Length(A)));
  SetLength(Rows,Length(Brows));
  SetLength(RowActive, Length(Brows));
  nManual := 0;
  for i := 0 to High(Brows) do begin
    Brows[i] := i+1;
    Rows[i] := i;
    RowActive[i] := False;
    if i <= High(A) then
      for j := 0 to High(A[i]) do
        if Trim(A[i][j]) <> '' then begin
          RowActive[i] := True;
          Break;
        end;
    if RowActive[i] then begin
      if i >= B.RowCount then
        B.SetRowCountAndClear(i+1,True);
      Sum := 0;
      for j := 0 to B.ColCount-1 do begin
        if j < Length(A[i]) then
          s := Trim(A[i][j])
        else
          s := '';
        if s = '' then
          V := 0
        else
          V := HandleValue(i,j);
        B[i+1,j+1] := V;
        Sum := Sum + Sqr(V);
      end;
      Sum := Sqrt(Sum);
      if Normalize and (Sum > ElectronicState.Molecule.SmallCoefficient*B.ColCount) then
        for j := 0 to B.ColCount-1 do
          B[i+1,j+1] := B[i+1,j+1]/Sum;
      Inc(nManual);
    end else
      if i <= High(A) then
        SetLength(A[i],0);
  end;
  if nManual = 0 then
    InPrincipalAxis := InternalCoords.InPrincipalAxis
  else if (nManual < B.RowCount) and (InternalCoords.InPrincipalAxis <> InPrincipalAxis) then
    raise EReportException.Create('Internal Co-ordinates and Cartesian Displacement not in same axis system for '+UniqueName);
  if Verbose and (nManual > 0) then begin
    Write(Log, 'B matrix after direct input of '+IntToStr(nManual)+' rows');
    if InPrincipalAxis then
      WriteLn(Log, ' in Principal Axis System')
    else
      WriteLn(Log, ' in Original Axis System');
    B.Write(Log, 'B   ', InternalCoords.nameR, State.nameNucleusXYZ, Precision, Precision-4);
  end;
  if B.RowCount = 0 then
    HandleError(-1,-1,'No Valence or Cartesian internal modes specified');
//  if Verbose and Assigned(InternalCoords.F) then begin
//    WriteLn(Log, ' f Matrix in Cartesian Co-ordinates');
//    Fcart := B.Transpose;
//    Fcart.PostMultiply(InternalCoords.F);
//    Fcart.PostMultiply(B);
//    Fcart.Write(Log, 'f', State.nameNucleusXYZ, State.nameNucleusXYZ)
//  end;
end;

function TCartesianDisplacements.nameB(Mode: Integer): string;
begin
  Result := '';
  if Assigned(InternalCoords) and (Mode <= InternalCoords.nInternal) then
    Result := InternalCoords.nameR(Mode)
  else if Assigned(B) and (Mode > B.RowCount) and Assigned(zMatrix) then
    if Mode <= B.RowCount + zMatrix.nRotations then
      Result := 'R' + zMatrix.RotationName[Mode-B.RowCount]
    else if Mode <= B.RowCount + zMatrix.nRotations + 3 then
      Result := 'T' + 'xyz'[Mode - B.RowCount - zMatrix.nRotations];
  if Result = '' then
    Result := 'B' + IntToStr(Mode);
end;

procedure TCartesianDisplacements.Prepare(Verbose: Boolean);
begin
  { Use for partial calculations from user interface;
    Make sure dependent nodes in order }
  if InternalCoords.ApplyOK(Verbose) then
    Apply(Verbose)
  else
    LastError := InternalCoords.LastError;
end;

procedure TCartesianDisplacements.ShowCartesianDisplacements(Sender: TObject);
begin
  {$IFNDEF CONSOLE}
  ConnectCartesianDisplacementsForm(Self);
  {$ENDIF}
end;

procedure TCartesianDisplacements.UpdateBMatrix(C: TMatrix; const Changed: array of Boolean; Verbose: Boolean);
var
  i, j, u: Integer;
begin
  SetLength(A, C.RowCount);
  u := 0;
  for i := 1 to C.RowCount do begin
    if Changed[u] then begin
      SetLength(A[i-1], C.ColCount);
      for j := 1 to C.ColCount do
        A[i-1][j-1] := FloatToStr(C[i,j]);
    end;
    if u < High(Changed) then
      Inc(u);
  end;
  MakeBMatrix(ElectronicState, Verbose);
end;

procedure TCartesianDisplacements.UpdateObjectChain;
begin
  inherited;
  FInternalCoords := nil;
end;

{ TSymmetryCoords }

procedure TSymmetryCoords.AddHandlers;
begin
  inherited;
  {$IFNDEF CONSOLE}
  AddHandler('View...', 'View and edit symmetry co-ordinates', ShowSymmetryCoords);
  {$ENDIF}
end;

procedure TSymmetryCoords.MakeLMatrix(State: TElectronicState; Verbose, ApplyOperators, CreateOperators, ForDisplay: Boolean);
var
  i, j, r, n, info: Integer;
  BsT, Lm1T, Gm1, AL: TMatrix;
  m, Biggest: Double;
  uplo: AnsiChar;
  BiggestNeg: Boolean;
  MappedMode: array of Integer;
  oldlambda: TColumnVector;
  procedure PostDivM(B: TMatrix);
  var
    i, j: Integer;
    m: Double;
  begin
    { Post Multiply by 1/M = matrix of inverse masses }
    for j := 1 to B.ColCount do begin
      m := 1/(State.Nucleus((j-1) div 3) as TMolecularNucleus).Mass;
      for i := 1 to B.RowCount do
        B[i,j] := B[i,j]*m;
    end;
  end;
  procedure ShowPED(const Title: string; ShowName: TShowName; L, F: TMatrix);
  var
    r, i, j, n, nVal: Integer;
    P: array of Double;
  begin
    if not Assigned(F) then
      Exit;
    WriteLn(Log,'Potential energy distribution '+Title);
    WriteLn(Log,'(Zero rows omitted)');
    n := L.RowCount;
    SetLength(P, n);
    for r := 1 to State.Molecule.nModes do begin
      WriteLn(Log, State.Mode(r-1).Name);
      for i := 1 to n do begin
        nVal := 0; { Skip zero rows }
        for j := 1 to n do begin
          P[j-1] := L[i,r]*L[j,r]*F[i,j]*2/lambda[r]*100;
          if Abs(P[j-1]) > 0.02 then
            Inc(nVal);
        end;
        P[i-1] := P[i-1]/2;
        if nVal > 0 then begin
          Write(Log, ShowName(i):5);
          for j := 0 to n-1 do
            Write(Log, P[j]:6:1);
          WriteLn(Log);
        end;
      end;
    end;
  end;
begin
  ClearArrays;
  try
    CartesianDisplacements.MakeBMatrix(State, Verbose);
    LastError := CartesianDisplacements.LastError;
  except
    on E: Exception do
      LastError := E.Message;
  end;
  if LastError <> '' then
    Exit;
  MakeU(CartesianDisplacements);
  Bs := TMatrix.Create(U);
  Bs.PostMultiply(CartesianDisplacements.B);
  if Verbose and (High(A) > 0) then begin
    G := TMatrix.Create(CartesianDisplacements.B);
    PostDivM(G);
    BsT := CartesianDisplacements.B.Transpose;
    G.PostMultiply(BsT);
    BsT.Free;
    WriteLn(Log, 'G matrix in internal co-ordinates');
    G.Write(Log, 'Gint', CartesianDisplacements.InternalCoords.nameR, CartesianDisplacements.InternalCoords.nameR, Precision);
    WriteLn(Log, 'Transformation between internal and symmetry co-ordinates');
    U.Write(Log, 'U   ', nameS, CartesianDisplacements.InternalCoords.nameR, Precision, Precision-4);
    WriteLn(Log, 'Symmetrised B matrix');
    Bs.Write(Log, 'B   ', nameS, State.nameNucleusXYZ, Precision, Precision-4);
    G.Free;
  end;
  G := TMatrix.Create(Bs);
  PostDivM(G);
  BsT := Bs.Transpose;
  G.PostMultiply(BsT);
  BsT.Free;
  if Verbose then
    G.Write(Log, 'G   ', nameS, nameS, Precision);
  { Avoid spurious mode mixing by forcing small elements to zero }
  Biggest := ElectronicState.Molecule.SmallCoefficient;
  i := TrimSmall(G, Biggest, j, True);
  if (i > 0) and Verbose then begin
    WriteLn(Log, 'G matrix with ', i, ' elements < ', Biggest, ' removed');
    G.Write(Log, 'G   ', nameS, nameS, Precision);
  end;
  for i := 1 to G.ColCount do
    if G[i,i] = 0 then begin
      HandleError(i-1,-1,'Symmetry co-ordinate '+IntToStr(i) +' has no net displacement' );
      if Verbose then
        WriteLn(Log, LastError);
      Exit;
    end;
  { Decompose G matrix }
  n := G.RowCount;
  GA := TMatrix.Create(G);
  uplo := 'L';
  info := 0;
  if n <> 0 then
    dpotrf_(uplo, n, GA.FirstItem^, n, info, 1);
  if info <> 0 then begin
    HandleError(-1,-1,'dpotrf fails (code '+IntToStr(info)+') - redundant coordinates?' );
    if Verbose then begin
      WriteLn(Log, LastError);
      if info > 0 then
        try
          WriteLn(Log, 'Partial decomposition of G matrix so that G = CC^T');
          GA.Write(Log, 'C   ', nameS, nameS, Precision);
          WriteLn(Log, 'Eigenvalues of G matrix');
          GA := TMatrix.Create(G);
          lambda := TColumnVector.Create(n);
          Eigen(lambda, GA, True, ElectronicState.Molecule.SmallCoefficient, Verbose, nameS, Precision);
        finally
          FreeAndNil(lambda);
          FreeAndNil(GA);
        end;
    end;
    Exit;
  end;
  if ForDisplay then
    Exit; { Only need Bs, but calculation continued to here to allow display of dpotrf error }
  { Zero fill entries left by dpotrf }
  for i := 1 to n
    do for j := i+1 to n do
      GA[i,j] := 0;
  if Verbose then begin
    WriteLn(Log, 'G matrix decomposed so that G = CC^T');
    GA.Write(Log, 'C   ', nameS, nameS, Precision);
  end;
  { Avoid spurious mode mixing by forcing small elements to zero }
  Biggest := ElectronicState.Molecule.SmallCoefficient;
  i := TrimSmall(GA, Biggest, j, True);
  if (i > 0) and Verbose then begin
    WriteLn(Log, 'C matrix with ', i ,' elements < ', Biggest, ' removed');
    GA.Write(Log, 'C   ', nameS, nameS, Precision);
  end;
  { Require F matrix to be completely absent (all entries blank) in which case the matrix is worked out from the
    InternalCoords f, or the complete matrix must be supplied }
  i := TrimArray(AF);
  if i > 1 then begin
    { F matrix given in symmetry co-ordinates }
    if Assigned(CartesianDisplacements.InternalCoords.F) then
      LastWarning := LastWarning + ' ' + 'Warning - force constant matrix given in both internal and symmetry coordinates';
    MakeFMatrix(G.ColCount, 'F   ', Verbose, nameS);
    if not Assigned(F) then
      Exit;
    if Verbose then begin
      WriteLn(Log, 'U^TFU = F matrix transformed to internal co-ordinates');
      BsT := U.Transpose;
      BsT.PostMultiply(F);
      BsT.PostMultiply(U);
      BsT.Write(Log, 'U^TFU', CartesianDisplacements.InternalCoords.nameR, CartesianDisplacements.InternalCoords.nameR, Precision);
      BsT.Free;
    end;
  end else if Assigned(CartesianDisplacements.InternalCoords.F) then begin
    BsT := InvertU(Verbose);
    if Assigned(BsT) then begin
      F := BsT.Transpose;
      F.PostMultiply(CartesianDisplacements.InternalCoords.F);
      F.PostMultiply(BsT);
      FreeAndNil(BsT);
    end else
      Exit;
    if Verbose then
      F.Write(Log, 'F   ', nameS, nameS, Precision);
    n := F.ColCount;
    SetLength(ComputedF,(n*(n+1)) div 2);
    r := 0;
    for i := 1 to n do
      if F[i,i] <> 0 then begin
        ComputedF[r].V := F[i,i];
        ComputedF[r].Coords[1] := i;
        ComputedF[r].Coords[2] := i;
        Inc(r);
      end;
    for i := 1 to n do
      for j := 1 to i-1 do
        if F[i,j] <> 0 then begin
          ComputedF[r].V := F[i,j];
          ComputedF[r].Coords[1] := i;
          ComputedF[r].Coords[2] := j;
          Inc(r);
        end;
    SetLength(ComputedF,r);
  end else
    Exit; { No force constant matrix available }
  bigL := GA.Transpose;
  bigL.PostMultiply(F);
  bigL.PostMultiply(GA);
  if Verbose then
    bigL.Write(Log, 'C^TFC', nameS, nameS, Precision);
  { Avoid spurious mode mixing by forcing small elements to zero }
  Biggest := ElectronicState.Molecule.SmallCoefficient;
  i := TrimSmall(bigL, Biggest, j, True);
  if (i > 0) and Verbose then begin
    WriteLn(Log, 'C^TFC with ', i ,' elements < ', Biggest, ' removed');
    bigL.Write(Log, 'C^TFC', nameS, nameS, Precision);
  end;
  try
    EigenGF(bigL, lambda, ElectronicState.Molecule.SmallCoefficient, Verbose, nameS, Precision);
  except
    on E:Exception do begin
      HandleError(-1,-1,E.Message);
      FreeAndNil(bigL);
      FreeAndNil(lambda);
    end;
  end;
  if not Assigned(lambda) then
    Exit;
  BsT := TMatrix.Create(GA);
  BsT.PostMultiply(bigL); { AV }
  bigL.Free;
  bigL := BsT;
  { Force largest coefficient to be positive, along the lines of suggestion on p111 of
    S.J.Cyvin, "Molecular Vibrations and Mean Square Amplitudes", Elsevier, Amsterdam (1968) }
  BiggestNeg := False;
  for i := 1 to n do begin
    Biggest := 0;
    for j := 1 to bigL.ColCount do
      if Abs(bigL[j,i]) > Biggest then begin
        Biggest := Abs(bigL[j,i]);
        BiggestNeg := bigL[j,i] < 0;
      end;
    if BiggestNeg then
      for j := 1 to bigL.ColCount do
        bigL[j,i] := -bigL[j,i];
  end;
  if PreserveS then begin
    if Verbose then
      lambda.Write(Log, 'Original lambda', Precision);
    { Try and keep normal modes in the same order as the symmetry co-ordinates, matching
      the largest coefficient. }
    SetLength(MappedMode,n);
    for i := 0 to n-1 do
      MappedMode[i] := 0;
    for i := 1 to n do begin
      Biggest := 0;
      r := -1;
      for j := 1 to n do
        if (MappedMode[j-1] = 0) and (Abs(bigL[i,j]) > Biggest) then begin
          Biggest := Abs(bigL[i,j]);
          r := j;
        end;
      if r > 0 then
        MappedMode[r-1] := i
      else begin
        WriteLn(Log, 'PreserveS fails for mode ',i);
        Break;
      end;
    end;
    if r > 0 then begin
      bigL := TMatrix.Create(BsT);
      oldlambda := lambda;
      lambda := TColumnVector.Create(oldlambda);
      for i := 1 to n do
        if MappedMode[i-1] <> i then begin
          r := MappedMode[i-1];
          lambda[r] := oldlambda[i];
          for j := 1 to n do
            bigL[j,r] := BsT[j,i];
        end;
      BsT.Free;
      oldlambda.Free;
    end;
  end;
  if Verbose then begin
    lambda.Write(Log, 'lambda', Precision);
    WriteLn(Log, 'L Matrix');
    Write(Log, '/cm-1');
    for i := 1 to n do
      Write(Log, Sign(lambda[i])*sqrt(Abs(lambda[i])/LambdaConversion):Precision+5:Precision-6);
    WriteLn(Log);
    bigL.Write(Log, 'L   ', nameS, State.nameMode, Precision+5, Precision);
    Lm1T := TMatrix.Create(bigL);
    for i := 1 to n do begin
      m := 1/Sqrt(gammaConversion*Sqrt(Abs(lambda[i])));
      for j := 1 to Lm1T.RowCount do
        Lm1T[j,i] := Lm1T[j,i] * m;
    end;
    WriteLn(Log, 'Transformation between dimensionless normal coordinates and Symmetry coordinates');
    Lm1T.Write(Log, 'dsym', nameS, State.nameMode, State.Precision+11, State.Precision+6);
    Lm1T.Free;
  end;
  for i := 1 to n do
    if lambda[i] < 0 then begin
      HandleError(-1,-1,'Negative Frequency for '+State.nameMode(i));
      Exit;
    end;
  Lm1T := TMatrix.Create(F);
  Lm1T.PostMultiply(bigL);
  for j := 1 to Lm1T.ColCount do begin
    m := 1/lambda[j];
    for i := 1 to Lm1T.RowCount do
      Lm1T[i,j] := Lm1T[i,j]*m;
  end;
  if Verbose then
    Lm1T.Write(Log, 'L^-1T', nameS, State.nameMode, Precision, Precision-4);
  if Verbose then begin
    Gm1 := TMatrix.Create(Lm1T);
    AL := Lm1T.Transpose;
    Gm1.PostMultiply(AL);
    AL.Free;
    Gm1.Write(Log, 'G^-1', nameS, nameS, Precision);
    Gm1.Free;
  end;
  if Verbose then
    { Potential energy distribution for each mode }
    ShowPED('',nameS,bigL,F);
  AL := Bs.Transpose;
  AL.PostMultiply(Lm1T);
  l := TMatrix.Create(AL);
  { Pre multiply by 1/M = matrix of inverse masses for AL
    Pre Multiply by 1/Sqrt(M) for l }
  for i := 1 to AL.RowCount do begin
    m := 1/(State.Nucleus((i-1) div 3) as TMolecularNucleus).Mass;
    for j := 1 to AL.ColCount do begin
      AL[i,j] := AL[i,j]*m;
      l[i,j] := l[i,j]*Sqrt(m);
    end
  end;
  if Verbose then begin
    if High(A) > 0 then begin
      BsT := TMatrix.Create(CartesianDisplacements.B);
      BsT.PostMultiply(AL);
      WriteLn(Log, 'L matrix in internal co-ordinates');
      BsT.Write(Log, 'Lint', CartesianDisplacements.InternalCoords.nameR, State.nameMode, Precision);
      ShowPED('in internal co-ordinates',CartesianDisplacements.InternalCoords.nameR,BsT,CartesianDisplacements.InternalCoords.F);
      BsT.Free;
    end;
    AL.Write(Log, 'AL  ', State.nameNucleusXYZ, State.nameMode, Precision);
  end;
  if Verbose then
    l.Write(Log, 'l   ', State.nameNucleusXYZ, State.nameMode, Precision);
  ShowDimemsionless(TMatrix.Create(l), CartesianDisplacements.B, lambda, State, CartesianDisplacements.nameB, Verbose, ApplyOperators, CreateOperators);
  Lm1T.Free;
  AL.Free;
end;

procedure TSymmetryCoords.MakeLMatrix(State: TElectronicState; Verbose: Boolean; ACartesianDisplacements: TCartesianDisplacements);
begin
  FCartesianDisplacements := ACartesianDisplacements;
  try
    MakeLMatrix(State, Verbose, False, False, False);
  finally
    FCartesianDisplacements := nil;
  end;
end;

procedure TSymmetryCoords.MakeU(CartesianDisplacements: TCartesianDisplacements);
var
  i,j, n, nInternal: Integer;
  V, Sum: Double;
begin
  nInternal := CartesianDisplacements.B.RowCount;
  TrimArray(A);
  if Length(A) > 0 then
    n := Length(A)
  else
    n := nInternal;
  FreeAndNil(U);
  U := TMatrix.Create(n, nInternal);
  U.Zero;
  SetLength(Brows, n);
  SetLength(Rows, n);
  if n = 0 then
    Exit;
  for i := 0 to n-1 do begin
    Brows[i] := i+1;
    Rows[i] := i;
  end;
  if High(A) > 0 then begin
    for i := 0 to n-1 do begin
      Sum := 0;
      for j := 0 to Min(High(A[i]), nInternal-1) do begin
        V := HandleValue(i,j,False,True);
        Sum := Sum + Sqr(V);
        U[i+1,j+1] := V;
      end;
      Sum := Sqrt(Sum);
      if Normalize and (Sum > ElectronicState.Molecule.SmallCoefficient*nInternal) then
        for j := 1 to nInternal do
          U[i+1,j] := U[i+1,j]/Sum;
    end;
  end else
    { U = 1 }
    for i := 1 to n do
      U[i,i] := 1;
end;

function TSymmetryCoords.Apply(Verbose: Boolean):string;
begin
  MakeLMatrix(ElectronicState, Verbose, True, False, False);
  Result := LastError;
  if not Assigned(l) then
    Exit;
  ElectronicState.ApplylMatrix(Self, CartesianDisplacements.InPrincipalAxis, Verbose);
end;

function TSymmetryCoords.nameS(Mode: Integer): string;
begin
  Result := 'S' + IntToStr(Mode);
end;

procedure TSymmetryCoords.Prepare(Verbose: Boolean);
begin
  { Use for partial calculations from user interface;
    Make sure dependent nodes in order }
  if not CartesianDisplacements.ApplyOK(Verbose) then
    LastError := CartesianDisplacements.LastError;
end;

procedure TSymmetryCoords.CalculateF(Verbose: Boolean; DoApply: Boolean);
begin
  // Was: Prepare(Verbose);
  CartesianDisplacements.Makel(Verbose, True, False);
  LastError := CartesianDisplacements.LastError;
  MakeU(CartesianDisplacements);
  inherited CalculateF(CartesianDisplacements.InternalCoords.nameR, nameS, CartesianDisplacements.B, U,
                       CartesianDisplacements.InPrincipalAxis, Verbose, DoApply);
end;

procedure TSymmetryCoords.ClearArrays;
begin
  FreeAndNil(U);
  FreeAndNil(Bs);
  FreeAndNil(G);
  FreeAndNil(GA);
  FreeAndNil(bigL);
  SetLength(ComputedF,0);
  inherited;
end;

function TSymmetryCoords.GetCartesianDisplacements: TCartesianDisplacements;
begin
  if not Assigned(FCartesianDisplacements) then
    FCartesianDisplacements := ElectronicState.ForceFieldNode(TCartesianDisplacements) as TCartesianDisplacements;
  Result := FCartesianDisplacements;
end;

function TSymmetryCoords.HelpOnSetting(AIndex: Integer): string;
begin
  case AIndex of
    2: Result := 'If true, normalize input symmetry co-ordinate coefficients';
    3: Result := 'Try and keep normal modes in same order as symmetry co-ordinates';
  else
    Result := inherited HelpOnSetting(AIndex);
  end;
end;

function TSymmetryCoords.InvertU(Verbose: Boolean): TMatrix;
begin
  Result := nil;
  if U.RowCount <> U.ColCount then begin
    { Can't invert U, so assume orthonormal }
    if U.CheckOrthogonal(Log, 'UU^T', False, ElectronicState.Molecule.SmallCoefficient, Verbose) > ElectronicState.Molecule.SmallCoefficient*U.RowCount then
      LastError := 'UU^T check fails'
    else
      Result := U.Transpose;
  end else begin
    Result := TMatrix.Create(U);
    { Try and invert U }
    try
      Invert(Result);
      if Verbose then
        Result.Write(Log, 'U^-1', nil, nil, Precision);
    except
      on E:Exception do begin
        HandleError(-1,-1,E.Message);
        FreeAndNil(Result);
      end;
    end;
  end;
end;

procedure TSymmetryCoords.ShowSymmetryCoords(Sender: TObject);
begin
  {$IFNDEF CONSOLE}
  ConnectSymmetryCoordsForm(Self);
  {$ENDIF}
end;

procedure TSymmetryCoords.UpdateObjectChain;
begin
  inherited;
  FCartesianDisplacements := nil;
end;

{ TMNodeWithFMatrix }

procedure TMNodeWithFMatrix.CalculateF(RowLabel, SRowLabel: TShowName; B, U: TMatrix; InPrincipalAxis, Verbose, DoApply: Boolean);
var
  State: TElectronicState;
  C, Linv, Fcalc: TMatrix;
  i, j: Integer;
  m: Double;
  procedure TryInvert(ForwardName, BackName, CName, FLabel, FName: string; RLabel: TShowName);
  var
    i, j: Integer;
  begin
    if Verbose then begin
      WriteLn(Log, ForwardName);
      C.Write(Log, CName, RLabel, State.nameMode, Precision);
    end;
    if C.RowCount = C.ColCount then begin
      Linv := TMatrix.Create(C);
      try
        Invert(Linv);
      except
        on E: Exception do begin
          LastError := E.Message;
          raise EReportException.Create(LastError);
        end;
      end;
      if Verbose then begin
        WriteLn(Log, BackName);
        Linv.Write(Log, CName+'^-1', State.nameMode, RLabel, Precision);
      end;
      Fcalc := Linv.Transpose;
    end else begin
      m := C.CheckOrthogonal(Log, CName, False, State.Molecule.SmallCoefficient, Verbose, RLabel, RLabel);
      if m <= State.Molecule.SmallCoefficient then begin
        { Can invert transformation }
        Linv := C.Transpose;
        Fcalc := TMatrix.Create(C);
      end else if Verbose then
        WriteLn(Log, 'Can''t invert transformation');
    end;
    if Assigned(Fcalc) then begin
      { Post Multiply by diagonal matrix of Lambda }
      for j := 1 to Fcalc.ColCount do begin
        m := Sqr(State.Mode(j-1).Omega)*LambdaConversion;
        for i := 1 to Fcalc.RowCount do
          Fcalc[i,j] := Fcalc[i,j]*m;
      end;
      Fcalc.PostMultiply(Linv);
      if Verbose then begin
        WriteLn(Log, FLabel);
        Fcalc.Write(Log, Fname, RLabel, RLabel, Precision);
      end;
    end;
  end;
begin
  { Back calculate f matrix from l matrix }
  { Key Equations (* denotes transpose):
      2V = S* F S = Q* Lambda Q
    Substituting for symmetry coordinates expressed as:
      S = U R
    where the internal co-ordinates, R are expressed as:
      R = B Dx where Dx are cartesian displacements
    we have:
      S* F S = Dx* B* U* F U B Dx
    The displacements can in turn be expressed in terms of the (mass adjusted) normal co-ordinates:
      Dx = M^-1/2 l Q
    where M is the matrix with masses on the diagonal. Substituting:
      S* F S = Q* l* M^-1/2 B* U* F U B M^-1/2 l Q
      = Q* Lambda Q
    so:
      Lambda = l* M^-1/2 B* U* F U B M^-1/2 l
    This can be simplified by introducing the L matrix:
      L = U B M^-1/2 l
    which relates S and Q:
      S = L Q
    we have
      Lambda = L* F L
    If L is square, and can thus be inverted:
      F = L^-1* Lambda L^-1
    If L is not square, the redundant co-ordinates case, it is possible to proceed if L L* = 1:
      L Lambda L* = L L* F L L* = F
    For internal co-ordinates, R, where the potential energy expression is similar:
      2V = R* f R
    a similar set of equations applies, simply omitting U. We define
      C = B M^-1/2 l
    which relates the internal and nromal co-ordinates:
      R = C Q
  }
  State := ElectronicState;
  State.PrepareMatrices(False);
  if not Assigned(State.adm) or not Assigned(B) then
    Exit;
  C := nil;
  Linv := nil;
  Fcalc := nil;
  try
    { Evaluate  C = B M^-1/2 l }
    C := TMatrix.Create(B);
    if InPrincipalAxis then begin
      { Post Multiply by M^1/2 = diagonal matrix of Sqrt(masses) }
      for j := 1 to C.ColCount do begin
        m := Sqrt((State.Nucleus((j-1) div 3) as TMolecularNucleus).Mass);
        for i := 1 to B.RowCount do
          C[i,j] := C[i,j]*m;
      end;
      C.PostMultiply(State.lMatrix);
    end else
      C.PostMultiply(State.adm);
    C.ColCount := (State.Owner.Owner as TVibratingMolecule).nModes; { Discard rotations and translations, if present }
    TryInvert('Transformation between normal and internal co-ordinates, C = BM^1/2l',
              'Transformation between internal and normal co-ordinates', 'C',
              'Force constant matrix in internal co-ordinates', 'f', RowLabel);
    if Assigned(U) then begin
      FreeAndNil(Linv);
      Linv := TMatrix.Create(U);
      Linv.PostMultiply(C);
      FreeAndNil(C);
      C := Linv;
      Linv := nil;
      FreeAndNil(Fcalc);
      TryInvert('Transformation between normal and symmetry co-ordinates, L = UBM^1/2l',
                'Transformation between symmetry and normal co-ordinates', 'L',
                'Force constant matrix in symmetry co-ordinates', 'F', SRowLabel);
    end;
    if DoApply and Assigned(Fcalc) then
      LoadF(Fcalc);
  finally
    C.Free;
    Linv.Free;
    Fcalc.Free;
  end;
end;

procedure TMNodeWithFMatrix.ClearArrays;
begin
  FreeAndNil(F);
  FreeAndNil(Flinear);
  SetLength(ExtraTerms,0);
  inherited;
end;

procedure TMNodeWithFMatrix.LoadF(NewF: TMatrix);
var
  nFC, i, j: Integer;
  procedure AddValue(V: Double; m1, m2: Integer);
  begin
    if Abs(V) < ElectronicState.Molecule.SmallCoefficient*NewF.ColCount then
      Exit;
    if nFC >= Length(AF) then
      SetLength(AF, nFC+20);
    SetLength(AF[nFC],3);
    AF[nFC][0] := FloatToStr(V);
    AF[nFC][1] := IntToStr(m1);
    AF[nFC][2] := IntToStr(m2);
    Inc(nFC);
  end;
begin
  nFC := 0;
  for i := 1 to NewF.RowCount do
    AddValue(NewF[i,i],i,i);
  for i := 1 to NewF.RowCount do
    for j := 1 to i-1 do
      AddValue(NewF[j,i],j,i);
  SetLength(AF, nFC);
end;

procedure TMNodeWithFMatrix.MakeFMatrix(nModes: Integer; const Caption: string; Verbose: Boolean; ShowName: TShowName);
var
  V: Double;
  i, j, p, Order, Mode: Integer;
  Modes: TIntegerArray;
  OpElem: string;
  PT: TPotentialTerm;
  function FindState(StateName: string):TElectronicState;
  begin
    Result := nil;
    StateName := Trim(StateName);
    if StateName <> '' then begin
      Result := (ElectronicState.Owner as TManifold).StateWithName(StateName) as TElectronicState;
      if not Assigned(Result) then
        HandleError(i,j,'Bad state name ('+StateName+') in '+AF[i][j]);
    end;
    if Result = nil then
      Result := ElectronicState; { Default for blank name }
  end;
begin
  { Construct F matrix, if present }
  i := TrimArray(AF);
  if i > 1 then begin
    F := TMatrix.Create(nModes,nModes);
    F.Zero;
    SetLength(Modes,i-1);
    for i := 0 to High(AF) do
      if Length(AF[i]) > 1 then begin
        V := HandleValue(i,0,True);
        if V <> 0 then begin
          Order := 0;
          OpElem := '';
          for j := 1 to High(AF[i]) do
            if AF[i][j] <> '' then
              if TryStrToInt(AF[i][j], Mode) and (Mode > 0) and (Mode <= nModes) then begin
                Modes[Order] := Mode;
                Inc(Order)
              end else if (AF[i][j][1] = '<') and (OpElem = '') then begin
                { Convert potential term to pertirbation operator acting between electronic states }
                OpElem := Trim(AF[i][j]);
                PT.Bra := ElectronicState;
                PT.Ket := ElectronicState;
                p := Pos('|', OpElem);
                if p > 0 then begin
                  PT.Bra := FindState(Copy(OpElem,2,p-2));
                  PT.Ket := FindState(Copy(OpElem,p+1,Length(OpElem)-p-1));
                end;
              end else
                HandleError(i,j,'Bad atom number '+AF[i][j]);
          if OpElem <> '' then begin
            PT.Value := V;
            PT.Modes := Modes;
            SetLength(PT.Modes, Order); { Note forces copy }
            SetLength(ExtraTerms, Length(ExtraTerms)+1);
            ExtraTerms[High(ExtraTerms)] := PT;
          end else if Order = 2 then begin
            if F[Modes[0],Modes[1]] <> 0 then
              HandleError(i,-1,'Duplicate force constant:'+IntToStr(Modes[0])+','+IntToStr(Modes[1]));
            F[Modes[0],Modes[1]] := V;
            F[Modes[1],Modes[0]] := V;
          end else if Order = 1 then begin
            if not Assigned(Flinear) then begin
              Flinear := TMatrix.Create(1,nModes);
              Flinear.Zero;
            end;
            if Flinear[1, Modes[0]] <> 0 then
              HandleError(i,-1,'Duplicate linear force constant:'+IntToStr(Modes[0]));
            Flinear[1, Modes[0]] := V;
          end else
            WriteLn(Log,'Warning - only linear and quadratic modes implemented');
        end;
      end;
    if Verbose then begin
      F.Write(Log, Caption, ShowName, ShowName, Precision);
      if Assigned(Flinear) then begin
        WriteLn(Log, 'Linear potential terms');
        Flinear.Write(Log, Caption, nil, ShowName, Precision);
      end;
      if Length(ExtraTerms) > 0 then begin
        WriteLn(Log, 'Additional Potential terms');
        WriteLn(Log, 'Bra  Ket                         Modes');
        for i := 0 to High(ExtraTerms) do
          with ExtraTerms[i] do begin
            Write(Log, Bra.Name, ' ':5-Length(Bra.Name), Ket.Name, ' ':5-Length(Ket.Name), FloatToStr(Value):20);
            for j := 0 to High(Modes) do
              Write(Log, ' ', Modes[j]:3);
            WriteLn(Log);
          end;
      end;
    end;
  end;
end;

type

  { TFSorter }

  TFSorter = class(TShellSorter)
    FC: TMNodeWithFMatrix;
    ByVal: Boolean;
    function InOrder(I,J:Integer):Boolean; override;
  end;

{ TFSorter }

function TFSorter.InOrder(I, J: Integer): Boolean;
var
  nt, n1, n2: array [0..1] of Integer;
  symbolic: array [0..1] of Boolean;
  value: array [0..1] of Double;
  s: array [0..1] of string;
  procedure Load(n, k: Integer);
  begin
    nt[n] := Length(FC.AF[k]);
    n1[n] := -1;
    n2[n] := -1;
    if nt[n] > 0 then begin
      s[n] := FC.AF[k][0];
      symbolic[n] := not TryStrToFloat(s[n], value[n]);
      if nt[n] > 1 then begin
        if not TryStrToInt(FC.AF[k][1], n1[n]) then
          n1[n] := -1;
        if nt[n] > 2 then begin
          if not TryStrToInt(FC.AF[k][2], n2[n]) then
            n2[n] := -1;
        end;
      end;
    end;
    if nt[n] = 0 then
      nt[n] := High(Integer); { Blank at end }
  end;
begin
  { Diagonal ones first }
  Load(0,I);
  Load(1,J);
  if nt[0] <> nt[1] then
    Result := nt[0] < nt[1] { Quadratic first }
  else if (n1[0] = n2[0]) <> (n1[1] = n2[1]) then
    Result := n1[0] = n2[0] { Diagonal first }
  else if n1[0] = n2[0] then
    Result := n1[0] <= n2[0] { Diagonal in mode order }
  else if ByVal then begin
    if symbolic[0] <> symbolic[1] then
      Result := symbolic[0]
    else if symbolic[0] then
      Result := s[0] <= s[1]
    else
      Result := Abs(value[0]) >= Abs(Value[1]);
  end else begin
    { Sort by index }
    if n1[0] <> n1[1] then
      Result := n1[0] < n1[1]
    else
      Result := n2[0] <= n2[1]
  end;
  if not Result then
    SwapValues(FC.AF[I], FC.AF[J], SizeOf(FC.AF[I]));
end;

procedure TMNodeWithFMatrix.SortF(ByValue: Boolean);
begin
  TrimArray(AF);
  with TFSorter.Create do begin
    FC := Self;
    ByVal := ByValue;
    Sort(Length(AF));
    Free;
  end;
end;

initialization
  TInternalCoords.RegisterClass(TElectronicState);
  TCartesianDisplacements.RegisterClass(TElectronicState);
  TSymmetryCoords.RegisterClass(TElectronicState);
end.
