unit JPLCat;

{$I utils.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}

{ Read JPL line catalog files as defined in
  H.M. Pickett, R.L.Poynter, E.A. Cohen, M.L. Delitsky, J.C. Pearson and H.S.P. Muller
  J. Quant. Spectrosc. Radiat. Transfer 60, 883 (1998)
  See also http://spec.jpl.nasa.gov/
}

interface

uses Classes, Mixture, TextStrm, Obs, ExpPlotLoader, MNode, MathSup, QuantumNumber, StaticField;

type

  { TSPFITReader }

  TSPFITReader = class(TObservationsReader)
    DoubleABLevels: Boolean;
    nBlendErr: Integer;
    function HandleDirective(var Handled: Boolean):Boolean; override;
    constructor Create; override;
    constructor Create(AReader: TObservationsReader); override;
    destructor Destroy; override;
    function DoReadOneLine: Boolean; override;
  end;

  TJPLCatReader = class(TSPFITReader)
    LastBuffer: string;
    LastnObs: Integer;
    constructor Create; override;
    function DoReadOneLine: Boolean; override;
  end;

  TJPLCatLoader = class(TExperimentalPlotLoader)
  public
    function Load: Boolean; override;
  end;


function CheckJPLCatFile(Mixture: TMixture; S: TTextStream; var nTried: Integer; Tolerance: Double; var Failures: Integer; Offset:Double; SkipELowerCheck, DoubleABlevels: Boolean; QNOmask: Integer): Double;

implementation

uses SysUtils, UCompat, Graphics, MatrixVec, Math, Units, NextTok, DRintensity,
     HundA, OAsymT, SymT, LineStrm, Worker, SimpleQnos;

type
  TJPLCatLine = record
    FREQ, ERR, LGINT, ELO: Double;
    DR, GUP, TAG, Q, H, NQN: Integer;
    Qnos: TSimpleQnos;
  end;

function LoadLine(const Line: AnsiString; var JPLCatLine: TJPLCatLine; out LastErr: string):Boolean;
var
  p, i, NonBlankCount: Integer;
  function GetField(Width: Integer): AnsiString;
  begin
    Result := Copy(Line, p, Width);
    if Trim(Result) <> '' then
      Inc(NonBlankCount)
    else
      Result := '0'; { Blank fields are read as zero to avoid spurious crashes }
    Inc(p, Width);
  end;
  function NextQno: Integer;
  var
    s: string[2];
  begin
    s := GetField(2);
    if s[1] in ['A'..'Z'] then
      Result := (Ord(s[1]) - Ord('A') + 10)*10 + StrToInt(s[2])
    else if s[1] in ['a'..'z'] then
      Result := -(Ord(s[1]) - Ord('a') + 1)*10 - StrToInt(s[2])
    else
      Result := StrToInt(s);
  end;
begin
  p := 1;
  LastErr := '';
  NonBlankCount := 0;
  try
    with JPLCatLine, Qnos do begin
      Result := TryStrToFloatEx(GetField(13), FREQ) and (NonBlankCount <> 0);
      if not Result then
        Exit; { First item not recognized - probably comment that can be ignored }
      ERR   := StrToFloatEx(GetField(8));
      LGINT := StrToFloatEx(GetField(8));
      DR    := StrToInt(GetField(2));
      ELO   := StrToFloatEx(GetField(10));
      GUP   := StrToInt(GetField(3));
      TAG   := StrToInt(GetField(7));
      Q     := StrToInt(GetField(2));
      H     := StrToInt(GetField(1));
      NQN   := StrToInt(GetField(1));
      if (NQN <= 0) or (NQN > 10) then
        Exit;
      NonBlankCount := 0;
      for i := 0 to NQN-1 do
        QNP[i] := NextQno;
      for i := 1 to 10 - NQN do
        if Copy(Line, p, 2) = '  ' then
          Inc(p,2)
        else
          Break;
      for i := 0 to NQN-1 do
        QN[i] := NextQno;
      Result := NonBlankCount = 2*NQN;
    end;
  except
    on E: Exception do begin
      Result := False;
      LastErr := E.Message;
    end;
  end;
end;

function SameQuantumNumbers(Line1, Line2: string):Boolean;
begin
  { Can't just test for string equality as line frequencies can differ in last
    significant digit }
  Result := Copy(TrimRight(Line1),45,35) = Copy(TrimRight(Line2),45,35);
end;

procedure AdjustQuantumNumbers(var Line: TJPLCatLine);
var
  i, Mask: Integer;
begin
  { Double (and subtract 1 if necessary) the quantum numbers }
  with Line, Qnos do begin
    Mask := H and 7;
    for i := NQN-1 downto 0 do begin
      QNP[i] := QNP[i]*2 - (Mask and 1);
      QN[i] := QN[i]*2 - (Mask and 1);
      Mask := Mask shr 1;
    end;
  end;
end;

function CheckJPLCatFile(Mixture: TMixture; S: TTextStream; var nTried: Integer; Tolerance: Double; var Failures: Integer; Offset:Double; SkipELowerCheck, DoubleABlevels:Boolean; QNOmask: Integer): Double;
var
  Buffer, NextBuffer, Es: string;
  CurrentLine: Integer;
  CatLine:  TJPLCatLine;
  Line: TLine;
  Jp,     J:   TQuantumNumber;
  symp,   sym: Shortint;
  np,     n:   Smallint;
  StateL, StateU: TState;
  Failed: Boolean;
  Molecule: TMolecule;
  Simulation: TSimulation;
  TransitionMoments: TTransitionMoments;
  cstr: Double;
  NextLineLoaded: Boolean;
  MapLower, MapUpper: TIntegerArray;
  QNOs: array of TSimpleQnos;
  Problem: Exception;

  procedure Fail(Name: string);
  begin
    Inc(Failures);
    if Failures > 0 then begin
      WriteLn(Log, 'Line ', CurrentLine, ':', Name);
      WriteLn(Log, Buffer);
    end;
    Failed := True;
  end;
  procedure CheckNo(n1, n2, tol: Double; Name: string; AddToLargestError: Boolean = True);
  begin
    if isNanOrInf(n2) then
      Fail(Format('%s fails: %g Nan', [Name, n1]))
    else if Abs(n1-n2) > tol then
      Fail(Format('%s out of tolerance: %g %g %g %d %d %d %d %d %d', [Name, n1, n2, n1-n2, Jp, symp, np, J, sym, n]));
    if AddToLargestError and (Abs(n1-n2) > Result) then
      Result := Abs(n1-n2);
  end;
  function MakeDefaultMap(Manifold: TManifold): TIntegerArray;
  var
    i, nE, n: Integer;
  begin
    Result := nil;
    nE := 0; { Count number of degenerate states }
    for i := 0 to Manifold.nStates - 1 do
      if Manifold.Molecule.SymName((Manifold.State[i] as TSymmetricTop).Symmetry)[1] = 'E' then
        Inc(nE);
    if nE = 0 then
      Exit;
    SetLength(Result,Manifold.nStates+nE+1);
    n := 1;
    for i := 0 to Manifold.nStates - 1 do begin
      Result[n] := i;
      Inc(n);
      if Manifold.Molecule.SymName((Manifold.State[i] as TSymmetricTop).Symmetry)[1] = 'E' then begin
        Result[n] := 1000+i;
        Inc(n);
      end;
    end;
  end;

  procedure ReadMap(var Map: TIntegerArray);
  var
    s: string;
    n: Integer;
  begin
    s := Buffer;
    NextToken(s); { Discard Map }
    n := StrToInt(NextToken(s));
    AddToMap(Map, n, NextToken(s));
  end;

  procedure MakeCATQnoList(LineList: TLineList; nV: Integer);
  var
    i: Integer;
    UpperStateMap, LowerStateMap: TStateMap;
    procedure MakeQNO(var Q: T10Integers; Eigenvalues: TEigenvalues; Index: Integer; StateMap: TStateMap);
    var
      s: string;
      Manifold: TManifold;
      BasisState: TBasisState;
      State: TState;
      i, p, p0, nQ: Integer;
      Qno: TQuantumNumber;
      function NextInteger: Boolean;
      begin
        Result := False;
        while (p <= Length(s)) and (s[p] = ' ') do
          Inc(p);
        Qno := 0;
        p0 := p;
        while (p <= Length(s)) and (s[p] <> ' ') do
          Inc(p);
        if p = p0 then
          Exit;
        if not TryStrToQuantumNumber(Copy(s,p0,p-p0),Qno, Mixture.QuantumNumberFormat) then
          raise Exception.Create('Invalid QNO'+s);
        Q[nQ] := Qno;
        Inc(nQ);
        Result := True;
      end;
    begin
      FillChar(Q, SizeOf(Q), 0);
      BasisState := Eigenvalues.DominantBasis[Index];
      State := BasisState.State;
      Manifold := State.Manifold;
      s := BasisState.StateLabel(qftJ);
      Molecule.FixSimpleFormat(s);
      if Molecule.nManifolds > 1 then
        Delete(s, 1, Length(Manifold.Name)+1);
      Delete(s, 1, Length(State.Name)+1);
      p := 1;
      nQ := 0;
      while NextInteger do
        if nQ = nV-1 then begin
          for i := 0 to High(StateMap) do
            if StateMap[i] = State then begin
              Q[nQ] := 2*i;
              Break;
            end;
          Inc(nQ);
        end;
    end;
    procedure MakeStateMap(var StateMap: TStateMap; const Map: TIntegerArray);
    var
      i: Integer;
      Manifold: TManifold;
    begin
      SetLength(StateMap, Length(Map));
      for i := 0 to High(Map) do begin
        Manifold := Molecule.Manifold[Map[i] div 2000-1];
        StateMap[i] := Manifold.State[Map[i] mod 1000];
      end;
    end;
  begin
    SetLength(QNOs, LineList.nLines);
    MakeStateMap(UpperStateMap, MapUpper);
    MakeStateMap(LowerStateMap, MapLower);
    for i := 0 to LineList.nLines-1 do begin
      MakeQno(QNOs[i].QNP, LineList.Line[i].EigenvaluesU, LineList.Line[i].IndexU, UpperStateMap);
      MakeQno(QNOs[i].QN,  LineList.Line[i].EigenvaluesL, LineList.Line[i].IndexL, LowerStateMap);
    end;
  end;
  function FindCatLine(LineList: TLineList): Boolean;
  var
    i, j, mask: Integer;
    Line: TLine;
    cstr: Double;
  begin
    for i := 0 to LineList.nLines-1 do begin
      mask := QNOmask;
      Result := True;
      for j := 0 to CatLine.NQN-1 do begin
        if (mask and 1) = 0 then begin
          Result := Result and (QNOs[i].QNP[j] = CatLine.Qnos.QNP[j]) and (QNOs[i].QN[j] = CatLine.Qnos.QN[j]);
          if not Result then
            Break;
        end;
        mask := mask shr 1;
      end;
      if not Result then Continue;
      Line := LineList.Line[i];
      Result := Result and (Abs(CatLine.ELO - Line.EigenvaluesL.E[Line.IndexL]*ConversionFactor(Mixture.Units,cm1)+Offset) < 1.5e-4);
      if CatLine.TAG >= 0 then
        Result := Result and (Abs(CatLine.FREQ - Line.Frequency*ConversionFactor(Mixture.Units,MHz)) < 1.5e-4)
      else
        Result := Result and (Abs(CatLine.FREQ - Line.Frequency*ConversionFactor(Mixture.Units,MHz)) < 3*CatLine.ERR);
      if Result then begin
        cstr := Simulation.LineIntensity(Line, Mixture.Units, Problem);
        if Assigned(Problem) then begin
          Fail(Problem.Message);
          Problem.Free;
        end;
        Result := Result and (Abs(CatLine.LGINT - log10(cstr)) < 1.5e-4);
        if not Result then
          WriteLn(Log, i:3, ' ', CatLine.LGINT - log10(cstr), ' ', Line.Strength);
      end;
      if Result and (Mixture.Correlations.nEntries > 0) and (CatLine.ERR < 999.9999) then
        CheckNo(CatLine.ERR, EstErr(Line.EigenvaluesU, Line.EigenvaluesL, Line.IndexU, Line.IndexL), Tolerance, 'Est Err', True);
      if Result then
        Break;
    end;
    //if not Result then begin
    //  for i := 0 to LineList.nLines-1 do begin
    //    Line := LineList.Line[i];
    //    if Abs(CatLine.FREQ - Line.Frequency*ConversionFactor(Mixture.Units,MHz)) < 1.5e-4 then begin
    //      Write(Log, i:3);
    //      for j := 0 to CatLine.NQN-1 do
    //        Write(Log, QNOs[i].QNP[j]:2, CatLine.Qnos.QNP[j]:2);
    //      Write(Log, ' - ');
    //      for j := 0 to CatLine.NQN-1 do
    //        Write(Log, QNOs[i].QN[j]:2, CatLine.Qnos.QN[j]:2);
    //    end;
    //  end;
    //end;
  end;

var
  ManifoldU, ManifoldL: TManifold;
  MyS: TStream;
  NotKZero, CATQnoListMade: Boolean;
  i, nV, Qcalc: Integer;
  LastErr: string;
begin
  Result := 0;
  ClearnnTableCache;
  CATQnoListMade := False;
  Molecule := Mixture.Species[0].Isotopomer[0];
  { Clear AutoQConverge, as SPCAT does not use this mechanism }
  for i := 0 to Molecule.nManifolds-1 do
    Molecule.Manifold[i].AutoQConverge := False;
  StateL := Molecule.Manifold[0].State[0];
  StateU := StateL;
  if Molecule.nManifolds > 1 then
    StateU := Molecule.Manifold[1].State[0];
  if Molecule is TSymmetricMolecule then begin
    MapUpper := MakeDefaultMap(StateU.Manifold);
    MapLower := MakeDefaultMap(StateL.Manifold);
  end else
    NameIsStateNumber(Molecule,MapUpper,MapLower);

  Simulation := Mixture.Simulations[0];
  Problem := nil;
  if QNOmask <> 0 then begin
    Simulation.MakeLineList(Problem);
    if Assigned(Problem) then
      raise Problem;
  end;
  MyS := nil;
  NextLineLoaded := False;
  while NextLineLoaded or not S.AtEndOfFile do begin
    if NextLineLoaded then begin
      Buffer := NextBuffer;
      Inc(CurrentLine);
    end else begin
      Buffer := S.RestOfLine;
      CurrentLine := S.Line;
      S.NextLine;
    end;
    if Trim(Buffer) = 'stop' then
      Exit;
    if not S.AtEndOfFile then begin
      NextBuffer := S.RestOfLine;
      S.NextLine;
      NextLineLoaded := True;
    end else
      NextLineLoaded := False;
    if Trim(Buffer) = '' then
      { Skip blank lines }
    else if LoadLine(buffer, CatLine, LastErr) then begin
      if NextLineLoaded and SameQuantumNumbers(Buffer,NextBuffer) then begin
        { Work around spcat oddity - selected transitions come out as two half intensity transitions }
        NextLineLoaded := False;
        CatLine.LGINT := CatLine.LGINT + log10(2.0);
      end;
      Inc(nTried);
      AdjustQuantumNumbers(CatLine);
      Line.CachedName := Buffer;
      if Molecule.nManifolds > 2 then begin
        StateL := nil;
        StateU := nil;
      end;
      ManifoldU := nil;
      ManifoldL := nil;
      NotKZero := (CatLine.Qnos.QN[1] <> 0) and (CatLine.Qnos.QNP[1] <> 0);  { Symmetric tops only }
      if QNOmask <> 0 then begin
        if not CATQnoListMade then begin
          Qcalc := ComputeQ(nil, nil, Molecule, Mixture, CatLine.NQN, 0, -1, True, nV);
          CheckNo(Qcalc, CatLine.Q, 0, 'Quantum number format, Q');
          MakeCATQnoList(Mixture.LineList, nV);
        end;
        CATQnoListMade := True;
        if not FindCatLine(Mixture.LineList) then
          Fail('Can''t find line');
      end else begin
        Failed := ConvertQuantumNumbers(ManifoldU, StateU, Jp, symp, np, ManifoldL, StateL, J, sym, n, Line.CachedName, Molecule,
                                        CatLine.Q, CatLine.NQN, CatLine.Qnos, Simulation.ExField, [], MapUpper, MapLower,
                                        Mixture.QuantumNumberFormat, 0, 2, -1, False);
        if Failed then
          Fail('Quantum number assignment')
        else begin
          if Length(MapLower) > 0 then
            n := 0;
          if Length(MapUpper) > 0 then
            np := 0;
          Line.Width := 0;
          Line.Width := 0;
          Line.EigenvaluesU := StateU.Manifold.Eigenvalues(Jp, symp, [meHrot, meWidth, meFloatedDerivatives { For checking predicted error in file}], ZeroField);
          if np <= 0 then
            Line.IndexU := FindIndex(Line.EigenvaluesU, Line.CachedName, False, CurrentLine, 'JPL .cat file', Mixture.QuantumNumberFormat, True)
          else
            Line.IndexU := np;
          if (Line.IndexU > Line.EigenvaluesU.Size) or (Line.IndexU <= 0) then
            Fail(Format('Upper eigenvalue (%d) > %d', [np, Line.EigenvaluesU.Size]));

          Line.EigenvaluesL := StateL.Manifold.Eigenvalues(J, sym, [meHrot, meWidth, meFloatedDerivatives], ZeroField);
          if n <= 0 then
            Line.IndexL := FindIndex(Line.EigenvaluesL, Line.CachedName, True, CurrentLine, 'JPL .cat file', Mixture.QuantumNumberFormat, True)
          else
            Line.IndexL := n;
          if (Line.IndexL > Line.EigenvaluesL.Size) or (Line.IndexL <= 0) then
            Fail(Format('Lower eigenvalue (%d) > %d', [n, Line.EigenvaluesL.Size]))
          else if not SkipELowerCheck then
            CheckNo(CatLine.ELO, Line.EigenvaluesL.E[Line.IndexL]*ConversionFactor(Mixture.Units,cm1)+Offset, Tolerance, 'E"');

          if not failed then begin
            Line.Frequency := Line.EigenvaluesU.E[Line.IndexU]-Line.EigenvaluesL.E[Line.IndexL];
            if CatLine.TAG >= 0 then
              CheckNo(CatLine.FREQ, Line.Frequency*ConversionFactor(Mixture.Units,MHz), Tolerance, 'Frequency')
            else
              CheckNo(CatLine.FREQ, Line.Frequency*ConversionFactor(Mixture.Units,MHz), 3*CatLine.ERR, 'Frequency', False);
            if (Mixture.Correlations.nEntries > 0) and (CatLine.ERR < 999.9999) then
              CheckNo(CatLine.ERR, EstErr(Line.EigenvaluesU, Line.EigenvaluesL, Line.IndexU, Line.IndexL), Tolerance, 'Est Err', True);
            TransitionMoments := Molecule.FindTransitionMoments(StateU.Manifold, StateL.Manifold);
            if not Assigned(TransitionMoments) then
              Fail('Transition moment not found for '+StateU.UniqueName+' - '+StateL.UniqueName)
            else begin
              TransitionMoments.MakeLine(Line,Line.EigenvaluesL,Line.EigenvaluesU,Line.IndexL,Line.IndexU,Simulation,True);
              Problem := nil;
              cstr := Simulation.LineIntensity(Line, Mixture.Units, Problem);
              if Assigned(Problem) then begin
                Fail(Problem.Message);
                Problem.Free;
              end;
              if DoubleABlevels and ((Molecule as TSymmetricMolecule).CharacterTable.IReps[sym].g = 0) and NotKZero then
                cstr := cstr*2; { A/B rotational levels occur in pairs, and CALPGM (sometimes) only calculates one of the pair }
              if cstr <> 0 then begin
                Es := '(Ratio = ' + FloatToStr(Power(10,CatLine.LGINT)/cstr) + ') ';
                CheckNo(CatLine.LGINT, log10(cstr), Tolerance, Es+'Intensity');
              end else begin
                CheckNo(CatLine.LGINT, -1000, Tolerance, '(Calculated = 0) Intensity');
              end;
              // if Power(10,CatLine.LGINT)/cstr > 1.1 then
                // WriteLn(Log,Line.EigenvaluesU.E[Line.IndexU], ' ', Line.EigenvaluesL.E[Line.IndexL]);
              // WriteLn(Log,Line.EigenvaluesL.J, Line.EigenvaluesL.E[Line.IndexL], ' ', Power(10,CatLine.LGINT)/cstr);
              // WriteLn(Log,Line.Strength*1e12:8:4, ' ', Mixture.LineIntensity(Line, Mixture.Temperature, Mixture.Tvib)*1e12:8:4);
            end;
          end;
        end;
      end;
    end else if Pos('#',Buffer) > 0 then
      { Skip line }
    else if (Pos('OverrideWeight',Buffer) = 1) or (Pos('Constrain',Buffer) = 1) or (Pos('IndexOffsets',Buffer) = 1) then
    { Skip line }
    else if (Buffer = 'TakeAll') or (Buffer = 'Intensity') or (Buffer = 'TakeCalc') then
    { Skip line }
    else if Pos('MapUpper ',Buffer) > 0 then
      ReadMap(MapUpper)
    else if Pos('MapLower ',Buffer) > 0 then
      ReadMap(MapLower)
    else if Pos('DoubleABlevels ',Buffer) > 0 then
      DoubleABlevels := True
    else if Pos('include ',Buffer) > 0 then begin
      NextToken(Buffer);
      FreeAndNil(MyS);
      S := TTextStream.Create(NextToken(Buffer), True);
      MyS := S; { Flag locally create dstream }
      NextLineLoaded := False;
    end else if Buffer <> '' then
      Fail('Bad line:'+LastErr);
  end;
      FreeAndNil(MyS);
end;

{ TJPLCatReader }

constructor TJPLCatReader.Create;
begin
  inherited;
  LastBuffer := '';
  LastnObs := -1;
end;

function TJPLCatReader.DoReadOneLine: Boolean;
var
  CatLine: TJPLCatLine;
  Buffer, LastErr: string;
  AStateL, AStateU: TState;
  NotKZero: Boolean;
begin
  Result := False;
  Buffer := StringOfChar(' ',Max(f.Column-Length(f.Token),0)) + f.Token;
  Buffer := Buffer + f.RestOfLine;
  AStateL := DefaultStateL;
  AStateU := DefaultStateU;
  if LoadLine(buffer, CatLine, LastErr) then begin
    if SameQuantumNumbers(Buffer,LastBuffer) and (LastnObs+1 = nObs) then begin
      { Merge two identical observations }
      if ObsType = obIntensity then
        Observations[LastnObs].Frequency := 2*Observations[LastnObs].Frequency
    end else begin
      AdjustQuantumNumbers(CatLine);
      case ObsType of
        obLine, obCDLower, obCDUpper:
                     Obs.Frequency := CatLine.FREQ;
        obIntensity: Obs.Frequency := Power(10,CatLine.LGINT);
        obELower:    Obs.Frequency := Convert(CatLine.ELO,cm1,MHz);
        obEUpper:    Obs.Frequency := Convert(CatLine.ELO,cm1,MHz)+CatLine.FREQ;
      end;
      Obs.Weight := Abs(CatLine.ERR);
      Obs.LineNumber := f.Line;
      Obs.FileName := FileName;
      Obs.Name := Buffer;
      NotKZero := (CatLine.Qnos.QN[1] <> 0) and (CatLine.Qnos.QNP[1] <> 0); { Symmetric tops only }
      if ConvertQuantumNumbers(Obs.ManifoldU, AStateU, Obs.JU, Obs.SymU, Obs.IndexU, Obs.ManifoldL, AStateL, Obs.JL, Obs.SymL, Obs.IndexL, Obs.Name, DefaultMolecule,
                               CatLine.Q, CatLine.NQN, CatLine.Qnos, ExField, StateMap, UpperMap, LowerMap,
                               QuantumNumberFormat, RemoveSpins, BranchRank, DefaultSymmetry, False) then
        ReportError('Bad Quantum numbers')
      else
        case Take of
          tkObs:  Result := CatLine.TAG < 0;
          tkCalc: Result := CatLine.TAG >= 0;
          tkAll:  Result := True;
        end;
      if Result and NotKZero and (ObsType = obIntensity) and DoubleABLevels and (Obs.ManifoldU is TSymmetricManifold)
         and ((Obs.ManifoldU.Molecule as TSymmetricMolecule).CharacterTable.IReps[Obs.SymL].g = 0) then
        Obs.Frequency { Actually intensity } := Obs.Frequency / 2; { Line in file is sum of both lines of a degenerate pair }
    end;
    LastnObs := nObs;
    if Result then
      LastBuffer := Buffer
    else
      LastBuffer := '';
  end else if Trim(buffer) <> '' then
    ReportError('Line not recognized:'+LastErr+':'+buffer);
end;

{ TJPLCatLoader }

function TJPLCatLoader.Load: Boolean;
var
  CatLine: TJPLCatLine;
  Buffer, LastBuffer: AnsiString;
  S: TLineStream;
  nPoints: Integer;
  LastIntensity: Double;
  LengthSet: Integer;
  LastErr: string;
begin
  Result := False;
  if Ext <> '.cat' then
    Exit;
  nPoints := 0;
  LengthSet := 0;
  LastBuffer := '';
  S := TLineStream.Create(Stream);
  try
    while S.NextLine(Buffer) do
      if LoadLine(Buffer,CatLine,LastErr) then
        if SameQuantumNumbers(Buffer,LastBuffer) then begin
          { Merge adjacent lines }
          SetValue(-1,1,nPoints,2*LastIntensity);
          LastBuffer := '';
        end else begin
          if LengthSet <= nPoints then begin
            LengthSet := nPoints+10000;
            Connector.SetLength(0,LengthSet,False,True,True);
          end;
          SetValue(-1,0,nPoints,CatLine.FREQ);
          LastIntensity := Power(10,CatLine.LGINT);
          SetValue(-1,1,nPoints,LastIntensity);
          Inc(nPoints);
          LastBuffer := Buffer;
        end
      else if LastErr <> '' then begin
        LineNumber := S.LastLine;
        FileErr(LastErr);
      end;
  finally
    S.Free;
  end;
  Dec(nPoints);
  Result := nPoints > 0;
  if Result then begin
    Connector.SetLineList(0,True);
    Connector.SetLength(0,nPoints,False,True,True);
    Connector.SetUnits(0,Mhz);
    Connector.SetIntensityUnits(nm2MHzperMolecule);
    Connector.SetScale(1);
  end;
end;

{ TSPFITReader }

constructor TSPFITReader.Create;
begin
  inherited;
  Units := MHz;
  ClearnnTableCache;
end;

constructor TSPFITReader.Create(AReader: TObservationsReader);
begin
  inherited Create(AReader);
  while AReader <> nil do begin
    if AReader is TSPFITReader then begin
      DoubleABLevels := (AReader as TSPFITReader).DoubleABLevels;
      Break;
    end;
    AReader := AReader.IncludedFrom;
  end;
end;

destructor TSPFITReader.Destroy;
begin
  if nBlendErr > 1 then
    WriteLn(Log, nBlendErr, ' Blends detected');
  inherited;
end;

function TSPFITReader.DoReadOneLine: Boolean;
var
  CatLine: TJPLCatLine;
  Buffer, tok: string;
  AStateL, AStateU: TState;
  p, i, Sum, nV: Integer;
  qns: array [0..11] of Integer;
  BlendWt: Double;
  Molecule: TMolecule;
  ErrSeen: Boolean;
  function NextQno:Integer;
  begin
    tok := Trim(Copy(buffer,p,3));
    if tok <> '' then begin
      Inc(CatLine.NQN);
      if not TryStrToInt(tok, Result) then begin
        if not ErrSeen then
          ReportError('Bad integer:'+tok);
        ErrSeen := True;
      end;
    end else
      Result := 0;
    Inc(p,3);
  end;
  function NextField: string;
  var
    s: Integer;
  begin
    while (p <= Length(Buffer)) and (Buffer[p] <= ' ') do
      Inc(p);
    s := p;
    while (p <= Length(Buffer)) and (Buffer[p] > ' ') do
      Inc(p);
    if (p > s) and (Buffer[s] <> '/') then
      Result := Copy(Buffer,s,p-s)
    else
      Result := '';
  end;
begin
  { Set defaults }
  AStateL := DefaultStateL;
  AStateU := DefaultStateU;
  p := 1;
  ObsType := obLine;
  ErrSeen := False;
  if GlobalNQN > 0 then begin
    { Free format }
    with CatLine, Qnos do begin
      NQN := GlobalNQN;
      QNP[0] := f.IntegerToken;
      for i := 1 to NQN-1 do
        QNP[i] := f.NextInteger(0);
      for i := 0 to NQN-1 do
        QN[i] := f.NextInteger(0);
    end;
    Result := f.GetFreeItem;
    if Result then begin
      Obs.Frequency := f.RealToken;
      if f.GetFreeItem then begin
        Obs.Weight := f.RealToken;
        if Obs.Weight < 0 then begin
          Obs.Weight := Convert(-Obs.Weight,cm1,Units);
          Obs.Frequency := Convert(Obs.Frequency,cm1,Units);
        end;
        if f.GetFreeItem and f.IsRealToken(BlendWt) then begin
          if nBlendErr = 0 then
            WriteLn(Log, 'Warning - Blend fitting not implemented (first line ',f.Line,' of ',f.StreamName,')');
          Inc(nBlendErr);
        end;
      end else
        Obs.Weight := 1;
    end;
  end else begin
    { Fixed format }
    { Reconstitute input line }
    Buffer := StringOfChar(' ',Max(f.Column-Length(f.Token),0)) + f.Token;
    Buffer := Buffer + f.RestOfLine;
    { Read quantum numbers }
    Result := Length(Buffer) > 36;
    if not Result then
      Exit;
    with CatLine, Qnos do begin
      NQN := 0;
      for i := 0 to 11 do
        qns[i] := NextQno;
      Result := not ErrSeen;
      if not Result then
        Exit;
      if GlobalNQN < 0 then
        NQN := -GlobalNQN
      else
        NQN := NQN div 2;
      for i := 0 to NQN-1 do
        QNP[i] := qns[i];
      for i := 0 to NQN-1 do
        QN[i] := qns[NQN+i];
      tok := NextField;
      if tok <> '' then begin
        Result := TryStrToFloatEx(tok, Obs.Frequency);
        if not Result then begin
          ReportError('Bad frequency:'+tok);
          Exit;
        end;
        tok := NextField;
        if tok <> '' then begin
          Obs.Weight := StrToFloatEx(tok);
          if Obs.Weight < 0 then begin
            Obs.Weight := Convert(-Obs.Weight,cm1,Units);
            Obs.Frequency := Convert(Obs.Frequency,cm1,Units);
          end;
          tok := NextField;
          if (tok <> '') and TryStrToFloatEx(tok,BlendWt) then begin
            if nBlendErr = 0 then
              WriteLn(Log, 'Warning - Blend fitting not implemented (first line ',f.Line,' of ',f.StreamName,')');
            Inc(nBlendErr);
          end;
        end else
          Obs.Weight := 1;
      end else
        Result := False;
    end;
  end;
  with CatLine do begin
    if (AStateL = nil) or (AStateU = nil) then
      Molecule := DefaultMolecule
    else
      Molecule := AStateL.Molecule;
    if Qfmt <= 0 then
      Q := ComputeQ(AStateU, AStateL, Molecule, Mixture, NQN, RemoveSpins, -1, True, nV)
    else begin
      nV := nVFromQ(Qfmt);
      Q := Qfmt;
    end;
    if (AStateL = nil) or (AStateU = nil) then
      with CatLine.Qnos do
        if (nv > 0) and (QNP[nV-1] >= 0) and (QNP[nV-1] < Length(StateMap)) and (StateMap[QNP[nV-1]] <> nil) then
          Molecule := StateMap[QNP[nV-1]].Molecule
        else begin
          ReportError('MapState needed for state ');
          Exit;
        end;
    with CatLine do
      if (Qnos.QNP[NQN-1] < 0) or (Qnos.QN[NQN-1] < 0) then begin
        Inc(Obs.RemoveSpins);
        Dec(NQN);
        i := Molecule.nNuclei-Obs.RemoveSpins;
        if (i > 0) and Molecule.Nuclei[i-1].AsNext then begin
          Inc(Obs.RemoveSpins);
          Dec(NQN);
        end;
      end;
    H := 0; { Flag non-integer spins }
    if (AStateL = nil) or (AStateU = nil) then begin
      { Check for excess quantum numbers }
      i := nV;
      if not Molecule.AllSinglet then
        Inc(i);
      if i + Molecule.nNuclei-Obs.RemoveSpins < CatLine.NQN then
        CatLine.NQN := i + Molecule.nNuclei-Obs.RemoveSpins;
      if Molecule.OddElectrons then
        H := 1;
    end else if Odd(AstateL.S) then
      H := 1;
    Sum := H; { Track if current spin total is half integer }
    for i := 0 to Molecule.nNuclei - 1 do begin
      H := H * 2;
      if Odd(Molecule.Nuclei[i].Spin) then
        Sum := 1-Sum;
      if not Molecule.Nuclei[i].AsNext then
        H := H + Sum; { Itot must be integer }
    end;
    H := H shr Obs.RemoveSpins;
  end;
  if Result then begin
    AdjustQuantumNumbers(CatLine);
    with CatLine.Qnos do
      if QN[0] < 0 then begin
        for i := 0 to High(QN) do
          QN[i] := -1; { Flag energy level fit }
        Obs.ObsType := obEUpper;
      end;
    Obs.LineNumber := f.Line;
    Obs.FileName := FileName;
    Obs.Name := Buffer;
    Result := not ConvertQuantumNumbers(Obs.ManifoldU, AStateU, Obs.JU, Obs.SymU, Obs.IndexU, Obs.ManifoldL, AStateL, Obs.JL, Obs.SymL, Obs.IndexL, Obs.Name, nil,
                                        CatLine.Q, CatLine.NQN, CatLine.Qnos, ExField, StateMap, UpperMap, LowerMap,
                                        QuantumNumberFormat, Obs.RemoveSpins, BranchRank, DefaultSymmetry, False);
    if not Result then
      ReportError('Bad Quantum numbers');
  end;
end;

function TSPFITReader.HandleDirective(var Handled: Boolean): Boolean;
begin
  Result := inherited HandleDirective(Handled);
  if Handled then
    Exit;
  if LowerCase(s) = 'doubleablevels' then begin
    if f.GetFreeItem then
      DoubleABLevels := f.BooleanToken
    else
      DoubleABLevels := True;
  end;
end;

initialization
  RegisterPlotLoader(TJPLCatLoader, 'JPL Line lists', '*.cat;*.mrg');
end.

