unit LevelLister;
{ Helper functions for level lists }
{$I utils.inc}{Copyright (C) 2010 C M Western; see commandline.pas for license}
interface

uses SysUtils, Math, QuantumNumber, MNode, Mixture;

type

  TLevelLister = class
  public
    ShowMoleculeLabel: Boolean;
    ShowManifoldLabel: Boolean;
    ShowPartitionAccum: Boolean;
    Emin, Emax, Q, QSum, QSump, Qscale: Double;
    nLevels, nPrinted: Integer;
    LastMolecule: TMolecule;
    LogHeaderShown: Boolean;
    ShowExternalField: Boolean;
    ShowNumericalPopulations: Boolean;
    QTable: Boolean;
    ESweep: Boolean;
    FieldSummary: Boolean;
    HasWidths: Boolean;
    Mixture: TMixture;
    BaseNode: TMNode;
    Simulation: TSimulation;
    Overflow: Exception;
    LabelLength: Integer;
    LevelPlotter: TObject; { Should be TLevelLister, but avoid circular reference }
    LastInfo: string;
    constructor Create(ALevelPlotter: TObject; ABaseNode: TMNode; AShowExternalField, AESweep, AFieldSummary: Boolean; AEmax: Double);
    procedure ShowLevel(Ev: TEigenvalues; i:Integer);
    procedure ShowManifold(Manifold: TManifold; State: TState = nil; IncludePerturbations: Boolean = True; Symmetry: Integer = -1; XMin: TQuantumNumber = -1; XMax: TQuantumNumber = -1; RemoveSpins: Byte = 0);
    procedure CheckHeaderShown(ALabelLength: Integer; ALabel:string);
    procedure PrepareExtraInfo(Ev: TEigenValues); virtual;
    procedure ShowExtraHeader; virtual;
    procedure ShowExtraInfo(Ev: TEigenvalues; i:Integer); virtual;
    procedure ShowTotals;
    procedure ShowErrors;
  end;
  TLevelListType = (lltBySymmetry, lltByEnergy, lltQTable);

procedure ShowLevelList(BaseNode: TMNode; LevelListType: TLevelListType = lltBySymmetry; Emax: Double = MaxDouble; Tstep: Double = 100);

implementation

uses MathSup, UCompat, Worker, Units, StaticField, LevelPlotter, Profile,
     PartFunc, Progress, Global;

var
  FindNextLargestTime: string = 'Find Next Largest Level';
  ShowLevelTime: string = 'ShowLevelTime';

procedure ShowLevelList(BaseNode: TMNode; LevelListType: TLevelListType; Emax: Double; Tstep: Double);
var
  LevelLister: TLevelLister;
  LastEv: TEigenvalues;
  LastIndex, nEvs: Integer;
  Evs: array of TEigenvalues;
  LastIndices: array of Integer;
  function CompareOther(Ev1, Ev2: TEigenvalues):Boolean;
  begin
    if Ev1.Manifold <> Ev2.Manifold then
      Result := CompareStr(Ev1.Manifold.FullHeirarchicName, Ev2.Manifold.FullHeirarchicName) > 0
    else if Ev1.Sym <> Ev2.Sym then
      Result := Ev1.Sym > Ev2.Sym
    else
      Result := Ev1.J > Ev2.J;
  end;

  procedure LoadManifolds(MNode: TMNode);
  var
    i, Sym: Integer;
    J, XMax: TQuantumNumber;
    Manifold: TManifold;
    Ev: TEigenvalues;
  begin
    if MNode is TManifold then begin
      Manifold := MNode as TManifold;
      Manifold.Molecule.PartitionFunction(LevelLister.Simulation,LevelLister.Overflow,LevelLister.Emin); // Force precalculation
      XMax := Manifold.FindJmax(-1);
      if LevelLister.Simulation.HasFields then
        J := -XMax
      else
        J := Manifold.FindJmin(-1);
      while J <= XMax do begin
        for Sym := 0 to Manifold.nSym-1 do begin
          Ev := Manifold.Eigenvalues(J, Sym, [meHrot, meWidth, meAllDerivatives, meFloatedDerivatives], LevelLister.Simulation.ExField, True, plNone);
          Inc(LevelLister.nLevels, Ev.Size);
          if Ev.Size > 0 then begin
            if Length(Evs) <= nEvs then
              SetLength(Evs,(Length(Evs)+100)*2);
            Evs[nEvs] := Ev;
            Inc(nEvs);
          end;
          if LevelLister.Simulation.HasFields then
            Break; { Symmetry not used for external fields }
        end;
        Inc(J,2);
      end;
    end else if MNode is TState then
      LoadManifolds(MNode.Owner as TMNode)
    else if not(MNode is TPartitionFunction) then
      for i := 0 to MNode.ComponentCount - 1 do
        LoadManifolds(MNode.Components[i] as TMNode);
    SetLength(LastIndices, nEvs);
  end;

  procedure FindNextLargest(State: TState);
  var
    i, j, CurrentIndex: Integer;
    Ev: TEigenvalues;
    LastE: Double;
    Take: Boolean;
    CurrentEv: TEigenvalues;
  begin
    repeat
      CurrentEv := nil;
      if Assigned(LastEv) then
        LastE := LastEv.Energy[LastIndex];
      for j := 0 to nEvs - 1 do begin
        Ev := Evs[j];
        if Ev.Size <= 0 then
          Continue;
        if LastEv = Ev then begin
          LastIndices[j] := LastIndex;
          i := LastIndex+1;
        end else if Assigned(LastEv) then begin
          i := LastIndices[j]+1;
          while (i <= Ev.Size) and (Ev.Energy[i] < LastE) do
            Inc(i);
          while (i <= Ev.Size) and (Ev.Energy[i] = LastE) and not CompareOther(Ev, LastEv) do
            Inc(i);
        end else begin
          LastIndices[j] := 0;
          i := 1;
        end;
        if i <= Ev.Size then begin
          Take := True;
          if Assigned(CurrentEv) then
            if CurrentEv.Energy[CurrentIndex] = Ev.Energy[i] then begin
              Take := CompareOther(CurrentEv,Ev)
            end else
              Take := CurrentEv.Energy[CurrentIndex] > Ev.Energy[i];
          if Take then begin
            CurrentEv := Ev;
            CurrentIndex := i;
          end;
        end;
      end;
      LastEv := CurrentEv;
      LastIndex := CurrentIndex;
    until (LastEv = nil) or (State = nil) or (CurrentEv.DominantBasis[LastIndex].State = State);
  end;

  procedure ShowMNode(MNode: TMNode);
  var
    i: Integer;
  begin
    if MNode is TManifold then
      LevelLister.ShowManifold(MNode as TManifold)
    else if MNode is TState then
      LevelLister.ShowManifold(MNode.Owner as TManifold, MNode as TState)
    else
      for i := 0 to MNode.ComponentCount - 1 do
        ShowMNode(MNode.Components[i] as TMNode);
  end;

  procedure ShowSorted(Qtable: Boolean);
  var
    SelState: TState;
  begin
    if BaseNode is TState then
      SelState := BaseNode as TState
    else
      SelState := nil;
    LastEv := nil;
    LastIndex := 0;
    LevelLister.nPrinted := 0;
    LevelLister.LastMolecule := nil;
    repeat
      TStart(FindNextLargestTime);
      FindNextLargest(SelState);
      TEnd(FindNextLargestTime);
      if not Assigned(LastEv) then
        Break;
      if not Qtable then
        LevelLister.CheckHeaderShown(0,LastEv.StateLabelKey(1));
      Tstart(ShowLevelTime);
      LevelLister.ShowLevel(LastEv, LastIndex);
      Tend(ShowLevelTime);
    until LastEv = nil;
    LevelLister.ShowTotals;
  end;

  procedure ShowQtable;
  var
    Tmax, T: Double;
    QMnode: TInterpolatedPartitionFunction;
    nPoints: Integer;
    s: string;
    E: Exception;
  begin
    Tmax := LevelLister.Simulation.Temperature;
    QMnode := TInterpolatedPartitionFunction.Create(nil);
    try
      PushProgress('Partition Function', Round(Tmax));
      WriteLn(Log,'Partition function for ',BaseNode.UniqueName);
      WriteLn(Log,'       T/K Qsum       Qsum/Q     Qthresh/Q   Qthresh');
      LevelLister.QTable := True;
      if Tstep < 0 then begin
        nPoints := Round(-TStep+4);
        T := Power(TMax,-1/Tstep);
        Tstep := -T;
      end else begin
        nPoints := Round(Tmax/Tstep + 100); // Allow room
        T := Tstep/100; { Initial value near to zero }
      end;
      QMnode.Name:= 'Intepolated Partition Function';
      QMnode.Comment := 'Generated from '+BaseNode.RootNode.Name;
      QMnode.Multiply := Emax < MaxDouble/2;
      with QMnode.Plot do begin
        SetLength(Frequency,nPoints);
        SetLength(Intensity,nPoints);
      end;
      nPoints := 0;
      repeat
        LastEv := nil;
        LastIndex := 0;
        LevelLister.Simulation.Temperature := T;
        try
          ShowSorted(True);
          if not Assigned(LevelLister.Overflow) then begin
            with QMnode.Plot do begin
              Frequency[nPoints] := T;
              Intensity[nPoints] := LevelLister.Q/LevelLister.Qscale; { Was Qsum/Qscale }
              if QMnode.Multiply then
                Intensity[nPoints] := Intensity[nPoints]/LevelLister.QSump;
            end;
            Inc(nPoints);
          end;
        except
          on E: EAbort do
            Exit;
          on E: Exception do
            if not Assigned(LevelLister.Overflow) then
              LevelLister.Overflow := EReportException.Create(E.Message);
        end;
        if Assigned(LevelLister.Overflow) then begin
          WriteLn(Log,DtosM(T,LevelLister.Mixture.Precision+6,-1):LevelLister.Mixture.Precision+6, ' ', LevelLister.Overflow.Message);
          FreeAndNil(LevelLister.Overflow);
        end;
        SetProgress(Round(T));
        if Tstep < 0 then
          T := -TStep * T
        else if T < Tstep then begin
          T := T*2;
          if T > Tstep then
            T := Tstep
        end else
          T := T + Tstep;
      until (T > Tmax);
      with QMnode.Plot do begin
        SetLength(Frequency,nPoints);
        SetLength(Intensity,nPoints);
      end;
      WriteLn(Log, LevelLister.LastInfo);
      if Emax < MaxDouble/2 then begin
        s := 'Qthresh includes the ' + IntToStr(LevelLister.nPrinted) + ' levels with E < '
             + DtosM(Emax,0,LevelLister.Mixture.Precision) + ', '
             + DtosM(Emax-LevelLister.Emin,0,LevelLister.Mixture.Precision) + ' above the origin';
        WriteLn(Log, s);
        QMnode.Comment := QMnode.Comment + '; ' + s;
      end;
      {$IFNDEF CONSOLE}
      QMnode.CopyToClipboard;
      {$ENDIF}
    finally
      LevelLister.Simulation.Temperature := Tmax;
      QMnode.Free;
      PopProgress;
    end;
  end;
begin
  BaseNode.RootNode.CheckProperties;
  if Emax = 0 then
    Emax := MaxDouble;
  LevelLister := nil;
  if not BeginCalculation(False) then
    try
      LevelLister := TLevelLister.Create(nil, BaseNode, False, False, False, Emax);
      if LevelListType <> lltBySymmetry then begin
        LevelLister.ShowPartitionAccum := True;
        nEvs := 0;
        LoadManifolds(BaseNode);
        if LevelListType = lltByEnergy then
          ShowSorted(False)
        else
          ShowQtable;
      end else begin
        ShowMNode(BaseNode);
        LevelLister.ShowTotals;
      end;
      LevelLister.ShowErrors;
    finally
      LevelLister.Free;
      EndCalculation;
    end;
end;

constructor TLevelLister.Create(ALevelPlotter: TObject; ABaseNode: TMNode; AShowExternalField, AESweep, AFieldSummary: Boolean; AEmax: Double);
var
  nMolecules, nManifolds: Integer;
  procedure CheckForWidths(MNode: TMNode);
  var
    i: Integer;
  begin
    if MNode is TManifold then begin
      HasWidths := HasWidths or TManifold(MNode).HasWidth;
      Inc(nManifolds);
      ShowNumericalPopulations := ShowNumericalPopulations or TManifold(MNode).UsePopParams;
    end else if MNode is TState then
      HasWidths := HasWidths or TState(MNode).HasWidth
    else begin
      if MNode is TMolecule then
        Inc(nMolecules);
      for i := 0 to MNode.ComponentCount - 1 do
        CheckForWidths(MNode.Components[i] as TMNode);
    end;
  end;
begin
  inherited Create;
  LevelPlotter := ALevelPlotter;
  BaseNode := ABaseNode;
  Mixture := BaseNode.RootNode as TMixture;
  Simulation := Mixture.Simulations[0];
  ShowNumericalPopulations := Simulation.Temperature <= 0;
  HasWidths := False;
  nManifolds := 0;
  nMolecules := 0;
  nLevels := 0;
  CheckForWidths(BaseNode);
  ShowMoleculeLabel := nMolecules > 1;
  ShowManifoldLabel := nManifolds > 1;
  ShowExternalField := AShowExternalField;
  ESweep := AESweep;
  FieldSummary := AFieldSummary;
  Emax := AEmax;
  QTable := False;
  Qscale := 1;
  LogHeaderShown := False;
end;

procedure TLevelLister.PrepareExtraInfo(Ev: TEigenValues);
begin
  {}
end;

procedure TLevelLister.ShowExtraHeader;
begin
  WriteLn(Log);
end;

procedure TLevelLister.ShowExtraInfo(Ev: TEigenvalues; i: Integer);
begin
  WriteLn(Log);
end;

procedure TLevelLister.CheckHeaderShown(ALabelLength: Integer; ALabel:string);
var
  SL: Integer;
  s, FieldUnits, DipoleUnits: string;
begin
  if not LogHeaderShown then begin
    { Header for list }
    LabelLength := Max(ALabelLength, Length(ALabel));
    SL := Max(Mixture.LongestSymLength,4);
    if Simulation.Temperature > 0 then begin
      s := ' at T = '+FloatToStr(Simulation.Temperature);
      if Simulation.Tvib >0 then
        s := s + ', Tvib = '+FloatToStr(Simulation.Tvib);
      if Simulation.Tspin >0 then
        s := s + ', Tspin = '+FloatToStr(Simulation.Tspin);
    end else
      s := '';
    if not ShowExternalField and IsNonZero(Simulation.ExField) then
      s :=  s + ' at a field of '+FieldToStr(Simulation.ExField);
    WriteLn(Log, 'Energy Level List for ',BaseNode.FullHeirarchicName,s);
    if FieldSummary then
      with LevelPlotter as TLevelPlotter do begin
        if ESweep then begin
          FieldUnits := '(Vm-1)';
          DipoleUnits := 'Debye';
        end else begin
          FieldUnits := 'T';
          DipoleUnits := 'muB';
        end;
        WriteLn(Log);
        WriteLn(Log, 'Coefficients for fit to a field sweep of ',FloatToStr(MinField),' to ',FloatToStr(MaxField), ' ',FieldUnits);
        WriteLn(Log, ' Linear coefficient is in ',Mixture.UnitsName,'/',FieldUnits);
        WriteLn(Log, ' Dipole is effective equivalent in ',DipoleUnits);
        WriteLn(Log, ' Quadratic coefficient is in ',Mixture.UnitsName,'/',FieldUnits,'^2');
        WriteLn(Log, ' Two Level fits (where present) are to E +/- Sqrt(Delta^2/4 + C^2*F^2) where F is field');
        WriteLn(Log, '   E and Delta are in ',Mixture.UnitsName);
        WriteLn(Log, '   C is in ',Mixture.UnitsName,'/',FieldUnits,' and Dipole2 is the equivalent dipole in ',DipoleUnits,'; the sign of the square root is taken from this');
        WriteLn(Log, '   %ages indicate the maximum error in energy from the fitted function as a percentage of the total energy change over the sweep');
        WriteLn(Log);
      end;
    if ShowMoleculeLabel or ShowManifoldLabel or (Alabel = '') then
      ALabel := 'Label';
    if ShowMoleculeLabel then
      Write(Log, Copy('Molecule',1,Length(Mixture.LongestMoleculeName)):Length(Mixture.LongestMoleculeName), ' ');
    if ShowManifoldLabel then
      Write(Log, Copy('Manifold',1,Length(Mixture.LongestManifoldName)):Length(Mixture.LongestManifoldName), ' ');
    if ShowExternalField or IsNonZero(Simulation.ExField) then
      Write(Log, '   M ')
    else
      Write(Log, '   J ');
    Write(Log, 'Sym':SL, '   # ');
    if ShowExternalField and not FieldSummary then
      if ESweep then
        Write(Log,'Field(V/m)')
      else
        Write(Log,'Field(T)');
    if not FieldSummary then
      Write(Log, '        Energy ');
    if Mixture.ShowEstUnc then
      Write(Log, 'Est Unc   ');
    if HasWidths then
      Write(Log, 'Width     ');
    Write(Log, '   g ');
    Write(Log, 'Population ');
    if ShowNumericalPopulations then begin
      Write(Log, 'Pop Param  ');
      Write(Log, 'Std Dev    ');
    end;
    if ShowPartitionAccum then
      Write(Log, 'Q          ');
    Write(Log, ALabel, ' ':LabelLength-Length(ALabel));
    if FieldSummary then
      Write(Log, '       Energy','  Linear         ','Dipole     Err','  Quadratic       Err  Two Level Delta     C              Dipole2    Err');
    ShowExtraHeader;
    LogHeaderShown := True;
  end;
end;

procedure TLevelLister.ShowErrors;
begin
  if Assigned(Overflow) then
    WriteLn(Log,'Note: ',Overflow.Message);
  FreeAndNil(Overflow);
end;

procedure TLevelLister.ShowLevel(Ev: TEigenvalues; i:Integer);
var
  Manifold: TManifold;
  SL, Offset, PIndex: Integer;
  Pop, W: Double;
  Lbl: string;
begin
  Manifold := Ev.Manifold;
  Pop := Ev.Degeneracy(i) * Manifold.Population(Ev,i,Simulation,Overflow,0);
  if LastMolecule <> Manifold.Molecule then begin
    ShowTotals;
    if QTable and Assigned(LastMolecule) then
      raise EReportException.Create('Please select a single species to calculate a partition function table');
    if IsNormalized(Simulation.IntensityUnits) then
      Q := Manifold.Molecule.PartitionFunction(Simulation, Overflow, Emin, Ev.RemovedSpins)
    else
      Q := 1;
    QSum := 0;
    QSump := 0;
    LastMolecule := Manifold.Molecule;
    Qscale := Manifold.bf(Simulation.Temperature,Emin-LastMolecule.ActualAssumedOrigin);
  end;
  QSum := QSum + Pop;
  if Ev.Energy[i] > Emax then
    Exit;
  QSump := QSump + Pop/Qscale;
  Inc(nPrinted);
  if (nPrinted > nLevels) and (nLevels > 0) then { nLevels = 0 is used in field sweeps }
    raise Exception.CreateFmt('More than %d levels printed',[nLevels]);
  if QTable then
    Exit;
  Lbl := Ev.DominantLabel(i, Mixture.QuantumNumberFormat);
  CheckHeaderShown(Length(Lbl)+5,Ev.StateLabelKey(i));
  SL := Max(Mixture.LongestSymLength,4);
  if ShowMoleculeLabel then
    Write(Log, Ev.Manifold.Molecule.Name:Length(Mixture.LongestMoleculeName), ' ');
  if ShowManifoldLabel then
    Write(Log, Ev.Manifold.Name:Length(Mixture.LongestManifoldName), ' ');
  Write(Log, Manifold.QuantumNumberToStr(Ev.J):4, ' ', Manifold.Molecule.JAdjustedSymName(Ev.J,Ev.Sym):SL, i:4);
  if ShowExternalField and not FieldSummary then
    if ESweep then
      Write(Log, DtosMA(Ev.ExField.E,10,-4))
    else
      Write(Log, DtosMA(Ev.ExField.B,10,-4));
  if not FieldSummary then
    Write(Log, ' ', Ev.Energy[i]:14:Manifold.Precision);
  if Mixture.ShowEstUnc then
    Write(Log, DtosMA(EstErr(nil,Ev,-1,i),10,-4));
  if HasWidths then begin
    if Assigned(Ev.Widths) then
      W := Ev.Widths[i]
    else
      W := 0;
    Write(Log, DtosMA(W,10,-4));
  end;
  Write(Log,Ev.Degeneracy(i):5);
  Write(Log,DtosMA(Pop/Q,11,-1));
  if ShowNumericalPopulations then begin
    PIndex := -1;
    if Manifold.UsePopParams or (Simulation.Temperature <=0) then
      PIndex := Manifold.PopulationIndex(Ev.J,Ev.Sym,i); { TODO -cField: Fix with Field }
    if PIndex > 0 then begin
      Write(Log,DtosMA(Manifold.Parameters[PIndex],11,-1));
      Write(Log,DtosMA(Manifold.StdDev[PIndex],11,-1));
    end else
      Write(Log,' ':22);
  end;
  if ShowPartitionAccum then
    Write(Log,DtosMA(QSump,11,-1));
  Write(Log, ' ', Lbl, ' ':Max(LabelLength-Length(Lbl),0));
  if Mixture.ShowDerivatives and (Length(Ev.Derivatives) > 0) then begin
    if Manifold.Count = 1 then
      Offset := Length(Manifold.State[0].FullHeirarchicName(False,True))+2
    else
      Offset := Length(Manifold.FullHeirarchicName(False,True))+2;
    for SL := 0 to High(Ev.Derivatives) do
      with Ev.Derivatives[SL] do
        Write(Log, ' ', Copy(MNode.FullParameterName(Index),Offset,255), DtosMA(Values[i], 12, -1) );
  end;
  if not FieldSummary then
    ShowExtraInfo(Ev, i);
end;

procedure TLevelLister.ShowManifold(Manifold: TManifold; State: TState; IncludePerturbations: Boolean; Symmetry: Integer; XMin: TQuantumNumber; XMax: TQuantumNumber; RemoveSpins: Byte);
var
  Sym, AdjSym, i, Pass: Integer;
  J: TQuantumNumber;
  Ev: TEigenValues;
  SelState: TState;
begin
  if XMax < 0 then
    XMax := Manifold.FindJmax(XMax, RemoveSpins);
  if XMin < 0 then
    if Simulation.HasFields then
      XMin := -XMax
    else
      XMin := Manifold.FindJmin(XMin, RemoveSpins);
  for Pass := 0 to 1 do begin
    if Simulation.HasFields then
      J := XMin
    else
      J := Manifold.FindJmin(XMin, RemoveSpins);
    while J <= XMax do begin
      for Sym := 0 to Manifold.nSym-1 do
        if (Symmetry < 0) or (Symmetry = Sym) then begin
          AdjSym := Manifold.Molecule.JAdjustedSym(J, Sym);
          if Assigned(State) and not IncludePerturbations then
            SelState := State
          else
            SelState := nil; { State specified, but want complete calculation }
          Ev := Manifold.Eigenvalues(J, AdjSym, [meHrot, meWidth, meAllDerivatives, meFloatedDerivatives], Simulation.ExField, True, plNone, SelState, IncludePerturbations, RemoveSpins);
          { Two passes needed to ensure derivatives all calculated if needed }
          if Pass = 1 then begin
            PrepareExtraInfo(Ev);
            Inc(nLevels, Ev.Size);
            for i := 1 to Ev.Size do
              if (SelState = State) or (Ev.DominantBasis[i].State = State) then             
                ShowLevel(Ev,i);
          end;
          if Simulation.HasFields then
            Break; { Symmetry not used }
        end;
      Inc(J, 2);
    end;
  end;
  WriteLn(Log);
end;

procedure TLevelLister.ShowTotals;
var
  Prec: Integer;
begin
  if not Assigned(LastMolecule) then
    Exit;
  Prec := Mixture.Precision+7;
  LastInfo := IntToStr(nLevels) + ' levels, actual origin = ' + DtosM(Emin,0,Mixture.Precision)
                                + ', assumed origin = ' + DtosM(LastMolecule.ActualAssumedOrigin,0,Mixture.Precision)
                                + ', Q from ' + LastMolecule.ActivePartitionFunction.UniqueName;
  if QTable then begin
    if not Assigned(Overflow) then begin
      Write(Log,DtosM(Simulation.Temperature,Prec-1,-1):Prec-1,DtosMA(QSum/Qscale,Prec,-1),DtosMA(QSum/Q,Prec,-1));
      if Emax < MaxDouble/2 then
        Write(Log, DtosMA(QSump/Q*Qscale,Prec,-1), DtosMA(QSump,Prec,-1));
      WriteLn(Log);
    end; { else error message displayed in calling routine }
  end else if nLevels > 0 then begin
    Write(Log, 'Partition function sums: ');
    WriteLn(Log, LastInfo, ' = ', DtosMA(Q/Qscale,Prec,-1));
    WriteLn(Log, '  All Levels :   Qsum =', DtosMA(QSum/Qscale,Prec,-1), ', Qsum/Q = ', DtosMA(QSum/Q,Prec,-1));
    if Emax < MaxDouble/2 then
      WriteLn(Log, '  ', nPrinted, ' levels < ', DtosM(Emax,0,Mixture.Precision), ' : Qsum/Q = ', DtosMA(QSump/Q*Qscale,Prec,-1), ', Qsum = ',DtosMA(QSump,Prec,-1));
  end;
  LastMolecule := nil;
end;

end.
