unit MNodeFunc;
{ Unit to handle scripting of TMNode functions; in separate file as rather compiler dependent. }
{$I utils.inc}{Copyright (C) 2013 C M Western; see commandline.pas for license}

interface

uses TypInfo, Classes, MNode;

procedure CheckMNodeFunction(MNode: TMNodeClass; const AName: string; TypeInfo: PTypeInfo; AltTypeInfo: PTypeInfo);

procedure CallMNodeFunction(MNode: TMNode; MethodAddress: Pointer; TypeData: PTypeData; Arguments: string);

implementation

uses SysUtils, UCompat, MathSup, PropHelp, MNodeExpr, Mixture, zMatrix, InternalCoords;

type
// From typinfo.pas FPC compiler source file TTypeData record
// As the short strings are simply concatenated, a proper structure definition is not possible
//  RTTIParameter = Record
//    Flags: TParamFlags;
//    ParamName: ShortString;
//    TypeName: ShortString;
//   end;
//  PRTTIParameter= ^RTTIParameter;
  PParamFlags = ^TParamFlags;

  TMethodString = procedure (const Arg: string) of object;
  TMethodStringString = procedure (const Arg1, Arg2: string) of object;
  TMethodStringStringString = procedure (const Arg1, Arg2, Arg3: string) of object;
  TMethodStringStringStringInt = procedure (const Arg1, Arg2, Arg3: string; I: Integer) of object;
  TMethodBoolBool = procedure (B1, B2: Boolean) of object;
  TMethodNone = procedure of object;
  TMethodStringStringIntBool = procedure (const Arg1, Arg2: string; I: Integer; B: Boolean) of object;
  TMethodDouble = procedure (const Arg: Double) of object;

function BumpParamPointer(var P: PByte; out TypeName: ShortString):TParamFlags;
var
  S: PShortString;
begin
  Result := PParamFlags(P)^;
  Inc(P, SizeOf(TParamFlags));
  S := PShortString(P); { Argument name }
  Inc(P, Length(S^)+1);
  S := PShortString(P); { Type name }
  Inc(P, Length(S^)+1);
  TypeName := S^;
end;

procedure CheckMNodeFunction(MNode: TMNodeClass; const AName: string; TypeInfo: PTypeInfo; AltTypeInfo: PTypeInfo);
var
  Data, AltData: PTypeData;
  Flags: TParamFlags;
  D, AD: PByte;
  i: Integer;
  S, S1: ShortString;
  OK: Boolean;
begin
  if (MNode.MethodAddress(AName) = nil) or (TypeInfo^.Kind <> tkMethod) then
    raise Exception.Create(AName);
  Data := GetTypeData(TypeInfo);
  D := PByte(@Data^.ParamList);
  AD := nil;
  if AltTypeInfo <> nil then begin
    AltData := GetTypeData(AltTypeInfo);
    if (AltData^.ParamCount <> Data^.ParamCount) or (AltTypeInfo^.Kind <> tkMethod) then
      raise Exception.Create(AName);
    AD := PByte(@AltData^.ParamList);
  end;
  { Check parameters - require simple value for each }
  for i := 0 to Data^.ParamCount-1 do begin
    Flags := BumpParamPointer(D, S);
    if S = {$IFDEF FPC} 'AnsiString' {$ELSE} 'string' {$ENDIF} then
      OK := Flags = [pfConst]
    else if S = 'Boolean' then
      OK := (Flags = [pfConst]) or (Flags = [])
    else if S = 'Double' then
      OK := (Flags = [pfConst]) or (Flags = [])
    else if S = {$IFDEF FPC} 'LongInt' {$ELSE} 'Integer' {$ENDIF} then
      OK := (Flags = [pfConst]) or (Flags = [])
    else
      OK := False;
    if OK and (AD <> nil) then begin
      OK := BumpParamPointer(AD, S1) = Flags;
      OK := OK and (S = S1);
    end;
    if not OK then
      raise Exception.Create(AName);
  end;
end;

procedure CallMNodeFunction(MNode: TMNode; MethodAddress: Pointer; TypeData: PTypeData; Arguments: string);
var
  Method: TMethod;
  Args: array of string;
  i, S, P: Integer;
  Quote: Char;
  function StringArg(n: Integer):string;
  begin
    Result := Args[n];
    if (Result <> '') and (Result[1] in ['''', '"']) and (Result[Length(Result)] = Result[1]) then
      Result := Copy(Result,2,Length(Result)-2);
  end;
  function BoolArg(n: Integer):Boolean;
  begin
    if not TryStrToBoolean(Args[n],Result) then
      raise Exception.Create('Boolean not recognized:'+Args[n]);
  end;
  function IntArg(n: Integer):Integer;
  begin
    if not TryStrToInt(Args[n],Result) then
      raise Exception.Create('Integer not recognized:'+Args[n]);
  end;
  function DoubleArg(n: Integer):Double;
  var
    s: string;
  begin
    if not TryStrToFloatEx(Args[n],Result) then begin
      s := EvaluateExpression(Args[n], MNode.RootNode, Result);
      if s <> '' then
        raise Exception.Create('Number not recognized:'+Args[n]+' - '+s);
    end
  end;
begin
  Method.Data := MNode;
  Method.Code := MethodAddress;
  SetLength(Args, TypeData^.ParamCount);
  P := 1;
  for i := 0 to High(Args) do begin
    if P > Length(Arguments) then
      raise EReportException.Create('Argument missing for '+MNode.UniqueName+'.'+MNode.MethodName(MethodAddress)+'('+Arguments+')');
    while (P <= Length(Arguments)) and (Arguments[P] = ' ') do
      Inc(P);
    S := P;
    while P <= Length(Arguments) do
      if (Arguments[P] in ['''', '"']) then begin
        { Quoted string }
        Quote := Arguments[P];
        Inc(P);
        repeat
          while (P <= Length(Arguments)) and (Arguments[P] <> Quote) do
            Inc(P);
          Inc(P);
          { Merge repeated quote marks }
          if (P <= Length(Arguments)) and (Arguments[P] = Quote) then
            Delete(Arguments,P,1)
          else
            Break;
        until False;
      end else if Arguments[P] = ',' then
        Break
      else
        Inc(P);
    Args[i] := Trim(Copy(Arguments,S,P-S));
    Inc(P);
  end;
  if P <= Length(Arguments) then
    raise EReportException.Create('Too many arguments in call to '+MNode.UniqueName+'.'+MNode.MethodName(MethodAddress)+'('+Arguments+')');
  if TypeData = GetTypeData(TypeInfo(TMethodString)) then
    TMethodString(Method)(StringArg(0))
  else if TypeData = GetTypeData(TypeInfo(TMethodStringString)) then
      TMethodStringString(Method)(StringArg(0), StringArg(1))
  else if TypeData = GetTypeData(TypeInfo(TMethodBoolBool)) then
    TMethodBoolBool(Method)(BoolArg(0), BoolArg(1))
  else if TypeData = GetTypeData(TypeInfo(TMethodNone)) then
    TMethodNone(Method)()
  else if TypeData = GetTypeData(TypeInfo(TMethodStringStringString)) then
    TMethodStringStringString(Method)(StringArg(0), StringArg(1), StringArg(2))
  else if TypeData = GetTypeData(TypeInfo(TMethodStringStringStringInt)) then
    TMethodStringStringStringInt(Method)(StringArg(0), StringArg(1), StringArg(2), IntArg(3))
  else if TypeData = GetTypeData(TypeInfo(TMethodStringStringIntBool)) then
    TMethodStringStringIntBool(Method)(StringArg(0), StringArg(1), IntArg(2), BoolArg(3))
  else if TypeData = GetTypeData(TypeInfo(TMethodDouble)) then
    TMethodDouble(Method)(DoubleArg(0))
  else
    raise Exception.Create(MNode.MethodName(MethodAddress));
end;

initialization
  { Placed here to avoid problems with @ operator in InternalCoords and TypeInfo built-in }
  TMNode.RegisterExternalFunction('AddNew', TypeInfo(TMethodString) {$IFDEF FPC}, TypeInfo(@TMNode.AddNew){$ENDIF});
  TMNode.RegisterExternalFunction('Make', TypeInfo(TMethodStringString) {$IFDEF FPC}, TypeInfo(@TMNode.Make){$ENDIF});
  TMNode.RegisterExternalFunction('DeleteNode', TypeInfo(TMethodBoolBool) {$IFDEF FPC}, TypeInfo(@TMNode.DeleteNode){$ENDIF});
  TMNode.RegisterExternalFunction('DeleteDynamicParameter', TypeInfo(TMethodString) {$IFDEF FPC}, TypeInfo(@TMNode.DeleteDynamicParameter){$ENDIF});
  TCartesianDisplacements.RegisterExternalFunction('CalculateFromL', TypeInfo(TMethodBoolBool) {$IFDEF FPC}, TypeInfo(@TCartesianDisplacements.CalculateFromL){$ENDIF});
  TSymmetryCoords.RegisterExternalFunction('CalculateF', TypeInfo(TMethodBoolBool) {$IFDEF FPC}, TypeInfo(@TSymmetryCoords.CalculateF){$ENDIF});
  TzMatrix.RegisterExternalFunction('LoadFromState', TypeInfo(TMethodString) {$IFDEF FPC}, TypeInfo(@TzMatrix.LoadFromState){$ENDIF});
  TMNodeWithArray.RegisterExternalFunction('AddRow', TypeInfo(TMethodString) {$IFDEF FPC}, TypeInfo(@TMNodeWithArray.AddRow){$ENDIF});
  TMNodeWithArray.RegisterExternalFunction('AddFRow', TypeInfo(TMethodString) {$IFDEF FPC}, TypeInfo(@TMNodeWithArray.AddFRow){$ENDIF});
  TMNodeWithArray.RegisterExternalFunction('ClearArray', TypeInfo(TMethodNone) {$IFDEF FPC}, TypeInfo(@TMNodeWithArray.ClearArray){$ENDIF});
  TMNodeWithArray.RegisterExternalFunction('ClearFArray', TypeInfo(TMethodNone) {$IFDEF FPC}, TypeInfo(@TMNodeWithArray.ClearFArray){$ENDIF});
  TInternalCoords.RegisterExternalFunction('CalculateFfromF', TypeInfo(TMethodBoolBool) {$IFDEF FPC}, TypeInfo(@TInternalCoords.CalculateFfromF){$ENDIF});
  TTransitionMoments.RegisterExternalFunction('AddTM', TypeInfo(TMethodStringStringStringInt) {$IFDEF FPC}, TypeInfo(@TTransitionMoments.AddTM){$ENDIF});
  TTransitionMoments.RegisterExternalFunction('AddForbidden', TypeInfo(TMethodStringStringIntBool) {$IFDEF FPC}, TypeInfo(@TTransitionMoments.AddForbidden){$ENDIF});
  TMNode.RegisterExternalFunction('MakePair', TypeInfo(TMethodStringStringString) {$IFDEF FPC}, TypeInfo(@TMNode.MakePair){$ENDIF});
  TMNode.RegisterExternalFunction('Rename', TypeInfo(TMethodString) {$IFDEF FPC}, TypeInfo(@TMNode.Rename){$ENDIF});
  TManifold.RegisterExternalFunction('OffsetBy', TypeInfo(TMethodDouble) {$IFDEF FPC}, TypeInfo(@TManifold.OffsetBy){$ENDIF});
end.
