unit MatrixElements;
{$I utils.inc}{Copyright (C) 2006 C M Western; see commandline.pas for license}
interface

uses Expr, QuantumNumber, MathSup;

function rlE(J, K: TExprNode; dK: Integer): TExprNode; overload;
function rlE(J: TExprNode; Kp, K: TQuantumNumber): TExprNode; overload;
function rlxE(n: Integer; J, K: TExprNode; dK: TQuantumNumber):TExprNode;
function rlyE(n: Integer; J, K: TExprNode; dK: TQuantumNumber):TExprNode;
function Halve(Q: TQuantumNumber):TExprNode;
function ReducedMatrixElementE(Rank: TQuantumNumber; J:TExprNode): TExprNode;

type
  PExprNode = ^TExprNode;

  TMatrixElement = record
    Signature: TIntegerArray; { Key quantum numbers (or change in them) to allow matrix element to be identified }
    Expr: TExprNode;
    Evaluator: TEvaluator;
    CEvaluator: TEvaluator; { Evaluator for purely imaginary functions - a bodge at the moment }
    BraPhase, KetPhase: Shortint; { Phases of wavefunctions; not always set at the moment. 0 => single function }
  end;

  TMatrixElements = array of TMatrixElement;
  PTMatrixElements = ^TMatrixElements;

  TSortedVariableNode = class(TVariableNode)
    Index: Integer;
    function EvalDependsOn: String; override;
    function EvalSortText(ConstantsHigh:Boolean = False): string; override;
  end;
  PSortedVariableNode = ^TSortedVariableNode;

  TSortedVariable = class (TVariable)
    Template: TSortedVariableNode;
    Index: Integer;
    function CreateExprNode:TExprNode; override;
    constructor Create(s:String; AIndex: Integer);
    destructor Destroy; override;
  end;

var
  ExprChecking: Boolean = False; { Set to evaluate matrix elements symbolically as well as numerically }
  HyperfineVariableList: TVariableList;

function MakeSortedVariables(const Names: array of string; const Nodes: array of PSortedVariableNode):TVariableList;
function eFn(n: Integer):TSortedVariableNode;
procedure DiscardMatrixElements(var MatrixElements: TMatrixElements);
procedure MergeElements(var A, B: TMatrixElements; ScaleB: Integer; Compile, Simplify: Boolean);
procedure MergeInMatrixElements(var ME: TMatrixElements; var M: TMatrixElements; Variable: TSortedVariableNode; Duplicate: Boolean);
function ShowElementPhases(const MatrixElement: TMatrixElement; const H: string; ShowPhases: Boolean):string;
procedure DumpMatrixElements(const M: TMatrixElements; const Title: string);
procedure CompileMatrixElements(MatrixElements: TMatrixElements; Variable: TExprNode);
function SignaturesMatch(const A, B: TIntegerArray):Boolean;

implementation

uses SysUtils, Math, Worker;

{ TSortedVariableNode }

function TSortedVariableNode.EvalDependsOn: String;
begin
  { DependsOn is used for sorting, so force order here by prefixing with number }
  Result := IntToHex(Index,8) + inherited EvalDependsOn;
end;

function TSortedVariableNode.EvalSortText(ConstantsHigh:Boolean = False): string;
begin
  Result := IntToHex(Index,8) + inherited EvalSortText(ConstantsHigh);
end;

{ TSortedVariable }

function TSortedVariable.CreateExprNode: TExprNode;
var
  R: TSortedVariableNode;
begin
  R := TSortedVariableNode.Create(Self);
  R.Index := Index;
  Result :=  R;
end;

constructor TSortedVariable.Create(s: String; AIndex: Integer);
begin
  inherited Create(s);
  Index := AIndex;
  Template := CreateExprNode as TSortedVariableNode;
end;

destructor TSortedVariable.Destroy;
begin
  Template.Free;
  inherited Destroy;
end;
  
function MakeSortedVariables(const Names: array of string; const Nodes: array of PSortedVariableNode):TVariableList;
var
  i: Integer;
  V: TSortedVariable;
begin
  Result := TVariableList.Create;
  for i := 0 to High(Names) do begin
    V := TSortedVariable.Create(Names[i], 1000+i);
    Result.Add(V);
    Nodes[i]^ := V.Template;
  end;
end;

var
  HyperfineVariables: array of TSortedVariableNode;

function eFn(n: Integer):TSortedVariableNode;
var
  s: string;
  i: Integer;
  Vbl: TSortedVariable;
begin
  { Hyperfine variable Fn }
  if n > High(HyperfineVariables) then begin
    for i := Length(HyperfineVariables) to n-1 do
      eFn(i); { Create complete set }
    s := 'F';
    if n = 0 then begin
      HyperfineVariableList := TVariableList.Create;
      i := 900
    end else begin
      s := s + IntToStr(n);
      i := 999-n;
    end;
    SetLength(HyperfineVariables, n+1);
    Vbl := TSortedVariable.Create(s,i);
    HyperfineVariableList.Add(Vbl);
    HyperfineVariables[n] := Vbl.Template;
    HyperfineVariables[n].Value := 5.5; {Arbitrary }
  end;
  Result := HyperFineVariables[n];
end;

procedure DiscardMatrixElements(var MatrixElements: TMatrixElements);
var
  i: Integer;
begin
  for i := 0 to High(MatrixElements) do begin
    FreeAndNil(MatrixElements[i].Expr);
    FreeAndNil(MatrixElements[i].Evaluator);
    FreeAndNil(MatrixElements[i].CEvaluator);
  end;
  SetLength(MatrixElements,0);
end;

procedure MergeElements(var A, B: TMatrixElements; ScaleB: Integer; Compile, Simplify: Boolean);
var
  i, j, k: Integer;
  Matches: Boolean;
begin
  { Merge two sets of matrix elements into A, discarding B }
  for i := 0 to High(B) do begin { For each matrix element in B }
    FreeAndNil(B[i].Evaluator);
    FreeAndNil(B[i].CEvaluator);
    j := -1;
    Matches := False;
    while (j < High(A)) and not Matches do begin
      { look through each matrix element in A }
      Inc(j);
      Matches := SignaturesMatch(A[j].Signature, B[i].Signature);
      Matches := Matches and (A[j].BraPhase = B[i].BraPhase) and (A[j].KetPhase = B[i].KetPhase);
    end;
    if Matches then
      { Combine two matrix elements }
      if Simplify then
        A[j].Expr := EN([B[i].Expr, '*', ScaleB, '+', A[J].Expr])
      else begin
        if A[j].Expr.Base <> OperatorPlus then
          A[j].Expr := OperatorPlus.CreateExprNode(A[j].Expr);
        if ScaleB <> 1 then
          A[j].Expr.AddNode(EN([B[i].Expr, '*', ScaleB]))
        else
          A[j].Expr.AddNode(B[i].Expr);
      end
    else begin
      { Add another element }
      j := Length(A);
      Setlength(A,j+1);
      A[j] := B[i];
      A[j].Expr := EN([B[i].Expr, '*', ScaleB]);
    end;
    FreeAndNil(A[j].Evaluator);
    FreeAndNil(A[j].CEvaluator);
  end;
  { Remove any zero results }
  for i := High(A) downto 0 do
    if A[i].Expr.EqualTo(0) then begin
      A[i].Expr.Free;
      A[i].Evaluator.Free;
      A[i].CEvaluator.Free;
      for k := i to High(A)-1 do
        A[k] := A[k+1];
      SetLength(A,Length(A)-1);
    end;
  if Compile then
    for i := 0 to High(A) do
      if not Assigned(A[i].Evaluator) then
        A[i].Evaluator := A[i].Expr.Compile([evoZeroOverZero]);
  B := nil;
end;

procedure MergeInMatrixElements(var ME: TMatrixElements; var M: TMatrixElements; Variable: TSortedVariableNode; Duplicate: Boolean);
var
  j, k, l, w: Integer;
  Match: Boolean;
  MExpr: TExprNode;
begin
  for j := 0 to High(M) do begin { For each set of changes of quantum numbers }
    w := Length(ME);
    for k := 0 to High(ME) do begin
      { Do we already have some matching changes? }
      Match := SignaturesMatch(ME[k].Signature, M[j].Signature);
      Match := Match and (ME[k].BraPhase = M[j].BraPhase) and (ME[k].KetPhase = M[j].KetPhase);
      if Match then begin
        w := k;
        Break;
      end;
    end;
    MExpr := M[j].Expr;
    if Duplicate then
      MExpr := MExpr.Duplicate;
    MExpr := OperatorTimes.CreateExprNode(Variable.Duplicate,MExpr);
    if w = Length(ME) then begin
      { No Match - add new }
      SetLength(ME, w+1);
      ME[w] := M[j];
      ME[w].Expr := MExpr;
      ME[w].Evaluator := nil;
      ME[w].CEvaluator := nil;
    end else
      { Same set of changes - add expression, but do not simplify }
      with ME[w] do begin
        if Expr.Base <> OperatorPlus then
          Expr := OperatorPlus.CreateExprNode(Expr);
        Expr.AddNode(MExpr);
      end;
  end;
end;

function ShowElementPhases(const MatrixElement: TMatrixElement; const H: string; ShowPhases: Boolean):string;
  function SP(Phase: Integer):string;
  begin
    if ShowPhases then
      case Phase of
         0: Result := '';
         1: Result := ' + ';
        -1: Result := ' - ';
      else
        raise Exception.Create('Signature Phase Error');
      end
    else
      Result := '';
  end;
begin
  Result := SP(MatrixElement.BraPhase) + '|' + H + '|' + SP(MatrixElement.KetPhase);
end;

function rlE(J, K: TExprNode; dK: Integer): TExprNode;
{ Work out matrix element of arbitrary power of raising/lowering operator }
{ J, K are Ket variables; dK is change in K }
var
  JSq, Kp, Node: TExprNode;
  i: Integer;
begin
  Result := TIntegerNode.Create(1);
  JSq := EB([J, '+', 1, '*', J]);
  K := K.Duplicate;
  Kp := nil;
  for i := 1 to Abs(dK) do begin
    if dK > 0 then
      Kp := EB([K,'+',1])
    else
      Kp := EB([K,'-',1]);
    Kp := Kp.Simplify;
    Node := EB([K, '*', Kp]);
    Node := OperatorMinus.CreateExprNode(JSq.Duplicate, Node);
    Result := OperatorTimes.CreateExprNode(Result, Node);
    K.Free;
    K := Kp;
  end;
  Result := SqrtFunction.CreateExprNode(Result);
  JSq.Free;
  Kp.Free;
end;

function Halve(Q: TQuantumNumber):TExprNode;
begin
  Result := EN([Q, '/', 2]);
end;

function rlE(J: TExprNode; Kp, K: TQuantumNumber): TExprNode;
{ Work out matrix element of arbitrary power of raising/lowering operator }
var
  JSq, Node: TExprNode;
  dK: TQuantumNumber;
begin
  Result := TIntegerNode.Create(1);
  JSq := EB([J, '+', 1, '*', J]);
  dK := Sign(Kp-K)*2;
  while Kp <> K do begin
    Node := EN([Halve(K), '*', Halve(K+dK)]);
    Node := OperatorMinus.CreateExprNode(JSq.Duplicate, Node);
    Result := OperatorTimes.CreateExprNode(Result, Node);
    Inc(K, dK);
  end;
  Result := SqrtFunction.CreateExprNode(Result);
  JSq.Free;
end;

function rlxy(n: Integer; J, K: TExprNode; dK: Integer; Sign:Integer):TExprNode;
{ Helper function for matrix elements of Jx^n and Jy^n; K is Ket }
var
  Node, eK: TExprNode;
begin
  if (n < Abs(dK)) or Odd(n+dK) then
    Result := TIntegerNode.Create(0)
  else if n = Abs(dK) then
    { Easy case - equivalent to all raising/all lowering operator }
    if n = 0 then
      Result := TIntegerNode.Create(1)
    else begin
      Result := rlE(J,K,dK);
      Result := OperatorDivide.CreateExprNode(Result, EB([2, '^', n]));
      if (dK < 0) and Odd(n) and (Sign < 0) then
        Result := MonadicMinus.CreateExprNode(Result);
    end
  else begin
    { Need to consider combinations;
      e.g. Jx^2 can give DeltaK = 0, +/-2 Matrix elements }
      eK := EB([K,'+',1]);
      Result := EN([rlxy(n-1,J,eK,dK-1,Sign),'*',rlxy(1,J,K,1,Sign)]);
      eK.Free;
      eK := EB([K,'-',1]);
      Node := EN([rlxy(n-1,J,eK,dK+1,Sign),'*',rlxy(1,J,K,-1,Sign)]);
      Result := OperatorPlus.CreateExprNode(Result,Node);
      eK.Free;
  end;
end;

function rlxE(n: Integer; J, K: TExprNode; dK: TQuantumNumber):TExprNode;
{ Matrix elements of Jx^n; K is Ket }
begin
  Result := rlxy(n,J,K,dK,1);
end;

function rlyE(n: Integer; J, K: TExprNode; dK: TQuantumNumber):TExprNode;
{ Matrix elements of Jy^n; K is Ket; for odd powers the result is multiplied by i }
var
  Phase: Integer;
begin
  Result := rlxy(n,J,K,dK,-1); { Purely real here - multiply by factor of i^n to give true result }
  { Multiply by i^n }
  if Odd(n) then
    Phase := Minus(n+3) { Multiply by -i to give a real result }
  else
    Phase := Minus(n);
  if Phase < 0 then
    Result := MonadicMinus.CreateExprNode(Result);
end;

function ReducedMatrixElementE(Rank: TQuantumNumber; J:TExprNode): TExprNode;
{ Return the reduced matrix element <J||T^k(J)||J> formed by
  coupling J with J k times. }
begin
//  case Rank of
//    2: Result := EN([EB([J, '+', 1, '*', J]), '*', EB([2, '*', J, '+', 1])]);
//    4: Result := EN([EB([2, '*', J, '-', 1, '*', 2, '*', J]),
//                     '*', EB([2, '*', J, '+', 1]), '*', EB([2, '*', J, '+', 2]),
//                     '*', EB([2, '*', J, '+', 3]), '/', 24]);
//  else
//    raise Exception.Create('ReducedMatrixElementE');
//  end;
//  Result := SqrtFunction.CreateExprNode(Result);
//  Exit;
  { Brown and Carrington 5.134 }
  Result := EN([ 2, '^', Rank div 2]);
  Result := EN([ Result,
                 '*', EB([J, '*', 2, '-', Rank div 2, 'factorial']),
                 '*', EN([Rank, 'factorial'])],[],1);
  Result := EN([EB([J, '*', 2, '+', Rank div 2, '+', 1, 'factorial']),
               '/', Result, 'sqrt']);
  Result := EN([Rank div 2, 'factorial', '*', Result]);
{ Specific results for reference:
<J||T^1(J)||J>=sqrt(J*(2*J+1)*(2*J+2))
<J||T^2(J)||J>=sqrt(J*(2*J-1)*(2*J+1)*(2*J+2)*(2*J+3))/(2*sqrt(3))
<J||T^3(J)||J>=sqrt(J*(2*J-1)*(2*J-2)*(2*J+1)*(2*J+2)*(2*J+3)*(2*J+4))/(4*sqrt(5))
<J||T^4(J)||J>=sqrt(J*(2*J-1)*(2*J-2)*(2*J-3)*(2*J+1)*(2*J+2)*(2*J+3)*(2*J+4)*(2*J+5))/(4*sqrt(35))
<J||T^5(J)||J>=sqrt(J*(2*J-1)*(2*J-2)*(2*J-3)*(2*J-4)*(2*J+1)*(2*J+2)*(2*J+3)*(2*J+4)*(2*J+5)*(2*J+6))/(24*sqrt(7))
<J||T^6(J)||J>=sqrt(J*(2*J-1)*(2*J-2)*(2*J-3)*(2*J-4)*(2*J-5)*(2*J+1)*(2*J+2)*(2*J+3)*(2*J+4)*(2*J+5)*(2*J+6)*(2*J+7))/(8*sqrt(462))
}
end;

procedure DumpMatrixElements(const M: TMatrixElements; const Title: string);
var
  i, j: Integer;
begin
  for i := 0 to High(M) do
    with M[i] do begin
      Write(Log, Title,':',i,BraPhase:3, KetPhase:3);
      for j := 0 to High(Signature) do
        Write(Log, Signature[j]:3);
      WriteLn(Log, ' ', Expr.Text(True));
    end;
end;

procedure CompileMatrixElements(MatrixElements: TMatrixElements; Variable: TExprNode);
var
  i: Integer;
begin
  for i := 0 to High(MatrixElements) do begin
    if Assigned(Variable) then
      MatrixElements[i].Expr := OperatorTimes.CreateExprNode(Variable.Duplicate,MatrixElements[i].Expr);
    MatrixElements[i].Evaluator := MatrixElements[i].Expr.Compile([evoZeroOverZero]);
  end;
end;

function SignaturesMatch(const A, B: TIntegerArray):Boolean;
var
  i, Mi: Integer;
begin
  Result := False;
  Mi := Min(High(A), High(B));
  for i := 0 to Mi do
    if A[i] <> B[i] then
      Exit;
  { Implicit zero extend - at most one loop will run }
  for i := Mi+1 to High(A) do
    if A[i] <> 0 then
      Exit;
  for i := Mi+1 to High(B) do
    if B[i] <> 0 then
      Exit;
  Result := True;
end;

initialization
finalization
  FreeAndNil(HyperfineVariableList);
end.
