unit OAsymT;

{ Open Shell Asymmetric Top }

{$DEFINE NOAT}{$I utils.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface
{.$DEFINE JTEST}
{ JTEST flags an experiment with modelling random perturbations by just
  including a single K value from the perturbing state in the model }

uses SysUtils, Classes, MNode, Mixture, QuantumNumber, QuantumMath, Expr,
     MatrixElements, Math, MathSup, Abelian, pgDOM;

const
  FirstWidthPar = LSP+40;

type

  TAsymmetricMolecule = class;

  TSphericalTensor = array [0..2] of array [-2..2] of Double;
  TSecondRankTensor = array[-2..2] of Complex;

  { TAsymmetricTop }

  TAsymmetricTop = class (TState)
  private
    FKinc: Integer;
    TSR: TSphericalTensor; { Spherical tensor equivalent of spin rotation terms }
    QuarticSpinRotation: Boolean; { True if any Quartic Spin Rotation term non zero }
    {$IFDEF JTEST}
    JTestOrigin: Double;
    {$ENDIF}
    { Next two record the mapping from overall sym number to rotational eo symmetry;
      Implicitly contains more than one value for subgroups of C2v.
      Index is overall RveSym and eoSymStep gives the increment to find the next one - 2, 4 or 8
      Contains -1=No Such state, 0=ee, 1=eo, 2=oe, 3=00 }
    eoSyms: array [0..7] of Integer;
    eoSymsStep: Integer;
    FMolecule: TAsymmetricMolecule;
    Izz, Iyy, Ixx, IxxpIyy, IxxmIyy: Double; { Cached intermediate values }
  public
    property Kinc: Integer read FKinc;
    property Molecule: TAsymmetricMolecule read FMolecule;
    constructor Create(AOwner: TComponent); override;
    class function SpinSpinType: TSpinSpinClass; override;
    function AddOneJ(Eigenvalues: TEigenvalues; J, F, M: TQuantumNumber; Sym: Integer; HyperfineState:THyperfineBasis; Seffective: TQuantumNumber): Integer; override;
    function AddOneN(Eigenvalues: TEigenvalues; N, J, F, M: TQuantumNumber; Sym: Integer; HyperfineState:THyperfineBasis; Seffective: TQuantumNumber; out StatWtGroup: Integer): Integer;
    function FixupProperties: Exception; override;
    function ReadProperties(Node: TDOMNode; CheckUnread: Boolean = True):Integer; override;
    function HRotA(J, Kp, K: TQuantumNumber): Double;
    function HRotS(J, Kp, K: TQuantumNumber): Double;
    {$IFDEF JTEST}
    function HJtest(J, Kp, K: TQuantumNumber): Double;
    {$ENDIF}
    function HSR(J, Np, Kp, N, K:TQuantumNumber): Complex;
    function HasWidth: Boolean; override;
    function HWidth(J, Kp, K: TQuantumNumber): Double;
    function HDip(TransitionMoment: TSingleTransitionMoment; Jp, Kp, J, K: TQuantumNumber): Complex;
    { Key Matrix element functions }
    function RotationalMatrixElement(ABra, AKet: TBasisState; SignA, SignB: Integer; MNode: TMnode; Flag: Integer): Complex; override;
    function RotationalPerturbationMatrixElement(ABra, AKet: TBasisState; SignA, SignB: Integer; MNode: TMNode; Flag: Integer):Complex; override;
    function HyperfineMatrixElement(ABra, AKet: TBasisState; Signp, Sign: Integer; MNode: TMNode; RankWanted: TQuantumNumber): Complex; override;
    function RotationalTransitionMatrixElement(ABra, AKet: TBasisState; SignA, SignB: Integer; TransitionMoment: TRootedMNodePair; F: TQuantumNumber; out dK: TQuantumNumber):Complex; override;
    function SymbolicRotationalTransitionMatrixElements(TransitionMoment: TTransitionMoment; AddHyperfineCoupling: Boolean; IncompleteOK: Boolean = False): TMatrixElements; override;
    function SymbolicRotationalMatrixElement(AIndex: Integer; RankWanted: Integer):TMatrixElements; override; { For given parameter }
    function TransitionMatrixElement(TransitionMoment: TTransitionMoment; A, B: TBasisState): Complex; override;

    function GetSettingsNames(AIndex: Integer; out LowestValue: Integer): PPChar; override;
    function HelpOnSetting(AIndex: Integer): string; override;
    function HelpOnParameter(AIndex: Integer): string; override;
    function IsSettingQuantum(AIndex: Integer):Boolean; override;
    procedure SymmetryTable(var F: Text); override;
    function ConvertParameter(AIndex: Integer; Zero: Boolean):TPerturbation; override;
    function MatrixElementType(const TheMatrixElement: TMatrixElement; const H: string; ShowPhases: Boolean = False):string; override;
    function zIndex: Integer; { Setting index of zAxis rotational constant }
    function BumpBranchQno(Branch: string; AIncrement: TQuantumNumber; QnoType:TQnoType):string; override;
    function ReadBranchFormat(Text: string; var Obs: TObservation; UpperState: TState; Reader: TObservationsReaderBase):Boolean; override;
    function SettingToStr(AIndex: Integer): string; override;
    procedure SetSettingFromStr(AIndex: Integer; const Str: string); override;
    procedure AdjustDefaults; override;
  published
    property {Vibronic}Symmetry: Integer index LSI   read GetSetting write SetSetting;
    property Kmin:     TQuantumNumber    index LSI+1 read GetSetting write SetSetting;
    property Kmax:     TQuantumNumber    index LSI+2 read GetSetting write SetSetting;
    {$IFDEF JTEST}
    property JTest:    TQuantumNumber    index LSI+3 read GetSetting write SetSetting;
    {$ENDIF}

    property A: Double index LSP   read GetParameter write SetParameter;
    property B: Double index LSP+1 read GetParameter write SetParameter;
    property C: Double index LSP+2 read GetParameter write SetParameter;
    { Alternate form for rotational constants }
    property BBar:  Double index LSP+3 read GetParameter write SetParameter;
    property BDelta:Double index LSP+4 read GetParameter write SetParameter;
    { Quartic Centrifugal Distortion }
    property DK:      Double index LSP+ 5 read GetParameter write SetParameter;
    property DJK:     Double index LSP+ 6 read GetParameter write SetParameter;
    property DJ:      Double index LSP+ 7 read GetParameter write SetParameter;
    property deltaK:  Double index LSP+ 8 read GetParameter write SetParameter;
    property deltaJ:  Double index LSP+ 9 read GetParameter write SetParameter;
    { Sextic Centrifugal Distortion }
    property HK:    Double index LSP+10 read GetParameter write SetParameter;
    property HKJ:   Double index LSP+11 read GetParameter write SetParameter;
    property HJK:   Double index LSP+12 read GetParameter write SetParameter;
    property HJ:    Double index LSP+13 read GetParameter write SetParameter;
    property phiK:  Double index LSP+14 read GetParameter write SetParameter;
    property phiJK: Double index LSP+15 read GetParameter write SetParameter;
    property phiJ:  Double index LSP+16 read GetParameter write SetParameter;
    { Octic Centrifugal Distortion }
    property LK:    Double index LSP+17 read GetParameter write SetParameter;
    property LKKJ:  Double index LSP+18 read GetParameter write SetParameter;
    property LJK:   Double index LSP+19 read GetParameter write SetParameter;
    property LJJK:  Double index LSP+20 read GetParameter write SetParameter;
    property LJ:    Double index LSP+21 read GetParameter write SetParameter;
    property llK:   Double index LSP+22 read GetParameter write SetParameter;
    property llKJ:  Double index LSP+23 read GetParameter write SetParameter;
    property llJK:  Double index LSP+24 read GetParameter write SetParameter;
    property llJ:   Double index LSP+25 read GetParameter write SetParameter;
    { Spin-Rotation Parameters }
    property eaa:    Double index LSP+26 read GetParameter write SetParameter;
    property ebb:    Double index LSP+27 read GetParameter write SetParameter;
    property ecc:    Double index LSP+28 read GetParameter write SetParameter;
    property eabbar: Double index LSP+29 read GetParameter write SetParameter; { (eab + eba)/2 }
    property eacbar: Double index LSP+30 read GetParameter write SetParameter; { (eac + eca)/2 }
    property ebcbar: Double index LSP+31 read GetParameter write SetParameter; { (ebc + ecb)/2 }
      { Anti symmetric combinations like eab - eba omitted as they are not dererminable }
    { Quartic spin rotation parameters - use S reduction names for parameters }
    {   A reduction names given as comments }
    property DsN:   { DeltasN }  Double index LSP+32 read GetParameter write SetParameter;
    property DsNK:  { DeltasNK } Double index LSP+33 read GetParameter write SetParameter;
    property DsKN:  { DeltasKN } Double index LSP+34 read GetParameter write SetParameter;
    property DsK:   { DeltasK }  Double index LSP+35 read GetParameter write SetParameter;
    property ds1:   { deltasN }  Double index LSP+36 read GetParameter write SetParameter;
    property ds2:   { deltasK }  Double index LSP+37 read GetParameter write SetParameter;
    { Spin-Spin interaction }
    property alpha: Double index LSP+38 read GetParameter write SetParameter;
    property beta:  Double index LSP+39 read GetParameter write SetParameter;

    { Linewidth parameters - initial stab }
    property wKa:   Double index FirstWidthPar read GetParameter write SetParameter;
    property wKb:   Double index FirstWidthPar+1 read GetParameter write SetParameter;
    property wKc:   Double index FirstWidthPar+2 read GetParameter write SetParameter;
    property wBar:  Double index FirstWidthPar+3 read GetParameter write SetParameter;
    property wDelta:Double index FirstWidthPar+4 read GetParameter write SetParameter;

  end;


  TPertEquiv = record
    Name:   string;
    JPower: Integer;
    aPower: Integer;
    bPower: Integer;
    cPower: Integer;
    pPower: Integer;
    zPower: Integer; { maps to aPower, bPower or cPower }
    Scale4: Integer; { 4 * Scale factor }
    SaPower:Integer;
    SbPower:Integer;
    ScPower:Integer;
    SpPower:Integer;
    SzPower:Integer; { maps to SaPower, SbPower or ScPower }
    SPower: Integer;
    NSPower:Integer;
  end;

  { TAsymmetricTopPerturbation }

  TAsymmetricTopPerturbation = class (TPerturbation)
  public
    constructor Create(AOwner: TComponent); override;
    function ReadProperties(Node: TDOMNode; CheckUnread: Boolean = True):Integer; override;
    function MatrixElement(J, S, Np, Kp, N, K: TQuantumNumber): Complex;
    function Molecule: TAsymmetricMolecule;
    function PairName: string; override;
    function SymbolicMatrixElements(WithHyperfine: Boolean):TMatrixElements; override;
    function SettingToStr(AIndex: Integer): string; override;
    procedure SetSettingFromStr(AIndex: Integer; const s: string); override;
    procedure SetFromPertEquiv(const PE: TPertEquiv);
    function HelpOnSetting(AIndex: Integer): string; override;
    function IsSettingQuantum(AIndex: Integer):Boolean; override;
  published
    property JPower:Integer         index LPI    read GetSetting write SetSetting;
    property aPower:Integer         index LPI+1  read GetSetting write SetSetting;
    property bPower:Integer         index LPI+2  read GetSetting write SetSetting;
    property cPower:Integer         index LPI+3  read GetSetting write SetSetting;
    property pPower:Integer         index LPI+4  read GetSetting write SetSetting;
    property KSelect:TQuantumNumber index LPI+5  read GetSetting write SetSetting;
    property SaPower:Integer        index LPI+6  read GetSetting write SetSetting;
    property SbPower:Integer        index LPI+7  read GetSetting write SetSetting;
    property ScPower:Integer        index LPI+8  read GetSetting write SetSetting;
    property SpPower:Integer        index LPI+9  read GetSetting write SetSetting;
    property SPower:Integer         index LPI+10 read GetSetting write SetSetting;
    property NSPower:Integer        index LPI+11 read GetSetting write SetSetting;
  end;

  TAsymmetricTopIntensityPerturbation = class (TAsymmetricTopPerturbation)
  public
    constructor Create(AOwner: TComponent); override;
    function HelpOnSetting(AIndex: Integer): string; override;
    function BraOwner: TMNode; override;
    function KetOwner: TMNode; override;
    function IsAntiSym: Boolean; override;
  published
    property AntiSym:Boolean      index LPI+12 read GetFlag write SetFlag;
  end;

  { TAsymmetricTopNucleus }

  TAsymmetricTopNucleus = class(TNucleusWithSpin)
  private
    TT:  TSecondRankTensor; { Spherical tensor equivalent of dipole - dipole terms }
    CHI: TSecondRankTensor; { Spherical tensor equivalent of quadrupole terms }
    QuadI: Double; { = 1/thrj(I, 4, I, -I, 0, I) }
    TC00, TC20, TC22: Double;
  public
    function FixupProperties: Exception; override;
    function HelpOnParameter(AIndex: Integer): string; override;
    function HasIdentical: Boolean; override;
  published
    property aF:       Double index  1 read GetParameter write SetParameter;
    property Tzz:      Double index  2 read GetParameter write SetParameter;
    property Txxmyy:   Double index  3 read GetParameter write SetParameter;
    property Tab:      Double index  4 read GetParameter write SetParameter;
    property Tac:      Double index  5 read GetParameter write SetParameter;
    property Tbc:      Double index  6 read GetParameter write SetParameter;
    { Quadrupole }
    property CHIzz:    Double index  7 read GetParameter write SetParameter;
    property CHIxxmyy: Double index  8 read GetParameter write SetParameter;
    property CHIab:    Double index  9 read GetParameter write SetParameter;
    property CHIac:    Double index 10 read GetParameter write SetParameter;
    property CHIbc:    Double index 11 read GetParameter write SetParameter;
    { Nuclear Spin - Rotation }
    property Caa:      Double index 12 read GetParameter write SetParameter;
    property Cbb:      Double index 13 read GetParameter write SetParameter;
    property Ccc:      Double index 14 read GetParameter write SetParameter;
  end;

  { TAsymmetricTopSpinSpin }

  TAsymmetricTopSpinSpin = class(TSpinSpin)
  private
    D: TSecondRankTensor;
  public
    function FixupProperties: Exception; override;
  published
    property J:      Double index  1 read GetParameter write SetParameter;
    property Dzz:    Double index  2 read GetParameter write SetParameter;
    property Dxxmyy: Double index  3 read GetParameter write SetParameter;
    property Dab:    Double index  4 read GetParameter write SetParameter;
    property Dac:    Double index  5 read GetParameter write SetParameter;
    property Dbc:    Double index  6 read GetParameter write SetParameter;
  end;

  { TAsymmetricTopBasisState }

  TAsymmetricTopBasisState = class(TBasisState)
  public
    K: TQuantumNumber;
    Ka, Kc: TQuantumNumber;
    function TrackIndex: Integer; override;
    function BasisLabel: string; override; { Basis State Label }
    function BasisLabelKey: string; override; { Explanation of Basis State Label }
    function StateLabel(AQuantumNumberFormat:TQuantumNumberFormat): string; override; { State name, assuming energy ordering preserved }
    function StateLabelKey: string; override; { Explanation of State Name }
    function Qno(QnoType: TQnoType): TQuantumNumber; override; { extra quantum numbers - Ka, Kc }
    function MakeSignature(SignA:Integer; Ket: TBasisState; SignB: Integer; MNode: TMNode; Symmetrised: Boolean): TIntegerArray; override;
    procedure SetSymbolicVariables(SignA:Integer; Ket: TBasisState; SignB: Integer; MNode: TMNode); override;
    function SymbolicVariableValues: string; override;
    function TxName(Upper, Lower: TBasisState): string; override;    { Make transition label such as qRK(3) }
    function TxNameKey(Upper, Lower: TBasisState): string; override; { String explaining transition label }
    procedure CopyFrom(OtherState: TBasisState; Sign: Integer); override;
    function NextBasis(var LoopState: TBasisState; TransitionMoment: TTransitionMoment; Sign: Integer):Boolean; override;
  end;

  { TAsymmetricManifold }

  TAsymmetricManifold = class(TManifold)
  public
    constructor Create(AOwner: TComponent); override;
    function nSym: Integer; override; { Number of different symmetries }
  end;

  TRepresentation = (Ir=0, IIr, IIIr, Il, IIl, IIIl);

  { TAsymmetricMolecule }

  TAsymmetricMolecule = class(TMolecule)
  private
    FOddRankSym: Integer;
    { Support for JAdjustedSymName for Cs, C2 and C2h }
    SymNames, JAdjustedSymNames: array of string;
    FJAdjustedSym: array of Integer;
    JAlternate: Boolean;
    function GetPointGroup(Index: integer): TAbelianPointGroup;
    function GetRepresentation(Index: integer): TRepresentation;
    function GetAxis(const Index: Integer): TabcAxis;
    procedure SetPointGroup(Index: integer; const AValue: TAbelianPointGroup);
    procedure SetRepresentation(Index: integer; const AValue: TRepresentation);
    procedure SetAxis(const Index: Integer; const AValue: TabcAxis);
  protected
    procedure MakeSpherical(MNode: TMNode; var S: TSphericalTensor; Prefix: string);
    procedure MakeSecondRank(MNode: TMNode; var S: TSecondRankTensor; Prefix: string);
  public
    StatisticalWeights, StatisticalWeightGroup: array of Integer;
    constructor Create(AOwner: TComponent); override;
    function Allowed(Rank: TQuantumNumber; Sym1, Sym2: Integer): Boolean; overload;
    function Allowed(TransitionMoment: TTransitionMoment; Sym1, Sym2: Integer): Boolean; overload; override;
    function eoSym(RveSym: Integer): integer;
    function FixupProperties: Exception; override;
    function ReadProperties(Node: TDOMNode; CheckUnread: Boolean = True):Integer; override;
    function IsValidChildClass(MNodeClass: TMNodeClass):Boolean; override;
    procedure PrecomputeSymNames; override;
    function GetSettingsNames(AIndex: Integer; out LowestValue: Integer): PPChar; override;
    function SymName(Sym: Integer):string; override;
    function JAdjustedSymNameBase(J: TQuantumNumber; Sym: Integer): string; override;
    function JAdjustedSym(J: TQuantumNumber; Sym: Integer): Integer; override;
    function WangName(J: TQuantumNumber; rSym: Integer): string;
    function LastSym: Integer; override;
    property OddRankSym: Integer read FOddRankSym; { Symmetry of dipole moment }
    function QnoInfo(QnoType: TQnoType): TQnoInfo; override;
    function HelpOnSetting(AIndex: Integer): string; override;
    function xyzAxis(abcAxis: TabcAxis): TxyzAxis; { Given a b or c, return corresponding x y or z axis }
    function abcAxis(AxyzAxis: TxyzAxis): TabcAxis; { Given x y or z, return corresponding a b or c axis }
    function FixSym(J: TQuantumNumber; Sym:Integer; InputPgopherVersion: Integer):Integer; override;
    procedure MakeeoSyms(out eoSyms: array of Integer; out eoSymsStep: Integer; VibronicSymmetry: Integer);
  published
    property PointGroup:     TAbelianPointGroup index LMI read GetPointGroup write SetPointGroup;
    property Representation: TRepresentation index LMI+1 read GetRepresentation write SetRepresentation;
    property SReduction:     Boolean  index LMI+2 read GetFlag write SetFlag; { Use Watson's S reduction (as opposed to A) }
    property eeWt:           Integer  index LMI+3 read GetSetting write SetSetting;
    property eoWt:           Integer  index LMI+4 read GetSetting write SetSetting;
    property oeWt:           Integer  index LMI+5 read GetSetting write SetSetting;
    property ooWt:           Integer  index LMI+6 read GetSetting write SetSetting;
    property C2zAxis:        TabcAxis index LMI+7 read GetAxis write SetAxis;
    property C2xAxis:        TabcAxis index LMI+8 read GetAxis write SetAxis;
    property PseudoC2v:      Boolean  index LMI+9 read GetFlag write SetFlag; { Ignore any operators breaking the C2v symmetry }
    property FakeSym:        Boolean  index LMI+10 read GetFlag write SetFlag; { Turn off transition moment test }
    property PhaseAdjust:    Boolean  index LMI+11 read GetFlag write SetFlag; { Add phase factor to minimize complex matrix elements }
  end;

  TDeltaK2TransitionMoment = class(TTransitionMoment)
    function CheckProjections(K1, K2: TQuantumNumber):Boolean; override;
  end;

  TAsymmetricTopOrbitalTransitionMoment = class(TDeltaK2TransitionMoment)
  private
    g: TSphericalTensor;
  public
    constructor Create(AOwner: TComponent); overload; override;
    function FixupProperties: Exception; override;
    function PairName: string; override;
  published
    property gsaa:    Double index 1 read GetParameter write SetParameter;
    property gsbb:    Double index 2 read GetParameter write SetParameter;
    property gscc:    Double index 3 read GetParameter write SetParameter;
    property gsabbar: Double index 4 read GetParameter write SetParameter; { (gab + gba)/2 }
    property gsacbar: Double index 5 read GetParameter write SetParameter; { (gac + gca)/2 }
    property gsbcbar: Double index 6 read GetParameter write SetParameter; { (gbc + gcb)/2 }
    property gsamb:   Double index 7 read GetParameter write SetParameter; { gab - gba }
    property gsamc:   Double index 8 read GetParameter write SetParameter; { gac - gca }
    property gsbmc:   Double index 9 read GetParameter write SetParameter; { gbc - gcb }
  end;

  TAsymmetricTopRotationalTransitionMoment = class(TDeltaK2TransitionMoment)
  private
    g: TSphericalTensor;
  public
    constructor Create(AOwner: TComponent); overload; override;
    function PairName: string; override;
    function FixupProperties: Exception; override;
  published { Note that the off diagonal terms are typically not experimentally determinable }
    property graa:    Double index 1 read GetParameter write SetParameter;
    property grbb:    Double index 2 read GetParameter write SetParameter;
    property grcc:    Double index 3 read GetParameter write SetParameter;
    property grabbar: Double index 4 read GetParameter write SetParameter; { (gab + gba)/2 }
    property gracbar: Double index 5 read GetParameter write SetParameter; { (gac + gca)/2 }
    property grbcbar: Double index 6 read GetParameter write SetParameter; { (gbc + gcb)/2 }
    property gramb:   Double index 7 read GetParameter write SetParameter; { gab - gba }
    property gramc:   Double index 8 read GetParameter write SetParameter; { gac - gca }
    property grbmc:   Double index 9 read GetParameter write SetParameter; { gbc - gcb }
  end;

const
  EmptyPertEquiv:TPertEquiv = (Name:''; JPower:0; aPower:0; bPower:0; cPower:0; pPower:0; zPower: 0; Scale4:4; SaPower:0; SbPower:0; ScPower:0; SpPower:0; SzPower:0; SPower: 0; NSPower:0);

function FindParamEquiv(const PE: TPertEquiv; var SReduction: Boolean; out OneOfPair: Boolean; out Scale: Double):string;

implementation

uses CODATA, nj, Global, Profile, ShellSort, Units, UCompat, Worker, StaticField, SymmetricTop, WignerD;

var
  AsymTTransitionMatrixElement: string = 'OAsymT Transition Matrix Element';

  eK, eN, eJ, eS: TSortedVariableNode;
  VList: TVariableList;

const
  FirstStatisticalWeight = LMI+3;

{ Two symmetry types are in use here. For the symmetry of the rotational wavefunction,
  (in the V (=D2?) group, the coding is:
    ee=00 eo=01 oe=10 oo=11
  e and o specify the oddness/evenness of Ka and Kc respectively.
  Multiplication of symmetry expressed like this is simply by bitwise xor.

  For the symmetry of the vibronic wavefunction, and the overall wavefunction, (the Symmetry variable) the coding
  depends on the point group, but multiplication of symmetries is still expressed by bitwise xor.

  The axis for which K is used in the basis is specified by the representation
  Ir: z=a and so on
  To use the same ideas for the vibronic symmetry, we use the identification (for C2v):
  R2z' = C2
  R2x' = sigma(y'z')
  R2y' = sigma(x'z')
  where yet another axis system is being introduced for the point group. This is
  defined by the postion of C2 axis (z') and the out of plane axis (x'). The
  location of these wrt the inertial axes is defined by the C2zAxis and C2xAxis
  variables.

  For D2h and D2 the C2zAxis and C2xAxis variable specify the axes corresponding to
  the C2z and C2x operators in the character table

  For Cs we have:
  R2z' = sigma(x'y')
  where C2zAxis is now the out of plane axis, and C2xAxis is unused
  C2 and C2h are effectively the same, the matching point group operator is C2

  For C1, no simplification is possible
  Ci is similar, with the proviso that all rotational functions are g

}

{ TAsymmetricTop }

const
  C2vCharacterTable: array [0..3] of array [0..3] of Integer =
   (      { E C2 s(zx)s*(yz) }
          { E (12) (12)*  E* }
          {   Rz   Ry     Rx  }
   {0:A1} ( 1, 1,   1,     1), {z}
   {1:A2} ( 1, 1,  -1,    -1),
   {2:B1} ( 1,-1,   1,    -1), {x}
   {3:B2} ( 1,-1,  -1,     1));{y}
  CsCharacterTable: array [0..1] of array [0..1] of Integer =
   (      { E s(xy) } { z out of plane }
          {    Rz   }
   {0:A'} ( 1,  1   ), {x, y}
   {1:A"} ( 1, -1)  ); {z}

  D2hCharacterTable: array [0..7] of array [0..4] of Integer =
   (       { E C2z C2y C2x i }
           {   Rz  Ry  Rx  E* }
   {0:Ag}  ( 1, 1,  1,  1,  1),
   {1:B1g} ( 1, 1, -1, -1,  1),  {Rz}
   {2:B2g} ( 1,-1,  1, -1,  1),  {Ry}
   {3:B3g} ( 1,-1, -1,  1,  1),  {Rx}
   {4:Au}  ( 1, 1,  1,  1, -1),
   {5:B1u} ( 1, 1, -1, -1, -1),  {z}
   {6:B2u} ( 1,-1,  1, -1, -1),  {y}
   {7:B3u} ( 1,-1, -1,  1, -1)); {x}
  { D2 is the same as ignoring g/u in D2h }

   C2hCharacterTable: array [0..3] of array [0..2] of Integer =
   (       { E C2z i }
           {   Rz  E* }
   {0:Ag}  ( 1, 1, 1),
   {1:Bg}  ( 1,-1, 1),
   {2:Au}  ( 1, 1,-1),  {z}
   {3:Bu}  ( 1,-1,-1)); {x,y}
  { C2 is the same as ignoring g/u in C2h }
  { Ci only has g/u, and rotations are always g }



  { Antisymmetric wrt rotation about the inertial axis implies the following symmetries: }
  AntiSymWrt: array [TabcAxis] of Integer = ( 2 {a - 10}, 3 {b - 11}, 1 {c - 01} );
  { Non totally symmetric function that is symmetric wrt the given axis: }
  SymWrt: array [TabcAxis] of Integer = ( 1 {a - 01}, 3 {b - 11}, 2 {c - 10} );
  { Mask to switch to functions that are antisymmetric wrt reflection in Cs }
  CsSwitch: array [TabcAxis] of Integer = ( 2 {a - 10}, 1 {b - 10 or 01; (arbitrary) }, 1 {c - 01} );

const
  AxisNames: array [TabcAxis] of string = ('a', 'b', 'c');

{ Mapping between x/y/z and a/b/c axes }
const                                             {   Ir,   IIr,  IIIr,    Il,   IIl,  IIIl}
  ZAxisFor: array [TRepresentation] of TabcAxis = (aAxis, bAxis, cAxis, aAxis, bAxis, cAxis);
  XAxisFor: array [TRepresentation] of TabcAxis = (bAxis, cAxis, aAxis, cAxis, aAxis, bAxis);
  YAxisFor: array [TRepresentation] of TabcAxis = (cAxis, aAxis, bAxis, bAxis, cAxis, aAxis);
  { Reverse Mapping }
  aAxisFor: array [TRepresentation] of TxyzAxis = (zAxis, yAxis, xAxis, zAxis, xAxis, yAxis);
  bAxisFor: array [TRepresentation] of TxyzAxis = (xAxis, zAxis, yAxis, yAxis, zAxis, xAxis);
  cAxisFor: array [TRepresentation] of TxyzAxis = (yAxis, xAxis, zAxis, xAxis, yAxis, zAxis);

{ TAsymmetricTopSpinSpin }

function TAsymmetricTopSpinSpin.FixupProperties: Exception;
begin
  (AncestorWithClass(TAsymmetricMolecule) as TAsymmetricMolecule).MakeSecondRank(Self, D, 'D');
  Result := inherited FixupProperties;
end;

function TAsymmetricMolecule.xyzAxis(abcAxis: TabcAxis): TxyzAxis;
{ Given a b or c, return corresponding x y or z axis }
begin
  if ZAxisFor[Representation] = abcAxis then
    Result := zAxis
  else if YAxisFor[Representation] = abcAxis then
    Result := yAxis
  else
    Result := xAxis;
end;

function TAsymmetricMolecule.abcAxis({Representation: TRepresentation;} AxyzAxis: TxyzAxis): TabcAxis;
{ Given x y or z, return corresponding a b or c axis }
begin
  if aAxisFor[Representation] = AxyzAxis then
    Result := aAxis
  else if bAxisFor[Representation] = AxyzAxis then
    Result := bAxis
  else
    Result := cAxis;
end;

procedure TAsymmetricMolecule.MakeSpherical(MNode: TMNode; var S: TSphericalTensor; Prefix: string);
  function efind(Suffix: string):Double;
  var
    i: Integer;
  begin
    i := MNode.IndexOfParameter(Prefix+Suffix);
    if i > 0 then
      Result := MNode.GetParameter(i)
    else
      Result := 0;
  end;
var
  xx, yy, zz, xy, yz, xz, xmy, xmz, ymz: Double;
  x, y, z: Char;
begin
  x := AxisNames[XAxisFor[Representation]][1];
  y := AxisNames[YAxisFor[Representation]][1];
  z := AxisNames[ZAxisFor[Representation]][1];
  xx := efind(x+x);
  yy := efind(y+y);
  zz := efind(z+z);
  xy := efind(x+y+'bar') + efind(y+x+'bar'); { Only one of the two will be non-zero }
  xz := efind(x+z+'bar') + efind(z+x+'bar');
  yz := efind(y+z+'bar') + efind(z+y+'bar');
  xmy := efind(x+'m'+y) - efind(y+'m'+x);
  xmz := efind(x+'m'+z) - efind(z+'m'+x);
  ymz := efind(y+'m'+z) - efind(z+'m'+y);

  S[0][0] := -(xx+yy+zz)/sqrt(Three);
  S[1][ 0] := -xmy/Sqrt(Two); { Imaginary part of S[1][0] }
  S[1][ 1] := xmz/2; { Real part of S[2][1] }
  S[1][-1] := ymz/2 ; { Imaginary part of S[2][1] }
  S[2][ 0] := (2*zz - xx - yy)/Sqrt(Six);
  S[2][ 1] := -xz; { Real part of S[2][1] }
  S[2][-1] := -yz; { Imaginary part of S[2][1] }
  S[2][ 2] :=  (xx - yy)/2; { Real part of S[2][2] }
  S[2][-2] := xy; { Imaginary part of S[2][2] } { TODO: Check sign change }
end;

procedure TAsymmetricMolecule.MakeSecondRank(MNode: TMNode; var S: TSecondRankTensor; Prefix: string);
var
  xy, xz, yz: Double;
  function efind(Axis1, Axis2: TabcAxis; Stem: string):Double;
  var
    s: string;
  begin
    if Axis1 < Axis2 then
      s := AxisNames[Axis1] + AxisNames[Axis2]
    else
      s := AxisNames[Axis2] + AxisNames[Axis1];
    Result := MNode.GetParameterByName(Stem+s)
  end;
begin
  xy := efind(XAxisFor[Representation], YAxisFor[Representation], Prefix);
  xz := efind(XAxisFor[Representation], ZAxisFor[Representation], Prefix);
  yz := efind(YAxisFor[Representation], ZAxisFor[Representation], Prefix);

  S[ 0].re := MNode.GetParameterByName(Prefix+'zz')/2;
  S[ 0].im := 0;
  S[ 1].re := -xz/Sqrt(Six);
  S[-1].re :=  xz/Sqrt(Six);
  S[ 1].im := -yz/Sqrt(Six);
  S[-1].im := -yz/Sqrt(Six);
  S[ 2].re :=  MNode.GetParameterByName(Prefix+'xxmyy')/Sqrt(4*Six);
  S[-2].re :=  S[2].re;
  S[ 2].im :=  xy/Sqrt(Six);
  S[-2].im := -xy/Sqrt(Six);
end;

{ For use with TSphericalTransitionMoment prepared with MakeSpherical }
procedure AddSpherical(var Result: Complex; Rank, q: Integer; { Both Undoubled } V: Double; const T: TSphericalTensor);
begin
  if q = 0 then
    if Rank = 1 then
      Result.im := Result.im + V * T[Rank][q]
    else
      Result.re := Result.re + V * T[Rank][q]
  else begin
    { Real part }
    if q < 0 then
      V := V*Minus(2*(Rank-q));
    Result.re := Result.re + V * T[Rank][Abs(q)];
    { Imaginary part }
    if q < 0 then
      V := -V;
    Result.im := Result.im + V * T[Rank][-Abs(q)];
  end;
end;

function TAsymmetricTop.FixupProperties: Exception;
var
  i: Integer;
begin
  { Force legal symmetry, in case point group changed }
  Symmetry := Min(Symmetry, Molecule.LastSym);
  FFullRotationalSymmetry := Molecule.PseudoC2v or (Molecule.PointGroup in [D2, C2v, D2h]);
  Molecule.MakeSpherical(Self, TSR, 'e'); { Convert spin rotation terms to spherical tensor form }

  { Find if any Quartic Spin Rotation }
  QuarticSpinRotation := False;
  for i := IndexOfParameter('DsN') to IndexOfParameter('ds2') do
    QuarticSpinRotation := QuarticSpinRotation or (GetParameter(i) <> 0);

  Molecule.MakeeoSyms(eoSyms, eoSymsStep, Symmetry);
  { Generate K states in order of increasing energy.
    Note that this could generate odd results if the centrifugal distortion
    is dominant, or if B ~ (A+C)/2. }
  if BBar > 0 then
    if A > BBar then
      FKinc := 4
    else
      FKinc := -4
  else
    case ZAxisFor[Molecule.Representation] of
      { Prolate - definately increasing with K }
      aAxis: FKinc := 4;
      { Oblate - definately decreasing with K }
      cAxis: FKinc := -4;
      { K is intermediate axis - depends on whether B is closer to A or C }
      bAxis: if B < (A+C)/2 then
               FKinc := -4
             else
               FKinc := 4;
    end;

  if BBar > 0 then begin
    Izz := A;
    IxxpIyy := BBar*2;
    IxxmIyy := BDelta
  end else begin
    Izz := GetParameter(LSP+Ord(ZAxisFor[Molecule.Representation]));
    Ixx := GetParameter(LSP+Ord(XAxisFor[Molecule.Representation]));
    Iyy := GetParameter(LSP+Ord(YAxisFor[Molecule.Representation]));
    IxxpIyy := Ixx + Iyy;
    IxxmIyy := Ixx-Iyy;
  end;

  VList.CheckThreadVars;
  Result := inherited FixupProperties;
  if Assigned(Result) then
    Exit;
  { TODO: Derive effective symmetries from transition moments? }
  {$IFDEF JTEST}
  JTestOrigin := 0;
  JTestOrigin := Origin - HJtest(JTest,0,0);
  {$ENDIF}
end;

function TAsymmetricTop.GetSettingsNames(AIndex: Integer; out LowestValue: Integer): PPChar;
begin
  Result := inherited GetSettingsNames(AIndex,LowestValue);
  if not Assigned(Owner) then begin { Should never happen }
    Result := IrepNames(C2v);
    Exit;
  end;
  if AIndex = LSI then
    Result := IrepNames(Molecule.PointGroup);
end;

const
  AReductionEquivs: array [0..45] of TPertEquiv = (
   (Name:'Origin'; JPower:0; aPower:0; bPower:0; cPower:0; pPower:0; zPower: 0; Scale4:4; SaPower:0; SbPower:0; ScPower:0; SpPower:0; SzPower:0; SPower: 0; NSPower:0),
   (Name:'A'; JPower:0; aPower:2; bPower:0; cPower:0; pPower:0; zPower: 0; Scale4:4; SaPower:0; SbPower:0; ScPower:0; SpPower:0; SzPower:0; SPower: 0; NSPower:0),
   (Name:'B'; JPower:0; aPower:0; bPower:2; cPower:0; pPower:0; zPower: 0; Scale4:4; SaPower:0; SbPower:0; ScPower:0; SpPower:0; SzPower:0; SPower: 0; NSPower:0),
   (Name:'C'; JPower:0; aPower:0; bPower:0; cPower:2; pPower:0; zPower: 0; Scale4:4; SaPower:0; SbPower:0; ScPower:0; SpPower:0; SzPower:0; SPower: 0; NSPower:0),
   (Name:'BBar'; JPower:0; aPower:0; bPower:2; cPower:0; pPower:0; zPower: 0; Scale4:4; SaPower:0; SbPower:0; ScPower:0; SpPower:0; SzPower:0; SPower: 0; NSPower:0),
   (Name:'BBar'; JPower:0; aPower:0; bPower:0; cPower:2; pPower:0; zPower: 0; Scale4:4; SaPower:0; SbPower:0; ScPower:0; SpPower:0; SzPower:0; SPower: 0; NSPower:0),
   (Name:'BDelta'; JPower:0; aPower:0; bPower:0; cPower:0; pPower:2; zPower: 0; Scale4:1; SaPower:0; SbPower:0; ScPower:0; SpPower:0; SzPower:0; SPower: 0; NSPower:0),

   (Name:'DJ';  JPower:4; aPower:0; bPower:0; cPower:0; pPower:0; zPower: 0; Scale4:-4; SaPower:0; SbPower:0; ScPower:0; SpPower:0; SzPower:0; SPower: 0; NSPower:0),
   (Name:'DJK'; JPower:2; aPower:0; bPower:0; cPower:0; pPower:0; zPower: 2; Scale4:-4; SaPower:0; SbPower:0; ScPower:0; SpPower:0; SzPower:0; SPower: 0; NSPower:0),
   (Name:'DK';  JPower:0; aPower:0; bPower:0; cPower:0; pPower:0; zPower: 4; Scale4:-4; SaPower:0; SbPower:0; ScPower:0; SpPower:0; SzPower:0; SPower: 0; NSPower:0),
   (Name:'deltaJ'; JPower:2; aPower:0; bPower:0; cPower:0; pPower:2; zPower: 0; Scale4:-4; SaPower:0; SbPower:0; ScPower:0; SpPower:0; SzPower:0; SPower: 0; NSPower:0),
   (Name:'deltaK'; JPower:0; aPower:0; bPower:0; cPower:0; pPower:2; zPower: 2; Scale4:-2; SaPower:0; SbPower:0; ScPower:0; SpPower:0; SzPower:0; SPower: 0; NSPower:0),

   (Name:'HJ';  JPower:6; aPower:0; bPower:0; cPower:0; pPower:0; zPower: 0; Scale4:4; SaPower:0; SbPower:0; ScPower:0; SpPower:0; SzPower:0; SPower: 0; NSPower:0),
   (Name:'HJK'; JPower:4; aPower:0; bPower:0; cPower:0; pPower:0; zPower: 2; Scale4:4; SaPower:0; SbPower:0; ScPower:0; SpPower:0; SzPower:0; SPower: 0; NSPower:0),
   (Name:'HKJ'; JPower:2; aPower:0; bPower:0; cPower:0; pPower:0; zPower: 4; Scale4:4; SaPower:0; SbPower:0; ScPower:0; SpPower:0; SzPower:0; SPower: 0; NSPower:0),
   (Name:'HK';  JPower:0; aPower:0; bPower:0; cPower:0; pPower:0; zPower: 6; Scale4:4; SaPower:0; SbPower:0; ScPower:0; SpPower:0; SzPower:0; SPower: 0; NSPower:0),
   (Name:'phiJ';  JPower:4; aPower:0; bPower:0; cPower:0; pPower:2; zPower: 0; Scale4:4; SaPower:0; SbPower:0; ScPower:0; SpPower:0; SzPower:0; SPower: 0; NSPower:0),
   (Name:'phiJK'; JPower:2; aPower:0; bPower:0; cPower:0; pPower:2; zPower: 2; Scale4:2; SaPower:0; SbPower:0; ScPower:0; SpPower:0; SzPower:0; SPower: 0; NSPower:0),
   (Name:'phiK';  JPower:0; aPower:0; bPower:0; cPower:0; pPower:2; zPower: 4; Scale4:2; SaPower:0; SbPower:0; ScPower:0; SpPower:0; SzPower:0; SPower: 0; NSPower:0),

   (Name:'LJ';   JPower:8; aPower:0; bPower:0; cPower:0; pPower:0; zPower: 0; Scale4:4; SaPower:0; SbPower:0; ScPower:0; SpPower:0; SzPower:0; SPower: 0; NSPower:0),
   (Name:'LJJK'; JPower:6; aPower:0; bPower:0; cPower:0; pPower:0; zPower: 2; Scale4:4; SaPower:0; SbPower:0; ScPower:0; SpPower:0; SzPower:0; SPower: 0; NSPower:0),
   (Name:'LJK';  JPower:4; aPower:0; bPower:0; cPower:0; pPower:0; zPower: 4; Scale4:4; SaPower:0; SbPower:0; ScPower:0; SpPower:0; SzPower:0; SPower: 0; NSPower:0),
   (Name:'LKKJ'; JPower:2; aPower:0; bPower:0; cPower:0; pPower:0; zPower: 6; Scale4:4; SaPower:0; SbPower:0; ScPower:0; SpPower:0; SzPower:0; SPower: 0; NSPower:0),
   (Name:'LK';   JPower:0; aPower:0; bPower:0; cPower:0; pPower:0; zPower: 8; Scale4:4; SaPower:0; SbPower:0; ScPower:0; SpPower:0; SzPower:0; SPower: 0; NSPower:0),
   (Name:'llJ';  JPower:6; aPower:0; bPower:0; cPower:0; pPower:2; zPower: 0; Scale4:4; SaPower:0; SbPower:0; ScPower:0; SpPower:0; SzPower:0; SPower: 0; NSPower:0),
   (Name:'llJK'; JPower:4; aPower:0; bPower:0; cPower:0; pPower:2; zPower: 2; Scale4:2; SaPower:0; SbPower:0; ScPower:0; SpPower:0; SzPower:0; SPower: 0; NSPower:0),
   (Name:'llKJ'; JPower:2; aPower:0; bPower:0; cPower:0; pPower:2; zPower: 4; Scale4:2; SaPower:0; SbPower:0; ScPower:0; SpPower:0; SzPower:0; SPower: 0; NSPower:0),
   (Name:'llK';  JPower:0; aPower:0; bPower:0; cPower:0; pPower:2; zPower: 6; Scale4:2; SaPower:0; SbPower:0; ScPower:0; SpPower:0; SzPower:0; SPower: 0; NSPower:0),

   (Name:'eaa';  JPower:0; aPower:1; bPower:0; cPower:0; pPower:0; zPower: 0; Scale4:2; SaPower:1; SbPower:0; ScPower:0; SpPower:0; SzPower:0; SPower: 0; NSPower:0),
   (Name:'ebb';  JPower:0; aPower:0; bPower:1; cPower:0; pPower:0; zPower: 0; Scale4:2; SaPower:0; SbPower:1; ScPower:0; SpPower:0; SzPower:0; SPower: 0; NSPower:0),
   (Name:'ecc';  JPower:0; aPower:0; bPower:0; cPower:1; pPower:0; zPower: 0; Scale4:2; SaPower:0; SbPower:0; ScPower:1; SpPower:0; SzPower:0; SPower: 0; NSPower:0),

   (Name:'eabbar';  JPower:0; aPower:1; bPower:0; cPower:0; pPower:0; zPower: 0; Scale4:2; SaPower:0; SbPower:1; ScPower:0; SpPower:0; SzPower:0; SPower: 0; NSPower:0),
   (Name:'eabbar';  JPower:0; aPower:0; bPower:1; cPower:0; pPower:0; zPower: 0; Scale4:2; SaPower:1; SbPower:0; ScPower:0; SpPower:0; SzPower:0; SPower: 0; NSPower:0),
   (Name:'eacbar';  JPower:0; aPower:1; bPower:0; cPower:0; pPower:0; zPower: 0; Scale4:2; SaPower:0; SbPower:0; ScPower:1; SpPower:0; SzPower:0; SPower: 0; NSPower:0),
   (Name:'eacbar';  JPower:0; aPower:0; bPower:0; cPower:1; pPower:0; zPower: 0; Scale4:2; SaPower:1; SbPower:0; ScPower:0; SpPower:0; SzPower:0; SPower: 0; NSPower:0),
   (Name:'ebcbar';  JPower:0; aPower:0; bPower:1; cPower:0; pPower:0; zPower: 0; Scale4:2; SaPower:0; SbPower:0; ScPower:1; SpPower:0; SzPower:0; SPower: 0; NSPower:0),
   (Name:'ebcbar';  JPower:0; aPower:0; bPower:0; cPower:1; pPower:0; zPower: 0; Scale4:2; SaPower:0; SbPower:1; ScPower:0; SpPower:0; SzPower:0; SPower: 0; NSPower:0),

   (Name:'DsN';   JPower:2; aPower:0; bPower:0; cPower:0; pPower:0; zPower: 0; Scale4:4; SaPower:0; SbPower:0; ScPower:0; SpPower:0; SzPower:0; SPower: 0; NSPower:1),
   (Name:'DsNK';  JPower:2; aPower:0; bPower:0; cPower:0; pPower:0; zPower: 1; Scale4:2; SaPower:0; SbPower:0; ScPower:0; SpPower:0; SzPower:1; SPower: 0; NSPower:0),
   (Name:'DsKN';  JPower:0; aPower:0; bPower:0; cPower:0; pPower:0; zPower: 2; Scale4:4; SaPower:0; SbPower:0; ScPower:0; SpPower:0; SzPower:0; SPower: 0; NSPower:1),
   (Name:'DsK';   JPower:0; aPower:0; bPower:0; cPower:0; pPower:0; zPower: 3; Scale4:2; SaPower:0; SbPower:0; ScPower:0; SpPower:0; SzPower:1; SPower: 0; NSPower:0),
   (Name:'ds1';   JPower:0; aPower:0; bPower:0; cPower:0; pPower:2; zPower: 0; Scale4:4; SaPower:0; SbPower:0; ScPower:0; SpPower:0; SzPower:0; SPower: 0; NSPower:1),
   (Name:'ds2';   JPower:0; aPower:0; bPower:0; cPower:0; pPower:2; zPower: 1; Scale4:2; SaPower:0; SbPower:0; ScPower:0; SpPower:0; SzPower:1; SPower: 0; NSPower:0),

   (Name:'alpha'; JPower:0; aPower:0; bPower:0; cPower:0; pPower:0; zPower: 0; Scale4:-4;SaPower:0; SbPower:0; ScPower:0; SpPower:-2;SzPower:0; SPower: 0; NSPower:0),
   (Name:'alpha'; JPower:0; aPower:0; bPower:0; cPower:0; pPower:0; zPower: 0; Scale4:8; SaPower:0; SbPower:0; ScPower:0; SpPower:0; SzPower:2; SPower: 0; NSPower:0),
   (Name:'beta';  JPower:0; aPower:0; bPower:0; cPower:0; pPower:0; zPower: 0; Scale4:2; SaPower:0; SbPower:0; ScPower:0; SpPower:2; SzPower:0; SPower: 0; NSPower:0)

  );

{ Parameters that are different in the S Reduction only }
  SReductionEquivs: array [0..7] of TPertEquiv = (
   (Name:'deltaJ'; JPower:2; aPower:0; bPower:0; cPower:0; pPower:2; zPower: 0; Scale4:4; SaPower:0; SbPower:0; ScPower:0; SpPower:0; SzPower:0; SPower: 0; NSPower:0),
   (Name:'deltaK'; JPower:0; aPower:0; bPower:0; cPower:0; pPower:4; zPower: 0; Scale4:4; SaPower:0; SbPower:0; ScPower:0; SpPower:0; SzPower:0; SPower: 0; NSPower:0),

   (Name:'phiJK'; JPower:2; aPower:0; bPower:0; cPower:0; pPower:4; zPower: 0; Scale4:4; SaPower:0; SbPower:0; ScPower:0; SpPower:0; SzPower:0; SPower: 0; NSPower:0),
   (Name:'phiK';  JPower:0; aPower:0; bPower:0; cPower:0; pPower:6; zPower: 0; Scale4:4; SaPower:0; SbPower:0; ScPower:0; SpPower:0; SzPower:0; SPower: 0; NSPower:0),

   (Name:'llJK'; JPower:4; aPower:0; bPower:0; cPower:0; pPower:4; zPower: 0; Scale4:4; SaPower:0; SbPower:0; ScPower:0; SpPower:0; SzPower:0; SPower: 0; NSPower:0),
   (Name:'llKJ'; JPower:2; aPower:0; bPower:0; cPower:0; pPower:6; zPower: 0; Scale4:4; SaPower:0; SbPower:0; ScPower:0; SpPower:0; SzPower:0; SPower: 0; NSPower:0),
   (Name:'llK';  JPower:0; aPower:0; bPower:0; cPower:0; pPower:8; zPower: 0; Scale4:4; SaPower:0; SbPower:0; ScPower:0; SpPower:0; SzPower:0; SPower: 0; NSPower:0),

   (Name:'ds2';   JPower:0; aPower:0; bPower:0; cPower:0; pPower:3; zPower: 0; Scale4:2; SaPower:0; SbPower:0; ScPower:0; SpPower:1; SzPower:0; SPower: 0; NSPower:0)
  );

function FindPertEquiv(const AName: string; SReduction: Boolean; Representation: TRepresentation; out PE1, PE2: TPertEquiv):TPertEquiv;
  function FindParameter(const Equivs: array of TPertEquiv):TPertEquiv;
  var
    i: Integer;
  begin
    Result.Name := '';
    for i := Low(Equivs) to High(Equivs) do
      if CompareText(Equivs[i].Name,AName) = 0 then begin
        Result := Equivs[i];
        if (i < High(Equivs)) and (CompareText(Equivs[i+1].Name,AName) = 0) then
          PE1 := Equivs[i+1] { Second parameter in constrained pair }
        else
          PE1.Name := '';
        if Result.SpPower = -2 then begin
          { SpPower = -2 is converted to x^2 (in Result) + y^2 (in PE1) }
          PE2 := PE1; { Third operator }
          Result.SpPower := 0;
          PE1 := Result;
          if ZAxisFor[Representation] <> aAxis then
            Result.SaPower := 2;
          if ZAxisFor[Representation] <> bAxis then
            if Result.SaPower = 0 then
              Result.SbPower := 2
            else
              PE1.SbPower := 2;
          if ZAxisFor[Representation] <> cAxis then
            PE1.ScPower := 2;
        end;
        Exit;
      end;
  end;
begin
  if SReduction then
    Result := FindParameter(SReductionEquivs)
  else
    Result.Name := '';
  if Result.Name = '' then
    Result := FindParameter(AReductionEquivs);
end;

function FindParamEquiv(const PE: TPertEquiv; var SReduction: Boolean; out OneOfPair: Boolean; out Scale: Double):string;
  function Scan(const Equivs: array of TPertEquiv):string;
  var
    PE1: TPertEquiv;
    i: Integer;
  begin
    Result := '';
    for i := Low(Equivs) to High(Equivs) do begin
      PE1 := Equivs[i];
      PE1.Name := PE.Name; { To allow easy compare }
      PE1.Scale4 := PE.Scale4;
      if CompareMem(@PE, @PE1, SizeOf(PE)) then begin
        Result := Equivs[i].Name;
        Scale := PE.Scale4/Equivs[i].Scale4;
        OneOfPair := ((i < High(Equivs)) and (CompareText(Equivs[i+1].Name,Result) = 0))
                     or ((i > 0) and (CompareText(Equivs[i-1].Name,Result) = 0));
        Break;
      end;
    end;
  end;
var
  i: Integer;
begin
  { Given PertEquivalent, find equivalent parameter, if any }
  OneOfPair := False;
  if SReduction then begin
    Result := Scan(SReductionEquivs);
    if Result = '' then begin
      Result := Scan(AReductionEquivs);
      if Result <> '' then begin
        for i := 0 to High(SReductionEquivs) do
          if Result = SReductionEquivs[i].Name then begin
            { Only force reduction change if one of the parameters with different operators in the two reductions }
            SReduction := False;
            Break;
          end;
      end;
    end;
  end else begin
    Result := Scan(AReductionEquivs);
    if Result = '' then begin
      Result := Scan(SReductionEquivs);
      if Result <> '' then
        SReduction := True;
    end;
  end;
end;

function TAsymmetricTop.ConvertParameter(AIndex: Integer; Zero: Boolean): TPerturbation;

  procedure SetUpPerturbation(PE: TPertEquiv);
  var
    p: TAsymmetricTopPerturbation;
  begin
    p := TAsymmetricTopPerturbation.Create(Owner);
    p.SetFromPertEquiv(PE);
    with PE do begin
      if Scale4 = -4 then
        p.Comment := '-'+Name
      else if Scale4 <> 4 then
        p.Comment := FloatToStr(Scale4/4)+Name
      else
        p.Comment := Name;
      p.Value := Scale4*Parameters[AIndex]/4;
    end;
    p.SetBraKetName(Name,Name);
    Result := p;
  end;

var
  PE, PE1, PE2: TPertEquiv;
begin
  Result := nil;
  PE := FindPertEquiv(ParameterNames[AIndex], Molecule.SReduction, Molecule.Representation, PE1, PE2);
  if PE.Name <> '' then begin
    SetUpPerturbation(PE);
    if PE1.Name <> '' then begin
      SetUpPerturbation(PE1);
      Result.ScalePrev := PE1.Scale4 div PE.Scale4;
    end;
    if PE2.Name <> '' then begin
      SetUpPerturbation(PE2);
      Result.ScalePrev := PE2.Scale4 div PE.Scale4;
    end;
    if Zero then
      Parameters[AIndex] := 0;
  end;
end;

constructor TAsymmetricTop.Create(AOwner: TComponent);
begin
  FMolecule := AOwner.Owner as TAsymmetricMolecule;
  SavedProperties := SavedProperties + ',Kmin=all,Kmax=all';
  inherited;
end;

class function TAsymmetricTop.SpinSpinType: TSpinSpinClass;
begin
  Result := TAsymmetricTopSpinSpin;
end;

type
  TBasisShellSorter = class(TShellSorter)
    Eigenvalues: TEigenvalues;
    Base: Integer;
    function InOrder(I,J:Integer):Boolean; override;
  end;

{ TBasisShellSorter }

function TBasisShellSorter.InOrder(I, J: Integer): Boolean;
var
  Diff: Integer;
begin
  Diff := Eigenvalues.Basis[I+Base+1].Qno(qtA) - Eigenvalues.Basis[J+Base+1].Qno(qtA);
  if Diff = 0 then
    Diff := - Eigenvalues.Basis[I+Base+1].Qno(qtA) + Eigenvalues.Basis[J+Base+1].Qno(qtA);
  Result := Diff <= 0;
  if not Result then
    SwapValues(Eigenvalues.Basis[I+Base+1], Eigenvalues.Basis[J+Base+1], SizeOf(TBasisState));
end;

function AntiSymmetricWrtAxis(RoSym: Integer; Axis: TabcAxis): Boolean;
{ Given axis, determine whether a function of rotational symmetry RoSym is symmetric or antisymmetric
  wrt C2 about that axis }
begin
  case Axis of
    aAxis: Result := (RoSym and 2) > 0;
    bAxis: Result := (((RoSym shr 1) xor RoSym) and 1) > 0;
    cAxis: Result := (RoSym and 1) > 0;
  end;
end;

function TAsymmetricTop.AddOneN(Eigenvalues: TEigenvalues; N, J, F, M: TQuantumNumber; Sym: Integer; HyperfineState:THyperfineBasis; Seffective: TQuantumNumber; out StatWtGroup: Integer): Integer;
var
  K, Khigh, Ka, Kc: TQuantumNumber;
  cSym, Phase, i: Integer;
  procedure Add(K: TQuantumNumber);
  var
    BasisState: TAsymmetricTopBasisState;
  begin
    if ((Kmin >= 0) and (Abs(K) < Kmin))
       or ((Kmax >= 0) and (Abs(K) > Kmax)) then
      Exit;
    {$IFDEF JTEST}
    if (JTest <> 0) and (Molecule.JAdjustedSym(J,Symmetry) <> Sym) then
      Exit;
    {$ENDIF}
    BasisState := TAsymmetricTopBasisState.Create(Eigenvalues,F,M,Sym);
    BasisState.State := Self;
    BasisState.K := K;
    BasisState.N := N;
    if K = 0 then
      BasisState.Phase := 0
    else
      BasisState.Phase := Phase;
    if Molecule.PhaseAdjust then    
      if Phase < 0 then
        BasisState.iPower := 2 + Ord(not Odd((N+K) div 2))
      else
        BasisState.iPower := Ord(Odd((N+K) div 2));
    BasisState.Ka := Ka;
    BasisState.Kc := Kc;
    BasisState.SubBasisIndex := Eigenvalues.LastSubBasisIndex;
    BasisState.J := J;
    BasisState.HyperfineState := HyperfineState;
    BasisState.S := Seffective;
    BasisState.StatWtGroup := StatWtGroup;
    BasisState.StatisticalWeight := Result;
  end;
var
  PauliPhase, EffSym: Integer;
begin
  EffSym := Sym;
  if (Length(HyperfineState) > 0) and (Molecule.PointGroup in [C2,C2v,D2,D2h,C2h] ) then begin
    { TODO: D2/D2h handling is restricted to 2 equivalent nuclei, and one specifc axis choice }
    PauliPhase := 0;
    { Look for equivalent nuclei }
    for i := 0 to High(HyperfineState) do
      if Molecule.Nuclei[i].AsNext then begin
        if PauliPhase = 0 then
          PauliPhase := 1;
          //case Molecule.PointGroup of
          //  //C2v: PauliPhase := C2vCharacterTable[Sym,1];
          //  //C2, C2h: PauliPhase := C2hCharacterTable[Sym,1];
          //else
          //  raise NotYet; { TODO: D2, D2h - more than one operation }
          //end;
        { Pair of equivalent nuclei - HyperfineState[i] is I12 = I1+I2 }
        PauliPhase := PauliPhase * Minus(HyperfineState[i]);
      end;
    if PauliPhase < 0 then
      case Molecule.PointGroup of
        C2v:     EffSym := EffSym xor 3;
        C2, C2h: EffSym := EffSym xor 1;
      else
        raise NotYet; { TODO: D2, D2h - more than one operation interchanges identical nuclei }
      end;
  end;
  StatWtGroup := Molecule.StatisticalWeightGroup[EffSym];
  Result := Molecule.StatisticalWeights[EffSym];
  if (Length(HyperfineState) < Molecule.nNuclei)  and (Molecule.PointGroup in [C2,C2v,D2,D2h,C2h] ) and (Result = 0) then begin
    { If spins not included in calculation, statistical weights need adjusting;
      for now, just make sure all states present, as this mode is only used in
      fitting. }
    { TODO: calculate statistical weight adjustment correctly }
    for i := Length(HyperfineState) to Molecule.nNuclei-1 do
      if Molecule.Nuclei[i].AsNext then
        Result := 1;
  end;
  if Result = 0 then
    Exit;
  { Not used: StartIndex := Eigenvalues.Size; }
  i := Sym;
  {$IFDEF JTEST} Extra test needed here {$ENDIF}
  while i < 8 do begin
    { Symmetry of rotational wavefunction }
    cSym := eoSyms[i];
    if cSym < 0 then Exit;
    if AntiSymmetricWrtAxis(cSym, ZAxisFor[Molecule.Representation]) then
      K := 2
    else
      K := 0;
    if AntiSymmetricWrtAxis(cSym, XAxisFor[Molecule.Representation]) xor (Minus(N) < 0) then
      Phase := -1
    else
      Phase := 1;
    if Minus(N-K) < 0 then
      Khigh := N-2
    else
      Khigh := N;
    if Kinc < 0 then
      K := Khigh;
    { Work out Ka and Kc for lowest state }
    { Kc is simply J or J-1 to match xo or xe }
    if (Minus(N) < 0) xor ((cSym and 1) <> 0) then
      Kc := N-2
    else
      Kc := N;
    { Ka is 1 for oe or oo }
    if (cSym and 2) <> 0 then
      Ka := 2 { oe or oo }
    else begin { ee or eo - use the requirement that Ka+Kc = J or J+1 }
      Ka := N - Kc;
      if Minus(Ka) < 0 then
        Inc(Ka, 2);
    end;
    while (K <= N) and (K >= 0) do begin
      if (K <> 0) or (Phase = 1) then begin
        { E- matrix has K = 0 missing }
        Add(K);
        Inc(Ka,4);
        Dec(Kc,4);
      end;
      Inc(K,Kinc);
    end;
    Inc(i, eoSymsStep);
    Inc(Eigenvalues.LastSubBasisIndex);
  end;
  {$IFDEF NOTUSED}
  if nRotSym <> 4 then begin
    { Sort in standard asymmetric top order }
    { This makes it more likely that (near) degenerate states with no matrix
      elements between them will end up mixed due to numerical errors in the
      LAPACK diagonalizer, so deleted for the time being }
    Eigenvalues.Fixup;
    Sorter := TBasisShellSorter.Create;
    Sorter.Eigenvalues := Eigenvalues;
    Sorter.Base := StartIndex;
    Sorter.Sort(Eigenvalues.Size-StartIndex);
    for i := StartIndex to Eigenvalues.Size-1 do
      Eigenvalues.Basis[i+1].ComponentIndex := i;
    Sorter.Free;
  end;
  {$ENDIF}
end;

function TAsymmetricTop.AddOneJ(Eigenvalues: TEigenvalues; J, F, M: TQuantumNumber; Sym: Integer; HyperfineState:THyperfineBasis; Seffective: TQuantumNumber): Integer;
var
  N: TQuantumNumber;
  StatWt, LastStatWtGroup: Integer;
  StatWtGroup: SmallInt;
begin
  Result := -1;
  StatWtGroup := -1;
  N := Abs(J-Seffective);
  while N <= J+Seffective do begin
    StatWt := AddOneN(Eigenvalues,N,J,F,M,Sym,HyperfineState,Seffective, LastStatWtGroup);
    if StatWtGroup < 0 then begin
      StatWtGroup := LastStatWtGroup;
      Result := StatWt;
    end else if StatWtGroup <> LastStatWtGroup then
      raise Exception.Create('AddBasis: Inconsistent Statistical Weights in '+Name);
    Inc(N,2);
  end;
end;

function TAsymmetricTop.HWidth(J, Kp, K: TQuantumNumber): Double;
{ Linewidth parameters - as a first stab, go for <Ka^2>, etc. }
{ TODO: This is straight copy of HRot - can this be generalised? }
var
  Izz, Iyy, Ixx, JSq, KSq, IxxpIyy, IxxmIyy: Double;
begin
  if K = Kp then begin
    Izz := GetParameter(FirstWidthPar+Ord(ZAxisFor[Molecule.Representation]));
    Ixx := GetParameter(FirstWidthPar+Ord(XAxisFor[Molecule.Representation]));
    Iyy := GetParameter(FirstWidthPar+Ord(YAxisFor[Molecule.Representation]));
    JSq := J*(J+2) / 4;
    KSq := K*K / 4;
    IxxpIyy := Ixx + Iyy + wBar*2;
    Result := Width + Izz*KSq  + IxxpIyy*(JSq - KSq)/2;
  end else if Abs(K-Kp) = 4 then begin
    Ixx := GetParameter(FirstWidthPar+Ord(XAxisFor[Molecule.Representation]));
    Iyy := GetParameter(FirstWidthPar+Ord(YAxisFor[Molecule.Representation]));
    IxxmIyy := Ixx-Iyy + wDelta;
    JSq := J*(J+2) / 4;
    Result := IxxmIyy/4 * Sqrt((JSq - K*(K+Kp)/8)*(JSq - Kp*(K+Kp)/8));
  end else
    Result := 0;
end;

function TAsymmetricTop.HRotA(J, Kp, K: TQuantumNumber): Double;
{ Hamiltonian using Watson's A reduction }
var
  JSq, KSq, KpSq: Double;
begin
  if K = Kp then begin
    JSq := J*(J+2) / 4;
    KSq := K*K / 4;
    Result := LJ*Sqr(Sqr(JSq)) + LJJK*Sqr(JSq)*JSq*KSq + LJK*Sqr(JSq)*Sqr(KSq)
              + LKKJ*JSq*KSq*Sqr(KSq) + LK*Sqr(Sqr(KSq));
    Result := Result + HJ*Sqr(JSq)*JSq + HJK*Sqr(JSq)*KSq + HKJ*Jsq*Sqr(KSq) + HK*Sqr(KSq)*KSq;
    Result := Result - DJ*Sqr(JSq) - DJK*JSq*KSq - DK*Sqr(KSq);
    Result := Result + Izz*KSq  + IxxpIyy*(JSq - KSq)/2 + Origin;
  end else if Abs(K-Kp) = 4 then begin
    JSq := J*(J+2) / 4;
    KSq := K*K / 4;
    KpSq := Kp*Kp / 4;
    Result := llJ*JSq*Sqr(JSq) + llJK*Sqr(JSq)*(KSq+KpSq)/2 + llKJ*Jsq*(Sqr(KSq)+Sqr(KpSq))/2 + llK*(Sqr(KSq)*KSq+Sqr(KpSq)*KpSq)/2;
    Result := Result + phiJ*Sqr(JSq) + phiJK*Jsq*(KSq+KpSq)/2 + phiK*(Sqr(KSq)+Sqr(KpSq))/2;
    Result := Result + IxxmIyy/4 - deltaJ*JSq - deltaK*(KSq+KpSq)/2;
    Result := Result * Sqrt((JSq - K*(K+Kp)/8)*(JSq - Kp*(K+Kp)/8));
  end else
    Result := 0;
end;

function TAsymmetricTop.HRotS(J, Kp, K: TQuantumNumber): Double;
{ Hamiltonian using Watson's S reduction }
var
  JSq: Double;
begin
  if K = Kp then
    Result := HRotA(J, KP, K)
  else if Abs(K-Kp) = 4 then begin
    JSq := J*(J+2) / 4;
    Result := IxxmIyy/4 + (deltaJ + (phiJ + llJ*JSq)*JSq)*JSq;
    Result := Result * Sqrt((JSq - K*(K+Kp)/8)*(JSq - Kp*(K+Kp)/8));
  end else if Abs(K-Kp) = 8 then begin
    JSq := J*(J+2) / 4;
    Result := (deltaK + (phiJK + llJK*JSq)*JSq)*rl(J, K, Kp)
  end else if Abs(K-Kp) = 12 then begin
    JSq := J*(J+2) / 4;
    Result := (phiK + llKJ*JSq)*rl(J, K, Kp)
  end else if Abs(K-Kp) = 16 then
    Result := llK*rl(J, K, Kp)
  else
    Result := 0;
end;

{$IFDEF JTEST}
function TAsymmetricTop.HJtest(J, Kp, K: TQuantumNumber): Double;
var
  JSq: Double;
begin
  if K = Kp then begin
    JSq := J*(J+2) / 4;
    Result := JTestOrigin + B*JSq - DJ*Sqr(JSq) + HJ*Sqr(JSq)*JSq + LJ*Sqr(Sqr(JSq))
  end else
    Result := 0;
end;
{$ENDIF}

function TAsymmetricTop.HSR(J, Np, Kp, N, K:TQuantumNumber): Complex;
var
  i, rank, q: Integer;
  r, r1, Ksq, Nsq, sgn, V: Double;
  NN: TQuantumNumber;
  function Gamma(N, S, J: TQuantumNumber):Double;
  begin
    Result := (J*(J+2) - N*(N+2) - S*(S+2))/4;
  end;
  function phi(N,S,J:TQuantumNumber):Double;
  begin
    Result := (N - J + S);
    Result := Result * (N + J + S + 2) * (S + J - N + 2);
    Result := Sqrt(Result * (N + J - S) / (N - One) / (N + 1) / 16);
  end;
begin
  Result := z0;
  if Abs(Np-N) > 4 then Exit;
  if S > 1 then begin
    { Matrix elements of Spin-Spin Coupling from Hirota eq 2.3.36, 2.3.38 }
    { Adapted taking into account equation 38 of T.J Sears, Comp. Phys. Rep, 2, 1 (1984) }
    { The reduced matrix element <S||T2(S)||S> is also included. Hirota assumes this is 1, which
      is appropriate for a triplet state while Sears assumes a more complex dependence depending
      on the specific make-up of the spins, which will (presumably) also be 1 for a triplet state.
      The values for the lowest few multiplicies are:
        S  <S||T2(S)||S>^2
        1      1
        1.5    6
        2     21
        2.5   56
        3    126
    }
    q := Kp-K;
    case q of
      0: Result.re := Sqrt(Six)*alpha;    // Before 7.1.323 this was Sqrt(5*Six)
      4, -4: Result.re := beta;           // Before 7.1.323 this was Sqrt(Six)
    end;
    if Result.re <> 0 then
      Result.re := Result.re * Minus(N+S+J+Np-Kp) * sixj(J, S, Np, 4, N, S)
                             * Sqrt((N+1)*(Np+One)) * threej(Np, -Kp, 4, q, N, K)
                             * ReducedMatrixElement(4, S);
  end;
  if Abs(Np-N) > 2 then Exit;
  r := Sqrt(S*(S+2)*(S+1)*(N+1)*(Np+One)/4)*Minus(J+S+Np)*sixj(N, S, J, S, Np, 2);
  if Abs(K-Kp) <= 4 then
    for i := 0 to 1 do begin
      { Brown and Sears, JMS 75, 111 (1979) suggests the antisymmetric components
        (rank = 1) do not contribute. However, it does not seem to be possible to
        reproduce a simulation done with all 9 components non-zero with just 6 components,
        as suggested by the paper. (For the prolate example used, the antisymmetic components
        seem to be required to lift the degeneracy between the +-Ka levels at high K.)
        The antisymmetric components can be added using a pair of perturbations for each,
        with ScaleNext = -1. The antisymmetric components do not produce values
        consistent with SPCAT/SPFIT; this seems to be due to an inconsistent phase
        difference in some off-diagonal matrix elements where both (eab+eba) and (eab-eab)
        contribute to the same matrix element. }
      Rank := 4*i;
      q := (Kp - K) div 2;
      if Abs(2*q) <= rank then begin
        r1 := (Minus(rank)*Sqrt(N*(N+2)*(N+One)/4)*sixj(2, 2, rank, Np, N, N)
               +Sqrt(Np*(Np+2)*(Np+One)/4)*sixj(2, 2, Rank, N, Np, Np))*Sqrt(Rank+One)/2;
        V := r*r1*Minus(Np-Kp)*threej(Np,-Kp, rank, 2*q, N, K);
        AddSpherical(Result,2*i,q,V,TSR);
      end;
      { Seems to be a phase problem with the delta N = +/1 delta K +/- 1 matrix elements,
        compared with ACDD. However, it checks OK with explicit evaluation with Maple,
        and ACDD seems to have a spcific problem with these matrix elements, in that they
        have not had their sign changed with respect to W. T. Raynes, JCP 41, 3020 (1964),
        and the ACDD documentation suggests that they should }
    end;
  { Quartic spin rotation - Matrix elements from J. M. Brown and T. J. Sears, JMS, 75, 111 (1979) }
  if QuarticSpinRotation then begin
    Nsq := N*(N+2)/4;
    if Np = N then begin
      if N <> 0 then begin
        if K = Kp then begin
          Ksq := Sqr(K)/4;
          Result.re := Result.re + Gamma(N,S,J)/2/Nsq*( DsK*Sqr(Ksq) + (DsNK+DsKN)*Ksq*Nsq + DsN*Sqr(Nsq) );
        end else if Abs(K-Kp) = 4 then
          if Molecule.SReduction then
            Result.re := Result.re + ds1/2 * Gamma(N,S,J) * rl(N,K,Kp)
          else
            Result.re := Result.re + (ds1{=dsN}*Nsq +  ds2{=dsK}*(Sqr(K)+Sqr(Kp))/8) * Gamma(N,S,J)/2/Nsq * rl(N,K,Kp)
        else if (Abs(K-Kp) = 8) and Molecule.SReduction then
          Result.re := Result.re + ds2/2 * Gamma(N,S,J)/Nsq * rl(N,K,Kp);
      end;
    end else { Abs(N-Np) = 2 }
      if K = Kp then begin
        NN := Max(N,Np);
        Ksq := Sqr(K)/4;
        Result.re := Result.re - K/2/NN * (DsK*Ksq + DsNK*Sqr(NN)/4)*phi(NN,S,J)*Sqrt(((Sqr(NN))-Sqr(K))/Four);
      end else if (Abs(K-Kp) = 4) and not Molecule.SReduction then begin
        NN := Max(N,Np);
        sgn := (Kp-K) div 4;
        if NN = Np then begin
          { <N+1|HSR|N> matrix element wanted - swap }
          K := Kp;
          sgn := -sgn;
        end;
        Result.re := Result.re - ds2{=dsK} / NN / 2 * phi(NN,S,J)
                                 *(K*(NN+sgn*K) + (K+sgn*4)*(NN+sgn*K+4))
                                 *Sqrt((NN-sgn*K-Two) * (NN+sgn*K+2) * (NN-sgn*K-4) * (NN-sgn*K)) / 16;
      end else if (Abs(K-Kp) = 8) and Molecule.SReduction then begin
        NN := Max(N,Np);
        sgn := (Kp-K) div 8;
        if NN = Np then begin
          { <N+1|HSR|N> matrix element wanted - swap }
          K := Kp;
          sgn := -sgn;
        end;
        Result.re := Result.re - sgn * ds2 / NN * phi(NN,S,J)
                                 *Sqrt((NN-sgn*K) * (Sqr(NN)-Sqr(K+Two*sgn)) * (Sqr(NN)-Sqr(K+4*sgn))
                                       * (Sqr(NN)-Sqr(K+6*sgn)) * (NN-sgn*K-8) / 256 )
      end;
  end;
end;

function TAsymmetricTop.HDip(TransitionMoment: TSingleTransitionMoment; Jp, Kp, J, K: TQuantumNumber): Complex;
var
  q: TQuantumNumber;
  Phase: Integer;
  IsComplex: Boolean;
begin
  Result.re := TransitionMoment.Strength;
  Result.im := 0;
  Phase := 0;
  IsComplex := False;
  if TransitionMoment is TSphericalTransitionMoment then begin
    q := TSphericalTransitionMoment(TransitionMoment).ActualComponent;
    { To match symmetries, treat + and - q as actually T(k,|q|) +- T(k,-|q|) }
    if (Kp < K) and (q < 0) then
      Phase := 2;
  end else if TransitionMoment is TCartesianTransitionMoment then begin
    case Molecule.xyzAxis(TCartesianTransitionMoment(TransitionMoment).Axis) of
      zAxis: q := 0;
      yAxis: begin; q := 2; IsComplex := True; Phase := 2; end;
      xAxis: begin
               q := 2;
               if K < Kp then
                 Phase := 2;
             end;
    end;
    if q <> 0 then
      Result.re := Result.re / Sqrt(Two);
  end else
    raise NotYet;
  if Abs(q) = Abs(Kp-K) then begin
    Result.re := Result.re * Sqrt((J+1)*(Jp+1))
                           * Minus(Jp-Kp+phase) * threej(Jp, -Kp, TransitionMoment.Rank, Kp-K, J, K);
    if IsComplex then begin
      Result.im := Result.re;
      Result.re := 0;
    end;
  end else
    Result.re := 0;
end;

function HDipMatrixElements(TransitionMoment: TTransitionMoment; IncompleteOK: Boolean): TMatrixElements;
var
  q, qInitial, dN: Integer;
  nMatrixElements: Integer;
  AlternateSign: Boolean;
  Phase, E, eNp, eKp: TExprNode;
begin
  Phase := nil;
  AlternateSign := False;
  if TransitionMoment is TSphericalTransitionMoment then begin
    q := TSphericalTransitionMoment(TransitionMoment).ActualComponent div 2;
    { To match symmetries, treat + and - q as actually T(k,|q|) +- T(k,-|q|) }
    AlternateSign := q < 0;
  end else if TransitionMoment is TCartesianTransitionMoment then begin
    case (TransitionMoment.AncestorWithClass(TMolecule) as TAsymmetricMolecule).xyzAxis(TCartesianTransitionMoment(TransitionMoment).Axis) of
      zAxis: q := 0;
      yAxis: begin;
               q := -1;
               Phase := EB([-1, 'sqrt', '*', -1]);
             end;
      xAxis: begin
               q := -1;
               Phase := TIntegerNode.Create(-1);
               AlternateSign := True;
             end;
    end;
    if q <> 0 then
      Phase := Expr.EN([Phase, '/', EB([2, 'sqrt'])],[],1)
  end else
    raise NotYet;
  if not IncompleteOK and (TransitionMoment.ActiveTransitionFunction <> nil) then
      raise NotYet;
  if Phase = nil then
    Phase := TIntegerNode.Create(1);
  nMatrixElements := TransitionMoment.Rank + 1;
  if q <> 0 then
    nMatrixElements := nMatrixElements*2;
  qInitial := -Abs(q);
  SetLength(Result,nMatrixElements);
  nMatrixElements := 0;
  for dN := -TransitionMoment.Rank div 2 to TransitionMoment.Rank div 2 do begin
    eNp := EB([eN,'+',dN]);
    q := qInitial;
    repeat
      eKp := EB([eK,'+',q]);
      E := Expr.EN([EB([2, '*', eNp, '+', 1]), '*', EB([2, '*', eN, '+', 1]), 'sqrt']);
      E := Expr.EN([-1, '^', EB([eNp, '-', eKp]), '*', E]);
      E := Expr.EN([Expr.EN(['threej', eNp.Duplicate, MonadicMinus.CreateExprNode(eKp), TransitionMoment.Rank div 2, q, eN.Duplicate, eK.Duplicate]),
                   '*', E, '*', Phase.Duplicate]);
      if AlternateSign and (q < 0) then
        E := Expr.EN([E, '*', -1]);
      if Assigned(E) then begin
        E := E.Simplify;
        Result[nMatrixElements].Expr := E;
        Result[nMatrixElements].Evaluator := nil;
        Result[nMatrixElements].CEvaluator := nil;
        SetLength(Result[nMatrixElements].Signature,2);
        Result[nMatrixElements].Signature[0] := q; // Component = Delta K
        Result[nMatrixElements].Signature[1] := dN;
        Inc(nMatrixElements);
      end;
      q := -q;
    until q <= 0;
    eNp.Free;
  end;
  Phase.Free;
  SetLength(Result, nMatrixElements);
end;

function TAsymmetricTop.HyperfineMatrixElement(ABra, AKet: TBasisState; Signp,
  Sign: Integer; MNode: TMNode; RankWanted: TQuantumNumber): Complex;
var
  Bra, Ket: TAsymmetricTopBasisState;
  Nuc, Nuc1: TAsymmetricTopNucleus;
  SpinSpin: TAsymmetricTopSpinSpin;
  JP, J, Np, N, Kp, K, S, Sp, I12, I12p: TQuantumNumber;
  t, Sum: Double;
  Rank, q: Integer;
begin
  Bra := ABra as TAsymmetricTopBasisState;
  Ket := AKet as TAsymmetricTopBasisState;
  KP := SignP*Bra.K;
  K := Sign*Ket.K;
  Result := z0;
  if Abs(K-Kp) > 4 then
    Exit;
  JP := Bra.J;
  Sp := Bra.S;
  J := Ket.J;
  NP := Bra.N;
  N := Ket.N;
  S := Ket.S;
  if MNode is TNucleus then begin
    Nuc := MNode as TAsymmetricTopNucleus;
    if RankWanted = 2 then begin
      { Magnetic Dipole }
      if (Np = N) and (Kp=K) then
        Result.re := Result.re + Minus(N+Sp+JP+2)
                        * sixj(Sp, Jp, N, J, Sp, 2) * Nuc.aF;
      if Abs(K-Kp) <= 4 then begin
        t := ninej(Np, N, 4, Sp, Sp, 2, Jp, J, 2); { FPC Code generation problem? }
        t := t * threej(Np, -Kp, 4, Kp-K, N, K);
        t := - t * Minus(Np-Kp) * Sqrt(30 * (N+1)*(NP+One));
        Result.re := Result.re + t * Nuc.TT[(Kp-K) div 2].re;
        Result.im := Result.im + t * Nuc.TT[(Kp-K) div 2].im;
      end;
      CMul(Result,Sqrt((J+1)*(JP+One)) * ReducedMatrixElement(2, Sp) * ReducedMatrixElement(2, Nuc.Spin));
      { Nuclear Spin - Rotation }
      Sum := 0;
      for Rank := 0 to 1 do begin { Actually a flag: Rank = 1 is k=2 }
        t := 0;
        if K = Kp then
          if Rank = 0 then
            t := Nuc.TC00
          else
            t := Nuc.TC20
        else if Abs(K-Kp) = 4 then
          t := Nuc.TC22;
        if t <> 0 then
          Sum := Sum
               + t * Sqrt(4*Rank+One) * Minus(Np-Kp)
                 * threej(Np, -Kp, Rank*4, Kp-K, N, K)
                 *(ReducedMatrixElement(2,N)*sixj(Rank*4, 2, 2, N, Np, N)
                   +ReducedMatrixElement(2,Np)*sixj(Rank*4, 2, 2, Np, N, Np))
                   { TODO: Line above is inconsistent with Hirota 2.3.82; check another source }
                   { It does check out OK on OIO test data though }
      end;
      if Sum <> 0 then
        Result.re := Result.re - Sum / 2 * sixj(N, J, Sp, Jp, Np, 2)
                                 * Sqrt((J+1)*(JP+1)*(N+One)*(Np+One))
                           * Minus(N+Sp+J) * ReducedMatrixElement(2, Nuc.Spin)
    end else if RankWanted = 4 then begin
      { Quadrupole }
      if Abs(K-Kp) <= 4 then begin
        t := sixj(NP, JP, Sp, J, N, 4)* threej(Np, -Kp, 4, Kp-K, N, K);
        t := t * Sqrt((N+1)*(NP+One)) / 2
               * Minus(NP+S+J+Np-Kp)
               * Sqrt((J+1)*(JP+One)) * Nuc.QuadI;
        Result.re := Result.re + t * Nuc.CHI[(Kp-K) div 2].re;
        Result.im := Result.im + t * Nuc.CHI[(Kp-K) div 2].im;
      end;
    end else
      raise NotYet
  end else if MNode is TAsymmetricTopSpinSpin then begin
    if (S <> 0) or (SP <> 0) then
      raise NotYet;
    SpinSpin := MNode as TAsymmetricTopSpinSpin;
    Nuc := SpinSpin.Bra as TAsymmetricTopNucleus;
    Nuc1 := SpinSpin.Ket as TAsymmetricTopNucleus;
    if Nuc.NodeIndex > Nuc1.NodeIndex then begin
      Nuc1 := Nuc;
      Nuc := SpinSpin.Ket as TAsymmetricTopNucleus;
    end;
    if Nuc.AsNext then begin
      { I12 = I1 + I2 Coupling scheme }
      I12P := Bra.HyperfineState[Nuc.NodeIndex];
      I12  := Ket.HyperfineState[Nuc.NodeIndex];
      if RankWanted = 0 then begin
        if (I12 = I12P) and (K = KP) then
          Result.re := SpinSpin.J*(I12*(I12+2) - Nuc.Spin*(Nuc.Spin+2) - Nuc1.Spin*(Nuc1.Spin+2))/8;
      end else if RankWanted = 4 then begin
        if Abs(K-Kp) <= 4 then begin
          { W. G. Read and W. H. Flygare, JCP 76, 2238 (1982); doi: 10.1063/1.443297 }
          { See also Carrington and Brown 8.30 - 8.33; sixj part handled elsewhere }
          q := (Kp-K) div 2;
          if not IsZero(SpinSpin.D[q]) then begin
            t := SpinSpin.Factors[I12P-SpinSpin.I12Min, I12-SpinSpin.I12Min];
            if t <> 0 then begin
              t := Sqrt(Six) * Minus(NP-Kp) * Sqrt((NP+1)*(N+One))
                   * threej(NP, -KP, 4, 0, N, K);
              Result.re := t * SpinSpin.D[q].re;
              Result.im := t * SpinSpin.D[q].im;
            end;
          end;
        end;
        { TODO: Could also have Delta Lambda = 2 term }
      end else
        raise NotYet;
    end else
      { F1 = J + I1; F = F1 + I2 coupling scheme }
      raise NotYet;
  end else
    raise NotYet;
end;

function TAsymmetricTop.IsSettingQuantum(AIndex: Integer): Boolean;
begin
  Result := AIndex in [LSI+1..LSI+3];
  if not Result then
    Result := inherited IsSettingQuantum(AIndex);
end;

function TAsymmetricTop.BumpBranchQno(Branch: string; AIncrement: TQuantumNumber; QnoType: TQnoType): string;
begin
  Result := ''; { Default is to indicate increment not possible }
  case QnoType of
    //qtN: ;
    qtA: Result := BumpBeforeBracketedBranch(Branch, AIncrement, 1);
    qtB: Result := BumpBeforeBracketedBranch(Branch, AIncrement, 0);
  else
    Result := inherited BumpBranchQno(Branch, AIncrement, QnoType);
  end;
end;

function TAsymmetricTop.ReadBranchFormat(Text: string; var Obs: TObservation;
  UpperState: TState; Reader: TObservationsReaderBase): Boolean;
var
  Ka, Kc, DelK, DelJ, JU, JL: TQuantumNumber;
  p, pe, Sym: Integer;
  Lname: string;
  HyperfineL, HyperfineU: THyperfineBasis;
  HasFields: Boolean;
  AlternateSym: ShortInt;
begin
  Result := False;
  Obs.Name := ' ' + Text;
  Text := Trim(Text);
  if not IsDeltaJName(Text[1],DelK) then
    Exit
  else
    Delete(Text,1,1);
  if (Length(Text) = 0) or not IsDeltaJName(Text[1],DelJ) then begin
    DelJ := DelK;
    DelK := 0;
  end else
    Delete(Text,1,1);
  { Now require Ka, Kc }
  p := Pos(',',Text);
  if p <= 0 then
    Exit; { TODO: Could assume either possible Kc? }
  { TODO: Check for two commas => spin }
  if Reader.ReadQuantumNumber(Ka,Copy(Text,1,p-1)) then
    Exit; { Bad quantum number }
  Delete(Text,1,p);
  p := Pos('(',Text);
  pe := Pos(')',Text);
  if (p > 1) and (pe > p+1)
     and not Reader.ReadQuantumNumber(Kc,Copy(Text,1,p-1))
     and not Reader.ReadQuantumNumber(Obs.JL,Copy(Text,p+1,pe-p-1))
      then begin
    Delete(Text,1,pe);
    Obs.JU := Obs.JL+DelJ;
    JL := Obs.JL;
    JU := Obs.JU;
    HasFields := IsNonZero(Obs.ExField);
    if HasFields or (Molecule.nNuclei > Reader.RemoveSpins) then begin
      { Some hyperfine structure to be found }
      if ReadHyperfineBranch(Text, Obs, HyperfineU, HyperfineL, Reader, HasFields) then
        Exit;
    end else if Text <> '' then
      Exit;
    Lname := Manifold.FindStateLabel(Obs.IndexL, Obs.SymL, AlternateSym, Obs.JL,
                                     [Ka,Kc], Obs.QuantumNumberFormat, Obs.ExField, -1, JL, -1, Self, Reader.RemoveSpins, HyperfineL);
    if Obs.SymL >= 0 then begin
      for Sym := 0 to Molecule.LastSym do
        if Molecule.Allowed(Reader.BranchRank, Sym, Obs.SymL) then begin
          Obs.SymU := Sym;
          Break;
        end;
      Obs.Name := Obs.Name + ' : '
                  + UpperState.Manifold.FindStateLabel(Obs.IndexU, Obs.SymU, AlternateSym, Obs.JU,
                                                       [Ka+DelK,Low(TQuantumNumber)], Obs.QuantumNumberFormat, Obs.ExField, Obs.SymU, JU, -1, UpperState, Reader.RemoveSpins, HyperfineU)
                  + ' - ' + Lname;

      if Obs.SymU >= 0 then begin
        Result := True;
        Obs.IndexU := 0;
        Obs.IndexL := 0;
        Exit;
      end;
    end else
      Obs.Name := Obs.Name + Lname
  end;
  if Result then
    Reader.ReportError('Bad Branch format');
end;

function TAsymmetricTop.ReadProperties(Node: TDOMNode; CheckUnread: Boolean): Integer;
begin
  Result := inherited ReadProperties(Node, CheckUnread);
  if (ReadVersion <= 7001323) and ((alpha <> 0) or (beta <> 0)) then begin
    alpha := alpha*Sqrt(5);
    beta := beta*Sqrt(Six);
    {$IFNDEF CONSOLE}
      ShowInfo('Warning: definition of alpha and beta has changed; values have been changed to reflect this in '+UniqueName);
    {$ENDIF}
  end;
end;

function TAsymmetricTop.RotationalMatrixElement(ABra, AKet: TBasisState; SignA, SignB: Integer; MNode: TMnode; Flag: Integer): Complex;
var
  Bra, Ket: TAsymmetricTopBasisState;
  K, Kp: TQuantumNumber;
  H: function (J, Kp, K: TQuantumNumber): Double of object;
begin
  Bra := ABra as TAsymmetricTopBasisState;
  Ket := AKet as TAsymmetricTopBasisState;
  if Bra.J <> Ket.J then
    raise Exception.Create('TAsymmetricTop.RotationalMatrixElement');
  Kp := SignA*Bra.K;
  K := SignB*Ket.K;
  if (Flag = Ord(meHrot)) { TODO: Spin dependent width model? }
     and (Bra.S <> 0) and (Abs(Bra.N-Ket.N) <= 4) then begin
    { TODO: C1 symmetry requires all four matrix elements or explicit test for identical phase? }
    { Spin rotation can connect states with different Wang symmetries, if molecular symmetry is sufficiently low }
    Result := HSR(Bra.J, Bra.N, Kp, Ket.N, K);
  end else
    Result := z0;
  if Bra.N = Ket.N then begin
    case TMatElemType(Flag) of
      meHrot:  {$IFDEF JTEST}
               if JTest <> 0 then
                 H := {$IFDEF FPC}@{$ENDIF}HJtest
               else
               {$ENDIF}
               if Molecule.SReduction then
                 H := HrotS
               else
                 H := HrotA;
      meWidth: H := HWidth;
    else
      raise NotYet;
    end;
    Result.re := Result.re + H(Bra.N, Kp, K);
  end;
end;

function TAsymmetricTop.RotationalTransitionMatrixElement(ABra, AKet: TBasisState; SignA, SignB: Integer; TransitionMoment: TRootedMNodePair; F: TQuantumNumber; out dK: TQuantumNumber):Complex;
  function grHelp(Np, N, k: TQuantumNumber):Double;
  begin
    Result := sixj(Np, N, 2, 2, k, N) * (N+1) * Sqrt(N*(N+2)*(Np+One)/4)
  end;
var
  Bra, Ket: TAsymmetricTopBasisState;
  Rank, q: Integer;
  V: Double;
  Kp, K, Kpp: TQuantumNumber;
  Z: Complex;
  TM: TTransitionMoment;
begin
  Bra := ABra as TAsymmetricTopBasisState;
  Ket := AKet as TAsymmetricTopBasisState;
  Kp := SignA*Bra.K;
  K := SignB*Ket.K;
  dK := Kp - K;
  if TransitionMoment is TPerturbation then begin
    TM := TransitionMoment.Owner as TTransitionMoment;
    raise NotYet;
  end else
    TM := TransitionMoment as TTransitionMoment;
  if TM.Magnetic then begin
    Result := z0;
    q := (Kp - K) div 2;
    if TM is TAsymmetricTopOrbitalTransitionMoment then begin
      for Rank := 0 to 2 do
        if Abs(q) <= Rank then begin { Hirota 2.5.16-17 }
          V := -Minus(2*Rank)*Sqrt(2*Rank+One)*Sqrt((Bra.J+1)*(Ket.J+1))
               * ninej(Bra.N, Ket.N, 2*Rank, Bra.S, Ket.S, 2, Bra.J, Ket.J, 2)
               * Minus(Bra.N-Kp) * threej(Bra.N, -Kp, 2*Rank, Kp-K, Ket.N, K)
               * Sqrt((Bra.N+1)*(Ket.N+1)) * ReducedMatrixElement(2,Bra.S);
          AddSpherical(Result,Rank,q,V,TAsymmetricTopOrbitalTransitionMoment(TM).g);
        end;
      Exit; { Acts on S, so don't want code below }
    end else if TM is TAsymmetricTopRotationalTransitionMoment then begin
      for Rank := 0 to 2 do
        if Abs(q) <= Rank then begin { Hirota 2.5.18 }
          V := Minus(2*Rank)*Sqrt(2*Rank+One) { Hirota 2.5.16 }
               / -2
               * Minus(Ket.N+Kp) * threej(Bra.N, -Kp, 2*Rank, Kp-K, Ket.N, K)
               * (grHelp(Bra.N, Ket.N, 2*Rank) + Minus(2*Rank)*grHelp(Ket.N, Bra.N, 2*Rank));
          AddSpherical(Result,Rank,q,V,TAsymmetricTopRotationalTransitionMoment(TM).g);
        end;
{ Don't use plain spin transition moment as this has the same matrix elements as the T(0,0) component of the
  OrbitalTransitionMoment times -1/Sqrt(3) }
//    end else if TransitionMoment is TSpinTransitionMoment then begin
//      if (q = 0) and (Bra.S = Ket.S) and (Bra.N = Ket.N) then
//        { Hirota 2.5.15 }
//        Result.re := TSpinTransitionMoment(TransitionMoment).Strength
//                     * Minus(Ket.N+Bra.S+Bra.J+2)
//                     * Sqrt((Bra.J+1)*(Ket.J+One)) * ReducedMatrixElement(TransitionMoment.Rank,Bra.S)
//                     * sixj(Ket.S, Bra.J, Bra.N, Ket.J, Ket.S, TransitionMoment.Rank);
//      Exit; { Acts on S, so don't want code below }
    end else if TM is TNuclearMagneticMoment then begin
      if (q=0) and (Bra.N=Ket.N) and (Bra.J=Ket.J) then
        Result.re :=1;
      Exit;
    end else
      raise NotYet;
  end else
    if TM.AxisSwitcher <> nil then begin
      Result := z0;
      Kpp := K-TM.Rank;
      while Kpp <= K+TM.Rank do begin
        if Abs(Kpp) <= Bra.N then begin
          Z := Djmn(Bra.N, Kpp, Kp, DegToRad(TM.AxisSwitcher.Chi), DegToRad(TM.AxisSwitcher.Theta), DegToRad(TM.AxisSwitcher.Phi));
          CMul(Z, HDip(TM as TSingleTransitionMoment, Bra.N, Kpp, Ket.N, K));
          CAdd(Result, Z);
        end;
        Inc(Kpp,2);
      end;
    end else
      Result := HDip(TM as TSingleTransitionMoment, Bra.N, Kp, Ket.N, K);
  if (Ket.S > 0) and not MathSup.IsZero(Result) then { TODO: Scope for pushing this one level up, as it will be the same for all K combinations }
    CMul(Result, Minus(Bra.N+Ket.S+Ket.J+TM.Rank) * Sqrt((Bra.J+1)*(Ket.J+One)) * sixj(Bra.N, Bra.J, Ket.S, Ket.J, Ket.N, TM.Rank));
end;

function TAsymmetricTop.SymbolicRotationalTransitionMatrixElements(TransitionMoment: TTransitionMoment; AddHyperfineCoupling: Boolean; IncompleteOK: Boolean): TMatrixElements;
begin
  if TransitionMoment.Magnetic or (Molecule.nNuclei > 0) then
    raise NotYet;
  Result := HDipMatrixElements(TransitionMoment, IncompleteOK);
  if S > 0 then
    AddOpenShellIntensity(eN, eJ, eS, TransitionMoment, 1, Result);
end;

{.$DEFINE VERYSLOWTEST}
function TAsymmetricTop.TransitionMatrixElement(TransitionMoment: TTransitionMoment; A, B: TBasisState): Complex;
var
  Bra, Ket: TAsymmetricTopBasisState;
begin
  { Optimization - only call full routine if selection rule on K is likely to be satisfied }
  Bra := A as TAsymmetricTopBasisState;
  Ket := B as TAsymmetricTopBasisState;
  if TransitionMoment.CheckProjections(Bra.K, Ket.K) then
    Result := inherited TransitionMatrixElement(TransitionMoment, A, B)
  else
    {$IFDEF VERYSLOWTEST}
    if not MathSup.IsZero(inherited TransitionMatrixElement(TransitionMoment, A, B)) then begin
      TransitionMoment.CheckProjections(Bra.K, Ket.K);
      raise Exception.Create('TAsymmetricTop.TransitionMatrixElement'+TransitionMoment.UniqueName+' '+ClassName)
    end else
    {$ENDIF}
      Result := z0;
end;

function TAsymmetricTop.RotationalPerturbationMatrixElement(ABra, AKet: TBasisState; SignA, SignB: Integer; MNode: TMNode; Flag: Integer):Complex;
var
  Bra, Ket: TAsymmetricTopBasisState;
  P: TAsymmetricTopPerturbation;
begin
  if Flag <> Ord(meHrot) then
    raise NotYet;
  Bra := ABra as TAsymmetricTopBasisState;
  Ket := AKet as TAsymmetricTopBasisState;
  P := MNode as TAsymmetricTopPerturbation;
  if Bra.J <> Ket.J then
    raise Exception.Create('TAsymmetricTop.PerturbationMatrixElement');
  { TODO: Allow for off-diagonal in S? }
  Result := P.MatrixElement(Bra.J, Bra.S, Bra.N, SignA*Bra.K, Ket.N, SignB*Ket.K);
end;

function TAsymmetricTop.HasWidth: Boolean;
var
  i: Integer;
begin
  Result := inherited HasWidth;
  if not Result then
    for i := FirstWidthPar to nParameters do
      if Parameters[i] <> 0 then begin
        Result := True;
        Exit;
      end;
end;

const
  AsymmetricTopSettingNames: array[LSI..LSI+2] of string = (
'Symmetry',
'Minimum |K| to use in basis; "all" (default) for no lower limit.',
'Maximum |K| to use in basis; "all" (default) for no upper limit.'
);

function TAsymmetricTop.HelpOnSetting(AIndex: Integer): string;
begin
  if (AIndex >= Low(AsymmetricTopSettingNames))
     and (AIndex <= High(AsymmetricTopSettingNames)) then
    Result := AsymmetricTopSettingNames[AIndex]
  else
    Result := inherited HelpOnSetting(AIndex);
end;

const
 AsymmetricTopParameterNames: array [LSP..FirstWidthPar+4] of string = (
'A rotational constant',
'B rotational constant',
'C rotational constant',
'Average of rotational constants perpendicular to K axis - Alternate form for rotational constants',
'Difference between rotational constants perpendicular to K axis - Alternate form for rotational constants',
'K^4 Quartic Centrifugal Distortion',
'J(J+1)K^2 Quartic Centrifugal Distortion',
'J^2(J+1)^2 Quartic Centrifugal Distortion',
'deltaK or d2 - A/S reduction dependent off diagonal Quartic Centrifugal Distortion',
'deltaJ or d1 - A/S reduction dependent off diagonal Quartic Centrifugal Distortion',
'K^6 Sextic Centrifugal Distortion',
'J(J+1)K^4 Sextic Centrifugal Distortion',
'J^2(J+1)^2K^2 Sextic Centrifugal Distortion',
'J^3(J+1)^3 Sextic Centrifugal Distortion',
'phiK or h3- A/S reduction dependent off diagonal Sextic Centrifugal Distortion',
'phiJK or h2 - A/S reduction dependent off diagonal Sextic Centrifugal Distortion',
'phiJ or h1 - A/S reduction dependent off diagonal Sextic Centrifugal Distortion',
    { Octic Centrifugal Distortion }
'K^8 Octic Centrifugal Distortion',
'J(J+1)K^6 Octic Centrifugal Distortion',
'J^2(J+1)^2K^4 Octic Centrifugal Distortion',
'J^3(J+1)^3K^2 Octic Centrifugal Distortion',
'J^4(J+1)^4 Octic Centrifugal Distortion',
'llK - A/S reduction dependent off diagonal Octic Centrifugal Distortion',
'llKJ - A/S reduction dependent off diagonal Octic Centrifugal Distortion',
'llJK - A/S reduction dependent off diagonal Octic Centrifugal Distortion',
'llJ - A/S reduction dependent off diagonal Octic Centrifugal Distortion',
    { Spin-Rotation Parameters }
'aa spin-rotation Parameter',
'bb spin-rotation Parameter',
'cc spin-rotation Parameter',
'(eab + eba)/2',
'(eac + eca)/2',
'(ebc + ecb)/2',
'(DeltasN in A reduction) quartic spin rotation parameter',
'(DeltasNK in A reduction) quartic spin rotation parameter',
'(DeltasKN in A reduction) quartic spin rotation parameter',
'(DeltasK in A reduction) quartic spin rotation parameter',
'(deltasN in A reduction) quartic spin rotation parameter',
'(deltasK in A reduction) quartic spin rotation parameter',
'Diagonal spin-spin coupling constant (=D/3)',
'Off-Diagonal spin-spin coupling constant (=E)',
'multiplier of <Ja^2> for linewidth',
'multiplier of <Jb^2> for linewidth',
'multiplier of <Jc^2> for linewidth',
'multiplier of <J(J+1)-K^2> for linewidth',
'multiplier of <J+^2+J-^2>/4 for linewidth'
);

function TAsymmetricTop.HelpOnParameter(AIndex: Integer): string;
begin
  if (AIndex >= Low(AsymmetricTopParameterNames))
     and (AIndex <= High(AsymmetricTopParameterNames)) then
    Result := AsymmetricTopParameterNames[AIndex]
  else
    Result := inherited HelpOnParameter(AIndex);
end;

function TAsymmetricTop.zIndex: Integer;
begin
  Result := LSP + Ord(zAxisFor[Molecule.Representation]);
end;

{ TAsymmetricManifold }

function TAsymmetricManifold.nSym: Integer;
begin
  Result := TAsymmetricMolecule(Owner).LastSym+1;
end;

constructor TAsymmetricManifold.Create(AOwner: TComponent);
begin
  SavedProperties := SavedProperties + ',LimitSearch=True'; { Default to give better quantum numbers }
  inherited Create(AOwner);
end;

{ TAsymmetricMolecule }

function TAsymmetricMolecule.GetPointGroup(Index: integer): TAbelianPointGroup;
begin
  Result := TAbelianPointGroup(GetSetting(Index));
end;

function TAsymmetricMolecule.GetRepresentation(Index: integer): TRepresentation;
begin
  Result := TRepresentation(GetSetting(Index));
end;

procedure TAsymmetricMolecule.SetPointGroup(Index: integer; const AValue: TAbelianPointGroup);
begin
  SetSetting(Index, Ord(AValue));
end;

procedure TAsymmetricMolecule.SetRepresentation(Index: integer; const AValue: TRepresentation);
begin
  SetSetting(Index, Ord(AValue));
end;

constructor TAsymmetricMolecule.Create(AOwner: TComponent);
begin
  SavedProperties := SavedProperties + ',PointGroup=C1,Representation=Ir,eeWt=1,eoWt=1,oeWt=1,ooWt=1,C2xAxis=c';
  inherited Create(AOwner);
end;

function TAsymmetricMolecule.Allowed(Rank: TQuantumNumber; Sym1, Sym2: Integer): Boolean;
begin
  if FakeSym then begin
    { Check for equal Statistical Weights }
    Result := StatisticalWeights[Sym1] = StatisticalWeights[Sym2];
    Exit;
  end else if PseudoC2v then
    case PointGroup of
      C1:          begin Sym1 := 0; Sym2 := 0; end;
      C2, Cs, C2h: begin Sym1 := Sym1 div 2; Sym2 := Sym2 div 2; end;
      Ci:          begin Sym1 := Sym1 div 4; Sym2 := Sym2 div 4; end;
      D2, C2v, D2h: { no change needed }
    end;
  if Minus(Rank) < 0 then
    { This is essentially testing for a parity change }
    Result := (OddRankSym xor Sym1 xor Sym2) = 0
  else
    Result := Sym1 = Sym2;
end;

function TAsymmetricMolecule.Allowed(TransitionMoment: TTransitionMoment; Sym1, Sym2: Integer): Boolean;
begin
  if TransitionMoment.Magnetic then
    Result := Allowed(0,Sym1,Sym2)
  else
    Result := Allowed(TransitionMoment.Rank, Sym1, Sym2);
end;

function TAsymmetricMolecule.FixupProperties: Exception;
var
  i, Ieosym, wt, iug: Integer;
  eoSyms: array [0..7] of Integer;
  eoSymsStep: Integer;
  j: Integer;
begin
  PrecomputeSymNames;
  case PointGroup of
    C1:  FOddRankSym := 0;
    Ci:  FOddRankSym := 1;
    C2:  FOddRankSym := 0;
    Cs:  FOddRankSym := 1;
    D2:  FOddRankSym := 0;
    C2v: FOddRankSym := 1; { A2 }
    C2h: FOddRankSym := 2; { Au }
    D2h: FOddRankSym := 4; { Au }
  else
    raise NotYet;
  end;
  { Work out statistical weights }
  SetLength(StatisticalWeights,LastSym+1);
  for i := 0 to LastSym do
    StatisticalWeights[i] := -1; { Start with them all unknown }
  for iug := 0 to 1 do begin
    MakeeoSyms(eoSyms, eoSymsStep, iug*((LastSym+1) div 2));
    for i := 0 to LastSym do begin
      Ieosym := i;
      while (Ieosym <= 7) and (eoSyms[Ieosym] >= 0) do begin
        wt := GetSetting(FirstStatisticalWeight+eoSyms[Ieosym]);
        if StatisticalWeights[i] < 0 then
          StatisticalWeights[i] := wt
        else if StatisticalWeights[i] <> wt then
          raise Exception.CreateFmt('Inconsistent statistical weights in molecule %s: %d %d', [UniqueName, wt, StatisticalWeights[i]]);
        Inc(Ieosym, eoSymsStep);
      end;
    end;
    if not (PointGroup in [Ci, C2h, D2h]) then
      Break;
  end;
  { For an initial implementation, figure out statistical weight groups from weights;
    this will fail if different spin species have the same statistical weight }
  { TODO: Allow different spin species to have the same statistical weight;
          C2H4 has 4 species with weights 7:3:3:3 }
  SetLength(StatisticalWeightGroup,LastSym+1);
  StatisticalWeightGroup[0] := 1;
  FnSpinSpecies := 1;
  for i := 1 to LastSym do begin
    StatisticalWeightGroup[i] := 0;
    for j := 0 to i - 1 do
      if StatisticalWeights[i] = StatisticalWeights[j] then begin
        StatisticalWeightGroup[i] := StatisticalWeightGroup[j];
        Break;
      end;
    if StatisticalWeightGroup[i] <= 0 then begin
      Inc(FnSpinSpecies);
      StatisticalWeightGroup[i] := nSpinSpecies;
    end;
  end;
  Result := Inherited FixupProperties;
  if not Assigned(Result) and (PointGroup in [C2v, D2, D2h]) and (C2xAxis = C2zAxis) then
    Result := EReportException.Create('C2xAxis must not equal C2zAxis in '+Name);
  if not Assigned(Result) and JAdjustSym and (nNuclei > 0) then
    for i := 0 to nNuclei - 1 do
      if (Nuclei[i].Spin > 0) and not Assigned(Result) then
        Result := EReportException.Create('J adjusted symmetry for '+Name+' not implemented for hyperfine structure');
end;

function TAsymmetricMolecule.ReadProperties(Node: TDOMNode; CheckUnread: Boolean): Integer;
var
  HasIdentical: Boolean;
  i: Integer;
begin
  Result := inherited ReadProperties(Node, False);
  { For backwards compatibility - fix for correct handling of multiple equivalent nuclei
    needs weights = 0 for previously working cases. }
  if (ReadVersion < 7001143) and (nNuclei > 0) and (PointGroup in [C2,C2v, C2h, D2, D2h]) then begin
    HasIdentical := False;
    for i := 0 to nNuclei - 1 do
      HasIdentical := HasIdentical or Nuclei[i].AsNext;
    if HasIdentical then begin
      case PointGroup of
        C2v, C2, C2h:
             begin
               SetSetting(FirstStatisticalWeight+CsSwitch[C2zAxis],0);
               SetSetting(FirstStatisticalWeight+(SymWrt[C2zAxis] xor CsSwitch[C2zAxis]),0);
             end;
      else
        raise NotYet;
      end;
    end;
  end;
  if CheckUnread then
    Result := Result + nUnreadProperties(Node);
end;

function TAsymmetricMolecule.GetAxis(const Index: Integer): TabcAxis;
begin
  Result := TabcAxis(GetSetting(Index));
end;

procedure TAsymmetricMolecule.SetAxis(const Index: Integer; const AValue: TabcAxis);
begin
  SetSetting(Index, Ord(AValue));
end;

function TAsymmetricMolecule.GetSettingsNames(AIndex: Integer; out LowestValue: Integer): PPChar;
const
  AxisNames: array [0..3] of PChar = ('a', 'b', 'c', nil);
begin
  Result := inherited GetSettingsNames(AIndex,LowestValue);
  if AIndex in [LMI+7, LMI+8] then
    Result := @AxisNames[0];
end;

function TAsymmetricMolecule.SymName(Sym: Integer): string;
begin
  if Sym < 0 then
    Result := '-'
  else
    Result := SymNames[Sym];
end;

{ TODO: Next two functions for half integral J }
function TAsymmetricMolecule.WangName(J: TQuantumNumber; rSym: Integer): string;
begin
  { Given rotational symmetry, work out Wang matrix }
  SetLength(Result,2);
  if AntiSymmetricWrtAxis(rSym, ZAxisFor[Representation]) then
    Result[1] := 'O'
  else
    Result[1] := 'E';
  if AntiSymmetricWrtAxis(rSym, XAxisFor[Representation]) xor (Minus(J) < 0) then
    Result[2] := '-'
  else
    Result[2] := '+';
end;

function TAsymmetricMolecule.JAdjustedSymNameBase(J: TQuantumNumber; Sym: Integer): string;
begin
  if Sym < 0 then
    Result := '-'
  else if not Odd(HalfSpin) then
    Result := JAdjustedSymNames[JAdjustedSym(J,Sym)]
  else
    { TODO: Check odd spin implementation; Check PseudoC2v }
    Result := inherited JAdjustedSymNameBase(J, Sym)
end;

{$IFDEF OldStyle}
const
                                                   {Ir, IIr, IIIr, Il, IIl, IIIl}
  JAdjustMask: array[TRepresentation] of Integer = ( 1,   3,    2,  1,   3,    2);
{$ENDIF}

function TAsymmetricMolecule.JAdjustedSym(J: TQuantumNumber; Sym: Integer): Integer;
begin
  Result := Sym;
  if (Sym >= 0) and JAdjustSym and not Odd(HalfSpin) { TODO: Implement for odd spin } and (Minus(J) < 0) then
    if PseudoC2v then
      Result := FJAdjustedSym[Sym]
    else
      case PointGroup of
        C1, Ci: ; { No adjustment }
        C2, Cs, C2h: if JAlternate then
                       Result := Sym xor 1;
        D2, C2v, D2h: Result := FJAdjustedSym[Sym];
      end;
    {$IFDEF OldStyle}
    if PseudoC2v then
      Result := Result xor JAdjustMask[Representation]
    else
      case PointGroup of
        D2, C2v, D2h: Result := Result xor JAdjustMask[Representation];
        { Three cases for Cs:
          1. The J alternation flips bewteen the two C2v symmetries that are in the same matrix - don't alternate
          2. The J alternation flips to the other matrix for both symmetries - alternate
          3. The J alternation flips one symmetry but not the other to the other matrix - In this case
             neither choice is likely to make sense, so alternate, so user has choice }
        Cs, C2h, C2: if JAdjustMask[Representation] <> SymWrt[C2zAxis] then
                       Result := Result xor 1;
        C1, Ci: { All matrices the same }
      end;
    {$ENDIF}
end;

function TAsymmetricMolecule.LastSym: Integer;
begin
  if PseudoC2v then
    case PointGroup of
      C1,C2,Cs,D2,C2v: Result := 3;
      Ci,C2h,D2h: Result := 7;
    end
  else
    case PointGroup of
      C1: Result := 0;
      Ci,C2,Cs: Result := 1;
      D2,C2v,C2h: Result := 3;
      D2h: Result := 7;
    end;
end;

const
  AsymmetricMoleculeSettingNames: array[LMI..LMI+11] of string = (
'Point group - C1, Ci, C2, Cs, D2, C2v, C2h or D2h',
'Representation - Ir, IIr, IIIr, Il, IIl or IIIl',
'Use Watson''s S reduction (as opposed to A).',
'Statistical weight for ee levels',
'Statistical weight for eo levels',
'Statistical weight for oe levels',
'Statistical weight for oo levels',
'a, b, or c; C2 axis in C2v, C2 and C2h; perpendicular to symmetry plane in Cs',
'a, b, or c; typically the out of plane axis in C2v',
'Ignore any operators breaking C2v symmetry.',
'Turn off transition moment test.',
'Add phase factor to minimize complex matrix elements'
);

function TAsymmetricMolecule.HelpOnSetting(AIndex: Integer): string;
begin
  if (AIndex >= Low(AsymmetricMoleculeSettingNames))
     and (AIndex <= High(AsymmetricMoleculeSettingNames)) then
    Result := AsymmetricMoleculeSettingNames[AIndex]
  else
    Result := inherited HelpOnSetting(AIndex);
end;

function TAsymmetricMolecule.IsValidChildClass(MNodeClass: TMNodeClass): Boolean;
begin
  Result := inherited IsValidChildClass(MNodeClass);
  if Result and MNodeClass.InheritsFrom(TTransitionMoment) then
    Result := (MNodeClass = TAsymmetricTopOrbitalTransitionMoment)
              or (MNodeClass = TAsymmetricTopRotationalTransitionMoment)
              or (MNodeClass = TSphericalTransitionMoment)
              or (MNodeClass = TCartesianTransitionMoment)
              or (MNodeClass = TNuclearMagneticMoment)
              or (MNodeClass = TQuadrupoleTransitionMoment);
end;

procedure TAsymmetricMolecule.PrecomputeSymNames;
  procedure ComputeC2vAdjust;
  var
    Sym, ASym, Col: Integer;
  begin
    SetLength(FJAdjustedSym, LastSym+1);
    { Find column in C2v Character table that corresponds to R(Pi) around
      the axis chosen for K quantization. This is the column that will not
      change with J }
    if ZAxisFor[Representation] = C2zAxis then
      Col := 1
    else if ZAxisFor[Representation] = C2xAxis then
      Col := 3
    else
      Col := 2;
    for Sym := 0 to LastSym do begin
      JAdjustedSymNames[Sym] := WangName(0,eoSym(Sym));
      if LastSym > 3 then
        if Sym >= 4 then
          JAdjustedSymNames[Sym] := JAdjustedSymNames[Sym] + 'u'
        else
          JAdjustedSymNames[Sym] := JAdjustedSymNames[Sym] + 'g';
      for Asym := 0 to 3 do
        if (ASym <> (Sym and 3))
           and (C2vCharacterTable[ASym][Col] = C2vCharacterTable[Sym and 3][Col]) then
          FJAdjustedSym[Sym] := Asym or (Sym and 4); { Sym and 4 is needed for D2h }
    end;
  end;
var
  i: Integer;
  P: PPChar;
begin
  SetLength(SymNames, LastSym+1);
  if PseudoC2v then
    if Length(SymNames) = 8 then
      P := IrepNames(D2h)
    else
      P := IrepNames(C2v)
  else
    P := IrepNames(PointGroup);
  for i := 0 to LastSym do begin
    SymNames[i] := P^;
    Inc(P);
  end;

  { Pre-compute J alternation and name }
  SetLength(JAdjustedSymNames, LastSym+1);
  case PointGroup of
    C1, Ci: if PseudoC2v then
              ComputeC2vAdjust
            else
              { No alternation normally }
              for i := 0 to LastSym do
                JAdjustedSymNames[i] := SymName(i);
    C2, Cs, C2h:
      if PseudoC2v then
        ComputeC2vAdjust
      else begin
        { Does single symmtery operator correspond to R(z), R(x) or R(y)? }
        case xyzAxis(C2zAxis) of
          { Symmetry wrt operator is (-1)^K - split into E/O; no J alternation }
          zAxis: begin JAlternate := False; JAdjustedSymNames[0] := 'E'; JAdjustedSymNames[1] := 'O'; end;
          { Symmetry wrt operator is +/-(-1)^J - split into +/- and alternate with J }
          xAxis: begin JAlternate := True; JAdjustedSymNames[0] := '+'; JAdjustedSymNames[1] := '-'; end;
          { Symmetry wrt operator is 1/-(-1)^(J-K) - split into E+O-/E-O+, and alternate with J }
          yAxis: begin JAlternate := True; JAdjustedSymNames[0] := 'E+O-'; JAdjustedSymNames[1] := 'E-O+'; end;
        end;
        if PointGroup = C2h then
          for i := 0 to 1 do begin
            JAdjustedSymNames[i+2] := JAdjustedSymNames[i] + 'u';
            JAdjustedSymNames[i] := JAdjustedSymNames[i] + 'g';
          end;
      end;
    D2, C2v, D2h: ComputeC2vAdjust;
  end;
end;

function TAsymmetricMolecule.QnoInfo(QnoType: TQnoType): TQnoInfo;
begin
  Result := inherited QnoInfo(QnoType);
  case QnoType of
    qtA: Result.Name := 'Ka';
    qtB: Result.Name := 'Kc';
  end;
end;

function TAsymmetricMolecule.FixSym(J: TQuantumNumber; Sym: Integer;
  InputPgopherVersion: Integer): Integer;
begin
  Result := Sym;
  if InputPgopherVersion < 5002000 then
    if PseudoC2v or (PointGroup in [D2, D2h, C2v]) then
      case C2zAxis of
           aAxis: case C2xAxis of
                    bAxis: ;
                    cAxis: if (Sym and 2) <> 0 then Result := Sym xor 1;
                  end;
           bAxis: case C2xAxis of
                    aAxis: if Odd(Sym) then Result := Sym xor 2;
                    cAxis: if (Sym and 3) <> 0 then Result := (((Sym and 3) mod 3)+1) or (Sym and 4);
                  end;
           cAxis: case C2xAxis of
                    bAxis: if Sym in [1,2,5,6] then Result := Sym xor 3;
                    aAxis: if (Sym and 3) <> 0 then Result := (((Sym and 3)+4) mod 3 + 1) or (Sym and 4);
                  end;
      end;
end;

function TAsymmetricMolecule.eoSym(RveSym: Integer): integer;
{ Convert from D2, C2v and D2h point group labels to eo symmetry labels }
{ For the other point groups, a particular symmetry will give more than one eo combination }
var
  Sym: Integer;
begin
  Result := 0;
  RveSym := RveSym and 3; { Discard E*/i for D2h }
  { Character table ordering is the same for the rotation operations for D2, C2v and D2h }
  { We know directly from eo label the symmetry wrt Ra and Rc, but not Rb, so
    must use only these two only to match irreducible representation labels }
  Sym := AntiSymWrt[C2zAxis];
  if (Sym <> 3) then { if not b axis }
    if C2vCharacterTable[RveSym][1{=C2}] < 0 then
      Result := Sym;
  Sym := AntiSymWrt[C2zAxis] xor AntiSymWrt[C2xAxis]; { = AntiSymWrt[C2yAxis] }
  if (Sym <> 3) then { if not b axis }
    if C2vCharacterTable[RveSym][2{=sigma(zx)}] < 0 then
      Result := Result or Sym;
  Sym := AntiSymWrt[C2xAxis];
  if (Sym <> 3) then { if not b axis }
    if C2vCharacterTable[RveSym][3{=sigma(yz)}] < 0 then
      Result := Result or Sym;
end;

procedure TAsymmetricMolecule.MakeeoSyms(out eoSyms: array of Integer; out eoSymsStep: Integer; VibronicSymmetry: Integer);
var
  i: Integer;
begin
  { Defaults - ee ee eo eo oe oe oe oo - need adjusting in some cases }
  for i := 0 to 7 do
    eoSyms[i] := i div 2;
  eoSymsStep := 8; { Only take one rotational symmetery }

  case PointGroup of
    C1: if PseudoC2v then
          for i := 0 to 3 do
            eoSyms[i] := i
        else
          eoSymsStep := 2; { Take all four possibile rotational symmetries always, as only one rovibrational symmetry }
    Ci: if PseudoC2v then begin
          for i := 0 to 7 do
            if (i >= 4) xor (VibronicSymmetry > 0) then
              eoSyms[i] := -1 { Half the states do not exist }
            else
              eoSyms[i] := i and 3
        end else begin
          eoSymsStep := 2;
          for i := 0 to 3 do
            eoSyms[2*i+(1-VibronicSymmetry)] := -1; { Half the states do not exist, as Rotational wavefunctions are always g }
        end;
    Cs, C2, C2h: begin
      if PseudoC2v then begin
        { The logic below ensures that the true symmetry is the effective symmetry / 2 }
        { Pair of symmetric functions }
        eoSyms[0] := 0;
        eoSyms[1] := SymWrt[C2zAxis];
        { Pair of antisymmetric functions }
        eoSyms[2] := eoSyms[0] xor CsSwitch[C2zAxis];
        eoSyms[3] := eoSyms[1] xor CsSwitch[C2zAxis];
        for i := 4 to LastSym do begin
          { Handle g/u in C2h }
          if VibronicSymmetry >= 2 then begin
            eoSyms[i] := eoSyms[i-4]; { Duplicate set for C2h }
            eoSyms[i-4] := -1; { but half don't exist }
          end else
            eoSyms[i] := -1;
        end;
      end else begin
        eoSymsStep := 4;
        { Pair of symmetric functions }
        eoSyms[0] := 0;
        eoSyms[4] := SymWrt[C2zAxis];
        { Pair of antisymmetric functions }
        eoSyms[1] := eoSyms[0] xor CsSwitch[C2zAxis];
        eoSyms[5] := eoSyms[4] xor CsSwitch[C2zAxis];
        for i := 2 to LastSym do begin
          { Handle g/u in C2h }
          if VibronicSymmetry >= 2 then begin
            eoSyms[i] := eoSyms[i-2]; { Duplicate set for C2h }
            eoSyms[i-2] := -1; { but half don't exist }
            eoSyms[i+4] := eoSyms[i+2];
            eoSyms[i+2] := -1; { but half don't exist }
          end else begin
            eoSyms[i] := -1;
            eoSyms[i+4] := -1;
          end;
        end;
      end;
      if {$IFDEF JTEST} (JTest <> 0) or {$ENDIF} ((VibronicSymmetry and 1) <> 0) then
        { Antisymmetric (wrt R2z') vibronic state - switch over }
        for i := 0 to 7 do
          if eoSyms[i] >= 0 then
            eoSyms[i] := eoSyms[i] xor CsSwitch[C2zAxis];
    end;
    C2v, D2, D2h: begin
      for i := 0 to LastSym do
        if (i >= 4) xor (VibronicSymmetry >= 4) then
          eoSyms[i] := -1
        else
          eoSyms[i] := eoSym(VibronicSymmetry xor i);
      {$IFDEF OldStyle}
      for i := 0 to LastSym do
        eoSyms[i] := (eoSym(VibronicSymmetry) xor i) and 3;
      {$ENDIF}
      {$IFDEF JTEST}
      if JTest <> 0 then
        for i := 0 to LastSym do
          eoSyms[i] := (eoSym(0) xor i) and 3;
      {$ENDIF}
    end;
  end;
end;

{ TAsymmetricTopBasisState }

function TAsymmetricTopBasisState.BasisLabel: string;
begin
  Result := inherited BasisLabel + ' ';
  if State.S <> 0 then
    Result := Result + State.Manifold.Mixture.QuantumNumberToStr(N) + ' ';
  Result := Result + State.Manifold.Mixture.QuantumNumberToStr(K)
            + HyperfineStateLabel;
end;

function TAsymmetricTopBasisState.BasisLabelKey: string;
begin
  Result := inherited BasisLabelKey;
  if State.S <> 0 then
    Result := Result + ' N';
  Result := Result + ' K' + AxisNames[ZAxisFor[TAsymmetricMolecule(State.Manifold.Molecule).Representation]]
            + HyperfineStateLabelKey;
end;

procedure TAsymmetricTopBasisState.CopyFrom(OtherState: TBasisState; Sign: Integer);
var
  B: TAsymmetricTopBasisState;
begin
  inherited;
  B := OtherState as TAsymmetricTopBasisState;
  K := Sign*B.K;
  Ka := B.Ka;
  Kc := B.Kc;
end;

function TAsymmetricTopBasisState.Qno(QnoType: TQnoType): TQuantumNumber;
begin
  case QnoType of
    qtA: Result := Ka;
    qtB: Result := Kc;
  else
    Result := inherited Qno(QnoType);
  end;
end;

function TAsymmetricTopBasisState.MakeSignature(SignA:Integer; Ket: TBasisState; SignB: Integer; MNode: TMNode; Symmetrised: Boolean): TIntegerArray;
var
  KetK: TAsymmetricTopBasisState;
begin
  if S > 0 then begin
    SetLength(Result,3);
    Result[1] := (N - Ket.N) div 2
  end else
    SetLength(Result,2);
  KetK := Ket as TAsymmetricTopBasisState;
  Result[0] := (K*SignA - KetK.K*SignB) div 2;
  Result[High(Result)] := (J - Ket.J) div 2;
end;

threadvar
  DummyState: TAsymmetricTopBasisState;

function TAsymmetricTopBasisState.NextBasis(var LoopState: TBasisState; TransitionMoment: TTransitionMoment; Sign: Integer):Boolean;
var
  DState: TAsymmetricTopBasisState;
begin
  repeat
    if LoopState = nil then begin
      if DummyState = nil then
        DummyState := TAsymmetricTopBasisState.Create(nil,F,M,Sym);
      DState := DummyState;
      LoopState := DummyState;
      DState.CopyFrom(Self,Sign);
      DState.K := Sign*K - TransitionMoment.Rank;
    end else begin
      DState := LoopState as TAsymmetricTopBasisState;
      Inc(DState.K, 2);
    end;
    Result := DState.K <= Sign*K+TransitionMoment.Rank;
  until not Result or (Abs(DState.K) <= N);
end;

procedure TAsymmetricTopBasisState.SetSymbolicVariables(SignA:Integer; Ket: TBasisState; SignB: Integer; MNode: TMNode);
var
  KetK: TAsymmetricTopBasisState;
begin
  KetK := Ket as TAsymmetricTopBasisState;
  eN.Value := Ket.N / 2;
  eK.Value := KetK.K*SignB / 2;
  eS.Value := Ket.S / 2;
  eJ.Value := Ket.J / 2;
end;

function TAsymmetricTopBasisState.StateLabel(AQuantumNumberFormat:TQuantumNumberFormat): string;
begin
  {$IFDEF FPC} { As of Feb 2012, this is 3-4 times faster }
  WriteStr(Result, inherited StateLabel(AQuantumNumberFormat),
                   ' ', QuantumNumberToStr(N,AQuantumNumberFormat):2,
                   ' ', QuantumNumberToStr(Ka,AQuantumNumberFormat):2,
                   ' ', QuantumNumberToStr(Kc,AQuantumNumberFormat):2);
  {$ELSE}
  Result := Format('%s %2s %2s %2s', [inherited StateLabel(AQuantumNumberFormat), QuantumNumberToStr(N,AQuantumNumberFormat),
                                      QuantumNumberToStr(Ka,AQuantumNumberFormat), QuantumNumberToStr(Kc,AQuantumNumberFormat)]);
  {$ENDIF}
  if State.S <> 0 then
    Result := Result + Format(' %2s', [QuantumNumberToStr(J,AQuantumNumberFormat)]);
  Result := Result + HyperfineStateLabel;
end;

function TAsymmetricTopBasisState.StateLabelKey: string;
begin
  Result := inherited StateLabelKey;
  if State.S <> 0 then
    Result := Result + ' N Ka Kc J'
  else
    Result := Result + ' J Ka Kc';
  Result := Result + HyperfineStateLabelKey;
end;

function TAsymmetricTopBasisState.SymbolicVariableValues: string;
begin
  Result := VList.VariableValues;
end;

function TAsymmetricTopBasisState.TrackIndex: Integer;
begin
 Result := Abs(K);
 if K < 0 then Inc(K);
end;

function TAsymmetricTopBasisState.TxName(Upper, Lower: TBasisState): string;
var
  Molecule: TAsymmetricMolecule;
  U, L: TAsymmetricTopBasisState;
  US, LS: TAsymmetricTop;
  fU, fL: string;
  function Fnumber(State: TAsymmetricTopBasisState):string;
  var
    S: TQuantumNumber;
  begin
    S := (State.State as TAsymmetricTop).S;
    if S > 0 then
      Result := IntToStr((State.N - State.J + S + 2) div 2)
    else
      Result := '';
  end;
begin
  Molecule := State.Manifold.Molecule as TASymmetricMolecule;
  U := Upper as TASymmetricTopBasisState;
  US := U.State as TASymmetricTop;
  L := Lower as TASymmetricTopBasisState;
  LS := L.State as TASymmetricTop;
  Result := LowerCase(DeltaJName(Abs(U.K)-Abs(L.K)))+DeltaJName(U.N-L.N);
  if (US.S > 0) or (LS.S > 0) then begin
    { Add spin label }
    fU := Fnumber(U);
    fL := Fnumber(L);
    Result := Result + fU;
    if fU <> fL then { Add second spin label if different }
      Result := Result + fL;
    Result := Result + ',';
  end;
  Result := Result + Molecule.QuantumNumberToStr(L.Ka) + ',' + Molecule.QuantumNumberToStr(L.Kc);
  Result := Result + Format('(%s)',[Molecule.QuantumNumberToStr(L.N)]);
  if (Length(Lower.HyperfineState) > 0) or (Owner as TEigenvalues).HasFields then
    Result := Result + Upper.HyperfineStateLabel(False) + ',' + Lower.HyperfineStateLabel(False);
end;

function TAsymmetricTopBasisState.TxNameKey(Upper, Lower: TBasisState): string;
begin
  Result := 'deltaK DeltaN';
  if Upper.State.Manifold.MaxSpin > 0 then
    Result := Result + ' Fn''';
  if Lower.State.Manifold.MaxSpin > 0 then
    Result := Result + ' Fn",';
  Result := Result + ' Ka",Kc" (';
  if Upper.State.Manifold.MaxSpin + Lower.State.Manifold.MaxSpin > 0 then
    Result := Result + 'N'
  else
    Result := Result + 'J';
  Result := Result + '")' + HyperfineTxNameKey(False);
end;

procedure TAsymmetricTop.SetSettingFromStr(AIndex: Integer; const Str: string);
begin
  if ((AIndex = LSI+1) or (AIndex = LSI+2)) and (Str = 'all') then
    SetSetting(AIndex,-9999)
  else
    inherited;
end;

procedure TAsymmetricTop.AdjustDefaults;
begin
  inherited AdjustDefaults;
  if C = 0 then
    C := Convert(1, cm1, (RootNode as TMixture).Units);
  if B = 0 then
    B := 1.2 * C;
  if A = 0 then
    A := 9 * C;
end;

function TAsymmetricTop.SettingToStr(AIndex: Integer): string;
begin
  if ((AIndex = LSI+1) or (AIndex = LSI+2)) and (GetSetting(AIndex) < 0) then
      Result := 'all'
  else
    Result := inherited SettingToStr(AIndex);
end;

procedure ShowSpherical(var F:Text; x, y, z: string);
var
  nPhot, Rank, q, p, p1: Integer;
  Vector: array [-1..+1] of TExprNode;
  Current, Previous: array of TExprNode;
  VList: TVariableList;
  ex, ey, ez: TSortedVariableNode;
begin
  { Currently fails sqrt(-1)*sqrt(-1) -> 1: line 3283 expr.pas }
  VList := MakeSortedVariables([x, y, z], [@ex, @ey, @ez]);
  Vector[0] := EB([ez]);
  Vector[1] := EB([SqrtMinusOne.CreateExprNode, '*', ey, '+', ex, '*', -1, '/', Expr.EN([2, 'sqrt'])]);
  Vector[-1] := EB([SqrtMinusOne.CreateExprNode, '*', ey, '*', -1, '+', ex, '/', Expr.EN([2, 'sqrt'])]);
  for nPhot := 1 to 2 do begin
    WriteLn(F, nPhot, ' Photon(s)');
    for Rank := 0 to nPhot do begin
      SetLength(Current,2*Rank+1);
      if nPhot = 1 then begin
        if Rank = 0 then
          Continue;
        for q := 0 to 2 do
          Current[q] := Vector[q-1]
      end else begin
        { Coupling of two tensor operators }
        for q := 0 to 2*Rank do begin
          Current[q] := EB([0]);
          for p := -1 to 1 do begin
            p1 := q-p-Rank+1;
            if (p1 >= 0) and (p1 <= 2 {???}) then
              Current[q] := Expr.EN([EB(['threej', 1, p, 1{Rank-1}, q-Rank-p, Rank, Rank-q, '*', Vector[p], '*', Previous[p1]]), '+', Current[q]]);
          end;
          Current[q] := Expr.EN([2*Rank+1, 'sqrt', '*', Current[q], '*', Minus(2*(q-Rank))]).Simplify([siExpand]).Simplify;
        end;
      end;
      for q := 0 to High(Current) do
        WriteLn(F, 'T(',Rank,',',-Rank+q,') = ', Current[q].Text(True));
    end;
    Previous := Current;
  end;
  FreeAndNil(VList);
end;

procedure TAsymmetricTop.SymmetryTable(var F: Text);
var
  Sym, OldSym, Ieosym, eosym, Jeo: Integer;
  function eo(i: Integer):string;
  begin
    if Odd(i) then
      Result := 'o'
    else
      Result := 'e';
  end;
  function Ka_Kc(eosym: Integer; J: TQuantumNumber):string;
  begin
    if Odd(eosym) xor Odd(eosym div 2) xor Odd(J div 2) then
      Result := 'Ka+Kc=N+1'
    else
      Result := 'Ka+Kc=N  '
  end;
begin
  Write(F, 'Symmetry Table for ', SettingToStr(IndexOfSetting('Symmetry')), ' ',
           Molecule.SettingToStr(Molecule.IndexOfSetting('PointGroup')), ' PseudoC2v:', BoolToStr(Molecule.PseudoC2v, True), ' ',
           Molecule.SettingToStr(Molecule.IndexOfSetting('Representation')));
  if Molecule.PseudoC2v or not (Molecule.PointGroup in [C1, Ci]) then
    Write(F, ' C2zaxis:', AxisNames[Molecule.C2zAxis]);
  if Molecule.PseudoC2v or not (Molecule.PointGroup in [C1, Ci, C2, Cs, C2h]) then
    Write(F, ' C2xaxis:', AxisNames[Molecule.C2xAxis]);
  WriteLn(F);
  for Sym := 0 to Manifold.nSym-1 do begin
    Write(F, Molecule.SymName(Sym):3, Sym:2);
    Ieosym := Sym;
    while (Ieosym <= 7) and (eoSyms[Ieosym] >= 0) do begin
      eosym := eoSyms[Ieosym];
      Write(F, ' ', eo(eosym div 2), eo(eosym), ' ');
      Write(F, Molecule.WangName(0,eosym), ' ', Ka_Kc(eosym,0), ' (Even N) ',
               Molecule.WangName(2,eosym), ' ', Ka_Kc(eosym,2), ' (Odd N) ');
      Inc(Ieosym, eoSymsStep);
    end;
    OldSym := Molecule.FixSym(0,Sym,0);
    if OldSym <> Sym then
      Write(F, ' (Old Sym ', OldSym, ')');
    WriteLn(F);
  end;
  { Repeat for J adjusted }
  WriteLn(F);
  WriteLn(F, 'J adjusted Symmetry Table');
  for Sym := 0 to Manifold.nSym-1 do begin
    Write(F, Molecule.JAdjustedSymName(0, Sym):3, ': ');
    if eoSyms[Molecule.JDeAdjustedSym(0, Sym)] >= 0 then
      for Jeo := 0 to 1 do begin
        if Jeo = 0 then
          Write(F, 'Even N: ')
        else
          Write(F, '  Odd N: ');
        Ieosym := Molecule.JDeAdjustedSym(Jeo*2, Sym);
        Write(F, Molecule.SymName(Ieosym), Ieosym:2, ' ');
        while (Ieosym <= 7) and (eoSyms[Ieosym] >= 0) do begin
          eosym := eoSyms[Ieosym];
          Write(F, ' ', eo(eosym div 2), eo(eosym),
                   ' ', Molecule.WangName(Jeo*2, eosym),
                   ' ', Ka_Kc(eosym,Jeo*2));
          Inc(Ieosym, eoSymsStep);
        end;
      end;
    WriteLn(F);
  end;
  WriteLn(F);
  with Molecule as TAsymmetricMolecule do
    ShowSpherical(F, AxisNames[abcAxis(xAxis)], AxisNames[abcAxis(yAxis)], AxisNames[abcAxis(zAxis)]);
  WriteLn(F);
end;

function TAsymmetricTop.MatrixElementType(const TheMatrixElement: TMatrixElement; const H: string; ShowPhases: Boolean):string;
var
  RHS: string;
  procedure AddVariable(const Variable: string; AIndex: Integer);
  begin
    with TheMatrixElement do
      if AIndex <= High(Signature) then begin
        if Result <> '' then begin
          Result := Result + ',';
          RHS := RHS + ',';
        end;
        Result := Result + Variable;
        RHS := RHS + Variable;
        if Signature[AIndex] > 0 then
          Result := Result + '+'
        else if Signature[AIndex] = 0 then
          Result := Result + '  '; { Nice spacing }
        if Signature[AIndex] <> 0 then
          Result := Result + IntToStr(Signature[AIndex]); { Note changes are not doubled }
      end;
  end;
begin
  Result := '';
  AddVariable('J',2);
  AddVariable('N',1);
  AddVariable('K',0);
  Result := '<' + Result + ShowElementPhases(TheMatrixElement,H,ShowPhases) + RHS + '>';
end;

{ TAsymmetricTopPerturbation }

{ Perturbation matrix elements are applied in the order:
    N^2 Nx Ny (N+S+ + N-S-) Nz Sx Sy Sz N.S
  If any off diagonal operators are present, then the operator:
    1/2(N^2 Nx Ny (N+S+ + N-S-) Nz Sx Sy Sz N.S + N.S Sz Sy Sx Nz (N+S+ + N-S-) Ny Nx N^2)
  is implied }

type
  TopType = (opFull, opForward, opReverse);

function TAsymmetricTopPerturbation.MatrixElement(J, S, Np, Kp, N, K: TQuantumNumber): Complex;
var
  xPow, yPow, zPow: Integer;

  function RotationalElement(N, Kp, K: TQuantumNumber; OpType: TopType): Double;
  var
    MaxPow: Integer;
    dK: Integer;
    Kpp, KppMax: TQuantumNumber;
  begin
    Result := 0;
    dK := Abs(K-Kp) div 2;
    if (dK <= xPow + yPow + pPower) and not Odd(dK+xPow+yPow+pPower) then begin
      { Actually work out [Jz,[Jy,Jx]+]+ }
      if pPower > 0 then begin
        if (xPow or yPow) <> 0 then
          raise EReportException.Create('AsymmetricTopPerturbation: pPower can''t be used with xPower or yPower');
        if dK = pPower then
          Result := rl(N, Kp, K)
        else
          Result := 0;
      end else if xPow > 0 then
        if yPow > 0 then begin
          MaxPow := Max(xPow, yPow);
          KppMax := Min(Kp+2*MaxPow, K+2*MaxPow);
          Kpp := Max(Kp-2*MaxPow, K-2*MaxPow);
          while Kpp <= KppMax do begin
            Result := Result + rlyRe(yPow,N,Kp,Kpp)*rlx(xPow,N,Kpp,K)
                             + rlx(xPow,N,Kp,Kpp)*rlyRe(yPow,N,Kpp,K);
            Inc(Kpp,2);
          end;
        end else
          Result := rlx(xPow,N,Kp,K)
      else if yPow > 0 then
        Result := rlyRe(yPow,N,Kp,K)
      else Result := 1;
      if zPow > 0 then
        if ((xPow or yPow or pPower) <> 0) and (OpType <> opForward) then
          if OpType = opReverse then
            Result := Result * IntPower(Kp,zPow)
          else
            Result := Result * (IntPower(K,zPow) + IntPower(Kp,zPow))
        else
          Result := Result * IntPower(K,zPow);
      {$IFDEF JTEST}
      if Odd(JPower) then
        Result := Result * IntPower(N, JPower)
      else
      {$ENDIF}
      if JPower > 0 then
        Result := Result * IntPower(N*(N+2), JPower div 2);
      Result := Result / IntPower(2,zPow + JPower);
  //    if (xPow = 1) and (yPow = 0) and (zPow = 1) then begin
  //      Result := Value * (K + Kp)*rl(N, K, Kp);
    end;
  end;

  function SOperator(q, Np, Kp, N, K: TQuantumNumber; NSPower, SPower: Integer; OpPower: Integer; OpType: TopType):Double;
  begin
    Result := T1Sn(Abs(OpPower), q, J, S, Np, Kp, N, K);
    if NSPower > 0 then
      case OpType of
        opFull:    Result := Result * (NSOperator(J, N, S, NSPower, SPower)+NSOperator(J, Np, S, NSPower, SPower));
        opForward: Result := Result * NSOperator(J, N, S, NSPower, SPower);
        opReverse: Result := Result * NSOperator(J, Np, S, NSPower, SPower);
      end;
  end;

var
  SxPow, SyPow, SzPow: Integer;
  dK, dMax, dNmax, q: Integer;
  HasRotPart: Boolean;
  delK, SxyPow: TQuantumNumber;
begin
  Result := z0;
  { K selection }
  if KSelect > -9999 then begin
    { For the moment, require K selection to be diagonal in |K| }
    { TODO: Allow K select to work for elements off diagonal in K }
    if (Abs(K) <> KSelect) or (Abs(Kp) <> KSelect) then
      Exit;
  end;

  { Get powers of S and N operators }
  SxPow := GetSetting(LPI+6+Ord(XAxisFor[Molecule.Representation]));
  SyPow := GetSetting(LPI+6+Ord(YAxisFor[Molecule.Representation]));
  SzPow := GetSetting(LPI+6+Ord(ZAxisFor[Molecule.Representation]));
  xPow := GetSetting(LPI+1+Ord(XAxisFor[Molecule.Representation]));
  yPow := GetSetting(LPI+1+Ord(YAxisFor[Molecule.Representation]));
  zPow := GetSetting(LPI+1+Ord(ZAxisFor[Molecule.Representation]));
  dK := Abs(K-Kp) div 2;
  dMax := xPow + yPow + pPower + SxPow + SyPow + Abs(SpPower);
  if (dK > dMax) or Odd(dK + dMax) then
    Exit;
  dNmax := SxPow + SyPow + SzPow + Abs(SpPower);
  if Abs(N-Np) > 2*dNmax then
    Exit;
  HasRotPart := (JPower or xPow or yPow or zPow or pPower) <> 0;
  if dNmax = 0 then
    Result.re := RotationalElement(N, Kp, K, opFull) * NSOperator(J, N, S, NSPower, SPower)
  else begin
    { Initial implementation - only allow single S operator }
    if SzPow > 0 then
      if HasRotPart then
        Result.re := (RotationalElement(Np,Kp,K,opForward)*SOperator(0,Np,K,N,K,NSPower,SPower,SzPow,opForward)
                      +SOperator(0,Np,Kp,N,Kp,NSPower,SPower,SzPow,opReverse)*RotationalElement(N,Kp,K,opReverse))
      else
        Result.re := SOperator(0,Np,Kp,N,K,NSPower,SPower,SzPow,opFull)
    else if (SxPow > 0) or (SyPow > 0) then
      if not HasRotPart then begin
          Result.re := Sxyn(SxPow,SyPow,J,S,Np,Kp,N,K);
          if (NSPower or SPower) <> 0 then
            Result.re := Result.re*(NSOperator(J,Np,S,NSPower,SPower)+NSOperator(J,N,S,NSPower,SPower));
      end else begin
        Result.re := 0;
        SxyPow := (SxPow + SyPow)*2;
        delK := -SxyPow;
        repeat { Possible delta K values are -Power .. + Power in steps of 2 }
          Result.re := Result.re
                       + RotationalElement(Np,Kp,K+delK,opForward)*Sxyn(SxPow,SyPow,J,S,Np,K+delK,N,K)*NSOperator(J,N,S,NSPower,SPower)
                       + NSOperator(J,Np,S,NSPower,SPower)*Sxyn(SxPow,SyPow,J,S,Np,Kp,N,Kp-delK)*RotationalElement(N,Kp-delK,K,opReverse);
          //if HasRotPart then
          //  Result.re := -RotationalElement(Np,Kp,K-2*SxPow,opForward)*SOperator(2,Np,K-2*SxPow,N,K,NSPower,SxPow,opForward)
          //               -SOperator(2,Np,Kp,N,Kp+2*SxPow,NSPower,SxPow,opReverse)*RotationalElement(N,Kp+2*SxPow,K,opReverse)
          //               +RotationalElement(Np,Kp,K+2*SxPow,opForward)*SOperator(-2,Np,K+2*SxPow,N,K,NSPower,SxPow,opForward)
          //               +SOperator(-2,Np,Kp,N,Kp-2*SxPow,NSPower,SxPow,opReverse)*RotationalElement(N,Kp-2*SxPow,K,opReverse)
          //else
          //  Result.re := -SOperator(2,Np,Kp,N,K,NSPower,SxPow,opFull)+SOperator(-2,Np,Kp,N,K,NSPower,SxPow,opFull)
          Inc(delK, 4);
        until delK > SxyPow;
      end
    //else if SyPow > 0 then
    //  if HasRotPart then
    //    Result.re := (RotationalElement(Np,Kp,K-2,opForward)*SOperator(2,Np,K-2,N,K,NSPower,opForward)
    //                  +SOperator(2,Np,Kp,N,Kp+2,NSPower,opReverse)*RotationalElement(N,Kp+2,K,opReverse)
    //                  +RotationalElement(Np,Kp,K+2,opForward)*SOperator(-2,Np,K+2,N,K,NSPower,opForward)
    //                  +SOperator(-2,Np,Kp,N,Kp-2,NSPower,opReverse)*RotationalElement(N,Kp-2,K,opReverse)
    //                 )/Sqrt(2)
    //  else
    //    Result.re := (SOperator(2,Np,Kp,N,K,NSPower,opFull)+SOperator(-2,Np,Kp,N,K,NSPower,opFull))/Sqrt(2)
    else if SpPower <> 0 then begin
      if Abs(xPow) + Abs(yPow) > 0 then
        raise NotYet; { This implementation does not make sense if combined with Nx or Ny }
      if dK = pPower + SpPower then begin
        if Kp > K then
          q := -1
        else
          q := 1;
        if HasRotPart then
          Result.re := ( RotationalElement(Np,Kp,K-2*q,opForward) * SOperator(2*q,Np,K-2*q,N,K,NSPower,SPower,SpPower,opForward)
                         +SOperator(2*q,Np,Kp,N,Kp+2*q,NSPower,SPower,SpPower,opReverse) * RotationalElement(N,Kp+2*q,K,opReverse))
        else
          Result.re := SOperator(2*q,Np,Kp,N,K,NSPower,SPower,SpPower,opFull);
        Result.re := Math.IntPower(-Sqrt(Two)*q,Abs(SpPower)) * Result.re; { The multiplication by q converts from standard tensor to S+- }
      end;
    end else
      raise NotYet;
  end;
  Result.re := Result.re*Value;
  CMuliPower(Result,Abs(yPow)+Abs(SyPow));
end;

function SymbolicRotationalMatrixElement(PE: TPertEquiv; Molecule: TAsymmetricMolecule): TMatrixElements;
var
  xPow, yPow, zPow: Integer;

  function RotationalMatrixElement(delK:Integer; eN, eK: TExprNode; OpType: TopType):TExprNode;
  var
    eKp, eKpp, Node: TExprNode;
    MaxPow, ddelK: Integer;
  begin
    Result := EB([eN, '+', 1, '*', eN, '^', PE.JPower div 2]);
    eKp := EB([eK, '+', delK]);
    if PE.pPower > 0 then begin
      if (xPow or yPow) <> 0 then
        raise EReportException.Create('AsymmetricTopPerturbation: pPower can''t be used with xPower or yPower');
      if Abs(delK) = PE.pPower then
        Result := OperatorTimes.CreateExprNode(Result,rlE(eN, eK, delK))
      else
        FreeAndNil(Result);
    end else if xPow > 0 then
      if yPow > 0 then begin
        MaxPow := Max(xPow,YPow);
        ddelK := -MaxPow;
        Node := OperatorPlus.CreateExprNode(TIntegerNode.Create(0));
        while ddelK <= MaxPow do begin
          eKpp := EB([eK, '+', ddelK]);
          Node.AddNode(OperatorTimes.CreateExprNode(rlyE(yPow,eN,eKpp,delK-ddelK), rlxE(xPow,eN,eK,ddelK)));
          Node.AddNode(OperatorTimes.CreateExprNode(rlxE(xPow,eN,eKpp,delK-ddelK), rlyE(yPow,eN,eK,ddelK)));
          eKpp.Free;
          Inc(ddelK,1);
        end;
        Result := OperatorTimes.CreateExprNode(Result, Node);
      end else
        Result := OperatorTimes.CreateExprNode(Result, rlxE(xPow,eN,eK,delK))
    else if yPow > 0 then { Note if Odd(yPow), return true result multiplied by -i }
      Result := OperatorTimes.CreateExprNode(Result, rlyE(yPow,eN,eK,delK))
    else if delK <> 0 then
      FreeAndNil(Result);
    if Assigned(Result) then begin
      if zPow > 0 then begin
        if ((xPow or yPow or PE.pPower) <> 0) and (OpType <> opForward) then
          if OpType = opReverse then
            Result := Expr.EN([eKp.Duplicate, '^', Zpow, '*', Result])
          else begin
            Node := Expr.EN([eKp.Duplicate, '^', Zpow, '*', Result.Duplicate]);
            Result := Expr.EN([eK.Duplicate, '^', Zpow, '*', Result, '+', Node]);
          end
        else
          Result := Expr.EN([eK.Duplicate, '^', Zpow, '*', Result]);
      end;
      Result := Result.Simplify;
    end;
    eKp.Free;
  end;

  function SOperator(q: Integer; eNp, eKp, eN, eK, ScaleFactor: TExprNode; NSPower, SPower: Integer; OpPower: Integer; OpType: TopType):TExprNode;
  begin
    Result := T1SnE(Abs(OpPower), q, eJ, eS, eNp, eKp, eN, eK, ScaleFactor);
    if NSPower > 0 then
      case OpType of
        opFull: Result := Expr.EN([NSOperatorE(eJ, eN, eS, NSPower, SPower),
                                   '+', NSOperatorE(eJ, eNp, eS, NSPower, SPower),
                                   '*', Result]);
        opForward: Result := Expr.EN([Result, '*' , NSOperatorE(eJ, eN, eS, NSPower, SPower)]);
        opReverse: Result := Expr.EN([Result, '*', NSOperatorE(eJ, eNp, eS, NSPower, SPower)]);
      end;
  end;

var
  delK, delKmax, delN, dNmax, nMatrixElements: Integer;
  SxPow, SyPow, SzPow, q: Integer;
  E, eNp, eKp, eKpp: TExprNode;
  HasRotPart: Boolean;
  function GetPower(Axis: TabcAxis):Integer;
  begin
    case Axis of
      aAxis: Result := PE.aPower;
      bAxis: Result := PE.bPower;
      cAxis: Result := PE.cPower;
    end;
  end;
  function GetSPower(Axis: TabcAxis):Integer;
  begin
    case Axis of
      aAxis: Result := PE.SaPower;
      bAxis: Result := PE.SbPower;
      cAxis: Result := PE.ScPower;
    end;
  end;
var
  SxyPower, Kinc: Integer;
  ScaleFactor, Scale1: TExprNode;
begin

  SetLength(Result,0);
  xPow := GetPower(XAxisFor[Molecule.Representation]);
  yPow := GetPower(YAxisFor[Molecule.Representation]);
  zPow := GetPower(ZAxisFor[Molecule.Representation]) + PE.zPower;
  SxPow := GetSPower(XAxisFor[Molecule.Representation]);
  SyPow := GetSPower(YAxisFor[Molecule.Representation]);
  SzPow := GetSPower(ZAxisFor[Molecule.Representation]) + PE.SzPower;

  dNmax := SxPow + SyPow + SzPow + Abs(PE.SpPower);
  delKmax := xPow + yPow + PE.pPower + SxPow + SyPow + Abs(PE.SpPower);
  delK := -delKmax;
  SetLength(Result, (2*Abs(delK) + 1)*(2*dNmax+1));
  HasRotPart := (PE.JPower or xPow or yPow or zPow or PE.pPower) <> 0;
  { Calculate scale factor needed to correct for: }
  { Parameter definition}
  ScaleFactor := Expr.EB([PE.Scale4, '/', 4]);
  { undo effect of extra multiplication by -i below }
  if Odd(Abs(SyPow)+Abs(yPow)) and Molecule.Mixture.AllowComplex then
    ScaleFactor := Expr.EN([-1, 'sqrt', '*', ScaleFactor]);
  if SyPow <> 0 then begin
    if Odd(SyPow) then
      { Complex matrix elements - Return actual matrix element times -i }
      ScaleFactor := MonadicMinus.CreateExprNode(ScaleFactor).Simplify;
      { Note: If Odd(yPow), then E needs to be divided by -i to correct for rlyE,
        and then multiplied by i to correct for missing factor here so:
        result = E/(-i)*i = -E, so above also works }
    if Odd(SyPow div 2) then
      { Add factors of i^2 sign }
      ScaleFactor := MonadicMinus.CreateExprNode(ScaleFactor).Simplify;
  end;
  nMatrixElements := 0;
  repeat { For each change in K }
    eKp := EB([eK, '+', delK]);
    delN := -dNmax;
    repeat { For each change in N }
      eNp := EB([eN, '+', delN]);
      if dNmax = 0 then
        E := Expr.EN([RotationalMatrixElement(delK, eN, eK, opFull), '*', NSOperatorE(eJ, eN, eS, PE.NSPower, PE.SPower), '*', ScaleFactor.Duplicate])
      else begin
        { Initial implementation - only allow single S operator (to power 1) }
        if SzPow > 0 then
          if HasRotPart then begin
            E := Expr.EN([RotationalMatrixElement(delK,eNp,eK,opForward), '*', SOperator(0,eNp,eK,eN,eK,ScaleFactor,PE.NSPower,PE.SPower,SzPow,opForward)]);
            E := Expr.EN([SOperator(0,eNp,eKp,eN,eKp,ScaleFactor,PE.NSPower,PE.SPower,SzPow,opReverse), '*', RotationalMatrixElement(delK,eN,eK,opReverse), '+', E]);
          end else
            E := SOperator(0,eNp,eKp,eN,eK,ScaleFactor,PE.NSPower,PE.SPower,SzPow,opFull)
        else if (SxPow > 0) or (SyPow > 0) then begin
          if not HasRotPart then begin
            E := SxynE(SxPow,SyPow,delK,eJ,eS,eNp,eKp,eN,eK,ScaleFactor);
            if (PE.NSPower or PE.SPower) <> 0 then
              E := Expr.EN([NSOperatorE(eJ,eNp,eS,PE.NSPower,PE.SPower), '+', NSOperatorE(eJ,eN,eS,PE.NSPower,PE.SPower), '*', E]);
          end else begin
            E := nil;
            SxyPower := SxPow + SyPow;
            Kinc := -SxyPower;
            repeat { Possible delta K values are -Power .. + Power in steps of 2 }
              eKpp := EB([eK, '+', Kinc]);
              E := Expr.EN([RotationalMatrixElement(delK-Kinc,eNp,eKpp,opForward), '*', SxynE(SxPow,SyPow,Kinc,eJ,eS,eNp,eKpp,eN,eK,ScaleFactor), '*', NSOperatorE(eJ,eN,eS,PE.NSPower,PE.SPower), '+', E]);
              eKpp.Free;
              eKpp := EB([eKp, '-', Kinc]);
              E := Expr.EN([NSOperatorE(eJ,eNp,eS,PE.NSPower,PE.SPower), '*', SxynE(SxPow,SyPow,Kinc,eJ,eS,eNp,eKp,eN,eKpp,ScaleFactor), '*', RotationalMatrixElement(delK-Kinc,eN,eK,opReverse), '+', E]);
              eKpp.Free;
              Inc(Kinc, 2);
            until Kinc > SxyPower;
          end;
          //if HasRotPart then begin
          //  eKpp := EB([eK, '-', 1]);
          //  E := Expr.EN([RotationalMatrixElement(delK+1,eNp,eKpp,opForward), '*', SOperator(1,eNp,eKpp,eN,eK,PE.NSPower,opForward)]);
          //  eKpp.Free;
          //  eKpp := EB([eKp, '+', 1]);
          //  E := Expr.EN([SOperator(1,eNp,eKp,eN,eKpp,PE.NSPower,opReverse), '*', RotationalMatrixElement(delK+1,eN,eK,opReverse), '+', E]);
          //  eKpp.Free;
          //  eKpp := EB([eK, '+', 1]);
          //  if SxPow > 0 then
          //    Op := '-'
          //  else
          //    Op := '+';
          //  E := Expr.EN([RotationalMatrixElement(delK-1,eNp,eKpp,opForward), '*', SOperator(-1,eNp,eKpp,eN,eK,PE.NSPower,opForward), Op, E]);
          //  eKpp.Free;
          //  eKpp := EB([eKp, '-', 1]);
          //  E := Expr.EN([SOperator(-1,eNp,eKp,eN,eKpp,PE.NSPower,opReverse), '*', RotationalMatrixElement(delK-1,eN,eK,opReverse), '+', E]);
          //  eKpp.Free;
          //end else begin
          //  E := SOperator(1,eNp,eKp,eN,eK,PE.NSPower,opFull);
          //  if SxPow > 0 then
          //    Op := '-'
          //  else
          //    Op := '+';
          //  E := Expr.EN([SOperator(-1,eNp,eKp,eN,eK,PE.NSPower,opFull), Op, E]);
          //end;
          //E := Expr.EN([E, '/', Expr.EN([2, 'sqrt'])]);
          { Factor of i^SyPow missed out above }
        end else if Abs(PE.SpPower) > 0 then begin
          if Abs(xPow) + Abs(yPow) > 0 then
            raise NotYet; { This implementation does not make sense if combined with Nx or Ny }
          if Abs(delK) = PE.pPower + PE.SpPower then begin
            if delK > 0 then
              q := -1
            else
              q := 1;
            { The multiplication by q converts from standard tensor to S+- }
            Scale1 := Expr.EN([2, 'sqrt', '*', -q, '^', Abs(PE.SpPower), '*', ScaleFactor.Duplicate]);
            if HasRotPart then begin
              eKpp := EB([eK, '-', q]);
              E := Expr.EN([RotationalMatrixElement(delK+q,eNp,eKpp,opForward), '*', SOperator(q,eNp,eKpp,eN,eK,Scale1,PE.NSPower,PE.SPower,PE.SpPower,opForward)]);
              eKpp.Free;
              eKpp := EB([eKp, '+', q]);
              E := Expr.EN([SOperator(q,eNp,eKp,eN,eKpp,Scale1,PE.NSPower,PE.SPower,PE.SpPower,opReverse), '*', RotationalMatrixElement(delK+q,eN,eK,opReverse), '+', E]);
              eKpp.Free;
            end else
              E := SOperator(q,eNp,eKp,eN,eK,Scale1,PE.NSPower,PE.SPower,PE.SpPower,opFull);
            Scale1.Free;
          end else
            E := nil;
        end else
          raise NotYet;
      end;
      if Assigned(E) then begin
        if E.EqualTo(0) then
          FreeAndNil(E);
      end;
      if Assigned(E) then begin
        Result[nMatrixElements].Expr := E;
        Result[nMatrixElements].Evaluator := nil;
        Result[nMatrixElements].CEvaluator := nil;
        SetLength(Result[nMatrixElements].Signature,2);
        Result[nMatrixElements].Signature[0] := delK;
        Result[nMatrixElements].Signature[1] := delN;
        Inc(nMatrixElements);
      end;
      eNp.Free;
      Inc(delN,1);
    until delN > dNmax;
    eKp.Free;
    Inc(delK,1);
  until delK > delKmax;
  ScaleFactor.Free;
  SetLength(Result,nMatrixElements)
end;

function TAsymmetricTop.SymbolicRotationalMatrixElement(AIndex: Integer; RankWanted: Integer): TMatrixElements;
var
  PE, PE1, PE2: TPertEquiv;
  M: TMatrixElements;
begin
  PE := FindPertEquiv(ParameterNames[AIndex], Molecule.SReduction, Molecule.Representation, PE1, PE2);
  if PE.Name = '' then
    Result := inherited SymbolicRotationalMatrixElement(AIndex, RankWanted)
  else begin
    Result := OAsymT.SymbolicRotationalMatrixElement(PE, Molecule);
    if PE1.Name <> '' then begin
      M := OAsymT.SymbolicRotationalMatrixElement(PE1, Molecule);
      MergeElements(Result, M, 1, False, True);
    end;
    if PE2.Name <> '' then begin
      M := OAsymT.SymbolicRotationalMatrixElement(PE2, Molecule);
      MergeElements(Result, M, 1, False, True);
    end;
  end;
end;

function TAsymmetricTopPerturbation.Molecule: TAsymmetricMolecule;
begin
  if Owner.Owner is TAsymmetricMolecule then
    Result := TAsymmetricMolecule(Owner.Owner)
  else
    Result := Owner.Owner.Owner as TAsymmetricMolecule;
end;

function TAsymmetricTopPerturbation.PairName: string;
  function Show(Letter: string; Power: Integer): string;
  begin
    case Power of
      0: Result := '';
      1: Result := Letter;
    else
      Result := Letter + '^' + IntToStr(Power)
    end;
  end;
begin
  Result := Show('J',JPower)
            + Show('a',aPower) + Show('b',bPower) + Show('c',cPower)
            + Show('+',pPower)
            + Show('Sa',SaPower) + Show('Sb',SbPower) + Show('Sc',ScPower)
            + Show('S+',SpPower) + Show('S',SPower) + Show('NS',NSPower);
  if SymSelect >= 0 then
    Result := Result + Molecule.JAdjustedSymName(0,SymSelect);
  if KSelect > -9999 then
    Result := Result + '(K=' + Molecule.Mixture.QuantumNumberToStr(KSelect) + ')';
end;

function TAsymmetricTopPerturbation.ReadProperties(Node: TDOMNode; CheckUnread: Boolean): Integer;
var
  NyPower, SyPower, Phase: Integer;
begin
  Result := inherited ReadProperties(Node, CheckUnread);
  if ReadVersion < 7001138 then begin
    { Apply sign change to operator to compensate for introduction of approximate handling of imaginary operators before 7.1.138 }
    SyPower := GetSetting(LPI+6+Ord(YAxisFor[Molecule.Representation]));
    NyPower := GetSetting(LPI+1+Ord(YAxisFor[Molecule.Representation]));
    Phase := 1;
    if SyPower > 0 then
      Phase := -1;
    if (SyPower mod 4) > 2 then
      Phase := -Phase;
    if (NyPower mod 4) > 2 then
      Phase := -Phase;
    if Odd(NyPower) then
      Phase := Minus(NyPower+1)*Phase
    else
      Phase := Minus(NyPower)*Phase;
    if Phase < 0 then begin
      Value := -Value;
      {$IFNDEF CONSOLE}
      Writeln(Log, 'Sign of ', FullHeirarchicName, ' perturbation changed to account for change in handling of imaginary operators');
      {$ENDIF}
    end;
  end;
end;

const
  AsymmetricTopPerturbationSettingNames: array[LPI..LPI+11] of string = (
'Twice power of J(J+1)',
'Power of Ja',
'Power of Jb',
'Power of Jc',
'Power of J+-',
'K the perturbation applies to',
'Power of Sa',
'Power of Sb',
'Power of Sc',
'Power of S+-',
'Twice power of S(S+1)',
'Power of N.S'
);

constructor TAsymmetricTopPerturbation.Create(AOwner: TComponent);
begin
  SavedProperties := SavedProperties + ',KSelect=all';
  inherited;
end;

function TAsymmetricTopPerturbation.HelpOnSetting(AIndex: Integer): string;
begin
  if (AIndex >= Low(AsymmetricTopPerturbationSettingNames))
     and (AIndex <= High(AsymmetricTopPerturbationSettingNames)) then
    Result := AsymmetricTopPerturbationSettingNames[AIndex]
  else
    Result := inherited HelpOnSetting(AIndex);
end;

function TAsymmetricTopPerturbation.IsSettingQuantum(AIndex: Integer): Boolean;
begin
  Result := AIndex = LPI+5;
  if not Result then
    Result := inherited IsSettingQuantum(AIndex);
end;

procedure TAsymmetricTopPerturbation.SetSettingFromStr(AIndex: Integer;
  const s: string);
begin
  if (AIndex = LPI+5) and (s = 'all') then
    KSelect := -9999
  else
    inherited;
end;

procedure TAsymmetricTopPerturbation.SetFromPertEquiv(const PE: TPertEquiv);
var
  abcIndex: Integer;
begin
  JPower := PE.JPower;
  aPower := PE.aPower;
  bPower := PE.bPower;
  cPower := PE.cPower;
  pPower := PE.pPower;
  if PE.zPower > 0 then begin
    abcIndex := Ord(ZAxisFor[Molecule.Representation]) - Ord(aAxis) + IndexOfSetting('aPower');
    SetSetting(abcIndex, GetSetting(abcIndex) + PE.zPower);
  end;
  SaPower := PE.SaPower;
  SbPower := PE.SbPower;
  ScPower := PE.ScPower;
  SpPower := PE.SpPower;
  if PE.SzPower > 0 then begin
    abcIndex := Ord(ZAxisFor[Molecule.Representation]) - Ord(aAxis) + IndexOfSetting('SaPower');
    SetSetting(abcIndex, GetSetting(abcIndex) + PE.SzPower);
  end;
  SPower := PE.SPower;
  NSPower := PE.NSPower;
end;

function TAsymmetricTopPerturbation.SettingToStr(AIndex: Integer): string;
begin
  if (AIndex = LPI+5) and (KSelect <= -9999) then
      Result := 'all'
  else
    Result := inherited SettingToStr(AIndex);
end;

function TAsymmetricTopPerturbation.SymbolicMatrixElements(WithHyperfine: Boolean): TMatrixElements;
var
  PE: TPertEquiv;
begin
  PE.aPower := aPower;
  PE.bPower := bPower;
  PE.cPower := cPower;
  PE.JPower := JPower;
  PE.pPower := pPower;
  PE.zPower := 0;
  PE.Scale4 := 4;
  PE.Name := PairName;
  PE.SaPower := SaPower;
  PE.SbPower := SbPower;
  PE.ScPower := ScPower;
  PE.SpPower := SpPower;
  PE.SPower := SPower;
  PE.NSPower := NSPower;
  PE.SzPower := 0;
  if (SymSelect >= 0) or (KSelect > -9999) then
    raise NotYet; { TODO: Add to TPertEquiv }
  Result := OAsymT.SymbolicRotationalMatrixElement(PE,Molecule);
end;

{ TAsymmetricTopNucleus }

const
  AsymmetricTopNucleusParameterNames: array[1..14] of string = (
'Nuclear spin - electron spin Fermi Contact interaction (I.S).',
'Tzz = -(Txx+Tyy) component of Nuclear spin - electron spin dipole dipole interaction.',
'Txx-Tyy component of Nuclear spin - electron spin dipole dipole interaction.',
'ab component of Nuclear spin - electron spin dipole dipole interaction.',
'ac component of Nuclear spin - electron spin dipole dipole interaction.',
'bc component of Nuclear spin - electron spin dipole dipole interaction.',
'zz component of Nuclear Quadrupole interaction.',
'xx-yy component of Nuclear Quadrupole interaction.',
'ab component of Nuclear Quadrupole interaction.',
'ac component of Nuclear Quadrupole interaction.',
'bc component of Nuclear Quadrupole interaction.',
'aa component of Nuclear Spin - Rotation interaction.',
'bb component of Nuclear Spin - Rotation interaction.',
'cc component of Nuclear Spin - Rotation interaction.'
);

function TAsymmetricTopNucleus.FixupProperties: Exception;
  function efind(Axis1, Axis2: TabcAxis; Stem: string):Double;
  var
    s: string;
  begin
    if Axis1 < Axis2 then
      s := AxisNames[Axis1] + AxisNames[Axis2]
    else
      s := AxisNames[Axis2] + AxisNames[Axis1];
    Result := GetParameterByName(Stem+s)
  end;
var
  Representation: TRepresentation;
begin
  (Molecule as TAsymmetricMolecule).MakeSecondRank(Self, TT, 'T');
  (Molecule as TAsymmetricMolecule).MakeSecondRank(Self, CHI, 'CHI');

  if Spin > 1 then
    QuadI := 1/threej(Spin, -Spin, 4, 0, Spin, Spin)
  else
    QuadI := 0;

  { Nuclear spin rotation
    Note more components could be non-zero, but will rarely contribute.
    (They cause effects such as ortho-para conversion.) }
  Representation := (Molecule as TAsymmetricMolecule).Representation;
  TC00 := -(Caa + Cbb + Ccc);
  TC20 := (3*efind(ZAxisFor[Representation], ZAxisFor[Representation], 'C') + TC00)/Sqrt(Six);
  TC00 := TC00/Sqrt(Three);
  TC22 := (efind(XAxisFor[Representation], XAxisFor[Representation], 'C')
           -efind(YAxisFor[Representation], YAxisFor[Representation], 'C'))/2;

  Result:=inherited FixupProperties;
end;

function TAsymmetricTopNucleus.HasIdentical: Boolean;
begin
  Result := (Molecule as TAsymmetricMolecule).PointGroup in [C2,C2v,D2,D2h,C2h];
end;

function TAsymmetricTopNucleus.HelpOnParameter(AIndex: Integer): string;
begin
  if (AIndex >= Low(AsymmetricTopNucleusParameterNames))
     and (AIndex <= High(AsymmetricTopNucleusParameterNames)) then
    Result := AsymmetricTopNucleusParameterNames[AIndex]
  else
    Result := inherited HelpOnParameter(AIndex);
end;

{ TAsymmetricTopOrbitalTransitionMoment }

constructor TAsymmetricTopOrbitalTransitionMoment.Create(AOwner: TComponent);
begin
  FNuclearMagneton := False;
  FMagnetic := True;
  { Default to including the isotropic spin term from the electron(s) }
  SavedProperties := SavedProperties + ',gsaa='+FloatToStr(ge)+',gsbb='+FloatToStr(ge)+',gscc='+FloatToStr(ge);
  DimensionlessParameters := DimensionlessParameters + ',gsaa,gsbb,gscc,gsabbar,gsacbar,gsbcbar,gsamb,gsamc,gsbmc';
  inherited;
end;

function TAsymmetricTopOrbitalTransitionMoment.FixupProperties: Exception;
begin
  Result := inherited FixupProperties;
  (Owner.Owner as TAsymmetricMolecule).MakeSpherical(Self,g,'gs');
end;

function TAsymmetricTopOrbitalTransitionMoment.PairName: string;
begin
  Result := 'gs';
end;

{ TAsymmetricTopRotationalTransitionMoment }

constructor TAsymmetricTopRotationalTransitionMoment.Create(AOwner: TComponent);
begin
  FNuclearMagneton := True;
  FMagnetic := True;
  DimensionlessParameters := DimensionlessParameters + ',graa,grbb,grcc,grabbar,gracbar,grbcbar,gramb,gramc,grbmc';
  inherited;
end;

function TAsymmetricTopRotationalTransitionMoment.FixupProperties: Exception;
begin
  Result := inherited FixupProperties;
  (Owner.Owner as TAsymmetricMolecule).MakeSpherical(Self,g,'gr');
end;

function TAsymmetricTopRotationalTransitionMoment.PairName: string;
begin
  Result := 'gr';
end;

{ TDeltaK2TransitionMoment }

function TDeltaK2TransitionMoment.CheckProjections(K1, K2: TQuantumNumber): Boolean;
begin
  { Need special handling, as overall rank one tensor in terms of external field, but
    K changes determined by rank 2 tensor. (The rank 2 tensor is coupled with N or S to
    give overall rank 1}
  Result := (Abs(K1-K2) <= 4) or (Abs(K1+K2) <= 4);
end;

{ TAsymmetricTopIntensityPerturbation }

function TAsymmetricTopIntensityPerturbation.IsAntiSym: Boolean;
begin
  Result := AntiSym;
end;

function TAsymmetricTopIntensityPerturbation.BraOwner: TMNode;
begin
  Result := (Owner as TTransitionMoment).BraOwner;
end;

function TAsymmetricTopIntensityPerturbation.KetOwner: TMNode;
begin
  Result := (Owner as TTransitionMoment).KetOwner;
end;

constructor TAsymmetricTopIntensityPerturbation.Create(AOwner: TComponent);
var
  T: TTransitionMoment;
begin
  T := AOwner as TTransitionMoment;
  FBraName := T.FBraName;
  FKetName := T.FKetName;
  inherited;
end;

function TAsymmetricTopIntensityPerturbation.HelpOnSetting(AIndex: Integer): string;
begin
  if AIndex = LPI+12 then
    Result := 'Use Dipole Op - Op Dipole if true'
  else
    Result := inherited HelpOnSetting(AIndex);
end;

initialization
  TAsymmetricMolecule.RegisterClass(TSpecies);
  TAsymmetricManifold.RegisterClass(TAsymmetricMolecule);
  TAsymmetricTop.RegisterClass(TAsymmetricManifold);
  TAsymmetricTopPerturbation.RegisterClass(TAsymmetricManifold);
  TAsymmetricTopNucleus.RegisterClass(TAsymmetricTop);
  TAsymmetricTopSpinSpin.RegisterClass(TAsymmetricTop);
  TAsymmetricTopOrbitalTransitionMoment.RegisterClass(TTransitionMoments);
  TAsymmetricTopRotationalTransitionMoment.RegisterClass(TTransitionMoments);
  { TODO: Commented out above }
//  TAsymmetricTopIntensityPerturbation.RegisterClass(TCartesianTransitionMoment); { TODO: More transition moment types? }
//  TAsymmetricTopIntensityPerturbation.RegisterClass(TSphericalTransitionMoment);
  VList := MakeSortedVariables(['N','K', 'J', 'S'], [@eN, @eK, @eJ, @eS]);
  { For use if CHECKSIMPILFY set in Expr }
  eN.Value := 0;
  eK.Value := 0;
  eJ.Value := 1;
  eS.Value := 1;
finalization
  FreeAndNil(VList);
end.
