unit OscFields;
{$I utils.inc}{Copyright (C) 2009 C M Western; see commandline.pas for license}

{ Objects under TSimulation to describe polarization of applied field }

interface

uses Classes, MNode, QuantumNumber, Mixture;

type

  TOscillatingField = class(TMNode)
  public
    constructor Create(AOwner: TComponent); override;
    function HelpOnParameter(AIndex: Integer): string; override;
    procedure SetModified(const Value: Boolean; Initiator: TMNode); override;
  published
    property Strength:  Double index 1 read GetParameter write SetParameter;
  end;

  TDirection = (poRandom, poParallel, poPerpendicular, poCircular);

  TPolarization = class(TOscillatingField)
  private
    function GetDirection(const Index: Integer): TDirection;
    procedure SetDirection(const Index: Integer; const Value: TDirection);
  public
    function GetSettingsNames(AIndex: Integer; out LowestValue:Integer): PPChar; override;
  published
    property Direction: TDirection index 1 read GetDirection write SetDirection;
  end;

function Mfactor(Jp, Mp, Rank, J, M: TQuantumNumber; Simulation:TSimulation):Double;

implementation

uses Math, MathSup, Global, nj;

{ TOscillatingField }

constructor TOscillatingField.Create(AOwner: TComponent);
begin
  SavedProperties := SavedProperties + ',Strength=1';
  DimensionlessParameters := DimensionlessParameters + ',Strength';
  inherited;
end;

function TOscillatingField.HelpOnParameter(AIndex: Integer): string;
begin
  if AIndex = 1 then
    Result := 'Relative strength of field - note that the relative intensity is proportional to the square of this value.'
  else
    Result := inherited HelpOnParameter(AIndex);
end;

procedure TOscillatingField.SetModified(const Value: Boolean; Initiator: TMNode);
begin
  inherited;
  if Value then
    RootNode.ClearCache(True);
end;

{ TPolarization }

const
  DirectionNames: array [0..4] of PChar = ('Random', 'Parallel', 'Perpendicular', 'Circular', nil);

function TPolarization.GetDirection(const Index: Integer): TDirection;
begin
  Result := TDirection(GetSetting(Index));
end;

function TPolarization.GetSettingsNames(AIndex: Integer; out LowestValue: Integer): PPChar;
begin
  Result := inherited GetSettingsNames(AIndex,LowestValue);
  if SettingNames[AIndex] = 'Direction' then
    Result := @DirectionNames[0];
end;

procedure TPolarization.SetDirection(const Index: Integer; const Value: TDirection);
begin
  SetSetting(Index, Ord(Value));
end;

{ Computation of parallel/perpendicular ratio }
var
  ParallelFactors: array of Double;
  PerpendicularFactors: array of TDoubleArray;

{ qth component of rank k vector formed by coupling rank 1 electric field vector
  with itself, with the rank 1 vector being a unit vector along the z (parallel)
  or x (perpendicular) axes. }
function TE(k, q: TQuantumNumber; Parallel: Boolean): Double;
var
  p, q1: TQuantumNumber;
  i: Integer;
  T: TDoubleArray;
  function TEpart(p, q: TQuantumNumber): Double;
  begin
    Result := Minus(4-k+q) * Sqrt(k+One) * threej(2,p,k-2,-p-q,k,q)*TE(2,p,Parallel)*TE(k-2,-p-q,Parallel);
  end;
begin
  i := k div 2 - 1;
  Result := 0;
  if Parallel then begin
    if q <> 0 then
      Exit;
    if i >= Length(ParallelFactors) then begin
      Result := TEpart(0,0);
      SetLength(ParallelFactors, i+1);
      ParallelFactors[i] := Result;
    end else
      Result := ParallelFactors[i];
  end else if Abs(q) <= k then begin
    if i >= Length(PerpendicularFactors) then begin
      SetLength(T, k+1);
      q1 := -k;
      repeat
        Result := 0;
        p := -2;
        repeat
          Result := Result + TEpart(p,q1);
          Inc(p,2);
        until p > 2;
        T[(q1+k) div 2] := Result;
        Inc(q1,2);
      until q1 > k;
      SetLength(PerpendicularFactors, i+1);
      PerpendicularFactors[i] := T;
    end;
    Result := PerpendicularFactors[i][(q + k) div 2];
  end;
end;

{ M dependence to transtion moment }
function Mfactor(Jp, Mp, Rank, J, M: TQuantumNumber; Simulation:TSimulation):Double;
  function Tkp(p: TQuantumNumber):Double;
  begin
    Result := Minus(Jp-Mp) * threej(Jp, -Mp, Rank, p, J, M);
  end;
  function PolFactor(Direction: TDirection):Double;
  begin
    { Choose right values of p; scale factors to give same total
      absorption independent of polarization. Scale factors come from the
      fact that the sum over a given component, p, is independent of p.
      The scale factors squared should then sum to one. }
    Result := 0;
    case Direction of
      poRandom:
        if Abs(Mp-M) <= Rank then begin
          Result := Tkp(Mp-M);
          Result := Result/Sqrt(Rank+1);
        end;
      poParallel: if M = Mp then
                    Result := Tkp(0)*TE(Rank,0,True);
      poPerpendicular:
//        if Mp > M then
//          Result := Tkp(Rank)/Sqrt(Two)
//        else if (Rank = 4) and (Mp = M) then { TODO: Generalize }
//          Result := Tkp(0)*Sqrt(Two) { TODO: *Sqrt(Two) is a guess, and overall factor will need adjustment }
//        else
//          Result := Tkp(-Rank)/Sqrt(Two);
        Result := Tkp(Mp-M)*TE(Rank,Mp-M,False);
      poCircular: Result := Tkp(Rank); { TODO: left/right combinations, and other mixtures }
    end;
  end;
var
  i: Integer;
begin
  if Assigned(Simulation) then
    if Simulation.ComponentCount > 0 then begin
      Result := 0;
      for i := 0 to Simulation.ComponentCount - 1 do
        if Simulation.Components[i] is TPolarization then
          with TPolarization(Simulation.Components[i]) do
            Result := Result + Strength * PolFactor(Direction)
    end else
      Result := PolFactor(poRandom) { Default if polarisation not specified }
  else
    Result := Tkp(0) { Normally simple static field }
end;

initialization
  { T(1,0) }
  SetLength(ParallelFactors, 1);
  ParallelFactors[0 { Rank div 2 - 1}] := 1;
  { T(1,+/-1) }
  SetLength(PerpendicularFactors, 1, 3);
  PerpendicularFactors[0][0] :=  1/Sqrt(Two);
  PerpendicularFactors[0][2] := -1/Sqrt(Two);
  TPolarization.RegisterClass(TSimulation);
end.

{ For testing:
Output is:
Parallel
k  q
1	 0	1
2	 0	0.816496581 = sqrt(2/3)
3	 0	0.632455532 = sqrt(2/5)
4	 0	0.478091444

Perpendicular
1	-1	 0.707106781 =  1/sqrt(2)
1	 1	-0.707106781 = -1/sqrt(2)

2	-2	1/2
2	 0	-0.40824829  = -1/sqrt(6)
2	 2	1/2

3	-3	-0.353553391 = -1/sqrt(8)
3	-1	 0.273861279 =  sqrt(3/40)
3	 1	-0.273861279 = -sqrt(3/40)
3	 3	 0.353553391 =  1/sqrt(8)

4	-4	-1/4
4	-2	 0.188982237
4	 0	-0.179284291
4	 2	 0.188982237
4	 4	-1/4

procedure ShowParPerpScaleFactors;
var
  k, q: TQuantumNumber;
begin
  k := 2;
  repeat
    q := -k;
    repeat
      WriteLn(k div 2:3, q div 2:3, ' ', TE(k,q,True):20:13, ' ', TE(k,q,False):20:13);
      Inc(q,2);
    until q > k;
    Inc(k, 2);
  until k > 8;
end;
}


