unit PDFreader;
{$H+}{Copyright (C) 2011 C M Western; see commandline.pas for license}
interface
{ Read images from a pdf file into an overlay }

uses Classes, UCompat, {$IFNDEF FPC} pgmd5 {$ELSE} md5 {$ENDIF};

type

  TOnImageStream = procedure (Sender: TObject; Stream: TStream; const Ext: string) of Object;

  TTokType = (pdfNumber, pdfString, pdfMarker, pdfName, pdfEOF);
  TObjType = (oNumber, oReal, oIndirect, oDictionary, oArray, oString, oStream);
  TPDFDictionary = class;
  TPDFObject = record
    case oType: TObjType of
      oNumber, oIndirect: (Value: Integer);
      oReal: (Rvalue: Double);
      oDictionary, oArray, oString: (Dictionary: TPDFDictionary);
    end;
  TPDFReader = class;

  { TPDFDictionary }

  TPDFDictionary = class
    Reader: TPDFReader;
    Keys: array of string; { Empty for array }
    Values: array of TPDFObject; { Empty for string }
    Number, Gen, Position: Integer;
    constructor Create(AReader: TPDFReader); overload;
    constructor Create(S: string); overload;
    destructor Destroy; override;
    procedure Show(const Indent: string);
    function Find(const Name: string):TPDFObject;
    function FindStr(const Name: string):string;
    function FindDictionary(const Name: string):TPDFDictionary;
    function FindNumber(const Name: string; Deflt: Integer = Low(Integer)):Integer;
    function FindArray(const Name: string):TPDFDictionary;
  end;

  { TPDFReader }

  TPDFReader = class
    Stream: TStream;
    Token: string;
    Ch: AnsiChar;
    Eof: Boolean;
    XRef, XRefIndex: array of Integer;
    Trailer, Catalog, Encrypt, PageTree: TPDFDictionary;
    nPages: Integer;
    Pages: array of TPDFDictionary;
    LastToken: TTokType;
    FileId, OwnerKey, UserKey: string;
    FileKey: TMD5Digest;
    PermFlags, EncRevision, KeyLength: Integer;
    EncryptMetadata: Boolean;
    FileName: string;
  public
    constructor Create(AFileName: string);
    destructor Destroy; override;
    procedure Deconstruct(AStream: TStream);
    function NextCh:AnsiChar;
    function HexVal:Integer;
    function NextToken:TTokType;
    function NextNumber:Double;
    function ReadObj(TokType: TTokType):TPDFObject;
    function ReadDictionary:TPDFDictionary;
    function ReadArray: TPDFDictionary;
    procedure ReadXref;
    procedure ReadXrefStream(Obj:TPDFObject);
    function ReadLabelledObj(Number: Integer):TPDFObject;
    procedure ReadTrailer;
    procedure LoadIndirect(var Obj: TPDFObject);
    procedure ReadEncryption;
    procedure ProcessPage(n: Integer);
    procedure FindImages(n: Integer; OnImageStream: TOnImageStream);
    procedure ProcessImage(var Obj: TPDFObject; OnImageStream: TOnImageStream);
    procedure SaveImageToFile(Sender: TObject; AStream: TSTream; const Ext: string);
    function ConstructStream(var Obj: TPDFObject; out UnhandledFilters: string; NilOK: Boolean = False): TStream;
    function IsDictionaryOrNull(Obj: TPDFObject):Boolean;
    procedure ForAllImages(OnImageStream:TOnImageStream);
  end;

const
  nilObject: TPDFObject = (oType:oArray; Dictionary: nil);

implementation

uses SysUtils, Math, PDFStream, Flate, ASCII85, CCITT, LZW, RC4, Worker { TODO: testing only};

procedure NotYet;
begin
  raise Exception.Create('Not implemented part of pdf reader');
end;

procedure ShowPDFObject(O: TPDFObject; const Indent: string = '');
begin
  case O.oType of
    oNumber:     WriteLn(Indent, 'Number: ', O.Value);
    oReal:       WriteLn(Indent, 'Real: ', O.Rvalue);
    oIndirect:   WriteLn(Indent, 'Indirect: ', O.Value);
    oDictionary: begin
                   WriteLn(Indent, '<<');
                   O.Dictionary.Show(Indent+'  ');
                   Writeln(Indent, '>>');
                 end;
    oArray:      begin
                   WriteLn(Indent, '[');
                   O.Dictionary.Show(Indent+'  ');
                   Writeln(Indent, ']');
                 end;
    oString:     WriteLn(Indent,'String: ',O.Dictionary.Keys[0]);
    oStream:     begin
                   WriteLn(Indent,'Stream:<<',O.Dictionary.Position);
                   O.Dictionary.Show(Indent+'  ');
                   Writeln(Indent, '>>');
                 end;
  end;
end;

procedure FreePDFObject(var O:TPDFObject);
begin
  case O.oType of
    oNumber, oReal, oIndirect: ;
    oDictionary, oArray, oString, oStream:
      FreeAndNil(O.Dictionary);
  end;
end;

{ TPDFReader }

function TPDFReader.NextNumber:Double;
begin
  if (NextToken <> pdfNumber) or not TryStrToFloat(Token,Result) then
    raise Exception.Create('PDF: Expected number found '+Token);
end;

function TPDFReader.ReadDictionary:TPDFDictionary;
var
  TokType: TTokType;
  TokRead: Boolean;
  n: Integer;
begin
  Result := TPDFDictionary.Create(Self);
  TokRead := False;
  n := 0;
  repeat
    if not TokRead then
      TokType := NextToken;
    TokRead := False;
    if (TokType = pdfMarker) and (Token = '>') then
      Exit;
    if TokType <> pdfName then
      raise Exception.Create('PDF: Expected name found:'+Token);
    Setlength(Result.Keys, n+1);
    Result.Keys[n] := Token;;
    TokType := NextToken;
    SetLength(Result.Values, n+1);
    Result.Values[n] := nilObject;
    if (TokType = pdfMarker) and (Token = '>') then
      Exit; { Allow missing last entry }
    Result.Values[n] := ReadObj(TokType);
    TokType := NextToken;
    if TokType <> pdfNumber then
      TokRead := True
    else begin
      if Token <> '0' then
        NotYet;
      if (NextToken <> pdfMarker) or (Token <> 'R') then
        raise Exception.Create('PDF: Expected R found '+Token);
      Result.Values[n].oType := oIndirect;
    end;
    Inc(n);
  until Eof;
end;

function TPDFReader.ReadArray: TPDFDictionary;
var
  TokType: TTokType;
  n: Integer;
begin
  Result := TPDFDictionary.Create;
  n := 0;
  repeat
    TokType := NextToken;
    if TokType = pdfMarker then
      if Token = ']' then
        Break
      else if Token = 'R' then
        { indirect object - preceeding two elements should be numbers }
        if (n > 1) and (Result.Values[n-2].oType = oNumber) and (Result.Values[n-1].oType = oNumber) then begin
          Result.Values[n-2].oType := oIndirect;
          Dec(n);
          Continue;
        end else
          raise Exception.Create('PDF: Invalid indirect object');
    SetLength(Result.Values,n+1);
    Result.Values[n] := ReadObj(TokType);
    Inc(n);
  until eof;
  SetLength(Result.Values,n);
end;

procedure TPDFReader.ReadXref;
var
  FirstEntry, nEntries, Offset, i: Integer;
begin
  while NextToken = pdfNumber do begin
    FirstEntry := StrToInt(Token);
    nEntries := Round(NextNumber);
    if FirstEntry+nEntries > Length(XRef) then
      SetLength(XRef,FirstEntry+nEntries);
    for i := 0 to nEntries-1 do begin
      Offset := Round(NextNumber);
      { Gen := } NextNumber;
      if NextCh = 'n' then begin
        XRef[FirstEntry+i] := Offset;
      end;
      NextCh;
    end;
  end;
end;

procedure TPDFReader.ReadXrefStream(Obj: TPDFObject);
var
  XRefStream: TStream;
  Wd: TPDFDictionary;
  W, V: array [0..2] of Integer;
  Size, N, i, j: Integer;
  EoS: Boolean;
  B: Byte;
  UnhandledFilters: string; { Not used }
begin
  XRefStream := ConstructStream(Obj,UnhandledFilters);
  try
    Wd := Obj.Dictionary.FindArray('W');
    Size := Obj.Dictionary.FindNumber('Size');
    if Length(XRef) < Size then
      SetLength(XRef,Size);
    SetLength(XRefIndex,Length(Xref));
    if Length(Wd.Values) <> 3 then
      NotYet;
    for i := 0 to 2 do
      if Wd.Values[i].oType <> oNumber then
        NotYet
      else
        W[i] := Wd.Values[i].Value;
    Eos := False;
    N := 0;
    while not Eos do begin
      for i := 0 to 2 do begin
        V[i] := 0;
        for j := 0 to W[i]-1 do
          if XRefStream.Read(B,1) <> 1 then begin
            Eos := True;
            Break;
          end else
            V[i] := (V[i] shl 8) or B;
        if Eos then
          Break;
      end;
      if Eos then
        Break;
      if W[0] = 0 then
        V[0] := 1;
      // WriteLn(N:4, V[0]:2, V[1]:15, V[2]:10);
      case V[0] of
        0: { Free object };
        1: begin XRef[N] := V[1]; end;
        2: begin XRef[N] := -V[1]; XRefIndex[N] := V[2]; end { Object in compressed stream }
      else
        NotYet;
      end;
      Inc(N);
    end;
  finally
    Deconstruct(XRefStream);
  end;
end;

procedure TPDFReader.SaveImageToFile(Sender: TObject; AStream: TStream; const Ext: string);
var
  Name: string;
  XObject: TPDFDictionary;
  Number: Integer;
begin
  XObject := Sender as TPDFDictionary;
  Name := ExtractFileName(FileName) + XObject.FindStr('Name');
  Number := 1;
  if FileExists(Name+Ext) then begin
    while FileExists(Name+IntToStr(Number)+Ext) do
      Inc(Number);
    Name := Name+IntToStr(Number)+Ext;
  end else
    Name := Name+Ext;
  (AStream as TMemoryStream).SaveToFile(Name);
end;

function TPDFReader.ReadObj(TokType: TTokType):TPDFObject;
begin
  case TokType of
    pdfNumber: begin
                 Result.oType := oNumber;
                 if not TryStrToInt(Token, Result.Value) then begin
                   Result.oType := oReal;
                   Result.Rvalue := StrToFloat(Token);
                 end
               end;
    pdfMarker: if Token = '[' then begin
                 Result.oType := oArray;
                 Result.Dictionary := ReadArray;
               end else if Token = '<' then begin
                 Result.oType := oDictionary;
                 Result.Dictionary := ReadDictionary;
               end else if Token = 'true' then begin
                 Result.oType := oNumber;
                 Result.Value := 1;
               end else if Token = 'false' then begin
                 Result.oType := oNumber;
                 Result.Value := 0;
               end else if Token = 'null' then
                 Result := nilObject
               else
                 raise Exception.Create('PDF: Unknown marker:'+Token);
    pdfName, pdfString:
               begin
                 Result.oType := oString;
                 Result.Dictionary := TPDFDictionary.Create(Token);
               end
  else
    raise Exception.Create('PDF: Unknown object:'+Token);
  end;
end;

function TPDFReader.ReadLabelledObj(Number:Integer):TPDFObject;
var
  TokType: TTokType;
  Gen: Integer;
begin
  Gen := Round(NextNumber); { Discard generation }
  if (NextToken <> pdfMarker) or (Token <> 'obj') then
    raise Exception.Create('PDF: Expected obj found '+Token);
  Result := ReadObj(NextToken);
  TokType := NextToken;
  if (TokType = pdfMarker) and (Token = 'stream') then begin
    if Result.oType <> oDictionary then
      raise Exception.Create('PDF:Unexpected stream');
    Result.oType := oStream;
    with Result do begin
      oType := oStream;
      Dictionary.Number := Number;
      Dictionary.Gen := Gen;
      Dictionary.Position := Stream.Position;
    end;
    Exit;
  end;
  if (TokType <> pdfMarker) or (Token <> 'endobj') then
    raise Exception.Create('PDF: Expected endobj found '+Token);
end;

procedure TPDFReader.ReadTrailer;
const
  BufSize:Integer = 50;
var
  buffer: string[50];
  p: Integer;
  NewTrailer: TPDFDictionary;
  TokType: TTokType;
  Obj: TPDFObject;
begin
  Stream.Seek(-BufSize, soFromEnd);
  SetLength(Buffer,Stream.Read(Buffer[1],BufSize));
  p := Pos('startxref',Buffer);
  if p <= 0 then
    raise Exception.Create('PDF: Can''t read startxref');
  Stream.Seek(-BufSize+p-1, soFromEnd);
  NextCh;
  if (NextToken <> pdfMarker) or (Token <> 'startxref') then
    raise Exception.Create('PDF: Can''t find startxref');
  p := Round(NextNumber);
  repeat
    Stream.Seek(p,soFromBeginning);
    NextCh;
    TokType := NextToken;
    if TokType = pdfNumber then begin
      { Xref table in object }
      Obj := ReadLabelledObj(StrToInt(Token));
      if Obj.oType <> oStream then
        NotYet;
      NewTrailer := Obj.Dictionary;
      try
        ReadXrefStream(Obj);
      except
        on Exception do begin
          NewTrailer.Free;
          raise;
        end;
      end;
    end else begin
      if (TokType <> pdfMarker) or (Token <> 'xref') then
        raise Exception.Create('PDF: Can''t find xref');
      ReadXref;
      if Token <> 'trailer' then
        raise Exception.Create('PDF: Can''t find trailer');
      if (NextToken <> pdfMarker) or (Token <> '<') then
        raise Exception.Create('PDF: Expected dictionary in trailer found '+Token);
      NewTrailer := ReadDictionary;
    end;
    p := NewTrailer.FindNumber('Prev');
    if not Assigned(Trailer) then
      Trailer := NewTrailer
    else
      NewTrailer.Free;
  until p = Low(Integer);
end;

procedure TPDFReader.LoadIndirect(var Obj: TPDFObject);
var
  StreamObj: TPDFObject;
  SavedStream: TStream;
  ObjStream: TStream;
  i, Off, nObjects: Integer;
  UnhandledFilters: string;
begin
  if Obj.oType = oIndirect then begin
    if XRef[Obj.Value] < 0 then begin
      { in object stream }
      StreamObj.oType := oIndirect;
      StreamObj.Value := -XRef[Obj.Value];
      SavedStream := Stream;
      ObjStream := nil;
      try
        LoadIndirect(StreamObj);
        nObjects := StreamObj.Dictionary.FindNumber('N',0);
        if StreamObj.oType <> oStream then
             NotYet;
        ObjStream := ConstructStream(StreamObj,UnhandledFilters);
        Stream := ObjStream;
        NextCh;
        for i := 0 to nObjects-1 do begin
          if i = XRefIndex[Obj.Value] then begin
            { Idx := } NextNumber;
            Off := Round(NextNumber);
          end else begin
            NextNumber; NextNumber;
          end;
        end;
        for i := 0 to Off do
          NextCh; { Skip to object start }
        Obj := ReadObj(NextToken);
      finally
        Deconstruct(ObjStream);
        Stream := SavedStream;
        FreePDFObject(StreamObj);
      end;
    end else begin
      Stream.Position := XRef[Obj.Value];
      NextCh;
      if Round(NextNumber) <> Obj.Value then
        raise Exception.Create('PDF: unexpected object:'+Token);
      Obj := ReadLabelledObj(Obj.Value);
    end;
  end;
end;

procedure TPDFReader.ReadEncryption;
var
  V: TPDFObject;
  buffer: AnsiString;
const
  PasswordPad: AnsiString = #$28#$bf#$4e#$5e#$4e#$75#$8a#$41#$64#$00#$4e#$56#$ff#$fa#$01#$08#$2e#$2e#$00#$b6#$d0#$68#$3e#$80#$2f#$0c#$a9#$fe#$64#$53#$69#$7a;
begin
  Encrypt := Trailer.FindDictionary('Encrypt');
  if not Assigned(Encrypt) then
    Exit; { Not Encrypted }
  if Encrypt.FindStr('Filter') <> 'Standard' then
    raise Exception.Create('PDF:Non standard security');
  V := Trailer.Find('ID');
  if (V.oType = oArray) and (V.Dictionary.Values[0].oType = oString) then
    FileID := V.Dictionary.Values[0].Dictionary.Keys[0];
  OwnerKey := Encrypt.FindStr('O');
  UserKey  := Encrypt.FindStr('U');
  PermFlags := NtoLE(Encrypt.FindNumber('P'));
  EncRevision := Encrypt.FindNumber('R');
  EncryptMetadata := Encrypt.FindNumber('EncryptMetadata') <> 0;
  KeyLength := Encrypt.FindNumber('Length', 40) div 8;
  if KeyLength < 5 then
    KeyLength := 5;
  buffer := PasswordPad+OwnerKey+'    '+FileId;
  Move(PermFlags, buffer[65],4);
  if not EncryptMetadata then
    buffer := buffer + #$ff#$ff#$ff#$ff;
  FileKey := MD5String(buffer);
  if EncRevision = 3 then
    NotYet;
  //for i := 1 to Length(Buffer) do begin
  //  Write(IntToHex(Ord(Buffer[i]),2):3);
  //  if i mod 8 = 0 then
  //    WriteLn;
  //end;
  //WriteLn;
end;

procedure TPDFReader.ProcessPage(n: Integer);
var
  Contents: TPDFObject;
  SourceStream: TStream;
  Buffer: array of Byte;
  i: Integer;
  UnhandledFilters: string;
begin
  Contents := Pages[n].Find('Contents');
  if Contents.oType = oStream then begin
    SourceStream := ConstructStream(Contents,UnhandledFilters);
    with TMemoryStream.Create do begin
      SetLength(Buffer,1000);
      repeat
        i := SourceStream.Read(Buffer[0], 1000);
        Write(Buffer[0],i);
      until i = 0;
      SaveToFile('/tmp/page'+IntToStr(n));
      Free;
    end;
  end else begin
    ShowPDFObject(Contents);
    NotYet;
  end;
end;

procedure TPDFReader.FindImages(n: Integer; OnImageStream: TOnImageStream);
var
  Resources, XObject: TPDFDictionary;
  i, j: Integer;
begin
  Resources := Pages[n].FindDictionary('Resources');
  if not Assigned(Resources) then
    Exit;
  for i := 0 to High(Resources.Keys) do
    if Resources.Keys[i] = 'XObject' then begin
      if Resources.Values[i].oType = oIndirect then
        LoadIndirect(Resources.Values[i]);
      if Resources.Values[i].oType = oDictionary then begin
        XObject := Resources.Values[i].Dictionary;
        for j := 0 to High(XObject.Keys) do begin
          // LoadIndirect(XObject.Values[j]);
          // ShowPDFObject(XObject.Values[j]);
          ProcessImage(XObject.Values[j], OnImageStream)
        end;
      end;
    end;
end;

procedure TPDFReader.ForAllImages(OnImageStream: TOnImageStream);
var
  i: Integer;
begin
  for i := 0 to nPages-1 do
    FindImages(i,OnImageStream);
end;

procedure TPDFReader.ProcessImage(var Obj: TPDFObject; OnImageStream: TOnImageStream);
var
  XObject: TPDFDictionary;
  SourceStream: TStream;
  Dest: TStream;
  buffer: AnsiString;
  Width, Height, c, i, nSet, StreamLength: Integer;
  Invert, First: Boolean;
  BitsSet: array [0..255] of Byte;
  UnhandledFilters: string;
  function CountBits(B: Byte):Integer;
  begin
    Result := 0;
    while B <> 0 do begin
      if Odd(B) then
        Inc(Result);
      B := B shr 1;
    end;
  end;
begin
  LoadIndirect(Obj);
  if Obj.oType <> oStream then
    NotYet;
  XObject := Obj.Dictionary;
  Invert := XObject.FindNumber('ImageMask',0) = 1;
  //if XObject.FindStr('ColorSpace') = 'DeviceGray' then
  //  Invert := not Invert;
  Width := XObject.FindNumber('Width',0);
  Height := XObject.FindNumber('Height',0);
  if (Width < 60) or (Height < 60) then
    Exit; { probably too small }
  { Bits := } XObject.FindNumber('BitsPerComponent',1);
  StreamLength := XObject.FindNumber('Length',0);
  SourceStream := ConstructStream(Obj, UnhandledFilters, True);
  Dest := nil;
  try
    if UnhandledFilters = '' then begin
      Dest := TMemoryStream.Create;
      buffer := Format('P4'#10'%d %d'#10,[Width, Height]);
      Dest.Write(buffer[1],Length(Buffer));
      SetLength(Buffer,100000);
      First := True;
      repeat
        c := SourceStream.Read(buffer[1], Length(Buffer));
        if First then begin
          { Invert image if most pixels black }
          for i := 0 to 255 do
            BitsSet[i] := CountBits(i);
          nSet := 0;
          for i := 1 to c do
            Inc(nSet,BitsSet[Byte(buffer[i])]);
          if (nSet > c*4) xor Invert then
            Invert := not Invert;
          First := False;
        end;
        if Invert then
          for i := 1 to c do
            buffer[i] := AnsiChar(not Ord(buffer[i]));
        Dest.Write(Buffer[1],c);
      until c = 0;
      Dest.Position := 0;
      OnImageStream(XObject, Dest, '.pbm');
    end else begin
      if UnhandledFilters = 'DCTDecode' then begin
        Dest := TMemoryStream.Create;
        Dest.CopyFrom(SourceStream,StreamLength);
        Dest.Position := 0;
        OnImageStream(XObject, Dest, '.jpg');
      end else begin
        {$IFDEF CONSOLE}
        WriteLn('Unknown filter(s):',UnhandledFilters);
        XObject.Show('');
        WriteLn(XObject.FindStr('Name'));
        //XObject.FindNumber('Length',0);
        //for i := 1 to Len do
        //  Dest.WriteByte(Stream.ReadByte);
        {$ENDIF}
      end;
    end;
  finally
    Deconstruct(SourceStream);
    Dest.Free;
  end;
end;

function TPDFReader.ConstructStream(var Obj: TPDFObject; out UnhandledFilters: string; NilOK: Boolean): TStream;
var
  buffer: array of byte;
  ObjKey: TMD5Digest;
  ObjKeyLength: Integer;
  i: Integer;
  Filter, DecodeParams: TPDFObject;
  procedure AddFilter(const Name: string; Params: TPDFDictionary);
  begin
    if Params.FindNumber('Predictor',0) <> 0 then
      if NilOk then
        UnhandledFilters:= UnhandledFilters+'Predictor with '+name
      else
        raise Exception.Create('PDF:Predictor not implemented')
    else if Name = 'LZWDecode' then
      Result := TLZWDecompressionStream.Create(Result,Params)
    else if Name = 'CCITTFaxDecode' then
      Result := TCCITTFaxDecompressionStream.Create(Result,Params)
    else if Name = 'FlateDecode' then
      Result := TFlateDecompressionStream.Create(Result,Params)
    else if Name = 'ASCII85Decode' then
      Result := TASCII85DecompressionStream.Create(Result,Params)
    else if NilOk then
      UnhandledFilters:= UnhandledFilters+name
    else
      raise Exception.Create('PDF:Unknown filter:'+Name);
  end;
begin
  UnhandledFilters := '';
  LoadIndirect(Obj);
  if Obj.oType <> oStream then
    raise Exception.Create('PDF:Stream Expected');
  Result := Stream;
  Result.Position := Obj.Dictionary.Position;
  try
    if Assigned(Encrypt) then begin
      SetLength(buffer, KeyLength+5);
      Move(FileKey[0],buffer[0],KeyLength);
      i := Obj.Dictionary.Number;
      NtoLE(i);
      Move(i,buffer[KeyLength],3);
      i := Obj.Dictionary.Gen;
      NtoLE(i);
      Move(i,buffer[KeyLength+3],2);
      //for i := 0 to Length(Buffer)-1 do begin
        //Write(IntToHex(Ord(Buffer[i]),2):3);
        //if i mod 8 = 0 then
        //  WriteLn;
      //end;
      ObjKey := MD5Buffer(buffer[0],KeyLength+5);
      ObjKeyLength := Min(KeyLength+5,16);
      Result := TRC4DecryptionStream.Create(Result, ObjKey, ObjKeyLength);
    end;
    Filter := Obj.Dictionary.Find('Filter');
    if Filter.oType = oString then
      AddFilter(Filter.Dictionary.Keys[0], Obj.Dictionary.FindDictionary('DecodeParms'))
    else if Filter.oType = oArray then begin
      if not Assigned(Filter.Dictionary) then
        Exit; { No filter }
      DecodeParams := Obj.Dictionary.Find('DecodeParms');
      if DecodeParams.oType <> oArray then
        NotYet;
      for i := 0 to High(Filter.Dictionary.Values) do begin
        if Filter.Dictionary.Values[i].oType <> oString then
          NotYet;
        if Assigned(DecodeParams.Dictionary) then begin
          if not IsDictionaryOrNull(DecodeParams.Dictionary.Values[i]) then
            NotYet;
          AddFilter(Filter.Dictionary.Values[i].Dictionary.Keys[0],DecodeParams.Dictionary.Values[i].Dictionary);
        end else
          AddFilter(Filter.Dictionary.Values[i].Dictionary.Keys[0],nil);
        if not Assigned(Result) then
          Break;
      end;
    end else
      NotYet;
  except
    on Exception do begin
      Deconstruct(Result);
      raise;
    end;
  end;
end;

function TPDFReader.IsDictionaryOrNull(Obj: TPDFObject): Boolean;
begin
  case Obj.oType of
    oDictionary: Result := True;
    oArray:      Result := not Assigned(Obj.Dictionary);
  else
    Result := False;
  end;
end;

constructor TPDFReader.Create(AFileName: string);
  procedure ReadPageTree(Node: TPDFDictionary);
  var
    i: Integer;
    Kids: TPDFDictionary;
  begin
    if Node.FindStr('Type') = 'Page' then begin
      Pages[nPages] := Node;
      Inc(nPages);
    end else begin
      Kids := Node.FindArray('Kids');
      for i := 0 to High(Kids.Values) do begin
        LoadIndirect(Kids.Values[i]);
        if Kids.Values[i].oType = oDictionary then
          ReadPageTree(Kids.Values[i].Dictionary)
        else
          raise Exception.Create('PDF:Expected Dictionary');
      end;
    end;
  end;
begin
  FileName := AFileName;
  Stream := TMemoryStream.Create;
  UnlockedLoadFromFile(TMemoryStream(Stream), FileName);
  ReadTrailer;
  ReadEncryption;
  Catalog := Trailer.FindDictionary('Root');
  if not Assigned(Catalog) then
    raise Exception.Create('PDF: Can''t find catalog');
  PageTree := Catalog.FindDictionary('Pages');
  if not Assigned(PageTree) then
    raise Exception.Create('PDF: Can''t find page tree');
  SetLength(Pages,PageTree.FindNumber('Count'));
  nPages := 0;
  ReadPageTree(PageTree);
  if nPages <> PageTree.FindNumber('Count') then
    raise Exception.CreateFmt('PDF: Bad page count %d %d',[nPages,PageTree.FindNumber('Count')]);
end;

destructor TPDFReader.Destroy;
begin
  Trailer.Free;
  Stream.Free;
  inherited Destroy;
end;

procedure TPDFReader.Deconstruct(AStream: TStream);
begin
  if AStream is TPDFStream then
    AStream.Free;
end;

function TPDFReader.NextCh: AnsiChar;
begin
  Eof := Stream.Read(Ch,1) <> 1;
  if Eof then
    Ch := #10;
  Result := Ch;
end;

function TPDFReader.HexVal: Integer;
begin
  case Ch of
    '0'..'9': Result := Ord(Ch)-Ord('0');
    'a'..'f': Result := Ord(Ch)-Ord('a')+10;
    'A'..'F': Result := Ord(Ch)-Ord('A')+10;
  else
    raise Exception.Create('Bad hex char in pdf:'+Ch);
  end;
end;

function TPDFReader.NextToken: TTokType;
var
  Depth, Val, I: Integer;
  CharAvailable: Boolean;
begin
  Token := '';
  Result := pdfEOF;
  while not Eof and (Ch in [#0, #9, #10, #12, #13, ' ']) do
    NextCh;
  if Eof then
    {}
  else if Ch = '%' then begin
    { Comment }
    while not (Ch in [#10, #13]) do
      NextCh;
    Result := NextToken;
  end else if Ch in ['0'..'9', '.', '-'] then begin
    { Number }
    Token := Ch;
    while NextCh in ['0'..'9', '.', '-', 'E'] do
      Token := Token + Ch;
    Result := pdfNumber;
  end else if Ch = '(' then begin
    { Literal string }
    Depth := 1;
    while not Eof do begin
      NextCh;
      CharAvailable := True;
      while CharAvailable do begin
        CharAvailable := False;
        case Ch of
          '(': Inc(Depth);
          ')': begin Dec(Depth); if Depth < 1 then Break; end;
          #13: if NextCh <> #10 then begin
                 Token := Token + #10;
                 CharAvailable := True;
               end else
                 Ch := #10;
          '\': begin
                 NextCh;
                 case Ch of
                   'n': Ch := #10;
                   'r': Ch := #13;
                   't': Ch := #9;
                   'b': Ch := #8;
                   'f': Ch := #12;
                   #10: NextCh;
                   #13: begin
                          NextCh;
                          if Ch = #10 then
                            NextCh;
                        end;
                   '0'..'7':
                        begin
                          Val := 0;
                          for I := 0 to 2 do begin
                            Val := Val * 8 + Ord(Ch)-Ord('0');
                            if not (NextCh in ['0'..'7']) then
                              Break;
                          end;
                          Token := Token + Char(Val);
                          CharAvailable := True;
                        end;
                 end;
               end;
        end;
      end;
      if Depth < 1 then
        Break;
      Token := Token + Ch;
    end;
    Result := pdfString;
    NextCh;
  end else if Ch = '<' then begin
    NextCh;
    if Ch = '<' then begin
      { Dictionary }
      Token := Ch;
      Result := pdfMarker;
      NextCh;
    end else begin
      { Hexadecimal string }
      while Ch <= ' ' do
        NextCh;
      repeat
        Val := HexVal*16;
        while NextCh <= ' ' do ;
        if Ch <> '>' then begin
          Val := Val + HexVal;
          while NextCh <= ' ' do ;
        end;
        Token := Token + Char(Val);
      until Ch = '>';
      NextCh;
      Result := pdfString;
    end;
  end else if Ch = '>' then begin
    Token := NextCh;
    if Ch <> '>' then
      raise Exception.Create('PDF:Unexpected end of dictionary:>'+Ch);
    Result := pdfMarker;
    NextCh;
  end else if Ch = '/' then begin
    { Name }
    while NextCh > ' ' do begin
      if Ch in ['[','/','>','<',']','('] then
        Break;
      if Ch = '#' then begin
        Val := HexVal*16;
        NextCh;
        Ch := AnsiChar(HexVal+Val);
      end;
      Token := Token + Ch;
    end;
    Result := pdfName;
  end else if Ch = '[' then begin
    { Array start }
    Token := Ch;
    Result := pdfMarker;
    NextCh;
  end else if Ch = ']' then begin
    { Array end }
    Token := Ch;
    Result := pdfMarker;
    NextCh;
  end else if Ch in ['a'..'z', 'A'..'Z'] then begin
    Token := Ch;
    while NextCh in ['a'..'z'] do
      Token := Token + Ch;
    if Ch = #13 then
      NextCh;
    Result := pdfMarker;
  end else
    NotYet;
  LastToken := Result;
  WriteLn(Log, Ord(Result):2, ' ', Token);
end;

{ TPDFDictionary }

constructor TPDFDictionary.Create(AReader: TPDFReader);
begin
  Reader := AReader;
end;

constructor TPDFDictionary.Create(S: string);
begin
  inherited Create;
  SetLength(Keys,1);
  Keys[0] := S;
end;

destructor TPDFDictionary.Destroy;
var
  i: Integer;
begin
  for i := 0 to High(Values) do
    FreePDFObject(Values[i]);
  inherited;
end;

procedure TPDFDictionary.Show(const Indent: string);
var
  i: Integer;
begin
  if Self = nil then begin
    WriteLn(Indent,'null');
    Exit;
  end;
  for i := 0 to Max(High(Keys), High(Values)) do begin
    if i < Length(Keys) then
      Write(Indent,Keys[i]:15)
    else
      Write(Indent,' ':15);
    if i < Length(Values) then
      ShowPDFObject(Values[i],Indent+'  ')
    else
      WriteLn;
  end;
end;

function TPDFDictionary.Find(const Name: string): TPDFObject;
var
  i: Integer;
begin
  for i := 0 to High(Keys) do
    if Keys[i] = Name then begin
      Result := Values[i];
      if Result.oType = oIndirect then
        if not Assigned(Reader) then
          raise Exception.Create('PDF:Reader not set')
        else begin
          Reader.LoadIndirect(Values[i]);
          Result := Values[i];
        end;
      Exit;
    end;
  Result := nilObject;
end;

function TPDFDictionary.FindStr(const Name: string): string;
var
  V: TPDFObject;
begin
  Result := '';
  V := Find(Name);
  if V.oType = oString then
    Result := V.Dictionary.Keys[0];
end;

function TPDFDictionary.FindDictionary(const Name: string): TPDFDictionary;
var
  V: TPDFObject;
begin
  Result := nil;
  V := Find(Name);
  if V.oType = oDictionary then
    Result := V.Dictionary;
end;

function TPDFDictionary.FindNumber(const Name: string; Deflt: Integer = Low(Integer)): Integer;
var
  V: TPDFObject;
begin
  Result := Deflt;
  if not Assigned(Self) then
    Exit;
  V := Find(Name);
  if V.oType = oNumber then
    Result := V.Value;
end;

function TPDFDictionary.FindArray(const Name: string): TPDFDictionary;
var
  V: TPDFObject;
begin
  Result := nil;
  V := Find(Name);
  if V.oType = oArray then
    Result := V.Dictionary;
end;

end.

procedure LoadImagesFromPDF(FileName: string {$IFNDEF CONSOLE}; Overlay: TOverlays{$ENDIF});
var
  i: Integer;
begin
  with TPDFReader.Create(FileName) do
    try
      for i := 0 to nPages-1 do begin
        // ProcessPage(i);
        FindImages(i,ExtractFileName(FileName),nImages);
      end;
    finally
      Free;
    end;
end;


