unit PeakFinders;
{$I utils.inc}{ Copyright (C) 2015 C M Western; see UCompat.pas for license}
interface

uses MathSup;

type
  TPeakInfo = record
    Position, nAbove: Integer;
    Centre: Double;
    Width: Double;
    Area: Double;
    Baseline: Double;
  end;

function FindCentreOfMass(out PeakInfo: TPeakInfo;
                          const y: array of Double;
                          const x: array of Double; { May be empty, in which case point numebrs are used }
                          const b: array of Double; { Baseline - May be empty, in which case y used as is }
                          Threshold: Double = 0.5;
                          First: Integer = 0; Last: Integer = -1; Old: Boolean = False):Boolean;

function MakeFloatingBaseline(out NoiseEst: Double; const y: array of Double; Window: Integer; NoiseMult: Double; Separation: Integer = 1; Dense: Boolean = False;  Inverted: Boolean = False; Verbose: Boolean = False; First: Integer = 0; Last: Integer = -1):TDoubleArray;
function EstimateNoise(const y: array of Double; out Baseline:Double; Separation: Integer = 1; Passes: Integer = 2; NoiseMult: Double = 5; First: Integer = 0; Last: Integer = -1):Double;
function ClosestPoint(f: Double; const x: array of Double; Sorted: Boolean = False; First: Integer = 0; Last: Integer = -1): Integer;
function Convolute(const y: array of Double; const c: array of Double; FirstPoint: Double = 0; LastPoint: Double = 0; First: Integer = 0; Last: Integer = -1):TDoubleArray;

implementation

uses SysUtils, Math, UCompat, Worker, SavitzkyGolay, Progress, Profile, Lapack;

var
  SmoothTime: string = 'Baseline Smoothing';

function Sum(const x: array of Double; First: Integer = 0; Last: Integer = -1):Double;
var
  i: Integer;
begin
  Result := 0;
  if Last < 0 then
    Last := High(x);
  for i := First to Last do
    Result := Result + x[i];
end;

function ClosestPoint(f: Double; const x: array of Double; Sorted: Boolean = False; First: Integer = 0; Last: Integer = -1): Integer;
var
  i, l, h: Integer;
  Descending: Boolean;
begin
  { Find closest point to peak }
  if Last < 0 then
    Last := High(x);
  if Sorted then begin
    Descending := x[Last] < x[First];
    while Last-First > 1 do begin
      Result := (Last + First) div 2;
      if x[Result] = f then
        Exit
      else if (x[Result] > f) xor Descending then
        Last := Result
      else
        First := Result;
    end;
  end;
  Result := First;
  for i := First+1 to Last do
    if Abs(x[i]-f) < Abs(x[Result]-f) then
      Result := i;
end;

function FindCentreOfMass(out PeakInfo: TPeakInfo;
                          const y: array of Double;
                          const x: array of Double; { May be empty, in which case point numebrs are used }
                          const b: array of Double; { Baseline - May be empty, in which case y used as is }
                          Threshold: Double = 0.5;
                          First: Integer = 0; Last: Integer = -1; Old: Boolean = False):Boolean;
var
  yMin, yMax: Double;
  Thresh, SumI, SumIX, f, v: Double;
  i, nB, LeftEdge, RightEdge: Integer;
  function df(i: Integer):Double;
  begin
    if Length(x) = 0 then
      Result := 1
    else if i = First then
      Result := Abs(x[1] - x[0])
    else begin
      Result := Abs(x[i] - x[i-1]);
      if i < Last then
        Result := (Result + Abs(x[i+1]-x[i]))/2;
    end;
  end;
  function yo(i: Integer): Double;
  begin
    Result := y[i] - Thresh;
    if Length(b) > 0 then
      Result := Result - b[i];
  end;
begin
  Result := False;
  FillChar(PeakInfo, SizeOf(PeakInfo), 0);
  PeakInfo.Position := First;
  if Last < 0 then
    Last := High(y);
  if Last <= First then
    Exit; { No points }
  if Length(b) > 0 then begin
    yMin := y[First] - b[First];
    yMax := yMin;
    for i := First+1 to Last do begin
      v := y[i] - b[i];
      yMin := Min(yMin, v);
      yMax := Max(yMax, v);
    end;
  end else
    Limits(y, yMin, yMax, First, Last);
  PeakInfo.Baseline := yMin;
  if yMin = yMax then begin
    { Flat function - return simple averages, but area zero }
    if Length(x) = 0  then
      PeakInfo.Centre := (First+Last)/2
    else
      PeakInfo.Centre := Sum(x, First, Last)/(Last-First+1);
    PeakInfo.Width := Abs(x[Last]-x[First]);
    Exit;
  end;
  Thresh := yMin + Threshold*(yMax-yMin);
  SumI := 0;
  SumIX := 0;
  for i := First to Last do begin
    if Length(x) = 0 then
      f := i
    else
      f := x[i];
    v := yo(i);
    if v > 0 then begin
      SumI := SumI + v;
      SumIX := SumIX + v * f;
      Inc(PeakInfo.nAbove);
    end;
  end;
  if SumI <= 0 then
    Exit; { Probably never }
  PeakInfo.Centre := SumIX / SumI;
  if Length(x) = 0 then
    PeakInfo.Position := Min(Max(First,Round(PeakInfo.Centre)),Last)
  else
    PeakInfo.Position := ClosestPoint(PeakInfo.Centre, x, False, First, Last);
  { Estimate area }
  { Contribution from central point }
  SumI := Max(yo(PeakInfo.Position)*df(PeakInfo.Position), 0);
  { Scan upwards }
  i := PeakInfo.Position + 1;
  while (i <= Last) and (yo(i) > 0) do begin
    SumI := SumI + yo(i)*df(i);
    Inc(i);
  end;
  RightEdge := i-1;
  i := PeakInfo.Position - 1;
  while (i > 0) and (yo(i) > 0) do begin
    SumI := SumI + yo(i)*df(i);
    Dec(i);
  end;
  LeftEdge := i+1;
  if SumI = 0 then begin
    { Noisy or dual peaks - take all points above threshold }
    { May be more appropriate to measure as two peaks, but this routine assumes one peak in range }
    LeftEdge := PeakInfo.Position;
    for i := First to PeakInfo.Position-1 do
      if yo(i) > 0 then begin
        if i < LeftEdge then
          LeftEdge := i;
        SumI := SumI + yo(i)*df(i);
      end;
    RightEdge := PeakInfo.Position;
    for i := RightEdge+1 to Last do
      if yo(i) > 0 then begin
        if i > RightEdge then
          RightEdge := i;
        SumI := SumI + yo(i)*df(i);
      end;
  end;
  if RightEdge = LeftEdge then
    PeakInfo.Width := df(PeakInfo.Position) { Single point }
  else if High(x) = 0 then
    PeakInfo.Width := RightEdge-LeftEdge
  else
    PeakInfo.Width := Abs(x[RightEdge]-x[LeftEdge]);
  PeakInfo.Area := SumI/Sqr(Threshold); { Correction by Threshold^2 is exact for a triangle }
  Result := True;
  if Length(b) > 0 then
    Exit;
  { Estimate baseline by averaging over points outside peak }
  { Move more outside peak if possible }
  i := (RightEdge-LeftEdge+1) div 2;
  if i <= 1 then
    i := 2;
  if LeftEdge > First + 2*i then
    Dec(LeftEdge, i);
  if RightEdge < Last - 2*i then
    Inc(RightEdge, i);
  SumI := Sum(y,First,LeftEdge-1) + Sum(y,RightEdge+1,Last);
  nB := LeftEdge-First + Last-RightEdge;
  if nB > 0 then
    PeakInfo.Baseline := SumI/nB;
end;

function MakeFloatingBaseline(out NoiseEst: Double; const y: array of Double; Window: Integer;
                              NoiseMult: Double; Separation: Integer; Dense: Boolean;
                              Inverted: Boolean; Verbose: Boolean;
                              First: Integer; Last: Integer):TDoubleArray;
var
  i, j, nLeft, nRight, Pass, nNoise, nNoPoints, PassWindow, CurrentWindow, dWindow, SubPasses: Integer;
  SumLeft, SumRight, SumDelta, InitialNoise: Double;
  FirstPass, Smooth: TDoubleArray;
  function OK(Index: Integer):Boolean;
  begin
    Result := (Index >= First) and (Index <= Last);
    if Result and (Pass >= 2) then begin
      if Dense then
        if Inverted then
          Result := FirstPass[Index]-y[Index] < NoiseEst
        else
          Result := y[Index]-FirstPass[Index] < NoiseEst
      else
        Result := Abs(y[Index]-FirstPass[Index]) < NoiseEst;
      if Result and (Index > First) then
        Result := y[Index] - y[Index-1] < NoiseEst;
      if Result and (Index < Last) then
        Result := y[Index] - y[Index+1] < NoiseEst;
    end;
  end;
begin
  if Last < 0 then
    Last := High(y);
  PushProgress('Baseline',4);
  NoiseEst := 0;
  PassWindow := Min(Window*16,Last-First+1);
  SetLength(Result, Length(y));
  FirstPass := nil;
  for Pass := 1 to 3 do begin
    InitialNoise := NoiseEst;
    SubPasses := 0;
    repeat
      Inc(SubPasses);
      SumLeft := 0;
      nLeft := 0;
      SumRight := 0;
      nRight := 0;
      nNoise := 0;
      SumDelta := 0;
      CurrentWindow := 0;
      Result[First] := y[First];
      Result[Last] := y[Last];
      nNoPoints := 0;
      for i := First+1 to Last-1 do begin
        dWindow := Min(Min(i-First, Last-i), PassWindow)-CurrentWindow;
        if dWindow > 0 then begin
          { Increase window }
          if OK(i+CurrentWindow) then begin
            Inc(nRight);
            SumRight := SumRight + y[i+CurrentWindow];
          end;
          Inc(CurrentWindow);
        end else if dWindow < 0 then begin
          if OK(i-CurrentWindow-1) then begin
            SumLeft := SumLeft - y[i-CurrentWindow-1];
            Dec(nLeft);
          end;
          Dec(CurrentWindow);
        end;
        if (dWindow <= 0) and OK(i-CurrentWindow-1) then begin
          SumLeft := SumLeft - y[i-CurrentWindow-1];
          Dec(nLeft);
        end;
        if (dWindow >= 0) and OK(i+CurrentWindow) then begin
          Inc(nRight);
          SumRight := SumRight + y[i+CurrentWindow];
        end;
        if OK(i-1) then begin
          Inc(nLeft);
          SumLeft := SumLeft + y[i-1];
        end;
        if OK(i) then begin
          { Note that this excludes the current point from the average }
          { Including it in SumRight would give a bias }
          if OK(i-Separation) then begin
            SumDelta := SumDelta + Abs(y[i]-y[i-Separation]);
            Inc(nNoise);
          end;
          Dec(nRight);
          SumRight := SumRight - y[i];
        end;
        if nLeft = 0 then begin
          if nRight = 0 then begin
            Result[i] := FirstPass[i];
            Inc(nNoPoints);
          end else
            Result[i] := SumRight/nRight
        end else if nRight = 0 then
          Result[i] := SumLeft/nLeft
        else
          Result[i] := (SumLeft/nLeft+SumRight/nRight)/2;
      end;
      if Verbose then
        WriteLn(Log, 'Pass ', Pass, '.', SubPasses, ', ', nNoPoints, ' points skipped, OldNoise = ', DtosM(NoiseEst,10,-1), ', NewNoise = ', DtosM(SumDelta/nNoise*NoiseMult,10,-1), ', Window = ', PassWindow);
      NoiseEst := NoiseEst * 2;
    until (Pass = 1) or (nNoPoints = 0) or (SubPasses > 5);
    if (nNoise > 0) and (Dense or (Pass = 1)) then
      NoiseEst := SumDelta/nNoise*NoiseMult
    else
      NoiseEst := InitialNoise;
    SetProgress(Pass);
    nLeft := PassWindow div 2;
    Smooth := SmoothingCoefficients(nLeft, nLeft);
    TStart(SmoothTime);
    FirstPass := Result;
    Result := Convolute(FirstPass, Smooth, FirstPass[First], FirstPass[Last], First, Last);
    TEnd(SmoothTime);
    PassWindow := Max(PassWindow div 4, Window);
  end;
  { Final pass to refine noise estimate }
  nNoise := 0;
  SumDelta := 0;
  Pass := 2;
  for i := First+Separation to Last do
    if OK(i) and OK(i-Separation) then begin
      SumDelta := SumDelta + Abs(y[i]-y[i-Separation]);
      Inc(nNoise);
    end;
  if nNoise > 0 then
    NoiseEst := SumDelta/nNoise;
  PopProgress;
end;

function EstimateNoise(const y: array of Double; out Baseline:Double; Separation: Integer = 1; Passes: Integer = 2; NoiseMult: Double = 5; First: Integer = 0; Last: Integer = -1):Double;
{ Estimate noise from the average difference between points Separation apart;
  The baseline is estimated from the average value of the points.
  repeat Passes times, but on passes other than the first discard points NoiseMult*Noise away
  from the baseline.
  The baseline estimate is only unseful for flat baselines }
var
  i, n, nBase, Pass: Integer;
  Sum, SumDelta: Double;
begin
  if Last < 0 then
    Last := High(y);
  if Last-First < Separation then
    raise Exception.CreateFmt('Too few points(%d) in EstimateNoise: Separation=%d', [Last-First+1, Separation]);
  Sum := 0;
  SumDelta := 0;
  for i := First to Last do begin
    if i-Separation >= First then
      SumDelta := SumDelta + Abs(y[i]-y[i-Separation]);
    Sum := Sum + y[i];
  end;
  Result := SumDelta/(Last-First+1-Separation);
  Baseline := Sum/(Last-First+1);
  for Pass := 2 to Passes do begin
    n := 0;
    nBase := 0;
    Sum := 0;
    SumDelta := 0;
    for i := First to Last do
      if Abs(y[i]-Baseline) < Result*NoiseMult then begin
        if (i-Separation >= First) and (Abs(y[i-Separation]-Baseline) < Result*NoiseMult) then begin
          SumDelta := SumDelta + Abs(y[i]-y[i-Separation]);
          Inc(n)
        end;
        Inc(nBase);
        Sum := Sum + y[i];
      end;
    if n > 1 then begin
      Result := SumDelta/n;
      Baseline := Sum/nBase;
    end else
      Exit; { Take result from previous pass }
  end;
end;

function Convolute(const y: array of Double; const c: array of Double; FirstPoint: Double; LastPoint: Double; First: Integer; Last: Integer):TDoubleArray;
var
  i, j, n, lc, i1, Milestone, Interval: Integer;
  Sum: Double;
begin
  if Last < 0 then
    Last := High(y);
  if not Odd(Length(c)) then
    raise Exception.Create('Convolute: number of points must be odd');
  lc := Length(c);
  i1 := 1;
  n := Length(c) div 2;
  { Show progress every 10^7 calculations }
  Interval := 10000000 div lc;
  if (Last - First) div Interval > 3 then begin
    Milestone := First + Interval;
    PushProgress('Convoluting', Last-First);
  end else begin
    Interval := 0;
    Milestone := Last;
  end;
  SetLength(Result, Length(y));
  for i := First to Last do begin
    Sum := 0;
    if (i < First+n) or (i > Last-n) then
      for j := -n to n do
        if i+j < First then
          Sum := Sum + FirstPoint*c[j+n]
        else if i+j > Last then
          Sum := Sum + LastPoint*c[j+n]
        else
          Sum := Sum + y[i+j]*c[j+n]
    else
      //for j := -n to n do
      //  Sum := Sum + y[i+j]*c[j+n];
      Sum := ddot_(lc, PDouble(@y[i-n])^, i1, PDouble(@c[0])^, i1);
    Result[i] := Sum;
    if i > Milestone then begin
      SetProgress(i-First);
      Milestone := i + Interval;
    end;
  end;
  if Interval > 0 then
    PopProgress;
end;

end.
