unit Plotter;
{$I utils.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses Classes, MNode, Mixture, Global, GraphM, GraphRange, Overlay, SpectrumPlot, ExpCan;

procedure MakePlot(Canvas: TExportableCanvas);
function MakeUpperPlot(G: TGraphM; Simulation: TSimulation; out LastFortrat: TFortratPlot; var FortratPlotRanges: TPlotRanges): Boolean;
function MakeLowerPlot(G: TGraphM; Mixture: TMixture; Simulation: TSimulation; Overlays: TOverlays;
                       var SpectrumPlotRanges: TPlotRanges; var PlotColours: TColours):string;
procedure MakePlotFile(FileName: string; IsEPlot: Boolean);
procedure MakePlotStream(IsEPlot: Boolean; Stream: TStream; CClass: TStreamCanvasClass);

implementation

uses SysUtils, Math, Units, Obs, Fitter, MathSup, Worker, ExportG, Profile, LevelPlotter;

function XLabel(Simulation: TSimulation):string;
begin
  if Simulation.Foffset = 0 then
    XLabel := TypeOfUnit(Simulation.PlotUnits) + '/' + NameOfUnit(Simulation.PlotUnits)
  else
    XLabel := Format('(%s-%g)/%s', [TypeOfUnit(Simulation.PlotUnits), Simulation.Foffset, NameOfUnit(Simulation.PlotUnits)]);
end;

function MakeUpperPlot(G: TGraphM; Simulation: TSimulation; out LastFortrat: TFortratPlot; var FortratPlotRanges: TPlotRanges): Boolean;
var
  Problem: Exception;
  xl, xu: Double;
begin
  Result := True;
  G.Start;
  xl := Simulation.Fmin-Simulation.Foffset;
  xu := Simulation.Fmax-Simulation.Foffset;
  SetLength(FortratPlotRanges,0);
  LastFortrat := TFortratPlot.Create(Simulation.PlotUnits);
  with LastFortrat do
    try
      MakeFromSimulation(Simulation, not Simulation.Mixture.NoLineList, Problem);
      if Assigned(Problem) then
        G.Title1(Problem.Message);
      G.Map(xl, xu, IMin-1, IMax+1);
      G.Axis(xl, xu, IMin-1, true, XLabel(Simulation));
      Show(G, NullPlotRange, Simulation.GraphStyle in [gsOffsetAll, gsInvertedAll], False, FortratPlotRanges);
      Result := not Assigned(Problem);
    finally
      Problem.Free;
    end;
  G.Finish;
end;

var
  MakeLowerPlotTime: string = 'Make Lower Plot';
  OverlayFixupTime: string = 'Overlay Fixup';

function MakeLowerPlot(G: TGraphM; Mixture: TMixture; Simulation: TSimulation; Overlays: TOverlays;
                       var SpectrumPlotRanges: TPlotRanges; var PlotColours: TColours):string;
var
  xl, xu: Double;
  procedure ShowObservedLines;
  var
    i, nPlotted, nErrors: Integer;
    v, vo, vcalc: Double;
    s: string;
  begin
    nPlotted := 0;
    nErrors := 0;
    for i := 0 to High(Observations) do begin
      if Observations[i].ObsType in [obParameter,obExpression] then
        Continue;
      if IsPopDist(Mixture.Simulations[0].IntensityUnits) then begin
        if Observations[i].ObsType <> obELower then
          Continue;
      end else if Observations[i].ObsType <> obLine then
        Continue;
      v := Observations[i].Frequency;
      vcalc := Observations[i].Calculated;
      if nErrors < 20 then
        try
          if CalculateObs(Observations[i],vo,s,False,Observations[i].ObsType) then
            vcalc := vo
          else
            Inc(nErrors);
        except
          Inc(nErrors);
        end
      else if nErrors = 20 then begin
        WriteLn(Log, 'Supressing further output on replot - refit or View, ClearObserved to remove this recalculation');
        Inc(nErrors);
      end;
      v := Convert(v,Mixture.Units,Simulation.PlotUnits);
      vcalc := Convert(vcalc,Mixture.Units,Simulation.PlotUnits) - Mixture.Simulations[0].Foffset;
      if Within(v,xl,xu) or Within(vcalc,xl,xu) then begin
          G.StyleP(FindObservationColour(Observations[i],True), G.LineStyle.Width, G.LineStyle.Pattern);
          G.Point(v, 1.05);
          G.Join(vcalc, 1.02);
          Inc(nPlotted);
        end;
    end;
    if nPlotted > 0 then
      AddPlotRange(SpectrumPlotRanges, PlotRange(G, 1.02, 1.05, 1.02, 1.05, -2));
  end;
var
  yi, yx, lowerFrac, f: Double;
  AnySpectrum, UseOverlayAxis, AllArbitrary: Boolean;
  SpectrumPlot: TSpectrumPlot;
  Problem: Exception;
  IU: TIntensityUnits;
  PR: TPlotRange;
begin
  Result := '';
  TStart(MakeLowerPlotTime);
  SetLength(SpectrumPlotRanges,0);
  xl := Simulation.Fmin-Simulation.Foffset;
  xu := Simulation.Fmax-Simulation.Foffset;
  AnySpectrum := Simulation.ShowSum or Simulation.ShowParts;
  if Simulation.ShowParts then
    SpectrumPlot := TMultiColouredSpectrumPlot.Create(Simulation)
  else
    SpectrumPlot := TSpectrumPlot.Create(Simulation.PlotUnits,IsPopDist(Simulation.IntensityUnits));
  if Simulation.PlotSplit > 0 then
    lowerFrac := Max(Min(Simulation.PlotSplit,0.99),0.01)
  else
    lowerFrac := Min(Max(Simulation.PlotSplit,-0.99),-0.01);
  try
    if AnySpectrum or not Assigned(Overlays) then
      SpectrumPlot.MakeFromSimulation(Simulation, not Simulation.Mixture.NoLineList, Problem)
    else
      Problem := nil;
    SpectrumPlot.GetColoursUsed(PlotColours);
    G.Start;
    if Assigned(Problem) then begin
      G.Title1(Problem.Message);
      Problem.Free;
      AnySpectrum := True;
    end else
      AnySpectrum := SpectrumPlot.nLinesInRange > 0;
    G.Map(xl, xu, -0.05, 1.02); { Upper limit should strictly be 1.05 to allow for ticks; 1.02 allows some spill over into title area }
    if Assigned(Overlays) then begin
      TStart(OverlayFixupTime);
      Overlays.FixupProperties;
      TEnd(OverlayFixupTime);
      G.Style(1);
      UseOverlayAxis := Simulation.GraphStyle <> gsNoOffset; { Overlays separate? }
      IU := Simulation.IntensityUnits;
      if not UseOverlayAxis then begin
        { Overlays on simulation, so normally scale on simulation }
        AnySpectrum := AnySpectrum and SpectrumPlot.FindLimits(Simulation);
        UseOverlayAxis := not AnySpectrum; { no simulation, so scale on overlay }
      end else if AnySpectrum then begin
        Overlays.AdjustPlotFraction(lowerFrac);
        f := 1+Abs(lowerFrac)/(1-Abs(lowerFrac)); { total range of simulation to give Overlay in 0..1 occupying lowerFrac }
        if lowerFrac >= 0 then
          G.Map(xl, xu, 1-f*21/20, 1)  { Overlay will be in 0..1 range, simulation in -1..0 if PlotSplit is 0.5 }
        else
          G.Map(xl, xu, -f/20, f); { Overlay will be in 0..1 range, simulation in  1..2 if PlotSplit is 0.5 }
      end;
      if UseOverlayAxis then begin
        IU := Overlays.IntensityUnits;
        PR := PlotRange(G, 0, 1, 0, 1, 1);
      end else
        if IsArbitraryStyle(IU)
           or IsArbitraryStyle(Overlays.IntensityUnits)
           or not Overlays.LimitsInRange(Simulation,Simulation.PlotUnits,yi,yx,IU,AllArbitrary)
           or AllArbitrary
        then
          PR := PlotRange(G, 0, 1, SpectrumPlot.IMin, SpectrumPlot.IMax)
        else
          PR := PlotRange(G, 0, 1, Min(SpectrumPlot.IMin,yi), Max(SpectrumPlot.IMax,yx));
      Overlays.Show(G, Simulation, Simulation.PlotUnits, IU, PR, SpectrumPlotRanges, UseOverlayAxis, Simulation.GraphStyle in [gsOffsetAll, gsInvertedAll]);
    end else
      UseOverlayAxis := True; { Actually not, but force recompute }
    if UseOverlayAxis then begin
      SpectrumPlot.FindLimits(Simulation);
      PR := PlotRange(G, 0, 1, SpectrumPlot.IMin, SpectrumPlot.IMax);
    end;

    ShowObservedLines;
    f := 1+(1-Abs(lowerFrac))/Abs(lowerFrac);
    if Assigned(Overlays) then
      case Simulation.GraphStyle of
        gsOffset, gsOffsetAll:
          if lowerFrac >= 0 then
            G.Map(xl, xu, -f/20, f) // Was G.Map(xl, xu, -0.1, 2)
          else
            G.Map(xl, xu, 1-f*21/20, 1);
        gsInverted, gsInvertedAll: 
          if lowerFrac >= 0 then
            G.Map(xl, xu, 1+f/20, 1-f) // Was G.Map(xl, xu, 1.1, -1);
          else
            G.Map(xl, xu, f*21/20, 0)
        { Default is -0.05, 1}
      end
    else
      case Simulation.GraphStyle of
        gsInverted, gsInvertedAll: G.Map(xl, xu, 1.05, 0);
                                   { Default is -0.05, 1}
      end;
    G.Style(0);
    G.Axis(xl, xu, G.Ymin, true, XLabel(Simulation));
    if AnySpectrum then begin
      if Simulation.GraphStyle <> gsNoOffset then begin
        PR := PlotRange(G, 0, 1, SpectrumPlot.IMin, SpectrumPlot.IMax, -9);
        SpectrumPlot.Show(G, PR, Simulation.GraphStyle in [gsOffsetAll, gsInvertedAll], True, SpectrumPlotRanges);
      end else if Simulation.ShowParts then begin
        PR.Tag := -9;
        SpectrumPlot.Show(G, PR, False, True, SpectrumPlotRanges)
      end else begin
        PR.Tag := -9;
        SpectrumPlot.Show(G, PR, False, True, SpectrumPlotRanges);
      end;
      Result := Format('%d lines in range with total intensity %s',
                              [SpectrumPlot.nLinesInRange, DtosM(SpectrumPlot.IntensityInRange,12,-1)]);
      if HasUnits(Simulation.IntensityUnits) then
        Result := Result + ' ' + NameOfUnit(Simulation.IntensityUnits, True, Simulation.PlotUnits);
    end;
    G.Finish;
  finally
    SpectrumPlot.Free;
  end;
  TEnd(MakeLowerPlotTime);
end;

procedure MakePlot(Canvas: TExportableCanvas);
var
  Simulation: TSimulation;
  AnySpectrum: Boolean;
  G: TGraphM;
  xl, xu, yl, yu: Double;
  SpectrumPlotRanges: TPlotRanges;
  PlotColours: TColours;
  LastFortrat: TFortratPlot;
begin
  Simulation := GlobalMixture.Simulations[0];
  AnySpectrum := Simulation.ShowSum or Simulation.ShowParts;
  if not (AnySpectrum or Simulation.ShowFortrat) then
    Exit;
  if GlobalMixture.PrintLevel >= plDetail then
    GlobalMixture.PrettyPrint(Log, GlobalMixture.PrintLevel);
  G := TGraphM.Create(nil);
  try
    if Canvas is TNullCanvas then
      G.SetTextStream(TNullCanvas(Canvas).Stream, False);
    G.CurrentCanvas := Canvas;
    G.PspaceArg(xl,xu,yl,yu);
    if AnySpectrum then begin
      if Simulation.ShowFortrat then
        G.Pspace(xl,xu,yl,0.5-(1-yu));
      MakeLowerPlot(G,GlobalMixture,Simulation,GlobalOverlays,SpectrumPlotRanges,PlotColours);
    end;
    if Simulation.ShowFortrat then begin
      if AnySpectrum then
        G.Pspace(xl,xu,0.5+yl,yu);
      MakeUpperPlot(G, Simulation, LastFortrat, SpectrumPlotRanges);
      LastFortrat.Free; { Not used }
    end;
  finally
    G.Free;
  end;
end;

procedure MakePlotStream(IsEPlot: Boolean; Stream: TStream; CClass: TStreamCanvasClass);
var
  G: TGraphM;
  Canvas: TExportableCanvas;
begin
  Canvas := CClass.Create(Stream,250,179);
  if IsEPlot then
    try
      G := TGraphM.Create(nil);
      if CClass = TNullCanvas then
        G.SetTextStream(Stream,False);
      G.CurrentCanvas := Canvas;
      ReadLevelPlotter(GlobalMixture).PlotItem(G,GlobalMixture,True,False)
    finally
      G.CurrentCanvas := nil;
      G.SetTextStream(nil);
      G.Free;
    end
  else
    Plotter.MakePlot(Canvas);
  Canvas.Free;
end;

procedure MakePlotFile(FileName: string; IsEPlot: Boolean);
var
  Stream: TFileStream;
  CClass: TStreamCanvasClass;
begin
  Stream := TFileStream.Create(FileName,fmCreate);
  CClass := GetStreamCanvasClassFromFilename(FileName);
  if CClass = TTextCanvas then
    CClass := TNullCanvas;
  try
    MakePlotStream(IsEPlot, Stream, CClass);
  finally
    Stream.Free;
  end;
end;

end.
