unit SimpleQnos;
{$I utils.inc}{Copyright (C) 2010 C M Western; see commandline.pas for license}
{ Simple quantum number and spfit quantum number helper routines }
interface

uses QuantumNumber, StaticField, Mixture, MathSup;

type
  T10Integers = array [0..9] of Integer;
  TSimpleQnos = record
    QNP, QN: T10Integers;
  end;
  TStateMap = array of TState;

function ComputeQ(AStateU, AStateL: TState; Molecule: TMolecule; Mixture: TMixture; NQN, RemoveSpins: Integer; SpecialQN: Integer; ForceLambda: Boolean; out nV: Integer): Integer;
function nVFromQ(Q: Integer):Integer;
function ConvertQuantumNumbers(var ManifoldU: TManifold; var StateU: TState; var JU:TQuantumNumber; var SymU: Shortint; var IndexU: Smallint;
                               var ManifoldL: TManifold; var StateL: TState; var JL:TQuantumNumber; var SymL: Shortint; var IndexL: Smallint;
                               var Name: string; Molecule: TMolecule;
                               Q, NQN: Integer; Qnos: TSimpleQnos;
                               const ExField: TStaticFields;
                               const StateMap: array of TState;
                               const MapUpper, MapLower: TIntegerArray;
                               const QuantumNumberFormat: TQuantumNumberFormat;
                               const RemoveSpins: Integer;
                               BranchRank: TQuantumNumber; { Used for linear and symmetric tops }
                               { Parameters below only needed/used for symmetric tops }
                               const DefaultSymmetry: Integer;
                               SignOfKisl: Boolean
                               ):Boolean;
procedure AddToMap(var Map: TIntegerArray; n: Integer; ms: string);
function NameIsStateNumber(Molecule:TMolecule; var MapUpper, MapLower: TIntegerArray):Boolean;

type
  TnnTable = array of TQuantumNumberArray;
  { First index is NN; second is quantum number }
function MakennTable(ElectronSpin:TQuantumNumber; Molecule: TMolecule; RemoveSpins: Integer): TnnTable;
procedure ClearnnTableCache;

implementation

uses SysUtils, Math, OAsymT, SymT, HundA, Worker, UCompat;

function NameIsStateNumber(Molecule:TMolecule; var MapUpper, MapLower: TIntegerArray):Boolean;
var
  Map: TIntegerArray;
  i, k, Sname: Integer;
begin
  { make default map assuming state name is state number, if names in appropriate pattern }
  Result := True;
  Setlength(Map, Molecule.nStates);
  for i := 0 to Molecule.nStates-1 do
    Map[i] := -1;
  for i := 0 to Molecule.nManifolds-1 do
    for k := 0 to Molecule.Manifold[i].nStates-1 do begin
      if TryStrToInt(Molecule.Manifold[i].State[k].Name, Sname)
         and (Sname >= 0) and (Sname < Molecule.nStates)
         and (Map[Sname] < 0) then
        Map[Sname] := (i+1)*2000 + k
      else
        Result := False;
      if not Result then
        Break;
    end;
  if Result then begin
    MapLower := Map;
    MapUpper := Copy(MapLower)
  end;
end;

var { Cache results of last call to MakennTable for convenience } { TODO: not thread safe }
  LastElectronSpin: TQuantumNumber;
  LastMolecule: TMolecule;
  nnTable: TnnTable;
  LastRemoveSpins: Integer;

procedure ClearnnTableCache;
begin
  LastMolecule := nil;
end;
  
{ To limit the number of quantum numbers for multiple spins,
  a composite quantum number, nn, is used, which is essentially
  an index into a table containing offsets from the total quantum number F.
  For example, for the case S=1/2, I1=1, I2=I3=1/2 with I2 and I3 equivalent
  and coupled to give Itot = 0 and 1, the table is:
    nn    N     J     F1   Itot
     0  -2.5  -2.0  -1.0   1.0
     1  -1.5  -1.0   0.0   0.0
     2  -1.5  -1.0   0.0   1.0
     3  -1.5  -1.0  -1.0   1.0
     4  -1.5  -2.0  -1.0   1.0
     5  -0.5   0.0   0.0   0.0
     6  -0.5  -1.0   0.0   0.0
     7  -0.5   0.0   1.0   1.0
     8  -0.5   0.0   0.0   1.0
     9  -0.5  -1.0   0.0   1.0
    10  -0.5   0.0  -1.0   1.0
    11  -0.5  -1.0  -1.0   1.0
    12   0.5   1.0   0.0   0.0
    13   0.5   0.0   0.0   0.0
    14   0.5   1.0   1.0   1.0
    15   0.5   0.0   1.0   1.0
    16   0.5   1.0   0.0   1.0
    17   0.5   0.0   0.0   1.0
    18   0.5   0.0  -1.0   1.0
    19   1.5   1.0   0.0   0.0
    20   1.5   2.0   1.0   1.0
    21   1.5   1.0   1.0   1.0
    22   1.5   1.0   0.0   1.0
    23   2.5   2.0   1.0   1.0
  The order of the table is:
    N = min .. max
      Itot = min .. max
        and all other spins max .. min
}

function MakennTable(ElectronSpin:TQuantumNumber; Molecule: TMolecule; RemoveSpins: Integer): TnnTable;
var
  SpinTotal, N: TQuantumNumber;
  i, nPerm, Perm, nItot, nSpins, NucOffset: Integer;
  Itots, Itot: TQuantumNumberArray;
  Spins, Offsets, BaseValue, HighValue: TQuantumNumberArray;
  ItotIndex: array of Integer;
  OK: Boolean;
  function OffsetOK(i:Integer):Boolean;
  { Is Offset i OK ? }
  begin
    if ItotIndex[i-1] < 0 then begin
      { Check spin before }
      Result := Abs(Offsets[i]-Offsets[i-1]) <= Spins[i-1];
      if Result and (i<nSpins-1) then
        { Check spin after }
        if ItotIndex[i] < 0 then
          Result := Abs(Offsets[i+1]-Offsets[i]) <= Spins[i];
    end else
      Result := True; { Enforced elsewhere }
  end;
begin
  { TODO: Needs testing for RemoveSpins > 0 }
  nSpins := Molecule.nNuclei - RemoveSpins;
  NucOffset := 0;
  if nSpins < 0 then
    nSpins := 0
  else if ElectronSpin > 0 then begin
    Inc(nSpins);
    Inc(NucOffset);
  end;
  SetLength(Spins, nSpins); { All spins in one place for simplicity }
  SetLength(Offsets, nSpins); { Offsets from base spin }
  SetLength(BaseValue, nSpins); { Lowest possible value }
  SetLength(HighValue, nSpins); { Lowest possible value }
  Setlength(ItotIndex, nSpins); { Index into Itot(s) arrays, or -1 }
  for i := 0 to nSpins - 1 do
    ItotIndex[i] := -1; { Initially flag no spins as equivalent }
  Spins[0] := ElectronSpin;
  SpinTotal := ElectronSpin;
  nPerm := ElectronSpin + 1; { Number of combinations of spin additions/projections }
  nItot := 0; { Number of sets of equivalent spins }
  for i := 0 to Molecule.nNuclei - 1  - RemoveSpins do begin
    SpinTotal := SpinTotal + Molecule.Nuclei[i].Spin;
    nPerm := nPerm*(Molecule.Nuclei[i].Spin+1);
    Spins[NucOffset+i] := Molecule.Nuclei[i].Spin;
    if Molecule.Nuclei[i].AsNext then begin
      if (i = 0) or not Molecule.Nuclei[i-1].AsNext then begin
        Inc(nItot); { First nucleus of equivalent set }
        SetLength(Itots, nItot);
        Itots[nItot-1] := Molecule.Nuclei[i].Spin;
      end;
      Inc(Itots[nItot-1],Molecule.Nuclei[i].Spin);
      ItotIndex[NucOffset+i] := nItot-1;
      ItotIndex[NucOffset+i+1] := nItot-1;
    end;
  end;
  SetLength(Result, nPerm);
  Perm := 0;
  SetLength(Offsets, nSpins);
  SetLength(Itot, nItot);
  N := -SpinTotal-2; { Acutally offset of N from F }
  while N <= SpinTotal do begin
    Inc(N,2);
    { Cycle through possible Itot values }
    for i := 0 to nItot - 1 do { Initialize to all at smallest values }
      Itot[i] := SmallestProjection(Itots[i]);
    repeat
      Offsets[0] := N;
      BaseValue[0] := N;
      HighValue[0] := N;
      { Intitialise offsets to highest possible values }
      { and record highest and lowest possible values }
      for i := 0 to nSpins - 2 do begin
        { This logic requires assumes equivalent spins at the end }
        if ItotIndex[i] >= 0 then begin
          Offsets[i+1] := Itot[ItotIndex[i]];
          BaseValue[i+1] := Offsets[i+1];
        end else begin
          Offsets[i+1] := Offsets[i] + Spins[i];
          BaseValue[i+1] := BaseValue[i] - Spins[i];
        end;
        HighValue[i] := Offsets[i];
      end;
      repeat
        { Do we have a valid set of offsets? }
        OK := True;
        for i := 1 to nSpins - 1 do begin
          OK := OffsetOK(i);
          if not OK then
            Break;
        end;
        if OK then
          { check on last spin }
          if ItotIndex[nSpins-1] >= 0 then
            { TODO: This test assumes only 2 equivalent spins }
            OK := Itot[ItotIndex[nSpins-1]] >= Abs(Offsets[nSpins-2])
          else
            OK := Spins[nSpins-1] >= Abs(Offsets[nSpins-1]);
        //if OK then begin
        //  Write(Log, OK:5, Perm:5);
        //  for i := 0 to High(Offsets) do
        //    Write(Log, QuantumNumberToStr(Offsets[i], qftJ):6);
        //  WriteLn(Log);
        //end;
        if OK then begin
          { Add to target list }
          Result[Perm] := Copy(Offsets, 0, nSpins);
          Inc(Perm);
        end;
        { Bump up to next combinations of offsets }
        i := 1;
        repeat
          if ItotIndex[i-1] < 0 then begin
            Dec(Offsets[i],2);
            if Offsets[i] >= BaseValue[i] then
              Break;
            { Need to step next spin - reset this one }
            Offsets[i] := HighValue[i];  
          end;
          Inc(i);
        until i = nSpins;
      until i = nSpins;
      { Bump up to next combination of Itot values }
      i := 0;
      repeat
        Inc(Itot[i],2);
        if Itot[i] > Itots[i] then begin
          Itot[i] := SmallestProjection(Itots[i]);
          Inc(i);
        end else
          Break;
      until i = nItot; { Stop if last Itot reached maximum }
    until i = nItot;
  end;
  if Perm <> nPerm then { Should not happen }
    raise Exception.CreateFmt('MakennTable: %d %d', [Perm, nPerm]);
end;

procedure AddToMap(var Map: TIntegerArray; n: Integer; ms: string);
var
  v, c, j: Integer;
begin
  c := Pos('.',ms);
  if c > 0 then
    { Manifold.State format }
    v := (StrToInt(Copy(ms,1,c-1))+1)*2000 + StrToInt(Copy(ms,c+1,Length(ms)))
  else
    v := StrToInt(ms);
  c := Length(Map);
  if c <= n then begin
    SetLength(Map,n+1);
    for j := c to n - 1 do
      Map[j] := j;
  end;
  Map[n] := v;
end;

function ComputeQ(AStateU, AStateL: TState; Molecule: TMolecule; Mixture: TMixture; NQN, RemoveSpins: Integer; SpecialQN: Integer; ForceLambda: Boolean; out nV: Integer): Integer;
var
  nSpins, nBase: Integer;
  HasSpin: Boolean;
  procedure HandleMolecule(Molecule: TMolecule);
  begin
    if Molecule is TAsymmetricMolecule then
      nV := 3
    else if Molecule is TSymmetricMolecule then
      nV := 2
    else if Molecule is TLinearMolecule then
      if not ForceLambda or TLinearMolecule(Molecule).AllSigma then
        nV := 1
      else
        nV := 2 { Room for Lambda, required for some SPFIT linear cases }
    else
      raise NotYet;
    if (Result > 0) and (nV <> Result) then
      raise EReportException.CreateFmt('Inconsistent nV values: %d <> %d', [nV, Result]);
    Result := nV;
    nSpins := Max(nSpins,Molecule.nNuclei - RemoveSpins);
    HasSpin := HasSpin or not Molecule.AllSinglet;
  end;
var
  i, j: Integer;
begin
  { Work out spfit quantum number format index, Q }
  { SpecialQN indicates a PGOPHER extension has been used }
  Result := -1;
  HasSpin := False;
  nSpins := -RemoveSpins;
  Result := -1;
  if Assigned(Molecule) then begin
    HandleMolecule(Molecule);
    if Assigned(AStateL) and Assigned(AStateU) then
      HasSpin := (AStateL.S > 0) or (AStateU.S > 0);
  end else
    for i := 0 to Mixture.nSpecies-1 do
      with Mixture.Species[i] do
        for j := 0 to ComponentCount-1 do
          HandleMolecule(Isotopomer[j]);
  if HasSpin then
    Inc(nSpins);
  { TODO: Does not yet handle equivalent spins }
  nBase := nV; { track number of non-hyperfine quantum numbers }
  if HasSpin then
    Inc(nBase); { Allow for J and N quantum numbers }
  if SpecialQN >= 0 then
    nBase := Max(nBase, SpecialQN+1);
  if (Molecule = nil) or (((Molecule.nStates > 1) or (nSpins + nBase < NQN) ) and (NQN > nBase + nSpins)) then begin
    Inc(Result,11); { vibrational quantum number present }
    nV := nV+1; { Quantum number (starting from 1) of vibrational quantum number }
  end else
    nV := -1; { Flag not present }
  if nSpins + nBase > 6 then
    Inc(Result,40);
  if SpecialQN >= 0 then
    Inc(Result, 10000);
end;

function nVFromQ(Q: Integer):Integer;
begin
  case Q of
    12, 32, 52: Result := 2;
    14, 19, 54: Result := 4;
    13, 33, 53, 73: Result := 3;
  else
    Result := -1
  end;
end;

function ConvertQuantumNumbers(var ManifoldU: TManifold; var StateU: TState; var JU:TQuantumNumber; var SymU: Shortint; var IndexU: Smallint;
                               var ManifoldL: TManifold; var StateL: TState; var JL:TQuantumNumber; var SymL: Shortint; var IndexL: Smallint;
                               var Name: string; Molecule: TMolecule;
                               Q, NQN: Integer; Qnos: TSimpleQnos;
                               const ExField: TStaticFields;
                               const StateMap: array of TState;
                               const MapUpper, MapLower: TIntegerArray;
                               const QuantumNumberFormat: TQuantumNumberFormat;
                               const RemoveSpins: Integer;
                               BranchRank: TQuantumNumber;
                               { Parameters below only needed/used for symmetric tops }
                               const DefaultSymmetry: Integer;
                               SignOfKisl: Boolean
                               ):Boolean;

  function MatchHyperfine(BasisState: TBasisState; const Q: array of Integer):Boolean;
  var
    i, ioff: Integer;
    FF, Fa : TQuantumNumber;
  begin
    Result := False;
    ioff := NQN-1;
    FF := Q[ioff];
    for i := StateU.Molecule.nNuclei-1-RemoveSpins downto 0 do begin
      if StateU.Molecule.Nuclei[i].AsNext then
        Fa := BasisState.HyperfineState[i]
      else begin
        Fa := FF;
        Dec(FF,BasisState.HyperfineState[i]);
      end;
      if Q[ioff] <> Fa then
        Exit;
      Dec(ioff);
    end;
    Result := True;
  end;

  function GetQnos(
             vflag: Integer;            { 1 if v quantum number present, 0 otherwise }
             AggregateQnos:Boolean;     { If nn aggregate hyperfine quantum number present }
             nBase:Integer;             { The number of non J, v and hyperfine quantum numbers }
                                        { At least 1 for linear, 2 for symmetric top, 3 for asymmetric top }
             SpecialQN: Boolean = False { If PGOPHER extension present }
            ): Boolean;
  var
    gOther, kOther: TQuantumNumber;
    function GetSym(var F: TQuantumNumber; const Q: array of Integer; var Manifold: TManifold; var State: TState;
                    var ASym: Shortint; var Number: Smallint; var AName: string; const Map: TIntegerArray;
                    OtherSym: Integer; var DuplicatesOK: Boolean):Boolean;
    var
      OtherQno: array [0..1] of Integer;
      Hyperfine: THyperfineBasis;
      N, J, FF, gU1, gU2, gDipole: TQuantumNumber;
      nn, SpinOffset, i, Sindex, SQN, nBaseQ, Sign1, Sign2, RemoveSpinsEff: Integer;
      klToggle, ForceSearch: Boolean;
      AlternateSym, OrgSym: ShortInt;
      OName: string;
    begin
      Result := True;
      AName := '';
      { Check for all values -1, which implies state rather than transition specified, with one state missing }
      for i := 0 to NQN-1 do begin
        Result := Result and (Q[i] = -1);
        if not Result then
          Break;
      end;
      if Result then
        Exit; { State not specified }
      if not Assigned(Molecule) and Assigned(Manifold) then
        Molecule := Manifold.Molecule;
      if not Assigned(Molecule) and Assigned(State) then
        Molecule := State.Molecule;
      Result := False;
      { Find state }
      if Assigned(State) then
        Sindex := State.NodeIndex
      else
        Sindex := 0;
      klToggle := False; { Symmetric tops only }
      ForceSearch := False;
      if vflag > 0 then begin
        Sindex := Q[nBase] div 2;
        if Sindex < 0 then begin
          WriteLn(Log, 'State Index must be positive');
          Exit;
        end else if (Sindex <= High(StateMap)) and (StateMap[Sindex] <> nil) then begin
          State := StateMap[Sindex];
          Manifold := State.Manifold;
          Sindex := State.NodeIndex;
        end else if Sindex <= High(Map) then begin
          Sindex := Map[Sindex];
          if Sindex >= 1000 then begin
            i := Sindex div 1000;
            klToggle := Odd(i); { Symmetric tops only - flag for other half of state }
            if i > 1 then
              Manifold := Molecule.Manifold[(i div 2)-1];
          end;
          Sindex := Sindex mod 1000;
        end else begin
          if not Assigned(Molecule) then begin
            WriteLn(Log, 'Must specify Molecule');
            Exit;
          end;
          { Default map }
          for i := 0 to Molecule.nManifolds-1 do begin
            Manifold := Molecule.Manifold[i];
            if Sindex < Manifold.nStates then
              Break
            else
              Dec(Sindex, Manifold.nStates);
          end;
        end;
      end;
      if not Assigned(Manifold) then begin
        WriteLn(Log, 'Must specify Manifold or State');
        Exit;
      end;
      if Sindex >= Manifold.nStates then begin
        WriteLn(Log, 'Requested State ', Sindex, ' but only ', Manifold.nStates, ' available');
        Exit;
      end;
      State := Manifold.State[Sindex];
      N := Q[0];
      F := Q[NQN-1];
      { Discard zero spin nuclei }
      RemoveSpinsEff := Max(RemoveSpins, State.Molecule.nNuclei-High(State.Molecule.HyperfineBasisStates));
      nBaseQ := NQN-vflag-Max(State.Molecule.nNuclei-RemoveSpinsEff,0); { Real number of base quantum numbers; nBase may be different if both J and N present }
      SetLength(Hyperfine, Max(State.Molecule.nNuclei-RemoveSpinsEff,0));
      if AggregateQnos then begin
        if (State.Molecule <> LastMolecule) or (State.S <> LastElectronSpin) or (RemoveSpinsEff <> LastRemoveSpins) then begin
          nnTable := MakennTable(State.S, State.Molecule, RemoveSpinsEff);
          LastElectronSpin := State.S;
          LastMolecule := State.Molecule;
          LastRemoveSpins := RemoveSpinsEff;
        end;
        nn := Q[nBase+vflag] div 2; { Was assumed to be a quantum number and doubled }
        if N <> F+nnTable[nn][0] then begin
          WriteLn(Log,'Inconsistent nn quantum number: ', nn, ' ', QuantumNumberToStr(N,qftJ), ' ', QuantumNumberToStr(F,qftJ));
          Exit;
        end;
        if State.S > 0 then
          SpinOffset := 1
        else
          SpinOffset := 0;
        J := F + nnTable[nn][SpinOffset];
        for i := 0 to State.Molecule.nNuclei - 2 - RemoveSpinsEff do
          if State.Molecule.Nuclei[i].AsNext then
            Hyperfine[i] := nnTable[nn][i+SpinOffset+1]
          else
            Hyperfine[i] := F + nnTable[nn][i+SpinOffset+1];
      end else begin
        if (nBaseQ = nBase) or (SpecialQN and (nBaseQ = nBase+1)) then begin
          J := Q[0]; { J Not given, so same as N }
        end else begin
          J := Q[nBase+vflag];
          if SpecialQN and (J < 0) then
            J := Q[nBase+vflag-1];
        end;
        if State.Molecule.nNuclei-RemoveSpinsEff <= 0 then
          F := J; { F not given }
        { Copy remaining hyperfine quantum numbers }
        for i := 0 to State.Molecule.nNuclei - 2 - RemoveSpinsEff do
          Hyperfine[i] := Q[NQN+i-State.Molecule.nNuclei];
      end;
      { Convert from quantum numbers to differences }
      FF := J;
      if State.Molecule.nNuclei-RemoveSpinsEff > 0 then
        Hyperfine[High(Hyperfine)] := F;
      for i := 0 to State.Molecule.nNuclei - 1 - RemoveSpinsEff do
        if State.NucleusWithSpin(i).AsNext then
          { Leave quantum number as is }
        else begin
          Hyperfine[i] := Hyperfine[i]-FF;
          FF := Hyperfine[i]+FF; { Value of qunantum number }
        end;
      Asym := -1; { Search all symmetries by default }
      OtherQno[0] := Low(TQuantumNumber);
      OtherQno[1] := Low(TQuantumNumber);
      for i := 0 to nBase-2 do
        OtherQno[i] := Q[i+1];
      if State is TSymmetricTop then begin
        SQN := Q[1];
        with TSymmetricTop(State) do
          ForceSearch := Molecule.IsDegenerate(Symmetry);
        if (SQN < 0) and SignOfKisl then
          { Force encoded sign - a simple "-3" does not force the special encoding }
          SQN := (Abs(SQN)+100000)*1600 + 200;
        if SQN > 100000 then begin
          { K and (possibly) symmetry in one quantum number }
          Asym := SQN mod 100 - 1;
          if Asym > Molecule.LastSym then
            raise Exception.Create('Bad encoded symmetry');
          Sign1 := SQN div 400 mod 4;
          Sign2 := SQN div 100 mod 4;
          if (Sign1 = 3) or (Sign2 = 3) then
            raise Exception.Create('Bad Sym Special');
          SQN := SQN div 1600 - 100000;
          with TSymmetricTop(State) do
            if Molecule.CharacterTable.HasSeparableDegeneracy then begin
              if (Sign1 = 0) and (Sign2 <> 0) and (not Molecule.IsDegenerate(Symmetry) or (SQN = 0)) then begin
                Sign1 := Sign2; { Sign is phase, not l }
                Sign2 := 0;
              end;
              case Sign2 of
                0: if Sign1 <> 0 then
                     if (SQN = 0) and Molecule.IsDegenerate(Symmetry) then
                       OtherQno[1] := 2
                     else
                       OtherQno[1] := 0;
                1: begin OtherQno[1] :=  2; ForceSearch := False; end;
                2: begin OtherQno[1] := -2; ForceSearch := False; end;
              end;
              if Sign1 <> 0 then
                OtherQno[1] := (OtherQno[1] and not 12) or Sign1*4;
            end else
              case Sign2 of
                0: if ForceSearch then
                     ForceSearch := False { l unspecified }
                   else
                     OtherQno[1] := 0; { l must be zero }
                1: begin OtherQno[1] :=  2; ForceSearch := False; end;
                2: begin OtherQno[1] := -2; ForceSearch := False; end;
              end;
        end;
        OtherQno[0] := Abs(SQN); { |K| }
        //NotKZero := NotKZero and (OtherQno[0] <> 0);
        if (OtherQno[0] > 0) and ForceSearch then
          if SignOfKisl then
            { Sign not specified } // Was: OtherQno[1] := 2*Sign(SQN)
          else
            if klToggle then
              OtherQno[1] := 2
            else
              OtherQno[1] := -2
      end else if State is TLinear then begin
        ASym := -1;
        if nBase = 3 then begin
          { J, Omega, Lambda given; Lambda<0 for negative parity }
          if Q[2] < 0 then
            Asym := 1;
          OtherQno[1] := Low(TQuantumNumber);
          N := -1;
        end else if nBase = 2 then begin
          { N, Lambda given; <0 for negative parity }
          if Q[1] < 0 then
            Asym := 1
          else if Q[1] > 0 then
            Asym := 0;
          { Leave parity unspecified for Sigma states }
          OtherQno[0] := Low(TQuantumNumber);
        end else if nBase = 1 then begin
          if SpecialQN then begin
            SQN := -Q[nBaseQ-1];
            if nBaseQ = 4 then
              SQN := -Q[nBaseQ-2]
            else if nBaseQ = 3 then begin
              { N J Special or J Omega Fn }
              if Q[1] < 0 then begin
                { J Omega Fn }
                J := N;
                N := -1;
                SQN := -Q[nBaseQ-2];
              end;
            end else if nBaseQ = 2 then begin
              { Only one of J and N given - assume J for this case }
              N := -1;
            end;
            if (nBaseQ = 4) or ((nBaseQ = 3) and (Q[1] < 0)) then begin
              { Both Omega and F1/F2/F2 given (redundant) }
              { Assume first quantum number is Omega }
              OtherQno[0] := SQN;
              if SQN > 1000 then begin { parity specified in Omega }
                OtherQno[0] := (SQN div 1000)-1000;
                if (SQN mod 10) > 0 then
                  ASym := SQN and 1;
              end else
                OtherQno[0] := -SQN; { Only omega given }
              SQN := -Q[nBaseQ-1];
              if SQN > 1000 then begin
                WriteLn(Log, 'Omega specified twice');
                Exit;
              end;
            end;
            if SQN > 1000 then { Omega given }
              OtherQno[0] := (SQN div 1000)-1000
            else if SQN > 9 then
              OtherQno[1] := SQN div 10; { F1/F2/F3... }
            { else just parity }
            if (SQN mod 10) > 0 then begin
              if (ASym >= 0) and (Asym <> SQN and 1) then begin
                WriteLn(Log, 'Inconsistent Parity');
                Exit;
              end;
              ASym := SQN and 1;
            end;
            if (Asym >= 0) and ((J and 2) <> 0) then
              ASym := 1 - ASym; { Convert e/f -> +/- }
          end;// else
          //  { Sigma state }
          //  if Odd(N div 2) xor (TLinear(State).Lambda < 0) then
          //    ASym := 1 { - parity }
        end;
      end;
      OrgSym := ASym;
      AName := Manifold.FindStateLabel(Number,ASym,AlternateSym,F,OtherQno,QuantumNumberFormat,ExField,OrgSym,J,N,State,RemoveSpinsEff,Hyperfine);
      if (Asym < 0) and (OrgSym < 0) and (AlternateSym < 0) and (Molecule is TSymmetricMolecule) and not DuplicatesOK and (OtherSym >= 0) and (BranchRank >= 0) then begin
        { Symmetry not pinned down - force allowed transition }
        if Odd(BranchRank div 2) then begin
          for i := 0 to Molecule.LastSym do
            if TSymmetricMolecule(Molecule).DipoleAllowed(OtherSym,i) then
              OrgSym := i;
        end else
          OrgSym := OtherSym;
        AName := Manifold.FindStateLabel(Number,ASym,AlternateSym,F,OtherQno,QuantumNumberFormat,ExField,OrgSym,J,N,State,RemoveSpinsEff,Hyperfine);
      end;
      Result := ASym >= 0;
      if not Result then begin
        if (Number = -3) and (AlternateSym >= 0) then begin
          { Two matches with different symmetry - look for allowed transition or other way of resolving the ambiguity }
          if State is TSymmetricTop then begin
            if (BranchRank >= 0) and (OtherSym >= 0) then begin
              if OtherSym >= 0 then begin
                if Odd(BranchRank div 2) then
                  if TSymmetricMolecule(Molecule).DipoleAllowed(OtherSym,-ASym-1) then
                    ASym := -ASym-1
                  else { Assume the alternate choice is allowed }
                    ASym := AlternateSym
                else
                  ASym := OtherSym;
                DuplicatesOK := False;
              end;
            end else if not DuplicatesOK then begin
              { fallback }
              DuplicatesOK := False;
              if SignOfKisl then begin
                { Two matches - use default symmetry }
                Asym := Molecule.ConvertDefaultSymmetry(J, DefaultSymmetry);
              end else
                { Two matches - assume sign of K is given in file, and take second if K < 0 }
                if (SQN >= 0) xor (TSymmetricMolecule(Manifold.Molecule).CharacterTable.IReps[TSymmetricTop(State).Symmetry].rho > 0) then
                  ASym := -ASym-1 { Take lowest symmetry number if K >= 0 }
                else
                  ASym := AlternateSym; { Otherwise take higher symmetry number }
            end else if not ForceSearch and (SQN < 0) then begin
              { See K logic above }
              ASym := -ASym-1;
              if TSymmetricMolecule(Manifold.Molecule).CharacterTable.IReps[TSymmetricTop(State).Symmetry].rho > 0 then
                ASym := -ASym-1
              else
                ASym := AlternateSym;
            end;
            if ASym >= 0 then
              AName := Manifold.FindStateLabel(Number,ASym,AlternateSym,F,OtherQno,QuantumNumberFormat,ExField,Asym,J,N,State,RemoveSpinsEff,Hyperfine);
            Result := ASym >= 0;
          end else if State is TLinear then begin
            DuplicatesOK := False;
            if BranchRank >= 0 then
              if OtherSym >= 0 then begin
                { Transition, and other symmetry known }
                if Odd(BranchRank div 2) then
                  ASym := 1-OtherSym
                else
                  ASym := OtherSym;
                AName := Manifold.FindStateLabel(Number,ASym,AlternateSym,F,OtherQno,QuantumNumberFormat,ExField,Asym,J,N,State,RemoveSpinsEff,Hyperfine);
                Result := ASym >= 0;
              end else
                DuplicatesOK := True;
          end else
            DuplicatesOK := False;
        end else if (Number = -3) and (Molecule is TSymmetricMolecule) and TSymmetricMolecule(Molecule).IsDegenerate((State as TSymmetricTop).Symmetry) and (BranchRank >= 0) then begin
          if not DuplicatesOK and (gOther <> Low(TQuantumNumber)) and (kOther + OtherQno[0] > BranchRank) then begin
            { Try using change in g to make selection }
            with State as TSymmetricTop do begin
              gU1 := Molecule.g(OtherQno[0], 2, Symmetry);
              gU2 := Molecule.g(OtherQno[0], -2, Symmetry);
              if gU1 <> gU2 then begin
                if Odd(BranchRank div 2) then
                  gDipole := Molecule.CharacterTable.IReps[Molecule.CharacterTable.DipoleSym].g
                else
                  gDipole := 0;
                gOther := AddG(gOther+gDipole, Molecule.CharacterTable.n);
                if gOther = gU1 then
                  OtherQno[1] := 2
                else if gOther = gU2 then
                  OtherQno[1] := -2
                else
                  raise EReportException.CreateFmt('Fail in SymT SimpleQnos:%d %d %d %d',[gOther, gU1, gU2, gDipole]);
                OName := AName;
                AName := Manifold.FindStateLabel(Number,ASym,AlternateSym,F,OtherQno,QuantumNumberFormat,ExField,Asym,J,N,State,RemoveSpinsEff,Hyperfine);
                if ASym < 0 then begin
                  OtherQno[1] := Low(TQuantumNumber); { put back so that error message mages sense }
                  AName := OName;
                end else
                  Result := True;
              end;
            end;
          end; { else require other state to be pinned down }
        end else if (Asym < 0) and (Molecule is TSymmetricMolecule) and DuplicatesOK then
          { Symmetry not pinned down - maybe lower state will be pinned down }
        else
          DuplicatesOK := False;
        if not Result and not DuplicatesOK then begin
          WriteLn(Log,AName);
          Manifold.FindStateLabel(Number,ASym,AlternateSym,F,OtherQno,QuantumNumberFormat,ExField,-1,J,N,State,RemoveSpinsEff,Hyperfine,True);
        end;
      end;
      kOther := OtherQno[0];
      gOther := Low(TQuantumNumber);
      if Result and (Molecule is TSymmetricMolecule) then
        with State as TSymmetricTop do begin
          if (OtherQno[1] = Low(TQuantumNumber)) and not Molecule.IsDegenerate(Symmetry) then
            OtherQno[1] := 0;
          if OtherQno[1] <> Low(TQuantumNumber) then
            gOther := Molecule.g(OtherQno[0], OtherQno[1], Symmetry)
        end;
      if Result then
        Number := -Number; { Force eigenvalue search for all transitions }
      //if Result and ((Manifold.nStates > 1) or (NQN > nBase+vflag) or ForceSearch) then
      //  Number := -Number; { Force eigenvalue search for hyperfine }
    end;
    var
      NameU, NameL: string;
      DuplicatesOK: Boolean;
  begin
    //NotKZero := True;
    DuplicatesOK := True;
    NameL := '';
    gOther := Low(TQuantumNumber);
    Result := GetSym(JU, Qnos.QNP, ManifoldU, StateU, SymU, IndexU, NameU, MapUpper, -1, DuplicatesOK);
    if not Result and DuplicatesOK then begin
      { Try finding states in other order }
      DuplicatesOK := False;
      Result := GetSym(JL, Qnos.QN, ManifoldL, StateL, SymL, IndexL, NameL, MapLower, -1, DuplicatesOK);
      if Result then
        Result := GetSym(JU, Qnos.QNP, ManifoldU, StateU, SymU, IndexU, NameU, MapUpper, SymL, DuplicatesOK);
    end else if Result then begin
      DuplicatesOK := False;
      Result := GetSym(JL, Qnos.QN, ManifoldL, StateL, SymL, IndexL, NameL, MapLower, SymU, DuplicatesOK);
    end;
    Name := ':' + NameU + ' - ' + NameL + ':' + Name;
  end;

begin
  if Assigned(StateU) and not Assigned(ManifoldU) then
    ManifoldU := StateU.Manifold;
  if Assigned(StateL) and not Assigned(ManifoldL) then
    ManifoldL := StateL.Manifold;
  Result := True; { Assume fail }
  { This will need adjustment }
  case Q of
       { Symmetric tops and some linear molecules }
    13,33: Result := not GetQnos(1,False,2);
     2,22: Result := not GetQnos(0,False,2);
// ns=??  62,73: Result := not GetQnos(0,False,3);
       53: Result := not GetQnos(1,True,2);
       42: Result := not GetQnos(0,True,2);

       { Linear Sigma States }
    12,32: Result := not GetQnos(1,False,1);
  1,72,21: Result := not GetQnos(0,False,1);
// ns=??  61: Result := not GetQnos(0,False,1);
       52: Result := not GetQnos(1,True,1);
       41: Result := not GetQnos(0,True,1);

       { Linear Case (a) }
     8,19: with Qnos do begin
             QN[0] := QN[0]-1;
             QN[1] := QN[1]-1;
             QNP[0] := QNP[0]-1;
             QNP[1] := QNP[1]-1;
             if Q = 19 then
               Result := not GetQnos(1,False,3)
             else
               Result := not GetQnos(0,False,3)
           end;

       { Linear with PGOPHER extension }
    10001,
    10002,
    10041: Result := not GetQnos(0,False,1,True);

       { Asymmetric tops }
       14: Result := not GetQnos(1,False,3);
        3: if (ManifoldU is TAsymmetricManifold) or (ManifoldU is TAsymmetricManifold) then
             Result := not GetQnos(0,False,3)
           else if (ManifoldU = nil) and (ManifoldL = nil) then
             WriteLn(Log, 'Must specify manifold or state')
           else
             raise NotYet; { Hund's case (a) }
       23: Result := not GetQnos(0,False,3);
       54: Result := not GetQnos(1,True,3);
       { 63: Documentation suggests 63 should have 6 quantum numbers, but NH2 sample has only 5 }
       43: Result := not GetQnos(0,True,3);
  else
    WriteLn(Log, 'Quantum number format ', Q, ' not implemented');
  end;
end;
{ In test suite:
Sym Top:   2y 13y 22  33  42 52
Linear:    1y  8  19  12y 21 32 41 52 72 10001y 10002
Asym tops: 3y 14y 23y 43y 54
}
end.
