unit SpectrumPlot;
{$I utils.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses GStyle, Gmarks, GraphI, GraphM, GraphRange,
     Units, Graphics, MNode, Mixture, Classes, Life, Expr, MNodeExpr,
     SysUtils, QuantumNumber, MathSup, QuantumMath;

type

  { TAbstractPlot }

  TAbstractPlot = class(TSpectrum)
  private
    FFOffset: Double;
    FRefWidth: Double;
    FLifeModel: TLifeModel;
    FIntensityUnits: TIntensityUnits;
  protected
    procedure AdjustStrengthAndWidth(var Width, Strength: Double);
  public
    IMin, IMax: Double; { Vertical limits used in plotting }
    procedure GetColoursUsed(var Colours: TColours); virtual;
    procedure InitializeFromSimulation(Simulation: TSimulation; ZeroWidth: Boolean = False; PrepareLineshape: Boolean = True); virtual;
    function FindLimits(Simulation:TSimulation):Boolean; virtual; { Return True if not Flat }
    procedure ReadOrWriteLimits(Simulation:TSimulation);
    procedure MakeFromSimulation(Simulation: TSimulation; MakeLineList: Boolean; out Problem: Exception); virtual;
    property FOffset: Double read FFOffset write FFOffset;
    property RefWidth: Double read FRefWidth write FRefWidth;
    property LifeModel: TLifeModel read FLifeModel write FLifeModel;
    property IntensityUnits: TIntensityUnits read FIntensityUnits write FIntensityUnits;
    procedure Show(G: TGraphM; PlotRange:TPlotRange; Offset: Boolean; ShowYLabel: Boolean; var PlotRanges:TPlotRanges; xShift: Double = 0); virtual; abstract;
  end;

  { TLineshape }

  TLineshape = class(TMNodeExprVars)
  private
    xVar, GauVar, LorVar, StickVar, WidthVar: TVariable;
  protected
    procedure SetUpVariables; override;
  public
    constructor Create(AOwner: TComponent); override;
  published
    property Width: Double index 1 read GetParameter write SetParameter;
  end;


  { TConvolutionParameters }

  TConvolutionParameters = record
    Gaussian: Double;
    Lorentzian: Double;
    WidthMult: Integer;
    SMargin: Double;
    Doppler: Double;
    LineShape: TLineshape;
    procedure SetFromSimulation(ASimulation: TSimulation);
    procedure Zero;
    function LineshapeValue(Offset, ExtraLor: Double; Stick: Boolean):Double;
    procedure CheckSensible;
    function IsStickSpectrum(df: Double):Boolean;
    function SumWidths: Double;
  end;

  { TSpectrumPlot }

  TSpectrumPlot = class(TAbstractPlot)
  private
    FIntensityInRange: Double;
    FnLinesInRange: Integer;
    FnDF: Integer; { Number of points spectrum required at }
    WorknDF: Integer; { Total number of points spectrum will be calculated at }
    FnExtra: Integer; { Number of additional points calculated at either side }
    LineShape: TDoubleArray;
    FConvoluted: Boolean;
    FNoConvolution, FStickSpectrum: Boolean;
    FPlot: TDoubleArray; { The Spectrum }
    function GetPlot(AIndex: Integer): Double;
  protected
    OThreshold: Double;
    procedure SetnDF(const Value: Integer); virtual;
  public
    NoBaseLine: Boolean; { If a zero width spectrum is shown with a baseline }
    ConvolutionParameters:TConvolutionParameters;
    function Another(Simulation: TSimulation):TSpectrumPlot; virtual;
    procedure AddInPlot(ThePlot: TSpectrumPlot); virtual;
    procedure InitializeFromSimulation(Simulation: TSimulation; ZeroWidth: Boolean = False; PrepareLineshape: Boolean = True); override;
    property nDF: Integer read FnDF write SetnDF; { Number of points to simulate spectrum at }
    property Convoluted: Boolean read FConvoluted;
    procedure AddLine(CentreFrequency, AStrength, AWidth: Double; AColour: TColor; ATransitionType: TTransitionType; AUnits: TResonanceUnits; LineshapeFunction: TLineShapeFunction); overload; override;
    procedure AddLine(const Line: TLine; Simulation: TSimulation; AUnits: TResonanceUnits; var Problem: Exception; CustomLineshapes: Boolean); overload; override;
    procedure AddIn(LineList: TLineList; Simulation: TSimulation; var Problem: Exception; FirstLine: Integer = 0; Increment: Integer = -1); override;
    procedure Clear; override;
    procedure Convolute;
    function FindLimits(Simulation:TSimulation):Boolean; override;
    procedure MakeLineShape(const AConvolutionParameters:TConvolutionParameters);
    function ConsistentUnits: Boolean;
    function IntegratedUints: Boolean;
    procedure Show(G: TGraphM; PlotRange:TPlotRange; Offset: Boolean; ShowYLabel: Boolean; var PlotRanges:TPlotRanges; xShift: Double = 0); override;
    property nLinesInRange: Integer read FnLinesInRange;
    property IntensityInRange: Double read FIntensityInRange;
    property Plot[AIndex: Integer]:Double read GetPlot;
    property NoConvolution: Boolean read FNoConvolution;
    property StickSpectrum: Boolean read FStickSpectrum;
  end;

  { TMultiColouredSpectrumPlot }

  TMultiColouredSpectrumPlot = class(TSpectrumPlot)
  private
    FSimulation: TSimulation;
    Plots: array of TSpectrumPlot;
    function PlotForColour(AColour: TColor; UsePlot: TSpectrumPlot = nil): TSpectrumPlot;
  public
    constructor Create(ASimulation: TSimulation);
    function Another(Simulation: TSimulation):TSpectrumPlot; override;
    procedure AddInPlot(ThePlot: TSpectrumPlot); override;
    destructor Destroy; override;
    procedure Clear; override;
    procedure AddLine(AFrequency, AStrength, AWidth: Double; AColour: TColor; ATransitionType: TTransitionType; AUnits: TResonanceUnits; LineshapeFunction: TLineShapeFunction); override;
    function FindLimits(Simulation:TSimulation):Boolean; override;
    procedure GetColoursUsed(var Colours: TColours); override;
    procedure Show(G: TGraphM; PlotRange:TPlotRange; Offset: Boolean; ShowYLabel: Boolean; var PlotRanges:TPlotRanges; xShift: Double = 0); override;
  end;

  { TFortratPlot }

  TFortratPlot = class(TAbstractPlot)
  private
    JMin, JMax: Integer;
    Stream: TMemoryStream;
    MaxIntensity: Double;
    MaxSym: Integer;
    MaxDeltaJ: Integer;
    MaxStateNo: Smallint;
    MoleculeType: TMoleculeClass; { If only one molecue, use that for symmetry labels }
    SampleMolecule: TMolecule;
    ColoursUsed: TColours;
    GraphStyle: TGraphStyle;
    PlotOffset: array of Double;
  public
    MinI: Double;
    ShowSymmetry: Boolean;
    ShowDeltaJ: Boolean;
    ScaleMarkSize: Boolean;
    UseUpper: Boolean;
    UseSymmetry: Boolean;
    UseStateNumber: Boolean;
    PlotQno: TQnoType;
    constructor Create(AUnits: TResonanceUnits);
    destructor Destroy; override;
    procedure Clear; override;
    procedure AddLine(const Line: TLine; Simulation: TSimulation; AUnits: TResonanceUnits; var Problem: Exception; CustomLineshapes: Boolean); override;
    procedure GetColoursUsed(var Colours: TColours); override;
    procedure InitializeFromSimulation(Simulation: TSimulation; ZeroWidth: Boolean = False; PrepareLineshape: Boolean = True); override;
    procedure MakeFromSimulation(Simulation: TSimulation; MakeLineList: Boolean; out Problem: Exception); override;
    {$IFNDEF CONSOLE}
    procedure GetPlotY(Line: TLine; out PlotNumber: Integer; out Y: Double);
    {$ENDIF}
    procedure Show(G: TGraphM; PlotRange:TPlotRange; Offset: Boolean; ShowYLabel: Boolean; var PlotRanges:TPlotRanges; xShift: Double = 0); override;
  end;

  { TFitPlot }

  TFitPlot = class(TSpectrumPlot)
  private
    FnDFLockout: Boolean;
  protected
    procedure SetnDF(const Value: Integer); override;
  public
    MaxIntensity: Double;
    Frequency: TDoubleArray; { The required frequencies }
    function Another(Simulation: TSimulation):TSpectrumPlot; override;
    procedure Clear; override;
    procedure AddLine(AFrequency, AStrength, AWidth: Double; AColour: TColor; ATransitionType: TTransitionType; AUnits: TResonanceUnits; LineShapeFunction: TLineShapeFunction); override;
    procedure InitializeFromSimulation(Simulation: TSimulation; ZeroWidth: Boolean = False; PrepareLineshape: Boolean = True); override;
    procedure AddIn(LineList: TLineList; Simulation: TSimulation; var Problem: Exception; FirstLine: Integer = 0; Increment: Integer = -1); override;
  end;


implementation

uses {$IFNDEF CONSOLE} ClipBrd, Huge, {$IFDEF MSWINDOWS} Windows, {$ENDIF} {$ENDIF}
     PopFunc, Progress, TextStrm, Math, Wofz, MNodeColour, Worker, Sysinfo, PGWork, UCompat;

{ TAbstractPlot }

function TAbstractPlot.FindLimits(Simulation:TSimulation):Boolean;
begin
  { Would be abstract, only no action might well be OK }
  Result := False;
  ReadOrWriteLimits(Simulation);
end;

procedure TAbstractPlot.ReadOrWriteLimits(Simulation: TSimulation);
begin
  if Assigned(Simulation) then begin
    if Simulation.AutoMin then
      Simulation.ModifyParameter(Simulation.IndexOfParameter('Ymin'),IMin,False,False)
    else
      IMin := Simulation.Ymin;
    if Simulation.AutoMax then
      Simulation.ModifyParameter(Simulation.IndexOfParameter('Ymax'),IMax,False,False)
    else
      IMax := Simulation.Ymax;
  end;
end;

procedure TAbstractPlot.GetColoursUsed(var Colours: TColours);
begin
  Colours := nil;
end;

procedure TAbstractPlot.InitializeFromSimulation(Simulation: TSimulation; ZeroWidth: Boolean; PrepareLineshape: Boolean);
begin
  FOffset := Simulation.FOffset;
  FMin := Simulation.FMin;
  FMax := Simulation.FMax;
  Temperature := Simulation.Temperature;
  RefWidth := Simulation.RefWidth;
  LifeModel := Simulation.LifeModel;
  Units := Simulation.PlotUnits;
  FIntensityUnits := Simulation.IntensityUnits;
  Clear;
end;

procedure TAbstractPlot.MakeFromSimulation(Simulation: TSimulation; MakeLineList: Boolean; out Problem: Exception);
begin
  PushProgress('Simulating Spectrum');
  try
    Problem := nil;
    if not Assigned(Simulation.Mixture.LineList) and MakeLineList then
      Simulation.MakeLineList(Problem);
    InitializeFromSimulation(Simulation);
    if not Assigned(Problem) then
      if Assigned(Simulation.Mixture.LineList) then begin
        // Not useful: Simulation.Mixture.LineList.SortOrder := lsoPosition;
        // Even with a very large linelist sorting is much slower than a simple scan
        AddIn(Simulation.Mixture.LineList, Simulation, Problem);
      end else
        Simulation.Mixture.AddToSpectrum(Self, Problem, Simulation);
  finally
    PopProgress;
  end;
end;

procedure TAbstractPlot.AdjustStrengthAndWidth(var Width, Strength: Double);
begin
  case LifeModel of
    lmProduct, lmProductWidth:
      { Scale Peak intensity as width/(width + RefWidth) }
      if RefWidth > 0 then
        Strength := Strength * Width / (Width + RefWidth)
      else
        Strength := Strength * Width;
    lmParent, lmParentWidth:
      { Scale Peak intensity as 1/(width + RefWidth) }
      if Width + RefWidth > 0 then
        Strength := Strength / ( Width + RefWidth);
    lmGate:
      { Scale Peak intensity as exp(-width * RefWidth) }
      Strength := Strength * Exp(-Width*RefWidth);
  end;
  if LifeModel in [lmNone, lmProduct, lmParent, lmGate] then
    Width := 0; { Discard width if required }
end;

{ TSpectrumPlot }

procedure TSpectrumPlot.AddLine(CentreFrequency, AStrength, AWidth: Double; AColour: TColor; ATransitionType: TTransitionType; AUnits: TResonanceUnits; LineshapeFunction: TLineShapeFunction);
var
  Centre, nl, i, Lines, nParts: Integer;
  df, a, w1, Offset, AFrequency: Double;
begin
  if (AStrength = 0) or (AColour = clHide) then Exit;
  if LineshapeFunction <> nil then
    with LineshapeFunction as TCustomLineshapeFunction do begin
      StrengthVariable.Value := AStrength;
      FrequencyVariable.Value := AFrequency;
    end;
  AdjustStrengthAndWidth(AWidth, AStrength);
  AWidth := ConvertWidth(AWidth, CentreFrequency, AUnits, Units);
  AFrequency := Convert(CentreFrequency, AUnits, Units);
  df := (FMax-FMin)/(nDF-1);
  { Add each line twice if Doppler split }
  if ConvolutionParameters.Doppler <> 0 then begin
    if IsEnergyUnit(Units) then
      Offset := AFrequency*ConvolutionParameters.Doppler
    else begin
      a := Convert(AFrequency,Units,cm1);
      Offset := ConvertWidth(a*ConvolutionParameters.Doppler,a,cm1,Units);
    end;
    AFrequency := AFrequency - Offset;
    nParts := 2;
    AStrength := AStrength / 2;
  end else begin
    Offset := 0;
    nParts := 1;
  end;
  for Lines := 1 to nParts do begin
    if Lines = 2 then
      AFrequency := AFrequency + 2*Offset;
    a := (AFrequency - FMin)/df + FnExtra;
    { Prevent overflows on conversion to integer }
    if Abs(a) < High(Integer) div 2 then
      Centre := Round(a)
    else
      Centre := High(Integer) div 2;
    { Note that line centre is moved onto grid }
    if AWidth > Abs(3*df) then begin
      { Add in Lorentzian to plot }
      nl := Round(ConvolutionParameters.WidthMult * 2 * AWidth / Abs(df) + 4);
      if (Centre >= -nl) and (Centre-nl < WorknDF) then begin
        FSomeWidthsSet := True;
        if LineShapeFunction <> nil then begin
          df := ConvertWidth(df, AFrequency, Units, AUnits);
          if Length(LineShape) = 1 then
            a := ConvertWidth(1, AFrequency, Units, AUnits) / LineShape[0]
          else
            a := Abs(df);
//WriteLn(Log, df, ' ', ConvertWidth(df, AFrequency, Units, AUnits));
//WriteLn(Log,AFrequency-((Centre - FnExtra)*df+FMin), ' ', df);
//AFrequency := (Centre - FnExtra)*df+FMin;
//CentreFrequency := Convert(AFrequency, Units, AUnits);
//WriteLn(Log, 'c', Length(LineShape):6, ' ', CentreFrequency:20:6, ' ', AStrength, ' ', AWidth, ' ', a*LineShapeFunction.LineshapeValue(0));
          for i := Max(Centre-nl,0) to Min(Centre+nl,WorknDF-1) do
            FPlot[i] := FPlot[i] + a*LineShapeFunction.LineshapeValue((i-Centre)*df);
        end else begin
          a := AStrength * AWidth / 2 / Pi / Abs(df);
          if Length(LineShape) = 1 then
            a := a / Abs(df) / LineShape[0]; { Full convolution done here }
          w1 := Sqr(AWidth) / 4 / df / df;
//WriteLn(Log, 'n', Length(LineShape):6, ' ', CentreFrequency:20:6, ' ', AStrength, ' ', AWidth, ' ', a/w1);
          for i := Max(Centre-nl,0) to Min(Centre+nl,WorknDF-1) do
            FPlot[i] := FPlot[i] + a/(Sqr(i-Centre) + w1);
        end;
      end else
        Continue;
    end else if (Centre >= 0) and (Centre < WorknDF) then begin
      if (AWidth > 0) and (Length(LineShape) = 1) then begin
        { Stick spectrum - make intensities consistent by taking intensity at centre of peak }
        { Do full convolution here to take account of case where both widths non-zero,
          but still a stick spectrum }
        //AStrength := AStrength * Voigt(0,AWidth+ConvolutionParameters.Lorentzian, ConvolutionParameters.Gaussian) / LineShape[0];
        if LineshapeFunction <> nil then begin
          AStrength := LineshapeFunction.LineshapeValue(0) * ConvertWidth(1, AFrequency, Units, AUnits);
          { Estimate additional intensity lost at the line centre from ratio of linewidths }
          AStrength := AStrength * AWidth / (ConvolutionParameters.SumWidths + AWidth);
        end else begin
          { Two possibilities }
          { 1. Evaluate Lorentzian at line centre only }
          {    Estimate additional intensity lost at the line centre from ratio of linewidths }
            //AStrength := Two * AStrength / Pi / AWidth;
            //AStrength := AStrength * AWidth / (ConvolutionParameters.SumWidths + AWidth);
          { 2. Compute complete function at line centre, and divide by LineShape[0] to reverse effects of subsequent convolution }
          { More exact, but results will be different if a LineshapeFunction is used }
            AStrength := AStrength * ConvolutionParameters.LineshapeValue(0,AWidth,True);
        end;
        { Divide by LineShape[0] to reverse effects of subsequent convolution }
        AStrength := AStrength / LineShape[0] ;
      end else
        { Will be convoluted with something broader, so ignore spread }
        { Approximate if widths similar }
        FSomeWidthsNotSet := True;
//WriteLn(Log, Length(LineShape):6, ' ', CentreFrequency:20:6, ' ', AStrength, ' ', AWidth);
      FPlot[Centre] := FPlot[Centre] + AStrength;
    end else
      Continue;
    Inc(FnLinesInRange);
    FIntensityInRange := FIntensityInRange + AStrength;
  end;
end;

procedure TSpectrumPlot.AddLine(const Line: TLine; Simulation: TSimulation; AUnits: TResonanceUnits; var Problem: Exception; CustomLineshapes: Boolean);
var
  Centre, df: Double;
  LineShapeFunction, LS2: TLineShapeFunction;
  State: TState;
begin
  { Speed up only - throw away out of range lines quickly, without calculating intensity }
  if Line.Width = 0 then begin
    Centre := Convert(Line.Frequency, AUnits, Units);
    df := (FMax-FMin)/(nDF-1);
    Centre := (Centre - FMin)/df + FnExtra;
    if (Centre < 0) or (Centre >= WorknDF) then
      Exit;
  end else if Line.Width < 0 then begin
    if not Assigned(Problem) then
      Problem := EWarningException.Create('Lines with negative width discarded');
    Exit;
  end;
  LineShapeFunction := nil;
  if CustomLineshapes then begin
    if Line.EigenvaluesU <> nil then begin
      LineShapeFunction := Line.EigenvaluesU.DominantBasis[Line.IndexU].State.ActiveLineshapeFunction;
      if LineShapeFunction <> nil then
        LineShapeFunction.PrepareLineshape(Line.EigenvaluesU,Line.IndexU);
    end;
    if Line.EigenvaluesL <> nil then begin
      LS2 := Line.EigenvaluesL.DominantBasis[Line.IndexL].State.ActiveLineshapeFunction;
      if LS2 <> nil then
        LS2.PrepareLineshape(Line.EigenvaluesL,Line.IndexL);
    end;
    if LS2 <> nil then
      if LineShapeFunction <> nil then begin
        if not Assigned(Problem) then
          Problem := EWarningException.Create('Custom lineshapes for both upper and lower states');
        Exit;
      end else
        LineShapeFunction := LS2;
  end;
  AddLine(Line.Frequency, Simulation.LineIntensity(Line, AUnits, Problem), Line.Width, Line.Colour, Line.TransitionType, AUnits, LineShapeFunction);
end;

function TSpectrumPlot.Another(Simulation: TSimulation):TSpectrumPlot;
begin
  Result := TSpectrumPlot.Create(Units, IsDistribution);
  Result.InitializeFromSimulation(Simulation);
  Result.nDF := nDF;
  Result.Clear;
end;

procedure TSpectrumPlot.AddInPlot(ThePlot: TSpectrumPlot);
var
  i: Integer;
begin
  for i := 0 to High(FPlot) do
    FPlot[i] := FPlot[i] + ThePlot.FPlot[i];
  FnLinesInRange := FnLinesInRange +ThePlot.nLinesInRange;
  FIntensityInRange := FIntensityInRange + ThePlot.IntensityInRange;
  FSomeWidthsSet := FSomeWidthsSet or ThePlot.FSomeWidthsSet;
  FSomeWidthsNotSet := FSomeWidthsNotSet or ThePlot.FSomeWidthsNotSet;
end;

procedure TSpectrumPlot.AddIn(LineList: TLineList; Simulation: TSimulation;
  var Problem: Exception; FirstLine: Integer; Increment: Integer);
var
  I: Integer;
  Worker: TPgopherWorker;
  nPlots: Integer;
  Plots: array of TSpectrumPlot;
  SubProblem: Exception;
begin
  nPlots := Min(nWorkers, nCPUs);
  if (nPlots <= 1) or (Increment > 1) then
    inherited
  else begin
    { Parallel algorithm }
    { Heuristic - much more than two threads doesn't seem to help }
    //if not LineList.SomeWidthsSet then
    //  nPlots := Min(nPlots,Max(nCPUs div 4,2));
    if IsNormalized(Simulation.IntensityUnits) then
      Simulation.CalculatePartitionFunctions(Problem);
    SetLength(Plots, nPlots);
    Plots[0] := Self;
    SubProblem := nil;
    try
      for I := 1 to nPlots - 1 do begin
        Plots[I] := Another(Simulation);
        Worker := FindWorkerWait as TPgopherWorker;
        if Assigned(Worker) then
          Worker.AddIn(Plots[I], LineList, Simulation, SubProblem, I, nPlots)
        else begin
          { Something has failed }
          if Assigned(Worker) then
            Worker.Skip;
          Worker := nil; { Flag error }
          Break;
        end;
      end;
      { Use main thread for one list, so that progress bar can be used }
      try
        if Assigned(Worker) then
          inherited AddIn(LineList, Simulation, Problem, 0, nPlots)
        else
          AbortCalculation;
      finally
        WaitNoBusyWorkers;
      end;
      for I := 1 to nPlots - 1 do
        AddInPlot(Plots[I]);
    finally
      { Combine exception flag }
      if not Assigned(Problem) then
        Problem := SubProblem
      else if Assigned(SubProblem) then
        FreeAndNil(SubProblem);
      for I := 1 to nPlots - 1 do
        FreeAndNil(Plots[I]);
    end;
  end;
end;

procedure TSpectrumPlot.Clear;
begin
  inherited;
  FConvoluted := False;
  if Length(FPlot) > 0 then
    FillChar(FPlot[0], Length(FPlot)*SizeOf(Double), 0);
  FnLinesInRange := 0;
  FIntensityInRange := 0;
end;

procedure TSpectrumPlot.Convolute;
var
  x: TDoubleArray;
  i, k, nLw: Integer;
begin
  if Convoluted or (Length(FPlot) = 0) then Exit;
  nLw := (Length(LineShape)-1) div 2;
  SetLength(x, Length(FPlot));
  FillChar(x[0], Length(FPlot)*SizeOf(Double), 0);
  for i := 0 to High(FPlot) do
    if FPlot[i] <> 0 then
      for k := Max(i-nLw,0) to Min(i+nLw, High(FPlot)) do
        x[k] := x[k] + FPlot[i]*LineShape[nLW+k-i];
  for i := 0 to High(FPlot) do
    FPlot[i] := x[i];
  FConvoluted := True;
end;

function TSpectrumPlot.FindLimits(Simulation:TSimulation):Boolean;
begin
  Convolute;
  Limits(FPlot, IMin, IMax, FnExtra, FnExtra + nDf - 1);
  Result := IMax <> IMin;
  if not Result then
    IMax := IMin + 1; { Flat spectrum - quick fix to avoid divide by zero }
  inherited FindLimits(Simulation); { Override limits if necessary }
end;

function TSpectrumPlot.GetPlot(AIndex: Integer): Double;
begin
  Result := FPlot[AIndex+FnExtra];
end;

procedure TSpectrumPlot.InitializeFromSimulation(Simulation: TSimulation; ZeroWidth: Boolean; PrepareLineshape: Boolean);
begin
  inherited;
  OThreshold := Simulation.OThreshold;
  ConvolutionParameters.SetFromSimulation(Simulation);
  if ZeroWidth then
    ConvolutionParameters.Zero;
  if PrepareLineshape then
    SetnDF(Simulation.nDF);
end;

procedure TSpectrumPlot.MakeLineShape(const AConvolutionParameters:TConvolutionParameters);
var
  n, i: Integer;
  df: Double;
begin
  AConvolutionParameters.CheckSensible;
  df := (FMax-FMin)/(nDF-1);
  FNoConvolution := False;
  FStickSpectrum := AConvolutionParameters.IsStickSpectrum(df);
  if StickSpectrum then begin
    SetLength(LineShape, 1);
    { Makes transition between stick and full lineshape consistent by }
    if AConvolutionParameters.SumWidths > 0 then
      //LineShape[0] := Voigt(0, AConvolutionParameters.Lorentzian, AConvolutionParameters.Gaussian) { approximating stick by value at centre }
      LineShape[0] := ConvolutionParameters.LineshapeValue(0,0,True) { approximating stick by value at centre }
    else begin
      LineShape[0] := 1;
      FNoConvolution := True;
    end;
  end else begin
    n := Round(AConvolutionParameters.WidthMult*AConvolutionParameters.SumWidths/ABS(df)+4);
    if n > nDF + FnExtra then
      n := nDF + FnExtra + 1; { Very large lineshape; however only nDF points either
                                side of lineshape will be used }
    SetLength(LineShape, 2*n+1);
    for i := 0 to n do begin
      //LineShape[i+n] := Voigt(i*df, AConvolutionParameters.Lorentzian, AConvolutionParameters.Gaussian);
      LineShape[i+n] := ConvolutionParameters.LineshapeValue(i*df,0,False);
      if Assigned(ConvolutionParameters.LineShape) then
        LineShape[n-i] := ConvolutionParameters.LineshapeValue(-i*df,0,False)
      else
        LineShape[n-i] := LineShape[i+n]; { Optimization - built in lineshape symmetrical }
    end;
    // FSomeWidthsSet := True;
  end;
end;

function TSpectrumPlot.ConsistentUnits: Boolean;
begin
  Result := not (FNoConvolution and FSomeWidthsSet and FSomeWidthsNotSet);
end;

function TSpectrumPlot.IntegratedUints: Boolean;
begin
  Result := FNoConvolution and FSomeWidthsNotSet;
end;

procedure TSpectrumPlot.SetnDF(const Value: Integer);
var
  df, dp: Double;
begin
  if Value > 1 then begin
    FnDF := Value;
    df := (FMax-FMin)/(nDF-1);
    if ConvolutionParameters.SMargin >= 0 then
      FnExtra := Round(ConvolutionParameters.SMargin)
    else begin
      dp := Max(Convert(FMin,Units,cm1),Convert(FMin,Units,cm1));
      dp := ConvertWidth(dp*ConvolutionParameters.Doppler,dp,cm1,Units);
      FnExtra := -Round(ConvolutionParameters.WidthMult*(ConvolutionParameters.SumWidths+dp)/ABS(df)*ConvolutionParameters.SMargin);
    end;
    if FnExtra > 0 then
      FnExtra := Max(FnExtra, 5);
    { TODO: Consider switching algorithms if FnExtra much larger than nDF }
    WorknDF := nDF+2*FnExtra;
    SetLength(FPlot, WorknDF);
    { This is called here to allow stick spectrum corrections to work properly }
    MakeLineShape(ConvolutionParameters);
  end else
    raise EReportException.Create('nDF must be > 1');
end;

procedure TSpectrumPlot.Show(G: TGraphM; PlotRange:TPlotRange; Offset: Boolean; ShowYLabel: Boolean; var PlotRanges:TPlotRanges; xShift: Double);
var
  i: Integer;
  df, x: Double;
  xl, xu, yl, yu: Double;
  AutoScale: Boolean;
begin
  if Colour = clHide then Exit;
  Convolute;
  if Colour <> clNone then
    G.StyleP(Colour, G.LineStyle.Width, G.LineStyle.Pattern);
  G.MapArg(xl,xu,yl,yu);
  AutoScale := IsNullPlotRange(PlotRange);
  if AutoScale then begin
    FindLimits(nil);
    PlotRange := GraphRange.PlotRange(G, 0, 1, Imin, Imax, -9);
  end;
  MapFromPlotRange(G,PlotRange);
  df := (FMax-FMin)/(nDF-1);
  x := FMin-Foffset-xShift;
  if not StickSpectrum or SomeWidthsSet then begin
    G.ClipPoint(x, FPlot[FnExtra]);
    for i := FnExtra+1 to FnExtra+nDF-1 do begin
      x := x + df;
      G.ClipJoin(x, FPlot[i]);
    end;
  end else begin
    { Special case for stick spectra - make sure lines are vertical and single }
    if not NoBaseLine then
     G.ClipPoint(x, 0);
    for i := FnExtra to FnExtra+nDF-1 do begin
      if FPlot[i] > 0 then
        if NoBaseline then begin
          G.ClipPoint(x, 0);
          G.ClipJoin(x, FPlot[i]);
        end else begin
          G.ClipJoin(x, 0);
          G.ClipJoin(x, FPlot[i]);
          G.ClipJoin(x, 0);
        end;
      x := x + df;
    end;
    if not NoBaseLine then
      G.ClipJoin(x-df, 0);
  end;
  G.Map(xl,xu,yl,yu);
  if ShowYLabel and ConsistentUnits then
    AxisFromPlotRange(PlotRange, G, G.Xmin, NameOfUnit(IntensityUnits, IntegratedUints, Units));
  AddPlotRange(PlotRanges, PlotRange);
end;

{ TFitPlot }

procedure TFitPlot.AddIn(LineList: TLineList; Simulation: TSimulation; var Problem: Exception; FirstLine: Integer; Increment: Integer);
var
  I: Integer;
  Worker: TPgopherWorker;
  nPlots: Integer;
  Plots: array of TFitPlot;
  j: Integer;
  SubProblem: Exception;
begin
  nPlots := Min(nWorkers, nCPUs);
  if (nPlots <= 1) or (Increment > 1) then
    inherited
  else begin
    { Parallel algorithm }
    if IsNormalized(Simulation.IntensityUnits) then
      Simulation.CalculatePartitionFunctions(Problem);
    SetLength(Plots, nPlots);
    Plots[0] := Self;
    SubProblem := nil;
    try
      for I := 1 to nPlots - 1 do begin
        Plots[I] := TFitPlot.Create(Units, IsDistribution);
        Plots[I].InitializeFromSimulation(Simulation);
        Plots[I].Frequency := Frequency;
        Plots[I].nDF := nDF;
        Plots[I].Clear;
        Worker := FindWorkerWait as TPgopherWorker;
        if Assigned(Worker) then
          Worker.AddIn(Plots[I], LineList, Simulation, SubProblem, I, nPlots)
        else begin
          { Something has failed }
          if Assigned(Worker) then
            Worker.Skip;
          Worker := nil; { Flag error }
          Break;
        end;
      end;
      { Use main thread for one list, so that progress bar can be used }
      try
        if Assigned(Worker) then
          inherited AddIn(LineList, Simulation, Problem, 0, nPlots)
        else
          AbortCalculation;
      finally
        WaitNoBusyWorkers;
      end;
      for I := 1 to nPlots - 1 do
        for j := 0 to High(FPlot) do
          FPlot[j] := FPlot[j] + Plots[I].FPlot[j];
    finally
      { Combine exception flag }
      if not Assigned(Problem) then
        Problem := SubProblem
      else if Assigned(SubProblem) then
        FreeAndNil(SubProblem);
      for I := 1 to nPlots - 1 do
        FreeAndNil(Plots[I]);
    end;
  end;
end;

procedure TFitPlot.AddLine(AFrequency, AStrength, AWidth: Double; AColour: TColor; ATransitionType: TTransitionType; AUnits: TResonanceUnits; LineShapeFunction: TLineShapeFunction);
var
  i: Integer;
  x: Double; { Smallest offset from peak centre }
  y: Double; { Correspondig height }
  Offset: Double; { For Doppler splitting convolution }
  Line, nParts: Integer;
begin
  if AColour = clHide then
    Exit;
  AdjustStrengthAndWidth(AWidth, AStrength);
  AWidth := ConvertWidth(AWidth, AFrequency, AUnits, Units);
  AFrequency := Convert(AFrequency, AUnits, Units);
  if AStrength <= MaxIntensity then
    Exit;
  if LineShapeFunction <> nil then
    raise NotYet;
  if ConvolutionParameters.Doppler <> 0 then begin
    if IsEnergyUnit(Units) then
     Offset := AFrequency*ConvolutionParameters.Doppler
    else begin
      Offset := Convert(AFrequency,Units,cm1);
      Offset := ConvertWidth(Offset*ConvolutionParameters.Doppler,Offset,cm1,Units);
    end;
    AFrequency := AFrequency - Offset;
    AStrength := AStrength / 2;
    nParts := 2;
  end else begin
    Offset := 0;
    nParts := 1;
  end;
  for Line := 1 to nParts do begin
    if Line = 2 then
      AFrequency := AFrequency + 2*Offset;
    if AFrequency <= FMin then
      x := FMin - AFrequency
    else if AFrequency >= FMax then
      x := AFrequency - FMax
    else
      x := 0;
    //y := AStrength*Voigt(x, AWidth + ConvolutionParameters.Lorentzian, ConvolutionParameters.Gaussian);
    y := AStrength*ConvolutionParameters.LineshapeValue(x, AWidth, False);
    if IsNanOrInf(y) then
      raise EReportException.Create('Invalid linewidth or intensity for line at '+FloatToStr(AFrequency)+' - widths must not be zero for contour fitting');
    if y <= MaxIntensity then
      Continue;
    if y*OThreshold > MaxIntensity then
      MaxIntensity := y*OThreshold;
    { TODO: Lots of possible optimizations here }
    for i := 0 to High(FPlot) do
      //FPlot[i] := FPlot[i] + AStrength*Voigt(Frequency[i]-AFrequency, AWidth + ConvolutionParameters.Lorentzian, ConvolutionParameters.Gaussian);
      FPlot[i] := FPlot[i] + AStrength*ConvolutionParameters.LineshapeValue(Frequency[i]-AFrequency, AWidth, False);
  end;
end;

procedure TFitPlot.Clear;
begin
  inherited Clear;
  Limits(Frequency, Fmin, FMax);
  MaxIntensity := 0;
end;

procedure TFitPlot.InitializeFromSimulation(Simulation: TSimulation; ZeroWidth: Boolean; PrepareLineshape: Boolean);
begin
  FnDFLockout := True;
  try
    inherited;
  finally
    FnDFLockout := False;
  end;
end;

procedure TFitPlot.SetnDF(const Value: Integer);
begin
  if FnDFLockout then Exit;
  ConvolutionParameters.SMargin := 0;
  inherited;
  SetLength(Frequency, nDF);
end;

function TFitPlot.Another(Simulation: TSimulation): TSpectrumPlot;
begin
  Result := TFitPlot.Create(Units, IsDistribution);
  Result.InitializeFromSimulation(Simulation);
  (Result as TFitPlot).Frequency := Frequency;
  Result.nDF := nDF;
  Result.Clear;
end;

{ TMultiColouredSpectrumPlot }

function TMultiColouredSpectrumPlot.PlotForColour(AColour: TColor; UsePlot: TSpectrumPlot): TSpectrumPlot;
var
  i, j: Integer;
begin
  i := 0;
  while (i < Length(Plots)) and (Plots[i].Colour < AColour) do
    Inc(i);
  if (i > High(Plots)) or (Plots[i].Colour <> AColour) then begin
    { Make new plot, or insert given plot }
    SetLength(Plots, Length(Plots)+1);
    for j := High(Plots) downto i+1 do
      Plots[j] := Plots[j-1];
    if UsePlot = nil then begin
      Result := TSpectrumPlot.Create(FSimulation.PlotUnits, IsDistribution);
      Result.InitializeFromSimulation(FSimulation);
      Result.Colour := AColour;
    end else
      Result := UsePlot;
    Plots[i] := Result;
  end else
    Result := Plots[i];
end;

procedure TMultiColouredSpectrumPlot.AddLine(AFrequency, AStrength, AWidth: Double; AColour: TColor; ATransitionType: TTransitionType; AUnits: TResonanceUnits; LineshapeFunction: TLineShapeFunction);
var
  CurrentPlot: TSpectrumPlot;
begin
  if AColour = clHide then
    Exit;
  inherited; { Always calculate sum, even if the sum is not displayed so that scaling is consistent }
  CurrentPlot := PlotForColour(AColour, nil);
  CurrentPlot.AddLine(AFrequency, AStrength, AWidth, AColour, ATransitionType, AUnits, LineshapeFunction);
end;

procedure TMultiColouredSpectrumPlot.Clear;
var
  i: Integer;
begin
  inherited;
  for i := 0 to High(Plots) do
    Plots[i].Free;
  SetLength(Plots,0);
end;

procedure TMultiColouredSpectrumPlot.GetColoursUsed(var Colours: TColours);
var
  i: Integer;
begin
  SetLength(Colours, Length(Plots));
  for i := 0 to High(Plots) do
    Colours[i] := Plots[i].Colour;
end;

constructor TMultiColouredSpectrumPlot.Create(ASimulation: TSimulation);
begin
  inherited Create(ASimulation.PlotUnits, IsPopDist(ASimulation.IntensityUnits));
  FSimulation := ASimulation;
end;

function TMultiColouredSpectrumPlot.Another(Simulation: TSimulation): TSpectrumPlot;
begin
  Result := TMultiColouredSpectrumPlot.Create(Simulation);
  Result.InitializeFromSimulation(Simulation);
  Result.nDF := nDF;
  Result.Clear;
end;

procedure TMultiColouredSpectrumPlot.AddInPlot(ThePlot: TSpectrumPlot);
var
  MC: TMultiColouredSpectrumPlot;
  CurrentPlot: TSpectrumPlot;
  i: Integer;
begin
  inherited AddInPlot(ThePlot); { Adds main plot }
  MC := ThePlot as TMultiColouredSpectrumPlot;
  for i := 0 to High(MC.Plots) do begin
    CurrentPlot := PlotForColour(MC.Plots[i].Colour,MC.Plots[i]);
    if CurrentPlot = MC.Plots[i] then
      MC.Plots[i] := nil { Ownership transferred }
    else
      CurrentPlot.AddInPlot(MC.Plots[i]);
  end;
end;

destructor TMultiColouredSpectrumPlot.Destroy;
begin
  Clear;
  inherited;
end;

function TMultiColouredSpectrumPlot.FindLimits(Simulation:TSimulation): Boolean;
var
  i: Integer;
begin
  Result := inherited FindLimits(nil);
  if not Result then
    IMax := IMin;
  for i := 0 to High(Plots) do begin
    if Plots[i].FindLimits(nil) then
      IMax := Max(IMax, Plots[i].IMax);
    IMin := Min(IMin, Plots[i].IMin);
  end;
  Result := IMax <> IMin;
  If not Result then
    IMax := IMin + 1; { Flat spectrum - quick fix to avoid divide by zero }
  ReadOrWriteLimits(Simulation); { Override with final limits if necessary }
end;

procedure TMultiColouredSpectrumPlot.Show(G: TGraphM; PlotRange:TPlotRange;
            Offset: Boolean; ShowYLabel: Boolean; var PlotRanges:TPlotRanges; xShift: Double);
var
  i, nPlots: Integer;
  AutoScale: Boolean;
  function PlotRangeFor(nPlot: Integer; Tag:Integer):TPlotRange;
  begin
    if Offset then
      if AutoScale then
        Result := GraphRange.PlotRange(G, nPlot/nPlots, (nPlot+0.95)/nPlots, Imin, Imax, Tag)
      else
        Result := GraphRange.PlotRange(G, nPlot/nPlots, (nPlot+0.95)/nPlots, PlotRange.yi, PlotRange.yx, Tag)
    else if AutoScale then
      Result := GraphRange.PlotRange(G, 0, 1, Imin, Imax, Tag)
    else
      Result := PlotRange
  end;
begin
  AutoScale := IsNullPlotRange(PlotRange);
  if AutoScale then
    FindLimits(nil);
  { Calculate number of plots }
  nPlots := Length(Plots);
  if FSimulation.ShowSum then
    Inc(nPlots);
  if nPlots = 0 then Exit;
  for i := 0 to High(Plots) do
    Plots[i].Show(G, PlotRangeFor(i,-10-i), False, ShowYLabel and (Offset or (nPlots = 1) or ConsistentUnits), PlotRanges, xShift);
  if FSimulation.ShowSum then begin
    G.Style(0);
    inherited Show(G, PlotRangeFor(nPlots-1,-9), Offset, ShowYLabel, PlotRanges, xShift);
  end;
end;

{ TFortratPlot }

type
  TFortratRec = packed record
    Frequency: Double;
    Strength:  Double;
    Colour:    TColor;
    J:         Smallint;
    StateNo:   Smallint;
    DeltaJ:    Shortint;
    Sym:       Shortint;
  end;

procedure TFortratPlot.AddLine(const Line: TLine; Simulation: TSimulation; AUnits: TResonanceUnits; var Problem: Exception; CustomLineshapes: Boolean);
var
  c: Double;
  FortratRec: TFortratRec;
  Width, f: Double;
  i: Integer;
  Molecule: TMolecule;
  procedure LoadRec(Ev: TEigenvalues; Index: Integer);
  begin
    Molecule := Ev.Manifold.Molecule;
    FortratRec.Sym := Molecule.JAdjustedSym(Ev.J, Ev.Sym);
    FortratRec.StateNo := Index;
    FortratRec.J := Ev.DominantBasis[Index].Qno(PlotQno);
  end;
begin
  if Line.Colour = clHide then
    Exit;
  FHasCustomLineshapeFunctions := FHasCustomLineshapeFunctions or CustomLineshapes;
  f := Convert(Line.Frequency, AUnits, Units);
  c := (f - FMin)/(FMax - FMin);
  if (c >= 0) and (c <= 1) and Assigned(Line.EigenvaluesL) then begin
    FortratRec.Frequency := f - FOffset;
    if UseUpper and Assigned(Line.EigenvaluesU) then
      LoadRec(Line.EigenvaluesU, Line.IndexU)
    else if Assigned(Line.EigenvaluesL) then
      LoadRec(Line.EigenvaluesL, Line.IndexL);
    if MoleculeType = nil then begin
      MoleculeType := TMoleculeClass(Molecule.ClassType);
      SampleMolecule := Molecule;
    end else if Assigned(SampleMolecule) and ((MoleculeType <> Molecule.ClassType) or (Molecule.LastSym <> SampleMolecule.LastSym)) then
      SampleMolecule := nil;
    if Assigned(Line.EigenvaluesU) and Assigned(Line.EigenvaluesL) then begin
      i := Line.EigenvaluesU.DominantBasis[Line.IndexU].Qno(PlotQno)
                           - Line.EigenvaluesL.DominantBasis[Line.IndexL].Qno(PlotQno);
      FortratRec.DeltaJ := Min(Max(i, Low(Shortint)), High(Shortint));
    end else
      FortratRec.DeltaJ := 0;
    if Abs(FortratRec.DeltaJ) > MaxDeltaJ then MaxDeltaJ := Abs(FortratRec.DeltaJ);
    if FortratRec.StateNo > MaxStateNo then MaxStateNo := FortratRec.StateNo;
    FortratRec.Colour := Line.Colour;
    if FortratRec.Colour = clNone then
      FortratRec.Colour := clBlack;
    if MaxSym < FortratRec.Sym then
      MaxSym := FortratRec.Sym;
    if not Assigned(Simulation) then
      Simulation := Line.EigenvaluesL.Manifold.Mixture.Simulations[0];
    FortratRec.Strength := Simulation.LineIntensity(Line, AUnits, Problem);
    Width := Line.Width;
    if Width > 0 then
      FSomeWidthsSet := True
    else
      FSomeWidthsNotSet := True;
    AdjustStrengthAndWidth(Width, FortratRec.Strength);
    // Not needed as only Strength required: Width := ConvertWidth(Width, Line.Frequency, Mixture.Units, Units);
    if FortratRec.Strength > MaxIntensity then
      MaxIntensity := FortratRec.Strength
    else if FortratRec.Strength < MinI * MaxIntensity then
      Exit; { Skip this line - note that some of the first weak lines will not be excluded, so a second test is used when plotting }
    Stream.Write(FortratRec, SizeOf(FortratRec));
    if (FnLines = 0) or (JMin > FortratRec.J) then begin
      JMin := FortratRec.J;
      IMin := FortratRec.J / 2.;
    end;
    if (FnLines = 0) or (FortratRec.J > JMax) then begin
      JMax := FortratRec.J;
      IMax := JMax / 2.;
    end;
    Inc(FnLines);
    for i := 0 to High(ColoursUsed) do
      if FortratRec.Colour = ColoursUsed[i] then
        Exit;
    SetLength(ColoursUsed, Length(ColoursUsed)+1);
    ColoursUsed[Length(ColoursUsed)-1] := FortratRec.Colour;
  end;
end;

procedure TFortratPlot.Clear;
begin
  inherited;
  IMin := 0;
  IMax := 0;
  JMax := -1;
  MaxIntensity := 0;
  MaxDeltaJ := 0;
  MaxSym := -1;
  MaxStateNo := 0;
  Stream.Free;
  Stream := TMemoryStream.Create;
  SetLength(ColoursUsed, 0);
  MoleculeType := nil;
  SampleMolecule := nil;
end;

constructor TFortratPlot.Create(AUnits: TResonanceUnits);
begin
  inherited Create(AUnits,False);
end;

destructor TFortratPlot.Destroy;
begin
  Clear;
  FreeAndNil(Stream);
  inherited;
end;

procedure TFortratPlot.GetColoursUsed(var Colours: TColours);
begin
  Colours := ColoursUsed;
end;

procedure TFortratPlot.InitializeFromSimulation(Simulation: TSimulation; ZeroWidth: Boolean; PrepareLineshape: Boolean);
begin
  inherited;
  MinI := Simulation.MinI;
  ShowSymmetry := Simulation.ShowSymmetry;
  ShowDeltaJ := Simulation.ShowDeltaJ;
  ScaleMarkSize := Simulation.ScaleMarkSize;
  UseUpper := Simulation.UseUpper;
  UseSymmetry := Simulation.UseSymmetry;
  UseStateNumber := Simulation.UseStateNumber;
  PlotQno := Simulation.FortratQno;
  GraphStyle := Simulation.GraphStyle;
end;

procedure TFortratPlot.MakeFromSimulation(Simulation: TSimulation; MakeLineList: Boolean; out Problem: Exception);
var
  FortratRec: TFortratRec;
  First: Boolean;
begin
  inherited MakeFromSimulation(Simulation, MakeLineList, Problem);
  if not Assigned(Problem) and (MinI > 0) and (nLines > 0) then begin
    { Second pass needed to determine plot range }
    Stream.Position := 0;
    First := True;
    while Stream.Read(FortratRec, SizeOf(FortratRec)) = SizeOf(FortratRec) do
      with FortratRec do
        if Strength >= MinI*MaxIntensity then
          if First then begin
            IMin := FortratRec.J;
            IMax := FortratRec.J;
            First := False;
          end else begin
            if FortratRec.J < IMin then
              IMin := FortratRec.J;
            if FortratRec.J > IMax then
              IMax := FortratRec.J;
          end;
    IMin := IMin / 2;
    IMax := IMax / 2;
  end;
end;

{$IFNDEF CONSOLE}
procedure TFortratPlot.GetPlotY(Line: TLine; out PlotNumber: Integer; out Y: Double);
var
  i: Integer;
  Offset: Boolean;
begin
  if (UseUpper and not Assigned(Line.EigenvaluesU))
     or (not UseUpper and not Assigned(Line.EigenvaluesL)) then
    Exit; { Eigenvalues not available }
  { Compute plot number }
  PlotNumber := 0;
  Offset := GraphStyle in [gsOffsetAll, gsInvertedAll];
  if Offset then
    for i := 0 to High(ColoursUsed) do
      if Line.Colour = ColoursUsed[i] then begin
        PlotNumber := i;
        Break;
      end;
  if UseSymmetry and (MaxSym >= 0) then
    PlotNumber := PlotNumber*(MaxSym+1) + IfThen(UseUpper, Line.EigenvaluesU.Sym, Line.EigenvaluesL.Sym);
  if UseStateNumber then
    PlotNumber := PlotNumber*MaxStateNo + IfThen(UseUpper, Line.IndexU, Line.IndexL) - 1;
  { Compute effective J }
  if UseUpper then
    Y := Line.EigenvaluesU.DominantBasis[Line.IndexU].Qno(PlotQno)
  else
    Y := Line.EigenvaluesL.DominantBasis[Line.IndexL].Qno(PlotQno);
  Y := (Y/2 - PlotOffset[0])/Length(PlotOffset) + PlotOffset[PlotNumber];
end;
{$ENDIF}

procedure TFortratPlot.Show(G: TGraphM; PlotRange:TPlotRange; Offset: Boolean; ShowYLabel: Boolean; var PlotRanges:TPlotRanges; xShift: Double);
var
  FortratRec: TFortratRec;
  MarkStyle: TMarkStyle;
  BaseSize: Single;
  i, m, iSym, MarkNumber, nPlots: Integer;
  o: Double;
  s: string;
  function SymLabel(Sym: Integer): string;
  begin
    if Assigned(SampleMolecule) then begin
      Result := SampleMolecule.JAdjustedSymName(0,Sym);
      if Result <> IntToStr(Sym) then
        Result := IntToStr(Sym) + '/' + Result;
    end else
      Result := IntToStr(Sym);
  end;
begin
  if nLines = 0 then
    Exit;
  if not IsNullPlotRange(PlotRange) then
    raise NotYet; { Actually not required }
  m := 0;
  if ShowSymmetry and (MaxSym >= 0) then begin
    s := 'Symmetry: ';
    for iSym := 0 to MaxSym do
      if ShowDeltaJ then
        for i := -MaxDeltaJ div 2 to MaxDeltaJ div 2 do begin
          s := s + #1 + Char(Ord(FirstMark)+m) + #0':' + SymLabel(iSym) + DeltaJName(i*2) + ' ';
          Inc(m);
        end
      else
        s := s + #1 + Char(Ord(FirstMark)+iSym) + #0':' + SymLabel(iSym) + ' ';
  end else if ShowDeltaJ then begin
    s := 'DeltaJ: ';
    for i := -MaxDeltaJ div 2 to MaxDeltaJ div 2 do begin
      s := s + #1 + Char(Ord(FirstMark)+m) + #0':' + DeltaJName(i*2) + ' ';
      Inc(m);
    end;
  end;
  G.Title2(s);
  Stream.Position := 0;
  MarkStyle := TMarkStyle.Create;
  MarkStyle.DotAtCentre := True;
  MarkStyle.Size := 2;
  BaseSize := G.MarkStyles[0].Size;
  nPlots := 1;
  if Offset then
    nPlots := Length(ColoursUsed);
  if UseSymmetry and (MaxSym >= 0) then
    nPlots := nPlots * (1+MaxSym);
  if UseStateNumber then
    nPlots := nPlots * MaxStateNo;
  SetLength(PlotOffset, nPlots);
  o := G.Ymin;
  for i := 0 to High(PlotOffset) do begin
    PlotOffset[i] := o;
    o := o + (G.Ymax-G.Ymin)/nPlots;
    G.ScaledAxis(IMin, G.Ymax, PlotOffset[i] + (IMin-G.Ymin)/nPlots, o, G.Xmin, False, QnoTypeNames[Ord(PlotQno)]);
    AddPlotRange(PlotRanges, GraphRange.PlotRange(G, PlotOffset[i] + (IMin-G.Ymin)/nPlots, o, IMin, G.Ymax, -10-i));
    { TODO: Separate Y axis labels here }
  end;
  if nPlots = 1 then
    PlotRanges[0].Tag := -9; { only one plot, so need not select on this }
  o := G.Ymin;
  while Stream.Read(FortratRec, SizeOf(FortratRec)) = SizeOf(FortratRec) do
    with FortratRec do
      if Strength >= MinI*MaxIntensity then begin
        MarkStyle.Colour := Colour;
        if ShowDeltaJ then begin
          MarkNumber := (DeltaJ + MaxDeltaJ) div 2;
          if ShowSymmetry then
            Inc(MarkNumber, Sym * (MaxDeltaJ + 1));
        end else if ShowSymmetry then
          MarkNumber := Max(Sym,0)
        else
          MarkNumber := 0;
        MarkStyle.MarkType := TMarkType(Ord(FirstMark) + MarkNumber mod (Ord(LastMark)-Ord(FirstMark)+1));
        if ScaleMarkSize and (MaxIntensity > 0) then
          MarkStyle.Size := 2 * BaseSize * Sqrt(Strength / MaxIntensity);
        if nPlots > 1 then begin
          { If the logic changes here, change GetPlotY to match }
          m := 0;
          if Offset then
            for i := 0 to High(ColoursUsed) do
              if Colour = ColoursUsed[i] then begin
                m := i;
                Break;
              end;
          if UseSymmetry and (MaxSym >= 0) then
            m := m*(MaxSym+1) + Sym;
          if UseStateNumber then
            m := m*MaxStateNo + FortratRec.StateNo - 1;
          o := PlotOffset[m];
        end;
        G.SymbolT(Frequency-xShift, (J/2.-G.Ymin)/nPlots+o, MarkStyle);
      end;
  MarkStyle.Free;
  { Currently only plot once, so free up space }
  FreeAndNil(Stream);
end;

{ TConvolutionParameters }

procedure TConvolutionParameters.SetFromSimulation(ASimulation: TSimulation);
begin
  SMargin := ASimulation.SMargin;
  Lorentzian := ASimulation.Lorentzian;
  Gaussian := ASimulation.Gaussian;
  WidthMult := ASimulation.WidthMult;
  Doppler := ASimulation.Doppler;
  LineShape := ActiveMnodeExpr(ASimulation, TLineshape) as TLineshape;
  if Assigned(LineShape) and not Assigned(LineShape.Expr) then
    LineShape := nil; { Invalid expression }
end;

function TConvolutionParameters.SumWidths: Double;
begin
  Result := Lorentzian + Gaussian;
  if Assigned(LineShape) then
    Result := Result + Abs(LineShape.Width);
end;

procedure TConvolutionParameters.Zero;
begin
  Lorentzian := 0;
  Gaussian := 0;
  Doppler := 0;
  LineShape := nil;
end;

function TConvolutionParameters.IsStickSpectrum(df: Double): Boolean;
begin
  { Is spectrum to be simulated as a stick spectrum (even though widths may be <> 0) }
  Result := SumWidths <= Abs(3*df);
end;

function TConvolutionParameters.LineshapeValue(Offset, ExtraLor: Double; Stick: Boolean): Double;
begin
  if Assigned(LineShape) then begin
    LineShape.xVar.Value := Offset;
    LineShape.LorVar.Value := Lorentzian+ExtraLor;
    LineShape.GauVar.Value := Gaussian;
    LineShape.StickVar.Value := Ord(Stick);
    LineShape.WidthVar.Value := LineShape.Width;
    Result := LineShape.Evaluate;
  end else
    Result := Voigt(Offset, Lorentzian+ExtraLor, Gaussian);
end;

procedure TConvolutionParameters.CheckSensible;
begin
  if (Lorentzian < 0) or (Gaussian < 0) then
    raise EReportException.Create('Negative linewidths');
end;

{ TLineshape }

constructor TLineshape.Create(AOwner: TComponent);
begin
  SavedProperties := SavedProperties + ',Width=1';
  DimensionlessParameters := DimensionlessParameters + ',Width';
  inherited;
end;

procedure TLineshape.SetUpVariables;
begin
  inherited SetUpVariables;
  V('x',xVar);
  V('Lor',LorVar);
  V('Gau',GauVar);
  V('Stick',StickVar);
  V('Width',WidthVar);
end;

initialization
  TLineshape.RegisterClass(TSimulation);
  TTable.RegisterClass(TLineshape);
end.
