unit SymT;
{ Symmetric Top }

{$DEFINE NOAT}{$I utils.inc}{Copyright (C) 2005-13 C M Western; see commandline.pas for license}
interface

uses Math, MathSup, Classes, SysUtils, MNode, Mixture, QuantumNumber, QuantumMath,
     MatrixElements, pgDOM;

const
  WTI = LMI+4; { Index of first statistical weight }
  FirstWidthPar = LSP+33;

type

  TPointGroup = (C3v=0, C4v, C5v, C6v, C7v, C8v, D3h, D4h, D5h, D6h, D7h, D8h, D2d, D3d, D4d, D5d, D6d, D7d, D8d,
                 C3h, C4h, C5h, C6h, S4, S6, S8, S10, S12, D3, D4, D5, D6, D7, D8);

  TCharacterTableEntry = record
    Name: String;
    g, rho, eta: Integer;
  end;

  TCharacterTable = record
    GroupName: string;
    Size: Integer;
    n: Integer;         { Highest order rotation axis }
    gKmult: Integer;    { multiplier of k in expression for g}
    rhoJmult: Integer;  { multiplier of J in expression for rho}
    rhoKmult: Integer;  { multiplier of k in expression for rho}
    etaKmult: Integer;  { multiplier of k in expression for eta}
    DipoleSym: Integer; { Symmetry of space fixed dipole moment component }
    HasSeparableDegeneracy: Boolean; { See C3h discussion below }
    LongestSymName: Integer; { Length of longest symmetry name }
    Names: array of PChar;
    IReps: array of TCharacterTableEntry;
    En: array of Integer; { 1 for E/E1, 2 for E2, ... }
    Group: array of Integer; { Statistical weight group number for each irep }
    GroupNames: array of string;
  end;

  TSymmetricMolecule = class;

  { TSymmetricTop }

  TSymmetricTop = class (TState)
  private
    FMolecule: TSymmetricMolecule;
    FProlate: Boolean;
    TSR: array [0..2] of Double; { q=0 Spherical tensor equivalent of spin rotation terms }
    QuarticSpinRotation: Boolean;
    function HDip(TransitionMoment: TSingleTransitionMoment; Jp, Kp, lp, J, K, l: TQuantumNumber; Drho: Integer): Double;
  public
    property Prolate: Boolean read FProlate;
    constructor Create(AOwner: TComponent); override;
    function FixupProperties:Exception; override;
    function ReadProperties(Node: TDOMNode; CheckUnread: Boolean = True):Integer; override;
    function AddOneJ(Eigenvalues: TEigenvalues; J, F, M: TQuantumNumber; Sym: Integer; HyperfineState:THyperfineBasis; Seffective: TQuantumNumber): Integer; override;
    function AddOneN(Eigenvalues: TEigenvalues; N, J, F, M: TQuantumNumber; Sym: Integer; HyperfineState:THyperfineBasis; Seffective: TQuantumNumber; out StatWtGroup: Integer): Integer;
    function GetSettingsNames(AIndex: Integer; out LowestValue: Integer): PPChar; override;
    function HRot(J, Kp, lp, K, l: TQuantumNumber): Double;
    function HSR(J, Np, Kp, lp, N, K, l:TQuantumNumber): Complex;
    function HasWidth: Boolean; override;
    function HWidth(J, Kp, lp, K, l: TQuantumNumber): Double;
    function HyperfineMatrixElement(ABra, AKet: TBasisState; Signp, Sign: Integer; MNode: TMNode; RankWanted: TQuantumNumber): Complex; override;
    function RotationalMatrixElement(ABra, AKet: TBasisState; SignA, SignB: Integer; MNode: TMnode; Flag: Integer): Complex; override;
    function RotationalPerturbationMatrixElement(ABra, AKet: TBasisState; SignA, SignB: Integer; MNode: TMNode; Flag: Integer):Complex; override;
    function SymbolicRotationalMatrixElement(AIndex: Integer; RankWanted: Integer):TMatrixElements; override; { For given parameter }

    function RotationalTransitionMatrixElement(ABra, AKet: TBasisState; SignA, SignB: Integer; TransitionMoment: TRootedMNodePair; F: TQuantumNumber; out dK: TQuantumNumber):Complex; override;
    function SymbolicRotationalTransitionMatrixElements(TransitionMoment: TTransitionMoment; AddHyperfineCoupling: Boolean; IncompleteOK: Boolean): TMatrixElements; override;
    function HDipMatrixElements(TransitionMoment: TTransitionMoment; IncompleteOK: Boolean): TMatrixElements;
    function AllowedSphericalComponent(TransitionMoment: TSphericalTransitionMoment; State: TState; out Component: TQuantumNumber):Exception; override;
    function ConvertParameter(AIndex: Integer; Zero: Boolean): TPerturbation; override;
    procedure ConvertToPerturbations(Nuclei: Boolean); override;
    function MatrixElementType(const TheMatrixElement: TMatrixElement; const H: string; ShowPhases: Boolean = False):string; override;

    property Molecule: TSymmetricMolecule read FMolecule;
    function HelpOnSetting(AIndex: Integer): string; override;
    function HelpOnParameter(AIndex: Integer): string; override;
    function IsSettingQuantum(AIndex: Integer):Boolean; override;
    function BumpBranchQno(Branch: string; AIncrement: TQuantumNumber; QnoType:TQnoType):string; override;
    function ReadBranchFormat(Text: string; var Obs: TObservation; UpperState: TState; Reader: TObservationsReaderBase):Boolean; override;
    function SettingToStr(AIndex: Integer): string; override;
    procedure SetSettingFromStr(AIndex: Integer; const Str: string); override;
    procedure AdjustDefaults; override;
  published
    property Symmetry: Integer index LSI read GetSetting write SetSetting;
    property Kmin:     TQuantumNumber  index LSI+1 read GetSetting write SetSetting;
    property Kmax:     TQuantumNumber  index LSI+2 read GetSetting write SetSetting;

    property B:      Double index LSP   read GetParameter write SetParameter;
    property C:      Double index LSP+1 read GetParameter write SetParameter;
    { Quartic Centrifugal Distortion }
    property DJ:     Double index LSP+2 read GetParameter write SetParameter;
    property DJK:    Double index LSP+3 read GetParameter write SetParameter;
    property DK:     Double index LSP+4 read GetParameter write SetParameter;
    { Coriolis }
    property zeta:   Double index LSP+5 read GetParameter write SetParameter;
    property etaJ:   Double index LSP+6 read GetParameter write SetParameter;
    property etaK:   Double index LSP+7 read GetParameter write SetParameter;
    { l doubling }
    property qplus:  Double index LSP+ 8 read GetParameter write SetParameter;
    property qminus: Double index LSP+ 9 read GetParameter write SetParameter;
    property r:      Double index LSP+10 read GetParameter write SetParameter;
    property DqJ:    Double index LSP+11 read GetParameter write SetParameter;
    property DqK:    Double index LSP+12 read GetParameter write SetParameter;
    property DrJ:    Double index LSP+13 read GetParameter write SetParameter;
    property DrK:    Double index LSP+14 read GetParameter write SetParameter;
    { Sextic Centrifugal Distortion }
    property HJ:     Double index LSP+15 read GetParameter write SetParameter;
    property HJK:    Double index LSP+16 read GetParameter write SetParameter;
    property HKJ:    Double index LSP+17 read GetParameter write SetParameter;
    property HK:     Double index LSP+18 read GetParameter write SetParameter;
    { Octic Centrifugal Distortion }
    property LJ:     Double index LSP+19 read GetParameter write SetParameter;
    property LJJK:   Double index LSP+20 read GetParameter write SetParameter;
    property LJK:    Double index LSP+21 read GetParameter write SetParameter;
    property LKKJ:   Double index LSP+22 read GetParameter write SetParameter;
    property LK:     Double index LSP+23 read GetParameter write SetParameter;

    { Spin-Rotation Parameters }
    property ebb:    Double index LSP+24 read GetParameter write SetParameter;
    property ecc:    Double index LSP+25 read GetParameter write SetParameter;
    { Quartic spin rotation parameters }
    property DsN:    Double index LSP+26 read GetParameter write SetParameter;
    property DsNK:   Double index LSP+27 read GetParameter write SetParameter;
    property DsKN:   Double index LSP+28 read GetParameter write SetParameter;
    property DsK:    Double index LSP+29 read GetParameter write SetParameter;
    { Spin-Spin interaction }
    property alpha:  Double index LSP+30 read GetParameter write SetParameter;
    property beta:   Double index LSP+31 read GetParameter write SetParameter;
    { Effective spin-orbit parameter }
    property aeff:   Double index LSP+32 read GetParameter write SetParameter;

    { Linewidth parameters - initial stab }
    property wK:     Double index LSP+33 read GetParameter write SetParameter;
    property wJ:     Double index LSP+34 read GetParameter write SetParameter;
  end;

  TPertEquiv = record
    Name: string;
    JPower:Integer;
    zPower:Integer;
    pPower:Integer;
    lPower:Integer;
    lChange:Integer;
    SPower:Integer;
    SzPower:Integer;
    SpPower:Integer;
    NSPower:Integer;
    Scale4:Integer; { 4 * Scale factor }
  end;

  { TSymmetricTopPerturbation }

  TSymmetricTopPerturbation = class (TPerturbation)
  public
    constructor Create(AOwner: TComponent); override;
    function MatrixElement(J, S, Np, Kp, lp, N, K, l: TQuantumNumber): Double;
    function PairName: string; override;
    function HelpOnSetting(AIndex: Integer): string; override;
    function IsSettingQuantum(AIndex: Integer):Boolean; override;
    function SettingToStr(AIndex: Integer): string; override;
    procedure SetSettingFromStr(AIndex: Integer; const s: string); override;
    procedure SetFromPertEquiv(const PE: TPertEquiv);
    function SymbolicMatrixElements(WithHyperfine: Boolean):TMatrixElements; override;
  published
    property JPower:Integer  index LPI   read GetSetting write SetSetting;
    property zPower:Integer  index LPI+1 read GetSetting write SetSetting;
    property pPower:Integer  index LPI+2 read GetSetting write SetSetting;
    property lPower:Integer  index LPI+3 read GetSetting write SetSetting;
    property lChange:Integer index LPI+4 read GetSetting write SetSetting;
    property KSelect:TQuantumNumber index LPI+5 read GetSetting write SetSetting;
    property SPower:Integer         index LPI+6 read GetSetting write SetSetting;
    property SzPower:Integer        index LPI+7 read GetSetting write SetSetting;
    property SpPower:Integer        index LPI+8 read GetSetting write SetSetting;
    property NSPower:Integer        index LPI+9 read GetSetting write SetSetting;
  end;

  TsymmetricTopNucleus = class(TNucleusWithSpin)
  private
    CHI: array[-2..2] of Double; { Spherical tensor equivalent of quadrupole terms }
    QuadI: Double; { = 1/thrj(I, 4, I, -I, 0, I) }
    TC00, TC20: Double;
  public
    function FixupProperties: Exception; override;
    function HelpOnParameter(AIndex: Integer): string; override;
    function HasIdentical: Boolean; override;
  published
    { Quadrupole }
    property eQq:                Double index  1 read GetParameter write SetParameter;
    property eQq2:               Double index  2 read GetParameter write SetParameter;
    { Nuclear Spin - Rotation }
    property Czz:                Double index  3 read GetParameter write SetParameter;
    property Cperp:{=(Cxx+Cyy)/2}Double index  4 read GetParameter write SetParameter;
  end;

  TSymmetricTopBasisState = class(TBasisState)
  public
    K, l: TQuantumNumber;
    function klLabel(out ExtraTag: string): string;
    function TrackIndex: Integer; override;
    function BasisLabel: string; override; { Basis State Label }
    function BasisLabelKey: string; override; { Explanation of Basis State Label }
    function StateLabel(AQuantumNumberFormat: TQuantumNumberFormat): string; override; { State name, assuming energy ordering preserved }
    function StateLabelKey: string; override; { Explanation of State Label }
    function Qno(QnoType: TQnoType): TQuantumNumber; override; { extra quantum numbers - |K|, l }
    function TxName(Upper, Lower: TBasisState): string; override;    { Make transition label such as qRK(3) }
    function TxNameKey(Upper, Lower: TBasisState): string; override; { String explaining transition label }
    function MakeSignature(SignA:Integer; Ket: TBasisState; SignB: Integer; MNode: TMNode; Symmetrised: Boolean): TIntegerArray; override;
    procedure SetSymbolicVariables(SignA:Integer; Ket: TBasisState; SignB: Integer; MNode: TMNode); override;
    function SymbolicVariableValues: string; override;
  end;

  TSymmetricManifold = class(TManifold)
  public
    function nSym: Integer; override; { Number of different symmetries }
  end;

  { TSymmetricMolecule }
  TSymmetryLabels = (Always, Required,  Never);

  TSymmetricMolecule = class(TMolecule)
  private
    FCharacterTable: TCharacterTable;
    function GetPointGroup(const Index: Integer): TPointGroup;
    function GetSymmetryLabels(AIndex: Integer): TSymmetryLabels;
    procedure SetPointGroup(const Index: Integer; const AValue: TPointGroup);
    procedure SetSymmetryLabels(AIndex: Integer; AValue: TSymmetryLabels);
  protected
    function GetSettingNames(AIndex: Integer): string; override;
  public
    { Symmetric Top Specific }
    property CharacterTable: TCharacterTable read FCharacterTable;
    function IsDegenerate(Sym: Integer):Boolean;
    function g(K, l: TQuantumNumber; EVSym: Integer):Integer;
    { General }
    constructor Create(AOwner: TComponent); override;
    function ReadProperties(Node: TDOMNode; CheckUnread: Boolean = True):Integer; override;
    function FixupProperties: Exception; override;
    function IsValidChildClass(MNodeClass: TMNodeClass):Boolean; override;
    function Allowed(TransitionMoment: TTransitionMoment; Sym1, Sym2: Integer): Boolean; override;
    function DipoleAllowed(Sym1, Sym2: Integer): Boolean;
    function AdjustableSym(Sym: Integer): Boolean;
    function JAdjustedSym(J: TQuantumNumber; Sym: Integer): Integer; override;
    function JAdjustedSymNameBase(J: TQuantumNumber; Sym: Integer): string; override;
    function IsPartialSymmetry(SymS: string):Boolean;
    function LastSym: Integer; override;
    function SymName(Sym: Integer): string; override;
    function TryStrToSimpleQuantumNumber(const str: string; var Value:TQuantumNumber; AQuantumNumberFormat: TQuantumNumberFormat; var SpecialQN: Integer; QnoIndex: Integer):Boolean; override;
    procedure FixSimpleFormat(var s: string); override;
    function QnoInfo(QnoType: TQnoType): TQnoInfo; override;
    function HelpOnSetting(AIndex: Integer): string; override;
    function VibronicallyAllowed(TransitionMoment: TSphericalTransitionMoment):TVibronicallyAllowed; overload;
    function VibronicallyAllowed(TransitionMoment: TSphericalTransitionMoment; Component: TQuantumNumber; BraSym, KetSym: Integer):TVibronicallyAllowed; overload;
  published
    property PointGroup:     TPointGroup     index LMI   read GetPointGroup write SetPointGroup;
    property AlternateBasis: Boolean         index LMI+1 read GetFlag write SetFlag;
    property SymmetryLabels: TSymmetryLabels index LMI+2 read GetSymmetryLabels write SetSymmetryLabels;
    property CompactKLabels: Boolean         index LMI+3 read GetFlag write SetFlag;
    property Wt0:            Integer  index WTI+0  read GetSetting write SetSetting;
    property Wt1:            Integer  index WTI+1  read GetSetting write SetSetting;
    property Wt2:            Integer  index WTI+2  read GetSetting write SetSetting;
    property Wt3:            Integer  index WTI+3  read GetSetting write SetSetting;
    property Wt4:            Integer  index WTI+4  read GetSetting write SetSetting;
    property Wt5:            Integer  index WTI+5  read GetSetting write SetSetting;
    property Wt6:            Integer  index WTI+6  read GetSetting write SetSetting;
    property Wt7:            Integer  index WTI+7  read GetSetting write SetSetting;
    property Wt8:            Integer  index WTI+8  read GetSetting write SetSetting;
    property Wt9:            Integer  index WTI+9  read GetSetting write SetSetting;
    property Wt10:           Integer  index WTI+10 read GetSetting write SetSetting;
    property Wt11:           Integer  index WTI+11 read GetSetting write SetSetting;
    property Wt12:           Integer  index WTI+12 read GetSetting write SetSetting;
    property Wt13:           Integer  index WTI+13 read GetSetting write SetSetting;
  end;

const
  MaxnStatWeight = 14; { Maximum number of statistical weights }

const
  EmptyPertEquiv:TPertEquiv = (Name:''; JPower:0; zPower: 0; pPower:0; lPower: 0; lChange: 0; SPower: 0; SzPower: 0; SpPower: 0; NSPower: 0; Scale4: 4;);

function FindParamEquiv(const PE: TPertEquiv; out OneOfPair: Boolean; out Scale: Double):string;
function AddG(G,n:Integer):Integer; { Should probably be a class function of TSymmetricMolecule }

implementation

uses StrUtils, nj, Worker, UCompat, StaticField, Global, Expr, SymmetricTop, Units;

var
  el, eK, eN, eJ, eS: TSortedVariableNode;
  VList: TVariableList;

{ This implements Table 12-14 and the surrounding logic of
  P.R. Bunker and Per Jensen, Molecular Symmetry and Spectroscopy, 2nd Edition.
  See also http://symmetry.jacobs-university.de/
}

const

  CharacterTableNames: array [TPointGroup] of PChar = ('C3v', 'C4v', 'C5v', 'C6v', 'C7v', 'C8v',
                                                       'D3h', 'D4h', 'D5h', 'D6h', 'D7h', 'D8h',
                                                       'D2d', 'D3d', 'D4d', 'D5d', 'D6d', 'D7d', 'D8d',
                                                       'C3h', 'C4h', 'C5h', 'C6h',
                                                       'S4',  'S6',  'S8',  'S10', 'S12',
                                                       'D3',  'D4',  'D5',  'D6',  'D7',  'D8');

  C3vCharacterTable: array [0..2] of TCharacterTableEntry =
  ((Name:'A1'; g:0; rho:0; eta:0),
   (Name:'A2'; g:0; rho:1; eta:0),
   (Name:'E';  g:1; rho:0; eta:0));

  C5vCharacterTable: array [0..3] of TCharacterTableEntry =
  ((Name:'A1'; g:0; rho:0; eta:0),
   (Name:'A2'; g:0; rho:1; eta:0),
   (Name:'E1'; g:1; rho:0; eta:0),
   (Name:'E2'; g:2; rho:0; eta:0));

  C7vCharacterTable: array [0..4] of TCharacterTableEntry =
  ((Name:'A1'; g:0; rho:0; eta:0),
   (Name:'A2'; g:0; rho:1; eta:0),
   (Name:'E1'; g:1; rho:0; eta:0),
   (Name:'E2'; g:2; rho:0; eta:0),
   (Name:'E3'; g:3; rho:0; eta:0));

  C4vCharacterTable: array [0..4] of TCharacterTableEntry =
  ((Name:'A1'; g:0; rho:0; eta:0),
   (Name:'A2'; g:0; rho:1; eta:0),
   (Name:'B1'; g:2; rho:0; eta:0),
   (Name:'B2'; g:2; rho:1; eta:0),
   (Name:'E';  g:1; rho:0; eta:0));

  C6vCharacterTable: array [0..5] of TCharacterTableEntry =
  ((Name:'A1'; g:0; rho:0; eta:0),
   (Name:'A2'; g:0; rho:1; eta:0),
   (Name:'B1'; g:3; rho:0; eta:0),
   (Name:'B2'; g:3; rho:1; eta:0),
   (Name:'E1'; g:1; rho:0; eta:0),
   (Name:'E2'; g:2; rho:0; eta:0));

  C8vCharacterTable: array [0..6] of TCharacterTableEntry =
  ((Name:'A1'; g:0; rho:0; eta:0),
   (Name:'A2'; g:0; rho:1; eta:0),
   (Name:'B1'; g:4; rho:0; eta:0),
   (Name:'B2'; g:4; rho:1; eta:0),
   (Name:'E1'; g:1; rho:0; eta:0),
   (Name:'E2'; g:2; rho:0; eta:0),
   (Name:'E3'; g:3; rho:0; eta:0));

  D3hCharacterTable: array [0..5] of TCharacterTableEntry =
  ((Name:'A1'''; g:0; rho:0; eta:0),
   (Name:'A2'''; g:0; rho:1; eta:0),
   (Name:'E''';  g:1; rho:0; eta:0),
   (Name:'A1"';  g:0; rho:0; eta:1),
   (Name:'A2"';  g:0; rho:1; eta:1),
   (Name:'E"';   g:1; rho:0; eta:1));

  D5hCharacterTable: array [0..7] of TCharacterTableEntry =
  ((Name:'A1'''; g:0; rho:0; eta:0),
   (Name:'A2'''; g:0; rho:1; eta:0),
   (Name:'E1'''; g:1; rho:0; eta:0),
   (Name:'E2'''; g:2; rho:0; eta:0),
   (Name:'A1"';  g:0; rho:0; eta:1),
   (Name:'A2"';  g:0; rho:1; eta:1),
   (Name:'E1"';  g:1; rho:0; eta:1),
   (Name:'E2"';  g:2; rho:0; eta:1));

  D7hCharacterTable: array [0..9] of TCharacterTableEntry =
  ((Name:'A1'''; g:0; rho:0; eta:0),
   (Name:'A2'''; g:0; rho:1; eta:0),
   (Name:'E1'''; g:1; rho:0; eta:0),
   (Name:'E2'''; g:2; rho:0; eta:0),
   (Name:'E3'''; g:3; rho:0; eta:0),
   (Name:'A1"';  g:0; rho:0; eta:1),
   (Name:'A2"';  g:0; rho:1; eta:1),
   (Name:'E1"';  g:1; rho:0; eta:1),
   (Name:'E2"';  g:2; rho:0; eta:1),
   (Name:'E3"';  g:3; rho:0; eta:1));

  D4hCharacterTable: array [0..9] of TCharacterTableEntry =
  ((Name:'A1g'; g:0; rho:0; eta:0),
   (Name:'A2g'; g:0; rho:1; eta:0),
   (Name:'B1g'; g:2; rho:0; eta:0),
   (Name:'B2g'; g:2; rho:1; eta:0),
   (Name:'Eg';  g:1; rho:0; eta:0),
   (Name:'A1u'; g:0; rho:0; eta:1),
   (Name:'A2u'; g:0; rho:1; eta:1),
   (Name:'B1u'; g:2; rho:0; eta:1),
   (Name:'B2u'; g:2; rho:1; eta:1),
   (Name:'Eu';  g:1; rho:0; eta:1));

  D6hCharacterTable: array [0..11] of TCharacterTableEntry =
  ((Name:'A1g'; g:0; rho:0; eta:0),
   (Name:'A2g'; g:0; rho:1; eta:0),
   (Name:'B1g'; g:3; rho:0; eta:0),
   (Name:'B2g'; g:3; rho:1; eta:0),
   (Name:'E1g'; g:1; rho:0; eta:0),
   (Name:'E2g'; g:2; rho:0; eta:0),
   (Name:'A1u'; g:0; rho:0; eta:1),
   (Name:'A2u'; g:0; rho:1; eta:1),
   (Name:'B1u'; g:3; rho:0; eta:1),
   (Name:'B2u'; g:3; rho:1; eta:1),
   (Name:'E1u'; g:1; rho:0; eta:1),
   (Name:'E2u'; g:2; rho:0; eta:1));

  D8hCharacterTable: array [0..13] of TCharacterTableEntry =
  ((Name:'A1g'; g:0; rho:0; eta:0),
   (Name:'A2g'; g:0; rho:1; eta:0),
   (Name:'B1g'; g:4; rho:0; eta:0),
   (Name:'B2g'; g:4; rho:1; eta:0),
   (Name:'E1g'; g:1; rho:0; eta:0),
   (Name:'E2g'; g:2; rho:0; eta:0),
   (Name:'E3g'; g:3; rho:0; eta:0),
   (Name:'A1u'; g:0; rho:0; eta:1),
   (Name:'A2u'; g:0; rho:1; eta:1),
   (Name:'B1u'; g:4; rho:0; eta:1),
   (Name:'B2u'; g:4; rho:1; eta:1),
   (Name:'E1u'; g:1; rho:0; eta:1),
   (Name:'E2u'; g:2; rho:0; eta:1),
   (Name:'E3u'; g:3; rho:0; eta:1));

  D3dCharacterTable: array [0..5] of TCharacterTableEntry =
  ((Name:'A1g'; g:0; rho:0; eta:0),
   (Name:'A2g'; g:0; rho:1; eta:0),
   (Name:'Eg';  g:1; rho:0; eta:0),
   (Name:'A1u'; g:0; rho:0; eta:1),
   (Name:'A2u'; g:0; rho:1; eta:1),
   (Name:'Eu';  g:1; rho:0; eta:1));

  D5dCharacterTable: array [0..7] of TCharacterTableEntry =
  ((Name:'A1g'; g:0; rho:0; eta:0),
   (Name:'A2g'; g:0; rho:1; eta:0),
   (Name:'E1g'; g:1; rho:0; eta:0),
   (Name:'E2g'; g:2; rho:0; eta:0),
   (Name:'A1u'; g:0; rho:0; eta:1),
   (Name:'A2u'; g:0; rho:1; eta:1),
   (Name:'E1u'; g:1; rho:0; eta:1),
   (Name:'E2u'; g:2; rho:0; eta:1));

  D7dCharacterTable: array [0..9] of TCharacterTableEntry =
  ((Name:'A1g'; g:0; rho:0; eta:0),
   (Name:'A2g'; g:0; rho:1; eta:0),
   (Name:'E1g'; g:1; rho:0; eta:0),
   (Name:'E2g'; g:2; rho:0; eta:0),
   (Name:'E3g'; g:3; rho:0; eta:0),
   (Name:'A1u'; g:0; rho:0; eta:1),
   (Name:'A2u'; g:0; rho:1; eta:1),
   (Name:'E1u'; g:1; rho:0; eta:1),
   (Name:'E2u'; g:2; rho:0; eta:1),
   (Name:'E3u'; g:3; rho:0; eta:1));

  D2dCharacterTable: array [0..4] of TCharacterTableEntry =
  ((Name:'A1'; g:0; rho:0; eta:0),
   (Name:'A2'; g:0; rho:1; eta:0),
   (Name:'B1'; g:2; rho:0; eta:0),
   (Name:'B2'; g:2; rho:1; eta:0),
   (Name:'E';  g:1; rho:0; eta:0));

  D4dCharacterTable: array [0..6] of TCharacterTableEntry =
  ((Name:'A1'; g:0; rho:0; eta:0),
   (Name:'A2'; g:0; rho:1; eta:0),
   (Name:'B1'; g:4; rho:0; eta:0),
   (Name:'B2'; g:4; rho:1; eta:0),
   (Name:'E1'; g:1; rho:0; eta:0),
   (Name:'E2'; g:2; rho:0; eta:0),
   (Name:'E3'; g:3; rho:0; eta:0));

  D6dCharacterTable: array [0..8] of TCharacterTableEntry =
  ((Name:'A1'; g:0; rho:0; eta:0),
   (Name:'A2'; g:0; rho:1; eta:0),
   (Name:'B1'; g:6; rho:0; eta:0),
   (Name:'B2'; g:6; rho:1; eta:0),
   (Name:'E1'; g:1; rho:0; eta:0),
   (Name:'E2'; g:2; rho:0; eta:0),
   (Name:'E3'; g:3; rho:0; eta:0),
   (Name:'E4'; g:4; rho:0; eta:0),
   (Name:'E5'; g:5; rho:0; eta:0));

  D8dCharacterTable: array [0..10] of TCharacterTableEntry =
  ((Name:'A1'; g:0; rho:0; eta:0),
   (Name:'A2'; g:0; rho:1; eta:0),
   (Name:'B1'; g:8; rho:0; eta:0),
   (Name:'B2'; g:8; rho:1; eta:0),
   (Name:'E1'; g:1; rho:0; eta:0),
   (Name:'E2'; g:2; rho:0; eta:0),
   (Name:'E3'; g:3; rho:0; eta:0),
   (Name:'E4'; g:4; rho:0; eta:0),
   (Name:'E5'; g:5; rho:0; eta:0),
   (Name:'E6'; g:6; rho:0; eta:0),
   (Name:'E7'; g:7; rho:0; eta:0));

  D3CharacterTable: array [0..2] of TCharacterTableEntry =
  ((Name:'A1'; g:0; rho:0; eta:0),
   (Name:'A2'; g:0; rho:1; eta:0),
   (Name:'E';  g:1; rho:0; eta:0));

  D5CharacterTable: array [0..3] of TCharacterTableEntry =
  ((Name:'A1'; g:0; rho:0; eta:0),
   (Name:'A2'; g:0; rho:1; eta:0),
   (Name:'E1'; g:1; rho:0; eta:0),
   (Name:'E2'; g:2; rho:0; eta:0));

  D7CharacterTable: array [0..4] of TCharacterTableEntry =
  ((Name:'A1'; g:0; rho:0; eta:0),
   (Name:'A2'; g:0; rho:1; eta:0),
   (Name:'E1'; g:1; rho:0; eta:0),
   (Name:'E2'; g:2; rho:0; eta:0),
   (Name:'E3'; g:3; rho:0; eta:0));

  D4CharacterTable: array [0..4] of TCharacterTableEntry =
  ((Name:'A1'; g:0; rho:0; eta:0),
   (Name:'A2'; g:0; rho:1; eta:0),
   (Name:'B1'; g:2; rho:0; eta:0),
   (Name:'B2'; g:2; rho:1; eta:0),
   (Name:'E';  g:1; rho:0; eta:0));

  D6CharacterTable: array [0..5] of TCharacterTableEntry =
  ((Name:'A1'; g:0; rho:0; eta:0),
   (Name:'A2'; g:0; rho:1; eta:0),
   (Name:'B1'; g:3; rho:0; eta:0),
   (Name:'B2'; g:3; rho:1; eta:0),
   (Name:'E1'; g:1; rho:0; eta:0),
   (Name:'E2'; g:2; rho:0; eta:0));

  D8CharacterTable: array [0..6] of TCharacterTableEntry =
  ((Name:'A1'; g:0; rho:0; eta:0),
   (Name:'A2'; g:0; rho:1; eta:0),
   (Name:'B1'; g:4; rho:0; eta:0),
   (Name:'B2'; g:4; rho:1; eta:0),
   (Name:'E1'; g:1; rho:0; eta:0),
   (Name:'E2'; g:2; rho:0; eta:0),
   (Name:'E3'; g:3; rho:0; eta:0));

{ For point groups like C3h E functions are actually pairs
   of separate functions that are complex conjugates of each other;
   The E functions must therefore be counted twice, as they are
   independent functions and the nuclear spin function can be chosen
   independently for each function.
   Also, there is no R- operation, which converts K -> -K
   This has the consequence that for K a multiple of n, the functions
   have the same symmetry, and are not necessarily degenerate.
   (The E functions still are, but this arises from time reversal symmetry,
    not standard point group theory).
 }
 C3hCharacterTable: array [0..3] of TCharacterTableEntry =
  ((Name:'A'''; g:0; rho:0; eta:0),  { Don't use rho at all; no R'+ }
   (Name:'E'''; g:1; rho:0; eta:0),
   (Name:'A"';  g:0; rho:0; eta:1),
   (Name:'E"';  g:1; rho:0; eta:1));

 C4hCharacterTable: array [0..5] of TCharacterTableEntry =
  ((Name:'Ag'; g:0; rho:0; eta:0),  { Don't use rho at all; R'+ = i }
   (Name:'Bg'; g:2; rho:0; eta:0),
   (Name:'Eg'; g:1; rho:0; eta:0),
   (Name:'Au'; g:0; rho:0; eta:1),
   (Name:'Bu'; g:2; rho:0; eta:1),
   (Name:'Eu'; g:1; rho:0; eta:1));

 C5hCharacterTable: array [0..5] of TCharacterTableEntry =
  ((Name:'A''';  g:0; rho:0; eta:0), { Don't use rho at all; R'+ = sigma }
   (Name:'E1'''; g:1; rho:0; eta:0),
   (Name:'E2'''; g:2; rho:0; eta:0),
   (Name:'A"';   g:0; rho:0; eta:1),
   (Name:'E1"';  g:1; rho:0; eta:1),
   (Name:'E2"';  g:2; rho:0; eta:1));

 C6hCharacterTable: array [0..7] of TCharacterTableEntry =
  ((Name:'Ag';  g:0; rho:0; eta:0),  { Don't use rho at all; R'+ = i }
   (Name:'Bg';  g:3; rho:0; eta:0),
   (Name:'E1g'; g:1; rho:0; eta:0),
   (Name:'E2g'; g:2; rho:0; eta:0),
   (Name:'Au';  g:0; rho:0; eta:1),
   (Name:'Bu';  g:3; rho:0; eta:1),
   (Name:'E1u'; g:1; rho:0; eta:1),
   (Name:'E2u'; g:2; rho:0; eta:1));

 S4CharacterTable: array [0..2] of TCharacterTableEntry =
  ((Name:'A'; g:0; rho:0; eta:0),  { Don't use rho at all; no R'+ }
   (Name:'B'; g:2; rho:0; eta:0),
   (Name:'E'; g:1; rho:0; eta:0));

 S6CharacterTable: array [0..3] of TCharacterTableEntry =
  ((Name:'Ag'; g:0; rho:0; eta:0),  { Don't use rho at all; R'+ = i }
   (Name:'Eg'; g:1; rho:0; eta:0),
   (Name:'Au'; g:0; rho:0; eta:1),
   (Name:'Eu'; g:1; rho:0; eta:1));

 S8CharacterTable: array [0..4] of TCharacterTableEntry =
  ((Name:'A';  g:0; rho:0; eta:0),  { Don't use rho at all; no R'+ }
   (Name:'B';  g:4; rho:0; eta:0),
   (Name:'E1'; g:1; rho:0; eta:0),
   (Name:'E2'; g:2; rho:0; eta:0),
   (Name:'E3'; g:3; rho:0; eta:0));

 S10CharacterTable: array [0..5] of TCharacterTableEntry =
  ((Name:'Ag';  g:0; rho:0; eta:0),  { Don't use rho at all; R'+ = i }
   (Name:'E1g'; g:1; rho:0; eta:0),
   (Name:'E2g'; g:2; rho:0; eta:0),
   (Name:'Au';  g:0; rho:0; eta:1),
   (Name:'E1u'; g:1; rho:0; eta:1),
   (Name:'E2u'; g:2; rho:0; eta:1));

 S12CharacterTable: array [0..6] of TCharacterTableEntry =
  ((Name:'A';  g:0; rho:0; eta:0),  { Don't use rho at all; no R'+ }
   (Name:'B';  g:6; rho:0; eta:0),
   (Name:'E1'; g:1; rho:0; eta:0),
   (Name:'E2'; g:2; rho:0; eta:0),
   (Name:'E3'; g:3; rho:0; eta:0),
   (Name:'E4'; g:4; rho:0; eta:0),
   (Name:'E5'; g:5; rho:0; eta:0));

var
  CharacterTables: array [TPointGroup] of TCharacterTable;

function AddG(G,n:Integer):Integer;
begin { Reduce G quantum number }
  if g < 0 then
    g := g - g*n; { Want result of mod on next line to be positive }
  g := g mod n;
  if g > n div 2 then
    Dec(g,n);
  Result := g;
end;

{ TSymmetricMolecule }

function TSymmetricMolecule.DipoleAllowed(Sym1, Sym2: Integer): Boolean;
var
  Gdipole, G1, G2: TCharacterTableEntry;
  g: Integer;
begin
  Result := False;
  { Does Sym1 * DipoleSym = Sym2? }
  { En Match }
  { TODO: Speed up? Gdipole.g is only non-zero for D2d, D4d, ..., S4, S8, S12, ... }
  Gdipole := CharacterTable.IReps[CharacterTable.DipoleSym];
  G1 := CharacterTable.IReps[Sym1];
  G2 := CharacterTable.IReps[Sym2];
  g := AddG(Gdipole.g + G1.g, CharacterTable.n);
  if Abs(g) <> G2.g then Exit;
  { g/u '/'' Match }
  if Odd(Gdipole.eta + G1.eta + G2.eta) then Exit;
  if (CharacterTable.En[sym2] = 0) { Not degenerate }
     and Odd(Gdipole.rho + G1.rho + G2.rho) then Exit;
  Result := True;
end;

function TSymmetricMolecule.FixupProperties: Exception;
var
  I: Integer;
begin
  Result := Inherited FixupProperties;
  if not Assigned(Result) and JAdjustSym and (nNuclei > 0) then
    for i := 0 to nNuclei - 1 do
      if Nuclei[i].Spin > 0 then
        Result := EReportException.Create('J adjusted symmetry for '+Name+' not implemented for hyperfine structure');
end;

function TSymmetricMolecule.g(K, l: TQuantumNumber; EVSym: Integer): Integer;
begin
  Result := CharacterTable.IReps[EVSym].g;
  if l < 0 then
    Result := AddG(CharacterTable.gKmult*K div 2 - Result, CharacterTable.n)
  else
    Result := AddG(CharacterTable.gKmult*K div 2 + Result, CharacterTable.n);
end;

function TSymmetricMolecule.Allowed(TransitionMoment: TTransitionMoment; Sym1, Sym2: Integer): Boolean;
begin
  Result := False;
  case TransitionMoment.VibronicallyAllowed of
    vaMaybe: if (TransitionMoment is TSphericalTransitionMoment)
                and (VibronicallyAllowed(TSphericalTransitionMoment(TransitionMoment)) = vaNo) then begin
               TransitionMoment.VibronicallyAllowed := vaNo;
               Exit;
             end else
               TransitionMoment.VibronicallyAllowed := vaYes;
    vaNo: Exit;
  end;

  if not TransitionMoment.Magnetic and  (Minus(TransitionMoment.Rank) < 0) then
    Result := DipoleAllowed(Sym1, Sym2)
  else
    Result := Sym1 = Sym2;
end;

constructor TSymmetricMolecule.Create(AOwner: TComponent);
var
  i: Integer;
begin
  SavedProperties := SavedProperties + ',PointGroup=C3v,SymmetryLabels=Required';
  for i := 0 to MaxnStatWeight-1 do
    SavedProperties := SavedProperties + Format(',wt%d=1', [i]);
  inherited Create(AOwner);
end;

function TSymmetricMolecule.ReadProperties(Node: TDOMNode; CheckUnread: Boolean): Integer;
var
  s: string;
  p: Integer;
begin
  { AlternateBasis was represented in point group name before 6.0.199, read it here }
  s := ReadValue(Node, 'PointGroup', 'C3v', '', False);
  p := Pos('alt', LowerCase(s));
  if p > 0 then begin
    s := Copy(s,1,Length(s)-3);
    WriteValue(Node,'PointGroup',s,'C3v');
  end;
  Result:=inherited ReadProperties(Node, CheckUnread);
  if p > 0 then
    AlternateBasis := True;
end;

function TSymmetricMolecule.GetSettingNames(AIndex: Integer): string;
begin
  Result := inherited GetSettingNames(AIndex);
  if (AIndex >= WTI) and (AIndex < WTI+MaxnStatWeight) then
    if AIndex <= WTI+High(CharacterTable.GroupNames) then
      Result := CharacterTable.GroupNames[AIndex-WTI]
    else
      Result := '(' + Result + ')';
end;

function TSymmetricMolecule.IsDegenerate(Sym: Integer): Boolean;
begin
  Result := CharacterTable.En[Sym] <> 0;
end;

function TSymmetricMolecule.IsPartialSymmetry(SymS: string): Boolean;
var
  i: Integer;
begin
  Result := Length(SymS) = 1;
  if Result then
    for i := 0 to High(CharacterTable.IReps) do
      if (Length(CharacterTable.IReps[i].Name) > 1) and (CharacterTable.IReps[i].Name[1] = SymS) then
        Exit;
  Result := False;
end;

function TSymmetricMolecule.GetPointGroup(const Index: Integer): TPointGroup;
begin
  Result := TPointGroup(GetSetting(Index));
end;

function TSymmetricMolecule.GetSymmetryLabels(AIndex: Integer): TSymmetryLabels;
begin
  Result := TSymmetryLabels(GetSetting(AIndex));
end;

procedure TSymmetricMolecule.SetPointGroup(const Index: Integer; const AValue: TPointGroup);
begin
  FCharacterTable := CharacterTables[AValue];
  FnSpinSpecies := Length(CharacterTable.GroupNames);
  if PointGroup = AValue then Exit;
  SetSetting(Index, Ord(AValue));
  PerformMNodeNotification(Self, mnPropNameChange, Index);
end;

procedure TSymmetricMolecule.SetSymmetryLabels(AIndex: Integer;  AValue: TSymmetryLabels);
begin
  SetSetting(AIndex, Ord(AValue));
end;

function TSymmetricMolecule.AdjustableSym(Sym: Integer): Boolean;
begin
  { Is Sym A1/A2 B1/B2, and thus adjustable? }
  Result := (Sym >= 0) and JAdjustSym and not Odd(HalfSpin) { TODO: Implement for odd spin } and not IsDegenerate(Sym);
end;

function TSymmetricMolecule.JAdjustedSym(J: TQuantumNumber; Sym: Integer): Integer;
begin
  Result := Sym;
  if AdjustableSym(Sym) and (Minus(J) < 0) then
    if CharacterTable.IReps[Sym].rho = 1 then
      Dec(Result)
    else if CharacterTable.IReps[Result+1].rho = 1 then
      { No J adjustment for Separably Degenerate groups }
      Inc(Result);
end;

function TSymmetricMolecule.JAdjustedSymNameBase(J: TQuantumNumber; Sym: Integer): string;
begin
  if Sym < 0 then
    Result := '-'
  else if not Odd(HalfSpin) then begin
    Result := SymName(Sym);
    if AdjustableSym(Sym) and (Length(Result) > 1) and (Result[2] in ['1'..'2']) then
      if (Result[2] = '1') xor (Minus(J) < 0) then
        Result[2] := '+' { A1 even J, A2 Odd J }
      else
        Result[2] := '-'; { A2 even J, A1 Odd J }
  end else
    { TODO: Check odd spin implementation; Check PseudoC2v }
    Result := inherited JAdjustedSymNameBase(J, Sym)
end;

function TSymmetricMolecule.SymName(Sym: Integer): string;
begin
  if Sym < 0 then
    Result := '-'
  else
    Result := CharacterTable.Names[Sym];
end;

function TSymmetricMolecule.TryStrToSimpleQuantumNumber(const str: string;
  var Value: TQuantumNumber; AQuantumNumberFormat: TQuantumNumberFormat;
  var SpecialQN: Integer; QnoIndex: Integer): Boolean;
var
  i, p, Sym, Sign1, Sign2: Integer;
begin
  Result := inherited TryStrToSimpleQuantumNumber(str, Value, AQuantumNumberFormat, SpecialQN, QnoIndex);
  { Allow specification of symmetry after K, and sign of kl as sign of K }
  if not Result and (str <> '') and (QnoIndex > 0) then begin
    Sign1 := 0;
    Sign2 := 0;
    p := 1;
    while (p <= Length(str)) and (str[p] in ['-','+']) do begin
      if Sign1 <> 0 then
        Exit; { Three signs given - not valid }
      Sign1 := Sign2;
      if str[p] = '+' then
        Sign2 := 1
      else
        Sign2 := 2;
      Inc(p);
    end;
    i := p;
    while (i <= Length(str)) and (str[i] in ['0'..'9','.', '/']) do
      Inc(i);
    Result := TryStrToQuantumNumber(Copy(str, p, i-p), Value, AQuantumNumberFormat);
    Sym := -1; { Flag symmetry unspecified }
    if Result and (i <= Length(str)) and not TryStrToAnySymName(0, Copy(str,i,Length(str)), Sym) then
      Result := False
    else begin
      SpecialQN := Max(SpecialQN, QnoIndex);
      Value := (Value+100000)*1600 + Sign1*400 + Sign2*100 + Sym + 1;
    end;
  end;
end;

procedure TSymmetricMolecule.FixSimpleFormat(var s: string);
var
  p, q, i: Integer;
begin
  { TODO: Something extra required for separable degeneracy }
  { Find location of symmetry label }
  p := Length(s);
  while (p > 0) and (s[p] in ['0'..'9',',',' ']) do
    Dec(p);
  if p < 2 then
    Exit; { No symmetry label }
  q := p-1;
  while (q > 0) and (s[q] = ' ') do
    Dec(q);
  { Delete spaces in front of symmetry }
  Delete(s, q+1, p-q-1);
  p := q+1;
  { Blank out symmetry if E }
  if s[p] = 'E' then
    if Length(s) < p+CharacterTable.LongestSymName-1 then begin
      { Space pad symmetry label if neccessary }
      SetLength(s, p+CharacterTable.LongestSymName-1);
      for i := 0 to CharacterTable.LongestSymName-1 do
        s[p+i] := ' ';
    end else
      repeat
        s[p] := ' ';
        Inc(p)
      until (p > Length(s)) or (s[p] = ' ');
  { Merge kl into sign of K if '+1' or '-1' present }
  if (q > 3) and (s[q] = '1') and (s[q-2] = ' ') and (s[q-3] in ['0'..'9']) then
    if s[q-1] = '+' then
      Delete(s,q-2,3) { Delete ' +1' }
    else if s[q-1] = '-' then begin
      { Delete ' -1', and insert a '-' sign }
      Delete(s,q-2,3); { Delete ' -1' }
      q := q-4;
      while (q > 0) and (s[q] in ['0'..'9']) do
        Dec(q);
      if (q > 1) and (s[q-1] = ' ') then
        s[q] := '-'
      else
        Insert('-', s, q+1);
    end;
end;

function TSymmetricMolecule.VibronicallyAllowed(TransitionMoment: TSphericalTransitionMoment): TVibronicallyAllowed;
var
  Bra, Ket: TSymmetricTop;
  Component: TQuantumNumber;
begin
  Bra := TransitionMoment.Bra as TSymmetricTop;
  Ket := TransitionMoment.Ket as TSymmetricTop;
  if (Bra.S = Ket.S) then begin
    Component := TransitionMoment.Component;
    if Component <= -9999 then
      Component := TransitionMoment.ActualComponent;
    Result := VibronicallyAllowed(TransitionMoment, Component, Bra.Symmetry, Ket.Symmetry);
  end else
    Result := vaMaybe;
end;

function TSymmetricMolecule.VibronicallyAllowed(TransitionMoment: TSphericalTransitionMoment; Component: TQuantumNumber; BraSym, KetSym: Integer):TVibronicallyAllowed;
var
  g, rho, eta, g1: Integer;
begin
  Result := vaNo;
  { Symmetry of Rovibronic Transition Moment }
  if (Minus(TransitionMoment.Rank) < 0) and not (TransitionMoment is TMagneticTransitionMoment) then begin
    g := CharacterTable.IReps[CharacterTable.DipoleSym].g;
    rho := CharacterTable.IReps[CharacterTable.DipoleSym].rho;
    eta := CharacterTable.IReps[CharacterTable.DipoleSym].eta;
  end else begin
    g := 0;
    rho := 0;
    eta := 0;
  end;
  g := AddG(CharacterTable.gKmult*Component div 2 + g, CharacterTable.n);
  eta := eta + CharacterTable.etaKmult*Component div 2;
  if Odd(eta + CharacterTable.IReps[BraSym].eta + CharacterTable.IReps[KetSym].eta) then
    Exit;
  g1 := AddG(g + CharacterTable.IReps[BraSym].g, CharacterTable.n);
  if Abs(g1) <> CharacterTable.IReps[KetSym].g then begin
    { Check alternate - identical unless ExE }
    g1 := AddG(g - CharacterTable.IReps[BraSym].g, CharacterTable.n);
    if Abs(g1) <> CharacterTable.IReps[KetSym].g then
      Exit;
  end;
  if (g1 = 0) or (2*g1 = CharacterTable.n) then begin
    if (g = 0) or (2*g = CharacterTable.n) then begin
      rho := rho + (CharacterTable.rhoJmult*TransitionMoment.Rank + CharacterTable.rhoKmult*Component) div 2;
      if Odd(rho + CharacterTable.IReps[BraSym].rho + CharacterTable.IReps[KetSym].rho) then
        Exit;
    end;
    { E x E gives both values of rho, so don't check }
  end;
  Result := vaYes;
end;

function TSymmetricMolecule.LastSym: Integer;
begin
  Result := CharacterTable.Size-1;
end;

function TSymmetricMolecule.HelpOnSetting(AIndex: Integer): string;
begin
  if AIndex = LMI then
    Result := 'Point group of molecule.'
  else if AIndex = LMI+1 then
    Result := 'Alternate Basis - Separably degenerate groups only.'
  else if AIndex = LMI+2 then
    Result := 'When to show rovibronic symmetry labels'
  else if AIndex = LMI+3 then
    Result := 'Include sign of l and symmetry in K label'
  else if (AIndex >= WTI) and (AIndex < WTI+MaxnStatWeight) then
    if AIndex <= WTI+High(CharacterTable.GroupNames) then begin
      Result := CharacterTable.GroupNames[AIndex-WTI];
      Result := 'Statistical weight for ' + Copy(Result,4,Length(Result)-4) + ' rotational levels.'
    end else
      Result := 'Unused'
  else
    Result := inherited HelpOnSetting(AIndex);
end;

function TSymmetricMolecule.IsValidChildClass(MNodeClass: TMNodeClass): Boolean;
begin
  Result := inherited IsValidChildClass(MNodeClass);
  if Result and MNodeClass.InheritsFrom(TTransitionMoment) then
    Result := (MNodeClass = TSphericalTransitionMoment)
               or (MNodeClass = TCartesianTransitionMoment)
               or (MNodeClass = TOrbitalTransitionMoment)
               or (MNodeClass = TQuadrupoleTransitionMoment);
end;

function TSymmetricMolecule.QnoInfo(QnoType: TQnoType): TQnoInfo;
begin
  Result := inherited QnoInfo(QnoType);
  case QnoType of
    qtA: Result.Name := '|K|';
    qtB: Result.Name := 'Kl';
  end;
end;

{ TSymmetricManifold }

function TSymmetricManifold.nSym: Integer;
begin
  Result := TSymmetricMolecule(Owner).CharacterTable.Size;
end;

{ TSymmetricTop }

function SameStatisticalWeightGroup(PointGroup: TPointGroup; sym, sym1: Integer):Boolean;
var
  CharacterTable: TCharacterTable;
  function Twog(sym: Integer):Integer;
  begin
    { Apply g for applying R+ twice }
    Result := 2*CharacterTable.IReps[sym].g mod CharacterTable.n;
    if Result > CharacterTable.n div 2 then
      Dec(Result,CharacterTable.n);
    Result := Abs(Result);
  end;
begin
  CharacterTable := CharacterTables[PointGroup];
  { For each symmetric top point group, one of the generating operations corresponds to
    a permutation/inversion, so the symmetry of the nuclear+rovibronic wavefunction is not selected
    by this operation. Symmetries that only differ in this operation are therefore in the same
    statistical weight group }
  { R+ Operation }
  if PointGroup in [D2d, D4d, D6d, D8d, S4, S8, S12] then
    { Operation is Sn; that need not match but Sn^2 must }
    Result := Twog(sym) = Twog(sym1)
  else { Operation is Cn - just match }
    Result := CharacterTable.IReps[sym].g = CharacterTable.IReps[sym1].g;
  { Does R'+ Operation only interchange equivalent nuclei? } { TODO: Always false? - G12 is exception? }
  if not (PointGroup in [D3h, D4h, D5h, D6h, D7h, D8h, D3d, D5d, D7d, C3h, C4h, C5h, C6h, S4, S6, S8, S10, S12]) then
    Result := Result and (CharacterTable.IReps[sym].eta = CharacterTable.IReps[sym1].eta);
  if not (PointGroup in [C3v, C4v, C5v, C6v, C7v, C8v]) then { R- Operation }
    Result := Result and (CharacterTable.IReps[sym].rho = CharacterTable.IReps[sym1].rho);
end;

procedure DetermineStatisticalWeightGroups(PointGroup: TPointGroup);
var
  sym, sym1, nGroups: Integer;
begin
  with CharacterTables[PointGroup] do begin
    SetLength(Group, Size);
    SetLength(GroupNames, Size); { Initial guess - contract later }
    nGroups := 0;
    FillChar(Group[0], SizeOf(Group), 0);
    for sym := 0 to Size-1 do { For each irreducible representation }
      if Group[sym] = 0 then begin { if not assigned to statistical weight group }
        GroupNames[nGroups] := 'wt(' + Names[sym];
        Group[sym] := nGroups+1;
        for sym1 := sym+1 to Size-1 do
          if SameStatisticalWeightGroup(PointGroup, sym, sym1) then begin
            Group[sym1] := nGroups+1;
            GroupNames[nGroups] := GroupNames[nGroups] + ',' + Names[sym1];
          end;
        GroupNames[nGroups] := GroupNames[nGroups] + ')';
        Inc(nGroups);
      end;
    SetLength(GroupNames,nGroups);
//    WriteLn(GroupName, ': ', IReps[DipoleSym].g, ' ', Names[DipoleSym], ' : ');
//    for i := 0 to nGroups-1 do
//      Write(' ', GroupNames[i]);
//    Writeln;
  end;
end;

function FindIrep(PointGroup: TPointGroup; g, rho, eta:Integer):Integer;
var
  i: Integer;
begin
  with CharacterTables[PointGroup] do
    for i := 0 to Size-1 do
      if (Ireps[i].g = g) and (Ireps[i].rho = rho) and (Ireps[i].eta = eta) then begin
        Result := i;
        Exit;
      end;
   raise Exception.Create('FindIrep');
end;

procedure DetermineDipoleSymmetry(PointGroup: TPointGroup);
begin
  with CharacterTables[PointGroup] do
    if PointGroup in [C3v, C4v, C5v, C6v, C7v, C8v] then
      DipoleSym := FindIrep(PointGroup, 0, 1, 0)
    else if PointGroup in [D2d, D4d, D6d, D8d, S4, S8, S12] then
      DipoleSym := FindIrep(PointGroup, n div 2, 0, 0)
    else if PointGroup in [D3, D4, D5, D6, D7, D8] then
      DipoleSym := FindIrep(PointGroup, 0, 0, 0)
    else
      DipoleSym := FindIrep(PointGroup, 0, 0, 1);
end;

procedure AddTables;
var
  p: TPointGroup;
  {$DEFINE CHECK}
  {$IFDEF CHECK}
  procedure SetStatWtSets(PointGroup: TPointGroup; DipoleName: string; GroupsDef: array of PChar);
  var
    i, p, sym: Integer;
    s, Irep: string;
    function IrepIndex(Name: string): Integer;
    var
      j: Integer;
    begin
      with CharacterTables[PointGroup] do
        for j := 0 to Size-1 do
          if Names[j] = Name then begin
            Result := j;
            Exit;
          end;
      raise Exception.Create('SymT:3');
    end;
  begin
    with CharacterTables[PointGroup] do begin
      {$IFDEF CHECK}
        if Length(Group) <> Size then raise NotYet;
        if Length(GroupNames) <> Length(GroupsDef) then raise NotYet;
      {$ELSE}
        SetLength(Group, Size);
        SetLength(GroupNames, Length(GroupsDef));
        FillChar(Group[0], SizeOf(Group[0])*Size, 0);
      {$ENDIF}
      for i := 0 to High(GroupsDef) do begin
        s := GroupsDef[i];
        {$IFDEF CHECK}
          if GroupNames[i] <> 'wt(' + s + ')' then raise NotYet;
        {$ELSE}
          GroupNames[i] := 'wt(' + s + ')';
        {$ENDIF}
        while s <> '' do begin
          Irep := s;
          p := Pos(',',s);
          if p > 0 then begin
            Delete(Irep, p, Length(Irep));
            Delete(s,1,p);
          end else
            s := '';
          sym := IrepIndex(Irep);
          {$IFDEF CHECK}
          if Group[sym] <> i+1 then raise NotYet;
          {$ELSE}
          if Group[sym] <> 0 then
            raise Exception.Create('SymT:1')
          else
            Group[sym] := i+1;
          {$ENDIF}
        end;
      end;
      { Check for unset groups }
      for sym := 0 to Size-1 do
        if Group[sym] = 0 then
          raise Exception.Create('SymT:2');
      {$IFDEF CHECK}
      if DipoleSym <> IrepIndex(DipoleName) then raise NotYet;
      {$ELSE}
      DipoleSym := IrepIndex(DipoleName);
      {$ENDIF}
    end;
  end;
  {$ENDIF}

  procedure A(T: array of TCharacterTableEntry; Order:Integer; rhomult: Integer; gMult:Integer = -1; etamult: Integer = 0;
              ArhoJmult: Integer = 1; AHasSeparableDegeneracy: Boolean = False);
  var
    i: Integer;
  begin
    with CharacterTables[p] do begin
      GroupName := CharacterTableNames[p];
      Size := Length(T);
      n := Order;
      gKmult := gMult;
      rhoJmult := ArhoJmult;
      rhoKmult := rhomult;
      etaKmult := etamult;
      HasSeparableDegeneracy := AHasSeparableDegeneracy;
      Setlength(IReps, Size);
      Setlength(Names, Size+1);
      SetLength(En,Size);
      Names[Size] := nil;
      LongestSymName := 1;
      for i := 0 to Size-1 do begin
        Names[i] := PChar(T[i].Name);
        LongestSymName := Max(LongestSymName, Length(Names[i]));
        IReps[i] := T[i];
        if Names[i][0] = 'E' then
          if Names[i][1] in ['1'..'9'] then
            En[i] := StrToInt(Names[i][1])
          else
            En[i] := 1 { E = E1 }
        else
          En[i] := 0; { A/B }
      end;
      DetermineDipoleSymmetry(p);
      DetermineStatisticalWeightGroups(p);
      if p < High(TPointGroup) then
        Inc(p);
    end;
  end;
begin
  p := Low(TPointGroup);
  { Must be in ordinal order }
  A(C3vCharacterTable,3,1); A(C4vCharacterTable,4,1); A(C5vCharacterTable,5,1); A(C6vCharacterTable,6,1); A(C7vCharacterTable,7,1); A(C8vCharacterTable,8,1);
  A(D3hCharacterTable,3,0,-1,1); A(D4hCharacterTable,4,0); A(D5hCharacterTable,5,0,-1,1); A(D6hCharacterTable,6,0); A(D7hCharacterTable,7,0,-1,1); A(D8hCharacterTable,8,0);
  A(D2dCharacterTable, 4,0,-3);  A(D3dCharacterTable,3,0); A(D4dCharacterTable, 8,0,-5);  A(D5dCharacterTable,5,0);
  A(D6dCharacterTable,12,0,-7);  A(D7dCharacterTable,7,0); A(D8dCharacterTable,16,0,-9);
  A(C3hCharacterTable,3,0,-1,1,0,True);
  A(C4hCharacterTable,4,0,-1,0,0,True);
  A(C5hCharacterTable,4,0,-1,1,0,True);
  A(C6hCharacterTable,6,0,-1,0,0,True);
  A(S4CharacterTable,4,0,-1,0,0,True);
  A(S6CharacterTable,3,0,-1,0,0,True);
  A(S8CharacterTable,8,0,-1,0,0,True);
  A(S10CharacterTable,5,0,-1,0,0,True);
  A(S12CharacterTable,12,0,-1,0,0,True);
  A(D3CharacterTable,3,0); A(D4CharacterTable,4,0); A(D5CharacterTable,5,0); A(D6CharacterTable,6,0); A(D7CharacterTable,7,0); A(D8CharacterTable,8,0);
  {$IFDEF CHECK}
  SetStatWtSets(C3v, 'A2',  ['A1,A2', 'E']);
  SetStatWtSets(D3h, 'A1"', ['A1'',A1"', 'A2'',A2"', 'E'',E"']);
  SetStatWtSets(D6h, 'A1u', ['A1g,A1u', 'A2g,A2u', 'B1g,B1u', 'B2g,B2u', 'E1g,E1u', 'E2g,E2u']);
  {$ENDIF}
end;

{$IFDEF OFF}
procedure DumpSymTables;
var
  I: TPointGroup;
  J: Integer;
{ Used for generating documentation only }
begin
  for I := Low(TPointGroup) to High(TPointGroup) do
    with CharacterTables[I] do begin
      Write(Log, GroupName, ' ', IReps[DipoleSym].Name);
      for J := 0 to High(GroupNames) do
        Write(Log, ' ', GroupNames[J]);
      Writeln(Log);
    end;
end;
{$ENDIF}

function TSymmetricTop.AddOneJ(Eigenvalues: TEigenvalues; J, F, M: TQuantumNumber; Sym: Integer; HyperfineState:THyperfineBasis; Seffective: TQuantumNumber): Integer;
var
  N: TQuantumNumber;
  StatWt, LastStatWtGroup: Integer;
  StatWtGroup: SmallInt;
begin
  Result := -1;
  StatWtGroup := -1;
  N := Abs(J-Seffective);
  while N <= J+Seffective do begin
    StatWt := AddOneN(Eigenvalues,N,J,F,M,Sym,HyperfineState,Seffective, LastStatWtGroup);
    if StatWtGroup < 0 then begin
      StatWtGroup := LastStatWtGroup;
      Result := StatWt;
    end else if StatWtGroup <> LastStatWtGroup then
      raise Exception.Create('AddBasis: Inconsistent Statistical Weights in '+Name);
    Inc(N,2);
  end;
end;

function TSymmetricTop.AddOneN(Eigenvalues: TEigenvalues; N, J, F, M: TQuantumNumber; Sym: Integer; HyperfineState: THyperfineBasis; Seffective: TQuantumNumber; out StatWtGroup: Integer): Integer;
var
  EVsym, EVRSym: TCharacterTableEntry;
  l, K, Kinc: TQuantumNumber;
  CharacterTable: TCharacterTable;
  procedure Consider(K, l: Integer; ConsiderBoth: Boolean);
  var
    BasisState: TSymmetricTopBasisState;
    g, rho, eta, Phase: Integer;
  begin
    if (Kmin >= 0) and (Abs(K) < Kmin) then
      Exit;
    if (Kmax >= 0) and (Abs(K) > Kmax) then
      Exit;
    { Does EVsym * Sym(K) = EVRSym? }
    { En Match }
    g := Molecule.g(K, l, Symmetry);
    if Abs(g) <> EVRsym.g then Exit;
    { g/u '/'' Match }
    eta := K*CharacterTable.etaKmult div 2 + EVsym.eta;
    if Odd(eta+EVRSym.eta) then Exit;
    if CharacterTable.HasSeparableDegeneracy then begin
      { Point group operations do not change sign of K, so need to explicitly
        include, say K=1 and K=-1 for non E functions }
      Phase := 0;
      if ((K <> 0) or (l <> 0)) and ((g=0) or (2*g = CharacterTable.n)) then begin
        if Molecule.AlternateBasis then begin
          { Take |K,l> and |-K,-l> as basis functions }
          if ConsiderBoth then
            Consider(-K,-l,False); { Add other K function }
        end else
          { To give better match with C3v style calculations, take
            |K,l> +/- |-K,-l> as basis functions for non E functions;
            this makes K=1 labels in particular more meaningful }
          if ConsiderBoth then begin
            Phase := -1;
            Consider(K,l,False);
          end else
            Phase := 1;
      end else if g < 0 then begin
        { E symmetry - see comments below }
        K := -K;
        l := -l;
      end;
    end else if (g = 0) or (2*g = CharacterTable.n) then begin
      { A1 or A2 / B1 or B2 }
      rho := (CharacterTable.rhoJmult*N+K*CharacterTable.rhoKmult) div 2 + EVsym.rho;
      if Odd(rho+EVRSym.rho) then begin
        Phase := -1;
        if (l=0) and (K=0) then
          Exit; { |J00>-|J00> }
      end else begin
        Phase := 1;
        if (l=0) and (K=0) then
          Phase := 0; { |J00>+|J00> }
      end;
    end else begin
      { E - two levels, but only work out the energy of one of them }
      Phase := 0;
      { In general, choose the basis state with g > 0, preserving sign of kl }
      { There should be no Hamiltonian matrix elements off diagonal in g }
      { This also fixes transition matrix elements (apart from D2nd, S4n)
        as these are also diagonal in G. For D2nd, S4n the dipole transition moment changes g
        by n (or n/2 for S4n), so special action is required }

      { Special action  for D2d, D4d, D6d, D8d, S4, S8, S12 as Transition moment changes G by n }
      { Flipping the sign of G for B states or the unique E state means that dipole transitions
        are calculated with the correct combination of K values; this is handled by conditional
        above; here just ensure g is positive. }
      if g < 0 then begin
        K := -K;
        l := -l;
      end;
    end;
    BasisState := TSymmetricTopBasisState.Create(Eigenvalues,F,M,Sym);
    BasisState.State := Self;
    BasisState.J := J;
    BasisState.N := N;
    BasisState.K := K;
    BasisState.l := l;
    BasisState.Phase := Phase;
    BasisState.SubBasisIndex := Eigenvalues.LastSubBasisIndex;
    { For l <> 0, split into two groups. Ordering, is determined by:
        (C-B)K^2 - 2CzetaKl
      There are two sequences with guaranteed order, corresponding to l = +/- 1,
      but the ordering between them is not guaranteed, so split into two sub-bases.
      Tricky part is K = 0, as it will be on the same sequence as states with (C-B)zetaKL < 0
    }
    if (K <> 0) and (l <> 0) then
      if (K*l > 0) xor (zeta >= 0) xor Prolate then
        { This form of the test forces the split when zeta = 0 }
        Inc(BasisState.SubBasisIndex);
    BasisState.HyperfineState := HyperfineState;
    BasisState.S := Seffective;
    BasisState.StatWtGroup := StatWtGroup;
    BasisState.StatisticalWeight := Result;
  end;
begin
  { K=2 missing for C4v }
  CharacterTable := Molecule.CharacterTable;
  StatWtGroup := CharacterTable.Group[Sym];
  Result := Molecule.GetSetting(WTI+StatWtGroup-1);
  if Result = 0 then
    Exit;
  EVsym := CharacterTable.IReps[Symmetry];
  if EVsym.Name[1] = 'E' then
    l := 2
  else
    l := 0;
  EVRsym := CharacterTable.IReps[Sym];
  if Prolate then begin
    K := 0;
    Kinc := 2
  end else begin
    Kinc := -2;
    K := N;
  end;
  while (K <= N) and (K >= 0) do begin
    Consider(K,l,True);
    if (l<>0) and (K>0) then
      Consider(-K,l,True);
    Inc(K,Kinc);
  end;
  Inc(Eigenvalues.LastSubBasisIndex);
  if l <> 0 then
    Inc(Eigenvalues.LastSubBasisIndex);
end;

function TSymmetricTop.GetSettingsNames(AIndex: Integer; out LowestValue: Integer): PPChar;
begin
  Result := inherited GetSettingsNames(AIndex,LowestValue);
  if AIndex = LSI then
    if Molecule = nil then { Should never happen }
      Result := @CharacterTables[D3h].Names[0]
    else
      Result := @CharacterTables[Molecule.PointGroup].Names[0];
end;

function TSymmetricTop.AllowedSphericalComponent(
  TransitionMoment: TSphericalTransitionMoment; State: TState;
  out Component: TQuantumNumber): Exception;
var
  Ket: TSymmetricTop;
  q, qMatch: TQuantumNumber;
begin
  Ket := State as TSymmetricTop;
  if S <> Ket.S then
    Result := EReportException.Create(TransitionMoment.UniqueName+': Spin forbidden transition - component must be specified')
  else if TransitionMoment is TMagneticTransitionMoment then
    Result := EReportException.Create(TransitionMoment.UniqueName+': auto not implemented for magnetic dipole transitions')
  else begin
    q := 0;
    qMatch := -1;
    Result := nil;
    while q <= TransitionMoment.Rank do begin
      if Molecule.VibronicallyAllowed(TransitionMoment, q, Symmetry, Ket.Symmetry) = vaYes then
        if qMatch >= 0 then begin
          if not Assigned(Result) then
            Result := EReportException.Create(TransitionMoment.UniqueName+': More than one possible component for an allowed transition - please set component to ' + QuantumNumberToStr(qMatch,qftJ));
          Result.Message := Result.Message + ' or ' + QuantumNumberToStr(q,qftJ);
        end else
          qMatch := q;
      Inc(q,2);
    end;
    if not Assigned(Result) then
      if qMatch >= 0 then
        Component := qMatch
      else
        Result := EReportException.Create(TransitionMoment.UniqueName+': No allowed component');
  end;
end;

const
  Equivs: array [0..35] of TPertEquiv = (
   (Name:'Origin';JPower:0; zPower: 0; pPower:0; lPower: 0; lChange: 0; SPower: 0; SzPower: 0; SpPower: 0; NSPower: 0; Scale4: 4;),
   (Name:'B';     JPower:2; zPower: 0; pPower:0; lPower: 0; lChange: 0; SPower: 0; SzPower: 0; SpPower: 0; NSPower: 0; Scale4: 4;),
   (Name:'B';     JPower:0; zPower: 2; pPower:0; lPower: 0; lChange: 0; SPower: 0; SzPower: 0; SpPower: 0; NSPower: 0; Scale4:-4;),
   (Name:'C';     JPower:0; zPower: 2; pPower:0; lPower: 0; lChange: 0; SPower: 0; SzPower: 0; SpPower: 0; NSPower: 0; Scale4: 4;),
   (Name:'DJ';    JPower:4; zPower: 0; pPower:0; lPower: 0; lChange: 0; SPower: 0; SzPower: 0; SpPower: 0; NSPower: 0; Scale4:-4;),
   (Name:'DJK';   JPower:2; zPower: 2; pPower:0; lPower: 0; lChange: 0; SPower: 0; SzPower: 0; SpPower: 0; NSPower: 0; Scale4:-4;),
   (Name:'DK';    JPower:0; zPower: 4; pPower:0; lPower: 0; lChange: 0; SPower: 0; SzPower: 0; SpPower: 0; NSPower: 0; Scale4:-4;),
   (Name:'HJ';    JPower:6; zPower: 0; pPower:0; lPower: 0; lChange: 0; SPower: 0; SzPower: 0; SpPower: 0; NSPower: 0; Scale4: 4;),
   (Name:'HJK';   JPower:4; zPower: 2; pPower:0; lPower: 0; lChange: 0; SPower: 0; SzPower: 0; SpPower: 0; NSPower: 0; Scale4: 4;),
   (Name:'HKJ';   JPower:2; zPower: 4; pPower:0; lPower: 0; lChange: 0; SPower: 0; SzPower: 0; SpPower: 0; NSPower: 0; Scale4: 4;),
   (Name:'HK';    JPower:0; zPower: 6; pPower:0; lPower: 0; lChange: 0; SPower: 0; SzPower: 0; SpPower: 0; NSPower: 0; Scale4: 4;),
   (Name:'LJ';    JPower:8; zPower: 0; pPower:0; lPower: 0; lChange: 0; SPower: 0; SzPower: 0; SpPower: 0; NSPower: 0; Scale4: 4;),
   (Name:'LJJK';  JPower:6; zPower: 2; pPower:0; lPower: 0; lChange: 0; SPower: 0; SzPower: 0; SpPower: 0; NSPower: 0; Scale4: 4;),
   (Name:'LJK';   JPower:4; zPower: 4; pPower:0; lPower: 0; lChange: 0; SPower: 0; SzPower: 0; SpPower: 0; NSPower: 0; Scale4: 4;),
   (Name:'LKKJ';  JPower:2; zPower: 6; pPower:0; lPower: 0; lChange: 0; SPower: 0; SzPower: 0; SpPower: 0; NSPower: 0; Scale4: 4;),
   (Name:'LK';    JPower:0; zPower: 8; pPower:0; lPower: 0; lChange: 0; SPower: 0; SzPower: 0; SpPower: 0; NSPower: 0; Scale4: 4;),
   (Name:'zeta';  JPower:0; zPower: 1; pPower:0; lPower: 1; lChange: 0; SPower: 0; SzPower: 0; SpPower: 0; NSPower: 0; Scale4:-8;),
   (Name:'etaJ';  JPower:2; zPower: 1; pPower:0; lPower: 1; lChange: 0; SPower: 0; SzPower: 0; SpPower: 0; NSPower: 0; Scale4: 4;),
   (Name:'etaK';  JPower:0; zPower: 3; pPower:0; lPower: 1; lChange: 0; SPower: 0; SzPower: 0; SpPower: 0; NSPower: 0; Scale4: 4;),
   (Name:'qplus'; JPower:0; zPower: 0; pPower:2; lPower: 0; lChange: 2; SPower: 0; SzPower: 0; SpPower: 0; NSPower: 0; Scale4: 2;),
   (Name:'DqJ';   JPower:2; zPower: 0; pPower:2; lPower: 0; lChange: 2; SPower: 0; SzPower: 0; SpPower: 0; NSPower: 0; Scale4: 2;),
   (Name:'DqK';   JPower:0; zPower: 2; pPower:2; lPower: 0; lChange: 2; SPower: 0; SzPower: 0; SpPower: 0; NSPower: 0; Scale4: 2;),
   (Name:'qminus';JPower:0; zPower: 0; pPower:2; lPower: 0; lChange:-2; SPower: 0; SzPower: 0; SpPower: 0; NSPower: 0; Scale4: 2;),
   (Name:'r';     JPower:0; zPower: 1; pPower:1; lPower: 0; lChange:-2; SPower: 0; SzPower: 0; SpPower: 0; NSPower: 0; Scale4: 4;),
   (Name:'DrJ';   JPower:2; zPower: 1; pPower:1; lPower: 0; lChange:-2; SPower: 0; SzPower: 0; SpPower: 0; NSPower: 0; Scale4: 4;),
   (Name:'DrK';   JPower:0; zPower:-3; pPower:1; lPower: 0; lChange:-2; SPower: 0; SzPower: 0; SpPower: 0; NSPower: 0; Scale4: 4;),
   (Name:'ebb';   JPower:0; zPower: 0; pPower:1; lPower: 0; lChange: 0; SPower: 0; SzPower: 0; SpPower:-1; NSPower: 0; Scale4: 1;),
   (Name:'ecc';   JPower:0; zPower: 1; pPower:0; lPower: 0; lChange: 0; SPower: 0; SzPower: 1; SpPower: 0; NSPower: 0; Scale4: 2;),
   (Name:'DsN';   JPower:2; zPower: 0; pPower:0; lPower: 0; lChange: 0; SPower: 0; SzPower: 0; SpPower: 0; NSPower: 1; Scale4: 4;),
   (Name:'DsNK';  JPower:2; zPower: 1; pPower:0; lPower: 0; lChange: 0; SPower: 0; SzPower: 1; SpPower: 0; NSPower: 0; Scale4: 2;),
   (Name:'DsKN';  JPower:0; zPower: 2; pPower:0; lPower: 0; lChange: 0; SPower: 0; SzPower: 0; SpPower: 0; NSPower: 1; Scale4: 4;),
   (Name:'DsK';   JPower:0; zPower: 3; pPower:0; lPower: 0; lChange: 0; SPower: 0; SzPower: 1; SpPower: 0; NSPower: 0; Scale4: 2;),
   (Name:'alpha'; JPower:0; zPower: 0; pPower:0; lPower: 0; lChange: 0; SPower: 2; SzPower: 0; SpPower: 0; NSPower: 0; Scale4:-4;),
   (Name:'alpha'; JPower:0; zPower: 0; pPower:0; lPower: 0; lChange: 0; SPower: 0; SzPower: 2; SpPower: 0; NSPower: 0; Scale4:12;),
   (Name:'beta';  JPower:0; zPower: 0; pPower:0; lPower: 0; lChange: 0; SPower: 0; SzPower: 0; SpPower: 2; NSPower: 0; Scale4: 2;),
   (Name:'aeff';  JPower:0; zPower: 0; pPower:0; lPower: 1; lChange: 0; SPower: 0; SzPower: 1; SpPower: 0; NSPower: 0; Scale4: 4;)
  );

function FindPertEquiv(const AName: string; out PE1: TPertEquiv):TPertEquiv;
var
  i: Integer;
begin
  Result.Name := '';
  for i := Low(Equivs) to High(Equivs) do
    if CompareText(Equivs[i].Name,AName) = 0 then begin
      Result := Equivs[i];
      if (i < High(Equivs)) and (CompareText(Equivs[i+1].Name,AName) = 0) then
        PE1 := Equivs[i+1] { Second parameter in constrained pair }
      else
        PE1.Name := '';
      Exit;
    end;
end;

function FindParamEquiv(const PE: TPertEquiv; out OneOfPair: Boolean; out Scale: Double):string;
var
  i: Integer;
  PE1: TPertEquiv;
begin
  { Given PertEquivalent, find equivalent parameter, if any }
  OneOfPair := False;
  for i := Low(Equivs) to High(Equivs) do begin
    PE1 := Equivs[i];
    PE1.Name := PE.Name; { To allow easy compare }
    PE1.Scale4 := PE.Scale4;
    if CompareMem(@PE, @PE1, SizeOf(PE)) then begin
      Result := Equivs[i].Name;
      Scale := PE.Scale4/Equivs[i].Scale4;
      OneOfPair := ((i < High(Equivs)) and (CompareText(Equivs[i+1].Name,Result) = 0))
                   or ((i > 0) and (CompareText(Equivs[i-1].Name,Result) = 0));
      Break;
    end;
  end;
end;


function TSymmetricTop.ConvertParameter(AIndex: Integer; Zero: Boolean): TPerturbation;
var
  NewValue: Double;

  procedure SetUpPerturbation(PE: TPertEquiv);
  var
    P: TSymmetricTopPerturbation;
  begin
    p := TSymmetricTopPerturbation.Create(Owner);
    with PE do begin
      if Scale4 = -4 then
        p.Comment := '-'+Name
      else if Scale4 <> 4 then
        p.Comment := FloatToStr(Scale4/4)+Name
      else
        p.Comment := Name;
      p.Value := Scale4*NewValue/4;
    end;
    p.SetFromPertEquiv(PE);
    p.SetBraKetName(Name,Name);
    Result := p;
  end;

var
  PE, PE1: TPertEquiv;
begin
  Result := nil;
  PE := FindPertEquiv(ParameterNames[AIndex], PE1);
  NewValue := Parameters[AIndex];
  if PE.Name = 'zeta' then
    if C = 0 then
      PE.Name := '' { Can only convert if C is present }
    else
      NewValue := zeta*C;
  if PE.Name <> '' then begin
    SetUpPerturbation(PE);
    if PE1.Name <> '' then begin
      SetUpPerturbation(PE1);
      Result.ScalePrev := Round(PE1.Scale4/PE.Scale4);
      if not SameValue(Result.ScalePrev,PE1.Scale4/PE.Scale4) then
        raise NotYet; { Must be integer ratio between constrained parameters }
    end;
    if Zero then
      Parameters[AIndex] := 0;
  end;
end;

procedure TSymmetricTop.ConvertToPerturbations(Nuclei: Boolean);
begin
  { zeta is actually C*zeta in Hamiltonian, so must convert before C }
  if zeta <> 0 then
    ConvertParameter(IndexOfParameter('zeta'),true);
  inherited;
end;

constructor TSymmetricTop.Create(AOwner: TComponent);
begin
  if Assigned(AOwner) then
    FMolecule := AOwner.Owner as TSymmetricMolecule;
  SavedProperties := SavedProperties + ',Kmin=all,Kmax=all';
  DimensionlessParameters := DimensionlessParameters + ',zeta';
  inherited;
end;

function TSymmetricTop.FixupProperties: Exception;
var
  i: Integer;
begin
  { Force legal symmetry, in case point group changed }
  Symmetry := Min(Symmetry, Molecule.LastSym);
  FProlate := C > B; { Precomputation - not a speed up, but work(ed) round a FPC PPC bug }
  FFullRotationalSymmetry := not Molecule.CharacterTable.HasSeparableDegeneracy;
  { Find if any Quartic Spin Rotation }
  QuarticSpinRotation := False;
  for i := IndexOfParameter('DsN') to IndexOfParameter('DsK') do
    QuarticSpinRotation := QuarticSpinRotation or (GetParameter(i) <> 0);
  TSR[0] := -(2*ebb+ecc)/sqrt(Three);
  TSR[1] := 0;
  TSR[2] := (2*ecc - 2*ebb)/Sqrt(Six);
  VList.CheckThreadVars;
  Result:=inherited FixupProperties;
end;

{ TSymmetricTopBasisState }

function TSymmetricTopBasisState.BasisLabel: string;
begin
  Result := inherited BasisLabel + ' ';
  if State.S <> 0 then
    Result := Result + State.Manifold.Mixture.QuantumNumberToStr(N) + ' ';
  Result := Result + State.Manifold.Mixture.QuantumNumberToStr(K);
  if l <> 0 then
    Result := Result + Format(' %s', [State.Manifold.Mixture.QuantumNumberToStr(l)]);
  Result := Result + HyperfineStateLabel;
end;

function TSymmetricTopBasisState.BasisLabelKey: string;
begin
  Result := inherited BasisLabelKey;
  if State.S <> 0 then
    Result := Result + ' N';
  Result := Result + '  K l' + HyperfineStateLabelKey;
end;

function TSymmetricTopBasisState.klLabel(out ExtraTag: string): string;
begin
  if l <> 0 then begin
    if K = 0 then
      Result := '  '
    else begin
      //kl := K*l;
      //CharacterTable := TSymmetricTop(State).Molecule.CharacterTable;
      //if CharacterTable.IReps[TSymmetricTop(State).Symmetry].g
      //   mod 3 = 2 then
      //  kl := -kl;
      { this is sgn(k*sin(2 pi Gve/n)) as per Hoy and Mills J Mol Spec 1973 46, 333 }
      if K*l > 0 then
        Result := '+1'
      else
        Result := '-1'
    end;
  end else
    Result := '';
  with TSymmetricTop(State) do
    if Molecule.CharacterTable.HasSeparableDegeneracy then
      if Molecule.AlternateBasis then begin
        with Molecule.CharacterTable.IReps[Sym] do
          if (K <> 0) and ((g = 0) or (2*g = Molecule.CharacterTable.n)) then
            if K > 0 then
              ExtraTag := '+'
            else
              ExtraTag := '-'
          else
            ExtraTag := ' ';
      end else begin
        if Phase > 0 then
          ExtraTag := '+'
        else if Phase < 0 then
          ExtraTag := '-'
        else
          ExtraTag := ' ';
      end
    else
      ExtraTag := '';
end;

procedure TSymmetricTopBasisState.SetSymbolicVariables(SignA: Integer; Ket: TBasisState; SignB: Integer; MNode: TMNode);
var
  KetK: TSymmetricTopBasisState;
  TransitionMoment: TTransitionMoment;
  Molecule: TSymmetricMolecule;
begin
  KetK := Ket as TSymmetricTopBasisState;
  if MNode is TTransitionMoment then begin
    { Special case where the signs of the quantum numbers of one of the states are flipped
      before evaluating the matrix elements. See comments in TSymmetricTop.RotationalTransitionMatrixElement
      for rationale. }
    TransitionMoment := TTransitionMoment(MNode);
    Molecule := TransitionMoment.AncestorWithClass(TSymmetricMolecule) as TSymmetricMolecule;
    if not TransitionMoment.Magnetic and (Minus(TransitionMoment.Rank) < 0)
       and (Molecule.CharacterTable.IReps[Molecule.CharacterTable.DipoleSym].g > 0)  { B Symmetry dipole rather than A }
       and (Phase = 0) and (KetK.Phase = 0)
       and Molecule.IsDegenerate(Sym)
       and Molecule.IsDegenerate(Ket.Sym) then
      SignB := -SignB; { Choose right combination of signs of l for E - E transitions }
  end;
  eJ.Value := Ket.J / 2;
  eN.Value := Ket.N / 2;
  eK.Value := KetK.K*SignB / 2;
  el.Value := KetK.l*SignB / 2;
  eS.Value := Ket.S / 2;
end;

function TSymmetricTopBasisState.MakeSignature(SignA: Integer; Ket: TBasisState; SignB: Integer; MNode: TMNode; Symmetrised: Boolean): TIntegerArray;
var
  KetK: TSymmetricTopBasisState;
  TransitionMoment: TTransitionMoment;
  Molecule: TSymmetricMolecule;
begin
  KetK := Ket as TSymmetricTopBasisState;
  if S > 0 then begin
    SetLength(Result,4);
    Result[3] := (J-KetK.J) div 2;
  end else
    SetLength(Result,3);
  if MNode is TTransitionMoment then begin
    { Special case where the signs of the quantum numbers of one of the states are flipped
      before evaluating the matrix elements. See comments in TSymmetricTop.RotationalTransitionMatrixElement
      for rationale. }
    TransitionMoment := TTransitionMoment(MNode);
    Molecule := TransitionMoment.AncestorWithClass(TSymmetricMolecule) as TSymmetricMolecule;
    if not TransitionMoment.Magnetic and (Minus(TransitionMoment.Rank) < 0)
       and (Molecule.CharacterTable.IReps[Molecule.CharacterTable.DipoleSym].g > 0)  { B Symmetry dipole rather than A }
       and (Phase = 0) and (KetK.Phase = 0)
       and Molecule.IsDegenerate(Sym)
       and Molecule.IsDegenerate(Ket.Sym) then
      SignB := -SignB; { Choose right combination of signs of l for E - E transitions }
    Result[0] := 1000+((l*SignA+2)*10+KetK.l*SignB+2) div 2; { Encoding: Transition moments 1000 + both values of l+1 }
  end else { Note also different signature encoding for Transition moments }
    Result[0] := (l*SignA-KetK.l*SignB) div 2;              { Otherwise change in l as for othe quantum numbers }
  Result[1] := (K*SignA-KetK.K*SignB) div 2;
  Result[2] := (N-KetK.N) div 2;
end;

function TSymmetricTopBasisState.Qno(QnoType: TQnoType): TQuantumNumber;
var
  ExtraTag: string;
begin
  case QnoType of
    qtA: Result := Abs(K);
    qtB: begin
           Result := 0;
           if l <> 0 then
             if K*l >= 0 then
               Result := 2
             else
               Result := -2;
           with TSymmetricTop(State) do
             if Molecule.CharacterTable.HasSeparableDegeneracy then begin
               klLabel(ExtraTag);
               { Encode extra sign in higher bits }
               if ExtraTag = '+' then
                 Result := (Result and not 12) or 4
               else if ExtraTag = '-' then
                 Result := (Result and not 12) or 8
             end;
        end
  else
    Result := inherited Qno(QnoType);
  end;
end;

function TSymmetricTopBasisState.StateLabel(AQuantumNumberFormat: TQuantumNumberFormat): string;
var
  kl, ExtraTag: string;
  Molecule: TSymmetricMolecule;
  KString: {$IFDEF FPC}ShortString{$ELSE}string{$ENDIF};
  Klength: Integer;
begin
  Molecule := (State as TSymmetricTop).Molecule;
  kl := kllabel(ExtraTag);
  if Molecule.CompactKLabels then begin
    Klength := 3+Molecule.CharacterTable.LongestSymName;
    KString := QuantumNumberToStr(Abs(K),AQuantumNumberFormat);
    if K*l <> 0 then
      KString := kl[1] + KString;
    if Molecule.CharacterTable.HasSeparableDegeneracy then begin
      Inc(Klength);
      KString := ExtraTag + KString;
    end;
    if Molecule.SymmetryLabels = Never then
      Dec(Klength, Molecule.CharacterTable.LongestSymName)
    else
      if (Molecule.SymmetryLabels = Always) or (not Molecule.IsDegenerate(Sym) and (K <> 0)) then
        KString := KString + PadRight(Molecule.SymName(Sym),Molecule.CharacterTable.LongestSymName)
      else
        KString := KString + StringOfChar(' ',Molecule.CharacterTable.LongestSymName)
  end else begin
    KString := QuantumNumberToStr(Abs(K),AQuantumNumberFormat);
    Klength := 2;
  end;
  Result := Format('%s %2s %*s', [inherited StateLabel(AQuantumNumberFormat),
                                  QuantumNumberToStr(N,AQuantumNumberFormat),
                                  Klength, KString]);
  if not Molecule.CompactKLabels then begin
    Result := Result + ExtraTag;
    if l<>0 then
      Result := Result + ' ' + kl;
  end;
  if State.S <> 0 then
    Result := Result + Format(' %*s', [2+Ord(Odd(J))*2, QuantumNumberToStr(J,AQuantumNumberFormat)]);
  if not Molecule.CompactKLabels then
    Result := Result + Format(' %-*s', [Molecule.CharacterTable.LongestSymName, Molecule.SymName(Sym)]);
  Result := Result + HyperfineStateLabel;
end;

function TSymmetricTopBasisState.StateLabelKey: string;
begin
  Result := inherited StateLabelKey;
  if (State as TSymmetricTop).Molecule.CompactKLabels then
   if State.S <> 0 then
     Result := Result + ' N Ksym J'
   else
     Result := Result + ' J Ksym'
  else
    if State.S <> 0 then
      Result := Result + ' N K (kl) J Sym'
    else
      Result := Result + ' J K (kl) Sym';
  Result := Result + HyperfineStateLabelKey;
end;

function TSymmetricTopBasisState.SymbolicVariableValues: string;
begin
  Result := Vlist.VariableValues;
end;

function TSymmetricTopBasisState.TrackIndex: Integer;
begin
  Result := Abs(K); { TODO: May want to include l also }
end;

function TSymmetricTopBasisState.TxName(Upper, Lower: TBasisState): string;
var
  Molecule: TSymmetricMolecule;
  U, LL: TSymmetricTopBasisState;
  Utag, LTag, ExtraTag: string;
  fU, fL: string;
  function Fnumber(State: TSymmetricTopBasisState):string;
  var
    S: TQuantumNumber;
  begin
    S := State.State.S;
    if S > 0 then
      Result := IntToStr((State.N - State.J + S + 2) div 2)
    else
      Result := '';
  end;
begin
  Molecule := State.Manifold.Molecule as TSymmetricMolecule;
  U := Upper as TSymmetricTopBasisState;
  LL := Lower as TSymmetricTopBasisState;
  Result := LowerCase(DeltaJName(Abs(U.K)-Abs(LL.K)))+DeltaJName(U.J-LL.J);
  if (U.State.S > 0) or (LL.State.S > 0) then begin
    { Add spin label }
    fU := Fnumber(U);
    fL := Fnumber(LL);
    Result := Result + fU;
    if fU <> fL then { Add second spin label if different }
      Result := Result + fL;
    Result := Result + ',';
  end;
  if Molecule.CharacterTable.HasSeparableDegeneracy then begin
    U.kllabel(Utag);
    LL.klLabel(LTag);
    if Utag = Ltag then
      ExtraTag := Utag
    else
      ExtraTag := UTag+LTag;
    Result := Result + Trim(ExtraTag);
  end;
  if LL.l <> 0 then
    Result := Result + Trim(Copy(LL.klLabel(LTag),1,1) { Only take sign of kl label }
                     + Molecule.QuantumNumberToStr(Abs(LL.K)))
  else
    Result := Result + Molecule.QuantumNumberToStr(Abs(LL.K));
  if (Molecule.SymmetryLabels = Always) or ((Molecule.SymmetryLabels = Required) and (LL.K <> 0) and (U.K <> 0) and not Molecule.IsDegenerate(LL.Sym)) then
    Result := Result + Molecule.SymName(LL.Sym);
  Result := Result + Format('(%s)',[Molecule.QuantumNumberToStr(LL.N)]);
  if (Length(Lower.HyperfineState) > 0) or (Owner as TEigenvalues).HasFields then
    Result := Result + Upper.HyperfineStateLabel(False) + ',' + Lower.HyperfineStateLabel(False);
end;

function TSymmetricTopBasisState.TxNameKey(Upper, Lower: TBasisState): string;
begin
  Result := 'delta|K| deltaN';
  if Upper.State.Manifold.MaxSpin > 0 then
    Result := Result + ' Fn''';
  if Lower.State.Manifold.MaxSpin > 0 then
    Result := Result + ' Fn",';
  Result := Result + ' K"(';
  if Upper.State.Manifold.MaxSpin + Lower.State.Manifold.MaxSpin > 0 then
    Result := Result + 'N'
  else
    Result := Result + 'J';
  Result := Result + '")' + HyperfineTxNameKey(False);
end;

function TSymmetricTop.HWidth(J, Kp, lp, K, l: TQuantumNumber): Double;
var
  JSq, KSq: Double;
begin
  if (K = Kp) and (l = lp) then begin
    JSq := J*(J+2) / 4;
    KSq := K*K / 4;
    Result := Width + wJ*JSq + (wK-wJ)*KSq
  end else
    Result := 0;
end;

function TSymmetricTop.HyperfineMatrixElement(ABra, AKet: TBasisState; Signp, Sign: Integer; MNode: TMNode; RankWanted: TQuantumNumber): Complex;
{ TODO: Abstract more from asymmetric top equivalent }
var
  Bra, Ket: TSymmetricTopBasisState;
  Nuc: TSymmetricTopNucleus;
  JP, J, Np, N, Kp, K, lp, l, Sp, S: TQuantumNumber;
  t, Sum: Double;
  Rank: Integer;
begin
  Bra := ABra as TSymmetricTopBasisState;
  Ket := AKet as TSymmetricTopBasisState;
  Nuc := MNode as TSymmetricTopNucleus;
  JP := Bra.J;
  J := Ket.J;
  NP := Bra.N;
  N := Ket.N;
  KP := SignP*Bra.K;
  K := Sign*Ket.K;
  lp := Signp*Bra.l;
  l := Sign*Ket.l;

  Sp := Bra.S;
  S := Ket.S;
  Result := z0;
  if Abs(K-Kp) > 4 then
    Exit;
  { Section below mainly copoed from OAsymT }
  if (RankWanted = 2) and (l=lp) then begin
    { Magnetic Dipole }
    { TODO }
    { Nuclear Spin - Rotation }
    Sum := 0;
    for Rank := 0 to 1 do begin { Actually a flag: Rank = 1 is k=2 }
      if K = Kp then begin
        if Rank = 0 then
          t := Nuc.TC00
        else
          t := Nuc.TC20; { Note T22 component missing for a symmetric top }
        Sum := Sum
               + t * Sqrt(4*Rank+One) * Minus(Np-Kp)
                 * threej(Np, -Kp, Rank*4, Kp-K, N, K)
                 *(ReducedMatrixElement(2,N)*sixj(Rank*4, 2, 2, N, Np, N)
                   +ReducedMatrixElement(2,Np)*sixj(Rank*4, 2, 2, Np, N, Np))
      end;
    end;
    if Sum <> 0 then
      Result.re := Result.re - Sum / 2 * sixj(N, J, Sp, Jp, Np, 2)
                               * Sqrt((J+1)*(JP+1)*(N+One)*(Np+One))
                         * Minus(N+Sp+J) * ReducedMatrixElement(2, Nuc.Spin);
  end else if RankWanted = 4 then begin
    { Quadrupole }
    if (Abs(K-Kp) <= 4) and ((K-Kp) = (l-lp)) then begin
      t := sixj(NP, JP, Sp, J, N, 4)* threej(Np, -Kp, 4, Kp-K, N, K);
      Result.re := Result.re + t * Sqrt((N+1)*(NP+One)) / 2
                               * Minus(NP+S+J+Np-Kp) * Nuc.CHI[(Kp-K) div 2];
      Result.re := Result.re * Sqrt((J+1)*(JP+One)) * Nuc.QuadI;
    end;
  end;
  if IsNan(Result.re) then
    raise Exception.Create('Help');
end;

function TSymmetricTop.HRot(J, Kp, lp, K, l: TQuantumNumber): Double;
var
  JSq, KSq: Double;
begin
  if (K = Kp) and (l = lp) then begin
    JSq := J*(J+2) / 4;
    KSq := K*K / 4;
    Result := Origin + B*JSq + (C-B)*KSq
              - DJ*Sqr(JSq) - DJK*JSq*KSq - DK*Sqr(KSq)
              + HJ*Sqr(JSq)*JSq + HJK*Sqr(JSq)*KSq + HKJ*Jsq*Sqr(KSq) + HK*Sqr(KSq)*KSq;
    Result := Result + LJ*Sqr(JSq)*Sqr(JSq) + LJJK*Sqr(JSq)*JSq*KSq + LJK*Sqr(Jsq)*Sqr(KSq) + LKKJ*Sqr(KSq)*KSq*JSq + LK*Sqr(KSq)*Sqr(KSq);
    Result := Result + (-2*C*zeta + etaj*JSq + etak*KSq) * l*K/4;
  end else if (Abs(K-Kp) = 4) and (K-Kp = l-lp) then begin
    JSq := J*(J+2) / 4;
    KSq := K*K / 4;
    { Note that NH3 Rydberg state work used opposite sign convention }
    Result := (qplus + DqJ*JSq + DqK*(Sqr(Kp)/4+KSq))/2;
    Result := Result * Sqrt((JSq - K*(K+Kp)/8)*(JSq - Kp*(K+Kp)/8));
  end else if (Abs(K-Kp) = 4) and (K-Kp = -(l-lp)) then begin
    JSq := J*(J+2) / 4;
    Result := qminus/2;
    Result := Result * Sqrt((JSq - K*(K+Kp)/8)*(JSq - Kp*(K+Kp)/8));
  end else if (Abs(K-Kp) = 2) and (2*(K-Kp) = -(l-lp)) then begin
    JSq := J*(J+2) / 4;
    Result := r + DrJ*JSq + DrK*(Sqr(Kp)+Sqr(K))/4;
    Result := Result * Sqrt((JSq - K*Kp/4)) * (K+Kp) / 2;
  end else
    Result := 0;
end;

function TSymmetricTop.HSR(J, Np, Kp, lp, N, K, l:TQuantumNumber): Complex;
{ Almost verbaitim copy of asymmetric top fucnction with delta K <> 0 terms removed }
var
  i, rank, q: Integer;
  rr, r1, Ksq, Nsq, V: Double;
  NN: TQuantumNumber;
  function Gamma(N, S, J: TQuantumNumber):Double;
  begin
    Result := (J*(J+2) - N*(N+2) - S*(S+2))/4;
  end;
  function phi(N,S,J:TQuantumNumber):Double;
  begin
    Result := (N - J + S);
    Result := Result * (N + J + S + 2) * (S + J - N + 2);
    Result := Sqrt(Result * (N + J - S) / (N - One) / (N + 1) / 16);
  end;
begin
  Result := z0;
  if (Abs(Np-N) > 4) or (l <> lp) then Exit; { TODO: l <> lp operators? }
  if S > 1 then begin
    { Matrix elements of Spin-Spin Coupling from Hirota eq 2.3.36, 2.3.37 }
    { Adapted taking into account equation 38 of T.J Sears, Comp. Phys. Rep, 2, 1 (1984) }
    { The reduced matrix element <S||T2(S)||S> is also included here. Hirota assumes this is 1, which
      is appropriate for a triplet state while Sears assumes a more complex dependence depending
      on the specific make-up of the spins, which will (presumably) also be 1 for a triplet state.
      The values for the lowest few multiplicies are:
        S  <S||T2(S)||S>^2
        1      1
        1.5    6
        2     21
        2.5   56
        3    126
    }
    q := Kp-K;
    case q of
      0: Result.re := Sqrt(Six)*alpha;    // Before 7.1.323 this was Sqrt(5*Six)
      4, -4: Result.re := beta;           // Before 7.1.323 this was Sqrt(Six)
    end;
    if Result.re <> 0 then
      Result.re := Result.re * Minus(N+S+J+Np-Kp) * sixj(J, S, Np, 4, N, S)
                             * Sqrt((N+1)*(Np+One)) * threej(Np, -Kp, 4, q, N, K)
                             * ReducedMatrixElement(4, S);
  end;
  if Abs(Np-N) > 2 then Exit;
  rr := Sqrt(S*(S+2)*(S+1)*(N+1)*(Np+One)/4)*Minus(J+S+Np)*sixj(N, S, J, S, Np, 2);
  if K = Kp then begin
    for i := 0 to 1 do begin
      { Brown and Sears, JMS 75, 111 (1979) suggests the antisymmetric components
        (rank = 1) do not contribute. However, it does not seem to be possible to
        reproduce a simulation done with all 9 components non-zero with just 6 components,
        as suggested by the paper. (For the prolate example used, the antisymmetic components
        seem to be required to lift the degeneracy between the +-Ka levels at high K.)
        The antisymmetric components can be added using a pair of perturbations for each,
        with ScaleNext = -1. The antisymmetric components do not produce values
        consistent with SPCAT/SPFIT; this seems to be due to an inconsistent phase
        difference in some off-diagonal matrix elements where both (eab+eba) and (eab-eab)
        contribute to the same matrix element. }
      Rank := 4*i;
      q := (Kp - K) div 2;
      if Abs(2*q) <= rank then begin
        r1 := (Minus(rank)*Sqrt(N*(N+2)*(N+One)/4)*sixj(2, 2, rank, Np, N, N)
               +Sqrt(Np*(Np+2)*(Np+One)/4)*sixj(2, 2, Rank, N, Np, Np))*Sqrt(Rank+One)/2;
        V := rr*r1*Minus(Np-Kp)*threej(Np,-Kp, rank, 2*q, N, K);
        Result.re := Result.re + V * TSR[i*2];
      end;
      { Seems to be a phase problem with the delta N = +/1 delta K +/- 1 matrix elements,
        compared with ACDD. However, it checks OK with explicit evaluation with Maple,
        and ACDD seems to have a spcific problem with these matrix elements, in that they
        have not had their sign changed with respect to W. T. Raynes, JCP 41, 3020 (1964),
        and the ACDD documentation suggests that they should }
    end;
    { aeff - effective spin-orbit coupling for doublet states from K Kawaguchi, E Hirota, T Ishiwata and I Tanaka, JCP 93, 951 (1990) }
    { TODO: This assumes doublet states - is this simply the Sz operator? }
    //if N = Np then
    //  if J = N+1 then
    //    Result.re := Result.re + aeff*K*l/(2*N + 4) { F1 }
    //  else { J = N - 1 => F2 }
    //    Result.re := Result.re - aeff*K*l / 2 / N
    //else if Np = N+2 then
    //  Result.re := Result.re + aeff*l*Sqrt(Np*Np + K*K) / 2 / Np
    //else if Np = N-2 then
    //  Result.re := Result.re + aeff*l*Sqrt(N*N + K*K) / 2 / N;
  end;
  Result.re := Result.re + aeff*l*T1S(0,J,S,Np,Kp,N,K);
  { Quartic spin rotation - Matrix elements from J. M. Brown and T. J. Sears, JMS, 75, 111 (1979) }
  if QuarticSpinRotation then begin
    Nsq := N*(N+2)/4;
    if Np = N then begin
      if N <> 0 then begin
        if K = Kp then begin
          Ksq := Sqr(K)/4;
          Result.re := Result.re + Gamma(N,S,J)/2/Nsq*( DsK*Sqr(Ksq) + (DsNK+DsKN)*Ksq*Nsq + DsN*Sqr(Nsq) );
        end
      end;
    end else { Abs(N-Np) = 2 }
      if K = Kp then begin
        NN := Max(N,Np);
        Ksq := Sqr(K)/4;
        Result.re := Result.re - K/2/NN * (DsK*Ksq + DsNK*Sqr(NN)/4)*phi(NN,S,J)*Sqrt(((Sqr(NN))-Sqr(K))/Four);
      end;
  end;
end;

function TSymmetricTop.IsSettingQuantum(AIndex: Integer): Boolean;
begin
  Result := AIndex in [LSI+1..LSI+2];
  if not Result then
    Result := inherited IsSettingQuantum(AIndex);
end;

function TSymmetricTop.MatrixElementType(const TheMatrixElement: TMatrixElement; const H: string; ShowPhases: Boolean): string;
var
  RHS: string;
  procedure AddVariable(const Variable: string; AIndex: Integer);
  begin
    with TheMatrixElement do
      if AIndex <= High(Signature) then begin
        if Result <> '' then begin
          Result := Result + ',';
          RHS := RHS + ',';
        end;
        Result := Result + Variable;
        RHS := RHS + Variable;
        if Signature[AIndex] > 0 then
          Result := Result + '+'
        else if Signature[AIndex] = 0 then
          Result := Result + '  '; { Nice spacing }
        if Signature[AIndex] <> 0 then
          Result := Result + IntToStr(Signature[AIndex]); { Note changes are not doubled }
      end;
  end;
  function lString(l: Integer):string;
  begin
    l := (l mod 10)-1;
    if l <> 0 then
      Result := ',l='+IntToStr(l)
    else
      Result := '';
  end;
begin
  Result := '';
  RHS := '';
  AddVariable('J',3);
  AddVariable('N',2);
  AddVariable('K',1);
  if TheMatrixElement.Signature[0] > 100 then begin
    { Transition moment: Both l values encoded }
    Result := Result + lString(TheMatrixElement.Signature[0] div 10);
    RHS := RHS + lString(TheMatrixElement.Signature[0] mod 10);
  end else  { change in quantum number, just like the others }
    AddVariable('l',0);
  Result := '<' + Result
                + ShowElementPhases(TheMatrixElement,H,ShowPhases)
                + RHS + '>';
end;

function TSymmetricTop.ReadBranchFormat(Text: string; var Obs: TObservation;
  UpperState: TState; Reader: TObservationsReaderBase): Boolean;
var
  K, DeltaK, DeltaJ, kl, JU, JL, NU, NL, klU: TQuantumNumber;
  p, pe, SymW, gL, gU1, gU2, gDipole: Integer;
  Lname, Uname, Uname2, fText, SymText: string;
  HyperfineL, HyperfineU: THyperfineBasis;
  HasFields, FirstOK: Boolean;
  AlternateSym, SymU2, DefSym: ShortInt;
  IndexU2: SmallInt;
  Signs: array [0..2] of Integer;
  function fAdj(fNo: Char; var J: TQuantumNumber):Boolean;
  var
    O: Integer;
  begin
    O := 2*(Ord(fNo)-Ord('1'))-S;
    if (O < -S) or (O > S) then
      Result := True
    else begin
      Result := False;
      J := J - O;
    end;
  end;
  function TryLower:Boolean;
  begin
    Lname := Manifold.FindStateLabel(Obs.IndexL, Obs.SymL, AlternateSym, Obs.JL,
                                     [K,kl], Obs.QuantumNumberFormat, Obs.ExField,
                                     SymW, JL, NL, Self, Reader.RemoveSpins, HyperfineL);
    Result := Obs.SymL >= 0;
  end;
  function TryUpper(SymL: ShortInt; out UName: string; out IndexU: SmallInt; out SymU: ShortInt; ShowPossible: Boolean):Boolean;
  var
    AlternateSym: ShortInt;
    Sym: ShortInt;
  begin
    if Odd(Reader.BranchRank div 2) then begin
      for Sym := 0 to Molecule.LastSym do
        if Molecule.DipoleAllowed(Sym, SymL) then begin
          SymU := Sym;
          Break;
        end;
    end else
      SymU := SymL;
    UName := UpperState.Manifold.FindStateLabel(IndexU, SymU, AlternateSym, Obs.JU,
                                                [K+DeltaK,klU], Obs.QuantumNumberFormat, Obs.ExField,
                                                SymU, JU, NU, UpperState, Reader.RemoveSpins, HyperfineU, ShowPossible);
    Result := SymU >= 0;
  end;
  function RotateSigns(GResult:Boolean):Boolean;
  begin
    Result := Signs[0] <> 0;
    if Result then begin
      if GResult then
        Reader.ReportError('More than three signs');
      Exit;
    end;
    Signs[0] := Signs[1];
    Signs[1] := Signs[2];
    Signs[2] := 0;
  end;
begin
  Result := False;
  Obs.Name := ' ' + Text;
  Text := Trim(Text);
  if not IsDeltaJName(Text[1],DeltaK) then
    Exit
  else
    Delete(Text,1,1);
  if (Length(Text) = 0) or not IsDeltaJName(Text[1],DeltaJ) then begin
    DeltaJ := DeltaK;
    DeltaK := 0;
  end else
    Delete(Text,1,1);
  if Length(Text) = 0 then begin
    Result := True; { Force Branch }
    Text := Reader.NextItem;
  end;
  p := Pos('(',Text);
  pe := Pos(',',Text);
  if (pe > 1) and (pe < p) then begin
    fText := Copy(Text,1,pe-1);
    Delete(Text,1,pe);
  end else
    fText := '';
  Signs[0] := 0; Signs[1] := 0; Signs[2] := 0;
  while (Text <> '') and (Text[1] in ['+','-']) do begin
    if RotateSigns(Result) then
      Exit;
    if Text[1] = '+' then
      Signs[2] := 2
    else
      Signs[2] := -2;
    Delete(Text,1,1);
  end;
  if (Signs[1] <> 0) and not Molecule.CharacterTable.HasSeparableDegeneracy then begin
    if Result then
      Reader.ReportError('More than one sign');
    Exit;
  end;
  pe := Pos('(',Text);
  p := 1;
  while (p < pe) and (Text[p] in ['0'..'9']) do
    Inc(p);
  if p <> pe then begin
    SymText := Copy(Text,p,pe-p); { Symmetry specified in branch }
    Delete(Text,p,pe-p);
  end else
    SymText := '';
  p := Pos('(',Text);
  pe := Pos(')',Text);
  if (p > 1) and (pe > p+1)
     and not Reader.ReadQuantumNumber(K,Copy(Text,1,p-1))
     and not Reader.ReadQuantumNumber(JL,Copy(Text,p+1,pe-p-1))
      then begin
    if SymText <> '' then begin
      if not Molecule.TryStrToAnySymName(JL, SymText, SymW) then
        Reader.ReportError('Bad Symmetry:'+SymText);
    end else
      SymW := -1;
    Delete(Text,1,pe);
    JU := JL+DeltaJ;
    NL := JL;
    NU := JU;
    if (fText <> '') and (S > 0) then begin
      if Length(fText) = 1 then
        fText := fText + fText
      else if Length(fText) <> 2 then
        Exit;
      if fAdj(fText[1],JU) or fAdj(fText[2],JL) then
        Exit;
    end;
    Obs.JU := JU;
    Obs.JL := JL;
    HasFields := IsNonZero(Obs.ExField);
    if not Molecule.IsDegenerate(Symmetry) or (K=0) then begin
      if RotateSigns(Result) then
        Exit;
    end;
    if Signs[0] = 0 then
      Signs[0] := Signs[1]; { First two signs not duplicated }
    if Molecule.IsDegenerate(Symmetry) and (K=0) then
      kl := 2
    else if Signs[2] = 0 then
      kl := 0
    else
      kl := Signs[2];
    if Signs[1] > 0 then
      kl := kl and not 12 or 4
    else if Signs[1] < 0 then
      kl := kl and not 12 or 8;
    if HasFields or (Molecule.nNuclei > Reader.RemoveSpins) then begin
      { Some hyperfine structure to be found }
      if ReadHyperfineBranch(Text, Obs, HyperfineU, HyperfineL, Reader, HasFields) then
        Exit;
    end else if Text <> '' then
      Exit;
    if not TryLower and (Obs.IndexL = -1) and (kl <> Low(TQuantumNumber)) then begin
      { No matches - try without kl }
      kl := Low(TQuantumNumber);
      TryLower;
    end;
    Uname := '';
    klU := Low(TQuantumNumber);
    if not Molecule.IsDegenerate((UpperState as TSymmetricTop).Symmetry) and (K+DeltaK <> 0) then
      if Signs[0] > 0 then
        klU := 4
      else if Signs[0] < 0 then
        klU := 8;
    if (Obs.IndexL = -3) and (SymW < 0) and (AlternateSym >= 0) then begin
      { Two matches of two different symmetries - look for a single allowed transition }
      SymW := -Obs.SymL-1; { One of the possible symmetries }
      if TryUpper(SymW, Uname, Obs.IndexU, Obs.SymU, False) then begin
        { First symmetry OK - how about second? }
        if TryUpper(AlternateSym, Uname2, IndexU2, SymU2, False) then begin
          { Both match - only OK if default symmetry used }
          DefSym := Molecule.ConvertDefaultSymmetry(JL, Reader.DefaultSymmetry);
          if SymW = DefSym then
            { OK with first symmetry }
          else if AlternateSym = DefSym then begin
            { Take second symmetry }
            Uname := Uname2;
            Obs.IndexU := IndexU2;
            Obs.SymU := SymU2;
            SymW := AlternateSym;
          end else
            Obs.SymU := -1; { Fails }
        end
      end else if TryUpper(AlternateSym, Uname, Obs.IndexU, Obs.SymU, False) then begin
        { Second symmetry only OK }
        SymW := AlternateSym;
      end else
        Uname := ''; { both fail }
      if Obs.SymU >= 0 then
        TryLower; { Find specific lower state }
    end else if Obs.SymL >= 0 then
      if not TryUpper(Obs.SymL, Uname, Obs.IndexU, Obs.SymU, False) then begin
        SymW := (UpperState as TSymmetricTop).Symmetry;
        { Two matches - see if it is possible to pin down kl }
        if (Obs.IndexU = -3) { Two matches } and Molecule.IsDegenerate(SymW) { kl applies to state } then
          if Signs[0] <> 0 { +/- sign specified - try two alternate signs of l } then begin
            if Signs[0] > 0 then
              klU := 6
            else
              klU := 10;
            if TryUpper(Obs.SymL, Uname, Obs.IndexU, Obs.SymU, False) then begin
              { First OK - try other l }
              klU := -2 and not 12 or (klU and 12);
              if TryUpper(Obs.SymL, Uname2, IndexU2, SymU2, False) then begin
                { Both OK, so force fail }
                klU := Low(TQuantumNumber);
                Obs.SymU := -1;
              end
            end else begin
              klU := -2 and not 12 or (klU and 12);
              TryUpper(Obs.SymL, Uname, Obs.IndexU, Obs.SymU, False)
            end;
          end else if (2*K + DeltaK > Reader.BranchRank) { Other sign of K gives too big a change } then begin
            gU1 := Molecule.g(K + DeltaK, 2, SymW);
            gU2 := Molecule.g(K + DeltaK, -2, SymW);
            if gU1 <> gU2 then begin
              gL := Molecule.g(K, kl, Symmetry);
              gDipole := Molecule.CharacterTable.IReps[Molecule.CharacterTable.DipoleSym].g;
              gL := AddG(gl+gDipole, Molecule.CharacterTable.n);
              if gL = gU1 then
                klU := 2
              else if gL = gU2 then
                klU := -2
              else
                raise EReportException.CreateFmt('Fail in SymT Readbranch:%d %d %d %d',[gL, gU1, gU2, gDipole]);
              if not TryUpper(Obs.SymL, Uname, Obs.IndexU, Obs.SymU, False) then
                klU := Low(TQuantumNumber); { Fail - put back general form for error message }
            end;
          end;
        if Obs.SymU < 0 then { Fails - show more detail }        
          TryUpper(Obs.SymL, Uname, Obs.IndexU, Obs.SymU, True);
      end;
    Obs.Name := Obs.Name + ' : ' + Uname + ' - ' + Lname;
    if (Obs.SymL >= 0) and (Obs.SymU >= 0) then begin
      Result := True;
      Obs.IndexU := 0;
      Obs.IndexL := 0;
      Exit;
    end;
  end;
  if Result then
    Reader.ReportError('Bad Branch format');
    //if (Obs.IndexL = -3) and (Reader.DefaultSymmetry >= 0) and (SymW < 0) then
    //  { Two matches - try using default symmetry }
    //  Lname := Manifold.FindStateLabel(Obs.IndexL, Obs.SymL, AlternateSym, Obs.JL,
    //                                   [K,kl], Obs.QuantumNumberFormat, Obs.ExField,
    //                                   Reader.DefaultSymmetry, JL, NL, Self, Reader.RemoveSpins, HyperfineL);
    { TODO: This section needs to look for allowed transitions to select sign of kl }
    { See e.g. tests/C3v_A1_E.pgo tests/C3v_A1_E.lin which should be readable, but currently are not }
    //if Obs.IndexL = -1 then
    //  { No matches - try without kl }
    //  Lname := Manifold.FindStateLabel(Obs.IndexL, Obs.SymL, AlternateSym, Obs.JL,
    //                                   [K,Low(TquantumNumber)], Obs.QuantumNumberFormat, Obs.ExField,
    //                                   -1, JL, NL, Self, Reader.RemoveSpins, HyperfineL);
  //  if Obs.SymL >= 0 then begin
  //    if Odd(Reader.BranchRank div 2) then begin
  //      for Sym := 0 to Molecule.LastSym do
  //        if Molecule.DipoleAllowed(Sym, Obs.SymL) then begin
  //          Obs.SymU := Sym;
  //          Break;
  //        end;
  //    end else
  //      Obs.SymU := Obs.SymL;
  //    Obs.Name := Obs.Name + ' : '
  //                + UpperState.Manifold.FindStateLabel(Obs.IndexU, Obs.SymU, AlternateSym, Obs.JU,
  //                                                     [K+DeltaK,Low(TQuantumNumber)], Obs.QuantumNumberFormat, Obs.ExField,
  //                                                     Obs.SymU, JU, NU, UpperState, Reader.RemoveSpins, HyperfineU)
  //                + ' - ' + Lname;
  //
  //    if Obs.SymU >= 0 then begin
  //      Result := True;
  //      Obs.IndexU := 0;
  //      Obs.IndexL := 0;
  //      Exit;
  //    end;
  //  end else
  //    Obs.Name := Obs.Name + ' : ' + Lname
  //end;
  //if Result then
  //  Reader.ReportError('Bad Branch format');
end;

function TSymmetricTop.BumpBranchQno(Branch: string; AIncrement: TQuantumNumber; QnoType: TQnoType): string;
begin
  Result := ''; { Default is to indicate increment not possible }
  case QnoType of
    //qtN: ;
    qtA: Result := BumpBeforeBracketedBranch(Branch, AIncrement, 1);
    //qtB: ;
  else
    Result := inherited BumpBranchQno(Branch, AIncrement, QnoType);
  end;
end;

function TSymmetricTop.ReadProperties(Node: TDOMNode; CheckUnread: Boolean): Integer;
begin
  Result := inherited ReadProperties(Node, CheckUnread);
  if (ReadVersion <= 7001323) and ((alpha <> 0) or (beta <> 0)) then begin
    alpha := alpha*Sqrt(5);
    beta := beta*Sqrt(Six);
    {.$IFNDEF CONSOLE}
    ShowInfo('Warning: definition of alpha and beta has changed; values have been changed to reflect this in '+UniqueName);
    {.$ENDIF}
  end;
end;

function TSymmetricTop.RotationalMatrixElement(ABra, AKet: TBasisState; SignA, SignB: Integer; MNode: TMnode; Flag: Integer): Complex;
var
  Bra, Ket: TSymmetricTopBasisState;
  H: function (J, Kp, lp, K, l: TQuantumNumber): Double of object;
  Kp, lp, K, l: TQuantumNumber;
begin
  Bra := ABra as TSymmetricTopBasisState;
  Ket := AKet as TSymmetricTopBasisState;
  case TMatElemType(Flag) of
    meHrot:  H := Hrot;
    meWidth: H := HWidth;
  else
    raise NotYet;
  end;
  if Bra.J <> Ket.J then
    raise Exception.Create('TSymmetricTop.RotationalMatrixElement');
  Kp := SignA*Bra.K;
  lp := SignA*Bra.l;
  K := SignB*Ket.K;
  l := SignB*Ket.l;
  if (Flag = Ord(meHrot)) { TODO: Spin dependent width model? }
     and (Bra.S <> 0) and (Abs(Bra.N-Ket.N) <= 4) then begin
    Result := HSR(Bra.J, Bra.N, Kp, lp, Ket.N, K, l);
  end else
    Result := z0;
  { TODO: may require all four matrix elements or explicit test for identical phase? }
  if Bra.N = Ket.N then
    Result.re := Result.re + H(Bra.N, Kp, lp, K, l);
  { TODO: S <> 0 terms }
end;

function TSymmetricTop.HasWidth: Boolean;
var
  i: Integer;
begin
  Result := inherited HasWidth;
  if not Result then
    for i := FirstWidthPar to nParameters do
      if Parameters[i] <> 0 then begin
        Result := True;
        Exit;
      end;
end;

function TSymmetricTop.HDip(TransitionMoment: TSingleTransitionMoment; Jp, Kp, lp, J, K, l: TQuantumNumber; Drho: Integer): Double;
var
  q: TQuantumNumber;
  Phase: Integer;
begin
  Phase := 0;
  Result := TransitionMoment.Strength;
  if TransitionMoment is TSphericalTransitionMoment then begin
    q := TSphericalTransitionMoment(TransitionMoment).ActualComponent;
  end else if TransitionMoment is TCartesianTransitionMoment then begin
    { Convert to Spherical equivalent }
    q := 2;
    case TCartesianTransitionMoment(TransitionMoment).Axis of
      aAxis: if Prolate then q := 0;
      cAxis: if not Prolate then q := 0;
    end;
    if q = 2 then
      Result := Result / Sqrt(Two);
  end else
    raise NotYet;
  if Abs(q) <> Abs(Kp-K) then begin
    Result := 0;
    Exit;
  end;
  { If the other sign of q is required, apply symmtery operation }
  { For Delta K = 0, need to choose a sign of l; in this case choose:
      l' > l or l' > 0
    as the defined one }
  if (q <> K-Kp)
     or ((q = 0) and ( (lp < l) or (lp < 0) ) ) then
    phase := Drho*2 + TransitionMoment.Rank*Molecule.CharacterTable.rhoJmult + q*Molecule.CharacterTable.rhoKmult;
  Result := Result * Sqrt((J+1)*(Jp+1))
            * Minus(Jp-Kp+phase) * threej(Jp, -Kp, TTransitionMoment(TransitionMoment).Rank, Kp-K, J, K);
end;

function TSymmetricTop.HDipMatrixElements(TransitionMoment: TTransitionMoment; IncompleteOK: Boolean): TMatrixElements;
var
  { Note quantum numbers not doubled here }
  q, qq, lBra, llBra, lKet, llKet, dN, Drho, g: Integer;
  Phase, E, eNp, eKp: TExprNode;
  nMatrixElements: Integer;
  Bra, Ket: TSymmetricTop;
begin
  Phase := TIntegerNode.Create(1);
  if TransitionMoment is TSphericalTransitionMoment then begin
    q := TSphericalTransitionMoment(TransitionMoment).ActualComponent div 2;
  end else if TransitionMoment is TCartesianTransitionMoment then begin
    { Convert to Spherical equivalent }
    q := 1;
    case TCartesianTransitionMoment(TransitionMoment).Axis of
      aAxis: if Prolate then q := 0;
      cAxis: if not Prolate then q := 0;
    end;
    if q = 1 then
      Phase := Expr.EN([Phase, '/', EB([2, 'sqrt'])]);
  end else
    raise NotYet;
  if not IncompleteOK and (TransitionMoment.ActiveTransitionFunction <> nil) then
    raise NotYet;
  nMatrixElements := TransitionMoment.Rank + 1;
  if q <> 0 then
    nMatrixElements := nMatrixElements*2;
  { Find possible delta l values }
  Bra := TransitionMoment.Bra as TSymmetricTop;
  Ket := TransitionMoment.Ket as TSymmetricTop;
  if Molecule.IsDegenerate(Bra.Symmetry) then begin
    lBra := -1;
    nMatrixElements := nMatrixElements*2;
  end else
    lBra := 0;
  if Molecule.IsDegenerate(Ket.Symmetry) then begin
    lKet := -1;
    nMatrixElements := nMatrixElements*2;
  end else
    lKet := 0;
  SetLength(Result,nMatrixElements);
  nMatrixElements := 0;
  for dN := -TransitionMoment.Rank div 2 to TransitionMoment.Rank div 2 do begin
    eNp := EB([eN,'+',dN]);
    llBra := lBra;
    repeat
      llKet := lKet;
      repeat
        qq := -Abs(q);
        repeat
          { Need to enforce kl selection rules here - work out Delta g
            = gKmult*DeltaK + gEV' - gEV }
          { -gKet }
          g := -Molecule.CharacterTable.IReps[Ket.Symmetry].g;
          if llKet < 0 then
            g := -g;
          { add gBra }
          if llBra < 0 then
            g := g - Molecule.CharacterTable.IReps[Bra.Symmetry].g
          else
            g := g + Molecule.CharacterTable.IReps[Bra.Symmetry].g;
          g := AddG(Molecule.CharacterTable.gKmult*qq+g, Molecule.CharacterTable.n);
          if not TransitionMoment.Magnetic and  (Minus(TransitionMoment.Rank) < 0) then begin
            Drho := Molecule.CharacterTable.IReps[Molecule.CharacterTable.DipoleSym].rho;
            g := AddG(g+Molecule.CharacterTable.IReps[Molecule.CharacterTable.DipoleSym].g, Molecule.CharacterTable.n);
          end else
            Drho := 0;
          if g = 0 then begin
            eKp := EB([eK,'+',qq]);
            E := Expr.EN([EB([2, '*', eNp, '+', 1]), '*', EB([2, '*', eN, '+', 1]), 'sqrt']);
            E := Expr.EN([-1, '^', EB([eNp, '-', eKp]), '*', E]);
            E := Expr.EN([Expr.EN(['threej', eNp.Duplicate, MonadicMinus.CreateExprNode(eKp), TransitionMoment.Rank div 2, qq, eN.Duplicate, eK.Duplicate]),
                         '*', E, '*', Phase.Duplicate]);
            if (qq <> -q)
               or ((q = 0) and ( (llBra < llKet) or (llBra < 0) ) ) then begin
              { Need "other" vibronic matrix element }
              Drho := Molecule.CharacterTable.IReps[Bra.Symmetry].rho
                      - Molecule.CharacterTable.IReps[Ket.Symmetry].rho
                      + Drho;
              Drho := Drho*2 + TransitionMoment.Rank*Molecule.CharacterTable.rhoJmult + 2*qq*Molecule.CharacterTable.rhoKmult;
              if Minus(Drho) < 0 then
                E := Expr.EN([E, '*', -1]);
            end;
            if Assigned(E) then begin
              E := E.Simplify;
              Result[nMatrixElements].Expr := E;
              Result[nMatrixElements].Evaluator := nil;
              Result[nMatrixElements].CEvaluator := nil;
              SetLength(Result[nMatrixElements].Signature,3);
              Result[nMatrixElements].Signature[0] := 1000+(llBra+1)*10+llKet+1; // l
              Result[nMatrixElements].Signature[1] := qq; // Component = Delta K
              Result[nMatrixElements].Signature[2] := dN;
              Inc(nMatrixElements);
            end;
          end;
          qq := -qq;
        until qq <= 0;
        llKet := -llKet;
      until llKet <= 0;
      llBra := -llBra;
    until llBra <= 0;
    eNp.Free;
  end;
  Phase.Free;
  SetLength(Result, nMatrixElements);
end;

function TSymmetricTop.RotationalTransitionMatrixElement(ABra, AKet: TBasisState; SignA, SignB: Integer; TransitionMoment: TRootedMNodePair; F: TQuantumNumber; out dK: TQuantumNumber):Complex;
var
  Bra, Ket: TSymmetricTopBasisState;
  Kp, lp, K, l: TQuantumNumber;
  Drho: Integer;
  TM: TTransitionMoment;
begin
  Bra := ABra as TSymmetricTopBasisState;
  Ket := AKet as TSymmetricTopBasisState;
  if TransitionMoment is TPerturbation then begin
    TM := TransitionMoment.Owner as TTransitionMoment;
    raise NotYet;
  end else
    TM := TransitionMoment as TTransitionMoment;
  { The transition moment uses tensor notation T(k,q) for evaluation, and a key
    point is that the matrix elements of T(k,q) and T(k,-q) are symmetry related.
    This is worked out here as follows. The matrix element required is of the form:
        <A'|Htran|A>
    which is split into separate vibronic (EV) and rotational(R) parts:
        <A'|Htran|A> = <EV'|HE(k,q)|EV><R'|HR(k,q)|R>
    <EV'|HE(k,q)|EV> is the strength parameter of the transition moment. HE(k,q) is the body
    fixed transition operator and HR(k,q) is essentially the Wigner D matrix transforming
    it from the space to body fixed axis systems. Now, if Hdip finds q has the wrong
    sign to give a non zero matrix element, it actually works out:
        <EV'|HE(k,q)|EV><R'|HR(k,-q)|R>
    whereas what is actually wanted is
        <-EV'|HE(k,-q)|-EV><R'|HR(k,-q)|R>
    The -EV indicates that some quanum numbers (l) in the vibronic part may also need
    changing to suit. We can generate <EV'|HE(k,q)|EV> by using the R- operator,
    which flips the sign of q. The characters with respect to this operator are rho
    in the CharaterTable.Irep array. A rotational function transforms as:
        R- |J,K> = rhoJK(J,K)|J,-K>
    where rhoJK(J,K) = (-1)^(rhoJmult*J+rhoKmult*K)
          where rhoKmult is 1 for Cnv and 0 otherwise.
    HR(K,q) transforms as the rotational function |K,q> so
        R- HR(K,q) = rhoJK(K,q)HR(K,-q)
    Now the overall transition moment operator transforms as:
        R- Htran = rhoTran Htran
    where rhoTran is given by Molecule.CharacterTable.DipoleSym for dipole transitions
    and 1 otherwise. As
        Htran = HE(k,q) HR(k,q)
    we have
        R- Htran = rhoTran Htran(-) = R- HE(k,q) R- HR(k,q)
                 = rhoHE HE(k,-q) rhoJK(K,q)HR(K,-q)
    so that rhoHE, the multiplying factor on HE(k,q) is
        rhoHE = rhoTran rhoJK(K,q)
    Applying R- to a number has no effect, so
        <EV'|HE(k,q)|EV> = R- <EV'|HE(k,q)|EV> = <R- EV'|R- HE(k,q)|R- EV>
                         = rho' rhoTran rhoJK(K,q) rho <-EV'|HE(k,-q)|-EV>
    This gives the required phase factor; rhoJK(K,q) is actually included
    in HDip.
    Note that the above does not apply to C3h...,S4... as there is no operation that
    transforms K -> -K.
  }
  { Note about symmetry check:  this is normally enforced elsewhere, but not in the presence of external fields
    Additional test is required because overall zero value arises from the combination
    of two non zero values for non-degenerate symmetries, so do not short cut process
    when checking }
  Result := z0;
  dK := Low(TQuantumNumber);
  if not TM.Magnetic and  (Minus(TM.Rank) < 0) then begin
    if not Molecule.DipoleAllowed(Bra.Sym, Ket.Sym) then
      { Overall transition dipole has A or B symmetry }
      if not ExprChecking or Molecule.IsDegenerate(Bra.Sym) or Molecule.IsDegenerate(Ket.Sym) then
        Exit;
    { For E rovibronic symmetry states, the signs of K and l have been chosen to
      give the overall quantum number, g > 0, as there are be no Hamiltonian matrix
      elements off diagonal in g. Transition matrix elements are also diagonal in g,
      apart from D2nd, S4n, where the dipole transtion moment changes g by n (or n/2 for S4n).
      The g < 0 combination needs to be taken for one of the states in this case,
      so flip the signs of K and l for one of the states. }
    Drho := Molecule.CharacterTable.IReps[Molecule.CharacterTable.DipoleSym].rho;
    if (Molecule.CharacterTable.IReps[Molecule.CharacterTable.DipoleSym].g > 0)  { B Symmetry dipole rather than A }
       and (Bra.Phase = 0) and (Ket.Phase = 0)
       and Molecule.IsDegenerate(Bra.Sym)
       and Molecule.IsDegenerate(Ket.Sym) then begin
      SignB := -SignB; { Choose right combination of signs of l for E - E transitions }
    end;
  end else begin
    { Transition dipole totally symmetric }
    if Bra.Sym <> Ket.Sym then
      if not ExprChecking or Molecule.IsDegenerate(Bra.Sym) or Molecule.IsDegenerate(Ket.Sym) then
        Exit;
    Drho := 0;
  end;
  Drho := Molecule.CharacterTable.IReps[(Bra.State as TSymmetricTop).Symmetry].rho
          - Molecule.CharacterTable.IReps[(Ket.State as TSymmetricTop).Symmetry].rho
          + Drho;
  Kp := SignA*Bra.K;
  lp := SignA*Bra.l;
  K := SignB*Ket.K;
  l := SignB*Ket.l;
  dK := Kp-K;
  Result.re := HDip(TM as TSingleTransitionMoment, Bra.N, Kp, lp, Ket.N, K, l, Drho);
  if (Ket.S > 0) and not MathSup.IsZero(Result) then { TODO: Scope for pushing this one level up, as it will be the same for all K combinations }
    CMul(Result, Minus(Bra.N+Ket.S+Ket.J+TM.Rank) * Sqrt((Bra.J+1)*(Ket.J+One)) * sixj(Bra.N, Bra.J, Ket.S, Ket.J, Ket.N, TM.Rank));
end;

function TSymmetricTop.SymbolicRotationalTransitionMatrixElements(TransitionMoment: TTransitionMoment; AddHyperfineCoupling: Boolean; IncompleteOK: Boolean): TMatrixElements;
begin
  if Molecule.nNuclei > 0 then
    raise NotYet; { TODO: Implement }
  Result := HDipMatrixElements(TransitionMoment, IncompleteOK);
  if S > 0 then
    AddOpenShellIntensity(eN, eJ, eS, TransitionMoment, 2, Result);
end;

procedure TSymmetricTop.SetSettingFromStr(AIndex: Integer; const Str: string);
begin
  if ((AIndex = LSI+1) or (AIndex = LSI+2)) and (Str = 'all') then
    SetSetting(AIndex,-9999)
  else
    inherited;
end;

procedure TSymmetricTop.AdjustDefaults;
begin
  inherited AdjustDefaults;
  if B = 0 then
    B := Convert(0.5, cm1, (RootNode as TMixture).Units);
  if C = 0 then
    C := 10*B;
end;

function TSymmetricTop.SettingToStr(AIndex: Integer): string;
begin
  if ((AIndex = LSI+1) or (AIndex = LSI+2)) and (GetSetting(AIndex) < 0) then
      Result := 'all'
  else
    Result := inherited SettingToStr(AIndex);
end;

function SymbolicRotationalMatrixElement(PE: TPertEquiv): TMatrixElements;

  function RotationalMatrixElement(delK, dell:Integer; eN, eKp, eK: TExprNode):TExprNode;
  var
    Node: TExprNode;
  begin
    Result := nil;
    if Abs(delK) <> Abs(PE.pPower) then
      Exit;
    Result := EB([eN, '+', 1, '*', eN, '^', PE.JPower div 2]); { N(N+1)^JPower }
    if delK <> 0 then
      Result := OperatorTimes.CreateExprNode(Result,rlE(eN, eK, delK)); { N+-^pPower }
    if Assigned(Result) then begin
      if PE.zPower <> 0 then
        if (PE.zPower < -2) and Odd(PE.zPower) and (PE.pPower <> 0) then begin
          { [Nz^(zPower-1),[Nz, N+-^pPower]+]+ }
          Result := Expr.EN([eK.Duplicate, '+', eKp.Duplicate, '*', Result]);
          Node := EB([eK, '^', Abs(PE.zPower)-1]);
          Result := EB([eKp.Duplicate, '^', Abs(PE.zPower)-1, '+', Node, '*', Result],[],0,False);
        end else begin
          Node := OperatorPower.CreateExprNode(eK.Duplicate,TIntegerNode.Create(Abs(PE.zPower)));
          if PE.pPower <> 0 then
            Result := EB([eKp.Duplicate, '^', Abs(PE.zPower), '+', Node, '*', Result],[],0,False) { [Nz^zPower, N+-^pPower]+ }
          else
            Result := OperatorTimes.CreateExprNode(Result, Node) { Nz^zPower }
        end;
      if PE.lPower <> 0 then
        if dell <> 0 then
          Result := OperatorTimes.CreateExprNode(Result,TIntegerNode.Create(Sign(dell)*PE.lPower))
        else
          Result := Expr.EN([Result, '*', eL.Duplicate, '*', PE.lPower]);
      Result := Result.Simplify;
      Result := Result.Simplify;
      if Result.EqualTo(0) then
        FreeAndNil(Result);
    end;
  end;

var
  delK, dell, delN, delNMax, q, nMatrixElements: Integer;
  E, eKp, eKpp, eNp: TExprNode;
  ScaleFactor: TExprNode;
  HasRotPart: Boolean;
begin
  nMatrixElements := 1;
  if (PE.pPower + PE.SpPower <> 0) or (PE.lChange <> 0) then
    Inc(nMatrixElements);
  delNMax := PE.SzPower + Abs(PE.SpPower);
  nMatrixElements := nMatrixElements * (2*delNMax+1);
  SetLength(Result,nMatrixElements);
  nMatrixElements := 0;
  ScaleFactor := Expr.EN([PE.Scale4, '/', 4]);
  HasRotPart := (PE.JPower or PE.zPower or PE.pPower) <> 0;
  dell := -PE.lChange;
  repeat
    { If dell <> 0, PE.lChange and PE.pPower indicate relative signs,
      and only one cycle of inner loop is required }
    if dell*PE.lChange <= 0 then
      delK := -(PE.pPower+PE.SpPower)
    else
      delK := PE.pPower+PE.SpPower;
    repeat
      eKp := OperatorPlus.CreateExprNode(eK.Duplicate, TIntegerNode.Create(delK)).Simplify;
      for delN := -delNMax to delNMax do begin
        eNp := OperatorPlus.CreateExprNode(eN.Duplicate, TIntegerNode.Create(delN)).Simplify;
        if delNMax = 0  then begin
          E := Expr.EN([RotationalMatrixElement(delK, dell, eN, eKp, eK), '*', ScaleFactor.Duplicate]);
          if Assigned(E) and ((PE.NSPower or PE.SPower) <> 0) then
            E := Expr.EN([NSOperatorE(eJ, eN, eS, PE.NSPower, PE.SPower), '*', E]);
        end else if PE.SzPower > 0 then begin
          if (PE.SpPower or PE.NSPower or PE.SPower) <> 0 then
            raise NotYet;
          if HasRotPart then begin
            E := Expr.EN([RotationalMatrixElement(delK, dell, eNp, eKp, eK),
                          '*', T1SnE(PE.SzPower, 0, eJ, eS, eNp, eKp, eN, eK, ScaleFactor)]);
            E := Expr.EN([T1SnE(PE.SzPower, 0, eJ, eS, eNp, eKp, eN, eK, ScaleFactor),
                          '*', RotationalMatrixElement(delK, dell, eN, eKp, eK), '+', E]);
          end else begin
            E := T1SnE(PE.SzPower, 0, eJ, eS, eNp, eKp, eN, eK, ScaleFactor);
          end;
        end else {PE.SzPower <> 0} begin
          if PE.SpPower > 0 then
            q := 1
          else
            q := -1;
          E := nil;
          if PE.pPower + PE.SpPower = delK then begin
            eKpp := EB([eK, '+', PE.SpPower]);
            E := Expr.EN([RotationalMatrixElement(PE.pPower, dell, eNp, eKp, eKpp), '*', rlSnE(PE.SpPower,-q,eJ,eS,eNp,eKpp,eN,eK,ScaleFactor)]);
            eKpp.Free;
            eKpp := EB([eK, '+', PE.pPower]);
            E := Expr.EN([rlSnE(PE.SpPower,-q,eJ,eS,eNp,eKp,eN,eKpp,ScaleFactor), '*', RotationalMatrixElement(PE.pPower, dell, eN, eKpp, eK), '+', E]);
            eKpp.Free;
          end;
          if PE.pPower + PE.SpPower = -delK then begin
            eKpp := EB([eK, '+', -PE.SpPower]);
            E := Expr.EN([RotationalMatrixElement(-PE.pPower, dell, eNp, eKp, eKpp), '*', rlSnE(PE.SpPower,q,eJ,eS,eNp,eKpp,eN,eK,ScaleFactor), '+', E]);
            eKpp.Free;
            eKpp := EB([eK, '+', -PE.pPower]);
            E := Expr.EN([rlSnE(PE.SpPower,q,eJ,eS,eNp,eKp,eN,eKpp,ScaleFactor), '*', RotationalMatrixElement(-PE.pPower, dell, eN, eKpp, eK), '+', E]);
            eKpp.Free;
          end;
        end;
//          if PE.pPower + PE.SpPower = delK then begin
//            eKpp := OperatorPlus.CreateExprNode(eK.Duplicate, TIntegerNode.Create(PE.SpPower)).Simplify;
//            E := Expr.EN([RotationalMatrixElement(PE.pPower, dell, eNp, eKp, eKpp), '*', rlSE(-PE.SpPower,eJ,eS,eNp,eKpp,eN,eK)]);
//            eKpp.Free;
//            eKpp := OperatorPlus.CreateExprNode(eK.Duplicate, TIntegerNode.Create(PE.pPower)).Simplify;
//            E := Expr.EN([rlSE(-PE.SpPower,eJ,eS,eNp,eKp,eN,eKpp), '*', RotationalMatrixElement(PE.pPower, dell, eN, eKpp, eK), '+', E]);
//            eKpp.Free;
//          end;
//          if PE.pPower + PE.SpPower = -delK then begin
//            eKpp := OperatorPlus.CreateExprNode(eK.Duplicate, TIntegerNode.Create(-PE.SpPower)).Simplify;
//            E := Expr.EN([RotationalMatrixElement(-PE.pPower, dell, eNp, eKp, eKpp), '*', rlSE(PE.SpPower,eJ,eS,eNp,eKpp,eN,eK), '+', E]);
//            eKpp.Free;
//            eKpp := OperatorPlus.CreateExprNode(eK.Duplicate, TIntegerNode.Create(-PE.pPower)).Simplify;
//            E := Expr.EN([rlSE(PE.SpPower,eJ,eS,eNp,eKp,eN,eKpp), '*', RotationalMatrixElement(-PE.pPower, dell, eN, eKpp, eK), '+', E]);
//            eKpp.Free;
//          end;
//        end else begin
//          E := RotationalMatrixElement(delK, dell, eN, eKp, eK);
//          if Assigned(E) then
//            if PE.NSPower > 0 then
//              E := Expr.EN([NSOperatorE(eJ, eN, eS, PE.NSPower), '*', E])
//            else if PE.SzPower <> 0 then begin
//              E := Expr.EN([E, '*', T1SE(0,eJ,eS,eNp,eKp,eN,eKp)]);
//              E := Expr.EN([RotationalMatrixElement(delK, dell, eNp, eKp, eK), '*', T1SE(0,eJ,eS,eNp,eK,eN,eK), '+', E]);
//            end;
//        end;
        if Assigned(E) then begin
          if E.EqualTo(0) then
            FreeAndNil(E);
        end;
        if Assigned(E) then begin
          if High(Result) < nMatrixElements then
            SetLength(Result, 2*nMatrixElements);
          Result[nMatrixElements].Expr := E;
          Result[nMatrixElements].Evaluator := nil;
          Result[nMatrixElements].CEvaluator := nil;
          SetLength(Result[nMatrixElements].Signature,3);
          Result[nMatrixElements].Signature[0] := dell;
          Result[nMatrixElements].Signature[1] := delK;
          Result[nMatrixElements].Signature[2] := delN;
          Inc(nMatrixElements);
        end;
        eNp.Free;
      end;
      eKp.Free;
      delK := -DelK
    until (delK = -(PE.pPower+PE.SpPower)) or (dell <> 0);
    dell := -dell
  until dell = -PE.lChange;
  ScaleFactor.Free;
  SetLength(Result, nMatrixElements);
end;

function TSymmetricTop.SymbolicRotationalMatrixElement(AIndex: Integer; RankWanted: Integer): TMatrixElements;
var
  PE, PE1: TPertEquiv;
  M: TMatrixElements;
  i: Integer;
begin
  PE := FindPertEquiv(ParameterNames[AIndex], PE1);
  if PE.Name = '' then
    Result := inherited SymbolicRotationalMatrixElement(AIndex, RankWanted)
  else begin
    Result := SymT.SymbolicRotationalMatrixElement(PE);
    if PE.Name = 'zeta' then { Want C*zeta... }
      for i := 0 to High(Result) do
        Result[i].Expr := OperatorTimes.CreateExprNode(ParameterVariable(IndexOfParameter('C')).Duplicate,Result[i].Expr);
    if PE1.Name <> '' then begin
      M := SymT.SymbolicRotationalMatrixElement(PE1);
      MergeElements(Result, M, 1, False, True);
    end;
  end;
end;

function TSymmetricTop.RotationalPerturbationMatrixElement(ABra, AKet: TBasisState;
  SignA, SignB: Integer; MNode: TMNode; Flag: Integer): Complex;
var
  Bra, Ket: TSymmetricTopBasisState;
  P: TSymmetricTopPerturbation;
begin
  if Flag <> Ord(meHrot) then
    raise NotYet;
  Bra := ABra as TSymmetricTopBasisState;
  Ket := AKet as TSymmetricTopBasisState;
  P := MNode as TSymmetricTopPerturbation;
  if Bra.J <> Ket.J then
    raise Exception.Create('TSymmetricTop.PerturbationMatrixElement');
  Result.re := P.MatrixElement(Bra.J, Bra.S, Bra.N, SignA*Bra.K, SignA*Bra.l,
                                             Ket.N, SignB*Ket.K, SignB*Ket.l);
  Result.im := 0;
end;

const
  SymmetricTopSettingNames: array[LSI..LSI+2] of string = (
'Symmetry',
'Minimum |K| to use; "all" (default) for no lower limit.',
'Maximum |K| to use; "all" (default) for no upper limit.'
);

function TSymmetricTop.HelpOnSetting(AIndex: Integer): string;
begin
  if (AIndex >= Low(SymmetricTopSettingNames))
     and (AIndex <= High(SymmetricTopSettingNames)) then
    Result := SymmetricTopSettingNames[AIndex]
  else
    Result := inherited HelpOnSetting(AIndex);
end;

const
 SymmetricTopParameterNames: array [LSP..LSP+34] of string = (
'Rotational constant perpendicular to symmetry axis.',
'Rotational constant about symmetry axis (A or C).',
'J^2(J+1)^2 Quartic Centrifugal Distortion',
'J(J+1)K^2 Quartic Centrifugal Distortion',
'K^4 Quartic Centrifugal Distortion',
'Coriolis coupling constant.',
'J(J+1) dependence of Coriolis coupling constant.',
'K^2 dependence of Coriolis coupling constant.',
'l doubling constant.',
'l doubling constant.',
'l doubling constant.',
'J(J+1) dependence of qplus l doubling constant.',
'K^2 dependence of qplus l doubling constant.',
'J(J+1) dependence of r l doubling constant.',
'K^2 dependence of r l doubling constant.',
'J^3(J+1)^3 Sextic Centrifugal Distortion',
'J^2(J+1)^2K^2 Sextic Centrifugal Distortion',
'J(J+1)K^4 Sextic Centrifugal Distortion',
'K^6 Sextic Centrifugal Distortion',
'J^4(J+1)^4 Octic Centrifugal Distortion',
'J^3(J+1)^3K^2 Octic Centrifugal Distortion',
'J^2(J+1)^2K^4 Octic Centrifugal Distortion',
'J(J+1)K^6 Octic Centrifugal Distortion',
'K^8 Octic Centrifugal Distortion',
'xx/yy Spin-Rotation',
'zz Spin-Rotation',
'Quartic centrifugal distortion of the spin-rotation',
'Quartic centrifugal distortion of the spin-rotation',
'Quartic centrifugal distortion of the spin-rotation',
'Quartic centrifugal distortion of the spin-rotation',
'Diagonal spin-spin coupling constant (=D/3)',
'Off-Diagonal spin-spin coupling constant (=E)',
'Effective spin-orbit parameter',
'multiplier of <K^2> for linewidth',
'multiplier of <J(J+1)-K^2> for linewidth'
);

function TSymmetricTop.HelpOnParameter(AIndex: Integer): string;
begin
  if (AIndex >= Low(SymmetricTopParameterNames))
     and (AIndex <= High(SymmetricTopParameterNames)) then
    Result := SymmetricTopParameterNames[AIndex]
  else
    Result := inherited HelpOnParameter(AIndex);
end;

{ TSymmetricTopPerturbation }

const
  SymmetricTopPerturbationSettingNames: array[LPI..LPI+9] of string = (
'Twice power of J(J+1)',
'Power of Jz; negative implies [[Jz,J+-^pPower]+,Jz^(zPower-1)]+ ',
'Power of J+-',
'Power of l',
'Power of l+-; < 0 implies opposite sign to change in K',
'K the perturbation applies to',
'Twice power of S(S+1)',
'Power of Sz',
'Power of S+-',
'Power of N.S'
);

constructor TSymmetricTopPerturbation.Create(AOwner: TComponent);
begin
  SavedProperties := SavedProperties + ',KSelect=all';
  inherited;
end;

function TSymmetricTopPerturbation.HelpOnSetting(AIndex: Integer): string;
begin
  if (AIndex >= Low(SymmetricTopPerturbationSettingNames))
     and (AIndex <= High(SymmetricTopPerturbationSettingNames)) then
    Result := SymmetricTopPerturbationSettingNames[AIndex]
  else
    Result := inherited HelpOnSetting(AIndex);
end;

function TSymmetricTopPerturbation.IsSettingQuantum(AIndex: Integer): Boolean;
begin
  Result := AIndex = LPI+5;
  if not Result then
    Result := inherited IsSettingQuantum(AIndex);
end;

type
  TopType = (opFull, opForward, opReverse);

function TSymmetricTopPerturbation.MatrixElement(J, S, Np, Kp, lp, N, K, l: TQuantumNumber): Double;

  function RotationalMatrixElement(N, Kp, K: TQuantumNumber{; OpType: TopType}):Double;
  begin
    Result := rl(N, Kp, K);
    if zPower <> 0 then
      if pPower <> 0 then
        if (zPower < -2) and Odd(zPower) then
          Result := Result * (K+Kp) * (IntPower(K,Abs(zPower)-1) + IntPower(Kp,Abs(zPower)-1))
        else
          Result := Result * (IntPower(K,Abs(zPower)) + IntPower(Kp,Abs(zPower)))
      else
        Result := Result * IntPower(K,Abs(zPower));
    if JPower > 0 then
      Result := Result * IntPower(N*(N+2), JPower div 2);
    Result := Result / IntPower(2,Abs(zPower) + JPower)
  end;
var
  dNMax: Integer;
  HasRotPart: Boolean;
  q: TQuantumNumber;
begin
  Result := 0;
  if KSelect > -9999 then begin
    { For the moment, require K selection to be diagonal in |K| }
    { TODO: Allow K select to work for elements off diagonal in K }
    if (Abs(K) <> KSelect) or (Abs(Kp) <> KSelect) then
      Exit;
  end;
  if Abs(lChange)*2 <> Abs(l-lp) then
    Exit;
  if (pPower+SpPower)*2 <> Abs(K-Kp) then
    Exit;
  if lChange*pPower*(l-lp)*(K-Kp) < 0 then
    Exit; { lChange < 0 => Change in l has opposite sign to change in K }
  dNMax := SzPower + Abs(SpPower);
  if Abs(Np-N) > 2*dNMax then
    Exit;
  HasRotPart := (JPower or zPower or pPower) <> 0;
  if dNMax = 0 then begin
    Result := RotationalMatrixElement(N, Kp, K);
    if (NSPower or SPower) <> 0 then
      Result := Result * NSOperator(J, N, S, NSPower, SPower)
  end else if SzPower > 0 then begin
    if (SpPower or NSPower or SPower) <> 0 then
      raise NotYet;
    if HasRotPart then
      Result := (RotationalMatrixElement(Np,Kp,K)*T1Sn(SzPower,0,J,S,Np,K,N,K)
                    +T1Sn(SzPower,0,J,S,Np,Kp,N,Kp)*RotationalMatrixElement(N,Kp,K))
    else
      Result := T1Sn(SzPower,0,J,S,Np,Kp,N,K);
  end else {SzPower <> 0} begin
    if (NSPower or SPower) <> 0 then
      raise NotYet;
    if SpPower <> 0 then begin
      if SpPower > 0 then
        q := 2
      else
        q := -2;
      if pPower+SpPower = -2*(Kp-K) then
        Result := RotationalMatrixElement(Np,Kp,K-2*SpPower)*rlSn(SpPower,q,J,S,Np,K-2*SpPower,N,K)
                  + rlSn(SpPower,q,J,S,Np,Kp,N,Kp+2*SpPower)*RotationalMatrixElement(N,Kp+2*SpPower,K)
      else
        Result := 0;
      if pPower+SpPower = 2*(Kp-K) then
        Result := Result + RotationalMatrixElement(Np,Kp,K+2*SpPower)*rlSn(SpPower,-q,J,S,Np,K+2*SpPower,N,K)
                         + rlSn(SpPower,-q,J,S,Np,Kp,N,Kp-2*SpPower)*RotationalMatrixElement(N,Kp-2*SpPower,K)
    end;
  end;
//  if dNMax + Ord(NSPower > 0) > 1 then
//    raise NotYet; { Initial implementation - only allow single spin operator }
//  if SpPower <> 0 then begin
//    if pPower+SpPower = -2*(Kp-K) then
//      Result := RotationalMatrixElement(Np,Kp,K-2*SpPower)*rlS(2*SpPower,J,S,Np,K-2*SpPower,N,K)
//                + rlS(2*SpPower,J,S,Np,Kp,N,Kp+2*SpPower)*RotationalMatrixElement(N,Kp+2*SpPower,K)
//    else
//      Result := 0;
//    if pPower+SpPower = 2*(Kp-K) then
//      Result := Result + RotationalMatrixElement(Np,Kp,K+2*SpPower)*rlS(-2*SpPower,J,S,Np,K+2*SpPower,N,K)
//                       + rlS(-2*SpPower,J,S,Np,Kp,N,Kp-2*SpPower)*RotationalMatrixElement(N,Kp-2*SpPower,K)
//  end else begin
//    Result := RotationalMatrixElement(N, Kp, K);
//    if NSPower > 0 then
//      Result := Result * NSOperator(J, N, S, NSPower)
//    else if SzPower <> 0 then
//      Result := T1S(0,J,S,Np,Kp,N,Kp)*Result
//                + RotationalMatrixElement(Np, Kp, K)*T1S(0,J,S,Np,K,N,K)
//  end;
  if lPower <> 0 then
    if lChange <> 0 then
      Result := Result * Sign(lp-l) * lPower { TODO: This may give non-Hermitian matrix elements }
    else
      Result := Result * Sign(lp) * lPower;
  Result := Result * Value;
end;

function TSymmetricTopPerturbation.PairName: string;
  function Show(const Letter: string; Power: Integer): string;
  begin
    case Power of
      0: Result := '';
      1: Result := Letter;
    else
      Result := Letter + '^' + IntToStr(Power)
    end;
  end;
begin
  Result := Show('J',JPower);
  if (zPower < -2) and Odd(zPower) then
    Result := Result + Show('Jz',Abs(zPower)-1) + 'Jz'
  else
    Result := Result + Show('Jz',Abs(zPower));
  Result := Result + Show('J+-', pPower) + Show('l', lPower);
  Result := Result + Show('Sz',SzPower) + Show('NS',NSPower) + Show('S',Spower);
  if SpPower < 0 then
    Result := Result + Show('S-+',-SpPower)
  else
    Result := Result + Show('S+-',SpPower);
  if lChange < 0 then
    Result := Result + Show('l-+',Abs(lChange))
  else
    Result := Result + Show('l+-',lChange);
  if KSelect > -9999 then
    Result := Result + '(K=' + (RootNode as TMixture).QuantumNumberToStr(KSelect) + ')';
end;

procedure TSymmetricTopPerturbation.SetSettingFromStr(AIndex: Integer; const s: string);
begin
  if (AIndex = LPI+5) and (s = 'all') then
    KSelect := -9999
  else
    inherited;
end;

procedure TSymmetricTopPerturbation.SetFromPertEquiv(const PE: TPertEquiv);
begin
  JPower := PE.JPower;
  pPower := PE.pPower;
  zPower := PE.zPower;
  lPower := PE.lPower;
  lChange := PE.lChange;
  SPower := PE.SPower;
  SzPower := PE.SzPower;
  SpPower := PE.SpPower;
  NSPower := PE.NSPower;
end;

function TSymmetricTopPerturbation.SettingToStr(AIndex: Integer): string;
begin
  if (AIndex = LPI+5) and (KSelect <= -9999) then
      Result := 'all'
  else
    Result := inherited SettingToStr(AIndex);
end;

function TSymmetricTopPerturbation.SymbolicMatrixElements(WithHyperfine: Boolean): TMatrixElements;
var
  PE: TPertEquiv;
begin
  PE.JPower := JPower;
  PE.pPower := pPower;
  PE.zPower := zPower;
  PE.lPower := lPower;
  PE.lChange := lChange;
  PE.SPower := SPower;
  PE.SzPower := SzPower;
  PE.SpPower := SpPower;
  PE.NSPower := NSPower;
  PE.Scale4 := 4;
  PE.Name := PairName;
  if (SymSelect >= 0) or (KSelect > -9999) then
    raise NotYet; { TODO: Add to TPertEquiv }
  Result := SymT.SymbolicRotationalMatrixElement(PE);
end;

{ TSymmetricTopNucleus }

function TsymmetricTopNucleus.FixupProperties: Exception;
begin
  CHI[ 0] := eQq/2;
  CHI[ 2] := eQq2*Sqrt(Six)/12;
  CHI[-2] := CHI[2];
  if Spin > 1 then
    QuadI := 1/threej(Spin, -Spin, 4, 0, Spin, Spin)
  else
    QuadI := 0;
  { Nuclear spin rotation }
  { Note that off-axis nuclei do not require Cxx = Cyy, though
    the difference between them will rarely be determinable }
  TC00 := -(Czz + 2*Cperp);
  TC20 := (3*Czz + TC00)/Sqrt(Six);
  TC00 := TC00/Sqrt(Three);
  Result:=inherited FixupProperties;
end;

function TsymmetricTopNucleus.HasIdentical: Boolean;
begin
  Result := True;
end;

const
  SymmetricTopNucleusParameterNames: array[1..4] of string = (
'Qudrupole Coupling Constant',
'Extra Qudrupole Coupling Constant for degenerate states',
'Component along symmetry axis of Nuclear Spin - Rotation interaction.',
'Average of components perpendicular to symmetry axis of Nuclear Spin - Rotation interaction.'
);

function TsymmetricTopNucleus.HelpOnParameter(AIndex: Integer): string;
begin
  if (AIndex >= Low(SymmetricTopNucleusParameterNames))
     and (AIndex <= High(SymmetricTopNucleusParameterNames)) then
    Result := SymmetricTopNucleusParameterNames[AIndex]
  else
    Result := inherited HelpOnParameter(AIndex);
end;

initialization
  AddTables;
  TSymmetricMolecule.RegisterClass(TSpecies);
  TSymmetricManifold.RegisterClass(TSymmetricMolecule);
  TSymmetricTop.RegisterClass(TSymmetricManifold);
  TSymmetricTopPerturbation.RegisterClass(TSymmetricManifold);
  TSymmetricTopNucleus.RegisterClass(TSymmetricTop);
  VList := MakeSortedVariables(['J', 'N', 'S', 'K', 'l'], [@eJ, @eN, @eS, @eK, @el]);
finalization
  FreeAndNil(VList);
end.

