unit SymmetricTop;
{$I utils.inc}{Copyright (C) 2011 C M Western; see commandline.pas for license}
interface
{ Functions common to symmetric and asymmetric tops }

uses QuantumNumber, Expr, Mixture, MatrixElements;

function T1S(q, J, S, Np, Kp, N, K: TQuantumNumber):Double;
function T1Sn(OpPower: Integer; q, J, S, Np, Kp, N, K: TQuantumNumber):Double;
function Sxyn(SxPower, SyPower: Integer; J, S, Np, Kp, N, K: TQuantumNumber):Double;
function rlSn(OpPower: Integer; q, J, S, Np, Kp, N, K: TQuantumNumber):Double;
function T1SE(q: Integer; J, S, Np, Kp, N, K: TExprNode):TExprNode;
function T1SnE(OpPower: Integer; q: Integer; J, S, Np, Kp, N, K, ScaleFactor: TExprNode):TExprNode;
function SxynE(SxPower, SyPower: Integer; delK: {Kp-K} Integer; J, S, Np, Kp, N, K, ScaleFactor: TExprNode):TExprNode;
function rlSnE(OpPower: Integer; q: Integer; J, S, Np, Kp, N, K, ScaleFactor: TExprNode):TExprNode;

function NSOperator(J, N, S: TQuantumNumber; NSPower, SPower: Integer):Double;
function NSOperatorE(J, N, S: TExprNode; NSPower, SPower: Integer):TExprNode;

procedure AddOpenShellIntensity(N, J, S: TExprNode; TransitionMoment: TTransitionMoment; Nsig: Integer; var Elements: TMatrixElements);

implementation

uses SysUtils, Math, MathSup, nj, njExpr;

function T1S(q, J, S, Np, Kp, N, K: TQuantumNumber):Double;
begin
  { Molecule fixed matrix elements of electron spin in tensor notation }
  Result := Minus(q+J-Np+N+S+Kp)*Sqrt((Np+1)*(N+1)*(S+1)*S*(S+2))/Two
            * threej(Np,-Kp,2,-q,N,K) * sixj(N, J, S, S, 2, Np);
end;

function T1Sn(OpPower: Integer; q, J, S, Np, Kp, N, K: TQuantumNumber):Double;
var
  Kpp, Npp: TQuantumNumber;
begin
  { Molecule fixed matrix elements of (electron spin)^n in tensor notation }
  if OpPower = 1 then
    Result := T1S(q, J, S, Np, Kp, N, K)
  else begin
    Kpp := Kp+q;
    Npp := Np - 2;
    Result := 0;
    repeat
      if Npp >= 0 then
        Result := Result + T1S(q, J, S, Np, Kp, Npp, Kpp) * T1Sn(OpPower-1, q, J, S, Npp, Kpp, N, K);
      Inc(Npp,2);
    until Npp > Np + 2;
  end;
end;

function Sxyn(SxPower, SyPower: Integer; J, S, Np, Kp, N, K: TQuantumNumber):Double;
var
  OpSign, dKp: Integer;
  Kpp, Npp: TQuantumNumber;
begin
  { Molecule fixed matrix elements of Sx^n or Sy^n; factor of i ommitted }
  if SxPower > 0 then begin
    OpSign := -1; { Want -T(1,1)+T(1,-1) }
    dKp := -SxPower;
    if SyPower > 0 then
      raise NotYet;
  end else begin
    OpSign := 1;  { Want T(1,1)+T(1,-1) }
    dKp := -SyPower;
  end;
  if dKp = -1 then begin
    { Single operator - essentially S+ or S- }
    Result := T1S(K-Kp, J, S, Np, Kp, N, K);
    if K > Kp then
      Result := OpSign*Result;
    Result := Result/Sqrt(2);
  end else begin
    Result := 0;
    Kpp := Kp-2;
    repeat { loop over K" = K-1, K+1 }
      Npp := Np-2;
      repeat { loop over N" = N'-1, N', N'+1 }
        if Npp >= 0 then
          if OpSign < 0 then
            Result := Result + Sxyn(1,0,J,S,Np,Kp,Npp,Kpp)*Sxyn(SxPower-1,0,J,S,Npp,Kpp,N,K)
          else
            Result := Result + Sxyn(0,1,J,S,Np,Kp,Npp,Kpp)*Sxyn(0,SyPower-1,J,S,Npp,Kpp,N,K);
        Inc(Npp, 2);
      until Npp > Np+2;
      Inc(Kpp, 4);
    until Kpp > Kp+2;
  end;
end;

function T1SE(q: Integer; J, S, Np, Kp, N, K: TExprNode):TExprNode;
var
  Node: TExprNode;
begin
  { Molecule fixed matrix elements of electron spin in tensor notation }
  Result := EN([-1, '^', EB([q, '+', J, '-', Np, '+', N, '+', S, '+', Kp])]);
  Result := EN([EB([2, '*', N, '+', 1]), '*', EB([2, '*', Np, '+', 1]),
                '*', EB([2, '*', S, '+', 1, '*', S]), '*', EB([S, '+', 1]), 'sqrt', '*', Result]);
  Node := EN(['threej', Np.Duplicate, MonadicMinus.CreateExprNode(Kp.Duplicate).Simplify, 1, -q, N.Duplicate, K.Duplicate]);
  Result := EN([Node, '*', Result]);
  Node := EB(['sixj', N, J, S, S, 1, Np]);
  { Cancel out common phase factor - J+N+S must be integer }
  Result := EN([-1, '^', EB([J, '+', N, '+', S, '*', -2]), '*', Result]);
  Result := EN([Node, '*', Result]);
  if Assigned(Result) then
    Result := Result.Simplify([siForce]);
end;

function T1SnE(OpPower: Integer; q: Integer; J, S, Np, Kp, N, K, ScaleFactor: TExprNode):TExprNode;
var
  Kpp, Npp: TExprNode;
  Ninc: Integer;
  Node: TExprNode;
begin
  { Molecule fixed matrix elements of (electron spin)^n in tensor notation }
  if OpPower = 1 then begin
    Result := T1SE(q, J, S, Np, Kp, N, K);
    if Assigned(ScaleFactor) then
      Result := EN([Result, '*', ScaleFactor.Duplicate]);
  end else begin
    Kpp := EB([Kp, '+', q]);
    Result := OperatorPlus.CreateExprNode;
    { The addition is done this way to defeat the simplification combining terms, as the combined
      terms will not necessarily give the correct results for cases where the intermediate quantum
      numbers are invalid }
    for Ninc := -1 to 1 do begin
      Npp := EB([Np, '+', Ninc]);
      Node := EN([T1SnE(1, q, J, S, Np, Kp, Npp, Kpp, ScaleFactor), '*', T1SnE(OpPower-1, q, J, S, Npp, Kpp, N, K, nil)]);
      if not Node.EqualTo(0) then
        Result.AddNode(Node)
      else
        Node.Free;
      Npp.Free;
    end;
    Kpp.Free;
    if Result.nChildren = 0 then
      FreeAndNil(Result)
    else if Result.nChildren = 1 then begin
      Node := Result;
      Result := Result.MoveChild(0);
      Node.Free;
    end;
    Node.Simplified := [];
  end;
end;

function SxynE(SxPower, SyPower: Integer; delK: {Kp-K} Integer; J, S, Np, Kp, N, K, ScaleFactor: TExprNode):TExprNode;
var
  OpSign, dKp, Kinc, Ninc: Integer;
  Npp, Kpp, Node: TExprNode;
begin
  { Molecule fixed matrix elements of Sx^n or Sy^n; factor of i ommitted }
  if SxPower > 0 then begin
    OpSign := -1; { Want -T(1,1)+T(1,-1) }
    dKp := -SxPower;
    if SyPower > 0 then
      raise NotYet;
  end else begin
    OpSign := 1;  { Want T(1,1)+T(1,-1) }
    dKp := -SyPower;
  end;
  if dKp = -1 then begin
    Result := nil;
    { Single operator - essentially S+ or S- }
    if Abs(delK) <> 1 then
      Exit;
    Result := T1SE(-delK, J, S, Np, Kp, N, K);
    if (delK < 0) and (OpSign < 0) then
      Result := MonadicMinus.CreateExprNode(Result);
    Result := EN([Result, '/', EN([2, 'sqrt'])]);
    if Assigned(ScaleFactor) then
      Result := EN([Result, '*', ScaleFactor.Duplicate]);
  end else begin
    Kinc := -1;
    Result := OperatorPlus.CreateExprNode;
    { The addition is done this way to defeat the simplification combining terms, as the combined
      terms will not necessarily give the correct results for cases where the intermediate quantum
      numbers are invalid }
    repeat { loop over K" = K-1, K+1 }
      Kpp := EB([Kp, '+', Kinc]);
      for Ninc := -1 to 1 do begin { loop over N" = N'-1, N', N'+1 }
        Npp := EB([Np, '+', Ninc]);
        if OpSign < 0 then
          Node := EN([SxynE(1,0,-KInc,J,S,Np,Kp,Npp,Kpp,ScaleFactor), '*', SxynE(SxPower-1,0,delK+Kinc,J,S,Npp,Kpp,N,K,nil)])
        else
          Node := EN([SxynE(0,1,-KInc,J,S,Np,Kp,Npp,Kpp,ScaleFactor), '*', SxynE(0,SyPower-1,delK+Kinc,J,S,Npp,Kpp,N,K,nil)]);
        Npp.Free;
        if not Node.EqualTo(0) then
          Result.AddNode(Node)
        else
          Node.Free;
      end;
      Kpp.Free;
      Inc(Kinc, 2);
    until Kinc > 1;
    if Result.nChildren = 0 then
      FreeAndNil(Result)
    else if Result.nChildren = 1 then begin
      Node := Result;
      Result := Result.MoveChild(0);
      Node.Free;
    end;
    // Node.Simplified := [];
  end;
end;

function rlSn(OpPower: Integer; q, J, S, Np, Kp, N, K: TQuantumNumber):Double;
  { Molecule fixed matrix elements of S+-, Sz  }
begin
  OpPower := Abs(OpPower);
  Result := T1Sn(OpPower, q, J, S, Np, Kp, N, K);
  if q <> 0 then
    Result := Math.IntPower(-q/Sqrt(Two),OpPower) * Result; { Note q is quantum number }
end;

function rlSnE(OpPower: Integer; q: Integer; J, S, Np, Kp, N, K, ScaleFactor: TExprNode):TExprNode;
  { Molecule fixed matrix elements of S+-, Sz  }
begin
  OpPower := Abs(OpPower);
  if q <> 0 then
    ScaleFactor := Expr.EB([2, 'sqrt', '*', -q, '^', Abs(OpPower), '*', ScaleFactor]);
  Result := T1SnE(OpPower, q, J, S, Np, Kp, N, K, ScaleFactor);
  if q <> 0 then
    ScaleFactor.Free;
end;

function NSOperator(J, N, S: TQuantumNumber; NSPower, SPower: Integer):Double; { N.S }
begin
  if NSPower > 0 then
    Result := Math.IntPower((J*(J+2)-S*(S+2)-N*(N+2))/8, NSPower)
  else
    Result := 1;
  if SPower > 0 then
    Result := Math.IntPower(S*(S+2)/Four, SPower div 2)
end;

function NSOperatorE(J, N, S: TExprNode; NSPower, SPower: Integer):TExprNode;
var
  Node1: TExprNode;
begin
  if NSPower > 0 then begin
    Result := EB([J, '+', 1, '*', J]);
    Node1 := EB([N, '+', 1, '*', N]);
    Result := EN([Result, '-', Node1]);
    Node1 := EB([S, '+', 1, '*', S]);
    Result := EN([Result, '-', Node1, '/', 2, '^', NSPower]);
  end else
    Result := TIntegerNode.Create(1);
  if SPower > 0 then begin
    Node1 := EB([S, '+', 1, '*', S]);
    Result := EN([Node1, '*', Result]);
  end;
end;

procedure AddOpenShellIntensity(N, J, S: TExprNode; TransitionMoment: TTransitionMoment; Nsig: Integer; var Elements: TMatrixElements);
{ N and J dependent part of transition moment matrix element }
var
  R: TMatrixElements;
  nMatrixElements: Integer;
  i, dj, sigs: Integer;
  Jp, Np, E, sj: TExprNode;
begin
  SetLength(R, Length(Elements)*(TransitionMoment.Rank+1));
  nMatrixElements := 0;
  for i := 0 to High(Elements) do begin
    Np := EB([N,'+',Elements[i].Signature[Nsig]]);
    for dj := -TransitionMoment.Rank div 2 to TransitionMoment.Rank div 2 do begin
      Jp := EB([J,'+',dJ]);
      E := Expr.EN([EB([2, '*', Jp, '+', 1]), '*', EB([2, '*', J, '+', 1]), 'sqrt']);
      E := Expr.EN([-1, '^', EB([Np, '+', S, '+', J]), '*', E]);
      sj := Expr.EB(['sixj', Np, Jp, S, J, N, TransitionMoment.Rank div 2]);
      E := Expr.EN([sj, '*', E]);
      if Minus(TransitionMoment.Rank) < 0 then
        E := Expr.EN([E, '*', -1]);
      E := Expr.EN([E, '*', Elements[i].Expr.Duplicate]);
      if Assigned(E) then begin
        R[nMatrixElements].Expr := E.Simplify;
        R[nMatrixElements].Evaluator := nil;
        R[nMatrixElements].CEvaluator := nil;
        sigs := Length(Elements[i].Signature);
        { Copy signature, adding Delta J }
        R[nMatrixElements].Signature := Elements[i].Signature;
        SetLength(R[nMatrixElements].Signature,sigs+1);
        R[nMatrixElements].Signature[sigs] := dj;
        Inc(nMatrixElements);
      end;
      Jp.Free;
    end;
    Np.Free;
  end;
  SetLength(R, nMatrixElements);
  DiscardMatrixElements(Elements);
  Elements := R;
end;

end.

