unit zMatrix;
{$I utils.inc}{Copyright (C) 2013 C M Western; see commandline.pas for license}

{ Internal representation of a Z matrix }

interface

uses SysUtils, MatrixVec, MNode, Mixture, NuclearInfo, MNodeExpr;

{ Simple vector functions }
type
  Txyz = array [1..3] of Double;
  function Diff(out V: Txyz; const A, B: Txyz):Double;
  procedure DivBy(var V: Txyz; Mag: Double);
  function NDiff(out V: Txyz; const A, B: Txyz):Double;
  function Cross(out V: Txyz; const A, B: Txyz):Double;
  function Dot(const A, B: Txyz):Double;
  function Angle(A, B, C: Txyz):Double;

  function TrimSmall(M: TMatrix; var Threshold: Double; out nNonZero: Integer; Symmetric: Boolean):Integer;
  procedure AngleIndex(Index, n: Integer; out a1, a2, a3: Word);

type

  TzMatrix = class;
  T4Integers = array [1..4] of Integer;

  TAtom = record
    Info: TNuclearInfo;
    Mass: Double;
    Index: Integer;    { Nucleus number for state; < 0 for dummy atoms }
    Name: string;      { Full name including mass number if present - 13C2, for example }
    ShortName: string; { Name excluding mass number - C2, for example }
    xyz:  Txyz; { co-ordinates in original axis system }
    pxyz: Txyz; { co-ordinates in principal axis system }
  end;
  TAtomarray = array of TAtom;

  TNotifyingActiveMNodeWithArray = class(TActiveMNodeWithArray)
  public
    Precision: Integer;
    procedure ClearArrays; override;
    procedure SetSetting(AIndex: Integer; const AValue: Integer); override;
  end;

  { TzMatrix }

  TzMatrix = class(TNotifyingActiveMNodeWithArray)
  private
    nDummy: Integer;
    NoRotate: Boolean;
  public
    Atoms: TAtomArray;
    CofM: Txyz; { In original axis system }
    Mass: Double;
    Ixx, RotConsts: TColumnVector;
    Itrans: TMatrix;
    nRotations: Integer;
    RotationName: array [1..3] of Char;
    procedure ClearArrays; override;
    procedure ShowZmatrix(Sender: TObject);
    procedure AddHandlers; override;
    procedure NumberAtoms;
    function nCalculatedParameters: Integer; override;
    function CalculatedParameter(Index: Integer):Double; override;
    function CalculatedParameterName(Index: Integer):string; override;
    function FindCartesian(Verbose: Boolean): Boolean;
    function FindCofM(Verbose: Boolean): Boolean;
    function FindPrincipalAxes(Verbose: Boolean): Boolean;
    procedure MassScale(B: TMatrix; MassPower:Integer);
    procedure AddRT(B: TMatrix; FirstRow: Integer; InPrincipalAxis, MassScaled:Boolean);
    function CheckAtomNames: Boolean;
    function FindNuclearInfo(s: string; out MassNumber: Integer; out Symbol: string; out Index: Integer):TNuclearInfo;
    procedure CheckEmpty(Index, Col: Integer);
    function FindAtom(Index, Col: Integer):Integer; overload;
    function FindAtom(const s: string):Integer; overload;
    procedure NextRealAtom(var ia: Integer);
    function ApplyToStateEx(Verbose: Boolean; UsePrincipalAxis: Boolean; State: TState):Boolean;
    procedure SetFromState(State: TState; Verbose: Boolean);
    procedure ShowXYZ(const s: string; TheMass: Double; xyz: Txyz);
    function Apply(Verbose: Boolean):string; overload; override;
    function Apply(Verbose: Boolean; UsePrincipalAxis: Boolean):string; overload;
//    function nameNucleusXYZ(Index: Integer):string;
    function nameXYZ(Index: Integer):string;
    function HelpOnSetting(AIndex: Integer): string; override;
    destructor Destroy; override;
  published
    property AverageMass: Boolean index 2 read GetFlag write SetFlag;
    procedure LoadFromState(const State: string);
  end;

implementation

uses {$IFNDEF CONSOLE} zMatrixFrm, {$ENDIF}
     Math, MathSup, EigLA, MatrixOps, Units, Worker, Vib;

function TryStrToIntEx(const s: string; out Value: Integer): Boolean;
begin
  { Do not accept hex numbers }
  if (Length(s) > 0) and (s[1] in ['0'..'9']) then
    Result := TryStrToInt(s, Value)
  else
    Result := False;
end;

function Diff(out V: Txyz; const A, B: Txyz):Double;
var
  xyz: Integer;
begin
  Result := 0;
  for xyz := 1 to 3 do begin
    V[xyz] := A[xyz] - B[xyz];
    Result := Result + Sqr(V[xyz]);
  end;
  Result := Sqrt(Result);
end;

procedure DivBy(var V: Txyz; Mag: Double);
var
  xyz: Integer;
begin
  if Abs(Mag) > LinearTest then
    for xyz := 1 to 3 do
      V[xyz] := V[xyz]/Mag;
end;

function NDiff(out V: Txyz; const A, B: Txyz):Double;
begin
  Result := Diff(V, A, B);
  DivBy(V, Result);
end;

function Cross(out V: Txyz; const A, B: Txyz):Double;
begin
  V[1] := A[2]*B[3] - A[3]*B[2];
  V[2] := A[3]*B[1] - A[1]*B[3];
  V[3] := A[1]*B[2] - A[2]*B[1];
  Result := Sqrt(Sqr(V[1])+Sqr(V[2])+Sqr(V[3]));
end;

function Dot(const A, B: Txyz):Double;
begin
  Result := A[1]*B[1] + A[2]*B[2] + A[3]*B[3];
end;

function Angle(A, B, C: Txyz):Double;
var
  T: Txyz;
  RCB, RAB, DotRes: Double;
begin
  RCB := Diff(C,C,B);
  RAB := Diff(A,A,B);
  Result := Dot(C,A)/(RCB*RAB);
  if not IsNanOrInf(Result) then begin
    if Abs(Result) > 0.75 then begin
      { ArcCos very numerically unstable for values around 1, so use cross product instead }
      DotRes := Result;
      Result := Cross(T,C,A)/(RCB*RAB);
      Result := Abs(ArcSin(Result));
      if DotRes < 0 then
        Result := Pi-Result;
    end else
      Result := ArcCos(Result);
  end;
end;

function TrimSmall(M: TMatrix; var Threshold: Double; out nNonZero: Integer; Symmetric: Boolean):Integer;
var
  t: Double;
  p, q: Integer;
  function Lim(r: Integer):Integer;
  begin
    if Symmetric then
      Result := r
    else
      Result := M.ColCount;
  end;
begin
  { To avoid spurious rotations, force small elements to zero }
  nNonZero := 0;
  { Find largest element }
  t := 0;
  for p := 1 to M.RowCount do
    for q := 1 to Lim(p) do
      t := Max(t, Abs(M[p,q]));
  { Discard small elements }
  Threshold := t * Threshold;
  Result := 0; { Number of elements discarded }
  for p := 1 to M.RowCount do
    for q := 1 to M.ColCount do
      { Note: must scan entire matrix, as small elements do not necessarily follow
        symmetry constraint }
      if (Abs(M[p,q]) < Threshold) then begin
        if M[p,q] <> 0 then begin
          Inc(Result);
          M[p,q] := 0;
          if Symmetric then
            M[q,p] := 0;
        end;
      end else
        Inc(nNonZero);
end;

{ TzMatrix }

procedure TzMatrix.AddHandlers;
begin
  inherited;
  {$IFNDEF CONSOLE}
  AddHandler('View...', 'View and edit Z Matrix', {$IFDEF FPC}@{$ENDIF}ShowZmatrix);
  {$ENDIF}
end;

procedure TzMatrix.AddRT(B: TMatrix; FirstRow: Integer; InPrincipalAxis, MassScaled: Boolean);
var
  i, r, k, k1, k2: Integer;
  RI: TColumnVector;
  M: Double;
begin
  if Mass = 0 then
    Exit;
  { Add rotations and translations to matrix }
  B.RowCount := FirstRow+nRotations+2;
  RI := TColumnVector.Create(3);
  for i := 0 to High(Atoms) do
    if Atoms[i].Mass > 0 then begin
      r := 0;
      if MassScaled then
        M := Sqrt(Atoms[i].Mass)
      else
        M := 1;
      for k := 1 to 3 do
        { Check for zero moments of Inertia }
        if Abs(Ixx[k]) > LinearTest then begin
          k1 := k mod 3 + 1;
          k2 := k1 mod 3 + 1;
          RI[k] := 0;
          RI[k2] := Atoms[i].pxyz[k1]/Sqrt(Ixx[k])*M;
          RI[k1] := -Atoms[i].pxyz[k2]/Sqrt(Ixx[k])*M;
          if not InPrincipalAxis then
            RI.PreMultiply(Itrans, False); { Transform rotation back to input axis system }
          for k1 := 1 to 3 do
            B[FirstRow+r,3*i+k1] := RI[k1];
          Inc(r);
        end;
      for k := 0 to 2 do
        for k1 := 0 to 2 do
          if k = k1 then
            B[FirstRow+r+k,3*i+k1+1] := M/Sqrt(Mass)
          else
            B[FirstRow+r+k,3*i+k1+1] := 0;
    end;
  RI.Free;
end;

function TzMatrix.Apply(Verbose: Boolean):string;
begin
  { TODO: zMatrix should probably have an InPrincipalAxis flag to avoid this ambiguous function }
  Result := Apply(Verbose, False);
end;

function TzMatrix.Apply(Verbose: Boolean; UsePrincipalAxis: Boolean): string;
begin
  LastError := '';
  if EffectiveOwner = Owner then
    ApplyToStateEx(Verbose,UsePrincipalAxis,nil)
  else
    SetFromState(EffectiveOwner as TElectronicState, Verbose); { Dummy object }
  Result := LastError;
end;

function TzMatrix.nameXYZ(Index: Integer): string;
begin
  Result := 'xyz'[(Index-1) mod 3 + 1];
end;

//function TzMatrix.nameNucleusXYZ(Index: Integer): string;
//begin
//  Result := nameXYZ(Index);
//  Index := (Index-1) div 3;
//  if Index > High(Atoms) then
//    Result := Result + 'Nuc' + IntToStr(Index+1) { Probably never called }
//  else
//    Result := Result + Atoms[Index].Name;
//end;

function TzMatrix.ApplyToStateEx(Verbose: Boolean; UsePrincipalAxis: Boolean; State: TState): Boolean;
var
  nAtoms, i: Integer;
  N: TMolecularNucleus;
begin
  Result := FindPrincipalAxes(Verbose) or (Length(A) = 0);
  if Result then begin
    nAtoms := 0;
    for i := 0 to High(A) do
      if Atoms[i].Mass > 0 then
        Inc(nAtoms);
    if not Assigned(State) then
      State := Owner as TState; { In fact must be TElectronicState }
    (State.Owner.Owner as TVibratingMolecule).nNuclei := nAtoms;
    nAtoms := 0;
    for i := 0 to High(Atoms) do
      if Atoms[i].Mass > 0 then begin
        N := State.Nucleus(nAtoms) as TMolecularNucleus;
        N.Mass := Atoms[i].Mass;
        N.Name := Atoms[i].Name;
        if UsePrincipalAxis then begin
          N.x := Atoms[i].pxyz[1];
          N.y := Atoms[i].pxyz[2];
          N.z := Atoms[i].pxyz[3];
        end else begin
          N.x := Atoms[i].xyz[1];
          N.y := Atoms[i].xyz[2];
          N.z := Atoms[i].xyz[3];
        end;
        Inc(nAtoms);
      end;
  end;
end;

function TzMatrix.nCalculatedParameters: Integer;
var
  n: Integer;
begin
  n := Length(Atoms);
  Result := Sqr(n) + n*(n-1)*(n-2);
end;

procedure AngleIndex(Index, n: Integer; out a1, a2, a3: Word);
var
  a3Bump: Boolean;
begin
  { Given angle index and number of atoms (n), decode
    the three atom numbers }
  DivMod(Index, n-2, a1, a3);
  DivMod(a1, n-1, a1, a2);
  if a2 >= a1 then
    Inc(a2);
  a3Bump := a3 >= a1;
  if a3Bump then
    Inc(a3);
  if a3 >= a2 then
    Inc(a3);
  if not a3Bump and (a3 >= a1) then
    Inc(a3);
end;

function TzMatrix.CalculatedParameter(Index: Integer): Double;
var
  n, a1, a2, a3: Word;
  d: Txyz;
begin
  n := Length(Atoms);
  if Index <= Sqr(n) then begin
    DivMod(Index-1, n, a1, a2);
    Result := Diff(d, Atoms[a1].xyz, Atoms[a2].xyz)
  end else begin
    Dec(Index, Sqr(n)+1);
    AngleIndex(Index, n, a1, a2, a3);
    Result := radtodeg(Angle(Atoms[a1].xyz, Atoms[a2].xyz, Atoms[a3].xyz));
  end;
end;

function TzMatrix.CalculatedParameterName(Index: Integer): string;
var
  n, a1, a2, a3: Word;
begin
  n := Length(Atoms);
  if Index <= Sqr(n) then begin
    DivMod(Index-1, n, a1, a2);
    Result := 'r['+Atoms[a1].Name+','+Atoms[a2].Name+']';
  end else begin
    Dec(Index, Sqr(n)+1);
    AngleIndex(Index, n, a1, a2, a3);
    Result := 'angle['+Atoms[a1].Name+','+Atoms[a2].Name+','+Atoms[a3].Name+']';
  end;
end;

function TzMatrix.CheckAtomNames: Boolean;
var
  i, Index, n, MassNumber: Integer;
  Nuc: TNuclearInfo;
  Symbol: string;
begin
  Result := True;
  n := 0;
  nDummy := 0;
  for i := 0 to High(A) do
    if A[i][0] <> '' then begin
      Nuc := FindNuclearInfo(Trim(A[i][0]), MassNumber, Symbol, Index);
      if Nuc.Mass < 0 then begin
        HandleError(i,0, 'Bad atom '+IntToStr(i+1)+':'+A[i][0]+' ');
        Result := False;
      end else begin
        SetLength(Atoms, n+1);
        SetLength(Rows, n+1);
        Atoms[n].Info := Nuc;
        Atoms[n].Name := Trim(A[i][0]);
        if AverageMass and (MassNumber < 0) then
          Atoms[n].Mass := Nuc.StandardWeight
        else
          Atoms[n].Mass := Nuc.Mass;
        if Atoms[n].Mass <= 0 then begin
          Inc(nDummy);
          Atoms[n].Index := -1
        end else
          Atoms[n].Index := n-nDummy;
        Atoms[n].Name := Atoms[n].Name;
        if Index > 0 then
          Atoms[n].ShortName := Nuc.Symbol+IntToStr(Index)
        else
          Atoms[n].ShortName := Nuc.Symbol;
        Rows[n] := i;
        Inc(n);
      end;
    end;
end;

procedure TzMatrix.CheckEmpty(Index, Col: Integer);
var
  i: Integer;
begin
  Index := Rows[Index];
  for i := Col to High(A[Index]) do
    if A[Index][i] <> '' then
      HandleError(Index, Col, 'Extra Items');
end;

procedure TzMatrix.ClearArrays;
begin
  NoRotate := False;
  FreeAndNil(RotConsts);
  FreeAndNil(Ixx);
  FreeAndNil(Itrans);
  inherited;
end;

destructor TzMatrix.Destroy;
begin
  inherited;
end;

function TzMatrix.FindAtom(Index, Col: Integer): Integer;
var
  r: Integer;
  s: string;
begin
  Result := -1;
  r := Rows[Index];
  if Col > High(A[r]) then begin
    HandleError(r,Col,'Missing Atom');
    Exit;
  end;
  s := A[r][Col];
  Result := FindAtom(s);
  if Result < 0 then
    HandleError(r,Col,'Atom '+s+' not found');
end;

function TzMatrix.FindAtom(const s: string): Integer;
var
  i: Integer;
begin
  Result := -1;
  if TryStrToIntEx(s, Result) then begin
    if (Result <= 0) or (Result > Length(Atoms)) then
      Result := -1;
    Dec(Result);
  end else begin
    Result := -1;
    for i := 0 to High(Atoms) do
      if Atoms[i].Name = s then begin
        Result := i;
        Break;
      end;
    if Result < 0 then
      for i := 0 to High(Atoms) do
        if Atoms[i].ShortName = s then begin
          Result := i;
          Break;
        end;
  end;
end;

function TzMatrix.FindCartesian(Verbose: Boolean): Boolean;
var
  PlaneSpecifier: string;
  function IsCartesian(Index: Integer): Boolean;
  var
    s: string;
    r, i, xyz: Integer;
  begin
    Result := False;
    r := Rows[Index];
    if High(A[r]) < 1 then
      Exit;
    s := A[r][1];
    if (Index = 0) and (Copy(LowerCase(s),1,5) = 'norot') then begin
      s := '';
      NoRotate := True;
    end;
    { Blank or negative first atom number indicates cartesians }
    { But first row can have a plane specifier }
    if (Index = 0) and (High(A[r]) = 1) then begin
      PlaneSpecifier := s;
      Exit;
    end;
    if s <> '' then
      Exit;
    if TryStrToInt(s, i) and (i > 0) then
      Exit;
    Result := True; { Must be Cartesian }
    xyz := 1;
    for i := 2 to 6 do
      if (i <= High(A[r])) and (A[r][i] <> '')  then
        if xyz > 3 then
          HandleError(r, i, 'More than 3 Cartesian co-ordinates')
        else begin
          Atoms[Index].xyz[xyz] := HandleValue(Index, i);
          Inc(xyz);
        end;
    if xyz <> 4 then
      HandleError(r, 1, 'Less than 3 Cartesian co-ordinates')
  end;
var
  Atom1, Atom2, Atom3, xyz, q, Index: Integer;
  Distance, Angle, Dihedral, mab, mbc, mn: Double;
  s, c: Extended;
  bc, ab, n, D2: Txyz;
  function PrepAtom(Index: Integer):Boolean;
    procedure AtomErr(Col: Integer; Msg: string);
    begin
      HandleError(Index,Col,'In z Matrix for '+Atoms[Index].Name+': '+Msg);
    end;
    function BadAtom(Atom: Integer; Col: Integer):Boolean;
    begin
      Result := True;
      if Atom = Index then
        AtomErr(Col,'Atom refers to itself')
      else if Atom > Index then
        AtomErr(Col,'Atom refers to a later atom')
      else if Atom >= 0 then
        Result := False; { OK }
    end;
  begin
    Result := False; { no more calculations should be done }
    if IsCartesian(Index) then
      Exit;
    if Index > 0 then begin
      Atom1 := FindAtom(Index,1);
      if BadAtom(Atom1, 1) then
        Exit;
      Distance := HandleValue(Index,2);
      if Index > 1 then begin
        Atom2 := FindAtom(Index,3);
        if BadAtom(Atom2, 3) then
          Exit;
        if Atom1 = Atom2 then begin
          AtomErr(3,'Atoms not different');
          Exit;
        end;
        Angle := HandleValue(Index,4);
        if Index > 2 then begin
          Atom3 := FindAtom(Index,5);
          if BadAtom(Atom3, 5) then
            Exit;
          if (Atom3 = Atom2) or (Atom3 = Atom1) then begin
            AtomErr(5,'Atoms not different');
            Exit;
          end;
          Dihedral := HandleValue(Index,6);
        end else
          CheckEmpty(Index,5)
      end else
        CheckEmpty(Index, 3)
    end else
      CheckEmpty(Index,1);
    Result := True;
  end;
var
  Atom2Disp, Atom3Disp: Integer;
begin
  { Set up array of real atom coordinates }
  ClearArrays;
  TrimArray(A);
  Result := CheckAtomNames;
  if not Result then
    Exit;
  if Length(A) = 0 then
    Exit;
//  for Index := 1 to High(A) do begin
//    if PrepAtom(Index) then begin
//      if Index >=2 then begin
//        SinCos(DegToRad(Angle), sa, ca);
//        if Index >= 3 then
//          SinCos(DegToRad(Dihedral), sd, cd)
//        else begin
//          cd := 1;
//          sd := 0;
//        end;
//      end else begin
//        ca := 1;
//        sa := 0;
//        cd := -1;
//        sd := 0;
//      end;
//      A[Index].xyz[1] := -ca*(A[Atom1].xyz[1]+Distance) - sa*A[Atom1].xyz[2];
//      A[Index].xyz[2] := sa*cd*(A[Atom1].xyz[1]+Distance) - ca*cd*A[Atom1].xyz[2] - sd*A[Atom1].xyz[3];
//      A[Index].xyz[3] := sa*sd*(A[Atom1].xyz[1]+Distance) - ca*sd*A[Atom1].xyz[2] + cd*A[Atom1].xyz[3];
//    end;
//    WriteLn(Log, A[Index].Name:4, A[Index].xyz[1]:10, A[Index].xyz[2]:10, A[Index].xyz[3]:10);
//  end;
  PlaneSpecifier := '';
  Atom2Disp := 1;
  Atom3Disp := 0;
  if not IsCartesian(0) then begin
    Atoms[0].xyz[1] := 0;
    Atoms[0].xyz[2] := 0;
    Atoms[0].xyz[3] := 0;
    if Length(PlaneSpecifier) in [1,2] then begin
      CheckEmpty(0,2);
      { First character specifies axis for first bond }
      Atom2Disp := Ord(PlaneSpecifier[1])-Ord('w');
      { Optional second character specifies plane for first three atoms }
      if Length(PlaneSpecifier) = 2 then
        Atom3Disp := Ord(PlaneSpecifier[2])-Ord('w');
      if (Atom3Disp = Atom2Disp) or not(Atom2Disp in [1,2,3]) or not (Atom3Disp in [0,1,2,3]) then
        HandleError(0,1,'Bad molecular plane specification:'+PlaneSpecifier);
    end else
      CheckEmpty(0,1);
  end;
  Atoms[0].pxyz := Atoms[0].xyz;
  Result := LastError = '';
  if Length(Atoms) = 1 then
    Exit;
  if not IsCartesian(1) then begin
    Atom1 := FindAtom(1,1);
    if Atom1 <> 0 then
      HandleError(Rows[1],1,'Unexpected atom');
    Atoms[1].xyz := Atoms[0].xyz;
    Atoms[1].xyz[Atom2Disp] := Atoms[1].xyz[Atom2Disp]+HandleValue(1,2);
    CheckEmpty(Rows[1],3);
  end;
  Atoms[1].pxyz := Atoms[1].xyz;
  if (Length(Atoms) > 2) and not IsCartesian(2) and PrepAtom(2) then begin
    SinCos(DegToRad(Angle), s, c);
    for xyz := 1 to 3 do
      Atoms[2].xyz[xyz] := Atoms[Atom1].xyz[xyz]-Distance*c*Sign(Atoms[Atom1].xyz[xyz]-Atoms[Atom2].xyz[xyz]);
    if Atom3Disp = 0 then
      for xyz := 1 to 3 do
        if Atoms[Atom2].xyz[xyz] = Atoms[Atom1].xyz[xyz] then begin
          Atom3Disp := xyz;
          Break;
        end;
    if Atom3Disp = 0 then
      HandleError(Rows[2],1,'Atoms not in simple plane');
    Atoms[2].xyz[Atom3Disp] := Atoms[2].xyz[Atom3Disp] + Atoms[Atom1].xyz[Atom3Disp]+Distance*s;
    Atoms[2].pxyz := Atoms[2].xyz;
  end;
  for Index := 3 to High(Atoms) do begin
    if not PrepAtom(Index) then
      Continue;
    { NeRF algorithm - Journal of Computational Chemistry 26, 1063 (2005) }
    SinCos(DegToRad(Angle), s, c);
    D2[1] := -Distance*c;
    D2[2] := Distance*s;
    if Dihedral = 180 then begin
      D2[3] := 0; { ... keeps molecules more accurately planar }
      D2[2] := -D2[2];
    end else begin
      SinCos(DegToRad(Dihedral), s, c);
      D2[3] := D2[2]*s;
      D2[2] := D2[2]*c;
    end;
    mbc := NDiff(bc, Atoms[Atom1].xyz, Atoms[Atom2].xyz);
    if mbc < LinearTest then
      HandleError(Index,1,'Atoms 1 and 2 coincident')
    else begin
      if Atom3 >= 0 then
        mab := Diff(ab, Atoms[Atom2].xyz, Atoms[Atom3].xyz)
      else begin
        HandleError(Rows[Index],5,'No atom 3');
        mab := 0;
      end;
      if mab < LinearTest then
        HandleError(Rows[Index],1,'Atoms 2 and 3 coincident')
      else begin
        mn := Cross(n, ab, bc);
        DivBy(n,mn);
        if ((Cross(ab, n, bc) <= LinearTest) or (mn <= LinearTest)) and (Angle <> 180) then
          { Note - linear case is accepted here as a special case }
          HandleError(Rows[Index], 1, 'Warning: atoms 12 and 23 collinear');
        for xyz := 1 to 3 do
          Atoms[Index].xyz[xyz] := bc[xyz]*D2[1] + ab[xyz]*D2[2] + n[xyz]*D2[3] + Atoms[Atom1].xyz[xyz];
      end;
    end;
    Atoms[Index].pxyz := Atoms[Index].xyz;
  end;
  if Verbose then begin
    WriteLn(Log, 'Z Matrix as input');
    WriteLn(Log, 'Atom    Mass       x', 'y':Precision, 'z':Precision);
    for q := 0 to High(A) do
      ShowXYZ(Atoms[q].Name, Atoms[q].Mass, Atoms[q].xyz);
  end;
  Result := LastError = '';
end;

function TzMatrix.FindCofM(Verbose: Boolean): Boolean;
var
  n, xyz: Integer;
begin
  LastError := '';
  Result := False;
  { Shift to Centre of mass }
  CofM[1] := 0; CofM[2] := 0; CofM[3] := 0;
  Mass := 0;
  nRotations := 0;
  if not FindCartesian(Verbose) or (High(Atoms) = 0) then
    Exit;
  for n := 0 to High(Atoms) do
    if Atoms[n].Mass > 0 then begin
      for xyz := 1 to 3 do
        CofM[xyz] := CofM[xyz] + Atoms[n].Mass*Atoms[n].xyz[xyz];
      Mass := Mass + Atoms[n].Mass;
    end;
  if Mass <= 0 then
    Exit;
  DivBy(CofM, Mass);
  if Verbose then
    ShowXYZ('CofM', Mass, CofM); { In original axis system }
  Result := True;
end;

function TzMatrix.FindNuclearInfo(s: string; out MassNumber: Integer; out Symbol: string; out Index: Integer): TNuclearInfo;
var
  p: string;
  M: Double;
  AtomicNumber: Integer;
begin
  Result.Mass := -1;
  Result.StandardWeight := -2;
  Index := -1;
  MassNumber := -1;
  Symbol := '';
  if TryStrToIntEx(s, AtomicNumber) then
    { Integer is atomic number }
    Result := FindMostAbundantNucleus(AtomicNumber)
  else if TryStrToFloat(s, M) then begin
    { Decimal point implies use given mass }
    Result := FindNucleus(M);
    if (Pos('.', s) > 0) or (Result.Mass < 0) then begin
      Result.Mass := M;
      Result.StandardWeight := M;
    end;
    if Result.Mass < 0 then begin
      { Not found - just take given mass }
      if M = 0 then
        Result.Symbol := 'X'
      else
        Result.Symbol := '?';
    end;
  end else begin
    p := '';
    while s[1] in ['0'..'9'] do begin
      { e.g. 15N }
      p := p + s[1];
      Delete(s,1,1);
    end;
    if p = '' then begin
      Result := FindMostAbundantNucleus(s,True);
    end else if TryStrToIntEx(p, MassNumber) then
      Result := FindNucleus(MassNumber, s);
    Symbol := s; { Name without isotope specifier }
    Index := TrimNucleusNumber(Symbol);
  end;
end;

function TzMatrix.FindPrincipalAxes(Verbose: Boolean): Boolean;
var
  LargestNeg, nNeg: Integer;
  Small: Double;
  procedure CountNegative;
  var
    p, q, x: Integer;
    V, mV: Double;
  begin
    { For each column, find the element with the largest absolute value, and
      count the number of columns for which this is negative. Additionally keep
      track of the largest negative value }
    nNeg := 0;
    LargestNeg := 0;
    mV := -10*Small;
    for q := 1 to 3 do begin
      x := 1;
      V := Abs(Itrans[1,q]);
      for p := 2 to 3 do
        if Abs(Itrans[p,q]) > V then begin
          x := p;
          V := Abs(Itrans[p,q])
        end;
      if Itrans[x,q] < -10*Small then begin
        Inc(nNeg);
        if Itrans[x,q] < mV then begin
          mV := Itrans[x,q];
          LargestNeg := q;
        end;
      end;
    end;
  end;
var
  Thresh, Sum: Double;
  p, q, n, nNonZero, llNeg: Integer;
begin
  Result := FindCofM(Verbose);
  if not Result then
    Exit;
  Small := (EffectiveOwner as TElectronicState).Molecule.SmallCoefficient;
  { Moment of inertia tensor }
  Itrans := TMatrix.Create(3,3);
  Itrans.Zero;
  for n := 0 to High(Atoms) do
    with Atoms[n] do begin
      for p := 1 to 3 do begin
        for q := 1 to p-1 do begin
          Itrans[p,q] := Itrans[p,q] - Mass * (xyz[p]-CofM[p])*(xyz[q]-CofM[q]);
          Itrans[q,p] := Itrans[p,q];
        end;
      for q := 1 to 3 do
        if p <> q then
          Itrans[p,p] := Itrans[p,p] + Mass * Sqr(xyz[q]-CofM[q]);
    end;
  end;
  if Verbose then begin
    WriteLn(Log, 'Moment of Inertia Tensor');
    Itrans.Write(Log, 'I', {$IFDEF FPC}@{$ENDIF}nameXYZ, {$IFDEF FPC}@{$ENDIF}nameXYZ, Precision);
  end;
  { To avoid spurious rotations, force small elements to zero }
  Thresh := Small;
  n := TrimSmall(Itrans, Thresh, nNonZero, True);
  if Verbose and (n > 0) then begin
    WriteLn(Log, 'Moment of Inertia Tensor with ',n,' elements < ', Thresh, ' removed');
    Itrans.Write(Log, 'I', {$IFDEF FPC}@{$ENDIF}nameXYZ, {$IFDEF FPC}@{$ENDIF}nameXYZ, Precision);
  end;
  Ixx := TColumnVector.Create(3);
  { Try and preserve axis system }
  { If matrix already OK, leave as is }
  if (nNonZero <= 3) and not NoRotate then
    { Are inertial axes already in order? }
    for p := 2 to 3 do
      if Itrans[p-1,p-1] > Itrans[p,p]+Thresh then
        nNonZero := 4; { No }
  if nNonZero <= 3 then begin
    for p := 1 to 3 do
      Ixx[p] := Itrans[p,p];
    Itrans.Zero;
    Itrans.AddDiagonal(1);
  end else
    try
      Eigen(Ixx,Itrans,True);
    except
      on E: Exception do begin
        HandleError(0,0,E.Message);
        FreeAndNil(RotConsts)
      end;
    end;
  if not Assigned(Ixx) then
    Exit;
  { To minimize rotation, minimze the number of negative diagonal elements }
  { TODO: This algorithm not very useful if matrix not approximately diagonal }
  CountNegative;
  if Determinant(Itrans) < 0 then begin
    { Need to switch sign of an odd number of columns }
    if nNeg = 3 then begin
      if Verbose then
        WriteLn(Log, 'Switching sign of all elements of transformation matrix to ensure it is a rotation');
      Itrans.Scale(-1);
    end else begin
      if LargestNeg = 0 then
        LargestNeg := 2;
      if Verbose then
        WriteLn(Log, 'Switching sign of column ', LargestNeg, ' of transformation matrix to ensure it is a rotation');
      for q := 1 to 3 do
        Itrans[q,LargestNeg] := -Itrans[q,LargestNeg];
    end;
  end else if nNeg >= 2 then begin
    { Switch the sign of an even number of columns }
    llNeg := LargestNeg;
    for p := 1 to 2 do begin
      for q := 1 to 3 do
        Itrans[q,LargestNeg] := -Itrans[q,LargestNeg];
      CountNegative;
    end;
    if Verbose then
      WriteLn(Log, 'Switching sign of columns ', Min(llNeg,LargestNeg), ' and ', Max(llNeg,LargestNeg), ' of transformation matrix to minimize rotation');
  end;
  if Verbose then
    Ixx.Write(Log,'Principal Moments of Intertia',Precision+5);
  RotConsts := TColumnVector.Create(Ixx);
  for p := 1 to 3 do begin
    if RotConsts[p] <> 0 then
      RotConsts[p] := MomentOfInertiaTimesRotationalConstant[cm1]/RotConsts[p];
    RotationName[p] := '?';
    if Ixx[p] >= LinearTest then begin
      Inc(nRotations);
      RotationName[nRotations] := 'xyz'[p];
    end;
  end;
  if Verbose then begin
    RotConsts.Write(Log,'Rotational Constants / cm-1',Precision+5,Precision-2);
    WriteLn(Log,'Transformation to principal axis system');
    Itrans.Write(Log,' ',nil,nil,Precision);
  end;
  { Now transform to principal axis system }
  for n := 0 to High(Atoms) do begin
    for p := 1 to 3 do begin
      Sum := 0;
      for q := 1 to 3 do
        Sum := Sum + Itrans[q,p]*(Atoms[n].xyz[q]-CofM[q]);
      Atoms[n].pxyz[p] := Sum;
    end;
  end;
  if Verbose then begin
    WriteLn(Log, 'Co-ordinates in Principal Axis system');
    WriteLn(Log, 'Atom    Mass       a', 'b':Precision, 'c':Precision);
    for q := 0 to High(A) do
      ShowXYZ(Atoms[q].Name, Atoms[q].Mass, Atoms[q].pxyz);
  end;
  Result := True;
end;

function TzMatrix.HelpOnSetting(AIndex: Integer): string;
begin
  case AIndex of
    2: Result := 'True to use average nuclear masses rather than most abundant mass if isotope not specified';
  else
    Result := inherited HelpOnSetting(AIndex);
  end;
end;

procedure TzMatrix.MassScale(B: TMatrix; MassPower: Integer);
var
  i, j, k: Integer;
  M: Double;
begin
  for i := 0 to High(Atoms) do
    if Atoms[i].Mass > 0 then begin
      M := Atoms[i].Mass;
      if MassPower < 0 then
        M := 1/M;
      if Odd(Abs(MassPower)) then
        M := Sqrt(M);
      if Abs(MassPower) > 1 then
        M := Math.IntPower(M, Abs(MassPower) div 2);
      for k := 1 to 3 do
        for j := 1 to B.RowCount do
          B[j,3*i+k] := B[j,3*i+k] * M;
    end;
end;

procedure TzMatrix.NextRealAtom(var ia: Integer);
begin
  repeat
    Inc(ia)
  until Atoms[ia].Mass > 0;
end;

procedure TzMatrix.NumberAtoms;
var
  nChanges: Integer;
  procedure CheckName(s: string);
  var
    n, i: Integer;
    V: Double;
  begin
    n := 0;
    for i := 0 to High(Atoms) do
      if A[Rows[i]][0] = s then
        Inc(n);
    if n < 2 then
      Exit;
    Inc(nChanges);
    n := 0;
    for i := 0 to High(Atoms) do
      if A[Rows[i]][0] = s then begin
        Inc(n);
        if TryStrToFloat(s,V) then
          if Atoms[i].Info.Symbol <> '?' then
            A[Rows[i]][0] := IntToStr(Atoms[i].Info.MassNumber) + Atoms[i].Name + IntToStr(n)
          else
            A[Rows[i]][0] := s + '-' + IntToStr(n)
        else
          A[Rows[i]][0] := s + IntToStr(n);
      end;
  end;
var
  i: Integer;
begin
  if CheckAtomNames then
    repeat
      nChanges := 0;
      for i := 0 to High(Atoms) do
        CheckName(A[Rows[i]][0]);
      CheckAtomNames;
    until nChanges = 0;
end;

procedure TzMatrix.SetFromState(State: TState; Verbose: Boolean);
var
  i, n, Index, MassNumber: Integer;
  MassDecided: Boolean;
  Info: TNuclearInfo;
  s, Symbol: string;
begin
  n := (State.Owner.Owner as TVibratingMolecule).nNuclei;
  SetLength(A, n, 7);
  MassDecided := False;
  for i := 0 to n-1 do
    with State.Nucleus(i) as TMolecularNucleus do begin
      Info := FindNuclearInfo(Name, MassNumber, Symbol, Index);
      if SameValue(Info.Mass, Mass) or (MassNumber > 0) then begin
        { Exact mass }
        if not MassDecided and not (MassNumber > 0) and not SameValue(Info.Mass, Info.StandardWeight) then begin
          MassDecided := True;
          AverageMass := False;
        end;
        if (AverageMass or (Info.Abundance < 50)) and not SameValue(Info.Mass, Info.StandardWeight) then
          MassNumber := Info.MassNumber; { isotope must be shown to preserve mass }
      end else if SameValue(Info.StandardWeight, Mass) then begin
        { Average mass }
        if not MassDecided then begin
          MassDecided := True;
          AverageMass := True;
        end else if not AverageMass then
          Symbol := ''; { Force explicit specification of mass }
      end else
        Symbol := ''; { Force explicit specification of mass }

      if Symbol = '' then
        s := FloatToStr(Mass) { Specify mass, as mass is non-standard or symbol not recognized }
      else begin
        s := Symbol;
        if MassNumber > 0 then
          s := IntToStr(Info.MassNumber)+s;
        if Index >= 0 then
          s := s + IntToStr(Index);
      end;
      A[i][0] := s;
      A[i][1] := '';
      A[i][2] := FloatToStr(x);
      A[i][3] := '';
      A[i][4] := FloatToStr(y);
      A[i][5] := '';
      A[i][6] := FloatToStr(z);
    end;
  FindPrincipalAxes(Verbose);
//  SetLength(Atoms,n);
//  for i := 0 to n-1 do
//    with State.Nucleus(i) as TMolecularNucleus do begin
//      Atoms[i].xyz[1] := x;
//      Atoms[i].xyz[2] := y;
//      Atoms[i].xyz[3] := z;
//      Atoms[i].pxyz := Atoms[i].xyz;
//      Atoms[i].Index := i;
//    end;
end;

procedure TzMatrix.LoadFromState(const State: string);
begin
  if State = '' then
    SetFromState(Owner as TElectronicState, False)
  else
    raise NotYet;
end;

procedure TzMatrix.ShowXYZ(const s: string; TheMass: Double; xyz: Txyz);
begin
  WriteLn(Log, s, ' ':5-Length(s), TheMass:11:6, xyz[1]:Precision:Precision-5, xyz[2]:Precision:Precision-5, xyz[3]:Precision:Precision-5);
end;

procedure TzMatrix.ShowZmatrix(Sender: TObject);
begin
  {$IFNDEF CONSOLE}
  ConnectZmatrixForm(Self);
  {$ENDIF}
end;

{ TNotifyingActiveMNodeWithArray }

procedure TNotifyingActiveMNodeWithArray.ClearArrays;
begin
  if RootNode is TMixture then
    Precision := (RootNode as TMixture).Precision + 6
  else
    Precision := 10;
end;

procedure TNotifyingActiveMNodeWithArray.SetSetting(AIndex: Integer; const AValue: Integer);
begin
  inherited;
  if (AIndex = 1) and Assigned(EffectiveOwner) then
    (EffectiveOwner as TElectronicState).UpdateObjectChain;
end;

initialization
  TzMatrix.RegisterClass(TElectronicState);
end.
