unit BigSet;
{$I utils.inc}{ Copyright (C) 2014 C M Western; see UCompat.pas for license}
interface

{ Implementation of a simple arbitrary sized set }
type

  TBigSet = record
  private
    FSize: Integer;
    procedure SetSize(const Value: Integer);
  public
    s: array of Int64;
    function AndNonZero(const p: TBigSet):Boolean;
    procedure Clear;
    procedure CopySet(const p: TBigSet);
    procedure OrIn(const p: TBigSet);
    function OrInChanged(const p: TBigSet):Boolean;
    procedure RotateLeft;
    procedure RotateRight;
    procedure SetBitn(n: Cardinal);
    property Size:Integer read FSize write SetSize;
  end;

implementation

uses Math;

{ TBigSet }

procedure TBigSet.Clear;
var
  i: Integer;
begin
  for i := 0 to High(s) do
    s[i] := 0;
end;

procedure TBigSet.CopySet(const p: TBigSet);
var
  i: Integer;
begin
  Size := p.Size;
  for i := 0 to High(s) do
    s[i] := p.s[i];
end;

procedure TBigSet.OrIn(const p: TBigSet);
var
  i: Integer;
begin
  for i := 0 to High(s) do
    s[i] := s[i] or p.s[i];
end;

function TBigSet.OrInChanged(const p: TBigSet): Boolean;
var
  i: Integer;
begin
  Result := False;
  for i := 0 to High(s) do
    if (s[i] or p.s[i]) <> s[i] then begin
      s[i] := s[i] or p.s[i];
      Result := True;
    end;
end;

procedure TBigSet.SetBitn(n: Cardinal);
var
  i: Integer;
begin
  i := n div (SizeOf(Int64)*8);
  n := n mod (SizeOf(Int64)*8);
  s[i] := s[i] or (Int64(1) shl n);
end;

procedure TBigSet.SetSize(const Value: Integer);
begin
  FSize := Value;
  if Value = 0 then
    SetLength(s, 0)
  else
    SetLength(s, (Value-1) div (SizeOf(Int64)*8) + 1);
end;

{$IFDEF FPC}
{$ASMMODE INTEL}
{$ENDIF}
{$DEFINE PUREPASCAL}
{ If assembler routines can't be used, about 70% of the gain in speed can be attained
  by compiling the pascal part with R- and Q- }

{$IFDEF CPUX64}
{$UNDEF PUREPASCAL}
procedure TBigSet.RotateLeft; assembler;
asm
  {$IFDEF DELPHI}
  .NOFRAME
  {$ENDIF}
  // Self in rdi or rcx (Windows)
  mov rax,Self.S
  test rax,rax
  je @Lexit       // Null pointer
  mov rcx,[rax-8] // High or Length(Delphi)
  {$IFDEF FPC}
  test rcx,rcx
  {$ELSE}
  sub  rcx,1
  {$ENDIF}
  js @Lexit       // Zero Size
  sal qword ptr [rax],1
  jmp @LoopTail
@Loop:
  lea rax,[rax+8]
  rcl qword ptr [rax],1
@LoopTail:
  dec rcx
  jns @Loop
@Lexit:
end;

procedure TBigSet.RotateRight; assembler;
asm
  {$IFDEF DELPHI}
  .NOFRAME
  {$ENDIF}
  // Self in rdi or rcx (Windows)
  mov rax,Self.S
  test rax,rax
  je @Lexit       // Null pointer
  mov rcx,[rax-8] // High or Length(Delphi)
  {$IFDEF FPC}
  test rcx,rcx
  {$ELSE}
  sub  rcx,1
  {$ENDIF}
  js @Lexit       // Zero Size
  shr qword ptr [rax+rcx*8],1
  jmp @LoopTail
@Loop:
  rcr qword ptr [rax+rcx*8],1
@LoopTail:
  dec rcx
  jns @Loop
@Lexit:
end;

function TBigSet.AndNonZero(const p: TBigSet):Boolean; assembler;
asm
  {$IFDEF DELPHI}
  .NOFRAME
  {$ENDIF}
  // Self in rdi or rcx (Windows)
  // p in rsi/rdx (UNIX)
  xor  rax,rax      // Assume false
  mov  r8,Self.S
  test r8,r8
  je   @False       // Null pointer
  {$IFDEF MSWINDOWS}
  mov  rdx,[p.S]
  {$ENDIF}
  test rdx,rdx      // p
  je   @False       // Null pointer
  mov  rcx,[r8-8]   // High(S)
  cmp  rcx,[rdx-8]  // Compare with High(P.S)
  jge  @NotHigher
  mov  rcx,[rdx-8]
@NotHigher:
  {$IFDEF FPC}
  test rcx,rcx
  {$ELSE}
  sub  rcx,1
  {$ENDIF}
  js   @False       // Zero Size
@Loop:
  mov  r9,[r8+rcx*8]
  and  r9,[rdx+rcx*8]
  jne  @True
  dec  rcx
  jns  @Loop
  dec  rax
@True:
  inc  rax
@False:
end;
{$ENDIF}

{$IFDEF CPU386}
{$UNDEF PUREPASCAL}
procedure TBigSet.RotateLeft; assembler;
asm
  // Self in eax
  mov eax,Self.S
  test eax,eax
  je @Lexit       // Null pointer
  mov ecx,[eax-4] // High or Length(Delphi)
  {$IFDEF FPC}
  test ecx,ecx
  {$ELSE}
  sub  ecx,1
  {$ENDIF}
  js @Lexit       // Zero Size
  sal dword ptr [eax],1
  rcl dword ptr [eax+4],1
  jmp @LoopTail
  @Loop:
  lea eax,[eax+8]
  rcl dword ptr [eax],1
  rcl dword ptr [eax+4],1
  @LoopTail:
  dec ecx
  jns @Loop
  @Lexit:
end;

procedure TBigSet.RotateRight; assembler;
asm
  // Self in eax
  mov eax,Self.S
  test eax,eax
  je @Lexit       // Null pointer
  mov ecx,[eax-4] // High or Length(Delphi)
  {$IFDEF FPC}
  test ecx,ecx
  {$ELSE}
  sub  ecx,1
  {$ENDIF}
  js @Lexit       // Zero Size
  shr dword ptr [eax+ecx*8+4],1
  rcr dword ptr [eax+ecx*8],1
  jmp @LoopTail
  @Loop:
  rcr dword ptr [eax+ecx*8+4],1
  rcr dword ptr [eax+ecx*8],1
  @LoopTail:
  dec ecx
  jns @Loop
  @Lexit:
end;

function TBigSet.AndNonZero(const p: TBigSet):Boolean; assembler;
asm
  // Self in eax
  // p in edx[4]
  push edi
  push ebx
  mov  edi,Self.S
  xor  eax,eax      // Assume false
  test edi,edi
  je   @False       // Null pointer
  {$IFDEF DELPHI}
  mov  edx,p.S      // p
  {$ELSE}
  mov  edx,[edx+4]
  {$ENDIF}
  test edx,edx
  je   @False       // Null pointer
  mov  ecx,[edi-4]  // High(S)
  cmp  ecx,[edx-4]  // Compare with High(P.S)
  jge  @NotHigher
  mov  ecx,[edx-4]
@NotHigher:
  {$IFDEF FPC}
  test ecx,ecx
  {$ELSE}
  sub  ecx,1
  {$ENDIF}
  js   @False       // Zero Size
@Loop:
  mov  ebx,[edi+ecx*8]
  and  ebx,[edx+ecx*8]
  jne  @True
  mov  ebx,[edi+ecx*8+4]
  and  ebx,[edx+ecx*8+4]
  jne  @True
  dec  ecx
  jns  @Loop
  dec  eax
@True:
  inc  eax
@False:
  pop  ebx
  pop  edi
end;
{$ENDIF}

{$IFDEF PUREPASCAL}
function TBigSet.AndNonZero(const p: TBigSet): Boolean;
var
  i: Integer;
begin
  if Length(p.s) <> 0 then begin
    Result := True;
    for i := 0 to Min(High(s), High(P.s)) do
      if (s[i] and p.s[i]) <> 0 then
        Exit;
  end;
  Result := False;
end;

procedure TBigSet.RotateLeft;
var
  i, c, nc: Integer;
begin
  if Length(s) = 1 then
    s[0] := s[0] shl 1
  else begin
    c := 0;
    for i := 0 to High(s) do begin
      if s[i] < 0 then
        nc := 1
      else
        nc := 0;
      s[i] := (s[i] shl 1) or c;
      c := nc;
    end;
  end;
end;

procedure TBigSet.RotateRight;
var
  i, c, nc: Integer;
begin
  if Length(s) = 1 then
    s[0] := s[0] shr 1
  else begin
    c := 0;
    for i := High(s) downto 0 do begin
      nc := s[i] and 1;
      s[i] := (s[i] shr 1) or (Int64(c) shl (SizeOf(Int64)*8-1));
      c := nc;
    end;
  end;
end;
{$ENDIF}


end.
