unit ComplexMatrix;
{$I utils.inc}{ Copyright (C) 2009 C M Western; see UCompat.pas for license}
interface

uses Vector, MatrixVec, MathSup, ShellSort;

type

  { Implementation of Complex Matrix; imaginary part normally stored after real part
    and only allocated if necessary }
  TComplexMatrixMode = (cmReal,    { Only real part present }
                        cmComplex, { Imaginary part present after real part }
                        cmNComplex { Imaginary part alternates with real part (to use with e.g. fortran routines) }
                       );

  { TComplexMatrix }

  TComplexMatrix = class (TMatrix)
  private
    FMode: TComplexMatrixMode;
    function GetColCount: Longint;
    function GetIm(ARow, ACol: Longint): Double;
    function GetRe(ARow, ACol: Longint): Double;
    function GetSqrMod(ARow, ACol: Longint): Double;
    function GetValue(ARow, ACol: Longint): Complex;
    procedure SetColCount(AValue: Longint);
    procedure SetIm(ARow, ACol: Longint; const AValue: Double);
    procedure SetMode(const NewMode: TComplexMatrixMode);
    procedure SetRe(ARow, ACol: Longint; const AValue: Double);
    procedure SetValue(ARow, ACol: Longint; const AValue: Complex);
    procedure NotComplex;
  public
    constructor Create(Rows, Cols: Longint; AMode: TComplexMatrixMode); overload;
    constructor Create(AMatrix: TMatrix); overload;
    procedure Write(var F: Text; const Title: string; RowLabel: TShowName = nil; ColLabel: TShowName = nil; Width: Integer = 10; Fixed: Integer = -1); override;
    function ReIndex(ARow, ACol: Longint): Integer;
    function ImIndex(ARow, ACol: Longint): Integer;
    procedure SetReal(ARow, ACol: Longint; AValue: Double);
    procedure AssignAndNil(var Source: TVector); override;
    function Transpose: TComplexMatrix;
    function ConjugateTranspose: TComplexMatrix;
    procedure PostMultiply(M: TMatrix);
    function SameValues(M: TMatrix; Epsilon: Double = 0): Boolean;
    procedure AutoSetValue(ARow, ACol: Longint; const AValue: Complex);
    procedure ForceReal; override;
    procedure Sqr; override;

    property Mode: TComplexMatrixMode read FMode write SetMode;
    property ColCount: Longint read GetColCount write SetColCount; { Hides property in ancestor }
    property Re[ARow, ACol:Longint]:Double read GetRe write SetRe;
    property Im[ARow, ACol:Longint]:Double read GetIm write SetIm;
    property SqrMod[ARow, ACol:Longint]:Double read GetSqrMod;
    property Value[ARow, ACol:Longint]:Complex read GetValue write SetValue; default;
  end;

  { TEigenvectorSorter }
  { Produce a sorted list of indices to the eigenvector in order of decreasing magnitude }
  { nIndices selects how many are required; use -1 for all }

  TEigenvectorSorter = class(TShellSorter)
    Eigenvalues: TMatrix;
    Indices: array of Integer;
    State: Integer;
    IsComplex, IsInColumns: Boolean;
    function InOrder(I,J:Integer):Boolean; override;
    procedure Process(c, nIndices: Integer);
    constructor Create(TheEigenvalues: TMatrix; InColumns: Boolean = True);
  end;

  function IsComplexMatrix(Matrix: TMatrix):Boolean;

implementation

uses SysUtils, Math, UCompat, Lapack;

function IsComplexMatrix(Matrix: TMatrix):Boolean;
begin
  Result := (Matrix is TComplexMatrix) and (TComplexMatrix(Matrix).Mode <> cmReal);
end;

{ TComplexMatrix }

procedure TComplexMatrix.SetMode(const NewMode: TComplexMatrixMode);
var
  s, d, b: PDouble;
  i, n: Integer;
  c: SizeInt;
  buffer: array of Double;
begin
  if FMode = NewMode then exit;
  c := ColCount; { Copy to avoid confusion while changing modes }
  if (c <> 0) and (RowCount <> 0) then
    case Mode of
      cmReal:
        begin
          inherited ColCount := 2*c; { Allocate space }
          if NewMode = cmComplex then begin
            { Clear Complex Area }
            d := inherited Item[RowCount,c];
            Inc(d);
            FillChar(d^,SizeOf(Double)*RowCount*c,0);
          end else begin
            s := inherited Item[RowCount,c];
            d := inherited Item[RowCount,2*c];
            Dec(d);
            for i := 1 to RowCount*c-1 do begin
              d^ := s^;
              Dec(d,2);
              Dec(s);
            end;
            { Clear complex part to zero }
            s := inherited Item[1,1];
            Inc(s);
            for i := 1 to RowCount*c do begin
              s^ := 0;
              Inc(s,2);
            end;
          end;
        end;
      cmComplex:
        if NewMode = cmReal then
          inherited ColCount := c
        else begin
          { Rearrange number layout }
          { May be a way requiring less memory }
          n := RowCount*c;
          { Save even numbered reals }
          SetLength(buffer,n div 2);
          s := inherited Item[1,1];
          Inc(s);
          for i := 0 to High(buffer) do begin
            Buffer[i] := s^;
            Inc(s,2);
          end;
          { Copy imaginary parts in place }
          d := inherited Item[1,1];
          Inc(d);
          s := inherited Item[RowCount,c];
          Inc(s);
          for i := 1 to RowCount*c-1 do begin
            d^ := s^;
            Inc(d,2);
            Inc(s);
          end;
          { Copy odd numbered reals in place }
          s := inherited Item[RowCount,c];
          d := inherited Item[RowCount,2*c];
          Dec(d);
          if not Odd(n) then begin
            { Last element not odd numbered }
            Dec(s);
            Dec(d,2);
          end;
          while PAnsiChar(d) > PAnsiChar(s) do begin
            d^ := s^;
            Dec(d,2);
            Dec(s);
          end;
          { Copy even numbered reals in place }
          d := inherited Item[1,1];
          Inc(d,2);
          for i := 0 to High(buffer) do begin
            d^ := buffer[i];
            Inc(d,4);
          end;
        end;
      cmNComplex:
        if NewMode = cmReal then begin
          s := inherited Item[1,1];
          d := s;
          for i := 1 to RowCount*c do begin
            d^ := s^;
            Inc(s,2);
            Inc(d);
          end;
          inherited ColCount := c;
        end else begin
          { Rearrange number layout }
          { May be a way requiring less memory }
          n := RowCount*c;
          { Save reals in second half of matrix }
          SetLength(buffer, n div 2);
          b := inherited Item[1,c+1];
          if Odd(n) then
            Inc(b);
          s := b;
          for i := 0 to High(buffer) do begin
            buffer[i] := s^;
            Inc(s,2);
          end;
          { Copy imaginary parts in place }
          d := inherited Item[RowCount,2*c];
          s := d;
          for i := 1 to n-1 do begin
            Dec(d);
            Dec(s,2);
            d^ := s^;
          end;
          { Copy lower reals in place }
          d := inherited Item[1,1];
          s := d;
          while PAnsiChar(s) < PAnsiChar(b) do begin
            d^ := s^;
            Inc(s,2);
            Inc(d)
          end;
          { Copy upper reals in place }
          for i := 0 to High(buffer) do begin
            d^ := buffer[i];
            Inc(d);
          end;
        end;
    end;
  FMode := NewMode;
end;

procedure TComplexMatrix.Write(var F: Text; const Title: string; RowLabel: TShowName; ColLabel: TShowName; Width: Integer; Fixed: Integer);
var
  r, c, lt: Integer;
  s: string;
begin
  if Mode = cmReal then
    inherited Write(F,Title,RowLabel,ColLabel,Width,Fixed)
  else begin
    if not Assigned(ColLabel) then
      ColLabel := {$IFDEF FPC}@{$ENDIF}DefaultLabel;
    if not Assigned(RowLabel) then
      RowLabel := {$IFDEF FPC}@{$ENDIF}DefaultLabel;
    lt := Length(Title);
    for r := 1 to RowCount do
      lt := Max(lt, Length(RowLabel(r)));
    System.Write(F, Title:lt);
    if Fixed < 0 then
      System.Write(F, '  ')
    else
      System.Write(F, ' ':Width-Fixed-1);
    for c := 1 to ColCount do begin
      s := ColLabel(c);
      System.Write(F, s, ' ':2*Width+1-Length(s));
    end;
    WriteLn(F);
    for r := 1 to RowCount do begin
      System.Write(F, RowLabel(r):lt, ' ');
      for c := 1 to ColCount do
        System.Write(F, DTosMA(Re[r,c],Width,Fixed), ',', DTosMA(Im[r,c],Width,Fixed));
      WriteLn(F)
    end;
  end;
end;

function TComplexMatrix.GetColCount: Longint;
begin
  Result := inherited ColCount;
  if Mode <> cmReal then
    Result := Result div 2;
end;

procedure TComplexMatrix.SetColCount(AValue: Longint);
begin
  if Mode <> cmReal then
    AValue := 2*AValue;
  inherited ColCount := AValue
end;

function TComplexMatrix.ReIndex(ARow, ACol: Longint): Integer;
begin
  {$IFOPT R+}
  if (ARow <= 0) or (ACol <= 0) or (ARow > RowCount) or (ACol > ColCount) then
    raise Exception.CreateFmt('TComplexMatrix[%d,%d]: Bound check failure on Re (%d)', [ARow, ACol, RowCount]);
  {$ENDIF}
  if Mode = cmNComplex then
    Result := SizeInt(ACol-1)*RowCount*2+2*ARow-1
  else
    Result := SizeInt(ACol-1)*RowCount+ARow;
end;

function TComplexMatrix.GetRe(ARow, ACol: Longint): Double;
begin
  Result := TVector(Self).Value[ReIndex(ARow,ACol)];
end;

function TComplexMatrix.GetSqrMod(ARow, ACol: Integer): Double;
begin
  if Mode = cmReal then
    Result := System.Sqr(GetRe(ARow,ACol))
  else
    Result := MathSup.SqrMod(GetValue(ARow,ACol));
end;

procedure TComplexMatrix.SetRe(ARow, ACol: Longint; const AValue: Double);
begin
  TVector(Self).Value[ReIndex(ARow,ACol)] := AValue;
end;

function TComplexMatrix.ImIndex(ARow, ACol: Longint): Integer;
begin
  {$IFOPT R+}
  if (ARow <= 0) or (ACol <= 0) or (ARow > RowCount) or (ACol > ColCount) then
    raise Exception.CreateFmt('TComplexMatrix[%d,%d]: Bound check failure on Im (%d)', [ARow, ACol, RowCount]);
  {$ENDIF}
  case Mode of
    cmReal:     NotComplex;
    cmComplex:  Result := SizeInt(ACol-1)*RowCount+ARow + Size div 2;
    cmNComplex: Result := SizeInt(ACol-1)*RowCount*2+2*ARow;
  end;
end;

procedure TComplexMatrix.SetReal(ARow, ACol: Longint; AValue: Double);
var
  i: Integer;
begin
  case Mode of
    cmReal:
      inherited Value[ARow,ACol] := AValue;
    cmComplex:
      begin;
        i := ReIndex(ARow,ACol);
        TVector(Self).Value[i] := AValue;
        TVector(Self).Value[i+Size div 2] := 0;
      end;
    cmNComplex:
      begin;
        i := ReIndex(ARow,ACol);
        TVector(Self).Value[i] := AValue;
        TVector(Self).Value[i+1] := 0;
      end;
  end;
end;

procedure TComplexMatrix.AssignAndNil(var Source: TVector);
begin
  if Source is TComplexMatrix then
    FMode := TComplexMatrix(Source).Mode
  else
    FMode := cmReal;
  inherited AssignAndNil(Source);
end;

function TComplexMatrix.Transpose: TComplexMatrix;
var
  i, j: Integer;
begin
  Result := TComplexMatrix.Create(ColCount,RowCount,Mode);
  for i := 1 to RowCount do
    for j := 1 to ColCount do
      Result.Value[j,i] := Value[i,j];
end;

procedure TComplexMatrix.AutoSetValue(ARow, ACol: Longint; const AValue: Complex);
begin
  // Mode := cmComplex;
  { Automatically switch to full complex represenataion if required }
  if Mode = cmReal then
    if AValue.im <> 0 then begin
      Mode := cmComplex;
      Value[ARow,ACol] := AValue;
    end else
      Re[ARow,ACol] := AValue.re
  else
    Value[ARow,ACol] := AValue;
end;

procedure TComplexMatrix.ForceReal;
begin
  Mode := cmReal;
end;

function TComplexMatrix.ConjugateTranspose: TComplexMatrix;
var
  i, j: Integer;
  z: Complex;
begin
  Result := TComplexMatrix.Create(ColCount,RowCount,Mode);
  for i := 1 to RowCount do
    for j := 1 to ColCount do begin
      z := Value[i,j];
      z.im := -z.im;
      Result.Value[j,i] := z;
    end;
end;

function TComplexMatrix.SameValues(M: TMatrix; Epsilon: Double): Boolean;
var
  OtherMode: TComplexMatrixMode;
  z, z1: Complex;
  i, j: Integer;
begin
  Result := False;
  z1 := z0;
  if M is TComplexMatrix then
    OtherMode := TComplexMatrix(M).Mode
  else
    OtherMode := cmReal;
  if Mode = OtherMode then
    Result := inherited SameValues(M,Epsilon) { Same memory layout }
  else begin
    for i := 1 to ColCount do
      for j := 1 to RowCount do begin
        z := Value[j,i];
        if OtherMode = cmReal then
          z1.re := M.Value[j,i]
        else
          z1 := TComplexMatrix(M).Value[j,i];
        if not SameValue(z,z1,Epsilon) then
          Exit;
      end;
    Result := True;
  end;
end;

procedure TComplexMatrix.PostMultiply(M: TMatrix);
var
  OtherMode: TComplexMatrixMode;
  W: array of Complex;
  mm, k, N: Integer;
  alpha, beta: Complex;
  A: Pcomplex;
  No: AnsiChar;
  OtherMatrix: TComplexMatrix;
begin
  if M is TComplexMatrix then
    OtherMode := TComplexMatrix(M).Mode
  else
    OtherMode := cmReal;
  if (Mode = cmReal) and (OtherMode = cmReal) then
    inherited PostMultiply(M)
  else begin
    if Size = 0 then
      Exit;
    OtherMatrix := nil;
    try
      if (OtherMode = cmReal) or (OtherMode = cmComplex) then begin
        OtherMatrix := TComplexMatrix.Create(M);
        OtherMatrix.Mode := cmNComplex;
        M := OtherMatrix;
        N := OtherMatrix.ColCount;
      end else
        N  := (M as TComplexMatrix).ColCount;
      Mode := cmNComplex;
      alpha := z0;
      beta := z0;
      alpha.re := 1;
      No := 'N';
      mm := RowCount;
      k := ColCount;
      if ColCount = N then begin
        { No change in size }
        Size := Size * 2; { Allocate workspace }
        A := Pcomplex(PAnsiChar(FirstItem) + (Size div 2)*SizeOf(Double));
        System.Move(FirstItem^, A^, (Size div 2)*SizeOf(Double));
        zgemm_(No, No, mm, N, N, alpha, A^, mm, PComplex(M.FirstItem)^, N, beta, PComplex(FirstItem)^, mm, 1, 1);
        Size := Size div 2;
      end else begin
        SetLength(W,Size);
        System.Move(FirstItem^, w[0], 2*Size*SizeOf(Double));
        ColCount := N;
        zgemm_(No, No, mm, N, k, alpha, W[0], mm, PComplex(M.FirstItem)^, k, beta, PComplex(FirstItem)^, mm, 1, 1);
      end;
    finally
      OtherMatrix.Free;
    end;
  end;
end;

function TComplexMatrix.GetIm(ARow, ACol: Longint): Double;
begin
  Result := TVector(Self).Value[ImIndex(ARow,ACol)];
end;

procedure TComplexMatrix.SetIm(ARow, ACol: Longint; const AValue: Double);
begin
  TVector(Self).Value[ImIndex(ARow,ACol)] := AValue;
end;

function TComplexMatrix.GetValue(ARow, ACol: Longint): Complex;
var
  i: Integer;
begin
  case Mode of
    cmReal:
      begin;
        Result.re := inherited Value[ARow,ACol];
        Result.im := 0;
      end;
    cmComplex:
      begin;
        i := ReIndex(ARow,ACol);
        Result.re := TVector(Self).Value[i];
        Result.im := TVector(Self).Value[i+Size div 2];
      end;
    cmNComplex:
        Result := Pcomplex(TVector(Self).Item[ReIndex(ARow,ACol)])^;
  end;
end;

procedure TComplexMatrix.SetValue(ARow, ACol: Longint; const AValue: Complex);
var
  i: Integer;
begin
  case Mode of
    cmReal:
      begin;
        inherited Value[ARow,ACol] := AValue.re;
        if AValue.im <> 0 then
          NotComplex;
      end;
    cmComplex:
      begin;
        i := ReIndex(ARow,ACol);
        TVector(Self).Value[i] := AValue.re;
        TVector(Self).Value[i+Size div 2] := AValue.im;
      end;
    cmNComplex:
      Pcomplex(TVector(Self).Item[ReIndex(ARow,ACol)])^ := AValue;
  end;
end;

procedure TComplexMatrix.Sqr;
var
  r, c: PDouble;
  i: SizeInt;
begin
  case Mode of
    cmReal: inherited;
    cmComplex:
      begin
        r := FirstItem;
        c := inherited Item[RowCount,ColCount];
        Inc(c);
        for i := 1 to SizeInt(RowCount)*ColCount do begin
          r^ := System.Sqr(r^) + System.Sqr(c^);
          Inc(r);
          Inc(c)
        end;
      end;
    cmNComplex:
      begin
        r := FirstItem;
        c := r;
        Inc(c);
        for i := 1 to SizeInt(RowCount)*ColCount do begin
          r^ := System.Sqr(r^) + System.Sqr(c^);
          Inc(r,2);
          Inc(c,2)
        end;
      end;
  end;
  Mode := cmReal;
end;

procedure TComplexMatrix.NotComplex;
begin
  raise Exception.Create('TComplexMatrix: Not Complex');
end;

constructor TComplexMatrix.Create(Rows, Cols: Longint; AMode: TComplexMatrixMode);
begin
  FMode := AMode;
  if Mode <> cmReal then
    Cols := Cols*2;
  inherited Create(Rows,Cols);
end;

constructor TComplexMatrix.Create(AMatrix: TMatrix);
begin
  if AMatrix is TComplexMatrix then
    FMode := TComplexMatrix(AMatrix).Mode;
  inherited Create(AMatrix);
end;

{ TEigenvectorSorter }

function TEigenvectorSorter.InOrder(I, J: Integer): Boolean;
begin
  if IsInColumns then
    if IsComplex then
      Result := TComplexMatrix(Eigenvalues).SqrMod[Indices[I],State] >= TComplexMatrix(Eigenvalues).SqrMod[Indices[J],State]
    else
      Result := Abs(Eigenvalues[Indices[I],State]) >= Abs(Eigenvalues[Indices[J],State])
  else
    if IsComplex then
      Result := TComplexMatrix(Eigenvalues).SqrMod[State, Indices[I]] >= TComplexMatrix(Eigenvalues).SqrMod[State, Indices[J]]
    else
      Result := Abs(Eigenvalues[State, Indices[I]]) >= Abs(Eigenvalues[State, Indices[J]]);
  if not Result then
    SwapValues(Indices[I], Indices[J], SizeOf(Integer));
end;

procedure TEigenvectorSorter.Process(c, nIndices: Integer);
var
  i, nTotal: Integer;
begin
  State := c;
  IsComplex := Eigenvalues is TComplexMatrix;
  if IsInColumns then
    nTotal := Eigenvalues.RowCount
  else
    nTotal := Eigenvalues.ColCount;
  if nIndices < 0 then
    nIndices := nTotal;
  if nIndices < nTotal div 4 then begin
    { Use limited scan algorithm }
    SetLength(Indices, nIndices+1); { One extra for scanning }
    for i := 0 to nIndices-1 do
      Indices[i] := i+1;
    Sort(nIndices); { Sort first nIndices }
    for i := nIndices to nTotal - 1 do begin
      Indices[nIndices] := i+1;
      if not InOrder(nIndices-1, nIndices) then begin
        Indices[nIndices-1] := i+1; { Replace smallest }
        Sort(nIndices); { Sort first nIndices }
      end;
    end;
  end else begin
    SetLength(Indices, nTotal);
    for i := 0 to High(Indices) do
      Indices[i] := i+1;
    Sort(Length(Indices));
  end;
  SetLength(Indices, nIndices);
end;

constructor TEigenvectorSorter.Create(TheEigenvalues: TMatrix; InColumns: Boolean = True);
begin
  Eigenvalues := TheEigenvalues;
  IsInColumns := InColumns;
end;

end.

