unit Contour;
{ Set of routines for generating contours from 2D data;
  based on routines from Gnuplot }

{$I utils.inc}{ Copyright (C) 2005 C M Western; see UCompat.pas for license}
interface

uses Vector, Interp, MathSup;

type

  TContourType = (ctLinear, ctCubic, ctBspline, ctOther);

  TContourOptions = record
    CType: TContourType;
    nInterp: Integer; { Number of additional points to add }
    Order: Integer; { Order of interpolation for BSplines }
  end;

  TContour = record
    Level: Double;
    Points: TVector2;
    Index: Integer;
    Closed: Boolean;
  end;

  TContours = array of TContour;

function Contours(CNValues: TVector; { Contour Values }
                  XValues, YValues: TDoubleArray; { Grid positions }
                  ZValues: TDoubleArray; { Function Values on grid in one dimensional array;
                                           x index varying fastest }
                  const ContourOptions: TContourOptions
                  ):TContours;

procedure SortContours(var C: TContours);

function Different(const a, b: TContourOptions): Boolean;

implementation

uses SysUtils;

function Different(const a, b: TContourOptions): Boolean;
begin
  Result := (a.CType <> b.CType) or (a.nInterp <> b.nInterp) or (a.Order <> b.Order);
end;

{
  Subdivide the region into triangles:

      V---V
      |\  |
      | \ |
      |  \|
      V---V

  There are:
    nX * nY vertices
    nY * (nX - 1) Horizontal edges
    nX * (nY - 1) Vertical edges
    (nX - 1) * (nY -1) Diagonals
  Ignoring the lower and right edges, we have three edges and two triangles that
  we can associate with each vertex:
        1
      V---
      |\ 1|
     0| \ | (The diagonal is edge 2)
      |0 \|
       ---
  so edges can be indexed as [Y][X][0..2] expressed as a single index e + (Y * nX + X)*4
  and triangles by           [Y][X][0..1] expressed as a single index t + (Y * nX + X)*2
  Given a triangle number, the edges are:
    for even triangles: [y][x][0], [y][x][2] and [y+1][x][1]
    for odd  triangles: [y][x][2], [y][x][1] and [x+1][x][0]
  Given an edge number, the triangles are:
    0: [y][x][0], [y][x-1][1]
    1: [y][x][1], [y-1][x][0]
    2: [y][x][0], [y][x][1]
  (index 3 is used to flag edges that )

  At each contour level, we need to consider each edge that has a value at each
  end that straddle the contour value. A contour is generated by following connected
  edges (using the triangles) ticking off each edge as we go. The edges are swept until all
  edges have been ticked off.



}

function Contours(CNValues: TVector; { Contour Values }
                  XValues, YValues: TDoubleArray; { Grid positions }
                  ZValues: TDoubleArray; { Function Values on grid in one dimensional array;
                                           x index varying fastest }
                  const ContourOptions: TContourOptions
                  ):TContours;
var
  nX, nY, i, cn: Integer;
  Active: array of Boolean; { non zero if edge current contour value }
  nActive: Integer; { Number elements of Active set }
  CurrentLevel: Double;
  SmallX, SmallY: Double; { for AlmostEqual }
  function AlmostEqual(x1, y1, x2, y2: Double): Boolean;
  begin
    Result := (Abs(x1-x2) < SmallX) and (Abs(y1-y2) < SmallY);
  end;
  procedure SetEdges(Level:Double);
  var
    ThisX, NextX, { Top two } ThisX1, NextX1 { Next row }: Boolean;
    iX, iY, i, ib: Integer;
  begin
    FillChar(Active[0], Length(Active), 0);
    nActive := 0;
    i := 0;  { Index Z value array }
    ib := 0; { Index Active array }
    for iX := 0 to nX-2 do
      Byte(Active[iX*4+3]) := 2; { Flag top edge }
    for iY := 0 to nY-2 do begin
      Inc(Byte(Active[ib+3]), 1); { Flag left edge }
      ThisX := ZValues[i] >= Level;
      ThisX1 := ZValues[i+nX] >= Level;
      for iX := 0 to nX-2 do begin
        NextX := ZValues[i+1] >= Level;
        NextX1 := ZValues[i+1+nX] >= Level;
        { Edge 0 - vertical }
        if ThisX xor ThisX1 then begin
          Active[ib] := True;
          Inc(nActive);
        end;
        { Edge 1 - horizontal }
        if ThisX xor NextX then begin
          Active[ib+1] := True;
          Inc(nActive);
        end;
        { Edge 2 - diagonal }
        if ThisX xor NextX1 then begin
          Active[ib+2] := True;
          Inc(nActive);
        end;
        Inc(ib, 4);
        Inc(i);
        ThisX := NextX;
        ThisX1 := NextX1;
      end;
      { Only vertical edge associated with last point on row }
      if ThisX xor ThisX1 then begin
        Active[ib] := True;
        Inc(nActive);
      end;
      Inc(Byte(Active[ib+3]), 4); { Flag right edge }
      Inc(ib, 4);
      Inc(i);
    end;
    { Only horizontal edge associated with bottom row }
    ThisX := ZValues[i] >= Level;
    for iX := 0 to nX-2 do begin
      NextX := ZValues[i+1] >= Level;
      { Edge 1 - horizontal }
      if ThisX xor NextX then begin
        Active[ib+1] := True;
        Inc(nActive);
      end;
      ThisX := NextX;
      Inc(Byte(Active[ib+3]), 8); { Flag bottom edge }
      Inc(ib, 4);
      Inc(i);
    end;
  end;
  procedure TrackContour(Edge: Integer; AtEdge: Boolean);
  var
    Tri, Tri0, Tri1, LastEdge, StartEdge: Integer;
    nPoints, e: Integer;
    ThisContour: TContour;
    ResetFirst: Boolean;
    procedure AddEdge(Edge: Integer);
    var
      x, y, z0, z1, t: Double;
      iX, iY, iZ: Integer;
    begin
      Active[Edge] := False;
      Dec(nActive);
      if ((Edge and 3) = 2) and (nPoints > 0) then
        Exit; { Leave diagonal edges out of contour }
      { Indices of start of Edge }
      iZ := Edge div 4;
      iX := iZ mod nX;
      iY := iZ div nX;
      x := XValues[iX];
      y := YValues[iY];
      { Find Z values at each end of edge }
      z0 := ZValues[iZ];
      case Edge and 3 of
        0: z1 := ZValues[iZ+nX];
        1: z1 := ZValues[iZ+1];
        2: z1 := ZValues[iZ+nX+1];
      end;
      t := (CurrentLevel - z0) / (z1-z0); { Should never have z1 = z0 as this would not be an active edge }
      if t < 0 then { Sanity check }
        t := 0
      else if t > 1 then
        t := 1;
      case Edge and 3 of
        0: y := y + t * (YValues[iY+1]-y);
        1: x := x + t * (XValues[iX+1]-x);
        2: begin
             x := x + t * (XValues[iX+1]-x);
             y := y + t * (YValues[iY+1]-y);
           end;
      end;
      if (nPoints > 0) and AlmostEqual(x, y, ThisContour.Points[nPoints-1][0], ThisContour.Points[nPoints-1][1]) then
        Exit;
      { TODO: Optimize allocation }
      { TODO: fuzzy equal }
      SetLength(ThisContour.Points, nPoints+1);
      ThisContour.Points[nPoints][0] := x;
      ThisContour.Points[nPoints][1] := y;
      Inc(nPoints);
    end;
  begin
    Tri := -1;
    ThisContour.Level := CurrentLevel;
    ThisContour.Index := cn;
    SetLength(ThisContour.Points, 0);
    nPoints := 0;
    StartEdge := Edge;
    ResetFirst := not AtEdge;
    AddEdge(Edge);
    repeat
      { Find the two triangles
        Given an edge number, the triangles are:
          0: [y][x][0], [y][x-1][1]
          1: [y][x][1], [y-1][x][0]
          2: [y][x][0], [y][x][1]
      }
      Tri0 := Edge and not 3; { [y][x][0] }
      if Active[Tri0+3] then
        { Need to look for edge problems }
        case Edge and 3 of
          0: if (Byte(Active[Tri0+3]) and 1) <> 0 then
               Tri1 := -1 { At left edge }
             else if (Byte(Active[Tri0+3]) and 4) <> 0 then begin
               Tri1 := Tri0 - 3; { At right edge }
               Tri0 := -1;
             end else
               Tri1 := Tri0 - 3;
          1: if (Byte(Active[Tri0+3]) and 2) <> 0 then begin
               Tri1 := -1; { At top edge }
               Tri0 := Edge;
             end else if (Byte(Active[Tri0+3]) and 8) <> 0 then begin
               Tri1 := Tri0 - nX*4; { at bottom edge }
               Tri0 := -1;
             end else begin
               Tri1 := Tri0 - nX*4;
               Inc(Tri0);
             end;
          2: if (Byte(Active[Tri0+3]) and 4) <> 0 then
               Tri1 := -1
             else
               Tri1 := Tri0 + 1;
        else
          Assert(True, 'Help0');
        end
      else
        { Not at edge, so easy }
        case Edge and 3 of
          0: Tri1 := Tri0 - 4 + 1;
          1: begin Tri1 := Tri0 - nX*4; Inc(Tri0); end;
          2: Tri1 := Tri0 + 1;
        else
          Assert(True, 'Help1');
        end;
      if Tri0 = Tri then
        Tri := Tri1
      else
        Tri := Tri0;
      LastEdge := Edge;
      Edge := -1;
      if Tri < 0 then begin
        Break;
      end else begin
        { Find the other active edge in triangle
          - curent edge has already been marked inactive
          Given a triangle number, the edges are:
           for even triangles: [y][x][0], [y][x][2] and [y+1][x][1]
           for odd  triangles: [y][x][2], [y][x][1] and [x+1][x][0] }
        if (Tri and 1) = 0 then begin
          if Active[Tri] then { [y][x][0] }
            Edge := Tri
          else if Active[Tri+2] then { [y][x][2] }
            Edge := Tri+2
          else if Active[Tri+nX*4+1] then { [y+1][x][1] }
            Edge := Tri+nX*4+1
        end else begin
          if Active[Tri] then { [y][x][1] }
            Edge := Tri
          else if Active[Tri+1] then { [y][x][2] }
            Edge := Tri+1
          else if Active[Tri+3] then { [y][x+1][0] }
            Edge := Tri+3
        end;
        if Edge < 0 then begin
          { Getting back to start is a legitimate point to stop }
          if LastEdge = StartEdge then
            Break;
          { Hitting a boundary is also a legitimate point to stop }
          e := Byte(Active[LastEdge or 3]); { Edge value }
          case LastEdge and 3 of
            0: if (e = 1) or (e = 4) then Break;
            1: if (e = 2) or (e = 8) then Break;
          end;
          raise Exception.Create('Unexpected end to contour');
        end else begin
          AddEdge(Edge);
          if ResetFirst then begin
            { Allow first point to figure twice to detect closed curves }
            Active[StartEdge] := True;
            Inc(nActive);
            ResetFirst := False;
          end;
        end;
      end;
    until Edge < 0;
    if nPoints > 1 then begin
      SetLength(Result, Length(Result)+1);
      ThisContour.Closed := LastEdge = StartEdge;
                            // AlmostEqual(ThisContour.Points[0][0], ThisContour.Points[0][1],
                            //            ThisContour.Points[nPoints-1][0], ThisContour.Points[nPoints-1][1]);
      Result[Length(Result)-1] := ThisContour;
      case ContourOptions.CType of
        { ctLinear: } { No action }
        ctCubic:   Result[Length(Result)-1].Points := CubicSpline(ThisContour.Points, ThisContour.Closed, XValues[0]-XValues[nx-1],
                                                                  YValues[0]-YValues[ny-1], ContourOptions.nInterp);
        ctBSpline: Result[Length(Result)-1].Points := BSpline(ThisContour.Points, ThisContour.Closed, ContourOptions.nInterp, ContourOptions.Order);
      end;
    end;
  end;
  procedure DoEdge(First, Number, Stride: Integer);
  var
    i: Integer;
  begin
    if nActive = 0 then
      Exit;
    for i := 1 to Number do begin
      if Active[First] then
        TrackContour(First, True);
      Inc(First, Stride);
    end;
  end;
begin
  nX := Length(XValues);
  nY := Length(YValues);
  SetLength(Active, nX*nY*4);
  SmallX := Abs(XValues[0]-XValues[nx-1])*1e-5;
  SmallY := Abs(YValues[0]-YValues[ny-1])*1e-5;

  for cn := CNValues.Offset to CNValues.LastIndex do begin
    CurrentLevel := CNValues[cn];
    SetEdges(CurrentLevel);
    { Look for contours on the edges first }
    DoEdge(0, nY-1, nX*4); { Vertical left edge }
    DoEdge(1, nX-1, 4);    { Horizontal top edge }
    DoEdge((nX-1)*4, nY-1, nX*4);   { Vertical right edge }
    DoEdge(1+(nY-1)*nX*4, nX-1, 4); { Horizontal bottom edge }
    { And then any that are left }
    i := 0;
    while (i < Length(Active)) and (nActive > 0) do begin
      if Active[i] then { Track contours starting from any vertical }
        TrackContour(i, False);
      if Active[i+1] then { Track contours starting from any horizontal }
        TrackContour(i+1, False);
      { Contours starting from diagonals not needed,
        as any remaining after horizontals and verticals had gone
        would be single point contours
        if Active[i+2] then
          TrackContour(i+2, False);
      }
      Inc(i, 4);
    end;
    if nActive > 0 then
      raise Exception.Create('Warning: Active edges left');
  end;
end;

{ Point in polygon algorithm from
  http://www.ecse.rpi.edu/Homepages/wrf/research/geom/pnpoly.html
  by Wm. Randolph Franklin
}
function PointInPolygon(const V: TVector2; x, y: Double):Boolean;
var
  i, j, npol: Integer;
begin
  npol := Length(V);
  Result := False;
  j := npol - 1; { Reverse tracking }
  for i := 0 to npol - 1 do
    if ( ( ( (V[i][1]<=y) and (y<V[j][1]) )
            or ( (V[j][1]<=y) and (y<V[i][1]) )
         )
         and (x < (V[j][0] - V[i][0]) * (y - V[i][1]) / (V[j][1] - V[i][1]) + V[i][0])) then
      Result := not Result;
end;

function PolygonArea(const V: TVector2):Double;
var
  n, i:Integer;
begin
  n := length(V);
  Result := V[0][0]*(V[1][1]-V[n-1][1]);
  if n > 1 then
    Result := Result + V[n-1][0]*(V[0][1]-V[n-2][1]);
  for i := 1 to n - 2 do
    Result := Result + V[i][0]*(V[i+1][1]-V[i-1][1]);
  Result := Result / 2;
end;

procedure SortContours(var C: TContours);
var
  N, I, J, Offset: Integer;
  IsInOrder: Boolean;
  T: TContour;
begin
  { ShellSort }
  N := Length(C);
  Offset := N;
  while Offset > 1 do begin
    Offset := Offset shr 1;
    repeat
      IsInOrder := True;
      for J := 0 to N - Offset - 1 do begin
        I := J + Offset;
        { if PointInPolygon(C[I].Points, C[J].Points[0][0], C[J].Points[0][1]) then begin  Fails }
        if Abs(PolygonArea(C[I].Points)) > Abs(PolygonArea(C[J].Points)) then begin
          IsInOrder := False;
          T := C[I];
          C[I] := C[J];
          C[J] := T;
        end;
      end;
    until IsInOrder;
  end;
end;

end.
