unit EigBlock;
{$I utils.inc}{ Copyright (C) 2014 C M Western; see UCompat.pas for license}
interface

uses Vector, MatrixVec;

procedure Eigen(Eigenvals: TColumnVector; Matrix: TMatrix; EvReqd: Boolean; Small: Double; Verbose: Boolean; BasisLabel: TShowName = nil; Width: Integer = 10);
{ Return eigenvalues and optionally eigenvectors of Matrix;
  Matrix is blocked taking values below small*Largest as zero to avoid unwanted mixing between states.
  Eigenvalue matrix (E) overwrites original matrix (M);
  Each column of E contains an eigenvector and Transpose(E) * M * E is a diagonal matrix
}

implementation

uses SysUtils, BigSet, EigLA, Worker, MathSup;

type

  TEigenHelper = class
  public
    Small: Double;
    Matrix: TMatrix;
    Size, nBlocks, CurBlockBase: Integer;
    Connections: array of TBigSet;
    Matrices: array of TMatrix;
    Perm: array of Integer;
    Evals: TColumnVector;
    FOnShowLabel: TShowName;
    Width: Integer;
    procedure MakeConnections;
    function CountBlocks:Integer;
    procedure FactorMatrix;
    procedure EigBlocks(EvReqd, Verbose: Boolean; TheEigenvals: TColumnVector);
    function ShowLabel(Index: Integer): string;
    function ShowEv(Index: Integer): string;
    destructor Destroy; override;
  end;

procedure Eigen(Eigenvals: TColumnVector; Matrix: TMatrix; EvReqd: Boolean; Small: Double; Verbose: Boolean; BasisLabel: TShowName = nil; Width: Integer = 10);
var
  EH: TEigenHelper;
begin
  EH := TEigenHelper.Create;
  EH.Width := Width;
  if Small < 0 then
    EH.Small := Abs(Small)
  else
    EH.Small := Matrix.MaxAbs*Small;
  EH.Matrix := Matrix;
  EH.FOnShowLabel := BasisLabel;
  try
    EH.MakeConnections;
    if EH.CountBlocks > 1 then begin
      EH.FactorMatrix;
      EH.EigBlocks(EvReqd, Verbose, Eigenvals);
    end else begin
      EigLA.Eigen(Eigenvals, Matrix, EvReqd);
      if Verbose then begin
        WriteLn(Log, 'Eigenvalues');
        Matrix.Write(Log, 'Ev', BasisLabel, {$IFDEF FPC}@{$ENDIF}EH.ShowEv, Width);
        WriteLn(Log);
      end;
    end;
  finally
    EH.Free;
  end;
end;

{ TEigenHelper }

function TEigenHelper.CountBlocks: Integer;
{ Count how many blocks to factor into }
var
  b, bb, bNext: Integer;
  MaskChanged: Boolean;
  CurMask: TBigSet;
begin
  { Work backwards from last row, as this will have most connections set }
  b := Size;
  Result := 0;
  repeat
    CurMask := Connections[b];
    Inc(Result);
    repeat
      bNext := 0;
      MaskChanged := False;
      for bb := b-1 downto 1 do begin
        if CurMask.AndNonZero(Connections[bb]) then begin
          MaskChanged := CurMask.OrInChanged(Connections[bb]) or MaskChanged;
          Connections[bb].Size := 0;
        end else if (bNext = 0) and (Connections[bb].Size <> 0) then
          bNext := bb; { Highest unused basis state not involved in the current matrix }
      end;
    until not MaskChanged;
    b := bNext;
  until bNext = 0;
  nBlocks := Result;
end;

destructor TEigenHelper.Destroy;
var
  i: Integer;
begin
  FreeAndNil(Evals);
  for i := 0 to High(Matrices) do
    FreeAndNil(Matrices[i]);
  inherited;
end;

procedure TEigenHelper.EigBlocks(EvReqd, Verbose: Boolean; TheEigenvals: TColumnVector);
var
  BlockSize, i, k, kk, Block, BlockNo: Integer;
  BlockStarts, BlockEnds, Ind, BlockSource, BlockP: array of Integer;
  EMin: Double;
begin
  if Verbose then
    WriteLn(Log,nBlocks,' Blocks');
  CurBlockBase := 0;
  SetLength(BlockStarts, Length(Matrices));
  SetLength(BlockEnds, Length(Matrices));
  Evals := TColumnVector.Create(Size);
  { Diagonalize each block }
  for i := 0 to High(Matrices) do begin
    BlockStarts[i] := CurBlockBase+1;
    BlockSize := Matrices[i].RowCount;
    BlockEnds[i] := CurBlockBase+BlockSize+1;
    if Verbose then begin
      WriteLn(Log, 'Block #', i+1);
      Matrices[i].Write(Log, '#'+IntToStr(i+1), {$IFDEF FPC}@{$ENDIF}ShowLabel, {$IFDEF FPC}@{$ENDIF}ShowLabel, Width);
    end;
    EigLA.Eigen(Evals, Matrices[i], EvReqd, CurBlockBase);
    if Verbose then begin
      WriteLn(Log, 'Eigenvalues of Block #', i+1);
      Matrices[i].Write(Log, 'Ev', {$IFDEF FPC}@{$ENDIF}ShowLabel, {$IFDEF FPC}@{$ENDIF}ShowEv, Width);
      WriteLn(Log);
    end;
    Inc(CurBlockBase, BlockSize);
  end;
  { Now merge sort eigenvalues, recording indices as we go }
  SetLength(Ind, Size);
  SetLength(BlockSource, Size);
  BlockP := Copy(BlockStarts); { First unused eigenvalue in each block }
  i := 0;
  while i < Size do begin
    BlockNo := -1;
    { Find next lowest energy by searching individual blocks }
    for Block := 0 to High(Matrices) do
      if (BlockP[Block] < BlockEnds[Block])
         and ((BlockNo < 0) or (Evals[BlockP[Block]] < Emin)) then begin
        BlockNo := Block;
        Emin := Evals[BlockP[Block]];
      end;
    Ind[i] := BlockP[BlockNo];
    BlockSource[i] := BlockNo;
    Inc(BlockP[BlockNo]);
    Inc(i);
  end;
  { Permute energies }
  TheEigenvals.Size := Size;
  for i := 1 to Size do
    TheEigenvals[i] := Evals[Ind[i-1]];
  FreeAndNil(Evals);
  if EvReqd then begin
    { Copy eigenvectors into position }
    Matrix.Zero;
    for i := 1 to Size do begin
      Block := BlockSource[i-1];
      CurBlockBase := BlockStarts[Block]-1;
      k := Ind[i-1]-CurBlockBase;
      for kk := 1 to Matrices[Block].RowCount do
        Matrix[Perm[CurBlockBase+kk-1],i] :=  Matrices[Block][kk,k];
    end;
  end;
end;

procedure TEigenHelper.FactorMatrix;
{ Given the Matrix and a bitmap of non-zero elements, rearrange elements to block-diagonalize }
var
  b, bb, bbb, bNext, i, k, CIndex, BlockSize, NextBlock: Integer;
  CurMask, Mask, bMask: TBigSet;
  Block: TMatrix;
  Locations: array of Integer;
  V: Double;
begin
  { Now split matrix }
  SetLength(Matrices,nBlocks);
  SetLength(Perm, Size);
  for i := 0 to nBlocks - 1 do
    Matrices[i] := nil;
  b := Size;
  CIndex := Size-1;
  Mask.Size := Connections[Size].Size;
  bMask.Size := Mask.Size;
  SetLength(Locations, Size); { Note not all entries used }
  NextBlock := nBlocks-1;
  repeat
    BlockSize := 0;
    bNext := 0;
    Mask.Clear;
    Mask.SetBitn(b-1);
    CurMask := Connections[b];
    for bb := b downto 1 do begin
      if CurMask.AndNonZero(Mask) then begin
        { Rearrange basis by block }
        Perm[CIndex] := bb; { Allows original index to be found }
        Dec(CIndex);
        Inc(BlockSize);
      end else if (bNext = 0) and (Connections[bb].Size <> 0) then
        bNext := bb; { Highest unused basis state not involved in the current matrix }
      Mask.RotateRight;
    end;
    { Create next block matrix }
    Block := TMatrix.Create(BlockSize,BlockSize);
    Matrices[NextBlock] := Block;
    Dec(NextBlock);
    { Copy elements }
    Mask.Clear;
    Mask.SetBitn(b-1);
    i := BlockSize;
    Locations[1] := -1; { Flag not set up }
    for bb := b downto 1 do begin
      if CurMask.AndNonZero(Mask) then begin
        { On first hit, make array of contributing states
          to avoid multiple set operations }
        if Locations[1] < 0 then begin
          bMask.CopySet(Mask);
          k := i;
          for bbb := bb downto 1 do begin
            { For each other basis state }
            if CurMask.AndNonZero(bMask) then begin
              Locations[k] := bbb;
              Dec(k);
            end;
            bMask.RotateRight;
          end;
        end;
        Block[i,i] := Matrix[bb,Locations[i]];
        for k := 1 to i-1 do begin
          V := Matrix[bb,Locations[k]];
          Block[i,k] := V;
          Block[k,i] := V;
        end;
//        { This basis state contributes }
//        bMask.CopySet(Mask);
//        k := i;
//        for bbb := bb downto 1 do begin
//          { For each other basis state }
//          if CurMask.AndNonZero(bMask) then begin
//            Block[i,k] := Matrix[bb,bbb];
//            if i <> k then
//              Block[k,i] := Matrix[bbb,bb];
//            Dec(k);
//          end;
//          bMask.RotateRight;
//        end;
        Dec(i);
      end;
      Mask.RotateRight;
    end;
    b := bNext;
  until b = 0;
  { Sanity check }
  bb := 0;
  for b := 0 to High(Matrices) do
    Inc(bb,Matrices[b].RowCount);
  if bb <> Size then begin
    raise Exception.Create('EigBlock.FactorMatrix');
    //WriteLn(Log, bb);
    //Clear(Mask);
    //for b := 1 to High(Connections) do
    //  if High(Connections[b]) > 0 then
    //    OrIn(Mask,Connections[b]);
    //MatrixElements(Matrices, True, meHrot);
    //for b := 1 to High(Connections) do begin
    //  Write(Log, b:4, ' ');
    //  for bb := Size downto 1 do
    //    Write(Log, ((Connections[b][0] shr (bb-1)) and 1));
    //  WriteLn(Log);
    //end;
  end;
end;

procedure TEigenHelper.MakeConnections;
var
  Mask: TBigSet;
  i, j: Integer;
begin
  { Scan matrix, setting up connections }
  Size := Matrix.RowCount;
  SetLength(Connections,Matrix.Size+1);
  Mask.Size := Matrix.Size;
  Mask.SetBitn(0);
  for i := 1 to Size do begin
    Connections[i].CopySet(Mask); { Diagonal elements set }
    Mask.RotateLeft;
    for j := 1 to i-1 do
      if Abs(Matrix[i,j]) > Small then
        Connections[i].OrIn(Connections[j]);
  end;
end;

function TEigenHelper.ShowEv(Index: Integer): string;
begin
  Result := DtosM(Evals[CurBlockBase+Index], Width-1, -1);
end;

function TEigenHelper.ShowLabel(Index: Integer): string;
begin
  Index := Perm[CurBlockBase+Index-1];
  if Assigned(FOnShowLabel) then
    Result := FOnShowLabel(Index)
  else
    Result := IntToStr(Index);
end;

end.
