unit Huge;
{ Note that the HugeClip unit must also be used for the clipboard routines to work }
interface
{$I utils.inc}{ Copyright (C) 2005 C M Western; see UCompat.pas for license}
uses SysUtils, Classes, UCompat;

type
  TGHandle = class(TPersistent)
  private
    function GetPointer(n:SizeUInt):Pointer; inline;
    procedure SetSize(n:SizeUInt);
  protected
    FSize:SizeUInt;
    FBase:Pointer;
    {$IFDEF MSWINDOWS}
    FHandle:THandle;
      {$IFDEF VCL_}
      FromClipboard:Boolean;
      {$ENDIF}
    {$ENDIF}
  public
    constructor Create;
    destructor Destroy; override;
    procedure Clear;
    procedure Zero;
    property Size:SizeUInt read FSize write SetSize;
    procedure CopyToClipboard(Format:Cardinal); { Notionally TClipboardFormat }
    procedure SetFromClipboard(Format:Cardinal);
    property Item[n:SizeUInt]:Pointer read GetPointer;
    property Base: Pointer read FBase;
    procedure Assign(Source: TPersistent); override;
    procedure AssignAndNil(var Source: TGHandle);
  end;

  TGVectorN = class(TGHandle)
  private
    FItemSize: Cardinal;
    Mask: Cardinal;
    ShiftCount: Cardinal;
    procedure SetSize(n:Integer);
    function GetPointer(n:Integer):Pointer; inline;
    function GetHigh: Integer;
  protected
    function GetSize:Integer;
  public
    Offset: Integer;
    constructor Create(AItemSize:Cardinal);
    procedure Move(Src, Dst:Integer);
    property Size:Integer read GetSize write SetSize;
    property ItemSize:Cardinal read FItemSize;
    property Item[n:Integer]:Pointer read GetPointer;
    property High: Integer read GetHigh;
    procedure Assign(Source: TPersistent); override;
  end;

  { THugeMemoryStream }

  THugeMemoryStream = Class(TCustomMemoryStream)
  { Much as TMemoryStream, except that intenal buffer is a TGHandle object }
  private
    Buffer:TGHandle;
  public
    constructor Create; overload;
    constructor Create(GHandle: TGHandle); overload;
    constructor Create(AString: AnsiString); overload;
    destructor Destroy; override;
    function GetCapacity:Longint;
    procedure SetCapacity(n:Longint);
    function Write(const Buf; Count: Longint): Longint;  override;
    procedure CopyToClipboard(Format:Cardinal);
    procedure SetFromClipboard(Format:Cardinal; ForceUTF8: Boolean = False);
    property Capacity:Longint read GetCapacity write SetCapacity;
    procedure SetSize(NewSize: Longint); override;
    procedure Disconnect;  { Disconnect from underlying GHandle }
  end;

var
  GMemoryCount: SizeUInt; { Note this is only approximate, as the count is not maintained in a thread safe way }
  CopyHandleToClipboard: procedure(GHandle: TGHandle; Format:Cardinal) = nil;
  SetHandleFromClipboard: procedure(GHandle: TGHandle; Format:Cardinal) = nil;

implementation

uses {$IFDEF MSWINDOWS} Windows, {$ENDIF}
     Math;

{ THugeMemoryStream }

constructor THugeMemoryStream.Create(GHandle: TGHandle);
begin
  inherited Create;
  Buffer := GHandle;
  SetPointer(GHandle.Base, GHandle.Size);
end;

constructor THugeMemoryStream.Create;
begin
  Create(TGHandle.Create);
end;

constructor THugeMemoryStream.Create(AString: AnsiString);
begin
  Create;
  Size := Length(AString);
  if Size > 0 then
    Write(AString[1], Length(AString));
end;

destructor THugeMemoryStream.Destroy;
begin
  FreeAndNil(Buffer);
  inherited Destroy;
end;

procedure THugeMemoryStream.SetSize(NewSize: Longint);
begin
  Buffer.Size := NewSize; { Tuncate buffer contents }
  SetPointer(Buffer.Base, Buffer.Size);
end;

procedure THugeMemoryStream.Disconnect;
begin
  Buffer := nil;
end;

function THugeMemoryStream.GetCapacity:Longint;
begin
  Result := Buffer.Size;
end;

procedure THugeMemoryStream.SetCapacity(n:Longint);
begin
  Buffer.Size := n;
  SetPointer(Buffer.Base, Min(n,Size));
end;

function THugeMemoryStream.Write(const Buf; Count: Longint): Longint;
var
  Pos, NewPos, NewSize: SizeUInt;
begin
  Pos := Position;
  NewPos := Pos+Count;
  if NewPos > Size then begin
    if NewPos > Buffer.Size then begin
      NewSize := Buffer.Size * 5 div 4;
      if NewPos > NewSize then
        NewSize := NewPos;
      Buffer.Size := (NewSize or $ffff) + 1;
    end;
    SetPointer(Buffer.Base, NewPos);
  end;
  Move(Buf, Buffer.Item[Pos]^, Count);
  Position := NewPos;
  Result := Count;
end;

procedure THugeMemoryStream.CopyToClipboard(Format:Cardinal);
begin
  {$IFDEF MSWINDOWS}
  if (Format = CF_PREFTEXT) and (PChar(Buffer.Item[Size - SizeOf(Char)])^ <> #0) then begin
    { Ensure null character terminator present }
    Size := Size + SizeOf(Char);
    PChar(Buffer.Item[Size - SizeOf(Char)])^ := #0;
  end else
  {$ENDIF}
    Buffer.Size := Size;
  Buffer.CopyToClipboard(Format);
end;

procedure THugeMemoryStream.SetFromClipboard(Format:Cardinal; ForceUTF8: Boolean);
{$IFDEF UNICODE}
var
  nChars: Integer;
  OldBuffer: TGHandle;
{$ENDIF}
begin
  Buffer.SetFromClipboard(Format);
  SetPointer(Buffer.Base, Buffer.Size);
  {$IFDEF MSWINDOWS}
    if (Size > 0) and ((Format = CF_PREFTEXT) or ForceUTF8) then begin
      if PChar(Buffer.Item[Size-SizeOf(Char)])^ = #0 then
        Size := Size - SizeOf(Char); { Stream should not include null terminator }
      {$IFDEF UNICODE}
        nChars := WideCharToMultiByte(CP_UTF8, 0, Buffer.Base, Size div SizeOf(Char), nil, 0, nil, nil);
        OldBuffer := Buffer;
        Buffer := TGHandle.Create;
        Size := nChars;
        WideCharToMultiByte(CP_UTF8, 0, OldBuffer.Base, OldBuffer.Size div SizeOf(Char), Buffer.Base, nChars, nil, nil);
        OldBuffer.Free;
      {$ENDIF}
    end;
  {$ENDIF}
end;

{ TGHandle }

constructor TGHandle.Create;
begin
  FSize := 0;
  {$IFDEF MSWINDOWS}
  {$IFDEF VCL_}
  FromClipboard := False;
  {$ENDIF}
  FHandle := 0;
  {$ENDIF}
  FBase := nil;
end;

destructor TGHandle.Destroy;
begin
  Clear;
  inherited Destroy;
end;

procedure TGHandle.CopyToClipboard(Format:Cardinal);
begin
  if Assigned(CopyHandleToClipboard) then
    CopyHandleToClipboard(Self, Format);
end;

procedure TGHandle.SetFromClipboard(Format:Cardinal);
begin
  if Assigned(SetHandleFromClipboard) then
    SetHandleFromClipboard(Self, Format);
end;

procedure TGHandle.Clear;
begin
  {$IFDEF MSWINDOWS}
    if (FBase <> nil) and (FHandle <> 0) then
      GlobalUnlock(FHandle);
    if (FHandle <> 0) then
      {$IFDEF VCL_}
        if FromClipboard then
          SetHandleFromClipboard(nil,0) { Forces Clipboard.Close }
        else
      {$ENDIF}
          GlobalFree(FHandle);
    FHandle := 0;
    {$IFDEF VCL_}
      FromClipboard := False;
    {$ENDIF}
  {$ENDIF}
  {$IFDEF UNIX}
    if (FBase) <> nil then
      FreeMem(FBase);
  {$ENDIF}
  {$IFOPT R+}
    {$R-}{$Q-} { Must turn off checks, as addition is not thread safe }
    Dec(GMemoryCount, FSize);
    {$R+}{$Q+}
  {$ELSE}
    Dec(GMemoryCount, FSize);
  {$ENDIF}
  FBase := nil;
  FSize := 0;
end;

procedure TGHandle.Zero;
begin
  FillChar(Fbase^, FSize, 0);
end;

procedure TGHandle.SetSize(n:SizeUInt);
begin
  if n = 0 then
    Clear
  else if n <> FSize then begin
    {$IFOPT R+}
      {$R-}{$Q-}
      Dec(GMemoryCount, FSize);
      {$R+}{$Q+}
    {$ELSE}
      Dec(GMemoryCount, FSize);
    {$ENDIF}
    {$IFDEF MSWINDOWS}
    FBase := nil;
    if FSize = 0 then
      FHandle := GlobalAlloc(GMEM_MOVEABLE or GMEM_ZEROINIT, n)
    else begin
      GlobalUnlock(FHandle);
      FHandle := GlobalRealloc(FHandle, n, GMEM_MOVEABLE or GMEM_ZEROINIT);
    end;
    if FHandle <> 0 then
      FBase := GlobalLock(FHandle)
    else
      OutOfMemoryError;
    if Fbase = nil then
      raise Exception.Create('GlobalLock Fails');
    {$ENDIF}
    {$IFDEF UNIX}
    if FSize = 0 then
      GetMem(FBase, n)
    else begin
      FSize := 0;  { In case an exception raised }
      ReallocMem(FBase, n);
    end;
    {$ENDIF}
    FSize := n;
    {$IFOPT R+}
      {$R-}{$Q-}
      Inc(GMemoryCount, FSize);
      {$R+}{$Q+}
    {$ELSE}
      Inc(GMemoryCount, FSize);
    {$ENDIF}
  end;
end;

function TGHandle.GetPointer(n:SizeUInt):Pointer;
begin
{$IFOPT R+}
  if n >= FSize then
    raise Exception.CreateFmt('TGHandle.GetPointer(%d) Size=%d', [n,FSize])
  else
{$ENDIF}
    Result := PAnsiChar(FBase)+n;
end;

procedure TGHandle.Assign(Source: TPersistent);
begin
  if Source is TGHandle then begin
    Size := TGHandle(Source).Size;
    Move(TGHandle(Source).Base^, Base^, Size);
  end else
    inherited Assign(Source);
end;

procedure TGHandle.AssignAndNil(var Source: TGHandle);
begin
  Clear;
  FSize := Source.Size;
  Source.FSize := 0;
  FBase := Source.Base;
  Source.FBase := nil;
  {$IFDEF MSWINDOWS}
    FHandle := Source.FHandle;
    Source.FHandle := 0;
    {$IFDEF VCL_}
      FromClipboard := Source.FromClipboard;
    {$ENDIF}
  {$ENDIF}
  FreeAndNil(Source);
end;

constructor TGVectorN.Create(AItemSize:Cardinal);
begin
  inherited Create;
  Offset := 0;
  FItemSize := AItemSize;
  Mask := AItemSize - 1;
  ShiftCount := 0;
  while AItemSize > 1 do begin
    AItemSize := AItemSize div 2;
    Inc(ShiftCount);
  end;
  if Cardinal(1 shl ShiftCount) <> FItemSize then
    raise Exception.CreateFmt('TGVectorN: ItemSize(&d) not a power of 2', [FItemSize]);
end;

function TGVectorN.GetSize:Integer;
begin
  Result := (inherited Size) shr ShiftCount;
end;

procedure TGVectorN.SetSize(n:Integer);
begin
  inherited Size := n * FItemSize;
end;

procedure TGVectorN.Move(Src, Dst:Longint);
var
  Bold: array [0..255] of Byte;
  P: PAnsiChar;
  i: Longint;
begin
  if ItemSize > 256 then
    raise Exception.CreateFmt('TGVectorN.Move: ItemSize=%d', [ItemSize]);
  P := GetPointer(Src);
  System.Move(P^, Bold, ItemSize);
  if Src > Dst then begin
    for i := 1 to Src-Dst do begin
      System.Move((P-ItemSize)^, P^, ItemSize);
      Dec(P,ItemSize);
    end;
  end else begin
    for i := 1 to Dst-Src do begin
      System.Move((P+ItemSize)^, P^, ItemSize);
      Inc(P,ItemSize);
    end;
  end;
  System.Move(Bold, P^, ItemSize);
end;

function TGVectorN.GetPointer(n:Integer):Pointer;
begin
  n := (n-Offset) shl ShiftCount;
  {$IFOPT R+}
  if (n < 0) or (n >= FSize) or ((n and Mask) <> 0) then
    raise Exception.CreateFmt('TGVectorN.GetPointer(%d) Size=%d', [n,Size])
  else
  {$ENDIF}
    Result := PAnsiChar(FBase)+n;
end;

procedure TGVectorN.Assign(Source: TPersistent);
var
  s: TGVectorN;
begin
  if Source is TGVectorN then begin
    s := TGVectorN(Source);
    if s.ItemSize <> ItemSize then
      raise Exception.CreateFmt('TGVectorN.Assign: ItemSize: %d <> %d', [s.ItemSize, ItemSize]);
    Offset := s.Offset;
    inherited Assign(Source); { Do the copy }
  end else
    inherited Assign(Source);
end;

function TGVectorN.GetHigh: Integer;
begin
  Result := Offset + Size - 1;
end;

end.
