unit MatrixOps;
{$I utils.inc}{ Copyright (C) 2007,8 C M Western; see UCompat.pas for license}
interface

uses Vector, MatrixVec;

procedure Invert(A: TMatrix);
function Determinant(A: TMatrix):Double;

implementation

{.$DEFINE TEST}

uses {$IFDEF TEST} Worker, {$ENDIF}
     SysUtils, Lapack, EigLA;

{ Replaces A with its inverse }
procedure Invert(A: TMatrix);
var
  n, info, lwork: Integer;
  ipiv: array of Integer;
  work: array of Double;
  nt: Double;
  {$IFDEF TEST}
  T: TMatrix;
  i,j: Integer;
  {$ENDIF}
begin
  {$IFDEF TEST}
  T := TMatrix.Create(A);
  {$ENDIF}
  n := A.RowCount;
  if n <> A.ColCount then
    raise Exception.Create('MatrixOps.Invert: Matrix must be square');
  SetLength(ipiv,n);
  dgetrf_(n, n, A.FirstItem^, n, ipiv[0], info);
  if info < 0 then
    raise Exception.CreateFmt('MatrixOps.Invert: Illegal argument: %d', [-info]);
  if info > 0 then
    raise Exception.CreateFmt('MatrixOps.Invert: U(%d,%d)=0', [info, info]);
  lwork := -1; { Compute workspace }
  SetLength(work,n);
  dgetri_(n, A.FirstItem^, n, ipiv[0], nt, lwork, info);
  if info <> 0 then
    raise Exception.CreateFmt('MatrixOps.Invert: Illegal argument to dgetri: %d', [-info]);
  lwork := Round(nt);
  SetLength(work, lwork);
  dgetri_(n, A.FirstItem^, n, ipiv[0], work[0], lwork, info);
  if info <> 0 then
    raise Exception.CreateFmt('MatrixOps.Invert: Illegal argument to dgetri: %d', [-info]);
  {$IFDEF TEST}
  T.Write(Log, 'Original');
  A.Write(Log, 'Inverse');
  T.PostMultiply(A);
  T.Write(Log, 'Product');
  for i := 1 to n do begin
    if not SameValue(T[i,i],1) then
      raise Exception.CreateFmt('Fail %g',[A[i,i]]);
    for j := 1 to n do
      if (i <> j) and (Abs(T[i,j]) > 1e-14) then
        raise Exception.Create('Fail');
  end;
  T.Free;
  {$ENDIF}
end;

function Determinant(A: TMatrix):Double;
var
  n, info: Integer;
  ipiv: array of Integer;
  w: array of Double;
  i: Integer;
  {$IFDEF TEST}
  T: TMatrix;
  V: TColumnVector;
  P: Double;
  {$ENDIF}
begin
  n := A.RowCount;
  {$IFDEF TEST}
  { This test only works on symmetric matrices }
  A.Write(Log,'Determinant of');
  T := TMatrix.Create(A);
  V := TColumnVector.Create(A.RowCount);
  Eigen(V,T,False);
  P := V[1];
  for i := 2 to n do
    P := P * V[i];
  {$ENDIF}
  if n <> A.ColCount then
    raise Exception.Create('MatrixOps.Determinant: Matrix must be square');
  SetLength(ipiv,n);
  SetLength(w, n*n);
  Move(A.FirstItem^,w[0],n*n*SizeOf(Double));
  dgetrf_(n, n, w[0], n, ipiv[0], info);
  if info < 0 then
    raise Exception.CreateFmt('MatrixOps.Determinant: Illegal argument: %d', [-info]);
  Result := 1;
  { Method for extraction from LU decomposition based on determinant function from Octave package }
  for i := 0 to n-1 do begin
    if ipiv[i] <> i+1 then
      Result := -Result;
    Result := Result*w[i+i*n];
  end;
  {$IFDEF TEST}
  WriteLn(Log, Result, ' From Diag= ', P);
  // if not SameValue(Result,P) then
    // raise Exception.CreateFmt('Fail %g %g', [Result, P]);
  {$ENDIF}
end;

{$IFDEF NOTLAPACK}
{ Replaces A with its inverse, and, optionally b, by a**-1*b }
procedure GaussJordan(A: TMatrix; B: TColumnVector);
var
  iPiv, indxr, indxc: TIntColumnVector;
  N: Integer;
  big, dum, pivinv: Double;
  iRow, iCol, I, J, K, L, LL: Longint;
begin
  N := A.ColCount;
  iPiv := nil;
  indxr := nil;
  indxc := nil;
  try
    iPiv := TIntColumnVector.Create(N);
    indxr := TIntColumnVector.Create(N);
    indxc := TIntColumnVector.Create(N);
    { All above initialised to zero }
    for I := 1 to N do begin
      BIG := 0;
      for J := 1 to N do
        if iPiv[J] <> 1 then
          for K := 1 to N do
            if iPiv[K] = 0 then
              if Abs(a[J,K]) >= BIG then begin
                BIG := Abs(a[J,K]);
                IROW := J;
                ICOL := K;
              end
            else if iPiv[K] > 1 then
              raise Exception.Create('GaussJordan: Singular Matrix');
      { Ignore compiler warning here - >= BIG above implies ICOL, IROW always set }
      iPiv[ICOL] := iPiv[ICOL] + 1;
      if IROW <> ICOL then begin
        for L := 1 to N do begin
          DUM := a[IROW ,L];
          a[IROW,L] := a[ICOL,L];
          a[ICOL,L] := DUM;
        end;
        if Assigned(b) then begin
          DUM := b[IROW];
          b[IROW] := b[ICOL];
          b[ICOL] := DUM;
        end;
      end;
      indxr[I] := IROW;
      indxc[I] := ICOL;
      if a[ICOL,ICOL] = 0 then
        raise Exception.Create('GaussJordan: Singular Matrix');
      PIVINV := 1./a[ICOL,ICOL];
      for L := 1 to N do
        a[ICOL,L] := a[ICOL,L]*PIVINV;
      a[ICOL,ICOL] := PIVINV;
      if Assigned(b) then
        b[ICOL] := b[ICOL]*PIVINV;
      for LL := 1 to N do
        if LL <> ICOL then begin
          DUM := a[LL,ICOL];
          a[LL,ICOL] := 0;
          for L := 1 to N do
            a[LL,L] := a[LL,L]-a[ICOL,L]*DUM;
          if Assigned(b) then
            b[LL] := b[LL]-b[ICOL]*DUM;
        end;
    end;
    for L := N downto 1 do
      if indxr[L] <> indxc[L] then
        for K := 1 to N do begin
          DUM := a[K,indxr[L]];
          a[K,indxr[L]] := a[K,indxc[L]];
          a[K,indxc[L]] := DUM;
        end;
  finally
    iPiv.Free;
    indxr.Free;
    indxc.Free;
  end;
end;
{$ENDIF}

end.
