unit QuantumNumber;
{$I utils.inc}{ Copyright (C) 2005 C M Western; see UCompat.pas for license}
interface

type
  TQuantumNumber = Integer;
  TQuantumNumberArray = array of TQuantumNumber;

function IsTriangle(J1, J2, J3: TQuantumNumber): Boolean;
procedure MatchOdd(A, B: TQuantumNumber);
function SmallestProjection(A: TQuantumNumber): TQuantumNumber;
procedure BadMinus(A: TQuantumNumber);
function Minus(A: TQuantumNumber): Integer; inline; { (-1)**A }
function BinomialCoefficient(n, k: Integer):Integer;

type
  TQuantumNumberFormat = (qftJ=0, qft2J=1, qftJph=2, qftJmh=3);

const
  QuantumNumberFormatNames: array [0..4] of PChar = ('J', '2J', 'J+1/2', 'J-1/2', nil);
  MaxQuantumNumber: TQuantumNumber = 200000;

function TryStrToQuantumNumber(const str: string; out Value: TQuantumNumber; QuantumNumberFormat: TQuantumNumberFormat):Boolean;
function QuantumNumberToStr(const Value:TQuantumNumber; QuantumNumberFormat: TQuantumNumberFormat):{$IFDEF FPC}ShortString{$ELSE}string{$ENDIF};
function QuantumNumberToDisplayDouble(const Value:TQuantumNumber; QuantumNumberFormat: TQuantumNumberFormat):Double;

implementation

uses SysUtils, UCompat;

function IsTriangle(J1, J2, J3: TQuantumNumber): Boolean;
{ Do J1, J2, J3 satisfy the triangle rule for angular momentum addition? }
begin
  if Odd(J1+J2+J3) then
    raise Exception.CreateFmt('Triangle: %d %d %d', [J1, J2, J3]);
  Result := (Abs(J1-J2) <= J3) and (J1+J2 >= J3);
end;

procedure MatchOdd(A, B: TQuantumNumber);
begin
  if Odd((A xor B) and 1) then
    raise EReportException.CreateFmt('Mix of integral and half integral quantum numbers: %d %d', [A, B]);
end;

function SmallestProjection(A: TQuantumNumber): TQuantumNumber;
begin
  Result := A and 1;
end;

procedure BadMinus(A: TQuantumNumber);
begin
  { Do not inline this function }
  raise Exception.CreateFmt('Minus(%d)', [A]);
end;

function Minus(A: TQuantumNumber): Integer; { (-1)**A }
begin
  if (A and 1) <> 0 then
    BadMinus(A);
  if (A and 2) <> 0 then
    Result := -1
  else
    Result := 1;
end;

function BinomialCoefficient(n, k: Integer):Integer;
var
  i: Integer;
begin
  { n!/((n-k)!k!) }
  if k > n-k then  { Make k small }
    k := n-k;
  Result := 1;
  for i := n-k+1 to n do
    Result := Result * i;
  for i := 2 to k do
    Result := Result div i;
end;

function TryStrToQuantumNumber(const str: string; out Value: TQuantumNumber; QuantumNumberFormat: TQuantumNumberFormat):Boolean;
var
  Minus, V, Half: Integer;
  p: PChar;
  noDigits: Boolean;
begin
  Result := False;
  if Length(Str) = 0 then
    Exit;
  p := @str[1];
  while p^ = ' ' do
    Inc(p);
  if p^ = '-' then begin
    Inc(p);
    Minus := -1;
  end else
    Minus := 1;
  noDigits := p^ = '.';
  if not (noDigits or (p^ in ['0'..'9'])) then
    Exit; { No digits }
  V := 0;
  while p^ in ['0'..'9'] do begin
    V := V*10 + Ord(p^) - Ord('0');
    if V > MaxQuantumNumber then
      Exit;
    Inc(p);
  end;
  Half := 0;
  if p^ = '.' then begin
    { Accept .5 or .0 only }
    Inc(p);
    if p^ = '5' then begin
      Half := 1;
      Inc(p);
    end else if noDigits and (p^ <> '0') then
      Exit; { Reject '.' on it's own }
    while p^ = '0' do
      Inc(p);
  end else if p^ = '/' then begin
    { also accept /2 notation for qftJ only }
    Inc(p);
    if (p^ <> '2') or (QuantumNumberFormat <> qftJ) then
      Exit;
    Inc(p);
    QuantumNumberFormat := qft2J; { Value is already doubled }
  end;
  while p^ = ' ' do
    Inc(p);
  if p^ = #0 then begin
    case QuantumNumberFormat of
      qftJ:   V := 2*V;
      qft2J:  if Half > 0 then Exit; { Can't have 0.5 if doubled }
      qftJph: V := 2*V-1;
      qftJmh: V := 2*V+1;
    end;
    Value := Minus*(V+Half);
    Result := True;
  end;
end;


{$IFDEF FPC}
{ Shortstring version currently (Feb 2012) rather quicker for FPC }
function QuantumNumberToStr(const Value:TQuantumNumber; QuantumNumberFormat: TQuantumNumberFormat):ShortString;
begin
  case QuantumNumberFormat of
    qft2J: Str(Value, Result);
    qftJ:  if Value < 0 then
             Result := '-'+QuantumNumberToStr(-Value,qftJ)
           else begin
             Str(Value div 2, Result);
             if Odd(Value) then begin
               Setlength(Result, Length(Result)+2);
               Result[Length(Result)-1] := '.';
               Result[Length(Result)] := '5';
             end;
           end;
    qftJph: Result := QuantumNumberToStr(Value+1, qftJ);
    qftJmh: Result := QuantumNumberToStr(Value-1, qftJ);
  end;
end;
{$ELSE}
function QuantumNumberToStr(const Value:TQuantumNumber; QuantumNumberFormat: TQuantumNumberFormat):string;
begin
  case QuantumNumberFormat of
    qft2J: Result := IntToStr(Value);
    qftJ:  if Value < 0 then
             Result := '-'+QuantumNumberToStr(-Value,qftJ)
           else begin
             Result := IntToStr(Value div 2);
             if Odd(Value) then
               Result := Result + '.5';
           end;
    qftJph: Result := QuantumNumberToStr(Value+1, qftJ);
    qftJmh: Result := QuantumNumberToStr(Value-1, qftJ);
  end;
end;
{$ENDIF}

function QuantumNumberToDisplayDouble(const Value:TQuantumNumber; QuantumNumberFormat: TQuantumNumberFormat):Double;
begin
  case QuantumNumberFormat of
    qftJ:   Result := Value/2;
    qft2J:  Result := Value;
    qftJph: Result := Value + 0.5;
    qftJmh: Result := Value - 0.5;
  end;
end;

end.
