unit StackTrace;
{$I utils.inc}{ Copyright (C) 2005, 2013 C M Western; see UCompat.pas for license}
{ Routines to produce a Stack Trace.
  Extracted and adapted from MemCheck, by: Jean Marc Eber & Vincent Mahon, Socit Gnrale, INFI/SGOP/R&D
    http://v.mahon.free.fr/pro/freeware/memcheck
  Several additional checks needed to work under unicode based Delphi;
  64 bit stack walk based on FPC }

interface

const
  StoredCallStackDepth = 26;
  {Size of the call stack stored; must be an EVEN number for MemCheck }

{$IF CompilerVersion < 23.0}
type
  NativeUInt = Cardinal;
  PNativeUInt = PCardinal;
{$IFEND}

type
  TCallStack = array[0..StoredCallStackDepth] of NativeUInt;
  PCallStack = ^TCallStack;

function TextualDebugInfoForAddress(const TheAddress: NativeUInt): ShortString;

procedure FillCallStack(var St: TCallStack; const NbLevelsToExclude: integer; TopAddress: Pointer = nil);
//Fills St with the call stack

function CallStackTextualRepresentation(const S: TCallStack; const LineHeader: string): string;
//Will contain CR/LFs

implementation

uses {$IFDEF FASTMM4} FastMM4, {$ENDIF}
     Windows, Math, SysUtils;

function CallStackTextualRepresentation(const S: TCallStack; const LineHeader: string): string;
var
  i: integer;
begin
  i := 0;
  Result := '';
  while (i <= StoredCallStackDepth) and (S[i] <> 0) do begin
    Result := Result + LineHeader + 'call stack - ' + IntToStr(i) + ' : ' + TextualDebugInfoForAddress(S[i]) + #13#10;
    i := i + 1;
  end;
end;

{$STACKFRAMES OFF}
function StackTop:NativeUInt;
asm
  {$IFDEF WIN32}
        mov EAX, FS:[4]
  {$ELSE}
        mov RAX, GS:[ABS 8]
  {$ENDIF}
end;

function StackFrame: NativeUInt;
asm
  {$IFDEF WIN32}
        mov EAX, EBP
  {$ELSE}
        mov RAX, RBP
  {$ENDIF}
end;
{$STACKFRAMES ON}

{ Options }
const
    UseDebugInfos = True;
    //Should use the debug informations which are in the executable ?
{ End of Options }

const
    NoDebugInfo = '(no debug info)';

type
    TAddressToLine = class
    public
        Address: Cardinal;
        Line: Cardinal;

        constructor Create(const AAddress, ALine: Cardinal);
    end;

    PAddressesArray = ^TAddressesArray;
    TAddressesArray = array[0..MaxInt div 16 - 1] of TAddressToLine;

    TUnitDebugInfos = class
    public
        Name: AnsiString;
        Addresses: array of TAddressToLine;

        constructor Create(const AName: AnsiString; const NbLines: Cardinal);

        function LineWhichContainsAddress(const Address: Cardinal): string;
    end;

    TRoutineDebugInfos = class
    public
        Name: AnsiString;
        StartAddress: Cardinal;
        EndAddress: Cardinal;

        constructor Create(const AName: AnsiString; const AStartAddress: Cardinal; const ALength: Cardinal);
    end;

var
    Routines: array of TRoutineDebugInfos;
    RoutinesCount: integer;
    Units: array of TUnitDebugInfos;
    UnitsCount: integer;

var
  Displ: Cardinal;
  {Displ is the displacement of the code in the executable file. The code in SetDispl was written by Juan Vidal Pich}

{$IFDEF DCC}
  {$IFDEF WIN64}
    type PImageOptionalHeader = PImageOptionalHeader64;
  {$ELSE}
    type PImageOptionalHeader = PImageOptionalHeader32;
  {$ENDIF}
{$ENDIF}
procedure SetDispl;
var
    NTHeader: PImageFileHeader;
    NTOptHeader: PImageOptionalHeader;
begin
    //-> If you have a compilation error in this routine and you are compiling with delphi 4, I'd say that you did not install the Delphi update pack 3
    NTHeader := PImageFileHeader(NativeUInt(PImageDosHeader(HInstance)._lfanew) + HInstance + 4); {SizeOf(IMAGE_NT_SIGNATURE) = 4}
    NTOptHeader := PImageOptionalHeader(NativeUInt(NTHeader) + IMAGE_SIZEOF_FILE_HEADER);
    Displ := HInstance + NTOptHeader.BaseOfCode;
    //Result := HInstance + PImageNtHeaders(LongInt(HInstance)+PImageDosHeader(HInstance)^._lfanew)^.OptionalHeader.BaseOfCode;
end;

{$IFDEF WIN32}

procedure FillCallStack(var St: TCallStack; const NbLevelsToExclude: integer; TopAddress: Pointer);
    {Works only with stack frames - Without, St contains correct info, but is not as deep as it should
    I just don't know a general rule for walking the stack when they are not there}
var
  StackStart: Cardinal;
  StackMax: Cardinal; //the stack can never go beyond - http://msdn.microsoft.com/library/periodic/period96/S2CE.htm
  CurrentFrame: Cardinal;
  Base: Cardinal;
  Count: integer;
begin
  FillChar(St, SizeOf(St), 0);
  StackMax := StackTop;
  StackStart := StackFrame;
  CurrentFrame:= StackStart;
  if TopAddress <> nil then begin
    St[0] := Cardinal(TopAddress);
    Count := 1;
  end else
    Count:= -NbLevelsToExclude;
  while (CurrentFrame >= StackStart) and (CurrentFrame < StackMax) and (Count <= StoredCallStackDepth) do begin
    Base := PCardinal(CurrentFrame + SizeOf(Cardinal))^;
    if Base <= SizeOf(Cardinal) then
      Exit;
    if Base - SizeOf(Cardinal) = Cardinal(TopAddress) then
      Count := 0;
    if Count >= 0 then
      St[Count] := Base - SizeOf(Cardinal);
    CurrentFrame:= PCardinal(CurrentFrame)^;
    Inc(Count);
  end;
end;
{$ENDIF}

const
    MaxListSize = MaxInt div 16 - 1;

type
    SRCMODHDR = packed record
        _cFile: Word;
        _cSeg: Word;
        _baseSrcFile: array[0..MaxListSize] of Integer;
    end;

    SRCFILE = packed record
        _cSeg: Word;
        _nName: Integer;
        _baseSrcLn: array[0..MaxListSize] of Integer;
    end;

    SRCLN = packed record
        _Seg: Word;
        _cPair: Word;
        _Offset: array[0..MaxListSize] of Cardinal;
    end;

    PSRCMODHDR = ^SRCMODHDR;
    PSRCFILE = ^SRCFILE;
    PSRCLN = ^SRCLN;

    TArrayOfByte = array[0..MaxListSize] of Byte;
    PArrayOfByte = ^TArrayOfByte;
    TArrayOfWord = array[0..MaxListSize] of Word;
    PArrayOfWord = ^TArrayOfWord;
    TArrayOfPointer = array[0..MaxListSize] of PArrayOfByte;
    PArrayOfPointer = ^TArrayOfPointer;

procedure AddRoutine(const Name: AnsiString; const Start, Len: Cardinal);
begin
  if Length(Routines) <= RoutinesCount then
    SetLength(Routines, Max(RoutinesCount * 2, 1000));
  Routines[RoutinesCount] := TRoutineDebugInfos.Create(Name, Start, Len);
  RoutinesCount := RoutinesCount + 1;
end;

procedure AddUnit(const U: TUnitDebugInfos);
begin
  if Length(Units) <= UnitsCount then
    SetLength(Units, Max(UnitsCount * 2, 1000));
  Units[UnitsCount] := U;
  UnitsCount := UnitsCount + 1;
end;

procedure dumpsymbols(NameTbl: PArrayOfPointer; sstptr: PArrayOfByte; size: integer);
//Copyright (C) Tenth Planet Software Intl., Clive Turvey 1998. All rights reserved. - Reused & modified by SG with permission
var
  len, sym: integer;
  seg: Word;
begin
  while size > 0 do begin
    len := PWord(@sstptr^[0])^;
    sym := PWord(@sstptr^[2])^;
    seg := PWord(@sstptr^[32])^; // Assume segment 1 is main code segment

    INC(len, 2);

    if ((sym = $205) or (sym = $204)) and (seg = 1) and (PInteger(@sstptr^[40])^ > 0) then begin
//      WriteLn(log, IntToHex(sym,4), ' ', seg, ' ', IntToHex(PInteger(@sstptr^[28])^,8), ' ', IntToHex(PInteger(@sstptr^[16])^,8), PAnsiChar(NameTbl^[PInteger(@sstptr^[40])^ - 1]));
      AddRoutine(PAnsiChar(NameTbl^[PInteger(@sstptr^[40])^ - 1]), PInteger(@sstptr^[28])^, PInteger(@sstptr^[16])^);
    end;
    if (len = 2) then
      size := 0
    else begin
      sstptr := PArrayOfByte(@sstptr^[len]);
      DEC(size, len);
    end;
  end;
end;

procedure dumplines(NameTbl: PArrayOfPointer; sstptr: PArrayOfByte; size: integer);
//Copyright (C) Tenth Planet Software Intl., Clive Turvey 1998. All rights reserved. - Reused & modified by SG with permission
var
  srcmodhdr: PSRCMODHDR;
  i: Word;
  srcfile: PSRCFILE;
  srcln: PSRCLN;
  k, seg: Word;
  CurrentUnit: TUnitDebugInfos;
  LastAd: Cardinal;
begin
  LastAd := 0;
  if size > 0 then begin
    srcmodhdr := PSRCMODHDR(sstptr);
    for i := 0 to pred(srcmodhdr^._cFile) do begin
      srcfile := PSRCFILE(@sstptr^[srcmodhdr^._baseSrcFile[i]]);
      if srcfile^._nName > 0 then
        for seg := 0 to Pred(srcfile^._cSeg) do begin
          srcln := PSRCLN(@sstptr^[srcfile^._baseSrcLn[0]]);
          CurrentUnit := TUnitDebugInfos.Create(ExtractFileName(PAnsiChar(NameTbl^[srcfile^._nName - 1])), srcln^._cPair);
          if srcln^._Seg <> 1 then
            Continue; // Assumes first segment contains the code
          if Integer(srcln^._Offset[0]) <= 0 then
            Continue; { Weird address - skip }
          if LastAd = srcln^._Offset[0] then
            Continue; { Duplicated address - skip }
          AddUnit(CurrentUnit);
          LastAd := srcln^._Offset[0];
          for k := 0 to pred(srcln^._cPair) do
            CurrentUnit.Addresses[k] := TAddressToLine.Create(srcln^._Offset[k], PArrayOfWord(@srcln^._Offset[srcln^._cPair])^[k]);
        end;
    end;
  end;
end;

procedure GetProjectInfos;
//Copyright (C) Tenth Planet Software Intl., Clive Turvey 1998. All rights reserved. - Reused & modified by SG with permission
var
  AHeader: packed record
    Signature: array[0..3] of AnsiChar;
    AnInteger: Integer;
  end;
  k: integer;
  j: Word;
  lfodir: Integer;
  SstFrameSize: integer;
  SstFrameElem: PArrayOfByte;
  ssttype, sstsize, sstbase: Integer;
  x, y, z: Integer;
  sstbuf: PArrayOfByte;
  OldFileMode: integer;
  AFileOfByte: file of Byte;
  Names: PArrayOfByte;
  NameTbl: PArrayOfPointer;
  SstFrame: PArrayOfByte;
  ifabase: Integer;
  cdir, cbdirentry: word;
  FileName: string;
begin
    RoutinesCount := 0;
    UnitsCount := 0;
    OldFileMode := FileMode;
    FileMode := 0;
    SetLength(FileName, MAX_PATH + 1);
    SetLength(FileName, GetModuleFileName(HInstance, PChar(FileName), MAX_PATH));
    AssignFile(AFileOfByte, FileName);
    Reset(AFileOfByte);
    Names := nil;
    NameTbl := nil;
    Seek(AFileOfByte, FileSize(AFileOfByte) - SizeOf(AHeader));
    BlockRead(AFileOfByte, AHeader, SizeOf(AHeader));
    if (AHeader.Signature = 'FB09') or (AHeader.Signature = 'FB0A') then
        begin
            ifabase := FilePos(AFileOfByte) - AHeader.AnInteger;
            Seek(AFileOfByte, ifabase);
            BlockRead(AFileOfByte, AHeader, SizeOf(AHeader));
            if (AHeader.Signature = 'FB09') or (AHeader.Signature = 'FB0A') then
                begin
                    lfodir := ifabase + AHeader.AnInteger;
                    if lfodir >= ifabase then
                        begin
                            Seek(AFileOfByte, lfodir);
                            BlockRead(AFileOfByte, j, SizeOf(Word));
                            BlockRead(AFileOfByte, cbdirentry, SizeOf(Word));
                            BlockRead(AFileOfByte, cdir, SizeOf(Word));
                            Seek(AFileOfByte, lfodir + j);
                            SstFrameSize := cdir * cbdirentry;
                            getmem(SstFrame, SstFrameSize);
                            BlockRead(AFileOfByte, SstFrame^, SstFrameSize);
                            for k := 0 to pred(cdir) do
                                begin
                                    SstFrameElem := PArrayOfByte(@SstFrame^[k * cbdirentry]);
                                    ssttype := PWord(@SstFrameElem^[0])^;
                                    if (ssttype = $0130) then
                                        begin
                                            sstbase := ifabase + PInteger(@SstFrameElem^[4])^;
                                            sstsize := PInteger(@SstFrameElem^[8])^;
                                            getmem(Names, sstsize);
                                            Seek(AFileOfByte, sstbase);
                                            BlockRead(AFileOfByte, Names^, sstsize);
                                            y := PInteger(@Names^[0])^;
                                            getmem(NameTbl, sizeof(Pointer) * y);
                                            z := 4;
                                            for x := 0 to pred(y) do
                                                begin
                                                    NameTbl^[x] := PArrayOfByte(@Names^[z + 1]);
                                                    z := z + Names^[z] + 2;
                                                end;
                                        end;
                                end;
                            for k := 0 to pred(cdir) do
                                begin
                                    SstFrameElem := PArrayOfByte(@SstFrame^[k * cbdirentry]);
                                    ssttype := PWord(@SstFrameElem^[0])^;
                                    sstbase := ifabase + PInteger(@SstFrameElem^[4])^;
                                    sstsize := PInteger(@SstFrameElem^[8])^;
                                    getmem(sstbuf, sstsize);
                                    Seek(AFileOfByte, sstbase);
                                    BlockRead(AFileOfByte, sstbuf^, sstsize);
                                    if (ssttype = $0125) then
                                        dumpsymbols(NameTbl, PArrayOfByte(@sstbuf^[4]), sstsize - 4);
                                    if (ssttype = $0127) then
                                        dumplines(NameTbl, sstbuf, sstsize);
                                    FreeMem(sstbuf);
                                end;
                            FreeMem(Names);
                            FreeMem(NameTbl);
                            FreeMem(SstFrame);
                        end;
                end;
        end;
    CloseFile(AFileOfByte);
    FileMode := OldFileMode;
end;

var
  StackCheckInitialized: Boolean = False;

procedure InitializeOnce;
begin
  if not StackCheckInitialized then begin
    {$IF CompilerVersion >= 18.0}
      ReportMemoryLeaksOnShutdown := False;
    {$ELSE}
      {$IFDEF FASTMM4}
        ReportMemoryLeaks := False;
      {$ENDIF}
    {$IFEND}
    SetDispl;
    GetProjectInfos;
    StackCheckInitialized := True;
  end;
end;

constructor TUnitDebugInfos.Create(const AName: AnsiString; const NbLines: Cardinal);
begin
  Name := AName;
  SetLength(Addresses, NbLines);
end;

constructor TRoutineDebugInfos.Create(const AName: AnsiString; const AStartAddress: Cardinal; const ALength: Cardinal);
begin
  Name := AName;
  StartAddress := AStartAddress;
  EndAddress := StartAddress + ALength - 1;
end;

constructor TAddressToLine.Create(const AAddress, ALine: Cardinal);
begin
  Address := AAddress;
  Line := ALine
end;

function TUnitDebugInfos.LineWhichContainsAddress(const Address: Cardinal): string;
var
  Start, Finish, Pivot: Cardinal;
begin
    if Addresses[0].Address > Address then
        Result := ''
    else
        begin
            Start := 0;
            Finish := Length(Addresses) - 1;

            while Start < Finish - 1 do
                begin
                    Pivot := Start + (Finish - Start) div 2;

                    if Addresses[Pivot].Address = Address then
                        begin
                            Start := Pivot;
                            Finish := Start
                        end
                    else
                        if Addresses[Pivot].Address > Address then
                            Finish := Pivot
                        else
                            Start := Pivot
                end;

            Result := ' Line ' + IntToStr(Addresses[Start].Line);
        end;
end;

function UnitWhichContainsAddress(const Address: Cardinal): TUnitDebugInfos;
var
    Start, Finish, Pivot: integer;
begin
    Start := 0;
    Finish := UnitsCount - 1;
    Result := nil;

    while Start <= Finish do
        begin
            Pivot := Start + (Finish - Start) div 2;

            if Units[Pivot].Addresses[0].Address > Address then
                Finish := Pivot - 1
            else
                if Units[Pivot].Addresses[Length(Units[Pivot].Addresses) - 1].Address < Address then
                    Start := Pivot + 1
                else
                    begin
                        Result := Units[Pivot];
                        Start := Finish + 1;
                    end;
        end;
end;

function RoutineWhichContainsAddress(const Address: Cardinal): string;
var
  Start, Finish, Pivot: integer;
begin
  Start := 0;
  Result := NoDebugInfo;
  Finish := RoutinesCount - 1;
  while Start <= Finish do begin
    Pivot := Start + (Finish - Start) div 2;
    if Routines[Pivot].StartAddress > Address then
      Finish := Pivot - 1
    else
      if Routines[Pivot].EndAddress < Address then
        Start := Pivot + 1
      else begin
        Result := ' Routine ' + Routines[Pivot].Name;
        Exit;
      end;
  end;
  { Routines can overlap - as a quick fix, just look around the finishing point }
  Start := Max(0, Start-50);
  Finish := Min(RoutinesCount-1, Start+50);
  while Start <= Finish do begin
    if (Address >= Routines[Start].StartAddress) and (Address <= Routines[Start].EndAddress) then begin
      Result := ' Routine ' + Routines[Start].Name;
      Exit;
    end;
    Inc(Start);
  end;
end;

{$IFDEF WIN32}
function TextualDebugInfoForAddress(const TheAddress: NativeUInt): ShortString;
var
  U: TUnitDebugInfos;
  AddressInDebugInfos: Cardinal;
begin
  InitializeOnce;
  if UseDebugInfos and (TheAddress > Displ) then begin
    AddressInDebugInfos := TheAddress - Displ;
    U := UnitWhichContainsAddress(AddressInDebugInfos);
    if U <> nil then
      Result := 'Module ' + U.Name + RoutineWhichContainsAddress(AddressInDebugInfos) + U.LineWhichContainsAddress(AddressInDebugInfos)
    else
      Result := RoutineWhichContainsAddress(AddressInDebugInfos);
  end else
    Result := NoDebugInfo;
  Result := Result+'('+IntToHex(TheAddress,8)+' = '+IntToHex(AddressInDebugInfos,8)+')';
end;
{$ENDIF}

{$IFNDEF WIN32}
type
  QWord = UInt64;
  PQWord = PUInt64;

{ Taken from fpc seh64.inc }

const

  EXCEPTION_NONCONTINUABLE  = $01;
  EXCEPTION_UNWINDING       = $02;
  EXCEPTION_EXIT_UNWIND     = $04;
  EXCEPTION_STACK_INVALID   = $08;
  EXCEPTION_NESTED_CALL     = $10;
  EXCEPTION_TARGET_UNWIND   = $20;
  EXCEPTION_COLLIDED_UNWIND = $40;
  EXCEPTION_UNWIND          = $66;

  UNWIND_HISTORY_TABLE_SIZE = 12;

  UNW_FLAG_NHANDLER         = 0;

type

  PM128A=^M128A;
  M128A = record
    Low : QWord;
    High : Int64;
  end;

  PContext = ^TContext;
  TContext = record
    P1Home : QWord;
    P2Home : QWord;
    P3Home : QWord;
    P4Home : QWord;
    P5Home : QWord;
    P6Home : QWord;
    ContextFlags : DWord;
    MxCsr : DWord;
    SegCs : word;
    SegDs : word;
    SegEs : word;
    SegFs : word;
    SegGs : word;
    SegSs : word;
    EFlags : DWord;
    Dr0 : QWord;
    Dr1 : QWord;
    Dr2 : QWord;
    Dr3 : QWord;
    Dr6 : QWord;
    Dr7 : QWord;
    Rax : QWord;
    Rcx : QWord;
    Rdx : QWord;
    Rbx : QWord;
    Rsp : QWord;
    Rbp : QWord;
    Rsi : QWord;
    Rdi : QWord;
    R8 : QWord;
    R9 : QWord;
    R10 : QWord;
    R11 : QWord;
    R12 : QWord;
    R13 : QWord;
    R14 : QWord;
    R15 : QWord;
    Rip : QWord;
    Header : array[0..1] of M128A;
    Legacy : array[0..7] of M128A;
    Xmm0 : M128A;
    Xmm1 : M128A;
    Xmm2 : M128A;
    Xmm3 : M128A;
    Xmm4 : M128A;
    Xmm5 : M128A;
    Xmm6 : M128A;
    Xmm7 : M128A;
    Xmm8 : M128A;
    Xmm9 : M128A;
    Xmm10 : M128A;
    Xmm11 : M128A;
    Xmm12 : M128A;
    Xmm13 : M128A;
    Xmm14 : M128A;
    Xmm15 : M128A;
    VectorRegister : array[0..25] of M128A;
    VectorControl : QWord;
    DebugControl : QWord;
    LastBranchToRip : QWord;
    LastBranchFromRip : QWord;
    LastExceptionToRip : QWord;
    LastExceptionFromRip : QWord;
  end;

  { This is a simplified definition, only array part of unions }
  PKNONVOLATILE_CONTEXT_POINTERS=^KNONVOLATILE_CONTEXT_POINTERS;
  KNONVOLATILE_CONTEXT_POINTERS=record
    FloatingContext: array[0..15] of PM128A;
    IntegerContext: array[0..15] of PQWord;
  end;

  EXCEPTION_DISPOSITION=(
    ExceptionContinueExecution,
    ExceptionContinueSearch,
    ExceptionNestedException,
    ExceptionCollidedUnwind
  );

  PExceptionPointers = ^TExceptionPointers;
  TExceptionPointers = record
    ExceptionRecord   : PExceptionRecord;
    ContextRecord     : PContext;
  end;


  EXCEPTION_ROUTINE = function(
    var ExceptionRecord: TExceptionRecord;
    EstablisherFrame: Pointer;
    var ContextRecord: TContext;
    DispatcherContext: Pointer ): EXCEPTION_DISPOSITION;

  PRUNTIME_FUNCTION=^RUNTIME_FUNCTION;
  RUNTIME_FUNCTION=record
    BeginAddress: DWORD;
    EndAddress: DWORD;
    UnwindData: DWORD;
  end;

  UNWIND_HISTORY_TABLE_ENTRY=record
    ImageBase: QWord;
    FunctionEntry: PRUNTIME_FUNCTION;
  end;

  PUNWIND_HISTORY_TABLE=^UNWIND_HISTORY_TABLE;
  UNWIND_HISTORY_TABLE=record
    Count: DWORD;
    Search: Byte;
    RaiseStatusIndex: Byte;
    Unwind: Byte;
    Exception: Byte;
    LowAddress: QWord;
    HighAddress: QWord;
    Entry: array[0..UNWIND_HISTORY_TABLE_SIZE-1] of UNWIND_HISTORY_TABLE_ENTRY;
  end;

  PDispatcherContext = ^TDispatcherContext;
  TDispatcherContext = record
    ControlPc: QWord;
    ImageBase: QWord;
    FunctionEntry: PRUNTIME_FUNCTION;
    EstablisherFrame: QWord;
    TargetIp: QWord;
    ContextRecord: PContext;
    LanguageHandler: EXCEPTION_ROUTINE;
    HandlerData: Pointer;
    HistoryTable: PUNWIND_HISTORY_TABLE;
    ScopeIndex: DWord;
    Fill0: DWord;
  end;

procedure RtlCaptureContext(var ctx: TContext); stdcall;
  external 'kernel32.dll' name 'RtlCaptureContext';

function RtlCaptureStackBackTrace(
  FramesToSkip: DWORD;
  FramesToCapture: DWORD;
  var BackTrace: Pointer;
  BackTraceHash: PDWORD): Word; stdcall;
  external 'kernel32.dll' name 'RtlCaptureStackBackTrace';

function RtlLookupFunctionEntry(
  ControlPC: QWord;
  out ImageBase: QWord;
  HistoryTable: PUNWIND_HISTORY_TABLE): PRUNTIME_FUNCTION;
  external 'kernel32.dll' name 'RtlLookupFunctionEntry';

function RtlVirtualUnwind(
  HandlerType: DWORD;
  ImageBase: QWord;
  ControlPc: QWord;
  FunctionEntry: PRUNTIME_FUNCTION;
  var ContextRecord: TContext;
  HandlerData: PPointer;
  EstablisherFrame: PQWord;
  ContextPointers: PKNONVOLATILE_CONTEXT_POINTERS): EXCEPTION_ROUTINE;
  external 'kernel32.dll' name 'RtlVirtualUnwind';

procedure RtlUnwindEx(
  TargetFrame: Pointer;
  TargetIp: Pointer;
  ExceptionRecord: PExceptionRecord;
  ReturnValue: Pointer;
  OriginalContext: PContext;  { scratch space, initial contents ignored }
  HistoryTable: PUNWIND_HISTORY_TABLE);
  external 'kernel32.dll' name 'RtlUnwindEx';

procedure RaiseException(
  dwExceptionCode: DWORD;
  dwExceptionFlags: DWORD;
  dwArgCount: DWORD;
  lpArguments: Pointer);  // msdn: *ULONG_PTR
  external 'kernel32.dll' name 'RaiseException';


function TextualDebugInfoForAddress(const TheAddress: NativeUInt): ShortString;
var
  AddressInDebugInfos: Cardinal;
  U: TUnitDebugInfos;
begin
  InitializeOnce;
  if (TheAddress > Displ) and (TheAddress - Displ < High(Cardinal)) then begin
    AddressInDebugInfos := Cardinal(TheAddress - Displ);
    U := UnitWhichContainsAddress(AddressInDebugInfos);
    if U <> nil then
        Result := 'Module ' + U.Name + RoutineWhichContainsAddress(AddressInDebugInfos) + U.LineWhichContainsAddress(AddressInDebugInfos)
    else
        Result := RoutineWhichContainsAddress(AddressInDebugInfos);
  end else
    Result := '';
  Result := Result + '(' + IntToHex(TheAddress,16) + ' = ' + IntToHex(AddressInDebugInfos,8) + ')';
end;

procedure FillCallStack(var St: TCallStack; const NbLevelsToExclude: integer; TopAddress: Pointer);
var
  Context: TContext;
  UnwindHistory: UNWIND_HISTORY_TABLE;
  RuntimeFunction: PRUNTIME_FUNCTION;
  ImageBase: QWord;
  HandlerData: Pointer;
  EstablisherFrame: QWord;
  Count: integer;
begin
  FillChar(St, SizeOf(St), 0);
  RtlCaptureContext(Context);
  FillChar(UnwindHistory,sizeof(UNWIND_HISTORY_TABLE),0);
  UnwindHistory.Unwind:=1;
  Count := -NbLevelsToExclude;
  if TopAddress <> nil then begin
    St[0] := QWord(TopAddress);
    Count := 1;
  end else
    Count:= -NbLevelsToExclude;
  repeat
    RuntimeFunction := RtlLookupFunctionEntry(Context.Rip, ImageBase, @UnwindHistory);
    if Assigned(RuntimeFunction) then
      RtlVirtualUnwind(UNW_FLAG_NHANDLER, ImageBase, Context.Rip, RuntimeFunction, Context, @HandlerData, @EstablisherFrame, nil)
    else begin { a leaf function }
      Context.Rip := PQWord(Context.Rsp)^;
      Inc(Context.Rsp, sizeof(Pointer));
    end;
    if Context.Rip = QWord(TopAddress) then
      Count := 0;
    if Count >= 0 then
      St[Count] := Context.Rip;
    Inc(Count);
  until (Context.Rip = 0) or (Count > High(St));
end;

{$ENDIF}

{$IFDEF FASTMM4OFF}
procedure GetMM4DebugInfo(AReturnAddress: Pointer; APDebugInfo: PAnsiChar; var ANumChars: integer);
var
  s: ShortString;
begin
  s := TextualDebugInfoForAddress(Cardinal(AReturnAddress) - 1);
  System.Move(s[1], APDebugInfo^, Length(s));
  ANumChars := Length(s);
end;

initialization
  DebugInfoProc := GetMM4DebugInfo;
{$ENDIF}
end.
