unit TempStrm;
{$I utils.inc}{ Copyright (C) 2005 C M Western; see UCompat.pas for license}
interface

uses Classes;

{ This stream provides a safe way of saving a file, leaving a back up copy
  and not loosing the original on disk full or other error. To use, create
  with the desired filename, write to the stream (which writes to a temporary file
  on disk) and then call Save (which does some file renaming). The destructor will
  discard the temporary file if Save has not been called.
  Typical usage:
    Steam := TBackupStream.Create(FileName);
    try
      Text.SaveToStream(Steam);
      Steam.Save;
    finally
      Steam.Free;
      Raise;
    end;
}

type
  TBackupStream = class (THandleStream)
  private
    FTmpName: string;
    FFileName: string;
    {$IFDEF FPC}
    HandleClosed: Boolean;
    {$ENDIF}
  public
    constructor Create(const AFileName: string);
    destructor Destroy; override;
    procedure Save;
    property TmpName: string read FTmpName;
    property FileName: string read FFileName;
  end;

function TempStreamFor(Name: string; var TmpName: string): THandleStream;
function TmpFileHandleFor(Name: string; var TmpName: string): Integer;

implementation

{$IFDEF UNIX}
uses BaseUnix, SysUtils;
const
  INVALID_HANDLE_VALUE = THandle(-1);

{$LINKLIB libc}
function mkstemp(Template: PChar):Integer; cdecl; external;

function TmpFileHandleFor(Name: string; var TmpName: string): Integer;
begin
  TmpName := Concat(ExpandFilename(Name), 'XXXXXX');
  UniqueString(TmpName);
  Result := mkstemp(PChar(TmpName));
  if Result = -1 then
     RaiseLastOSError;
end;
{$ENDIF}

{$IFDEF MSWINDOWS}
uses Windows, SysUtils;
{$IFNDEF FPC}
  {$IF CompilerVersion < 23.0}
    { Cardinal <> Integer range check issue }
    const INVALID_HANDLE_VALUE = Integer(Windows.INVALID_HANDLE_VALUE);
    const INVALID_FILE_ATTRIBUTES = DWORD($FFFFFFFF);
  {$IFEND}
{$ELSE}
    const INVALID_FILE_ATTRIBUTES = DWORD($FFFFFFFF);
{$ENDIF}


function TmpFileHandleFor(Name: string; var TmpName: string): Integer;
var
  Buffer: array [0..MAX_PATH+1] of char;
begin
  {$IFNDEF offLCL_}
  TmpName := ExtractFilePath(ExpandFilename(Name));
  if Windows.GetTempFileName(PChar(TmpName), 'txt', 0, Buffer) = 0 then
    raise Exception.CreateFmt('Can''t create temporary file in %s', [TmpName]);
  TmpName := Buffer;
  {$ELSE}
  TmpName := GetTempFileName(ExtractFilePath(ExpandFilename(Name)), 'txt');
  {$ENDIF}
  Result := FileCreate(TmpName);
  if Result < 0 then
    RaiseLastOSError;
end;
{$ENDIF}

function TempStreamFor(Name: string; var TmpName: string): THandleStream;
begin
  Result := THandleStream.Create(TmpFileHandleFor(Name,TmpName));
end;

{ TBackupStream }

procedure TBackupStream.Save;
var
  BackFile: string;
begin
  if (Handle <> INVALID_HANDLE_VALUE) {$IFDEF FPC} and not HandleCLosed {$ENDIF} then begin
    FileClose(Handle);
    {$IFDEF FPC}
      HandleClosed := True;
    {$ELSE}
      FHandle := INVALID_HANDLE_VALUE;
    {$ENDIF}
    BackFile := Concat(FileName, '.bak');
    if FileExists(BackFile) then
      if not DeleteFile(BackFile) then
        raise Exception.CreateFmt('Can''t delete old backup file %s', [BackFile]);
    if FileExists(FileName) and not RenameFile(FileName, BackFile) then
        raise Exception.CreateFmt('Can''t rename old file (%s) to %s', [FileName, BackFile]);
    if not RenameFile(TmpName, FileName) then
      raise Exception.CreateFmt('Can''t rename new file (%s) to %s', [TmpName, FileName]);
  end;
end;

constructor TBackupStream.Create(const AFileName: string);
var
  {$IFDEF MSWINDOWS}
  Attr: DWORD;
  {$ENDIF}
  {$IFDEF UNIX}
  Info: BaseUnix.Stat;
  {$ENDIF}
begin
  {$IFDEF FPC}
    HandleClosed := False;
  {$ELSE}
    FHandle := INVALID_HANDLE_VALUE;
  {$ENDIF}
  {$IFDEF UNIX}
  Info.st_mode := 0;
  if (FpStat(AFileName, Info) >= 0) and { (Libc.Access(PChar(AFileName),W_OK)<> 0) } FileIsReadOnly(AFileName) then
      raise Exception.Create(AFileName + ' is read only');
  {$ENDIF}
  {$IFDEF MSWINDOWS}
  Attr := GetFileAttributes(PChar(AFileName));
  if (Attr <> INVALID_FILE_ATTRIBUTES) and ((Attr and FILE_ATTRIBUTE_READONLY) <> 0) then
    raise Exception.Create(AFileName + ' is read only');
  {$ENDIF}
  inherited Create(TmpFileHandleFor(AFileName, FTmpName));
  {$IFDEF UNIX}
  if (Info.st_mode <> 0) and (FpChmod(FTmpName, Info.st_mode) < 0) then
    RaiseLastOSError;
  {$ENDIF}
  FFileName := AFileName;
end;

destructor TBackupStream.Destroy;
begin
  if (Handle <> INVALID_HANDLE_VALUE) {$IFDEF FPC} and not HandleClosed {$ENDIF} then begin
    FileClose(Handle);
    {$IFDEF FPC}
      HandleClosed := True;
    {$ELSE}
      FHandle := INVALID_HANDLE_VALUE;
    {$ENDIF}
    DeleteFile(TmpName);
  end;
  inherited Destroy;
end;

end.


