unit TextAssign;
{ Connect TextFile output to a Stream or TStrings object }
{$I utils.inc}{ Copyright (C) 2005 C M Western; see UCompat.pas for license}
interface

uses Classes, SysUtils;

procedure AssignStrings(var F: TextFile; Strings: TStrings);
procedure AssignStream(var F: TextFile; Stream: TStream; PlatformDefaultCodePage: Boolean = True);

{ Note that each thread must have it's own TextFile; Write and friends to same
  file from different threads can easily fail }
procedure AssignTextFile(var F: TextFile; var Source:TextFile);

type
  PTextRec = ^TTextRec;
  PPTextRec = ^PTextRec;

implementation

uses {$IFDEF MSWINDOWS} Windows, {$ELSE} Unix, {$ENDIF} { For EnterCriticalSection et al }
     UCompat, Math;

{ Note that no error checking is done - all the actions are assumed to succeed }

function NoAction(Var t:TTextRec):Integer;
begin
  Result := 0;
end;

procedure ResetTextFile(var F: TextFile);
begin
  FillChar(F,SizeOf(TTextRec),0);
  with TTextRec(F) do begin
    Handle := {$IFDEF FPC} UnusedHandle {$ELSE} -1 {$ENDIF};
    Mode := fmClosed;
    OpenFunc := @NoAction;
    CloseFunc := @NoAction;
    InOutFunc := @NoAction;
    FlushFunc := @NoAction;
    BufSize := SizeOf(Buffer);
    BufPtr := @Buffer[0];
    { Force platform default line ending }
    {$IFDEF FPC}
      LineEnd := LineEnding;
    {$ELSE}
      Flags := tfCRLF * Byte(DefaultTextLineBreakStyle);
    {$ENDIF}
  end;
end;

type
  PStream = ^TStream;

function StreamReadFunc(var T:TTextRec):Integer;
begin
  T.BufEnd := PStream(@T.UserData)^.Write(T.BufPtr^, T.BufSize);
  T.BufPos := 0;
  Result := 0;
end;

function StreamWriteFunc(var T:TTextRec):Integer;
begin
  T.BufPos := T.BufPos - PStream(@T.UserData)^.Write(T.BufPtr^, T.BufPos);
  Result := 0;
end;

function StreamOpenFunc(var T:TTextRec):Integer;
begin
  with T do begin
    if Mode = fmInput then
      InOutFunc := @StreamReadFunc
    else begin
      InOutFunc := @StreamWriteFunc;
      FlushFunc := InOutFunc;
      if Mode = fmOutput then
        PStream(@T.UserData)^.Position := 0
      else { Mode = fmInOut }
        Mode := fmOutput;
    end;
  end;
  Result := 0;
end;

procedure AssignStream(var F: TextFile; Stream: TStream; PlatformDefaultCodePage: Boolean);
begin
  ResetTextFile(F);
  with TTextRec(F) do begin
    PStream(@UserData)^ := Stream;
    {$IFNDEF FPC}
    Flags := tfCRLF * Byte(DefaultTextLineBreakStyle);
    {$ENDIF}
    Name := 'Stream';
    OpenFunc := @StreamOpenFunc;
    {$IFDEF UNICODE}
      if PlatformDefaultCodePage then
        CodePage := DefaultUnicodeCodePage
      else
        CodePage := CP_UTF8;
    {$ENDIF}
  end;
end;

type
  TTextFileWriteFunc = function (var T:TTextRec; Start: PChar; Count:Integer;
                                 TwoCharNewLine: Boolean {$IFDEF UNICODE}; CodePage: Word{$ENDIF}):Integer;

  TextFileData = record
    CurrentLine: AnsiString; { Used as buffer if built in buffer too small }
    ScanPos: Integer; { Current position in scan for end of line }
    DoWriteFunc: TTextFileWriteFunc;
    ExtraData: Pointer;
  end;
  PTextFileData = ^TextFileData;

var
  TextFileFlushFuncLock: TRTLCriticalSection;

function TextFileWrite(var T:TTextRec; Start: PChar; Count:Integer; TwoCharNewLine: Boolean{$IFDEF UNICODE}; CodePage: Word{$ENDIF}):Integer;
var
  PT: PTextRec;
  OldPtr: PAnsiChar;
  OldPos: Integer;
begin
  if Count > 0 then begin
    {$IFDEF UNICODE}
      if CodePage <> T.CodePage then
        raise Exception.Create('TextFileWrite:'+IntToStr(CodePage));
    {$ENDIF}
    try
      EnterCriticalSection(TextFileFlushFuncLock);
      with T, PTextFileData(@UserData)^ do begin
        PT := ExtraData;
        OldPtr := Pointer(PT^.BufPtr);
        OldPos := PT^.BufPos;
        PT^.BufPos := Count;
        PT^.BufPtr := Pointer(Start);
        Result := TTextIOFunc(PT^.inoutfunc)(PT^);
        PT^.BufPtr := Pointer(OldPtr);
        PT^.BufPos := OldPos;
      end;
    finally
      LeaveCriticalSection(TextFileFlushFuncLock);
    end
  end else
    Result := 0;
end;

{$IFDEF UNICODE}
function AnsiTextFileWriteFunc(var T:TTextRec):Integer; forward;

function UnicodeTextFileWriteFunc(var T:TTextRec):Integer;
var
  s, p, e: PChar;
  NewLine: Boolean;
begin
  if T.CodePage <> DefaultUnicodeCodePage then begin
    T.InOutFunc := @AnsiTextFileWriteFunc;
    if T.Mode <> fmInput then
      T.FlushFunc := T.InOutFunc;
    Result := AnsiTextFileWriteFunc(T);
    Exit;
  end;
  Result := 0;
  T.UTF16Buffer[0] := #0; { DELPHI bug work-around }
  with T, PTextFileData(@UserData)^ do begin
    { Is there anything to do? }
    if BufPos > 0 then begin
      s := Pointer(BufPtr);
      p := s+Min(ScanPos, BufPos div SizeOf(Char));
      e := s + BufPos div SizeOf(Char);
      NewLine := False;
      while p < e do begin
        NewLine := p^ in [#10,#13];
        if NewLine and not ((p+1=e)
                             and {$IFDEF FPC} (Length(LineEnd) > 1) {$ELSE} ((t.flags and tfCRLF) <> 0) {$ENDIF}) then begin
          { Purge output - elaborate test is to allow for two character end of line marker to fill buffer }
          { Test for two character newline }
          NewLine := (p+1 < e) and ((p+1)^ in [#10,#13]) and ((p+1)^ <> p^);
          if NewLine then
            Inc(p); { Include other newline character }
          Result := DoWriteFunc(T,Pointer(s),p-s+1,Newline,T.CodePage);
          s := p + 1;
          p := s;
          NewLine := False; { Flag not in split newline sequence }
          Continue;
        end else
          Inc(p);
      end;
      { Slide any remaining characters to the start of the buffer }
      BufPos := (p-s)*SizeOf(Char);
      if (BufPos > 0) and (s <> Pointer(BufPtr)) then
        Move(s^, BufPtr^, (p-s)*SizeOf(Char));
      ScanPos := p-s;
      if NewLine then
        Dec(ScanPos); { first part of newline fragment }
    end;
    if BufPos = BufSize then begin
      { Make sure there is some space in the buffer }
      if BufPtr <> PAnsiChar(CurrentLine) then
        { Copy string to new buffer if necessary }
        SetString(CurrentLine, BufPtr, BufSize);
      SetLength(CurrentLine, BufSize+256);
      BufPtr := PAnsiChar(CurrentLine);
      BufSize := Length(CurrentLine);
    end;
  end;
end;
{$ENDIF}

function AnsiTextFileWriteFunc(var T:TTextRec):Integer;
var
  s, p, e: PAnsiChar;
  NewLine: Boolean;
begin
  {$IFDEF UNICODE}
  if T.CodePage = DefaultUnicodeCodePage then begin
    T.InOutFunc := @UnicodeTextFileWriteFunc;
    if T.Mode <> fmInput then
      T.FlushFunc := T.InOutFunc;
    Result := UnicodeTextFileWriteFunc(T);
    Exit;
  end;
  {$ENDIF}
  Result := 0;
  with T, PTextFileData(@UserData)^ do begin
    { Is there anything to do? }
    if BufPos > 0 then begin
      s := Pointer(BufPtr);
      p := s+Min(ScanPos,BufPos);
      e := s + BufPos;
      NewLine := False;
      while p < e do begin
        NewLine := p^ in [#10,#13];
        if NewLine and not ((p+1=e)
                             and {$IFDEF FPC} (Length(LineEnd) > 1) {$ELSE} ((t.flags and tfCRLF) <> 0) {$ENDIF}) then begin
          { Purge output - elaborate test is to allow for two character end of line marker to fill buffer }
          { Test for two character newline }
          NewLine := (p+1 < e) and ((p+1)^ in [#10,#13]) and ((p+1)^ <> p^);
          if NewLine then
            Inc(p); { Include other newline character }
          Result := DoWriteFunc(T,Pointer(s),p-s+1,Newline {$IFDEF UNICODE}, T.CodePage{$ENDIF});
          s := p + 1;
          p := s;
          NewLine := False; { Flag not in split newline sequence }
          Continue;
        end else
          Inc(p);
      end;
      { Slide any remaining characters to the start of the buffer }
      BufPos := p-s;
      if (BufPos > 0) and (Pointer(s) <> Pointer(BufPtr)) then
        Move(s^, BufPtr^, p-s);
      ScanPos := BufPos;
      if NewLine then
        Dec(ScanPos); { first part of newline fragment }
    end;
    if BufPos = BufSize then begin
      { Make sure there is some space in the buffer }
      if Pointer(BufPtr) <> PAnsiChar(CurrentLine) then
        { Copy string to new buffer if necessary }
        SetString(CurrentLine, PAnsiChar(BufPtr), BufSize);
      SetLength(CurrentLine, BufSize+256);
      BufPtr := Pointer(PAnsiChar(CurrentLine));
      BufSize := Length(CurrentLine);
    end;
  end;
end;

function TextFileCloseFunc(var T:TTextRec):Integer;
begin
  if (T.Mode <> fmInput) and (T.BufPos > 0) then
    Writeln(Text(T)); { Write any remaining characters in buffer }
  SetLength(PTextFileData(@T.UserData)^.CurrentLine,0);
  T.Mode := fmClosed;
  Result := 0;
end;

{ Connect text file to a string - Thread safe on write only at the moment }

type
  PTextBuf = {$IFDEF FPC} ^TextBuf {$ELSE} PChar {$ENDIF};
  
function StringsReadFunc(var T:TTextRec):Integer;
var
  Strings: TStrings;
begin
  with T, PTextFileData(@UserData)^ do begin
    Strings := TStrings(ExtraData);
    if ScanPos >= Strings.Count then
      BufEnd := 0
    else begin
      CurrentLine := Strings[ScanPos] + LineEnd;
      {$IFDEF UNICODE}
      BufPtr := Pointer(CurrentLine);
      {$ELSE}
      BufPtr := PTextBuf(CurrentLine);
      {$ENDIF}
      BufEnd := Length(CurrentLine);
      Inc(ScanPos);
    end;
    BufPos := 0;
  end;
  Result := 0;
end;

{ TStringsWriter - helper class }

type
  TStringsWriter = class
    Str: string;
    Strings: TStrings;
    ThreadID: TThreadID;
    procedure DoStringWrite;
    constructor Create(AStrings: TStrings);
  end;

constructor TStringsWriter.Create(AStrings: TStrings);
begin
  inherited Create;
  Strings := AStrings;
  ThreadID := GetCurrentThreadId;
end;

procedure TStringsWriter.DoStringWrite;
begin
  Strings.Add(Str);
end;

function StringsWrite(var T:TTextRec; Start: PChar; Count:Integer; TwoCharNewLine: Boolean {$IFDEF UNICODE}; CodePage: Word{$ENDIF}):Integer;
var
  W: TStringsWriter;
begin
  if Count > 0 then begin
    {$IFDEF UNICODE}
      if CodePage <> T.CodePage then
        raise Exception.Create('StringsWrite:'+IntToStr(CodePage));
    {$ENDIF}
    W := TStringsWriter(PTextFileData(@T.UserData)^.ExtraData);
    { Discard any newline }
    if TwoCharNewLine then
      Dec(Count);
    SetString(W.Str, Start, Count-1);
    if W.ThreadID <> GetCurrentThreadId then
      raise Exception.Create('Multiple use of StringsWrite');
    if GetCurrentThreadID = MainThreadID then
      W.DoStringWrite
    else
      TThread.Synchronize(nil,{$IFDEF FPC}@{$ENDIF}W.DoStringWrite);
  end;
  Result := 0;
end;

function StringsWriteCloseFunc(var T:TTextRec):Integer;
var
  W: TStringsWriter;
begin
  with T, PTextFileData(@T.UserData)^ do begin
    if BufPos > 0 then
      Writeln(Text(T)); { Write any remaining characters in buffer }
    SetLength(CurrentLine,0);
    W := TStringsWriter(ExtraData);
    ExtraData := W.Strings;
    W.Free;
    Mode := fmClosed;
  end;
  Result := 0;
end;

function StringsOpenFunc(var T:TTextRec):Integer;
begin
  with T, PTextFileData(@UserData)^ do begin
    if Mode = fmInput then begin
      InOutFunc := @StringsReadFunc;
      ScanPos := 0;
      CloseFunc := @TextFileCloseFunc; { Needed to free line buffer }
    end else begin
      InOutFunc := @AnsiTextFileWriteFunc;
      FlushFunc := InOutFunc;
      CloseFunc := @StringsWriteCloseFunc;
      Mode := fmOutput; { fmInOut not implemented }
      DoWriteFunc := {$IFDEF FPC}@{$ENDIF}StringsWrite;
      ExtraData := TStringsWriter.Create(TStrings(ExtraData));
    end;
  end;
  Result := 0;
end;

procedure AssignStrings(var F: TextFile; Strings: TStrings);
begin
  ResetTextFile(F);
  with TTextRec(F) do begin
    PTextFileData(@UserData)^.ExtraData := Strings;
    {$IFDEF UNICODE}
      CodePage := DefaultUnicodeCodePage; // CP_UTF16
    {$ENDIF}
    Name := 'TStrings';
    OpenFunc := @StringsOpenFunc;
  end;
end;

procedure AssignTextFile(var F: TextFile; var Source:TextFile);
var
  Obj: TObject;
begin
  ResetTextFile(F);
  if TTextRec(Source).OpenFunc = Pointer(@StringsOpenFunc) then begin
    { As a convenience, if applied to a TextFile diverted to TStrings,
      switch directly to strings to avoid problems with redundant locks }
    Obj := TObject(PTextFileData(@TTextRec(Source).UserData)^.ExtraData);
    if Obj is TStringsWriter then
      AssignStrings(F,TStringsWriter(Obj).Strings)
    else
      AssignStrings(F,Obj as TStrings)
  end else
    with TTextRec(F), PTextFileData(@UserData)^ do begin
      Mode := TTextRec(Source).Mode;
      Name := TTextRec(Source).Name;
      {$IFDEF UNICODE}
        CodePage := TTextRec(Source).CodePage;
      {$ENDIF}
      {$IFNDEF FPC}
      Flags := TTextRec(Source).Flags;
      {$ELSE}
      LineEnd := TTextRec(Source).LineEnd;
      {$ENDIF}
      DoWriteFunc := {$IFDEF FPC}@{$ENDIF}TextFileWrite;
      ExtraData := @TTextRec(Source);
      if Mode <> fmInput then begin
        { Output only implemented }
        InOutFunc := @AnsiTextFileWriteFunc;
        FlushFunc := InOutFunc;
        CloseFunc := @TextFileCloseFunc;
      end;
    end;
end;

initialization
  InitCriticalSection(TextFileFlushFuncLock);
finalization
  DoneCriticalSection(TextFileFlushFuncLock);
end.
