unit Handles;
{ C M Western: Modified from original (see below) mainly for Kylix and FPC
  compatability }

{ TStretchHandles is a transparent control to implement runtime grab handles
  for Forms Designer-like projects.  It paints the handles on its own canvas,
  maintains a list of the controls it is supposed to manage, and traps mouse
  and keyboard events to move/resize itself and its child controls.  See the
  accompanying README file for more information.

  Distributed by the author as freeware, please do not sell.

  Anthony Scott
  CIS: 75567,3547

  OnMoved event added by John Biddiscombe Jan 97
  needed to know when control was moved but not resized.
}

interface

{$I utils.inc}

uses
  {$IFDEF MSWINDOWS} Windows, {$ENDIF} {$IFDEF LCL_} LCLType, {$ENDIF}
  {$IFNDEF CLX_}  Messages, Graphics, Controls, Forms, Menus, StdCtrls, Dialogs,
  {$ELSE} QControls, Types, QGraphics, QForms, Qt, {$ENDIF}
  Classes, SysUtils;

{$IFDEF LCLcarbon}
{$DEFINE SETPARENT}
{$ENDIF}
                                       { miscellaneous type declarations }
type
  TDragStyle = (dsMove, dsSizeTopLeft, dsSizeTopRight, dsSizeBottomLeft, dsSizeBottomRight,
                dsSizeTop, dsSizeLeft, dsSizeBottom, dsSizeRight);
  TForwardMessage = (fmMouseDown, fmMouseUp);
  GridValues = 1..32;
  EBadChild = class(Exception);
                                       { TStretchHandle component declaration }
type
  TStretchHandle = class(TCustomControl)
  private
    FDragOffset: TPoint;
    FDragStyle: TDragStyle;
    FDragging: boolean;
    FDragRect: TRect;
    FLocked: boolean;
    FPrimaryColor: TColor;
    FSecondaryColor: TColor;
    FGridX, FGridY: GridValues;
    FChildList: TList;
    FOnMoved : TNotifyEvent;    { Added by JAB }
    FOnResized : TNotifyEvent;
    FAspectRatio: Double;
    FMouseDownAspectRatio: Double; { Aspect ratio before changing size } 
    {$IFDEF MSWINDOWS}
    procedure ForwardMessage(FwdMsg: TForwardMessage; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    {$ELSE}
      {$IFDEF XBASED}
      FMask: TPixmap; { Control shape }
      FFreezeCount: Integer;
      FStaleMask: boolean;
      procedure FreezeMask;
      procedure ThawMask;
      procedure CheckMask;
      {$ENDIF}
    {$ENDIF}
    {$IFNDEF CLX_}
    procedure WMEraseBkgnd(var Message: TWMEraseBkgnd); message WM_ERASEBKGND;
    procedure WMGetDLGCode(var Message: TMessage); message WM_GETDLGCODE;
    {$ENDIF}
    procedure Rubberband(XPos, YPos: Integer; ShowBox: boolean);
    procedure SetPrimaryColor(AColor: TColor);
    procedure SetSecondaryColor(AColor: TColor);
    procedure SetGridState(Value: boolean);
    function GetGridState: boolean;
    function GetChildCount: Integer;
    function GetChildControl(idx: Integer): TControl;
    function GetModifiedRect(XPos, YPos: Integer; DragStyle: TDragStyle): TRect;
    function PointOverChild(P: TPoint): boolean;
    function XGridAdjust(X: integer): Integer;
    function YGridAdjust(Y: integer): Integer;
    function IsAttached: boolean;
    function FixWidth(AWidth, AHeight: Integer; AAspectRatio: Double):Integer;
    function FixHeight(AWidth, AHeight: Integer; AAspectRatio: Double):Integer;
    procedure SaveAspectRatio;
  protected
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure KeyDown(var key: Word; Shift: TShiftState); override;
    {$IFDEF MSWINDOWS}
    procedure CreateParams(var Params: TCreateParams); override;
    {$ENDIF}
    {$IFDEF CLX_}
    procedure InitWidget; override;
    function WidgetFlags: Integer; override;
    procedure DrawMask(ACanvas: TCanvas); override;
    procedure AlignControls(AControl: TControl; var Rect: TRect); override;
    function EventFilter(Sender: QObjectH; Event: QEventH): Boolean; override;
    {$ENDIF}
    {$IFDEF XBASED}
    procedure Resize; override;
    {$ENDIF}
    procedure Paint; override;
    procedure DrawHandles(ACanvas: TCanvas);
    property Canvas;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Attach(ChildControl: TControl);
    procedure Detach;
    procedure ReleaseChild(ChildControl: TControl);
    procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); override;
    procedure BringToFront;
    procedure SendToBack;
    procedure SetColors(Color1, Color2: TColor);
    function IndexOf(ChildControl: TControl): integer;
                                       { new run-time only properties }
    property Attached: boolean read IsAttached;
    property ChildCount: integer read GetChildCount;
    property Children[idx: integer]: TControl read GetChildControl;
  published
                                       { new properties }
    property Color: TColor read FPrimaryColor write SetPrimaryColor default clBlack;
    property SecondaryColor: TColor read FSecondaryColor write SetSecondaryColor default clGray;
    property Locked: boolean read FLocked write FLocked default False;
    property GridX: GridValues read FGridX write FGridX default 8;
    property GridY: GridValues read FGridY write FGridY default 8;
    property SnapToGrid: boolean read GetGridState write SetGridState default False;
                                       { inherited properties }
    {$IFDEF MSWINDOWS}
    property DragCursor;
    {$ENDIF}
    property Enabled;
    property Hint;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property Visible;
                                       { defined events }
    property OnClick;
    property OnDblClick;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnKeyDown;
    property OnKeyUp;
    property OnKeyPress;
    property OnMoved : TNotifyEvent read FOnMoved write FOnMoved;
    property OnResized : TNotifyEvent read FOnResized write FOnResized;
    { Width / Height - only enforced on mouse movements }
    property AspectRatio: Double read FAspectRatio write FAspectRatio;
  end;

{$IFDEF SINGLE}
procedure Register;
{$ENDIF}

implementation

uses {$IFDEF LCL_} LCLIntf,
       {$IFDEF LCLgtk1} gtk, gdk, GTKProc, GTKDef, {$ENDIF}
       {$IFDEF LCLgtk2} gtk2, gdk2, Gtk2Def, Gtk2Proc, {$ENDIF}
       {$IFDEF LCLgtk3}  {$ENDIF}
     {$ENDIF}
     UCompat;

{$IFDEF SINGLE}
procedure Register;
begin
  RegisterComponents('CMW', [TStretchHandle]);
end;

{$IFDEF VER80}
{$R Handles.d16}
{$ELSE}
{$R Handles.d32}
{$ENDIF}
{$ENDIF}

{ TODO : Mouse drag of handles (When mouse grab works) }
{$IFDEF CLX_}
procedure TStretchHandle.DrawMask(ACanvas: TCanvas);
begin
  if csDesigning in ComponentState then begin
    ACanvas.Brush.Color := clDontMask;
    ACanvas.FillRect(ClientRect);
  end else begin
    ACanvas.Brush.Color := clDontMask;
    DrawHandles(Canvas);
  end;
end;

procedure TStretchHandle.InitWidget;
begin
  inherited InitWidget;
  Masked := True;
end;

function TStretchHandle.WidgetFlags: Integer;
begin
  Result := (inherited WidgetFlags) or Integer(WidgetFlags_WMouseNoMask);
end;

procedure TStretchHandle.AlignControls(AControl: TControl; var Rect: TRect);
begin
  inherited AlignControls(AControl, Rect);
  MaskChanged;
end;


function TStretchHandle.EventFilter(Sender: QObjectH; Event: QEventH): Boolean;
begin
  Result := False;
  if Sender <> Handle then
    case QEvent_type(Event) of
      QEventType_MouseButtonRelease:
        if FDragging and (QMouseEvent_button(QMouseEventH(Event)) = ButtonState_LeftButton) then begin
          inherited EventFilter(Sender, Event);
          Result := True;
        end;
      QEventType_MouseButtonPress:
        if not FLocked and (QMouseEvent_button(QMouseEventH(Event)) = ButtonState_LeftButton) then begin
          inherited EventFilter(Sender, Event);
          Result := True;
        end;
      QEventType_MouseMove:
        begin
          inherited EventFilter(Sender, Event);
          Result := FDragging or (FDragStyle <> dsMove);
        end;
    end
  else
    Result := inherited EventFilter(Sender, Event);
end;
{$ENDIF}

function MinInt(a, b: integer): integer;
begin
  if a < b then
    Result := a
  else
    Result := b;
end;

function MaxInt(a, b: integer): integer;
begin
  if a > b then
    Result := a
  else
    Result := b;
end;

constructor TStretchHandle.Create(AOwner: TComponent);
begin

  inherited Create(AOwner);
                                       { create storage for child objects }
  FChildList := TList.Create;
                                       { initialize default properties }
  Width := 24;
  Height := 24;
  FPrimaryColor := clBlack;
  FSecondaryColor := clGray;
                                       { a value of 1 is used to effectively disable the snap-to grid }
  FGridX := 1;
  FGridY := 1;
                                       { doesn't do anything until it is Attached to something else }
  Enabled := False;
  Visible := False;

  AspectRatio := 0;
  ControlStyle := ControlStyle - [csOpaque] + [csCaptureMouse];
  {$IFDEF CLX_}
  InputKeys := [ikArrows];
  {$ENDIF}
  {$IFDEF XBASED}
  FMask := nil;
  FFreezeCount := 0;
  FStaleMask := False;
  {$ENDIF}
end;

destructor TStretchHandle.Destroy;
begin
                                       { tidy up carefully }
  FChildList.Free;
  {$IFDEF XBASED}
  FreeAndNil(FMask);
  {$ENDIF}
  inherited Destroy;

end;


{$IFDEF MSWINDOWS}
procedure TStretchHandle.CreateParams(var Params: TCreateParams);
begin
                                       { set default Params values }
  inherited CreateParams(Params);
                                       { then add transparency; ensures correct repaint order }
  Params.ExStyle := Params.ExStyle + WS_EX_TRANSPARENT;

end;
{$ENDIF}

{$IFNDEF CLX_}
procedure TStretchHandle.WMGetDLGCode(var Message: TMessage);
begin
                                       { get arrow key press events }
  Message.Result := DLGC_WANTARROWS;

end;

procedure TStretchHandle.WMEraseBkgnd(var Message: TWMEraseBkgnd);
begin
                                       { completely fake erase, don't call inherited, don't collect $200 }
  Message.Result := 1;

end;

{$IFDEF MSWINDOWS}
procedure TStretchHandle.ForwardMessage(FwdMsg: TForwardMessage; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  i: integer;
  Msg: Word;
  ARect: TRect;
  AControl: TControl;
  AMessage: TMessage;
begin
                                       { construct the message to be sent }
  case FwdMsg of
    fmMouseDown:
      case Button of
        mbLeft:
          Msg := WM_LBUTTONDOWN;
        mbMiddle:
          Msg := WM_MBUTTONDOWN;
        mbRight:
          Msg := WM_RBUTTONDOWN;
      end;
    fmMouseUp:
      case Button of
        mbLeft:
          Msg := WM_LBUTTONUP;
        mbMiddle:
          Msg := WM_MBUTTONUP;
        mbRight:
          Msg := WM_RBUTTONUP;
      end;
  end;

  AMessage.WParam := 0;
                                       { determine whether X, Y is over any other windowed control }
  AControl := nil;
  for i := 0 to Parent.ControlCount - 1 do begin
    AControl := TControl(Parent.Controls[i]);
    if (AControl is TWinControl) and not (AControl is TStretchHandle) then begin
      ARect := Rect(AControl.Left,
                    AControl.Top,
                    AControl.Left + AControl.Width,
                    AControl.Top + AControl.Height);
                                      { X, Y are relative to Parent }
      if PtInRect(ARect, Point(X, Y)) then
        break
      else
        AControl := nil;
    end else
      AControl := nil;
  end;                                    { forward the message to the control if found, else to the Parent }
  if AControl <> nil then begin
    AMessage.LParamLo := X - AControl.Left;
    AMessage.LParamHi := Y - AControl.Top;
    SendMessage(TWinControl(AControl).Handle, Msg, AMessage.WParam, AMessage.LParam);
  end else begin
    AMessage.LParamLo := X;
    AMessage.LParamHi := Y;
    SendMessage(Parent.Handle, Msg, AMessage.WParam, AMessage.LParam);
  end;

end;
{$ENDIF}


{$ENDIF}

{$IFDEF XBASED}
{$IFDEF LCLGTK3}
procedure TStretchHandle.CheckMask;
begin
  { TODO: GTk3 }
end;
{$ELSE}
procedure TStretchHandle.CheckMask;
var
  GdiObject: PgdiObject;
  gdkBitmap: pgdkBitmap;
begin
  if not (csDesigning in ComponentState) and (FFreezeCount = 0) and Visible and HandleAllocated and (FStaleMask or (FMask = nil) or (FMask.Width <> Width) or (FMask.Height <> Height)) then begin
    gtk_widget_shape_combine_mask(GetFixedWidget(PGTKWidget(Handle)), nil, 0, 0);
    FreeAndNil(FMask);  { Not sure why it is not possible to re-use bitmaps... }
    if FMask = nil then begin
      FMask := TPixmap.Create();
      FMask.Monochrome := True;
    end;
    FMask.Height := Height;
    FMask.Width := Width;
    FMask.Canvas.Brush.Color := clBlack;
    FMask.Canvas.FillRect(0,0,Width,Height);
    FMask.Canvas.Brush.Color := clWhite;
    DrawHandles(FMask.Canvas);
    GdiObject := PgdiObject(FMask.Handle);
    if GdiObject^.GDIBitmapType = gbBitmap then
      gdkBitmap := GdiObject^.GDIBitmapObject
    else begin
      FMask.Mask(clWhite); { Bitmap not suitable, so try creating mask }
      GdiObject := PgdiObject(FMask.MaskHandle);
      if GdiObject^.GDIBitmapType = gbBitmap then
        gdkBitmap := GdiObject^.GDIBitmapObject
      else
        gdkBitmap := nil;
    end;
    gtk_widget_shape_combine_mask(GetFixedWidget(PGTKWidget(Handle)), gdkBitmap, 0, 0);
    FStaleMask := False;
  end;
end;
{$ENDIF}

procedure TStretchHandle.FreezeMask;
begin
  Inc(FFreezeCount);
end;

procedure TStretchHandle.ThawMask;
begin
  Dec(FFreezeCount);
  FStaleMask := True;
  if FFreezeCount = 0 then
    CheckMask;
end;
{$ENDIF}

procedure TStretchHandle.Attach(ChildControl: TControl);
var
  L, T, W, H: integer;
  MousePos: TPoint;
begin
 try
  {$IFDEF XBASED}
  FreezeMask;
  {$ENDIF}
                                       { definitely not allowed! }
  if ChildControl is TForm then
    raise EBadChild.Create('Handles can not be attached to a Form!');
                                       { add child component to unique list managed by TStretchHandle }
  if (ChildControl <> nil) and (FChildList.IndexOf(TObject(ChildControl)) = -1) then
    begin
                                       { make sure new child's Parent matches siblings }
      if (FChildList.Count > 0) and (ChildControl.Parent <> Parent) then
        Detach;
                                       { initialize when first child is attached }
      if FChildList.Count = 0 then
        begin
          Parent := ChildControl.Parent;
                                       { only make it visible now, to avoid color flashing, & accept events }
          FDragRect := Rect(0, 0, 0, 0);
          Enabled := True;
          Visible := True;
          inherited SetBounds(ChildControl.Left - 2, ChildControl.Top - 2, ChildControl.Width + 5, ChildControl.Height + 5);

        end
      else
        begin
                                       { set size to bound all children, plus room for handles }
          L := MinInt(Left, ChildControl.Left - 2);
          T := MinInt(Top, ChildControl.Top - 2);
          W := Maxint(Left + Width - 3, ChildControl.Left + ChildControl.Width) - L + 3;
          H := Maxint(Top + Height - 3, ChildControl.Top + ChildControl.Height) - T + 3;
          inherited SetBounds(L, T, W, H);
        end;
      {$IFDEF SETPARENT}
      ChildControl.Parent := Self;
      ChildControl.Left := ChildControl.Left - Left;
      ChildControl.Top := ChildControl.Top - Top;
      {$ENDIF}
                                       { add to list of active Children }
      FChildList.Add(TObject(ChildControl));

      SaveAspectRatio;
                                       { re-set DragStyle }
      FDragStyle := dsMove;
                                       { use old BringToFront so as not to change Child's Z-order }
      if not (csDesigning in ComponentState) then
        begin
          inherited BringToFront;

          if HandleAllocated then begin  { CMW }
            if Visible and Enabled and Parent.Visible { CMW } then begin
                                       { allow us to get Mouse events immediately! }
              {$IFDEF VER80}
              GetCursorPos(MousePos);
              {$ELSE}
               MousePos := Mouse.CursorPos;
              {$ENDIF}
              if PtInRect(ClientRect, ScreenToClient(MousePos)) then
                SetCaptureControl(Self);
              SetFocus;                { get keyboard events }
            end;
            {$IFDEF CLX_}
            if ChildControl is TWidgetControl then
              QObject_installEventFilter(TWidgetControl(ChildControl).Handle, QObjectH(Hooks));
            MaskChanged;
            {$ENDIF}
          end;
        end;

    end;
 finally
   {$IFDEF XBASED}
   ThawMask;
   {$ENDIF}
 end;
end;

procedure TStretchHandle.Detach;
begin
 try
  {$IFDEF XBASED}
  FreezeMask;
  {$ENDIF}
                                       { remove all Child components from list }
  if FChildList.Count > 0 then
    with FChildList do
      repeat
        {$IFDEF CLX_}
        if Children[0] is TWidgetControl then
          QObject_removeEventFilter(TWidgetControl(Children[0]).Handle, QObjectH(Hooks));
        {$ENDIF}
        {$IFDEF SETPARENT}
        TControl(Children[0]).Parent := Parent;
        TControl(Children[0]).Left := TControl(Children[0]).Left + Left;
        TControl(Children[0]).Top := TControl(Children[0]).Top + Top;
        {$ENDIF}
        Delete(0);
      until Count = 0;
                                       { disable & hide StretchHandle }
  FLocked := False;
  Visible := False;
  Width := 24;
  Height := 24;
  Enabled := False;
  Parent := nil;
  FDragRect := Rect(0, 0, 0, 0);
 finally
   {$IFDEF XBASED}
   ThawMask;
   {$ENDIF}
 end;
end;

procedure TStretchHandle.ReleaseChild(ChildControl: TControl);
var
  idx, L, T, W, H: integer;
  AControl: TControl;
begin
 try
  {$IFDEF XBASED}
  FreezeMask;
  {$ENDIF}
  {$IFDEF CLX_}
  if ChildControl is TWidgetControl then
    QObject_removeEventFilter(TWidgetControl(ChildControl).Handle, QObjectH(Hooks));
  {$ENDIF}
                                       { delete the Child if it exists in the list }
  idx := FChildList.IndexOf(TObject(ChildControl));
  if (ChildControl <> nil) and (idx >= 0) then begin
    {$IFDEF SETPARENT}
    TControl(Children[idx]).Parent := Parent;
    TControl(Children[idx]).Left := TControl(Children[0]).Left + Left;
    TControl(Children[idx]).Top := TControl(Children[0]).Top + Top;
    {$ENDIF}
    FChildList.Delete(idx);
  end;
                                       { disable & hide StretchHandle if no more children }
  if FChildList.Count = 0 then
    begin
      FLocked := False;
      Enabled := False;
      Visible := False;
      Parent := nil;
      FDragRect := Rect(0, 0, 0, 0);
    end
  else
    begin
                                       { set size to bound remaining children, plus room for handles }
      L := TControl(FChildList.Items[0]).Left - 2;
      T := TControl(FChildList.Items[0]).Top - 2;
      W := TControl(FChildList.Items[0]).Width + 3;
      H := TControl(FChildList.Items[0]).Height + 3;

      for idx := 0 to FChildList.Count - 1 do
        begin
          AControl := TControl(FChildList.Items[idx]);
          L := MinInt(L, AControl.Left - 2);
          T := MinInt(T, AControl.Top - 2);
          W := Maxint(L + W - 3, AControl.Left + AControl.Width) - L + 3;
          H := Maxint(T + H - 3, AControl.Top + AControl.Height) - T + 3;
        end;

      inherited SetBounds(L, T, W, H);

    end;
 finally
   {$IFDEF XBASED}
   ThawMask;
   {$ENDIF}
 end;
end;

function TStretchHandle.IndexOf(ChildControl: TControl): integer;
begin
                                       { simply pass on the result... }
  Result := FChildList.IndexOf(TObject(ChildControl));

end;

procedure TStretchHandle.BringToFront;
var
  i: integer;
begin
                                       { do nothing if not Attached }
  if Attached and not Locked then
    begin
                                       { take care of Children first, in Attach order }
      for i := 0 to FChildList.Count - 1 do
        begin
          TControl(FChildList[i]).BringToFront;
        end;
                                       { make sure keyboard focus is restored }
      inherited BringToFront;
      if Visible and Enabled then
        SetFocus;
    end;

end;

procedure TStretchHandle.SendToBack;
var
  i: integer;
begin
                                       { do nothing if not Attached }
  if Attached and not Locked then
    begin
                                       { take care of Children first, in Attach order }
      for i := 0 to FChildList.Count - 1 do
        begin
          TControl(FChildList[i]).SendToBack;
        end;
                                       { Handles stay in front of everything, always }
      inherited BringToFront;
      if Visible and Enabled then
        SetFocus;
    end;

end;

procedure TStretchHandle.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  // Writeln('MouseDown', FDragging, ' ', LongWord(GetCaptureControl()));
                                       { only process MouseDown if it is over a Child, else forward }
  if PointOverChild(Point({$IFNDEF SETPARENT}Left +{$ENDIF} X, {$IFNDEF SETPARENT}Top{$ENDIF} + Y)) then begin
    if (Button = mbLeft) and not FLocked then begin
      FDragOffset := Point(X, Y);
      FDragging := True;
      SaveAspectRatio;
      {$IFDEF SETPARENT}
      MouseCapture := True;
      {$ENDIF}
    end;
    inherited MouseDown(Button, Shift, X, Y);
  end
{$IFDEF MSWINDOWS}
  else begin
    Cursor := crDefault;
    SetCursor({$IFDEF VCL_}Screen.Cursors[Cursor]{$ELSE}Cursor{$ENDIF});
    ForwardMessage(fmMouseDown, Button, Shift, Left + X, Top + Y);
  end;
{$ENDIF}
end;

procedure TStretchHandle.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  ARect: TRect;
  movedflag, resizedflag : boolean;
begin
  // Writeln('MouseUP', FDragging);
                                       { resize, reposition if anything changed }
  if FDragging and (Button = mbLeft) then
    begin
                                       { disallow drop off Parent }
      if (Left + X) < 0 then
        X := -Left;
      if (Top + Y) < 0 then
        Y := -Top;
      if (Left + X) > Parent.Width then
        X := Parent.Width - Left;
      if (Top + Y) > Parent.Height then
        Y := Parent.Height - Top;

      RubberBand(0, 0, False);         { clear drag outline CMW: Was later, but events could destroy state info }

                                       { force Paint when size doesn't change but position does }
      if (X <> FDragOffset.X) or (Y <> FDragOffset.Y) then
        begin
          Invalidate;
          ARect := GetModifiedRect(X, Y, FDragStyle);
          { if all 4 corners have moved - then we have moved rather than resized - Added by JAB }
          resizedflag := (Arect.Right <> Width) or (Arect.bottom <> Height);
          movedflag := (not resizedflag) and ( (Arect.Left <> Left) or (Arect.top <> top) );
          SetBounds(ARect.Left, ARect.Top, ARect.Right, ARect.Bottom);
          {$IFDEF UNIX}
          Invalidate;
          {$ENDIF}
          if movedflag and Assigned(FOnMoved) then FOnMoved(self);
          if resizedflag and Assigned(FOnResized) then FOnResized(self);
        end;
                                       { seem to need this for keyboard events }
      if Visible and Enabled then
        SetFocus;

      FDragging := False;
      Cursor := crDefault;
      SetCaptureControl(nil);
{$IFDEF UNIX}
      Screen.Cursor := crDefault;
{$ENDIF}
                                       { perform default processing }
      inherited MouseUp(Button, Shift, X, Y);

    end
{$IFDEF MSWINDOWS}
  else
    ForwardMessage(fmMouseUp, Button, Shift, Left + X, Top + Y);
{$ENDIF}
end;

procedure TStretchHandle.MouseMove(Shift: TShiftState; X, Y: Integer);
var
  ARect: TRect;
  DragStyle: TDragStyle;
begin
  // Writeln('MouseMove', FDragging, ' ', LongWord(GetCaptureControl()));
                                       { this may be a move immediately on Attach instead of MouseDown }
  if (ssLeft in Shift) and not FDragging and not FLocked then begin
    FDragOffset := Point(X, Y);
    FDragging := True;
    SetCaptureControl(Self);
    SaveAspectRatio;          
  end else begin
                                     { only recognize move after simulated MouseDown }
                                     { let's not hog mouse events unnecessarily }
    if not (ssLeft in Shift) and MouseCapture then
      SetCaptureControl(nil);        { to conterbalance Capture in Attach }
                                     { default to drag cursor only when dragging }
    DragStyle := dsMove;
                                     { disallow resize if multiple children }
    if not FDragging and (FChildList.Count = 1) then begin
      Cursor := crDefault;
      ARect := GetClientRect;
      if Abs(X - ARect.Left) < 5 then begin
        if Abs(Y - ARect.Top) < 5 then begin
          DragStyle := dsSizeTopLeft;
          Cursor := crSizeNWSE;
        end else if Abs(Y - ARect.Bottom) < 5 then begin
          DragStyle := dsSizeBottomLeft;
          Cursor := crSizeNESW;
        end else if Abs(Y - (ARect.Bottom - ARect.Top) div 2) < 3 then begin
          DragStyle := dsSizeLeft;
          Cursor := crSizeWE;
        end;
      end else if Abs(X - ARect.Right) < 5 then begin
        if Abs(Y - ARect.Bottom) < 5 then begin
          DragStyle := dsSizeBottomRight;
          Cursor := crSizeNWSE;
        end else if Abs(Y - ARect.Top) < 5 then begin
          DragStyle := dsSizeTopRight;
          Cursor := crSizeNESW;
        end else if Abs(Y - (ARect.Bottom - ARect.Top) div 2) < 3 then begin
          DragStyle := dsSizeRight;
          Cursor := crSizeWE;
        end;
      end else if (Abs(Y - ARect.Top) < 5) and (Abs(X - (ARect.Right - ARect.Left) div 2) < 3) then begin
        DragStyle := dsSizeTop;
        Cursor := crSizeNS;
      end else if (Abs(Y - ARect.Bottom) < 5) and (Abs(X - (ARect.Right - ARect.Left) div 2) < 3) then begin
        DragStyle := dsSizeBottom;
        Cursor := crSizeNS;
      end;

    end;
                                     { if position-locked, override cursor change }
    if FLocked then
      Cursor := crNoDrop;

    if FDragging then begin
                                     { disallow drag off Parent }
      if (Left + X) < 0 then
        X := -Left;
      if (Top + Y) < 0 then
        Y := -Top;
      if (Left + X) > Parent.Width then
        X := Parent.Width - Left;
      if (Top + Y) > Parent.Height then
        Y := Parent.Height - Top;
                                     { display cursor & drag outline }
{$IFDEF MSWINDOWS}
      if FDragStyle = dsMove then
        Cursor := DragCursor;
      SetCursor({$IFDEF VCL_}Screen.Cursors[Cursor]{$ELSE}Cursor{$ENDIF});
{$ELSE}
      if FDragStyle = dsMove then
        Cursor := crDrag;
      Screen.Cursor := Cursor;
{$ENDIF}
      RubberBand(X, Y, True);

    end else
      FDragStyle := DragStyle;

  end;
                                       { perform default processing }
  inherited MouseMove(Shift, X, Y);

end;

procedure TStretchHandle.KeyDown(var Key: Word; Shift: TShiftState);
var
 L, T: Integer;
 R: TRect;
 DragStyle: TDragStyle;
begin
                                       { process arrow keys to move/resize Handles & Child, also move siblings }
  L := 0;
  T := 0;
  case Key of
    VK_UP:    T := - GridY;
    VK_DOWN:  T := GridY;
    VK_LEFT:  L := - GridX;
    VK_RIGHT: L := GridX;
  else
    inherited KeyDown(Key, Shift);
    Exit;
  end;
  Invalidate;
  if ssShift in Shift then begin
    if ChildCount = 1 then begin
      if T <> 0 then
        DragStyle := dsSizeBottom
      else
        DragStyle := dsSizeRight;
      if ssCtrl in Shift then begin
        DragStyle := dsSizeBottomRight;
        if T = 0 then
          T := L
        else
          L := T;
      end;
      R := GetModifiedRect(FDragOffset.X+L, FDragOffset.Y+T, DragStyle);
      SetBounds(R.Left, R.Top, R.Right, R.Bottom);
      if Assigned(FOnResized) then FOnResized(self);
    end;
  end else begin
    R := GetModifiedRect(FDragOffset.X+L, FDragOffset.Y+T, dsMove);
    SetBounds(R.Left, R.Top, R.Right, R.Bottom);
    if Assigned(FOnMoved) then FOnMoved(self);
  end;
  inherited KeyDown(Key, Shift);
end;

{$IFDEF XBASED}
procedure TStretchHandle.Resize;
begin
  inherited Resize;
  FStaleMask := True;
  CheckMask;
end;
{$ENDIF}

function TStretchHandle.GetModifiedRect(XPos, YPos: integer; DragStyle: TDragStyle): TRect;
var
  ARect: TRect;
begin
                                       { compute new position/size, depending on FDragStyle}
  case DragStyle of

    dsSizeTopLeft:
      begin
        ARect.Left := Left + XGridAdjust(XPos - FDragOffset.X); { XGridAdjust(Left + (XPos - FDragOffset.X)) - 2; }
        ARect.Top := Top + YGridAdjust(YPos - FDragOffset.Y); { YGridAdjust(Top + (YPos - FDragOffset.Y)) - 2; }
        ARect.Right := Width - (ARect.Left - Left);
        ARect.Bottom := Height - (ARect.Top - Top);
        if FMouseDownAspectRatio * ARect.Bottom > ARect.Right then begin
          ARect.Bottom := FixHeight(ARect.Right, ARect.Bottom, FMouseDownAspectRatio);
          Arect.Top := Top + Height - ARect.Bottom;
        end else begin
          ARect.Right := FixWidth(ARect.Right, ARect.Bottom, FMouseDownAspectRatio);
          Arect.Left := Left + Width - ARect.Right;
        end;
      end;

    dsSizeTopRight:
      begin
        ARect.Left := Left;
        ARect.Top := Top + YGridAdjust(YPos - FDragOffset.Y); { YGridAdjust(Top + (YPos - FDragOffset.Y)) - 2; }
        ARect.Right := Width + XGridAdjust(XPos - FDragOffset.X); { XGridAdjust(Width + (XPos - FDragOffset.X)) - 3; }
        ARect.Bottom := Height - (ARect.Top - Top);
        if FMouseDownAspectRatio * ARect.Bottom > ARect.Right then begin
          ARect.Bottom := FixHeight(ARect.Right, ARect.Bottom, FMouseDownAspectRatio);
          Arect.Top := Top + Height - ARect.Bottom;
        end else begin
          ARect.Right := FixWidth(ARect.Right, ARect.Bottom, FMouseDownAspectRatio);
        end;
      end;

    dsSizeBottomLeft:
      begin
        ARect.Left := Left + XGridAdjust(XPos - FDragOffset.X); { XGridAdjust(Left + (XPos - FDragOffset.X)) - 2; }
        ARect.Top := Top;
        ARect.Right := Width - (ARect.Left - Left);
        ARect.Bottom := Height + YGridAdjust(YPos - FDragOffset.Y); { YGridAdjust(Height + (YPos - FDragOffset.Y)) - 3;}
        if FMouseDownAspectRatio * ARect.Bottom > ARect.Right then
          ARect.Bottom := FixHeight(ARect.Right, ARect.Bottom, FMouseDownAspectRatio)
        else begin
          ARect.Right := FixWidth(ARect.Right, ARect.Bottom, FMouseDownAspectRatio);
          Arect.Left := Left + Width - ARect.Right;
        end;
      end;

    dsSizeBottomRight:
      begin
        ARect.Left := Left;
        ARect.Top := Top;
        ARect.Right := Width + XGridAdjust(XPos - FDragOffset.X); { XGridAdjust(Width + (XPos - FDragOffset.X)) - 3; }
        ARect.Bottom := Height + YGridAdjust(YPos - FDragOffset.Y); { YGridAdjust(Height + (YPos - FDragOffset.Y)) - 3; }
        if FMouseDownAspectRatio * ARect.Bottom > ARect.Right then
          ARect.Bottom := FixHeight(ARect.Right, ARect.Bottom, FMouseDownAspectRatio)
        else
          ARect.Right := FixWidth(ARect.Right, ARect.Bottom, FMouseDownAspectRatio);
      end;

    dsSizeTop:
      begin
        ARect.Left := Left;
        ARect.Top := Top + YGridAdjust(YPos - FDragOffset.Y); { YGridAdjust(Top + (YPos - FDragOffset.Y)) - 2; }
        ARect.Bottom := Height - (ARect.Top - Top);
        ARect.Right := FixWidth(Width, ARect.Bottom, AspectRatio);
      end;

    dsSizeBottom:
      begin
        ARect.Left := Left;
        ARect.Top := Top;
        ARect.Bottom := Height + YGridAdjust(YPos - FDragOffset.Y); { YGridAdjust(Height + (YPos - FDragOffset.Y)) - 3; }
        ARect.Right := FixWidth(Width, ARect.Bottom, AspectRatio);
      end;

    dsSizeLeft:
      begin
        ARect.Left := Left + XGridAdjust(XPos - FDragOffset.X); { CMW: was XGridAdjust(Left + (XPos - FDragOffset.X)) - 2; }
        ARect.Top := Top;
        ARect.Right := Width - (ARect.Left - Left);
        ARect.Bottom := FixHeight(ARect.Right, Height, AspectRatio);
      end;

    dsSizeRight:
      begin
        ARect.Left := Left;
        ARect.Top := Top;
        ARect.Right := Width + XGridAdjust(XPos - FDragOffset.X); { XGridAdjust(Width + (XPos - FDragOffset.X)) - 3; }
        ARect.Bottom := FixHeight(ARect.Right, Height, AspectRatio);
      end;

  else
                                       { keep size, move to new position }
    ARect.Left := Left + XGridAdjust(XPos - FDragOffset.X); { CMW was: XGridAdjust(Left + (XPos - FDragOffset.X)) - 2; }
    ARect.Top := Top + YGridAdjust(YPos - FDragOffset.Y); { YGridAdjust(Top + (YPos - FDragOffset.Y)) - 2; }
    ARect.Right := Width;
    ARect.Bottom := Height;

  end;
                                       { impose a minimum size for sanity }
  if ARect.Right < 5 then
    ARect.Right := 5;
  if ARect.Bottom < 5 then
    ARect.Bottom := 5;

  Result := ARect;

end;

procedure TStretchHandle.Rubberband(XPos, YPos: integer; ShowBox: boolean);
{$IFDEF VERYOLD}
var
  NewRect: TRect;
  PtA, PtB: TPoint;
  ScreenDC: HDC;
begin
                                       { outline is drawn over all windows }
  ScreenDC := GetDC(0);
                                       { erase previous rectangle, if any, & adjust for handle's position }
  if (FDragRect.Left <> 0) or (FDragRect.Top <> 0) or (FDragRect.Right <> 0) or (FDragRect.Bottom <> 0) then
    begin
      PtA := Parent.ClientToScreen(Point(FDragRect.Left + 2, FDragRect.Top + 2));
      PtB := Parent.ClientToScreen(Point(FDragRect.Left + FDragRect.Right - 3, FDragRect.Top + FDragRect.Bottom - 3));
      DrawFocusRect(ScreenDC, Rect(PtA.X, PtA.Y, PtB.X, PtB.Y));
      FDragRect := Rect(0, 0, 0, 0);
    end;
                                       { draw new rectangle unless this is a final erase }
  if ShowBox then
    begin
      NewRect := GetModifiedRect(XPos, YPos, FDragStyle);
      PtA := Parent.ClientToScreen(Point(NewRect.Left + 2, NewRect.Top + 2));
      PtB := Parent.ClientToScreen(Point(NewRect.Left + NewRect.Right - 3, NewRect.Top + NewRect.Bottom - 3));
      DrawFocusRect(ScreenDC, Rect(PtA.X, PtA.Y, PtB.X, PtB.Y));
      FDragRect := NewRect;
    end
  else
    begin
      Parent.Repaint;
      Repaint;
    end;

  ReleaseDC(0, ScreenDC);
{$ENDIF}
{$IFDEF OLD}
begin
  if not (Parent is {$IFDEF VER80} TForm {$ELSE} TCustomForm {$ENDIF} ) then Exit;
                                       { outline is drawn over all windows }
                                       { erase previous rectangle, if any, & adjust for handle's position }
  if (FDragRect.Left <> 0) or (FDragRect.Top <> 0) or (FDragRect.Right <> 0) or (FDragRect.Bottom <> 0) then begin
    {$IFDEF VER80} TForm {$ELSE} TCustomForm {$ENDIF}(Parent)
        .Canvas.DrawFocusRect(Rect(FDragRect.Left + 2, FDragRect.Top + 2,
                              FDragRect.Left + FDragRect.Right - 3, FDragRect.Top + FDragRect.Bottom - 3));
    FDragRect := Rect(0, 0, 0, 0);
  end;
                                       { draw new rectangle unless this is a final erase }
  if ShowBox then begin
    FDragRect := GetModifiedRect(XPos, YPos, FDragStyle);
    {$IFDEF VER80} TForm {$ELSE} TCustomForm {$ENDIF}(Parent)
        .Canvas.DrawFocusRect(Rect(FDragRect.Left + 2, FDragRect.Top + 2,
                              FDragRect.Left + FDragRect.Right - 3, FDragRect.Top + FDragRect.Bottom - 3));
  end else begin
    Parent.Invalidate;
    Repaint;
  end;
{$ENDIF}
{$IFDEF MSWINDOWS}
var
  LineDC: HDC;
  procedure DoFocusRect;
  var
    R: TRect;
  begin
    R := Rect(FDragRect.Left + 2, FDragRect.Top + 2,
                                  FDragRect.Left + FDragRect.Right - 3,
                                  FDragRect.Top + FDragRect.Bottom - 3);
    DrawFocusRect(LineDC, R);
  end;
begin
  LineDC := GetDCEx(Parent.Handle, 0, DCX_CACHE or DCX_CLIPSIBLINGS or DCX_LOCKWINDOWUPDATE);
  try
    if (FDragRect.Left <> 0) or (FDragRect.Top <> 0) or (FDragRect.Right <> 0) or (FDragRect.Bottom <> 0) then begin
      DoFocusRect;
      FDragRect := Rect(0, 0, 0, 0);
    end;
    if ShowBox then begin
      FDragRect := GetModifiedRect(XPos, YPos, FDragStyle);
      DoFocusRect;
    end else begin
      Parent.Invalidate;
      Invalidate;
    end;
  finally
    ReleaseDC(Parent.Handle, LineDC);
  end;
{$ELSE}
{$IFDEF CLX_}
var
  ParentWidget: QOpenWidgetH;
  PaintUnclipped: Boolean;
  Painter: QPainterH;
begin
  ParentWidget := QOpenWidgetH(Parent.Handle);
  PaintUnclipped := QOpenWidget_getWFlags(ParentWidget) and Integer(WidgetFlags_WPaintUnclipped) <> 0;
  if not PaintUnclipped then
    QOpenWidget_setWFlags(ParentWidget, Integer(WidgetFlags_WPaintUnclipped));
  try
    Painter := QPainter_create;
    try
      QPainter_begin(Painter, QWidget_to_QPaintDevice(ParentWidget));
      try
        if (FDragRect.Left <> 0) or (FDragRect.Top <> 0) or (FDragRect.Right <> 0) or (FDragRect.Bottom <> 0) then begin
          QPainter_drawWinFocusRect(Painter, FDragRect.Left + 2, FDragRect.Top + 2,
                                             FDragRect.Right - 3, FDragRect.Bottom - 3);
          FDragRect := Rect(0, 0, 0, 0);
        end;
        if ShowBox then begin
          FDragRect := GetModifiedRect(XPos, YPos, FDragStyle);
          QPainter_drawWinFocusRect(Painter, FDragRect.Left + 2, FDragRect.Top + 2,
                                             FDragRect.Right - 3, FDragRect.Bottom - 3);
        end else begin
          Parent.Invalidate;
          Invalidate;
        end;
      finally
        QPainter_end(Painter);
      end;
    finally
      QPainter_destroy(Painter);
    end;
  finally
    if not PaintUnclipped then
      QOpenWidget_clearWFlags(ParentWidget, Integer(WidgetFlags_WPaintUnclipped));
  end;
{$ENDIF}
{$IFDEF LCL_}
begin
  if not (Parent is TCustomControl) then Exit;
                                       { outline is drawn over all windows }
                                       { erase previous rectangle, if any, & adjust for handle's position }
  if (FDragRect.Left <> 0) or (FDragRect.Top <> 0) or (FDragRect.Right <> 0) or (FDragRect.Bottom <> 0) then begin
    TCustomControl(Parent)
        .Canvas.DrawFocusRect(Rect(FDragRect.Left + 2, FDragRect.Top + 2,
                          FDragRect.Left + FDragRect.Right - 3, FDragRect.Top + FDragRect.Bottom - 3));
    FDragRect := Rect(0, 0, 0, 0);
  end;
                                       { draw new rectangle unless this is a final erase }
  if ShowBox then begin
    FDragRect := GetModifiedRect(XPos, YPos, FDragStyle);
    TCustomControl(Parent)
        .Canvas.DrawFocusRect(Rect(FDragRect.Left + 2, FDragRect.Top + 2,
                          FDragRect.Left + FDragRect.Right - 3, FDragRect.Top + FDragRect.Bottom - 3));
  end else begin
    Parent.Invalidate;
    Repaint;
  end;
{$ENDIF}
{$ENDIF}
end;

procedure TStretchHandle.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
var
  WasVisible: boolean;
  i: integer;
  AControl: TControl;
begin
                                       { hide & preserve fixed size in design mode }
  WasVisible := Visible;
  if csDesigning in ComponentState then
    begin
      Visible := False;
      inherited SetBounds(ALeft, ATop, 24, 24);
    end
  else                                 { move child also, if any (but only if not locked) }
    if not FLocked then
      begin
        {$IFNDEF SETPARENT}
        for i := 0 to FChildList.Count - 1 do
          begin
            AControl := TControl(FChildList[i]);
            AControl.SetBounds(AControl.Left - Left + ALeft,
                               AControl.Top - Top + ATop,
                               AControl.Width - Width + AWidth,
                               AControl.Height - Height + AHeight);
          end;
        {$ENDIF}
        inherited SetBounds(ALeft, ATop, AWidth, AHeight);
      end;
                                       { restore visibility }
  if Visible = False then
    Visible := WasVisible;

end;

procedure TStretchHandle.DrawHandles(ACanvas: TCanvas);
var
  AControl: TControl;
  ARect, BoxRect: TRect;
  i: Integer;
begin
  for i := 0 to FChildList.Count - 1 do begin
    AControl := Children[i];
    {$IFDEF SETPARENT}
    ARect := Rect(AControl.Left - 2,
                  AControl.Top - 2,
                  AControl.Left + AControl.Width + 2,
                  AControl.Top + AControl.Height + 2);
    {$ELSE}
    ARect := Rect(AControl.Left - Left - 2,
                  AControl.Top - Top - 2,
                  AControl.Left - Left + AControl.Width + 2,
                  AControl.Top - Top + AControl.Height + 2);
    {$ENDIF}
    with ACanvas do begin
                                 { draw corner boxes (assuming Canvas is minimum 5x5) }
      BoxRect := Rect(ARect.Left, ARect.Top, ARect.Left + 5, ARect.Top + 5);
      FillRect(BoxRect);
      BoxRect := Rect(ARect.Right - 5, ARect.Top, ARect.Right, ARect.Top + 5);
      FillRect(BoxRect);
      BoxRect := Rect(ARect.Right - 5, ARect.Bottom - 5, ARect.Right, ARect.Bottom);
      FillRect(BoxRect);
      BoxRect := Rect(ARect.Left, ARect.Bottom - 5, ARect.Left + 5, ARect.Bottom);
      FillRect(BoxRect);
                               { only for single Children, draw center boxes }
      if FChildList.Count = 1 then begin
        BoxRect := Rect(ARect.Left + (ARect.Right - ARect.Left) div 2 - 2,
                        ARect.Top,
                        ARect.Left + (ARect.Right - ARect.Left) div 2 + 3,
                        ARect.Top + 5);
        FillRect(BoxRect);
        BoxRect := Rect(ARect.Left + (ARect.Right - ARect.Left) div 2 - 2,
                        ARect.Bottom - 5,
                        ARect.Left + (ARect.Right - ARect.Left) div 2 + 3,
                        ARect.Bottom);
        FillRect(BoxRect);
        BoxRect := Rect(ARect.Left,
                        ARect.Top + (ARect.Bottom - ARect.Top) div 2 - 2,
                        ARect.Left + 5,
                        ARect.Top + (ARect.Bottom - ARect.Top) div 2 + 3);
        FillRect(BoxRect);
        BoxRect := Rect(ARect.Right - 5,
                        ARect.Top + (ARect.Bottom - ARect.Top) div 2 - 2,
                        ARect.Right,
                        ARect.Top + (ARect.Bottom - ARect.Top) div 2 + 3);
        FillRect(BoxRect);
      end;
    end;
  end;
end;

procedure TStretchHandle.Paint;
var
   BoxRect: TRect;
begin

  inherited Paint;
                                        { do it differently at design time... }
  if csDesigning in ComponentState then begin
    Canvas.Brush.Color := FPrimaryColor;
    BoxRect := Rect(0, 0, 5, 5);
    Canvas.FillRect(BoxRect);
    BoxRect := Rect(19, 0, 24, 5);
    Canvas.FillRect(BoxRect);
    BoxRect := Rect(19, 19, 24, 24);
    Canvas.FillRect(BoxRect);
    BoxRect := Rect(0, 19, 5, 24);
    Canvas.FillRect(BoxRect);
  end else begin
                                     { set color to primary if only one child, else secondary }
    if FChildList.Count = 1 then
      Canvas.Brush.Color := FPrimaryColor
    else
      Canvas.Brush.Color := FSecondaryColor;
    DrawHandles(Canvas);

  end;

end;

procedure TStretchHandle.SetPrimaryColor(AColor: TColor);
begin
                                       { set single select color, repaint immediately }
  FPrimaryColor := AColor;
  Repaint;

end;

procedure TStretchHandle.SetSecondaryColor(AColor: TColor);
begin
                                       { set multiple select color, repaint immediately }
  FSecondaryColor := AColor;
  Repaint;

end;

procedure TStretchHandle.SetColors(Color1, Color2: TColor);
begin
                                       { set single/multiple select colors, repaint }
  FPrimaryColor := Color1;
  FSecondaryColor := Color2;
  Repaint;

end;

procedure TStretchHandle.SetGridState(Value: boolean);
begin
                                       { a value of 1 effectively disables a grid axis }
  if Value then
    begin
      FGridX := 8;
      FGridY := 8;
    end
  else
    begin
      FGridX := 1;
      FGridY := 1;
    end;

end;

function TStretchHandle.GetGridState: boolean;
begin

  if (FGridX > 1) or (FGridY > 1) then
    Result := True
  else
    Result := False;

end;

function TStretchHandle.GetChildCount: integer;
begin
  Result := FChildList.Count;
end;

function TStretchHandle.GetChildControl(idx: integer): TControl;
begin
  if (FChildList.Count > 0) and (idx >= 0) then
    Result := TControl(FChildList[idx])
  else
    Result := nil;
end;

function TStretchHandle.IsAttached: boolean;
begin
  Result := FChildList.Count > 0;
end;

function TStretchHandle.PointOverChild(P: TPoint): boolean;
var
  i: integer;
  ARect: TRect;
  AControl: TControl;
begin
                                       { determine whether X, Y is over any child (for dragging) }
  Result := False;
  for i := 0 to FChildList.Count - 1 do
    begin
      AControl := TControl(FChildList[i]);
      ARect := Rect(AControl.Left - 2,
                    AControl.Top - 2,
                    AControl.Left + AControl.Width + 2,
                    AControl.Top + AControl.Height + 2);
                                       { P is relative to the Parent }
      if PtInRect(ARect, P) then
        begin
          Result := True;
          break;
        end;
    end;

end;

function TStretchHandle.XGridAdjust(X: integer): integer;
begin
  Result := (X div FGridX) * FGridX;
end;

function TStretchHandle.YGridAdjust(Y: integer): integer;
begin
  Result := (Y div FGridY) * FGridY;
end;

function TStretchHandle.FixWidth(AWidth, AHeight: Integer; AAspectRatio: Double): Integer;
begin
  if AAspectRatio > 0 then
    Result := Round(AHeight * AAspectRatio)
  else
    Result := AWidth;
end;

function TStretchHandle.FixHeight(AWidth, AHeight: Integer; AAspectRatio: Double): Integer;
begin
  if AAspectRatio > 0 then
    Result := Round(AWidth / AAspectRatio)
  else
    Result := AHeight;
end;

procedure TStretchHandle.SaveAspectRatio;
var
  Control: TControl;
begin
  if AspectRatio <> 0 then
    FMouseDownAspectRatio := AspectRatio
  else if ChildCount = 1 then begin
    FMouseDownAspectRatio := 0;
    Control := Children[0];
    if Control.Height > 0 then
      FMouseDownAspectRatio := Control.Width / Control.Height
  end;
end;

end.
