unit PrintPre;

interface
{$I utils.inc}{ Copyright (C) 2005 C M Western; see PcontrolsReg.pas for license}
uses
  {$IFNDEF CLX_} { WinTypes, WinProcs, } Messages, Graphics, Controls, Forms, Dialogs, Printers,
  {$ELSE} QGraphics, QControls, QForms, QDialogs, QPrinters, {$ENDIF}
  SysUtils, Classes, Prevwdlg, ExpCan;

type
  TPrintPreview = class(TComponent)
  private
    FOrientation: TPrinterOrientation;
    FXSize, FYSize,                        { Page size in mm }
    FXMargin, FYMargin: Double;            { Printing Offset in mm; set very negative to centre }
    FPages: Integer;
    FOnShowPreview: TShowPreviewEvent;
    FOptions: TPrintPreviewOptions;
    FDefaultDrawing: Boolean;
    FMaxPages: Integer;
    FOnChange: TNotifyEvent;
    FDefaultFraction: Double;
    FAspectRatio: Double;
    FOnSettingsClick: TNotifyEvent;
    procedure SetOptions(const Value: TPrintPreviewOptions);
    procedure SetDefaultDrawing(const Value: Boolean);
    procedure SetMaxPages(const Value: Integer);
    procedure SetDefaultFraction(const Value: Double);
    procedure SetAspectRatio(const Value: Double);
    procedure SetOnSettingsClick(const Value: TNotifyEvent);
  protected
    PreviewForm: TPrintPreviewForm;
    procedure PInvalidate(Sender: TObject);
    procedure ReadBack(Sender: TObject); virtual;
    procedure PreviewGoing(Sender: TObject);
    procedure SetUpPreviewForm;
    procedure SetXSize(v:Double);
    procedure SetYSize(v:Double);
    procedure SetXMargin(v:Double);
    procedure SetYMargin(v:Double);
    procedure SetPages(v:Integer);
    procedure SetOrientation(AOrientation: TPrinterOrientation);
    procedure DoShowPreview(Sender:TObject; Canvas:TExportableCanvas; X, Y, AWidth, AHeight:Longint;
                            APage: Integer; OnPrinter:Boolean); virtual;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure SetMargins(AXSize, AYSize, AXMargin, AYMargin:Double); virtual;
    procedure Show;
    procedure ShowModal;
    procedure Print;
    procedure PrinterSetup;
  published
    property Pages:Integer read FPages write SetPages default 1;
    property XSize:Double read FXSize write SetXSize;
    property YSize:Double read FYSize write SetYSize;
    property XMargin:Double read FXMargin write SetXMargin;
    property YMargin:Double read FYMargin write SetYMargin;
    property Orientation: TPrinterOrientation read FOrientation write SetOrientation default poLandscape;
    property OnShowPreview:TShowPreviewEvent read FOnShowPreview write FOnShowPreview;
    property OnChange:TNotifyEvent read FOnChange write FOnChange;
    property Options: TPrintPreviewOptions read FOptions write SetOptions;
    property DefaultDrawing:Boolean read FDefaultDrawing write SetDefaultDrawing;
    property MaxPages:Integer read FMaxPages write SetMaxPages default 99;
    property DefaultFraction: Double read FDefaultFraction write SetDefaultFraction;
    { Width / Height of image }
    property AspectRatio: Double read FAspectRatio write SetAspectRatio;
    property OnSettingsClick: TNotifyEvent read FOnSettingsClick write SetOnSettingsClick;
  end;

implementation

constructor TPrintPreview.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FXMargin := -2*LargeMargin;
  FYMargin := -2*LargeMargin;
  FOrientation := poLandscape;
  FXSize := -1;
  FYSize := -1;
  FPages := 1;
  FOptions := [ppShowPrint, ppShowPrintSetup, ppShowLandPort, ppShowPages,
               ppShowCentre, ppShowDefault, ppShowZoom, ppShowClose];
  FMaxPages := 99;
  FDefaultFraction := 0.9;
end;

destructor TPrintPreview.Destroy;
begin
  if Assigned(PreviewForm) then
    PreviewForm.Free;
  inherited Destroy;
end;

procedure TPrintPreview.SetXSize(v:Double);
begin
  SetMargins(v, YSize, XMargin, YMargin);
end;

procedure TPrintPreview.SetYSize(v:Double);
begin
  SetMargins(XSize, v, XMargin, YMargin);
end;

procedure TPrintPreview.SetXMargin(v:Double);
begin
  SetMargins(XSize, YSize, v, YMargin);
end;

procedure TPrintPreview.SetYMargin(v:Double);
begin
  SetMargins(XSize, YSize, XMargin, v);
end;

procedure TPrintPreview.SetMargins(AXSize, AYSize, AXMargin, AYMargin:Double);
begin
  FXSize := AXSize;
  FYSize := AYSize;
  FXMargin := AXMargin;
  FYMargin := AYMargin;
  if Assigned(PreviewForm) then
    PreviewForm.SetMargins(FXSize, FYSize, FXMargin, FYMargin);
end;

procedure TPrintPreview.SetOrientation(AOrientation: TPrinterOrientation);
begin
  FOrientation := AOrientation;
  if Assigned(PreviewForm) then
    PreviewForm.Orientation := AOrientation;
end;

procedure TPrintPreview.PInvalidate(Sender: TObject);
begin
  if Assigned(PreviewForm) then
    PreviewForm.Invalidate;
end;

procedure TPrintPreview.SetPages(v:Integer);
begin
  if Assigned(PreviewForm) then
    PreviewForm.Pages := v;
  FPages := v;
end;

procedure TPrintPreview.ReadBack(Sender: TObject);
begin
  SetMargins(PreviewForm.XSize, PreviewForm.YSize, PreviewForm.XMargin, PreviewForm.YMargin);
  Pages := PreviewForm.Pages;
  Orientation := PreviewForm.Orientation;
  if Assigned(FOnChange) then FOnChange(Self);
end;

procedure TPrintPreview.PreviewGoing(Sender: TObject);
begin
  PreviewForm := nil;
end;

procedure TPrintPreview.Show;
begin
  SetUpPreviewForm;
  PreviewForm.Show;
end;

procedure TPrintPreview.ShowModal;
begin
  SetUpPreviewForm;
  PreviewForm.ShowModal;
end;

procedure TPrintPreview.Print;
begin
  SetUpPreviewForm;
  PreviewForm.Print;
end;

procedure TPrintPreview.PrinterSetup;
begin
  SetUpPreviewForm;
  PreviewForm.PrinterSetup;
end;

procedure TPrintPreview.SetUpPreviewForm;
begin
  if not Assigned(PreviewForm) then begin
    PreviewForm := TPrintPreviewForm.Create(Application);
    PreviewForm.Orientation := Orientation;
    PreviewForm.Options := Options;
    PreviewForm.DefaultDrawing := DefaultDrawing;
    PreviewForm.MaxPages := MaxPages;
    PreviewForm.DefaultFraction := DefaultFraction;
    PreviewForm.AspectRatio := AspectRatio;
    PreviewForm.Pages := Pages;
    SetMargins(FXSize, FYSize, FXMargin, FYMargin);
    PreviewForm.OnChange := {$IFDEF FPC}@{$ENDIF}ReadBack;
    PreviewForm.OnDestroy := {$IFDEF FPC}@{$ENDIF}PreviewGoing;
    PreviewForm.OnShowPreview := {$IFDEF FPC}@{$ENDIF}DoShowPreview;
    PreviewForm.OnSettingsClick := OnSettingsClick;
    ReadBack(Self);
  end;
end;

procedure TPrintPreview.DoShowPreview(Sender: TObject; Canvas: TExportableCanvas;
  X, Y, AWidth, AHeight:Longint; APage: Integer; OnPrinter: Boolean);
begin
  if Assigned(FOnShowPreview) then
    FOnShowPreview(Sender, Canvas, X, Y, AWidth, AHeight, APage, OnPrinter);
end;

procedure TPrintPreview.SetOptions(const Value: TPrintPreviewOptions);
begin
  FOptions := Value;
  if Assigned(PreviewForm) then
    PreviewForm.Options := Value;
end;

procedure TPrintPreview.SetDefaultDrawing(const Value: Boolean);
begin
  FDefaultDrawing := Value;
  if Assigned(PreviewForm) then
    PreviewForm.DefaultDrawing := Value;
end;

procedure TPrintPreview.SetMaxPages(const Value: Integer);
begin
  FMaxPages := Value;
  if Assigned(PreviewForm) then
    PreviewForm.MaxPages := Value;
end;

procedure TPrintPreview.SetDefaultFraction(const Value: Double);
begin
  FDefaultFraction := Value;
  if Assigned(PreviewForm) then
    PreviewForm.DefaultFraction := Value;
end;

procedure TPrintPreview.SetAspectRatio(const Value: Double);
begin
  FAspectRatio := Value;
  if Assigned(PreviewForm) then
    PreviewForm.AspectRatio := Value;
end;

procedure TPrintPreview.SetOnSettingsClick(const Value: TNotifyEvent);
begin
  FOnSettingsClick := Value;
  if Assigned(PreviewForm) then
    PreviewForm.OnSettingsClick := Value;
end;

end.
