unit PropStrm;
{$I utils.inc}{ Copyright (C) 2005 C M Western; see PcontrolsReg.pas for license}
interface

uses Sysutils, Classes, TypInfo;

type

  TPropReader = class(TObject)
  private
    FSections: TStrings;
    function GetSection(const Name: string): TStrings;
  public
    ErrorReportStyle: (erIgnore, erReport, erRaise);
    constructor Create(const FileName: string);
    destructor Destroy; override;
    procedure ReadProperties(AObject: TObject; Section, Names: String);
    procedure ReadProperty(AObject: TObject; const Prefix: string; Section: TStrings; PropInfo: PPropInfo);
    procedure ReadClassProperties(AObject: TObject; Prefix: string; Section: TStrings; PropInfo: PPropInfo);
    function ReadBool(const Section, Key: string; var Value: Boolean): Boolean;
    function ReadString(const Section, Key: string; var Value: string): Boolean;
    function ReadInteger(const Section, Key: string; var Value: Integer): Boolean;
    function ReadFloat(const Section, Key: string; var Value: Extended): Boolean;
    property Sections[const Name: string]: TStrings read GetSection;
  end;

procedure SavePropertiesToStream(AObject: TObject; s: TStream; Section, Names: String);

implementation

uses UCompat, PropHelp, Forms;

{ Mainly borrowed from RxProps unit, but some
  stuff now available in TypInfo, SysUtils unit, so
  hidden by COMPAT }

procedure SavePropertyToStream(AObject: TObject; const Prefix: string; s: TStream; PropInfo: PPropInfo); forward;

function NextName(var Names: string): string;
var
  i: Integer;
begin
  Names := Trim(Names);
  if Length(Names) = 0 then
    Result := ''
  else begin
    if Names[1] = '[' then
      Names := Trim(Copy(Names, 2, Length(Names)));
    if Names[Length(Names)] = ']' then
      Names := Trim(Copy(Names, 1, Length(Names)-1));
    i := Pos(',', Names);
    if i = 0 then begin
      Result := Names;
      Names := '';
    end else begin
      Result := Copy(Names, 1, i-1);
      Names := Copy(Names, i+1, Length(Names));
    end;
    Result := Trim(Result);
  end;
end;

function NextProperty(var Names: string; AObject: TObject; var BObject: TObject;
                      var Prefix: string; var PropInfo: PPropInfo): Boolean;
var
  Name, Tail: string;
  e: Integer;
begin
  Name := NextName(Names);
  Result := Name <> '';
  if Result then begin
    e := Pos('.', Name);
    if e > 0 then begin
      Tail := Copy(Name, e+1, Length(Name));
      Name := Copy(Name, 1, e-1);
      Prefix := Name + '.';
    end else
      Prefix := '';
    PropInfo := GetPropInfo(AObject.ClassInfo, Name);
    if (e>0) and (PropInfo = nil) and (AObject is TComponent) then begin
      BObject := TComponent(AObject).FindComponent(Name);
      PropInfo := GetPropInfo(BObject.ClassInfo, Tail);
    end else
      BObject := AObject;
    if PropInfo = nil then
      raise Exception.CreateFmt('Can''t find property: %s', [Name]);
  end;
end;

procedure SaveClassPropertiesToStream(AObject: TObject; Prefix: string; s: TStream; PropInfo: PPropInfo);
var
  line: string;
  i: Integer;
  List: PPropList;
  Count: Integer;
  Size: Integer;
begin
  Prefix := Concat(Prefix, PropInfo^.Name, '.');
  {$IFDEF FPC}
  AObject := GetObjectProp(AObject, PropInfo, nil);
  {$ELSE}
  AObject := TObject(GetOrdProp(AObject, PropInfo));
  {$ENDIF}
  if AObject = nil then
    raise Exception.CreateFmt('Component not found: %s', [Prefix])
  else if AObject is TStrings then begin
    FmtStr(line, '%scount=%d'+LineEnd, [Prefix, TStrings(AObject).Count]);
    s.Write(line[1], Length(Line));
    for i := 0 to TStrings(AObject).Count-1 do begin
      FmtStr(line, '%s%d=%s'+LineEnd, [Prefix, i, TStrings(AObject).Strings[i]]);
      s.Write(line[1], Length(Line));
    end;
  end else if AObject is TPersistent then begin
    List := nil;
    Size := 0;
    try
      Count := GetPropList(AObject.ClassInfo, tkProperties, nil);
      Size := Count * SizeOf(Pointer);
      GetMem(List, Size);
      GetPropList(AObject.ClassInfo, tkProperties, List);
      for i := 0 to Count - 1 do
          SavePropertyToStream(AObject, Prefix, s, List^[i]);
    finally
      FreeMem(List, Size);
    end
  end else
    raise Exception.CreateFmt('Unknown class type for: %s', [Prefix]);
end;

procedure SavePropertyToStream(AObject: TObject; const Prefix: string; s: TStream; PropInfo: PPropInfo);
var
  line: string;
begin
  if PropInfo^.PropType^.Kind = tkClass then begin
    SaveClassPropertiesToStream(AObject, Prefix, s, PropInfo);
  end else begin
    line := Concat(Prefix, PropInfo^.Name, '=', PropertyToStr(AObject, PropInfo), #13#10);
    s.Write(line[1], Length(line));
  end
end;

procedure SavePropertiesToStream(AObject: TObject; s: TStream; Section, Names: String);
var
  Prefix: string;
  PropInfo: PPropInfo;
  BObject: TObject;
begin
  if Section <> '' then begin
    FmtStr(Prefix, '[%s]'+LineEnd, [Section]);
    s.Write(Prefix[1], Length(Prefix));
  end;
  while NextProperty(Names, AObject, BObject, Prefix, PropInfo) do
    SavePropertyToStream(BObject, Prefix, s, PropInfo)
end;

{ TPropReader }

constructor TPropReader.Create(const FileName: string);
var
  Lines: TStringList;
  i: Integer;
  CurSection: TStringList;
  l: string;
begin
  inherited Create;
  ErrorReportStyle := erIgnore;
  Lines := nil;
  FSections := nil;
  try
    FSections := TStringList.Create;
    Lines := TStringList.Create;
    Lines.LoadFromFile(FileName);
    CurSection := nil;
    for i := 0 to Lines.Count - 1 do begin
      l := Trim(Lines[i]);
      if not ( (l = '') or (l[1] = ';') ) then
        if ( (l[1] = '[') and (l[Length(l)] = ']') ) then begin
          CurSection := TStringList.Create;
          FSections.AddObject(Copy(l, 2, Length(l)-2), CurSection);
        end else if CurSection <> nil then
          CurSection.Add(Lines[i]);
    end;
  finally
    Lines.Free;
  end;
end;

destructor TPropReader.Destroy;
var
  i: Integer;
begin
  if FSections <> nil then
    for i := 0 to FSections.Count - 1 do
      FSections.Objects[i].Free;
  FSections.Free;
  inherited Destroy;
end;

procedure TPropReader.ReadProperties(AObject: TObject; Section, Names: String);
var
  Prefix: string;
  SectionList: TStrings;
  PropInfo: PPropInfo;
  BObject: TObject;
begin
  SectionList := Sections[Section];
  if SectionList <> nil then
    while NextProperty(Names, AObject, BObject, Prefix, PropInfo) do
      ReadProperty(BObject, Prefix, SectionList, PropInfo)
end;

procedure TPropReader.ReadProperty(AObject: TObject; const Prefix: string;
  Section: TStrings; PropInfo: PPropInfo);
var
  s: string;
begin
  if PropInfo^.PropType^.Kind = tkClass then begin
    ReadClassProperties(AObject, Prefix, Section, PropInfo);
  end else begin
    s := Concat(Prefix, PropInfo^.Name);
    s := Section.Values[s];
    if Length(s) > 0 then
      case ErrorReportStyle of
        erIgnore, erReport:
          try
            SetPropertyFromStr(AObject, PropInfo, s);
          except
            on E: Exception do
              if ErrorReportStyle = erReport then
                Application.ShowException(E);
          end;
        erRaise:
          SetPropertyFromStr(AObject, PropInfo, s);
      end;
  end
end;

procedure TPropReader.ReadClassProperties(AObject: TObject; Prefix: string; Section: TStrings; PropInfo: PPropInfo);
var
  key, s: string;
  i: Integer;
  List: PPropList;
  Count: Integer;
  Size: Integer;
begin
  Prefix := Concat(Prefix, PropInfo^.Name, '.');
  {$IFDEF FPC}
  AObject := GetObjectProp(AObject, PropInfo, nil);
  {$ELSE}
  AObject := TObject(GetOrdProp(AObject, PropInfo));
  {$ENDIF}
  if AObject = nil then
    raise Exception.CreateFmt('Component not found: %s', [Prefix])
  else if AObject is TStrings then begin
    FmtStr(key, '%scount', [Prefix]);
    s := Section.Values[key];
    if Length(s) > 0 then begin
      TStrings(AObject).Clear;
      Size := StrToInt(s);
      for i := 0 to Size-1 do begin
        FmtStr(key, '%s%d', [Prefix, i]);
        s := Section.Values[key];
        if Length(s) > 0 then
          TStrings(AObject).Add(s);
      end;
    end;
  end else if AObject is TPersistent then begin
    List := nil;
    Size := 0;
    try
      Count := GetPropList(AObject.ClassInfo, tkProperties, nil);
      Size := Count * SizeOf(Pointer);
      GetMem(List, Size);
      GetPropList(AObject.ClassInfo, tkProperties, List);
      for i := 0 to Count - 1 do
          ReadProperty(AObject, Prefix, Section, List^[i]);
    finally
      FreeMem(List, Size);
    end
  end else
    raise Exception.CreateFmt('Unknown class type for: %s', [Prefix]);
end;

function TPropReader.GetSection(const Name: string): TStrings;
var
  i: Integer;
begin
  i := FSections.IndexOf(Name);
  if i >= 0 then
    Result := TStrings(FSections.Objects[i])
  else
    Result := nil;
end;

function TPropReader.ReadBool(const Section, Key: string; var Value: Boolean): Boolean;
var
  v: string;
begin
  Result := ReadString(Section, Key, v);
  if Result then
    if (CompareText(v, 'true') = 0) or (CompareText(v, 'yes') = 0) then
      Value := True
    else if (CompareText(v, 'false') = 0) or (CompareText(v, 'no') = 0) then
      Value := False
    else
      Value := StrToInt(v) <> 0;
end;

function TPropReader.ReadString(const Section, Key: string; var Value: string): Boolean;
var
  SectionStrings: TStrings;
  s: string;
begin
  Result := False;
  SectionStrings := Sections[Section];
  if SectionStrings <> nil then begin
    s := SectionStrings.Values[Key];
    if Length(s) > 0 then begin
      Value := s;
      Result := True;
    end;
  end;
end;

function TPropReader.ReadInteger(const Section, Key: string;  var Value: Integer): Boolean;
var
  v: string;
begin
  Result := ReadString(Section, Key, v);
  if Result then
    Value := StrToInt(v);
end;

function TPropReader.ReadFloat(const Section, Key: string;  var Value: Extended): Boolean;
var
  v: string;
begin
  Result := ReadString(Section, Key, v);
  if Result then
    Value := StrToFloat(v);
end;

end.

