unit Recent;
{$I utils.inc}{ Copyright (C) 2013 C M Western; see PcontrolsReg.pas for license}
interface

uses Classes, Menus;

type

  TRecentClick = procedure(Sender: TObject; AFilename: string) of object;
  TGetDisplayName = procedure(Sender: TObject; AFilename: string; var ADisplayName: string) of object;

  TRecentFiles = class(TComponent)
  private
    FAddAfter: TMenuItem;
    FAutoSaveKey: string;
    FAutoSaveName: string;
    FFiles: TStrings;
    FnRecentFileItems: Integer;
    FOnClick: TRecentClick;
    { These two only valid after SetupOK call }
    Base: Integer;
    ParentMenu: TMenuItem;
    FRemoveOnClick: Boolean;
    FMaxItems: Integer;
    FInsertSeparator: Boolean;
    FOnGetDisplayName: TGetDisplayName;
    procedure SetAddAfter(const Value: TMenuItem);
    procedure SetMaxItems(const Value: Integer);
    procedure SetInsertSeparator(const Value: Boolean);
  protected
    procedure Loaded; override;
    procedure MenuClicked(Sender: TObject);
    function SetupOK: Boolean;
  public
    property Files: TStrings read FFiles;
    property nRecentFileItems: Integer read FnRecentFileItems; { Number of recent file items created; unused items are hidden, rather then deleted }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure AddAll;
    procedure RemoveAll;
    procedure LoadFromRegistry;
    procedure SaveToRegistry;
    procedure AddItem(FileName: string);
  published
    property AddAfter : TMenuItem              read FAddAfter         write SetAddAfter;      { Menu items are added after this one }
    property AutoSaveName: string              read FAutoSaveName     write FAutoSaveName;    { The registry key to save to }
    property AutoSaveKey: string               read FAutoSaveKey      write FAutoSaveKey;     { The section to save to }
    property OnClick: TRecentClick             read FOnClick          write FOnClick;
    property OnGetDisplayName: TGetDisplayName read FOnGetDisplayName write FOnGetDisplayName;
    property RemoveOnClick: Boolean            read FRemoveOnClick    write FRemoveOnClick default true;
    property MaxItems: Integer                 read FMaxItems         write SetMaxItems default 5;
    property InsertSeparator: Boolean          read FInsertSeparator  write SetInsertSeparator default true;
  end;

implementation

uses {$IFDEF LCL_} LCLType, {$ENDIF} IniFiles, UCompat, FileCtrl, SysUtils;

{ TRecentFiles }

procedure TRecentFiles.AddAll;
var
  MenuItem: TMenuItem;
  i: Integer;
  Caption: string;
begin
  if not SetupOK then
    Exit;
  for i := 0 to Files.Count-1 do begin
    if i >= nRecentFileItems then begin
      MenuItem := TMenuItem.Create(Self);
      MenuItem.Tag := i;
      MenuItem.OnClick := {$IFDEF FPC}@{$ENDIF}MenuClicked;
      ParentMenu.Insert(Base+i,MenuItem);
      Inc(FnRecentFileItems);
    end else
      MenuItem := ParentMenu.Items[Base+i];
    Caption := Files[i];
    if i = 9 then
      Caption := '1&0 ' + Caption
    else if i < 9 then begin
      Caption := '&' + Char(Ord('1')+i) + ' ' + Caption;
      {$IFDEF DARWIN}
      MenuItem.ShortCut := ShortCut(VK_1+i, [ssMeta]);
      {$ENDIF}
    end;
    if Assigned(OnGetDisplayName) then
      OnGetDisplayName(Self, Files[i], Caption);
    MenuItem.Caption := Caption;
    MenuItem.Visible := True;
  end;
  for i := Files.Count to nRecentFileItems-1 do
    ParentMenu.Items[Base+i].Visible := False;
  if (ParentMenu.Count > Base+nRecentFileItems) and (ParentMenu[Base+nRecentFileItems].Owner = Self) then
    ParentMenu[Base+nRecentFileItems].Visible := InsertSeparator and (nRecentFileItems > 0)
  else if InsertSeparator and (nRecentFileItems > 0) then begin
    MenuItem := TMenuItem.Create(Self);
    MenuItem.Caption := '-';
    ParentMenu.Insert(Base+nRecentFileItems,MenuItem);
  end;
end;

procedure TRecentFiles.AddItem(FileName: string);
var
  p: Integer;
begin
  p := Files.IndexOf(FileName);
  if p >= 0 then
    Files.Move(p, 0)
  else
    Files.Insert(0,FileName);
  SetMaxItems(MaxItems); { Triggers redisplay }
end;

constructor TRecentFiles.Create(AOwner: TComponent);
begin
  inherited;
  FFiles := TStringList.Create;
  FAutoSaveKey := 'MRU Items';
  FRemoveOnClick := True;
  FMaxItems := 5;
  FInsertSeparator := True;
end;

destructor TRecentFiles.Destroy;
begin
  SaveToRegistry;
  FFiles.Free;
  inherited;
end;

procedure TRecentFiles.Loaded;
begin
  inherited;
  LoadFromRegistry;
end;

procedure TRecentFiles.LoadFromRegistry;
var
  IniSettings: TCustomIniFile;
  i, p, n: Integer;
  s: string;
begin
  if (AutoSaveName = '') or (AutoSaveKey = '') or (csDesigning in ComponentState) then
    exit;
  Files.Clear;
  IniSettings := GetIniFile(AutoSaveName, False);
  try
    IniSettings.ReadSectionValues(AutoSaveKey, Files);
  finally
    IniSettings.Free;
  end;
  { Strings are of the form "F1=file.txt" }
  n := 0;
  for i := 0 to Files.Count-1 do begin
    s := Files[i];
    p := Pos('=', s);
    if (p > 0) and (p < Length(s)) then begin
      Files[n] := Copy(Files[i],p+1,Length(Files[i]));
      Inc(n);
    end else
      Files[i] := '';
  end;
  AddAll;
end;

procedure TRecentFiles.SaveToRegistry;
var
  IniSettings: TCustomIniFile;
  i: Integer;
begin
  if (AutoSaveName = '') or (AutoSaveKey = '') or (csDesigning in ComponentState) then
    exit;
  IniSettings := GetIniFile(AutoSaveName, False);
  try
    IniSettings.EraseSection(AutoSaveKey);
    for i := 0 to Files.Count-1 do
      IniSettings.WriteString(AutoSaveKey, 'F'+IntToStr(i), Files[i]);
  finally
    IniSettings.Free;
  end;
end;

procedure TRecentFiles.MenuClicked(Sender: TObject);
var
  Index: Integer;
  FileName: string;
begin
  if not (Sender is TComponent) then
    Exit;
  Index := TComponent(Sender).Tag;
  if (Index < 0) or (Index >= Files.Count) then
    Exit;
  FileName := Files[Index];
  if RemoveOnClick then begin
    Files.Delete(Index);
    AddAll;
  end;
  if Assigned(OnClick) then
    OnClick(Self,FileName);
end;

procedure TRecentFiles.RemoveAll;
var
  i: Integer;
begin
  if not SetupOK then
    Exit;
  for i := nRecentFileItems-1 downto 0 do
    ParentMenu.Items[Base+i].Free;
  if (ParentMenu.Count > Base) and (ParentMenu[Base].Owner = Self) then
    ParentMenu.Items[Base].Free; { Separator Item }
  FnRecentFileItems := 0;
end;

procedure TRecentFiles.SetAddAfter(const Value: TMenuItem);
begin
  RemoveAll;
  FAddAfter := Value;
  AddAll;
end;

procedure TRecentFiles.SetInsertSeparator(const Value: Boolean);
begin
  FInsertSeparator := Value;
  AddAll;
end;

procedure TRecentFiles.SetMaxItems(const Value: Integer);
begin
  FMaxItems := Value;
  while Files.Count > MaxItems do
    Files.Delete(Files.Count-1);
  AddAll;
end;

function TRecentFiles.SetupOK: Boolean;
begin
  Result := False;
  if (csLoading in ComponentState) or (csDestroying in ComponentState) then
    Exit;
  if AddAfter = nil then
    Exit;
  Base := AddAfter.MenuIndex+1;
  ParentMenu := AddAfter.Parent;
  Result := True;
end;

end.
