unit SizeHelpers;
{$I utils.inc}{ Copyright (C) 2016 C M Western; see PcontrolsReg.pas for license}
interface

uses
  Classes, Controls, ExtCtrls, ComCtrls, Forms;

procedure AutoSizeStatusBar(StatusBar: TStatusBar; const Contents: array of string; GrowOnly: Boolean = False);
procedure ConstantSizeButtons(Form: TCustomForm);

type
  TControlArray = array of TControl;

procedure WrapPanel(Panel: TPanel;
                    var Controls: TControlArray; { Used to record routine state }
                    Gap: Integer = -1; { Between controls }
                    Border: Integer = -1; { At left, top and bottom }
                    MinLast: Integer = -1; { If > 0, expand last control to available space; MinLast is then minimum size }
                    MaxRows: Integer = 2  { Maximum number of rows of controls }
                    );

implementation

uses {$IFDEF MSWINDOWS} Windows, Messages, CommCtrl, {$ENDIF}
     Buttons, Math, ShellSort, StdCtrls, NewSpin;

procedure AutoSizeStatusBar(StatusBar: TStatusBar; const Contents: array of string; GrowOnly: Boolean = False);
var
  Borders: array[0..2] of Integer;
  i, w: Integer;
begin
  with StatusBar do begin
    SimplePanel := False;
    Canvas.Font := Font;
    {$IFDEF MSWINDOWS}
      SendMessage(Handle, SB_GETBORDERS, 0, LPARAM(@Borders));
    {$ELSE}
      Borders[1] := 5; { Guess }
    {$ENDIF}
    for i := 0 to Panels.Count - 1 do begin
      w := Canvas.TextWidth(Panels[i].Text);
      if i <= High(Contents) then
        w := Max(w, Canvas.TextWidth(Contents[i]));
      w := w + Borders[1]*2 + 3;
      if GrowOnly then
        w := Max(w, Panels[i].Width);
      Panels[i].Width := w;
    end;
  end;
end;

type

  TControlSorter = class(TShellSorter)
    Controls: TControlArray;
    function InOrder(I,J:Integer):Boolean; override;
  end;

function TControlSorter.InOrder(I, J: Integer): Boolean;
begin
  Result := Controls[I].Left <= Controls[J].Left;
  if not Result then
    SwapValues(Controls[I], Controls[J], SizeOf(Controls[I]));
end;


function IsButton(Control: TControl):Boolean;
begin
  {$IF Defined(VCL_) and not Defined(UNICODE)}
    Result := (Control is TButton) or (Control is TSpeedButton);
  {$ELSE}
    Result := (Control is TCustomButton) or (Control is TSpeedButton);
  {$IFEND}
end;

procedure WrapPanel(Panel: TPanel; var Controls: TControlArray; Gap: Integer; Border: Integer; MinLast: Integer; MaxRows: Integer);
var
  n, i, j, W, H, Pass, nRows, cRow, RowH, LastGap, Hbutton: Integer;
  LastControl: TControl;
  NewControls: TControlArray;
begin
  { Assume initial panel setup is one long line of Controls:
    - save initial order
    - Automatically space out controls
    - Make status bar more or less lines if it fits }
  Panel.AutoSize := False;
  if Gap < 0 then
    Gap := 3;
  if Border < 0 then
    Border := 2;
  { Make up current list of controls }
  n := 0;
  with Panel.Owner do begin
    SetLength(NewControls, ComponentCount);
    for i := 0 to ComponentCount-1 do
      if (Components[i] is TControl) and ((Components[i] as TControl).Parent = Panel) then begin
        NewControls[n] := Components[i] as TControl;
        NewControls[n].Anchors := [];
        Inc(n);
      end;
  end;
  SetLength(NewControls, n);
  if n = Length(Controls) then begin
    { Check for missing controls }
    for i := 0 to High(Controls) do begin
      j := 0;
      while j < Length(NewControls) do begin
        if Controls[i] = NewControls[j] then
          Break;
        Inc(j);
      end;
      if j = Length(NewControls) then begin
        Controls := nil; { Control has changed }
        Break;
      end;
    end;
  end else
    Controls := nil; { Extra or deleted controls }
  if Length(Controls) = 0 then begin
    n := Length(NewControls);
    with TControlSorter.Create do begin
      Controls := NewControls;
      Sort(n);
      Free;
    end;
    Controls := NewControls;
  end;
  Hbutton := 0;
  for Pass := 1 to 2 do begin
    { Pass1: Find total width and maximum height }
    { Pass2: Position controls }
    cRow := 1;
    W := Border;
    H := 0;
    LastControl := nil;
    for i := 0 to n-1 do begin
      if not Controls[i].Visible then
        Continue;
      if Assigned(LastControl) and IsButton(LastControl) and IsButton(Controls[i]) then
        LastGap := 0
      else
        LastGap := Gap;
      Inc(W, LastGap);
      if Pass = 1 then begin
        H := Max(H, Controls[i].Height);
        if IsButton(Controls[i]) then begin
          Hbutton := Max(Hbutton, Controls[i].Height);
          {$IFNDEF VCL_}
            Controls[i].AutoSize := False;
          {$ENDIF}
        end;
      end else begin
        if IsButton(Controls[i]) then
          Controls[i].Height := Hbutton;
        if (nRows > 1) and (cRow < MaxRows) and (W+Controls[i].Width > Panel.Width) and (i > 2) and (LastControl <> nil) then begin
          { Consider wrapping }
          W := Border;
          Inc(cRow);
          Inc(H, RowH-Border);
          if Controls[i-1] is TLabel then begin
            { Do not break between label and following control }
            Controls[i-1].Left := W;
            Controls[i-1].Top := Controls[i-1].Top + RowH;
            Inc(W, Controls[i-1].Width+Gap);
          end;
        end;
        Controls[i].Left := W;
        Controls[i].Top := H + (RowH - Controls[i].Height) div 2;
      end;
      LastControl := Controls[i];
      W := W + Controls[i].Width;
      { For controls with sub-controls, assume they are spin or edit buttons,
        so add to the control width }
      for j := 0 to Controls[i].ComponentCount-1 do
        Inc(W, (Controls[i].Components[j] as TControl).Width);
    end;
    if Pass = 2 then begin
      if MinLast > 0 then
        LastControl.Width := Max(MinLast, Panel.Width-LastControl.Left-Border);
      Break;
    end;
    if MinLast > 0 then
      Dec(W, LastControl.Width-MinLast);
    RowH := H + 2*Border;
    if (W > Panel.Width + 2*Gap) and (MaxRows > 1) then begin
      { Need to split }
      nRows := 2; { or possibly more }
      Panel.Height := 2*RowH - Border;
    end else begin
      { Single line }
      Panel.Height := RowH;
      nRows := 1;
    end;
  end;
end;

procedure ConstantSizeButtons(Form: TCustomForm);
var
  i, w, h, Pass: Integer;
begin
  w := 0;
  h := 0;
  with Form do begin
    DisableAlign;
    for Pass := 1 to 2 do
      for i := 0 to ComponentCount-1 do
        if Components[i] is TButton then
          with TButton(Components[i]) do
            if Pass = 1 then begin
              w := Max(w, Width);
              h := Max(h, Height);
            end else begin
              {$IFDEF LCL}
              AutoSize := False;
              {$ENDIF}
              Width := w;
              Height := h;
            end;
    EnableAlign;
  end;
end;

end.

