unit WhatsThis;
{ Various enhancements to allow more automatic use of help:
   1. Allow use of the "What is This" feature of online help
   2. Automatically generate HelpKeyword from component name when both
      HelpContext and HelpKeyword are blank
   3. Automatically generate Topic+Bookmark URL for menu items
}
{$I utils.inc}{ Copyright (C) 2005 C M Western; see PcontrolsReg.pas for license}
interface

uses {$IFDEF MSWINDOWS} Windows, Messages, {$ENDIF}
     {$IFDEF LCL_} HelpIntfs, LazHelpIntf, LazHelpHTML, {$ENDIF}
     Classes, Forms, Menus, Controls, FixForm;

type

  { THelpForm }

  THelpForm = class(TFixForm)
  private
    {$IFDEF MSWINDOWS}
    procedure WMSetCursor(var Message: TWMSetCursor); message WM_SETCURSOR;
    procedure WMInitMenu(var Message: TMessage); message WM_INITMENU;
    procedure WMHelp(var Message: TMessage); message WM_HELP;
    {$ENDIF}
  protected
    procedure DoShow; override;
  public
    HasBeenArranged: Boolean; { Remember if ArrangeForm called }
    constructor Create(AOwner: TComponent); override;
    procedure ArrangeForm(Screen: TRect); virtual;
    procedure ShowMenuItemHelp(MenuItem: TMenuItem; TopItem: TComponent = nil); virtual;
    procedure ShowControlHelp(Control: TControl); virtual;
    procedure ShowHelp; {$IFDEF FPC} override; {$ENDIF}
  end;

function InHelp: Boolean;
procedure SetInHelp(Value: Boolean);

var
  InContextHelp: Boolean;

{$IFDEF LCL_}
type

  { TSingleHTMLHelpDatabase }

  TSingleHTMLHelpDatabase = class(THTMLHelpDatabase)
  public
    constructor Create(TheOwner: TComponent); override;
    destructor Destroy; override;
    function GetNodesForKeyword(const HelpKeyword: string;
                                var ListOfNodes: THelpNodeQueryList;
                                var ErrMsg: string): TShowHelpResult; override;
    procedure LookForF1(Sender: TObject; var Key: Word; Shift: TShiftState);
  end;
{$ENDIF}

implementation

uses {$IFDEF MSWINDOWS} Types, {$ENDIF}
     {$IFDEF LCL_} LCLType, LCLIntf, {$ENDIF}
     WCompat, SysUtils, UCompat;

type

  { TWhatsThisHelper }

  TWhatsThisHelper = class
  private
    FHooked: Boolean;
    OldOnHelp: THelpEvent;
    {$IFDEF MSWINDOWS}
    procedure OnMessage (var Msg: TMsg; var Handled: Boolean);
    function OnHelp(Command: Word; Data: SizeInt; var CallHelp: Boolean): Boolean;
    {$ENDIF}
  public
    destructor Destroy; override;
    procedure Hook;
    procedure UnHook;
  end;

var
  FInHelp: Boolean;
  WhatsThisHelper: TWhatsThisHelper;
  {$IFDEF VCL_}
  CurrentOnMessage: TMessageEvent;
  {$ENDIF}

function InHelp: Boolean;
begin
  Result := FInHelp;
end;

procedure SetInHelp(Value: Boolean);
begin
  if Value <> FInHelp then
    if Value then begin
      if not Assigned(WhatsThisHelper) then
        WhatsThisHelper := TWhatsThisHelper.Create;
      {$IFDEF VCL_}
      CurrentOnMessage := Application.OnMessage;
      Application.OnMessage := WhatsThisHelper.OnMessage;
      {$ENDIF}
      Screen.Cursor := crHelp;
      // Similar to: DefWindowProc(Handle, WM_SYSCOMMAND, SC_CONTEXTHELP, 0);
      // but this does not work (at least) for re-sizable forms.
      { TODO: Trap above message? or will everything happen automatically? }
    end else begin
      Screen.Cursor := crDefault;
      {$IFDEF VCL_}
      Application.OnMessage := CurrentOnMessage;
      {$ENDIF}
    end;
  FInHelp := Value;
end;

function FindFormMenuItemHelp(Form: TCustomForm; ID: Integer): TMenuItem;
var
  MenuItem: TMenuItem;
begin
  { NB: May also need to try Message.HelpInfo.hItemHandle for WM_HELP messages }
  Result := nil;
  if Assigned(Form.Menu) then
    Result := Form.Menu.FindItem(ID,fkCommand);
  if not Assigned(Result) and (Form is TForm) and Assigned(TForm(Form).PopupMenu) then
    Result := TForm(Form).PopupMenu.FindItem(ID,fkCommand);
  if Assigned(Result) then begin
    if Result.HelpContext = 0 then begin
      { See if any Parent items have a non-zero HelpContext }
      MenuItem := Result;
      while Assigned(MenuItem) and (MenuItem.HelpContext = 0) do
        MenuItem := MenuItem.Parent;
      if Assigned(MenuItem) then
        Result := MenuItem;
    end;
    if Result.HelpContext = 0 then
      if Form is THelpForm then
        THelpForm(Form).ShowMenuItemHelp(Result)
      else
        Application.HelpKeyword(Result.Name)
    else
      Application.HelpContext(Result.HelpContext);
  end;
end;

function ControlHasHelp(const Control: TControl) : Boolean;
begin
  Result := ((Control.HelpType = htContext) and (Control.HelpContext <> 0))
            or ((Control.HelpType = htKeyword) and (Control.HelpKeyword <> ''));
end;

{$IFDEF MSWINDOWS}
function FindFormWindowHelp(Form: TCustomForm; var HelpInfo: THelpInfo): TControl;
var
  WinControl: TWinControl;
  Control: TControl;
begin
  WinControl := FindControl(HelpInfo.hItemHandle);
  if Assigned(WinControl) then begin
    Result := WinControl.ControlAtPos(WinControl.ScreenToClient(HelpInfo.MousePos), {$IFDEF VCL_}True, True, True{$ELSE}[capfAllowDisabled,capfAllowWinControls,capfRecursive]{$ENDIF});
    if not Assigned(Result) then
      Result := WinControl;
  end else
    Result := nil;
  Control := Result;
  while Control <> nil do begin
    if ControlHasHelp(Control) then
      if not (Control is TWinControl) then begin
        { Delphi will not find correct help }
        if Control.HelpType = htContext then
          Application.HelpContext(Control.HelpContext)
        else
          Application.HelpKeyword(Control.HelpKeyword);
        Exit; { Handled }
      end else begin
        Result := nil; { Delphi should find correct help - flag not handled }
        Exit;
      end;
    Control := Control.Parent;
  end;
  { No help found - defaults are: }
  if Assigned(Result) then
    if Form is THelpForm then
      THelpForm(Form).ShowControlHelp(Result) { for a specific control }
    else
      Application.HelpKeyword(Result.Name)
end;
{$ENDIF}

{ TWhatsThisHelper }

destructor TWhatsThisHelper.Destroy;
begin
  UnHook;
  SetInHelp(False);
  inherited;
end;

procedure TWhatsThisHelper.Hook;
begin
  if not FHooked then begin
    {$IFDEF MSWINDOWS}
    OldOnHelp := Application.OnHelp;
    Application.OnHelp := {$IFDEF FPC}@{$ENDIF}OnHelp;
    {$ENDIF}
  end;
  FHooked := True;
end;

{$IFDEF MSWINDOWS}
function TWhatsThisHelper.OnHelp(Command: Word; Data: SizeInt; var CallHelp: Boolean): Boolean;
begin
  if (Command = HELP_COMMAND) and (StrPos(PChar(Data),'#') <> nil) then begin
    {$IFDEF VCL_}
    Application.HelpSystem.ShowTopicHelp(PChar(Data), Application.CurrentHelpFile);
    CallHelp := False;
    {$ENDIF}
    Result := True;
  end else if Assigned(OldOnHelp) then
    Result := OldOnHelp(Command, Data, CallHelp)
  else
    Result := False;
end;

procedure TWhatsThisHelper.OnMessage(var Msg: TMsg; var Handled: Boolean);
var
  WinControl: TWinControl;
  HelpInfo: THelpInfo;
begin
  case Msg.message of
    { The next two are duplicates of THelpForm messages; seems to be called via SendMessage, so not called here }
    WM_INITMENU:  begin Windows.SetCursor(Screen.Cursors[crHelp]); Handled := True; end;
    WM_SETCURSOR: if LoWord(Msg.lParam) in [HTMENU,HTCAPTION] then begin
                    Windows.SetCursor(Screen.Cursors[crHelp]);
                    Handled := True;
                  end;
    WM_LBUTTONDOWN, WM_RBUTTONDOWN,
    WM_MBUTTONDOWN: begin
                      SetInHelp(False);
                      Handled := True;
                      InContextHelp := True;
                      HelpInfo.cbSize := SizeOf(THelpInfo);
                      HelpInfo.iContextType := HELPINFO_WINDOW;
                      HelpInfo.iCtrlId := 0; { Not used by Delphi help system }
                      HelpInfo.hItemHandle := Msg.hwnd;
                      HelpInfo.dwContextId := 0; { Not used by Delphi help system }
                      {$IFDEF VCL_}
                      HelpInfo.MousePos := SmallPointToPoint(TSmallPoint(Integer(Msg.lParam)));
                      {$ENDIF}
                      Windows.ClientToScreen(Msg.hwnd, HelpInfo.MousePos);
                      Windows.SendMessage(Msg.hwnd, WM_HELP, 0, Integer(@HelpInfo));
                      InContextHelp := False;
{                      WinControl := FindControl(Msg.hwnd);
                      if Assigned(WinControl) then begin
                        Control := WinControl.ControlAtPos(SmallPointToPoint(TSmallPoint(Msg.lParam)),True,True,True);
                        if not Assigned(Control) then
                          Control := WinControl;
                        while not ControlHasHelp(Control) and Assigned(Control.Parent) do
                          Control := Control.Parent;
                        if Control.HelpType = htContext then
                          Application.HelpContext(Control.HelpContext)
                        else
                          Application.HelpKeyword(Control.HelpKeyword);
                      end; }
                    end;
    WM_COMMAND:     if Msg.lParam = 0 then begin
                      SetInHelp(False);
                      Handled := True;
                      WinControl := FindControl(Msg.hwnd);
                      if Assigned(WinControl) and (WinControl is TCustomForm) then
                        FindFormMenuItemHelp(TCustomForm(WinControl),LoWord(Msg.wParam));
                    end;
    WM_KEYDOWN:     if Msg.wParam = VK_ESCAPE then begin
                      { TODO: Escape does not work in menus }
                      SetInHelp(False);
                      Handled := True;
                    end;

  end;
  {$IFDEF VCL_}
  if not Handled and Assigned(CurrentOnMessage) then
    CurrentOnMessage(Msg, Handled);
  {$ENDIF}
end;
{$ENDIF}

procedure TWhatsThisHelper.UnHook;
begin
  if FHooked and Assigned(Application) then begin
    Application.OnHelp := OldOnHelp;
    OldOnHelp := nil;
  end;
  FHooked := False;
end;

{ THelpForm }

constructor THelpForm.Create(AOwner: TComponent);
begin
  inherited;
  if not Assigned(WhatsThisHelper) then
    WhatsThisHelper := TWhatsThisHelper.Create;
  WhatsThisHelper.Hook;
  {$IFDEF VCL_}
  ScreenSnap := True;
  {$ENDIF}
end;

procedure THelpForm.DoShow;
var
  ARect: TRect;
begin
  inherited;
  ARect := ScreenAreaFor(Application.MainForm);
  if Self <> Application.MainForm then
    with Application.MainForm do
      if WindowState = wsMaximized then begin
        WindowState := wsNormal;
        if Application.MainForm is THelpForm then
          THelpForm(Application.MainForm).ArrangeForm(ARect);
      end;
  if not HasBeenArranged then
    ArrangeForm(ARect);
  HasBeenArranged := True;
end;

procedure THelpForm.ArrangeForm(Screen: TRect);
var
  BottomMargin: Integer;
begin
  { No default arrangement }
  {$IFDEF LCLcarbon}
  { Force sane arrangement; required as poDefault does not work }
  BottomMargin := (Screen.Bottom - Screen.Top) div 8; { Allow for autohiding dock position }
  if Top < Screen.Top then
    Top := Screen.Top;
  if Top >= Screen.Bottom - BottomMargin then
    Top := Screen.Bottom - BottomMargin;
  if Left < Screen.Left then
    Left := Screen.Left;
  if Left >= Screen.Right then
    Left := Screen.Bottom;
  if Height >= Screen.Bottom - Screen.Top then
    Height := Screen.Bottom - Screen.Top - GetSystemMetrics(SM_CYCAPTION) ;
  if Width >= Screen.Right - Screen.Left then
    Width := Screen.Right - Screen.Left;
  {$ENDIF}
  {$IFNDEF MSWINDOWS}
    BringToFront;
  {$ENDIF}
end;

procedure THelpForm.ShowControlHelp(Control: TControl);
begin
  {$IFDEF VCL_}
  Application.HelpSystem.ShowTopicHelp(LowerCase(Name)+'.htm#'+Control.Name,Application.CurrentHelpFile);
  {$ELSE}
  if not Assigned(Control) then
    Application.HelpKeyword(LowerCase(Name)+'.htm')
  else if Control.HelpKeyword <> '' then
    Control.ShowHelp
  else
    Application.HelpKeyword(LowerCase(Name)+'.htm#'+Control.Name);
  {$ENDIF}
end;

procedure THelpForm.ShowHelp;
begin
  ShowControlHelp(ActiveControl);
end;

procedure THelpForm.ShowMenuItemHelp(MenuItem: TMenuItem; TopItem: TComponent);
begin
  if not Assigned(TopItem) then
    TopItem := MenuItem;
  while (TopItem is TMenuItem) and Assigned(TMenuItem(TopItem).Parent) and (TMenuItem(TopItem).Parent.Name <> '') do
    TopItem := TMenuItem(TopItem).Parent;
  {$IFDEF VCL_}
  Application.HelpSystem.ShowTopicHelp(TopItem.Name+'.htm#'+MenuItem.Name,Application.CurrentHelpFile);
  {$ENDIF}
end;

{$IFDEF MSWINDOWS}
procedure THelpForm.WMHelp(var Message: TMessage);
begin
  if not (csDesigning in ComponentState) then begin
    if (PHELPINFO(Message.lParam)^.iContextType = HELPINFO_MENUITEM)
       and Assigned(FindFormMenuItemHelp(Self, PHELPINFO(Message.lParam)^.iCtrlID)) then
      Exit; { Handled }
    if (PHELPINFO(Message.lParam)^.iContextType = HELPINFO_WINDOW)
       and Assigned(FindFormWindowHelp(Self, PHELPINFO(Message.lParam)^)) then
      Exit; { Handled }
  end;
  {$IFNDEF FPC}
    inherited ;
  {$ENDIF}
end;

procedure THelpForm.WMInitMenu(var Message: TMessage);
begin
  if InHelp then begin
    Windows.SetCursor(Screen.Cursors[crHelp]);
    Message.Result := 0;
  end;
end;

procedure THelpForm.WMSetCursor(var Message: TWMSetCursor);
begin
  if InHelp and (Message.HitTest in [HTMENU {,HTCAPTION }]) then begin
    Windows.SetCursor(Screen.Cursors[crHelp]);
    Message.Result := 1;
  end else
    inherited;
end;

{$IFDEF VCL_}
type
  TPopupListHelp = class(TPopupList)
    procedure WndProc(var Message: TMessage); override;
  end;

{ TPopupListHelp }

procedure TPopupListHelp.WndProc(var Message: TMessage);
var
  i: Integer;
  MenuItem: TMenuItem;
  TopItem: TMenu;
  Owner: TComponent;
begin
  if Message.Msg = WM_HELP then
    with PHelpInfo(Message.LParam)^ do
      if iContextType = HELPINFO_MENUITEM then
        for i := 0 to Count - 1 do begin
          if hItemHandle = TMenu(Items[i]).Handle then
            MenuItem := TMenu(Items[i]).Items
          else
            MenuItem := TPopupMenu(Items[i]).FindItem(hItemHandle, fkHandle);
          if MenuItem <> nil then begin
            Owner := MenuItem;
            MenuItem := TMenu(Items[i]).FindItem(iCtrlID, fkCommand);
            TopItem := nil;
            while not (Owner is TCustomForm) and Assigned(Owner.Owner) do begin
              Owner := Owner.Owner;
              if Owner is TMenu then
                TopItem := Owner as TMenu;
            end;
            if Owner is THelpForm then begin
               THelpForm(Owner).ShowMenuItemHelp(MenuItem,TopItem);
              Exit; { Handled }
            end;
          end;
        end;
  inherited;
end;

initialization
{$IFNDEF INPACKAGE}
  FreeAndNil(PopupList);
  PopupList := TPopupListHelp.Create;
{$ENDIF}
{$ENDIF}
{$ENDIF}

{$IFDEF LCL_}
{ TSingleHTMLHelpDatabase }

constructor TSingleHTMLHelpDatabase.Create(TheOwner: TComponent);
begin
  inherited Create(TheOwner);
  Application.AddOnKeyDownHandler(@LookForF1);
end;

destructor TSingleHTMLHelpDatabase.Destroy;
begin
  Application.RemoveOnKeyDownHandler(@LookForF1);
  inherited Destroy;
end;

function TSingleHTMLHelpDatabase.GetNodesForKeyword(const HelpKeyword: string;
  var ListOfNodes: THelpNodeQueryList; var ErrMsg: string): TShowHelpResult;
var
  url, anchor: string;
  p: Integer;
begin
  { Assume a single set of help files available, and always add prefix }
  p := Pos('#', HelpKeyword);
  if p > 0 then begin
    url := LowerCase(Copy(HelpKeyword,1,p-1));
    anchor := Copy(HelpKeyword,p,Length(HelpKeyword));
  end else begin
    url := LowerCase(HelpKeyword);
    anchor := '';
  end;
  if Pos('.',url) = 0 then
    url := LowerCase(url) + '.htm';
  repeat
    p := Pos(' ',url);
    if p > 0 then
      Delete(url,p,1);
  until p <= 0;
  if Pos('/',url) = 0 then
    url := KeywordPrefix + url;
  Result:=inherited GetNodesForKeyword(url+anchor, ListOfNodes, ErrMsg);
end;

procedure TSingleHTMLHelpDatabase.LookForF1(Sender: TObject; var Key: Word;
  Shift: TShiftState);
var
  Form: TComponent;
begin
if (Shift=[]) and (Key=VK_F1) and (Sender is TControl) then begin
    Form := TComponent(Sender);
    while Assigned(Form) and not (Form is TCustomForm) do
      Form := Form.Owner;
    if Form is THelpForm then begin
      THelpForm(Form).ShowControlHelp(Sender as TControl);
      Key := 0;
    end;
  end;
end;

{$ENDIF}

finalization
  FreeAndNil(WhatsThisHelper);
end.
