unit Vib;
{ Vibrational Energy Levels }
{Copyright (C) 2008-13 C M Western; see commandline.pas for license}
{$DEFINE NOAT}{$I utils.inc}

{ TODO: Plot normal modes }
{ TODO: Update DushinskyFrm if calculate following change in l matrix }
{ TODO: Should allow direct entry of Dushinsky matrix }

interface

uses SysUtils, Classes, MNode, Mixture, QuantumNumber, Math, MathSup, pgDOM,
     Abelian, { Note here in uses list so point group names not confused }
     MatrixVec, Vector, MatrixElements, QuantumMath, Interlock, Expr, Units,
     MNodeExpr, zMatrix;

type

  TPointGroup = (C1=0, Ci, C2, Cs, D2, C2v, C2h, D2h, CInfinityv, DInfinityh);

  TPointGroupInfo = class(TPersistent)
  private
    FPointGroup: TPointGroup;
    FLastSymmetry: Integer;
    FHasDegenerateModes: Boolean;
  public
    Name: string;
    SettingsNames: PPChar;
    constructor Create(APointGroup: TPointGroup);
    function Allowed(Bra, Dipole, Ket: Integer; HalfSpin: Integer):Boolean; virtual; abstract;
    function TryStrToSymmetry(Str: string; out Symmetry:Integer; HalfSpin: TQuantumNumber):Boolean; virtual;
    function StrToSymmetry(Str: string; HalfSpin: TQuantumNumber):Integer;
    function SymmetryToStr(Symmetry:Integer; HalfSpin: TQuantumNumber):string; virtual;
    property HasDegenerateModes:Boolean read FHasDegenerateModes;
    property LastSymmetry:Integer read FLastSymmetry;
  published
    property PointGroup: TPointGroup read FPointGroup; { To allow conversion to string }
  end;

  TVibratingMolecule = class;

  TBaseVibrationalMode = class(TMNode)
  protected
    function GetName: TComponentName; override;
    procedure SetName(const Value: TComponentName); override;
  public
    constructor Create(AOwner: TComponent); override;
    function FixupProperties: Exception; override;
    function Molecule: TVibratingMolecule;
    function ReadProperties(Node: TDOMNode; CheckUnread: Boolean = True):Integer; override;
    procedure SetSetting(AIndex: Integer; const AValue: Integer); override;
    function GetSettingsNames(AIndex: Integer; out LowestValue: Integer): PPChar; override;
    function SettingToStr(AIndex: Integer): string; override;
    procedure SetSettingFromStr(AIndex: Integer; const Str: string); override;
    function HelpOnSetting(AIndex: Integer): string; override;
  published
    property Symmetry: Integer index 1 read GetSetting write SetSetting;
  end;

  TElectronicState = class;

  TVibrationalOperator = (x, q, qSigned, Gaussian, qp);

  TVibrationalPerturbation = class (TPerturbation)
  private
    FMolecule: TVibratingMolecule;
    FTotalChange: Integer;
    FRTMode: Integer; { Mode containing a Renner-Teller type operator mixing electronic states }
    FPreviousValue: Double;
    procedure SetOperator(const Index: Integer; const AValue: TVibrationalOperator);
    function GetOperator(const Index: Integer): TVibrationalOperator;
  public
    constructor Create(AOwner: TComponent); override;
    constructor CreateUnOwned(ABra, AKet: TState);
    function FixupProperties: Exception; override;
    function PairName: string; override;
    procedure AdjustSettingNames;
    procedure SetParameter(AIndex: Integer; const AValue: Double); override;
    function MatrixElement(const ABra, AKet: TIntegerArray; SignA, SignB: Integer):Complex;
    function CheckMatrixElementsAvailable: Boolean; override;
    function SymbolicMatrixElements(WithHyperfine: Boolean):TMatrixElements; override;
    function EffectiveQuantumNumber(v: array of Integer):Integer;
    function HelpOnSetting(AIndex: Integer): string; override;
    function HelpOnParameter(AIndex: Integer): string; override;
    property Molecule: TVibratingMolecule read FMolecule;
    property PreviousValue: Double read FPreviousValue;
  published
    { Exponent is only used for Gaussian type perturbation }
    property Exponent:  Double index 2 read GetParameter write SetParameter;
    property Op:TVibrationalOperator    index LPI read GetOperator write SetOperator;
    property EffectiveQno:Boolean       index LPI+1 read GetFlag write SetFlag;
    property FromForceField:Boolean     index LPI+2 read GetFlag write SetFlag;
  end;

  TVibPertEquiv = record
    Op:TVibrationalOperator;
    Powers: array of Integer;
  end;

  { TVibrationalMode }

  TVibrationalMode = class(TBaseVibrationalMode)
  private
    FState: TElectronicState;
  public
    constructor Create(AOwner: TComponent); override;
    procedure AdjustParameterNames;
    function PertEquiv(AIndex: Integer):TVibPertEquiv;
    procedure GlobalShift(Shift:Integer);
    procedure GlobalMoveUp(Sender: TObject);
    procedure GlobalMoveDown(Sender: TObject);
    procedure Sort(Sender: TObject);
    procedure AddHandlers; override;
    function HelpOnSetting(AIndex: Integer): string; override;
    function HelpOnParameter(AIndex: Integer): string; override;
    function nCalculatedParameters: Integer; override;
    function CalculatedParameter(Index: Integer):Double; override;
    function CalculatedParameterName(Index: Integer):string; override;
    property State: TElectronicState read FState;
  published
    property Omega:    Double index  1 read GetParameter write SetParameter;
    property xOmega:   Double index  2 read GetParameter write SetParameter;
    property yOmega:   Double index  3 read GetParameter write SetParameter;
    property zOmega:   Double index  4 read GetParameter write SetParameter;
    property gll:      Double index  5 read GetParameter write SetParameter;
    property epsOmega: Double index  6 read GetParameter write SetParameter;

    property vMin: Integer     index 2 read GetSetting write SetSetting;
    property vMax: Integer     index 3 read GetSetting write SetSetting;
  end;

  TlMatrixMNode = class(TNotifyingActiveMNodeWithArray)
  private
    function GetzMatrix: TzMatrix; { Error reporting helper }
  protected
    FzMatrix: TzMatrix;
    CurrentAtoms: TIntegerArray;
  published
  public
    l: TMatrix;            { Only present if complete calculation possible }
    lambda: TColumnVector; { Only present if complete calculation possible }
    lInPrincipalAxes: Boolean;
    LastWarning: string;
    Generation: Integer;
    procedure ClearArrays; override;
    function Percent(n:Integer):string; override; { Expand %% strings }
    function ElectronicState: TElectronicState;
    procedure UpdateObjectChain; virtual;
    property zMatrix: TzMatrix read GetzMatrix;
  end;

  TRepresentationCD = record
    DJ, DJK, DK, d1, d2: Double;
  end;

  TCentrifugal = record
    DJ, DJK, DK, deltaJ, R5, R6: Double;
    A, S: TRepresentationCD; { Values for A and S representation }
  end;


  { TElectronicState }

  TElectronicState = class (TState)
  public
    const nFixedCalculatedParameters = 112;
  private
    FMolecule: TVibratingMolecule;
    FModes: TMNodeArray;
    FEffectiveQuanta: array of TVibrationalPerturbation;
    function GetzMatrix: TzMatrix;
  protected
    FCalculatedParameters: array [1..nFixedCalculatedParameters] of Double;
    FCalculatedParameterNames: array [1..nFixedCalculatedParameters] of string;
    FnCalculatedFixedParameters: Integer;
    FixedCalculatedParametersSetUp: Boolean;
    ForceFieldObjects: array [0..3] of TActiveMNodeWithArray;
    DummyObjects: array [0..3] of TActiveMNodeWithArray;
  public
    adm: TMatrix;     { adm = M^-1/2 l - one mode in each column;
                        not necessarily in principal inertial axis system, but using
                        co-ordinate system as given for individual nuclei;
                        M = diagonal matrix with nuclear mass on diagonal }
    lMatrix: TMatrix; { l matrix - in principal inertial axis system; one mode in each column }
    Madm: TMatrix;
//    CofM: array [1..3] of Double; { Centre of mass }
    Ixx: array [1..3] of Double; { Diagonal moments of inertia }
    RotConsts: array [1..3] of Double { Rotational constants / cm-1 };

    Coriolis: array [1..3] of TMatrix;
    dIdQ: array { Mode } of array [1..3,1..3] {xx yy zz xy xz yz} of Double;
    alpha: array { Mode + 1 } of array [1..3] {x y z} of Double; { Sum in last row }
    Delta0: Double; { Interial defect for zero point level }
    { Centrifugal distortion terms / kHz }
    tau: array [1..3,1..3,1..3,1..3] of Double;
    taupxy, taupxz, taupyz: Double;
    CD: array [1..6] of TCentrifugal;

    Precision: Integer; { Used to control significant figures in printout }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    class function NucleusType: TNucleusClass; override;
    procedure ClearArrays;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    function FixupProperties:Exception; override;
    function FixupForceFieldCalc(Verbose:Boolean):Exception;
    function ForceFieldNode(FClass: TActiveMNodeWithArrayClass):TActiveMNodeWithArray;
    function LastError: string;
    procedure ClearObjectChain;
    function UpdateObjectChain:string;
    function GetSettingsNames(AIndex: Integer; out LowestValue: Integer): PPChar; override;
    function SettingToStr(AIndex: Integer): string; override;
    procedure SetSettingFromStr(AIndex: Integer; const Str: string); override;
    function HelpOnSetting(AIndex: Integer): string; override;
    function HelpOnParameter(AIndex: Integer): string; override;

    function AddOneJ(Eigenvalues: TEigenvalues; J, F, M: TQuantumNumber; Sym: Integer; HyperfineState:THyperfineBasis; Seffective: TQuantumNumber): Integer; override;
    function RotationalMatrixElement(ABra, AKet: TBasisState; SignA, SignB: Integer; MNode: TMnode; Flag: Integer): Complex; override;
    function RotationalTransitionMatrixElement(ABra, AKet: TBasisState; SignA, SignB: Integer; TransitionMoment: TRootedMNodePair; F: TQuantumNumber; out dK: TQuantumNumber):Complex; override;
    function RotationalPerturbationMatrixElement(ABra, AKet: TBasisState; SignA, SignB: Integer; MNode: TMNode; Flag: Integer):Complex; override;
    function ReadBranchFormat(Text: string; var Obs: TObservation; UpperState: TState; Reader: TObservationsReaderBase):Boolean; override;
    function EstimatedStateOrigin: Double; override;
    function SymmetryOf(v: array of Integer; out AlternateSym:Integer):Integer;

    function ConvertParameter(AIndex: Integer; Zero: Boolean):TPerturbation; override;
    function MatrixElementType(const TheMatrixElement: TMatrixElement; const H: string; ShowPhases: Boolean = False):string; override;
    function SymbolicRotationalMatrixElement(AIndex: Integer; RankWanted: Integer):TMatrixElements; override; { For given parameter }
    function ParameterVariable(AIndex: Integer; Suffix: string = ''):TSortedVariableNode; override;
    function nParameterVariables: Integer; override;

    function nCalculatedParameters: Integer; override;
    function CalculatedParameter(Index: Integer):Double; override;
    function CalculatedParameterName(Index: Integer):string; override;
    procedure ACP(Value: Double; const PName: string);

    procedure AddHandlers; override;
    procedure PrintMolecularInfo(Sender: TObject);
    procedure FixCentreOfMass(Sender: TObject);
    procedure Normalize(Sender: TObject);
    procedure Orthogonalize(Sender: TObject);
    procedure PrepareMomentsOfInteria;
    procedure SetLmatrix(Sender: TObject);
    procedure AdjustModes(FixNumber: Boolean);

    function PrepareMatrices(Print: Boolean; AllModes:Boolean = True):Boolean;
    procedure PrepareCoriolis(Print: Boolean);
    procedure PrepareCentrifugal(Print: Boolean);
    procedure PrepareAlphas(Print: Boolean; SecondDerivsOnly: Boolean);
    function CheckOrthogonal(Threshold: Double; Print: Boolean):Double;
    procedure DoCheckOrthogonal(Threshold: Double);
    procedure ApplylMatrix(lMatrixNode: TlMatrixMNode; InPrincipalAxes: Boolean; Verbose: Boolean);

    function AddMode:TVibrationalMode;
    function Mode(Index:Integer): TVibrationalMode;
    function nameMode(Index: Integer):string;
    function nameNucleusXYZ(Index: Integer):string;
    property Molecule: TVibratingMolecule read FMolecule;
    property zMatrix: TzMatrix read GetzMatrix;
  published
    property A:        Double index  LSP   read GetParameter write SetParameter; { Spin orbit coupling }
    property LambdaSS: Double index  LSP+1 read GetParameter write SetParameter; { Spin-spin coupling }
    property o:        Double index  LSP+2 read GetParameter write SetParameter; { Omega = 0 lambda doubling }
    property Emax:     Double index  LSP+3 read GetParameter write SetParameter; { Energy cut off for basis }

    property Symmetry: Integer index LSI   read GetSetting write SetSetting;
  end;

  { TMolecularNucleus }

  TMolecularNucleus = class(TNucleus)
  protected
    procedure SetComponentIndex(const Value: Integer); override;
  public
    constructor Create(AOwner: TComponent); override;
    function ReadProperties(Node: TDOMNode; CheckUnread: Boolean = True):Integer; override;
    procedure AddHandlers; override;
    procedure DoFindExactMass(Sender: TObject);
    procedure SetParameter(AIndex: Integer; const Value: Double); override;
    procedure GlobalShift(Shift:Integer); override;
    function HelpOnParameter(AIndex: Integer): string; override;
  published
    property Mass:     Double index  1 read GetParameter write SetParameter;
    property x:        Double index  2 read GetParameter write SetParameter;
    property y:        Double index  3 read GetParameter write SetParameter;
    property z:        Double index  4 read GetParameter write SetParameter;
  end;

  TVibrationalManifold = class(TManifold)
  public
    function nSym: Integer; override; { Number of different symmetries }
    function FindJmax(Default: Integer; RemovedSpins: Byte = 0):Integer; override;
    procedure ShowRotationalConstants(Sender: TObject);
    procedure AddHandlers; override;
  end;

  TModeInfo = record
    Symmetry: Integer;
    Name: string;
    dIndex: Integer; { Index in basis set v array of l quantum number }
    Degeneracy: Integer; { use v + Degeneracy / 2 in quantum number expressions }
    ShowQno: Boolean; { If this quantum number is to be displayed }
  end;

  { TVibratingMolecule }

  TVibratingMolecule = class(TMolecule)
  private
    FNuclearMasses: array of Double;
    FnQuantumNumbers: Integer;
    FHasElectronicAngularMomentum: Boolean;
    FMass: Double;
    function GetPointGroup(const Index: Integer): TPointGroup;
    procedure SetPointGroup(const Index: Integer; const AValue: TPointGroup);
    procedure SetnModes(const Index, Value: Integer);
  protected
    procedure SetnNuclei(const Index, Value: Integer); override;
    procedure FixMass;
  public
    ModeInfo: array of TModeInfo;
    constructor Create(AOwner: TComponent); override;
    function IsValidChildClass(MNodeClass: TMNodeClass):Boolean; override;
    function FixupProperties:Exception; override;
    function Allowed(TransitionMoment: TTransitionMoment; Sym1, Sym2: Integer): Boolean; override;
    function JAdjustedSym(J: TQuantumNumber; Sym: Integer): Integer; override;
    function LastSym: Integer; override;
    function QnoInfo(QnoType: TQnoType): TQnoInfo; override;
    function SymName(Sym: Integer):string; override;
    function HelpOnParameter(AIndex: Integer): string; override;
    function HelpOnSetting(AIndex: Integer): string; override;
    function IsDegenerate(Symmetry: Integer; AHalfSpin: Integer):Boolean;
    function AngularMomentum(Symmetry: Integer; AHalfSpin: Integer):TQuantumNumber;
    procedure SetUpTransitionFunctionVariables(TransitionFunction: TTransitionFunction; var ExtraVariables: TVariableArray); override;
    procedure SetTransitionFunctionVariables(TransitionFunction: TTransitionFunction; var ExtraVariables: TVariableArray; ABra, AKet: TBasisState; SignA, SignB: Integer); override;
    procedure SetUpStateFunctionVariables(StateFunction: TStateFunction; var ExtraVariables: TVariableArray); override;
    procedure SetStateFunctionVariables(StateFunction: TStateFunction; var ExtraVariables: TVariableArray; AState: TBasisState); override;
    property nQuantumNumbers: Integer read FnQuantumNumbers;
    property HasElectronicAngularMomentum: Boolean read FHasElectronicAngularMomentum;
    property Mass: Double read FMass;
  published
    property SmallCoefficient: Double      index 3 read GetParameter write SetParameter;
    property PointGroup:       TPointGroup index LMI read GetPointGroup write SetPointGroup;
    property nModes:           Integer     index LMI+1 read GetSetting write SetnModes;
    property ShowAllQnos:      Boolean     index LMI+2 read GetFlag write SetFlag;
  end;

  { TVibrationalBasisState }

  TVibrationalBasisState = class(TBasisState)
  private
    FOriginalPosition: Integer; { Position in array sorted by vibrational quantum numbers }
    FOffset: Integer; { Position within basis }
  public
    v: TIntegerArray; { v1,l1, v2,l2, ... vn, ln, Lambda, Omega;
                        l only present for degenerate modes;
                        Lambda, Omega only present if any state in molecule has Lambda > 0 or S > 0 }
    function Qno(QnoType: TQnoType): TQuantumNumber; override;
    function VibLabel: string;
    function VibLabelKey: string;
    function BasisLabel: string; override; { Basis State Label }
    function BasisLabelKey: string; override; { Explanation of Basis State Label }
    function StateLabel(AQuantumNumberFormat: TQuantumNumberFormat): string; override;
    function StateLabelKey: string; override; { Explanation of State Name }
    function TxName(Upper, Lower: TBasisState): string; override;
    function TxNameKey(Upper, Lower: TBasisState): string; override;
    function MakeSignature(SignA:Integer; AKet: TBasisState; SignB: Integer; MNode: TMNode; Symmetrised: Boolean): TIntegerArray; override;
    procedure SetSymbolicVariables(SignA:Integer; AKet: TBasisState; SignB: Integer; MNode: TMNode); override;
    function SymbolicVariableValues: string; override;
    property OriginalPosition: Integer read FOriginalPosition;
    property Offset: Integer read FOffset;
  end;

  TVibrationalModeChange = class(TBaseVibrationalMode)
  public
    constructor Create(AOwner: TComponent); override;
    procedure ScaleParameters(UFrom, UTo: TResonanceUnits); override;
    procedure AdjustParameterNames;
    function CanShift(Shift:Integer):Boolean; override;
    function HelpOnParameter(AIndex: Integer): string; override;
  published
    property Displacement:  Double index  1 read GetParameter write SetParameter;
  end;

  TCachedTransitionMatrix = class(TInterlockedComponent)
  public
    BraSample,KetSample: TVibrationalBasisState; { Only valid during construction }
    BraOffsets, KetOffsets: TIntegerArray;
    TransitionMatrix: TMatrix;
    TransitionMoment: TTransitionMoment;
    TheCacheIndex: Integer;
    destructor Destroy; override;
    procedure DoMake(RequiredAttributes: Pointer); override;
    function HasCorrectAttributes(Attributes: Pointer):Boolean; override;
  end;

  { TMultidimensionalFCF }

  TMultidimensionalFCF = class(TSingleTransitionMoment)
  private
    FMolecule: TVibratingMolecule;
    FModes: TMNodeArray;
    FActualSymmetry: Integer;
    FMatrixCache: array of TCachedTransitionMatrix;
    CacheLock: TMultiReadExclusiveWriteSynchronizer;
    FBraState: TElectronicState;
    FKetState: TElectronicState;
    procedure ClearMatrices;
    procedure SaveOffsets(out Offsets: TIntegerArray; SourceSample: TVibrationalBasisState; CurrentCacheIndex: Integer);
    function CacheIndex(v: TIntegerArray):Integer;
  public
    d: TColumnVector;
    S, Q1, P1, R: TMatrix;
    Rdelta, P1Delta: TColumnVector;
    I00: Double;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function FixupProperties:Exception; override;
    function ReadProperties(Node: TDOMNode; CheckUnread: Boolean = True):Integer; override;
    procedure ClearCache(ClearChildren: Boolean); override;
    procedure PrepareDushinsky(nModes: Integer; Print: Boolean; SetModeInfo: Boolean);
    procedure PrepareMatrices(Print: Boolean; SetModeInfo, AllModes: Boolean);
    procedure ShowDushinsky(Sender: TObject);
    procedure AddHandlers; override;
    procedure PrintInfo(Sender: TObject);
    procedure SetModeInfo(Sender: TObject);
    function PairName: string; override;
    procedure AdjustModes(FixNumber: Boolean);
    procedure SwapModes(nMode1, nMode2: Integer);
    function Mode(Index:Integer): TVibrationalModeChange;
    function GetSettingsNames(AIndex: Integer; out LowestValue: Integer): PPChar; override;
    function SettingToStr(AIndex: Integer): string; override;
    function SettingToPrettyStr(AIndex: Integer): string; override;
    procedure SetSettingFromStr(AIndex: Integer; const Str: string); override;
    function HelpOnSetting(AIndex: Integer): string; override;

    function Overlap(var ABra, AKet: array of Integer):Double;
    procedure InvertTransformation;
    function TransitionMatrix(BraSample,KetSample: TVibrationalBasisState; Cache: Boolean):TMatrix;

    property Molecule: TVibratingMolecule read FMolecule;
    property ActualSymmetry: Integer read FActualSymmetry;
    property BraState: TElectronicState read FBraState;
    property KetState: TElectronicState read FKetState;
  published
    property Symmetry:  Integer index 1 read GetSetting write SetSetting;
    property Calculate: Boolean index 2 read GetFlag write SetFlag;
  end;

//  TauFactor: Double; { (10^21 h L / (4pi^2) )^3 /(10 c^2) = 1148947468038.87 - Note kHz units }
//  aFactor: Double; { (1e3 h L /c)^(3/2) /(8e-27 pi^2) = 195.76711230607512;
//                     Would be Sqrt(TauFactor), but units are cm-1, not kHz, so aFactor = sqrt(TauFactor*10/c) }

const
  LinearTest: Double = 1e-10; { Used for linearity tests }

implementation

uses {$IFNDEF CONSOLE} LmatrixFrm, DushinskyFrm, {$ENDIF}
     NuclearInfo, Worker, MatrixOps, TextStrm, Profile, ComplexMatrix, BigSet,
     Global, PropHelp, EigLA, Decomp, Progress, StaticField, InternalCoords,
     LevelLister, UCompat;

var
  PointGroupInfo: array [TPointGroup] of TPointGroupInfo;

  eLambda, eOmega: TSortedVariableNode;
  elVariables: TVariableList; { Contains eLambda, eOmega }
  vVariables, lVariables: TVariableList; { These contain v1, v2, ... and l1, l2, ..., filled as required using the functions below }

function FindVariable(Prefix: Char; Index: Integer; var VariableList: TVariableList):TSortedVariableNode;
var
  Name: string;
  i: Integer;
begin
  if not Assigned(VariableList) then
    VariableList := TVariableList.Create;
  Name := Prefix+IntToStr(Index+1);
  i := VariableList.IndexOf(Name);
  if i < 0 then { Just create as needed }
    VariableList.Add(TSortedVariable.Create(Name,Index));
  i := VariableList.IndexOf(Name);
  Result := (VariableList.Objects[i] as TSortedVariable).Template;
end;

function vVariable(Index: Integer):TSortedVariableNode;
begin
  Result := FindVariable('v',Index,vVariables);
end;

function lVariable(Index: Integer):TSortedVariableNode;
begin
  Result := FindVariable('l',Index,lVariables);
end;

function vnvp(v, n, vp: Integer):Double;
{ Matrix elements of <v|q^n|v'> in a harmonic basis }
var
  vv: Double;
begin
  Result := 0;
  if (Abs(v-vp) > n) or Odd(v+n+vp) then
    Exit;
  if n < 0 then
    raise NotYet;
  case n of
    0: Result := 1;
    1: Result := Sqrt(Max(v,vp)/2);
    2: if v = vp then
         Result := v + 0.5
       else begin
         vv := Max(v, vp);
         vnvp := Sqrt( vv*(vv-1) / 4 );
       end;
    3: begin
         vv := Max(v, vp);
         if Abs(v-vp) = 1 then
           Result := 3 * Sqrt( vv*vv*vv / 8 )
         else
           Result := Sqrt( vv*(vv-1)*(vv-2) / 8 );
       end;
    4: begin
         vv := Max(v, vp);
         if v = vp then
           Result := 3/4 * (2*vv*vv + 2*vv + 1)
         else if Abs(v-vp) = 2 then
           Result := (2*vv-1) * sqrt( vv*(vv-1) /4 )
         else
           Result := Sqrt( vv*(vv-1)*(vv-2)*(vv-3) / 16 )
       end;
    else
      { Use recursion: }
      { <v|q^n|v'> = sqrt((v+1)/2) * <v+1|q^n-1|v'> + sqrt(v/2) * <v-1|q^n-1|v'> }
      Result := Sqrt((v+1)/2) * vnvp(v+1,n-1,vp);
      if v > 0 then
        Result := Result + Sqrt(v/2) * vnvp(v-1,n-1,vp);
    end;
end;

function vnvpE(n: Integer; v: TExprNode; dv: Integer):TExprNode;
{ Symbolic Matrix elements of <v+dv|q^n|v> in a harmonic basis }
var
  nm: TExprNode;
begin
  Result := nil;
  if (Abs(dv) > n) or Odd(dv+n) then
    Exit;
  if n < 0 then
    raise NotYet;
  case n of
    0: Result := TIntegerNode.Create(1);
    1: begin
         if dv > 0 then
           Result := EB([v, '+', dv, '/', 2])
         else
           Result := EB([v, '/', 2]);
         Result := SqrtFunction.CreateExprNode(Result);
       end;
  else
    { Use recursion }
    nm := EB([v, '+', dv-1]);
    Result := EN([vnvpE(1,nm,1), '*', vnvpE(n-1,v,dv-1)]);
    nm.Free;
    nm := EB([v, '+', dv+1]);
    Result := EN([vnvpE(1,nm,-1), '*', vnvpE(n-1,v,dv+1), '+', Result]);
    nm.Free;
  end;
end;

function vnvpl(v, n, vp, ll: Integer):Double;
{ Matrix elements of <v ll|q^n|vp ll> in a 2D harmonic basis. }
var
  vv: Integer;
begin
  Result := 0;
  if (Abs(v-vp) > n) or Odd(v+n+vp) then
    Exit;
  if (n < 0) or Odd(n) then
    raise NotYet;
  case n of
    0: Result := 1;
    2: if v = vp then
         Result := v+1
       else begin
         vv := Max(v, vp);
         Result := Sqrt((vv*vv-ll*ll) / 4 );
       end;
  else { Recurse }
    vv := Max(v-2, Abs(ll));
    while vv <= v+2 do begin
      Result := Result + vnvpl(v,2,vv,ll)*vnvpl(vv,n-2,vp,ll);
      Inc(vv,2);
    end;
  end;
end;

function vnvplE(n: Integer; v, ll: TExprNode; dv: Integer):TExprNode;
{ Symbolic matrix elements of <v+dv ll|q^n|v ll> in a 2D harmonic basis. }
var
  ddv: Integer;
  np: TExprNode;
begin
  Result := nil;
  if (Abs(dv) > n) or Odd(dv+n) then
    Exit;
  if (n < 0) or Odd(n) then
    raise NotYet;
  case n of
    0: Result := TIntegerNode.Create(1);
    2: if dv = 0 then
         Result := EB([v, '+', 1])
       else begin
         if dv < 0 then
           Result := v.Duplicate
         else
           Result := EB([v, '+', dv]);
         Result := EN([SqrFunction.CreateExprNode(Result), '-', SqrFunction.CreateExprNode(ll.Duplicate), '/', 4]);
         Result := SqrtFunction.CreateExprNode(Result);
       end;
  else { Recurse }
    for ddv := -1 to 1 do begin
      np := EB([v, '+', dv+ddv*2]);
      Result := EN([vnvplE(2,np,ll,-ddv*2), '*', vnvplE(n-2,v,ll,dv+ddv*2), '+', Result]);
      np.Free;
    end;
  end;
end;

function vnvpll(vp, llp, n, v, ll: Integer):Double;
{ Matrix elements of <v ll|q+-^n|vp llp> in a 2D harmonic basis. }
var
  vi, li: Integer;
begin
  Result := 0;
  if (Abs(v-vp) > n) or Odd(v+n+vp) then
    Exit;
  if n = 0 then begin
    if ll = llp then
      Result := 1
  end else if n = 1 then begin
    { q+- }
    if vp = v+1 then begin
      if llp = ll+1 then
        Result := sqrt((v+2+ll)/2)
      else if llp = ll-1 then
        Result := sqrt((v+2-ll)/2)
    end else if vp = v-1 then begin
      if llp = ll+1 then
        Result := sqrt((v-ll)/2)
      else if llp = ll-1 then
        Result := sqrt((v+ll)/2)
    end;
  end else begin
    { General case by recursion }
    vi := vp-1;
    while vi <= vp+1 do begin
      if llp > ll then
        li := llp - 1
      else
        li := llp + 1;
      if Abs(li) <= vi THEN
        Result := Result + vnvpll(vp,llp,1,vi,li)*vnvpll(vi,li,n-1,v,ll);
      Inc(vi,2);
    end;
  end;
end;

function vnvpllE(n: Integer; v, ll:TExprNode; dv, dl: Integer):TExprNode;
{ Matrix elements of <v+dv ll+dl|q+-^n|v ll> in a 2D harmonic basis. }
var
  ddl: Integer;
  n1, n2, np: TExprNode;
begin
  Result := nil;
  if (Abs(dv) > n) or Odd(dv+n) then
    Exit;
  if n < 0 then
    raise NotYet;
  case n of
    0: if dl = 0 then
         Result := TIntegerNode.Create(1);
    1: if dv = -1 then begin
         if dl = -1 then
           Result := EB([v, '+', ll, '/', 2, 'sqrt'])
         else if dl = 1 then
           Result := EB([v, '-', ll, '/', 2, 'sqrt']);
       end else if dv = 1 then begin
         if dl = -1 then
           Result := EB([v, '-', ll, '+', 2, '/', 2, 'sqrt'])
         else if dl = 1 then
           Result := EB([v, '+', ll, '+', 2, '/', 2, 'sqrt']);
       end;
  else
    { General case by recursion }
    if dl > 0 then
      ddl := 1
    else
      ddl := -1;
    n1 := EB([v, '+', dv-1]);
    n2 := EB([v, '+', dv+1]);
    np := EB([ll, '+', dl-ddl]);
    Result := EN([vnvpllE(1,n1,np,1,ddl), '*', vnvpllE(n-1,v,ll,dv-1,dl-ddl)]);
    Result := EN([vnvpllE(1,n2,np,-1,ddl), '*', vnvpllE(n-1,v,ll,dv+1,dl-ddl), '+', Result]);
    n1.Free;
    n2.Free;
    np.Free;
  end;
end;

{ Matrix elements of exp(-bq^2)
  Use recursion formula in
  J. B. Coon, N. W. Naugle and R. D. McKenzie
  JMS 20, 107 (1966)
  See also (Rather slower):
  S.I. Chan and D. Stelman. J. Chem. Phys. 39, 545 (1963)
}
{ TODO: This routine becomes very slow for larger values of v;
  may need to cache values }
function vGaussian(v, vp:Integer; b:Double):Double;
//var
//  t: Integer;
//  function vGO(n, m:Integer):Double;
//  begin
//    if n = 0 then
//      if m = 0 then
//        Result := 1/Sqrt(1+b)
//      else
//        Result := -vGO(0,m-2)*Sqrt((m-1)/m)*b/(1+b)
//    else begin
//      Result := Sqrt(m)*vGO(n-1,m-1)+Sqrt(m+1)*vGO(n-1,m+1);
//      if n > 1 then
//        Result := (Result-Sqrt(n-1)*vGO(n-2,m))/Sqrt(n)
//    end;
//  end;
  function vG(m, n:Integer):Double;
  begin
    if n = 0 then
      if m = 0 then
        Result := 1/Sqrt(1+b)
      else
        Result := -vG(m-2,0)*Sqrt((m-1)/m)*b/(1+b)
    else if m = 0 then
        Result := -vG(0,n-2)*Sqrt((n-1)/n)*b/(1+b)
    else begin
      Result := Sqrt(n/m)/(1+b)*vG(m-1,n-1);
      if m > 1 then
        Result := Result - Sqrt((m-1)/m)*b/(1+b)*vG(m-2,n);
    end;
  end;
begin
  Result := 0;
  if Odd(v-vp) then
    Exit;
  if v > vp then
    Result := vG(v,vp)
  else
    Result := vG(vp,v);
end;

{$IFDEF TEST}
function vnvptest(v, n, vp: Integer):Double;
begin
  Result := 0;
  if (Abs(v-vp) > n) or Odd(v+n+vp) then
    Exit;
  case n of
    0: Result := 1;
    1: Result := Sqrt(Max(v,vp)/2);
    else
      { Use recursion: }
      { <v|q^n|v'> = sqrt((v+1)/2) * <v+1|q^n-1|v'> + sqrt(v/2) * <v-1|q^n-1|v'> }
      Result := Sqrt((v+1)/2) * vnvp(v+1,n-1,vp);
      if v > 0 then
        Result := Result + Sqrt(v/2) * vnvp(v-1,n-1,vp);
    end;
end;

procedure TestVib;
var
  v, v1, n: Integer;
  o, p: Double;
begin
  for v := 0 to 10 do
    for n := 0 to 10 do
      for v1 := 0 to 10 do begin
        o := vnvp(v,n,v1);
        p := vnvptest(v,n,v1);
        if not SameValue(o,p) then begin
          WriteLn(Log,v,n,v1,o,p);
          Abort;
        end;
      end;
end;
{$ENDIF}

{ TVibrationalPerturbation }

procedure TVibrationalPerturbation.AdjustSettingNames;
var
  i: Integer;
begin
  for i := nSettings+1 to FirstDynamicSetting + Molecule.nModes - 1 do
    AddSetting(Molecule.ModeInfo[i-FirstDynamicSetting].Name,'0');
  for i := nSettings downto FirstDynamicSetting + Molecule.nModes do
    DeleteSetting(i);
  for i := FirstDynamicSetting to nSettings do
    SetSettingName(i, Molecule.ModeInfo[i-FirstDynamicSetting].Name);
end;

function TVibrationalPerturbation.CheckMatrixElementsAvailable: Boolean;
begin
  if Op = Gaussian then
    Result := False { Not implementable in closed form }
  else
    Result := inherited CheckMatrixElementsAvailable;
end;

constructor TVibrationalPerturbation.Create(AOwner: TComponent);
begin
  if Assigned(AOwner) then
    FMolecule := AOwner.Owner as TVibratingMolecule;
  inherited;
  if Assigned(AOwner) then
    AdjustSettingNames;
end;

constructor TVibrationalPerturbation.CreateUnOwned(ABra, AKet: TState);
begin
  inherited Create(nil);
  FMolecule := ABra.Molecule as TVibratingMolecule;
  FBra := ABra;
  FKet := AKet;
  AdjustSettingNames;
end;

function TVibrationalPerturbation.EffectiveQuantumNumber(v: array of Integer): Integer;
var
  m, p: Integer;
begin
  Result := 0;
  for m := 1 to Molecule.nModes do begin
    p := GetSetting(FirstDynamicSetting+m-1);
    if p <> 0 then
      Result := Result + v[m-1]*(FTotalChange-Abs(p));
  end;
end;

function TVibrationalPerturbation.FixupProperties: Exception;
var
  m, p: Integer;
begin
  Result := inherited FixupProperties;
  AdjustSettingNames;
  FTotalChange := 0;
  FRTMode := -1;
  for m := 0 to Molecule.nModes-1 do begin
    if Assigned(Result) then
      Break;
    p := GetSetting(FirstDynamicSetting+m);
    Inc(FTotalChange,Abs(p));
    if (Op = Gaussian) and (p <> 0) then begin
      if (FRTMode >= 0) or Molecule.IsDegenerate(Molecule.ModeInfo[m].Symmetry,0) then
        Result := EReportException.Create('Only 1D Gaussian barrier implemented in '+Name);
      FRTMode := m;
    end else if (p < 0) and (Op = q) then begin
      if FRTMode >= 0 then
        Result := EReportException.Create('Only one Renner-Teller mode implemented in '+Name)
      else if not Molecule.IsDegenerate(Molecule.ModeInfo[m].Symmetry,0) then
        Result := EReportException.Create('Renner-Teller operator must involve degenerate mode in '+Name)
      else
        FRTMode := m;
    end else if (Op = qp) and Molecule.IsDegenerate(Molecule.ModeInfo[m].Symmetry,0) then
      Result := EReportException.Create('qp perturbation type not implemented for degenerate modes in '+Name);
  end;
  if EffectiveQno and (FTotalChange = 1) then
    FTotalChange := 2; { This allows a single quantum number to be selected }
end;

function TVibrationalPerturbation.GetOperator(const Index: Integer): TVibrationalOperator;
begin
  Result := TVibrationalOperator(GetSetting(Index));
end;

const
 VibrationalPerturbationParameterNames: array [2..2] of string = (
'Exponent for Gaussian type perturbation.'
);

function TVibrationalPerturbation.HelpOnParameter(AIndex: Integer): string;
begin
  if (AIndex >= Low(VibrationalPerturbationParameterNames))
     and (AIndex <= High(VibrationalPerturbationParameterNames)) then
    Result := VibrationalPerturbationParameterNames[AIndex]
  else
    Result := inherited HelpOnParameter(AIndex);
end;

const
 VibrationalPerturbationSettingNames: array [LPI..LPI+2] of string = (
'Type of perturbation.',
'Set to group together states connected by this perturbation when assigning quantum numbers.',
'If set from force field specified in an Internal or Symmetry coordinates object'
);

function TVibrationalPerturbation.HelpOnSetting(AIndex: Integer): string;
begin
  if AIndex > High(VibrationalPerturbationSettingNames) then
    Result := 'Power of ' + (Bra as TElectronicState).Mode(AIndex-High(VibrationalPerturbationSettingNames)-1).Name
              + ' in perturbation.'
  else if AIndex >= Low(VibrationalPerturbationSettingNames) then
    Result := VibrationalPerturbationSettingNames[AIndex]
  else
    Result := inherited HelpOnSetting(AIndex);
end;

procedure TVibrationalPerturbation.SetOperator(const Index: Integer; const AValue: TVibrationalOperator);
begin
  SetSetting(Index, Ord(AValue));
end;

procedure TVibrationalPerturbation.SetParameter(AIndex: Integer; const AValue: Double);
begin
  { Allows "Other Potential Terms" to be set to zero and recalculated during
    FixupProperties }
  if FromForceField and (AIndex = 1) then
    { Put back previous value, without triggering change actions }
    ModifyParameter(1, FPreviousValue, False, False);
  inherited SetParameter(AIndex, AValue);
  if AIndex = 1 then  
    FPreviousValue := Value;
end;

function SymbolicMatrixElements(PertEquiv: TVibPertEquiv; Molecule: TVibratingMolecule): TMatrixElements;
var
  RTMode: Integer;
  procedure AddVibChanges(vn, p: Integer; InitialPhase: Integer);
    { Add all possible changes for given vibrational mode }
    { p is power of operator, and possible changes are -p .. p in steps of 2 }
  var
    Old, nNew, i, j, Base, dind, iPower: Integer;
    Expr: TExprNode;
  begin
    Old := Length(Result);
    nNew := Abs(p)+1; { Number of different matrix elements from this mode }
    if Old = 0 then
      SetLength(Result, nNew)
    else
      SetLength(Result, Length(Result)*nNew);
    for i := Max(Old-1,0) downto 0 do begin
      { For each old set of quantum numbers, duplicate and concatenate new set of changes }
      Base := i*nNew; { Start of block to be updated }
      if vn = 0 then begin
        SetLength(Result[0].Signature,Molecule.nQuantumNumbers);
        Result[0].Expr := TIntegerNode.Create(InitialPhase);
      end;
      for j := Base+nNew-1 downto Base+1 do begin
        Result[j].Signature := Copy(Result[i].Signature);
        if Assigned(Result[i].Expr) then
          Result[j].Expr := Result[i].Expr.Duplicate;
      end;
      if Base <> i then { This test should not be required, but FPC 2.2.3 generated bad code for this }
        Result[Base].Signature := Result[i].Signature; { Can use current memory for one }
      Result[Base].Expr := Result[i].Expr;
      for j := Base to Base+nNew-1 do begin
        Result[j].Signature[vn] := -Abs(p) + (j-Base)*2;
        if ((PertEquiv.Op = qSigned) and (Abs(p) <> Abs(Result[j].Signature[vn])))
           or not Assigned(Result[j].Expr) then begin
          { This filters out some of the extra elements for qSigned, those with
            the wrong magnitude of the change in v }
          FreeAndNil(Result[j].Expr);
          Expr := nil;
        end else begin
          if Molecule.IsDegenerate(Molecule.ModeInfo[vn].Symmetry,0) then begin
            dind := Molecule.ModeInfo[vn].dIndex;
            if (p < 0) and (PertEquiv.Op = q) then begin
              if RTMode < 0 then
                RTMode := vn
              else if RTMode <> vn then
                raise EReportException.Create('Mode than one RT active mode in operator');
              { Delta Lambda = - Delta p = +/- p }
              Result[j].Signature[dind] := -p;
              Result[j].Signature[Molecule.nQuantumNumbers-2] := p;
              Result[j].Signature[Molecule.nQuantumNumbers-1] := p; { Delta sigma = 0 }
              Expr := vnvpllE(Abs(p),vVariable(vn),lVariable(vn),Result[j].Signature[vn],-p);
              { Need to duplicate this with opposite signs - see DuplicateRTChanges }
            end else begin
              Expr := vnvplE(Abs(p),vVariable(vn),lVariable(vn),Result[j].Signature[vn]);
              Result[j].Signature[dind] := 0;
            end;
          end else begin
            Expr := vnvpE(Abs(p),vVariable(vn),Result[j].Signature[vn]);
            if (PertEquiv.op = qp) and (p > 0) then begin
              { Phase correction needed for power of p }
              iPower := -Result[j].Signature[vn];
              if iPower < 0 then
                iPower := -3*iPower;
              if (iPower and 1) <> 0 then
                if Molecule.Mixture.AllowComplex then
                  Expr := EN([-1, 'sqrt', '*', Expr])
                else
                  Inc(iPower); { Multiply imaginary matrix elements by i }
              if (iPower and 2) <> 0 then
                Expr := EN([-1, '*', Expr]);
            end;
          end;
          Expr := EN([Result[j].Expr, '*', Expr]);
        end;
        Result[j].Expr := Expr;
      end;
    end;
  end;
  procedure CheckSigns;
  var
    r, w, vn, p, RelSign: Integer;
  begin
    { Check for relative signes of changes for qSigned;
      also remove completely zero elements }
    w := 0;
    for r := 0 to High(Result) do begin
      if Assigned(Result[r].Expr) then begin
        RelSign := 0;
        for vn := 0 to Molecule.nModes-1 do begin
          p := PertEquiv.Powers[vn];
          if p <> 0 then begin
            if RelSign = 0 then
              if p*Result[r].Signature[vn] > 0 then
                RelSign := 1
              else
                RelSign := -1;
            if p * RelSign <> Result[r].Signature[vn] then
              FreeAndNil(Result[r].Expr);
          end;
        end;
        if Assigned(Result[r].Expr) then begin
          Result[w] := Result[r];
          Inc(w);
        end;
      end;
    end;
    SetLength(Result, w);
  end;
  procedure DuplicateRTChanges;
  var
    Old, r, p: Integer;
  begin
    { Each matrix element needs to be duplicated with opposite sign changes in the angular momenta }
    Old := Length(Result);
    SetLength(Result, Old*2);
    for r := Old to High(Result) do begin
      Result[r].Signature := Copy(Result[r-Old].Signature);
      p := -Result[r].Signature[Molecule.nQuantumNumbers-2];
      Result[r].Signature[Molecule.ModeInfo[RTMode].dIndex] := -p;
      Result[r].Signature[Molecule.nQuantumNumbers-2] := p;
      Result[r].Signature[Molecule.nQuantumNumbers-1] := p;
      Result[r].Expr := vnvpllE(Abs(p),vVariable(RTMode),lVariable(RTMode),Result[r].Signature[RTMode],-p);
    end;
  end;
  function xElements: TExprNode;
  var
    vn, p: Integer;
  begin
    Result := TIntegerNode.Create(1);
    for vn := 0 to Molecule.nModes-1 do begin
      p := PertEquiv.Powers[vn];
      if p > 0 then
        Result := EN([Result, '*', EB([Molecule.ModeInfo[vn].Degeneracy, '/', 2, '+', vVariable(vn), '^', p])]);
    end;
  end;
var
  vn, InitialPhase: Integer;
begin
  SetLength(Result,0);
  case PertEquiv.Op of
    x:begin
        SetLength(Result,1); { No change in any quantum number }
        SetLength(Result[0].Signature,Molecule.nQuantumNumbers);
        Result[0].Expr := xElements;
      end;
    q, qSigned, qp:
      begin
         RTMode := -1;
         InitialPhase := 1;
         if (PertEquiv.op = qSigned) and not Molecule.Mixture.AllowComplex then begin
           { Count the number of odd powers of p }
           InitialPhase := 0;
           for vn := 0 to Molecule.nModes-1 do
             if (PertEquiv.Powers[vn] > 0) and Odd(PertEquiv.Powers[vn]) then
               Inc(InitialPhase);
           { Adjust so overall matrix element multiplied by i, not each individual one }
           { Also corrects case where there are an even number of odd powers of p, which gives a real matrix element }
           if Odd(InitialPhase div 2) then
             InitialPhase := -1
           else
             InitialPhase := 1;
         end;
         for vn := 0 to Molecule.nModes-1 do
           AddVibChanges(vn, PertEquiv.Powers[vn], InitialPhase);
         if PertEquiv.Op = qSigned then
           CheckSigns
         else if RTMode >= 0 then
           DuplicateRTChanges;
      end;
    // Gaussian: ;
  else
    raise NotYet;
  end;
end;

function TVibrationalPerturbation.MatrixElement(const ABra, AKet: TIntegerArray; SignA, SignB: Integer): Complex;
var
  p, vn, dind, dv, iPower: Integer;
  RelSign, DLambda: Integer;
begin
  Result.re := Value;
  Result.im := 0;
  case Op of
    x: if (ABra = AKet) and (SignA = SignB) then
         for vn := 0 to Molecule.nModes-1 do begin
           p := GetSetting(FirstDynamicSetting+vn);
           if p > 0 then
             Result.re := Result.re * Math.IntPower(AKet[vn]+Molecule.ModeInfo[vn].Degeneracy/2,p);
           if Result.re = 0 then
             Exit;
         end
       else
         Result.re := 0;
    q, qSigned, qp:
       begin
         { Check electronic part first }
         if Molecule.HasElectronicAngularMomentum then begin
           p := High(ABra)-1;
           if SignA*(ABra[p+1]-ABra[p]) <> SignB*(AKet[p+1]-AKet[p]) then
             { Sigma must not change }
             Result.re := 0;
           DLambda := (SignA*ABra[p]-SignB*AKet[p]) div 2;
           if FRTMode >= 0 then begin
             { Require Delta lambda = - Delta l = +-p }
             { First work out delta lambda }
             p := Abs(GetSetting(FirstDynamicSetting+FRTMode)); { Power of operator }
             if p <> Abs(DLambda) then
               Result.re := 0
             else begin
               { check delta l }
               dind := Molecule.ModeInfo[FRTMode].dIndex;
               if (SignA*ABra[dind]-SignB*AKet[dind]) <> -DLambda then
                 Result.re := 0;
             end;
           end else if DLambda <> 0 then
             Result.re := 0; { lambda must not change }
           if Result.re = 0 then
             Exit;
         end;
         RelSign := 0;
         iPower := 0;
         for vn := 0 to Molecule.nModes-1 do begin
           p := GetSetting(FirstDynamicSetting+vn);
           if (p <> 0) and (Op = qSigned) then begin
             { Require specific change in v }
             if Abs(ABra[vn]-AKet[vn]) = Abs(p) then
               if RelSign = 0 then
                 if (ABra[vn] - AKet[vn])*p > 0 then
                   RelSign := 1
                 else
                   RelSign := -1;
             if (ABra[vn]-AKet[vn])*RelSign <> p then begin
               Result.re := 0;
               Exit;
             end;
             p := Abs(p);
           end;
           if Molecule.IsDegenerate(Molecule.ModeInfo[vn].Symmetry,0) then begin
             dind := Molecule.ModeInfo[vn].dIndex;
             if p < 0 then
               Result.re := Result.re * vnvpll(ABra[vn],SignA*ABra[dind],Abs(p),AKet[vn],SignB*AKet[dind])
             else if SignA*ABra[dind] = SignB*AKet[dind] then
               Result.re := Result.re * vnvpl(ABra[vn],p,AKet[vn],SignB*AKet[dind])
             else
               Result.re := 0;
           end else
             Result.re := Result.re * vnvp(ABra[vn],Abs(p),AKet[vn]);
           if Result.re = 0 then
             Exit;
           if (Op = qp) and (p > 0) then begin
             { Accumulate phase correction for p operators }
             dv := ABra[vn]-AKet[vn];
             if dv <= 0 then
               iPower := iPower - dv
             else
               iPower := iPower + 3*dv;
           end;
         end;
         CMuliPower(Result, iPower);
       end;
    Gaussian:
       begin
         { Electronic part must not change }
         if Molecule.HasElectronicAngularMomentum then begin
           p := High(ABra)-1;
           if (SignA*ABra[p] <> SignB*AKet[p]) or (SignA*ABra[p+1] <> SignB*AKet[p+1]) then begin
             Result.re := 0;
             Exit;
           end;
         end;
         for vn := 0 to Molecule.nModes-1 do begin
           p := GetSetting(FirstDynamicSetting+vn);
           if p <> 0 then
             Result.re := Result.re * vGaussian(ABra[vn],AKet[vn],Exponent)
           else if ABra[vn] <> AKet[vn] then begin
             Result.re := 0;
             Exit;
           end else if Molecule.IsDegenerate(Molecule.ModeInfo[vn].Symmetry,0) then begin
             dind := Molecule.ModeInfo[vn].dIndex;
             if SignA*ABra[dind] <> SignB*AKet[dind] then begin
               Result.re := 0;
               Exit;
             end;
           end;
         end;
       end;
  end;
end;

function TVibrationalPerturbation.SymbolicMatrixElements(WithHyperfine: Boolean): TMatrixElements;
var
  PertEquiv: TVibPertEquiv;
  vn: Integer;
begin
  if SymSelect >= 0 then
    raise NotYet;
  PertEquiv.Op := Op;
  SetLength(PertEquiv.Powers, Molecule.nModes);
  for vn := 0 to Molecule.nModes - 1 do
    PertEquiv.Powers[vn] := GetSetting(FirstDynamicSetting+vn);
  Result := Vib.SymbolicMatrixElements(PertEquiv, Molecule);
end;

function TVibrationalPerturbation.PairName: string;
var
  m, p: Integer;
begin
  case op of
    x: begin
         Result := 'x';
         for m := 1 to Molecule.nModes do begin
           p := GetSetting(FirstDynamicSetting+m-1);
           if p > 0 then
             if m < 10 then
               Result := Result + StringOfChar(IntToStr(m)[1],p)
             else begin
               Result := Result + '(' + IntToStr(m);
               if p > 1 then
                 Result := Result + '^' + IntToStr(p);
               Result := Result + ')';
             end;
         end;
       end;
    q: begin
         Result := '';
         for m := 1 to Molecule.nModes do begin
           p := GetSetting(FirstDynamicSetting+m-1);
           if p <> 0 then begin
             Result := Result + 'q' + IntToStr(m);
             if Abs(p) > 1 then
               Result := Result + '^' + IntToStr(Abs(p));
             if p < 0 then
               Result := Result + 'L^' + IntToStr(Abs(p));
           end;
         end;
       end;
    qp:begin
         Result := '';
         for m := 1 to Molecule.nModes do begin
           p := GetSetting(FirstDynamicSetting+m-1);
           if p <> 0 then begin
             if p < 0 then
               Result := Result + 'q'
             else
               Result := Result + 'p';
             Result := Result + IntToStr(m);
             if Abs(p) > 1 then
               Result := Result + '^' + IntToStr(Abs(p));
           end;
         end;
       end;
    qSigned:
       begin
         Result := '';
         for m := 1 to Molecule.nModes do begin
           p := GetSetting(FirstDynamicSetting+m-1);
           if p <> 0 then
             Result := Result + Format('v%d(%d)',[m,p]);
         end;
       end;
    Gaussian:
       begin
         Result := '';
         for m := 1 to Molecule.nModes do begin
           p := GetSetting(FirstDynamicSetting+m-1);
           if p <> 0 then
             if Result <> '' then
               Result := '??'
             else
               Result := Result + Format('exp(-q%d^2)',[m]);
         end;
       end;
  end;
end;

{ TVibratingMolecule }

function TVibratingMolecule.Allowed(TransitionMoment: TTransitionMoment; Sym1, Sym2: Integer): Boolean;
begin { Sym1, Sym2 can't include spin }
  if PointGroupInfo[PointGroup].HasDegenerateModes then
    Result := Abs(AngularMomentum(Sym1,0)-AngularMomentum(Sym2,0)) <= TransitionMoment.Rank
  else
    Result := True; { TODO: fix }
end;

procedure TVibratingMolecule.FixMass;
var
  i: Integer;
begin
  FMass := 0;
  for i := 0 to nNuclei - 1 do
    FMass := FMass + FNuclearMasses[i];
end;

function TVibratingMolecule.FixupProperties: Exception;
  procedure ForAllVibrationalPerturbations(Pass: Integer);
  var
    i, j: Integer;
  begin
    { This routine allows the "Other Potential Terms", such as Jahn-Teller terms
      to function correctly, specifically the calculated perturbations go if the exta
      potantial terms do. }
    for i := 0 to ComponentCount - 1 do
      if Components[i] is TVibrationalManifold then
        for j := 0 to Components[i].ComponentCount - 1 do
          if Components[i].Components[j] is TVibrationalPerturbation then
            with TVibrationalPerturbation(Components[i].Components[j]) do
              if FromForceField then
                if Pass = 1 then begin
                  { Clear parameter at start of fixup cycle }
                  FixupName.Free; { Required on loading }
                  ModifyParameter(1, 0, False, False);
                end else
                  Value := Value; { Trigger Modified/Notify ony if parameter changed by fixup }
  end;
var
  i: Integer;
begin
  nNuclei := nNuclei; { Trigger add/remove of nuclei as required }
  nModes := nModes; { Ensure quantum number info up to date }
  ForAllVibrationalPerturbations(1);
  Result := inherited FixupProperties;
  ForAllVibrationalPerturbations(2);
  for i := 0 to ComponentCount - 1 do
    if Components[i] is TTransitionMoments then
      TTransitionMoments(Components[i]).DisableRankCheck := True;
end;

function TVibratingMolecule.GetPointGroup(const Index: Integer): TPointGroup;
begin
  Result := TPointGroup(GetSetting(Index));
end;

const
  VibratingMoleculeSettingNames: array[LMI..LMI+2] of string = (
'Point group of molecule.',
'Number of vibrational modes to include in the calculation.',
'Set false to hide vibrational quantum numbers with vmin=vmax=0'
);

function TVibratingMolecule.HelpOnParameter(AIndex: Integer): string;
begin
  if Aindex = 3 then
    Result := 'Threshold for discarding normal modes and other values'
  else
    Result := inherited HelpOnParameter(AIndex);
end;

function TVibratingMolecule.HelpOnSetting(AIndex: Integer): string;
begin
  if (AIndex >= Low(VibratingMoleculeSettingNames))
     and (AIndex <= High(VibratingMoleculeSettingNames)) then
    Result := VibratingMoleculeSettingNames[AIndex]
  else if AIndex = IndexOfSetting('nNuclei') then
    Result := 'Number of atoms; Only required for l matrix calculations'
  else if AIndex = IndexOfSetting('Jmin') then
    Result := 'Not used'
  else if AIndex = IndexOfSetting('Jmax') then
    Result := 'Maximum vibrational angular momentum'
  else
    Result := inherited HelpOnSetting(AIndex);
end;

function TVibratingMolecule.AngularMomentum(Symmetry: Integer; AHalfSpin: Integer): TQuantumNumber;
begin
  { Note HalfSpin is used to distinguish between cases where symmetry includes or excludes spin }
  case PointGroup of
    CInfinityv: if AHalfSpin = 0 then
                  Result := Max(Symmetry-1,0)*2 { Sigma+, Sigma-, Pi, ... }
                else
                  Result := 2*Symmetry+1; { 1/2, 3/2, 5/4}
    DInfinityh: if AHalfSpin = 0 then
                  Result := Max(Symmetry div 2 - 1,0)*2
                else
                  Result := 2*(Symmetry div 2)+1;
  else
    Result := 0;
  end;
end;

constructor TVibratingMolecule.Create(AOwner: TComponent);
begin
  SavedProperties := SavedProperties + ',ShowAllQnos=True,SmallCoefficient=5e-15';
  inherited;
end;

function TVibratingMolecule.IsDegenerate(Symmetry: Integer; AHalfSpin: Integer): Boolean;
begin
  Result := AngularMomentum(Symmetry, AHalfSpin) > 0;
end;

function TVibratingMolecule.IsValidChildClass(MNodeClass: TMNodeClass): Boolean;
begin
  Result := inherited IsValidChildClass(MNodeClass);
  if Result and MNodeClass.InheritsFrom(TTransitionMoment) then
    Result := MNodeClass = TMultidimensionalFCF;
end;

function TVibratingMolecule.JAdjustedSym(J: TQuantumNumber; Sym: Integer): Integer;
begin
  Result := Sym;
end;

function TVibratingMolecule.LastSym: Integer;
begin
  Result := PointGroupInfo[PointGroup].LastSymmetry;
  if Result = High(Result) then
    Result := JMax
  else
    Exit; { Not a linear point group }
  if (Result < 0) and (Owner is TSpecies) then
    Result := Species.Jmax;
  if Result < 0 then
    Result := 20;
  Result := Result div 2;
end;

function TVibratingMolecule.QnoInfo(QnoType: TQnoType): TQnoInfo;
var
  i: TQnoType;
  vn: Integer;
begin
  { qtA is v1;
    order is v1, (l1), v2, (l2), .... Lambda, Omega }
  if QnoType < qtA then
    Result := inherited QnoInfo(QnoType)
  else begin
    { Find relevant quantum numbers }
    i := qtA;
    if Ord(QnoType)-Ord(qtA) < nModes then
      with Result do begin
        MinValue := 0;
        MaxValue := High(Integer);
        StepSize := 1;
        Name := ModeInfo[Ord(QnoType)-Ord(qtA)].Name;
        Exit;
      end;
    Inc(i,nModes);
    for vn := 0 to nModes - 1 do begin
      if IsDegenerate(ModeInfo[vn].Symmetry,0) then begin
        if i = QnoType then
          with Result do begin
            MinValue := -High(Integer);
            MaxValue := High(Integer);
            StepSize := 1;
            Name := 'l'+Copy(ModeInfo[vn].Name,2,255);
            Exit;
          end;
        Inc(i);
      end;
    end;
    if i = QnoType then
      with Result do begin
        MinValue := -High(Integer);
        MaxValue := High(Integer);
        StepSize := 1;
        Name := 'Lambda';
        Exit;
      end;
    Inc(i);
    if i = QnoType then
      with Result do begin
        MinValue := -High(Integer);
        MaxValue := High(Integer);
        StepSize := 1;
        Name := 'Omega';
        Exit;
      end;
    Inc(i);
    if i = QnoType then
      with Result do begin
        MinValue := -1;
        MaxValue := 1;
        StepSize := 0;
        Name := 'Phase';
        Exit;
      end;
    Result := inherited QnoInfo(QnoType); { Not found }
  end;
end;

procedure TVibratingMolecule.SetnModes(const Index, Value: Integer);
var
  i, j, vn: Integer;
begin
  if (Value < nModes) and FittingInProgress then
    raise EReportException.Create('Number of modes reduced during fit');
  inherited SetSetting(Index, Value);
  j := Length(ModeInfo);
  SetLength(ModeInfo,Value);
  for i := j to Value - 1 do
    ModeInfo[i].Name := Format('v%d', [i+1]); { Fill in default name }
  { Update computed infornation }
  FnQuantumNumbers := nModes;
  i := nModes;
  for vn := 0 to nModes - 1 do begin
    ModeInfo[vn].dIndex := -1;
    ModeInfo[vn].Degeneracy := 1;
    ModeInfo[vn].ShowQno := ShowAllQnos;
    if IsDegenerate(ModeInfo[vn].Symmetry,0) then begin
      ModeInfo[vn].dIndex := i;
      ModeInfo[vn].Degeneracy := 2;
      Inc(FnQuantumNumbers);
      Inc(i);
    end;
  end;
  FHasElectronicAngularMomentum := False;
  for i := 0 to nManifolds-1 do
    for j := 0 to Manifold[i].nStates - 1 do begin
      FHasElectronicAngularMomentum := FHasElectronicAngularMomentum
                                       or (AngularMomentum((Manifold[i].State[j] as TElectronicState).Symmetry,0) > 0)
                                       or (Manifold[i].State[j].S > 0);
      for vn := 0 to nModes - 1 do
        with (Manifold[i].State[j] as TElectronicState).Mode(vn) do
          ModeInfo[vn].ShowQno := ModeInfo[vn].ShowQno or (vMin <> 0) or (vMax <> 0);
    end;
  if HasElectronicAngularMomentum then
    Inc(FnQuantumNumbers,2);
  { Make sure each state has the right number of modes }
  for i := 0 to nManifolds - 1 do
    with Manifold[i] do begin
      for j := 0 to nStates - 1 do
        (State[j] as TElectronicState).AdjustModes(True);
      for j := 0 to nPerturbations - 1 do
        (Perturbation[j] as TVibrationalPerturbation).AdjustSettingNames;
    end;
  for i := 0 to ComponentCount-1 do
    if Items[i] is TTransitionMoments then
      with TTransitionMoments(Items[i]) do
        for j := 0 to ComponentCount - 1 do
          (TransitionMoment[j] as TMultidimensionalFCF).AdjustModes(True);
end;

procedure TVibratingMolecule.SetnNuclei(const Index, Value: Integer);
begin
  SetLength(FNuclearMasses, Value);
  inherited;
  nModes := nModes; { Trigger update of Mode information }
  FixMass;
end;

procedure TVibratingMolecule.SetPointGroup(const Index: Integer; const AValue: TPointGroup);
begin
  SetSetting(Index, Ord(AValue));
end;

procedure TVibratingMolecule.SetUpTransitionFunctionVariables(
  TransitionFunction: TTransitionFunction; var ExtraVariables: TVariableArray);
var
  nExtra, i: Integer;
begin
  nExtra := nQuantumNumbers-2;
  SetLength(ExtraVariables, 2*nExtra);
  with TransitionFunction do
    for i := 0 to nExtra-1 do begin
      V(QnoInfo(TQnoType(Ord(qtA)+i+2)).Name+'''', ExtraVariables[2*i]);
      V(QnoInfo(TQnoType(Ord(qtA)+i+2)).Name+'"',  ExtraVariables[2*i+1]);
    end;
  inherited;
end;

procedure TVibratingMolecule.SetTransitionFunctionVariables(
  TransitionFunction: TTransitionFunction; var ExtraVariables: TVariableArray;
  ABra, AKet: TBasisState; SignA, SignB: Integer);
var
  i: Integer;
begin
  with TransitionFunction do
    for i := 0 to nQuantumNumbers-3 do begin
      ExtraVariables[2*i].Value := ABra.Qno(TQnoType(Ord(qtA)+i+2));
      ExtraVariables[2*i+1].Value := AKet.Qno(TQnoType(Ord(qtA)+i+2));
    end;
  inherited;
end;

procedure TVibratingMolecule.SetUpStateFunctionVariables(
  StateFunction: TStateFunction; var ExtraVariables: TVariableArray);
var
  nExtra, i: Integer;
begin
  nExtra := nQuantumNumbers-2;
  SetLength(ExtraVariables, nExtra);
  with StateFunction do
    for i := 0 to nExtra-1 do
      V(QnoInfo(TQnoType(Ord(qtA)+i+2)).Name, ExtraVariables[i]);
  inherited;
end;

procedure TVibratingMolecule.SetStateFunctionVariables(
  StateFunction: TStateFunction; var ExtraVariables: TVariableArray;
  AState: TBasisState);
var
  i: Integer;
begin
  with StateFunction do
    for i := 0 to nQuantumNumbers-3 do
      ExtraVariables[i].Value := AState.Qno(TQnoType(Ord(qtA)+i+2));
  inherited;
end;

function TVibratingMolecule.SymName(Sym: Integer): string;
begin
  if Sym < 0 then
    Result := '-'
  else
    Result := PointGroupInfo[PointGroup].SymmetryToStr(Sym, HalfSpin);
end;

type

  TRotationalConstantsLister = class(TLevelLister)
    Mol: TVibratingMolecule;
    Perts: array of TVibrationalPerturbation;
    Ii: array [1..6] of TColumnVector;
    OK: array [1..6] of TBooleanArray;
    ExtraInfo: array of string;
    MaxElements: Integer;
    procedure PrepareExtraInfo(Ev: TEigenValues); override;
    procedure ShowExtraHeader; override;
    procedure ShowExtraInfo(Ev: TEigenvalues; i:Integer); override;
    destructor Destroy; override;
  end;

{ TRotationalConstantsLister }

destructor TRotationalConstantsLister.Destroy;
var
  i: Integer;
begin
  for i := 1 to 6 do
    FreeAndNil(Ii[i]);
  for i := 0 to High(Perts) do
    Perts[i].Free;
  inherited;
end;

procedure TRotationalConstantsLister.PrepareExtraInfo(Ev: TEigenValues);
var
  xyz, xyzp, xyzpp, i, vn, vp, alphabeta, nCalc: Integer;
  NewMat: Boolean;
  M: TComplexMatrix;
  Empty: array of TBigSet;
  cName: string;
  xyzch: Char;
  Sum, aFactor, bFactor, Ei, V: Double;
  muPi: array [1..3] of TComplexMatrix;
  procedure HandlePerturbation(i: Integer);
  var
    M1: TComplexMatrix;
    n: Integer;
  begin
    n := Ev.AddPerturbation(M, Perts[i], meHrot, Empty, Mol.SmallCoefficient);
    if (n > 0) and (Mixture.PrintLevel >= plAll) then begin
      M1 := TComplexMatrix.Create(Ev.Size, Ev.Size);
      M1.Zero;
      Ev.AddPerturbation(M1, Perts[i], meHrot, Empty, Mol.SmallCoefficient);
      Ev.ShowBasisMatrix(M1, cName+' '+Perts[i].Name, 0, MaxElements);
      M1.Free;
    end;
  end;
  procedure Bump(Incr, v: Integer);
  begin
    with Perts[i] do
      ModifySetting(FirstDynamicSetting+v,Settings[FirstDynamicSetting+v]+Incr,False,False);
  end;
  procedure SecondOrder(xyz, xyzp: Integer);
  var
    i, k: Integer;
    Sum, dE: Double;
    Vc: Complex;
    St: TElectronicState;
  begin
    SetLength(OK[alphabeta], Ev.Size+1);
    for i := 1 to Ev.Size do begin
      { Check at least one extra vibrational quantum in the basis;
        removes values that are obviously wrong, though not a guarantee }
      with Ev.DominantBasis[i] as TVibrationalBasisState do begin
        OK[alphabeta][i] := True;
        St := State as TElectronicState;
        for k := 0 to Mol.nModes - 1 do
          if v[k] >= St.Mode(k).vMax then begin
            OK[alphabeta][i] := False;
            Break;
          end;
      end;
      if not OK[alphabeta][i] then
        Continue;
      if xyz = xyzp then
        Sum := muPi[xyz].Re[i,i] { First order term }
      else
        Sum := 0;
      Ei := Ev.E[i];
      for k := 1 to Ev.Size do
        if i <> k then begin
          if xyz = xyzp then
            V := muPi[xyz].SqrMod[i,k]
          else begin
            CMul(Vc, muPi[xyz][i,k], muPi[xyzp][k,i]);
            V := Sqrt(SqrMod(Vc));
          end;
          if Abs(V) > 0 then begin
            dE := Ei-Ev.E[k];
            if Abs(dE) > 1 then
              Sum := Sum + V/dE
            else begin
              if ExtraInfo[i] = '' then
                ExtraInfo[i] := 'Excluded Terms:';
              ExtraInfo[i] := ExtraInfo[i] + ' '+ 'xyz'[xyz] + 'xyz'[xyzp] + ' Matrix Element=' + FloatToStr(V) + ',Delta E='+ FloatToStr(dE);
            end;
          end;
        end;
      Ii[alphabeta][i] := Ii[alphabeta][i] + Sum;
    end;
  end;
begin
  { TODO: This could also calculate rotational matrix elements off-diagonal in vibronic state. }
  { TODO: Would then need to consider terms between different symmetries }
  if Ev.Size = 0 then
    Exit;
  SetLength(ExtraInfo, Ev.Size+1);
  for xyz := 1 to 3 do
    muPi[xyz] := nil;
  for alphabeta := 1 to 6 do
    FreeAndNil(Ii[alphabeta]);
  M := TComplexMatrix.Create(Ev.Size, Ev.Size);
  alphabeta := 1;
  aFactor := Sqr(hcross)/cm1toJ*gammaMhalf/Math.IntPower(Sqrt(amuAngstoms2toSI),3); { 195.76711230607512 }
  bFactor := Sqr(hcross*gammaMhalf/amuAngstoms2toSI)/cm1toJ; { = 1136.71864814679825 }
  PushProgress('Rotational Constants',9+6*Sqr(Mol.nModes)*Length(Perts));
  nCalc := 0;
  try
    for xyz := 1 to 3 do begin
      xyzch := 'xyz'[xyz];
      for xyzp := 1 to xyz do begin { TODO: Consider 1 to 3? }
        M.Zero;
        if xyz = xyzp then
          for i := 1 to Ev.Size do
            M.Re[i,i] := (Ev.Basis[i].State as TElectronicState).RotConsts[xyz];
        cName := xyzch + 'xyz'[xyzp];
        for vn := 0 to Mol.nModes - 1 do begin
          if Mol.IsDegenerate(Mol.ModeInfo[vn].Symmetry,0) then
            raise NotYet;
          for i := 0 to High(Perts) do begin
            Bump(1,vn);
            with Perts[i].Bra as TElectronicState do
              { First derivative of moment of inertia tensor }
              Perts[i].ModifyParameter(1, - aFactor*dIdQ[vn][xyz,xyzp]/Sqrt(Convert(Mode(vn).Omega,Mixture.Units,cm1))/Ixx[xyz]/Ixx[xyzp]/Two, False, False);
            HandlePerturbation(i);
            for vp := 0 to Mol.nModes - 1 do begin
              SetProgress(nCalc);
              Inc(nCalc);
              Bump(1,vp);
              with Perts[i].Bra as TElectronicState do begin
                { Second derivative of moment of inertia tensor }
                Sum := 0;
                for xyzpp := 1 to 3 do
                  Sum := Sum + Three*dIdQ[vn][xyz,xyzpp]*dIdQ[vp][xyzp,xyzpp]/Ixx[xyzpp];
                Sum := Sum * bFactor / (8*Ixx[xyz]*Ixx[xyzp]*Sqrt(Convert(Mode(vn).Omega,Mixture.Units,cm1)*Convert(Mode(vp).Omega,Mixture.Units,cm1)));
                Perts[i].ModifyParameter(1, Sum, False, False);
                HandlePerturbation(i);
                if xyz = xyzp then begin
                  { Prepare matrix of mu_aa pi_a so second order correction can be calculated }
                  Sum := - Two * RotConsts[xyz]*Coriolis[xyz][vn+1,vp+1]*Sqrt(Mode(vp).Omega/Mode(vn).Omega);
                  if Abs(Sum) > Molecule.SmallCoefficient then begin
                    if not Assigned(muPi[xyz]) then begin
                      muPi[xyz] := TComplexMatrix.Create(Ev.Size, Ev.Size);
                      muPi[xyz].Zero;
                      NewMat := True;
                    end else
                      NewMat := False;
                    Perts[i].Op := qp;
                    Perts[i].ModifyParameter(1, Sum, False, False);
                    Bump(-2,vn); { Must be one, so just switch sign }
                    if (Ev.AddPerturbation(muPi[xyz], Perts[i], meHrot, Empty, Mol.SmallCoefficient) = 0) and NewMat then
                      FreeAndNil(muPi[xyz]);
                    Bump(2,vn); { Put back }
                    Perts[i].Op := Vib.q;
                  end;
                end;
              end;
              Bump(-1,vp);
            end;
            Bump(-1,vn);
          end;
        end;
        if Mixture.PrintLevel >= plMatrices then
          Ev.ShowBasisMatrix(M, cName, 0, MaxElements);
        Ii[alphabeta] := Ev.Transform(M); { Note that Ev.Transform() returns the diagonal elements only }
        Inc(alphabeta);
      end;
    end;
    alphabeta := 1;
    for xyz := 1 to 3 do begin
      if Assigned(muPi[xyz]) then begin
        { Apply second order correction via Coriolis term }
        xyzch := 'xyz'[xyz];
        if Mixture.PrintLevel >= plMatrices then
          Ev.ShowBasisMatrix(muPi[xyz], 'mu_'+xyzch+xyzch+' pi_'+xyzch, 0, MaxElements);
        Ev.TransformTransitionMatrix(muPi[xyz], Ev);
        if Mixture.PrintLevel >= plMatrices then
          Ev.ShowBasisMatrix(muPi[xyz], 'mu_'+xyzch+xyzch+' pi_'+xyzch+' Transformed', 0, MaxElements);
        SecondOrder(xyz, xyz);
      end;
      Inc(alphabeta, xyz+1);
    end;
    alphabeta := 1;
    for xyz := 1 to 3 do
      for xyzp := 1 to xyz do begin
        if (xyz <> xyzp) and Assigned(muPi[xyz]) and Assigned(muPi[xyzp]) then
          SecondOrder(xyz, xyzp);
        Inc(alphabeta);
      end;
  finally
    PopProgress;
    for xyz := 1 to 3 do
      FreeAndNil(muPi[xyz]);
    M.Free;
  end;
end;

procedure TRotationalConstantsLister.ShowExtraHeader;
var
  xyz, xyzp: Integer;
begin
  for xyz := 1 to 3 do
    for xyzp := 1 to xyz do
      Write(Log, 'xyz'[xyz]+'xyz'[xyzp],' ':6+Mol.Precision);
  WriteLn(Log);
end;

procedure TRotationalConstantsLister.ShowExtraInfo(Ev: TEigenvalues; i: Integer);
var
  j: Integer;
begin
  for j := 1 to 6 do
    if (Length(OK[j]) = 0) or OK[j][i] then
      Write(Log, DtosMA(Ii[j][i],8+Mol.Precision,-1))
    else
      Write(Log, '   -':8+Mol.Precision);
  WriteLn(Log);
  if ExtraInfo[i] <> '' then
    WriteLn(Log, ExtraInfo[i]);
end;

{ TVibrationalManifold }

procedure TVibrationalManifold.AddHandlers;
begin
  inherited;
  AddHandler('Show Rotational Constants', 'Tabuate Rotational Constants for each level', ShowRotationalConstants);
end;

function TVibrationalManifold.FindJmax(Default: Integer; RemovedSpins: Byte): Integer;
begin
  if PointGroupInfo[(Molecule as TVibratingMolecule).PointGroup].HasDegenerateModes then
    Result := inherited FindJmax(Default, RemovedSpins)
  else
    Result := HalfSpin(RemovedSpins);
end;

function TVibrationalManifold.nSym: Integer;
begin
  Result := Molecule.LastSym+1;
end;

procedure TVibrationalManifold.ShowRotationalConstants(Sender: TObject);
var
  i: Integer;
  Lister: TRotationalConstantsLister;
begin
  ReShowLogForm;
  Lister := TRotationalConstantsLister.Create(nil, Self, False, False, False, MaxDouble);
  try
    with Lister do begin
      Mol := Molecule as TVibratingMolecule;
      Mol.CheckProperties;
      SetLength(Perts, nStates);
      for i := 0 to nStates - 1 do begin
        Perts[i] := TVibrationalPerturbation.CreateUnOwned(State[i], State[i]);
        Perts[i].Op := Vib.q;
      end;
      MaxElements := 10;
      ShowManifold(Self);
    end;
  finally
    Lister.Free
  end;
end;

{ TMolecularNucleus }

procedure TMolecularNucleus.AddHandlers;
begin
  inherited;
  AddHandler('Set Exact Mass', 'Try to set exact mass from mass and name',DoFindExactMass);
end;

constructor TMolecularNucleus.Create(AOwner: TComponent);
begin
  inherited;
  { Mass should be a per molecule setting, but is actually
    in the per state object. Copy from the per molecule setting on
    initialisation }
  ModifyParameter(1, (AOwner.Owner.Owner as TVibratingMolecule).FNuclearMasses[NodeIndex], False);
end;

function TMolecularNucleus.ReadProperties(Node: TDOMNode; CheckUnread: Boolean): Integer;
begin
  { Discard properties that are no longer published }
  ReadValue(Node, 'Spin', '');
  ReadValue(Node, 'AsNext', '');
  Result := inherited ReadProperties(Node, CheckUnread);
end;

procedure TMolecularNucleus.DoFindExactMass(Sender: TObject);
var
  EMass: Double;
begin
  if Copy(Name,1,7) = 'Nucleus' then
    EMass := FindExactMass(Mass)
  else
    EMass := FindExactMass(Mass,Name);
  if Emass > 0 then
    Mass := EMass
  else
    raise EReportException.Create('Mass not found');
end;

procedure TMolecularNucleus.GlobalShift(Shift: Integer);
var
  Target: TMolecularNucleus;
begin
  if (NodeIndex+Shift < 0) or (NodeIndex+Shift >= Molecule.nNuclei) then
    Exit;
  Target := State.Nucleus(NodeIndex+Shift) as TMolecularNucleus;
  inherited;
  { Reset mass, to ensure copied to global array }
  SetParameter(1,GetParameter(1));
  Target.SetParameter(1,Target.GetParameter(1));
end;

const
 MolecularNucleusParameterNames: array [1..4] of string = (
'Mass of nucleus/amu.',
'Equilibrium x co-ordinate / Angstrom.',
'Equilibrium y co-ordinate / Angstrom.',
'Equilibrium z co-ordinate / Angstrom.'
);

function TMolecularNucleus.HelpOnParameter(AIndex: Integer): string;
begin
  if (AIndex >= Low(MolecularNucleusParameterNames))
     and (AIndex <= High(MolecularNucleusParameterNames)) then
    Result := MolecularNucleusParameterNames[AIndex]
  else
    Result := inherited HelpOnParameter(AIndex);
end;

procedure TMolecularNucleus.SetComponentIndex(const Value: Integer);
var
  Displacements: array of array of Double;
  Original: array of TNucleus;
  OldNuc, NewNuc, i, j, nModes: Integer;
  EState: TElectronicState;
begin
  { If nuclei numbering changed, move nuclear displacements also }
  { Save original }
  EState := State as TElectronicState;
  SetLength(Original,Molecule.nNuclei);
  nModes := (Molecule as TVibratingMolecule).nModes;
  for i := 0 to Molecule.nNuclei-1 do
    Original[i] := State.Nucleus(i);
  SetLength(Displacements,nModes,3*Molecule.nNuclei);
  for i := 0 to nModes-1 do
    with EState.Mode(i) do
      for j := 0 to 3*Molecule.nNuclei-1 do
        Displacements[i,j] := GetParameter(FirstDynamicParameter+j);
  inherited;
  { Look for changed nodes }
  for OldNuc := 0 to Molecule.nNuclei-1 do
    if Original[OldNuc].NodeIndex <> OldNuc then begin
      NewNuc := Original[OldNuc].NodeIndex;
      for i := 0 to nModes-1 do
        with EState.Mode(i) do
          for j := 0 to 2 do
            SetParameter(FirstDynamicParameter+j+NewNuc*3,Displacements[i,j+OldNuc*3]);
    end;
end;

procedure TMolecularNucleus.SetParameter(AIndex: Integer; const Value: Double);
var
  i, j, MyNodeIndex: Integer;
  Changed: Boolean;
  NewValue: Double;
begin
  inherited SetParameter(AIndex, Value);
  { See notes in Create method; set per molecule settting also }
  if NodeIndex < 0 then
    Exit;
  Changed := False;
  NewValue := GetParameter(AIndex); { Handle reset to default case }
  case AIndex of
    1: begin
         Changed := (Molecule as TVibratingMolecule).FNuclearMasses[NodeIndex] <> NewValue;
         (Molecule as TVibratingMolecule).FNuclearMasses[NodeIndex] := NewValue;
       end;
  end;
  if not Changed then
    Exit;
  { As GetParameter is not virtual, changes must be copied to all other nueclei }
  (Molecule as TVibratingMolecule).FixMass;
  MyNodeIndex := NodeIndex;
  for i := 0 to Molecule.nManifolds - 1 do
    with Molecule.Manifold[i] do
      for j := 0 to nStates - 1 do
        State[j].Nucleus(MyNodeIndex).ModifyParameter(AIndex, NewValue, True);
end;

{ TElectronicState }

procedure TElectronicState.AddHandlers;
begin
  inherited;
  if Molecule.nNuclei > 0 then begin
    AddHandler('l Matrix...', 'Enter/Edit l matrix', SetLmatrix);
    AddHandler('Print Information', 'Print molecular information', PrintMolecularInfo);
    AddHandler('Fix Centre of Mass', 'Adjust atomic co-ordinates so centre of mass is at the origin', FixCentreOfMass);
    AddHandler('Normalize', 'Normalize vibrational co-ordinates', Normalize);
    AddHandler('Orthogonalize', 'Adjust vibrational co-ordinates to be orthogonal and correctly normalized', Orthogonalize);
  end;
end;

function TElectronicState.AddMode:TVibrationalMode;
begin
  Result := TVibrationalMode.Create(Self);
end;

function TElectronicState.AddOneJ(Eigenvalues: TEigenvalues; J, F, M: TQuantumNumber; Sym: Integer; HyperfineState: THyperfineBasis; Seffective: TQuantumNumber): Integer;
var
  v: TIntegerArray;
  Position, Phase: Integer;
  EQMin, EQMax: array of Integer; { Range of effective quantum numbers }
  DummyBasisState: TVibrationalBasisState;
  procedure AddState;
  var
    BasisState: TVibrationalBasisState;
    FirstSym, AltSym, i, q: Integer;
  begin
    { Determine symmetry }
    Phase := 0;
    FirstSym := SymmetryOf(v, AltSym);
    if AltSym >=0 then
      Phase := 1;
    if FirstSym <> Sym then
      if AltSym = Sym then
        Phase := -1
      else
        Exit;
    if Assigned(DummyBasisState) then begin
      DummyBasisState.Phase := Phase;
      if Eigenvalues.DiagonalMatrixElement(meHrot, DummyBasisState, True) > Emax then
        Exit; { TODO: Inefficient, as tests every state }
    end;
    BasisState := TVibrationalBasisState.Create(Eigenvalues,F,M,Sym);
    BasisState.State := Self;
    BasisState.HyperfineState := HyperfineState;
    BasisState.SubBasisIndex := Eigenvalues.LastSubBasisIndex;
    BasisState.v := Copy(v,0,Length(v));
    BasisState.FOriginalPosition := Position;
    BasisState.Phase := Phase;
    BasisState.StatisticalWeight := Result;
    BasisState.StatWtGroup := 1; { Statistical weights not significant for vibration }
    for i := 0 to High(FEffectiveQuanta) do begin
      q := FEffectiveQuanta[i].EffectiveQuantumNumber(v);
      EQMin[i] := Min(EQMin[i],q);
      EQMax[i] := Max(EQMax[i],q);
    end;
  end;
var
  i, vn, StartIndex, nQ, dind, ApproxBasis, ProgCount: Integer;
  OK: Boolean;
begin
  if Molecule.IsDegenerate(Sym, Molecule.HalfSpin) then
    Result := 2
  else
    Result := 1;
  if J > 1 then
    Exit; { Jmax is used to indicate maxium vibrational angular momentum only, and only a single
            calculation for a given symmetry is required, so this routine should only be used
            for J = 0 or 1/2 }
  StartIndex := Max(Eigenvalues.Size,0);
  SetLength(v, Molecule.nQuantumNumbers);
  if Emax <= 1e99 then begin
    DummyBasisState := TVibrationalBasisState.Create(nil, F, M, Sym);
    DummyBasisState.State := Self;
    DummyBasisState.HyperfineState := HyperfineState;
    DummyBasisState.v := v;
    DummyBasisState.StatisticalWeight := Result;
    DummyBasisState.StatWtGroup := 1;
  end else
    DummyBasisState := nil;
  try
    { Cycle through all available vibrational quantum numbers }
    Position := 1;
    for i := 0 to Molecule.nQuantumNumbers - 1 do
      v[i] := 0;
    ApproxBasis := 1; { Produce estimate of basis set size }
    for vn := 0 to Molecule.nModes - 1 do
      ApproxBasis := ApproxBasis * (Mode(vn).vmax - Mode(vn).vmin + 1);
    ProgCount := 0;
    if ApproxBasis > 1000 then
      PushProgress(UniqueName+' Basis',ApproxBasis);
    SetLength(EQMin,Length(FEffectiveQuanta));
    for i := 0 to Length(FEffectiveQuanta) - 1 do
      EQMin[i] := High(Integer);
    SetLength(EQMax,Length(FEffectiveQuanta));
    repeat { For every mode combination }
      OK := True;
      for vn := 0 to Molecule.nModes - 1 do
        if v[vn] < Mode(vn).vmin then begin
          OK := False;
          Break;
        end;
      if OK then
        { Cycle though electronic angular momentum }
        if Molecule.HasElectronicAngularMomentum then begin
          dind := High(v)-1;
          v[dind] := Molecule.AngularMomentum(Symmetry,0);
          { Try both + and - values of electronic orbital angular momentum }
          repeat
            for i := 0 to S do begin
              v[dind+1] := v[dind] + 2*i - S; { Omega }
              AddState;
            end;
            v[dind] := -v[dind];
          until v[dind] >= 0;
        end else
          AddState;
      { Cycle though vibrational angular momentum; must do it here
        so that Postion variable is correct }
      vn := Molecule.nModes - 1;
      while vn >= 0 do begin
        if Molecule.IsDegenerate(Molecule.ModeInfo[vn].Symmetry,0) then begin
          { Bump up angular momentum }
          dind := Molecule.ModeInfo[vn].dIndex;
          Inc(v[dind],2);
          if v[dind] > v[vn] then begin
            { run out of l values, so reset and move to next }
            v[dind] := -v[vn];
          end else
            Break;
        end;
        Dec(vn);
      end;
      if vn >= 0 then
        Continue; { New combination of angular momenta; Add }
      Inc(Position);
      { Now find next set of v, cycling highest numbered modes fastest }
      vn := Molecule.nModes - 1;
      while vn >= 0 do begin
        Inc(v[vn]);
        if Molecule.IsDegenerate(Molecule.ModeInfo[vn].Symmetry,0) then begin
          dind := Molecule.ModeInfo[vn].dIndex;
          v[dind] := -v[vn]; { Reset angular momentum to minimum }
        end;
        if v[vn] > Mode(vn).vMax then begin
          v[vn] := 0; { Overflow to next }
          if Molecule.IsDegenerate(Molecule.ModeInfo[vn].Symmetry,0) then
            v[dind] := 0; { Reset angular momentum to minimum }
        end else begin
          Inc(ProgCount);
          if (ApproxBasis > 1000) and (ProgCount mod 256 = 0) then
            SetProgress(ProgCount);
          Break; { Combination OK }
        end;
        Dec(vn);
      end;
    until vn < 0;
    if ApproxBasis > 1000 then
      PopProgress;
    { Put effective quantum numbers in state index }
    nQ := 1;
    EigenValues.Fixup;
    if StartIndex+1 <= Eigenvalues.Size then
      for i := 0 to High(FEffectiveQuanta) do begin
        for j := StartIndex+1 to Eigenvalues.Size do
          with EigenValues.Basis[j] as TVibrationalBasisState do
            Inc(SubBasisIndex,(FEffectiveQuanta[i].EffectiveQuantumNumber(v)-EQMin[i])*nQ);
        nQ := nQ * (EQMax[i]-EQMin[i]+1);
      end;
    EigenValues.SortFrom(StartIndex,Emax);
    Inc(Eigenvalues.LastSubBasisIndex,nQ);
  finally
    DummyBasisState.Free;
  end;
end;

procedure TElectronicState.AdjustModes(FixNumber: Boolean);
var
  j, n: Integer;
begin
  if FixNumber then
    n := AdjustNodeIndexes(TVibrationalMode, FModes, nil, opInsert, Molecule.nModes)
  else
    n := AdjustNodeIndexes(TVibrationalMode, FModes);
  for j := 0 to n - 1 do
    Mode(j).AdjustParameterNames;
end;

procedure TElectronicState.ApplylMatrix(lMatrixNode: TlMatrixMNode; InPrincipalAxes: Boolean; Verbose: Boolean);
var
  i, j, nNonVibModes, Generation: Integer;
  M: TVibrationalMode;
  Um: TResonanceUnits;
  Mx: Double;
  WasModified: Boolean;
begin
  Um := (RootNode as TMixture).Units;
  Mx := lMatrixNode.lambda.Maximum;
  Generation := lMatrixNode.Generation;
  if Mx < LinearTest then begin
    if Verbose then
      WriteLn(Log, 'All vibrations zero frequency');
    Exit;
  end;
  nNonVibModes := 0;
  for i := 1 to lMatrixNode.lambda.Size do
    if lMatrixNode.lambda[i] < Molecule.SmallCoefficient*lMatrixNode.lambda[1] then
      Inc(nNonVibModes)
    else if nNonVibModes > 0 then begin
      if Verbose then     
        lMatrixNode.lambda.Write(Log, 'lambda - Modes out of order');
      Exit;
    end;
  if Verbose then
    WriteLn(Log, 'Discarding ', nNonVibModes, ' modes');
  Molecule.nModes := lMatrixNode.lambda.Size - nNonVibModes;
  AdjustModes(True);
  { Above can cause this routine to be reentered - just quit if this happens }
  if Generation <> lMatrixNode.Generation then
    Exit;
  WasModified := RootNode.Modified;
  for i := 1 to Molecule.nModes do begin
    M := Mode(i-1);
    M.Omega := Convert(sqrt(lMatrixNode.lambda[i]/LambdaConversion), cm1, Um);
    for j := 1 to 3*Molecule.nNuclei do
      M.Parameters[M.FirstDynamicParameter+j-1] := lMatrixNode.l[j,i]/Sqrt((Nucleus((j-1) div 3) as TMolecularNucleus).Mass);
  end;
  if not WasModified then
    RootNode.Modified := False; { Avoid spurious modified flag }
  lMatrixNode.lInPrincipalAxes := InPrincipalAxes;
  if InPrincipalAxes then
    { l Matrix must use same axis system as atomic co-ordinates; zMatrix will have updated state with co-ordinates
      in original axis system }
    lMatrixNode.zMatrix.ApplyToStateEx(Verbose, True, Self);
end;

function TElectronicState.ConvertParameter(AIndex: Integer; Zero: Boolean): TPerturbation;
var
  PertEquiv: TVibPertEquiv;
  vn, i: Integer;
begin
  Result := nil; { Leave by default }
  if AIndex > nParameters then begin
    { Parameter is mode specific }
    Dec(AIndex, nParameters+1);
    vn := AIndex div Mode(0).nParameters ;
    AIndex := AIndex mod Mode(vn).nParameters + 1;
    if AIndex = 1 then
      Exit; { Omega values must not be converted, as they determine the basis }
    PertEquiv := Mode(vn).PertEquiv(AIndex);
    if Length(PertEquiv.Powers) = 0 then
      Exit; { Not implemented - leave }
    Result := TVibrationalPerturbation.Create(Owner);
    Result.SetBraKetName(Name,Name);
    with Result as TVibrationalPerturbation do begin
      Op := PertEquiv.Op;
      for i := 0 to Molecule.nModes - 1 do
        SetSetting(FirstDynamicSetting+i,PertEquiv.Powers[i]);
      Value := Mode(vn).Parameters[AIndex];
    end;
    if Zero then
      Mode(vn).Parameters[AIndex] := 0;
  end;
  { TODO: Spin parameters as perturbations }
end;

constructor TElectronicState.Create(AOwner: TComponent);
begin
  SavedProperties := SavedProperties + ',Emax=1e99';
  if Assigned(AOwner) then
    FMolecule := AOwner.Owner as TVibratingMolecule;
  inherited;
end;

destructor TElectronicState.Destroy;
var
  i: Integer;
begin
  ClearArrays;
  for i := High(DummyObjects) downto 0 do
    FreeAndNil(DummyObjects[i]);
  inherited;
end;

class function TElectronicState.NucleusType: TNucleusClass;
begin
  Result := TNucleus;
end;

const
  ForceFieldClasses: array [0..3] of TActiveMNodeWithArrayClass = (TzMatrix, TInternalCoords, TCartesianDisplacements, TSymmetryCoords);

function TElectronicState.ForceFieldNode(FClass: TActiveMNodeWithArrayClass): TActiveMNodeWithArray;
  function MakeN(i:Integer):TActiveMNodeWithArray;
  begin
    Result := ForceFieldObjects[i];
    if not Assigned(Result) then begin
      if not Assigned(DummyObjects[i]) then
        DummyObjects[i] := ForceFieldClasses[i].MakeDummy(Self);
      Result := DummyObjects[i];
      if Result is TlMatrixMNode then
        TlMatrixMNode(Result).UpdateObjectChain;
      ForceFieldObjects[i] := Result;
    end;
  end;
var
  i: Integer;
begin
  Result := nil;
  if not Assigned(ForceFieldObjects[0]) then begin
    UpdateObjectChain;
    MakeN(0); { Ensure z Matrix available }
  end;
  for i := 0 to High(ForceFieldClasses) do
    if FClass = ForceFieldClasses[i] then begin
      Result := MakeN(i);
      Break;
    end;
end;

function TElectronicState.FixupForceFieldCalc(Verbose: Boolean): Exception;
var
  i: Integer;
  ErrMsg: string;
  LastObject: TActiveMNodeWithArray;
begin
  Result := nil;
  ErrMsg := UpdateObjectChain;
  LastObject := nil;
  if ErrMsg = '' then begin
    for i := High(ForceFieldObjects) downto 0 do
      if Assigned(ForceFieldObjects[i]) then begin
        LastObject := ForceFieldObjects[i];
        Break;
      end;
    if LastObject = nil then
      { Force existance of zMatrix }
      LastObject := ForceFieldNode(TzMatrix);
    if not LastObject.ApplyOK(Verbose) then
      ErrMsg := LastObject.UniqueName + ' - ' + LastObject.LastError;
  end;
  if ErrMsg <> '' then
    Result := EReportException.Create(ErrMsg);
end;

function TElectronicState.FixupProperties: Exception;
var
  nPQ, i: Integer;
begin
  Precision := (RootNode as TMixtureMNode).Precision;
  AdjustNuclei(True); { Explicit call here, so that nuclei typically appear before vibrations, though this order is not essential }
  AdjustModes(True);
  if Symmetry > Molecule.LastSym then
    Symmetry := Symmetry mod (Molecule.LastSym+1); { Force sane symmetry following point group change }
  Result := FixupForceFieldCalc(False);
  if Assigned(Result) then
    Exit;
  if (nWorkers > 0) and (Molecule.nModes > 0) then begin
    { Ensure symbolic variables set up }
    for i := 0 to Molecule.nModes-1 do begin
      vVariable(i);
      lVariable(i);
    end;
    vVariables.CheckThreadVars;
    lVariables.CheckThreadVars;
    elVariables.CheckThreadVars;
  end;
  Result := inherited FixupProperties;
  { TODO: called multiple times if FCF present; adjust fixup order so State.FixupProperties is called first }
  PrepareMatrices(False);
  nPQ := 0;
  for i := 0 to Manifold.Count - 1 do
    if Manifold.Components[i] is TVibrationalPerturbation then
      with TVibrationalPerturbation(Manifold.Components[i]) do
        if (Bra = Self) and (Ket = Self) and EffectiveQno then begin
          if Length(FEffectiveQuanta) <= nPQ then
            SetLength(FEffectiveQuanta,nPQ+1);
          FEffectiveQuanta[nPQ] := TVibrationalPerturbation(Manifold.Components[i]);
          Inc(nPQ);
        end;
  SetLength(FEffectiveQuanta,nPQ);
end;

function TElectronicState.MatrixElementType(const TheMatrixElement: TMatrixElement; const H: string; ShowPhases: Boolean): string;
  procedure AddVariables(Bra: Boolean);
    procedure AddVariable(const Variable: string; Change: Integer);
    begin
      Result := Result + Variable;
      if not Bra then
        Exit;
      if Change > 0 then
        Result := Result + '+';
      if Change <> 0 then
        Result := Result + IntToStr(Change);
    end;
    procedure AddNumberedVariable(const Variable: string; AIndex, Change: Integer);
    begin
      AddVariable(Variable+IntToStr(AIndex+1),Change);
    end;
  var
    vn, dind: Integer;
  begin
    for vn := 0 to Molecule.nModes - 1 do begin
      if vn > 0 then
        Result := Result + ',';
      AddNumberedVariable('v',vn,TheMatrixElement.Signature[vn]);
      if Molecule.IsDegenerate(Molecule.ModeInfo[vn].Symmetry,0) then begin
        dind := Molecule.ModeInfo[vn].dIndex;
        Result := Result + '(';
        AddNumberedVariable('l',vn,TheMatrixElement.Signature[dind]);
        Result := Result + ')';
      end;
    end;
    if Molecule.HasElectronicAngularMomentum then begin
      if Molecule.nModes > 0 then
        Result := Result + ',';
      AddVariable('Lambda',TheMatrixElement.Signature[Molecule.nQuantumNumbers-2]);
      Result := Result + ',';
      AddVariable('Omega',TheMatrixElement.Signature[Molecule.nQuantumNumbers-1]);
    end;
  end;
begin
  Result := '<';
  AddVariables(True);
  Result := Result + ShowElementPhases(TheMatrixElement,H,ShowPhases);
  AddVariables(False);
  Result := Result + '>';
end;

function TElectronicState.Mode(Index: Integer): TVibrationalMode;
begin
  if (Index >= Length(FModes)) and (Molecule.nModes <> Length(FModes)) then
    AdjustModes(True);
  Result := FModes[Index] as TVibrationalMode;
end;

function TElectronicState.nameMode(Index: Integer): string;
begin
  if Index <= Molecule.nModes then
   Result := Mode(Index-1).Name
  else begin
    Index := Index - Molecule.nModes;
    if Index > 3 then begin
      Result := 'T';
      Dec(Index,3);
      if Index > 3 then
        Index := Index mod 3 + 1; { Molecule not completely set up }
    end else
      Result := 'R';
    Result := Result + 'xyz'[Index];
  end;
end;

function TElectronicState.nameNucleusXYZ(Index: Integer): string;
begin
  Dec(Index);
  Result := 'xyz'[Index mod 3 + 1] + Nucleus(Index div 3).Name;
end;

function TElectronicState.nParameterVariables: Integer;
begin
  Result := nParameters;
  if Molecule.nModes > 0 then
    Result := Result + Molecule.nModes*Mode(0).nParameters;
end;

function TElectronicState.ParameterVariable(AIndex: Integer; Suffix: string): TSortedVariableNode;
var
  vn, i: Integer;
begin
  if AIndex <= nParameters then
    Result := inherited ParameterVariable(AIndex, Suffix)
  else begin
    if not Assigned(VariableList) then
      inherited ParameterVariable(1, Suffix); { Force creation of list }
    if VariableList.Count < nParameterVariables then
      for vn := 0 to Molecule.nModes - 1 do
        with Mode(vn) do
          for i := 1 to nParameters do
            VariableList.Add(TSortedVariable.Create(ParameterNames[i]+IntToStr(vn+1),VariableList.Count+1));
    Dec(AIndex,nParameters+1);
    vn := AIndex div Mode(0).nParameters;
    AIndex := AIndex mod Mode(vn).nParameters + 1;
    i := VariableList.IndexOf(Mode(vn).ParameterNames[AIndex]+IntToStr(vn+1));
    Result := (VariableList.Objects[i] as TSortedVariable).Template;
    Result.Value := Mode(vn).Parameters[AIndex];
  end;
end;

procedure TElectronicState.FixCentreOfMass(Sender: TObject);
var
  i, j, p: Integer;
  N: TMolecularNucleus;
begin
  if zMatrix.FindCofM(True) then
    for i := 0 to Molecule.nNuclei - 1 do begin
      N := Nucleus(i) as TMolecularNucleus;
      for j := 0 to 2 do begin
        p := N.IndexOfParameter('x')+j;
        N.SetParameter(p, N.GetParameter(p) - zMatrix.CofM[j+1]);
      end;
    end;
end;

function TElectronicState.CalculatedParameter(Index: Integer): Double;
  function NucInfo(Index: Integer):Txyz;
  var
    j: Integer;
  begin
    with Nucleus(Index) as TMolecularNucleus do
      for j := 1 to 3 do
        Result[j] := GetParameter(IndexOfParameter('x')+j-1);
  end;
var
  n, xyz, r, rr, a1, a2, a3: Word;
  R1, R2, R3: Txyz;
begin
  if (Index > nFixedCalculatedParameters) and (Index <= nCalculatedParameters) then begin
    Dec(Index, nFixedCalculatedParameters);
    n := 3*Molecule.nModes*Molecule.nModes;
    Result := 0;
    if Index <= n then begin
      { zetax[r,s] }
      n := Molecule.nModes;
      DivMod(Index-1, n*n, xyz, r);
      DivMod(r,n,rr,r);
      if Assigned(Coriolis[1]) then { if a Coriolis calculation has been done }
        Result := Coriolis[xyz+1][r+1,rr+1];
    end else begin
      Dec(Index, n);
      n := Sqr(Molecule.nNuclei);
      if Index <= n then begin
        { r[A1,A2] }
        DivMod(Index-1, Molecule.nNuclei, a2, a1);
        R1 := NucInfo(a1);
        R2 := NucInfo(a2);
        Result := Diff(R1, R1, R2);
      end else begin
        Dec(Index,n+1);
        { angle[A1,A2,A3] }
        AngleIndex(Index, Molecule.nNuclei, a1, a2, a3);
//        DivMod(Index-1, Molecule.nNuclei, a2, a1);
//        DivMod(a2, Molecule.nNuclei, a3, a2);
        R1 := NucInfo(a1);
        R2 := NucInfo(a2);
        R3 := NucInfo(a3);
        Result := radtodeg(Angle(R1, R2, R3));
      end;
    end;
  end else
    Result := FCalculatedParameters[Index];
end;

function TElectronicState.CalculatedParameterName(Index: Integer): string;
var
  n, xyz, q, a1, a2, a3: Word;
begin
  if (Index > nFixedCalculatedParameters) and (Index <= nCalculatedParameters) then begin
    Dec(Index, nFixedCalculatedParameters);
    n := 3*Molecule.nModes*Molecule.nModes;
    if Index <= n then begin
      { zetax[r,s] }
      n := Molecule.nModes;
      DivMod(Index-1, n*n, xyz, q);
      Result := 'zeta'+'xyz'[xyz+1]+'[';
      DivMod(q,n,xyz,q);
      Result := Result+Mode(q).Name+','+Mode(xyz).Name+']';
    end else begin
      Dec(Index, n);
      n := Sqr(Molecule.nNuclei);
      if Index <= n then begin
        { r[A1,A2] }
        DivMod(Index-1, Molecule.nNuclei, a2, a1);
        Result := 'r['+Nucleus(a1).Name+','+Nucleus(a2).Name+']';
      end else begin
        Dec(Index,n+1);
        AngleIndex(Index, Molecule.nNuclei, a1, a2, a3);
        { angle[A1,A2,A3] }
//        DivMod(Index-1, Molecule.nNuclei, a2, a1);
//        DivMod(a2, Molecule.nNuclei, a3, a2);
        Result := 'angle['+Nucleus(a1).Name+','+Nucleus(a2).Name+','+Nucleus(a3).Name+']';
      end;
    end;
  end else
    Result := FCalculatedParameterNames[Index];
end;

function TElectronicState.nCalculatedParameters: Integer;
var
  n: Integer;
begin
  n := Molecule.nNuclei;
  Result := nFixedCalculatedParameters
            + 3*Molecule.nModes*Molecule.nModes { zeta }
            + Sqr(n) { bond lengths }
            + n*(n-1)*(n-2) { bond angles }
            ;
end;

procedure TElectronicState.ACP(Value: Double; const PName: string);
begin
  Inc(FnCalculatedFixedParameters);
  { Bound check error here implies nFixedCalculatedParameters is wrong }
  FCalculatedParameters[FnCalculatedFixedParameters] := Value;
  FCalculatedParameterNames[FnCalculatedFixedParameters] := PName;
end;

procedure TElectronicState.ClearArrays;
var
  xyz: Integer;
begin
  FreeAndNil(adm);
  FreeAndNil(Madm);
  FreeAndNil(lMatrix);
  for xyz := 1 to 3 do
    FreeAndNil(Coriolis[xyz]);
end;

procedure TElectronicState.PrintMolecularInfo(Sender: TObject);
var
  E: Exception;
begin
  Molecule.FixupProperties.Free; { Required for "Other Potential Terms" to be calculated }
  ReShowLogForm;
  E := FixupForceFieldCalc(True);
  if Assigned(E) then
    raise E
  else
    PrepareMatrices(True);
end;

procedure TElectronicState.PrepareMomentsOfInteria;
var
  p: Integer;
begin
  SetCachedInformationPresent;
  if zMatrix = nil then
    ForceFieldNode(TzMatrix).Apply(False);
  FnCalculatedFixedParameters := 0;
  ClearArrays;
  for p := 1 to 3 do begin
    if Molecule.nNuclei = 0 then begin
      Ixx[p] := 0;
      RotConsts[p] := 0;
    end else begin
      Ixx[p] := zMatrix.Ixx[p];
      RotConsts[p] := zMatrix.RotConsts[p];
    end;
    ACP(Ixx[p], 'I'+'xyz'[p]+'xyz'[p]);
    ACP(RotConsts[p], 'B'+'xyz'[p]);
  end;
end;

procedure TElectronicState.Normalize(Sender: TObject);
var
  i, k, offset: Integer;
  T: TMatrix;
  Sc: Double;
begin
  PrepareMatrices(False,False);
  if not Assigned(adm) then
    Exit;
  T := Madm.Transpose;
  T.PostMultiply(adm);
  offset := Mode(0).FirstDynamicParameter;
  for i := 0 to Molecule.nModes - 1 do begin { For each vibration }
    Sc := T[i+1,i+1];
    for k := 0 to adm.RowCount-1 do
      Mode(i).Parameters[offset+k] := Mode(i).Parameters[offset+k] / Sqrt(Sc);
  end;
  T.Free;
end;

procedure TElectronicState.Notification(AComponent: TComponent; Operation: TOperation);
var
  i: Integer;
begin
  inherited;
  if not (AComponent is TlMatrixMNode) then
    Exit;
  if Operation = opRemove then
    for i := 0 to High(DummyObjects) do
      if DummyObjects[i] = AComponent then
        DummyObjects[i] := nil;
  if not (csDestroying in ComponentState) then
    ClearObjectChain;
end;

procedure TElectronicState.Orthogonalize(Sender: TObject);
var
  i, j, k, offset: Integer;
  T: TMatrix;
  Sc: Double;
  procedure UpdateT(All:Boolean);
  begin
    PrepareMatrices(False,All);
    T := Madm.Transpose;
    T.PostMultiply(adm);
  end;
begin
  T := nil;
  UpdateT(True);
  if not Assigned(adm) then
    Exit;
  offset := Mode(0).FirstDynamicParameter;
  { Remove components along T, R }
  for i := 0 to Molecule.nModes - 1 do { For each vibration }
    for j := Molecule.nModes to adm.ColCount-1 do begin
      { For each T or R motion }
      Sc := T[i+1,j+1];
      for k := 0 to adm.RowCount-1 do
        Mode(i).Parameters[offset+k] := Mode(i).Parameters[offset+k] - adm[k+1,j+1] * Sc;
    end;
  { Now remove components along other vibrations }
  for i := 0 to Molecule.nModes - 1 do begin { For each vibration }
    for j := 0 to i - 1 do begin
      UpdateT(False);
      Sc := T[i+1,j+1];
      for k := 0 to adm.RowCount-1 do
        Mode(i).Parameters[offset+k] := Mode(i).Parameters[offset+k] - adm[k+1,j+1] * Sc;
    end;
    UpdateT(False);
    Sc := T[i+1,i+1];
    for k := 0 to adm.RowCount-1 do
      Mode(i).Parameters[offset+k] := Mode(i).Parameters[offset+k] / Sqrt(Sc);
  end;
  T.Free;
  PrepareMatrices(True,True);
end;

procedure TElectronicState.PrepareCentrifugal(Print: Boolean);
var
  r, Nuc, xyz, xyzp, xyzq, xyzr, p, i, ia: Integer;
  Sum, TauFactor: Double;
  Mixture: TMixture;
  procedure Wr(p, q, r, s: Integer);
  var
    l: string;
  begin
    l := 'xyz';
    l := l[p]+l[q]+l[r]+l[s];
    Write(Log, l, tau[p,q,r,s]:17:8, '  ');
  end;
  function MakeCD(x, y, z: Integer; Rep: string):TCentrifugal;
  var
    Txx, Tyy, Tzz, Txy, Txz, Tyz, sigma: Double;
  begin
    Rep := '[' + Rep + ']';
    Txx := tau[x,x,x,x]/4;
    Tyy := tau[y,y,y,y]/4;
    Tzz := tau[z,z,z,z]/4;
    Txy := (tau[x,x,y,y] +2*tau[x,y,x,y])/4;
    Txz := (tau[x,x,z,z] +2*tau[x,z,x,z])/4;
    Tyz := (tau[y,y,z,z] +2*tau[y,z,y,z])/4;
    if Abs(Ixx[x]-Ixx[y]) < LinearTest then
      SetNan(sigma) { Force failure }
    else
      sigma := (2*RotConsts[z]-RotConsts[x]-RotConsts[y])/(RotConsts[x]-RotConsts[y]);
    with Result do begin
      DJ  := -(3*Txx + 3*Tyy + 2*Txy)/8;
      DJK :=  (3*Txx + 3*Tyy + 2*Txy - 4*Txz - 4*Tyz)/4;
      DK  := -(3*Txx + 3*Tyy + 2*Txy + 8*Tzz - 8*Txz - 8*Tyz)/8;
      deltaJ := -(Txx - Tyy)/4;
      R5 := -(Txx - Tyy - 2*Txz +2*Tyz)/8;
      R6 := (Txx + Tyy -2*Txy)/16;
      A.DJ := -(Txx + Tyy)/2;
      A.DJK := 3*(Txx + Tyy)/2 - (Txy + Tyz + Txz);
      A.DK := -(Txx + Tyy + Tzz) + (Txy + Tyz + Txz);
      A.d1 := deltaJ;
      A.d2 := -((sigma-1)*Txx + (sigma+1)*Tyy - 2*sigma*Txy + 2*(Txz-Tyz))/4;
      ACP(A.DJ, 'DeltaJ'+rep);
      ACP(A.DJK, 'DeltaJK'+rep);
      ACP(A.DK, 'DeltaK'+rep);
      ACP(A.d1, 'deltaJ'+rep);
      ACP(A.d2, 'deltaK'+rep);
      if IsNanOrInf(sigma) then
        sigma := 0
      else
        sigma := 1/sigma;
      S.DJ  := -((3+sigma)*Txx + (3-sigma)*Tyy + 2*Txy - 2*sigma*(Txz-Tyz))/8;
      S.DJK := ((3+3*sigma)*Txx + (3-3*sigma)*Tyy + 2*Txy - (4+6*sigma)*Txz - (4-6*sigma)*Tyz)/4;
      { Note multiplier of Tyy in next line is 3*sigma/5 in eq 36 of JMS 160 117 (1993))
        Multiplier given here is consistent with VIBCA and relationships between A and S reductions
        in Table 4 of Watson, Vib Spect Struc 6 1 (1977) }
      S.DK  := -((3+5*sigma)*Txx + (3-5*sigma) { Was 3*sigma/5 } *Tyy + 2*Txy + 8*Tzz - (8+10*sigma)*Txz - (8-10*sigma)*Tyz)/8;
      S.d1 := (Txx - Tyy)/4;
      S.d2 := ((1-sigma)*Txx + (1+sigma)*Tyy - 2*Txy + 2*sigma*(Txz-Tyz))/16;
      ACP(S.DJ, 'DJ'+rep);
      ACP(S.DJK, 'DJK'+rep);
      ACP(S.DK, 'DK'+rep);
      ACP(S.d1, 'd1'+rep);
      ACP(S.d2, 'd2'+rep);
    end;
  end;
begin
  SetLength(dIdQ, Molecule.nModes);
  if Molecule.nModes = 0 then
    if FixedCalculatedParametersSetUp then
      Exit
    else
      Print := False;
  TauFactor := Sqr(Sqr(hCross)*gammaMhalf/cm1toJ)/Math.IntPower(amuAngstoms2toSI,3);
  TauFactor := Convert(TauFactor, cm1, MHz)*1000; { Values in kHz, not cm-1 ; 1148947468038.87 }
  Mixture := RootNode as TMixture;
  for r := 1 to Molecule.nModes do begin
    { Diagonal elements }
    for p := 1 to 3 do
      dIdQ[r-1][p,p] := 0;
    for xyz := 1 to 3 do begin
      Sum := 0;
      ia := -1; { Separate index needed to skip dummy atoms }
      for Nuc := 0 to Molecule.nNuclei-1 do begin
        zMatrix.NextRealAtom(ia);
        with Nucleus(Nuc) as TMolecularNucleus do
          Sum := Sum + 2*Sqrt(Mass)*zMatrix.Atoms[ia].pxyz[xyz]*lMatrix[xyz+Nuc*3,r];
      end;
      for p := 1 to 3 do
        if p <> xyz then
          dIdQ[r-1][p,p] := dIdQ[r-1][p,p] + Sum;
    end;
    { Off-Diagonal elements }
    for xyz := 1 to 3 do
      for xyzp := 1 to xyz-1 do begin
        Sum := 0;
        ia := -1; { Separate index needed to skip dummy atoms }
        for Nuc := 0 to Molecule.nNuclei-1 do begin
          zMatrix.NextRealAtom(ia);
          with Nucleus(Nuc) as TMolecularNucleus do
            Sum := Sum - Sqrt(Mass)*(zMatrix.Atoms[ia].pxyz[xyz]*lMatrix[xyzp+Nuc*3,r] + zMatrix.Atoms[ia].pxyz[xyzp]*lMatrix[xyz+Nuc*3,r]);
        end;
        dIdQ[r-1][xyz,xyzp] := Sum;
        dIdQ[r-1][xyzp,xyz] := Sum;
      end;
  end;
  if Print then begin
    WriteLn(Log, 'Inertial Derivatives / (amu^1/2 Angstrom)');
    WriteLn(Log, 'Mode    xx          yy          zz          xy          xz          yz');
    for r := 0 to Molecule.nModes-1 do begin
      Write(Log, nameMode(r+1):5);
      for i := 1 to 3 do
        Write(Log, dIdQ[r][i,i]:12:7);
      Write(Log, dIdQ[r][1,2]:12:7);
      Write(Log, dIdQ[r][1,3]:12:7);
      WriteLn(Log, dIdQ[r][2,3]:12:7);
    end;
  end;
  for xyz := 1 to 3 do
    for xyzp := 1 to xyz do
      for xyzq := 1 to 3 do
        for xyzr := 1 to xyzq do begin
          Sum := 0;
          for r := 0 to Molecule.nModes-1 do
            Sum := Sum + dIdQ[r][xyz,xyzp]*dIdQ[r][xyzq,xyzr]/(2*Ixx[xyz]*Ixx[xyzp]*Ixx[xyzq]*Ixx[xyzr]*Sqr(Convert(Mode(r).Omega,Mixture.Units,cm1)));
          Sum := - Sum * TauFactor;
          tau[xyz,xyzp,xyzq,xyzr] := Sum;
          ACP(Sum, 'tau'+'xyz'[xyz]+'xyz'[xyzp]+'xyz'[xyzq]+'xyz'[xyzr]); { Only add first of four possibilities }
          tau[xyzp,xyz,xyzq,xyzr] := Sum;
          tau[xyz,xyzp,xyzr,xyzq] := Sum;
          tau[xyzp,xyz,xyzr,xyzq] := Sum;
        end;
  taupxy := tau[1,1,2,2] +2*tau[1,2,1,2]; ACP(taupxy, 'taupxy');
  taupxz := tau[1,1,3,3] +2*tau[1,3,1,3]; ACP(taupxz, 'taupxz');
  taupyz := tau[2,2,3,3] +2*tau[2,3,2,3]; ACP(taupyz, 'taupyz');
  for xyz := 1 to 3 do begin
    r := xyz mod 3 + 1;
    i := r mod 3 + 1;
    CD[xyz] := MakeCD(r,i,xyz, StringOfChar('I',xyz)+'r');
    CD[xyz+3] := MakeCD(i,r,xyz, StringOfChar('I',xyz)+'l');
  end;
  if Print then begin
    WriteLn(Log, 'Centrifugal Distortion, tau/kHz');
    for xyz := 1 to 3 do
      Wr(xyz,xyz,xyz,xyz);
    WriteLn(Log);
    for xyz := 1 to 3 do begin
      xyzp := xyz mod 3 + 1;
      Wr(xyz,xyz,xyzp,xyzp);
    end;
    WriteLn(Log);
    for xyz := 1 to 3 do begin
      xyzp := xyz mod 3 + 1;
      Wr(xyz,xyzp,xyz,xyzp);
    end;
    WriteLn(Log);
    Wr(1,1,2,3); Wr(2,2,1,3); Wr(3,3,1,2); WriteLn(Log);
    Wr(1,2,1,3); Wr(2,1,2,3); Wr(3,1,3,2); WriteLn(Log);
    Wr(1,1,1,2); Wr(1,1,1,3);
    Wr(2,2,2,1); Wr(2,2,2,3);
    Wr(3,3,3,1); Wr(3,3,3,2);
    WriteLn(Log);
    WriteLn(Log, 'tau''/kHz: ', 'xxyy ', taupxy:16:8, ' xxzz ', taupxz:16:8, ' yyzz ', taupyz:16:8);
    WriteLn(Log, 'Rep         DJ               DJK                DK                deltaJ            R5                R6                ',
                 'A.DeltaJ          A.DeltaJK         A.DeltaK          A.deltaJ          A.deltaK          ',
                 'S.DeltaJ          S.DeltaJK         S.DeltaK          S.d1              S.d2');
    for i := 1 to 6 do
      with CD[i] do
        WriteLn(Log, StringOfChar('I', (i-1) mod 3 + 1)+'rl'[i div 4+1]:4,' ',
                     DJ:17:7,' ',   DJK:17:7,' ',   DK:17:7,' ', deltaJ:17:7,' ', R5:17:7,' ', R6:17:7,' ',
                   A.DJ:17:7,' ', A.DJK:17:7,' ', A.DK:17:7,' ', A.d1:17:7,' ', A.d2:17:7,' ',
                   S.DJ:17:7,' ', S.DJK:17:7,' ', S.DK:17:7,' ', S.d1:17:7,' ', S.d2:17:7);
  end;
end;

procedure TElectronicState.PrepareAlphas(Print: Boolean; SecondDerivsOnly: Boolean);
var
  r, rp, xyz, rho: Integer;
  Sum, Sum0, OmR, OmS: Double;
  Sumv: array [1..3] of Double;
begin
  SetLength(alpha, Molecule.nModes+1);
  if Molecule.nModes = 0 then
    if FixedCalculatedParametersSetUp then
      Exit
    else
      Print := False;
  if Print and not SecondDerivsOnly and (Molecule.Mixture.PrintLevel = plAll) then
    PrepareAlphas(Print, True);
  Delta0 := 0;
  for xyz := 1 to 3 do begin
    { Separate sum for ground state, avoiding issues with degenerate modes, as
      reported by A. L. L. East, C. S. Johnson, and W. D. Allen, J. Chem. Phys. 98, 1299 (1993).
      See also JCP 116 6482 }
    Sumv[xyz] := 0;
    alpha[Molecule.nModes][xyz] := 0;
    for r := 0 to Molecule.nModes-1 do begin
      Sum := 0;
      OmR := Mode(r).Omega;
      for rho := 1 to 3 do
        if Ixx[rho] > Molecule.SmallCoefficient then
          Sum := Sum + 3 * Sqr(dIdQ[r][rho,xyz])/4/Ixx[rho];
      Sum0 := Sum;
      if not SecondDerivsOnly then
        for rp := 0 to Molecule.nModes-1 do begin
          OmS := Mode(rp).Omega;
          if (rp <> r) and (Abs(OmR-OmS) > 1) then
            Sum := Sum + Sqr(Coriolis[xyz][r+1,rp+1])*(Sqr(OmR+OmS)/OmS/(OmR-OmS) - Sqr(OmR-OmS)/OmS/(OmR+OmS))/2;
          Sum0 := Sum0 - Sqr(Coriolis[xyz][r+1,rp+1]) * Sqr(OmR-OmS)/OmS/(OmR+OmS)/2;
        end;
      Sum := -Sum * 2*Sqr(RotConsts[xyz])/OmR;
      alpha[r][xyz] := Sum;
      Sumv[xyz] := Sumv[xyz] + Sum;
      alpha[Molecule.nModes][xyz] := alpha[Molecule.nModes][xyz] - 2*Sum0*Sqr(RotConsts[xyz])/OmR;
    end;
    Sum0 := alpha[Molecule.nModes][xyz];
    if not SecondDerivsOnly then begin
      ACP(Sum0, 'alpha'+'xyz'[xyz]);
      ACP(RotConsts[xyz]-Sum0/2, 'B'+'xyz'[xyz]+'0');
    end;
    if RotConsts[xyz] > Molecule.SmallCoefficient then
      Sum0 := Ixx[xyz]*Sum0/RotConsts[xyz];
    if xyz <> 3 then
      Sum0 := -Sum0;
    Delta0 := Delta0 + Sum0/2; { Ic - Ia - Ib }
  end;
  if not SecondDerivsOnly then
    ACP(Delta0, 'Delta0');
  FixedCalculatedParametersSetUp := True;
  if Print then begin
    Write(Log, 'Vibrational dependence of rotational constants/cm-1');
    if SecondDerivsOnly then
      WriteLn(Log, ' from second derivative of inverse of moment of interia matrix only')
    else
      WriteLn(Log);
    WriteLn(Log, ' Mode    alphaA         alphaB         alphaC');
    for r := 0 to Molecule.nModes-1 do
      WriteLn(Log, nameMode(r+1):5, DTosMA(alpha[r][1],15,10), DTosMA(alpha[r][2],15,10), DTosMA(alpha[r][3],15,10));
    WriteLn(Log, '(Sum':5, Sumv[1]:15:10, Sumv[2]:15:10, Sumv[3]:15:10, ' excluding resonant contribution)');
    r := Molecule.nModes;
    WriteLn(Log, 'Sum':5, alpha[r][1]:15:10, alpha[r][2]:15:10, alpha[r][3]:15:10);
    WriteLn(Log, 'Rotational constants of zero point level');
    for xyz := 1 to 3 do      
      Write(Log, (RotConsts[xyz]-alpha[Molecule.nModes][xyz]/2):15:10);
    WriteLn(Log);
    WriteLn(Log, 'Delta0/(amu A^2):', Delta0:15:10);

  end;
end;

procedure TElectronicState.PrepareCoriolis(Print: Boolean);
var
  alph, beta, gamma, k, r, rr: Integer;
  Sum: Double;
begin
  if Print then
    WriteLn(Log, 'Coriolis Coupling Constants');
  for alph := 1 to 3 do begin
    if not Assigned(Coriolis[alph]) then
      Coriolis[alph] := TMatrix.Create(Molecule.nModes, Molecule.nModes);
    beta := alph mod 3 + 1;
    gamma := beta mod 3 + 1;
    for r := 1 to Molecule.nModes do
      for rr := 1 to Molecule.nModes do begin
        Sum := 0;
        for k := 0 to Molecule.nNuclei-1 do
          Sum := Sum + lMatrix[beta+k*3,r]*lMatrix[gamma+k*3,rr] - lMatrix[beta+k*3,rr]*lMatrix[gamma+k*3,r];
          // To compute from adm instead:  Sum := Sum + (adm[beta+k*3,r]*adm[gamma+k*3,rr] - adm[beta+k*3,rr]*adm[gamma+k*3,r])*(Nucleus(k) as TMolecularNucleus).Mass;
        Coriolis[alph][r,rr] := Sum;
        { Note that VIBCA seems to give exactly the opposite sign }
        { Done by an different method: ACP(Sum, 'zeta'+'xyz'[alpha]+'('+IntToStr(r)+','+IntToStr(rr)+')'); }
      end;
    if Print then
      Coriolis[alph].Write(Log, 'xyz'[alph]+'zeta', nameMode, nameMode, Precision+6);
  end;
end;

function TElectronicState.PrepareMatrices(Print: Boolean; AllModes:Boolean):Boolean;
var
  i, j, k, k1, k2, b, ia: Integer;
  N: TMolecularNucleus;
  T: TMatrix;
  xI, RI: TColumnVector;
  Makel, Transforml: Boolean;
begin
  Result := True;
  PrepareMomentsOfInteria;
  if Molecule.nNuclei = 0 then
    Exit;
  for i := High(ForceFieldObjects) downto 0 do
    if Assigned(ForceFieldObjects[i]) and (ForceFieldObjects[i] is TlMatrixMNode) then
      with TlMatrixMNode(ForceFieldObjects[i]) do begin
        lMatrix := l;
        Transforml := not lInPrincipalAxes;
        Result := Transforml;
        if Assigned(lMatrix) then
          Break;
      end;
  Makel := not Assigned(lMatrix);
  i := 3*Molecule.nNuclei;
  if AllModes then
    j := i
  else
    j := Molecule.nModes;
  adm := TMatrix.Create(i,j);
  Madm := TMatrix.Create(i,j);
  adm.Zero;
  xI := TColumnVector.Create(3);
  RI := TColumnVector.Create(3);
  if Makel then begin
    lMatrix := TMatrix.Create(i,j);
    lMatrix.Clear;
    Transforml := True;
  end else begin
    lMatrix := TMatrix.Create(lMatrix);
    lMatrix.SetColCountAndClear(j,True);
  end;
  ia := -1; { Separate index needed to skip dummy atoms }
  for i := 0 to Molecule.nNuclei - 1 do begin
    zMatrix.NextRealAtom(ia);
    N := Nucleus(i) as TMolecularNucleus;
    for j := 0 to 2 do
      for k := 0 to Molecule.nModes - 1 do
        with Mode(k) do
          if nParameters >= FirstDynamicParameter+3*i+j then
            adm[3*i+j+1,k+1] := GetParameter(FirstDynamicParameter+3*i+j);
    if AllModes then begin
      b := Max(Molecule.nModes+1,3*Molecule.nNuclei-5);
      { Add rotations }
      { See Hirota 2.2.9, though note adm matrix is mass weighted: el/sqrt(m) }
      { ... but need to do this in principal axis system }
      for k := 1 to 3 do
        xI[k] := zMatrix.Atoms[ia].pxyz[k];
      for k := 1 to 3 do
        { Check for zero moments of Inertia }
        if Abs(Ixx[k]) > LinearTest then begin
          k1 := k mod 3 + 1;
          k2 := k1 mod 3 + 1;
          RI[k] := 0;
          RI[k2] := xI[k1]/Sqrt(Ixx[k]);
          RI[k1] := -xI[k2]/Sqrt(Ixx[k]);
          for k1 := 1 to 3 do
            lMatrix[3*i+k1,b] := RI[k1]*Sqrt(N.Mass);
          if Makel or Transforml then
            RI.PreMultiply(zMatrix.Itrans, False); { Transform rotation back to input axis system }
          for k1 := 1 to 3 do
            adm[3*i+k1,b] := RI[k1];
          Inc(b);
        end;
      { Add translations }
      b := 3*Molecule.nNuclei-3; { This may overwrite one of the rotations }
      if Molecule.Mass > 0 then { Avoid errors during data entry }
        for k := 1 to 3 do begin
          adm[3*i+k,b+k] := 1/Sqrt(Molecule.Mass);
          for k1 := 1 to 3 do
            if k = k1 then
              lMatrix[3*i+k,b+k] := Sqrt(N.Mass/Molecule.Mass)
            else
              lMatrix[3*i+k1,b+k] := 0;
        end;
    end;
    for k := 1 to adm.ColCount do
      for j := 1 to 3 do
        Madm[3*i+j,k] := N.Mass * adm[3*i+j,k]; { TODO: Madm not really needed given adm }
    if Makel or Transforml then
      for k := 1 to adm.ColCount do
        if (k <= Molecule.nModes) or (k > adm.ColCount-3) then begin
          { Rotate mode to principal axis system }
          { Test above skips rotational modes, as these are already correct }
          for j := 1 to 3 do
            if Makel then
              xI[j] := adm[3*i+j,k]*Sqrt(N.Mass)
            else
              xI[j] := lMatrix[3*i+j,k];
          xI.PreMultiply(zMatrix.Itrans, True);
          for j := 1 to 3 do
            lMatrix[3*i+j,k] := xI[j];
        end;
  end;
  if Print then begin
    WriteLn(Log,'Matrices for ',UniqueName);
    adm.Write(Log, 'adm ', nameNucleusXYZ, nameMode, Precision+6);
    Madm.Write(Log, 'Madm', nameNucleusXYZ, nameMode, Precision+6);
    CheckOrthogonal(Molecule.SmallCoefficient, True);
    WriteLn(Log, 'lMatrix in Principal Axis System');
    lMatrix.Write(Log, 'l', nameNucleusXYZ, nameMode, Precision+8);
    lMatrix.CheckOrthogonal(Log, 'l^Tl', True, Molecule.SmallCoefficient, True, nameMode, nameMode);
  end;
  xI.Free;
  RI.Free;
  PrepareCoriolis(Print);
  PrepareCentrifugal(Print);
  PrepareAlphas(Print, False);
  if (Molecule.nModes > 0) and (FnCalculatedFixedParameters <> nFixedCalculatedParameters) then
    raise Exception.Create('nFixedCalculatedParameters');
end;

function TElectronicState.CheckOrthogonal(Threshold: Double; Print: Boolean): Double;
var
  T: TMatrix;
begin
  T := Madm.Transpose;
  T.PostMultiply(adm);
  Result := T.CheckUnit(Log,'Orthogonality check: Madm^T adm',Threshold,Print,nameMode,nameMode);
  T.Free;
end;

procedure TElectronicState.DoCheckOrthogonal(Threshold: Double);
var
  V: Double;
begin
  PrepareMatrices(False,True);
  V := CheckOrthogonal(100,False);
  if V < Threshold then
    Exit;
  ReShowLogForm;
  ShowInfo('Warning l Matrix for '+UniqueName+' fails orthogonality check:'+FloatToStr(V));
  PrepareMatrices(True,True);
end;

function TElectronicState.ReadBranchFormat(Text: string; var Obs: TObservation;
  UpperState: TState; Reader: TObservationsReaderBase): Boolean;
  function ReadQnos(var vs: TIntegerArray; State: TElectronicState): Boolean;
  var
    vn, p, dind: Integer;
    function Absorb(c: Char):Boolean;
    begin
      Result := (p <= Length(Text)) and (Text[p] = c);
      if Result then
        Inc(p);
    end;
    function NextNumber(var Value: Integer):Boolean;
    var
      e: Integer;
    begin
      e := p;
      Absorb('-');
      while (p <= Length(Text)) and (Text[p] in ['0'..'9','.']) do
        Inc(p);
      Result := (e = p) or Reader.ReadQuantumNumber(Value,Copy(Text,e,p-e));
    end;
  begin
    Result := False;
    p := 1;
    if not Absorb('(') then
      Exit;
    SetLength(vs,Molecule.nQuantumNumbers+1);
    for vn := 0 to Molecule.nModes - 1 do begin
      if NextNumber(vs[vn]) then
        Exit;
      vs[vn] := vs[vn] div 2;
      if Molecule.IsDegenerate(Molecule.ModeInfo[vn].Symmetry,0) then begin
        dind := Molecule.ModeInfo[vn].dIndex;
        if Absorb('(') then begin
          { Read optional l value }
          if NextNumber(vs[dind]) then
            Exit;
          vs[dind] := vs[dind] div 2;
          Absorb(')');
        end else
          vs[dind] := vs[vn] and 1; { Default to lowest possible value }
      end;
      Absorb(',');
    end;
    if Molecule.HasElectronicAngularMomentum then begin
      dind := Length(vs)-3; { Extra qno here (not present in normal basis) is phase }
      vs[dind] := 0; { Default to no electronic angular momentum }
      vs[dind+1] := 0;
      if Absorb('[') then begin
        if Molecule.IsDegenerate(State.Symmetry,0) then begin
          if Absorb('+') then
            vs[dind] := Molecule.AngularMomentum(State.Symmetry,0)
          else if Absorb('-') then
            vs[dind] := -Molecule.AngularMomentum(State.Symmetry,0)
          else
            Exit;
          Absorb(',');
        end;
        if State.S > 0 then begin
          if NextNumber(vs[dind+1]) then
            Exit;
        end else
          vs[dind+1] := vs[dind]; { Omega = Lambda }
        Absorb(']')
      end;
    end;
    dind := High(vs);
    if Absorb('+') then
      vs[dind] := 1
    else if Absorb('-') then
      vs[dind] := -1
    else
      vs[dind] := Low(Integer);
    if Absorb(')') then begin
      Delete(Text,1,p-1);
      Result := True;
    end;
  end;
  function FindState(AManifold:TManifold; var Index: Smallint; var Sym: Shortint; v: array of Integer; State: TElectronicState):string;
  var
    AltSym: Integer;
    AlternateSym: ShortInt;
  begin
    Sym := State.SymmetryOf(v,AltSym);
    Result := AManifold.FindStateLabel(Index, Sym, AlternateSym, 0, { Not used }
                                       v, Obs.QuantumNumberFormat, Obs.ExField,
                                       Sym, 0, -1, State, Reader.RemoveSpins);
    if (Sym < 0) and (AltSym >= 0) then begin
      Sym := AltSym;
      Result := AManifold.FindStateLabel(Index, Sym, AlternateSym, 0, { Not used }
                                         v, Obs.QuantumNumberFormat, Obs.ExField,
                                         Sym, 0, -1, State, Reader.RemoveSpins);
    end;
  end;
var
  Up, Lo: TIntegerArray;
  Lname: string;
begin
  Result := False;
  Obs.JL := 0;
  Obs.JU := 0;
  Text := Trim(Text);
  if (Length(Text) > 2) and (Text[1] = '-') and (Text[2] = '(') then begin
    { Lower state fit }
    Delete(Text,1,1);
    if not (ReadQnos(Lo, Self) and (Text = '')) then
      Exit;
    Obs.ObsType := obELower;
    Lname := FindState(Manifold, Obs.IndexL, Obs.SymL, Lo, Self);
    if Obs.SymL >= 0 then
      Obs.Name := Obs.Name + ' : . - ' + Lname
    else
      obs.Name := Obs.Name + LName;
    Obs.SymU := 0;
  end else if not ReadQnos(Up, UpperState as TElectronicState) then
    Exit
  else if (Text = '') or (Text = '-') then begin
    { Upper state fit }
    Obs.ObsType := obEUpper;
    Lname := FindState(UpperState.Manifold,Obs.IndexU, Obs.SymU,Up,UpperState as TElectronicState);
    if Obs.SymU >= 0 then
      Obs.Name := Obs.Name + ' : ' + Lname + ' - .'
    else
      obs.Name := Obs.Name + LName;
    Obs.SymL := 0;
  end else begin
    { Normal Transition fit }
    if Text[1] <> '-' then
      Exit;
    Delete(Text,1,1);
    if ReadQnos(Lo, Self) and (Text = '') then begin
      Lname := FindState(Manifold, Obs.IndexL, Obs.SymL, Lo, Self);
      if Obs.SymL >= 0 then begin
        Obs.Name := Obs.Name + ' : '
                    + FindState(UpperState.Manifold,Obs.IndexU, Obs.SymU,Up,UpperState as TElectronicState)
                     + ' - ' + Lname;
      end else
        Obs.Name := Obs.Name + LName;
    end;
  end;
  Result := (Obs.SymL >=0) and (Obs.SymU >= 0);
  if Result then begin
    Obs.IndexU := 0; { Force search }
    Obs.IndexL := 0;
  end;
end;

function TElectronicState.SymbolicRotationalMatrixElement(AIndex: Integer; RankWanted: Integer): TMatrixElements;
var
  Sigma, Svar: TExprNode;
  dLam, vn: Integer;
  PertEquiv: TVibPertEquiv;
begin
  if AIndex > nParameters then begin
    { Parameter is mode specific }
    Dec(AIndex, nParameters+1);
    vn := AIndex div Mode(0).nParameters;
    AIndex := AIndex mod Mode(vn).nParameters + 1;
    if AIndex >= Mode(vn).FirstDynamicParameter then
      SetLength(Result,0)
    else if AIndex = 5 then begin { Special case gll, as not implemented as perturbation (yet) }
      SetLength(Result,1);
      SetLength(Result[0].Signature,Molecule.nQuantumNumbers);
      Result[0].Expr := EB([lVariable(vn), '^', 2]);
    end else begin
      PertEquiv := Mode(vn).PertEquiv(AIndex);
      if Length(PertEquiv.Powers) = 0 then
        raise NotYet;
      Result := Vib.SymbolicMatrixElements(PertEquiv, Molecule);
    end;
  end else if Molecule.IsDegenerate(Symmetry,0) or (S > 0) or (AIndex = 1)then begin
    SetLength(Result,1);
    SetLength(Result[0].Signature,Molecule.nQuantumNumbers);
    Sigma := EB([eOmega, '-', eLambda]);
    case AIndex of
      1: { Origin } Result[0].Expr := TIntegerNode.Create(1) ;
      LSP: { A - Spin orbit coupling }
        if S > 0 then
          Result[0].Expr := EB([eLambda, '*', Sigma])
        else
          Result[0].Expr := TIntegerNode.Create(0);
      LSP+1: { LambdaSS - Spin-spin coupling }
        if S > 1 then begin
          Result[0].Expr := EB([Sigma, '^', 2]);
          Result[0].Expr := EN([3, '*', Result[0].Expr, '-', EN([S*(S+2), '/', 4]), '*', 2, '/', 3]);
        end else
          Result[0].Expr := TIntegerNode.Create(0);
      LSP+2: { o - Omega = 0 lambda doubling }
        if Molecule.AngularMomentum(Symmetry,0) >= S then begin
          { This changes sign of Lambda }
          dLam := Molecule.AngularMomentum(Symmetry,0);
          Result[0].Signature[Molecule.nQuantumNumbers-2] := dLam;
          Svar := EN([S, '/', 2]);
          Result[0].Expr := EN([rlE(Svar,dLam,-dLam), '/', 2]);
          Svar.Free;
          { Also need opposite sign change }
          SetLength(Result,2);
          Result[1].Signature := Copy(Result[0].Signature);
          Result[1].Signature[Molecule.nQuantumNumbers-2] := -dLam;
          Result[1].Expr := Result[0].Expr.Duplicate;
        end else
          Result[0].Expr := TIntegerNode.Create(0);
    else
      SetLength(Result,0);
    end;
    Sigma.Free;
  end else
    SetLength(Result,0);
end;

function TElectronicState.RotationalMatrixElement(ABra, AKet: TBasisState; SignA, SignB: Integer; MNode: TMnode; Flag: Integer): Complex;
var
  Bra, Ket: TVibrationalBasisState;
  vn, j, RTMode, dind: Integer;
  vh, Sum: Double;
  Lambda, Sigma: TQuantumNumber;
begin
  Result := z0;
  case TMatElemType(Flag) of
    meHrot: { fall though to routine below };
    meWidth: begin
               if (ABra = AKet) and (SignA = SignB) then
                 Result.re := Width;
               Exit;
             end;
  else
    raise NotYet;
  end;
  Bra := ABra as TVibrationalBasisState;
  Ket := AKet as TVibrationalBasisState;
  if (Bra.Sym <> Ket.Sym) or (Bra.J <> Ket.J) then
    raise Exception.Create('Bad TElectronicState.RotationalMatrixElement call');
  if (Bra.v = Ket.v) and (SignA = SignB) then begin
    { Diagonal Matrix elements }
    Result.re := Origin;
    for vn := 0 to Molecule.nModes - 1 do
      with Mode(vn) do begin
        vh := Bra.v[vn]+Molecule.ModeInfo[vn].Degeneracy/2;
        if Molecule.IsDegenerate(Symmetry,0) then
          Result.re := Result.re + gll*Sqr(Bra.v[Molecule.ModeInfo[vn].dIndex]); { g l^2; note that l is not doubled }
        Sum := 0;
        for j := 4 { zOmega } downto 1 { Omega } do
          Sum := (Sum + GetParameter(j))*vh;
        Result.re := Result.re + Sum;
      end;
    if Molecule.IsDegenerate(Symmetry,0) or (S > 0) then begin
      Lambda := SignA*Bra.v[High(Bra.v)-1];
      Sigma := SignA*Bra.v[High(Bra.v)] - Lambda;
      Result.re := Result.re + A*Lambda*Sigma / 4
                      { 2/3 lambda (3Sz^2-S^2) }
                      + 2/3 * LambdaSS*(3*Sigma*Sigma - S*(S+2)) / 4;
    end;
  end else if Molecule.IsDegenerate(Symmetry,0) then begin
    { Look for Renner-Teller style matrix elements }
    dind := High(Bra.v)-1;
    if SignA*Bra.v[dind] <> -SignB*Ket.v[dind] then
      Exit; { Lambda must change sign }
    if SignA*(Bra.v[dind+1]-Bra.v[dind]) <> SignB*(Ket.v[dind+1]-Ket.v[dind]) then begin
      { Sigma changes - possibly Lambda doubling term,
        for which vibrations must not change }
      if SignA*Bra.v[dind+1] <> SignB*Ket.v[dind+1] then
        Exit; { Omega changes }
      for vn := 0 to Molecule.nModes - 1 do begin
        if Molecule.IsDegenerate(Molecule.ModeInfo[vn].Symmetry,0) then begin
          dind := Molecule.ModeInfo[vn].dIndex;
          if SignA*Bra.v[dind] <> SignB*Ket.v[dind] then
            Exit;
        end;
        if Bra.v[vn] <> Ket.v[vn] then
          Exit;
      end;
      dind := High(Bra.v)-1;
      Result.re := o/2 * rl(S, SignA*(Bra.v[dind+1]-Bra.v[dind]), SignB*(Ket.v[dind+1]-Ket.v[dind]));
      Exit;
    end;
    { Now look for exactly one change in l of the right magnitude }
    RTMode := -1; { Flag no mode found yet }
    for vn := 0 to Molecule.nModes - 1 do begin
      if Molecule.IsDegenerate(Molecule.ModeInfo[vn].Symmetry,0) then begin
        dind := Molecule.ModeInfo[vn].dIndex;
        if SignA*Bra.v[dind] <> SignB*Ket.v[dind] then
          if RTMode = -1 then
            RTMode := vn { Found candidate }
          else
            RTMode := -2 { Off diagonal in more than one mode }
        else if Bra.v[dind] <> Ket.v[dind] then
          RTMode := -2; { non RT active mode changes }
      end else if Bra.v[vn] <> Ket.v[vn] then
        RTMode := -2; { non RT active mode changes }
      if RTMode = -2 then
        Break;
    end;
    if RTMode >= 0 then begin
      dind := Molecule.ModeInfo[RTMode].dIndex;
      Result.re := Mode(RTMode).epsOmega * vnvpll(Bra.v[RTMode],SignA*Bra.v[dind],Molecule.AngularMomentum(Symmetry,0),Ket.v[RTMode],SignB*Ket.v[dind]);
    end;
  end;
end;

function TElectronicState.RotationalPerturbationMatrixElement(ABra,
  AKet: TBasisState; SignA, SignB: Integer; MNode: TMNode;
  Flag: Integer): Complex;
var
  Bra, Ket: TVibrationalBasisState;
  P: TVibrationalPerturbation;
begin
  if Flag <> Ord(meHrot) then
    raise NotYet;
  Bra := ABra as TVibrationalBasisState;
  Ket := AKet as TVibrationalBasisState;
  P := MNode as TVibrationalPerturbation;
  if Bra.J <> Ket.J then
    raise Exception.Create('TElectronicState.PerturbationMatrixElement');
  Result := P.MatrixElement(Bra.v, Ket.v, SignA, SignB);
end;

{.$DEFINE VERYSLOWTEST}
function TElectronicState.RotationalTransitionMatrixElement(
  ABra, AKet: TBasisState; SignA, SignB: Integer;
  TransitionMoment: TRootedMNodePair; F: TQuantumNumber; out dK: TQuantumNumber): Complex;
var
  Left, Right: TVibrationalBasisState;
  Dmat: TMatrix;
  dind, vn: Integer;
  Lambda, LambdaP: TQuantumNumber;
  {$IFDEF VERYSLOWTEST}
  r1, r2: Double;
  BraV,KetV: TIntegerArray;
  {$ENDIF}
begin
  Left := ABra as TVibrationalBasisState;
  Right := AKet as TVibrationalBasisState;
  if not (TransitionMoment is TMultidimensionalFCF) then
    raise NotYet;
  Result := z0;
  dK := Low(TQuantumNumber);
  with TMultidimensionalFCF(TransitionMoment) do
    if (VibronicallyAllowed = vaYes) and PointGroupInfo[Molecule.PointGroup].Allowed(Left.Sym, ActualSymmetry, Right.Sym, Molecule.HalfSpin) then begin
      { First check for no change in electronic quantum numbers }
      if Molecule.HasElectronicAngularMomentum then begin
        dind := Length(Left.v)-2;
        LambdaP := SignA*Left.v[dind];
        Lambda  := SignB*Right.v[dind];
        if (SignA*Left.v[dind+1]-LambdaP) <> (SignB*Right.v[dind+1]-Lambda) then begin
          { Sigma changed, do forbbiden, but:
            For degenerate states, only calculating one component of state, so need to check for cases
            where transition to "other" component is allowed }
          { TODO: Test probably too specific }
          if (Left.Phase = 0) and (Right.Phase = 0) and (ActualSymmetry <> 0) and (Abs(LambdaP-Lambda) = ActualSymmetry) then
            if (SignA*Left.v[dind+1]-LambdaP) = (-SignB*Right.v[dind+1]+Lambda) then begin
              if Lambda <> 0 then begin
                Lambda := -Lambda;
                SignB := -SignB;
              end else begin
                LambdaP := -LambdaP;
                SignA := -SignA;
              end;
            end else
              exit;
        end;
        if Lambda*LambdaP < 0 then { Check for allowed change in Lambda }
          Exit; { TODO: This is only correct for one photon transitions }
      end;
      { Check for no change in l }
      for vn := 0 to Molecule.nModes - 1 do
        if Molecule.IsDegenerate(Molecule.ModeInfo[vn].Symmetry,0) then begin
          dind := Molecule.ModeInfo[vn].dIndex;
          if SignA*Left.v[dind] <> SignB*Right.v[dind] then
            Exit; { l must not change }
        end;
      Result.re := TMultidimensionalFCF(TransitionMoment).Strength;
      dK := 0;
      Dmat := TransitionMatrix(Left,Right,True);
      {$IFDEF VERYSLOWTEST}
      r1 := Dmat[Left.Offset,Right.Offset];
      Brav := Copy(Left.v);
      Ketv := Copy(Right.v);
      for vn := 0 to Molecule.nModes - 1 do
        if Molecule.IsDegenerate(Molecule.ModeInfo[vn].Symmetry) then begin
          dind := Molecule.ModeInfo[vn].dIndex;
          Brav[dind] := SignA*Brav[dind];
          Ketv[dind] := SignB*Ketv[dind];
        end;
      r2 := Overlap(Brav, Ketv);
      if not SameValue(r1, r2) then begin
        r2 := Overlap(Brav, Ketv);
        raise Exception.Create('Help '+Left.BasisLabel+Right.BasisLabel);
      end;
      {$ENDIF}
      Result.re := Result.re * Dmat[Left.Offset,Right.Offset];
    end;
  // No - reduction occurs outside this routine. Result.re := Result.re * Sqrt(Three); { Pretend it is a reduced matrix element }
end;

procedure TElectronicState.SetLmatrix(Sender: TObject);
begin
  {$IFNDEF CONSOLE}
  ConnectLmatrixFrm(Self)
  {$ENDIF}
end;

function TElectronicState.GetSettingsNames(AIndex: Integer; out LowestValue: Integer): PPChar;
begin
  if Aindex = LSI then begin
    LowestValue := 0;
    Result := PointGroupInfo[Molecule.PointGroup].SettingsNames;
  end else
    Result := inherited GetSettingsNames(AIndex, LowestValue);
end;

function TElectronicState.GetzMatrix: TzMatrix;
begin
  Result := ForceFieldObjects[0] as TzMatrix;
end;

const
 ElectronicStateParameterNames: array [LSP..LSP+3] of string = (
'Spin-orbit coupling constant.',
'Spin - spin coupling constant',
'Spin-spin coupling constant giving lambda doubling for Omega = 0 states.',
'Maximum energy to include in the vibrational basis.'
);

function TElectronicState.HelpOnParameter(AIndex: Integer): string;
begin
  if (AIndex >= Low(ElectronicStateParameterNames))
     and (AIndex <= High(ElectronicStateParameterNames)) then
    Result := ElectronicStateParameterNames[AIndex]
  else
    Result := inherited HelpOnParameter(AIndex);
end;

function TElectronicState.HelpOnSetting(AIndex: Integer): string;
begin
  if AIndex = LSI then
    Result := 'Symmetry of electronic state'
  else
    Result := inherited HelpOnSetting(AIndex)
end;

function TElectronicState.LastError: string;
var
  i: Integer;
begin
  Result := '';
  for i := High(ForceFieldObjects) downto 0 do
    if Assigned(ForceFieldObjects[i]) and (ForceFieldObjects[i] is TlMatrixMNode) then begin
      Result := ForceFieldObjects[i].LastError;
      Exit;
    end;
end;

procedure TElectronicState.SetSettingFromStr(AIndex: Integer; const Str: string);
begin
  if AIndex = LSI then
    Symmetry := PointGroupInfo[Molecule.PointGroup].StrToSymmetry(Str, 0)
  else
    inherited;
end;

function TElectronicState.SettingToStr(AIndex: Integer): string;
begin
  if AIndex = LSI then
    if Symmetry <= Molecule.LastSym then
      Result := PointGroupInfo[Molecule.PointGroup].SymmetryToStr(Symmetry, 0)
    else
      Result := IntToStr(Symmetry)
  else
    Result := inherited SettingToStr(AIndex);
end;

function TElectronicState.EstimatedStateOrigin: Double;
var
  Ket: TVibrationalBasisState;
  Sigma: TQuantumNumber;
  E: Double;
begin
  { Estimate energy of lowest level - Used to stop Boltzman factors blowing up
    at low temperatures, or if there are many modes, when the total zero point
    energy can become very large }
  { Just use the simple matrix elements, so this may fail if perturbations are large }
  Ket := TVibrationalBasisState.Create(nil,0,0,0);
  try
    SetLength(Ket.v, Molecule.nQuantumNumbers);
    { Estimate lowest spin orbit + spin-spin energy }
    Sigma := -S;
    if Molecule.HasElectronicAngularMomentum then
      Ket.v[High(Ket.v)-1] := Molecule.AngularMomentum(Symmetry,0); { Lambda }
    while Sigma <= S do begin
      if Molecule.HasElectronicAngularMomentum then
        Ket.v[High(Ket.v)] := Ket.v[High(Ket.v)-1] + Sigma; { Omega }
      E := RotationalMatrixElement(Ket,Ket,1,1,nil,0).Re;
      if (Sigma = -S) or (E < Result) then
        Result := E;
      Inc(Sigma, 2);
    end;
  finally
    Ket.Free;
  end;
end;

{ Return symmetry of given particular combination of quantum numbers in v.
  Result is -1 if particular combination is invalid
   -2 if redundant;
  for a redundant wavefunction the symmetry must be determined by changing the sign of the signed quantum numbers.
  Where two symmetries are possible, the alternate symmetry is returned in AlternateSym;
  for this case phase=1 will have the returned symmetry and phase=-1 will have the AlternateSym
}
function TElectronicState.SymmetryOf(v: array of Integer; out AlternateSym:Integer): Integer;
  function AbelianSymmetry: Integer;
  var
    i: Integer;
  begin
    { Abelian groups with no degenerate modes }
    Result := Symmetry;
    for i := 0 to Molecule.nModes - 1 do
      if Odd(v[i]) then
        Result := Result xor Molecule.ModeInfo[i].Symmetry;
  end;
  function Linear(Centre:Boolean): Integer;
  var
    vn, l, Phase: Integer;
    K, Sigma: TQuantumNumber;
    AllAngularMomentumZero: Boolean;
    LastProjectionIsPositive: Boolean;
  begin
    if Centre then
      Result := Symmetry and 1 { u/g symmetry of state }
    else
      Result := 0;
    if Molecule.HasElectronicAngularMomentum then begin
      K := v[Molecule.nQuantumNumbers-1];
      Sigma := Molecule.AngularMomentum(Symmetry,0) - K;
    end else begin
      K := 0;
      Sigma := 0;
    end;
    AllAngularMomentumZero := not Molecule.IsDegenerate(Symmetry,0) and (Sigma = 0);
    LastProjectionIsPositive := False;
    for vn := 0 to Molecule.nModes - 1 do begin
      if Centre and Odd(Molecule.ModeInfo[vn].Symmetry) and Odd(v[vn]) then
        Result := Result xor 1; { Track u/g symmetry }
      if Molecule.IsDegenerate(Molecule.ModeInfo[vn].Symmetry,0) then begin
        l := v[Molecule.ModeInfo[vn].dIndex]; { Vibrational angular momentum }
        Inc(K,2*l);
        AllAngularMomentumZero := AllAngularMomentumZero and (l = 0);
        if l <> 0 then
          LastProjectionIsPositive := l > 0;
      end;
    end;
    if K < 0 then begin
      { States with overall angular momentum < 0 are redundant - these will be the same as those with the other sign }
      Result := -2;
      Exit;
    end else if Odd(K) then
      K := K div 2
    else begin
      if K > 0 then
        K := K div 2 + 1
      else if AllAngularMomentumZero then
        { Must be Sigma+ or Sigma-, reflecting electronic state,
          as vibrations can't be Sigma- }
        if Centre then
          K := Symmetry div 2
        else
          K := Symmetry
      else begin
        { Sigma+ or Sigma- with non-zero angular momenta;
          In this case both +ve and -ve combinations muct be considered, so return two answers }
        { First, remove redundancies, choose the state with:
            Lambda > 0,
            else Omega > 0
            else with the highest numbered l value > 0; note that
              LastProjectionIsPositive is already set for this last condition }
          if Molecule.HasElectronicAngularMomentum then
            if v[High(v)-1] <> 0 then
              LastProjectionIsPositive := v[High(v)-1] > 0
            else if v[High(v)] <> 0 then
              LastProjectionIsPositive := v[High(v)] > 0;
          if not LastProjectionIsPositive then begin
            Result := -1; { Redundant }
            Exit;
          end;
        { Now work out correct phase factor to give Sigma+ or Sigma- }
        { Effect of Sigma(v) is (-1)^(l1+l2+...)(-1)^(Lambda+s)(-1)^(S-Sigma) * (state with all angular momenta reversed)
          where s is 1 for a Sigma- state, and zero otherwise.
          This is the same as (-1)^(K+s-S); if this is +ve then the + combination is Sigma+
          This section is only reached if K = 0 }
        if Molecule.AngularMomentum(Symmetry,0) = 0 then
          { Work out s }
          if Centre then
            Phase := Symmetry div 2
          else
            Phase := Symmetry
        else
          Phase := 0;
        Phase := Minus(2*Phase-S);
        if Phase < 0 then begin
          K := 1; { + combination is Sigma- }
          AlternateSym := Result; { Sigma+ g/u }
        end else
          if Centre then
            AlternateSym := 2 + Result { Sigma- g/u }
          else
            AlternateSym := 1;    { Sigma- }
      end;
    end;
    if Centre then  { Add back g/u symmetry }
      Result := K*2 + Result
    else
      Result := K;
  end;
begin
  AlternateSym := -1; { Flag only one result possible }
  case Molecule.PointGroup of
    CInfinityv: Result := Linear(False);
    DInfinityh: Result := Linear(True);
  else
    Result := AbelianSymmetry;
  end;
end;

procedure TElectronicState.ClearObjectChain;
var
  i: Integer;
begin
  for i := 0 to High(ForceFieldObjects) do
    ForceFieldObjects[i] := nil;
  for i := 0 to ComponentCount-1 do
    if Components[i] is TlMatrixMNode then
      TlMatrixMNode(Components[i]).UpdateObjectChain;
end;

function TElectronicState.UpdateObjectChain:string;
var
  i, j: Integer;
  Obj: TActiveMNodeWithArray;
begin
  Result := '';
  ClearObjectChain;
  for i := 0 to ComponentCount-1 do
    if (Components[i] is TActiveMNodeWithArray) and TActiveMNodeWithArray(Components[i]).ReallyActive then begin
      Obj := TActiveMNodeWithArray(Components[i]);
      for j := 0 to High(ForceFieldObjects) do
        if Obj is ForceFieldClasses[j] then
          if Assigned(ForceFieldObjects[j]) then begin
            Result := 'More than one active '+Obj.ClassName;
            Exit;
          end else begin
            ForceFieldObjects[j] := Obj;
            Break;
          end;
    end;
end;

{ TVibrationalMode }

procedure TVibrationalMode.AdjustParameterNames;
var
  i: Integer;
begin
  { Add/remove l matrix parameters as required }
  for i := nParameters+1 to FirstDynamicParameter + 3*Molecule.nNuclei - 1 do
    AddParameter(Format('l%s%d',[Char(Ord('x')+(i-FirstDynamicParameter) mod 3),(i-FirstDynamicParameter) div 3]));
  for i := nParameters downto FirstDynamicParameter + 3*Molecule.nNuclei do
    DeleteParameter(i);
end;

function TVibrationalMode.PertEquiv(AIndex: Integer):TVibPertEquiv;
begin
  SetLength(Result.Powers, Molecule.nModes);
  case AIndex of
    1..4: { Omega .. zOmega }
      begin
        Result.Op := x;
        Result.Powers[NodeIndex] := AIndex;
      end;
//    5: { gll }
    6: { epsOmega }
      begin
        Result.Op := q;
        Result.Powers[NodeIndex] := -Molecule.AngularMomentum(Symmetry,0);
      end;
  else
    SetLength(Result.Powers,0) { Flag not (yet) implemented }
  end;
end;

constructor TVibrationalMode.Create(AOwner: TComponent);
var
  n: Integer;
begin
  SavedProperties := SavedProperties + ',vMax=5';
  FState := AOwner as TElectronicState;
  inherited Create(AOwner);
  { To avoid large calculations by default, adjust default vMax depending on the number of other modes }
  if Molecule.nModes > 3 then
    n := 0
  else
    n := 3;
  ModifySetting(IndexOfSetting('vMax'), n, False);
end;

procedure TVibrationalMode.AddHandlers;
begin
  inherited;
  AddHandler('Sort', 'Sort modes by symmetry and frequency', Sort);
  AddHandler('Global Shift Up', 'Shift mode in all states', GlobalMoveUp);
  AddHandler('Global Shift Down', 'Shift mode in all states', GlobalMoveDown);
end;

procedure TVibrationalMode.Sort(Sender: TObject);
var
  i, d: Integer;
  Change: Boolean;
begin
  repeat
    Change := False;
    for i := 0 to Molecule.nModes - 2 do begin
      d := State.Mode(i).Symmetry - State.Mode(i+1).Symmetry;
      if (d > 0) or ((d=0) and (State.Mode(i).Omega < State.Mode(i+1).Omega)) then begin
        Change := True;
        State.Mode(i).GlobalShift(1);
      end;
    end;
  until not Change;
end;

procedure TVibrationalMode.GlobalMoveDown(Sender: TObject);
begin
  GlobalShift(1);
end;

procedure TVibrationalMode.GlobalMoveUp(Sender: TObject);
begin
  GlobalShift(-1);
end;

procedure TVibrationalMode.GlobalShift(Shift: Integer);
var
  Source, Target: TVibrationalMode;
  s: string;
  i, j, MyIndex, Temp: Integer;
begin
  if (NodeIndex+Shift < 0) or (NodeIndex+Shift >= Molecule.nModes) then
    Exit;
  Target := State.Mode(NodeIndex+Shift);
  { Move global values once }
  s := Name;
  if (s <> Format('v%d', [NodeIndex+1])) and (Target.Name <> Format('v%d', [Target.NodeIndex+1])) then begin
    { Only exchange names if using non-default names }
    Name := Target.Name;
    Target.Name := s;
  end;
  MyIndex := NodeIndex;
  { Now move all other settings }
  for i := 0 to Molecule.nManifolds - 1 do
    with Molecule.Manifold[i] do begin
      for j := 0 to nStates - 1 do
        with State[j] as TElectronicState do begin
          Source := Mode(MyIndex);
          Target := Mode(MyIndex+Shift);
          if Shift > 0 then
            Source.ComponentIndex := Target.ComponentIndex
          else
            Target.ComponentIndex := Source.ComponentIndex;
          AdjustModes(False);
        end;
      for j := 0 to nPerturbations - 1 do
        with Perturbation[j] do begin
          { Shift operator powers }
          Temp := Settings[FirstDynamicSetting+MyIndex];
          Settings[FirstDynamicSetting+MyIndex] := Settings[FirstDynamicSetting+MyIndex+Shift];
          Settings[FirstDynamicSetting+MyIndex+Shift] := Temp;
        end;
    end;
  for i := 0 to Molecule.nTransitionMoments - 1 do
    with Molecule.GetTransitionMoments(i) do
      for j := 0 to ComponentCount - 1 do
        if TransitionMoment[j] is TMultidimensionalFCF then
          with TMultidimensionalFCF(TransitionMoment[j]) do begin
            SwapModes(MyIndex,MyIndex+Shift);
            AdjustModes(False);
          end;
  { Reset symmetry, to ensure copied to global array }
  SetSetting(1,GetSetting(1));
  Target.SetSetting(1,Target.GetSetting(1));
end;

const
 VibrationalModeParameterNames: array [1..6] of string = (
'Vibrational Frequency.',
'(v+1/2)^2 anharmonicity.',
'(v+1/2)^3 anharmonicity.',
'(v+1/2)^4 anharmonicity.',
'l^2 term.',
'Renner-Teller term'
);

function TVibrationalMode.HelpOnParameter(AIndex: Integer): string;
begin
  if AIndex > High(VibrationalModeParameterNames) then begin
    AIndex := AIndex-High(VibrationalModeParameterNames)-1;
    Result := 'l matrix element for ' + Char(Ord('x') + AIndex mod 3)
              + ' co-ordinate for nucleus ' + (Owner as TState).Nucleus(AIndex div 3).Name
              + ' / Sqrt(amu)';
  end else if AIndex >= Low(VibrationalModeParameterNames) then
    Result := VibrationalModeParameterNames[AIndex]
  else
    Result := inherited HelpOnParameter(AIndex);
end;

function TVibrationalMode.nCalculatedParameters: Integer;
begin
  if NodeIndex < Length(State.alpha) then
    Result := 12
  else
    Result := 0;
end;

function TVibrationalMode.CalculatedParameter(Index: Integer): Double;
var
  xyz, xyzp: Word;
begin
  if Index in [1..3] then
    Result := State.alpha[NodeIndex][Index]
  else if Index in [4..12] then begin
    DivMod(Index-4,3,xyz,xyzp);
    Result := State.dIdQ[NodeIndex][xyzp+1,xyz+1];
  end else
    Result := inherited CalculatedParameter(Index);
end;

function TVibrationalMode.CalculatedParameterName(Index: Integer): string;
var
  xyz, xyzp: Word;
begin
  if Index in [1..3] then
    Result := 'alpha'+'xyz'[Index]
  else if Index in [4..12] then begin
    DivMod(Index-4,3,xyz,xyzp);
    Result := 'a'+'xyz'[xyzp+1]+'xyz'[xyz+1];
  end else
    Result := inherited CalculatedParameterName(Index);
end;

const
 VibrationalModeSettingNames: array [2..3] of string = (
'Minimum vibrational quantum number for this mode to include in the basis.',
'Maximum vibrational quantum number for this mode to include in the basis.'
);

function TVibrationalMode.HelpOnSetting(AIndex: Integer): string;
begin
  if (AIndex >= Low(VibrationalModeSettingNames))
     and (AIndex <= High(VibrationalModeSettingNames)) then
    Result := VibrationalModeSettingNames[AIndex]
  else
    Result := inherited HelpOnSetting(AIndex);
end;

{ TVibrationalBasisState }

function TVibrationalBasisState.VibLabel: string;
var
  vn, dind: Integer;
  Mol: TVibratingMolecule;
  FirstQno: Boolean;
begin
  Mol := (State as TElectronicState).Molecule;
  FirstQno := True;
  Result := '';
  for vn := 0 to Mol.nModes - 1 do
    if Mol.ModeInfo[vn].ShowQno then begin
      if not FirstQno then
        Result := Result + ',';
      FirstQno := False;
      Result := Result + IntToStr(v[vn]);
      if Mol.IsDegenerate(Mol.ModeInfo[vn].Symmetry,0) then
        Result := Result + '(' + IntToStr(v[Mol.ModeInfo[vn].dIndex]) +')';
    end;
  if Mol.HasElectronicAngularMomentum then begin { Lambda, Omega }
    dind := High(v)-1;
    if (v[dind] <> 0) or (State.S > 0)then begin
      Result := Result + '[';
      if v[dind] <> 0 then begin
        if v[dind] > 0 then
          Result := Result + '+'
        else
          Result := Result + '-';
        if State.S > 0 then
          Result := Result + ','
      end;
      if State.S > 0 then
        Result := Result + State.Manifold.Mixture.QuantumNumberToStr(v[dind+1]);
      Result := Result + ']';
    end;
  end else if FirstQno then
    Result := '0'; { Must put something here }
  if Phase > 0 then
    Result := Result + '+'
  else if Phase < 0 then
    Result := Result + '-';
  Result := Result + HyperfineStateLabel;
  { TODO: Make option  Result := Result + Format('%d',[StateIndex]); }
end;

function TVibrationalBasisState.BasisLabel: string;
begin
  Result := StateLabel(qftJ);
  if State.Mixture.ShowSubBasis then
    Result := '{'+IntToStr(SubBasisIndex)+'} ' + Result;
end;

function TVibrationalBasisState.BasisLabelKey: string;
begin
  Result := StateLabelKey;
  if State.Mixture.ShowSubBasis then
    Result := '{Sub} '+Result;
end;

function TVibrationalBasisState.StateLabel(AQuantumNumberFormat: TQuantumNumberFormat): string;
begin
  if State.Manifold.Molecule.nManifolds > 1 then
    Result := State.Manifold.Name + ' '
  else
    Result := '';
  Result := Result + State.Name + ' ' + VibLabel;
end;

function TVibrationalBasisState.StateLabelKey: string;
begin
  Result := 'Name ' + VibLabelKey;
end;

function TVibrationalBasisState.VibLabelKey: string;
var
  vn: Integer;
  Mol: TVibratingMolecule;
  FirstQno: Boolean;
begin
  Mol := (State as TElectronicState).Molecule;
  FirstQno := True;
  for vn := 0 to Mol.nModes - 1 do
    if Mol.ModeInfo[vn].ShowQno then begin
      if not FirstQno then
        Result := Result + ',';
      FirstQno := False;
      Result := Result + Mol.ModeInfo[vn].Name;
      if Mol.IsDegenerate(Mol.ModeInfo[vn].Symmetry,0) then begin
        Result := Result + '(l' + Copy(Mol.ModeInfo[vn].Name,2,255) +')';
      end;
    end;
  if Mol.HasElectronicAngularMomentum then
    Result := Result + '[Lambda,Omega]'
  else if FirstQno then
    Result := 'v';
  Result := Result + '+-';
end;

function TVibrationalBasisState.MakeSignature(SignA: Integer; AKet: TBasisState; SignB: Integer; MNode: TMNode; Symmetrised: Boolean): TIntegerArray;
var
  vn, i: Integer;
  Mol: TVibratingMolecule;
  Ket: TVibrationalBasisState;
begin
  { Signature is array of changes in vibronic quantum numbers }
  SetLength(Result,Length(v));
  Mol := (State as TElectronicState).Molecule;
  Ket := AKet as TVibrationalBasisState;
  for vn := 0 to Mol.nModes - 1 do begin
    Result[vn] := v[vn] - Ket.v[vn];
    if Mol.IsDegenerate(Mol.ModeInfo[vn].Symmetry,0) then begin
      i := Mol.ModeInfo[vn].dIndex;
      Result[i] := (SignA*v[i]-SignB*Ket.v[i]);
    end;
  end;
  if Mol.HasElectronicAngularMomentum then begin { Lambda, Omega }
    i := High(v)-1;
    Result[i] := (SignA*v[i]-SignB*Ket.v[i]) div 2;
    Result[i+1] := (SignA*v[i+1]-SignB*Ket.v[i+1]) div 2;
  end;
end;

function TVibrationalBasisState.Qno(QnoType: TQnoType): TQuantumNumber;
var
  o: Integer;
begin
  o := Ord(QnoType)-Ord(qtA);
  if o = Length(v) then
    Result := Phase
  else if (o < Length(v)) and (o >= 0) then
    Result := v[o]
  else
    Result := inherited Qno(QnoType);
end;

procedure TVibrationalBasisState.SetSymbolicVariables(SignA: Integer; AKet: TBasisState; SignB: Integer; MNode: TMNode);
var
  vn, dind: Integer;
  Mol: TVibratingMolecule;
  Ket: TVibrationalBasisState;
begin
  Mol := (State as TElectronicState).Molecule;
  Ket := AKet as TVibrationalBasisState;
  for vn := 0 to Mol.nModes - 1 do begin
    vVariable(vn).Value := Ket.v[vn];
    if Mol.IsDegenerate(Mol.ModeInfo[vn].Symmetry,0) then begin
      dind := Mol.ModeInfo[vn].dIndex;
      lVariable(vn).Value := SignB*Ket.v[dind];
    end;
  end;
  if Mol.HasElectronicAngularMomentum then begin { Lambda, Omega }
    dind := High(v)-1;
    eLambda.Value := SignB*Ket.v[dind] / 2;
    eOmega.Value := SignB*Ket.v[dind+1] / 2;
  end;
end;

function TVibrationalBasisState.SymbolicVariableValues: string;
begin
  Result := vVariables.VariableValues + ' ' + lVariables.VariableValues + ' ' + lVariables.VariableValues;
end;

function TVibrationalBasisState.TxName(Upper, Lower: TBasisState): string;
begin
  Result := '(' + (Upper as TVibrationalBasisState).VibLabel + ')-(' + (Lower as TVibrationalBasisState).VibLabel + ')';
end;

function TVibrationalBasisState.TxNameKey(Upper, Lower: TBasisState): string;
begin
  Result := '(' + (Upper as TVibrationalBasisState).VibLabelKey + ')-(' + (Lower as TVibrationalBasisState).VibLabelKey + ')';
end;

{ TMultidimensionalFCF }

procedure TMultidimensionalFCF.AddHandlers;
begin
  inherited;
  AddHandler('Dushinsky Matrix...', 'View and edit Dushinsky transformation', ShowDushinsky);
  AddHandler('Print Matrices', 'Print detailed information about Franck Condon calculation', PrintInfo);
  AddHandler('Set Modes From Calculated', 'Set mode displacements and rotations from l matrices', SetModeInfo);
end;

procedure TMultidimensionalFCF.SaveOffsets(out Offsets: TIntegerArray; SourceSample: TVibrationalBasisState;
    CurrentCacheIndex: Integer);
var
  i, o: Integer;
  SourceState: TElectronicState;
  Ev: TEigenValues;
  Basis: TVibrationalBasisState;
begin
  { Save mapping between original position in standard order and index into FC array }
  SourceState := SourceSample.State as TElectronicState;
  o := 1;
  for i := Molecule.nModes downto 1 do begin
    o := o*(SourceState.Mode(i-1).vmax+1);
  end;
  SetLength(Offsets, o+1);
  Ev := SourceSample.Owner as TEigenvalues;
  for i := 1 to High(Ev.Basis) do begin
    Basis := Ev.Basis[i] as TVibrationalBasisState;
    if (Basis.State = SourceState) and (CacheIndex(Basis.v)=CurrentCacheIndex) then
      Offsets[Basis.OriginalPosition] := Basis.Offset;
  end;
end;

procedure TMultidimensionalFCF.SetModeInfo(Sender: TObject);
begin
  PrepareMatrices(True,True,False);
end;

procedure TMultidimensionalFCF.AdjustModes(FixNumber: Boolean);
var
  j, n: Integer;
begin
  if FixNumber then
    n := AdjustNodeIndexes(TVibrationalModeChange, FModes, nil, opInsert, Molecule.nModes)
  else
    n := AdjustNodeIndexes(TVibrationalModeChange, FModes);
  for j := 0 to n - 1 do
    Mode(j).AdjustParameterNames;
end;

function TMultidimensionalFCF.CacheIndex(v: TIntegerArray): Integer;
var
  CacheSpan, vn, MaxSym, Sym: Integer;
begin
  Result := 0;
  Sym := 0;
  MaxSym := 0;
  for vn := Molecule.nModes - 1 downto 0 do
    if Molecule.IsDegenerate(Molecule.ModeInfo[vn].Symmetry,0) then begin
      CacheSpan := Min(BraState.Mode(vn).vMax, KetState.Mode(vn).vMax)+1;
      Result := Result*CacheSpan + Abs(v[Molecule.ModeInfo[vn].dIndex]);
    end else begin
      { TODO: The stuff below will require revsion for non-abelian groups }
      MaxSym := MaxSym or Molecule.ModeInfo[vn].Symmetry;
      if Odd(v[vn]) then
        Sym := Sym xor Molecule.ModeInfo[vn].Symmetry;
    end;
  Result := Result*(MaxSym+1) + Sym;
end;

procedure TMultidimensionalFCF.ClearCache(ClearChildren: Boolean);
var
  i: Integer;
begin
  inherited ClearCache(ClearChildren);
  ClearMatrices;
  for i := 0 to High(FMatrixCache) do
    FMatrixCache[i].Free;
  SetLength(FMatrixCache,0);
end;

procedure TMultidimensionalFCF.ClearMatrices;
begin
  FreeAndNil(Q1);
  FreeAndNil(P1);
  FreeAndNil(P1);
  FreeAndNil(R);
  FreeAndNil(Rdelta);
  FreeAndNil(P1Delta);
end;

constructor TMultidimensionalFCF.Create(AOwner: TComponent);
begin
  SavedProperties := SavedProperties + ',Symmetry=Auto';
  inherited;
  if Assigned(AOwner) then
    FMolecule := Owner.Owner as TVibratingMolecule;
end;

destructor TMultidimensionalFCF.Destroy;
begin
  ClearCache(False);
  FreeAndNil(d);
  FreeAndNil(S);
  FreeAndNil(CacheLock);
  inherited;
end;

function TMultidimensionalFCF.FixupProperties: Exception;
var
  i: Integer;
begin
  if (nWorkers > 0) and not Assigned(CacheLock) then
    CacheLock := TMultiReadExclusiveWriteSynchronizer.Create;
  Result := inherited FixupProperties;
  if Molecule.nNuclei = 0 then
    Calculate := False;
  AdjustModes(True);
  if not Assigned(Result) then
    try
      FBraState := Bra as TElectronicState;
      FKetState := Ket as TElectronicState;
      FActualSymmetry := Symmetry;
      if Symmetry < 0 then begin
        { Work out symmetry of transition moment }
        FActualSymmetry := -1;
        for i := 0 to Molecule.LastSym do
          if PointGroupInfo[Molecule.PointGroup].Allowed(BraState.Symmetry, i, KetState.Symmetry, 0) then begin
            if FActualSymmetry >= 0 then begin
              Result := EReportException.Create('Auto transition moment not possible for '+UniqueName);
              Exit;
            end else begin
              FActualSymmetry := i;
              VibronicallyAllowed := vaYes;
            end;
          end
      end else
        if PointGroupInfo[Molecule.PointGroup].Allowed(BraState.Symmetry, ActualSymmetry, KetState.Symmetry, 0) then
          VibronicallyAllowed := vaYes
        else
          VibronicallyAllowed := vaNo;
      I00 := 1;
      PrepareMatrices(False,False,True);
      if I00 = 0 then
        Result := EWarningException.Create('Zero Franck-Condon factor for origin band');
    except
      on E: Exception do begin
        Result := Exception.Create(E.Message);
      end;
    end;
end;

function TMultidimensionalFCF.GetSettingsNames(AIndex: Integer; out LowestValue: Integer): PPChar;
begin
  if Aindex = 1 then begin
    LowestValue := -1;
    Result := PointGroupInfo[Molecule.PointGroup].SettingsNames;
    Dec(Result); { Allow 'Auto' also }
  end else
    Result := inherited GetSettingsNames(AIndex, LowestValue);
end;

const
  MultidimensionalFCFSettingNames: array[1..2] of string = (
'Symmetry of electronic transition moment.',
'True to calculate normal co-ordinate displacements and Dushinsky matrix from l matrix.'
);

function TMultidimensionalFCF.HelpOnSetting(AIndex: Integer): string;
begin
  if (AIndex = IndexOfSetting('Symmetry')) and (Symmetry < 0) then
    Result := SettingToPrettyStr(AIndex)+'. '
  else
    Result := '';
  if (AIndex >= Low(MultidimensionalFCFSettingNames))
     and (AIndex <= High(MultidimensionalFCFSettingNames)) then
    Result := Result + MultidimensionalFCFSettingNames[AIndex]
  else
    Result := inherited HelpOnSetting(AIndex);
end;

procedure TMultidimensionalFCF.InvertTransformation;
begin
  { TODO: This has no effect, as d and S are overwritten } 
  PrepareMatrices(True,False,False);
  Invert(S);
  d.PreMultiply(S);
  d.Scale(-1);
  Writeln(Log,'Inverse Transform');
  d.Write(Log,'d');
  S.Write(Log,'S');
  Modified := True;
end;

function TMultidimensionalFCF.Mode(Index: Integer): TVibrationalModeChange;
begin
  if (Index >= Length(FModes)) and (Index < Molecule.nModes) then
    AdjustModes(True);
  Result := FModes[Index] as TVibrationalModeChange;
end;

{ Degenerate modes from:
  E V Doktorov, I A Malkin and V I Man'ko
  J Phys. B: Atom. Molec. Phys. 9, 507 (1976) }
{ Note that the sign of l is ignored in this rourine }
function TMultidimensionalFCF.Overlap(var ABra, AKet: array of Integer): Double;
var
  vp, v, l, vn, vn1: Integer;
  Scale: Double;
  OmegaP, Omega, Vm: Double;
begin
  { Note that arguments are not constants to avoid multiple copies of quantum numbers }
  { This makes it not thread safe unless routine called with a thread local copy } 
  if Aborted then { This routine can be called repeatedly recursively }
    Abort;
  Result := 0;
  for vn := 0 to Molecule.nModes-1 do
    { Pick a mode to reduce }
    if AKet[vn] > 0 then
      if Molecule.IsDegenerate(Molecule.ModeInfo[vn].Symmetry,0) then begin
        { Special handling for reducing degenerate modes }
        { TODO: currently does not take account of mode mixing for degenerate modes }
        v := AKet[vn];
        l := Abs(AKet[Molecule.ModeInfo[vn].dIndex]);
        if v > l then begin
          vp := ABra[vn];
          Omega := (Ket as TElectronicState).Mode(vn).Omega;
          OmegaP := (Bra as TElectronicState).Mode(vn).Omega;
          Scale := Omega + OmegaP;
          { Reduce v }
          if v > 3 then begin
            Vm := (Omega-OmegaP)/Scale*Sqrt((v-2-l)*(v-2+l)/(v-l)/(v+l));
            if Vm <> 0 then begin
              Dec(AKet[vn],4);
              Result := Vm*Overlap(ABra,AKet);
              Inc(AKet[vn],4);
            end;
          end;
          if vp > 1 then begin
            Vm := 2*Omega/Scale*Sqrt((vp-l)*(vp+l)/(v-l)/(v+l));
            if Vm <> 0 then begin
              Dec(AKet[vn],2);
              Dec(ABra[vn],2);
              Result := Result + Vm*Overlap(ABra,AKet);
              Inc(ABra[vn],2);
              Inc(AKet[vn],2);
            end;
          end;
          Dec(AKet[vn],2);
          Result := Result + 2*(OmegaP*(v-1)-Omega*(vp+1))/Scale/Sqrt((v-l)*(v+l))*Overlap(ABra,AKet);
          Inc(AKet[vn],2);
          Exit;
        end else if v < l then
          raise Exception.Create('Overlap1') { Sanity check }
        else
          Continue; { This degnerate mode at minmum value for l }
      end else begin
        { Non degenerate mode }
        Scale := AKet[vn];
        Dec(AKet[vn]);
        for vn1 := 0 to Molecule.nModes-1 do
          if ABra[vn1] > 0 then begin
            if (vn=vn1) or (Molecule.ModeInfo[vn1].Symmetry = 0) then begin
              Dec(ABra[vn1]);
              Result := Result + 2*R[vn+1,vn1+1]*Sqrt((ABra[vn1]+1)/Scale)*Overlap(ABra,AKet);
              Inc(ABra[vn1]);
            end;
          end;
        for vn1 := 0 to Molecule.nModes-1 do
          if AKet[vn1] > 0 then begin
            if (vn=vn1) or (Molecule.ModeInfo[vn1].Symmetry = 0) then begin
              Dec(AKet[vn1]);
              Result := Result + Q1[vn+1,vn1+1]*Sqrt((AKet[vn1]+1)/Scale)*Overlap(ABra,AKet);
              Inc(AKet[vn1]);
            end;
          end;
        if Molecule.ModeInfo[vn].Symmetry = 0 then
          Result := Result - Rdelta[vn+1]*Sqrt(2/Scale)*Overlap(ABra,AKet);
        Inc(AKet[vn]);
        Exit;
      end;
  { Ket all zero; reduce Bra }
  for vn := 0 to Molecule.nModes-1 do
    { Pick a mode to reduce }
    if ABra[vn] > 0 then begin
      if Molecule.IsDegenerate(Molecule.ModeInfo[vn].Symmetry,0) then begin
        { Special handling for reducing degenerate modes }
        vp := ABra[vn];
        l := Abs(ABra[Molecule.ModeInfo[vn].dIndex]);
        if vp > l then begin
          v := AKet[vn];
          Omega := (Ket as TElectronicState).Mode(vn).Omega;
          OmegaP := (Bra as TElectronicState).Mode(vn).Omega;
          Scale := OmegaP + Omega;
          { Reduce vp }
          if vp > 3 then begin
            Vm := (OmegaP-Omega)/Scale*Sqrt((vp-2-l)*(vp-2+l)/(vp-l)/(vp+l));
            if Vm <> 0 then begin
              Dec(ABra[vn],4);
              Result := Vm*Overlap(ABra,AKet);
              Inc(ABra[vn],4);
            end;
          end;
          if v > 1 then begin
            Vm := 2*OmegaP/Scale*Sqrt((v-l)*(v+l)/(vp-l)/(vp+l));
            if Vm <> 0 then begin
              Dec(AKet[vn],2);
              Dec(ABra[vn],2);
              Result := Result + Vm*Overlap(ABra,AKet);
              Inc(ABra[vn],2);
              Inc(AKet[vn],2);
            end;
          end;
          Dec(ABra[vn],2);
          Result := Result + 2*(Omega*(vp-1)-OmegaP*(v+1))/Scale/Sqrt((vp-l)*(vp+l))*Overlap(ABra,AKet);
          Inc(ABra[vn],2);
          Exit;
        end else if vp < l then
          raise Exception.Create('Overlap2') { Sanity check }
        else
          Continue; { This degnerate mode at minmum value for l }
      end else begin
        Scale := ABra[vn];
        Dec(ABra[vn]);
        for vn1 := 0 to Molecule.nModes-1 do { Note that this is redundant code }
          if AKet[vn1] > 0 then begin
            if (vn=vn1) or (Molecule.ModeInfo[vn1].Symmetry = 0) then begin
              Dec(AKet[vn1]);
              Result := Result + 2*R[vn+1,vn1+1]*Sqrt((AKet[vn1]+1)/Scale)*Overlap(ABra,AKet);
              Inc(AKet[vn1]);
            end;
          end;
        for vn1 := 0 to Molecule.nModes-1 do
          if ABra[vn1] > 0 then begin
            if (vn=vn1) or (Molecule.ModeInfo[vn1].Symmetry = 0) then begin
              Dec(ABra[vn1]);
              Result := Result + P1[vn+1,vn1+1]*Sqrt((ABra[vn1]+1)/Scale)*Overlap(ABra,AKet);
              Inc(ABra[vn1]);
            end;
          end;
        if Molecule.ModeInfo[vn].Symmetry = 0 then
          Result := Result + P1Delta[vn+1]*Sqrt(2/Scale)*Overlap(ABra,AKet);
        Inc(ABra[vn]);
        Exit;
      end;
    end;
  Result := I00;
  for vn := 0 to Molecule.nModes-1 do
    { Special action needed for degenerate modes - these will have v' = v = l, rather than 0;
      in addition I00 is wrong, as mode should have been counted twice }
    if Molecule.IsDegenerate(Molecule.ModeInfo[vn].Symmetry,0) then begin
      l := Abs(ABra[Molecule.ModeInfo[vn].dIndex]);
      if l <> Abs(AKet[Molecule.ModeInfo[vn].dIndex]) then begin
        Result := 0;
        Exit;
      end;
      l := Abs(l);
      Omega := (Ket as TElectronicState).Mode(vn).Omega;
      OmegaP := (Bra as TElectronicState).Mode(vn).Omega;
      Result := Result * Sqrt(Math.IntPower(2*Sqrt(Omega*Omegap)/(Omega+Omegap),2*l+1))
    end;
end;

function TMultidimensionalFCF.PairName: string;
begin
  Result := 'FCF';
end;

procedure TMultidimensionalFCF.PrepareDushinsky(nModes: Integer; Print: Boolean; SetModeInfo: Boolean);
var
  C, T, Rc, BestS: TMatrix;
  slam, Req, Rpeq, SumV: TColumnVector;
  BestTrace, Det: Double;
  nSMatrix, BestSMatrix, nNuclei: Integer;
  Left, Right: TElectronicState;
  zLeft, zRight: TzMatrix;
  function MakeDus(Z: TMatrix):TMatrix;
    procedure PostMult(Invert: Boolean);
    var
      n, i, j, ia: Integer;
      M: Double;
    begin
      ia := -1;
      for n := 0 to nNuclei - 1 do begin
        zLeft.NextRealAtom(ia);
        M := Sqrt(zLeft.Atoms[ia].Mass);
        if Invert then
          M := 1/M;
        for i := 1 to 3 do
          for j := 1 to Result.RowCount do
            Result[j,3*n+i] := Result[j,3*n+i] * M;
      end;
    end;
  begin
//    Result := Left.Madm.Transpose;
//    if Assigned(Z) then
//      Result.PostMultiply(Z);
//    Result.PostMultiply(Right.adm);
    Result := Left.lMatrix.Transpose;
    PostMult(False);
    if Assigned(Z) then
      Result.PostMultiply(Z);
    PostMult(True);
    Result.PostMultiply(Right.lMatrix);
  end;
  procedure TryS(const signs: array of Integer);
  var
    S, Cinv, Z, D: TMatrix;
    i, j, m: Integer;
    Trace: Double;
  begin
    { Eigenvalue check }
{
    if Print then begin
      S := Rc.Transpose;
      Cinv := C.Transpose;
      S.PostMultiply(Cinv);
      Cinv.Free;
      S.PostMultiply(C);
      S.PostMultiply(Rc);
      S.Write(Log,'RCtCRt');
      S.Free;
    end;
}
    S := TMatrix.Create(Rc);
    for i := 1 to 3 do
      for j := 1 to 3 do
        S[i,j] := S[i,j] * Sqrt(slam[j]) * signs[j-1];
//    if Print then
//      S.Write(Log,'Rt Lambda Sqrt(lambda)');
    Cinv := Rc.Transpose;
    S.PostMultiply(Cinv);
    Cinv.Free;
//    if Print then
//      S.Write(Log,'Rt Lambda Sqrt(lambda) R');
    Cinv := TMatrix.Create(C);
    Invert(Cinv);
    if Print then
      Cinv.Write(Log,'C-1');
    S.PostMultiply(Cinv);
    Cinv.Free;
    Trace := S[1,1] + S[2,2] + S[3,3];
    Det := Determinant(S);
    Inc(nSMatrix);
    if Print then begin
      S.Write(Log,'S trial #'+IntToStr(nSMatrix)+', trace = '+FloatToStr(Trace)+', determinant = '+FloatToStr(Det));
      Z := TMatrix.Create(3*nNuclei,3*nNuclei);
      Z.Zero;
      for m := 0 to nNuclei - 1 do
        for i := 1 to 3 do
          for j := 1 to 3 do
            Z[3*m + i, 3*m+j] := S[i,j];
      D := MakeDus(Z);
      D.Write(Log,'Trial Dushinsky');
      D.Free;
      Z.Free;
    end;
    if (Det > 0) and (not Assigned(BestS) or (Trace > BestTrace)) then begin
      BestS.Free;
      BestS := S;
      BestTrace := Trace;
      BestSMatrix := nSMatrix;
    end else
      S.Free;
  end;
var
  i, j, k, m, ia: Integer;
  Sum, v: Double;
  sn, cs: Extended;
  AllZero: Boolean;
  Z, De: TMatrix;
begin
  { Prepare Matrices as per Gruner and Brumer }
  ClearMatrices;
  if not Assigned(d) then begin
    { Initial creation }
    d := TColumnVector.Create(0);
    S := TMatrix.Create(0,0);
  end;
  { Update sizes as required }
  d.SetSizeAndClear(nModes,True);
  m := S.RowCount;
  S.Resize(nModes,nModes,True);
  if m < nModes then
    for i := m+1 to nModes do
      S[i,i] := 1;
  if Calculate and (Molecule.nNuclei > 0) then begin
    Left := Bra as TElectronicState;
    zLeft := Left.ForceFieldNode(TzMatrix) as TzMatrix;
    nNuclei := Left.Molecule.nNuclei;
    Right := Ket as TElectronicState;
    zRight := Right.ForceFieldNode(TzMatrix) as TzMatrix;
    { d ( = K Chen ) = displacement between normal co-ordinates in the two states }
    for m := 0 to nModes - 1 do begin
      sum := 0;
      ia := -1; { Separate index needed to skip dummy atoms }
      for i := 0 to nNuclei - 1 do begin
        zLeft.NextRealAtom(ia);
        for k := 0 to 2 do
          sum := sum + (zRight.Atoms[ia].pxyz[k+1] - zLeft.Atoms[ia].pxyz[k+1])
                       * Left.lMatrix[3*i+k+1,m+1] * sqrt(zLeft.Atoms[ia].Mass); // Was: Left.Madm[3*i+k+1,m+1];
      end;
      d[m+1] := sum;
      if SetModeInfo then
        Mode(m).Displacement := d[m+1];
    end;
    { Chen eq 22 }
    C := TMatrix.Create(3,3);
    for i := 0 to 2 do { x, y, z }
      for j := 0 to 2 do begin { x, y, z }
        Sum := 0;
        AllZero := True;
        ia := -1; { Separate index needed to skip dummy atoms }
        for m := 0 to nNuclei - 1 do begin
          { TODO: check i j order }
          zLeft.NextRealAtom(ia);
          v := zLeft.Atoms[ia].Mass * zLeft.Atoms[ia].pxyz[j+1] * zRight.Atoms[ia].pxyz[i+1];
          Sum := Sum + v;
          AllZero := AllZero and (Abs(v) < LinearTest)
        end;
        if AllZero and (i=j) then
          Sum := 1; { Avoid issues with planar molecules }
        C[i+1,j+1] := Sum;
      end;
    if Print then
      C.Write(Log,'C matrix');
    Rc := C.Transpose;
    Rc.PostMultiply(C);
    if Print then
      Rc.Write(Log,'CtC matrix');
    slam := TColumnVector.Create(3);
    Eigen(slam,Rc,True); { This generates Rt in Chen's notation }
    if Print then begin
      slam.Write(Log,'lambda');
      Rc.Write(Log,'Rt matrix');
    end;
    { Display 4 possible S matrices }
    BestS := nil;
    nSMatrix := 0;
    TryS([1,1,1]);
    TryS([-1,-1,1]);
    TryS([-1,1,-1]);
    TryS([1,-1,-1]);
    TryS([-1,1,1]);
    TryS([1,-1,1]);
    TryS([1,1,-1]);
    TryS([-1,-1,-1]);
    slam.Free;
    Rc.Free;
    C.Free;
    if Print then
      WriteLn(Log, 'Using matrix ' + IntToStr(BestSMatrix));
    if Molecule.PointGroup in [C1, Ci, Vib.Cs, C2h] then begin
      Z := TMatrix.Create(3*nNuclei,3*nNuclei);
      Z.Zero;
      for m := 0 to nNuclei - 1 do
        for i := 1 to 3 do
          for j := 1 to 3 do
            Z[3*m + i, 3*m+j] := BestS[i,j];
      if Print then begin
        Z.Write(Log, 'Z');
        { Check Chen eq 21 }
        SumV := TColumnVector.Create(3);
        SumV.Zero;
        Req := TColumnVector.Create(3);
        Rpeq := TColumnVector.Create(3);
        ia := -1; { Separate index needed to skip dummy atoms }
        for m := 0 to nNuclei - 1 do begin
          zLeft.NextRealAtom(ia);
          for i := 0 to 2 do begin
            Req[i+1] := zRight.Atoms[ia].pxyz[i+1];
            Rpeq[i+1] := zLeft.Atoms[ia].pxyz[i+1];
          end;
          Req.PreMultiply(BestS);
          Rpeq.CrossProduct(Req);
          for i := 1 to 3 do
            SumV[i] := SumV[i] + Rpeq[i] * zLeft.Atoms[ia].Mass;
        end;
        Sumv.Write(Log, 'Rotation Check');
        Req.Free;
        Rpeq.Free;
        SumV.Free;
      end;
    end else
      Z := nil;
    { S ( = J Chen ) = rotation between co-ordinate systems }
    S.Free;
    S := MakeDus(Z);
    Z.Free;
    BestS.Free;
  end else begin
    S.Zero;
    S.AddDiagonal(1);
    T := TMatrix.Create(nModes,nModes);
    for m := 1 to nModes do begin
      d[m] := Mode(m-1).Displacement;
      for i := 1 to m-1 do begin
        with Mode(m-1) do
          v := GetParameter(FirstDynamicParameter+i-1);
        if v <> 0 then begin
          SinCos(v*Pi/180,sn,cs);
          T.Zero;
          T.AddDiagonal(1);
          T[m,m] := cs;
          T[i,i] := cs;
          T[m,i] := sn;
          T[i,m] := -sn;
          S.PostMultiply(T);
        end;
      end;
    end;
    T.Free;
  end;
  if (Print and (nModes < 20)) or SetModeInfo then begin
    v := Determinant(S);
    if v < 0 then
      WriteLn(Log,'Determinant < 0 (', FloatToStr(v),') so skipping decomposition')
    else begin
      WriteLn(Log,'Decomposition into equivalent rotations');
      try
        De := FindRotations(S);
        if SetModeInfo then
          for m := 1 to nModes do
            for i := 1 to m-1 do
              with Mode(m-1) do
                SetParameter(FirstDynamicParameter+i-1,De[m,i]);
      except
        on E: EOutOfMemory do
          if SetModeInfo then
            raise
          else
            WriteLn(Log,' **** Skipped: too many modes');
      end;
      De.Free;
    end;
  end;
end;

procedure TMultidimensionalFCF.PrepareMatrices(Print: Boolean; SetModeInfo, AllModes: Boolean);
var
  Left, Right: TElectronicState;
  PreExp: Double;
  m, k, nModes: Integer;
  J: TMatrix;
begin
  Left := Bra as TElectronicState;
  Right := Ket as TElectronicState;
  Left.PrepareMatrices(Print, AllModes);
  Right.PrepareMatrices(Print, AllModes);
  nModes := Left.Molecule.nModes;

  PrepareDushinsky(nModes, Print, SetModeInfo);
  if Print then
    d.Write(Log,'Mode Displacements');
  if Print then
    S.Write(Log, 'Dushinsky Matrix');
  if Print then { Is the matrix necessarily Orthogonal? }
    S.CheckOrthogonal(Log, 'Dushinsky Orthogonality Check', True, Left.Molecule.SmallCoefficient);
  { J = lambda S lambda }
  J := TMatrix.Create(S);
  J.Resize(nModes, nModes, True); { Only want vibrations }
  for m := 0 to nModes - 1 do
    for k := 0 to nModes - 1 do
      if Right.Mode(k).Omega = 0 then
        J[m+1,k+1] := 0 { Avoid divide by sero }
      else
        J[m+1,k+1] := Sqrt(Abs(Left.Mode(m).Omega / Right.Mode(k).Omega))*J[m+1,k+1];
  { Q = Inverse(1-J Jt) }
  Q1 := J.Transpose;
  Q1.PostMultiply(J);
  Q1.AddDiagonal(1);
  if Print then
    Q1.Write(Log,'1+J Jt');
  Invert(Q1);
  if Print then
    Q1.Write(Log,'Q');
  { R = Q Jt}
  R := TMatrix.Create(Q1);
  P1 := J.Transpose;
  R.PostMultiply(P1);
  if Print then
    R.Write(Log,'R');
  { P = J Q Jt }
  P1.Assign(J);
  P1.PostMultiply(R);
  if Print then
    P1.Write(Log,'P');
  { delta = lambda' d }
  Rdelta := TColumnVector.Create(d);
  for m := 1 to nModes do
    { Abs is not strictly necessary in next line, but avoid unhelpful crash in later working
      if Omega values are accidentally set negative }
    Rdelta[m] := Rdelta[m]*Sqrt(Abs(Left.Mode(m-1).Omega*ReduceFrequency));
  if Print then
    Rdelta.Write(Log,'delta');
  { P1Delta = (1-P) delta }
  P1Delta := TColumnVector.Create(Rdelta);
  P1Delta.PreMultiply(P1);
  P1Delta.Scale(-1);
  P1Delta.Add(Rdelta);
  if Print then
    P1Delta.Write(Log,'(1-P)delta');
  { I00 = -1/2 delta^T (1-P) delta}
  I00 := -Rdelta.DotProduct(P1Delta)/2;
  PreExp := 1;
  for m := 0 to nModes - 1 do
    PreExp := PreExp * Left.Mode(m).Omega/Right.Mode(m).Omega;
  PreExp := Sqrt(IntPower(2,nModes)*Sqrt(PreExp)*Determinant(Q1));
  if Print then
    WriteLn(Log,'PreExp=', PreExp, ', ExpArg = ', I00, ', I00=', PreExp*Exp(I00));
  I00 := PreExp*Exp(I00);
  { TODO: If this comes out as zero due to underflow, take special action }
  RDelta.PreMultiply(R);
  if Print then
    RDelta.Write(Log,'R delta');
  Q1.Scale(2);
  Q1.AddDiagonal(-1);
  if Print then
    Q1.Write(Log,'2Q-1');
  P1.Scale(2);
  P1.AddDiagonal(-1);
  if Print then
    P1.Write(Log,'2P-1');
  J.Free;
end;

type
  TVibrationalBasisArray = array of TVibrationalBasisState;

{ Profiling Variables }
var
 FCFMatrixTime: string = 'Franck Condon Factors';

function TMultidimensionalFCF.TransitionMatrix(BraSample,KetSample: TVibrationalBasisState; Cache: Boolean):TMatrix;
{ This method relies on the vibrational quantum numbers begin sorted as follows:
    000
    001
    002
    010
    011
    012
    020
    ...
    For non-totally symmetric modes, there will be members in the sequence missing from a particular basis;
    "OriginalPosition" in the basis state indicates where in the complete sequence the basis lies, and thus some values will
     typically be missing
    The "Offset" value in the basis state indicates the row or column of TransitionMatrix result that
    the overlap integrals for that state lie; it runs continuously from 1, but is not necessarily in order
    because of the sorting of the basis state. This means that this routine must be called each time the basis
    state is changed.
}
{ TODO: Check cache cleared as required }
var
  BraOffsets, KetOffsets: TIntegerArray;
  OrderedBra, OrderedKet: TVibrationalBasisArray;
  EvLeft, EvRight: TEigenvalues;
  Ovr: TMatrix;
  CurrentCacheIndex: Integer;
  BraCopy, KetCopy: TIntegerArray;
  function PrepareOffsets(var Offsets: TIntegerArray; var OrderedStates: TVibrationalBasisArray; Ev:TEigenvalues; AState: TElectronicState):Integer;
  var
    i, o: Integer;
    State, oState: TVibrationalBasisState;
  begin
    SetLength(Offsets, Molecule.nModes+1);
    o := 1;
    for i := Molecule.nModes downto 1 do begin
      Offsets[i] := o;
      o := o*(AState.Mode(i-1).vmax+1);
    end;
    Offsets[0] := o;
    SetLength(OrderedStates, o+1);
    Result := 1;
    for i := 1 to High(Ev.Basis) do begin
      State := Ev.Basis[i] as TVibrationalBasisState;
      if (State.State = AState) and (CacheIndex(State.v)=CurrentCacheIndex) then begin
        oState := OrderedStates[State.OriginalPosition];
        if not Assigned(oState) then begin
          OrderedStates[State.OriginalPosition] := State;
          State.FOffset := Result;
          Inc(Result);
        end else begin
          { Basis state with the same set of vibrational quantum numbers,
            but different electronic ones }
          State.FOffset := oState.FOffset;
        end;
      end;
    end;
    Dec(Result);
  end;
  procedure SetOffsets(const Offsets: TIntegerArray; State: TState; Dest: TEigenValues);
  var
    i: Integer;
    Basis: TVibrationalBasisState;
  begin
    for i := 1 to High(Dest.Basis) do begin
      Basis := Dest.Basis[i] as TVibrationalBasisState;
      if (Basis.State = State) and (CacheIndex(Basis.v)=CurrentCacheIndex) then
        Basis.FOffset := Offsets[Basis.OriginalPosition];
    end;
  end;
//  procedure CopyOffsets(SourceSample: TVibrationalBasisState; Dest: TEigenvalues);
//  var
//    i, o: Integer;
//    OrderedStates: TVibrationalBasisArray;
//    SourceState: TElectronicState;
//    Basis: TVibrationalBasisState;
//    Ev: TEigenValues;
//  begin
//   { Copy offsets from one vibrational basis to another;
//      number of vibrational states should be the same, but ordering may differ }
//    SourceState := SourceSample.State as TElectronicState;
//    o := 1;
//    for i := Molecule.nModes downto 1 do begin
//      o := o*(SourceState.Mode(i-1).vmax+1);
//    end;
//    SetLength(OrderedStates, o+1);
//    Ev := SourceSample.Owner as TEigenvalues;
//    for i := 1 to High(Ev.Basis) do begin
//      Basis := Ev.Basis[i] as TVibrationalBasisState;
//      if (Basis.State = SourceState) and (CacheIndex(Basis.v)=CurrentCacheIndex) then
//        OrderedStates[Basis.OriginalPosition] := Basis;
//    end;
//    for i := 1 to High(Dest.Basis) do begin
//      Basis := Dest.Basis[i] as TVibrationalBasisState;
//      if (Basis.State = SourceState) and (CacheIndex(Basis.v)=CurrentCacheIndex) then
//        Basis.FOffset := OrderedStates[Basis.OriginalPosition].Offset;
//    end;
//  end;
  function FillOverlap(BB, KK: Integer): Double;
  var
    Bra, Ket: TVibrationalBasisState;
    Scale, Vm, Omega, OmegaP: Double;
    vn, vn1, Bn, Bnn, Kn, Knn, l, v, vp, v1: Integer;
    Reduced, Prepared: Boolean;
    procedure PrepareCall;
    begin
      if Prepared then
        Exit;
      { Make a copy of quantum number arrays so that values can be changes
        and therefore make routine threadsafe }
      Move(Bra.v[0],BraCopy[0],Length(BraCopy)*SizeOf(Integer));
      Move(Ket.v[0],KetCopy[0],Length(BraCopy)*SizeOf(Integer));
      Prepared := True;
    end;
  begin
    { TODO: Not thread safe }
    Bra := OrderedBra[BB];
    Ket := OrderedKet[KK];
    Result := 0;
    Reduced := False;
    Prepared := False;
    for vn := 0 to Molecule.nModes-1 do
      { Pick a mode to reduce }
      if Ket.v[vn] > 0 then
        if Molecule.IsDegenerate(Molecule.ModeInfo[vn].Symmetry,0) then begin
          { Special handling for reducing degenerate modes }
          { TODO: currently does not take account of mode mixing for degenerate modes }
          v := Ket.v[vn];
          l := Abs(Ket.v[Molecule.ModeInfo[vn].dIndex]);
          if v > l then begin
            vp := Bra.v[vn];
            Omega := KetState.Mode(vn).Omega;
            OmegaP := BraState.Mode(vn).Omega;
            Scale := Omega + OmegaP;
            { Reduce v }
            if v > 3 then begin
              Kn := Ket.OriginalPosition - 4*KetOffsets[vn+1];
              Vm := (Omega-OmegaP)/Scale*Sqrt((v-2-l)*(v-2+l)/(v-l)/(v+l));
              if Vm <> 0 then
                if Assigned(OrderedKet[Kn]) then
                  Result := Vm*Ovr[Bra.Offset,OrderedKet[Kn].Offset]
                else begin
                  PrepareCall;
                  KetCopy[vn] := Ket.v[vn]-4;
                  Result := Vm*Overlap(BraCopy,KetCopy);
                end;
            end;
            Kn := Ket.OriginalPosition - 2*KetOffsets[vn+1];
            if vp > 1 then begin
              Bn := Bra.OriginalPosition - 2*BraOffsets[vn+1];
              Vm := 2*Omega/Scale*Sqrt((vp-l)*(vp+l)/(v-l)/(v+l));
              if Vm <> 0 then
                if Assigned(OrderedBra[Bn]) and Assigned(OrderedKet[Kn]) then
                  Result := Result + Vm*Ovr[OrderedBra[Bn].Offset,OrderedKet[Kn].Offset]
                else begin
                  PrepareCall;
                  KetCopy[vn] := Ket.v[vn]-2;
                  BraCopy[vn] := Bra.v[vn]-2;
                  Result := Result + Vm*Overlap(BraCopy,KetCopy);
                end;
            end;
            Vm := 2*(OmegaP*(v-1)-Omega*(vp+1))/Scale/Sqrt((v-l)*(v+l));
            if Assigned(OrderedKet[Kn]) then
              Result := Result + Vm*Ovr[Bra.Offset,OrderedKet[Kn].Offset]
            else begin
              PrepareCall;
              KetCopy[vn] := Ket.v[vn]-2;
              Result := Result + Vm*Overlap(BraCopy,KetCopy);
            end;
            Reduced := True;
            Break;
          end else
            Continue; { This degnerate mode at minmum value for l }
        end else begin
          Scale := Ket.v[vn];
          Kn := Ket.OriginalPosition - KetOffsets[vn+1];
          for vn1 := 0 to Molecule.nModes-1 do
            if Bra.v[vn1] > 0 then begin
              if (vn=vn1) or (Molecule.ModeInfo[vn1].Symmetry = 0) then begin
                Vm := R[vn+1,vn1+1];
                if Vm <> 0 then begin
                  Bn := Bra.OriginalPosition - BraOffsets[vn1+1];
                  Vm := 2 * Vm * Sqrt(Bra.v[vn1]/Scale);
                  if Assigned(OrderedBra[Bn]) and Assigned(OrderedKet[Kn]) then begin
                    Result := Result + Vm*Ovr[OrderedBra[Bn].Offset,OrderedKet[Kn].Offset];
                  end else begin
                    PrepareCall;
                    KetCopy[vn] := Ket.v[vn]-1;
                    BraCopy[vn1] := Bra.v[vn1]-1;
                    Result := Result + Vm*Overlap(BraCopy,KetCopy);
                  end;
                end;
              end;
            end;
          for vn1 := 0 to Molecule.nModes-1 do begin
            v1 := Ket.v[vn1];
            if vn=vn1 then
              Dec(v1);
            if v1 > 0 then begin
              if (vn=vn1) or (Molecule.ModeInfo[vn1].Symmetry = 0) then begin
                Vm := Q1[vn+1,vn1+1];
                if Vm <> 0 then begin
                  Vm := Vm * Sqrt(v1/Scale);
                  Knn := Kn - KetOffsets[vn1+1];
                  if Assigned(OrderedKet[Knn]) then begin
                    Result := Result + Vm*Ovr[Bra.Offset,OrderedKet[Knn].Offset];
                  end else begin
                    PrepareCall;
                    KetCopy[vn] := Ket.v[vn]-1;
                    KetCopy[vn1] := v1-1; { This is correct for vn=vn1 case }
                    Result := Result + Vm*Overlap(BraCopy,KetCopy);
                  end;
                end;
              end;
            end;
          end;
          if Molecule.ModeInfo[vn].Symmetry = 0 then begin
            Vm := Rdelta[vn+1];
            if Vm <> 0 then begin
              Vm := - Vm * Sqrt(2/Scale);
              if Assigned(OrderedKet[Kn]) then
                Result := Result + Vm*Ovr[Bra.Offset,OrderedKet[Kn].Offset]
              else begin
                PrepareCall;
                KetCopy[vn] := Ket.v[vn]-1;
                Result := Result + Vm*Overlap(BraCopy,KetCopy);
              end;
            end;
          end;
          Reduced := True;
          Break;
        end;
    if not Reduced then
      { Ket all zero; reduce Bra }
      for vn := 0 to Molecule.nModes-1 do
        { Pick a mode to reduce }
        if Bra.v[vn] > 0 then
          if Molecule.IsDegenerate(Molecule.ModeInfo[vn].Symmetry,0) then begin
            { Special handling for reducing degenerate modes }
            vp := Bra.v[vn];
            l := Abs(Bra.v[Molecule.ModeInfo[vn].dIndex]);
            if vp > l then begin
              v := Ket.v[vn];
              Omega := KetState.Mode(vn).Omega;
              OmegaP := BraState.Mode(vn).Omega;
              Scale := OmegaP + Omega;
              { Reduce vp }
              if vp > 3 then begin
                Bn := Bra.OriginalPosition - 4*BraOffsets[vn+1];
                Vm := (OmegaP-Omega)/Scale*Sqrt((vp-2-l)*(vp-2+l)/(vp-l)/(vp+l));
                if Vm <> 0 then
                  if Assigned(OrderedBra[Bn]) then
                    Result := Vm*Ovr[OrderedBra[Bn].Offset,Ket.Offset]
                  else begin
                    PrepareCall;
                    BraCopy[vn] := Bra.v[vn]-4;
                    Result := Vm*Overlap(BraCopy,KetCopy);
                  end;
              end;
              Bn := Bra.OriginalPosition - 2*BraOffsets[vn+1];
              if v > 1 then begin
                Kn := Ket.OriginalPosition - 2*KetOffsets[vn+1];
                Vm := 2*OmegaP/Scale*Sqrt((v-l)*(v+l)/(vp-l)/(vp+l));
                if Vm <> 0 then
                  if Assigned(OrderedBra[Bn]) and Assigned(OrderedKet[Kn]) then
                    Result := Result + Vm*Ovr[OrderedBra[Bn].Offset,OrderedKet[Kn].Offset]
                  else begin
                    PrepareCall;
                    BraCopy[vn] := Bra.v[vn]-2;
                    KetCopy[vn] := Ket.v[vn]-2;
                    Result := Result + Vm*Overlap(BraCopy,KetCopy);
                  end;
              end;
              Vm := 2*(Omega*(vp-1)-OmegaP*(v+1))/Scale/Sqrt((vp-l)*(vp+l));
              if Assigned(OrderedBra[Bn]) then
                Result := Result + Vm*Ovr[OrderedBra[Bn].Offset,Ket.Offset]
              else begin
                PrepareCall;
                BraCopy[vn] := Bra.v[vn]-2;
                Result := Result + Vm*Overlap(BraCopy,KetCopy);
              end;
              Reduced := True;
              Break;
            end else
              Continue; { This degnerate mode at minmum value for l }
          end else begin
            Scale := Bra.v[vn];
            Bn := Bra.OriginalPosition - BraOffsets[vn+1];
            for vn1 := 0 to Molecule.nModes-1 do { Note that this is redundant code }
              if Ket.v[vn1] > 0 then begin
                if (vn=vn1) or (Molecule.ModeInfo[vn1].Symmetry = 0) then begin
                  Vm := R[vn+1,vn1+1];
                  if Vm <> 0 then begin
                    Vm := 2*Vm*Sqrt((Ket.v[vn1]+1)/Scale);
                    Kn := Ket.OriginalPosition - KetOffsets[vn1+1];
                    if Assigned(OrderedBra[Bn]) and Assigned(OrderedKet[Kn]) then
                      Result := Result + Vm*Ovr[OrderedBra[Bn].Offset,OrderedKet[Kn].Offset]
                    else begin
                      PrepareCall;
                      BraCopy[vn] := Bra.v[vn]-1;
                      KetCopy[vn1] := Ket.v[vn1]-1;
                      Result := Result + Vm*Overlap(BraCopy,KetCopy);
                    end;
                  end;
                end;
              end;
            for vn1 := 0 to Molecule.nModes-1 do begin
              v1 := Bra.v[vn1];
              if vn = vn1 then
                Dec(v1);
              if v1 > 0 then begin
                if (vn=vn1) or (Molecule.ModeInfo[vn1].Symmetry = 0) then begin
                  Vm := P1[vn+1,vn1+1];
                  if Vm <> 0 then begin
                    Vm := Vm * Sqrt(v1/Scale);
                    Bnn := Bn - BraOffsets[vn1+1];
                    if Assigned(OrderedBra[Bnn]) then
                      Result := Result + Vm*Ovr[OrderedBra[Bnn].Offset,Ket.Offset]
                    else begin
                      PrepareCall;
                      BraCopy[vn] := Bra.v[vn]-1;
                      BraCopy[vn1] := v1-1;
                      Result := Result + Vm*Overlap(BraCopy,KetCopy);
                    end;
                  end;
                end;
              end;
            end;
            if Molecule.ModeInfo[vn].Symmetry = 0 then begin
              Vm := P1Delta[vn+1];
              if Vm <> 0 then begin
                Vm := Vm*Sqrt(2/Scale);
                if Assigned(OrderedBra[Bn]) then
                  Result := Result + Vm*Ovr[OrderedBra[Bn].Offset,Ket.Offset]
                else begin
                  PrepareCall;
                  BraCopy[vn] := Bra.v[vn]-1;
                  Result := Result + Vm*Overlap(BraCopy,KetCopy);
                end;
              end;
            end;
            Reduced := True;
            Break;
          end;
    if not Reduced then begin
      Result := I00;
      if I00 = 0 then
        Result := 1e-99; 
      for vn := 0 to Molecule.nModes-1 do
        { Special action needed for degenerate modes - these will have v' = v = l, rather than 0;
          in addition I00 is wrong, as mode should have been counted twice }
        if Molecule.IsDegenerate(Molecule.ModeInfo[vn].Symmetry,0) then begin
          l := Abs(Bra.v[Molecule.ModeInfo[vn].dIndex]);
          Omega := (Ket.State as TElectronicState).Mode(vn).Omega;
          OmegaP := (Bra.State as TElectronicState).Mode(vn).Omega;
          Result := Result * Sqrt(Math.IntPower(2*Sqrt(Omega*Omegap)/(Omega+Omegap),2*l+1))
        end;
      end;
    Ovr[Bra.Offset,Ket.Offset] := Result;
//    if not SameValue(Result, Overlap(Bra.v,Ket.v)) then
    if IsNan(Result) then
      Raise Exception.Create('FCFOverlap');
  end;
var
  i, j, BraMax, Ketmax: Integer;
  CacheEntry: TCachedTransitionMatrix;
begin
  Result := nil;
  if Molecule.nModes = 0 then
    Exit;
  { Compute the vibrational overlap matrix; this depends on the vibrational quantum numbers only;
    it will have the form <v1',v2'...,l1,l2...|v1,v2,...,l1,l2>, i.e. diagonal in the angular momentum
    quantum numbers. The angular momentum quantum numbers are therfore used as the key to a cache.
    In the absence of mode mixing between the degenerate modes the sign of l should be unimportant,
    so each l can have the range 0..vmax. For the other modes the overall vibrational symmetry is
    used to determine the appropriate index }
  CurrentCacheIndex := CacheIndex(BraSample.v);
  if Cache then begin
    { The cache is implicitly a multidimensional array indexed by Abs(l1), Abs(l2), ... but
      collapse to a one dimensional array using fortran style indexing }
    if Assigned(CacheLock) then
      CacheLock.BeginRead;
    if (CurrentCacheIndex < Length(FMatrixCache)) and Assigned(FMatrixCache[CurrentCacheIndex]) then
      CacheEntry := FMatrixCache[CurrentCacheIndex]
    else
      CacheEntry := nil;
    if Assigned(CacheLock) then
      CacheLock.EndRead;
    if not Assigned(CacheEntry) then
      try
        if Assigned(CacheLock) then
          CacheLock.BeginWrite;
        if not((CurrentCacheIndex < Length(FMatrixCache)) and Assigned(FMatrixCache[CurrentCacheIndex])) then begin
          if Length(FMatrixCache) = 0 then
            SetCachedInformationPresent;
          CacheEntry := TCachedTransitionMatrix.Create(nil);
          SetLength(FMatrixCache,Max(Length(FMatrixCache),CurrentCacheIndex+1));
          FMatrixCache[CurrentCacheIndex] := CacheEntry;
          CacheEntry.BraSample := BraSample;
          CacheEntry.KetSample := KetSample;
          CacheEntry.TransitionMoment := Self;
          CacheEntry.TheCacheIndex := CurrentCacheIndex;
        end else
          CacheEntry := FMatrixCache[CurrentCacheIndex];
      finally
        if Assigned(CacheLock) then
          CacheLock.EndWrite;
      end;
    CacheEntry.Make(True);
    Result := CacheEntry.TransitionMatrix;
    { If offsets not set, then same basis used for more than one symmetry }
    { TODO: Can fail if BraSample/KetSample are deallocated, as by TranInfo }
    if BraSample.Offset = 0 then
      SetOffsets(CacheEntry.BraOffsets, BraSample.State, BraSample.Owner as TEigenvalues);
    if KetSample.Offset = 0 then
      SetOffsets(CacheEntry.KetOffsets, KetSample.State, KetSample.Owner as TEigenvalues);
    Exit;
  end;
  TStart(FCFMatrixTime);
  EvLeft := BraSample.Owner as TEigenvalues;
  EvRight := KetSample.Owner as TEigenvalues;
  BraMax := PrepareOffsets(BraOffsets, OrderedBra, EvLeft, BraState);
  KetMax := PrepareOffsets(KetOffsets, OrderedKet, EvRight, KetState);
  SetLength(BraCopy,Length(BraSample.v));
  SetLength(KetCopy,Length(KetSample.v));
  Ovr := TMatrix.Create(BraMax, KetMax);
  { Now fill matrix, starting from smallest quantum numbers }
  PushProgress('Franck Condon factors', High(OrderedBra));
  try
    for i := 1 to High(OrderedBra) do
      { Is this basis state present? }
      if Assigned(OrderedBra[i]) then begin
        SetProgress(i);
        for j := 1 to High(OrderedKet) do
          if Assigned(OrderedKet[j]) then
            FillOverlap(i,j);
      end;
  finally
    PopProgress;
  end;
  Result := Ovr;
  TEnd(FCFMatrixTime);
//  Result.Write(Log,'CacheIndex='+IntToStr(CurrentCacheIndex),6);
//  for i := 0 to High(BraOffsets) do
//    WriteLn(Log,i:4, BraOffsets[i]:8);
//  for i := 0 to High(OrderedBra) do
//    if Assigned(OrderedBra[i]) then
//      WriteLn(Log, i:4, OrderedBra[i].Offset:4, ' ', OrderedBra[i].BasisLabel, ' ', Result[OrderedBra[i].Offset,1]);
//  for i := 0 to High(KetOffsets) do
//    WriteLn(Log,i:4, KetOffsets[i]:8);
//  for i := 0 to High(OrderedKet) do
//    if Assigned(OrderedKet[i]) then
//      WriteLn(Log, i:4, OrderedKet[i].Offset:4, ' ', OrderedKet[i].BasisLabel, ' ', Result[1,OrderedKet[i].Offset]);
end;

procedure TMultidimensionalFCF.PrintInfo(Sender: TObject);
var
  nModes: Integer;
  DD: TColumnVector;
  C: TMatrix;
  procedure MakeGamma(var G, Gs: TMatrix; State: TElectronicState);
  var
    i: Integer;
  begin
    G := TMatrix.Create(nModes,nModes);
    G.Zero;
    Gs := TMatrix.Create(G);
    for i := 1 to nModes do begin
      G[i,i] := State.Mode(i-1).Omega*ReduceFrequency;
      Gs[i,i] := Sqrt(G[i,i]);
    end;
  end;
{$IFDEF TEST}
  function FCF(m, n: Integer):Double; { FCF from Chen paper }
  var
    p, q, r: Integer;
    S: Double;
    fac: array of Double;
  begin  { Chen Eq 19 for testing only }
    SetLength(fac, Max(m,n)+1);
    fac[0] := 1;
    for p := 1 to High(fac) do
      fac[p] := fac[p-1] * p;
    S := 0;
    for p := 0 to Min(m,n) do
      for q := 0 to (m-p) div 2 do
        for r := 0 to (n-p) div 2 do
          S := S + IntPower(2,p)*IntPower(DD[1],m-2*q-p)*IntPower(DD[2],n-2*r-p)
                   * IntPower(C[1,1],q) * IntPower(C[2,2],r) * IntPower(C[1,2],p)
                   /(Fac[p]*Fac[q]*Fac[r]*Fac[m-2*q-p]*Fac[n-2*r-p]);
    Result := S * Sqrt(Fac[m]*Fac[n]/IntPower(2,m+n));
  end;
{$ENDIF}
var
  Left, Right: TElectronicState;
  Strim, SS, GammaP, Gamma, sGammaP, sGamma, JSJ, T, T1: TMatrix;
  k: TColumnVector;
  m, i: Integer;
  PreExp, ExpArg: Double;
  {$IFDEF TEST}
  bb, kk: array of Integer;
  {$ENDIF}
begin
  ReShowLogForm;
  PrepareMatrices(True,False,True);
  Left := Bra as TElectronicState;
  Right := Ket as TElectronicState;
  nModes := Left.Molecule.nModes;

  { Show Ruhoff style info - CP 186 355 (1994) }
  MakeGamma(GammaP, sGammaP, Right);
  GammaP.Write(Log,'GammaP');
  MakeGamma(Gamma, sGamma, Left);
  Gamma.Write(Log,'Gamma');
  Strim := TMatrix.Create(S);
  Strim.Resize(nModes, nModes, True); { Discard rotations and vibrations }
  SS := Strim.Transpose; { J in Ruhoff notation }
  SS.Write(Log, 'Jt');
  SS.PostMultiply(GammaP);
  SS.Write(Log, 'S = Jt GammaP');
  SS.PostMultiply(Strim);
  for m := 1 to nModes do
    SS[m,m] := SS[m,m] + Gamma[m,m];
  SS.Write(Log, 'S = Jt GammaP J + Gamma');

  PreExp := 1;
  for m := 1 to nModes do
    PreExp := PreExp * Gamma[m,m] * GammaP[m,m];
  T := TMatrix.Create(Strim);
  T.PostMultiply(SS);
  PreExp := Sqrt(PreExp)/Determinant(T);
  PreExp := Sqrt(Abs(PreExp) * IntPower(2,nModes));
  T.Free;

  Invert(SS);
  SS.Write(Log, 'S^-1 = (Jt GammaP J + Gamma)^-1');
  JSJ := TMatrix.Create(Strim);
  JSJ.PostMultiply(SS);
  T := Strim.Transpose;
  JSJ.PostMultiply(T);
  JSJ.Write(Log, 'J S^-1 Jt');

  T.Assign(GammaP);
  T.PostMultiply(JSJ);
  for m := 1 to nModes do
    T[m,m] := T[m,m] - 1;
  T.Write(Log, 'GammaP J S^-1 Jt - 1');
  T.PostMultiply(GammaP);
  T.Write(Log, '(GammaP J S^-1 Jt - 1)GammaP');
  k := TColumnVector.Create(d);
  k.PreMultiply(T);
  ExpArg := k.DotProduct(d)/2;
  WriteLn(Log,'PreExp=', PreExp, ', ExpArg=', ExpArg, ', I00=', PreExp*Exp(ExpArg));

  T.Assign(sGammaP);
  T.PostMultiply(JSJ);
  T.PostMultiply(sGammaP);
  for m := 1 to nModes do begin
    for i := 1 to nModes do
      T[m,i] := 2*T[m,i];
    T[m,m] := T[m,m] - 1;
  end;
  T.Write(Log, 'A');

  T.Assign(JSJ);
  T.PostMultiply(GammaP);
  for m := 1 to nModes do
    T[m,m] := 1 - T[m,m];
  k.Assign(d);
  k.PreMultiply(T);
  k.PreMultiply(sGamma);
  k.Scale(2);
  k.Write(Log, 'b=');

  C := TMatrix.Create(sGamma);
  C.PostMultiply(SS);
  C.PostMultiply(sGamma);
  for m := 1 to nModes do begin
    for i := 1 to nModes do
      C[m,i] := 2*C[m,i];
    C[m,m] := C[m,m] - 1;
  end;
  C.Write(Log, 'C');

  DD := TColumnVector.Create(d);
  DD.PreMultiply(GammaP);
  T.Free;
  T := Strim.Transpose;
  DD.PreMultiply(T);
  DD.PreMultiply(SS);
  DD.PreMultiply(sGamma);
  DD.Scale(-2);
  DD.Write(Log, 'd=');

  T.Assign(sGamma);
  T.PostMultiply(SS);
  T1 := Strim.Transpose;
  T.PostMultiply(T1);
  T1.Free;
  T.PostMultiply(sGammaP);
  T.Write(Log, 'E');

{$IFDEF TEST}
  PreExp := PreExp*Exp(ExpArg);
  T.RowCount := 5;
  T.ColCount := 5;
  if nModes > 1 then begin
    for m := 1 to T.RowCount do
      for i := 1 to T.RowCount do
        T[m,i] := FCF(m-1,i-1)*PreExp;
    T.Write(Log, 'FCF');
    SetLength(bb,2);
    SetLength(kk,2);
    kk[0] := 0; kk[1] := 0;
    for m := 1 to T.RowCount do begin
      bb[0] := m-1;
      for i := 1 to T.RowCount do begin
        bb[1] := i-1;
        T[m,i] := Overlap(bb,kk);
      end;
    end;
    T.Write(Log, '<n''n|00>');
    for m := 1 to T.RowCount do begin
      bb[0] := m-1;
      for i := 1 to T.RowCount do begin
        bb[1] := i-1;
        T[m,i] := Overlap(kk,bb);
      end;
    end;
    T.Write(Log, '<00|n''n>');
    bb[1] := 0;
  end else begin
    SetLength(bb,1);
    SetLength(kk,1);
  end;

  for m := 1 to T.RowCount do begin
    bb[0] := m-1;
    for i := 1 to T.RowCount do begin
      kk[0] := i-1;
      T[i,m] := Overlap(kk,bb);
    end;
  end;
  T.Write(Log, '<v''|v>');
{$ENDIF}

  DD.Free; C.Free; Strim.Free;
  SS.Free; GammaP.Free; Gamma.Free; sGammaP.Free; sGamma.Free; JSJ.Free;

  k.Free;
  T.Free
end;

function TMultidimensionalFCF.ReadProperties(Node: TDOMNode; CheckUnread: Boolean): Integer;
begin
  Result := inherited ReadProperties(Node, CheckUnread);
  if ReadVersion < 7000106 then
    Strength := Strength * Sqrt(3);
end;

procedure TMultidimensionalFCF.SetSettingFromStr(AIndex: Integer; const Str: string);
begin
  if AIndex = 1 then
    if CompareText(Str, 'Auto') = 0 then
      Symmetry := -1
    else
      Symmetry := PointGroupInfo[Molecule.PointGroup].StrToSymmetry(Str,0)
  else
    inherited;
end;

function TMultidimensionalFCF.SettingToStr(AIndex: Integer): string;
begin
  if AIndex = 1 then
    if Symmetry < 0 then
      Result := 'Auto'
    else
      Result := PointGroupInfo[Molecule.PointGroup].SymmetryToStr(Symmetry,0)
  else
    Result := inherited SettingToStr(AIndex);
end;

function TMultidimensionalFCF.SettingToPrettyStr(AIndex: Integer): string;
begin
  Result := inherited SettingToPrettyStr(AIndex);
  if (AIndex = 1) and (Symmetry < 0) and (ActualSymmetry >= 0) and (ActualSymmetry <= PointGroupInfo[Molecule.PointGroup].LastSymmetry) then
    Result := Result + ' => ' + PointGroupInfo[Molecule.PointGroup].SymmetryToStr(ActualSymmetry,0);
end;

procedure TMultidimensionalFCF.ShowDushinsky(Sender: TObject);
begin
  {$IFNDEF CONSOLE}
  ConnectDushinskyForm(Self);
  {$ENDIF}
end;

procedure TMultidimensionalFCF.SwapModes(nMode1, nMode2: Integer);
var
  T: Double;
  i, f: Integer;
  Mode1, Mode2: TVibrationalModeChange;
begin
  if Abs(nMode1-nMode2) <> 1 then
    raise NotYet; { This routine will only exchange adjacent modes }
  Mode1 := Mode(nMode1);
  Mode2 := Mode(nMode2);
  { Given modes have been exchanged, so move correspondign values also }
  T := Mode1.Displacement;
  Mode1.Displacement := Mode2.Displacement;
  Mode2.Displacement := T;
  { Permute rotations }
  { TODO: may require adjustment to give mathematically equivalent results }
  f := Mode1.FirstDynamicParameter;
  for i := 0 to Min(nMode1,nMode2) - 1 do begin
    T := Mode1.Parameters[f+i];
    Mode1.Parameters[f+i] := Mode2.Parameters[f+i];
    Mode2.Parameters[f+i] := T;
  end;
  for i := Max(nMode1,nMode2)+1 to Molecule.nModes-1 do begin
    T := Mode(i).Parameters[f+nMode1];
    Mode(i).Parameters[f+nMode1] := Mode(i).Parameters[f+nMode2];
    Mode(i).Parameters[f+nMode2] := T;
  end;
end;

{ TBaseVibrationalMode }

constructor TBaseVibrationalMode.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  if AOwner is TElectronicState then
    TElectronicState(AOwner).AdjustModes(False)
  else
    (AOwner as TMultidimensionalFCF).AdjustModes(False);
  { Symmetry should be a per molecule setting, but is actually
    in the per state object. Copy from the per molecule setting on
    initialisation }
  ModifySetting(1, (AOwner.Owner.Owner as TVibratingMolecule).ModeInfo[NodeIndex].Symmetry, False);
end;

function TBaseVibrationalMode.FixupProperties: Exception;
begin
  if Symmetry > Molecule.LastSym then
    Symmetry := Symmetry mod (Molecule.LastSym+1);
  { Copy back symmetry from global variable }
  ModifySetting(1, Molecule.ModeInfo[NodeIndex].Symmetry, False);
  Result := inherited FixUpProperties;
end;

function TBaseVibrationalMode.GetName: TComponentName;
begin
  Result := Molecule.ModeInfo[NodeIndex].Name;
end;

function TBaseVibrationalMode.Molecule: TVibratingMolecule;
begin
  Result := Owner.Owner.Owner as TVibratingMolecule;
end;

function TBaseVibrationalMode.ReadProperties(Node: TDOMNode; CheckUnread: Boolean): Integer;
var
  i: Integer;
begin
  Result := inherited ReadProperties(Node, False);
  for i := FirstDynamicParameter to nParameters do
    ReadFloatableValue(Node,i,ParameterNames[i],'0');
  if CheckUnread then
    Result := Result + nUnreadProperties(Node)
end;

procedure TBaseVibrationalMode.SetName(const Value: TComponentName);
var
  i: Integer;
begin
  for i := 1 to Length(Value) do
    if not (Value[i] in ['0'..'9','a'..'z','A'..'Z','_','(',')',' ','''','"']) then
      raise EReportException.Create('Invalid character in vibrational mode name:'+Value[i]);
  inherited;
  if NodeIndex >= 0 then
    Molecule.ModeInfo[NodeIndex].Name := Value;
end;

procedure TBaseVibrationalMode.SetSetting(AIndex: Integer; const AValue: Integer);
var
  Changed: Boolean;
  procedure CopySetting(Node: TMNode; MyIndex: Integer);
  var
    i: Integer;
  begin
    if Node is TBaseVibrationalMode then
      with TBaseVibrationalMode(Node) do
        if NodeIndex = MyIndex then
          ModifySetting(AIndex, AValue, True);
    for i := 0 to Node.ComponentCount - 1 do
      CopySetting(Node.Components[i] as TMNode, MyIndex);
  end;
begin
  inherited SetSetting(AIndex, AValue);
  { See notes in Create method; set per molecule settting also }
  if NodeIndex < 0 then
    Exit;
  Changed := False;
  case AIndex of
    1: begin
         Changed := Molecule.ModeInfo[NodeIndex].Symmetry <> AValue;
         Molecule.ModeInfo[NodeIndex].Symmetry := AValue;
       end;
  end;
  if not Changed then
    Exit;
  { As GetSetting is not virtual, changes must be copied to all other copies of this mode }
  CopySetting(Molecule, NodeIndex);
end;

function TBaseVibrationalMode.GetSettingsNames(AIndex: Integer; out LowestValue: Integer): PPChar;
begin
  if Aindex = 1 then begin
    LowestValue := 0;
    Result := PointGroupInfo[Molecule.PointGroup].SettingsNames;
  end else
    Result := inherited GetSettingsNames(AIndex, LowestValue);
end;

function TBaseVibrationalMode.HelpOnSetting(AIndex: Integer): string;
begin
  if AIndex = 1 then
    Result := 'Symmetry of this vibrational mode. Note that this setting is shared with every state in the molecule.'
  else
    Result := inherited HelpOnSetting(AIndex);
end;

procedure TBaseVibrationalMode.SetSettingFromStr(AIndex: Integer; const Str: string);
begin
  if AIndex = 1 then
    Symmetry := PointGroupInfo[Molecule.PointGroup].StrToSymmetry(Str,0)
  else
    inherited;
end;

function TBaseVibrationalMode.SettingToStr(AIndex: Integer): string;
begin
  if AIndex = 1 then
    if Symmetry <= Molecule.LastSym then
      Result := PointGroupInfo[Molecule.PointGroup].SymmetryToStr(Symmetry,0)
    else
      Result := IntToStr(Symmetry)
  else
    Result := inherited SettingToStr(AIndex);
end;

{ TVibrationalModeChange }

procedure TVibrationalModeChange.AdjustParameterNames;
var
  i: Integer;
begin
  { Add/remove rotation matrix parameters as required }
  for i := nParameters+1 to FirstDynamicParameter + NodeIndex - 1 do
    AddParameter(Molecule.ModeInfo[i-FirstDynamicParameter].Name);
  for i := nParameters downto FirstDynamicParameter + NodeIndex do
    DeleteParameter(i);
  for i := FirstDynamicParameter to nParameters do
    SetParameterName(i, Molecule.ModeInfo[i-FirstDynamicSetting].Name);
end;

function TVibrationalModeChange.CanShift(Shift: Integer): Boolean;
begin
  Result := False;
end;

constructor TVibrationalModeChange.Create(AOwner: TComponent);
begin
  DimensionlessParameters := DimensionlessParameters + ',Displacement';
  inherited Create(AOwner);
end;

procedure TVibrationalModeChange.ScaleParameters(UFrom, UTo: TResonanceUnits);
begin
  inherited;
  { Displacement has sqrt(frequency) in units: TODO specify }
  Displacement := Displacement/Sqrt(ConversionFactor(UFrom, UTo));
end;

function TVibrationalModeChange.HelpOnParameter(AIndex: Integer): string;
begin
  if AIndex > 1 then
    Result := 'Mixing of this mode with ' + (Owner as TMultidimensionalFCF).Mode(AIndex-1).Name + ' /degrees'
  else if AIndex = 1 then
    Result := 'Displacement along normal mode between the two electronic states'
  else
    Result := inherited HelpOnParameter(AIndex);
end;

{ TPointGroupInfo }

constructor TPointGroupInfo.Create(APointGroup: TPointGroup);
begin
  FPointGroup := APointGroup;
  Name := PropertyToStr(Self,'PointGroup');
  PointGroupInfo[PointGroup] := Self;
  FHasDegenerateModes := False;
end;

procedure FreePointGroupInfo;
var
  I: TPointGroup;
begin
  for I := Low(TPointGroup) to High(TPointGroup) do
    PointGroupInfo[I].Free;
end;

function TPointGroupInfo.StrToSymmetry(Str: string; HalfSpin: TQuantumNumber): Integer;
begin
  if not TryStrToSymmetry(Str, Result, HalfSpin) then
    raise Exception.Create('Symmetry ' + Str + ' not valid for ' + Name + ' (' + QuantumNumberToStr(HalfSpin,qftJ) + ')');
end;

function TPointGroupInfo.SymmetryToStr(Symmetry: Integer; HalfSpin: TQuantumNumber): string;
var
  P: PPChar;
begin
  if HalfSpin <> 0 then
    raise EReportException.Create('1/2 integral spin not implemented');
  if (Symmetry < 0) or (Symmetry > LastSymmetry) or (HalfSpin <> 0) then
    raise Exception.CreateFmt('Bad Symmetry for %s: %d', [Name,Symmetry]);
  P := SettingsNames;
  Inc(P,Symmetry);
  Result := P^;
end;

function TPointGroupInfo.TryStrToSymmetry(Str: string; out Symmetry: Integer; HalfSpin: TQuantumNumber): Boolean;
var
  P: PPChar;
begin
  if HalfSpin <> 0 then
    raise NotYet; { Probably an error }
  P := SettingsNames;
  Symmetry := 0;
  while P^ <> nil do begin
    if CompareText(P^,Str) = 0 then begin
      Result := True;
      Exit;
    end;
    Inc(P);
    Inc(Symmetry);
  end;
  Result := False;
  Symmetry := -1;
end;

const
  LambdaNames: array [0..7] of PChar = ('Sigma+', 'Sigma-', 'Pi', 'Delta', 'Phi', 'Gamma', 'H', 'I');
  { Note that for half spin, symmetries are 1/2, 3/2, 5/2, ... }

{ TCinfinityv }

type
  TCinfinityv = class(TPointGroupInfo)
  private
    FIrepNames: array of string;
  public
    constructor Create(APointGroup: TPointGroup);
    function TryStrToSymmetry(Str: string; out Symmetry:Integer; HalfSpin: TQuantumNumber):Boolean; override;
    function SymmetryToStr(Symmetry:Integer; HalfSpin: TQuantumNumber):string; override;
    function Allowed(Bra, Dipole, Ket: Integer; HalfSpin: TQuantumNumber):Boolean; override;
  end;

function TCinfinityv.Allowed(Bra, Dipole, Ket: Integer; HalfSpin: TQuantumNumber): Boolean;
begin
  if HalfSpin <> 0 then begin
    { Dipole is not Omega, so convert to lambda }
    Dipole := 2*Max(Dipole-1,0); { Max(...0) maps Sigma+/- to Delta omega = 0 }
    { Map to standard quantum numbers, as need to allow Delta omgea = 1 transitions to connect +1/2 to -1/2,
      which otherwise looks like Delta omega = 0 }
    Bra := 2*Bra + 1;
    Ket := 2*Ket + 1;
    Result := (Bra+Dipole = Ket) or (Bra-Dipole = Ket) or (Bra-Dipole = -Ket);
  end else if Bra <= 1 then
    if Ket <= 1 then { Sigma - Sigma }
      Result := (Dipole <= 1) and (Bra xor Ket xor Dipole = 0)
    else if Dipole <= 1 then { Sigma Sigma not Sigma }
      Result := False
    else
      Result := Ket = Dipole { Sigma not Sigma not Sigma }
  else if Ket <= 1 then
    if Dipole <= 1 then
      Result := False { not Sigma Sigma Sigma }
    else
      { not Sigma not Sigma Sigma }
      Result := Bra = Dipole
  else if Dipole <= 1 then
    Result := Bra = Ket { not Sigma Sigma not Sigma }
  else { not Sigma not Sigma not Sigma }
    Result := (Abs(Bra-Dipole+1) = Ket) or (Abs(Bra+Dipole-1) = Ket);
end;

constructor TCinfinityv.Create(APointGroup: TPointGroup);
var
  i: Integer;
begin
  inherited;
  SetLength(FIrepNames, (Ord(APointGroup)-Ord(CInfinityv)+1)*Length(LambdaNames)+4);
  for i := 1 to High(FIrepNames) - 1 do
    FIRepNames[i] := SymmetryToStr(i-1,0);
  FIrepNames[0] := 'Auto';
  SettingsNames := PPChar(@FIrepNames[1]);
  FLastSymmetry := High(Integer);
  FHasDegenerateModes := True;
end;

function TCinfinityv.SymmetryToStr(Symmetry: Integer; HalfSpin: TQuantumNumber): string;
begin
  if HalfSpin <> 0 then
    Result := QuantumNumberToStr(2*Symmetry+1,qftJ)
  else if Symmetry <= High(LambdaNames) then
    Result := LambdaNames[Symmetry]
  else
    Result := QuantumNumberToStr(2*Symmetry,qftJ);
end;

function TCinfinityv.TryStrToSymmetry(Str: string; out Symmetry: Integer; HalfSpin: TQuantumNumber): Boolean;
var
  i: Integer;
begin
  Result := False;
  if HalfSpin = 0 then
    for i := 0 to High(LambdaNames) do
      if CompareText(Str,LambdaNames[i]) = 0 then begin
        Result := True;
        Symmetry := i;
        Exit;
      end;
  if TryStrToQuantumNumber(Str,Symmetry,qftJ) then begin
    Result := not Odd(Symmetry + HalfSpin);
    Symmetry := Symmetry div 2;
  end;
end;

{ TDInfinityh }

type
  TDInfinityh = class(TCinfinityv)
    function TryStrToSymmetry(Str: string; out Symmetry:Integer; HalfSpin: TQuantumNumber):Boolean; override;
    function SymmetryToStr(Symmetry:Integer; HalfSpin: TQuantumNumber):string; override;
    function Allowed(Bra, Dipole, Ket: Integer; HalfSpin: TQuantumNumber):Boolean; override;
  end;

function TDInfinityh.Allowed(Bra, Dipole, Ket: Integer; HalfSpin: TQuantumNumber): Boolean;
begin
  Result := not Odd(Bra xor Dipole xor Ket);
  if Result then
    Result := inherited Allowed(Bra div 2, Dipole div 2, Ket div 2, HalfSpin);
end;

function TDInfinityh.SymmetryToStr(Symmetry: Integer; HalfSpin: TQuantumNumber): string;
begin
  Result := inherited SymmetryToStr(Symmetry div 2, HalfSpin);
  if Odd(Symmetry) then
    Result := Result + 'u'
  else
    Result := Result + 'g'
end;

function TDInfinityh.TryStrToSymmetry(Str: string; out Symmetry: Integer; HalfSpin: TQuantumNumber): Boolean;
var
  ug: Integer;
begin
  Result := False;
  Symmetry := -1;
  if Length(Str) < 2 then
    Exit;
  case Str[Length(Str)] of
    'g', 'G': ug := 0;
    'u', 'U': ug := 1;
  else
    Exit;
  end;
  Delete(Str,Length(Str),1);
  Result := inherited TryStrToSymmetry(Str,Symmetry,HalfSpin);
  if Result then
    Symmetry := Symmetry*2 + ug;
end;

type
  TAbelianPointGroupInfo = class(TPointGroupInfo)
    constructor Create(APointGroup: TPointGroup);
    function Allowed(Bra, Dipole, Ket: Integer; HalfSpin: TQuantumNumber):Boolean; override;
  end;

{ TAbelianPointGroupInfo }

function TAbelianPointGroupInfo.Allowed(Bra, Dipole, Ket: Integer; HalfSpin: TQuantumNumber): Boolean;
begin
  if HalfSpin <> 0 then
    raise NotYet;
  Result := Bra xor Dipole xor Ket = 0;
end;

constructor TAbelianPointGroupInfo.Create(APointGroup: TPointGroup);
var
  P: PPChar;
begin
  inherited Create(APointGroup);
  SettingsNames := IrepNames(TAbelianPointGroup(PointGroup));
  P := SettingsNames;
  FLastSymmetry := -1;
  while P^ <> nil do begin
    Inc(P);
    Inc(FLastSymmetry);
  end;
end;

{ TCachedTransitionMatrix }

destructor TCachedTransitionMatrix.Destroy;
begin
  FreeAndNil(TransitionMatrix);
  inherited;
end;

procedure TCachedTransitionMatrix.DoMake(RequiredAttributes: Pointer);
begin
  TransitionMatrix := (TransitionMoment as TMultidimensionalFCF).TransitionMatrix(BraSample,KetSample,False);
  if Length(BraOffsets) = 0 then
    with TransitionMoment as TMultidimensionalFCF do begin
      SaveOffsets(BraOffsets,BraSample,TheCacheIndex);
      SaveOffsets(KetOffsets,KetSample,TheCacheIndex);
    end;
end;

function TCachedTransitionMatrix.HasCorrectAttributes(Attributes: Pointer): Boolean;
begin
  Result := Assigned(TransitionMatrix);
end;

{ TPercentMNode }

procedure TlMatrixMNode.ClearArrays;
begin
  FreeAndNil(l);
  FreeAndNil(lambda);
  LastWarning := '';
  LastError := '';
  {$IFOPT R+}
    {$R-}
    Inc(Generation);
    {$R+}
  {$ELSE}
    Inc(Generation);
  {$ENDIF}
  inherited;
end;

function TlMatrixMNode.ElectronicState: TElectronicState;
begin
  Result := EffectiveOwner as TElectronicState
end;

function TlMatrixMNode.GetzMatrix: TzMatrix;
begin
  if not Assigned(FzMatrix) then
    FzMatrix := ElectronicState.ForceFieldNode(TzMatrix) as TzMatrix;
  Result := FzMatrix;
end;

function TlMatrixMNode.Percent(n: Integer): string;
begin
  if (n > 0) and (n <= Length(CurrentAtoms)) then
    Result := zMatrix.Atoms[CurrentAtoms[n-1]].Name
  else
    Result := inherited Percent(n);
end;

procedure TlMatrixMNode.UpdateObjectChain;
begin
  FzMatrix := nil
end;

initialization
  TVibratingMolecule.RegisterClass(TSpecies);
  TVibrationalManifold.RegisterClass(TVibratingMolecule);
  TElectronicState.RegisterClass(TVibrationalManifold);
  TMolecularNucleus.RegisterClass(TElectronicState);
  TVibrationalMode.RegisterClass(TElectronicState);
  TMultidimensionalFCF.RegisterClass(TTransitionMoments);
  TVibrationalModeChange.RegisterClass(TMultidimensionalFCF);
  TVibrationalPerturbation.RegisterClass(TVibrationalManifold);

//  TauFactor := Math.IntPower(1e21 * Planck * Avogadro / 4 / Pi / Pi, 3) / 10 / SpeedOfLight / SpeedOfLight;
//  aFactor := Sqrt(Math.IntPower(1000 * Planck * Avogadro / SpeedOfLight, 3))/(8e-27 * Pi * Pi * Pi);

  TAbelianPointGroupInfo.Create(C1);
  TAbelianPointGroupInfo.Create(Ci);
  TAbelianPointGroupInfo.Create(C2);
  TAbelianPointGroupInfo.Create(Cs);
  TAbelianPointGroupInfo.Create(D2);
  TAbelianPointGroupInfo.Create(C2v);
  TAbelianPointGroupInfo.Create(C2h);
  TAbelianPointGroupInfo.Create(D2h);
  TCinfinityv.Create(CInfinityv);
  TDInfinityh.Create(DInfinityh);

  elVariables := MakeSortedVariables(['Lambda','Omega'], [@eLambda, @eOmega]);

finalization
  FreePointGroupInfo;
  FreeAndNil(vVariables);
  FreeAndNil(lVariables);
  FreeAndNil(elVariables);
end.
