unit NuclearInfo;
{$I utils.inc}{Copyright (C) 2008, 2015 C M Western; see commandline.pas for license}
{ Data from http://physics.nist.gov/PhysRefData/Compositions }

interface

type
  TNuclearInfo = record
    Symbol: string;
    AtomicNumber: Integer;
    MassNumber: Integer;
    Mass: Double;
    Abundance: Double;
    StandardWeight: Double;
  end;

function FindNucleus(ApproximateMass: Double; Name: string = ''; MinAbundance: Double = -1):TNuclearInfo;
function FindExactMass(ApproximateMass: Double; Name: string = ''; DummyOK: Boolean = False): Double;
function FindMostAbundantNucleus(Name: string; DummyOK: Boolean): TNuclearInfo; overload;
function FindMostAbundantNucleus(AtomicNumber: Integer): TNuclearInfo; overload;
function TrimNucleusNumber(var Name: string):Integer;
function StandardWeight(const Name: string): Double;
function TryFormulaToFloat(const s: string; out Value: Double):Boolean;

implementation

uses SysUtils;

{.$DEFINE PREVIOUSVERSION}
{$DEFINE Version2015}
{ From :
  Coursey, J.S., Schwab, D.J., Tsai, J.J., and Dragoset, R.A. (2015),
  Atomic Weights and Isotopic Compositions (version 4.1). [Online]
  Available: http://physics.nist.gov/PhysRefData/Compositions [2015, Nov, 5]. National Institute of Standards and Technology, Gaithersburg, MD.
}
{$IFDEF Version2015}
const
Nuclei: array [0..3353] of TNuclearInfo = (
(Symbol:'H'; AtomicNumber:1; MassNumber:1; Mass:1.00782503223{9} ; Abundance:99.9885; {m} StandardWeight:1.00794 {1.00784-1.00811}),
(Symbol:'D'; AtomicNumber:1; MassNumber:2; Mass:2.01410177812{12} ; Abundance:0.0115; {m} StandardWeight:2.01410177812 ),
(Symbol:'H'; AtomicNumber:1; MassNumber:2; Mass:2.01410177812{12} ; Abundance:0.0115; {m} StandardWeight:1.00794 {1.00784-1.00811}), { Alternate }
(Symbol:'T'; AtomicNumber:1; MassNumber:3; Mass:3.0160492779{24} ; Abundance:0; {m} StandardWeight:3.0160492779),
(Symbol:'H'; AtomicNumber:1; MassNumber:3; Mass:3.0160492779{24} ; Abundance:0; {m} StandardWeight:1.00794 {1.00784-1.00811}), { Alternate }
(Symbol:'H'; AtomicNumber:1; MassNumber:4; Mass:4.02643{11} ; Abundance:0; {m} StandardWeight:1.00794 {1.00784-1.00811}),
(Symbol:'H'; AtomicNumber:1; MassNumber:5; Mass:5.035311{96} ; Abundance:0; {m} StandardWeight:1.00794 {1.00784-1.00811}),
(Symbol:'H'; AtomicNumber:1; MassNumber:6; Mass:6.04496{27} ; Abundance:0; {m} StandardWeight:1.00794 {1.00784-1.00811}),
(Symbol:'H'; AtomicNumber:1; MassNumber:7; Mass:7.0527{11#} ; Abundance:0; {m} StandardWeight:1.00794 {1.00784-1.00811}),
(Symbol:'He'; AtomicNumber:2; MassNumber:3; Mass:3.0160293201{25} ; Abundance:0.000134; {g,r} StandardWeight:4.002602{2}),
(Symbol:'He'; AtomicNumber:2; MassNumber:4; Mass:4.00260325413{6} ; Abundance:99.999866; {g,r} StandardWeight:4.002602{2}),
(Symbol:'He'; AtomicNumber:2; MassNumber:5; Mass:5.012057{21} ; Abundance:0; {g,r} StandardWeight:4.002602{2}),
(Symbol:'He'; AtomicNumber:2; MassNumber:6; Mass:6.018885891{57} ; Abundance:0; {g,r} StandardWeight:4.002602{2}),
(Symbol:'He'; AtomicNumber:2; MassNumber:7; Mass:7.0279907{81} ; Abundance:0; {g,r} StandardWeight:4.002602{2}),
(Symbol:'He'; AtomicNumber:2; MassNumber:8; Mass:8.033934390{95} ; Abundance:0; {g,r} StandardWeight:4.002602{2}),
(Symbol:'He'; AtomicNumber:2; MassNumber:9; Mass:9.043946{50} ; Abundance:0; {g,r} StandardWeight:4.002602{2}),
(Symbol:'He'; AtomicNumber:2; MassNumber:10; Mass:10.05279{11} ; Abundance:0; {g,r} StandardWeight:4.002602{2}),
(Symbol:'Li'; AtomicNumber:3; MassNumber:3; Mass:3.0308{21#} ; Abundance:0; {m} StandardWeight:6.940 {6.938-6.997}),
(Symbol:'Li'; AtomicNumber:3; MassNumber:4; Mass:4.02719{23} ; Abundance:0; {m} StandardWeight:6.940 {6.938-6.997}),
(Symbol:'Li'; AtomicNumber:3; MassNumber:5; Mass:5.012538{54} ; Abundance:0; {m} StandardWeight:6.940 {6.938-6.997}),
(Symbol:'Li'; AtomicNumber:3; MassNumber:6; Mass:6.0151228874{16} ; Abundance:7.59; {m} StandardWeight:6.940 {6.938-6.997}),
(Symbol:'Li'; AtomicNumber:3; MassNumber:7; Mass:7.0160034366{45} ; Abundance:92.41; {m} StandardWeight:6.940 {6.938-6.997}),
(Symbol:'Li'; AtomicNumber:3; MassNumber:8; Mass:8.022486246{50} ; Abundance:0; {m} StandardWeight:6.940 {6.938-6.997}),
(Symbol:'Li'; AtomicNumber:3; MassNumber:9; Mass:9.02679019{20} ; Abundance:0; {m} StandardWeight:6.940 {6.938-6.997}),
(Symbol:'Li'; AtomicNumber:3; MassNumber:10; Mass:10.035483{14} ; Abundance:0; {m} StandardWeight:6.940 {6.938-6.997}),
(Symbol:'Li'; AtomicNumber:3; MassNumber:11; Mass:11.04372358{66} ; Abundance:0; {m} StandardWeight:6.940 {6.938-6.997}),
(Symbol:'Li'; AtomicNumber:3; MassNumber:12; Mass:12.052517{16} ; Abundance:0; {m} StandardWeight:6.940 {6.938-6.997}),
(Symbol:'Li'; AtomicNumber:3; MassNumber:13; Mass:13.06263{38} ; Abundance:0; {m} StandardWeight:6.940 {6.938-6.997}),
(Symbol:'Be'; AtomicNumber:4; MassNumber:5; Mass:5.0399{22#} ; Abundance:0;  StandardWeight:9.0121831{5}),
(Symbol:'Be'; AtomicNumber:4; MassNumber:6; Mass:6.0197264{58} ; Abundance:0;  StandardWeight:9.0121831{5}),
(Symbol:'Be'; AtomicNumber:4; MassNumber:7; Mass:7.016928717{76} ; Abundance:0;  StandardWeight:9.0121831{5}),
(Symbol:'Be'; AtomicNumber:4; MassNumber:8; Mass:8.005305102{37} ; Abundance:0;  StandardWeight:9.0121831{5}),
(Symbol:'Be'; AtomicNumber:4; MassNumber:9; Mass:9.012183065{82} ; Abundance:100;  StandardWeight:9.0121831{5}),
(Symbol:'Be'; AtomicNumber:4; MassNumber:10; Mass:10.013534695{86} ; Abundance:0;  StandardWeight:9.0121831{5}),
(Symbol:'Be'; AtomicNumber:4; MassNumber:11; Mass:11.02166108{26} ; Abundance:0;  StandardWeight:9.0121831{5}),
(Symbol:'Be'; AtomicNumber:4; MassNumber:12; Mass:12.0269221{20} ; Abundance:0;  StandardWeight:9.0121831{5}),
(Symbol:'Be'; AtomicNumber:4; MassNumber:13; Mass:13.036135{11} ; Abundance:0;  StandardWeight:9.0121831{5}),
(Symbol:'Be'; AtomicNumber:4; MassNumber:14; Mass:14.04289{14} ; Abundance:0;  StandardWeight:9.0121831{5}),
(Symbol:'Be'; AtomicNumber:4; MassNumber:15; Mass:15.05342{43#} ; Abundance:0;  StandardWeight:9.0121831{5}),
(Symbol:'Be'; AtomicNumber:4; MassNumber:16; Mass:16.06167{18} ; Abundance:0;  StandardWeight:9.0121831{5}),
(Symbol:'B'; AtomicNumber:5; MassNumber:6; Mass:6.0508{22#} ; Abundance:0; {m} StandardWeight:10.811 {10.806-10.821}),
(Symbol:'B'; AtomicNumber:5; MassNumber:7; Mass:7.029712{27} ; Abundance:0; {m} StandardWeight:10.811 {10.806-10.821}),
(Symbol:'B'; AtomicNumber:5; MassNumber:8; Mass:8.0246073{11} ; Abundance:0; {m} StandardWeight:10.811 {10.806-10.821}),
(Symbol:'B'; AtomicNumber:5; MassNumber:9; Mass:9.01332965{97} ; Abundance:0; {m} StandardWeight:10.811 {10.806-10.821}),
(Symbol:'B'; AtomicNumber:5; MassNumber:10; Mass:10.01293695{41} ; Abundance:19.9; {m} StandardWeight:10.811 {10.806-10.821}),
(Symbol:'B'; AtomicNumber:5; MassNumber:11; Mass:11.00930536{45} ; Abundance:80.1; {m} StandardWeight:10.811 {10.806-10.821}),
(Symbol:'B'; AtomicNumber:5; MassNumber:12; Mass:12.0143527{14} ; Abundance:0; {m} StandardWeight:10.811 {10.806-10.821}),
(Symbol:'B'; AtomicNumber:5; MassNumber:13; Mass:13.0177802{12} ; Abundance:0; {m} StandardWeight:10.811 {10.806-10.821}),
(Symbol:'B'; AtomicNumber:5; MassNumber:14; Mass:14.025404{23} ; Abundance:0; {m} StandardWeight:10.811 {10.806-10.821}),
(Symbol:'B'; AtomicNumber:5; MassNumber:15; Mass:15.031088{23} ; Abundance:0; {m} StandardWeight:10.811 {10.806-10.821}),
(Symbol:'B'; AtomicNumber:5; MassNumber:16; Mass:16.039842{26} ; Abundance:0; {m} StandardWeight:10.811 {10.806-10.821}),
(Symbol:'B'; AtomicNumber:5; MassNumber:17; Mass:17.04699{18} ; Abundance:0; {m} StandardWeight:10.811 {10.806-10.821}),
(Symbol:'B'; AtomicNumber:5; MassNumber:18; Mass:18.05566{18} ; Abundance:0; {m} StandardWeight:10.811 {10.806-10.821}),
(Symbol:'B'; AtomicNumber:5; MassNumber:19; Mass:19.06310{43#} ; Abundance:0; {m} StandardWeight:10.811 {10.806-10.821}),
(Symbol:'B'; AtomicNumber:5; MassNumber:20; Mass:20.07207{75#} ; Abundance:0; {m} StandardWeight:10.811 {10.806-10.821}),
(Symbol:'B'; AtomicNumber:5; MassNumber:21; Mass:21.08129{97#} ; Abundance:0; {m} StandardWeight:10.811 {10.806-10.821}),
(Symbol:'C'; AtomicNumber:6; MassNumber:8; Mass:8.037643{20} ; Abundance:0;  StandardWeight:12.0107 {12.0096-12.0116}),
(Symbol:'C'; AtomicNumber:6; MassNumber:9; Mass:9.0310372{23} ; Abundance:0;  StandardWeight:12.0107 {12.0096-12.0116}),
(Symbol:'C'; AtomicNumber:6; MassNumber:10; Mass:10.01685331{42} ; Abundance:0;  StandardWeight:12.0107 {12.0096-12.0116}),
(Symbol:'C'; AtomicNumber:6; MassNumber:11; Mass:11.0114336{10} ; Abundance:0;  StandardWeight:12.0107 {12.0096-12.0116}),
(Symbol:'C'; AtomicNumber:6; MassNumber:12; Mass:12.0000000{00} ; Abundance:98.93;  StandardWeight:12.0107 {12.0096-12.0116}),
(Symbol:'C'; AtomicNumber:6; MassNumber:13; Mass:13.00335483507{23} ; Abundance:1.07;  StandardWeight:12.0107 {12.0096-12.0116}),
(Symbol:'C'; AtomicNumber:6; MassNumber:14; Mass:14.0032419884{40} ; Abundance:0;  StandardWeight:12.0107 {12.0096-12.0116}),
(Symbol:'C'; AtomicNumber:6; MassNumber:15; Mass:15.01059926{86} ; Abundance:0;  StandardWeight:12.0107 {12.0096-12.0116}),
(Symbol:'C'; AtomicNumber:6; MassNumber:16; Mass:16.0147013{38} ; Abundance:0;  StandardWeight:12.0107 {12.0096-12.0116}),
(Symbol:'C'; AtomicNumber:6; MassNumber:17; Mass:17.022577{19} ; Abundance:0;  StandardWeight:12.0107 {12.0096-12.0116}),
(Symbol:'C'; AtomicNumber:6; MassNumber:18; Mass:18.026751{32} ; Abundance:0;  StandardWeight:12.0107 {12.0096-12.0116}),
(Symbol:'C'; AtomicNumber:6; MassNumber:19; Mass:19.03480{11} ; Abundance:0;  StandardWeight:12.0107 {12.0096-12.0116}),
(Symbol:'C'; AtomicNumber:6; MassNumber:20; Mass:20.04032{26} ; Abundance:0;  StandardWeight:12.0107 {12.0096-12.0116}),
(Symbol:'C'; AtomicNumber:6; MassNumber:21; Mass:21.04900{43#} ; Abundance:0;  StandardWeight:12.0107 {12.0096-12.0116}),
(Symbol:'C'; AtomicNumber:6; MassNumber:22; Mass:22.05753{26} ; Abundance:0;  StandardWeight:12.0107 {12.0096-12.0116}),
(Symbol:'C'; AtomicNumber:6; MassNumber:23; Mass:23.0689{11#} ; Abundance:0;  StandardWeight:12.0107 {12.0096-12.0116}),
(Symbol:'N'; AtomicNumber:7; MassNumber:10; Mass:10.04165{43} ; Abundance:0;  StandardWeight:14.00670 {14.00643-14.00728}),
(Symbol:'N'; AtomicNumber:7; MassNumber:11; Mass:11.026091{50} ; Abundance:0;  StandardWeight:14.00670 {14.00643-14.00728}),
(Symbol:'N'; AtomicNumber:7; MassNumber:12; Mass:12.0186132{11} ; Abundance:0;  StandardWeight:14.00670 {14.00643-14.00728}),
(Symbol:'N'; AtomicNumber:7; MassNumber:13; Mass:13.00573861{29} ; Abundance:0;  StandardWeight:14.00670 {14.00643-14.00728}),
(Symbol:'N'; AtomicNumber:7; MassNumber:14; Mass:14.00307400443{20} ; Abundance:99.636;  StandardWeight:14.00670 {14.00643-14.00728}),
(Symbol:'N'; AtomicNumber:7; MassNumber:15; Mass:15.00010889888{64} ; Abundance:0.364;  StandardWeight:14.00670 {14.00643-14.00728}),
(Symbol:'N'; AtomicNumber:7; MassNumber:16; Mass:16.0061019{25} ; Abundance:0;  StandardWeight:14.00670 {14.00643-14.00728}),
(Symbol:'N'; AtomicNumber:7; MassNumber:17; Mass:17.008449{16} ; Abundance:0;  StandardWeight:14.00670 {14.00643-14.00728}),
(Symbol:'N'; AtomicNumber:7; MassNumber:18; Mass:18.014078{20} ; Abundance:0;  StandardWeight:14.00670 {14.00643-14.00728}),
(Symbol:'N'; AtomicNumber:7; MassNumber:19; Mass:19.017022{18} ; Abundance:0;  StandardWeight:14.00670 {14.00643-14.00728}),
(Symbol:'N'; AtomicNumber:7; MassNumber:20; Mass:20.023366{60} ; Abundance:0;  StandardWeight:14.00670 {14.00643-14.00728}),
(Symbol:'N'; AtomicNumber:7; MassNumber:21; Mass:21.02711{10} ; Abundance:0;  StandardWeight:14.00670 {14.00643-14.00728}),
(Symbol:'N'; AtomicNumber:7; MassNumber:22; Mass:22.03439{21} ; Abundance:0;  StandardWeight:14.00670 {14.00643-14.00728}),
(Symbol:'N'; AtomicNumber:7; MassNumber:23; Mass:23.04114{32#} ; Abundance:0;  StandardWeight:14.00670 {14.00643-14.00728}),
(Symbol:'N'; AtomicNumber:7; MassNumber:24; Mass:24.05039{43#} ; Abundance:0;  StandardWeight:14.00670 {14.00643-14.00728}),
(Symbol:'N'; AtomicNumber:7; MassNumber:25; Mass:25.06010{54#} ; Abundance:0;  StandardWeight:14.00670 {14.00643-14.00728}),
(Symbol:'O'; AtomicNumber:8; MassNumber:12; Mass:12.034262{26} ; Abundance:0;  StandardWeight:15.99940 {15.99903-15.99977}),
(Symbol:'O'; AtomicNumber:8; MassNumber:13; Mass:13.024815{10} ; Abundance:0;  StandardWeight:15.99940 {15.99903-15.99977}),
(Symbol:'O'; AtomicNumber:8; MassNumber:14; Mass:14.00859636{12} ; Abundance:0;  StandardWeight:15.99940 {15.99903-15.99977}),
(Symbol:'O'; AtomicNumber:8; MassNumber:15; Mass:15.00306562{53} ; Abundance:0;  StandardWeight:15.99940 {15.99903-15.99977}),
(Symbol:'O'; AtomicNumber:8; MassNumber:16; Mass:15.99491461957{17} ; Abundance:99.757;  StandardWeight:15.99940 {15.99903-15.99977}),
(Symbol:'O'; AtomicNumber:8; MassNumber:17; Mass:16.99913175650{69} ; Abundance:0.038;  StandardWeight:15.99940 {15.99903-15.99977}),
(Symbol:'O'; AtomicNumber:8; MassNumber:18; Mass:17.99915961286{76} ; Abundance:0.205;  StandardWeight:15.99940 {15.99903-15.99977}),
(Symbol:'O'; AtomicNumber:8; MassNumber:19; Mass:19.0035780{28} ; Abundance:0;  StandardWeight:15.99940 {15.99903-15.99977}),
(Symbol:'O'; AtomicNumber:8; MassNumber:20; Mass:20.00407535{95} ; Abundance:0;  StandardWeight:15.99940 {15.99903-15.99977}),
(Symbol:'O'; AtomicNumber:8; MassNumber:21; Mass:21.008655{13} ; Abundance:0;  StandardWeight:15.99940 {15.99903-15.99977}),
(Symbol:'O'; AtomicNumber:8; MassNumber:22; Mass:22.009966{61} ; Abundance:0;  StandardWeight:15.99940 {15.99903-15.99977}),
(Symbol:'O'; AtomicNumber:8; MassNumber:23; Mass:23.015696{97} ; Abundance:0;  StandardWeight:15.99940 {15.99903-15.99977}),
(Symbol:'O'; AtomicNumber:8; MassNumber:24; Mass:24.01986{12} ; Abundance:0;  StandardWeight:15.99940 {15.99903-15.99977}),
(Symbol:'O'; AtomicNumber:8; MassNumber:25; Mass:25.02936{12} ; Abundance:0;  StandardWeight:15.99940 {15.99903-15.99977}),
(Symbol:'O'; AtomicNumber:8; MassNumber:26; Mass:26.03729{17} ; Abundance:0;  StandardWeight:15.99940 {15.99903-15.99977}),
(Symbol:'O'; AtomicNumber:8; MassNumber:27; Mass:27.04772{54#} ; Abundance:0;  StandardWeight:15.99940 {15.99903-15.99977}),
(Symbol:'O'; AtomicNumber:8; MassNumber:28; Mass:28.05591{75#} ; Abundance:0;  StandardWeight:15.99940 {15.99903-15.99977}),
(Symbol:'F'; AtomicNumber:9; MassNumber:14; Mass:14.034315{44} ; Abundance:0;  StandardWeight:18.998403163{6}),
(Symbol:'F'; AtomicNumber:9; MassNumber:15; Mass:15.018043{67} ; Abundance:0;  StandardWeight:18.998403163{6}),
(Symbol:'F'; AtomicNumber:9; MassNumber:16; Mass:16.0114657{89} ; Abundance:0;  StandardWeight:18.998403163{6}),
(Symbol:'F'; AtomicNumber:9; MassNumber:17; Mass:17.00209524{27} ; Abundance:0;  StandardWeight:18.998403163{6}),
(Symbol:'F'; AtomicNumber:9; MassNumber:18; Mass:18.00093733{50} ; Abundance:0;  StandardWeight:18.998403163{6}),
(Symbol:'F'; AtomicNumber:9; MassNumber:19; Mass:18.99840316273{92} ; Abundance:100;  StandardWeight:18.998403163{6}),
(Symbol:'F'; AtomicNumber:9; MassNumber:20; Mass:19.999981252{31} ; Abundance:0;  StandardWeight:18.998403163{6}),
(Symbol:'F'; AtomicNumber:9; MassNumber:21; Mass:20.9999489{19} ; Abundance:0;  StandardWeight:18.998403163{6}),
(Symbol:'F'; AtomicNumber:9; MassNumber:22; Mass:22.002999{13} ; Abundance:0;  StandardWeight:18.998403163{6}),
(Symbol:'F'; AtomicNumber:9; MassNumber:23; Mass:23.003557{54} ; Abundance:0;  StandardWeight:18.998403163{6}),
(Symbol:'F'; AtomicNumber:9; MassNumber:24; Mass:24.008115{78} ; Abundance:0;  StandardWeight:18.998403163{6}),
(Symbol:'F'; AtomicNumber:9; MassNumber:25; Mass:25.012199{81} ; Abundance:0;  StandardWeight:18.998403163{6}),
(Symbol:'F'; AtomicNumber:9; MassNumber:26; Mass:26.020038{83} ; Abundance:0;  StandardWeight:18.998403163{6}),
(Symbol:'F'; AtomicNumber:9; MassNumber:27; Mass:27.02644{20} ; Abundance:0;  StandardWeight:18.998403163{6}),
(Symbol:'F'; AtomicNumber:9; MassNumber:28; Mass:28.03534{21} ; Abundance:0;  StandardWeight:18.998403163{6}),
(Symbol:'F'; AtomicNumber:9; MassNumber:29; Mass:29.04254{54#} ; Abundance:0;  StandardWeight:18.998403163{6}),
(Symbol:'F'; AtomicNumber:9; MassNumber:30; Mass:30.05165{64#} ; Abundance:0;  StandardWeight:18.998403163{6}),
(Symbol:'F'; AtomicNumber:9; MassNumber:31; Mass:31.05971{56#} ; Abundance:0;  StandardWeight:18.998403163{6}),
(Symbol:'Ne'; AtomicNumber:10; MassNumber:16; Mass:16.025750{22} ; Abundance:0; {g,m} StandardWeight:20.1797{6}),
(Symbol:'Ne'; AtomicNumber:10; MassNumber:17; Mass:17.01771396{38} ; Abundance:0; {g,m} StandardWeight:20.1797{6}),
(Symbol:'Ne'; AtomicNumber:10; MassNumber:18; Mass:18.00570870{39} ; Abundance:0; {g,m} StandardWeight:20.1797{6}),
(Symbol:'Ne'; AtomicNumber:10; MassNumber:19; Mass:19.00188091{17} ; Abundance:0; {g,m} StandardWeight:20.1797{6}),
(Symbol:'Ne'; AtomicNumber:10; MassNumber:20; Mass:19.9924401762{17} ; Abundance:90.48; {g,m} StandardWeight:20.1797{6}),
(Symbol:'Ne'; AtomicNumber:10; MassNumber:21; Mass:20.993846685{41} ; Abundance:0.27; {g,m} StandardWeight:20.1797{6}),
(Symbol:'Ne'; AtomicNumber:10; MassNumber:22; Mass:21.991385114{18} ; Abundance:9.25; {g,m} StandardWeight:20.1797{6}),
(Symbol:'Ne'; AtomicNumber:10; MassNumber:23; Mass:22.99446691{11} ; Abundance:0; {g,m} StandardWeight:20.1797{6}),
(Symbol:'Ne'; AtomicNumber:10; MassNumber:24; Mass:23.99361065{55} ; Abundance:0; {g,m} StandardWeight:20.1797{6}),
(Symbol:'Ne'; AtomicNumber:10; MassNumber:25; Mass:24.997789{48} ; Abundance:0; {g,m} StandardWeight:20.1797{6}),
(Symbol:'Ne'; AtomicNumber:10; MassNumber:26; Mass:26.000515{20} ; Abundance:0; {g,m} StandardWeight:20.1797{6}),
(Symbol:'Ne'; AtomicNumber:10; MassNumber:27; Mass:27.007553{70} ; Abundance:0; {g,m} StandardWeight:20.1797{6}),
(Symbol:'Ne'; AtomicNumber:10; MassNumber:28; Mass:28.01212{10} ; Abundance:0; {g,m} StandardWeight:20.1797{6}),
(Symbol:'Ne'; AtomicNumber:10; MassNumber:29; Mass:29.01975{11} ; Abundance:0; {g,m} StandardWeight:20.1797{6}),
(Symbol:'Ne'; AtomicNumber:10; MassNumber:30; Mass:30.02473{30} ; Abundance:0; {g,m} StandardWeight:20.1797{6}),
(Symbol:'Ne'; AtomicNumber:10; MassNumber:31; Mass:31.0331{17} ; Abundance:0; {g,m} StandardWeight:20.1797{6}),
(Symbol:'Ne'; AtomicNumber:10; MassNumber:32; Mass:32.03972{54#} ; Abundance:0; {g,m} StandardWeight:20.1797{6}),
(Symbol:'Ne'; AtomicNumber:10; MassNumber:33; Mass:33.04938{64#} ; Abundance:0; {g,m} StandardWeight:20.1797{6}),
(Symbol:'Ne'; AtomicNumber:10; MassNumber:34; Mass:34.05673{55#} ; Abundance:0; {g,m} StandardWeight:20.1797{6}),
(Symbol:'Na'; AtomicNumber:11; MassNumber:18; Mass:18.02688{12} ; Abundance:0;  StandardWeight:22.98976928{2}),
(Symbol:'Na'; AtomicNumber:11; MassNumber:19; Mass:19.013880{11} ; Abundance:0;  StandardWeight:22.98976928{2}),
(Symbol:'Na'; AtomicNumber:11; MassNumber:20; Mass:20.0073544{12} ; Abundance:0;  StandardWeight:22.98976928{2}),
(Symbol:'Na'; AtomicNumber:11; MassNumber:21; Mass:20.99765469{30} ; Abundance:0;  StandardWeight:22.98976928{2}),
(Symbol:'Na'; AtomicNumber:11; MassNumber:22; Mass:21.99443741{18} ; Abundance:0;  StandardWeight:22.98976928{2}),
(Symbol:'Na'; AtomicNumber:11; MassNumber:23; Mass:22.9897692820{19} ; Abundance:100;  StandardWeight:22.98976928{2}),
(Symbol:'Na'; AtomicNumber:11; MassNumber:24; Mass:23.990962950{38} ; Abundance:0;  StandardWeight:22.98976928{2}),
(Symbol:'Na'; AtomicNumber:11; MassNumber:25; Mass:24.9899540{13} ; Abundance:0;  StandardWeight:22.98976928{2}),
(Symbol:'Na'; AtomicNumber:11; MassNumber:26; Mass:25.9926346{38} ; Abundance:0;  StandardWeight:22.98976928{2}),
(Symbol:'Na'; AtomicNumber:11; MassNumber:27; Mass:26.9940765{40} ; Abundance:0;  StandardWeight:22.98976928{2}),
(Symbol:'Na'; AtomicNumber:11; MassNumber:28; Mass:27.998939{11} ; Abundance:0;  StandardWeight:22.98976928{2}),
(Symbol:'Na'; AtomicNumber:11; MassNumber:29; Mass:29.0028771{79} ; Abundance:0;  StandardWeight:22.98976928{2}),
(Symbol:'Na'; AtomicNumber:11; MassNumber:30; Mass:30.0090979{51} ; Abundance:0;  StandardWeight:22.98976928{2}),
(Symbol:'Na'; AtomicNumber:11; MassNumber:31; Mass:31.013163{25} ; Abundance:0;  StandardWeight:22.98976928{2}),
(Symbol:'Na'; AtomicNumber:11; MassNumber:32; Mass:32.02019{13} ; Abundance:0;  StandardWeight:22.98976928{2}),
(Symbol:'Na'; AtomicNumber:11; MassNumber:33; Mass:33.02573{64#} ; Abundance:0;  StandardWeight:22.98976928{2}),
(Symbol:'Na'; AtomicNumber:11; MassNumber:34; Mass:34.03359{54#} ; Abundance:0;  StandardWeight:22.98976928{2}),
(Symbol:'Na'; AtomicNumber:11; MassNumber:35; Mass:35.04062{63#} ; Abundance:0;  StandardWeight:22.98976928{2}),
(Symbol:'Na'; AtomicNumber:11; MassNumber:36; Mass:36.04929{64#} ; Abundance:0;  StandardWeight:22.98976928{2}),
(Symbol:'Na'; AtomicNumber:11; MassNumber:37; Mass:37.05705{65#} ; Abundance:0;  StandardWeight:22.98976928{2}),
(Symbol:'Mg'; AtomicNumber:12; MassNumber:19; Mass:19.034169{54} ; Abundance:0;  StandardWeight:24.305 {24.304-24.307}),
(Symbol:'Mg'; AtomicNumber:12; MassNumber:20; Mass:20.018850{29} ; Abundance:0;  StandardWeight:24.305 {24.304-24.307}),
(Symbol:'Mg'; AtomicNumber:12; MassNumber:21; Mass:21.011716{18} ; Abundance:0;  StandardWeight:24.305 {24.304-24.307}),
(Symbol:'Mg'; AtomicNumber:12; MassNumber:22; Mass:21.99957065{34} ; Abundance:0;  StandardWeight:24.305 {24.304-24.307}),
(Symbol:'Mg'; AtomicNumber:12; MassNumber:23; Mass:22.99412421{74} ; Abundance:0;  StandardWeight:24.305 {24.304-24.307}),
(Symbol:'Mg'; AtomicNumber:12; MassNumber:24; Mass:23.985041697{14} ; Abundance:78.99;  StandardWeight:24.305 {24.304-24.307}),
(Symbol:'Mg'; AtomicNumber:12; MassNumber:25; Mass:24.985836976{50} ; Abundance:10;  StandardWeight:24.305 {24.304-24.307}),
(Symbol:'Mg'; AtomicNumber:12; MassNumber:26; Mass:25.982592968{31} ; Abundance:11.01;  StandardWeight:24.305 {24.304-24.307}),
(Symbol:'Mg'; AtomicNumber:12; MassNumber:27; Mass:26.984340624{53} ; Abundance:0;  StandardWeight:24.305 {24.304-24.307}),
(Symbol:'Mg'; AtomicNumber:12; MassNumber:28; Mass:27.9838767{22} ; Abundance:0;  StandardWeight:24.305 {24.304-24.307}),
(Symbol:'Mg'; AtomicNumber:12; MassNumber:29; Mass:28.988617{12} ; Abundance:0;  StandardWeight:24.305 {24.304-24.307}),
(Symbol:'Mg'; AtomicNumber:12; MassNumber:30; Mass:29.9904629{37} ; Abundance:0;  StandardWeight:24.305 {24.304-24.307}),
(Symbol:'Mg'; AtomicNumber:12; MassNumber:31; Mass:30.9966480{33} ; Abundance:0;  StandardWeight:24.305 {24.304-24.307}),
(Symbol:'Mg'; AtomicNumber:12; MassNumber:32; Mass:31.9991102{34} ; Abundance:0;  StandardWeight:24.305 {24.304-24.307}),
(Symbol:'Mg'; AtomicNumber:12; MassNumber:33; Mass:33.0053271{31} ; Abundance:0;  StandardWeight:24.305 {24.304-24.307}),
(Symbol:'Mg'; AtomicNumber:12; MassNumber:34; Mass:34.008935{31} ; Abundance:0;  StandardWeight:24.305 {24.304-24.307}),
(Symbol:'Mg'; AtomicNumber:12; MassNumber:35; Mass:35.01679{19} ; Abundance:0;  StandardWeight:24.305 {24.304-24.307}),
(Symbol:'Mg'; AtomicNumber:12; MassNumber:36; Mass:36.02188{49} ; Abundance:0;  StandardWeight:24.305 {24.304-24.307}),
(Symbol:'Mg'; AtomicNumber:12; MassNumber:37; Mass:37.03037{54#} ; Abundance:0;  StandardWeight:24.305 {24.304-24.307}),
(Symbol:'Mg'; AtomicNumber:12; MassNumber:38; Mass:38.03658{54#} ; Abundance:0;  StandardWeight:24.305 {24.304-24.307}),
(Symbol:'Mg'; AtomicNumber:12; MassNumber:39; Mass:39.04538{55#} ; Abundance:0;  StandardWeight:24.305 {24.304-24.307}),
(Symbol:'Mg'; AtomicNumber:12; MassNumber:40; Mass:40.05218{64#} ; Abundance:0;  StandardWeight:24.305 {24.304-24.307}),
(Symbol:'Al'; AtomicNumber:13; MassNumber:21; Mass:21.02897{43#} ; Abundance:0;  StandardWeight:26.9815385{7}),
(Symbol:'Al'; AtomicNumber:13; MassNumber:22; Mass:22.01954{43#} ; Abundance:0;  StandardWeight:26.9815385{7}),
(Symbol:'Al'; AtomicNumber:13; MassNumber:23; Mass:23.00724435{37} ; Abundance:0;  StandardWeight:26.9815385{7}),
(Symbol:'Al'; AtomicNumber:13; MassNumber:24; Mass:23.9999489{12} ; Abundance:0;  StandardWeight:26.9815385{7}),
(Symbol:'Al'; AtomicNumber:13; MassNumber:25; Mass:24.99042810{51} ; Abundance:0;  StandardWeight:26.9815385{7}),
(Symbol:'Al'; AtomicNumber:13; MassNumber:26; Mass:25.986891904{69} ; Abundance:0;  StandardWeight:26.9815385{7}),
(Symbol:'Al'; AtomicNumber:13; MassNumber:27; Mass:26.98153853{11} ; Abundance:100;  StandardWeight:26.9815385{7}),
(Symbol:'Al'; AtomicNumber:13; MassNumber:28; Mass:27.98191021{13} ; Abundance:0;  StandardWeight:26.9815385{7}),
(Symbol:'Al'; AtomicNumber:13; MassNumber:29; Mass:28.9804565{10} ; Abundance:0;  StandardWeight:26.9815385{7}),
(Symbol:'Al'; AtomicNumber:13; MassNumber:30; Mass:29.982960{15} ; Abundance:0;  StandardWeight:26.9815385{7}),
(Symbol:'Al'; AtomicNumber:13; MassNumber:31; Mass:30.983945{22} ; Abundance:0;  StandardWeight:26.9815385{7}),
(Symbol:'Al'; AtomicNumber:13; MassNumber:32; Mass:31.988085{13} ; Abundance:0;  StandardWeight:26.9815385{7}),
(Symbol:'Al'; AtomicNumber:13; MassNumber:33; Mass:32.990909{81} ; Abundance:0;  StandardWeight:26.9815385{7}),
(Symbol:'Al'; AtomicNumber:13; MassNumber:34; Mass:33.996705{74} ; Abundance:0;  StandardWeight:26.9815385{7}),
(Symbol:'Al'; AtomicNumber:13; MassNumber:35; Mass:34.999764{75} ; Abundance:0;  StandardWeight:26.9815385{7}),
(Symbol:'Al'; AtomicNumber:13; MassNumber:36; Mass:36.00639{11} ; Abundance:0;  StandardWeight:26.9815385{7}),
(Symbol:'Al'; AtomicNumber:13; MassNumber:37; Mass:37.01053{13} ; Abundance:0;  StandardWeight:26.9815385{7}),
(Symbol:'Al'; AtomicNumber:13; MassNumber:38; Mass:38.01740{27} ; Abundance:0;  StandardWeight:26.9815385{7}),
(Symbol:'Al'; AtomicNumber:13; MassNumber:39; Mass:39.02254{54#} ; Abundance:0;  StandardWeight:26.9815385{7}),
(Symbol:'Al'; AtomicNumber:13; MassNumber:40; Mass:40.03003{54#} ; Abundance:0;  StandardWeight:26.9815385{7}),
(Symbol:'Al'; AtomicNumber:13; MassNumber:41; Mass:41.03638{64#} ; Abundance:0;  StandardWeight:26.9815385{7}),
(Symbol:'Al'; AtomicNumber:13; MassNumber:42; Mass:42.04384{64#} ; Abundance:0;  StandardWeight:26.9815385{7}),
(Symbol:'Al'; AtomicNumber:13; MassNumber:43; Mass:43.05147{75#} ; Abundance:0;  StandardWeight:26.9815385{7}),
(Symbol:'Si'; AtomicNumber:14; MassNumber:22; Mass:22.03579{54#} ; Abundance:0;  StandardWeight:28.085 {28.084-28.086}),
(Symbol:'Si'; AtomicNumber:14; MassNumber:23; Mass:23.02544{54#} ; Abundance:0;  StandardWeight:28.085 {28.084-28.086}),
(Symbol:'Si'; AtomicNumber:14; MassNumber:24; Mass:24.011535{21} ; Abundance:0;  StandardWeight:28.085 {28.084-28.086}),
(Symbol:'Si'; AtomicNumber:14; MassNumber:25; Mass:25.004109{11} ; Abundance:0;  StandardWeight:28.085 {28.084-28.086}),
(Symbol:'Si'; AtomicNumber:14; MassNumber:26; Mass:25.99233384{11} ; Abundance:0;  StandardWeight:28.085 {28.084-28.086}),
(Symbol:'Si'; AtomicNumber:14; MassNumber:27; Mass:26.98670481{15} ; Abundance:0;  StandardWeight:28.085 {28.084-28.086}),
(Symbol:'Si'; AtomicNumber:14; MassNumber:28; Mass:27.97692653465{44} ; Abundance:92.223;  StandardWeight:28.085 {28.084-28.086}),
(Symbol:'Si'; AtomicNumber:14; MassNumber:29; Mass:28.97649466490{52} ; Abundance:4.685;  StandardWeight:28.085 {28.084-28.086}),
(Symbol:'Si'; AtomicNumber:14; MassNumber:30; Mass:29.973770136{23} ; Abundance:3.092;  StandardWeight:28.085 {28.084-28.086}),
(Symbol:'Si'; AtomicNumber:14; MassNumber:31; Mass:30.975363194{46} ; Abundance:0;  StandardWeight:28.085 {28.084-28.086}),
(Symbol:'Si'; AtomicNumber:14; MassNumber:32; Mass:31.97415154{32} ; Abundance:0;  StandardWeight:28.085 {28.084-28.086}),
(Symbol:'Si'; AtomicNumber:14; MassNumber:33; Mass:32.97797696{75} ; Abundance:0;  StandardWeight:28.085 {28.084-28.086}),
(Symbol:'Si'; AtomicNumber:14; MassNumber:34; Mass:33.978576{15} ; Abundance:0;  StandardWeight:28.085 {28.084-28.086}),
(Symbol:'Si'; AtomicNumber:14; MassNumber:35; Mass:34.984583{41} ; Abundance:0;  StandardWeight:28.085 {28.084-28.086}),
(Symbol:'Si'; AtomicNumber:14; MassNumber:36; Mass:35.986695{77} ; Abundance:0;  StandardWeight:28.085 {28.084-28.086}),
(Symbol:'Si'; AtomicNumber:14; MassNumber:37; Mass:36.992921{89} ; Abundance:0;  StandardWeight:28.085 {28.084-28.086}),
(Symbol:'Si'; AtomicNumber:14; MassNumber:38; Mass:37.995523{75} ; Abundance:0;  StandardWeight:28.085 {28.084-28.086}),
(Symbol:'Si'; AtomicNumber:14; MassNumber:39; Mass:39.002491{97} ; Abundance:0;  StandardWeight:28.085 {28.084-28.086}),
(Symbol:'Si'; AtomicNumber:14; MassNumber:40; Mass:40.00583{25} ; Abundance:0;  StandardWeight:28.085 {28.084-28.086}),
(Symbol:'Si'; AtomicNumber:14; MassNumber:41; Mass:41.01301{40} ; Abundance:0;  StandardWeight:28.085 {28.084-28.086}),
(Symbol:'Si'; AtomicNumber:14; MassNumber:42; Mass:42.01778{54#} ; Abundance:0;  StandardWeight:28.085 {28.084-28.086}),
(Symbol:'Si'; AtomicNumber:14; MassNumber:43; Mass:43.02480{64#} ; Abundance:0;  StandardWeight:28.085 {28.084-28.086}),
(Symbol:'Si'; AtomicNumber:14; MassNumber:44; Mass:44.03061{64#} ; Abundance:0;  StandardWeight:28.085 {28.084-28.086}),
(Symbol:'Si'; AtomicNumber:14; MassNumber:45; Mass:45.03995{75#} ; Abundance:0;  StandardWeight:28.085 {28.084-28.086}),
(Symbol:'P'; AtomicNumber:15; MassNumber:24; Mass:24.03577{54#} ; Abundance:0;  StandardWeight:30.973761998{5}),
(Symbol:'P'; AtomicNumber:15; MassNumber:25; Mass:25.02119{43#} ; Abundance:0;  StandardWeight:30.973761998{5}),
(Symbol:'P'; AtomicNumber:15; MassNumber:26; Mass:26.01178{21#} ; Abundance:0;  StandardWeight:30.973761998{5}),
(Symbol:'P'; AtomicNumber:15; MassNumber:27; Mass:26.999224{28} ; Abundance:0;  StandardWeight:30.973761998{5}),
(Symbol:'P'; AtomicNumber:15; MassNumber:28; Mass:27.9923266{12} ; Abundance:0;  StandardWeight:30.973761998{5}),
(Symbol:'P'; AtomicNumber:15; MassNumber:29; Mass:28.98180079{60} ; Abundance:0;  StandardWeight:30.973761998{5}),
(Symbol:'P'; AtomicNumber:15; MassNumber:30; Mass:29.97831375{34} ; Abundance:0;  StandardWeight:30.973761998{5}),
(Symbol:'P'; AtomicNumber:15; MassNumber:31; Mass:30.97376199842{70} ; Abundance:100;  StandardWeight:30.973761998{5}),
(Symbol:'P'; AtomicNumber:15; MassNumber:32; Mass:31.973907643{42} ; Abundance:0;  StandardWeight:30.973761998{5}),
(Symbol:'P'; AtomicNumber:15; MassNumber:33; Mass:32.9717257{12} ; Abundance:0;  StandardWeight:30.973761998{5}),
(Symbol:'P'; AtomicNumber:15; MassNumber:34; Mass:33.97364589{87} ; Abundance:0;  StandardWeight:30.973761998{5}),
(Symbol:'P'; AtomicNumber:15; MassNumber:35; Mass:34.9733141{20} ; Abundance:0;  StandardWeight:30.973761998{5}),
(Symbol:'P'; AtomicNumber:15; MassNumber:36; Mass:35.978260{14} ; Abundance:0;  StandardWeight:30.973761998{5}),
(Symbol:'P'; AtomicNumber:15; MassNumber:37; Mass:36.979607{41} ; Abundance:0;  StandardWeight:30.973761998{5}),
(Symbol:'P'; AtomicNumber:15; MassNumber:38; Mass:37.984252{93} ; Abundance:0;  StandardWeight:30.973761998{5}),
(Symbol:'P'; AtomicNumber:15; MassNumber:39; Mass:38.986227{98} ; Abundance:0;  StandardWeight:30.973761998{5}),
(Symbol:'P'; AtomicNumber:15; MassNumber:40; Mass:39.99133{12} ; Abundance:0;  StandardWeight:30.973761998{5}),
(Symbol:'P'; AtomicNumber:15; MassNumber:41; Mass:40.994654{86} ; Abundance:0;  StandardWeight:30.973761998{5}),
(Symbol:'P'; AtomicNumber:15; MassNumber:42; Mass:42.00108{23} ; Abundance:0;  StandardWeight:30.973761998{5}),
(Symbol:'P'; AtomicNumber:15; MassNumber:43; Mass:43.00502{40} ; Abundance:0;  StandardWeight:30.973761998{5}),
(Symbol:'P'; AtomicNumber:15; MassNumber:44; Mass:44.01121{54#} ; Abundance:0;  StandardWeight:30.973761998{5}),
(Symbol:'P'; AtomicNumber:15; MassNumber:45; Mass:45.01645{64#} ; Abundance:0;  StandardWeight:30.973761998{5}),
(Symbol:'P'; AtomicNumber:15; MassNumber:46; Mass:46.02446{75#} ; Abundance:0;  StandardWeight:30.973761998{5}),
(Symbol:'P'; AtomicNumber:15; MassNumber:47; Mass:47.03139{86#} ; Abundance:0;  StandardWeight:30.973761998{5}),
(Symbol:'S'; AtomicNumber:16; MassNumber:26; Mass:26.02907{64#} ; Abundance:0;  StandardWeight:32.065 {32.059-32.076}),
(Symbol:'S'; AtomicNumber:16; MassNumber:27; Mass:27.01828{43#} ; Abundance:0;  StandardWeight:32.065 {32.059-32.076}),
(Symbol:'S'; AtomicNumber:16; MassNumber:28; Mass:28.00437{17} ; Abundance:0;  StandardWeight:32.065 {32.059-32.076}),
(Symbol:'S'; AtomicNumber:16; MassNumber:29; Mass:28.996611{54} ; Abundance:0;  StandardWeight:32.065 {32.059-32.076}),
(Symbol:'S'; AtomicNumber:16; MassNumber:30; Mass:29.98490703{40} ; Abundance:0;  StandardWeight:32.065 {32.059-32.076}),
(Symbol:'S'; AtomicNumber:16; MassNumber:31; Mass:30.97955701{25} ; Abundance:0;  StandardWeight:32.065 {32.059-32.076}),
(Symbol:'S'; AtomicNumber:16; MassNumber:32; Mass:31.9720711744{14} ; Abundance:94.99;  StandardWeight:32.065 {32.059-32.076}),
(Symbol:'S'; AtomicNumber:16; MassNumber:33; Mass:32.9714589098{15} ; Abundance:0.75;  StandardWeight:32.065 {32.059-32.076}),
(Symbol:'S'; AtomicNumber:16; MassNumber:34; Mass:33.967867004{47} ; Abundance:4.25;  StandardWeight:32.065 {32.059-32.076}),
(Symbol:'S'; AtomicNumber:16; MassNumber:35; Mass:34.969032310{43} ; Abundance:0;  StandardWeight:32.065 {32.059-32.076}),
(Symbol:'S'; AtomicNumber:16; MassNumber:36; Mass:35.96708071{20} ; Abundance:0.01;  StandardWeight:32.065 {32.059-32.076}),
(Symbol:'S'; AtomicNumber:16; MassNumber:37; Mass:36.97112551{21} ; Abundance:0;  StandardWeight:32.065 {32.059-32.076}),
(Symbol:'S'; AtomicNumber:16; MassNumber:38; Mass:37.9711633{77} ; Abundance:0;  StandardWeight:32.065 {32.059-32.076}),
(Symbol:'S'; AtomicNumber:16; MassNumber:39; Mass:38.975134{54} ; Abundance:0;  StandardWeight:32.065 {32.059-32.076}),
(Symbol:'S'; AtomicNumber:16; MassNumber:40; Mass:39.9754826{43} ; Abundance:0;  StandardWeight:32.065 {32.059-32.076}),
(Symbol:'S'; AtomicNumber:16; MassNumber:41; Mass:40.9795935{44} ; Abundance:0;  StandardWeight:32.065 {32.059-32.076}),
(Symbol:'S'; AtomicNumber:16; MassNumber:42; Mass:41.9810651{30} ; Abundance:0;  StandardWeight:32.065 {32.059-32.076}),
(Symbol:'S'; AtomicNumber:16; MassNumber:43; Mass:42.9869076{53} ; Abundance:0;  StandardWeight:32.065 {32.059-32.076}),
(Symbol:'S'; AtomicNumber:16; MassNumber:44; Mass:43.9901188{56} ; Abundance:0;  StandardWeight:32.065 {32.059-32.076}),
(Symbol:'S'; AtomicNumber:16; MassNumber:45; Mass:44.99572{74} ; Abundance:0;  StandardWeight:32.065 {32.059-32.076}),
(Symbol:'S'; AtomicNumber:16; MassNumber:46; Mass:46.00004{54#} ; Abundance:0;  StandardWeight:32.065 {32.059-32.076}),
(Symbol:'S'; AtomicNumber:16; MassNumber:47; Mass:47.00795{54#} ; Abundance:0;  StandardWeight:32.065 {32.059-32.076}),
(Symbol:'S'; AtomicNumber:16; MassNumber:48; Mass:48.01370{64#} ; Abundance:0;  StandardWeight:32.065 {32.059-32.076}),
(Symbol:'S'; AtomicNumber:16; MassNumber:49; Mass:49.02276{72#} ; Abundance:0;  StandardWeight:32.065 {32.059-32.076}),
(Symbol:'Cl'; AtomicNumber:17; MassNumber:28; Mass:28.02954{64#} ; Abundance:0; {m} StandardWeight:35.453 {35.446-35.457}),
(Symbol:'Cl'; AtomicNumber:17; MassNumber:29; Mass:29.01478{43#} ; Abundance:0; {m} StandardWeight:35.453 {35.446-35.457}),
(Symbol:'Cl'; AtomicNumber:17; MassNumber:30; Mass:30.00477{21#} ; Abundance:0; {m} StandardWeight:35.453 {35.446-35.457}),
(Symbol:'Cl'; AtomicNumber:17; MassNumber:31; Mass:30.992414{54} ; Abundance:0; {m} StandardWeight:35.453 {35.446-35.457}),
(Symbol:'Cl'; AtomicNumber:17; MassNumber:32; Mass:31.98568464{60} ; Abundance:0; {m} StandardWeight:35.453 {35.446-35.457}),
(Symbol:'Cl'; AtomicNumber:17; MassNumber:33; Mass:32.97745199{42} ; Abundance:0; {m} StandardWeight:35.453 {35.446-35.457}),
(Symbol:'Cl'; AtomicNumber:17; MassNumber:34; Mass:33.973762485{52} ; Abundance:0; {m} StandardWeight:35.453 {35.446-35.457}),
(Symbol:'Cl'; AtomicNumber:17; MassNumber:35; Mass:34.968852682{37} ; Abundance:75.76; {m} StandardWeight:35.453 {35.446-35.457}),
(Symbol:'Cl'; AtomicNumber:17; MassNumber:36; Mass:35.968306809{38} ; Abundance:0; {m} StandardWeight:35.453 {35.446-35.457}),
(Symbol:'Cl'; AtomicNumber:17; MassNumber:37; Mass:36.965902602{55} ; Abundance:24.24; {m} StandardWeight:35.453 {35.446-35.457}),
(Symbol:'Cl'; AtomicNumber:17; MassNumber:38; Mass:37.96801044{11} ; Abundance:0; {m} StandardWeight:35.453 {35.446-35.457}),
(Symbol:'Cl'; AtomicNumber:17; MassNumber:39; Mass:38.9680082{19} ; Abundance:0; {m} StandardWeight:35.453 {35.446-35.457}),
(Symbol:'Cl'; AtomicNumber:17; MassNumber:40; Mass:39.970415{34} ; Abundance:0; {m} StandardWeight:35.453 {35.446-35.457}),
(Symbol:'Cl'; AtomicNumber:17; MassNumber:41; Mass:40.970685{74} ; Abundance:0; {m} StandardWeight:35.453 {35.446-35.457}),
(Symbol:'Cl'; AtomicNumber:17; MassNumber:42; Mass:41.97325{15} ; Abundance:0; {m} StandardWeight:35.453 {35.446-35.457}),
(Symbol:'Cl'; AtomicNumber:17; MassNumber:43; Mass:42.97389{10} ; Abundance:0; {m} StandardWeight:35.453 {35.446-35.457}),
(Symbol:'Cl'; AtomicNumber:17; MassNumber:44; Mass:43.97787{20} ; Abundance:0; {m} StandardWeight:35.453 {35.446-35.457}),
(Symbol:'Cl'; AtomicNumber:17; MassNumber:45; Mass:44.98029{11} ; Abundance:0; {m} StandardWeight:35.453 {35.446-35.457}),
(Symbol:'Cl'; AtomicNumber:17; MassNumber:46; Mass:45.98517{17} ; Abundance:0; {m} StandardWeight:35.453 {35.446-35.457}),
(Symbol:'Cl'; AtomicNumber:17; MassNumber:47; Mass:46.98916{43#} ; Abundance:0; {m} StandardWeight:35.453 {35.446-35.457}),
(Symbol:'Cl'; AtomicNumber:17; MassNumber:48; Mass:47.99564{54#} ; Abundance:0; {m} StandardWeight:35.453 {35.446-35.457}),
(Symbol:'Cl'; AtomicNumber:17; MassNumber:49; Mass:49.00123{64#} ; Abundance:0; {m} StandardWeight:35.453 {35.446-35.457}),
(Symbol:'Cl'; AtomicNumber:17; MassNumber:50; Mass:50.00905{64#} ; Abundance:0; {m} StandardWeight:35.453 {35.446-35.457}),
(Symbol:'Cl'; AtomicNumber:17; MassNumber:51; Mass:51.01554{75#} ; Abundance:0; {m} StandardWeight:35.453 {35.446-35.457}),
(Symbol:'Ar'; AtomicNumber:18; MassNumber:30; Mass:30.02307{54#} ; Abundance:0; {g,r} StandardWeight:39.948{1}),
(Symbol:'Ar'; AtomicNumber:18; MassNumber:31; Mass:31.01212{22#} ; Abundance:0; {g,r} StandardWeight:39.948{1}),
(Symbol:'Ar'; AtomicNumber:18; MassNumber:32; Mass:31.9976378{19} ; Abundance:0; {g,r} StandardWeight:39.948{1}),
(Symbol:'Ar'; AtomicNumber:18; MassNumber:33; Mass:32.98992555{43} ; Abundance:0; {g,r} StandardWeight:39.948{1}),
(Symbol:'Ar'; AtomicNumber:18; MassNumber:34; Mass:33.980270090{83} ; Abundance:0; {g,r} StandardWeight:39.948{1}),
(Symbol:'Ar'; AtomicNumber:18; MassNumber:35; Mass:34.97525759{80} ; Abundance:0; {g,r} StandardWeight:39.948{1}),
(Symbol:'Ar'; AtomicNumber:18; MassNumber:36; Mass:35.967545105{28} ; Abundance:0.3336; {g,r} StandardWeight:39.948{1}),
(Symbol:'Ar'; AtomicNumber:18; MassNumber:37; Mass:36.96677633{22} ; Abundance:0; {g,r} StandardWeight:39.948{1}),
(Symbol:'Ar'; AtomicNumber:18; MassNumber:38; Mass:37.96273211{21} ; Abundance:0.0629; {g,r} StandardWeight:39.948{1}),
(Symbol:'Ar'; AtomicNumber:18; MassNumber:39; Mass:38.9643130{54} ; Abundance:0; {g,r} StandardWeight:39.948{1}),
(Symbol:'Ar'; AtomicNumber:18; MassNumber:40; Mass:39.9623831237{24} ; Abundance:99.6035; {g,r} StandardWeight:39.948{1}),
(Symbol:'Ar'; AtomicNumber:18; MassNumber:41; Mass:40.96450057{37} ; Abundance:0; {g,r} StandardWeight:39.948{1}),
(Symbol:'Ar'; AtomicNumber:18; MassNumber:42; Mass:41.9630457{62} ; Abundance:0; {g,r} StandardWeight:39.948{1}),
(Symbol:'Ar'; AtomicNumber:18; MassNumber:43; Mass:42.9656361{57} ; Abundance:0; {g,r} StandardWeight:39.948{1}),
(Symbol:'Ar'; AtomicNumber:18; MassNumber:44; Mass:43.9649238{17} ; Abundance:0; {g,r} StandardWeight:39.948{1}),
(Symbol:'Ar'; AtomicNumber:18; MassNumber:45; Mass:44.96803973{55} ; Abundance:0; {g,r} StandardWeight:39.948{1}),
(Symbol:'Ar'; AtomicNumber:18; MassNumber:46; Mass:45.968083{44} ; Abundance:0; {g,r} StandardWeight:39.948{1}),
(Symbol:'Ar'; AtomicNumber:18; MassNumber:47; Mass:46.972935{96} ; Abundance:0; {g,r} StandardWeight:39.948{1}),
(Symbol:'Ar'; AtomicNumber:18; MassNumber:48; Mass:47.97591{32#} ; Abundance:0; {g,r} StandardWeight:39.948{1}),
(Symbol:'Ar'; AtomicNumber:18; MassNumber:49; Mass:48.98190{43#} ; Abundance:0; {g,r} StandardWeight:39.948{1}),
(Symbol:'Ar'; AtomicNumber:18; MassNumber:50; Mass:49.98613{54#} ; Abundance:0; {g,r} StandardWeight:39.948{1}),
(Symbol:'Ar'; AtomicNumber:18; MassNumber:51; Mass:50.99370{64#} ; Abundance:0; {g,r} StandardWeight:39.948{1}),
(Symbol:'Ar'; AtomicNumber:18; MassNumber:52; Mass:51.99896{64#} ; Abundance:0; {g,r} StandardWeight:39.948{1}),
(Symbol:'Ar'; AtomicNumber:18; MassNumber:53; Mass:53.00729{75#} ; Abundance:0; {g,r} StandardWeight:39.948{1}),
(Symbol:'K'; AtomicNumber:19; MassNumber:32; Mass:32.02265{54#} ; Abundance:0;  StandardWeight:39.0983{1}),
(Symbol:'K'; AtomicNumber:19; MassNumber:33; Mass:33.00756{21#} ; Abundance:0;  StandardWeight:39.0983{1}),
(Symbol:'K'; AtomicNumber:19; MassNumber:34; Mass:33.99869{32#} ; Abundance:0;  StandardWeight:39.0983{1}),
(Symbol:'K'; AtomicNumber:19; MassNumber:35; Mass:34.98800541{55} ; Abundance:0;  StandardWeight:39.0983{1}),
(Symbol:'K'; AtomicNumber:19; MassNumber:36; Mass:35.98130201{37} ; Abundance:0;  StandardWeight:39.0983{1}),
(Symbol:'K'; AtomicNumber:19; MassNumber:37; Mass:36.97337589{10} ; Abundance:0;  StandardWeight:39.0983{1}),
(Symbol:'K'; AtomicNumber:19; MassNumber:38; Mass:37.96908112{21} ; Abundance:0;  StandardWeight:39.0983{1}),
(Symbol:'K'; AtomicNumber:19; MassNumber:39; Mass:38.9637064864{49} ; Abundance:93.2581;  StandardWeight:39.0983{1}),
(Symbol:'K'; AtomicNumber:19; MassNumber:40; Mass:39.963998166{60} ; Abundance:0.0117;  StandardWeight:39.0983{1}),
(Symbol:'K'; AtomicNumber:19; MassNumber:41; Mass:40.9618252579{41} ; Abundance:6.7302;  StandardWeight:39.0983{1}),
(Symbol:'K'; AtomicNumber:19; MassNumber:42; Mass:41.96240231{11} ; Abundance:0;  StandardWeight:39.0983{1}),
(Symbol:'K'; AtomicNumber:19; MassNumber:43; Mass:42.96073470{44} ; Abundance:0;  StandardWeight:39.0983{1}),
(Symbol:'K'; AtomicNumber:19; MassNumber:44; Mass:43.96158699{45} ; Abundance:0;  StandardWeight:39.0983{1}),
(Symbol:'K'; AtomicNumber:19; MassNumber:45; Mass:44.96069149{56} ; Abundance:0;  StandardWeight:39.0983{1}),
(Symbol:'K'; AtomicNumber:19; MassNumber:46; Mass:45.96198159{78} ; Abundance:0;  StandardWeight:39.0983{1}),
(Symbol:'K'; AtomicNumber:19; MassNumber:47; Mass:46.9616616{15} ; Abundance:0;  StandardWeight:39.0983{1}),
(Symbol:'K'; AtomicNumber:19; MassNumber:48; Mass:47.96534119{83} ; Abundance:0;  StandardWeight:39.0983{1}),
(Symbol:'K'; AtomicNumber:19; MassNumber:49; Mass:48.96821075{86} ; Abundance:0;  StandardWeight:39.0983{1}),
(Symbol:'K'; AtomicNumber:19; MassNumber:50; Mass:49.9723800{83} ; Abundance:0;  StandardWeight:39.0983{1}),
(Symbol:'K'; AtomicNumber:19; MassNumber:51; Mass:50.975828{14} ; Abundance:0;  StandardWeight:39.0983{1}),
(Symbol:'K'; AtomicNumber:19; MassNumber:52; Mass:51.98224{43#} ; Abundance:0;  StandardWeight:39.0983{1}),
(Symbol:'K'; AtomicNumber:19; MassNumber:53; Mass:52.98746{54#} ; Abundance:0;  StandardWeight:39.0983{1}),
(Symbol:'K'; AtomicNumber:19; MassNumber:54; Mass:53.99463{64#} ; Abundance:0;  StandardWeight:39.0983{1}),
(Symbol:'K'; AtomicNumber:19; MassNumber:55; Mass:55.00076{75#} ; Abundance:0;  StandardWeight:39.0983{1}),
(Symbol:'K'; AtomicNumber:19; MassNumber:56; Mass:56.00851{86#} ; Abundance:0;  StandardWeight:39.0983{1}),
(Symbol:'Ca'; AtomicNumber:20; MassNumber:34; Mass:34.01487{32#} ; Abundance:0; {g} StandardWeight:40.078{4}),
(Symbol:'Ca'; AtomicNumber:20; MassNumber:35; Mass:35.00514{21#} ; Abundance:0; {g} StandardWeight:40.078{4}),
(Symbol:'Ca'; AtomicNumber:20; MassNumber:36; Mass:35.993074{43} ; Abundance:0; {g} StandardWeight:40.078{4}),
(Symbol:'Ca'; AtomicNumber:20; MassNumber:37; Mass:36.98589785{68} ; Abundance:0; {g} StandardWeight:40.078{4}),
(Symbol:'Ca'; AtomicNumber:20; MassNumber:38; Mass:37.97631922{21} ; Abundance:0; {g} StandardWeight:40.078{4}),
(Symbol:'Ca'; AtomicNumber:20; MassNumber:39; Mass:38.97071081{64} ; Abundance:0; {g} StandardWeight:40.078{4}),
(Symbol:'Ca'; AtomicNumber:20; MassNumber:40; Mass:39.962590863{22} ; Abundance:96.941; {g} StandardWeight:40.078{4}),
(Symbol:'Ca'; AtomicNumber:20; MassNumber:41; Mass:40.96227792{15} ; Abundance:0; {g} StandardWeight:40.078{4}),
(Symbol:'Ca'; AtomicNumber:20; MassNumber:42; Mass:41.95861783{16} ; Abundance:0.647; {g} StandardWeight:40.078{4}),
(Symbol:'Ca'; AtomicNumber:20; MassNumber:43; Mass:42.95876644{24} ; Abundance:0.135; {g} StandardWeight:40.078{4}),
(Symbol:'Ca'; AtomicNumber:20; MassNumber:44; Mass:43.95548156{35} ; Abundance:2.086; {g} StandardWeight:40.078{4}),
(Symbol:'Ca'; AtomicNumber:20; MassNumber:45; Mass:44.95618635{39} ; Abundance:0; {g} StandardWeight:40.078{4}),
(Symbol:'Ca'; AtomicNumber:20; MassNumber:46; Mass:45.9536890{24} ; Abundance:0.004; {g} StandardWeight:40.078{4}),
(Symbol:'Ca'; AtomicNumber:20; MassNumber:47; Mass:46.9545424{24} ; Abundance:0; {g} StandardWeight:40.078{4}),
(Symbol:'Ca'; AtomicNumber:20; MassNumber:48; Mass:47.95252276{13} ; Abundance:0.187; {g} StandardWeight:40.078{4}),
(Symbol:'Ca'; AtomicNumber:20; MassNumber:49; Mass:48.95566274{23} ; Abundance:0; {g} StandardWeight:40.078{4}),
(Symbol:'Ca'; AtomicNumber:20; MassNumber:50; Mass:49.9574992{17} ; Abundance:0; {g} StandardWeight:40.078{4}),
(Symbol:'Ca'; AtomicNumber:20; MassNumber:51; Mass:50.960989{24} ; Abundance:0; {g} StandardWeight:40.078{4}),
(Symbol:'Ca'; AtomicNumber:20; MassNumber:52; Mass:51.963217{64} ; Abundance:0; {g} StandardWeight:40.078{4}),
(Symbol:'Ca'; AtomicNumber:20; MassNumber:53; Mass:52.96945{43#} ; Abundance:0; {g} StandardWeight:40.078{4}),
(Symbol:'Ca'; AtomicNumber:20; MassNumber:54; Mass:53.97340{54#} ; Abundance:0; {g} StandardWeight:40.078{4}),
(Symbol:'Ca'; AtomicNumber:20; MassNumber:55; Mass:54.98030{54#} ; Abundance:0; {g} StandardWeight:40.078{4}),
(Symbol:'Ca'; AtomicNumber:20; MassNumber:56; Mass:55.98508{64#} ; Abundance:0; {g} StandardWeight:40.078{4}),
(Symbol:'Ca'; AtomicNumber:20; MassNumber:57; Mass:56.99262{64#} ; Abundance:0; {g} StandardWeight:40.078{4}),
(Symbol:'Ca'; AtomicNumber:20; MassNumber:58; Mass:57.99794{75#} ; Abundance:0; {g} StandardWeight:40.078{4}),
(Symbol:'Sc'; AtomicNumber:21; MassNumber:36; Mass:36.01648{32#} ; Abundance:0;  StandardWeight:44.955908{5}),
(Symbol:'Sc'; AtomicNumber:21; MassNumber:37; Mass:37.00374{32#} ; Abundance:0;  StandardWeight:44.955908{5}),
(Symbol:'Sc'; AtomicNumber:21; MassNumber:38; Mass:37.99512{21#} ; Abundance:0;  StandardWeight:44.955908{5}),
(Symbol:'Sc'; AtomicNumber:21; MassNumber:39; Mass:38.984785{26} ; Abundance:0;  StandardWeight:44.955908{5}),
(Symbol:'Sc'; AtomicNumber:21; MassNumber:40; Mass:39.9779673{30} ; Abundance:0;  StandardWeight:44.955908{5}),
(Symbol:'Sc'; AtomicNumber:21; MassNumber:41; Mass:40.969251105{88} ; Abundance:0;  StandardWeight:44.955908{5}),
(Symbol:'Sc'; AtomicNumber:21; MassNumber:42; Mass:41.96551653{18} ; Abundance:0;  StandardWeight:44.955908{5}),
(Symbol:'Sc'; AtomicNumber:21; MassNumber:43; Mass:42.9611505{20} ; Abundance:0;  StandardWeight:44.955908{5}),
(Symbol:'Sc'; AtomicNumber:21; MassNumber:44; Mass:43.9594029{19} ; Abundance:0;  StandardWeight:44.955908{5}),
(Symbol:'Sc'; AtomicNumber:21; MassNumber:45; Mass:44.95590828{77} ; Abundance:100;  StandardWeight:44.955908{5}),
(Symbol:'Sc'; AtomicNumber:21; MassNumber:46; Mass:45.95516826{78} ; Abundance:0;  StandardWeight:44.955908{5}),
(Symbol:'Sc'; AtomicNumber:21; MassNumber:47; Mass:46.9524037{21} ; Abundance:0;  StandardWeight:44.955908{5}),
(Symbol:'Sc'; AtomicNumber:21; MassNumber:48; Mass:47.9522236{53} ; Abundance:0;  StandardWeight:44.955908{5}),
(Symbol:'Sc'; AtomicNumber:21; MassNumber:49; Mass:48.9500146{29} ; Abundance:0;  StandardWeight:44.955908{5}),
(Symbol:'Sc'; AtomicNumber:21; MassNumber:50; Mass:49.952176{16} ; Abundance:0;  StandardWeight:44.955908{5}),
(Symbol:'Sc'; AtomicNumber:21; MassNumber:51; Mass:50.953592{21} ; Abundance:0;  StandardWeight:44.955908{5}),
(Symbol:'Sc'; AtomicNumber:21; MassNumber:52; Mass:51.95688{15} ; Abundance:0;  StandardWeight:44.955908{5}),
(Symbol:'Sc'; AtomicNumber:21; MassNumber:53; Mass:52.95909{29} ; Abundance:0;  StandardWeight:44.955908{5}),
(Symbol:'Sc'; AtomicNumber:21; MassNumber:54; Mass:53.96393{39} ; Abundance:0;  StandardWeight:44.955908{5}),
(Symbol:'Sc'; AtomicNumber:21; MassNumber:55; Mass:54.96782{50} ; Abundance:0;  StandardWeight:44.955908{5}),
(Symbol:'Sc'; AtomicNumber:21; MassNumber:56; Mass:55.97345{43#} ; Abundance:0;  StandardWeight:44.955908{5}),
(Symbol:'Sc'; AtomicNumber:21; MassNumber:57; Mass:56.97777{54#} ; Abundance:0;  StandardWeight:44.955908{5}),
(Symbol:'Sc'; AtomicNumber:21; MassNumber:58; Mass:57.98403{64#} ; Abundance:0;  StandardWeight:44.955908{5}),
(Symbol:'Sc'; AtomicNumber:21; MassNumber:59; Mass:58.98894{64#} ; Abundance:0;  StandardWeight:44.955908{5}),
(Symbol:'Sc'; AtomicNumber:21; MassNumber:60; Mass:59.99565{75#} ; Abundance:0;  StandardWeight:44.955908{5}),
(Symbol:'Sc'; AtomicNumber:21; MassNumber:61; Mass:61.00100{86#} ; Abundance:0;  StandardWeight:44.955908{5}),
(Symbol:'Ti'; AtomicNumber:22; MassNumber:38; Mass:38.01145{32#} ; Abundance:0;  StandardWeight:47.867{1}),
(Symbol:'Ti'; AtomicNumber:22; MassNumber:39; Mass:39.00236{22#} ; Abundance:0;  StandardWeight:47.867{1}),
(Symbol:'Ti'; AtomicNumber:22; MassNumber:40; Mass:39.99050{17} ; Abundance:0;  StandardWeight:47.867{1}),
(Symbol:'Ti'; AtomicNumber:22; MassNumber:41; Mass:40.983148{30} ; Abundance:0;  StandardWeight:47.867{1}),
(Symbol:'Ti'; AtomicNumber:22; MassNumber:42; Mass:41.97304903{30} ; Abundance:0;  StandardWeight:47.867{1}),
(Symbol:'Ti'; AtomicNumber:22; MassNumber:43; Mass:42.9685225{78} ; Abundance:0;  StandardWeight:47.867{1}),
(Symbol:'Ti'; AtomicNumber:22; MassNumber:44; Mass:43.95968995{75} ; Abundance:0;  StandardWeight:47.867{1}),
(Symbol:'Ti'; AtomicNumber:22; MassNumber:45; Mass:44.95812198{95} ; Abundance:0;  StandardWeight:47.867{1}),
(Symbol:'Ti'; AtomicNumber:22; MassNumber:46; Mass:45.95262772{35} ; Abundance:8.25;  StandardWeight:47.867{1}),
(Symbol:'Ti'; AtomicNumber:22; MassNumber:47; Mass:46.95175879{38} ; Abundance:7.44;  StandardWeight:47.867{1}),
(Symbol:'Ti'; AtomicNumber:22; MassNumber:48; Mass:47.94794198{38} ; Abundance:73.72;  StandardWeight:47.867{1}),
(Symbol:'Ti'; AtomicNumber:22; MassNumber:49; Mass:48.94786568{39} ; Abundance:5.41;  StandardWeight:47.867{1}),
(Symbol:'Ti'; AtomicNumber:22; MassNumber:50; Mass:49.94478689{39} ; Abundance:5.18;  StandardWeight:47.867{1}),
(Symbol:'Ti'; AtomicNumber:22; MassNumber:51; Mass:50.94661065{65} ; Abundance:0;  StandardWeight:47.867{1}),
(Symbol:'Ti'; AtomicNumber:22; MassNumber:52; Mass:51.9468930{76} ; Abundance:0;  StandardWeight:47.867{1}),
(Symbol:'Ti'; AtomicNumber:22; MassNumber:53; Mass:52.94973{11} ; Abundance:0;  StandardWeight:47.867{1}),
(Symbol:'Ti'; AtomicNumber:22; MassNumber:54; Mass:53.95105{13} ; Abundance:0;  StandardWeight:47.867{1}),
(Symbol:'Ti'; AtomicNumber:22; MassNumber:55; Mass:54.95527{17} ; Abundance:0;  StandardWeight:47.867{1}),
(Symbol:'Ti'; AtomicNumber:22; MassNumber:56; Mass:55.95791{15} ; Abundance:0;  StandardWeight:47.867{1}),
(Symbol:'Ti'; AtomicNumber:22; MassNumber:57; Mass:56.96364{27} ; Abundance:0;  StandardWeight:47.867{1}),
(Symbol:'Ti'; AtomicNumber:22; MassNumber:58; Mass:57.96660{43#} ; Abundance:0;  StandardWeight:47.867{1}),
(Symbol:'Ti'; AtomicNumber:22; MassNumber:59; Mass:58.97247{43#} ; Abundance:0;  StandardWeight:47.867{1}),
(Symbol:'Ti'; AtomicNumber:22; MassNumber:60; Mass:59.97603{54#} ; Abundance:0;  StandardWeight:47.867{1}),
(Symbol:'Ti'; AtomicNumber:22; MassNumber:61; Mass:60.98245{64#} ; Abundance:0;  StandardWeight:47.867{1}),
(Symbol:'Ti'; AtomicNumber:22; MassNumber:62; Mass:61.98651{75#} ; Abundance:0;  StandardWeight:47.867{1}),
(Symbol:'Ti'; AtomicNumber:22; MassNumber:63; Mass:62.99375{75#} ; Abundance:0;  StandardWeight:47.867{1}),
(Symbol:'V'; AtomicNumber:23; MassNumber:40; Mass:40.01276{43#} ; Abundance:0;  StandardWeight:50.9415{1}),
(Symbol:'V'; AtomicNumber:23; MassNumber:41; Mass:41.00021{32#} ; Abundance:0;  StandardWeight:50.9415{1}),
(Symbol:'V'; AtomicNumber:23; MassNumber:42; Mass:41.99182{32#} ; Abundance:0;  StandardWeight:50.9415{1}),
(Symbol:'V'; AtomicNumber:23; MassNumber:43; Mass:42.980766{46} ; Abundance:0;  StandardWeight:50.9415{1}),
(Symbol:'V'; AtomicNumber:23; MassNumber:44; Mass:43.97411{20} ; Abundance:0;  StandardWeight:50.9415{1}),
(Symbol:'V'; AtomicNumber:23; MassNumber:45; Mass:44.9657748{86} ; Abundance:0;  StandardWeight:50.9415{1}),
(Symbol:'V'; AtomicNumber:23; MassNumber:46; Mass:45.96019878{36} ; Abundance:0;  StandardWeight:50.9415{1}),
(Symbol:'V'; AtomicNumber:23; MassNumber:47; Mass:46.95490491{36} ; Abundance:0;  StandardWeight:50.9415{1}),
(Symbol:'V'; AtomicNumber:23; MassNumber:48; Mass:47.9522522{11} ; Abundance:0;  StandardWeight:50.9415{1}),
(Symbol:'V'; AtomicNumber:23; MassNumber:49; Mass:48.94851180{96} ; Abundance:0;  StandardWeight:50.9415{1}),
(Symbol:'V'; AtomicNumber:23; MassNumber:50; Mass:49.94715601{95} ; Abundance:0.25;  StandardWeight:50.9415{1}),
(Symbol:'V'; AtomicNumber:23; MassNumber:51; Mass:50.94395704{94} ; Abundance:99.75;  StandardWeight:50.9415{1}),
(Symbol:'V'; AtomicNumber:23; MassNumber:52; Mass:51.94477301{95} ; Abundance:0;  StandardWeight:50.9415{1}),
(Symbol:'V'; AtomicNumber:23; MassNumber:53; Mass:52.9443367{34} ; Abundance:0;  StandardWeight:50.9415{1}),
(Symbol:'V'; AtomicNumber:23; MassNumber:54; Mass:53.946439{16} ; Abundance:0;  StandardWeight:50.9415{1}),
(Symbol:'V'; AtomicNumber:23; MassNumber:55; Mass:54.94724{10} ; Abundance:0;  StandardWeight:50.9415{1}),
(Symbol:'V'; AtomicNumber:23; MassNumber:56; Mass:55.95048{19} ; Abundance:0;  StandardWeight:50.9415{1}),
(Symbol:'V'; AtomicNumber:23; MassNumber:57; Mass:56.95252{24} ; Abundance:0;  StandardWeight:50.9415{1}),
(Symbol:'V'; AtomicNumber:23; MassNumber:58; Mass:57.95672{14} ; Abundance:0;  StandardWeight:50.9415{1}),
(Symbol:'V'; AtomicNumber:23; MassNumber:59; Mass:58.95939{17} ; Abundance:0;  StandardWeight:50.9415{1}),
(Symbol:'V'; AtomicNumber:23; MassNumber:60; Mass:59.96431{24} ; Abundance:0;  StandardWeight:50.9415{1}),
(Symbol:'V'; AtomicNumber:23; MassNumber:61; Mass:60.96725{96} ; Abundance:0;  StandardWeight:50.9415{1}),
(Symbol:'V'; AtomicNumber:23; MassNumber:62; Mass:61.97265{32#} ; Abundance:0;  StandardWeight:50.9415{1}),
(Symbol:'V'; AtomicNumber:23; MassNumber:63; Mass:62.97639{43#} ; Abundance:0;  StandardWeight:50.9415{1}),
(Symbol:'V'; AtomicNumber:23; MassNumber:64; Mass:63.98264{43#} ; Abundance:0;  StandardWeight:50.9415{1}),
(Symbol:'V'; AtomicNumber:23; MassNumber:65; Mass:64.98750{54#} ; Abundance:0;  StandardWeight:50.9415{1}),
(Symbol:'V'; AtomicNumber:23; MassNumber:66; Mass:65.99398{64#} ; Abundance:0;  StandardWeight:50.9415{1}),
(Symbol:'Cr'; AtomicNumber:24; MassNumber:42; Mass:42.00670{43#} ; Abundance:0;  StandardWeight:51.9961{6}),
(Symbol:'Cr'; AtomicNumber:24; MassNumber:43; Mass:42.99753{43#} ; Abundance:0;  StandardWeight:51.9961{6}),
(Symbol:'Cr'; AtomicNumber:24; MassNumber:44; Mass:43.98536{32#} ; Abundance:0;  StandardWeight:51.9961{6}),
(Symbol:'Cr'; AtomicNumber:24; MassNumber:45; Mass:44.979050{38} ; Abundance:0;  StandardWeight:51.9961{6}),
(Symbol:'Cr'; AtomicNumber:24; MassNumber:46; Mass:45.968359{21} ; Abundance:0;  StandardWeight:51.9961{6}),
(Symbol:'Cr'; AtomicNumber:24; MassNumber:47; Mass:46.9628974{75} ; Abundance:0;  StandardWeight:51.9961{6}),
(Symbol:'Cr'; AtomicNumber:24; MassNumber:48; Mass:47.9540291{79} ; Abundance:0;  StandardWeight:51.9961{6}),
(Symbol:'Cr'; AtomicNumber:24; MassNumber:49; Mass:48.9513333{25} ; Abundance:0;  StandardWeight:51.9961{6}),
(Symbol:'Cr'; AtomicNumber:24; MassNumber:50; Mass:49.94604183{94} ; Abundance:4.345;  StandardWeight:51.9961{6}),
(Symbol:'Cr'; AtomicNumber:24; MassNumber:51; Mass:50.94476502{94} ; Abundance:0;  StandardWeight:51.9961{6}),
(Symbol:'Cr'; AtomicNumber:24; MassNumber:52; Mass:51.94050623{63} ; Abundance:83.789;  StandardWeight:51.9961{6}),
(Symbol:'Cr'; AtomicNumber:24; MassNumber:53; Mass:52.94064815{62} ; Abundance:9.501;  StandardWeight:51.9961{6}),
(Symbol:'Cr'; AtomicNumber:24; MassNumber:54; Mass:53.93887916{61} ; Abundance:2.365;  StandardWeight:51.9961{6}),
(Symbol:'Cr'; AtomicNumber:24; MassNumber:55; Mass:54.94083843{64} ; Abundance:0;  StandardWeight:51.9961{6}),
(Symbol:'Cr'; AtomicNumber:24; MassNumber:56; Mass:55.9406531{20} ; Abundance:0;  StandardWeight:51.9961{6}),
(Symbol:'Cr'; AtomicNumber:24; MassNumber:57; Mass:56.9436130{20} ; Abundance:0;  StandardWeight:51.9961{6}),
(Symbol:'Cr'; AtomicNumber:24; MassNumber:58; Mass:57.94435{22} ; Abundance:0;  StandardWeight:51.9961{6}),
(Symbol:'Cr'; AtomicNumber:24; MassNumber:59; Mass:58.94859{26} ; Abundance:0;  StandardWeight:51.9961{6}),
(Symbol:'Cr'; AtomicNumber:24; MassNumber:60; Mass:59.95008{23} ; Abundance:0;  StandardWeight:51.9961{6}),
(Symbol:'Cr'; AtomicNumber:24; MassNumber:61; Mass:60.95442{14} ; Abundance:0;  StandardWeight:51.9961{6}),
(Symbol:'Cr'; AtomicNumber:24; MassNumber:62; Mass:61.95610{16} ; Abundance:0;  StandardWeight:51.9961{6}),
(Symbol:'Cr'; AtomicNumber:24; MassNumber:63; Mass:62.96165{49} ; Abundance:0;  StandardWeight:51.9961{6}),
(Symbol:'Cr'; AtomicNumber:24; MassNumber:64; Mass:63.96408{32#} ; Abundance:0;  StandardWeight:51.9961{6}),
(Symbol:'Cr'; AtomicNumber:24; MassNumber:65; Mass:64.96996{32#} ; Abundance:0;  StandardWeight:51.9961{6}),
(Symbol:'Cr'; AtomicNumber:24; MassNumber:66; Mass:65.97366{54#} ; Abundance:0;  StandardWeight:51.9961{6}),
(Symbol:'Cr'; AtomicNumber:24; MassNumber:67; Mass:66.98016{54#} ; Abundance:0;  StandardWeight:51.9961{6}),
(Symbol:'Cr'; AtomicNumber:24; MassNumber:68; Mass:67.98403{75#} ; Abundance:0;  StandardWeight:51.9961{6}),
(Symbol:'Mn'; AtomicNumber:25; MassNumber:44; Mass:44.00715{54#} ; Abundance:0;  StandardWeight:54.938044{3}),
(Symbol:'Mn'; AtomicNumber:25; MassNumber:45; Mass:44.99449{43#} ; Abundance:0;  StandardWeight:54.938044{3}),
(Symbol:'Mn'; AtomicNumber:25; MassNumber:46; Mass:45.98609{43#} ; Abundance:0;  StandardWeight:54.938044{3}),
(Symbol:'Mn'; AtomicNumber:25; MassNumber:47; Mass:46.975775{34} ; Abundance:0;  StandardWeight:54.938044{3}),
(Symbol:'Mn'; AtomicNumber:25; MassNumber:48; Mass:47.96852{18} ; Abundance:0;  StandardWeight:54.938044{3}),
(Symbol:'Mn'; AtomicNumber:25; MassNumber:49; Mass:48.959595{11} ; Abundance:0;  StandardWeight:54.938044{3}),
(Symbol:'Mn'; AtomicNumber:25; MassNumber:50; Mass:49.95423778{95} ; Abundance:0;  StandardWeight:54.938044{3}),
(Symbol:'Mn'; AtomicNumber:25; MassNumber:51; Mass:50.94820847{94} ; Abundance:0;  StandardWeight:54.938044{3}),
(Symbol:'Mn'; AtomicNumber:25; MassNumber:52; Mass:51.9455639{20} ; Abundance:0;  StandardWeight:54.938044{3}),
(Symbol:'Mn'; AtomicNumber:25; MassNumber:53; Mass:52.94128889{68} ; Abundance:0;  StandardWeight:54.938044{3}),
(Symbol:'Mn'; AtomicNumber:25; MassNumber:54; Mass:53.9403576{12} ; Abundance:0;  StandardWeight:54.938044{3}),
(Symbol:'Mn'; AtomicNumber:25; MassNumber:55; Mass:54.93804391{48} ; Abundance:100;  StandardWeight:54.938044{3}),
(Symbol:'Mn'; AtomicNumber:25; MassNumber:56; Mass:55.93890369{49} ; Abundance:0;  StandardWeight:54.938044{3}),
(Symbol:'Mn'; AtomicNumber:25; MassNumber:57; Mass:56.9382861{16} ; Abundance:0;  StandardWeight:54.938044{3}),
(Symbol:'Mn'; AtomicNumber:25; MassNumber:58; Mass:57.9400666{29} ; Abundance:0;  StandardWeight:54.938044{3}),
(Symbol:'Mn'; AtomicNumber:25; MassNumber:59; Mass:58.9403911{25} ; Abundance:0;  StandardWeight:54.938044{3}),
(Symbol:'Mn'; AtomicNumber:25; MassNumber:60; Mass:59.9431366{25} ; Abundance:0;  StandardWeight:54.938044{3}),
(Symbol:'Mn'; AtomicNumber:25; MassNumber:61; Mass:60.9444525{25} ; Abundance:0;  StandardWeight:54.938044{3}),
(Symbol:'Mn'; AtomicNumber:25; MassNumber:62; Mass:61.94795{16#} ; Abundance:0;  StandardWeight:54.938044{3}),
(Symbol:'Mn'; AtomicNumber:25; MassNumber:63; Mass:62.9496647{40} ; Abundance:0;  StandardWeight:54.938044{3}),
(Symbol:'Mn'; AtomicNumber:25; MassNumber:64; Mass:63.9538494{38} ; Abundance:0;  StandardWeight:54.938044{3}),
(Symbol:'Mn'; AtomicNumber:25; MassNumber:65; Mass:64.9560198{40} ; Abundance:0;  StandardWeight:54.938044{3}),
(Symbol:'Mn'; AtomicNumber:25; MassNumber:66; Mass:65.960547{12} ; Abundance:0;  StandardWeight:54.938044{3}),
(Symbol:'Mn'; AtomicNumber:25; MassNumber:67; Mass:66.96424{43#} ; Abundance:0;  StandardWeight:54.938044{3}),
(Symbol:'Mn'; AtomicNumber:25; MassNumber:68; Mass:67.96962{54#} ; Abundance:0;  StandardWeight:54.938044{3}),
(Symbol:'Mn'; AtomicNumber:25; MassNumber:69; Mass:68.97366{64#} ; Abundance:0;  StandardWeight:54.938044{3}),
(Symbol:'Mn'; AtomicNumber:25; MassNumber:70; Mass:69.97937{75#} ; Abundance:0;  StandardWeight:54.938044{3}),
(Symbol:'Mn'; AtomicNumber:25; MassNumber:71; Mass:70.98368{75#} ; Abundance:0;  StandardWeight:54.938044{3}),
(Symbol:'Fe'; AtomicNumber:26; MassNumber:45; Mass:45.01442{43#} ; Abundance:0;  StandardWeight:55.845{2}),
(Symbol:'Fe'; AtomicNumber:26; MassNumber:46; Mass:46.00063{54#} ; Abundance:0;  StandardWeight:55.845{2}),
(Symbol:'Fe'; AtomicNumber:26; MassNumber:47; Mass:46.99185{54#} ; Abundance:0;  StandardWeight:55.845{2}),
(Symbol:'Fe'; AtomicNumber:26; MassNumber:48; Mass:47.98023{43#} ; Abundance:0;  StandardWeight:55.845{2}),
(Symbol:'Fe'; AtomicNumber:26; MassNumber:49; Mass:48.973429{26} ; Abundance:0;  StandardWeight:55.845{2}),
(Symbol:'Fe'; AtomicNumber:26; MassNumber:50; Mass:49.962975{64} ; Abundance:0;  StandardWeight:55.845{2}),
(Symbol:'Fe'; AtomicNumber:26; MassNumber:51; Mass:50.9568410{96} ; Abundance:0;  StandardWeight:55.845{2}),
(Symbol:'Fe'; AtomicNumber:26; MassNumber:52; Mass:51.9481131{70} ; Abundance:0;  StandardWeight:55.845{2}),
(Symbol:'Fe'; AtomicNumber:26; MassNumber:53; Mass:52.9453064{18} ; Abundance:0;  StandardWeight:55.845{2}),
(Symbol:'Fe'; AtomicNumber:26; MassNumber:54; Mass:53.93960899{53} ; Abundance:5.845;  StandardWeight:55.845{2}),
(Symbol:'Fe'; AtomicNumber:26; MassNumber:55; Mass:54.93829199{51} ; Abundance:0;  StandardWeight:55.845{2}),
(Symbol:'Fe'; AtomicNumber:26; MassNumber:56; Mass:55.93493633{49} ; Abundance:91.754;  StandardWeight:55.845{2}),
(Symbol:'Fe'; AtomicNumber:26; MassNumber:57; Mass:56.93539284{49} ; Abundance:2.119;  StandardWeight:55.845{2}),
(Symbol:'Fe'; AtomicNumber:26; MassNumber:58; Mass:57.93327443{53} ; Abundance:0.282;  StandardWeight:55.845{2}),
(Symbol:'Fe'; AtomicNumber:26; MassNumber:59; Mass:58.93487434{54} ; Abundance:0;  StandardWeight:55.845{2}),
(Symbol:'Fe'; AtomicNumber:26; MassNumber:60; Mass:59.9340711{37} ; Abundance:0;  StandardWeight:55.845{2}),
(Symbol:'Fe'; AtomicNumber:26; MassNumber:61; Mass:60.9367462{28} ; Abundance:0;  StandardWeight:55.845{2}),
(Symbol:'Fe'; AtomicNumber:26; MassNumber:62; Mass:61.9367918{30} ; Abundance:0;  StandardWeight:55.845{2}),
(Symbol:'Fe'; AtomicNumber:26; MassNumber:63; Mass:62.9402727{46} ; Abundance:0;  StandardWeight:55.845{2}),
(Symbol:'Fe'; AtomicNumber:26; MassNumber:64; Mass:63.9409878{54} ; Abundance:0;  StandardWeight:55.845{2}),
(Symbol:'Fe'; AtomicNumber:26; MassNumber:65; Mass:64.9450115{73} ; Abundance:0;  StandardWeight:55.845{2}),
(Symbol:'Fe'; AtomicNumber:26; MassNumber:66; Mass:65.9462500{44} ; Abundance:0;  StandardWeight:55.845{2}),
(Symbol:'Fe'; AtomicNumber:26; MassNumber:67; Mass:66.95054{23} ; Abundance:0;  StandardWeight:55.845{2}),
(Symbol:'Fe'; AtomicNumber:26; MassNumber:68; Mass:67.95295{39} ; Abundance:0;  StandardWeight:55.845{2}),
(Symbol:'Fe'; AtomicNumber:26; MassNumber:69; Mass:68.95807{43#} ; Abundance:0;  StandardWeight:55.845{2}),
(Symbol:'Fe'; AtomicNumber:26; MassNumber:70; Mass:69.96102{54#} ; Abundance:0;  StandardWeight:55.845{2}),
(Symbol:'Fe'; AtomicNumber:26; MassNumber:71; Mass:70.96672{64#} ; Abundance:0;  StandardWeight:55.845{2}),
(Symbol:'Fe'; AtomicNumber:26; MassNumber:72; Mass:71.96983{75#} ; Abundance:0;  StandardWeight:55.845{2}),
(Symbol:'Fe'; AtomicNumber:26; MassNumber:73; Mass:72.97572{75#} ; Abundance:0;  StandardWeight:55.845{2}),
(Symbol:'Fe'; AtomicNumber:26; MassNumber:74; Mass:73.97935{86#} ; Abundance:0;  StandardWeight:55.845{2}),
(Symbol:'Co'; AtomicNumber:27; MassNumber:47; Mass:47.01057{86#} ; Abundance:0;  StandardWeight:58.933194{4}),
(Symbol:'Co'; AtomicNumber:27; MassNumber:48; Mass:48.00093{86#} ; Abundance:0;  StandardWeight:58.933194{4}),
(Symbol:'Co'; AtomicNumber:27; MassNumber:49; Mass:48.98891{75#} ; Abundance:0;  StandardWeight:58.933194{4}),
(Symbol:'Co'; AtomicNumber:27; MassNumber:50; Mass:49.98091{64#} ; Abundance:0;  StandardWeight:58.933194{4}),
(Symbol:'Co'; AtomicNumber:27; MassNumber:51; Mass:50.970647{52} ; Abundance:0;  StandardWeight:58.933194{4}),
(Symbol:'Co'; AtomicNumber:27; MassNumber:52; Mass:51.96351{21#} ; Abundance:0;  StandardWeight:58.933194{4}),
(Symbol:'Co'; AtomicNumber:27; MassNumber:53; Mass:52.9542041{19} ; Abundance:0;  StandardWeight:58.933194{4}),
(Symbol:'Co'; AtomicNumber:27; MassNumber:54; Mass:53.94845987{54} ; Abundance:0;  StandardWeight:58.933194{4}),
(Symbol:'Co'; AtomicNumber:27; MassNumber:55; Mass:54.94199720{57} ; Abundance:0;  StandardWeight:58.933194{4}),
(Symbol:'Co'; AtomicNumber:27; MassNumber:56; Mass:55.93983880{63} ; Abundance:0;  StandardWeight:58.933194{4}),
(Symbol:'Co'; AtomicNumber:27; MassNumber:57; Mass:56.93629057{66} ; Abundance:0;  StandardWeight:58.933194{4}),
(Symbol:'Co'; AtomicNumber:27; MassNumber:58; Mass:57.9357521{13} ; Abundance:0;  StandardWeight:58.933194{4}),
(Symbol:'Co'; AtomicNumber:27; MassNumber:59; Mass:58.93319429{56} ; Abundance:100;  StandardWeight:58.933194{4}),
(Symbol:'Co'; AtomicNumber:27; MassNumber:60; Mass:59.93381630{56} ; Abundance:0;  StandardWeight:58.933194{4}),
(Symbol:'Co'; AtomicNumber:27; MassNumber:61; Mass:60.93247662{95} ; Abundance:0;  StandardWeight:58.933194{4}),
(Symbol:'Co'; AtomicNumber:27; MassNumber:62; Mass:61.934059{20} ; Abundance:0;  StandardWeight:58.933194{4}),
(Symbol:'Co'; AtomicNumber:27; MassNumber:63; Mass:62.933600{20} ; Abundance:0;  StandardWeight:58.933194{4}),
(Symbol:'Co'; AtomicNumber:27; MassNumber:64; Mass:63.935811{21} ; Abundance:0;  StandardWeight:58.933194{4}),
(Symbol:'Co'; AtomicNumber:27; MassNumber:65; Mass:64.9364621{22} ; Abundance:0;  StandardWeight:58.933194{4}),
(Symbol:'Co'; AtomicNumber:27; MassNumber:66; Mass:65.939443{15} ; Abundance:0;  StandardWeight:58.933194{4}),
(Symbol:'Co'; AtomicNumber:27; MassNumber:67; Mass:66.9406096{69} ; Abundance:0;  StandardWeight:58.933194{4}),
(Symbol:'Co'; AtomicNumber:27; MassNumber:68; Mass:67.94426{16} ; Abundance:0;  StandardWeight:58.933194{4}),
(Symbol:'Co'; AtomicNumber:27; MassNumber:69; Mass:68.94614{20} ; Abundance:0;  StandardWeight:58.933194{4}),
(Symbol:'Co'; AtomicNumber:27; MassNumber:70; Mass:69.94963{32} ; Abundance:0;  StandardWeight:58.933194{4}),
(Symbol:'Co'; AtomicNumber:27; MassNumber:71; Mass:70.95237{50} ; Abundance:0;  StandardWeight:58.933194{4}),
(Symbol:'Co'; AtomicNumber:27; MassNumber:72; Mass:71.95729{43#} ; Abundance:0;  StandardWeight:58.933194{4}),
(Symbol:'Co'; AtomicNumber:27; MassNumber:73; Mass:72.96039{54#} ; Abundance:0;  StandardWeight:58.933194{4}),
(Symbol:'Co'; AtomicNumber:27; MassNumber:74; Mass:73.96515{64#} ; Abundance:0;  StandardWeight:58.933194{4}),
(Symbol:'Co'; AtomicNumber:27; MassNumber:75; Mass:74.96876{75#} ; Abundance:0;  StandardWeight:58.933194{4}),
(Symbol:'Co'; AtomicNumber:27; MassNumber:76; Mass:75.97413{86#} ; Abundance:0;  StandardWeight:58.933194{4}),
(Symbol:'Ni'; AtomicNumber:28; MassNumber:48; Mass:48.01769{54#} ; Abundance:0; {r} StandardWeight:58.6934{4}),
(Symbol:'Ni'; AtomicNumber:28; MassNumber:49; Mass:49.00770{86#} ; Abundance:0; {r} StandardWeight:58.6934{4}),
(Symbol:'Ni'; AtomicNumber:28; MassNumber:50; Mass:49.99474{86#} ; Abundance:0; {r} StandardWeight:58.6934{4}),
(Symbol:'Ni'; AtomicNumber:28; MassNumber:51; Mass:50.98611{86#} ; Abundance:0; {r} StandardWeight:58.6934{4}),
(Symbol:'Ni'; AtomicNumber:28; MassNumber:52; Mass:51.97480{75#} ; Abundance:0; {r} StandardWeight:58.6934{4}),
(Symbol:'Ni'; AtomicNumber:28; MassNumber:53; Mass:52.968190{27} ; Abundance:0; {r} StandardWeight:58.6934{4}),
(Symbol:'Ni'; AtomicNumber:28; MassNumber:54; Mass:53.957892{54} ; Abundance:0; {r} StandardWeight:58.6934{4}),
(Symbol:'Ni'; AtomicNumber:28; MassNumber:55; Mass:54.95133063{85} ; Abundance:0; {r} StandardWeight:58.6934{4}),
(Symbol:'Ni'; AtomicNumber:28; MassNumber:56; Mass:55.94212855{57} ; Abundance:0; {r} StandardWeight:58.6934{4}),
(Symbol:'Ni'; AtomicNumber:28; MassNumber:57; Mass:56.93979218{71} ; Abundance:0; {r} StandardWeight:58.6934{4}),
(Symbol:'Ni'; AtomicNumber:28; MassNumber:58; Mass:57.93534241{52} ; Abundance:68.077; {r} StandardWeight:58.6934{4}),
(Symbol:'Ni'; AtomicNumber:28; MassNumber:59; Mass:58.93434620{52} ; Abundance:0; {r} StandardWeight:58.6934{4}),
(Symbol:'Ni'; AtomicNumber:28; MassNumber:60; Mass:59.93078588{52} ; Abundance:26.223; {r} StandardWeight:58.6934{4}),
(Symbol:'Ni'; AtomicNumber:28; MassNumber:61; Mass:60.93105557{52} ; Abundance:1.1399; {r} StandardWeight:58.6934{4}),
(Symbol:'Ni'; AtomicNumber:28; MassNumber:62; Mass:61.92834537{55} ; Abundance:3.6346; {r} StandardWeight:58.6934{4}),
(Symbol:'Ni'; AtomicNumber:28; MassNumber:63; Mass:62.92966963{56} ; Abundance:0; {r} StandardWeight:58.6934{4}),
(Symbol:'Ni'; AtomicNumber:28; MassNumber:64; Mass:63.92796682{58} ; Abundance:0.9255; {r} StandardWeight:58.6934{4}),
(Symbol:'Ni'; AtomicNumber:28; MassNumber:65; Mass:64.93008517{60} ; Abundance:0; {r} StandardWeight:58.6934{4}),
(Symbol:'Ni'; AtomicNumber:28; MassNumber:66; Mass:65.9291393{15} ; Abundance:0; {r} StandardWeight:58.6934{4}),
(Symbol:'Ni'; AtomicNumber:28; MassNumber:67; Mass:66.9315694{31} ; Abundance:0; {r} StandardWeight:58.6934{4}),
(Symbol:'Ni'; AtomicNumber:28; MassNumber:68; Mass:67.9318688{32} ; Abundance:0; {r} StandardWeight:58.6934{4}),
(Symbol:'Ni'; AtomicNumber:28; MassNumber:69; Mass:68.9356103{40} ; Abundance:0; {r} StandardWeight:58.6934{4}),
(Symbol:'Ni'; AtomicNumber:28; MassNumber:70; Mass:69.9364313{23} ; Abundance:0; {r} StandardWeight:58.6934{4}),
(Symbol:'Ni'; AtomicNumber:28; MassNumber:71; Mass:70.9405190{24} ; Abundance:0; {r} StandardWeight:58.6934{4}),
(Symbol:'Ni'; AtomicNumber:28; MassNumber:72; Mass:71.9417859{24} ; Abundance:0; {r} StandardWeight:58.6934{4}),
(Symbol:'Ni'; AtomicNumber:28; MassNumber:73; Mass:72.9462067{26} ; Abundance:0; {r} StandardWeight:58.6934{4}),
(Symbol:'Ni'; AtomicNumber:28; MassNumber:74; Mass:73.94798{43#} ; Abundance:0; {r} StandardWeight:58.6934{4}),
(Symbol:'Ni'; AtomicNumber:28; MassNumber:75; Mass:74.95250{32#} ; Abundance:0; {r} StandardWeight:58.6934{4}),
(Symbol:'Ni'; AtomicNumber:28; MassNumber:76; Mass:75.95533{54#} ; Abundance:0; {r} StandardWeight:58.6934{4}),
(Symbol:'Ni'; AtomicNumber:28; MassNumber:77; Mass:76.96055{54#} ; Abundance:0; {r} StandardWeight:58.6934{4}),
(Symbol:'Ni'; AtomicNumber:28; MassNumber:78; Mass:77.96336{86#} ; Abundance:0; {r} StandardWeight:58.6934{4}),
(Symbol:'Ni'; AtomicNumber:28; MassNumber:79; Mass:78.97025{86#} ; Abundance:0; {r} StandardWeight:58.6934{4}),
(Symbol:'Cu'; AtomicNumber:29; MassNumber:52; Mass:51.99671{86#} ; Abundance:0; {r} StandardWeight:63.546{3}),
(Symbol:'Cu'; AtomicNumber:29; MassNumber:53; Mass:52.98459{86#} ; Abundance:0; {r} StandardWeight:63.546{3}),
(Symbol:'Cu'; AtomicNumber:29; MassNumber:54; Mass:53.97666{54#} ; Abundance:0; {r} StandardWeight:63.546{3}),
(Symbol:'Cu'; AtomicNumber:29; MassNumber:55; Mass:54.96604{17} ; Abundance:0; {r} StandardWeight:63.546{3}),
(Symbol:'Cu'; AtomicNumber:29; MassNumber:56; Mass:55.95895{21#} ; Abundance:0; {r} StandardWeight:63.546{3}),
(Symbol:'Cu'; AtomicNumber:29; MassNumber:57; Mass:56.94921250{66} ; Abundance:0; {r} StandardWeight:63.546{3}),
(Symbol:'Cu'; AtomicNumber:29; MassNumber:58; Mass:57.94453305{70} ; Abundance:0; {r} StandardWeight:63.546{3}),
(Symbol:'Cu'; AtomicNumber:29; MassNumber:59; Mass:58.93949748{67} ; Abundance:0; {r} StandardWeight:63.546{3}),
(Symbol:'Cu'; AtomicNumber:29; MassNumber:60; Mass:59.9373645{18} ; Abundance:0; {r} StandardWeight:63.546{3}),
(Symbol:'Cu'; AtomicNumber:29; MassNumber:61; Mass:60.9334576{10} ; Abundance:0; {r} StandardWeight:63.546{3}),
(Symbol:'Cu'; AtomicNumber:29; MassNumber:62; Mass:61.93259541{75} ; Abundance:0; {r} StandardWeight:63.546{3}),
(Symbol:'Cu'; AtomicNumber:29; MassNumber:63; Mass:62.92959772{56} ; Abundance:69.15; {r} StandardWeight:63.546{3}),
(Symbol:'Cu'; AtomicNumber:29; MassNumber:64; Mass:63.92976434{56} ; Abundance:0; {r} StandardWeight:63.546{3}),
(Symbol:'Cu'; AtomicNumber:29; MassNumber:65; Mass:64.92778970{71} ; Abundance:30.85; {r} StandardWeight:63.546{3}),
(Symbol:'Cu'; AtomicNumber:29; MassNumber:66; Mass:65.92886903{72} ; Abundance:0; {r} StandardWeight:63.546{3}),
(Symbol:'Cu'; AtomicNumber:29; MassNumber:67; Mass:66.9277303{13} ; Abundance:0; {r} StandardWeight:63.546{3}),
(Symbol:'Cu'; AtomicNumber:29; MassNumber:68; Mass:67.9296109{17} ; Abundance:0; {r} StandardWeight:63.546{3}),
(Symbol:'Cu'; AtomicNumber:29; MassNumber:69; Mass:68.9294293{15} ; Abundance:0; {r} StandardWeight:63.546{3}),
(Symbol:'Cu'; AtomicNumber:29; MassNumber:70; Mass:69.9323921{12} ; Abundance:0; {r} StandardWeight:63.546{3}),
(Symbol:'Cu'; AtomicNumber:29; MassNumber:71; Mass:70.9326768{16} ; Abundance:0; {r} StandardWeight:63.546{3}),
(Symbol:'Cu'; AtomicNumber:29; MassNumber:72; Mass:71.9358203{15} ; Abundance:0; {r} StandardWeight:63.546{3}),
(Symbol:'Cu'; AtomicNumber:29; MassNumber:73; Mass:72.9366744{21} ; Abundance:0; {r} StandardWeight:63.546{3}),
(Symbol:'Cu'; AtomicNumber:29; MassNumber:74; Mass:73.9398749{66} ; Abundance:0; {r} StandardWeight:63.546{3}),
(Symbol:'Cu'; AtomicNumber:29; MassNumber:75; Mass:74.9415226{25} ; Abundance:0; {r} StandardWeight:63.546{3}),
(Symbol:'Cu'; AtomicNumber:29; MassNumber:76; Mass:75.9452750{72} ; Abundance:0; {r} StandardWeight:63.546{3}),
(Symbol:'Cu'; AtomicNumber:29; MassNumber:77; Mass:76.94792{16#} ; Abundance:0; {r} StandardWeight:63.546{3}),
(Symbol:'Cu'; AtomicNumber:29; MassNumber:78; Mass:77.95223{54} ; Abundance:0; {r} StandardWeight:63.546{3}),
(Symbol:'Cu'; AtomicNumber:29; MassNumber:79; Mass:78.95502{43#} ; Abundance:0; {r} StandardWeight:63.546{3}),
(Symbol:'Cu'; AtomicNumber:29; MassNumber:80; Mass:79.96089{64#} ; Abundance:0; {r} StandardWeight:63.546{3}),
(Symbol:'Cu'; AtomicNumber:29; MassNumber:81; Mass:80.96587{86#} ; Abundance:0; {r} StandardWeight:63.546{3}),
(Symbol:'Cu'; AtomicNumber:29; MassNumber:82; Mass:81.97244{86#} ; Abundance:0; {r} StandardWeight:63.546{3}),
(Symbol:'Zn'; AtomicNumber:30; MassNumber:54; Mass:53.99204{75#} ; Abundance:0; {r} StandardWeight:65.38{2}),
(Symbol:'Zn'; AtomicNumber:30; MassNumber:55; Mass:54.98398{75#} ; Abundance:0; {r} StandardWeight:65.38{2}),
(Symbol:'Zn'; AtomicNumber:30; MassNumber:56; Mass:55.97254{54#} ; Abundance:0; {r} StandardWeight:65.38{2}),
(Symbol:'Zn'; AtomicNumber:30; MassNumber:57; Mass:56.96506{22#} ; Abundance:0; {r} StandardWeight:65.38{2}),
(Symbol:'Zn'; AtomicNumber:30; MassNumber:58; Mass:57.954591{54} ; Abundance:0; {r} StandardWeight:65.38{2}),
(Symbol:'Zn'; AtomicNumber:30; MassNumber:59; Mass:58.94931266{89} ; Abundance:0; {r} StandardWeight:65.38{2}),
(Symbol:'Zn'; AtomicNumber:30; MassNumber:60; Mass:59.94184210{69} ; Abundance:0; {r} StandardWeight:65.38{2}),
(Symbol:'Zn'; AtomicNumber:30; MassNumber:61; Mass:60.939507{17} ; Abundance:0; {r} StandardWeight:65.38{2}),
(Symbol:'Zn'; AtomicNumber:30; MassNumber:62; Mass:61.93433397{73} ; Abundance:0; {r} StandardWeight:65.38{2}),
(Symbol:'Zn'; AtomicNumber:30; MassNumber:63; Mass:62.9332115{17} ; Abundance:0; {r} StandardWeight:65.38{2}),
(Symbol:'Zn'; AtomicNumber:30; MassNumber:64; Mass:63.92914201{71} ; Abundance:49.17; {r} StandardWeight:65.38{2}),
(Symbol:'Zn'; AtomicNumber:30; MassNumber:65; Mass:64.92924077{71} ; Abundance:0; {r} StandardWeight:65.38{2}),
(Symbol:'Zn'; AtomicNumber:30; MassNumber:66; Mass:65.92603381{94} ; Abundance:27.73; {r} StandardWeight:65.38{2}),
(Symbol:'Zn'; AtomicNumber:30; MassNumber:67; Mass:66.92712775{96} ; Abundance:4.04; {r} StandardWeight:65.38{2}),
(Symbol:'Zn'; AtomicNumber:30; MassNumber:68; Mass:67.92484455{98} ; Abundance:18.45; {r} StandardWeight:65.38{2}),
(Symbol:'Zn'; AtomicNumber:30; MassNumber:69; Mass:68.9265507{10} ; Abundance:0; {r} StandardWeight:65.38{2}),
(Symbol:'Zn'; AtomicNumber:30; MassNumber:70; Mass:69.9253192{21} ; Abundance:0.61; {r} StandardWeight:65.38{2}),
(Symbol:'Zn'; AtomicNumber:30; MassNumber:71; Mass:70.9277196{28} ; Abundance:0; {r} StandardWeight:65.38{2}),
(Symbol:'Zn'; AtomicNumber:30; MassNumber:72; Mass:71.9268428{23} ; Abundance:0; {r} StandardWeight:65.38{2}),
(Symbol:'Zn'; AtomicNumber:30; MassNumber:73; Mass:72.9295826{20} ; Abundance:0; {r} StandardWeight:65.38{2}),
(Symbol:'Zn'; AtomicNumber:30; MassNumber:74; Mass:73.9294073{27} ; Abundance:0; {r} StandardWeight:65.38{2}),
(Symbol:'Zn'; AtomicNumber:30; MassNumber:75; Mass:74.9328402{21} ; Abundance:0; {r} StandardWeight:65.38{2}),
(Symbol:'Zn'; AtomicNumber:30; MassNumber:76; Mass:75.9331150{16} ; Abundance:0; {r} StandardWeight:65.38{2}),
(Symbol:'Zn'; AtomicNumber:30; MassNumber:77; Mass:76.9368872{21} ; Abundance:0; {r} StandardWeight:65.38{2}),
(Symbol:'Zn'; AtomicNumber:30; MassNumber:78; Mass:77.9382892{21} ; Abundance:0; {r} StandardWeight:65.38{2}),
(Symbol:'Zn'; AtomicNumber:30; MassNumber:79; Mass:78.9426381{24} ; Abundance:0; {r} StandardWeight:65.38{2}),
(Symbol:'Zn'; AtomicNumber:30; MassNumber:80; Mass:79.9445529{28} ; Abundance:0; {r} StandardWeight:65.38{2}),
(Symbol:'Zn'; AtomicNumber:30; MassNumber:81; Mass:80.9504026{54} ; Abundance:0; {r} StandardWeight:65.38{2}),
(Symbol:'Zn'; AtomicNumber:30; MassNumber:82; Mass:81.95426{32#} ; Abundance:0; {r} StandardWeight:65.38{2}),
(Symbol:'Zn'; AtomicNumber:30; MassNumber:83; Mass:82.96056{54#} ; Abundance:0; {r} StandardWeight:65.38{2}),
(Symbol:'Zn'; AtomicNumber:30; MassNumber:84; Mass:83.96521{64#} ; Abundance:0; {r} StandardWeight:65.38{2}),
(Symbol:'Zn'; AtomicNumber:30; MassNumber:85; Mass:84.97226{75#} ; Abundance:0; {r} StandardWeight:65.38{2}),
(Symbol:'Ga'; AtomicNumber:31; MassNumber:56; Mass:55.99536{64#} ; Abundance:0;  StandardWeight:69.723{1}),
(Symbol:'Ga'; AtomicNumber:31; MassNumber:57; Mass:56.98320{32#} ; Abundance:0;  StandardWeight:69.723{1}),
(Symbol:'Ga'; AtomicNumber:31; MassNumber:58; Mass:57.97478{21#} ; Abundance:0;  StandardWeight:69.723{1}),
(Symbol:'Ga'; AtomicNumber:31; MassNumber:59; Mass:58.96353{18#} ; Abundance:0;  StandardWeight:69.723{1}),
(Symbol:'Ga'; AtomicNumber:31; MassNumber:60; Mass:59.95729{21#} ; Abundance:0;  StandardWeight:69.723{1}),
(Symbol:'Ga'; AtomicNumber:31; MassNumber:61; Mass:60.949399{41} ; Abundance:0;  StandardWeight:69.723{1}),
(Symbol:'Ga'; AtomicNumber:31; MassNumber:62; Mass:61.94419025{75} ; Abundance:0;  StandardWeight:69.723{1}),
(Symbol:'Ga'; AtomicNumber:31; MassNumber:63; Mass:62.9392942{14} ; Abundance:0;  StandardWeight:69.723{1}),
(Symbol:'Ga'; AtomicNumber:31; MassNumber:64; Mass:63.9368404{15} ; Abundance:0;  StandardWeight:69.723{1}),
(Symbol:'Ga'; AtomicNumber:31; MassNumber:65; Mass:64.93273459{88} ; Abundance:0;  StandardWeight:69.723{1}),
(Symbol:'Ga'; AtomicNumber:31; MassNumber:66; Mass:65.9315894{34} ; Abundance:0;  StandardWeight:69.723{1}),
(Symbol:'Ga'; AtomicNumber:31; MassNumber:67; Mass:66.9282025{13} ; Abundance:0;  StandardWeight:69.723{1}),
(Symbol:'Ga'; AtomicNumber:31; MassNumber:68; Mass:67.9279805{16} ; Abundance:0;  StandardWeight:69.723{1}),
(Symbol:'Ga'; AtomicNumber:31; MassNumber:69; Mass:68.9255735{13} ; Abundance:60.108;  StandardWeight:69.723{1}),
(Symbol:'Ga'; AtomicNumber:31; MassNumber:70; Mass:69.9260219{13} ; Abundance:0;  StandardWeight:69.723{1}),
(Symbol:'Ga'; AtomicNumber:31; MassNumber:71; Mass:70.92470258{87} ; Abundance:39.892;  StandardWeight:69.723{1}),
(Symbol:'Ga'; AtomicNumber:31; MassNumber:72; Mass:71.92636747{88} ; Abundance:0;  StandardWeight:69.723{1}),
(Symbol:'Ga'; AtomicNumber:31; MassNumber:73; Mass:72.9251747{18} ; Abundance:0;  StandardWeight:69.723{1}),
(Symbol:'Ga'; AtomicNumber:31; MassNumber:74; Mass:73.9269457{32} ; Abundance:0;  StandardWeight:69.723{1}),
(Symbol:'Ga'; AtomicNumber:31; MassNumber:75; Mass:74.9265002{26} ; Abundance:0;  StandardWeight:69.723{1}),
(Symbol:'Ga'; AtomicNumber:31; MassNumber:76; Mass:75.9288276{21} ; Abundance:0;  StandardWeight:69.723{1}),
(Symbol:'Ga'; AtomicNumber:31; MassNumber:77; Mass:76.9291543{26} ; Abundance:0;  StandardWeight:69.723{1}),
(Symbol:'Ga'; AtomicNumber:31; MassNumber:78; Mass:77.9316088{20} ; Abundance:0;  StandardWeight:69.723{1}),
(Symbol:'Ga'; AtomicNumber:31; MassNumber:79; Mass:78.9328523{20} ; Abundance:0;  StandardWeight:69.723{1}),
(Symbol:'Ga'; AtomicNumber:31; MassNumber:80; Mass:79.9364208{31} ; Abundance:0;  StandardWeight:69.723{1}),
(Symbol:'Ga'; AtomicNumber:31; MassNumber:81; Mass:80.9381338{35} ; Abundance:0;  StandardWeight:69.723{1}),
(Symbol:'Ga'; AtomicNumber:31; MassNumber:82; Mass:81.9431765{26} ; Abundance:0;  StandardWeight:69.723{1}),
(Symbol:'Ga'; AtomicNumber:31; MassNumber:83; Mass:82.9471203{28} ; Abundance:0;  StandardWeight:69.723{1}),
(Symbol:'Ga'; AtomicNumber:31; MassNumber:84; Mass:83.95246{43#} ; Abundance:0;  StandardWeight:69.723{1}),
(Symbol:'Ga'; AtomicNumber:31; MassNumber:85; Mass:84.95699{32#} ; Abundance:0;  StandardWeight:69.723{1}),
(Symbol:'Ga'; AtomicNumber:31; MassNumber:86; Mass:85.96301{75#} ; Abundance:0;  StandardWeight:69.723{1}),
(Symbol:'Ga'; AtomicNumber:31; MassNumber:87; Mass:86.96824{86#} ; Abundance:0;  StandardWeight:69.723{1}),
(Symbol:'Ge'; AtomicNumber:32; MassNumber:58; Mass:57.99172{43#} ; Abundance:0;  StandardWeight:72.630{8}),
(Symbol:'Ge'; AtomicNumber:32; MassNumber:59; Mass:58.98249{32#} ; Abundance:0;  StandardWeight:72.630{8}),
(Symbol:'Ge'; AtomicNumber:32; MassNumber:60; Mass:59.97036{21#} ; Abundance:0;  StandardWeight:72.630{8}),
(Symbol:'Ge'; AtomicNumber:32; MassNumber:61; Mass:60.96379{32#} ; Abundance:0;  StandardWeight:72.630{8}),
(Symbol:'Ge'; AtomicNumber:32; MassNumber:62; Mass:61.95502{15#} ; Abundance:0;  StandardWeight:72.630{8}),
(Symbol:'Ge'; AtomicNumber:32; MassNumber:63; Mass:62.949628{40} ; Abundance:0;  StandardWeight:72.630{8}),
(Symbol:'Ge'; AtomicNumber:32; MassNumber:64; Mass:63.9416899{40} ; Abundance:0;  StandardWeight:72.630{8}),
(Symbol:'Ge'; AtomicNumber:32; MassNumber:65; Mass:64.9393681{23} ; Abundance:0;  StandardWeight:72.630{8}),
(Symbol:'Ge'; AtomicNumber:32; MassNumber:66; Mass:65.9338621{26} ; Abundance:0;  StandardWeight:72.630{8}),
(Symbol:'Ge'; AtomicNumber:32; MassNumber:67; Mass:66.9327339{50} ; Abundance:0;  StandardWeight:72.630{8}),
(Symbol:'Ge'; AtomicNumber:32; MassNumber:68; Mass:67.9280953{20} ; Abundance:0;  StandardWeight:72.630{8}),
(Symbol:'Ge'; AtomicNumber:32; MassNumber:69; Mass:68.9279645{14} ; Abundance:0;  StandardWeight:72.630{8}),
(Symbol:'Ge'; AtomicNumber:32; MassNumber:70; Mass:69.92424875{90} ; Abundance:20.57;  StandardWeight:72.630{8}),
(Symbol:'Ge'; AtomicNumber:32; MassNumber:71; Mass:70.92495233{90} ; Abundance:0;  StandardWeight:72.630{8}),
(Symbol:'Ge'; AtomicNumber:32; MassNumber:72; Mass:71.922075826{81} ; Abundance:27.45;  StandardWeight:72.630{8}),
(Symbol:'Ge'; AtomicNumber:32; MassNumber:73; Mass:72.923458956{61} ; Abundance:7.75;  StandardWeight:72.630{8}),
(Symbol:'Ge'; AtomicNumber:32; MassNumber:74; Mass:73.921177761{13} ; Abundance:36.5;  StandardWeight:72.630{8}),
(Symbol:'Ge'; AtomicNumber:32; MassNumber:75; Mass:74.922858370{55} ; Abundance:0;  StandardWeight:72.630{8}),
(Symbol:'Ge'; AtomicNumber:32; MassNumber:76; Mass:75.921402726{19} ; Abundance:7.73;  StandardWeight:72.630{8}),
(Symbol:'Ge'; AtomicNumber:32; MassNumber:77; Mass:76.923549843{57} ; Abundance:0;  StandardWeight:72.630{8}),
(Symbol:'Ge'; AtomicNumber:32; MassNumber:78; Mass:77.9228529{38} ; Abundance:0;  StandardWeight:72.630{8}),
(Symbol:'Ge'; AtomicNumber:32; MassNumber:79; Mass:78.925360{40} ; Abundance:0;  StandardWeight:72.630{8}),
(Symbol:'Ge'; AtomicNumber:32; MassNumber:80; Mass:79.9253508{22} ; Abundance:0;  StandardWeight:72.630{8}),
(Symbol:'Ge'; AtomicNumber:32; MassNumber:81; Mass:80.9288329{22} ; Abundance:0;  StandardWeight:72.630{8}),
(Symbol:'Ge'; AtomicNumber:32; MassNumber:82; Mass:81.9297740{24} ; Abundance:0;  StandardWeight:72.630{8}),
(Symbol:'Ge'; AtomicNumber:32; MassNumber:83; Mass:82.9345391{26} ; Abundance:0;  StandardWeight:72.630{8}),
(Symbol:'Ge'; AtomicNumber:32; MassNumber:84; Mass:83.9375751{34} ; Abundance:0;  StandardWeight:72.630{8}),
(Symbol:'Ge'; AtomicNumber:32; MassNumber:85; Mass:84.9429697{40} ; Abundance:0;  StandardWeight:72.630{8}),
(Symbol:'Ge'; AtomicNumber:32; MassNumber:86; Mass:85.94658{32#} ; Abundance:0;  StandardWeight:72.630{8}),
(Symbol:'Ge'; AtomicNumber:32; MassNumber:87; Mass:86.95268{43#} ; Abundance:0;  StandardWeight:72.630{8}),
(Symbol:'Ge'; AtomicNumber:32; MassNumber:88; Mass:87.95691{54#} ; Abundance:0;  StandardWeight:72.630{8}),
(Symbol:'Ge'; AtomicNumber:32; MassNumber:89; Mass:88.96379{64#} ; Abundance:0;  StandardWeight:72.630{8}),
(Symbol:'Ge'; AtomicNumber:32; MassNumber:90; Mass:89.96863{75#} ; Abundance:0;  StandardWeight:72.630{8}),
(Symbol:'As'; AtomicNumber:33; MassNumber:60; Mass:59.99388{43#} ; Abundance:0;  StandardWeight:74.921595{6}),
(Symbol:'As'; AtomicNumber:33; MassNumber:61; Mass:60.98112{32#} ; Abundance:0;  StandardWeight:74.921595{6}),
(Symbol:'As'; AtomicNumber:33; MassNumber:62; Mass:61.97361{32#} ; Abundance:0;  StandardWeight:74.921595{6}),
(Symbol:'As'; AtomicNumber:33; MassNumber:63; Mass:62.96390{21#} ; Abundance:0;  StandardWeight:74.921595{6}),
(Symbol:'As'; AtomicNumber:33; MassNumber:64; Mass:63.95743{33#} ; Abundance:0;  StandardWeight:74.921595{6}),
(Symbol:'As'; AtomicNumber:33; MassNumber:65; Mass:64.949611{91} ; Abundance:0;  StandardWeight:74.921595{6}),
(Symbol:'As'; AtomicNumber:33; MassNumber:66; Mass:65.9441488{61} ; Abundance:0;  StandardWeight:74.921595{6}),
(Symbol:'As'; AtomicNumber:33; MassNumber:67; Mass:66.93925111{48} ; Abundance:0;  StandardWeight:74.921595{6}),
(Symbol:'As'; AtomicNumber:33; MassNumber:68; Mass:67.9367741{20} ; Abundance:0;  StandardWeight:74.921595{6}),
(Symbol:'As'; AtomicNumber:33; MassNumber:69; Mass:68.932246{34} ; Abundance:0;  StandardWeight:74.921595{6}),
(Symbol:'As'; AtomicNumber:33; MassNumber:70; Mass:69.930926{54} ; Abundance:0;  StandardWeight:74.921595{6}),
(Symbol:'As'; AtomicNumber:33; MassNumber:71; Mass:70.9271138{45} ; Abundance:0;  StandardWeight:74.921595{6}),
(Symbol:'As'; AtomicNumber:33; MassNumber:72; Mass:71.9267523{44} ; Abundance:0;  StandardWeight:74.921595{6}),
(Symbol:'As'; AtomicNumber:33; MassNumber:73; Mass:72.9238291{41} ; Abundance:0;  StandardWeight:74.921595{6}),
(Symbol:'As'; AtomicNumber:33; MassNumber:74; Mass:73.9239286{18} ; Abundance:0;  StandardWeight:74.921595{6}),
(Symbol:'As'; AtomicNumber:33; MassNumber:75; Mass:74.92159457{95} ; Abundance:100;  StandardWeight:74.921595{6}),
(Symbol:'As'; AtomicNumber:33; MassNumber:76; Mass:75.92239202{95} ; Abundance:0;  StandardWeight:74.921595{6}),
(Symbol:'As'; AtomicNumber:33; MassNumber:77; Mass:76.9206476{18} ; Abundance:0;  StandardWeight:74.921595{6}),
(Symbol:'As'; AtomicNumber:33; MassNumber:78; Mass:77.921828{11} ; Abundance:0;  StandardWeight:74.921595{6}),
(Symbol:'As'; AtomicNumber:33; MassNumber:79; Mass:78.9209484{58} ; Abundance:0;  StandardWeight:74.921595{6}),
(Symbol:'As'; AtomicNumber:33; MassNumber:80; Mass:79.9224746{36} ; Abundance:0;  StandardWeight:74.921595{6}),
(Symbol:'As'; AtomicNumber:33; MassNumber:81; Mass:80.9221323{29} ; Abundance:0;  StandardWeight:74.921595{6}),
(Symbol:'As'; AtomicNumber:33; MassNumber:82; Mass:81.9247412{46} ; Abundance:0;  StandardWeight:74.921595{6}),
(Symbol:'As'; AtomicNumber:33; MassNumber:83; Mass:82.9252069{30} ; Abundance:0;  StandardWeight:74.921595{6}),
(Symbol:'As'; AtomicNumber:33; MassNumber:84; Mass:83.9293033{34} ; Abundance:0;  StandardWeight:74.921595{6}),
(Symbol:'As'; AtomicNumber:33; MassNumber:85; Mass:84.9321637{33} ; Abundance:0;  StandardWeight:74.921595{6}),
(Symbol:'As'; AtomicNumber:33; MassNumber:86; Mass:85.9367015{37} ; Abundance:0;  StandardWeight:74.921595{6}),
(Symbol:'As'; AtomicNumber:33; MassNumber:87; Mass:86.9402917{32} ; Abundance:0;  StandardWeight:74.921595{6}),
(Symbol:'As'; AtomicNumber:33; MassNumber:88; Mass:87.94555{21#} ; Abundance:0;  StandardWeight:74.921595{6}),
(Symbol:'As'; AtomicNumber:33; MassNumber:89; Mass:88.94976{32#} ; Abundance:0;  StandardWeight:74.921595{6}),
(Symbol:'As'; AtomicNumber:33; MassNumber:90; Mass:89.95563{64#} ; Abundance:0;  StandardWeight:74.921595{6}),
(Symbol:'As'; AtomicNumber:33; MassNumber:91; Mass:90.96039{64#} ; Abundance:0;  StandardWeight:74.921595{6}),
(Symbol:'As'; AtomicNumber:33; MassNumber:92; Mass:91.96674{75#} ; Abundance:0;  StandardWeight:74.921595{6}),
(Symbol:'Se'; AtomicNumber:34; MassNumber:64; Mass:63.97109{54#} ; Abundance:0; {r} StandardWeight:78.971{8}),
(Symbol:'Se'; AtomicNumber:34; MassNumber:65; Mass:64.96440{64#} ; Abundance:0; {r} StandardWeight:78.971{8}),
(Symbol:'Se'; AtomicNumber:34; MassNumber:66; Mass:65.95559{32#} ; Abundance:0; {r} StandardWeight:78.971{8}),
(Symbol:'Se'; AtomicNumber:34; MassNumber:67; Mass:66.949994{72} ; Abundance:0; {r} StandardWeight:78.971{8}),
(Symbol:'Se'; AtomicNumber:34; MassNumber:68; Mass:67.94182524{53} ; Abundance:0; {r} StandardWeight:78.971{8}),
(Symbol:'Se'; AtomicNumber:34; MassNumber:69; Mass:68.9394148{16} ; Abundance:0; {r} StandardWeight:78.971{8}),
(Symbol:'Se'; AtomicNumber:34; MassNumber:70; Mass:69.9335155{17} ; Abundance:0; {r} StandardWeight:78.971{8}),
(Symbol:'Se'; AtomicNumber:34; MassNumber:71; Mass:70.9322094{30} ; Abundance:0; {r} StandardWeight:78.971{8}),
(Symbol:'Se'; AtomicNumber:34; MassNumber:72; Mass:71.9271405{21} ; Abundance:0; {r} StandardWeight:78.971{8}),
(Symbol:'Se'; AtomicNumber:34; MassNumber:73; Mass:72.9267549{80} ; Abundance:0; {r} StandardWeight:78.971{8}),
(Symbol:'Se'; AtomicNumber:34; MassNumber:74; Mass:73.922475934{15} ; Abundance:0.89; {r} StandardWeight:78.971{8}),
(Symbol:'Se'; AtomicNumber:34; MassNumber:75; Mass:74.922522870{78} ; Abundance:0; {r} StandardWeight:78.971{8}),
(Symbol:'Se'; AtomicNumber:34; MassNumber:76; Mass:75.919213704{17} ; Abundance:9.37; {r} StandardWeight:78.971{8}),
(Symbol:'Se'; AtomicNumber:34; MassNumber:77; Mass:76.919914154{67} ; Abundance:7.63; {r} StandardWeight:78.971{8}),
(Symbol:'Se'; AtomicNumber:34; MassNumber:78; Mass:77.91730928{20} ; Abundance:23.77; {r} StandardWeight:78.971{8}),
(Symbol:'Se'; AtomicNumber:34; MassNumber:79; Mass:78.91849929{24} ; Abundance:0; {r} StandardWeight:78.971{8}),
(Symbol:'Se'; AtomicNumber:34; MassNumber:80; Mass:79.9165218{13} ; Abundance:49.61; {r} StandardWeight:78.971{8}),
(Symbol:'Se'; AtomicNumber:34; MassNumber:81; Mass:80.9179930{14} ; Abundance:0; {r} StandardWeight:78.971{8}),
(Symbol:'Se'; AtomicNumber:34; MassNumber:82; Mass:81.9166995{15} ; Abundance:8.73; {r} StandardWeight:78.971{8}),
(Symbol:'Se'; AtomicNumber:34; MassNumber:83; Mass:82.9191186{36} ; Abundance:0; {r} StandardWeight:78.971{8}),
(Symbol:'Se'; AtomicNumber:34; MassNumber:84; Mass:83.9184668{21} ; Abundance:0; {r} StandardWeight:78.971{8}),
(Symbol:'Se'; AtomicNumber:34; MassNumber:85; Mass:84.9222608{28} ; Abundance:0; {r} StandardWeight:78.971{8}),
(Symbol:'Se'; AtomicNumber:34; MassNumber:86; Mass:85.9243117{27} ; Abundance:0; {r} StandardWeight:78.971{8}),
(Symbol:'Se'; AtomicNumber:34; MassNumber:87; Mass:86.9286886{24} ; Abundance:0; {r} StandardWeight:78.971{8}),
(Symbol:'Se'; AtomicNumber:34; MassNumber:88; Mass:87.9314175{36} ; Abundance:0; {r} StandardWeight:78.971{8}),
(Symbol:'Se'; AtomicNumber:34; MassNumber:89; Mass:88.9366691{40} ; Abundance:0; {r} StandardWeight:78.971{8}),
(Symbol:'Se'; AtomicNumber:34; MassNumber:90; Mass:89.94010{35} ; Abundance:0; {r} StandardWeight:78.971{8}),
(Symbol:'Se'; AtomicNumber:34; MassNumber:91; Mass:90.94596{54#} ; Abundance:0; {r} StandardWeight:78.971{8}),
(Symbol:'Se'; AtomicNumber:34; MassNumber:92; Mass:91.94984{64#} ; Abundance:0; {r} StandardWeight:78.971{8}),
(Symbol:'Se'; AtomicNumber:34; MassNumber:93; Mass:92.95629{86#} ; Abundance:0; {r} StandardWeight:78.971{8}),
(Symbol:'Se'; AtomicNumber:34; MassNumber:94; Mass:93.96049{86#} ; Abundance:0; {r} StandardWeight:78.971{8}),
(Symbol:'Se'; AtomicNumber:34; MassNumber:95; Mass:94.96730{86#} ; Abundance:0; {r} StandardWeight:78.971{8}),
(Symbol:'Br'; AtomicNumber:35; MassNumber:67; Mass:66.96465{54#} ; Abundance:0;  StandardWeight:79.904 {79.901-79.907}),
(Symbol:'Br'; AtomicNumber:35; MassNumber:68; Mass:67.95873{33#} ; Abundance:0;  StandardWeight:79.904 {79.901-79.907}),
(Symbol:'Br'; AtomicNumber:35; MassNumber:69; Mass:68.950497{40} ; Abundance:0;  StandardWeight:79.904 {79.901-79.907}),
(Symbol:'Br'; AtomicNumber:35; MassNumber:70; Mass:69.944792{16} ; Abundance:0;  StandardWeight:79.904 {79.901-79.907}),
(Symbol:'Br'; AtomicNumber:35; MassNumber:71; Mass:70.9393422{58} ; Abundance:0;  StandardWeight:79.904 {79.901-79.907}),
(Symbol:'Br'; AtomicNumber:35; MassNumber:72; Mass:71.9365886{72} ; Abundance:0;  StandardWeight:79.904 {79.901-79.907}),
(Symbol:'Br'; AtomicNumber:35; MassNumber:73; Mass:72.9316715{78} ; Abundance:0;  StandardWeight:79.904 {79.901-79.907}),
(Symbol:'Br'; AtomicNumber:35; MassNumber:74; Mass:73.9299102{63} ; Abundance:0;  StandardWeight:79.904 {79.901-79.907}),
(Symbol:'Br'; AtomicNumber:35; MassNumber:75; Mass:74.9258105{46} ; Abundance:0;  StandardWeight:79.904 {79.901-79.907}),
(Symbol:'Br'; AtomicNumber:35; MassNumber:76; Mass:75.924542{10} ; Abundance:0;  StandardWeight:79.904 {79.901-79.907}),
(Symbol:'Br'; AtomicNumber:35; MassNumber:77; Mass:76.9213792{30} ; Abundance:0;  StandardWeight:79.904 {79.901-79.907}),
(Symbol:'Br'; AtomicNumber:35; MassNumber:78; Mass:77.9211459{38} ; Abundance:0;  StandardWeight:79.904 {79.901-79.907}),
(Symbol:'Br'; AtomicNumber:35; MassNumber:79; Mass:78.9183376{14} ; Abundance:50.69;  StandardWeight:79.904 {79.901-79.907}),
(Symbol:'Br'; AtomicNumber:35; MassNumber:80; Mass:79.9185298{14} ; Abundance:0;  StandardWeight:79.904 {79.901-79.907}),
(Symbol:'Br'; AtomicNumber:35; MassNumber:81; Mass:80.9162897{14} ; Abundance:49.31;  StandardWeight:79.904 {79.901-79.907}),
(Symbol:'Br'; AtomicNumber:35; MassNumber:82; Mass:81.9168032{14} ; Abundance:0;  StandardWeight:79.904 {79.901-79.907}),
(Symbol:'Br'; AtomicNumber:35; MassNumber:83; Mass:82.9151756{41} ; Abundance:0;  StandardWeight:79.904 {79.901-79.907}),
(Symbol:'Br'; AtomicNumber:35; MassNumber:84; Mass:83.916496{28} ; Abundance:0;  StandardWeight:79.904 {79.901-79.907}),
(Symbol:'Br'; AtomicNumber:35; MassNumber:85; Mass:84.9156458{33} ; Abundance:0;  StandardWeight:79.904 {79.901-79.907}),
(Symbol:'Br'; AtomicNumber:35; MassNumber:86; Mass:85.9188054{33} ; Abundance:0;  StandardWeight:79.904 {79.901-79.907}),
(Symbol:'Br'; AtomicNumber:35; MassNumber:87; Mass:86.9206740{34} ; Abundance:0;  StandardWeight:79.904 {79.901-79.907}),
(Symbol:'Br'; AtomicNumber:35; MassNumber:88; Mass:87.9240833{34} ; Abundance:0;  StandardWeight:79.904 {79.901-79.907}),
(Symbol:'Br'; AtomicNumber:35; MassNumber:89; Mass:88.9267046{35} ; Abundance:0;  StandardWeight:79.904 {79.901-79.907}),
(Symbol:'Br'; AtomicNumber:35; MassNumber:90; Mass:89.9312928{36} ; Abundance:0;  StandardWeight:79.904 {79.901-79.907}),
(Symbol:'Br'; AtomicNumber:35; MassNumber:91; Mass:90.9343986{38} ; Abundance:0;  StandardWeight:79.904 {79.901-79.907}),
(Symbol:'Br'; AtomicNumber:35; MassNumber:92; Mass:91.9396316{72} ; Abundance:0;  StandardWeight:79.904 {79.901-79.907}),
(Symbol:'Br'; AtomicNumber:35; MassNumber:93; Mass:92.94313{48} ; Abundance:0;  StandardWeight:79.904 {79.901-79.907}),
(Symbol:'Br'; AtomicNumber:35; MassNumber:94; Mass:93.94890{43#} ; Abundance:0;  StandardWeight:79.904 {79.901-79.907}),
(Symbol:'Br'; AtomicNumber:35; MassNumber:95; Mass:94.95301{21#} ; Abundance:0;  StandardWeight:79.904 {79.901-79.907}),
(Symbol:'Br'; AtomicNumber:35; MassNumber:96; Mass:95.95903{32#} ; Abundance:0;  StandardWeight:79.904 {79.901-79.907}),
(Symbol:'Br'; AtomicNumber:35; MassNumber:97; Mass:96.96344{43#} ; Abundance:0;  StandardWeight:79.904 {79.901-79.907}),
(Symbol:'Br'; AtomicNumber:35; MassNumber:98; Mass:97.96946{43#} ; Abundance:0;  StandardWeight:79.904 {79.901-79.907}),
(Symbol:'Kr'; AtomicNumber:36; MassNumber:69; Mass:68.96518{43#} ; Abundance:0; {g,m} StandardWeight:83.798{2}),
(Symbol:'Kr'; AtomicNumber:36; MassNumber:70; Mass:69.95604{21#} ; Abundance:0; {g,m} StandardWeight:83.798{2}),
(Symbol:'Kr'; AtomicNumber:36; MassNumber:71; Mass:70.95027{14} ; Abundance:0; {g,m} StandardWeight:83.798{2}),
(Symbol:'Kr'; AtomicNumber:36; MassNumber:72; Mass:71.9420924{86} ; Abundance:0; {g,m} StandardWeight:83.798{2}),
(Symbol:'Kr'; AtomicNumber:36; MassNumber:73; Mass:72.9392892{71} ; Abundance:0; {g,m} StandardWeight:83.798{2}),
(Symbol:'Kr'; AtomicNumber:36; MassNumber:74; Mass:73.9330840{22} ; Abundance:0; {g,m} StandardWeight:83.798{2}),
(Symbol:'Kr'; AtomicNumber:36; MassNumber:75; Mass:74.9309457{87} ; Abundance:0; {g,m} StandardWeight:83.798{2}),
(Symbol:'Kr'; AtomicNumber:36; MassNumber:76; Mass:75.9259103{43} ; Abundance:0; {g,m} StandardWeight:83.798{2}),
(Symbol:'Kr'; AtomicNumber:36; MassNumber:77; Mass:76.9246700{21} ; Abundance:0; {g,m} StandardWeight:83.798{2}),
(Symbol:'Kr'; AtomicNumber:36; MassNumber:78; Mass:77.92036494{76} ; Abundance:0.355; {g,m} StandardWeight:83.798{2}),
(Symbol:'Kr'; AtomicNumber:36; MassNumber:79; Mass:78.9200829{38} ; Abundance:0; {g,m} StandardWeight:83.798{2}),
(Symbol:'Kr'; AtomicNumber:36; MassNumber:80; Mass:79.91637808{75} ; Abundance:2.286; {g,m} StandardWeight:83.798{2}),
(Symbol:'Kr'; AtomicNumber:36; MassNumber:81; Mass:80.9165912{15} ; Abundance:0; {g,m} StandardWeight:83.798{2}),
(Symbol:'Kr'; AtomicNumber:36; MassNumber:82; Mass:81.91348273{94} ; Abundance:11.593; {g,m} StandardWeight:83.798{2}),
(Symbol:'Kr'; AtomicNumber:36; MassNumber:83; Mass:82.91412716{32} ; Abundance:11.5; {g,m} StandardWeight:83.798{2}),
(Symbol:'Kr'; AtomicNumber:36; MassNumber:84; Mass:83.9114977282{44} ; Abundance:56.987; {g,m} StandardWeight:83.798{2}),
(Symbol:'Kr'; AtomicNumber:36; MassNumber:85; Mass:84.9125273{21} ; Abundance:0; {g,m} StandardWeight:83.798{2}),
(Symbol:'Kr'; AtomicNumber:36; MassNumber:86; Mass:85.9106106269{41} ; Abundance:17.279; {g,m} StandardWeight:83.798{2}),
(Symbol:'Kr'; AtomicNumber:36; MassNumber:87; Mass:86.91335476{26} ; Abundance:0; {g,m} StandardWeight:83.798{2}),
(Symbol:'Kr'; AtomicNumber:36; MassNumber:88; Mass:87.9144479{28} ; Abundance:0; {g,m} StandardWeight:83.798{2}),
(Symbol:'Kr'; AtomicNumber:36; MassNumber:89; Mass:88.9178355{23} ; Abundance:0; {g,m} StandardWeight:83.798{2}),
(Symbol:'Kr'; AtomicNumber:36; MassNumber:90; Mass:89.9195279{20} ; Abundance:0; {g,m} StandardWeight:83.798{2}),
(Symbol:'Kr'; AtomicNumber:36; MassNumber:91; Mass:90.9238063{24} ; Abundance:0; {g,m} StandardWeight:83.798{2}),
(Symbol:'Kr'; AtomicNumber:36; MassNumber:92; Mass:91.9261731{29} ; Abundance:0; {g,m} StandardWeight:83.798{2}),
(Symbol:'Kr'; AtomicNumber:36; MassNumber:93; Mass:92.9311472{27} ; Abundance:0; {g,m} StandardWeight:83.798{2}),
(Symbol:'Kr'; AtomicNumber:36; MassNumber:94; Mass:93.934140{13} ; Abundance:0; {g,m} StandardWeight:83.798{2}),
(Symbol:'Kr'; AtomicNumber:36; MassNumber:95; Mass:94.939711{20} ; Abundance:0; {g,m} StandardWeight:83.798{2}),
(Symbol:'Kr'; AtomicNumber:36; MassNumber:96; Mass:95.943017{22} ; Abundance:0; {g,m} StandardWeight:83.798{2}),
(Symbol:'Kr'; AtomicNumber:36; MassNumber:97; Mass:96.94909{14} ; Abundance:0; {g,m} StandardWeight:83.798{2}),
(Symbol:'Kr'; AtomicNumber:36; MassNumber:98; Mass:97.95243{32#} ; Abundance:0; {g,m} StandardWeight:83.798{2}),
(Symbol:'Kr'; AtomicNumber:36; MassNumber:99; Mass:98.95839{54#} ; Abundance:0; {g,m} StandardWeight:83.798{2}),
(Symbol:'Kr'; AtomicNumber:36; MassNumber:100; Mass:99.96237{43#} ; Abundance:0; {g,m} StandardWeight:83.798{2}),
(Symbol:'Kr'; AtomicNumber:36; MassNumber:101; Mass:100.96873{54#} ; Abundance:0; {g,m} StandardWeight:83.798{2}),
(Symbol:'Rb'; AtomicNumber:37; MassNumber:71; Mass:70.96532{54#} ; Abundance:0; {g} StandardWeight:85.4678{3}),
(Symbol:'Rb'; AtomicNumber:37; MassNumber:72; Mass:71.95908{54#} ; Abundance:0; {g} StandardWeight:85.4678{3}),
(Symbol:'Rb'; AtomicNumber:37; MassNumber:73; Mass:72.95053{11#} ; Abundance:0; {g} StandardWeight:85.4678{3}),
(Symbol:'Rb'; AtomicNumber:37; MassNumber:74; Mass:73.9442659{32} ; Abundance:0; {g} StandardWeight:85.4678{3}),
(Symbol:'Rb'; AtomicNumber:37; MassNumber:75; Mass:74.9385732{13} ; Abundance:0; {g} StandardWeight:85.4678{3}),
(Symbol:'Rb'; AtomicNumber:37; MassNumber:76; Mass:75.9350730{10} ; Abundance:0; {g} StandardWeight:85.4678{3}),
(Symbol:'Rb'; AtomicNumber:37; MassNumber:77; Mass:76.9304016{14} ; Abundance:0; {g} StandardWeight:85.4678{3}),
(Symbol:'Rb'; AtomicNumber:37; MassNumber:78; Mass:77.9281419{35} ; Abundance:0; {g} StandardWeight:85.4678{3}),
(Symbol:'Rb'; AtomicNumber:37; MassNumber:79; Mass:78.9239899{23} ; Abundance:0; {g} StandardWeight:85.4678{3}),
(Symbol:'Rb'; AtomicNumber:37; MassNumber:80; Mass:79.9225164{20} ; Abundance:0; {g} StandardWeight:85.4678{3}),
(Symbol:'Rb'; AtomicNumber:37; MassNumber:81; Mass:80.9189939{53} ; Abundance:0; {g} StandardWeight:85.4678{3}),
(Symbol:'Rb'; AtomicNumber:37; MassNumber:82; Mass:81.9182090{32} ; Abundance:0; {g} StandardWeight:85.4678{3}),
(Symbol:'Rb'; AtomicNumber:37; MassNumber:83; Mass:82.9151142{25} ; Abundance:0; {g} StandardWeight:85.4678{3}),
(Symbol:'Rb'; AtomicNumber:37; MassNumber:84; Mass:83.9143752{24} ; Abundance:0; {g} StandardWeight:85.4678{3}),
(Symbol:'Rb'; AtomicNumber:37; MassNumber:85; Mass:84.9117897379{54} ; Abundance:72.17; {g} StandardWeight:85.4678{3}),
(Symbol:'Rb'; AtomicNumber:37; MassNumber:86; Mass:85.91116743{21} ; Abundance:0; {g} StandardWeight:85.4678{3}),
(Symbol:'Rb'; AtomicNumber:37; MassNumber:87; Mass:86.9091805310{60} ; Abundance:27.83; {g} StandardWeight:85.4678{3}),
(Symbol:'Rb'; AtomicNumber:37; MassNumber:88; Mass:87.91131559{17} ; Abundance:0; {g} StandardWeight:85.4678{3}),
(Symbol:'Rb'; AtomicNumber:37; MassNumber:89; Mass:88.9122783{59} ; Abundance:0; {g} StandardWeight:85.4678{3}),
(Symbol:'Rb'; AtomicNumber:37; MassNumber:90; Mass:89.9147985{70} ; Abundance:0; {g} StandardWeight:85.4678{3}),
(Symbol:'Rb'; AtomicNumber:37; MassNumber:91; Mass:90.9165372{84} ; Abundance:0; {g} StandardWeight:85.4678{3}),
(Symbol:'Rb'; AtomicNumber:37; MassNumber:92; Mass:91.9197284{66} ; Abundance:0; {g} StandardWeight:85.4678{3}),
(Symbol:'Rb'; AtomicNumber:37; MassNumber:93; Mass:92.9220393{84} ; Abundance:0; {g} StandardWeight:85.4678{3}),
(Symbol:'Rb'; AtomicNumber:37; MassNumber:94; Mass:93.9263948{22} ; Abundance:0; {g} StandardWeight:85.4678{3}),
(Symbol:'Rb'; AtomicNumber:37; MassNumber:95; Mass:94.929260{22} ; Abundance:0; {g} StandardWeight:85.4678{3}),
(Symbol:'Rb'; AtomicNumber:37; MassNumber:96; Mass:95.9341334{36} ; Abundance:0; {g} StandardWeight:85.4678{3}),
(Symbol:'Rb'; AtomicNumber:37; MassNumber:97; Mass:96.9371771{21} ; Abundance:0; {g} StandardWeight:85.4678{3}),
(Symbol:'Rb'; AtomicNumber:37; MassNumber:98; Mass:97.9416869{37} ; Abundance:0; {g} StandardWeight:85.4678{3}),
(Symbol:'Rb'; AtomicNumber:37; MassNumber:99; Mass:98.94503{12} ; Abundance:0; {g} StandardWeight:85.4678{3}),
(Symbol:'Rb'; AtomicNumber:37; MassNumber:100; Mass:99.95003{21#} ; Abundance:0; {g} StandardWeight:85.4678{3}),
(Symbol:'Rb'; AtomicNumber:37; MassNumber:101; Mass:100.95404{23#} ; Abundance:0; {g} StandardWeight:85.4678{3}),
(Symbol:'Rb'; AtomicNumber:37; MassNumber:102; Mass:101.95952{32#} ; Abundance:0; {g} StandardWeight:85.4678{3}),
(Symbol:'Rb'; AtomicNumber:37; MassNumber:103; Mass:102.96392{43#} ; Abundance:0; {g} StandardWeight:85.4678{3}),
(Symbol:'Sr'; AtomicNumber:38; MassNumber:73; Mass:72.96570{43#} ; Abundance:0; {g,r} StandardWeight:87.62{1}),
(Symbol:'Sr'; AtomicNumber:38; MassNumber:74; Mass:73.95617{11#} ; Abundance:0; {g,r} StandardWeight:87.62{1}),
(Symbol:'Sr'; AtomicNumber:38; MassNumber:75; Mass:74.94995{24} ; Abundance:0; {g,r} StandardWeight:87.62{1}),
(Symbol:'Sr'; AtomicNumber:38; MassNumber:76; Mass:75.941763{37} ; Abundance:0; {g,r} StandardWeight:87.62{1}),
(Symbol:'Sr'; AtomicNumber:38; MassNumber:77; Mass:76.9379455{85} ; Abundance:0; {g,r} StandardWeight:87.62{1}),
(Symbol:'Sr'; AtomicNumber:38; MassNumber:78; Mass:77.9321800{80} ; Abundance:0; {g,r} StandardWeight:87.62{1}),
(Symbol:'Sr'; AtomicNumber:38; MassNumber:79; Mass:78.9297077{90} ; Abundance:0; {g,r} StandardWeight:87.62{1}),
(Symbol:'Sr'; AtomicNumber:38; MassNumber:80; Mass:79.9245175{37} ; Abundance:0; {g,r} StandardWeight:87.62{1}),
(Symbol:'Sr'; AtomicNumber:38; MassNumber:81; Mass:80.9232114{34} ; Abundance:0; {g,r} StandardWeight:87.62{1}),
(Symbol:'Sr'; AtomicNumber:38; MassNumber:82; Mass:81.9183999{64} ; Abundance:0; {g,r} StandardWeight:87.62{1}),
(Symbol:'Sr'; AtomicNumber:38; MassNumber:83; Mass:82.9175544{73} ; Abundance:0; {g,r} StandardWeight:87.62{1}),
(Symbol:'Sr'; AtomicNumber:38; MassNumber:84; Mass:83.9134191{13} ; Abundance:0.56; {g,r} StandardWeight:87.62{1}),
(Symbol:'Sr'; AtomicNumber:38; MassNumber:85; Mass:84.9129320{30} ; Abundance:0; {g,r} StandardWeight:87.62{1}),
(Symbol:'Sr'; AtomicNumber:38; MassNumber:86; Mass:85.9092606{12} ; Abundance:9.86; {g,r} StandardWeight:87.62{1}),
(Symbol:'Sr'; AtomicNumber:38; MassNumber:87; Mass:86.9088775{12} ; Abundance:7; {g,r} StandardWeight:87.62{1}),
(Symbol:'Sr'; AtomicNumber:38; MassNumber:88; Mass:87.9056125{12} ; Abundance:82.58; {g,r} StandardWeight:87.62{1}),
(Symbol:'Sr'; AtomicNumber:38; MassNumber:89; Mass:88.9074511{12} ; Abundance:0; {g,r} StandardWeight:87.62{1}),
(Symbol:'Sr'; AtomicNumber:38; MassNumber:90; Mass:89.9077300{28} ; Abundance:0; {g,r} StandardWeight:87.62{1}),
(Symbol:'Sr'; AtomicNumber:38; MassNumber:91; Mass:90.9101954{61} ; Abundance:0; {g,r} StandardWeight:87.62{1}),
(Symbol:'Sr'; AtomicNumber:38; MassNumber:92; Mass:91.9110382{37} ; Abundance:0; {g,r} StandardWeight:87.62{1}),
(Symbol:'Sr'; AtomicNumber:38; MassNumber:93; Mass:92.9140242{81} ; Abundance:0; {g,r} StandardWeight:87.62{1}),
(Symbol:'Sr'; AtomicNumber:38; MassNumber:94; Mass:93.9153556{18} ; Abundance:0; {g,r} StandardWeight:87.62{1}),
(Symbol:'Sr'; AtomicNumber:38; MassNumber:95; Mass:94.9193529{63} ; Abundance:0; {g,r} StandardWeight:87.62{1}),
(Symbol:'Sr'; AtomicNumber:38; MassNumber:96; Mass:95.9217066{93} ; Abundance:0; {g,r} StandardWeight:87.62{1}),
(Symbol:'Sr'; AtomicNumber:38; MassNumber:97; Mass:96.9263740{36} ; Abundance:0; {g,r} StandardWeight:87.62{1}),
(Symbol:'Sr'; AtomicNumber:38; MassNumber:98; Mass:97.9286888{40} ; Abundance:0; {g,r} StandardWeight:87.62{1}),
(Symbol:'Sr'; AtomicNumber:38; MassNumber:99; Mass:98.9328907{38} ; Abundance:0; {g,r} StandardWeight:87.62{1}),
(Symbol:'Sr'; AtomicNumber:38; MassNumber:100; Mass:99.935770{10} ; Abundance:0; {g,r} StandardWeight:87.62{1}),
(Symbol:'Sr'; AtomicNumber:38; MassNumber:101; Mass:100.940352{86} ; Abundance:0; {g,r} StandardWeight:87.62{1}),
(Symbol:'Sr'; AtomicNumber:38; MassNumber:102; Mass:101.943791{75} ; Abundance:0; {g,r} StandardWeight:87.62{1}),
(Symbol:'Sr'; AtomicNumber:38; MassNumber:103; Mass:102.94909{21#} ; Abundance:0; {g,r} StandardWeight:87.62{1}),
(Symbol:'Sr'; AtomicNumber:38; MassNumber:104; Mass:103.95265{32#} ; Abundance:0; {g,r} StandardWeight:87.62{1}),
(Symbol:'Sr'; AtomicNumber:38; MassNumber:105; Mass:104.95855{54#} ; Abundance:0; {g,r} StandardWeight:87.62{1}),
(Symbol:'Sr'; AtomicNumber:38; MassNumber:106; Mass:105.96265{64#} ; Abundance:0; {g,r} StandardWeight:87.62{1}),
(Symbol:'Sr'; AtomicNumber:38; MassNumber:107; Mass:106.96897{75#} ; Abundance:0; {g,r} StandardWeight:87.62{1}),
(Symbol:'Y'; AtomicNumber:39; MassNumber:76; Mass:75.95856{54#} ; Abundance:0;  StandardWeight:88.90584{2}),
(Symbol:'Y'; AtomicNumber:39; MassNumber:77; Mass:76.949781{65#} ; Abundance:0;  StandardWeight:88.90584{2}),
(Symbol:'Y'; AtomicNumber:39; MassNumber:78; Mass:77.94361{43#} ; Abundance:0;  StandardWeight:88.90584{2}),
(Symbol:'Y'; AtomicNumber:39; MassNumber:79; Mass:78.93735{48} ; Abundance:0;  StandardWeight:88.90584{2}),
(Symbol:'Y'; AtomicNumber:39; MassNumber:80; Mass:79.9343561{67} ; Abundance:0;  StandardWeight:88.90584{2}),
(Symbol:'Y'; AtomicNumber:39; MassNumber:81; Mass:80.9294556{58} ; Abundance:0;  StandardWeight:88.90584{2}),
(Symbol:'Y'; AtomicNumber:39; MassNumber:82; Mass:81.9269314{59} ; Abundance:0;  StandardWeight:88.90584{2}),
(Symbol:'Y'; AtomicNumber:39; MassNumber:83; Mass:82.922485{20} ; Abundance:0;  StandardWeight:88.90584{2}),
(Symbol:'Y'; AtomicNumber:39; MassNumber:84; Mass:83.9206721{46} ; Abundance:0;  StandardWeight:88.90584{2}),
(Symbol:'Y'; AtomicNumber:39; MassNumber:85; Mass:84.916433{20} ; Abundance:0;  StandardWeight:88.90584{2}),
(Symbol:'Y'; AtomicNumber:39; MassNumber:86; Mass:85.914886{15} ; Abundance:0;  StandardWeight:88.90584{2}),
(Symbol:'Y'; AtomicNumber:39; MassNumber:87; Mass:86.9108761{17} ; Abundance:0;  StandardWeight:88.90584{2}),
(Symbol:'Y'; AtomicNumber:39; MassNumber:88; Mass:87.9095016{20} ; Abundance:0;  StandardWeight:88.90584{2}),
(Symbol:'Y'; AtomicNumber:39; MassNumber:89; Mass:88.9058403{24} ; Abundance:100;  StandardWeight:88.90584{2}),
(Symbol:'Y'; AtomicNumber:39; MassNumber:90; Mass:89.9071439{24} ; Abundance:0;  StandardWeight:88.90584{2}),
(Symbol:'Y'; AtomicNumber:39; MassNumber:91; Mass:90.9072974{28} ; Abundance:0;  StandardWeight:88.90584{2}),
(Symbol:'Y'; AtomicNumber:39; MassNumber:92; Mass:91.9089451{99} ; Abundance:0;  StandardWeight:88.90584{2}),
(Symbol:'Y'; AtomicNumber:39; MassNumber:93; Mass:92.909578{11} ; Abundance:0;  StandardWeight:88.90584{2}),
(Symbol:'Y'; AtomicNumber:39; MassNumber:94; Mass:93.9115906{69} ; Abundance:0;  StandardWeight:88.90584{2}),
(Symbol:'Y'; AtomicNumber:39; MassNumber:95; Mass:94.9128161{74} ; Abundance:0;  StandardWeight:88.90584{2}),
(Symbol:'Y'; AtomicNumber:39; MassNumber:96; Mass:95.9158968{69} ; Abundance:0;  StandardWeight:88.90584{2}),
(Symbol:'Y'; AtomicNumber:39; MassNumber:97; Mass:96.9182741{75} ; Abundance:0;  StandardWeight:88.90584{2}),
(Symbol:'Y'; AtomicNumber:39; MassNumber:98; Mass:97.9223821{88} ; Abundance:0;  StandardWeight:88.90584{2}),
(Symbol:'Y'; AtomicNumber:39; MassNumber:99; Mass:98.9241480{74} ; Abundance:0;  StandardWeight:88.90584{2}),
(Symbol:'Y'; AtomicNumber:39; MassNumber:100; Mass:99.927715{12} ; Abundance:0;  StandardWeight:88.90584{2}),
(Symbol:'Y'; AtomicNumber:39; MassNumber:101; Mass:100.9301477{79} ; Abundance:0;  StandardWeight:88.90584{2}),
(Symbol:'Y'; AtomicNumber:39; MassNumber:102; Mass:101.9343277{44} ; Abundance:0;  StandardWeight:88.90584{2}),
(Symbol:'Y'; AtomicNumber:39; MassNumber:103; Mass:102.937243{12} ; Abundance:0;  StandardWeight:88.90584{2}),
(Symbol:'Y'; AtomicNumber:39; MassNumber:104; Mass:103.94196{43#} ; Abundance:0;  StandardWeight:88.90584{2}),
(Symbol:'Y'; AtomicNumber:39; MassNumber:105; Mass:104.94544{54#} ; Abundance:0;  StandardWeight:88.90584{2}),
(Symbol:'Y'; AtomicNumber:39; MassNumber:106; Mass:105.95056{54#} ; Abundance:0;  StandardWeight:88.90584{2}),
(Symbol:'Y'; AtomicNumber:39; MassNumber:107; Mass:106.95452{54#} ; Abundance:0;  StandardWeight:88.90584{2}),
(Symbol:'Y'; AtomicNumber:39; MassNumber:108; Mass:107.95996{64#} ; Abundance:0;  StandardWeight:88.90584{2}),
(Symbol:'Y'; AtomicNumber:39; MassNumber:109; Mass:108.96436{75#} ; Abundance:0;  StandardWeight:88.90584{2}),
(Symbol:'Zr'; AtomicNumber:40; MassNumber:78; Mass:77.95566{54#} ; Abundance:0; {g} StandardWeight:91.224{2}),
(Symbol:'Zr'; AtomicNumber:40; MassNumber:79; Mass:78.94948{43#} ; Abundance:0; {g} StandardWeight:91.224{2}),
(Symbol:'Zr'; AtomicNumber:40; MassNumber:80; Mass:79.9404{16} ; Abundance:0; {g} StandardWeight:91.224{2}),
(Symbol:'Zr'; AtomicNumber:40; MassNumber:81; Mass:80.93731{18} ; Abundance:0; {g} StandardWeight:91.224{2}),
(Symbol:'Zr'; AtomicNumber:40; MassNumber:82; Mass:81.93135{22#} ; Abundance:0; {g} StandardWeight:91.224{2}),
(Symbol:'Zr'; AtomicNumber:40; MassNumber:83; Mass:82.9292421{69} ; Abundance:0; {g} StandardWeight:91.224{2}),
(Symbol:'Zr'; AtomicNumber:40; MassNumber:84; Mass:83.9233269{59} ; Abundance:0; {g} StandardWeight:91.224{2}),
(Symbol:'Zr'; AtomicNumber:40; MassNumber:85; Mass:84.9214444{69} ; Abundance:0; {g} StandardWeight:91.224{2}),
(Symbol:'Zr'; AtomicNumber:40; MassNumber:86; Mass:85.9162972{38} ; Abundance:0; {g} StandardWeight:91.224{2}),
(Symbol:'Zr'; AtomicNumber:40; MassNumber:87; Mass:86.9148180{45} ; Abundance:0; {g} StandardWeight:91.224{2}),
(Symbol:'Zr'; AtomicNumber:40; MassNumber:88; Mass:87.9102213{58} ; Abundance:0; {g} StandardWeight:91.224{2}),
(Symbol:'Zr'; AtomicNumber:40; MassNumber:89; Mass:88.9088814{37} ; Abundance:0; {g} StandardWeight:91.224{2}),
(Symbol:'Zr'; AtomicNumber:40; MassNumber:90; Mass:89.9046977{20} ; Abundance:51.45; {g} StandardWeight:91.224{2}),
(Symbol:'Zr'; AtomicNumber:40; MassNumber:91; Mass:90.9056396{20} ; Abundance:11.22; {g} StandardWeight:91.224{2}),
(Symbol:'Zr'; AtomicNumber:40; MassNumber:92; Mass:91.9050347{20} ; Abundance:17.15; {g} StandardWeight:91.224{2}),
(Symbol:'Zr'; AtomicNumber:40; MassNumber:93; Mass:92.9064699{20} ; Abundance:0; {g} StandardWeight:91.224{2}),
(Symbol:'Zr'; AtomicNumber:40; MassNumber:94; Mass:93.9063108{20} ; Abundance:17.38; {g} StandardWeight:91.224{2}),
(Symbol:'Zr'; AtomicNumber:40; MassNumber:95; Mass:94.9080385{19} ; Abundance:0; {g} StandardWeight:91.224{2}),
(Symbol:'Zr'; AtomicNumber:40; MassNumber:96; Mass:95.9082714{21} ; Abundance:2.8; {g} StandardWeight:91.224{2}),
(Symbol:'Zr'; AtomicNumber:40; MassNumber:97; Mass:96.9109512{21} ; Abundance:0; {g} StandardWeight:91.224{2}),
(Symbol:'Zr'; AtomicNumber:40; MassNumber:98; Mass:97.9127289{93} ; Abundance:0; {g} StandardWeight:91.224{2}),
(Symbol:'Zr'; AtomicNumber:40; MassNumber:99; Mass:98.916667{11} ; Abundance:0; {g} StandardWeight:91.224{2}),
(Symbol:'Zr'; AtomicNumber:40; MassNumber:100; Mass:99.9180006{89} ; Abundance:0; {g} StandardWeight:91.224{2}),
(Symbol:'Zr'; AtomicNumber:40; MassNumber:101; Mass:100.9214480{91} ; Abundance:0; {g} StandardWeight:91.224{2}),
(Symbol:'Zr'; AtomicNumber:40; MassNumber:102; Mass:101.9231409{97} ; Abundance:0; {g} StandardWeight:91.224{2}),
(Symbol:'Zr'; AtomicNumber:40; MassNumber:103; Mass:102.927191{10} ; Abundance:0; {g} StandardWeight:91.224{2}),
(Symbol:'Zr'; AtomicNumber:40; MassNumber:104; Mass:103.929436{10} ; Abundance:0; {g} StandardWeight:91.224{2}),
(Symbol:'Zr'; AtomicNumber:40; MassNumber:105; Mass:104.934008{13} ; Abundance:0; {g} StandardWeight:91.224{2}),
(Symbol:'Zr'; AtomicNumber:40; MassNumber:106; Mass:105.93676{21#} ; Abundance:0; {g} StandardWeight:91.224{2}),
(Symbol:'Zr'; AtomicNumber:40; MassNumber:107; Mass:106.94174{32#} ; Abundance:0; {g} StandardWeight:91.224{2}),
(Symbol:'Zr'; AtomicNumber:40; MassNumber:108; Mass:107.94487{43#} ; Abundance:0; {g} StandardWeight:91.224{2}),
(Symbol:'Zr'; AtomicNumber:40; MassNumber:109; Mass:108.95041{54#} ; Abundance:0; {g} StandardWeight:91.224{2}),
(Symbol:'Zr'; AtomicNumber:40; MassNumber:110; Mass:109.95396{64#} ; Abundance:0; {g} StandardWeight:91.224{2}),
(Symbol:'Zr'; AtomicNumber:40; MassNumber:111; Mass:110.95968{75#} ; Abundance:0; {g} StandardWeight:91.224{2}),
(Symbol:'Zr'; AtomicNumber:40; MassNumber:112; Mass:111.96370{75#} ; Abundance:0; {g} StandardWeight:91.224{2}),
(Symbol:'Nb'; AtomicNumber:41; MassNumber:81; Mass:80.94960{43#} ; Abundance:0;  StandardWeight:92.90637{2}),
(Symbol:'Nb'; AtomicNumber:41; MassNumber:82; Mass:81.94396{32#} ; Abundance:0;  StandardWeight:92.90637{2}),
(Symbol:'Nb'; AtomicNumber:41; MassNumber:83; Mass:82.93729{32} ; Abundance:0;  StandardWeight:92.90637{2}),
(Symbol:'Nb'; AtomicNumber:41; MassNumber:84; Mass:83.93449{32#} ; Abundance:0;  StandardWeight:92.90637{2}),
(Symbol:'Nb'; AtomicNumber:41; MassNumber:85; Mass:84.9288458{44} ; Abundance:0;  StandardWeight:92.90637{2}),
(Symbol:'Nb'; AtomicNumber:41; MassNumber:86; Mass:85.9257828{59} ; Abundance:0;  StandardWeight:92.90637{2}),
(Symbol:'Nb'; AtomicNumber:41; MassNumber:87; Mass:86.9206937{73} ; Abundance:0;  StandardWeight:92.90637{2}),
(Symbol:'Nb'; AtomicNumber:41; MassNumber:88; Mass:87.918222{61} ; Abundance:0;  StandardWeight:92.90637{2}),
(Symbol:'Nb'; AtomicNumber:41; MassNumber:89; Mass:88.913445{25} ; Abundance:0;  StandardWeight:92.90637{2}),
(Symbol:'Nb'; AtomicNumber:41; MassNumber:90; Mass:89.9112584{38} ; Abundance:0;  StandardWeight:92.90637{2}),
(Symbol:'Nb'; AtomicNumber:41; MassNumber:91; Mass:90.9069897{37} ; Abundance:0;  StandardWeight:92.90637{2}),
(Symbol:'Nb'; AtomicNumber:41; MassNumber:92; Mass:91.9071881{26} ; Abundance:0;  StandardWeight:92.90637{2}),
(Symbol:'Nb'; AtomicNumber:41; MassNumber:93; Mass:92.9063730{20} ; Abundance:100;  StandardWeight:92.90637{2}),
(Symbol:'Nb'; AtomicNumber:41; MassNumber:94; Mass:93.9072788{20} ; Abundance:0;  StandardWeight:92.90637{2}),
(Symbol:'Nb'; AtomicNumber:41; MassNumber:95; Mass:94.90683240{71} ; Abundance:0;  StandardWeight:92.90637{2}),
(Symbol:'Nb'; AtomicNumber:41; MassNumber:96; Mass:95.9080973{35} ; Abundance:0;  StandardWeight:92.90637{2}),
(Symbol:'Nb'; AtomicNumber:41; MassNumber:97; Mass:96.9080959{19} ; Abundance:0;  StandardWeight:92.90637{2}),
(Symbol:'Nb'; AtomicNumber:41; MassNumber:98; Mass:97.9103265{58} ; Abundance:0;  StandardWeight:92.90637{2}),
(Symbol:'Nb'; AtomicNumber:41; MassNumber:99; Mass:98.911613{13} ; Abundance:0;  StandardWeight:92.90637{2}),
(Symbol:'Nb'; AtomicNumber:41; MassNumber:100; Mass:99.9143276{88} ; Abundance:0;  StandardWeight:92.90637{2}),
(Symbol:'Nb'; AtomicNumber:41; MassNumber:101; Mass:100.9153103{42} ; Abundance:0;  StandardWeight:92.90637{2}),
(Symbol:'Nb'; AtomicNumber:41; MassNumber:102; Mass:101.9180772{35} ; Abundance:0;  StandardWeight:92.90637{2}),
(Symbol:'Nb'; AtomicNumber:41; MassNumber:103; Mass:102.9194572{44} ; Abundance:0;  StandardWeight:92.90637{2}),
(Symbol:'Nb'; AtomicNumber:41; MassNumber:104; Mass:103.9228925{37} ; Abundance:0;  StandardWeight:92.90637{2}),
(Symbol:'Nb'; AtomicNumber:41; MassNumber:105; Mass:104.9249465{45} ; Abundance:0;  StandardWeight:92.90637{2}),
(Symbol:'Nb'; AtomicNumber:41; MassNumber:106; Mass:105.9289317{46} ; Abundance:0;  StandardWeight:92.90637{2}),
(Symbol:'Nb'; AtomicNumber:41; MassNumber:107; Mass:106.9315937{87} ; Abundance:0;  StandardWeight:92.90637{2}),
(Symbol:'Nb'; AtomicNumber:41; MassNumber:108; Mass:107.9360748{88} ; Abundance:0;  StandardWeight:92.90637{2}),
(Symbol:'Nb'; AtomicNumber:41; MassNumber:109; Mass:108.93922{56} ; Abundance:0;  StandardWeight:92.90637{2}),
(Symbol:'Nb'; AtomicNumber:41; MassNumber:110; Mass:109.94403{21#} ; Abundance:0;  StandardWeight:92.90637{2}),
(Symbol:'Nb'; AtomicNumber:41; MassNumber:111; Mass:110.94753{32#} ; Abundance:0;  StandardWeight:92.90637{2}),
(Symbol:'Nb'; AtomicNumber:41; MassNumber:112; Mass:111.95247{32#} ; Abundance:0;  StandardWeight:92.90637{2}),
(Symbol:'Nb'; AtomicNumber:41; MassNumber:113; Mass:112.95651{43#} ; Abundance:0;  StandardWeight:92.90637{2}),
(Symbol:'Nb'; AtomicNumber:41; MassNumber:114; Mass:113.96201{54#} ; Abundance:0;  StandardWeight:92.90637{2}),
(Symbol:'Nb'; AtomicNumber:41; MassNumber:115; Mass:114.96634{54#} ; Abundance:0;  StandardWeight:92.90637{2}),
(Symbol:'Mo'; AtomicNumber:42; MassNumber:83; Mass:82.94988{43#} ; Abundance:0; {g} StandardWeight:95.95{1}),
(Symbol:'Mo'; AtomicNumber:42; MassNumber:84; Mass:83.94149{43#} ; Abundance:0; {g} StandardWeight:95.95{1}),
(Symbol:'Mo'; AtomicNumber:42; MassNumber:85; Mass:84.938261{17} ; Abundance:0; {g} StandardWeight:95.95{1}),
(Symbol:'Mo'; AtomicNumber:42; MassNumber:86; Mass:85.9311748{40} ; Abundance:0; {g} StandardWeight:95.95{1}),
(Symbol:'Mo'; AtomicNumber:42; MassNumber:87; Mass:86.9281962{31} ; Abundance:0; {g} StandardWeight:95.95{1}),
(Symbol:'Mo'; AtomicNumber:42; MassNumber:88; Mass:87.9219678{41} ; Abundance:0; {g} StandardWeight:95.95{1}),
(Symbol:'Mo'; AtomicNumber:42; MassNumber:89; Mass:88.9194682{42} ; Abundance:0; {g} StandardWeight:95.95{1}),
(Symbol:'Mo'; AtomicNumber:42; MassNumber:90; Mass:89.9139309{38} ; Abundance:0; {g} StandardWeight:95.95{1}),
(Symbol:'Mo'; AtomicNumber:42; MassNumber:91; Mass:90.9117453{67} ; Abundance:0; {g} StandardWeight:95.95{1}),
(Symbol:'Mo'; AtomicNumber:42; MassNumber:92; Mass:91.90680796{84} ; Abundance:14.53; {g} StandardWeight:95.95{1}),
(Symbol:'Mo'; AtomicNumber:42; MassNumber:93; Mass:92.90680958{84} ; Abundance:0; {g} StandardWeight:95.95{1}),
(Symbol:'Mo'; AtomicNumber:42; MassNumber:94; Mass:93.90508490{48} ; Abundance:9.15; {g} StandardWeight:95.95{1}),
(Symbol:'Mo'; AtomicNumber:42; MassNumber:95; Mass:94.90583877{47} ; Abundance:15.84; {g} StandardWeight:95.95{1}),
(Symbol:'Mo'; AtomicNumber:42; MassNumber:96; Mass:95.90467612{47} ; Abundance:16.67; {g} StandardWeight:95.95{1}),
(Symbol:'Mo'; AtomicNumber:42; MassNumber:97; Mass:96.90601812{49} ; Abundance:9.6; {g} StandardWeight:95.95{1}),
(Symbol:'Mo'; AtomicNumber:42; MassNumber:98; Mass:97.90540482{49} ; Abundance:24.39; {g} StandardWeight:95.95{1}),
(Symbol:'Mo'; AtomicNumber:42; MassNumber:99; Mass:98.90770851{52} ; Abundance:0; {g} StandardWeight:95.95{1}),
(Symbol:'Mo'; AtomicNumber:42; MassNumber:100; Mass:99.9074718{11} ; Abundance:9.82; {g} StandardWeight:95.95{1}),
(Symbol:'Mo'; AtomicNumber:42; MassNumber:101; Mass:100.9103414{11} ; Abundance:0; {g} StandardWeight:95.95{1}),
(Symbol:'Mo'; AtomicNumber:42; MassNumber:102; Mass:101.9102834{91} ; Abundance:0; {g} StandardWeight:95.95{1}),
(Symbol:'Mo'; AtomicNumber:42; MassNumber:103; Mass:102.913079{10} ; Abundance:0; {g} StandardWeight:95.95{1}),
(Symbol:'Mo'; AtomicNumber:42; MassNumber:104; Mass:103.9137344{98} ; Abundance:0; {g} StandardWeight:95.95{1}),
(Symbol:'Mo'; AtomicNumber:42; MassNumber:105; Mass:104.916969{10} ; Abundance:0; {g} StandardWeight:95.95{1}),
(Symbol:'Mo'; AtomicNumber:42; MassNumber:106; Mass:105.918259{10} ; Abundance:0; {g} StandardWeight:95.95{1}),
(Symbol:'Mo'; AtomicNumber:42; MassNumber:107; Mass:106.922106{10} ; Abundance:0; {g} StandardWeight:95.95{1}),
(Symbol:'Mo'; AtomicNumber:42; MassNumber:108; Mass:107.924033{10} ; Abundance:0; {g} StandardWeight:95.95{1}),
(Symbol:'Mo'; AtomicNumber:42; MassNumber:109; Mass:108.928424{12} ; Abundance:0; {g} StandardWeight:95.95{1}),
(Symbol:'Mo'; AtomicNumber:42; MassNumber:110; Mass:109.930704{26} ; Abundance:0; {g} StandardWeight:95.95{1}),
(Symbol:'Mo'; AtomicNumber:42; MassNumber:111; Mass:110.935654{14} ; Abundance:0; {g} StandardWeight:95.95{1}),
(Symbol:'Mo'; AtomicNumber:42; MassNumber:112; Mass:111.93831{21#} ; Abundance:0; {g} StandardWeight:95.95{1}),
(Symbol:'Mo'; AtomicNumber:42; MassNumber:113; Mass:112.94335{32#} ; Abundance:0; {g} StandardWeight:95.95{1}),
(Symbol:'Mo'; AtomicNumber:42; MassNumber:114; Mass:113.94653{32#} ; Abundance:0; {g} StandardWeight:95.95{1}),
(Symbol:'Mo'; AtomicNumber:42; MassNumber:115; Mass:114.95196{43#} ; Abundance:0; {g} StandardWeight:95.95{1}),
(Symbol:'Mo'; AtomicNumber:42; MassNumber:116; Mass:115.95545{54#} ; Abundance:0; {g} StandardWeight:95.95{1}),
(Symbol:'Mo'; AtomicNumber:42; MassNumber:117; Mass:116.96117{54#} ; Abundance:0; {g} StandardWeight:95.95{1}),
(Symbol:'Tc'; AtomicNumber:43; MassNumber:85; Mass:84.95058{43#} ; Abundance:0;  StandardWeight:98),
(Symbol:'Tc'; AtomicNumber:43; MassNumber:86; Mass:85.94493{32#} ; Abundance:0;  StandardWeight:98),
(Symbol:'Tc'; AtomicNumber:43; MassNumber:87; Mass:86.9380672{45} ; Abundance:0;  StandardWeight:98),
(Symbol:'Tc'; AtomicNumber:43; MassNumber:88; Mass:87.93378{16} ; Abundance:0;  StandardWeight:98),
(Symbol:'Tc'; AtomicNumber:43; MassNumber:89; Mass:88.9276487{41} ; Abundance:0;  StandardWeight:98),
(Symbol:'Tc'; AtomicNumber:43; MassNumber:90; Mass:89.9240739{11} ; Abundance:0;  StandardWeight:98),
(Symbol:'Tc'; AtomicNumber:43; MassNumber:91; Mass:90.9184254{25} ; Abundance:0;  StandardWeight:98),
(Symbol:'Tc'; AtomicNumber:43; MassNumber:92; Mass:91.9152698{33} ; Abundance:0;  StandardWeight:98),
(Symbol:'Tc'; AtomicNumber:43; MassNumber:93; Mass:92.9102460{14} ; Abundance:0;  StandardWeight:98),
(Symbol:'Tc'; AtomicNumber:43; MassNumber:94; Mass:93.9096536{44} ; Abundance:0;  StandardWeight:98),
(Symbol:'Tc'; AtomicNumber:43; MassNumber:95; Mass:94.9076536{55} ; Abundance:0;  StandardWeight:98),
(Symbol:'Tc'; AtomicNumber:43; MassNumber:96; Mass:95.9078680{55} ; Abundance:0;  StandardWeight:98),
(Symbol:'Tc'; AtomicNumber:43; MassNumber:97; Mass:96.9063667{40} ; Abundance:0;  StandardWeight:98),
(Symbol:'Tc'; AtomicNumber:43; MassNumber:98; Mass:97.9072124{36} ; Abundance:0;  StandardWeight:98),
(Symbol:'Tc'; AtomicNumber:43; MassNumber:99; Mass:98.9062508{10} ; Abundance:0;  StandardWeight:98),
(Symbol:'Tc'; AtomicNumber:43; MassNumber:100; Mass:99.9076539{15} ; Abundance:0;  StandardWeight:98),
(Symbol:'Tc'; AtomicNumber:43; MassNumber:101; Mass:100.907309{26} ; Abundance:0;  StandardWeight:98),
(Symbol:'Tc'; AtomicNumber:43; MassNumber:102; Mass:101.9092097{99} ; Abundance:0;  StandardWeight:98),
(Symbol:'Tc'; AtomicNumber:43; MassNumber:103; Mass:102.909176{11} ; Abundance:0;  StandardWeight:98),
(Symbol:'Tc'; AtomicNumber:43; MassNumber:104; Mass:103.911425{27} ; Abundance:0;  StandardWeight:98),
(Symbol:'Tc'; AtomicNumber:43; MassNumber:105; Mass:104.911655{38} ; Abundance:0;  StandardWeight:98),
(Symbol:'Tc'; AtomicNumber:43; MassNumber:106; Mass:105.914358{13} ; Abundance:0;  StandardWeight:98),
(Symbol:'Tc'; AtomicNumber:43; MassNumber:107; Mass:106.9154606{93} ; Abundance:0;  StandardWeight:98),
(Symbol:'Tc'; AtomicNumber:43; MassNumber:108; Mass:107.9184957{94} ; Abundance:0;  StandardWeight:98),
(Symbol:'Tc'; AtomicNumber:43; MassNumber:109; Mass:108.920256{10} ; Abundance:0;  StandardWeight:98),
(Symbol:'Tc'; AtomicNumber:43; MassNumber:110; Mass:109.923744{10} ; Abundance:0;  StandardWeight:98),
(Symbol:'Tc'; AtomicNumber:43; MassNumber:111; Mass:110.925901{11} ; Abundance:0;  StandardWeight:98),
(Symbol:'Tc'; AtomicNumber:43; MassNumber:112; Mass:111.9299458{60} ; Abundance:0;  StandardWeight:98),
(Symbol:'Tc'; AtomicNumber:43; MassNumber:113; Mass:112.9325690{36} ; Abundance:0;  StandardWeight:98),
(Symbol:'Tc'; AtomicNumber:43; MassNumber:114; Mass:113.93691{11#} ; Abundance:0;  StandardWeight:98),
(Symbol:'Tc'; AtomicNumber:43; MassNumber:115; Mass:114.93998{21#} ; Abundance:0;  StandardWeight:98),
(Symbol:'Tc'; AtomicNumber:43; MassNumber:116; Mass:115.94476{32#} ; Abundance:0;  StandardWeight:98),
(Symbol:'Tc'; AtomicNumber:43; MassNumber:117; Mass:116.94806{43#} ; Abundance:0;  StandardWeight:98),
(Symbol:'Tc'; AtomicNumber:43; MassNumber:118; Mass:117.95299{43#} ; Abundance:0;  StandardWeight:98),
(Symbol:'Tc'; AtomicNumber:43; MassNumber:119; Mass:118.95666{54#} ; Abundance:0;  StandardWeight:98),
(Symbol:'Tc'; AtomicNumber:43; MassNumber:120; Mass:119.96187{54#} ; Abundance:0;  StandardWeight:98),
(Symbol:'Ru'; AtomicNumber:44; MassNumber:87; Mass:86.95069{43#} ; Abundance:0; {g} StandardWeight:101.07{2}),
(Symbol:'Ru'; AtomicNumber:44; MassNumber:88; Mass:87.94160{32#} ; Abundance:0; {g} StandardWeight:101.07{2}),
(Symbol:'Ru'; AtomicNumber:44; MassNumber:89; Mass:88.93762{32#} ; Abundance:0; {g} StandardWeight:101.07{2}),
(Symbol:'Ru'; AtomicNumber:44; MassNumber:90; Mass:89.9303444{40} ; Abundance:0; {g} StandardWeight:101.07{2}),
(Symbol:'Ru'; AtomicNumber:44; MassNumber:91; Mass:90.9267419{24} ; Abundance:0; {g} StandardWeight:101.07{2}),
(Symbol:'Ru'; AtomicNumber:44; MassNumber:92; Mass:91.9202344{29} ; Abundance:0; {g} StandardWeight:101.07{2}),
(Symbol:'Ru'; AtomicNumber:44; MassNumber:93; Mass:92.9171044{22} ; Abundance:0; {g} StandardWeight:101.07{2}),
(Symbol:'Ru'; AtomicNumber:44; MassNumber:94; Mass:93.9113429{34} ; Abundance:0; {g} StandardWeight:101.07{2}),
(Symbol:'Ru'; AtomicNumber:44; MassNumber:95; Mass:94.910406{10} ; Abundance:0; {g} StandardWeight:101.07{2}),
(Symbol:'Ru'; AtomicNumber:44; MassNumber:96; Mass:95.90759025{49} ; Abundance:5.54; {g} StandardWeight:101.07{2}),
(Symbol:'Ru'; AtomicNumber:44; MassNumber:97; Mass:96.9075471{30} ; Abundance:0; {g} StandardWeight:101.07{2}),
(Symbol:'Ru'; AtomicNumber:44; MassNumber:98; Mass:97.9052868{69} ; Abundance:1.87; {g} StandardWeight:101.07{2}),
(Symbol:'Ru'; AtomicNumber:44; MassNumber:99; Mass:98.9059341{11} ; Abundance:12.76; {g} StandardWeight:101.07{2}),
(Symbol:'Ru'; AtomicNumber:44; MassNumber:100; Mass:99.9042143{11} ; Abundance:12.6; {g} StandardWeight:101.07{2}),
(Symbol:'Ru'; AtomicNumber:44; MassNumber:101; Mass:100.9055769{12} ; Abundance:17.06; {g} StandardWeight:101.07{2}),
(Symbol:'Ru'; AtomicNumber:44; MassNumber:102; Mass:101.9043441{12} ; Abundance:31.55; {g} StandardWeight:101.07{2}),
(Symbol:'Ru'; AtomicNumber:44; MassNumber:103; Mass:102.9063186{12} ; Abundance:0; {g} StandardWeight:101.07{2}),
(Symbol:'Ru'; AtomicNumber:44; MassNumber:104; Mass:103.9054275{28} ; Abundance:18.62; {g} StandardWeight:101.07{2}),
(Symbol:'Ru'; AtomicNumber:44; MassNumber:105; Mass:104.9077476{28} ; Abundance:0; {g} StandardWeight:101.07{2}),
(Symbol:'Ru'; AtomicNumber:44; MassNumber:106; Mass:105.9073291{58} ; Abundance:0; {g} StandardWeight:101.07{2}),
(Symbol:'Ru'; AtomicNumber:44; MassNumber:107; Mass:106.9099720{93} ; Abundance:0; {g} StandardWeight:101.07{2}),
(Symbol:'Ru'; AtomicNumber:44; MassNumber:108; Mass:107.9101880{93} ; Abundance:0; {g} StandardWeight:101.07{2}),
(Symbol:'Ru'; AtomicNumber:44; MassNumber:109; Mass:108.9133260{96} ; Abundance:0; {g} StandardWeight:101.07{2}),
(Symbol:'Ru'; AtomicNumber:44; MassNumber:110; Mass:109.9140407{96} ; Abundance:0; {g} StandardWeight:101.07{2}),
(Symbol:'Ru'; AtomicNumber:44; MassNumber:111; Mass:110.917570{10} ; Abundance:0; {g} StandardWeight:101.07{2}),
(Symbol:'Ru'; AtomicNumber:44; MassNumber:112; Mass:111.918809{10} ; Abundance:0; {g} StandardWeight:101.07{2}),
(Symbol:'Ru'; AtomicNumber:44; MassNumber:113; Mass:112.922844{39} ; Abundance:0; {g} StandardWeight:101.07{2}),
(Symbol:'Ru'; AtomicNumber:44; MassNumber:114; Mass:113.9246136{38} ; Abundance:0; {g} StandardWeight:101.07{2}),
(Symbol:'Ru'; AtomicNumber:44; MassNumber:115; Mass:114.928820{71} ; Abundance:0; {g} StandardWeight:101.07{2}),
(Symbol:'Ru'; AtomicNumber:44; MassNumber:116; Mass:115.9312192{40} ; Abundance:0; {g} StandardWeight:101.07{2}),
(Symbol:'Ru'; AtomicNumber:44; MassNumber:117; Mass:116.93610{63} ; Abundance:0; {g} StandardWeight:101.07{2}),
(Symbol:'Ru'; AtomicNumber:44; MassNumber:118; Mass:117.93853{32#} ; Abundance:0; {g} StandardWeight:101.07{2}),
(Symbol:'Ru'; AtomicNumber:44; MassNumber:119; Mass:118.94357{32#} ; Abundance:0; {g} StandardWeight:101.07{2}),
(Symbol:'Ru'; AtomicNumber:44; MassNumber:120; Mass:119.94631{43#} ; Abundance:0; {g} StandardWeight:101.07{2}),
(Symbol:'Ru'; AtomicNumber:44; MassNumber:121; Mass:120.95164{43#} ; Abundance:0; {g} StandardWeight:101.07{2}),
(Symbol:'Ru'; AtomicNumber:44; MassNumber:122; Mass:121.95447{54#} ; Abundance:0; {g} StandardWeight:101.07{2}),
(Symbol:'Ru'; AtomicNumber:44; MassNumber:123; Mass:122.95989{54#} ; Abundance:0; {g} StandardWeight:101.07{2}),
(Symbol:'Ru'; AtomicNumber:44; MassNumber:124; Mass:123.96305{64#} ; Abundance:0; {g} StandardWeight:101.07{2}),
(Symbol:'Rh'; AtomicNumber:45; MassNumber:89; Mass:88.95058{39#} ; Abundance:0;  StandardWeight:102.90550{2}),
(Symbol:'Rh'; AtomicNumber:45; MassNumber:90; Mass:89.94422{43#} ; Abundance:0;  StandardWeight:102.90550{2}),
(Symbol:'Rh'; AtomicNumber:45; MassNumber:91; Mass:90.93688{43#} ; Abundance:0;  StandardWeight:102.90550{2}),
(Symbol:'Rh'; AtomicNumber:45; MassNumber:92; Mass:91.9323677{47} ; Abundance:0;  StandardWeight:102.90550{2}),
(Symbol:'Rh'; AtomicNumber:45; MassNumber:93; Mass:92.9259128{28} ; Abundance:0;  StandardWeight:102.90550{2}),
(Symbol:'Rh'; AtomicNumber:45; MassNumber:94; Mass:93.9217305{36} ; Abundance:0;  StandardWeight:102.90550{2}),
(Symbol:'Rh'; AtomicNumber:45; MassNumber:95; Mass:94.9158979{42} ; Abundance:0;  StandardWeight:102.90550{2}),
(Symbol:'Rh'; AtomicNumber:45; MassNumber:96; Mass:95.914453{11} ; Abundance:0;  StandardWeight:102.90550{2}),
(Symbol:'Rh'; AtomicNumber:45; MassNumber:97; Mass:96.911329{38} ; Abundance:0;  StandardWeight:102.90550{2}),
(Symbol:'Rh'; AtomicNumber:45; MassNumber:98; Mass:97.910708{13} ; Abundance:0;  StandardWeight:102.90550{2}),
(Symbol:'Rh'; AtomicNumber:45; MassNumber:99; Mass:98.9081282{73} ; Abundance:0;  StandardWeight:102.90550{2}),
(Symbol:'Rh'; AtomicNumber:45; MassNumber:100; Mass:99.908117{19} ; Abundance:0;  StandardWeight:102.90550{2}),
(Symbol:'Rh'; AtomicNumber:45; MassNumber:101; Mass:100.9061606{63} ; Abundance:0;  StandardWeight:102.90550{2}),
(Symbol:'Rh'; AtomicNumber:45; MassNumber:102; Mass:101.9068374{50} ; Abundance:0;  StandardWeight:102.90550{2}),
(Symbol:'Rh'; AtomicNumber:45; MassNumber:103; Mass:102.9054980{26} ; Abundance:100;  StandardWeight:102.90550{2}),
(Symbol:'Rh'; AtomicNumber:45; MassNumber:104; Mass:103.9066492{26} ; Abundance:0;  StandardWeight:102.90550{2}),
(Symbol:'Rh'; AtomicNumber:45; MassNumber:105; Mass:104.9056885{27} ; Abundance:0;  StandardWeight:102.90550{2}),
(Symbol:'Rh'; AtomicNumber:45; MassNumber:106; Mass:105.9072868{58} ; Abundance:0;  StandardWeight:102.90550{2}),
(Symbol:'Rh'; AtomicNumber:45; MassNumber:107; Mass:106.906748{13} ; Abundance:0;  StandardWeight:102.90550{2}),
(Symbol:'Rh'; AtomicNumber:45; MassNumber:108; Mass:107.908714{15} ; Abundance:0;  StandardWeight:102.90550{2}),
(Symbol:'Rh'; AtomicNumber:45; MassNumber:109; Mass:108.9087488{43} ; Abundance:0;  StandardWeight:102.90550{2}),
(Symbol:'Rh'; AtomicNumber:45; MassNumber:110; Mass:109.911079{19} ; Abundance:0;  StandardWeight:102.90550{2}),
(Symbol:'Rh'; AtomicNumber:45; MassNumber:111; Mass:110.9116423{74} ; Abundance:0;  StandardWeight:102.90550{2}),
(Symbol:'Rh'; AtomicNumber:45; MassNumber:112; Mass:111.914403{47} ; Abundance:0;  StandardWeight:102.90550{2}),
(Symbol:'Rh'; AtomicNumber:45; MassNumber:113; Mass:112.9154393{77} ; Abundance:0;  StandardWeight:102.90550{2}),
(Symbol:'Rh'; AtomicNumber:45; MassNumber:114; Mass:113.918718{77} ; Abundance:0;  StandardWeight:102.90550{2}),
(Symbol:'Rh'; AtomicNumber:45; MassNumber:115; Mass:114.9203116{78} ; Abundance:0;  StandardWeight:102.90550{2}),
(Symbol:'Rh'; AtomicNumber:45; MassNumber:116; Mass:115.924059{76} ; Abundance:0;  StandardWeight:102.90550{2}),
(Symbol:'Rh'; AtomicNumber:45; MassNumber:117; Mass:116.9260354{95} ; Abundance:0;  StandardWeight:102.90550{2}),
(Symbol:'Rh'; AtomicNumber:45; MassNumber:118; Mass:117.930340{26} ; Abundance:0;  StandardWeight:102.90550{2}),
(Symbol:'Rh'; AtomicNumber:45; MassNumber:119; Mass:118.932557{10} ; Abundance:0;  StandardWeight:102.90550{2}),
(Symbol:'Rh'; AtomicNumber:45; MassNumber:120; Mass:119.93686{21#} ; Abundance:0;  StandardWeight:102.90550{2}),
(Symbol:'Rh'; AtomicNumber:45; MassNumber:121; Mass:120.93942{32#} ; Abundance:0;  StandardWeight:102.90550{2}),
(Symbol:'Rh'; AtomicNumber:45; MassNumber:122; Mass:121.94399{32#} ; Abundance:0;  StandardWeight:102.90550{2}),
(Symbol:'Rh'; AtomicNumber:45; MassNumber:123; Mass:122.94685{43#} ; Abundance:0;  StandardWeight:102.90550{2}),
(Symbol:'Rh'; AtomicNumber:45; MassNumber:124; Mass:123.95151{43#} ; Abundance:0;  StandardWeight:102.90550{2}),
(Symbol:'Rh'; AtomicNumber:45; MassNumber:125; Mass:124.95469{54#} ; Abundance:0;  StandardWeight:102.90550{2}),
(Symbol:'Rh'; AtomicNumber:45; MassNumber:126; Mass:125.95946{54#} ; Abundance:0;  StandardWeight:102.90550{2}),
(Symbol:'Pd'; AtomicNumber:46; MassNumber:91; Mass:90.95032{54#} ; Abundance:0; {g} StandardWeight:106.42{1}),
(Symbol:'Pd'; AtomicNumber:46; MassNumber:92; Mass:91.94088{54#} ; Abundance:0; {g} StandardWeight:106.42{1}),
(Symbol:'Pd'; AtomicNumber:46; MassNumber:93; Mass:92.93651{43#} ; Abundance:0; {g} StandardWeight:106.42{1}),
(Symbol:'Pd'; AtomicNumber:46; MassNumber:94; Mass:93.9290376{46} ; Abundance:0; {g} StandardWeight:106.42{1}),
(Symbol:'Pd'; AtomicNumber:46; MassNumber:95; Mass:94.9248898{33} ; Abundance:0; {g} StandardWeight:106.42{1}),
(Symbol:'Pd'; AtomicNumber:46; MassNumber:96; Mass:95.9182151{45} ; Abundance:0; {g} StandardWeight:106.42{1}),
(Symbol:'Pd'; AtomicNumber:46; MassNumber:97; Mass:96.9164720{52} ; Abundance:0; {g} StandardWeight:106.42{1}),
(Symbol:'Pd'; AtomicNumber:46; MassNumber:98; Mass:97.9126983{51} ; Abundance:0; {g} StandardWeight:106.42{1}),
(Symbol:'Pd'; AtomicNumber:46; MassNumber:99; Mass:98.9117748{54} ; Abundance:0; {g} StandardWeight:106.42{1}),
(Symbol:'Pd'; AtomicNumber:46; MassNumber:100; Mass:99.908505{19} ; Abundance:0; {g} StandardWeight:106.42{1}),
(Symbol:'Pd'; AtomicNumber:46; MassNumber:101; Mass:100.9082864{49} ; Abundance:0; {g} StandardWeight:106.42{1}),
(Symbol:'Pd'; AtomicNumber:46; MassNumber:102; Mass:101.9056022{28} ; Abundance:1.02; {g} StandardWeight:106.42{1}),
(Symbol:'Pd'; AtomicNumber:46; MassNumber:103; Mass:102.9060809{27} ; Abundance:0; {g} StandardWeight:106.42{1}),
(Symbol:'Pd'; AtomicNumber:46; MassNumber:104; Mass:103.9040305{14} ; Abundance:11.14; {g} StandardWeight:106.42{1}),
(Symbol:'Pd'; AtomicNumber:46; MassNumber:105; Mass:104.9050796{12} ; Abundance:22.33; {g} StandardWeight:106.42{1}),
(Symbol:'Pd'; AtomicNumber:46; MassNumber:106; Mass:105.9034804{12} ; Abundance:27.33; {g} StandardWeight:106.42{1}),
(Symbol:'Pd'; AtomicNumber:46; MassNumber:107; Mass:106.9051282{13} ; Abundance:0; {g} StandardWeight:106.42{1}),
(Symbol:'Pd'; AtomicNumber:46; MassNumber:108; Mass:107.9038916{12} ; Abundance:26.46; {g} StandardWeight:106.42{1}),
(Symbol:'Pd'; AtomicNumber:46; MassNumber:109; Mass:108.9059504{12} ; Abundance:0; {g} StandardWeight:106.42{1}),
(Symbol:'Pd'; AtomicNumber:46; MassNumber:110; Mass:109.90517220{75} ; Abundance:11.72; {g} StandardWeight:106.42{1}),
(Symbol:'Pd'; AtomicNumber:46; MassNumber:111; Mass:110.90768968{86} ; Abundance:0; {g} StandardWeight:106.42{1}),
(Symbol:'Pd'; AtomicNumber:46; MassNumber:112; Mass:111.9073297{70} ; Abundance:0; {g} StandardWeight:106.42{1}),
(Symbol:'Pd'; AtomicNumber:46; MassNumber:113; Mass:112.9102610{75} ; Abundance:0; {g} StandardWeight:106.42{1}),
(Symbol:'Pd'; AtomicNumber:46; MassNumber:114; Mass:113.9103686{75} ; Abundance:0; {g} StandardWeight:106.42{1}),
(Symbol:'Pd'; AtomicNumber:46; MassNumber:115; Mass:114.913659{15} ; Abundance:0; {g} StandardWeight:106.42{1}),
(Symbol:'Pd'; AtomicNumber:46; MassNumber:116; Mass:115.9142970{77} ; Abundance:0; {g} StandardWeight:106.42{1}),
(Symbol:'Pd'; AtomicNumber:46; MassNumber:117; Mass:116.9179547{78} ; Abundance:0; {g} StandardWeight:106.42{1}),
(Symbol:'Pd'; AtomicNumber:46; MassNumber:118; Mass:117.9190667{27} ; Abundance:0; {g} StandardWeight:106.42{1}),
(Symbol:'Pd'; AtomicNumber:46; MassNumber:119; Mass:118.9233402{89} ; Abundance:0; {g} StandardWeight:106.42{1}),
(Symbol:'Pd'; AtomicNumber:46; MassNumber:120; Mass:119.9245511{25} ; Abundance:0; {g} StandardWeight:106.42{1}),
(Symbol:'Pd'; AtomicNumber:46; MassNumber:121; Mass:120.9289503{36} ; Abundance:0; {g} StandardWeight:106.42{1}),
(Symbol:'Pd'; AtomicNumber:46; MassNumber:122; Mass:121.930632{21} ; Abundance:0; {g} StandardWeight:106.42{1}),
(Symbol:'Pd'; AtomicNumber:46; MassNumber:123; Mass:122.93514{21#} ; Abundance:0; {g} StandardWeight:106.42{1}),
(Symbol:'Pd'; AtomicNumber:46; MassNumber:124; Mass:123.93714{32#} ; Abundance:0; {g} StandardWeight:106.42{1}),
(Symbol:'Pd'; AtomicNumber:46; MassNumber:125; Mass:124.94179{43#} ; Abundance:0; {g} StandardWeight:106.42{1}),
(Symbol:'Pd'; AtomicNumber:46; MassNumber:126; Mass:125.94416{54#} ; Abundance:0; {g} StandardWeight:106.42{1}),
(Symbol:'Pd'; AtomicNumber:46; MassNumber:127; Mass:126.94907{54#} ; Abundance:0; {g} StandardWeight:106.42{1}),
(Symbol:'Pd'; AtomicNumber:46; MassNumber:128; Mass:127.95183{64#} ; Abundance:0; {g} StandardWeight:106.42{1}),
(Symbol:'Ag'; AtomicNumber:47; MassNumber:93; Mass:92.95033{54#} ; Abundance:0; {g} StandardWeight:107.8682{2}),
(Symbol:'Ag'; AtomicNumber:47; MassNumber:94; Mass:93.94373{69#} ; Abundance:0; {g} StandardWeight:107.8682{2}),
(Symbol:'Ag'; AtomicNumber:47; MassNumber:95; Mass:94.93602{43#} ; Abundance:0; {g} StandardWeight:107.8682{2}),
(Symbol:'Ag'; AtomicNumber:47; MassNumber:96; Mass:95.930744{97} ; Abundance:0; {g} StandardWeight:107.8682{2}),
(Symbol:'Ag'; AtomicNumber:47; MassNumber:97; Mass:96.92397{12} ; Abundance:0; {g} StandardWeight:107.8682{2}),
(Symbol:'Ag'; AtomicNumber:47; MassNumber:98; Mass:97.921560{35} ; Abundance:0; {g} StandardWeight:107.8682{2}),
(Symbol:'Ag'; AtomicNumber:47; MassNumber:99; Mass:98.9176458{67} ; Abundance:0; {g} StandardWeight:107.8682{2}),
(Symbol:'Ag'; AtomicNumber:47; MassNumber:100; Mass:99.9161154{54} ; Abundance:0; {g} StandardWeight:107.8682{2}),
(Symbol:'Ag'; AtomicNumber:47; MassNumber:101; Mass:100.9126840{52} ; Abundance:0; {g} StandardWeight:107.8682{2}),
(Symbol:'Ag'; AtomicNumber:47; MassNumber:102; Mass:101.9117047{88} ; Abundance:0; {g} StandardWeight:107.8682{2}),
(Symbol:'Ag'; AtomicNumber:47; MassNumber:103; Mass:102.9089631{41} ; Abundance:0; {g} StandardWeight:107.8682{2}),
(Symbol:'Ag'; AtomicNumber:47; MassNumber:104; Mass:103.9086239{45} ; Abundance:0; {g} StandardWeight:107.8682{2}),
(Symbol:'Ag'; AtomicNumber:47; MassNumber:105; Mass:104.9065256{49} ; Abundance:0; {g} StandardWeight:107.8682{2}),
(Symbol:'Ag'; AtomicNumber:47; MassNumber:106; Mass:105.9066636{32} ; Abundance:0; {g} StandardWeight:107.8682{2}),
(Symbol:'Ag'; AtomicNumber:47; MassNumber:107; Mass:106.9050916{26} ; Abundance:51.839; {g} StandardWeight:107.8682{2}),
(Symbol:'Ag'; AtomicNumber:47; MassNumber:108; Mass:107.9059503{26} ; Abundance:0; {g} StandardWeight:107.8682{2}),
(Symbol:'Ag'; AtomicNumber:47; MassNumber:109; Mass:108.9047553{14} ; Abundance:48.161; {g} StandardWeight:107.8682{2}),
(Symbol:'Ag'; AtomicNumber:47; MassNumber:110; Mass:109.9061102{14} ; Abundance:0; {g} StandardWeight:107.8682{2}),
(Symbol:'Ag'; AtomicNumber:47; MassNumber:111; Mass:110.9052959{16} ; Abundance:0; {g} StandardWeight:107.8682{2}),
(Symbol:'Ag'; AtomicNumber:47; MassNumber:112; Mass:111.9070486{26} ; Abundance:0; {g} StandardWeight:107.8682{2}),
(Symbol:'Ag'; AtomicNumber:47; MassNumber:113; Mass:112.906573{18} ; Abundance:0; {g} StandardWeight:107.8682{2}),
(Symbol:'Ag'; AtomicNumber:47; MassNumber:114; Mass:113.9088230{49} ; Abundance:0; {g} StandardWeight:107.8682{2}),
(Symbol:'Ag'; AtomicNumber:47; MassNumber:115; Mass:114.908767{20} ; Abundance:0; {g} StandardWeight:107.8682{2}),
(Symbol:'Ag'; AtomicNumber:47; MassNumber:116; Mass:115.9113868{35} ; Abundance:0; {g} StandardWeight:107.8682{2}),
(Symbol:'Ag'; AtomicNumber:47; MassNumber:117; Mass:116.911774{15} ; Abundance:0; {g} StandardWeight:107.8682{2}),
(Symbol:'Ag'; AtomicNumber:47; MassNumber:118; Mass:117.9145955{27} ; Abundance:0; {g} StandardWeight:107.8682{2}),
(Symbol:'Ag'; AtomicNumber:47; MassNumber:119; Mass:118.915570{16} ; Abundance:0; {g} StandardWeight:107.8682{2}),
(Symbol:'Ag'; AtomicNumber:47; MassNumber:120; Mass:119.9187848{48} ; Abundance:0; {g} StandardWeight:107.8682{2}),
(Symbol:'Ag'; AtomicNumber:47; MassNumber:121; Mass:120.920125{13} ; Abundance:0; {g} StandardWeight:107.8682{2}),
(Symbol:'Ag'; AtomicNumber:47; MassNumber:122; Mass:121.923664{41} ; Abundance:0; {g} StandardWeight:107.8682{2}),
(Symbol:'Ag'; AtomicNumber:47; MassNumber:123; Mass:122.925337{33} ; Abundance:0; {g} StandardWeight:107.8682{2}),
(Symbol:'Ag'; AtomicNumber:47; MassNumber:124; Mass:123.92893{27} ; Abundance:0; {g} StandardWeight:107.8682{2}),
(Symbol:'Ag'; AtomicNumber:47; MassNumber:125; Mass:124.93105{64} ; Abundance:0; {g} StandardWeight:107.8682{2}),
(Symbol:'Ag'; AtomicNumber:47; MassNumber:126; Mass:125.93475{21#} ; Abundance:0; {g} StandardWeight:107.8682{2}),
(Symbol:'Ag'; AtomicNumber:47; MassNumber:127; Mass:126.93711{21#} ; Abundance:0; {g} StandardWeight:107.8682{2}),
(Symbol:'Ag'; AtomicNumber:47; MassNumber:128; Mass:127.94106{32#} ; Abundance:0; {g} StandardWeight:107.8682{2}),
(Symbol:'Ag'; AtomicNumber:47; MassNumber:129; Mass:128.94395{32#} ; Abundance:0; {g} StandardWeight:107.8682{2}),
(Symbol:'Ag'; AtomicNumber:47; MassNumber:130; Mass:129.95070{36#} ; Abundance:0; {g} StandardWeight:107.8682{2}),
(Symbol:'Cd'; AtomicNumber:48; MassNumber:95; Mass:94.94994{54#} ; Abundance:0; {g} StandardWeight:112.414{4}),
(Symbol:'Cd'; AtomicNumber:48; MassNumber:96; Mass:95.94034{43#} ; Abundance:0; {g} StandardWeight:112.414{4}),
(Symbol:'Cd'; AtomicNumber:48; MassNumber:97; Mass:96.93510{32#} ; Abundance:0; {g} StandardWeight:112.414{4}),
(Symbol:'Cd'; AtomicNumber:48; MassNumber:98; Mass:97.927389{56} ; Abundance:0; {g} StandardWeight:112.414{4}),
(Symbol:'Cd'; AtomicNumber:48; MassNumber:99; Mass:98.9249258{17} ; Abundance:0; {g} StandardWeight:112.414{4}),
(Symbol:'Cd'; AtomicNumber:48; MassNumber:100; Mass:99.9203488{18} ; Abundance:0; {g} StandardWeight:112.414{4}),
(Symbol:'Cd'; AtomicNumber:48; MassNumber:101; Mass:100.9185862{16} ; Abundance:0; {g} StandardWeight:112.414{4}),
(Symbol:'Cd'; AtomicNumber:48; MassNumber:102; Mass:101.9144820{18} ; Abundance:0; {g} StandardWeight:112.414{4}),
(Symbol:'Cd'; AtomicNumber:48; MassNumber:103; Mass:102.9134165{19} ; Abundance:0; {g} StandardWeight:112.414{4}),
(Symbol:'Cd'; AtomicNumber:48; MassNumber:104; Mass:103.9098564{18} ; Abundance:0; {g} StandardWeight:112.414{4}),
(Symbol:'Cd'; AtomicNumber:48; MassNumber:105; Mass:104.9094639{15} ; Abundance:0; {g} StandardWeight:112.414{4}),
(Symbol:'Cd'; AtomicNumber:48; MassNumber:106; Mass:105.9064599{12} ; Abundance:1.25; {g} StandardWeight:112.414{4}),
(Symbol:'Cd'; AtomicNumber:48; MassNumber:107; Mass:106.9066121{18} ; Abundance:0; {g} StandardWeight:112.414{4}),
(Symbol:'Cd'; AtomicNumber:48; MassNumber:108; Mass:107.9041834{12} ; Abundance:0.89; {g} StandardWeight:112.414{4}),
(Symbol:'Cd'; AtomicNumber:48; MassNumber:109; Mass:108.9049867{17} ; Abundance:0; {g} StandardWeight:112.414{4}),
(Symbol:'Cd'; AtomicNumber:48; MassNumber:110; Mass:109.90300661{61} ; Abundance:12.49; {g} StandardWeight:112.414{4}),
(Symbol:'Cd'; AtomicNumber:48; MassNumber:111; Mass:110.90418287{61} ; Abundance:12.8; {g} StandardWeight:112.414{4}),
(Symbol:'Cd'; AtomicNumber:48; MassNumber:112; Mass:111.90276287{60} ; Abundance:24.13; {g} StandardWeight:112.414{4}),
(Symbol:'Cd'; AtomicNumber:48; MassNumber:113; Mass:112.90440813{45} ; Abundance:12.22; {g} StandardWeight:112.414{4}),
(Symbol:'Cd'; AtomicNumber:48; MassNumber:114; Mass:113.90336509{43} ; Abundance:28.73; {g} StandardWeight:112.414{4}),
(Symbol:'Cd'; AtomicNumber:48; MassNumber:115; Mass:114.90543751{77} ; Abundance:0; {g} StandardWeight:112.414{4}),
(Symbol:'Cd'; AtomicNumber:48; MassNumber:116; Mass:115.90476315{17} ; Abundance:7.49; {g} StandardWeight:112.414{4}),
(Symbol:'Cd'; AtomicNumber:48; MassNumber:117; Mass:116.9072260{11} ; Abundance:0; {g} StandardWeight:112.414{4}),
(Symbol:'Cd'; AtomicNumber:48; MassNumber:118; Mass:117.906922{21} ; Abundance:0; {g} StandardWeight:112.414{4}),
(Symbol:'Cd'; AtomicNumber:48; MassNumber:119; Mass:118.909847{40} ; Abundance:0; {g} StandardWeight:112.414{4}),
(Symbol:'Cd'; AtomicNumber:48; MassNumber:120; Mass:119.9098681{40} ; Abundance:0; {g} StandardWeight:112.414{4}),
(Symbol:'Cd'; AtomicNumber:48; MassNumber:121; Mass:120.9129637{21} ; Abundance:0; {g} StandardWeight:112.414{4}),
(Symbol:'Cd'; AtomicNumber:48; MassNumber:122; Mass:121.9134591{25} ; Abundance:0; {g} StandardWeight:112.414{4}),
(Symbol:'Cd'; AtomicNumber:48; MassNumber:123; Mass:122.9168925{29} ; Abundance:0; {g} StandardWeight:112.414{4}),
(Symbol:'Cd'; AtomicNumber:48; MassNumber:124; Mass:123.9176574{32} ; Abundance:0; {g} StandardWeight:112.414{4}),
(Symbol:'Cd'; AtomicNumber:48; MassNumber:125; Mass:124.9212576{31} ; Abundance:0; {g} StandardWeight:112.414{4}),
(Symbol:'Cd'; AtomicNumber:48; MassNumber:126; Mass:125.9224291{27} ; Abundance:0; {g} StandardWeight:112.414{4}),
(Symbol:'Cd'; AtomicNumber:48; MassNumber:127; Mass:126.926472{14} ; Abundance:0; {g} StandardWeight:112.414{4}),
(Symbol:'Cd'; AtomicNumber:48; MassNumber:128; Mass:127.9278129{78} ; Abundance:0; {g} StandardWeight:112.414{4}),
(Symbol:'Cd'; AtomicNumber:48; MassNumber:129; Mass:128.93182{21#} ; Abundance:0; {g} StandardWeight:112.414{4}),
(Symbol:'Cd'; AtomicNumber:48; MassNumber:130; Mass:129.93394{18} ; Abundance:0; {g} StandardWeight:112.414{4}),
(Symbol:'Cd'; AtomicNumber:48; MassNumber:131; Mass:130.94060{21#} ; Abundance:0; {g} StandardWeight:112.414{4}),
(Symbol:'Cd'; AtomicNumber:48; MassNumber:132; Mass:131.94604{21#} ; Abundance:0; {g} StandardWeight:112.414{4}),
(Symbol:'Cd'; AtomicNumber:48; MassNumber:133; Mass:132.95285{32#} ; Abundance:0; {g} StandardWeight:112.414{4}),
(Symbol:'In'; AtomicNumber:49; MassNumber:97; Mass:96.94934{54#} ; Abundance:0;  StandardWeight:114.818{1}),
(Symbol:'In'; AtomicNumber:49; MassNumber:98; Mass:97.94214{21#} ; Abundance:0;  StandardWeight:114.818{1}),
(Symbol:'In'; AtomicNumber:49; MassNumber:99; Mass:98.93411{21#} ; Abundance:0;  StandardWeight:114.818{1}),
(Symbol:'In'; AtomicNumber:49; MassNumber:100; Mass:99.93096{20} ; Abundance:0;  StandardWeight:114.818{1}),
(Symbol:'In'; AtomicNumber:49; MassNumber:101; Mass:100.92634{32#} ; Abundance:0;  StandardWeight:114.818{1}),
(Symbol:'In'; AtomicNumber:49; MassNumber:102; Mass:101.9241071{49} ; Abundance:0;  StandardWeight:114.818{1}),
(Symbol:'In'; AtomicNumber:49; MassNumber:103; Mass:102.9198819{98} ; Abundance:0;  StandardWeight:114.818{1}),
(Symbol:'In'; AtomicNumber:49; MassNumber:104; Mass:103.9182145{62} ; Abundance:0;  StandardWeight:114.818{1}),
(Symbol:'In'; AtomicNumber:49; MassNumber:105; Mass:104.914502{11} ; Abundance:0;  StandardWeight:114.818{1}),
(Symbol:'In'; AtomicNumber:49; MassNumber:106; Mass:105.913464{13} ; Abundance:0;  StandardWeight:114.818{1}),
(Symbol:'In'; AtomicNumber:49; MassNumber:107; Mass:106.910290{12} ; Abundance:0;  StandardWeight:114.818{1}),
(Symbol:'In'; AtomicNumber:49; MassNumber:108; Mass:107.9096935{93} ; Abundance:0;  StandardWeight:114.818{1}),
(Symbol:'In'; AtomicNumber:49; MassNumber:109; Mass:108.9071514{43} ; Abundance:0;  StandardWeight:114.818{1}),
(Symbol:'In'; AtomicNumber:49; MassNumber:110; Mass:109.907170{12} ; Abundance:0;  StandardWeight:114.818{1}),
(Symbol:'In'; AtomicNumber:49; MassNumber:111; Mass:110.9051085{38} ; Abundance:0;  StandardWeight:114.818{1}),
(Symbol:'In'; AtomicNumber:49; MassNumber:112; Mass:111.9055377{46} ; Abundance:0;  StandardWeight:114.818{1}),
(Symbol:'In'; AtomicNumber:49; MassNumber:113; Mass:112.90406184{91} ; Abundance:4.29;  StandardWeight:114.818{1}),
(Symbol:'In'; AtomicNumber:49; MassNumber:114; Mass:113.90491791{94} ; Abundance:0;  StandardWeight:114.818{1}),
(Symbol:'In'; AtomicNumber:49; MassNumber:115; Mass:114.903878776{12} ; Abundance:95.71;  StandardWeight:114.818{1}),
(Symbol:'In'; AtomicNumber:49; MassNumber:116; Mass:115.90525999{24} ; Abundance:0;  StandardWeight:114.818{1}),
(Symbol:'In'; AtomicNumber:49; MassNumber:117; Mass:116.9045157{52} ; Abundance:0;  StandardWeight:114.818{1}),
(Symbol:'In'; AtomicNumber:49; MassNumber:118; Mass:117.9063566{83} ; Abundance:0;  StandardWeight:114.818{1}),
(Symbol:'In'; AtomicNumber:49; MassNumber:119; Mass:118.9058507{78} ; Abundance:0;  StandardWeight:114.818{1}),
(Symbol:'In'; AtomicNumber:49; MassNumber:120; Mass:119.907967{43} ; Abundance:0;  StandardWeight:114.818{1}),
(Symbol:'In'; AtomicNumber:49; MassNumber:121; Mass:120.907851{29} ; Abundance:0;  StandardWeight:114.818{1}),
(Symbol:'In'; AtomicNumber:49; MassNumber:122; Mass:121.910281{54} ; Abundance:0;  StandardWeight:114.818{1}),
(Symbol:'In'; AtomicNumber:49; MassNumber:123; Mass:122.910434{21} ; Abundance:0;  StandardWeight:114.818{1}),
(Symbol:'In'; AtomicNumber:49; MassNumber:124; Mass:123.913182{33} ; Abundance:0;  StandardWeight:114.818{1}),
(Symbol:'In'; AtomicNumber:49; MassNumber:125; Mass:124.913605{29} ; Abundance:0;  StandardWeight:114.818{1}),
(Symbol:'In'; AtomicNumber:49; MassNumber:126; Mass:125.916507{29} ; Abundance:0;  StandardWeight:114.818{1}),
(Symbol:'In'; AtomicNumber:49; MassNumber:127; Mass:126.917446{23} ; Abundance:0;  StandardWeight:114.818{1}),
(Symbol:'In'; AtomicNumber:49; MassNumber:128; Mass:127.92040{16} ; Abundance:0;  StandardWeight:114.818{1}),
(Symbol:'In'; AtomicNumber:49; MassNumber:129; Mass:128.9218053{29} ; Abundance:0;  StandardWeight:114.818{1}),
(Symbol:'In'; AtomicNumber:49; MassNumber:130; Mass:129.924977{41} ; Abundance:0;  StandardWeight:114.818{1}),
(Symbol:'In'; AtomicNumber:49; MassNumber:131; Mass:130.9269715{29} ; Abundance:0;  StandardWeight:114.818{1}),
(Symbol:'In'; AtomicNumber:49; MassNumber:132; Mass:131.933001{64} ; Abundance:0;  StandardWeight:114.818{1}),
(Symbol:'In'; AtomicNumber:49; MassNumber:133; Mass:132.93831{21#} ; Abundance:0;  StandardWeight:114.818{1}),
(Symbol:'In'; AtomicNumber:49; MassNumber:134; Mass:133.94454{32#} ; Abundance:0;  StandardWeight:114.818{1}),
(Symbol:'In'; AtomicNumber:49; MassNumber:135; Mass:134.95005{43#} ; Abundance:0;  StandardWeight:114.818{1}),
(Symbol:'Sn'; AtomicNumber:50; MassNumber:99; Mass:98.94853{54#} ; Abundance:0; {g} StandardWeight:118.710{7}),
(Symbol:'Sn'; AtomicNumber:50; MassNumber:100; Mass:99.93850{32} ; Abundance:0; {g} StandardWeight:118.710{7}),
(Symbol:'Sn'; AtomicNumber:50; MassNumber:101; Mass:100.93526{32} ; Abundance:0; {g} StandardWeight:118.710{7}),
(Symbol:'Sn'; AtomicNumber:50; MassNumber:102; Mass:101.93029{11} ; Abundance:0; {g} StandardWeight:118.710{7}),
(Symbol:'Sn'; AtomicNumber:50; MassNumber:103; Mass:102.928105{76} ; Abundance:0; {g} StandardWeight:118.710{7}),
(Symbol:'Sn'; AtomicNumber:50; MassNumber:104; Mass:103.9231052{62} ; Abundance:0; {g} StandardWeight:118.710{7}),
(Symbol:'Sn'; AtomicNumber:50; MassNumber:105; Mass:104.9212684{43} ; Abundance:0; {g} StandardWeight:118.710{7}),
(Symbol:'Sn'; AtomicNumber:50; MassNumber:106; Mass:105.9169574{55} ; Abundance:0; {g} StandardWeight:118.710{7}),
(Symbol:'Sn'; AtomicNumber:50; MassNumber:107; Mass:106.9157137{57} ; Abundance:0; {g} StandardWeight:118.710{7}),
(Symbol:'Sn'; AtomicNumber:50; MassNumber:108; Mass:107.9118943{58} ; Abundance:0; {g} StandardWeight:118.710{7}),
(Symbol:'Sn'; AtomicNumber:50; MassNumber:109; Mass:108.9112921{85} ; Abundance:0; {g} StandardWeight:118.710{7}),
(Symbol:'Sn'; AtomicNumber:50; MassNumber:110; Mass:109.907845{15} ; Abundance:0; {g} StandardWeight:118.710{7}),
(Symbol:'Sn'; AtomicNumber:50; MassNumber:111; Mass:110.9077401{58} ; Abundance:0; {g} StandardWeight:118.710{7}),
(Symbol:'Sn'; AtomicNumber:50; MassNumber:112; Mass:111.90482387{61} ; Abundance:0.97; {g} StandardWeight:118.710{7}),
(Symbol:'Sn'; AtomicNumber:50; MassNumber:113; Mass:112.9051757{18} ; Abundance:0; {g} StandardWeight:118.710{7}),
(Symbol:'Sn'; AtomicNumber:50; MassNumber:114; Mass:113.9027827{10} ; Abundance:0.66; {g} StandardWeight:118.710{7}),
(Symbol:'Sn'; AtomicNumber:50; MassNumber:115; Mass:114.903344699{16} ; Abundance:0.34; {g} StandardWeight:118.710{7}),
(Symbol:'Sn'; AtomicNumber:50; MassNumber:116; Mass:115.90174280{10} ; Abundance:14.54; {g} StandardWeight:118.710{7}),
(Symbol:'Sn'; AtomicNumber:50; MassNumber:117; Mass:116.90295398{52} ; Abundance:7.68; {g} StandardWeight:118.710{7}),
(Symbol:'Sn'; AtomicNumber:50; MassNumber:118; Mass:117.90160657{54} ; Abundance:24.22; {g} StandardWeight:118.710{7}),
(Symbol:'Sn'; AtomicNumber:50; MassNumber:119; Mass:118.90331117{78} ; Abundance:8.59; {g} StandardWeight:118.710{7}),
(Symbol:'Sn'; AtomicNumber:50; MassNumber:120; Mass:119.90220163{97} ; Abundance:32.58; {g} StandardWeight:118.710{7}),
(Symbol:'Sn'; AtomicNumber:50; MassNumber:121; Mass:120.9042426{10} ; Abundance:0; {g} StandardWeight:118.710{7}),
(Symbol:'Sn'; AtomicNumber:50; MassNumber:122; Mass:121.9034438{26} ; Abundance:4.63; {g} StandardWeight:118.710{7}),
(Symbol:'Sn'; AtomicNumber:50; MassNumber:123; Mass:122.9057252{26} ; Abundance:0; {g} StandardWeight:118.710{7}),
(Symbol:'Sn'; AtomicNumber:50; MassNumber:124; Mass:123.9052766{11} ; Abundance:5.79; {g} StandardWeight:118.710{7}),
(Symbol:'Sn'; AtomicNumber:50; MassNumber:125; Mass:124.9077864{11} ; Abundance:0; {g} StandardWeight:118.710{7}),
(Symbol:'Sn'; AtomicNumber:50; MassNumber:126; Mass:125.907659{11} ; Abundance:0; {g} StandardWeight:118.710{7}),
(Symbol:'Sn'; AtomicNumber:50; MassNumber:127; Mass:126.910390{11} ; Abundance:0; {g} StandardWeight:118.710{7}),
(Symbol:'Sn'; AtomicNumber:50; MassNumber:128; Mass:127.910507{19} ; Abundance:0; {g} StandardWeight:118.710{7}),
(Symbol:'Sn'; AtomicNumber:50; MassNumber:129; Mass:128.913465{21} ; Abundance:0; {g} StandardWeight:118.710{7}),
(Symbol:'Sn'; AtomicNumber:50; MassNumber:130; Mass:129.9139738{23} ; Abundance:0; {g} StandardWeight:118.710{7}),
(Symbol:'Sn'; AtomicNumber:50; MassNumber:131; Mass:130.9170450{65} ; Abundance:0; {g} StandardWeight:118.710{7}),
(Symbol:'Sn'; AtomicNumber:50; MassNumber:132; Mass:131.9178267{31} ; Abundance:0; {g} StandardWeight:118.710{7}),
(Symbol:'Sn'; AtomicNumber:50; MassNumber:133; Mass:132.9239134{26} ; Abundance:0; {g} StandardWeight:118.710{7}),
(Symbol:'Sn'; AtomicNumber:50; MassNumber:134; Mass:133.9286821{35} ; Abundance:0; {g} StandardWeight:118.710{7}),
(Symbol:'Sn'; AtomicNumber:50; MassNumber:135; Mass:134.9349086{33} ; Abundance:0; {g} StandardWeight:118.710{7}),
(Symbol:'Sn'; AtomicNumber:50; MassNumber:136; Mass:135.93999{43#} ; Abundance:0; {g} StandardWeight:118.710{7}),
(Symbol:'Sn'; AtomicNumber:50; MassNumber:137; Mass:136.94655{54#} ; Abundance:0; {g} StandardWeight:118.710{7}),
(Symbol:'Sn'; AtomicNumber:50; MassNumber:138; Mass:137.95184{64#} ; Abundance:0; {g} StandardWeight:118.710{7}),
(Symbol:'Sb'; AtomicNumber:51; MassNumber:103; Mass:102.93969{32#} ; Abundance:0; {g} StandardWeight:121.760{1}),
(Symbol:'Sb'; AtomicNumber:51; MassNumber:104; Mass:103.93648{13} ; Abundance:0; {g} StandardWeight:121.760{1}),
(Symbol:'Sb'; AtomicNumber:51; MassNumber:105; Mass:104.931276{23} ; Abundance:0; {g} StandardWeight:121.760{1}),
(Symbol:'Sb'; AtomicNumber:51; MassNumber:106; Mass:105.9286380{80} ; Abundance:0; {g} StandardWeight:121.760{1}),
(Symbol:'Sb'; AtomicNumber:51; MassNumber:107; Mass:106.9241506{45} ; Abundance:0; {g} StandardWeight:121.760{1}),
(Symbol:'Sb'; AtomicNumber:51; MassNumber:108; Mass:107.9222267{59} ; Abundance:0; {g} StandardWeight:121.760{1}),
(Symbol:'Sb'; AtomicNumber:51; MassNumber:109; Mass:108.9181411{57} ; Abundance:0; {g} StandardWeight:121.760{1}),
(Symbol:'Sb'; AtomicNumber:51; MassNumber:110; Mass:109.9168543{64} ; Abundance:0; {g} StandardWeight:121.760{1}),
(Symbol:'Sb'; AtomicNumber:51; MassNumber:111; Mass:110.9132182{95} ; Abundance:0; {g} StandardWeight:121.760{1}),
(Symbol:'Sb'; AtomicNumber:51; MassNumber:112; Mass:111.912400{19} ; Abundance:0; {g} StandardWeight:121.760{1}),
(Symbol:'Sb'; AtomicNumber:51; MassNumber:113; Mass:112.909375{18} ; Abundance:0; {g} StandardWeight:121.760{1}),
(Symbol:'Sb'; AtomicNumber:51; MassNumber:114; Mass:113.909290{23} ; Abundance:0; {g} StandardWeight:121.760{1}),
(Symbol:'Sb'; AtomicNumber:51; MassNumber:115; Mass:114.906598{17} ; Abundance:0; {g} StandardWeight:121.760{1}),
(Symbol:'Sb'; AtomicNumber:51; MassNumber:116; Mass:115.9067931{55} ; Abundance:0; {g} StandardWeight:121.760{1}),
(Symbol:'Sb'; AtomicNumber:51; MassNumber:117; Mass:116.9048415{91} ; Abundance:0; {g} StandardWeight:121.760{1}),
(Symbol:'Sb'; AtomicNumber:51; MassNumber:118; Mass:117.9055321{32} ; Abundance:0; {g} StandardWeight:121.760{1}),
(Symbol:'Sb'; AtomicNumber:51; MassNumber:119; Mass:118.9039455{83} ; Abundance:0; {g} StandardWeight:121.760{1}),
(Symbol:'Sb'; AtomicNumber:51; MassNumber:120; Mass:119.9050794{77} ; Abundance:0; {g} StandardWeight:121.760{1}),
(Symbol:'Sb'; AtomicNumber:51; MassNumber:121; Mass:120.9038120{30} ; Abundance:57.21; {g} StandardWeight:121.760{1}),
(Symbol:'Sb'; AtomicNumber:51; MassNumber:122; Mass:121.9051699{30} ; Abundance:0; {g} StandardWeight:121.760{1}),
(Symbol:'Sb'; AtomicNumber:51; MassNumber:123; Mass:122.9042132{23} ; Abundance:42.79; {g} StandardWeight:121.760{1}),
(Symbol:'Sb'; AtomicNumber:51; MassNumber:124; Mass:123.9059350{23} ; Abundance:0; {g} StandardWeight:121.760{1}),
(Symbol:'Sb'; AtomicNumber:51; MassNumber:125; Mass:124.9052530{28} ; Abundance:0; {g} StandardWeight:121.760{1}),
(Symbol:'Sb'; AtomicNumber:51; MassNumber:126; Mass:125.907253{34} ; Abundance:0; {g} StandardWeight:121.760{1}),
(Symbol:'Sb'; AtomicNumber:51; MassNumber:127; Mass:126.9069243{55} ; Abundance:0; {g} StandardWeight:121.760{1}),
(Symbol:'Sb'; AtomicNumber:51; MassNumber:128; Mass:127.909146{21} ; Abundance:0; {g} StandardWeight:121.760{1}),
(Symbol:'Sb'; AtomicNumber:51; MassNumber:129; Mass:128.909147{23} ; Abundance:0; {g} StandardWeight:121.760{1}),
(Symbol:'Sb'; AtomicNumber:51; MassNumber:130; Mass:129.911662{15} ; Abundance:0; {g} StandardWeight:121.760{1}),
(Symbol:'Sb'; AtomicNumber:51; MassNumber:131; Mass:130.9119888{23} ; Abundance:0; {g} StandardWeight:121.760{1}),
(Symbol:'Sb'; AtomicNumber:51; MassNumber:132; Mass:131.9145077{29} ; Abundance:0; {g} StandardWeight:121.760{1}),
(Symbol:'Sb'; AtomicNumber:51; MassNumber:133; Mass:132.9152732{34} ; Abundance:0; {g} StandardWeight:121.760{1}),
(Symbol:'Sb'; AtomicNumber:51; MassNumber:134; Mass:133.9205357{18} ; Abundance:0; {g} StandardWeight:121.760{1}),
(Symbol:'Sb'; AtomicNumber:51; MassNumber:135; Mass:134.9251851{31} ; Abundance:0; {g} StandardWeight:121.760{1}),
(Symbol:'Sb'; AtomicNumber:51; MassNumber:136; Mass:135.9307459{68} ; Abundance:0; {g} StandardWeight:121.760{1}),
(Symbol:'Sb'; AtomicNumber:51; MassNumber:137; Mass:136.93555{32} ; Abundance:0; {g} StandardWeight:121.760{1}),
(Symbol:'Sb'; AtomicNumber:51; MassNumber:138; Mass:137.94145{32#} ; Abundance:0; {g} StandardWeight:121.760{1}),
(Symbol:'Sb'; AtomicNumber:51; MassNumber:139; Mass:138.94655{43#} ; Abundance:0; {g} StandardWeight:121.760{1}),
(Symbol:'Sb'; AtomicNumber:51; MassNumber:140; Mass:139.95283{64#} ; Abundance:0; {g} StandardWeight:121.760{1}),
(Symbol:'Te'; AtomicNumber:52; MassNumber:105; Mass:104.94330{32} ; Abundance:0; {g} StandardWeight:127.60{3}),
(Symbol:'Te'; AtomicNumber:52; MassNumber:106; Mass:105.93750{11} ; Abundance:0; {g} StandardWeight:127.60{3}),
(Symbol:'Te'; AtomicNumber:52; MassNumber:107; Mass:106.935012{76} ; Abundance:0; {g} StandardWeight:127.60{3}),
(Symbol:'Te'; AtomicNumber:52; MassNumber:108; Mass:107.9293805{58} ; Abundance:0; {g} StandardWeight:127.60{3}),
(Symbol:'Te'; AtomicNumber:52; MassNumber:109; Mass:108.9273045{47} ; Abundance:0; {g} StandardWeight:127.60{3}),
(Symbol:'Te'; AtomicNumber:52; MassNumber:110; Mass:109.9224581{71} ; Abundance:0; {g} StandardWeight:127.60{3}),
(Symbol:'Te'; AtomicNumber:52; MassNumber:111; Mass:110.9210006{69} ; Abundance:0; {g} StandardWeight:127.60{3}),
(Symbol:'Te'; AtomicNumber:52; MassNumber:112; Mass:111.9167279{90} ; Abundance:0; {g} StandardWeight:127.60{3}),
(Symbol:'Te'; AtomicNumber:52; MassNumber:113; Mass:112.915891{30} ; Abundance:0; {g} StandardWeight:127.60{3}),
(Symbol:'Te'; AtomicNumber:52; MassNumber:114; Mass:113.912089{30} ; Abundance:0; {g} StandardWeight:127.60{3}),
(Symbol:'Te'; AtomicNumber:52; MassNumber:115; Mass:114.911902{30} ; Abundance:0; {g} StandardWeight:127.60{3}),
(Symbol:'Te'; AtomicNumber:52; MassNumber:116; Mass:115.908460{30} ; Abundance:0; {g} StandardWeight:127.60{3}),
(Symbol:'Te'; AtomicNumber:52; MassNumber:117; Mass:116.908646{14} ; Abundance:0; {g} StandardWeight:127.60{3}),
(Symbol:'Te'; AtomicNumber:52; MassNumber:118; Mass:117.905854{20} ; Abundance:0; {g} StandardWeight:127.60{3}),
(Symbol:'Te'; AtomicNumber:52; MassNumber:119; Mass:118.9064071{85} ; Abundance:0; {g} StandardWeight:127.60{3}),
(Symbol:'Te'; AtomicNumber:52; MassNumber:120; Mass:119.9040593{33} ; Abundance:0.09; {g} StandardWeight:127.60{3}),
(Symbol:'Te'; AtomicNumber:52; MassNumber:121; Mass:120.904944{28} ; Abundance:0; {g} StandardWeight:127.60{3}),
(Symbol:'Te'; AtomicNumber:52; MassNumber:122; Mass:121.9030435{16} ; Abundance:2.55; {g} StandardWeight:127.60{3}),
(Symbol:'Te'; AtomicNumber:52; MassNumber:123; Mass:122.9042698{16} ; Abundance:0.89; {g} StandardWeight:127.60{3}),
(Symbol:'Te'; AtomicNumber:52; MassNumber:124; Mass:123.9028171{16} ; Abundance:4.74; {g} StandardWeight:127.60{3}),
(Symbol:'Te'; AtomicNumber:52; MassNumber:125; Mass:124.9044299{16} ; Abundance:7.07; {g} StandardWeight:127.60{3}),
(Symbol:'Te'; AtomicNumber:52; MassNumber:126; Mass:125.9033109{16} ; Abundance:18.84; {g} StandardWeight:127.60{3}),
(Symbol:'Te'; AtomicNumber:52; MassNumber:127; Mass:126.9052257{16} ; Abundance:0; {g} StandardWeight:127.60{3}),
(Symbol:'Te'; AtomicNumber:52; MassNumber:128; Mass:127.90446128{93} ; Abundance:31.74; {g} StandardWeight:127.60{3}),
(Symbol:'Te'; AtomicNumber:52; MassNumber:129; Mass:128.90659646{93} ; Abundance:0; {g} StandardWeight:127.60{3}),
(Symbol:'Te'; AtomicNumber:52; MassNumber:130; Mass:129.906222748{12} ; Abundance:34.08; {g} StandardWeight:127.60{3}),
(Symbol:'Te'; AtomicNumber:52; MassNumber:131; Mass:130.908522213{65} ; Abundance:0; {g} StandardWeight:127.60{3}),
(Symbol:'Te'; AtomicNumber:52; MassNumber:132; Mass:131.9085467{37} ; Abundance:0; {g} StandardWeight:127.60{3}),
(Symbol:'Te'; AtomicNumber:52; MassNumber:133; Mass:132.9109688{39} ; Abundance:0; {g} StandardWeight:127.60{3}),
(Symbol:'Te'; AtomicNumber:52; MassNumber:134; Mass:133.9113940{30} ; Abundance:0; {g} StandardWeight:127.60{3}),
(Symbol:'Te'; AtomicNumber:52; MassNumber:135; Mass:134.9165557{29} ; Abundance:0; {g} StandardWeight:127.60{3}),
(Symbol:'Te'; AtomicNumber:52; MassNumber:136; Mass:135.9201006{26} ; Abundance:0; {g} StandardWeight:127.60{3}),
(Symbol:'Te'; AtomicNumber:52; MassNumber:137; Mass:136.9255989{27} ; Abundance:0; {g} StandardWeight:127.60{3}),
(Symbol:'Te'; AtomicNumber:52; MassNumber:138; Mass:137.9294722{47} ; Abundance:0; {g} StandardWeight:127.60{3}),
(Symbol:'Te'; AtomicNumber:52; MassNumber:139; Mass:138.9353672{38} ; Abundance:0; {g} StandardWeight:127.60{3}),
(Symbol:'Te'; AtomicNumber:52; MassNumber:140; Mass:139.939499{30} ; Abundance:0; {g} StandardWeight:127.60{3}),
(Symbol:'Te'; AtomicNumber:52; MassNumber:141; Mass:140.94580{43#} ; Abundance:0; {g} StandardWeight:127.60{3}),
(Symbol:'Te'; AtomicNumber:52; MassNumber:142; Mass:141.95022{54#} ; Abundance:0; {g} StandardWeight:127.60{3}),
(Symbol:'Te'; AtomicNumber:52; MassNumber:143; Mass:142.95676{54#} ; Abundance:0; {g} StandardWeight:127.60{3}),
(Symbol:'I'; AtomicNumber:53; MassNumber:107; Mass:106.94678{32#} ; Abundance:0;  StandardWeight:126.90447{3}),
(Symbol:'I'; AtomicNumber:53; MassNumber:108; Mass:107.94348{14} ; Abundance:0;  StandardWeight:126.90447{3}),
(Symbol:'I'; AtomicNumber:53; MassNumber:109; Mass:108.9380853{61} ; Abundance:0;  StandardWeight:126.90447{3}),
(Symbol:'I'; AtomicNumber:53; MassNumber:110; Mass:109.935089{54} ; Abundance:0;  StandardWeight:126.90447{3}),
(Symbol:'I'; AtomicNumber:53; MassNumber:111; Mass:110.9302692{51} ; Abundance:0;  StandardWeight:126.90447{3}),
(Symbol:'I'; AtomicNumber:53; MassNumber:112; Mass:111.928005{11} ; Abundance:0;  StandardWeight:126.90447{3}),
(Symbol:'I'; AtomicNumber:53; MassNumber:113; Mass:112.9236501{86} ; Abundance:0;  StandardWeight:126.90447{3}),
(Symbol:'I'; AtomicNumber:53; MassNumber:114; Mass:113.92185{32#} ; Abundance:0;  StandardWeight:126.90447{3}),
(Symbol:'I'; AtomicNumber:53; MassNumber:115; Mass:114.918048{31} ; Abundance:0;  StandardWeight:126.90447{3}),
(Symbol:'I'; AtomicNumber:53; MassNumber:116; Mass:115.91681{10} ; Abundance:0;  StandardWeight:126.90447{3}),
(Symbol:'I'; AtomicNumber:53; MassNumber:117; Mass:116.913648{28} ; Abundance:0;  StandardWeight:126.90447{3}),
(Symbol:'I'; AtomicNumber:53; MassNumber:118; Mass:117.913074{21} ; Abundance:0;  StandardWeight:126.90447{3}),
(Symbol:'I'; AtomicNumber:53; MassNumber:119; Mass:118.910074{30} ; Abundance:0;  StandardWeight:126.90447{3}),
(Symbol:'I'; AtomicNumber:53; MassNumber:120; Mass:119.910087{16} ; Abundance:0;  StandardWeight:126.90447{3}),
(Symbol:'I'; AtomicNumber:53; MassNumber:121; Mass:120.9074051{58} ; Abundance:0;  StandardWeight:126.90447{3}),
(Symbol:'I'; AtomicNumber:53; MassNumber:122; Mass:121.9075888{56} ; Abundance:0;  StandardWeight:126.90447{3}),
(Symbol:'I'; AtomicNumber:53; MassNumber:123; Mass:122.9055885{40} ; Abundance:0;  StandardWeight:126.90447{3}),
(Symbol:'I'; AtomicNumber:53; MassNumber:124; Mass:123.9062090{26} ; Abundance:0;  StandardWeight:126.90447{3}),
(Symbol:'I'; AtomicNumber:53; MassNumber:125; Mass:124.9046294{16} ; Abundance:0;  StandardWeight:126.90447{3}),
(Symbol:'I'; AtomicNumber:53; MassNumber:126; Mass:125.9056233{41} ; Abundance:0;  StandardWeight:126.90447{3}),
(Symbol:'I'; AtomicNumber:53; MassNumber:127; Mass:126.9044719{39} ; Abundance:100;  StandardWeight:126.90447{3}),
(Symbol:'I'; AtomicNumber:53; MassNumber:128; Mass:127.9058086{39} ; Abundance:0;  StandardWeight:126.90447{3}),
(Symbol:'I'; AtomicNumber:53; MassNumber:129; Mass:128.9049837{34} ; Abundance:0;  StandardWeight:126.90447{3}),
(Symbol:'I'; AtomicNumber:53; MassNumber:130; Mass:129.9066702{34} ; Abundance:0;  StandardWeight:126.90447{3}),
(Symbol:'I'; AtomicNumber:53; MassNumber:131; Mass:130.90612630{69} ; Abundance:0;  StandardWeight:126.90447{3}),
(Symbol:'I'; AtomicNumber:53; MassNumber:132; Mass:131.9079935{44} ; Abundance:0;  StandardWeight:126.90447{3}),
(Symbol:'I'; AtomicNumber:53; MassNumber:133; Mass:132.9077970{50} ; Abundance:0;  StandardWeight:126.90447{3}),
(Symbol:'I'; AtomicNumber:53; MassNumber:134; Mass:133.9097588{59} ; Abundance:0;  StandardWeight:126.90447{3}),
(Symbol:'I'; AtomicNumber:53; MassNumber:135; Mass:134.9100488{58} ; Abundance:0;  StandardWeight:126.90447{3}),
(Symbol:'I'; AtomicNumber:53; MassNumber:136; Mass:135.914604{15} ; Abundance:0;  StandardWeight:126.90447{3}),
(Symbol:'I'; AtomicNumber:53; MassNumber:137; Mass:136.9180282{90} ; Abundance:0;  StandardWeight:126.90447{3}),
(Symbol:'I'; AtomicNumber:53; MassNumber:138; Mass:137.9227264{64} ; Abundance:0;  StandardWeight:126.90447{3}),
(Symbol:'I'; AtomicNumber:53; MassNumber:139; Mass:138.926506{31} ; Abundance:0;  StandardWeight:126.90447{3}),
(Symbol:'I'; AtomicNumber:53; MassNumber:140; Mass:139.93173{20} ; Abundance:0;  StandardWeight:126.90447{3}),
(Symbol:'I'; AtomicNumber:53; MassNumber:141; Mass:140.93569{21#} ; Abundance:0;  StandardWeight:126.90447{3}),
(Symbol:'I'; AtomicNumber:53; MassNumber:142; Mass:141.94120{40} ; Abundance:0;  StandardWeight:126.90447{3}),
(Symbol:'I'; AtomicNumber:53; MassNumber:143; Mass:142.94565{32#} ; Abundance:0;  StandardWeight:126.90447{3}),
(Symbol:'I'; AtomicNumber:53; MassNumber:144; Mass:143.95139{43#} ; Abundance:0;  StandardWeight:126.90447{3}),
(Symbol:'I'; AtomicNumber:53; MassNumber:145; Mass:144.95605{54#} ; Abundance:0;  StandardWeight:126.90447{3}),
(Symbol:'Xe'; AtomicNumber:54; MassNumber:109; Mass:108.95043{32} ; Abundance:0; {g,m} StandardWeight:131.293{6}),
(Symbol:'Xe'; AtomicNumber:54; MassNumber:110; Mass:109.94426{11} ; Abundance:0; {g,m} StandardWeight:131.293{6}),
(Symbol:'Xe'; AtomicNumber:54; MassNumber:111; Mass:110.941607{93} ; Abundance:0; {g,m} StandardWeight:131.293{6}),
(Symbol:'Xe'; AtomicNumber:54; MassNumber:112; Mass:111.9355590{89} ; Abundance:0; {g,m} StandardWeight:131.293{6}),
(Symbol:'Xe'; AtomicNumber:54; MassNumber:113; Mass:112.9332217{73} ; Abundance:0; {g,m} StandardWeight:131.293{6}),
(Symbol:'Xe'; AtomicNumber:54; MassNumber:114; Mass:113.927980{12} ; Abundance:0; {g,m} StandardWeight:131.293{6}),
(Symbol:'Xe'; AtomicNumber:54; MassNumber:115; Mass:114.926294{13} ; Abundance:0; {g,m} StandardWeight:131.293{6}),
(Symbol:'Xe'; AtomicNumber:54; MassNumber:116; Mass:115.921581{14} ; Abundance:0; {g,m} StandardWeight:131.293{6}),
(Symbol:'Xe'; AtomicNumber:54; MassNumber:117; Mass:116.920359{11} ; Abundance:0; {g,m} StandardWeight:131.293{6}),
(Symbol:'Xe'; AtomicNumber:54; MassNumber:118; Mass:117.916179{11} ; Abundance:0; {g,m} StandardWeight:131.293{6}),
(Symbol:'Xe'; AtomicNumber:54; MassNumber:119; Mass:118.915411{11} ; Abundance:0; {g,m} StandardWeight:131.293{6}),
(Symbol:'Xe'; AtomicNumber:54; MassNumber:120; Mass:119.911784{13} ; Abundance:0; {g,m} StandardWeight:131.293{6}),
(Symbol:'Xe'; AtomicNumber:54; MassNumber:121; Mass:120.911453{11} ; Abundance:0; {g,m} StandardWeight:131.293{6}),
(Symbol:'Xe'; AtomicNumber:54; MassNumber:122; Mass:121.908368{12} ; Abundance:0; {g,m} StandardWeight:131.293{6}),
(Symbol:'Xe'; AtomicNumber:54; MassNumber:123; Mass:122.908482{10} ; Abundance:0; {g,m} StandardWeight:131.293{6}),
(Symbol:'Xe'; AtomicNumber:54; MassNumber:124; Mass:123.9058920{19} ; Abundance:0.0952; {g,m} StandardWeight:131.293{6}),
(Symbol:'Xe'; AtomicNumber:54; MassNumber:125; Mass:124.9063944{20} ; Abundance:0; {g,m} StandardWeight:131.293{6}),
(Symbol:'Xe'; AtomicNumber:54; MassNumber:126; Mass:125.9042983{38} ; Abundance:0.089; {g,m} StandardWeight:131.293{6}),
(Symbol:'Xe'; AtomicNumber:54; MassNumber:127; Mass:126.9051829{44} ; Abundance:0; {g,m} StandardWeight:131.293{6}),
(Symbol:'Xe'; AtomicNumber:54; MassNumber:128; Mass:127.9035310{11} ; Abundance:1.9102; {g,m} StandardWeight:131.293{6}),
(Symbol:'Xe'; AtomicNumber:54; MassNumber:129; Mass:128.9047808611{60} ; Abundance:26.4006; {g,m} StandardWeight:131.293{6}),
(Symbol:'Xe'; AtomicNumber:54; MassNumber:130; Mass:129.903509349{10} ; Abundance:4.071; {g,m} StandardWeight:131.293{6}),
(Symbol:'Xe'; AtomicNumber:54; MassNumber:131; Mass:130.90508406{24} ; Abundance:21.2324; {g,m} StandardWeight:131.293{6}),
(Symbol:'Xe'; AtomicNumber:54; MassNumber:132; Mass:131.9041550856{56} ; Abundance:26.9086; {g,m} StandardWeight:131.293{6}),
(Symbol:'Xe'; AtomicNumber:54; MassNumber:133; Mass:132.9059108{26} ; Abundance:0; {g,m} StandardWeight:131.293{6}),
(Symbol:'Xe'; AtomicNumber:54; MassNumber:134; Mass:133.90539466{90} ; Abundance:10.4357; {g,m} StandardWeight:131.293{6}),
(Symbol:'Xe'; AtomicNumber:54; MassNumber:135; Mass:134.9072278{45} ; Abundance:0; {g,m} StandardWeight:131.293{6}),
(Symbol:'Xe'; AtomicNumber:54; MassNumber:136; Mass:135.907214484{11} ; Abundance:8.8573; {g,m} StandardWeight:131.293{6}),
(Symbol:'Xe'; AtomicNumber:54; MassNumber:137; Mass:136.91155778{11} ; Abundance:0; {g,m} StandardWeight:131.293{6}),
(Symbol:'Xe'; AtomicNumber:54; MassNumber:138; Mass:137.9141463{30} ; Abundance:0; {g,m} StandardWeight:131.293{6}),
(Symbol:'Xe'; AtomicNumber:54; MassNumber:139; Mass:138.9187922{23} ; Abundance:0; {g,m} StandardWeight:131.293{6}),
(Symbol:'Xe'; AtomicNumber:54; MassNumber:140; Mass:139.9216458{25} ; Abundance:0; {g,m} StandardWeight:131.293{6}),
(Symbol:'Xe'; AtomicNumber:54; MassNumber:141; Mass:140.9267872{31} ; Abundance:0; {g,m} StandardWeight:131.293{6}),
(Symbol:'Xe'; AtomicNumber:54; MassNumber:142; Mass:141.9299731{29} ; Abundance:0; {g,m} StandardWeight:131.293{6}),
(Symbol:'Xe'; AtomicNumber:54; MassNumber:143; Mass:142.9353696{50} ; Abundance:0; {g,m} StandardWeight:131.293{6}),
(Symbol:'Xe'; AtomicNumber:54; MassNumber:144; Mass:143.9389451{57} ; Abundance:0; {g,m} StandardWeight:131.293{6}),
(Symbol:'Xe'; AtomicNumber:54; MassNumber:145; Mass:144.944720{12} ; Abundance:0; {g,m} StandardWeight:131.293{6}),
(Symbol:'Xe'; AtomicNumber:54; MassNumber:146; Mass:145.948518{26} ; Abundance:0; {g,m} StandardWeight:131.293{6}),
(Symbol:'Xe'; AtomicNumber:54; MassNumber:147; Mass:146.95426{21#} ; Abundance:0; {g,m} StandardWeight:131.293{6}),
(Symbol:'Xe'; AtomicNumber:54; MassNumber:148; Mass:147.95813{21#} ; Abundance:0; {g,m} StandardWeight:131.293{6}),
(Symbol:'Cs'; AtomicNumber:55; MassNumber:112; Mass:111.950309{93} ; Abundance:0;  StandardWeight:132.90545196{6}),
(Symbol:'Cs'; AtomicNumber:55; MassNumber:113; Mass:112.9444291{93} ; Abundance:0;  StandardWeight:132.90545196{6}),
(Symbol:'Cs'; AtomicNumber:55; MassNumber:114; Mass:113.941296{76} ; Abundance:0;  StandardWeight:132.90545196{6}),
(Symbol:'Cs'; AtomicNumber:55; MassNumber:115; Mass:114.93591{32#} ; Abundance:0;  StandardWeight:132.90545196{6}),
(Symbol:'Cs'; AtomicNumber:55; MassNumber:116; Mass:115.93337{11#} ; Abundance:0;  StandardWeight:132.90545196{6}),
(Symbol:'Cs'; AtomicNumber:55; MassNumber:117; Mass:116.928617{67} ; Abundance:0;  StandardWeight:132.90545196{6}),
(Symbol:'Cs'; AtomicNumber:55; MassNumber:118; Mass:117.926560{14} ; Abundance:0;  StandardWeight:132.90545196{6}),
(Symbol:'Cs'; AtomicNumber:55; MassNumber:119; Mass:118.922377{15} ; Abundance:0;  StandardWeight:132.90545196{6}),
(Symbol:'Cs'; AtomicNumber:55; MassNumber:120; Mass:119.920677{11} ; Abundance:0;  StandardWeight:132.90545196{6}),
(Symbol:'Cs'; AtomicNumber:55; MassNumber:121; Mass:120.917227{15} ; Abundance:0;  StandardWeight:132.90545196{6}),
(Symbol:'Cs'; AtomicNumber:55; MassNumber:122; Mass:121.916108{36} ; Abundance:0;  StandardWeight:132.90545196{6}),
(Symbol:'Cs'; AtomicNumber:55; MassNumber:123; Mass:122.912996{13} ; Abundance:0;  StandardWeight:132.90545196{6}),
(Symbol:'Cs'; AtomicNumber:55; MassNumber:124; Mass:123.9122578{89} ; Abundance:0;  StandardWeight:132.90545196{6}),
(Symbol:'Cs'; AtomicNumber:55; MassNumber:125; Mass:124.9097280{83} ; Abundance:0;  StandardWeight:132.90545196{6}),
(Symbol:'Cs'; AtomicNumber:55; MassNumber:126; Mass:125.909446{11} ; Abundance:0;  StandardWeight:132.90545196{6}),
(Symbol:'Cs'; AtomicNumber:55; MassNumber:127; Mass:126.9074174{60} ; Abundance:0;  StandardWeight:132.90545196{6}),
(Symbol:'Cs'; AtomicNumber:55; MassNumber:128; Mass:127.9077487{58} ; Abundance:0;  StandardWeight:132.90545196{6}),
(Symbol:'Cs'; AtomicNumber:55; MassNumber:129; Mass:128.9060657{49} ; Abundance:0;  StandardWeight:132.90545196{6}),
(Symbol:'Cs'; AtomicNumber:55; MassNumber:130; Mass:129.9067093{90} ; Abundance:0;  StandardWeight:132.90545196{6}),
(Symbol:'Cs'; AtomicNumber:55; MassNumber:131; Mass:130.9054649{53} ; Abundance:0;  StandardWeight:132.90545196{6}),
(Symbol:'Cs'; AtomicNumber:55; MassNumber:132; Mass:131.9064339{21} ; Abundance:0;  StandardWeight:132.90545196{6}),
(Symbol:'Cs'; AtomicNumber:55; MassNumber:133; Mass:132.9054519610{80} ; Abundance:100;  StandardWeight:132.90545196{6}),
(Symbol:'Cs'; AtomicNumber:55; MassNumber:134; Mass:133.906718503{17} ; Abundance:0;  StandardWeight:132.90545196{6}),
(Symbol:'Cs'; AtomicNumber:55; MassNumber:135; Mass:134.9059770{11} ; Abundance:0;  StandardWeight:132.90545196{6}),
(Symbol:'Cs'; AtomicNumber:55; MassNumber:136; Mass:135.9073114{20} ; Abundance:0;  StandardWeight:132.90545196{6}),
(Symbol:'Cs'; AtomicNumber:55; MassNumber:137; Mass:136.90708923{36} ; Abundance:0;  StandardWeight:132.90545196{6}),
(Symbol:'Cs'; AtomicNumber:55; MassNumber:138; Mass:137.9110171{98} ; Abundance:0;  StandardWeight:132.90545196{6}),
(Symbol:'Cs'; AtomicNumber:55; MassNumber:139; Mass:138.9133638{34} ; Abundance:0;  StandardWeight:132.90545196{6}),
(Symbol:'Cs'; AtomicNumber:55; MassNumber:140; Mass:139.9172831{88} ; Abundance:0;  StandardWeight:132.90545196{6}),
(Symbol:'Cs'; AtomicNumber:55; MassNumber:141; Mass:140.9200455{98} ; Abundance:0;  StandardWeight:132.90545196{6}),
(Symbol:'Cs'; AtomicNumber:55; MassNumber:142; Mass:141.9242960{79} ; Abundance:0;  StandardWeight:132.90545196{6}),
(Symbol:'Cs'; AtomicNumber:55; MassNumber:143; Mass:142.927349{24} ; Abundance:0;  StandardWeight:132.90545196{6}),
(Symbol:'Cs'; AtomicNumber:55; MassNumber:144; Mass:143.932076{27} ; Abundance:0;  StandardWeight:132.90545196{6}),
(Symbol:'Cs'; AtomicNumber:55; MassNumber:145; Mass:144.935527{12} ; Abundance:0;  StandardWeight:132.90545196{6}),
(Symbol:'Cs'; AtomicNumber:55; MassNumber:146; Mass:145.940344{42} ; Abundance:0;  StandardWeight:132.90545196{6}),
(Symbol:'Cs'; AtomicNumber:55; MassNumber:147; Mass:146.944156{57} ; Abundance:0;  StandardWeight:132.90545196{6}),
(Symbol:'Cs'; AtomicNumber:55; MassNumber:148; Mass:147.94923{62} ; Abundance:0;  StandardWeight:132.90545196{6}),
(Symbol:'Cs'; AtomicNumber:55; MassNumber:149; Mass:148.95302{21#} ; Abundance:0;  StandardWeight:132.90545196{6}),
(Symbol:'Cs'; AtomicNumber:55; MassNumber:150; Mass:149.95833{32#} ; Abundance:0;  StandardWeight:132.90545196{6}),
(Symbol:'Cs'; AtomicNumber:55; MassNumber:151; Mass:150.96258{43#} ; Abundance:0;  StandardWeight:132.90545196{6}),
(Symbol:'Ba'; AtomicNumber:56; MassNumber:114; Mass:113.95066{12} ; Abundance:0;  StandardWeight:137.327{7}),
(Symbol:'Ba'; AtomicNumber:56; MassNumber:115; Mass:114.94737{54#} ; Abundance:0;  StandardWeight:137.327{7}),
(Symbol:'Ba'; AtomicNumber:56; MassNumber:116; Mass:115.94128{32#} ; Abundance:0;  StandardWeight:137.327{7}),
(Symbol:'Ba'; AtomicNumber:56; MassNumber:117; Mass:116.93814{21} ; Abundance:0;  StandardWeight:137.327{7}),
(Symbol:'Ba'; AtomicNumber:56; MassNumber:118; Mass:117.93306{21#} ; Abundance:0;  StandardWeight:137.327{7}),
(Symbol:'Ba'; AtomicNumber:56; MassNumber:119; Mass:118.93066{21} ; Abundance:0;  StandardWeight:137.327{7}),
(Symbol:'Ba'; AtomicNumber:56; MassNumber:120; Mass:119.92605{32} ; Abundance:0;  StandardWeight:137.327{7}),
(Symbol:'Ba'; AtomicNumber:56; MassNumber:121; Mass:120.92405{15} ; Abundance:0;  StandardWeight:137.327{7}),
(Symbol:'Ba'; AtomicNumber:56; MassNumber:122; Mass:121.919904{30} ; Abundance:0;  StandardWeight:137.327{7}),
(Symbol:'Ba'; AtomicNumber:56; MassNumber:123; Mass:122.918781{13} ; Abundance:0;  StandardWeight:137.327{7}),
(Symbol:'Ba'; AtomicNumber:56; MassNumber:124; Mass:123.915094{13} ; Abundance:0;  StandardWeight:137.327{7}),
(Symbol:'Ba'; AtomicNumber:56; MassNumber:125; Mass:124.914472{12} ; Abundance:0;  StandardWeight:137.327{7}),
(Symbol:'Ba'; AtomicNumber:56; MassNumber:126; Mass:125.911250{13} ; Abundance:0;  StandardWeight:137.327{7}),
(Symbol:'Ba'; AtomicNumber:56; MassNumber:127; Mass:126.911091{12} ; Abundance:0;  StandardWeight:137.327{7}),
(Symbol:'Ba'; AtomicNumber:56; MassNumber:128; Mass:127.9083420{56} ; Abundance:0;  StandardWeight:137.327{7}),
(Symbol:'Ba'; AtomicNumber:56; MassNumber:129; Mass:128.908681{11} ; Abundance:0;  StandardWeight:137.327{7}),
(Symbol:'Ba'; AtomicNumber:56; MassNumber:130; Mass:129.9063207{28} ; Abundance:0.106;  StandardWeight:137.327{7}),
(Symbol:'Ba'; AtomicNumber:56; MassNumber:131; Mass:130.9069410{28} ; Abundance:0;  StandardWeight:137.327{7}),
(Symbol:'Ba'; AtomicNumber:56; MassNumber:132; Mass:131.9050611{11} ; Abundance:0.101;  StandardWeight:137.327{7}),
(Symbol:'Ba'; AtomicNumber:56; MassNumber:133; Mass:132.9060074{11} ; Abundance:0;  StandardWeight:137.327{7}),
(Symbol:'Ba'; AtomicNumber:56; MassNumber:134; Mass:133.90450818{30} ; Abundance:2.417;  StandardWeight:137.327{7}),
(Symbol:'Ba'; AtomicNumber:56; MassNumber:135; Mass:134.90568838{29} ; Abundance:6.592;  StandardWeight:137.327{7}),
(Symbol:'Ba'; AtomicNumber:56; MassNumber:136; Mass:135.90457573{29} ; Abundance:7.854;  StandardWeight:137.327{7}),
(Symbol:'Ba'; AtomicNumber:56; MassNumber:137; Mass:136.90582714{30} ; Abundance:11.232;  StandardWeight:137.327{7}),
(Symbol:'Ba'; AtomicNumber:56; MassNumber:138; Mass:137.90524700{31} ; Abundance:71.698;  StandardWeight:137.327{7}),
(Symbol:'Ba'; AtomicNumber:56; MassNumber:139; Mass:138.90884110{31} ; Abundance:0;  StandardWeight:137.327{7}),
(Symbol:'Ba'; AtomicNumber:56; MassNumber:140; Mass:139.9106057{85} ; Abundance:0;  StandardWeight:137.327{7}),
(Symbol:'Ba'; AtomicNumber:56; MassNumber:141; Mass:140.9144033{57} ; Abundance:0;  StandardWeight:137.327{7}),
(Symbol:'Ba'; AtomicNumber:56; MassNumber:142; Mass:141.9164324{64} ; Abundance:0;  StandardWeight:137.327{7}),
(Symbol:'Ba'; AtomicNumber:56; MassNumber:143; Mass:142.9206253{74} ; Abundance:0;  StandardWeight:137.327{7}),
(Symbol:'Ba'; AtomicNumber:56; MassNumber:144; Mass:143.9229549{77} ; Abundance:0;  StandardWeight:137.327{7}),
(Symbol:'Ba'; AtomicNumber:56; MassNumber:145; Mass:144.9275184{91} ; Abundance:0;  StandardWeight:137.327{7}),
(Symbol:'Ba'; AtomicNumber:56; MassNumber:146; Mass:145.930284{22} ; Abundance:0;  StandardWeight:137.327{7}),
(Symbol:'Ba'; AtomicNumber:56; MassNumber:147; Mass:146.935304{21} ; Abundance:0;  StandardWeight:137.327{7}),
(Symbol:'Ba'; AtomicNumber:56; MassNumber:148; Mass:147.938171{68} ; Abundance:0;  StandardWeight:137.327{7}),
(Symbol:'Ba'; AtomicNumber:56; MassNumber:149; Mass:148.94308{21#} ; Abundance:0;  StandardWeight:137.327{7}),
(Symbol:'Ba'; AtomicNumber:56; MassNumber:150; Mass:149.94605{32#} ; Abundance:0;  StandardWeight:137.327{7}),
(Symbol:'Ba'; AtomicNumber:56; MassNumber:151; Mass:150.95127{32#} ; Abundance:0;  StandardWeight:137.327{7}),
(Symbol:'Ba'; AtomicNumber:56; MassNumber:152; Mass:151.95481{43#} ; Abundance:0;  StandardWeight:137.327{7}),
(Symbol:'Ba'; AtomicNumber:56; MassNumber:153; Mass:152.96036{43#} ; Abundance:0;  StandardWeight:137.327{7}),
(Symbol:'La'; AtomicNumber:57; MassNumber:116; Mass:115.95630{23#} ; Abundance:0; {g} StandardWeight:138.90547{7}),
(Symbol:'La'; AtomicNumber:57; MassNumber:117; Mass:116.94999{32#} ; Abundance:0; {g} StandardWeight:138.90547{7}),
(Symbol:'La'; AtomicNumber:57; MassNumber:118; Mass:117.94673{32#} ; Abundance:0; {g} StandardWeight:138.90547{7}),
(Symbol:'La'; AtomicNumber:57; MassNumber:119; Mass:118.94099{32#} ; Abundance:0; {g} StandardWeight:138.90547{7}),
(Symbol:'La'; AtomicNumber:57; MassNumber:120; Mass:119.93807{32#} ; Abundance:0; {g} StandardWeight:138.90547{7}),
(Symbol:'La'; AtomicNumber:57; MassNumber:121; Mass:120.93315{32#} ; Abundance:0; {g} StandardWeight:138.90547{7}),
(Symbol:'La'; AtomicNumber:57; MassNumber:122; Mass:121.93071{32#} ; Abundance:0; {g} StandardWeight:138.90547{7}),
(Symbol:'La'; AtomicNumber:57; MassNumber:123; Mass:122.92630{21#} ; Abundance:0; {g} StandardWeight:138.90547{7}),
(Symbol:'La'; AtomicNumber:57; MassNumber:124; Mass:123.924574{61} ; Abundance:0; {g} StandardWeight:138.90547{7}),
(Symbol:'La'; AtomicNumber:57; MassNumber:125; Mass:124.920816{28} ; Abundance:0; {g} StandardWeight:138.90547{7}),
(Symbol:'La'; AtomicNumber:57; MassNumber:126; Mass:125.919513{97} ; Abundance:0; {g} StandardWeight:138.90547{7}),
(Symbol:'La'; AtomicNumber:57; MassNumber:127; Mass:126.916375{28} ; Abundance:0; {g} StandardWeight:138.90547{7}),
(Symbol:'La'; AtomicNumber:57; MassNumber:128; Mass:127.915592{58} ; Abundance:0; {g} StandardWeight:138.90547{7}),
(Symbol:'La'; AtomicNumber:57; MassNumber:129; Mass:128.912694{23} ; Abundance:0; {g} StandardWeight:138.90547{7}),
(Symbol:'La'; AtomicNumber:57; MassNumber:130; Mass:129.912369{28} ; Abundance:0; {g} StandardWeight:138.90547{7}),
(Symbol:'La'; AtomicNumber:57; MassNumber:131; Mass:130.910070{30} ; Abundance:0; {g} StandardWeight:138.90547{7}),
(Symbol:'La'; AtomicNumber:57; MassNumber:132; Mass:131.910119{39} ; Abundance:0; {g} StandardWeight:138.90547{7}),
(Symbol:'La'; AtomicNumber:57; MassNumber:133; Mass:132.908218{30} ; Abundance:0; {g} StandardWeight:138.90547{7}),
(Symbol:'La'; AtomicNumber:57; MassNumber:134; Mass:133.908514{21} ; Abundance:0; {g} StandardWeight:138.90547{7}),
(Symbol:'La'; AtomicNumber:57; MassNumber:135; Mass:134.906984{10} ; Abundance:0; {g} StandardWeight:138.90547{7}),
(Symbol:'La'; AtomicNumber:57; MassNumber:136; Mass:135.907635{57} ; Abundance:0; {g} StandardWeight:138.90547{7}),
(Symbol:'La'; AtomicNumber:57; MassNumber:137; Mass:136.9064504{18} ; Abundance:0; {g} StandardWeight:138.90547{7}),
(Symbol:'La'; AtomicNumber:57; MassNumber:138; Mass:137.9071149{37} ; Abundance:0.08881; {g} StandardWeight:138.90547{7}),
(Symbol:'La'; AtomicNumber:57; MassNumber:139; Mass:138.9063563{24} ; Abundance:99.91119; {g} StandardWeight:138.90547{7}),
(Symbol:'La'; AtomicNumber:57; MassNumber:140; Mass:139.9094806{24} ; Abundance:0; {g} StandardWeight:138.90547{7}),
(Symbol:'La'; AtomicNumber:57; MassNumber:141; Mass:140.9109660{48} ; Abundance:0; {g} StandardWeight:138.90547{7}),
(Symbol:'La'; AtomicNumber:57; MassNumber:142; Mass:141.9140909{69} ; Abundance:0; {g} StandardWeight:138.90547{7}),
(Symbol:'La'; AtomicNumber:57; MassNumber:143; Mass:142.9160795{79} ; Abundance:0; {g} StandardWeight:138.90547{7}),
(Symbol:'La'; AtomicNumber:57; MassNumber:144; Mass:143.919646{14} ; Abundance:0; {g} StandardWeight:138.90547{7}),
(Symbol:'La'; AtomicNumber:57; MassNumber:145; Mass:144.921808{13} ; Abundance:0; {g} StandardWeight:138.90547{7}),
(Symbol:'La'; AtomicNumber:57; MassNumber:146; Mass:145.925875{36} ; Abundance:0; {g} StandardWeight:138.90547{7}),
(Symbol:'La'; AtomicNumber:57; MassNumber:147; Mass:146.928418{12} ; Abundance:0; {g} StandardWeight:138.90547{7}),
(Symbol:'La'; AtomicNumber:57; MassNumber:148; Mass:147.932679{21} ; Abundance:0; {g} StandardWeight:138.90547{7}),
(Symbol:'La'; AtomicNumber:57; MassNumber:149; Mass:148.93535{21} ; Abundance:0; {g} StandardWeight:138.90547{7}),
(Symbol:'La'; AtomicNumber:57; MassNumber:150; Mass:149.93947{21#} ; Abundance:0; {g} StandardWeight:138.90547{7}),
(Symbol:'La'; AtomicNumber:57; MassNumber:151; Mass:150.94232{21#} ; Abundance:0; {g} StandardWeight:138.90547{7}),
(Symbol:'La'; AtomicNumber:57; MassNumber:152; Mass:151.94682{32#} ; Abundance:0; {g} StandardWeight:138.90547{7}),
(Symbol:'La'; AtomicNumber:57; MassNumber:153; Mass:152.95036{32#} ; Abundance:0; {g} StandardWeight:138.90547{7}),
(Symbol:'La'; AtomicNumber:57; MassNumber:154; Mass:153.95517{43#} ; Abundance:0; {g} StandardWeight:138.90547{7}),
(Symbol:'La'; AtomicNumber:57; MassNumber:155; Mass:154.95901{43#} ; Abundance:0; {g} StandardWeight:138.90547{7}),
(Symbol:'Ce'; AtomicNumber:58; MassNumber:119; Mass:118.95271{54#} ; Abundance:0; {g} StandardWeight:140.116{1}),
(Symbol:'Ce'; AtomicNumber:58; MassNumber:120; Mass:119.94654{54#} ; Abundance:0; {g} StandardWeight:140.116{1}),
(Symbol:'Ce'; AtomicNumber:58; MassNumber:121; Mass:120.94335{43#} ; Abundance:0; {g} StandardWeight:140.116{1}),
(Symbol:'Ce'; AtomicNumber:58; MassNumber:122; Mass:121.93787{43#} ; Abundance:0; {g} StandardWeight:140.116{1}),
(Symbol:'Ce'; AtomicNumber:58; MassNumber:123; Mass:122.93528{32#} ; Abundance:0; {g} StandardWeight:140.116{1}),
(Symbol:'Ce'; AtomicNumber:58; MassNumber:124; Mass:123.93031{32#} ; Abundance:0; {g} StandardWeight:140.116{1}),
(Symbol:'Ce'; AtomicNumber:58; MassNumber:125; Mass:124.92844{21#} ; Abundance:0; {g} StandardWeight:140.116{1}),
(Symbol:'Ce'; AtomicNumber:58; MassNumber:126; Mass:125.923971{30} ; Abundance:0; {g} StandardWeight:140.116{1}),
(Symbol:'Ce'; AtomicNumber:58; MassNumber:127; Mass:126.922727{31} ; Abundance:0; {g} StandardWeight:140.116{1}),
(Symbol:'Ce'; AtomicNumber:58; MassNumber:128; Mass:127.918911{30} ; Abundance:0; {g} StandardWeight:140.116{1}),
(Symbol:'Ce'; AtomicNumber:58; MassNumber:129; Mass:128.918102{30} ; Abundance:0; {g} StandardWeight:140.116{1}),
(Symbol:'Ce'; AtomicNumber:58; MassNumber:130; Mass:129.914736{30} ; Abundance:0; {g} StandardWeight:140.116{1}),
(Symbol:'Ce'; AtomicNumber:58; MassNumber:131; Mass:130.914429{35} ; Abundance:0; {g} StandardWeight:140.116{1}),
(Symbol:'Ce'; AtomicNumber:58; MassNumber:132; Mass:131.911464{22} ; Abundance:0; {g} StandardWeight:140.116{1}),
(Symbol:'Ce'; AtomicNumber:58; MassNumber:133; Mass:132.911520{18} ; Abundance:0; {g} StandardWeight:140.116{1}),
(Symbol:'Ce'; AtomicNumber:58; MassNumber:134; Mass:133.908928{22} ; Abundance:0; {g} StandardWeight:140.116{1}),
(Symbol:'Ce'; AtomicNumber:58; MassNumber:135; Mass:134.909161{11} ; Abundance:0; {g} StandardWeight:140.116{1}),
(Symbol:'Ce'; AtomicNumber:58; MassNumber:136; Mass:135.90712921{41} ; Abundance:0.185; {g} StandardWeight:140.116{1}),
(Symbol:'Ce'; AtomicNumber:58; MassNumber:137; Mass:136.90776236{45} ; Abundance:0; {g} StandardWeight:140.116{1}),
(Symbol:'Ce'; AtomicNumber:58; MassNumber:138; Mass:137.905991{11} ; Abundance:0.251; {g} StandardWeight:140.116{1}),
(Symbol:'Ce'; AtomicNumber:58; MassNumber:139; Mass:138.9066551{78} ; Abundance:0; {g} StandardWeight:140.116{1}),
(Symbol:'Ce'; AtomicNumber:58; MassNumber:140; Mass:139.9054431{23} ; Abundance:88.45; {g} StandardWeight:140.116{1}),
(Symbol:'Ce'; AtomicNumber:58; MassNumber:141; Mass:140.9082807{23} ; Abundance:0; {g} StandardWeight:140.116{1}),
(Symbol:'Ce'; AtomicNumber:58; MassNumber:142; Mass:141.9092504{29} ; Abundance:11.114; {g} StandardWeight:140.116{1}),
(Symbol:'Ce'; AtomicNumber:58; MassNumber:143; Mass:142.9123921{29} ; Abundance:0; {g} StandardWeight:140.116{1}),
(Symbol:'Ce'; AtomicNumber:58; MassNumber:144; Mass:143.9136529{34} ; Abundance:0; {g} StandardWeight:140.116{1}),
(Symbol:'Ce'; AtomicNumber:58; MassNumber:145; Mass:144.917265{36} ; Abundance:0; {g} StandardWeight:140.116{1}),
(Symbol:'Ce'; AtomicNumber:58; MassNumber:146; Mass:145.918802{18} ; Abundance:0; {g} StandardWeight:140.116{1}),
(Symbol:'Ce'; AtomicNumber:58; MassNumber:147; Mass:146.9226899{92} ; Abundance:0; {g} StandardWeight:140.116{1}),
(Symbol:'Ce'; AtomicNumber:58; MassNumber:148; Mass:147.924424{12} ; Abundance:0; {g} StandardWeight:140.116{1}),
(Symbol:'Ce'; AtomicNumber:58; MassNumber:149; Mass:148.928427{11} ; Abundance:0; {g} StandardWeight:140.116{1}),
(Symbol:'Ce'; AtomicNumber:58; MassNumber:150; Mass:149.930384{13} ; Abundance:0; {g} StandardWeight:140.116{1}),
(Symbol:'Ce'; AtomicNumber:58; MassNumber:151; Mass:150.934272{19} ; Abundance:0; {g} StandardWeight:140.116{1}),
(Symbol:'Ce'; AtomicNumber:58; MassNumber:152; Mass:151.93660{21#} ; Abundance:0; {g} StandardWeight:140.116{1}),
(Symbol:'Ce'; AtomicNumber:58; MassNumber:153; Mass:152.94093{21#} ; Abundance:0; {g} StandardWeight:140.116{1}),
(Symbol:'Ce'; AtomicNumber:58; MassNumber:154; Mass:153.94380{32#} ; Abundance:0; {g} StandardWeight:140.116{1}),
(Symbol:'Ce'; AtomicNumber:58; MassNumber:155; Mass:154.94855{43#} ; Abundance:0; {g} StandardWeight:140.116{1}),
(Symbol:'Ce'; AtomicNumber:58; MassNumber:156; Mass:155.95183{43#} ; Abundance:0; {g} StandardWeight:140.116{1}),
(Symbol:'Ce'; AtomicNumber:58; MassNumber:157; Mass:156.95705{54#} ; Abundance:0; {g} StandardWeight:140.116{1}),
(Symbol:'Pr'; AtomicNumber:59; MassNumber:121; Mass:120.95532{54#} ; Abundance:0;  StandardWeight:140.90766{2}),
(Symbol:'Pr'; AtomicNumber:59; MassNumber:122; Mass:121.95175{54#} ; Abundance:0;  StandardWeight:140.90766{2}),
(Symbol:'Pr'; AtomicNumber:59; MassNumber:123; Mass:122.94596{43#} ; Abundance:0;  StandardWeight:140.90766{2}),
(Symbol:'Pr'; AtomicNumber:59; MassNumber:124; Mass:123.94294{43#} ; Abundance:0;  StandardWeight:140.90766{2}),
(Symbol:'Pr'; AtomicNumber:59; MassNumber:125; Mass:124.93770{32#} ; Abundance:0;  StandardWeight:140.90766{2}),
(Symbol:'Pr'; AtomicNumber:59; MassNumber:126; Mass:125.93524{21#} ; Abundance:0;  StandardWeight:140.90766{2}),
(Symbol:'Pr'; AtomicNumber:59; MassNumber:127; Mass:126.93071{21#} ; Abundance:0;  StandardWeight:140.90766{2}),
(Symbol:'Pr'; AtomicNumber:59; MassNumber:128; Mass:127.928791{32} ; Abundance:0;  StandardWeight:140.90766{2}),
(Symbol:'Pr'; AtomicNumber:59; MassNumber:129; Mass:128.925095{32} ; Abundance:0;  StandardWeight:140.90766{2}),
(Symbol:'Pr'; AtomicNumber:59; MassNumber:130; Mass:129.923590{69} ; Abundance:0;  StandardWeight:140.90766{2}),
(Symbol:'Pr'; AtomicNumber:59; MassNumber:131; Mass:130.920235{50} ; Abundance:0;  StandardWeight:140.90766{2}),
(Symbol:'Pr'; AtomicNumber:59; MassNumber:132; Mass:131.919255{61} ; Abundance:0;  StandardWeight:140.90766{2}),
(Symbol:'Pr'; AtomicNumber:59; MassNumber:133; Mass:132.916331{13} ; Abundance:0;  StandardWeight:140.90766{2}),
(Symbol:'Pr'; AtomicNumber:59; MassNumber:134; Mass:133.915697{22} ; Abundance:0;  StandardWeight:140.90766{2}),
(Symbol:'Pr'; AtomicNumber:59; MassNumber:135; Mass:134.913112{13} ; Abundance:0;  StandardWeight:140.90766{2}),
(Symbol:'Pr'; AtomicNumber:59; MassNumber:136; Mass:135.912677{12} ; Abundance:0;  StandardWeight:140.90766{2}),
(Symbol:'Pr'; AtomicNumber:59; MassNumber:137; Mass:136.9106792{87} ; Abundance:0;  StandardWeight:140.90766{2}),
(Symbol:'Pr'; AtomicNumber:59; MassNumber:138; Mass:137.910754{15} ; Abundance:0;  StandardWeight:140.90766{2}),
(Symbol:'Pr'; AtomicNumber:59; MassNumber:139; Mass:138.9089408{85} ; Abundance:0;  StandardWeight:140.90766{2}),
(Symbol:'Pr'; AtomicNumber:59; MassNumber:140; Mass:139.9090803{69} ; Abundance:0;  StandardWeight:140.90766{2}),
(Symbol:'Pr'; AtomicNumber:59; MassNumber:141; Mass:140.9076576{23} ; Abundance:100;  StandardWeight:140.90766{2}),
(Symbol:'Pr'; AtomicNumber:59; MassNumber:142; Mass:141.9100496{23} ; Abundance:0;  StandardWeight:140.90766{2}),
(Symbol:'Pr'; AtomicNumber:59; MassNumber:143; Mass:142.9108228{24} ; Abundance:0;  StandardWeight:140.90766{2}),
(Symbol:'Pr'; AtomicNumber:59; MassNumber:144; Mass:143.9133109{32} ; Abundance:0;  StandardWeight:140.90766{2}),
(Symbol:'Pr'; AtomicNumber:59; MassNumber:145; Mass:144.9145182{78} ; Abundance:0;  StandardWeight:140.90766{2}),
(Symbol:'Pr'; AtomicNumber:59; MassNumber:146; Mass:145.917680{37} ; Abundance:0;  StandardWeight:140.90766{2}),
(Symbol:'Pr'; AtomicNumber:59; MassNumber:147; Mass:146.919008{17} ; Abundance:0;  StandardWeight:140.90766{2}),
(Symbol:'Pr'; AtomicNumber:59; MassNumber:148; Mass:147.922130{16} ; Abundance:0;  StandardWeight:140.90766{2}),
(Symbol:'Pr'; AtomicNumber:59; MassNumber:149; Mass:148.923736{11} ; Abundance:0;  StandardWeight:140.90766{2}),
(Symbol:'Pr'; AtomicNumber:59; MassNumber:150; Mass:149.9266765{97} ; Abundance:0;  StandardWeight:140.90766{2}),
(Symbol:'Pr'; AtomicNumber:59; MassNumber:151; Mass:150.928309{13} ; Abundance:0;  StandardWeight:140.90766{2}),
(Symbol:'Pr'; AtomicNumber:59; MassNumber:152; Mass:151.931553{20} ; Abundance:0;  StandardWeight:140.90766{2}),
(Symbol:'Pr'; AtomicNumber:59; MassNumber:153; Mass:152.933904{13} ; Abundance:0;  StandardWeight:140.90766{2}),
(Symbol:'Pr'; AtomicNumber:59; MassNumber:154; Mass:153.93753{16} ; Abundance:0;  StandardWeight:140.90766{2}),
(Symbol:'Pr'; AtomicNumber:59; MassNumber:155; Mass:154.940509{18} ; Abundance:0;  StandardWeight:140.90766{2}),
(Symbol:'Pr'; AtomicNumber:59; MassNumber:156; Mass:155.94464{32#} ; Abundance:0;  StandardWeight:140.90766{2}),
(Symbol:'Pr'; AtomicNumber:59; MassNumber:157; Mass:156.94789{43#} ; Abundance:0;  StandardWeight:140.90766{2}),
(Symbol:'Pr'; AtomicNumber:59; MassNumber:158; Mass:157.95241{43#} ; Abundance:0;  StandardWeight:140.90766{2}),
(Symbol:'Pr'; AtomicNumber:59; MassNumber:159; Mass:158.95589{54#} ; Abundance:0;  StandardWeight:140.90766{2}),
(Symbol:'Nd'; AtomicNumber:60; MassNumber:124; Mass:123.95220{54#} ; Abundance:0; {g} StandardWeight:144.242{3}),
(Symbol:'Nd'; AtomicNumber:60; MassNumber:125; Mass:124.94890{43#} ; Abundance:0; {g} StandardWeight:144.242{3}),
(Symbol:'Nd'; AtomicNumber:60; MassNumber:126; Mass:125.94311{32#} ; Abundance:0; {g} StandardWeight:144.242{3}),
(Symbol:'Nd'; AtomicNumber:60; MassNumber:127; Mass:126.94038{32#} ; Abundance:0; {g} StandardWeight:144.242{3}),
(Symbol:'Nd'; AtomicNumber:60; MassNumber:128; Mass:127.93525{21#} ; Abundance:0; {g} StandardWeight:144.242{3}),
(Symbol:'Nd'; AtomicNumber:60; MassNumber:129; Mass:128.93310{22#} ; Abundance:0; {g} StandardWeight:144.242{3}),
(Symbol:'Nd'; AtomicNumber:60; MassNumber:130; Mass:129.928506{30} ; Abundance:0; {g} StandardWeight:144.242{3}),
(Symbol:'Nd'; AtomicNumber:60; MassNumber:131; Mass:130.927248{30} ; Abundance:0; {g} StandardWeight:144.242{3}),
(Symbol:'Nd'; AtomicNumber:60; MassNumber:132; Mass:131.923321{26} ; Abundance:0; {g} StandardWeight:144.242{3}),
(Symbol:'Nd'; AtomicNumber:60; MassNumber:133; Mass:132.922348{50} ; Abundance:0; {g} StandardWeight:144.242{3}),
(Symbol:'Nd'; AtomicNumber:60; MassNumber:134; Mass:133.918790{13} ; Abundance:0; {g} StandardWeight:144.242{3}),
(Symbol:'Nd'; AtomicNumber:60; MassNumber:135; Mass:134.918181{21} ; Abundance:0; {g} StandardWeight:144.242{3}),
(Symbol:'Nd'; AtomicNumber:60; MassNumber:136; Mass:135.914976{13} ; Abundance:0; {g} StandardWeight:144.242{3}),
(Symbol:'Nd'; AtomicNumber:60; MassNumber:137; Mass:136.914562{13} ; Abundance:0; {g} StandardWeight:144.242{3}),
(Symbol:'Nd'; AtomicNumber:60; MassNumber:138; Mass:137.911950{12} ; Abundance:0; {g} StandardWeight:144.242{3}),
(Symbol:'Nd'; AtomicNumber:60; MassNumber:139; Mass:138.911954{30} ; Abundance:0; {g} StandardWeight:144.242{3}),
(Symbol:'Nd'; AtomicNumber:60; MassNumber:140; Mass:139.909550{28} ; Abundance:0; {g} StandardWeight:144.242{3}),
(Symbol:'Nd'; AtomicNumber:60; MassNumber:141; Mass:140.9096147{38} ; Abundance:0; {g} StandardWeight:144.242{3}),
(Symbol:'Nd'; AtomicNumber:60; MassNumber:142; Mass:141.9077290{20} ; Abundance:27.152; {g} StandardWeight:144.242{3}),
(Symbol:'Nd'; AtomicNumber:60; MassNumber:143; Mass:142.9098200{20} ; Abundance:12.174; {g} StandardWeight:144.242{3}),
(Symbol:'Nd'; AtomicNumber:60; MassNumber:144; Mass:143.9100930{20} ; Abundance:23.798; {g} StandardWeight:144.242{3}),
(Symbol:'Nd'; AtomicNumber:60; MassNumber:145; Mass:144.9125793{20} ; Abundance:8.293; {g} StandardWeight:144.242{3}),
(Symbol:'Nd'; AtomicNumber:60; MassNumber:146; Mass:145.9131226{20} ; Abundance:17.189; {g} StandardWeight:144.242{3}),
(Symbol:'Nd'; AtomicNumber:60; MassNumber:147; Mass:146.9161061{20} ; Abundance:0; {g} StandardWeight:144.242{3}),
(Symbol:'Nd'; AtomicNumber:60; MassNumber:148; Mass:147.9168993{26} ; Abundance:5.756; {g} StandardWeight:144.242{3}),
(Symbol:'Nd'; AtomicNumber:60; MassNumber:149; Mass:148.9201548{26} ; Abundance:0; {g} StandardWeight:144.242{3}),
(Symbol:'Nd'; AtomicNumber:60; MassNumber:150; Mass:149.9209022{18} ; Abundance:5.638; {g} StandardWeight:144.242{3}),
(Symbol:'Nd'; AtomicNumber:60; MassNumber:151; Mass:150.9238403{18} ; Abundance:0; {g} StandardWeight:144.242{3}),
(Symbol:'Nd'; AtomicNumber:60; MassNumber:152; Mass:151.924692{26} ; Abundance:0; {g} StandardWeight:144.242{3}),
(Symbol:'Nd'; AtomicNumber:60; MassNumber:153; Mass:152.9277180{29} ; Abundance:0; {g} StandardWeight:144.242{3}),
(Symbol:'Nd'; AtomicNumber:60; MassNumber:154; Mass:153.92948{12} ; Abundance:0; {g} StandardWeight:144.242{3}),
(Symbol:'Nd'; AtomicNumber:60; MassNumber:155; Mass:154.9331357{98} ; Abundance:0; {g} StandardWeight:144.242{3}),
(Symbol:'Nd'; AtomicNumber:60; MassNumber:156; Mass:155.93508{21} ; Abundance:0; {g} StandardWeight:144.242{3}),
(Symbol:'Nd'; AtomicNumber:60; MassNumber:157; Mass:156.939386{27} ; Abundance:0; {g} StandardWeight:144.242{3}),
(Symbol:'Nd'; AtomicNumber:60; MassNumber:158; Mass:157.94197{32#} ; Abundance:0; {g} StandardWeight:144.242{3}),
(Symbol:'Nd'; AtomicNumber:60; MassNumber:159; Mass:158.94653{43#} ; Abundance:0; {g} StandardWeight:144.242{3}),
(Symbol:'Nd'; AtomicNumber:60; MassNumber:160; Mass:159.94940{43#} ; Abundance:0; {g} StandardWeight:144.242{3}),
(Symbol:'Nd'; AtomicNumber:60; MassNumber:161; Mass:160.95428{54#} ; Abundance:0; {g} StandardWeight:144.242{3}),
(Symbol:'Pm'; AtomicNumber:61; MassNumber:126; Mass:125.95792{54#} ; Abundance:0;  StandardWeight:145),
(Symbol:'Pm'; AtomicNumber:61; MassNumber:127; Mass:126.95192{43#} ; Abundance:0;  StandardWeight:145),
(Symbol:'Pm'; AtomicNumber:61; MassNumber:128; Mass:127.94870{32#} ; Abundance:0;  StandardWeight:145),
(Symbol:'Pm'; AtomicNumber:61; MassNumber:129; Mass:128.94323{32#} ; Abundance:0;  StandardWeight:145),
(Symbol:'Pm'; AtomicNumber:61; MassNumber:130; Mass:129.94053{21#} ; Abundance:0;  StandardWeight:145),
(Symbol:'Pm'; AtomicNumber:61; MassNumber:131; Mass:130.93567{21#} ; Abundance:0;  StandardWeight:145),
(Symbol:'Pm'; AtomicNumber:61; MassNumber:132; Mass:131.93384{16#} ; Abundance:0;  StandardWeight:145),
(Symbol:'Pm'; AtomicNumber:61; MassNumber:133; Mass:132.929782{54} ; Abundance:0;  StandardWeight:145),
(Symbol:'Pm'; AtomicNumber:61; MassNumber:134; Mass:133.928353{62} ; Abundance:0;  StandardWeight:145),
(Symbol:'Pm'; AtomicNumber:61; MassNumber:135; Mass:134.924823{70} ; Abundance:0;  StandardWeight:145),
(Symbol:'Pm'; AtomicNumber:61; MassNumber:136; Mass:135.923585{77} ; Abundance:0;  StandardWeight:145),
(Symbol:'Pm'; AtomicNumber:61; MassNumber:137; Mass:136.920480{14} ; Abundance:0;  StandardWeight:145),
(Symbol:'Pm'; AtomicNumber:61; MassNumber:138; Mass:137.919548{30} ; Abundance:0;  StandardWeight:145),
(Symbol:'Pm'; AtomicNumber:61; MassNumber:139; Mass:138.916800{15} ; Abundance:0;  StandardWeight:145),
(Symbol:'Pm'; AtomicNumber:61; MassNumber:140; Mass:139.916040{38} ; Abundance:0;  StandardWeight:145),
(Symbol:'Pm'; AtomicNumber:61; MassNumber:141; Mass:140.913555{15} ; Abundance:0;  StandardWeight:145),
(Symbol:'Pm'; AtomicNumber:61; MassNumber:142; Mass:141.912890{25} ; Abundance:0;  StandardWeight:145),
(Symbol:'Pm'; AtomicNumber:61; MassNumber:143; Mass:142.9109383{34} ; Abundance:0;  StandardWeight:145),
(Symbol:'Pm'; AtomicNumber:61; MassNumber:144; Mass:143.9125964{34} ; Abundance:0;  StandardWeight:145),
(Symbol:'Pm'; AtomicNumber:61; MassNumber:145; Mass:144.9127559{33} ; Abundance:0;  StandardWeight:145),
(Symbol:'Pm'; AtomicNumber:61; MassNumber:146; Mass:145.9147024{48} ; Abundance:0;  StandardWeight:145),
(Symbol:'Pm'; AtomicNumber:61; MassNumber:147; Mass:146.9151450{19} ; Abundance:0;  StandardWeight:145),
(Symbol:'Pm'; AtomicNumber:61; MassNumber:148; Mass:147.9174819{63} ; Abundance:0;  StandardWeight:145),
(Symbol:'Pm'; AtomicNumber:61; MassNumber:149; Mass:148.9183423{27} ; Abundance:0;  StandardWeight:145),
(Symbol:'Pm'; AtomicNumber:61; MassNumber:150; Mass:149.920991{22} ; Abundance:0;  StandardWeight:145),
(Symbol:'Pm'; AtomicNumber:61; MassNumber:151; Mass:150.9212175{51} ; Abundance:0;  StandardWeight:145),
(Symbol:'Pm'; AtomicNumber:61; MassNumber:152; Mass:151.923506{28} ; Abundance:0;  StandardWeight:145),
(Symbol:'Pm'; AtomicNumber:61; MassNumber:153; Mass:152.9241567{97} ; Abundance:0;  StandardWeight:145),
(Symbol:'Pm'; AtomicNumber:61; MassNumber:154; Mass:153.926472{48} ; Abundance:0;  StandardWeight:145),
(Symbol:'Pm'; AtomicNumber:61; MassNumber:155; Mass:154.9281370{51} ; Abundance:0;  StandardWeight:145),
(Symbol:'Pm'; AtomicNumber:61; MassNumber:156; Mass:155.9311175{39} ; Abundance:0;  StandardWeight:145),
(Symbol:'Pm'; AtomicNumber:61; MassNumber:157; Mass:156.9331214{75} ; Abundance:0;  StandardWeight:145),
(Symbol:'Pm'; AtomicNumber:61; MassNumber:158; Mass:157.936565{14} ; Abundance:0;  StandardWeight:145),
(Symbol:'Pm'; AtomicNumber:61; MassNumber:159; Mass:158.939287{11} ; Abundance:0;  StandardWeight:145),
(Symbol:'Pm'; AtomicNumber:61; MassNumber:160; Mass:159.94310{32#} ; Abundance:0;  StandardWeight:145),
(Symbol:'Pm'; AtomicNumber:61; MassNumber:161; Mass:160.94607{32#} ; Abundance:0;  StandardWeight:145),
(Symbol:'Pm'; AtomicNumber:61; MassNumber:162; Mass:161.95022{43#} ; Abundance:0;  StandardWeight:145),
(Symbol:'Pm'; AtomicNumber:61; MassNumber:163; Mass:162.95357{54#} ; Abundance:0;  StandardWeight:145),
(Symbol:'Sm'; AtomicNumber:62; MassNumber:128; Mass:127.95842{54#} ; Abundance:0; {g} StandardWeight:150.36{2}),
(Symbol:'Sm'; AtomicNumber:62; MassNumber:129; Mass:128.95476{54#} ; Abundance:0; {g} StandardWeight:150.36{2}),
(Symbol:'Sm'; AtomicNumber:62; MassNumber:130; Mass:129.94900{43#} ; Abundance:0; {g} StandardWeight:150.36{2}),
(Symbol:'Sm'; AtomicNumber:62; MassNumber:131; Mass:130.94618{43#} ; Abundance:0; {g} StandardWeight:150.36{2}),
(Symbol:'Sm'; AtomicNumber:62; MassNumber:132; Mass:131.94087{32#} ; Abundance:0; {g} StandardWeight:150.36{2}),
(Symbol:'Sm'; AtomicNumber:62; MassNumber:133; Mass:132.93856{32#} ; Abundance:0; {g} StandardWeight:150.36{2}),
(Symbol:'Sm'; AtomicNumber:62; MassNumber:134; Mass:133.93411{21#} ; Abundance:0; {g} StandardWeight:150.36{2}),
(Symbol:'Sm'; AtomicNumber:62; MassNumber:135; Mass:134.93252{17} ; Abundance:0; {g} StandardWeight:150.36{2}),
(Symbol:'Sm'; AtomicNumber:62; MassNumber:136; Mass:135.928276{13} ; Abundance:0; {g} StandardWeight:150.36{2}),
(Symbol:'Sm'; AtomicNumber:62; MassNumber:137; Mass:136.926971{46} ; Abundance:0; {g} StandardWeight:150.36{2}),
(Symbol:'Sm'; AtomicNumber:62; MassNumber:138; Mass:137.923244{13} ; Abundance:0; {g} StandardWeight:150.36{2}),
(Symbol:'Sm'; AtomicNumber:62; MassNumber:139; Mass:138.922297{12} ; Abundance:0; {g} StandardWeight:150.36{2}),
(Symbol:'Sm'; AtomicNumber:62; MassNumber:140; Mass:139.918995{13} ; Abundance:0; {g} StandardWeight:150.36{2}),
(Symbol:'Sm'; AtomicNumber:62; MassNumber:141; Mass:140.9184816{92} ; Abundance:0; {g} StandardWeight:150.36{2}),
(Symbol:'Sm'; AtomicNumber:62; MassNumber:142; Mass:141.9152044{36} ; Abundance:0; {g} StandardWeight:150.36{2}),
(Symbol:'Sm'; AtomicNumber:62; MassNumber:143; Mass:142.9146353{33} ; Abundance:0; {g} StandardWeight:150.36{2}),
(Symbol:'Sm'; AtomicNumber:62; MassNumber:144; Mass:143.9120065{21} ; Abundance:3.07; {g} StandardWeight:150.36{2}),
(Symbol:'Sm'; AtomicNumber:62; MassNumber:145; Mass:144.9134173{21} ; Abundance:0; {g} StandardWeight:150.36{2}),
(Symbol:'Sm'; AtomicNumber:62; MassNumber:146; Mass:145.9130470{35} ; Abundance:0; {g} StandardWeight:150.36{2}),
(Symbol:'Sm'; AtomicNumber:62; MassNumber:147; Mass:146.9149044{19} ; Abundance:14.99; {g} StandardWeight:150.36{2}),
(Symbol:'Sm'; AtomicNumber:62; MassNumber:148; Mass:147.9148292{19} ; Abundance:11.24; {g} StandardWeight:150.36{2}),
(Symbol:'Sm'; AtomicNumber:62; MassNumber:149; Mass:148.9171921{18} ; Abundance:13.82; {g} StandardWeight:150.36{2}),
(Symbol:'Sm'; AtomicNumber:62; MassNumber:150; Mass:149.9172829{18} ; Abundance:7.38; {g} StandardWeight:150.36{2}),
(Symbol:'Sm'; AtomicNumber:62; MassNumber:151; Mass:150.9199398{18} ; Abundance:0; {g} StandardWeight:150.36{2}),
(Symbol:'Sm'; AtomicNumber:62; MassNumber:152; Mass:151.9197397{18} ; Abundance:26.75; {g} StandardWeight:150.36{2}),
(Symbol:'Sm'; AtomicNumber:62; MassNumber:153; Mass:152.9221047{18} ; Abundance:0; {g} StandardWeight:150.36{2}),
(Symbol:'Sm'; AtomicNumber:62; MassNumber:154; Mass:153.9222169{20} ; Abundance:22.75; {g} StandardWeight:150.36{2}),
(Symbol:'Sm'; AtomicNumber:62; MassNumber:155; Mass:154.9246477{20} ; Abundance:0; {g} StandardWeight:150.36{2}),
(Symbol:'Sm'; AtomicNumber:62; MassNumber:156; Mass:155.925536{10} ; Abundance:0; {g} StandardWeight:150.36{2}),
(Symbol:'Sm'; AtomicNumber:62; MassNumber:157; Mass:156.9284187{48} ; Abundance:0; {g} StandardWeight:150.36{2}),
(Symbol:'Sm'; AtomicNumber:62; MassNumber:158; Mass:157.9299510{53} ; Abundance:0; {g} StandardWeight:150.36{2}),
(Symbol:'Sm'; AtomicNumber:62; MassNumber:159; Mass:158.9332172{64} ; Abundance:0; {g} StandardWeight:150.36{2}),
(Symbol:'Sm'; AtomicNumber:62; MassNumber:160; Mass:159.9353353{64} ; Abundance:0; {g} StandardWeight:150.36{2}),
(Symbol:'Sm'; AtomicNumber:62; MassNumber:161; Mass:160.9391602{73} ; Abundance:0; {g} StandardWeight:150.36{2}),
(Symbol:'Sm'; AtomicNumber:62; MassNumber:162; Mass:161.94146{21#} ; Abundance:0; {g} StandardWeight:150.36{2}),
(Symbol:'Sm'; AtomicNumber:62; MassNumber:163; Mass:162.94555{32#} ; Abundance:0; {g} StandardWeight:150.36{2}),
(Symbol:'Sm'; AtomicNumber:62; MassNumber:164; Mass:163.94836{32#} ; Abundance:0; {g} StandardWeight:150.36{2}),
(Symbol:'Sm'; AtomicNumber:62; MassNumber:165; Mass:164.95297{43#} ; Abundance:0; {g} StandardWeight:150.36{2}),
(Symbol:'Eu'; AtomicNumber:63; MassNumber:130; Mass:129.96369{54#} ; Abundance:0; {g} StandardWeight:151.964{1}),
(Symbol:'Eu'; AtomicNumber:63; MassNumber:131; Mass:130.95784{43#} ; Abundance:0; {g} StandardWeight:151.964{1}),
(Symbol:'Eu'; AtomicNumber:63; MassNumber:132; Mass:131.95467{43#} ; Abundance:0; {g} StandardWeight:151.964{1}),
(Symbol:'Eu'; AtomicNumber:63; MassNumber:133; Mass:132.94929{32#} ; Abundance:0; {g} StandardWeight:151.964{1}),
(Symbol:'Eu'; AtomicNumber:63; MassNumber:134; Mass:133.94640{32#} ; Abundance:0; {g} StandardWeight:151.964{1}),
(Symbol:'Eu'; AtomicNumber:63; MassNumber:135; Mass:134.94187{21#} ; Abundance:0; {g} StandardWeight:151.964{1}),
(Symbol:'Eu'; AtomicNumber:63; MassNumber:136; Mass:135.93962{21#} ; Abundance:0; {g} StandardWeight:151.964{1}),
(Symbol:'Eu'; AtomicNumber:63; MassNumber:137; Mass:136.93546{21#} ; Abundance:0; {g} StandardWeight:151.964{1}),
(Symbol:'Eu'; AtomicNumber:63; MassNumber:138; Mass:137.933709{30} ; Abundance:0; {g} StandardWeight:151.964{1}),
(Symbol:'Eu'; AtomicNumber:63; MassNumber:139; Mass:138.929792{14} ; Abundance:0; {g} StandardWeight:151.964{1}),
(Symbol:'Eu'; AtomicNumber:63; MassNumber:140; Mass:139.928088{55} ; Abundance:0; {g} StandardWeight:151.964{1}),
(Symbol:'Eu'; AtomicNumber:63; MassNumber:141; Mass:140.924932{14} ; Abundance:0; {g} StandardWeight:151.964{1}),
(Symbol:'Eu'; AtomicNumber:63; MassNumber:142; Mass:141.923442{32} ; Abundance:0; {g} StandardWeight:151.964{1}),
(Symbol:'Eu'; AtomicNumber:63; MassNumber:143; Mass:142.920299{12} ; Abundance:0; {g} StandardWeight:151.964{1}),
(Symbol:'Eu'; AtomicNumber:63; MassNumber:144; Mass:143.918820{12} ; Abundance:0; {g} StandardWeight:151.964{1}),
(Symbol:'Eu'; AtomicNumber:63; MassNumber:145; Mass:144.9162726{36} ; Abundance:0; {g} StandardWeight:151.964{1}),
(Symbol:'Eu'; AtomicNumber:63; MassNumber:146; Mass:145.9172110{65} ; Abundance:0; {g} StandardWeight:151.964{1}),
(Symbol:'Eu'; AtomicNumber:63; MassNumber:147; Mass:146.9167527{31} ; Abundance:0; {g} StandardWeight:151.964{1}),
(Symbol:'Eu'; AtomicNumber:63; MassNumber:148; Mass:147.918089{11} ; Abundance:0; {g} StandardWeight:151.964{1}),
(Symbol:'Eu'; AtomicNumber:63; MassNumber:149; Mass:148.9179378{44} ; Abundance:0; {g} StandardWeight:151.964{1}),
(Symbol:'Eu'; AtomicNumber:63; MassNumber:150; Mass:149.9197077{68} ; Abundance:0; {g} StandardWeight:151.964{1}),
(Symbol:'Eu'; AtomicNumber:63; MassNumber:151; Mass:150.9198578{18} ; Abundance:47.81; {g} StandardWeight:151.964{1}),
(Symbol:'Eu'; AtomicNumber:63; MassNumber:152; Mass:151.9217522{18} ; Abundance:0; {g} StandardWeight:151.964{1}),
(Symbol:'Eu'; AtomicNumber:63; MassNumber:153; Mass:152.9212380{18} ; Abundance:52.19; {g} StandardWeight:151.964{1}),
(Symbol:'Eu'; AtomicNumber:63; MassNumber:154; Mass:153.9229870{19} ; Abundance:0; {g} StandardWeight:151.964{1}),
(Symbol:'Eu'; AtomicNumber:63; MassNumber:155; Mass:154.9229011{19} ; Abundance:0; {g} StandardWeight:151.964{1}),
(Symbol:'Eu'; AtomicNumber:63; MassNumber:156; Mass:155.9247605{59} ; Abundance:0; {g} StandardWeight:151.964{1}),
(Symbol:'Eu'; AtomicNumber:63; MassNumber:157; Mass:156.9254334{46} ; Abundance:0; {g} StandardWeight:151.964{1}),
(Symbol:'Eu'; AtomicNumber:63; MassNumber:158; Mass:157.927799{11} ; Abundance:0; {g} StandardWeight:151.964{1}),
(Symbol:'Eu'; AtomicNumber:63; MassNumber:159; Mass:158.9291001{47} ; Abundance:0; {g} StandardWeight:151.964{1}),
(Symbol:'Eu'; AtomicNumber:63; MassNumber:160; Mass:159.931851{10} ; Abundance:0; {g} StandardWeight:151.964{1}),
(Symbol:'Eu'; AtomicNumber:63; MassNumber:161; Mass:160.933664{11} ; Abundance:0; {g} StandardWeight:151.964{1}),
(Symbol:'Eu'; AtomicNumber:63; MassNumber:162; Mass:161.936989{65} ; Abundance:0; {g} StandardWeight:151.964{1}),
(Symbol:'Eu'; AtomicNumber:63; MassNumber:163; Mass:162.939196{76} ; Abundance:0; {g} StandardWeight:151.964{1}),
(Symbol:'Eu'; AtomicNumber:63; MassNumber:164; Mass:163.94274{22#} ; Abundance:0; {g} StandardWeight:151.964{1}),
(Symbol:'Eu'; AtomicNumber:63; MassNumber:165; Mass:164.94559{35#} ; Abundance:0; {g} StandardWeight:151.964{1}),
(Symbol:'Eu'; AtomicNumber:63; MassNumber:166; Mass:165.94962{32#} ; Abundance:0; {g} StandardWeight:151.964{1}),
(Symbol:'Eu'; AtomicNumber:63; MassNumber:167; Mass:166.95289{43#} ; Abundance:0; {g} StandardWeight:151.964{1}),
(Symbol:'Gd'; AtomicNumber:64; MassNumber:133; Mass:132.96133{54#} ; Abundance:0; {g} StandardWeight:157.25{3}),
(Symbol:'Gd'; AtomicNumber:64; MassNumber:134; Mass:133.95566{43#} ; Abundance:0; {g} StandardWeight:157.25{3}),
(Symbol:'Gd'; AtomicNumber:64; MassNumber:135; Mass:134.95245{43#} ; Abundance:0; {g} StandardWeight:157.25{3}),
(Symbol:'Gd'; AtomicNumber:64; MassNumber:136; Mass:135.94730{32#} ; Abundance:0; {g} StandardWeight:157.25{3}),
(Symbol:'Gd'; AtomicNumber:64; MassNumber:137; Mass:136.94502{32#} ; Abundance:0; {g} StandardWeight:157.25{3}),
(Symbol:'Gd'; AtomicNumber:64; MassNumber:138; Mass:137.94025{21#} ; Abundance:0; {g} StandardWeight:157.25{3}),
(Symbol:'Gd'; AtomicNumber:64; MassNumber:139; Mass:138.93813{21#} ; Abundance:0; {g} StandardWeight:157.25{3}),
(Symbol:'Gd'; AtomicNumber:64; MassNumber:140; Mass:139.933674{30} ; Abundance:0; {g} StandardWeight:157.25{3}),
(Symbol:'Gd'; AtomicNumber:64; MassNumber:141; Mass:140.932126{21} ; Abundance:0; {g} StandardWeight:157.25{3}),
(Symbol:'Gd'; AtomicNumber:64; MassNumber:142; Mass:141.928116{30} ; Abundance:0; {g} StandardWeight:157.25{3}),
(Symbol:'Gd'; AtomicNumber:64; MassNumber:143; Mass:142.92675{22} ; Abundance:0; {g} StandardWeight:157.25{3}),
(Symbol:'Gd'; AtomicNumber:64; MassNumber:144; Mass:143.922963{30} ; Abundance:0; {g} StandardWeight:157.25{3}),
(Symbol:'Gd'; AtomicNumber:64; MassNumber:145; Mass:144.921713{21} ; Abundance:0; {g} StandardWeight:157.25{3}),
(Symbol:'Gd'; AtomicNumber:64; MassNumber:146; Mass:145.9183188{46} ; Abundance:0; {g} StandardWeight:157.25{3}),
(Symbol:'Gd'; AtomicNumber:64; MassNumber:147; Mass:146.9191014{25} ; Abundance:0; {g} StandardWeight:157.25{3}),
(Symbol:'Gd'; AtomicNumber:64; MassNumber:148; Mass:147.9181215{21} ; Abundance:0; {g} StandardWeight:157.25{3}),
(Symbol:'Gd'; AtomicNumber:64; MassNumber:149; Mass:148.9193481{38} ; Abundance:0; {g} StandardWeight:157.25{3}),
(Symbol:'Gd'; AtomicNumber:64; MassNumber:150; Mass:149.9186644{66} ; Abundance:0; {g} StandardWeight:157.25{3}),
(Symbol:'Gd'; AtomicNumber:64; MassNumber:151; Mass:150.9203560{35} ; Abundance:0; {g} StandardWeight:157.25{3}),
(Symbol:'Gd'; AtomicNumber:64; MassNumber:152; Mass:151.9197995{18} ; Abundance:0.2; {g} StandardWeight:157.25{3}),
(Symbol:'Gd'; AtomicNumber:64; MassNumber:153; Mass:152.9217580{18} ; Abundance:0; {g} StandardWeight:157.25{3}),
(Symbol:'Gd'; AtomicNumber:64; MassNumber:154; Mass:153.9208741{17} ; Abundance:2.18; {g} StandardWeight:157.25{3}),
(Symbol:'Gd'; AtomicNumber:64; MassNumber:155; Mass:154.9226305{17} ; Abundance:14.8; {g} StandardWeight:157.25{3}),
(Symbol:'Gd'; AtomicNumber:64; MassNumber:156; Mass:155.9221312{17} ; Abundance:20.47; {g} StandardWeight:157.25{3}),
(Symbol:'Gd'; AtomicNumber:64; MassNumber:157; Mass:156.9239686{17} ; Abundance:15.65; {g} StandardWeight:157.25{3}),
(Symbol:'Gd'; AtomicNumber:64; MassNumber:158; Mass:157.9241123{17} ; Abundance:24.84; {g} StandardWeight:157.25{3}),
(Symbol:'Gd'; AtomicNumber:64; MassNumber:159; Mass:158.9263970{17} ; Abundance:0; {g} StandardWeight:157.25{3}),
(Symbol:'Gd'; AtomicNumber:64; MassNumber:160; Mass:159.9270624{18} ; Abundance:21.86; {g} StandardWeight:157.25{3}),
(Symbol:'Gd'; AtomicNumber:64; MassNumber:161; Mass:160.9296775{21} ; Abundance:0; {g} StandardWeight:157.25{3}),
(Symbol:'Gd'; AtomicNumber:64; MassNumber:162; Mass:161.9309930{45} ; Abundance:0; {g} StandardWeight:157.25{3}),
(Symbol:'Gd'; AtomicNumber:64; MassNumber:163; Mass:162.9341769{90} ; Abundance:0; {g} StandardWeight:157.25{3}),
(Symbol:'Gd'; AtomicNumber:64; MassNumber:164; Mass:163.93583{21#} ; Abundance:0; {g} StandardWeight:157.25{3}),
(Symbol:'Gd'; AtomicNumber:64; MassNumber:165; Mass:164.93936{32#} ; Abundance:0; {g} StandardWeight:157.25{3}),
(Symbol:'Gd'; AtomicNumber:64; MassNumber:166; Mass:165.94146{64#} ; Abundance:0; {g} StandardWeight:157.25{3}),
(Symbol:'Gd'; AtomicNumber:64; MassNumber:167; Mass:166.94545{43#} ; Abundance:0; {g} StandardWeight:157.25{3}),
(Symbol:'Gd'; AtomicNumber:64; MassNumber:168; Mass:167.94808{43#} ; Abundance:0; {g} StandardWeight:157.25{3}),
(Symbol:'Gd'; AtomicNumber:64; MassNumber:169; Mass:168.95260{54#} ; Abundance:0; {g} StandardWeight:157.25{3}),
(Symbol:'Tb'; AtomicNumber:65; MassNumber:135; Mass:134.96476{43#} ; Abundance:0;  StandardWeight:158.92535{2}),
(Symbol:'Tb'; AtomicNumber:65; MassNumber:136; Mass:135.96129{54#} ; Abundance:0;  StandardWeight:158.92535{2}),
(Symbol:'Tb'; AtomicNumber:65; MassNumber:137; Mass:136.95602{54#} ; Abundance:0;  StandardWeight:158.92535{2}),
(Symbol:'Tb'; AtomicNumber:65; MassNumber:138; Mass:137.95312{32#} ; Abundance:0;  StandardWeight:158.92535{2}),
(Symbol:'Tb'; AtomicNumber:65; MassNumber:139; Mass:138.94833{32#} ; Abundance:0;  StandardWeight:158.92535{2}),
(Symbol:'Tb'; AtomicNumber:65; MassNumber:140; Mass:139.94581{86} ; Abundance:0;  StandardWeight:158.92535{2}),
(Symbol:'Tb'; AtomicNumber:65; MassNumber:141; Mass:140.94145{11} ; Abundance:0;  StandardWeight:158.92535{2}),
(Symbol:'Tb'; AtomicNumber:65; MassNumber:142; Mass:141.93928{75} ; Abundance:0;  StandardWeight:158.92535{2}),
(Symbol:'Tb'; AtomicNumber:65; MassNumber:143; Mass:142.935137{55} ; Abundance:0;  StandardWeight:158.92535{2}),
(Symbol:'Tb'; AtomicNumber:65; MassNumber:144; Mass:143.933045{30} ; Abundance:0;  StandardWeight:158.92535{2}),
(Symbol:'Tb'; AtomicNumber:65; MassNumber:145; Mass:144.92882{10} ; Abundance:0;  StandardWeight:158.92535{2}),
(Symbol:'Tb'; AtomicNumber:65; MassNumber:146; Mass:145.927253{48} ; Abundance:0;  StandardWeight:158.92535{2}),
(Symbol:'Tb'; AtomicNumber:65; MassNumber:147; Mass:146.9240548{87} ; Abundance:0;  StandardWeight:158.92535{2}),
(Symbol:'Tb'; AtomicNumber:65; MassNumber:148; Mass:147.924282{14} ; Abundance:0;  StandardWeight:158.92535{2}),
(Symbol:'Tb'; AtomicNumber:65; MassNumber:149; Mass:148.9232535{41} ; Abundance:0;  StandardWeight:158.92535{2}),
(Symbol:'Tb'; AtomicNumber:65; MassNumber:150; Mass:149.9236649{80} ; Abundance:0;  StandardWeight:158.92535{2}),
(Symbol:'Tb'; AtomicNumber:65; MassNumber:151; Mass:150.9231096{46} ; Abundance:0;  StandardWeight:158.92535{2}),
(Symbol:'Tb'; AtomicNumber:65; MassNumber:152; Mass:151.924083{43} ; Abundance:0;  StandardWeight:158.92535{2}),
(Symbol:'Tb'; AtomicNumber:65; MassNumber:153; Mass:152.9234424{44} ; Abundance:0;  StandardWeight:158.92535{2}),
(Symbol:'Tb'; AtomicNumber:65; MassNumber:154; Mass:153.924685{49} ; Abundance:0;  StandardWeight:158.92535{2}),
(Symbol:'Tb'; AtomicNumber:65; MassNumber:155; Mass:154.923511{11} ; Abundance:0;  StandardWeight:158.92535{2}),
(Symbol:'Tb'; AtomicNumber:65; MassNumber:156; Mass:155.9247552{43} ; Abundance:0;  StandardWeight:158.92535{2}),
(Symbol:'Tb'; AtomicNumber:65; MassNumber:157; Mass:156.9240330{18} ; Abundance:0;  StandardWeight:158.92535{2}),
(Symbol:'Tb'; AtomicNumber:65; MassNumber:158; Mass:157.9254209{20} ; Abundance:0;  StandardWeight:158.92535{2}),
(Symbol:'Tb'; AtomicNumber:65; MassNumber:159; Mass:158.9253547{19} ; Abundance:100;  StandardWeight:158.92535{2}),
(Symbol:'Tb'; AtomicNumber:65; MassNumber:160; Mass:159.9271756{19} ; Abundance:0;  StandardWeight:158.92535{2}),
(Symbol:'Tb'; AtomicNumber:65; MassNumber:161; Mass:160.9275778{20} ; Abundance:0;  StandardWeight:158.92535{2}),
(Symbol:'Tb'; AtomicNumber:65; MassNumber:162; Mass:161.929495{39} ; Abundance:0;  StandardWeight:158.92535{2}),
(Symbol:'Tb'; AtomicNumber:65; MassNumber:163; Mass:162.9306547{47} ; Abundance:0;  StandardWeight:158.92535{2}),
(Symbol:'Tb'; AtomicNumber:65; MassNumber:164; Mass:163.93336{11} ; Abundance:0;  StandardWeight:158.92535{2}),
(Symbol:'Tb'; AtomicNumber:65; MassNumber:165; Mass:164.93498{21#} ; Abundance:0;  StandardWeight:158.92535{2}),
(Symbol:'Tb'; AtomicNumber:65; MassNumber:166; Mass:165.937860{75} ; Abundance:0;  StandardWeight:158.92535{2}),
(Symbol:'Tb'; AtomicNumber:65; MassNumber:167; Mass:166.93996{21#} ; Abundance:0;  StandardWeight:158.92535{2}),
(Symbol:'Tb'; AtomicNumber:65; MassNumber:168; Mass:167.94340{32#} ; Abundance:0;  StandardWeight:158.92535{2}),
(Symbol:'Tb'; AtomicNumber:65; MassNumber:169; Mass:168.94597{32#} ; Abundance:0;  StandardWeight:158.92535{2}),
(Symbol:'Tb'; AtomicNumber:65; MassNumber:170; Mass:169.94984{43#} ; Abundance:0;  StandardWeight:158.92535{2}),
(Symbol:'Tb'; AtomicNumber:65; MassNumber:171; Mass:170.95273{54#} ; Abundance:0;  StandardWeight:158.92535{2}),
(Symbol:'Dy'; AtomicNumber:66; MassNumber:138; Mass:137.96250{43#} ; Abundance:0; {g} StandardWeight:162.500{1}),
(Symbol:'Dy'; AtomicNumber:66; MassNumber:139; Mass:138.95959{54#} ; Abundance:0; {g} StandardWeight:162.500{1}),
(Symbol:'Dy'; AtomicNumber:66; MassNumber:140; Mass:139.95402{54#} ; Abundance:0; {g} StandardWeight:162.500{1}),
(Symbol:'Dy'; AtomicNumber:66; MassNumber:141; Mass:140.95128{32#} ; Abundance:0; {g} StandardWeight:162.500{1}),
(Symbol:'Dy'; AtomicNumber:66; MassNumber:142; Mass:141.94619{78#} ; Abundance:0; {g} StandardWeight:162.500{1}),
(Symbol:'Dy'; AtomicNumber:66; MassNumber:143; Mass:142.943994{14} ; Abundance:0; {g} StandardWeight:162.500{1}),
(Symbol:'Dy'; AtomicNumber:66; MassNumber:144; Mass:143.9392695{77} ; Abundance:0; {g} StandardWeight:162.500{1}),
(Symbol:'Dy'; AtomicNumber:66; MassNumber:145; Mass:144.9374740{70} ; Abundance:0; {g} StandardWeight:162.500{1}),
(Symbol:'Dy'; AtomicNumber:66; MassNumber:146; Mass:145.9328445{72} ; Abundance:0; {g} StandardWeight:162.500{1}),
(Symbol:'Dy'; AtomicNumber:66; MassNumber:147; Mass:146.9310827{95} ; Abundance:0; {g} StandardWeight:162.500{1}),
(Symbol:'Dy'; AtomicNumber:66; MassNumber:148; Mass:147.927157{10} ; Abundance:0; {g} StandardWeight:162.500{1}),
(Symbol:'Dy'; AtomicNumber:66; MassNumber:149; Mass:148.927322{10} ; Abundance:0; {g} StandardWeight:162.500{1}),
(Symbol:'Dy'; AtomicNumber:66; MassNumber:150; Mass:149.9255933{48} ; Abundance:0; {g} StandardWeight:162.500{1}),
(Symbol:'Dy'; AtomicNumber:66; MassNumber:151; Mass:150.9261916{38} ; Abundance:0; {g} StandardWeight:162.500{1}),
(Symbol:'Dy'; AtomicNumber:66; MassNumber:152; Mass:151.9247253{51} ; Abundance:0; {g} StandardWeight:162.500{1}),
(Symbol:'Dy'; AtomicNumber:66; MassNumber:153; Mass:152.9257724{45} ; Abundance:0; {g} StandardWeight:162.500{1}),
(Symbol:'Dy'; AtomicNumber:66; MassNumber:154; Mass:153.9244293{80} ; Abundance:0; {g} StandardWeight:162.500{1}),
(Symbol:'Dy'; AtomicNumber:66; MassNumber:155; Mass:154.925759{10} ; Abundance:0; {g} StandardWeight:162.500{1}),
(Symbol:'Dy'; AtomicNumber:66; MassNumber:156; Mass:155.9242847{17} ; Abundance:0.056; {g} StandardWeight:162.500{1}),
(Symbol:'Dy'; AtomicNumber:66; MassNumber:157; Mass:156.9254707{57} ; Abundance:0; {g} StandardWeight:162.500{1}),
(Symbol:'Dy'; AtomicNumber:66; MassNumber:158; Mass:157.9244159{31} ; Abundance:0.095; {g} StandardWeight:162.500{1}),
(Symbol:'Dy'; AtomicNumber:66; MassNumber:159; Mass:158.9257470{22} ; Abundance:0; {g} StandardWeight:162.500{1}),
(Symbol:'Dy'; AtomicNumber:66; MassNumber:160; Mass:159.9252046{20} ; Abundance:2.329; {g} StandardWeight:162.500{1}),
(Symbol:'Dy'; AtomicNumber:66; MassNumber:161; Mass:160.9269405{20} ; Abundance:18.889; {g} StandardWeight:162.500{1}),
(Symbol:'Dy'; AtomicNumber:66; MassNumber:162; Mass:161.9268056{20} ; Abundance:25.475; {g} StandardWeight:162.500{1}),
(Symbol:'Dy'; AtomicNumber:66; MassNumber:163; Mass:162.9287383{20} ; Abundance:24.896; {g} StandardWeight:162.500{1}),
(Symbol:'Dy'; AtomicNumber:66; MassNumber:164; Mass:163.9291819{20} ; Abundance:28.26; {g} StandardWeight:162.500{1}),
(Symbol:'Dy'; AtomicNumber:66; MassNumber:165; Mass:164.9317105{20} ; Abundance:0; {g} StandardWeight:162.500{1}),
(Symbol:'Dy'; AtomicNumber:66; MassNumber:166; Mass:165.9328139{21} ; Abundance:0; {g} StandardWeight:162.500{1}),
(Symbol:'Dy'; AtomicNumber:66; MassNumber:167; Mass:166.935661{65} ; Abundance:0; {g} StandardWeight:162.500{1}),
(Symbol:'Dy'; AtomicNumber:66; MassNumber:168; Mass:167.93713{15} ; Abundance:0; {g} StandardWeight:162.500{1}),
(Symbol:'Dy'; AtomicNumber:66; MassNumber:169; Mass:168.94031{32} ; Abundance:0; {g} StandardWeight:162.500{1}),
(Symbol:'Dy'; AtomicNumber:66; MassNumber:170; Mass:169.94239{21#} ; Abundance:0; {g} StandardWeight:162.500{1}),
(Symbol:'Dy'; AtomicNumber:66; MassNumber:171; Mass:170.94612{32#} ; Abundance:0; {g} StandardWeight:162.500{1}),
(Symbol:'Dy'; AtomicNumber:66; MassNumber:172; Mass:171.94846{32#} ; Abundance:0; {g} StandardWeight:162.500{1}),
(Symbol:'Dy'; AtomicNumber:66; MassNumber:173; Mass:172.95283{43#} ; Abundance:0; {g} StandardWeight:162.500{1}),
(Symbol:'Ho'; AtomicNumber:67; MassNumber:140; Mass:139.96859{54#} ; Abundance:0;  StandardWeight:164.93033{2}),
(Symbol:'Ho'; AtomicNumber:67; MassNumber:141; Mass:140.96311{54#} ; Abundance:0;  StandardWeight:164.93033{2}),
(Symbol:'Ho'; AtomicNumber:67; MassNumber:142; Mass:141.96001{54#} ; Abundance:0;  StandardWeight:164.93033{2}),
(Symbol:'Ho'; AtomicNumber:67; MassNumber:143; Mass:142.95486{43#} ; Abundance:0;  StandardWeight:164.93033{2}),
(Symbol:'Ho'; AtomicNumber:67; MassNumber:144; Mass:143.9521097{91} ; Abundance:0;  StandardWeight:164.93033{2}),
(Symbol:'Ho'; AtomicNumber:67; MassNumber:145; Mass:144.9472674{80} ; Abundance:0;  StandardWeight:164.93033{2}),
(Symbol:'Ho'; AtomicNumber:67; MassNumber:146; Mass:145.9449935{71} ; Abundance:0;  StandardWeight:164.93033{2}),
(Symbol:'Ho'; AtomicNumber:67; MassNumber:147; Mass:146.9401423{54} ; Abundance:0;  StandardWeight:164.93033{2}),
(Symbol:'Ho'; AtomicNumber:67; MassNumber:148; Mass:147.937744{90} ; Abundance:0;  StandardWeight:164.93033{2}),
(Symbol:'Ho'; AtomicNumber:67; MassNumber:149; Mass:148.933803{16} ; Abundance:0;  StandardWeight:164.93033{2}),
(Symbol:'Ho'; AtomicNumber:67; MassNumber:150; Mass:149.933498{15} ; Abundance:0;  StandardWeight:164.93033{2}),
(Symbol:'Ho'; AtomicNumber:67; MassNumber:151; Mass:150.9316983{89} ; Abundance:0;  StandardWeight:164.93033{2}),
(Symbol:'Ho'; AtomicNumber:67; MassNumber:152; Mass:151.931724{14} ; Abundance:0;  StandardWeight:164.93033{2}),
(Symbol:'Ho'; AtomicNumber:67; MassNumber:153; Mass:152.9302064{56} ; Abundance:0;  StandardWeight:164.93033{2}),
(Symbol:'Ho'; AtomicNumber:67; MassNumber:154; Mass:153.9306068{89} ; Abundance:0;  StandardWeight:164.93033{2}),
(Symbol:'Ho'; AtomicNumber:67; MassNumber:155; Mass:154.929104{19} ; Abundance:0;  StandardWeight:164.93033{2}),
(Symbol:'Ho'; AtomicNumber:67; MassNumber:156; Mass:155.929706{64} ; Abundance:0;  StandardWeight:164.93033{2}),
(Symbol:'Ho'; AtomicNumber:67; MassNumber:157; Mass:156.928254{25} ; Abundance:0;  StandardWeight:164.93033{2}),
(Symbol:'Ho'; AtomicNumber:67; MassNumber:158; Mass:157.928946{29} ; Abundance:0;  StandardWeight:164.93033{2}),
(Symbol:'Ho'; AtomicNumber:67; MassNumber:159; Mass:158.9277197{36} ; Abundance:0;  StandardWeight:164.93033{2}),
(Symbol:'Ho'; AtomicNumber:67; MassNumber:160; Mass:159.928737{16} ; Abundance:0;  StandardWeight:164.93033{2}),
(Symbol:'Ho'; AtomicNumber:67; MassNumber:161; Mass:160.9278615{30} ; Abundance:0;  StandardWeight:164.93033{2}),
(Symbol:'Ho'; AtomicNumber:67; MassNumber:162; Mass:161.9291023{39} ; Abundance:0;  StandardWeight:164.93033{2}),
(Symbol:'Ho'; AtomicNumber:67; MassNumber:163; Mass:162.9287410{20} ; Abundance:0;  StandardWeight:164.93033{2}),
(Symbol:'Ho'; AtomicNumber:67; MassNumber:164; Mass:163.9302403{25} ; Abundance:0;  StandardWeight:164.93033{2}),
(Symbol:'Ho'; AtomicNumber:67; MassNumber:165; Mass:164.9303288{21} ; Abundance:100;  StandardWeight:164.93033{2}),
(Symbol:'Ho'; AtomicNumber:67; MassNumber:166; Mass:165.9322909{21} ; Abundance:0;  StandardWeight:164.93033{2}),
(Symbol:'Ho'; AtomicNumber:67; MassNumber:167; Mass:166.9331385{59} ; Abundance:0;  StandardWeight:164.93033{2}),
(Symbol:'Ho'; AtomicNumber:67; MassNumber:168; Mass:167.935522{32} ; Abundance:0;  StandardWeight:164.93033{2}),
(Symbol:'Ho'; AtomicNumber:67; MassNumber:169; Mass:168.936878{22} ; Abundance:0;  StandardWeight:164.93033{2}),
(Symbol:'Ho'; AtomicNumber:67; MassNumber:170; Mass:169.939625{54} ; Abundance:0;  StandardWeight:164.93033{2}),
(Symbol:'Ho'; AtomicNumber:67; MassNumber:171; Mass:170.94147{64} ; Abundance:0;  StandardWeight:164.93033{2}),
(Symbol:'Ho'; AtomicNumber:67; MassNumber:172; Mass:171.94473{21#} ; Abundance:0;  StandardWeight:164.93033{2}),
(Symbol:'Ho'; AtomicNumber:67; MassNumber:173; Mass:172.94702{32#} ; Abundance:0;  StandardWeight:164.93033{2}),
(Symbol:'Ho'; AtomicNumber:67; MassNumber:174; Mass:173.95095{32#} ; Abundance:0;  StandardWeight:164.93033{2}),
(Symbol:'Ho'; AtomicNumber:67; MassNumber:175; Mass:174.95362{43#} ; Abundance:0;  StandardWeight:164.93033{2}),
(Symbol:'Er'; AtomicNumber:68; MassNumber:142; Mass:141.97010{54#} ; Abundance:0; {g} StandardWeight:167.259{3}),
(Symbol:'Er'; AtomicNumber:68; MassNumber:143; Mass:142.96662{43#} ; Abundance:0; {g} StandardWeight:167.259{3}),
(Symbol:'Er'; AtomicNumber:68; MassNumber:144; Mass:143.96070{21#} ; Abundance:0; {g} StandardWeight:167.259{3}),
(Symbol:'Er'; AtomicNumber:68; MassNumber:145; Mass:144.95805{21#} ; Abundance:0; {g} StandardWeight:167.259{3}),
(Symbol:'Er'; AtomicNumber:68; MassNumber:146; Mass:145.9524184{72} ; Abundance:0; {g} StandardWeight:167.259{3}),
(Symbol:'Er'; AtomicNumber:68; MassNumber:147; Mass:146.949964{41} ; Abundance:0; {g} StandardWeight:167.259{3}),
(Symbol:'Er'; AtomicNumber:68; MassNumber:148; Mass:147.944735{11} ; Abundance:0; {g} StandardWeight:167.259{3}),
(Symbol:'Er'; AtomicNumber:68; MassNumber:149; Mass:148.942306{30} ; Abundance:0; {g} StandardWeight:167.259{3}),
(Symbol:'Er'; AtomicNumber:68; MassNumber:150; Mass:149.937916{18} ; Abundance:0; {g} StandardWeight:167.259{3}),
(Symbol:'Er'; AtomicNumber:68; MassNumber:151; Mass:150.937449{18} ; Abundance:0; {g} StandardWeight:167.259{3}),
(Symbol:'Er'; AtomicNumber:68; MassNumber:152; Mass:151.935057{10} ; Abundance:0; {g} StandardWeight:167.259{3}),
(Symbol:'Er'; AtomicNumber:68; MassNumber:153; Mass:152.935080{10} ; Abundance:0; {g} StandardWeight:167.259{3}),
(Symbol:'Er'; AtomicNumber:68; MassNumber:154; Mass:153.9327908{55} ; Abundance:0; {g} StandardWeight:167.259{3}),
(Symbol:'Er'; AtomicNumber:68; MassNumber:155; Mass:154.9332159{67} ; Abundance:0; {g} StandardWeight:167.259{3}),
(Symbol:'Er'; AtomicNumber:68; MassNumber:156; Mass:155.931067{26} ; Abundance:0; {g} StandardWeight:167.259{3}),
(Symbol:'Er'; AtomicNumber:68; MassNumber:157; Mass:156.931949{27} ; Abundance:0; {g} StandardWeight:167.259{3}),
(Symbol:'Er'; AtomicNumber:68; MassNumber:158; Mass:157.929893{27} ; Abundance:0; {g} StandardWeight:167.259{3}),
(Symbol:'Er'; AtomicNumber:68; MassNumber:159; Mass:158.9306918{42} ; Abundance:0; {g} StandardWeight:167.259{3}),
(Symbol:'Er'; AtomicNumber:68; MassNumber:160; Mass:159.929077{26} ; Abundance:0; {g} StandardWeight:167.259{3}),
(Symbol:'Er'; AtomicNumber:68; MassNumber:161; Mass:160.9300046{96} ; Abundance:0; {g} StandardWeight:167.259{3}),
(Symbol:'Er'; AtomicNumber:68; MassNumber:162; Mass:161.9287884{20} ; Abundance:0.139; {g} StandardWeight:167.259{3}),
(Symbol:'Er'; AtomicNumber:68; MassNumber:163; Mass:162.9300408{53} ; Abundance:0; {g} StandardWeight:167.259{3}),
(Symbol:'Er'; AtomicNumber:68; MassNumber:164; Mass:163.9292088{20} ; Abundance:1.601; {g} StandardWeight:167.259{3}),
(Symbol:'Er'; AtomicNumber:68; MassNumber:165; Mass:164.9307345{21} ; Abundance:0; {g} StandardWeight:167.259{3}),
(Symbol:'Er'; AtomicNumber:68; MassNumber:166; Mass:165.9302995{22} ; Abundance:33.503; {g} StandardWeight:167.259{3}),
(Symbol:'Er'; AtomicNumber:68; MassNumber:167; Mass:166.9320546{22} ; Abundance:22.869; {g} StandardWeight:167.259{3}),
(Symbol:'Er'; AtomicNumber:68; MassNumber:168; Mass:167.9323767{22} ; Abundance:26.978; {g} StandardWeight:167.259{3}),
(Symbol:'Er'; AtomicNumber:68; MassNumber:169; Mass:168.9345968{22} ; Abundance:0; {g} StandardWeight:167.259{3}),
(Symbol:'Er'; AtomicNumber:68; MassNumber:170; Mass:169.9354702{26} ; Abundance:14.91; {g} StandardWeight:167.259{3}),
(Symbol:'Er'; AtomicNumber:68; MassNumber:171; Mass:170.9380357{26} ; Abundance:0; {g} StandardWeight:167.259{3}),
(Symbol:'Er'; AtomicNumber:68; MassNumber:172; Mass:171.9393619{47} ; Abundance:0; {g} StandardWeight:167.259{3}),
(Symbol:'Er'; AtomicNumber:68; MassNumber:173; Mass:172.94240{21#} ; Abundance:0; {g} StandardWeight:167.259{3}),
(Symbol:'Er'; AtomicNumber:68; MassNumber:174; Mass:173.94423{32#} ; Abundance:0; {g} StandardWeight:167.259{3}),
(Symbol:'Er'; AtomicNumber:68; MassNumber:175; Mass:174.94777{43#} ; Abundance:0; {g} StandardWeight:167.259{3}),
(Symbol:'Er'; AtomicNumber:68; MassNumber:176; Mass:175.94994{43#} ; Abundance:0; {g} StandardWeight:167.259{3}),
(Symbol:'Er'; AtomicNumber:68; MassNumber:177; Mass:176.95399{54#} ; Abundance:0; {g} StandardWeight:167.259{3}),
(Symbol:'Tm'; AtomicNumber:69; MassNumber:144; Mass:143.97628{43#} ; Abundance:0;  StandardWeight:168.93422{2}),
(Symbol:'Tm'; AtomicNumber:69; MassNumber:145; Mass:144.97039{21#} ; Abundance:0;  StandardWeight:168.93422{2}),
(Symbol:'Tm'; AtomicNumber:69; MassNumber:146; Mass:145.96684{21#} ; Abundance:0;  StandardWeight:168.93422{2}),
(Symbol:'Tm'; AtomicNumber:69; MassNumber:147; Mass:146.9613799{73} ; Abundance:0;  StandardWeight:168.93422{2}),
(Symbol:'Tm'; AtomicNumber:69; MassNumber:148; Mass:147.958384{11} ; Abundance:0;  StandardWeight:168.93422{2}),
(Symbol:'Tm'; AtomicNumber:69; MassNumber:149; Mass:148.95289{32#} ; Abundance:0;  StandardWeight:168.93422{2}),
(Symbol:'Tm'; AtomicNumber:69; MassNumber:150; Mass:149.95009{21#} ; Abundance:0;  StandardWeight:168.93422{2}),
(Symbol:'Tm'; AtomicNumber:69; MassNumber:151; Mass:150.945488{21} ; Abundance:0;  StandardWeight:168.93422{2}),
(Symbol:'Tm'; AtomicNumber:69; MassNumber:152; Mass:151.944422{79} ; Abundance:0;  StandardWeight:168.93422{2}),
(Symbol:'Tm'; AtomicNumber:69; MassNumber:153; Mass:152.942040{16} ; Abundance:0;  StandardWeight:168.93422{2}),
(Symbol:'Tm'; AtomicNumber:69; MassNumber:154; Mass:153.941570{15} ; Abundance:0;  StandardWeight:168.93422{2}),
(Symbol:'Tm'; AtomicNumber:69; MassNumber:155; Mass:154.939210{11} ; Abundance:0;  StandardWeight:168.93422{2}),
(Symbol:'Tm'; AtomicNumber:69; MassNumber:156; Mass:155.938992{16} ; Abundance:0;  StandardWeight:168.93422{2}),
(Symbol:'Tm'; AtomicNumber:69; MassNumber:157; Mass:156.936944{28} ; Abundance:0;  StandardWeight:168.93422{2}),
(Symbol:'Tm'; AtomicNumber:69; MassNumber:158; Mass:157.936980{27} ; Abundance:0;  StandardWeight:168.93422{2}),
(Symbol:'Tm'; AtomicNumber:69; MassNumber:159; Mass:158.934975{30} ; Abundance:0;  StandardWeight:168.93422{2}),
(Symbol:'Tm'; AtomicNumber:69; MassNumber:160; Mass:159.935263{37} ; Abundance:0;  StandardWeight:168.93422{2}),
(Symbol:'Tm'; AtomicNumber:69; MassNumber:161; Mass:160.933549{30} ; Abundance:0;  StandardWeight:168.93422{2}),
(Symbol:'Tm'; AtomicNumber:69; MassNumber:162; Mass:161.934002{28} ; Abundance:0;  StandardWeight:168.93422{2}),
(Symbol:'Tm'; AtomicNumber:69; MassNumber:163; Mass:162.9326592{62} ; Abundance:0;  StandardWeight:168.93422{2}),
(Symbol:'Tm'; AtomicNumber:69; MassNumber:164; Mass:163.933544{26} ; Abundance:0;  StandardWeight:168.93422{2}),
(Symbol:'Tm'; AtomicNumber:69; MassNumber:165; Mass:164.9324431{26} ; Abundance:0;  StandardWeight:168.93422{2}),
(Symbol:'Tm'; AtomicNumber:69; MassNumber:166; Mass:165.933561{13} ; Abundance:0;  StandardWeight:168.93422{2}),
(Symbol:'Tm'; AtomicNumber:69; MassNumber:167; Mass:166.9328562{25} ; Abundance:0;  StandardWeight:168.93422{2}),
(Symbol:'Tm'; AtomicNumber:69; MassNumber:168; Mass:167.9341774{27} ; Abundance:0;  StandardWeight:168.93422{2}),
(Symbol:'Tm'; AtomicNumber:69; MassNumber:169; Mass:168.9342179{22} ; Abundance:100;  StandardWeight:168.93422{2}),
(Symbol:'Tm'; AtomicNumber:69; MassNumber:170; Mass:169.9358060{22} ; Abundance:0;  StandardWeight:168.93422{2}),
(Symbol:'Tm'; AtomicNumber:69; MassNumber:171; Mass:170.9364339{24} ; Abundance:0;  StandardWeight:168.93422{2}),
(Symbol:'Tm'; AtomicNumber:69; MassNumber:172; Mass:171.9384055{62} ; Abundance:0;  StandardWeight:168.93422{2}),
(Symbol:'Tm'; AtomicNumber:69; MassNumber:173; Mass:172.9396084{53} ; Abundance:0;  StandardWeight:168.93422{2}),
(Symbol:'Tm'; AtomicNumber:69; MassNumber:174; Mass:173.942173{48} ; Abundance:0;  StandardWeight:168.93422{2}),
(Symbol:'Tm'; AtomicNumber:69; MassNumber:175; Mass:174.943841{54} ; Abundance:0;  StandardWeight:168.93422{2}),
(Symbol:'Tm'; AtomicNumber:69; MassNumber:176; Mass:175.94700{11} ; Abundance:0;  StandardWeight:168.93422{2}),
(Symbol:'Tm'; AtomicNumber:69; MassNumber:177; Mass:176.94904{32#} ; Abundance:0;  StandardWeight:168.93422{2}),
(Symbol:'Tm'; AtomicNumber:69; MassNumber:178; Mass:177.95264{43#} ; Abundance:0;  StandardWeight:168.93422{2}),
(Symbol:'Tm'; AtomicNumber:69; MassNumber:179; Mass:178.95534{54#} ; Abundance:0;  StandardWeight:168.93422{2}),
(Symbol:'Yb'; AtomicNumber:70; MassNumber:148; Mass:147.96758{64#} ; Abundance:0; {g} StandardWeight:173.054{5}),
(Symbol:'Yb'; AtomicNumber:70; MassNumber:149; Mass:148.96436{54#} ; Abundance:0; {g} StandardWeight:173.054{5}),
(Symbol:'Yb'; AtomicNumber:70; MassNumber:150; Mass:149.95852{43#} ; Abundance:0; {g} StandardWeight:173.054{5}),
(Symbol:'Yb'; AtomicNumber:70; MassNumber:151; Mass:150.95540{32} ; Abundance:0; {g} StandardWeight:173.054{5}),
(Symbol:'Yb'; AtomicNumber:70; MassNumber:152; Mass:151.95027{17} ; Abundance:0; {g} StandardWeight:173.054{5}),
(Symbol:'Yb'; AtomicNumber:70; MassNumber:153; Mass:152.94932{21#} ; Abundance:0; {g} StandardWeight:173.054{5}),
(Symbol:'Yb'; AtomicNumber:70; MassNumber:154; Mass:153.946396{19} ; Abundance:0; {g} StandardWeight:173.054{5}),
(Symbol:'Yb'; AtomicNumber:70; MassNumber:155; Mass:154.945783{18} ; Abundance:0; {g} StandardWeight:173.054{5}),
(Symbol:'Yb'; AtomicNumber:70; MassNumber:156; Mass:155.942825{11} ; Abundance:0; {g} StandardWeight:173.054{5}),
(Symbol:'Yb'; AtomicNumber:70; MassNumber:157; Mass:156.942645{12} ; Abundance:0; {g} StandardWeight:173.054{5}),
(Symbol:'Yb'; AtomicNumber:70; MassNumber:158; Mass:157.9398705{86} ; Abundance:0; {g} StandardWeight:173.054{5}),
(Symbol:'Yb'; AtomicNumber:70; MassNumber:159; Mass:158.940055{19} ; Abundance:0; {g} StandardWeight:173.054{5}),
(Symbol:'Yb'; AtomicNumber:70; MassNumber:160; Mass:159.937557{17} ; Abundance:0; {g} StandardWeight:173.054{5}),
(Symbol:'Yb'; AtomicNumber:70; MassNumber:161; Mass:160.937907{17} ; Abundance:0; {g} StandardWeight:173.054{5}),
(Symbol:'Yb'; AtomicNumber:70; MassNumber:162; Mass:161.935774{17} ; Abundance:0; {g} StandardWeight:173.054{5}),
(Symbol:'Yb'; AtomicNumber:70; MassNumber:163; Mass:162.936340{17} ; Abundance:0; {g} StandardWeight:173.054{5}),
(Symbol:'Yb'; AtomicNumber:70; MassNumber:164; Mass:163.934495{17} ; Abundance:0; {g} StandardWeight:173.054{5}),
(Symbol:'Yb'; AtomicNumber:70; MassNumber:165; Mass:164.935270{28} ; Abundance:0; {g} StandardWeight:173.054{5}),
(Symbol:'Yb'; AtomicNumber:70; MassNumber:166; Mass:165.9338747{78} ; Abundance:0; {g} StandardWeight:173.054{5}),
(Symbol:'Yb'; AtomicNumber:70; MassNumber:167; Mass:166.9349530{47} ; Abundance:0; {g} StandardWeight:173.054{5}),
(Symbol:'Yb'; AtomicNumber:70; MassNumber:168; Mass:167.9338896{22} ; Abundance:0.123; {g} StandardWeight:173.054{5}),
(Symbol:'Yb'; AtomicNumber:70; MassNumber:169; Mass:168.9351825{22} ; Abundance:0; {g} StandardWeight:173.054{5}),
(Symbol:'Yb'; AtomicNumber:70; MassNumber:170; Mass:169.9347664{22} ; Abundance:2.982; {g} StandardWeight:173.054{5}),
(Symbol:'Yb'; AtomicNumber:70; MassNumber:171; Mass:170.9363302{22} ; Abundance:14.09; {g} StandardWeight:173.054{5}),
(Symbol:'Yb'; AtomicNumber:70; MassNumber:172; Mass:171.9363859{22} ; Abundance:21.68; {g} StandardWeight:173.054{5}),
(Symbol:'Yb'; AtomicNumber:70; MassNumber:173; Mass:172.9382151{22} ; Abundance:16.103; {g} StandardWeight:173.054{5}),
(Symbol:'Yb'; AtomicNumber:70; MassNumber:174; Mass:173.9388664{22} ; Abundance:32.026; {g} StandardWeight:173.054{5}),
(Symbol:'Yb'; AtomicNumber:70; MassNumber:175; Mass:174.9412808{22} ; Abundance:0; {g} StandardWeight:173.054{5}),
(Symbol:'Yb'; AtomicNumber:70; MassNumber:176; Mass:175.9425764{24} ; Abundance:12.996; {g} StandardWeight:173.054{5}),
(Symbol:'Yb'; AtomicNumber:70; MassNumber:177; Mass:176.9452656{24} ; Abundance:0; {g} StandardWeight:173.054{5}),
(Symbol:'Yb'; AtomicNumber:70; MassNumber:178; Mass:177.946651{11} ; Abundance:0; {g} StandardWeight:173.054{5}),
(Symbol:'Yb'; AtomicNumber:70; MassNumber:179; Mass:178.95004{21#} ; Abundance:0; {g} StandardWeight:173.054{5}),
(Symbol:'Yb'; AtomicNumber:70; MassNumber:180; Mass:179.95212{32#} ; Abundance:0; {g} StandardWeight:173.054{5}),
(Symbol:'Yb'; AtomicNumber:70; MassNumber:181; Mass:180.95589{32#} ; Abundance:0; {g} StandardWeight:173.054{5}),
(Symbol:'Lu'; AtomicNumber:71; MassNumber:150; Mass:149.97355{54#} ; Abundance:0; {g} StandardWeight:174.9668{1}),
(Symbol:'Lu'; AtomicNumber:71; MassNumber:151; Mass:150.96768{43#} ; Abundance:0; {g} StandardWeight:174.9668{1}),
(Symbol:'Lu'; AtomicNumber:71; MassNumber:152; Mass:151.96412{21#} ; Abundance:0; {g} StandardWeight:174.9668{1}),
(Symbol:'Lu'; AtomicNumber:71; MassNumber:153; Mass:152.95875{17} ; Abundance:0; {g} StandardWeight:174.9668{1}),
(Symbol:'Lu'; AtomicNumber:71; MassNumber:154; Mass:153.95736{22#} ; Abundance:0; {g} StandardWeight:174.9668{1}),
(Symbol:'Lu'; AtomicNumber:71; MassNumber:155; Mass:154.954321{21} ; Abundance:0; {g} StandardWeight:174.9668{1}),
(Symbol:'Lu'; AtomicNumber:71; MassNumber:156; Mass:155.953033{79} ; Abundance:0; {g} StandardWeight:174.9668{1}),
(Symbol:'Lu'; AtomicNumber:71; MassNumber:157; Mass:156.950127{16} ; Abundance:0; {g} StandardWeight:174.9668{1}),
(Symbol:'Lu'; AtomicNumber:71; MassNumber:158; Mass:157.949316{16} ; Abundance:0; {g} StandardWeight:174.9668{1}),
(Symbol:'Lu'; AtomicNumber:71; MassNumber:159; Mass:158.946636{40} ; Abundance:0; {g} StandardWeight:174.9668{1}),
(Symbol:'Lu'; AtomicNumber:71; MassNumber:160; Mass:159.946033{61} ; Abundance:0; {g} StandardWeight:174.9668{1}),
(Symbol:'Lu'; AtomicNumber:71; MassNumber:161; Mass:160.943572{30} ; Abundance:0; {g} StandardWeight:174.9668{1}),
(Symbol:'Lu'; AtomicNumber:71; MassNumber:162; Mass:161.943283{81} ; Abundance:0; {g} StandardWeight:174.9668{1}),
(Symbol:'Lu'; AtomicNumber:71; MassNumber:163; Mass:162.941179{30} ; Abundance:0; {g} StandardWeight:174.9668{1}),
(Symbol:'Lu'; AtomicNumber:71; MassNumber:164; Mass:163.941339{30} ; Abundance:0; {g} StandardWeight:174.9668{1}),
(Symbol:'Lu'; AtomicNumber:71; MassNumber:165; Mass:164.939407{28} ; Abundance:0; {g} StandardWeight:174.9668{1}),
(Symbol:'Lu'; AtomicNumber:71; MassNumber:166; Mass:165.939859{32} ; Abundance:0; {g} StandardWeight:174.9668{1}),
(Symbol:'Lu'; AtomicNumber:71; MassNumber:167; Mass:166.938270{34} ; Abundance:0; {g} StandardWeight:174.9668{1}),
(Symbol:'Lu'; AtomicNumber:71; MassNumber:168; Mass:167.938736{42} ; Abundance:0; {g} StandardWeight:174.9668{1}),
(Symbol:'Lu'; AtomicNumber:71; MassNumber:169; Mass:168.9376441{39} ; Abundance:0; {g} StandardWeight:174.9668{1}),
(Symbol:'Lu'; AtomicNumber:71; MassNumber:170; Mass:169.938478{18} ; Abundance:0; {g} StandardWeight:174.9668{1}),
(Symbol:'Lu'; AtomicNumber:71; MassNumber:171; Mass:170.9379170{27} ; Abundance:0; {g} StandardWeight:174.9668{1}),
(Symbol:'Lu'; AtomicNumber:71; MassNumber:172; Mass:171.9390891{30} ; Abundance:0; {g} StandardWeight:174.9668{1}),
(Symbol:'Lu'; AtomicNumber:71; MassNumber:173; Mass:172.9389340{23} ; Abundance:0; {g} StandardWeight:174.9668{1}),
(Symbol:'Lu'; AtomicNumber:71; MassNumber:174; Mass:173.9403409{23} ; Abundance:0; {g} StandardWeight:174.9668{1}),
(Symbol:'Lu'; AtomicNumber:71; MassNumber:175; Mass:174.9407752{20} ; Abundance:97.401; {g} StandardWeight:174.9668{1}),
(Symbol:'Lu'; AtomicNumber:71; MassNumber:176; Mass:175.9426897{20} ; Abundance:2.599; {g} StandardWeight:174.9668{1}),
(Symbol:'Lu'; AtomicNumber:71; MassNumber:177; Mass:176.9437615{20} ; Abundance:0; {g} StandardWeight:174.9668{1}),
(Symbol:'Lu'; AtomicNumber:71; MassNumber:178; Mass:177.9459580{29} ; Abundance:0; {g} StandardWeight:174.9668{1}),
(Symbol:'Lu'; AtomicNumber:71; MassNumber:179; Mass:178.9473309{57} ; Abundance:0; {g} StandardWeight:174.9668{1}),
(Symbol:'Lu'; AtomicNumber:71; MassNumber:180; Mass:179.949888{76} ; Abundance:0; {g} StandardWeight:174.9668{1}),
(Symbol:'Lu'; AtomicNumber:71; MassNumber:181; Mass:180.95191{17} ; Abundance:0; {g} StandardWeight:174.9668{1}),
(Symbol:'Lu'; AtomicNumber:71; MassNumber:182; Mass:181.95504{21#} ; Abundance:0; {g} StandardWeight:174.9668{1}),
(Symbol:'Lu'; AtomicNumber:71; MassNumber:183; Mass:182.957363{98} ; Abundance:0; {g} StandardWeight:174.9668{1}),
(Symbol:'Lu'; AtomicNumber:71; MassNumber:184; Mass:183.96091{32#} ; Abundance:0; {g} StandardWeight:174.9668{1}),
(Symbol:'Lu'; AtomicNumber:71; MassNumber:185; Mass:184.96362{32#} ; Abundance:0; {g} StandardWeight:174.9668{1}),
(Symbol:'Hf'; AtomicNumber:72; MassNumber:153; Mass:152.97069{54#} ; Abundance:0;  StandardWeight:178.49{2}),
(Symbol:'Hf'; AtomicNumber:72; MassNumber:154; Mass:153.96486{54#} ; Abundance:0;  StandardWeight:178.49{2}),
(Symbol:'Hf'; AtomicNumber:72; MassNumber:155; Mass:154.96311{32#} ; Abundance:0;  StandardWeight:178.49{2}),
(Symbol:'Hf'; AtomicNumber:72; MassNumber:156; Mass:155.95935{17} ; Abundance:0;  StandardWeight:178.49{2}),
(Symbol:'Hf'; AtomicNumber:72; MassNumber:157; Mass:156.95824{21#} ; Abundance:0;  StandardWeight:178.49{2}),
(Symbol:'Hf'; AtomicNumber:72; MassNumber:158; Mass:157.954801{19} ; Abundance:0;  StandardWeight:178.49{2}),
(Symbol:'Hf'; AtomicNumber:72; MassNumber:159; Mass:158.953996{18} ; Abundance:0;  StandardWeight:178.49{2}),
(Symbol:'Hf'; AtomicNumber:72; MassNumber:160; Mass:159.950691{11} ; Abundance:0;  StandardWeight:178.49{2}),
(Symbol:'Hf'; AtomicNumber:72; MassNumber:161; Mass:160.950278{24} ; Abundance:0;  StandardWeight:178.49{2}),
(Symbol:'Hf'; AtomicNumber:72; MassNumber:162; Mass:161.9472148{97} ; Abundance:0;  StandardWeight:178.49{2}),
(Symbol:'Hf'; AtomicNumber:72; MassNumber:163; Mass:162.947113{27} ; Abundance:0;  StandardWeight:178.49{2}),
(Symbol:'Hf'; AtomicNumber:72; MassNumber:164; Mass:163.944371{17} ; Abundance:0;  StandardWeight:178.49{2}),
(Symbol:'Hf'; AtomicNumber:72; MassNumber:165; Mass:164.944567{30} ; Abundance:0;  StandardWeight:178.49{2}),
(Symbol:'Hf'; AtomicNumber:72; MassNumber:166; Mass:165.942180{30} ; Abundance:0;  StandardWeight:178.49{2}),
(Symbol:'Hf'; AtomicNumber:72; MassNumber:167; Mass:166.942600{30} ; Abundance:0;  StandardWeight:178.49{2}),
(Symbol:'Hf'; AtomicNumber:72; MassNumber:168; Mass:167.940568{30} ; Abundance:0;  StandardWeight:178.49{2}),
(Symbol:'Hf'; AtomicNumber:72; MassNumber:169; Mass:168.941259{30} ; Abundance:0;  StandardWeight:178.49{2}),
(Symbol:'Hf'; AtomicNumber:72; MassNumber:170; Mass:169.939609{30} ; Abundance:0;  StandardWeight:178.49{2}),
(Symbol:'Hf'; AtomicNumber:72; MassNumber:171; Mass:170.940492{31} ; Abundance:0;  StandardWeight:178.49{2}),
(Symbol:'Hf'; AtomicNumber:72; MassNumber:172; Mass:171.939450{26} ; Abundance:0;  StandardWeight:178.49{2}),
(Symbol:'Hf'; AtomicNumber:72; MassNumber:173; Mass:172.940513{30} ; Abundance:0;  StandardWeight:178.49{2}),
(Symbol:'Hf'; AtomicNumber:72; MassNumber:174; Mass:173.9400461{28} ; Abundance:0.16;  StandardWeight:178.49{2}),
(Symbol:'Hf'; AtomicNumber:72; MassNumber:175; Mass:174.9415092{29} ; Abundance:0;  StandardWeight:178.49{2}),
(Symbol:'Hf'; AtomicNumber:72; MassNumber:176; Mass:175.9414076{22} ; Abundance:5.26;  StandardWeight:178.49{2}),
(Symbol:'Hf'; AtomicNumber:72; MassNumber:177; Mass:176.9432277{20} ; Abundance:18.6;  StandardWeight:178.49{2}),
(Symbol:'Hf'; AtomicNumber:72; MassNumber:178; Mass:177.9437058{20} ; Abundance:27.28;  StandardWeight:178.49{2}),
(Symbol:'Hf'; AtomicNumber:72; MassNumber:179; Mass:178.9458232{20} ; Abundance:13.62;  StandardWeight:178.49{2}),
(Symbol:'Hf'; AtomicNumber:72; MassNumber:180; Mass:179.9465570{20} ; Abundance:35.08;  StandardWeight:178.49{2}),
(Symbol:'Hf'; AtomicNumber:72; MassNumber:181; Mass:180.9491083{20} ; Abundance:0;  StandardWeight:178.49{2}),
(Symbol:'Hf'; AtomicNumber:72; MassNumber:182; Mass:181.9505612{68} ; Abundance:0;  StandardWeight:178.49{2}),
(Symbol:'Hf'; AtomicNumber:72; MassNumber:183; Mass:182.953530{32} ; Abundance:0;  StandardWeight:178.49{2}),
(Symbol:'Hf'; AtomicNumber:72; MassNumber:184; Mass:183.955446{43} ; Abundance:0;  StandardWeight:178.49{2}),
(Symbol:'Hf'; AtomicNumber:72; MassNumber:185; Mass:184.958862{98} ; Abundance:0;  StandardWeight:178.49{2}),
(Symbol:'Hf'; AtomicNumber:72; MassNumber:186; Mass:185.960897{59} ; Abundance:0;  StandardWeight:178.49{2}),
(Symbol:'Hf'; AtomicNumber:72; MassNumber:187; Mass:186.96477{32#} ; Abundance:0;  StandardWeight:178.49{2}),
(Symbol:'Hf'; AtomicNumber:72; MassNumber:188; Mass:187.96685{32#} ; Abundance:0;  StandardWeight:178.49{2}),
(Symbol:'Hf'; AtomicNumber:72; MassNumber:189; Mass:188.97084{32#} ; Abundance:0;  StandardWeight:178.49{2}),
(Symbol:'Ta'; AtomicNumber:73; MassNumber:155; Mass:154.97424{54#} ; Abundance:0;  StandardWeight:180.94788{2}),
(Symbol:'Ta'; AtomicNumber:73; MassNumber:156; Mass:155.97203{32#} ; Abundance:0;  StandardWeight:180.94788{2}),
(Symbol:'Ta'; AtomicNumber:73; MassNumber:157; Mass:156.96818{17} ; Abundance:0;  StandardWeight:180.94788{2}),
(Symbol:'Ta'; AtomicNumber:73; MassNumber:158; Mass:157.96654{22#} ; Abundance:0;  StandardWeight:180.94788{2}),
(Symbol:'Ta'; AtomicNumber:73; MassNumber:159; Mass:158.963023{21} ; Abundance:0;  StandardWeight:180.94788{2}),
(Symbol:'Ta'; AtomicNumber:73; MassNumber:160; Mass:159.961488{79} ; Abundance:0;  StandardWeight:180.94788{2}),
(Symbol:'Ta'; AtomicNumber:73; MassNumber:161; Mass:160.958452{27} ; Abundance:0;  StandardWeight:180.94788{2}),
(Symbol:'Ta'; AtomicNumber:73; MassNumber:162; Mass:161.957294{56} ; Abundance:0;  StandardWeight:180.94788{2}),
(Symbol:'Ta'; AtomicNumber:73; MassNumber:163; Mass:162.954337{41} ; Abundance:0;  StandardWeight:180.94788{2}),
(Symbol:'Ta'; AtomicNumber:73; MassNumber:164; Mass:163.953534{30} ; Abundance:0;  StandardWeight:180.94788{2}),
(Symbol:'Ta'; AtomicNumber:73; MassNumber:165; Mass:164.950781{15} ; Abundance:0;  StandardWeight:180.94788{2}),
(Symbol:'Ta'; AtomicNumber:73; MassNumber:166; Mass:165.950512{30} ; Abundance:0;  StandardWeight:180.94788{2}),
(Symbol:'Ta'; AtomicNumber:73; MassNumber:167; Mass:166.948093{30} ; Abundance:0;  StandardWeight:180.94788{2}),
(Symbol:'Ta'; AtomicNumber:73; MassNumber:168; Mass:167.948047{30} ; Abundance:0;  StandardWeight:180.94788{2}),
(Symbol:'Ta'; AtomicNumber:73; MassNumber:169; Mass:168.946011{30} ; Abundance:0;  StandardWeight:180.94788{2}),
(Symbol:'Ta'; AtomicNumber:73; MassNumber:170; Mass:169.946175{30} ; Abundance:0;  StandardWeight:180.94788{2}),
(Symbol:'Ta'; AtomicNumber:73; MassNumber:171; Mass:170.944476{30} ; Abundance:0;  StandardWeight:180.94788{2}),
(Symbol:'Ta'; AtomicNumber:73; MassNumber:172; Mass:171.944895{30} ; Abundance:0;  StandardWeight:180.94788{2}),
(Symbol:'Ta'; AtomicNumber:73; MassNumber:173; Mass:172.943750{30} ; Abundance:0;  StandardWeight:180.94788{2}),
(Symbol:'Ta'; AtomicNumber:73; MassNumber:174; Mass:173.944454{30} ; Abundance:0;  StandardWeight:180.94788{2}),
(Symbol:'Ta'; AtomicNumber:73; MassNumber:175; Mass:174.943737{30} ; Abundance:0;  StandardWeight:180.94788{2}),
(Symbol:'Ta'; AtomicNumber:73; MassNumber:176; Mass:175.944857{33} ; Abundance:0;  StandardWeight:180.94788{2}),
(Symbol:'Ta'; AtomicNumber:73; MassNumber:177; Mass:176.9444795{38} ; Abundance:0;  StandardWeight:180.94788{2}),
(Symbol:'Ta'; AtomicNumber:73; MassNumber:178; Mass:177.945678{56#} ; Abundance:0;  StandardWeight:180.94788{2}),
(Symbol:'Ta'; AtomicNumber:73; MassNumber:179; Mass:178.9459366{21} ; Abundance:0;  StandardWeight:180.94788{2}),
(Symbol:'Ta'; AtomicNumber:73; MassNumber:180; Mass:179.9474648{24} ; Abundance:0.01201;  StandardWeight:180.94788{2}),
(Symbol:'Ta'; AtomicNumber:73; MassNumber:181; Mass:180.9479958{20} ; Abundance:99.98799;  StandardWeight:180.94788{2}),
(Symbol:'Ta'; AtomicNumber:73; MassNumber:182; Mass:181.9501519{20} ; Abundance:0;  StandardWeight:180.94788{2}),
(Symbol:'Ta'; AtomicNumber:73; MassNumber:183; Mass:182.9513726{20} ; Abundance:0;  StandardWeight:180.94788{2}),
(Symbol:'Ta'; AtomicNumber:73; MassNumber:184; Mass:183.954008{28} ; Abundance:0;  StandardWeight:180.94788{2}),
(Symbol:'Ta'; AtomicNumber:73; MassNumber:185; Mass:184.955559{15} ; Abundance:0;  StandardWeight:180.94788{2}),
(Symbol:'Ta'; AtomicNumber:73; MassNumber:186; Mass:185.958551{64} ; Abundance:0;  StandardWeight:180.94788{2}),
(Symbol:'Ta'; AtomicNumber:73; MassNumber:187; Mass:186.960386{71} ; Abundance:0;  StandardWeight:180.94788{2}),
(Symbol:'Ta'; AtomicNumber:73; MassNumber:188; Mass:187.963916{71} ; Abundance:0;  StandardWeight:180.94788{2}),
(Symbol:'Ta'; AtomicNumber:73; MassNumber:189; Mass:188.96583{32#} ; Abundance:0;  StandardWeight:180.94788{2}),
(Symbol:'Ta'; AtomicNumber:73; MassNumber:190; Mass:189.96939{21#} ; Abundance:0;  StandardWeight:180.94788{2}),
(Symbol:'Ta'; AtomicNumber:73; MassNumber:191; Mass:190.97156{32#} ; Abundance:0;  StandardWeight:180.94788{2}),
(Symbol:'Ta'; AtomicNumber:73; MassNumber:192; Mass:191.97514{43#} ; Abundance:0;  StandardWeight:180.94788{2}),
(Symbol:'W'; AtomicNumber:74; MassNumber:157; Mass:156.97884{43#} ; Abundance:0;  StandardWeight:183.84{1}),
(Symbol:'W'; AtomicNumber:74; MassNumber:158; Mass:157.97456{54#} ; Abundance:0;  StandardWeight:183.84{1}),
(Symbol:'W'; AtomicNumber:74; MassNumber:159; Mass:158.97264{32#} ; Abundance:0;  StandardWeight:183.84{1}),
(Symbol:'W'; AtomicNumber:74; MassNumber:160; Mass:159.96846{17} ; Abundance:0;  StandardWeight:183.84{1}),
(Symbol:'W'; AtomicNumber:74; MassNumber:161; Mass:160.96720{21#} ; Abundance:0;  StandardWeight:183.84{1}),
(Symbol:'W'; AtomicNumber:74; MassNumber:162; Mass:161.963499{19} ; Abundance:0;  StandardWeight:183.84{1}),
(Symbol:'W'; AtomicNumber:74; MassNumber:163; Mass:162.962524{57} ; Abundance:0;  StandardWeight:183.84{1}),
(Symbol:'W'; AtomicNumber:74; MassNumber:164; Mass:163.958961{11} ; Abundance:0;  StandardWeight:183.84{1}),
(Symbol:'W'; AtomicNumber:74; MassNumber:165; Mass:164.958281{27} ; Abundance:0;  StandardWeight:183.84{1}),
(Symbol:'W'; AtomicNumber:74; MassNumber:166; Mass:165.955031{10} ; Abundance:0;  StandardWeight:183.84{1}),
(Symbol:'W'; AtomicNumber:74; MassNumber:167; Mass:166.954805{20} ; Abundance:0;  StandardWeight:183.84{1}),
(Symbol:'W'; AtomicNumber:74; MassNumber:168; Mass:167.951806{14} ; Abundance:0;  StandardWeight:183.84{1}),
(Symbol:'W'; AtomicNumber:74; MassNumber:169; Mass:168.951779{17} ; Abundance:0;  StandardWeight:183.84{1}),
(Symbol:'W'; AtomicNumber:74; MassNumber:170; Mass:169.949232{14} ; Abundance:0;  StandardWeight:183.84{1}),
(Symbol:'W'; AtomicNumber:74; MassNumber:171; Mass:170.949451{30} ; Abundance:0;  StandardWeight:183.84{1}),
(Symbol:'W'; AtomicNumber:74; MassNumber:172; Mass:171.947292{30} ; Abundance:0;  StandardWeight:183.84{1}),
(Symbol:'W'; AtomicNumber:74; MassNumber:173; Mass:172.947689{30} ; Abundance:0;  StandardWeight:183.84{1}),
(Symbol:'W'; AtomicNumber:74; MassNumber:174; Mass:173.946079{30} ; Abundance:0;  StandardWeight:183.84{1}),
(Symbol:'W'; AtomicNumber:74; MassNumber:175; Mass:174.946717{30} ; Abundance:0;  StandardWeight:183.84{1}),
(Symbol:'W'; AtomicNumber:74; MassNumber:176; Mass:175.945634{30} ; Abundance:0;  StandardWeight:183.84{1}),
(Symbol:'W'; AtomicNumber:74; MassNumber:177; Mass:176.946643{30} ; Abundance:0;  StandardWeight:183.84{1}),
(Symbol:'W'; AtomicNumber:74; MassNumber:178; Mass:177.945883{16} ; Abundance:0;  StandardWeight:183.84{1}),
(Symbol:'W'; AtomicNumber:74; MassNumber:179; Mass:178.947077{16} ; Abundance:0;  StandardWeight:183.84{1}),
(Symbol:'W'; AtomicNumber:74; MassNumber:180; Mass:179.9467108{20} ; Abundance:0.12;  StandardWeight:183.84{1}),
(Symbol:'W'; AtomicNumber:74; MassNumber:181; Mass:180.9481978{51} ; Abundance:0;  StandardWeight:183.84{1}),
(Symbol:'W'; AtomicNumber:74; MassNumber:182; Mass:181.94820394{91} ; Abundance:26.5;  StandardWeight:183.84{1}),
(Symbol:'W'; AtomicNumber:74; MassNumber:183; Mass:182.95022275{90} ; Abundance:14.31;  StandardWeight:183.84{1}),
(Symbol:'W'; AtomicNumber:74; MassNumber:184; Mass:183.95093092{94} ; Abundance:30.64;  StandardWeight:183.84{1}),
(Symbol:'W'; AtomicNumber:74; MassNumber:185; Mass:184.95341897{99} ; Abundance:0;  StandardWeight:183.84{1}),
(Symbol:'W'; AtomicNumber:74; MassNumber:186; Mass:185.9543628{17} ; Abundance:28.43;  StandardWeight:183.84{1}),
(Symbol:'W'; AtomicNumber:74; MassNumber:187; Mass:186.9571588{17} ; Abundance:0;  StandardWeight:183.84{1}),
(Symbol:'W'; AtomicNumber:74; MassNumber:188; Mass:187.9584862{36} ; Abundance:0;  StandardWeight:183.84{1}),
(Symbol:'W'; AtomicNumber:74; MassNumber:189; Mass:188.961763{44} ; Abundance:0;  StandardWeight:183.84{1}),
(Symbol:'W'; AtomicNumber:74; MassNumber:190; Mass:189.963091{42} ; Abundance:0;  StandardWeight:183.84{1}),
(Symbol:'W'; AtomicNumber:74; MassNumber:191; Mass:190.966531{48} ; Abundance:0;  StandardWeight:183.84{1}),
(Symbol:'W'; AtomicNumber:74; MassNumber:192; Mass:191.96817{21#} ; Abundance:0;  StandardWeight:183.84{1}),
(Symbol:'W'; AtomicNumber:74; MassNumber:193; Mass:192.97178{21#} ; Abundance:0;  StandardWeight:183.84{1}),
(Symbol:'W'; AtomicNumber:74; MassNumber:194; Mass:193.97367{32#} ; Abundance:0;  StandardWeight:183.84{1}),
(Symbol:'Re'; AtomicNumber:75; MassNumber:159; Mass:158.98418{54#} ; Abundance:0;  StandardWeight:186.207{1}),
(Symbol:'Re'; AtomicNumber:75; MassNumber:160; Mass:159.98182{32#} ; Abundance:0;  StandardWeight:186.207{1}),
(Symbol:'Re'; AtomicNumber:75; MassNumber:161; Mass:160.97757{17} ; Abundance:0;  StandardWeight:186.207{1}),
(Symbol:'Re'; AtomicNumber:75; MassNumber:162; Mass:161.97584{22#} ; Abundance:0;  StandardWeight:186.207{1}),
(Symbol:'Re'; AtomicNumber:75; MassNumber:163; Mass:162.972080{20} ; Abundance:0;  StandardWeight:186.207{1}),
(Symbol:'Re'; AtomicNumber:75; MassNumber:164; Mass:163.970453{79} ; Abundance:0;  StandardWeight:186.207{1}),
(Symbol:'Re'; AtomicNumber:75; MassNumber:165; Mass:164.967103{27} ; Abundance:0;  StandardWeight:186.207{1}),
(Symbol:'Re'; AtomicNumber:75; MassNumber:166; Mass:165.965761{78} ; Abundance:0;  StandardWeight:186.207{1}),
(Symbol:'Re'; AtomicNumber:75; MassNumber:167; Mass:166.962595{44#} ; Abundance:0;  StandardWeight:186.207{1}),
(Symbol:'Re'; AtomicNumber:75; MassNumber:168; Mass:167.961573{33} ; Abundance:0;  StandardWeight:186.207{1}),
(Symbol:'Re'; AtomicNumber:75; MassNumber:169; Mass:168.958766{12} ; Abundance:0;  StandardWeight:186.207{1}),
(Symbol:'Re'; AtomicNumber:75; MassNumber:170; Mass:169.958220{28} ; Abundance:0;  StandardWeight:186.207{1}),
(Symbol:'Re'; AtomicNumber:75; MassNumber:171; Mass:170.955716{30} ; Abundance:0;  StandardWeight:186.207{1}),
(Symbol:'Re'; AtomicNumber:75; MassNumber:172; Mass:171.955420{42} ; Abundance:0;  StandardWeight:186.207{1}),
(Symbol:'Re'; AtomicNumber:75; MassNumber:173; Mass:172.953243{30} ; Abundance:0;  StandardWeight:186.207{1}),
(Symbol:'Re'; AtomicNumber:75; MassNumber:174; Mass:173.953115{30} ; Abundance:0;  StandardWeight:186.207{1}),
(Symbol:'Re'; AtomicNumber:75; MassNumber:175; Mass:174.951381{30} ; Abundance:0;  StandardWeight:186.207{1}),
(Symbol:'Re'; AtomicNumber:75; MassNumber:176; Mass:175.951623{30} ; Abundance:0;  StandardWeight:186.207{1}),
(Symbol:'Re'; AtomicNumber:75; MassNumber:177; Mass:176.950328{30} ; Abundance:0;  StandardWeight:186.207{1}),
(Symbol:'Re'; AtomicNumber:75; MassNumber:178; Mass:177.950989{30} ; Abundance:0;  StandardWeight:186.207{1}),
(Symbol:'Re'; AtomicNumber:75; MassNumber:179; Mass:178.949989{26} ; Abundance:0;  StandardWeight:186.207{1}),
(Symbol:'Re'; AtomicNumber:75; MassNumber:180; Mass:179.950792{23} ; Abundance:0;  StandardWeight:186.207{1}),
(Symbol:'Re'; AtomicNumber:75; MassNumber:181; Mass:180.950058{14} ; Abundance:0;  StandardWeight:186.207{1}),
(Symbol:'Re'; AtomicNumber:75; MassNumber:182; Mass:181.95121{11} ; Abundance:0;  StandardWeight:186.207{1}),
(Symbol:'Re'; AtomicNumber:75; MassNumber:183; Mass:182.9508196{86} ; Abundance:0;  StandardWeight:186.207{1}),
(Symbol:'Re'; AtomicNumber:75; MassNumber:184; Mass:183.9525228{47} ; Abundance:0;  StandardWeight:186.207{1}),
(Symbol:'Re'; AtomicNumber:75; MassNumber:185; Mass:184.9529545{13} ; Abundance:37.4;  StandardWeight:186.207{1}),
(Symbol:'Re'; AtomicNumber:75; MassNumber:186; Mass:185.9549856{13} ; Abundance:0;  StandardWeight:186.207{1}),
(Symbol:'Re'; AtomicNumber:75; MassNumber:187; Mass:186.9557501{16} ; Abundance:62.6;  StandardWeight:186.207{1}),
(Symbol:'Re'; AtomicNumber:75; MassNumber:188; Mass:187.9581115{16} ; Abundance:0;  StandardWeight:186.207{1}),
(Symbol:'Re'; AtomicNumber:75; MassNumber:189; Mass:188.9592260{89} ; Abundance:0;  StandardWeight:186.207{1}),
(Symbol:'Re'; AtomicNumber:75; MassNumber:190; Mass:189.961744{76} ; Abundance:0;  StandardWeight:186.207{1}),
(Symbol:'Re'; AtomicNumber:75; MassNumber:191; Mass:190.963122{11} ; Abundance:0;  StandardWeight:186.207{1}),
(Symbol:'Re'; AtomicNumber:75; MassNumber:192; Mass:191.966088{82} ; Abundance:0;  StandardWeight:186.207{1}),
(Symbol:'Re'; AtomicNumber:75; MassNumber:193; Mass:192.967541{41} ; Abundance:0;  StandardWeight:186.207{1}),
(Symbol:'Re'; AtomicNumber:75; MassNumber:194; Mass:193.97076{21#} ; Abundance:0;  StandardWeight:186.207{1}),
(Symbol:'Re'; AtomicNumber:75; MassNumber:195; Mass:194.97254{32#} ; Abundance:0;  StandardWeight:186.207{1}),
(Symbol:'Re'; AtomicNumber:75; MassNumber:196; Mass:195.97580{32#} ; Abundance:0;  StandardWeight:186.207{1}),
(Symbol:'Re'; AtomicNumber:75; MassNumber:197; Mass:196.97799{32#} ; Abundance:0;  StandardWeight:186.207{1}),
(Symbol:'Re'; AtomicNumber:75; MassNumber:198; Mass:197.98160{43#} ; Abundance:0;  StandardWeight:186.207{1}),
(Symbol:'Os'; AtomicNumber:76; MassNumber:161; Mass:160.98903{43#} ; Abundance:0; {g} StandardWeight:190.23{3}),
(Symbol:'Os'; AtomicNumber:76; MassNumber:162; Mass:161.98443{54#} ; Abundance:0; {g} StandardWeight:190.23{3}),
(Symbol:'Os'; AtomicNumber:76; MassNumber:163; Mass:162.98241{32#} ; Abundance:0; {g} StandardWeight:190.23{3}),
(Symbol:'Os'; AtomicNumber:76; MassNumber:164; Mass:163.97802{17} ; Abundance:0; {g} StandardWeight:190.23{3}),
(Symbol:'Os'; AtomicNumber:76; MassNumber:165; Mass:164.97660{22#} ; Abundance:0; {g} StandardWeight:190.23{3}),
(Symbol:'Os'; AtomicNumber:76; MassNumber:166; Mass:165.972692{20} ; Abundance:0; {g} StandardWeight:190.23{3}),
(Symbol:'Os'; AtomicNumber:76; MassNumber:167; Mass:166.971549{78} ; Abundance:0; {g} StandardWeight:190.23{3}),
(Symbol:'Os'; AtomicNumber:76; MassNumber:168; Mass:167.967808{12} ; Abundance:0; {g} StandardWeight:190.23{3}),
(Symbol:'Os'; AtomicNumber:76; MassNumber:169; Mass:168.967018{27} ; Abundance:0; {g} StandardWeight:190.23{3}),
(Symbol:'Os'; AtomicNumber:76; MassNumber:170; Mass:169.963578{11} ; Abundance:0; {g} StandardWeight:190.23{3}),
(Symbol:'Os'; AtomicNumber:76; MassNumber:171; Mass:170.963174{19} ; Abundance:0; {g} StandardWeight:190.23{3}),
(Symbol:'Os'; AtomicNumber:76; MassNumber:172; Mass:171.960017{14} ; Abundance:0; {g} StandardWeight:190.23{3}),
(Symbol:'Os'; AtomicNumber:76; MassNumber:173; Mass:172.959808{16} ; Abundance:0; {g} StandardWeight:190.23{3}),
(Symbol:'Os'; AtomicNumber:76; MassNumber:174; Mass:173.957064{11} ; Abundance:0; {g} StandardWeight:190.23{3}),
(Symbol:'Os'; AtomicNumber:76; MassNumber:175; Mass:174.956945{13} ; Abundance:0; {g} StandardWeight:190.23{3}),
(Symbol:'Os'; AtomicNumber:76; MassNumber:176; Mass:175.954806{30} ; Abundance:0; {g} StandardWeight:190.23{3}),
(Symbol:'Os'; AtomicNumber:76; MassNumber:177; Mass:176.954966{17} ; Abundance:0; {g} StandardWeight:190.23{3}),
(Symbol:'Os'; AtomicNumber:76; MassNumber:178; Mass:177.953254{15} ; Abundance:0; {g} StandardWeight:190.23{3}),
(Symbol:'Os'; AtomicNumber:76; MassNumber:179; Mass:178.953817{18} ; Abundance:0; {g} StandardWeight:190.23{3}),
(Symbol:'Os'; AtomicNumber:76; MassNumber:180; Mass:179.952375{17} ; Abundance:0; {g} StandardWeight:190.23{3}),
(Symbol:'Os'; AtomicNumber:76; MassNumber:181; Mass:180.953247{27} ; Abundance:0; {g} StandardWeight:190.23{3}),
(Symbol:'Os'; AtomicNumber:76; MassNumber:182; Mass:181.952110{23} ; Abundance:0; {g} StandardWeight:190.23{3}),
(Symbol:'Os'; AtomicNumber:76; MassNumber:183; Mass:182.953125{53} ; Abundance:0; {g} StandardWeight:190.23{3}),
(Symbol:'Os'; AtomicNumber:76; MassNumber:184; Mass:183.9524885{14} ; Abundance:0.02; {g} StandardWeight:190.23{3}),
(Symbol:'Os'; AtomicNumber:76; MassNumber:185; Mass:184.9540417{14} ; Abundance:0; {g} StandardWeight:190.23{3}),
(Symbol:'Os'; AtomicNumber:76; MassNumber:186; Mass:185.9538350{16} ; Abundance:1.59; {g} StandardWeight:190.23{3}),
(Symbol:'Os'; AtomicNumber:76; MassNumber:187; Mass:186.9557474{16} ; Abundance:1.96; {g} StandardWeight:190.23{3}),
(Symbol:'Os'; AtomicNumber:76; MassNumber:188; Mass:187.9558352{16} ; Abundance:13.24; {g} StandardWeight:190.23{3}),
(Symbol:'Os'; AtomicNumber:76; MassNumber:189; Mass:188.9581442{17} ; Abundance:16.15; {g} StandardWeight:190.23{3}),
(Symbol:'Os'; AtomicNumber:76; MassNumber:190; Mass:189.9584437{17} ; Abundance:26.26; {g} StandardWeight:190.23{3}),
(Symbol:'Os'; AtomicNumber:76; MassNumber:191; Mass:190.9609264{17} ; Abundance:0; {g} StandardWeight:190.23{3}),
(Symbol:'Os'; AtomicNumber:76; MassNumber:192; Mass:191.9614770{29} ; Abundance:40.78; {g} StandardWeight:190.23{3}),
(Symbol:'Os'; AtomicNumber:76; MassNumber:193; Mass:192.9641479{29} ; Abundance:0; {g} StandardWeight:190.23{3}),
(Symbol:'Os'; AtomicNumber:76; MassNumber:194; Mass:193.9651772{30} ; Abundance:0; {g} StandardWeight:190.23{3}),
(Symbol:'Os'; AtomicNumber:76; MassNumber:195; Mass:194.968318{65} ; Abundance:0; {g} StandardWeight:190.23{3}),
(Symbol:'Os'; AtomicNumber:76; MassNumber:196; Mass:195.969641{43} ; Abundance:0; {g} StandardWeight:190.23{3}),
(Symbol:'Os'; AtomicNumber:76; MassNumber:197; Mass:196.97283{21#} ; Abundance:0; {g} StandardWeight:190.23{3}),
(Symbol:'Os'; AtomicNumber:76; MassNumber:198; Mass:197.97441{21#} ; Abundance:0; {g} StandardWeight:190.23{3}),
(Symbol:'Os'; AtomicNumber:76; MassNumber:199; Mass:198.97801{21#} ; Abundance:0; {g} StandardWeight:190.23{3}),
(Symbol:'Os'; AtomicNumber:76; MassNumber:200; Mass:199.97984{32#} ; Abundance:0; {g} StandardWeight:190.23{3}),
(Symbol:'Os'; AtomicNumber:76; MassNumber:201; Mass:200.98364{32#} ; Abundance:0; {g} StandardWeight:190.23{3}),
(Symbol:'Os'; AtomicNumber:76; MassNumber:202; Mass:201.98595{43#} ; Abundance:0; {g} StandardWeight:190.23{3}),
(Symbol:'Ir'; AtomicNumber:77; MassNumber:164; Mass:163.99191{34#} ; Abundance:0;  StandardWeight:192.217{3}),
(Symbol:'Ir'; AtomicNumber:77; MassNumber:165; Mass:164.98750{18#} ; Abundance:0;  StandardWeight:192.217{3}),
(Symbol:'Ir'; AtomicNumber:77; MassNumber:166; Mass:165.98566{22#} ; Abundance:0;  StandardWeight:192.217{3}),
(Symbol:'Ir'; AtomicNumber:77; MassNumber:167; Mass:166.981666{20} ; Abundance:0;  StandardWeight:192.217{3}),
(Symbol:'Ir'; AtomicNumber:77; MassNumber:168; Mass:167.979907{80} ; Abundance:0;  StandardWeight:192.217{3}),
(Symbol:'Ir'; AtomicNumber:77; MassNumber:169; Mass:168.976298{27} ; Abundance:0;  StandardWeight:192.217{3}),
(Symbol:'Ir'; AtomicNumber:77; MassNumber:170; Mass:169.974922{95#} ; Abundance:0;  StandardWeight:192.217{3}),
(Symbol:'Ir'; AtomicNumber:77; MassNumber:171; Mass:170.971640{42} ; Abundance:0;  StandardWeight:192.217{3}),
(Symbol:'Ir'; AtomicNumber:77; MassNumber:172; Mass:171.970607{35} ; Abundance:0;  StandardWeight:192.217{3}),
(Symbol:'Ir'; AtomicNumber:77; MassNumber:173; Mass:172.967506{12} ; Abundance:0;  StandardWeight:192.217{3}),
(Symbol:'Ir'; AtomicNumber:77; MassNumber:174; Mass:173.966861{30} ; Abundance:0;  StandardWeight:192.217{3}),
(Symbol:'Ir'; AtomicNumber:77; MassNumber:175; Mass:174.964150{13} ; Abundance:0;  StandardWeight:192.217{3}),
(Symbol:'Ir'; AtomicNumber:77; MassNumber:176; Mass:175.963650{22} ; Abundance:0;  StandardWeight:192.217{3}),
(Symbol:'Ir'; AtomicNumber:77; MassNumber:177; Mass:176.961301{21} ; Abundance:0;  StandardWeight:192.217{3}),
(Symbol:'Ir'; AtomicNumber:77; MassNumber:178; Mass:177.961082{21} ; Abundance:0;  StandardWeight:192.217{3}),
(Symbol:'Ir'; AtomicNumber:77; MassNumber:179; Mass:178.959120{10} ; Abundance:0;  StandardWeight:192.217{3}),
(Symbol:'Ir'; AtomicNumber:77; MassNumber:180; Mass:179.959229{23} ; Abundance:0;  StandardWeight:192.217{3}),
(Symbol:'Ir'; AtomicNumber:77; MassNumber:181; Mass:180.957625{28} ; Abundance:0;  StandardWeight:192.217{3}),
(Symbol:'Ir'; AtomicNumber:77; MassNumber:182; Mass:181.958076{23} ; Abundance:0;  StandardWeight:192.217{3}),
(Symbol:'Ir'; AtomicNumber:77; MassNumber:183; Mass:182.956840{26} ; Abundance:0;  StandardWeight:192.217{3}),
(Symbol:'Ir'; AtomicNumber:77; MassNumber:184; Mass:183.957476{30} ; Abundance:0;  StandardWeight:192.217{3}),
(Symbol:'Ir'; AtomicNumber:77; MassNumber:185; Mass:184.956698{30} ; Abundance:0;  StandardWeight:192.217{3}),
(Symbol:'Ir'; AtomicNumber:77; MassNumber:186; Mass:185.957944{18} ; Abundance:0;  StandardWeight:192.217{3}),
(Symbol:'Ir'; AtomicNumber:77; MassNumber:187; Mass:186.957542{30} ; Abundance:0;  StandardWeight:192.217{3}),
(Symbol:'Ir'; AtomicNumber:77; MassNumber:188; Mass:187.958828{10} ; Abundance:0;  StandardWeight:192.217{3}),
(Symbol:'Ir'; AtomicNumber:77; MassNumber:189; Mass:188.958715{14} ; Abundance:0;  StandardWeight:192.217{3}),
(Symbol:'Ir'; AtomicNumber:77; MassNumber:190; Mass:189.9605412{21} ; Abundance:0;  StandardWeight:192.217{3}),
(Symbol:'Ir'; AtomicNumber:77; MassNumber:191; Mass:190.9605893{21} ; Abundance:37.3;  StandardWeight:192.217{3}),
(Symbol:'Ir'; AtomicNumber:77; MassNumber:192; Mass:191.9626002{21} ; Abundance:0;  StandardWeight:192.217{3}),
(Symbol:'Ir'; AtomicNumber:77; MassNumber:193; Mass:192.9629216{21} ; Abundance:62.7;  StandardWeight:192.217{3}),
(Symbol:'Ir'; AtomicNumber:77; MassNumber:194; Mass:193.9650735{21} ; Abundance:0;  StandardWeight:192.217{3}),
(Symbol:'Ir'; AtomicNumber:77; MassNumber:195; Mass:194.9659747{21} ; Abundance:0;  StandardWeight:192.217{3}),
(Symbol:'Ir'; AtomicNumber:77; MassNumber:196; Mass:195.968397{41} ; Abundance:0;  StandardWeight:192.217{3}),
(Symbol:'Ir'; AtomicNumber:77; MassNumber:197; Mass:196.969655{22} ; Abundance:0;  StandardWeight:192.217{3}),
(Symbol:'Ir'; AtomicNumber:77; MassNumber:198; Mass:197.97228{21#} ; Abundance:0;  StandardWeight:192.217{3}),
(Symbol:'Ir'; AtomicNumber:77; MassNumber:199; Mass:198.973805{44} ; Abundance:0;  StandardWeight:192.217{3}),
(Symbol:'Ir'; AtomicNumber:77; MassNumber:200; Mass:199.97680{21#} ; Abundance:0;  StandardWeight:192.217{3}),
(Symbol:'Ir'; AtomicNumber:77; MassNumber:201; Mass:200.97864{21#} ; Abundance:0;  StandardWeight:192.217{3}),
(Symbol:'Ir'; AtomicNumber:77; MassNumber:202; Mass:201.98199{32#} ; Abundance:0;  StandardWeight:192.217{3}),
(Symbol:'Ir'; AtomicNumber:77; MassNumber:203; Mass:202.98423{43#} ; Abundance:0;  StandardWeight:192.217{3}),
(Symbol:'Ir'; AtomicNumber:77; MassNumber:204; Mass:203.98960{43#} ; Abundance:0;  StandardWeight:192.217{3}),
(Symbol:'Pt'; AtomicNumber:78; MassNumber:166; Mass:165.99486{54#} ; Abundance:0;  StandardWeight:195.084{9}),
(Symbol:'Pt'; AtomicNumber:78; MassNumber:167; Mass:166.99269{33#} ; Abundance:0;  StandardWeight:195.084{9}),
(Symbol:'Pt'; AtomicNumber:78; MassNumber:168; Mass:167.98813{17} ; Abundance:0;  StandardWeight:195.084{9}),
(Symbol:'Pt'; AtomicNumber:78; MassNumber:169; Mass:168.98657{22#} ; Abundance:0;  StandardWeight:195.084{9}),
(Symbol:'Pt'; AtomicNumber:78; MassNumber:170; Mass:169.982496{20} ; Abundance:0;  StandardWeight:195.084{9}),
(Symbol:'Pt'; AtomicNumber:78; MassNumber:171; Mass:170.981245{78} ; Abundance:0;  StandardWeight:195.084{9}),
(Symbol:'Pt'; AtomicNumber:78; MassNumber:172; Mass:171.977351{12} ; Abundance:0;  StandardWeight:195.084{9}),
(Symbol:'Pt'; AtomicNumber:78; MassNumber:173; Mass:172.976443{60} ; Abundance:0;  StandardWeight:195.084{9}),
(Symbol:'Pt'; AtomicNumber:78; MassNumber:174; Mass:173.972820{11} ; Abundance:0;  StandardWeight:195.084{9}),
(Symbol:'Pt'; AtomicNumber:78; MassNumber:175; Mass:174.972410{19} ; Abundance:0;  StandardWeight:195.084{9}),
(Symbol:'Pt'; AtomicNumber:78; MassNumber:176; Mass:175.968938{14} ; Abundance:0;  StandardWeight:195.084{9}),
(Symbol:'Pt'; AtomicNumber:78; MassNumber:177; Mass:176.968470{16} ; Abundance:0;  StandardWeight:195.084{9}),
(Symbol:'Pt'; AtomicNumber:78; MassNumber:178; Mass:177.965650{11} ; Abundance:0;  StandardWeight:195.084{9}),
(Symbol:'Pt'; AtomicNumber:78; MassNumber:179; Mass:178.9653590{86} ; Abundance:0;  StandardWeight:195.084{9}),
(Symbol:'Pt'; AtomicNumber:78; MassNumber:180; Mass:179.963032{12} ; Abundance:0;  StandardWeight:195.084{9}),
(Symbol:'Pt'; AtomicNumber:78; MassNumber:181; Mass:180.963098{16} ; Abundance:0;  StandardWeight:195.084{9}),
(Symbol:'Pt'; AtomicNumber:78; MassNumber:182; Mass:181.961172{14} ; Abundance:0;  StandardWeight:195.084{9}),
(Symbol:'Pt'; AtomicNumber:78; MassNumber:183; Mass:182.961597{17} ; Abundance:0;  StandardWeight:195.084{9}),
(Symbol:'Pt'; AtomicNumber:78; MassNumber:184; Mass:183.959915{17} ; Abundance:0;  StandardWeight:195.084{9}),
(Symbol:'Pt'; AtomicNumber:78; MassNumber:185; Mass:184.960614{28} ; Abundance:0;  StandardWeight:195.084{9}),
(Symbol:'Pt'; AtomicNumber:78; MassNumber:186; Mass:185.959351{23} ; Abundance:0;  StandardWeight:195.084{9}),
(Symbol:'Pt'; AtomicNumber:78; MassNumber:187; Mass:186.960617{26} ; Abundance:0;  StandardWeight:195.084{9}),
(Symbol:'Pt'; AtomicNumber:78; MassNumber:188; Mass:187.9593889{61} ; Abundance:0;  StandardWeight:195.084{9}),
(Symbol:'Pt'; AtomicNumber:78; MassNumber:189; Mass:188.960831{12} ; Abundance:0;  StandardWeight:195.084{9}),
(Symbol:'Pt'; AtomicNumber:78; MassNumber:190; Mass:189.9599297{63} ; Abundance:0.012;  StandardWeight:195.084{9}),
(Symbol:'Pt'; AtomicNumber:78; MassNumber:191; Mass:190.9616729{53} ; Abundance:0;  StandardWeight:195.084{9}),
(Symbol:'Pt'; AtomicNumber:78; MassNumber:192; Mass:191.9610387{32} ; Abundance:0.782;  StandardWeight:195.084{9}),
(Symbol:'Pt'; AtomicNumber:78; MassNumber:193; Mass:192.9629824{21} ; Abundance:0;  StandardWeight:195.084{9}),
(Symbol:'Pt'; AtomicNumber:78; MassNumber:194; Mass:193.9626809{10} ; Abundance:32.86;  StandardWeight:195.084{9}),
(Symbol:'Pt'; AtomicNumber:78; MassNumber:195; Mass:194.9647917{10} ; Abundance:33.78;  StandardWeight:195.084{9}),
(Symbol:'Pt'; AtomicNumber:78; MassNumber:196; Mass:195.96495209{99} ; Abundance:25.21;  StandardWeight:195.084{9}),
(Symbol:'Pt'; AtomicNumber:78; MassNumber:197; Mass:196.96734069{94} ; Abundance:0;  StandardWeight:195.084{9}),
(Symbol:'Pt'; AtomicNumber:78; MassNumber:198; Mass:197.9678949{23} ; Abundance:7.356;  StandardWeight:195.084{9}),
(Symbol:'Pt'; AtomicNumber:78; MassNumber:199; Mass:198.9705952{24} ; Abundance:0;  StandardWeight:195.084{9}),
(Symbol:'Pt'; AtomicNumber:78; MassNumber:200; Mass:199.971443{22} ; Abundance:0;  StandardWeight:195.084{9}),
(Symbol:'Pt'; AtomicNumber:78; MassNumber:201; Mass:200.974513{54} ; Abundance:0;  StandardWeight:195.084{9}),
(Symbol:'Pt'; AtomicNumber:78; MassNumber:202; Mass:201.975639{27} ; Abundance:0;  StandardWeight:195.084{9}),
(Symbol:'Pt'; AtomicNumber:78; MassNumber:203; Mass:202.97893{21#} ; Abundance:0;  StandardWeight:195.084{9}),
(Symbol:'Pt'; AtomicNumber:78; MassNumber:204; Mass:203.98076{21#} ; Abundance:0;  StandardWeight:195.084{9}),
(Symbol:'Pt'; AtomicNumber:78; MassNumber:205; Mass:204.98608{32#} ; Abundance:0;  StandardWeight:195.084{9}),
(Symbol:'Pt'; AtomicNumber:78; MassNumber:206; Mass:205.98966{32#} ; Abundance:0;  StandardWeight:195.084{9}),
(Symbol:'Au'; AtomicNumber:79; MassNumber:169; Mass:168.99808{32#} ; Abundance:0;  StandardWeight:196.966569{5}),
(Symbol:'Au'; AtomicNumber:79; MassNumber:170; Mass:169.99597{22#} ; Abundance:0;  StandardWeight:196.966569{5}),
(Symbol:'Au'; AtomicNumber:79; MassNumber:171; Mass:170.991876{22} ; Abundance:0;  StandardWeight:196.966569{5}),
(Symbol:'Au'; AtomicNumber:79; MassNumber:172; Mass:171.989942{81} ; Abundance:0;  StandardWeight:196.966569{5}),
(Symbol:'Au'; AtomicNumber:79; MassNumber:173; Mass:172.986241{26} ; Abundance:0;  StandardWeight:196.966569{5}),
(Symbol:'Au'; AtomicNumber:79; MassNumber:174; Mass:173.984717{95#} ; Abundance:0;  StandardWeight:196.966569{5}),
(Symbol:'Au'; AtomicNumber:79; MassNumber:175; Mass:174.981304{42} ; Abundance:0;  StandardWeight:196.966569{5}),
(Symbol:'Au'; AtomicNumber:79; MassNumber:176; Mass:175.980250{36} ; Abundance:0;  StandardWeight:196.966569{5}),
(Symbol:'Au'; AtomicNumber:79; MassNumber:177; Mass:176.976870{11} ; Abundance:0;  StandardWeight:196.966569{5}),
(Symbol:'Au'; AtomicNumber:79; MassNumber:178; Mass:177.976032{61} ; Abundance:0;  StandardWeight:196.966569{5}),
(Symbol:'Au'; AtomicNumber:79; MassNumber:179; Mass:178.973174{13} ; Abundance:0;  StandardWeight:196.966569{5}),
(Symbol:'Au'; AtomicNumber:79; MassNumber:180; Mass:179.972523{21} ; Abundance:0;  StandardWeight:196.966569{5}),
(Symbol:'Au'; AtomicNumber:79; MassNumber:181; Mass:180.970079{21} ; Abundance:0;  StandardWeight:196.966569{5}),
(Symbol:'Au'; AtomicNumber:79; MassNumber:182; Mass:181.969618{22} ; Abundance:0;  StandardWeight:196.966569{5}),
(Symbol:'Au'; AtomicNumber:79; MassNumber:183; Mass:182.967591{10} ; Abundance:0;  StandardWeight:196.966569{5}),
(Symbol:'Au'; AtomicNumber:79; MassNumber:184; Mass:183.967452{24} ; Abundance:0;  StandardWeight:196.966569{5}),
(Symbol:'Au'; AtomicNumber:79; MassNumber:185; Mass:184.965790{28} ; Abundance:0;  StandardWeight:196.966569{5}),
(Symbol:'Au'; AtomicNumber:79; MassNumber:186; Mass:185.965953{23} ; Abundance:0;  StandardWeight:196.966569{5}),
(Symbol:'Au'; AtomicNumber:79; MassNumber:187; Mass:186.964543{24} ; Abundance:0;  StandardWeight:196.966569{5}),
(Symbol:'Au'; AtomicNumber:79; MassNumber:188; Mass:187.965349{17} ; Abundance:0;  StandardWeight:196.966569{5}),
(Symbol:'Au'; AtomicNumber:79; MassNumber:189; Mass:188.963948{22} ; Abundance:0;  StandardWeight:196.966569{5}),
(Symbol:'Au'; AtomicNumber:79; MassNumber:190; Mass:189.964698{17} ; Abundance:0;  StandardWeight:196.966569{5}),
(Symbol:'Au'; AtomicNumber:79; MassNumber:191; Mass:190.963702{40} ; Abundance:0;  StandardWeight:196.966569{5}),
(Symbol:'Au'; AtomicNumber:79; MassNumber:192; Mass:191.964814{17} ; Abundance:0;  StandardWeight:196.966569{5}),
(Symbol:'Au'; AtomicNumber:79; MassNumber:193; Mass:192.9641373{93} ; Abundance:0;  StandardWeight:196.966569{5}),
(Symbol:'Au'; AtomicNumber:79; MassNumber:194; Mass:193.9654178{23} ; Abundance:0;  StandardWeight:196.966569{5}),
(Symbol:'Au'; AtomicNumber:79; MassNumber:195; Mass:194.9650352{15} ; Abundance:0;  StandardWeight:196.966569{5}),
(Symbol:'Au'; AtomicNumber:79; MassNumber:196; Mass:195.9665699{32} ; Abundance:0;  StandardWeight:196.966569{5}),
(Symbol:'Au'; AtomicNumber:79; MassNumber:197; Mass:196.96656879{71} ; Abundance:100;  StandardWeight:196.966569{5}),
(Symbol:'Au'; AtomicNumber:79; MassNumber:198; Mass:197.96824242{70} ; Abundance:0;  StandardWeight:196.966569{5}),
(Symbol:'Au'; AtomicNumber:79; MassNumber:199; Mass:198.96876528{70} ; Abundance:0;  StandardWeight:196.966569{5}),
(Symbol:'Au'; AtomicNumber:79; MassNumber:200; Mass:199.970756{29} ; Abundance:0;  StandardWeight:196.966569{5}),
(Symbol:'Au'; AtomicNumber:79; MassNumber:201; Mass:200.9716575{34} ; Abundance:0;  StandardWeight:196.966569{5}),
(Symbol:'Au'; AtomicNumber:79; MassNumber:202; Mass:201.973856{25} ; Abundance:0;  StandardWeight:196.966569{5}),
(Symbol:'Au'; AtomicNumber:79; MassNumber:203; Mass:202.9751544{33} ; Abundance:0;  StandardWeight:196.966569{5}),
(Symbol:'Au'; AtomicNumber:79; MassNumber:204; Mass:203.97783{22#} ; Abundance:0;  StandardWeight:196.966569{5}),
(Symbol:'Au'; AtomicNumber:79; MassNumber:205; Mass:204.97985{21#} ; Abundance:0;  StandardWeight:196.966569{5}),
(Symbol:'Au'; AtomicNumber:79; MassNumber:206; Mass:205.98474{32#} ; Abundance:0;  StandardWeight:196.966569{5}),
(Symbol:'Au'; AtomicNumber:79; MassNumber:207; Mass:206.98840{32#} ; Abundance:0;  StandardWeight:196.966569{5}),
(Symbol:'Au'; AtomicNumber:79; MassNumber:208; Mass:207.99345{32#} ; Abundance:0;  StandardWeight:196.966569{5}),
(Symbol:'Au'; AtomicNumber:79; MassNumber:209; Mass:208.99735{43#} ; Abundance:0;  StandardWeight:196.966569{5}),
(Symbol:'Au'; AtomicNumber:79; MassNumber:210; Mass:210.00250{43#} ; Abundance:0;  StandardWeight:196.966569{5}),
(Symbol:'Hg'; AtomicNumber:80; MassNumber:171; Mass:171.00353{33#} ; Abundance:0;  StandardWeight:200.592{3}),
(Symbol:'Hg'; AtomicNumber:80; MassNumber:172; Mass:171.99881{17} ; Abundance:0;  StandardWeight:200.592{3}),
(Symbol:'Hg'; AtomicNumber:80; MassNumber:173; Mass:172.99709{22#} ; Abundance:0;  StandardWeight:200.592{3}),
(Symbol:'Hg'; AtomicNumber:80; MassNumber:174; Mass:173.992865{21} ; Abundance:0;  StandardWeight:200.592{3}),
(Symbol:'Hg'; AtomicNumber:80; MassNumber:175; Mass:174.991441{78} ; Abundance:0;  StandardWeight:200.592{3}),
(Symbol:'Hg'; AtomicNumber:80; MassNumber:176; Mass:175.987361{14} ; Abundance:0;  StandardWeight:200.592{3}),
(Symbol:'Hg'; AtomicNumber:80; MassNumber:177; Mass:176.986277{81} ; Abundance:0;  StandardWeight:200.592{3}),
(Symbol:'Hg'; AtomicNumber:80; MassNumber:178; Mass:177.982484{12} ; Abundance:0;  StandardWeight:200.592{3}),
(Symbol:'Hg'; AtomicNumber:80; MassNumber:179; Mass:178.981831{29} ; Abundance:0;  StandardWeight:200.592{3}),
(Symbol:'Hg'; AtomicNumber:80; MassNumber:180; Mass:179.978260{14} ; Abundance:0;  StandardWeight:200.592{3}),
(Symbol:'Hg'; AtomicNumber:80; MassNumber:181; Mass:180.977819{17} ; Abundance:0;  StandardWeight:200.592{3}),
(Symbol:'Hg'; AtomicNumber:80; MassNumber:182; Mass:181.974689{11} ; Abundance:0;  StandardWeight:200.592{3}),
(Symbol:'Hg'; AtomicNumber:80; MassNumber:183; Mass:182.9744448{76} ; Abundance:0;  StandardWeight:200.592{3}),
(Symbol:'Hg'; AtomicNumber:80; MassNumber:184; Mass:183.971714{11} ; Abundance:0;  StandardWeight:200.592{3}),
(Symbol:'Hg'; AtomicNumber:80; MassNumber:185; Mass:184.971899{17} ; Abundance:0;  StandardWeight:200.592{3}),
(Symbol:'Hg'; AtomicNumber:80; MassNumber:186; Mass:185.969362{13} ; Abundance:0;  StandardWeight:200.592{3}),
(Symbol:'Hg'; AtomicNumber:80; MassNumber:187; Mass:186.969814{15} ; Abundance:0;  StandardWeight:200.592{3}),
(Symbol:'Hg'; AtomicNumber:80; MassNumber:188; Mass:187.967567{12} ; Abundance:0;  StandardWeight:200.592{3}),
(Symbol:'Hg'; AtomicNumber:80; MassNumber:189; Mass:188.968195{34} ; Abundance:0;  StandardWeight:200.592{3}),
(Symbol:'Hg'; AtomicNumber:80; MassNumber:190; Mass:189.966323{17} ; Abundance:0;  StandardWeight:200.592{3}),
(Symbol:'Hg'; AtomicNumber:80; MassNumber:191; Mass:190.967157{24} ; Abundance:0;  StandardWeight:200.592{3}),
(Symbol:'Hg'; AtomicNumber:80; MassNumber:192; Mass:191.965635{17} ; Abundance:0;  StandardWeight:200.592{3}),
(Symbol:'Hg'; AtomicNumber:80; MassNumber:193; Mass:192.966653{17} ; Abundance:0;  StandardWeight:200.592{3}),
(Symbol:'Hg'; AtomicNumber:80; MassNumber:194; Mass:193.9654491{31} ; Abundance:0;  StandardWeight:200.592{3}),
(Symbol:'Hg'; AtomicNumber:80; MassNumber:195; Mass:194.966721{25} ; Abundance:0;  StandardWeight:200.592{3}),
(Symbol:'Hg'; AtomicNumber:80; MassNumber:196; Mass:195.9658326{32} ; Abundance:0.15;  StandardWeight:200.592{3}),
(Symbol:'Hg'; AtomicNumber:80; MassNumber:197; Mass:196.9672128{35} ; Abundance:0;  StandardWeight:200.592{3}),
(Symbol:'Hg'; AtomicNumber:80; MassNumber:198; Mass:197.96676860{52} ; Abundance:9.97;  StandardWeight:200.592{3}),
(Symbol:'Hg'; AtomicNumber:80; MassNumber:199; Mass:198.96828064{46} ; Abundance:16.87;  StandardWeight:200.592{3}),
(Symbol:'Hg'; AtomicNumber:80; MassNumber:200; Mass:199.96832659{47} ; Abundance:23.1;  StandardWeight:200.592{3}),
(Symbol:'Hg'; AtomicNumber:80; MassNumber:201; Mass:200.97030284{69} ; Abundance:13.18;  StandardWeight:200.592{3}),
(Symbol:'Hg'; AtomicNumber:80; MassNumber:202; Mass:201.97064340{69} ; Abundance:29.86;  StandardWeight:200.592{3}),
(Symbol:'Hg'; AtomicNumber:80; MassNumber:203; Mass:202.9728728{18} ; Abundance:0;  StandardWeight:200.592{3}),
(Symbol:'Hg'; AtomicNumber:80; MassNumber:204; Mass:203.97349398{53} ; Abundance:6.87;  StandardWeight:200.592{3}),
(Symbol:'Hg'; AtomicNumber:80; MassNumber:205; Mass:204.9760734{39} ; Abundance:0;  StandardWeight:200.592{3}),
(Symbol:'Hg'; AtomicNumber:80; MassNumber:206; Mass:205.977514{22} ; Abundance:0;  StandardWeight:200.592{3}),
(Symbol:'Hg'; AtomicNumber:80; MassNumber:207; Mass:206.982300{32} ; Abundance:0;  StandardWeight:200.592{3}),
(Symbol:'Hg'; AtomicNumber:80; MassNumber:208; Mass:207.985759{33} ; Abundance:0;  StandardWeight:200.592{3}),
(Symbol:'Hg'; AtomicNumber:80; MassNumber:209; Mass:208.99072{16#} ; Abundance:0;  StandardWeight:200.592{3}),
(Symbol:'Hg'; AtomicNumber:80; MassNumber:210; Mass:209.99424{21#} ; Abundance:0;  StandardWeight:200.592{3}),
(Symbol:'Hg'; AtomicNumber:80; MassNumber:211; Mass:210.99933{21#} ; Abundance:0;  StandardWeight:200.592{3}),
(Symbol:'Hg'; AtomicNumber:80; MassNumber:212; Mass:212.00296{32#} ; Abundance:0;  StandardWeight:200.592{3}),
(Symbol:'Hg'; AtomicNumber:80; MassNumber:213; Mass:213.00823{32#} ; Abundance:0;  StandardWeight:200.592{3}),
(Symbol:'Hg'; AtomicNumber:80; MassNumber:214; Mass:214.01200{43#} ; Abundance:0;  StandardWeight:200.592{3}),
(Symbol:'Hg'; AtomicNumber:80; MassNumber:215; Mass:215.01740{43#} ; Abundance:0;  StandardWeight:200.592{3}),
(Symbol:'Hg'; AtomicNumber:80; MassNumber:216; Mass:216.02132{43#} ; Abundance:0;  StandardWeight:200.592{3}),
(Symbol:'Tl'; AtomicNumber:81; MassNumber:176; Mass:176.000624{81} ; Abundance:0;  StandardWeight:204.383 {204.382-204.385}),
(Symbol:'Tl'; AtomicNumber:81; MassNumber:177; Mass:176.996431{25} ; Abundance:0;  StandardWeight:204.383 {204.382-204.385}),
(Symbol:'Tl'; AtomicNumber:81; MassNumber:178; Mass:177.99485{11#} ; Abundance:0;  StandardWeight:204.383 {204.382-204.385}),
(Symbol:'Tl'; AtomicNumber:81; MassNumber:179; Mass:178.991111{43} ; Abundance:0;  StandardWeight:204.383 {204.382-204.385}),
(Symbol:'Tl'; AtomicNumber:81; MassNumber:180; Mass:179.990057{64} ; Abundance:0;  StandardWeight:204.383 {204.382-204.385}),
(Symbol:'Tl'; AtomicNumber:81; MassNumber:181; Mass:180.9862600{98} ; Abundance:0;  StandardWeight:204.383 {204.382-204.385}),
(Symbol:'Tl'; AtomicNumber:81; MassNumber:182; Mass:181.985713{63} ; Abundance:0;  StandardWeight:204.383 {204.382-204.385}),
(Symbol:'Tl'; AtomicNumber:81; MassNumber:183; Mass:182.982193{10} ; Abundance:0;  StandardWeight:204.383 {204.382-204.385}),
(Symbol:'Tl'; AtomicNumber:81; MassNumber:184; Mass:183.981886{22} ; Abundance:0;  StandardWeight:204.383 {204.382-204.385}),
(Symbol:'Tl'; AtomicNumber:81; MassNumber:185; Mass:184.978789{22} ; Abundance:0;  StandardWeight:204.383 {204.382-204.385}),
(Symbol:'Tl'; AtomicNumber:81; MassNumber:186; Mass:185.978651{24} ; Abundance:0;  StandardWeight:204.383 {204.382-204.385}),
(Symbol:'Tl'; AtomicNumber:81; MassNumber:187; Mass:186.9759063{88} ; Abundance:0;  StandardWeight:204.383 {204.382-204.385}),
(Symbol:'Tl'; AtomicNumber:81; MassNumber:188; Mass:187.976021{32} ; Abundance:0;  StandardWeight:204.383 {204.382-204.385}),
(Symbol:'Tl'; AtomicNumber:81; MassNumber:189; Mass:188.973588{12} ; Abundance:0;  StandardWeight:204.383 {204.382-204.385}),
(Symbol:'Tl'; AtomicNumber:81; MassNumber:190; Mass:189.973828{54#} ; Abundance:0;  StandardWeight:204.383 {204.382-204.385}),
(Symbol:'Tl'; AtomicNumber:81; MassNumber:191; Mass:190.9717842{79} ; Abundance:0;  StandardWeight:204.383 {204.382-204.385}),
(Symbol:'Tl'; AtomicNumber:81; MassNumber:192; Mass:191.972225{34} ; Abundance:0;  StandardWeight:204.383 {204.382-204.385}),
(Symbol:'Tl'; AtomicNumber:81; MassNumber:193; Mass:192.9705020{72} ; Abundance:0;  StandardWeight:204.383 {204.382-204.385}),
(Symbol:'Tl'; AtomicNumber:81; MassNumber:194; Mass:193.971081{15} ; Abundance:0;  StandardWeight:204.383 {204.382-204.385}),
(Symbol:'Tl'; AtomicNumber:81; MassNumber:195; Mass:194.969774{12} ; Abundance:0;  StandardWeight:204.383 {204.382-204.385}),
(Symbol:'Tl'; AtomicNumber:81; MassNumber:196; Mass:195.970481{13} ; Abundance:0;  StandardWeight:204.383 {204.382-204.385}),
(Symbol:'Tl'; AtomicNumber:81; MassNumber:197; Mass:196.969576{18} ; Abundance:0;  StandardWeight:204.383 {204.382-204.385}),
(Symbol:'Tl'; AtomicNumber:81; MassNumber:198; Mass:197.970483{86} ; Abundance:0;  StandardWeight:204.383 {204.382-204.385}),
(Symbol:'Tl'; AtomicNumber:81; MassNumber:199; Mass:198.969877{30} ; Abundance:0;  StandardWeight:204.383 {204.382-204.385}),
(Symbol:'Tl'; AtomicNumber:81; MassNumber:200; Mass:199.9709633{62} ; Abundance:0;  StandardWeight:204.383 {204.382-204.385}),
(Symbol:'Tl'; AtomicNumber:81; MassNumber:201; Mass:200.970822{15} ; Abundance:0;  StandardWeight:204.383 {204.382-204.385}),
(Symbol:'Tl'; AtomicNumber:81; MassNumber:202; Mass:201.972102{15} ; Abundance:0;  StandardWeight:204.383 {204.382-204.385}),
(Symbol:'Tl'; AtomicNumber:81; MassNumber:203; Mass:202.9723446{14} ; Abundance:29.52;  StandardWeight:204.383 {204.382-204.385}),
(Symbol:'Tl'; AtomicNumber:81; MassNumber:204; Mass:203.9738639{13} ; Abundance:0;  StandardWeight:204.383 {204.382-204.385}),
(Symbol:'Tl'; AtomicNumber:81; MassNumber:205; Mass:204.9744278{14} ; Abundance:70.48;  StandardWeight:204.383 {204.382-204.385}),
(Symbol:'Tl'; AtomicNumber:81; MassNumber:206; Mass:205.9761106{15} ; Abundance:0;  StandardWeight:204.383 {204.382-204.385}),
(Symbol:'Tl'; AtomicNumber:81; MassNumber:207; Mass:206.9774197{59} ; Abundance:0;  StandardWeight:204.383 {204.382-204.385}),
(Symbol:'Tl'; AtomicNumber:81; MassNumber:208; Mass:207.9820190{21} ; Abundance:0;  StandardWeight:204.383 {204.382-204.385}),
(Symbol:'Tl'; AtomicNumber:81; MassNumber:209; Mass:208.9853594{86} ; Abundance:0;  StandardWeight:204.383 {204.382-204.385}),
(Symbol:'Tl'; AtomicNumber:81; MassNumber:210; Mass:209.990074{12} ; Abundance:0;  StandardWeight:204.383 {204.382-204.385}),
(Symbol:'Tl'; AtomicNumber:81; MassNumber:211; Mass:210.993475{45} ; Abundance:0;  StandardWeight:204.383 {204.382-204.385}),
(Symbol:'Tl'; AtomicNumber:81; MassNumber:212; Mass:211.99834{22#} ; Abundance:0;  StandardWeight:204.383 {204.382-204.385}),
(Symbol:'Tl'; AtomicNumber:81; MassNumber:213; Mass:213.001915{29} ; Abundance:0;  StandardWeight:204.383 {204.382-204.385}),
(Symbol:'Tl'; AtomicNumber:81; MassNumber:214; Mass:214.00694{21#} ; Abundance:0;  StandardWeight:204.383 {204.382-204.385}),
(Symbol:'Tl'; AtomicNumber:81; MassNumber:215; Mass:215.01064{32#} ; Abundance:0;  StandardWeight:204.383 {204.382-204.385}),
(Symbol:'Tl'; AtomicNumber:81; MassNumber:216; Mass:216.01580{32#} ; Abundance:0;  StandardWeight:204.383 {204.382-204.385}),
(Symbol:'Tl'; AtomicNumber:81; MassNumber:217; Mass:217.01966{43#} ; Abundance:0;  StandardWeight:204.383 {204.382-204.385}),
(Symbol:'Tl'; AtomicNumber:81; MassNumber:218; Mass:218.02479{43#} ; Abundance:0;  StandardWeight:204.383 {204.382-204.385}),
(Symbol:'Pb'; AtomicNumber:82; MassNumber:178; Mass:178.003831{26} ; Abundance:0; {g,r} StandardWeight:207.2{1}),
(Symbol:'Pb'; AtomicNumber:82; MassNumber:179; Mass:179.002201{81} ; Abundance:0; {g,r} StandardWeight:207.2{1}),
(Symbol:'Pb'; AtomicNumber:82; MassNumber:180; Mass:179.997928{15} ; Abundance:0; {g,r} StandardWeight:207.2{1}),
(Symbol:'Pb'; AtomicNumber:82; MassNumber:181; Mass:180.996653{81} ; Abundance:0; {g,r} StandardWeight:207.2{1}),
(Symbol:'Pb'; AtomicNumber:82; MassNumber:182; Mass:181.992672{13} ; Abundance:0; {g,r} StandardWeight:207.2{1}),
(Symbol:'Pb'; AtomicNumber:82; MassNumber:183; Mass:182.991872{30} ; Abundance:0; {g,r} StandardWeight:207.2{1}),
(Symbol:'Pb'; AtomicNumber:82; MassNumber:184; Mass:183.988136{14} ; Abundance:0; {g,r} StandardWeight:207.2{1}),
(Symbol:'Pb'; AtomicNumber:82; MassNumber:185; Mass:184.987610{17} ; Abundance:0; {g,r} StandardWeight:207.2{1}),
(Symbol:'Pb'; AtomicNumber:82; MassNumber:186; Mass:185.984238{12} ; Abundance:0; {g,r} StandardWeight:207.2{1}),
(Symbol:'Pb'; AtomicNumber:82; MassNumber:187; Mass:186.9839109{55} ; Abundance:0; {g,r} StandardWeight:207.2{1}),
(Symbol:'Pb'; AtomicNumber:82; MassNumber:188; Mass:187.980875{11} ; Abundance:0; {g,r} StandardWeight:207.2{1}),
(Symbol:'Pb'; AtomicNumber:82; MassNumber:189; Mass:188.980807{37} ; Abundance:0; {g,r} StandardWeight:207.2{1}),
(Symbol:'Pb'; AtomicNumber:82; MassNumber:190; Mass:189.978082{13} ; Abundance:0; {g,r} StandardWeight:207.2{1}),
(Symbol:'Pb'; AtomicNumber:82; MassNumber:191; Mass:190.978276{41} ; Abundance:0; {g,r} StandardWeight:207.2{1}),
(Symbol:'Pb'; AtomicNumber:82; MassNumber:192; Mass:191.975775{13} ; Abundance:0; {g,r} StandardWeight:207.2{1}),
(Symbol:'Pb'; AtomicNumber:82; MassNumber:193; Mass:192.976173{53} ; Abundance:0; {g,r} StandardWeight:207.2{1}),
(Symbol:'Pb'; AtomicNumber:82; MassNumber:194; Mass:193.974012{19} ; Abundance:0; {g,r} StandardWeight:207.2{1}),
(Symbol:'Pb'; AtomicNumber:82; MassNumber:195; Mass:194.974543{25} ; Abundance:0; {g,r} StandardWeight:207.2{1}),
(Symbol:'Pb'; AtomicNumber:82; MassNumber:196; Mass:195.972774{15} ; Abundance:0; {g,r} StandardWeight:207.2{1}),
(Symbol:'Pb'; AtomicNumber:82; MassNumber:197; Mass:196.9734312{60} ; Abundance:0; {g,r} StandardWeight:207.2{1}),
(Symbol:'Pb'; AtomicNumber:82; MassNumber:198; Mass:197.972034{16} ; Abundance:0; {g,r} StandardWeight:207.2{1}),
(Symbol:'Pb'; AtomicNumber:82; MassNumber:199; Mass:198.972913{11} ; Abundance:0; {g,r} StandardWeight:207.2{1}),
(Symbol:'Pb'; AtomicNumber:82; MassNumber:200; Mass:199.971819{12} ; Abundance:0; {g,r} StandardWeight:207.2{1}),
(Symbol:'Pb'; AtomicNumber:82; MassNumber:201; Mass:200.972883{23} ; Abundance:0; {g,r} StandardWeight:207.2{1}),
(Symbol:'Pb'; AtomicNumber:82; MassNumber:202; Mass:201.9721520{40} ; Abundance:0; {g,r} StandardWeight:207.2{1}),
(Symbol:'Pb'; AtomicNumber:82; MassNumber:203; Mass:202.9733911{71} ; Abundance:0; {g,r} StandardWeight:207.2{1}),
(Symbol:'Pb'; AtomicNumber:82; MassNumber:204; Mass:203.9730440{13} ; Abundance:1.4; {g,r} StandardWeight:207.2{1}),
(Symbol:'Pb'; AtomicNumber:82; MassNumber:205; Mass:204.9744822{13} ; Abundance:0; {g,r} StandardWeight:207.2{1}),
(Symbol:'Pb'; AtomicNumber:82; MassNumber:206; Mass:205.9744657{13} ; Abundance:24.1; {g,r} StandardWeight:207.2{1}),
(Symbol:'Pb'; AtomicNumber:82; MassNumber:207; Mass:206.9758973{13} ; Abundance:22.1; {g,r} StandardWeight:207.2{1}),
(Symbol:'Pb'; AtomicNumber:82; MassNumber:208; Mass:207.9766525{13} ; Abundance:52.4; {g,r} StandardWeight:207.2{1}),
(Symbol:'Pb'; AtomicNumber:82; MassNumber:209; Mass:208.9810905{19} ; Abundance:0; {g,r} StandardWeight:207.2{1}),
(Symbol:'Pb'; AtomicNumber:82; MassNumber:210; Mass:209.9841889{16} ; Abundance:0; {g,r} StandardWeight:207.2{1}),
(Symbol:'Pb'; AtomicNumber:82; MassNumber:211; Mass:210.9887371{28} ; Abundance:0; {g,r} StandardWeight:207.2{1}),
(Symbol:'Pb'; AtomicNumber:82; MassNumber:212; Mass:211.9918977{23} ; Abundance:0; {g,r} StandardWeight:207.2{1}),
(Symbol:'Pb'; AtomicNumber:82; MassNumber:213; Mass:212.9965629{72} ; Abundance:0; {g,r} StandardWeight:207.2{1}),
(Symbol:'Pb'; AtomicNumber:82; MassNumber:214; Mass:213.9998059{25} ; Abundance:0; {g,r} StandardWeight:207.2{1}),
(Symbol:'Pb'; AtomicNumber:82; MassNumber:215; Mass:215.00474{11#} ; Abundance:0; {g,r} StandardWeight:207.2{1}),
(Symbol:'Pb'; AtomicNumber:82; MassNumber:216; Mass:216.00803{21#} ; Abundance:0; {g,r} StandardWeight:207.2{1}),
(Symbol:'Pb'; AtomicNumber:82; MassNumber:217; Mass:217.01314{32#} ; Abundance:0; {g,r} StandardWeight:207.2{1}),
(Symbol:'Pb'; AtomicNumber:82; MassNumber:218; Mass:218.01659{32#} ; Abundance:0; {g,r} StandardWeight:207.2{1}),
(Symbol:'Pb'; AtomicNumber:82; MassNumber:219; Mass:219.02177{43#} ; Abundance:0; {g,r} StandardWeight:207.2{1}),
(Symbol:'Pb'; AtomicNumber:82; MassNumber:220; Mass:220.02541{43#} ; Abundance:0; {g,r} StandardWeight:207.2{1}),
(Symbol:'Bi'; AtomicNumber:83; MassNumber:184; Mass:184.001275{84} ; Abundance:0;  StandardWeight:208.98040{1}),
(Symbol:'Bi'; AtomicNumber:83; MassNumber:185; Mass:184.997600{87#} ; Abundance:0;  StandardWeight:208.98040{1}),
(Symbol:'Bi'; AtomicNumber:83; MassNumber:186; Mass:185.996644{65} ; Abundance:0;  StandardWeight:208.98040{1}),
(Symbol:'Bi'; AtomicNumber:83; MassNumber:187; Mass:186.993147{11} ; Abundance:0;  StandardWeight:208.98040{1}),
(Symbol:'Bi'; AtomicNumber:83; MassNumber:188; Mass:187.992287{22} ; Abundance:0;  StandardWeight:208.98040{1}),
(Symbol:'Bi'; AtomicNumber:83; MassNumber:189; Mass:188.989195{22} ; Abundance:0;  StandardWeight:208.98040{1}),
(Symbol:'Bi'; AtomicNumber:83; MassNumber:190; Mass:189.988622{24} ; Abundance:0;  StandardWeight:208.98040{1}),
(Symbol:'Bi'; AtomicNumber:83; MassNumber:191; Mass:190.9857866{80} ; Abundance:0;  StandardWeight:208.98040{1}),
(Symbol:'Bi'; AtomicNumber:83; MassNumber:192; Mass:191.985469{33} ; Abundance:0;  StandardWeight:208.98040{1}),
(Symbol:'Bi'; AtomicNumber:83; MassNumber:193; Mass:192.982960{10} ; Abundance:0;  StandardWeight:208.98040{1}),
(Symbol:'Bi'; AtomicNumber:83; MassNumber:194; Mass:193.982785{54#} ; Abundance:0;  StandardWeight:208.98040{1}),
(Symbol:'Bi'; AtomicNumber:83; MassNumber:195; Mass:194.9806488{57} ; Abundance:0;  StandardWeight:208.98040{1}),
(Symbol:'Bi'; AtomicNumber:83; MassNumber:196; Mass:195.980667{26} ; Abundance:0;  StandardWeight:208.98040{1}),
(Symbol:'Bi'; AtomicNumber:83; MassNumber:197; Mass:196.9788651{89} ; Abundance:0;  StandardWeight:208.98040{1}),
(Symbol:'Bi'; AtomicNumber:83; MassNumber:198; Mass:197.979206{30} ; Abundance:0;  StandardWeight:208.98040{1}),
(Symbol:'Bi'; AtomicNumber:83; MassNumber:199; Mass:198.977673{11} ; Abundance:0;  StandardWeight:208.98040{1}),
(Symbol:'Bi'; AtomicNumber:83; MassNumber:200; Mass:199.978131{24} ; Abundance:0;  StandardWeight:208.98040{1}),
(Symbol:'Bi'; AtomicNumber:83; MassNumber:201; Mass:200.977010{16} ; Abundance:0;  StandardWeight:208.98040{1}),
(Symbol:'Bi'; AtomicNumber:83; MassNumber:202; Mass:201.977734{17} ; Abundance:0;  StandardWeight:208.98040{1}),
(Symbol:'Bi'; AtomicNumber:83; MassNumber:203; Mass:202.976893{14} ; Abundance:0;  StandardWeight:208.98040{1}),
(Symbol:'Bi'; AtomicNumber:83; MassNumber:204; Mass:203.9778361{99} ; Abundance:0;  StandardWeight:208.98040{1}),
(Symbol:'Bi'; AtomicNumber:83; MassNumber:205; Mass:204.9773867{55} ; Abundance:0;  StandardWeight:208.98040{1}),
(Symbol:'Bi'; AtomicNumber:83; MassNumber:206; Mass:205.9784993{82} ; Abundance:0;  StandardWeight:208.98040{1}),
(Symbol:'Bi'; AtomicNumber:83; MassNumber:207; Mass:206.9784710{26} ; Abundance:0;  StandardWeight:208.98040{1}),
(Symbol:'Bi'; AtomicNumber:83; MassNumber:208; Mass:207.9797425{25} ; Abundance:0;  StandardWeight:208.98040{1}),
(Symbol:'Bi'; AtomicNumber:83; MassNumber:209; Mass:208.9803991{16} ; Abundance:100;  StandardWeight:208.98040{1}),
(Symbol:'Bi'; AtomicNumber:83; MassNumber:210; Mass:209.9841207{16} ; Abundance:0;  StandardWeight:208.98040{1}),
(Symbol:'Bi'; AtomicNumber:83; MassNumber:211; Mass:210.9872697{59} ; Abundance:0;  StandardWeight:208.98040{1}),
(Symbol:'Bi'; AtomicNumber:83; MassNumber:212; Mass:211.9912860{21} ; Abundance:0;  StandardWeight:208.98040{1}),
(Symbol:'Bi'; AtomicNumber:83; MassNumber:213; Mass:212.9943851{56} ; Abundance:0;  StandardWeight:208.98040{1}),
(Symbol:'Bi'; AtomicNumber:83; MassNumber:214; Mass:213.998712{12} ; Abundance:0;  StandardWeight:208.98040{1}),
(Symbol:'Bi'; AtomicNumber:83; MassNumber:215; Mass:215.001770{16} ; Abundance:0;  StandardWeight:208.98040{1}),
(Symbol:'Bi'; AtomicNumber:83; MassNumber:216; Mass:216.006306{12} ; Abundance:0;  StandardWeight:208.98040{1}),
(Symbol:'Bi'; AtomicNumber:83; MassNumber:217; Mass:217.009372{19} ; Abundance:0;  StandardWeight:208.98040{1}),
(Symbol:'Bi'; AtomicNumber:83; MassNumber:218; Mass:218.014188{29} ; Abundance:0;  StandardWeight:208.98040{1}),
(Symbol:'Bi'; AtomicNumber:83; MassNumber:219; Mass:219.01748{21#} ; Abundance:0;  StandardWeight:208.98040{1}),
(Symbol:'Bi'; AtomicNumber:83; MassNumber:220; Mass:220.02235{32#} ; Abundance:0;  StandardWeight:208.98040{1}),
(Symbol:'Bi'; AtomicNumber:83; MassNumber:221; Mass:221.02587{32#} ; Abundance:0;  StandardWeight:208.98040{1}),
(Symbol:'Bi'; AtomicNumber:83; MassNumber:222; Mass:222.03078{32#} ; Abundance:0;  StandardWeight:208.98040{1}),
(Symbol:'Bi'; AtomicNumber:83; MassNumber:223; Mass:223.03450{43#} ; Abundance:0;  StandardWeight:208.98040{1}),
(Symbol:'Bi'; AtomicNumber:83; MassNumber:224; Mass:224.03947{43#} ; Abundance:0;  StandardWeight:208.98040{1}),
(Symbol:'Po'; AtomicNumber:84; MassNumber:186; Mass:186.004393{35} ; Abundance:0;  StandardWeight:209),
(Symbol:'Po'; AtomicNumber:84; MassNumber:187; Mass:187.003041{34} ; Abundance:0;  StandardWeight:209),
(Symbol:'Po'; AtomicNumber:84; MassNumber:188; Mass:187.999416{21} ; Abundance:0;  StandardWeight:209),
(Symbol:'Po'; AtomicNumber:84; MassNumber:189; Mass:188.998473{24} ; Abundance:0;  StandardWeight:209),
(Symbol:'Po'; AtomicNumber:84; MassNumber:190; Mass:189.995101{14} ; Abundance:0;  StandardWeight:209),
(Symbol:'Po'; AtomicNumber:84; MassNumber:191; Mass:190.9945585{76} ; Abundance:0;  StandardWeight:209),
(Symbol:'Po'; AtomicNumber:84; MassNumber:192; Mass:191.991336{12} ; Abundance:0;  StandardWeight:209),
(Symbol:'Po'; AtomicNumber:84; MassNumber:193; Mass:192.991026{37} ; Abundance:0;  StandardWeight:209),
(Symbol:'Po'; AtomicNumber:84; MassNumber:194; Mass:193.988186{14} ; Abundance:0;  StandardWeight:209),
(Symbol:'Po'; AtomicNumber:84; MassNumber:195; Mass:194.988126{41} ; Abundance:0;  StandardWeight:209),
(Symbol:'Po'; AtomicNumber:84; MassNumber:196; Mass:195.985526{14} ; Abundance:0;  StandardWeight:209),
(Symbol:'Po'; AtomicNumber:84; MassNumber:197; Mass:196.985660{53} ; Abundance:0;  StandardWeight:209),
(Symbol:'Po'; AtomicNumber:84; MassNumber:198; Mass:197.983389{19} ; Abundance:0;  StandardWeight:209),
(Symbol:'Po'; AtomicNumber:84; MassNumber:199; Mass:198.983667{25} ; Abundance:0;  StandardWeight:209),
(Symbol:'Po'; AtomicNumber:84; MassNumber:200; Mass:199.981799{15} ; Abundance:0;  StandardWeight:209),
(Symbol:'Po'; AtomicNumber:84; MassNumber:201; Mass:200.9822598{63} ; Abundance:0;  StandardWeight:209),
(Symbol:'Po'; AtomicNumber:84; MassNumber:202; Mass:201.980758{16} ; Abundance:0;  StandardWeight:209),
(Symbol:'Po'; AtomicNumber:84; MassNumber:203; Mass:202.9814161{93} ; Abundance:0;  StandardWeight:209),
(Symbol:'Po'; AtomicNumber:84; MassNumber:204; Mass:203.980310{12} ; Abundance:0;  StandardWeight:209),
(Symbol:'Po'; AtomicNumber:84; MassNumber:205; Mass:204.981203{22} ; Abundance:0;  StandardWeight:209),
(Symbol:'Po'; AtomicNumber:84; MassNumber:206; Mass:205.9804740{43} ; Abundance:0;  StandardWeight:209),
(Symbol:'Po'; AtomicNumber:84; MassNumber:207; Mass:206.9815938{72} ; Abundance:0;  StandardWeight:209),
(Symbol:'Po'; AtomicNumber:84; MassNumber:208; Mass:207.9812461{19} ; Abundance:0;  StandardWeight:209),
(Symbol:'Po'; AtomicNumber:84; MassNumber:209; Mass:208.9824308{20} ; Abundance:0;  StandardWeight:209),
(Symbol:'Po'; AtomicNumber:84; MassNumber:210; Mass:209.9828741{13} ; Abundance:0;  StandardWeight:209),
(Symbol:'Po'; AtomicNumber:84; MassNumber:211; Mass:210.9866536{14} ; Abundance:0;  StandardWeight:209),
(Symbol:'Po'; AtomicNumber:84; MassNumber:212; Mass:211.9888684{13} ; Abundance:0;  StandardWeight:209),
(Symbol:'Po'; AtomicNumber:84; MassNumber:213; Mass:212.9928576{33} ; Abundance:0;  StandardWeight:209),
(Symbol:'Po'; AtomicNumber:84; MassNumber:214; Mass:213.9952017{16} ; Abundance:0;  StandardWeight:209),
(Symbol:'Po'; AtomicNumber:84; MassNumber:215; Mass:214.9994201{27} ; Abundance:0;  StandardWeight:209),
(Symbol:'Po'; AtomicNumber:84; MassNumber:216; Mass:216.0019152{23} ; Abundance:0;  StandardWeight:209),
(Symbol:'Po'; AtomicNumber:84; MassNumber:217; Mass:217.0063182{67} ; Abundance:0;  StandardWeight:209),
(Symbol:'Po'; AtomicNumber:84; MassNumber:218; Mass:218.0089735{25} ; Abundance:0;  StandardWeight:209),
(Symbol:'Po'; AtomicNumber:84; MassNumber:219; Mass:219.013614{17} ; Abundance:0;  StandardWeight:209),
(Symbol:'Po'; AtomicNumber:84; MassNumber:220; Mass:220.016386{19} ; Abundance:0;  StandardWeight:209),
(Symbol:'Po'; AtomicNumber:84; MassNumber:221; Mass:221.021228{21} ; Abundance:0;  StandardWeight:209),
(Symbol:'Po'; AtomicNumber:84; MassNumber:222; Mass:222.024140{43} ; Abundance:0;  StandardWeight:209),
(Symbol:'Po'; AtomicNumber:84; MassNumber:223; Mass:223.02907{21#} ; Abundance:0;  StandardWeight:209),
(Symbol:'Po'; AtomicNumber:84; MassNumber:224; Mass:224.03211{21#} ; Abundance:0;  StandardWeight:209),
(Symbol:'Po'; AtomicNumber:84; MassNumber:225; Mass:225.03707{32#} ; Abundance:0;  StandardWeight:209),
(Symbol:'Po'; AtomicNumber:84; MassNumber:226; Mass:226.04031{43#} ; Abundance:0;  StandardWeight:209),
(Symbol:'Po'; AtomicNumber:84; MassNumber:227; Mass:227.04539{43#} ; Abundance:0;  StandardWeight:209),
(Symbol:'At'; AtomicNumber:85; MassNumber:191; Mass:191.004148{17} ; Abundance:0;  StandardWeight:210),
(Symbol:'At'; AtomicNumber:85; MassNumber:192; Mass:192.003152{35} ; Abundance:0;  StandardWeight:210),
(Symbol:'At'; AtomicNumber:85; MassNumber:193; Mass:192.999927{23} ; Abundance:0;  StandardWeight:210),
(Symbol:'At'; AtomicNumber:85; MassNumber:194; Mass:193.999236{29} ; Abundance:0;  StandardWeight:210),
(Symbol:'At'; AtomicNumber:85; MassNumber:195; Mass:194.9962685{98} ; Abundance:0;  StandardWeight:210),
(Symbol:'At'; AtomicNumber:85; MassNumber:196; Mass:195.995800{33} ; Abundance:0;  StandardWeight:210),
(Symbol:'At'; AtomicNumber:85; MassNumber:197; Mass:196.993189{55} ; Abundance:0;  StandardWeight:210),
(Symbol:'At'; AtomicNumber:85; MassNumber:198; Mass:197.992784{54#} ; Abundance:0;  StandardWeight:210),
(Symbol:'At'; AtomicNumber:85; MassNumber:199; Mass:198.9905277{58} ; Abundance:0;  StandardWeight:210),
(Symbol:'At'; AtomicNumber:85; MassNumber:200; Mass:199.990351{26} ; Abundance:0;  StandardWeight:210),
(Symbol:'At'; AtomicNumber:85; MassNumber:201; Mass:200.9884171{88} ; Abundance:0;  StandardWeight:210),
(Symbol:'At'; AtomicNumber:85; MassNumber:202; Mass:201.988630{30} ; Abundance:0;  StandardWeight:210),
(Symbol:'At'; AtomicNumber:85; MassNumber:203; Mass:202.986943{11} ; Abundance:0;  StandardWeight:210),
(Symbol:'At'; AtomicNumber:85; MassNumber:204; Mass:203.987251{24} ; Abundance:0;  StandardWeight:210),
(Symbol:'At'; AtomicNumber:85; MassNumber:205; Mass:204.986076{16} ; Abundance:0;  StandardWeight:210),
(Symbol:'At'; AtomicNumber:85; MassNumber:206; Mass:205.986657{16} ; Abundance:0;  StandardWeight:210),
(Symbol:'At'; AtomicNumber:85; MassNumber:207; Mass:206.985800{13} ; Abundance:0;  StandardWeight:210),
(Symbol:'At'; AtomicNumber:85; MassNumber:208; Mass:207.9866133{96} ; Abundance:0;  StandardWeight:210),
(Symbol:'At'; AtomicNumber:85; MassNumber:209; Mass:208.9861702{55} ; Abundance:0;  StandardWeight:210),
(Symbol:'At'; AtomicNumber:85; MassNumber:210; Mass:209.9871479{83} ; Abundance:0;  StandardWeight:210),
(Symbol:'At'; AtomicNumber:85; MassNumber:211; Mass:210.9874966{30} ; Abundance:0;  StandardWeight:210),
(Symbol:'At'; AtomicNumber:85; MassNumber:212; Mass:211.9907377{26} ; Abundance:0;  StandardWeight:210),
(Symbol:'At'; AtomicNumber:85; MassNumber:213; Mass:212.9929370{53} ; Abundance:0;  StandardWeight:210),
(Symbol:'At'; AtomicNumber:85; MassNumber:214; Mass:213.9963721{46} ; Abundance:0;  StandardWeight:210),
(Symbol:'At'; AtomicNumber:85; MassNumber:215; Mass:214.9986528{73} ; Abundance:0;  StandardWeight:210),
(Symbol:'At'; AtomicNumber:85; MassNumber:216; Mass:216.0024236{39} ; Abundance:0;  StandardWeight:210),
(Symbol:'At'; AtomicNumber:85; MassNumber:217; Mass:217.0047192{55} ; Abundance:0;  StandardWeight:210),
(Symbol:'At'; AtomicNumber:85; MassNumber:218; Mass:218.008695{12} ; Abundance:0;  StandardWeight:210),
(Symbol:'At'; AtomicNumber:85; MassNumber:219; Mass:219.0111618{42} ; Abundance:0;  StandardWeight:210),
(Symbol:'At'; AtomicNumber:85; MassNumber:220; Mass:220.015433{15} ; Abundance:0;  StandardWeight:210),
(Symbol:'At'; AtomicNumber:85; MassNumber:221; Mass:221.018017{15} ; Abundance:0;  StandardWeight:210),
(Symbol:'At'; AtomicNumber:85; MassNumber:222; Mass:222.022494{17} ; Abundance:0;  StandardWeight:210),
(Symbol:'At'; AtomicNumber:85; MassNumber:223; Mass:223.025151{15} ; Abundance:0;  StandardWeight:210),
(Symbol:'At'; AtomicNumber:85; MassNumber:224; Mass:224.029749{24} ; Abundance:0;  StandardWeight:210),
(Symbol:'At'; AtomicNumber:85; MassNumber:225; Mass:225.03263{32#} ; Abundance:0;  StandardWeight:210),
(Symbol:'At'; AtomicNumber:85; MassNumber:226; Mass:226.03716{32#} ; Abundance:0;  StandardWeight:210),
(Symbol:'At'; AtomicNumber:85; MassNumber:227; Mass:227.04024{32#} ; Abundance:0;  StandardWeight:210),
(Symbol:'At'; AtomicNumber:85; MassNumber:228; Mass:228.04475{43#} ; Abundance:0;  StandardWeight:210),
(Symbol:'At'; AtomicNumber:85; MassNumber:229; Mass:229.04812{43#} ; Abundance:0;  StandardWeight:210),
(Symbol:'Rn'; AtomicNumber:86; MassNumber:193; Mass:193.009708{27} ; Abundance:0;  StandardWeight:222),
(Symbol:'Rn'; AtomicNumber:86; MassNumber:194; Mass:194.006144{18} ; Abundance:0;  StandardWeight:222),
(Symbol:'Rn'; AtomicNumber:86; MassNumber:195; Mass:195.005422{54} ; Abundance:0;  StandardWeight:222),
(Symbol:'Rn'; AtomicNumber:86; MassNumber:196; Mass:196.002116{15} ; Abundance:0;  StandardWeight:222),
(Symbol:'Rn'; AtomicNumber:86; MassNumber:197; Mass:197.001585{38} ; Abundance:0;  StandardWeight:222),
(Symbol:'Rn'; AtomicNumber:86; MassNumber:198; Mass:197.998679{14} ; Abundance:0;  StandardWeight:222),
(Symbol:'Rn'; AtomicNumber:86; MassNumber:199; Mass:198.998390{68} ; Abundance:0;  StandardWeight:222),
(Symbol:'Rn'; AtomicNumber:86; MassNumber:200; Mass:199.995690{14} ; Abundance:0;  StandardWeight:222),
(Symbol:'Rn'; AtomicNumber:86; MassNumber:201; Mass:200.995628{53} ; Abundance:0;  StandardWeight:222),
(Symbol:'Rn'; AtomicNumber:86; MassNumber:202; Mass:201.993264{19} ; Abundance:0;  StandardWeight:222),
(Symbol:'Rn'; AtomicNumber:86; MassNumber:203; Mass:202.993388{25} ; Abundance:0;  StandardWeight:222),
(Symbol:'Rn'; AtomicNumber:86; MassNumber:204; Mass:203.991430{16} ; Abundance:0;  StandardWeight:222),
(Symbol:'Rn'; AtomicNumber:86; MassNumber:205; Mass:204.991719{54} ; Abundance:0;  StandardWeight:222),
(Symbol:'Rn'; AtomicNumber:86; MassNumber:206; Mass:205.990214{16} ; Abundance:0;  StandardWeight:222),
(Symbol:'Rn'; AtomicNumber:86; MassNumber:207; Mass:206.9907303{91} ; Abundance:0;  StandardWeight:222),
(Symbol:'Rn'; AtomicNumber:86; MassNumber:208; Mass:207.989635{12} ; Abundance:0;  StandardWeight:222),
(Symbol:'Rn'; AtomicNumber:86; MassNumber:209; Mass:208.990415{22} ; Abundance:0;  StandardWeight:222),
(Symbol:'Rn'; AtomicNumber:86; MassNumber:210; Mass:209.9896891{49} ; Abundance:0;  StandardWeight:222),
(Symbol:'Rn'; AtomicNumber:86; MassNumber:211; Mass:210.9906011{73} ; Abundance:0;  StandardWeight:222),
(Symbol:'Rn'; AtomicNumber:86; MassNumber:212; Mass:211.9907039{34} ; Abundance:0;  StandardWeight:222),
(Symbol:'Rn'; AtomicNumber:86; MassNumber:213; Mass:212.9938831{61} ; Abundance:0;  StandardWeight:222),
(Symbol:'Rn'; AtomicNumber:86; MassNumber:214; Mass:213.9953630{99} ; Abundance:0;  StandardWeight:222),
(Symbol:'Rn'; AtomicNumber:86; MassNumber:215; Mass:214.9987459{83} ; Abundance:0;  StandardWeight:222),
(Symbol:'Rn'; AtomicNumber:86; MassNumber:216; Mass:216.0002719{65} ; Abundance:0;  StandardWeight:222),
(Symbol:'Rn'; AtomicNumber:86; MassNumber:217; Mass:217.0039280{45} ; Abundance:0;  StandardWeight:222),
(Symbol:'Rn'; AtomicNumber:86; MassNumber:218; Mass:218.0056016{25} ; Abundance:0;  StandardWeight:222),
(Symbol:'Rn'; AtomicNumber:86; MassNumber:219; Mass:219.0094804{27} ; Abundance:0;  StandardWeight:222),
(Symbol:'Rn'; AtomicNumber:86; MassNumber:220; Mass:220.0113941{23} ; Abundance:0;  StandardWeight:222),
(Symbol:'Rn'; AtomicNumber:86; MassNumber:221; Mass:221.0155371{63} ; Abundance:0;  StandardWeight:222),
(Symbol:'Rn'; AtomicNumber:86; MassNumber:222; Mass:222.0175782{25} ; Abundance:0;  StandardWeight:222),
(Symbol:'Rn'; AtomicNumber:86; MassNumber:223; Mass:223.0218893{84} ; Abundance:0;  StandardWeight:222),
(Symbol:'Rn'; AtomicNumber:86; MassNumber:224; Mass:224.024096{11} ; Abundance:0;  StandardWeight:222),
(Symbol:'Rn'; AtomicNumber:86; MassNumber:225; Mass:225.028486{12} ; Abundance:0;  StandardWeight:222),
(Symbol:'Rn'; AtomicNumber:86; MassNumber:226; Mass:226.030861{11} ; Abundance:0;  StandardWeight:222),
(Symbol:'Rn'; AtomicNumber:86; MassNumber:227; Mass:227.035304{15} ; Abundance:0;  StandardWeight:222),
(Symbol:'Rn'; AtomicNumber:86; MassNumber:228; Mass:228.037835{19} ; Abundance:0;  StandardWeight:222),
(Symbol:'Rn'; AtomicNumber:86; MassNumber:229; Mass:229.042257{14} ; Abundance:0;  StandardWeight:222),
(Symbol:'Rn'; AtomicNumber:86; MassNumber:230; Mass:230.04514{21#} ; Abundance:0;  StandardWeight:222),
(Symbol:'Rn'; AtomicNumber:86; MassNumber:231; Mass:231.04987{32#} ; Abundance:0;  StandardWeight:222),
(Symbol:'Fr'; AtomicNumber:87; MassNumber:199; Mass:199.007259{45} ; Abundance:0;  StandardWeight:223),
(Symbol:'Fr'; AtomicNumber:87; MassNumber:200; Mass:200.006586{63} ; Abundance:0;  StandardWeight:223),
(Symbol:'Fr'; AtomicNumber:87; MassNumber:201; Mass:201.003867{77} ; Abundance:0;  StandardWeight:223),
(Symbol:'Fr'; AtomicNumber:87; MassNumber:202; Mass:202.003320{55#} ; Abundance:0;  StandardWeight:223),
(Symbol:'Fr'; AtomicNumber:87; MassNumber:203; Mass:203.0009407{67} ; Abundance:0;  StandardWeight:223),
(Symbol:'Fr'; AtomicNumber:87; MassNumber:204; Mass:204.000652{26} ; Abundance:0;  StandardWeight:223),
(Symbol:'Fr'; AtomicNumber:87; MassNumber:205; Mass:204.9985939{84} ; Abundance:0;  StandardWeight:223),
(Symbol:'Fr'; AtomicNumber:87; MassNumber:206; Mass:205.998666{30} ; Abundance:0;  StandardWeight:223),
(Symbol:'Fr'; AtomicNumber:87; MassNumber:207; Mass:206.996946{19} ; Abundance:0;  StandardWeight:223),
(Symbol:'Fr'; AtomicNumber:87; MassNumber:208; Mass:207.997138{12} ; Abundance:0;  StandardWeight:223),
(Symbol:'Fr'; AtomicNumber:87; MassNumber:209; Mass:208.995955{16} ; Abundance:0;  StandardWeight:223),
(Symbol:'Fr'; AtomicNumber:87; MassNumber:210; Mass:209.996422{16} ; Abundance:0;  StandardWeight:223),
(Symbol:'Fr'; AtomicNumber:87; MassNumber:211; Mass:210.995556{13} ; Abundance:0;  StandardWeight:223),
(Symbol:'Fr'; AtomicNumber:87; MassNumber:212; Mass:211.9962257{94} ; Abundance:0;  StandardWeight:223),
(Symbol:'Fr'; AtomicNumber:87; MassNumber:213; Mass:212.9961860{55} ; Abundance:0;  StandardWeight:223),
(Symbol:'Fr'; AtomicNumber:87; MassNumber:214; Mass:213.9989713{93} ; Abundance:0;  StandardWeight:223),
(Symbol:'Fr'; AtomicNumber:87; MassNumber:215; Mass:215.0003418{76} ; Abundance:0;  StandardWeight:223),
(Symbol:'Fr'; AtomicNumber:87; MassNumber:216; Mass:216.0031899{45} ; Abundance:0;  StandardWeight:223),
(Symbol:'Fr'; AtomicNumber:87; MassNumber:217; Mass:217.0046323{70} ; Abundance:0;  StandardWeight:223),
(Symbol:'Fr'; AtomicNumber:87; MassNumber:218; Mass:218.0075787{51} ; Abundance:0;  StandardWeight:223),
(Symbol:'Fr'; AtomicNumber:87; MassNumber:219; Mass:219.0092524{76} ; Abundance:0;  StandardWeight:223),
(Symbol:'Fr'; AtomicNumber:87; MassNumber:220; Mass:220.0123277{44} ; Abundance:0;  StandardWeight:223),
(Symbol:'Fr'; AtomicNumber:87; MassNumber:221; Mass:221.0142552{54} ; Abundance:0;  StandardWeight:223),
(Symbol:'Fr'; AtomicNumber:87; MassNumber:222; Mass:222.017552{23} ; Abundance:0;  StandardWeight:223),
(Symbol:'Fr'; AtomicNumber:87; MassNumber:223; Mass:223.0197360{25} ; Abundance:0;  StandardWeight:223),
(Symbol:'Fr'; AtomicNumber:87; MassNumber:224; Mass:224.023398{14} ; Abundance:0;  StandardWeight:223),
(Symbol:'Fr'; AtomicNumber:87; MassNumber:225; Mass:225.025573{13} ; Abundance:0;  StandardWeight:223),
(Symbol:'Fr'; AtomicNumber:87; MassNumber:226; Mass:226.029566{13} ; Abundance:0;  StandardWeight:223),
(Symbol:'Fr'; AtomicNumber:87; MassNumber:227; Mass:227.031869{14} ; Abundance:0;  StandardWeight:223),
(Symbol:'Fr'; AtomicNumber:87; MassNumber:228; Mass:228.035823{14} ; Abundance:0;  StandardWeight:223),
(Symbol:'Fr'; AtomicNumber:87; MassNumber:229; Mass:229.038298{15} ; Abundance:0;  StandardWeight:223),
(Symbol:'Fr'; AtomicNumber:87; MassNumber:230; Mass:230.042416{17} ; Abundance:0;  StandardWeight:223),
(Symbol:'Fr'; AtomicNumber:87; MassNumber:231; Mass:231.045158{27} ; Abundance:0;  StandardWeight:223),
(Symbol:'Fr'; AtomicNumber:87; MassNumber:232; Mass:232.04937{17#} ; Abundance:0;  StandardWeight:223),
(Symbol:'Fr'; AtomicNumber:87; MassNumber:233; Mass:233.05264{32#} ; Abundance:0;  StandardWeight:223),
(Symbol:'Ra'; AtomicNumber:88; MassNumber:201; Mass:201.01271{11#} ; Abundance:0;  StandardWeight:226),
(Symbol:'Ra'; AtomicNumber:88; MassNumber:202; Mass:202.009760{26} ; Abundance:0;  StandardWeight:226),
(Symbol:'Ra'; AtomicNumber:88; MassNumber:203; Mass:203.009304{86} ; Abundance:0;  StandardWeight:226),
(Symbol:'Ra'; AtomicNumber:88; MassNumber:204; Mass:204.006492{16} ; Abundance:0;  StandardWeight:226),
(Symbol:'Ra'; AtomicNumber:88; MassNumber:205; Mass:205.006268{76} ; Abundance:0;  StandardWeight:226),
(Symbol:'Ra'; AtomicNumber:88; MassNumber:206; Mass:206.003828{19} ; Abundance:0;  StandardWeight:226),
(Symbol:'Ra'; AtomicNumber:88; MassNumber:207; Mass:207.003799{59} ; Abundance:0;  StandardWeight:226),
(Symbol:'Ra'; AtomicNumber:88; MassNumber:208; Mass:208.001841{17} ; Abundance:0;  StandardWeight:226),
(Symbol:'Ra'; AtomicNumber:88; MassNumber:209; Mass:209.001990{54} ; Abundance:0;  StandardWeight:226),
(Symbol:'Ra'; AtomicNumber:88; MassNumber:210; Mass:210.000494{16} ; Abundance:0;  StandardWeight:226),
(Symbol:'Ra'; AtomicNumber:88; MassNumber:211; Mass:211.0008932{85} ; Abundance:0;  StandardWeight:226),
(Symbol:'Ra'; AtomicNumber:88; MassNumber:212; Mass:211.999787{12} ; Abundance:0;  StandardWeight:226),
(Symbol:'Ra'; AtomicNumber:88; MassNumber:213; Mass:213.000384{22} ; Abundance:0;  StandardWeight:226),
(Symbol:'Ra'; AtomicNumber:88; MassNumber:214; Mass:214.0000997{56} ; Abundance:0;  StandardWeight:226),
(Symbol:'Ra'; AtomicNumber:88; MassNumber:215; Mass:215.0027204{82} ; Abundance:0;  StandardWeight:226),
(Symbol:'Ra'; AtomicNumber:88; MassNumber:216; Mass:216.0035334{94} ; Abundance:0;  StandardWeight:226),
(Symbol:'Ra'; AtomicNumber:88; MassNumber:217; Mass:217.0063207{92} ; Abundance:0;  StandardWeight:226),
(Symbol:'Ra'; AtomicNumber:88; MassNumber:218; Mass:218.007141{12} ; Abundance:0;  StandardWeight:226),
(Symbol:'Ra'; AtomicNumber:88; MassNumber:219; Mass:219.0100855{89} ; Abundance:0;  StandardWeight:226),
(Symbol:'Ra'; AtomicNumber:88; MassNumber:220; Mass:220.0110259{89} ; Abundance:0;  StandardWeight:226),
(Symbol:'Ra'; AtomicNumber:88; MassNumber:221; Mass:221.0139177{50} ; Abundance:0;  StandardWeight:226),
(Symbol:'Ra'; AtomicNumber:88; MassNumber:222; Mass:222.0153748{49} ; Abundance:0;  StandardWeight:226),
(Symbol:'Ra'; AtomicNumber:88; MassNumber:223; Mass:223.0185023{27} ; Abundance:0;  StandardWeight:226),
(Symbol:'Ra'; AtomicNumber:88; MassNumber:224; Mass:224.0202120{23} ; Abundance:0;  StandardWeight:226),
(Symbol:'Ra'; AtomicNumber:88; MassNumber:225; Mass:225.0236119{32} ; Abundance:0;  StandardWeight:226),
(Symbol:'Ra'; AtomicNumber:88; MassNumber:226; Mass:226.0254103{25} ; Abundance:0;  StandardWeight:226),
(Symbol:'Ra'; AtomicNumber:88; MassNumber:227; Mass:227.0291783{25} ; Abundance:0;  StandardWeight:226),
(Symbol:'Ra'; AtomicNumber:88; MassNumber:228; Mass:228.0310707{26} ; Abundance:0;  StandardWeight:226),
(Symbol:'Ra'; AtomicNumber:88; MassNumber:229; Mass:229.034942{16} ; Abundance:0;  StandardWeight:226),
(Symbol:'Ra'; AtomicNumber:88; MassNumber:230; Mass:230.037055{11} ; Abundance:0;  StandardWeight:226),
(Symbol:'Ra'; AtomicNumber:88; MassNumber:231; Mass:231.041027{12} ; Abundance:0;  StandardWeight:226),
(Symbol:'Ra'; AtomicNumber:88; MassNumber:232; Mass:232.0434753{98} ; Abundance:0;  StandardWeight:226),
(Symbol:'Ra'; AtomicNumber:88; MassNumber:233; Mass:233.047582{17} ; Abundance:0;  StandardWeight:226),
(Symbol:'Ra'; AtomicNumber:88; MassNumber:234; Mass:234.050342{33} ; Abundance:0;  StandardWeight:226),
(Symbol:'Ra'; AtomicNumber:88; MassNumber:235; Mass:235.05497{32#} ; Abundance:0;  StandardWeight:226),
(Symbol:'Ac'; AtomicNumber:89; MassNumber:206; Mass:206.014452{77#} ; Abundance:0;  StandardWeight:227),
(Symbol:'Ac'; AtomicNumber:89; MassNumber:207; Mass:207.011966{54} ; Abundance:0;  StandardWeight:227),
(Symbol:'Ac'; AtomicNumber:89; MassNumber:208; Mass:208.011550{60} ; Abundance:0;  StandardWeight:227),
(Symbol:'Ac'; AtomicNumber:89; MassNumber:209; Mass:209.009495{54} ; Abundance:0;  StandardWeight:227),
(Symbol:'Ac'; AtomicNumber:89; MassNumber:210; Mass:210.009436{62} ; Abundance:0;  StandardWeight:227),
(Symbol:'Ac'; AtomicNumber:89; MassNumber:211; Mass:211.007732{57} ; Abundance:0;  StandardWeight:227),
(Symbol:'Ac'; AtomicNumber:89; MassNumber:212; Mass:212.007813{55} ; Abundance:0;  StandardWeight:227),
(Symbol:'Ac'; AtomicNumber:89; MassNumber:213; Mass:213.006609{56} ; Abundance:0;  StandardWeight:227),
(Symbol:'Ac'; AtomicNumber:89; MassNumber:214; Mass:214.006918{16} ; Abundance:0;  StandardWeight:227),
(Symbol:'Ac'; AtomicNumber:89; MassNumber:215; Mass:215.006475{13} ; Abundance:0;  StandardWeight:227),
(Symbol:'Ac'; AtomicNumber:89; MassNumber:216; Mass:216.008743{12} ; Abundance:0;  StandardWeight:227),
(Symbol:'Ac'; AtomicNumber:89; MassNumber:217; Mass:217.009344{12} ; Abundance:0;  StandardWeight:227),
(Symbol:'Ac'; AtomicNumber:89; MassNumber:218; Mass:218.011642{54} ; Abundance:0;  StandardWeight:227),
(Symbol:'Ac'; AtomicNumber:89; MassNumber:219; Mass:219.012421{54} ; Abundance:0;  StandardWeight:227),
(Symbol:'Ac'; AtomicNumber:89; MassNumber:220; Mass:220.0147549{66} ; Abundance:0;  StandardWeight:227),
(Symbol:'Ac'; AtomicNumber:89; MassNumber:221; Mass:221.015592{54} ; Abundance:0;  StandardWeight:227),
(Symbol:'Ac'; AtomicNumber:89; MassNumber:222; Mass:222.0178442{56} ; Abundance:0;  StandardWeight:227),
(Symbol:'Ac'; AtomicNumber:89; MassNumber:223; Mass:223.0191377{77} ; Abundance:0;  StandardWeight:227),
(Symbol:'Ac'; AtomicNumber:89; MassNumber:224; Mass:224.0217232{45} ; Abundance:0;  StandardWeight:227),
(Symbol:'Ac'; AtomicNumber:89; MassNumber:225; Mass:225.0232300{53} ; Abundance:0;  StandardWeight:227),
(Symbol:'Ac'; AtomicNumber:89; MassNumber:226; Mass:226.0260984{36} ; Abundance:0;  StandardWeight:227),
(Symbol:'Ac'; AtomicNumber:89; MassNumber:227; Mass:227.0277523{25} ; Abundance:0;  StandardWeight:227),
(Symbol:'Ac'; AtomicNumber:89; MassNumber:228; Mass:228.0310215{27} ; Abundance:0;  StandardWeight:227),
(Symbol:'Ac'; AtomicNumber:89; MassNumber:229; Mass:229.032956{13} ; Abundance:0;  StandardWeight:227),
(Symbol:'Ac'; AtomicNumber:89; MassNumber:230; Mass:230.036327{17} ; Abundance:0;  StandardWeight:227),
(Symbol:'Ac'; AtomicNumber:89; MassNumber:231; Mass:231.038393{14} ; Abundance:0;  StandardWeight:227),
(Symbol:'Ac'; AtomicNumber:89; MassNumber:232; Mass:232.042034{14} ; Abundance:0;  StandardWeight:227),
(Symbol:'Ac'; AtomicNumber:89; MassNumber:233; Mass:233.044346{14} ; Abundance:0;  StandardWeight:227),
(Symbol:'Ac'; AtomicNumber:89; MassNumber:234; Mass:234.048139{15} ; Abundance:0;  StandardWeight:227),
(Symbol:'Ac'; AtomicNumber:89; MassNumber:235; Mass:235.050840{15} ; Abundance:0;  StandardWeight:227),
(Symbol:'Ac'; AtomicNumber:89; MassNumber:236; Mass:236.054988{41} ; Abundance:0;  StandardWeight:227),
(Symbol:'Ac'; AtomicNumber:89; MassNumber:237; Mass:237.05827{43#} ; Abundance:0;  StandardWeight:227),
(Symbol:'Th'; AtomicNumber:90; MassNumber:208; Mass:208.017900{36} ; Abundance:0; {g} StandardWeight:232.0377{4}),
(Symbol:'Th'; AtomicNumber:90; MassNumber:209; Mass:209.017753{93} ; Abundance:0; {g} StandardWeight:232.0377{4}),
(Symbol:'Th'; AtomicNumber:90; MassNumber:210; Mass:210.015094{20} ; Abundance:0; {g} StandardWeight:232.0377{4}),
(Symbol:'Th'; AtomicNumber:90; MassNumber:211; Mass:211.014929{80} ; Abundance:0; {g} StandardWeight:232.0377{4}),
(Symbol:'Th'; AtomicNumber:90; MassNumber:212; Mass:212.012988{17} ; Abundance:0; {g} StandardWeight:232.0377{4}),
(Symbol:'Th'; AtomicNumber:90; MassNumber:213; Mass:213.013009{76} ; Abundance:0; {g} StandardWeight:232.0377{4}),
(Symbol:'Th'; AtomicNumber:90; MassNumber:214; Mass:214.011500{17} ; Abundance:0; {g} StandardWeight:232.0377{4}),
(Symbol:'Th'; AtomicNumber:90; MassNumber:215; Mass:215.0117248{95} ; Abundance:0; {g} StandardWeight:232.0377{4}),
(Symbol:'Th'; AtomicNumber:90; MassNumber:216; Mass:216.011056{13} ; Abundance:0; {g} StandardWeight:232.0377{4}),
(Symbol:'Th'; AtomicNumber:90; MassNumber:217; Mass:217.013117{22} ; Abundance:0; {g} StandardWeight:232.0377{4}),
(Symbol:'Th'; AtomicNumber:90; MassNumber:218; Mass:218.013276{11} ; Abundance:0; {g} StandardWeight:232.0377{4}),
(Symbol:'Th'; AtomicNumber:90; MassNumber:219; Mass:219.015537{54} ; Abundance:0; {g} StandardWeight:232.0377{4}),
(Symbol:'Th'; AtomicNumber:90; MassNumber:220; Mass:220.015748{24} ; Abundance:0; {g} StandardWeight:232.0377{4}),
(Symbol:'Th'; AtomicNumber:90; MassNumber:221; Mass:221.018184{10} ; Abundance:0; {g} StandardWeight:232.0377{4}),
(Symbol:'Th'; AtomicNumber:90; MassNumber:222; Mass:222.018469{13} ; Abundance:0; {g} StandardWeight:232.0377{4}),
(Symbol:'Th'; AtomicNumber:90; MassNumber:223; Mass:223.0208119{99} ; Abundance:0; {g} StandardWeight:232.0377{4}),
(Symbol:'Th'; AtomicNumber:90; MassNumber:224; Mass:224.021464{11} ; Abundance:0; {g} StandardWeight:232.0377{4}),
(Symbol:'Th'; AtomicNumber:90; MassNumber:225; Mass:225.0239514{55} ; Abundance:0; {g} StandardWeight:232.0377{4}),
(Symbol:'Th'; AtomicNumber:90; MassNumber:226; Mass:226.0249034{50} ; Abundance:0; {g} StandardWeight:232.0377{4}),
(Symbol:'Th'; AtomicNumber:90; MassNumber:227; Mass:227.0277042{27} ; Abundance:0; {g} StandardWeight:232.0377{4}),
(Symbol:'Th'; AtomicNumber:90; MassNumber:228; Mass:228.0287413{23} ; Abundance:0; {g} StandardWeight:232.0377{4}),
(Symbol:'Th'; AtomicNumber:90; MassNumber:229; Mass:229.0317627{30} ; Abundance:0; {g} StandardWeight:232.0377{4}),
(Symbol:'Th'; AtomicNumber:90; MassNumber:230; Mass:230.0331341{19} ; Abundance:0; {g} StandardWeight:232.0377{4}),
(Symbol:'Th'; AtomicNumber:90; MassNumber:231; Mass:231.0363046{19} ; Abundance:0; {g} StandardWeight:232.0377{4}),
(Symbol:'Th'; AtomicNumber:90; MassNumber:232; Mass:232.0380558{21} ; Abundance:100; {g} StandardWeight:232.0377{4}),
(Symbol:'Th'; AtomicNumber:90; MassNumber:233; Mass:233.0415823{21} ; Abundance:0; {g} StandardWeight:232.0377{4}),
(Symbol:'Th'; AtomicNumber:90; MassNumber:234; Mass:234.0436014{37} ; Abundance:0; {g} StandardWeight:232.0377{4}),
(Symbol:'Th'; AtomicNumber:90; MassNumber:235; Mass:235.047255{14} ; Abundance:0; {g} StandardWeight:232.0377{4}),
(Symbol:'Th'; AtomicNumber:90; MassNumber:236; Mass:236.049657{15} ; Abundance:0; {g} StandardWeight:232.0377{4}),
(Symbol:'Th'; AtomicNumber:90; MassNumber:237; Mass:237.053629{17} ; Abundance:0; {g} StandardWeight:232.0377{4}),
(Symbol:'Th'; AtomicNumber:90; MassNumber:238; Mass:238.05650{30#} ; Abundance:0; {g} StandardWeight:232.0377{4}),
(Symbol:'Th'; AtomicNumber:90; MassNumber:239; Mass:239.06077{43#} ; Abundance:0; {g} StandardWeight:232.0377{4}),
(Symbol:'Pa'; AtomicNumber:91; MassNumber:212; Mass:212.023203{80} ; Abundance:0;  StandardWeight:231.03588{2}),
(Symbol:'Pa'; AtomicNumber:91; MassNumber:213; Mass:213.021109{76} ; Abundance:0;  StandardWeight:231.03588{2}),
(Symbol:'Pa'; AtomicNumber:91; MassNumber:214; Mass:214.020918{82} ; Abundance:0;  StandardWeight:231.03588{2}),
(Symbol:'Pa'; AtomicNumber:91; MassNumber:215; Mass:215.019183{78} ; Abundance:0;  StandardWeight:231.03588{2}),
(Symbol:'Pa'; AtomicNumber:91; MassNumber:216; Mass:216.019109{57} ; Abundance:0;  StandardWeight:231.03588{2}),
(Symbol:'Pa'; AtomicNumber:91; MassNumber:217; Mass:217.018325{56} ; Abundance:0;  StandardWeight:231.03588{2}),
(Symbol:'Pa'; AtomicNumber:91; MassNumber:218; Mass:218.020059{20} ; Abundance:0;  StandardWeight:231.03588{2}),
(Symbol:'Pa'; AtomicNumber:91; MassNumber:219; Mass:219.019904{55} ; Abundance:0;  StandardWeight:231.03588{2}),
(Symbol:'Pa'; AtomicNumber:91; MassNumber:220; Mass:220.021705{55#} ; Abundance:0;  StandardWeight:231.03588{2}),
(Symbol:'Pa'; AtomicNumber:91; MassNumber:221; Mass:221.021875{55} ; Abundance:0;  StandardWeight:231.03588{2}),
(Symbol:'Pa'; AtomicNumber:91; MassNumber:222; Mass:222.023784{78#} ; Abundance:0;  StandardWeight:231.03588{2}),
(Symbol:'Pa'; AtomicNumber:91; MassNumber:223; Mass:223.023963{76} ; Abundance:0;  StandardWeight:231.03588{2}),
(Symbol:'Pa'; AtomicNumber:91; MassNumber:224; Mass:224.0256176{82} ; Abundance:0;  StandardWeight:231.03588{2}),
(Symbol:'Pa'; AtomicNumber:91; MassNumber:225; Mass:225.026131{76} ; Abundance:0;  StandardWeight:231.03588{2}),
(Symbol:'Pa'; AtomicNumber:91; MassNumber:226; Mass:226.027948{12} ; Abundance:0;  StandardWeight:231.03588{2}),
(Symbol:'Pa'; AtomicNumber:91; MassNumber:227; Mass:227.0288054{80} ; Abundance:0;  StandardWeight:231.03588{2}),
(Symbol:'Pa'; AtomicNumber:91; MassNumber:228; Mass:228.0310517{47} ; Abundance:0;  StandardWeight:231.03588{2}),
(Symbol:'Pa'; AtomicNumber:91; MassNumber:229; Mass:229.0320972{38} ; Abundance:0;  StandardWeight:231.03588{2}),
(Symbol:'Pa'; AtomicNumber:91; MassNumber:230; Mass:230.0345410{35} ; Abundance:0;  StandardWeight:231.03588{2}),
(Symbol:'Pa'; AtomicNumber:91; MassNumber:231; Mass:231.0358842{24} ; Abundance:100;  StandardWeight:231.03588{2}),
(Symbol:'Pa'; AtomicNumber:91; MassNumber:232; Mass:232.0385917{83} ; Abundance:0;  StandardWeight:231.03588{2}),
(Symbol:'Pa'; AtomicNumber:91; MassNumber:233; Mass:233.0402472{22} ; Abundance:0;  StandardWeight:231.03588{2}),
(Symbol:'Pa'; AtomicNumber:91; MassNumber:234; Mass:234.0433072{51} ; Abundance:0;  StandardWeight:231.03588{2}),
(Symbol:'Pa'; AtomicNumber:91; MassNumber:235; Mass:235.045399{15} ; Abundance:0;  StandardWeight:231.03588{2}),
(Symbol:'Pa'; AtomicNumber:91; MassNumber:236; Mass:236.048668{15} ; Abundance:0;  StandardWeight:231.03588{2}),
(Symbol:'Pa'; AtomicNumber:91; MassNumber:237; Mass:237.051023{14} ; Abundance:0;  StandardWeight:231.03588{2}),
(Symbol:'Pa'; AtomicNumber:91; MassNumber:238; Mass:238.054637{17} ; Abundance:0;  StandardWeight:231.03588{2}),
(Symbol:'Pa'; AtomicNumber:91; MassNumber:239; Mass:239.05726{21#} ; Abundance:0;  StandardWeight:231.03588{2}),
(Symbol:'Pa'; AtomicNumber:91; MassNumber:240; Mass:240.06098{32#} ; Abundance:0;  StandardWeight:231.03588{2}),
(Symbol:'Pa'; AtomicNumber:91; MassNumber:241; Mass:241.06408{43#} ; Abundance:0;  StandardWeight:231.03588{2}),
(Symbol:'U'; AtomicNumber:92; MassNumber:217; Mass:217.02466{11#} ; Abundance:0; {g,m} StandardWeight:238.02891{3}),
(Symbol:'U'; AtomicNumber:92; MassNumber:218; Mass:218.023523{20} ; Abundance:0; {g,m} StandardWeight:238.02891{3}),
(Symbol:'U'; AtomicNumber:92; MassNumber:219; Mass:219.024999{55} ; Abundance:0; {g,m} StandardWeight:238.02891{3}),
(Symbol:'U'; AtomicNumber:92; MassNumber:220; Mass:220.02462{11#} ; Abundance:0; {g,m} StandardWeight:238.02891{3}),
(Symbol:'U'; AtomicNumber:92; MassNumber:221; Mass:221.02628{11#} ; Abundance:0; {g,m} StandardWeight:238.02891{3}),
(Symbol:'U'; AtomicNumber:92; MassNumber:222; Mass:222.02600{11#} ; Abundance:0; {g,m} StandardWeight:238.02891{3}),
(Symbol:'U'; AtomicNumber:92; MassNumber:223; Mass:223.027739{76} ; Abundance:0; {g,m} StandardWeight:238.02891{3}),
(Symbol:'U'; AtomicNumber:92; MassNumber:224; Mass:224.027605{27} ; Abundance:0; {g,m} StandardWeight:238.02891{3}),
(Symbol:'U'; AtomicNumber:92; MassNumber:225; Mass:225.029391{13} ; Abundance:0; {g,m} StandardWeight:238.02891{3}),
(Symbol:'U'; AtomicNumber:92; MassNumber:226; Mass:226.029339{14} ; Abundance:0; {g,m} StandardWeight:238.02891{3}),
(Symbol:'U'; AtomicNumber:92; MassNumber:227; Mass:227.031157{18} ; Abundance:0; {g,m} StandardWeight:238.02891{3}),
(Symbol:'U'; AtomicNumber:92; MassNumber:228; Mass:228.031371{15} ; Abundance:0; {g,m} StandardWeight:238.02891{3}),
(Symbol:'U'; AtomicNumber:92; MassNumber:229; Mass:229.0335063{64} ; Abundance:0; {g,m} StandardWeight:238.02891{3}),
(Symbol:'U'; AtomicNumber:92; MassNumber:230; Mass:230.0339401{51} ; Abundance:0; {g,m} StandardWeight:238.02891{3}),
(Symbol:'U'; AtomicNumber:92; MassNumber:231; Mass:231.0362939{32} ; Abundance:0; {g,m} StandardWeight:238.02891{3}),
(Symbol:'U'; AtomicNumber:92; MassNumber:232; Mass:232.0371563{23} ; Abundance:0; {g,m} StandardWeight:238.02891{3}),
(Symbol:'U'; AtomicNumber:92; MassNumber:233; Mass:233.0396355{29} ; Abundance:0; {g,m} StandardWeight:238.02891{3}),
(Symbol:'U'; AtomicNumber:92; MassNumber:234; Mass:234.0409523{19} ; Abundance:0.0054; {g,m} StandardWeight:238.02891{3}),
(Symbol:'U'; AtomicNumber:92; MassNumber:235; Mass:235.0439301{19} ; Abundance:0.7204; {g,m} StandardWeight:238.02891{3}),
(Symbol:'U'; AtomicNumber:92; MassNumber:236; Mass:236.0455682{19} ; Abundance:0; {g,m} StandardWeight:238.02891{3}),
(Symbol:'U'; AtomicNumber:92; MassNumber:237; Mass:237.0487304{20} ; Abundance:0; {g,m} StandardWeight:238.02891{3}),
(Symbol:'U'; AtomicNumber:92; MassNumber:238; Mass:238.0507884{20} ; Abundance:99.2742; {g,m} StandardWeight:238.02891{3}),
(Symbol:'U'; AtomicNumber:92; MassNumber:239; Mass:239.0542935{20} ; Abundance:0; {g,m} StandardWeight:238.02891{3}),
(Symbol:'U'; AtomicNumber:92; MassNumber:240; Mass:240.0565934{57} ; Abundance:0; {g,m} StandardWeight:238.02891{3}),
(Symbol:'U'; AtomicNumber:92; MassNumber:241; Mass:241.06033{32#} ; Abundance:0; {g,m} StandardWeight:238.02891{3}),
(Symbol:'U'; AtomicNumber:92; MassNumber:242; Mass:242.06293{22#} ; Abundance:0; {g,m} StandardWeight:238.02891{3}),
(Symbol:'U'; AtomicNumber:92; MassNumber:243; Mass:243.06699{43#} ; Abundance:0; {g,m} StandardWeight:238.02891{3}),
(Symbol:'Np'; AtomicNumber:93; MassNumber:219; Mass:219.03143{21#} ; Abundance:0;  StandardWeight:237),
(Symbol:'Np'; AtomicNumber:93; MassNumber:220; Mass:220.03254{21#} ; Abundance:0;  StandardWeight:237),
(Symbol:'Np'; AtomicNumber:93; MassNumber:221; Mass:221.03204{21#} ; Abundance:0;  StandardWeight:237),
(Symbol:'Np'; AtomicNumber:93; MassNumber:222; Mass:222.03330{21#} ; Abundance:0;  StandardWeight:237),
(Symbol:'Np'; AtomicNumber:93; MassNumber:223; Mass:223.03285{21#} ; Abundance:0;  StandardWeight:237),
(Symbol:'Np'; AtomicNumber:93; MassNumber:224; Mass:224.03422{21#} ; Abundance:0;  StandardWeight:237),
(Symbol:'Np'; AtomicNumber:93; MassNumber:225; Mass:225.033911{77} ; Abundance:0;  StandardWeight:237),
(Symbol:'Np'; AtomicNumber:93; MassNumber:226; Mass:226.035188{95#} ; Abundance:0;  StandardWeight:237),
(Symbol:'Np'; AtomicNumber:93; MassNumber:227; Mass:227.034957{78} ; Abundance:0;  StandardWeight:237),
(Symbol:'Np'; AtomicNumber:93; MassNumber:228; Mass:228.036067{54} ; Abundance:0;  StandardWeight:237),
(Symbol:'Np'; AtomicNumber:93; MassNumber:229; Mass:229.036264{93} ; Abundance:0;  StandardWeight:237),
(Symbol:'Np'; AtomicNumber:93; MassNumber:230; Mass:230.037828{55} ; Abundance:0;  StandardWeight:237),
(Symbol:'Np'; AtomicNumber:93; MassNumber:231; Mass:231.038245{54} ; Abundance:0;  StandardWeight:237),
(Symbol:'Np'; AtomicNumber:93; MassNumber:232; Mass:232.04011{11#} ; Abundance:0;  StandardWeight:237),
(Symbol:'Np'; AtomicNumber:93; MassNumber:233; Mass:233.040741{55} ; Abundance:0;  StandardWeight:237),
(Symbol:'Np'; AtomicNumber:93; MassNumber:234; Mass:234.0428953{91} ; Abundance:0;  StandardWeight:237),
(Symbol:'Np'; AtomicNumber:93; MassNumber:235; Mass:235.0440635{21} ; Abundance:0;  StandardWeight:237),
(Symbol:'Np'; AtomicNumber:93; MassNumber:236; Mass:236.046570{54} ; Abundance:0;  StandardWeight:237),
(Symbol:'Np'; AtomicNumber:93; MassNumber:237; Mass:237.0481736{19} ; Abundance:0;  StandardWeight:237),
(Symbol:'Np'; AtomicNumber:93; MassNumber:238; Mass:238.0509466{19} ; Abundance:0;  StandardWeight:237),
(Symbol:'Np'; AtomicNumber:93; MassNumber:239; Mass:239.0529392{22} ; Abundance:0;  StandardWeight:237),
(Symbol:'Np'; AtomicNumber:93; MassNumber:240; Mass:240.056165{18} ; Abundance:0;  StandardWeight:237),
(Symbol:'Np'; AtomicNumber:93; MassNumber:241; Mass:241.058253{76} ; Abundance:0;  StandardWeight:237),
(Symbol:'Np'; AtomicNumber:93; MassNumber:242; Mass:242.06164{21} ; Abundance:0;  StandardWeight:237),
(Symbol:'Np'; AtomicNumber:93; MassNumber:243; Mass:243.064280{34#} ; Abundance:0;  StandardWeight:237),
(Symbol:'Np'; AtomicNumber:93; MassNumber:244; Mass:244.06785{32#} ; Abundance:0;  StandardWeight:237),
(Symbol:'Np'; AtomicNumber:93; MassNumber:245; Mass:245.07080{43#} ; Abundance:0;  StandardWeight:237),
(Symbol:'Pu'; AtomicNumber:94; MassNumber:228; Mass:228.038732{33} ; Abundance:0;  StandardWeight:244),
(Symbol:'Pu'; AtomicNumber:94; MassNumber:229; Mass:229.040144{55} ; Abundance:0;  StandardWeight:244),
(Symbol:'Pu'; AtomicNumber:94; MassNumber:230; Mass:230.039650{16} ; Abundance:0;  StandardWeight:244),
(Symbol:'Pu'; AtomicNumber:94; MassNumber:231; Mass:231.041102{28} ; Abundance:0;  StandardWeight:244),
(Symbol:'Pu'; AtomicNumber:94; MassNumber:232; Mass:232.041185{19} ; Abundance:0;  StandardWeight:244),
(Symbol:'Pu'; AtomicNumber:94; MassNumber:233; Mass:233.042998{54} ; Abundance:0;  StandardWeight:244),
(Symbol:'Pu'; AtomicNumber:94; MassNumber:234; Mass:234.0433174{75} ; Abundance:0;  StandardWeight:244),
(Symbol:'Pu'; AtomicNumber:94; MassNumber:235; Mass:235.045286{22} ; Abundance:0;  StandardWeight:244),
(Symbol:'Pu'; AtomicNumber:94; MassNumber:236; Mass:236.0460581{23} ; Abundance:0;  StandardWeight:244),
(Symbol:'Pu'; AtomicNumber:94; MassNumber:237; Mass:237.0484098{24} ; Abundance:0;  StandardWeight:244),
(Symbol:'Pu'; AtomicNumber:94; MassNumber:238; Mass:238.0495601{19} ; Abundance:0;  StandardWeight:244),
(Symbol:'Pu'; AtomicNumber:94; MassNumber:239; Mass:239.0521636{19} ; Abundance:0;  StandardWeight:244),
(Symbol:'Pu'; AtomicNumber:94; MassNumber:240; Mass:240.0538138{19} ; Abundance:0;  StandardWeight:244),
(Symbol:'Pu'; AtomicNumber:94; MassNumber:241; Mass:241.0568517{19} ; Abundance:0;  StandardWeight:244),
(Symbol:'Pu'; AtomicNumber:94; MassNumber:242; Mass:242.0587428{20} ; Abundance:0;  StandardWeight:244),
(Symbol:'Pu'; AtomicNumber:94; MassNumber:243; Mass:243.0620036{34} ; Abundance:0;  StandardWeight:244),
(Symbol:'Pu'; AtomicNumber:94; MassNumber:244; Mass:244.0642053{56} ; Abundance:0;  StandardWeight:244),
(Symbol:'Pu'; AtomicNumber:94; MassNumber:245; Mass:245.067826{15} ; Abundance:0;  StandardWeight:244),
(Symbol:'Pu'; AtomicNumber:94; MassNumber:246; Mass:246.070205{16} ; Abundance:0;  StandardWeight:244),
(Symbol:'Pu'; AtomicNumber:94; MassNumber:247; Mass:247.07419{21#} ; Abundance:0;  StandardWeight:244),
(Symbol:'Am'; AtomicNumber:95; MassNumber:230; Mass:230.04609{14#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Am'; AtomicNumber:95; MassNumber:231; Mass:231.04556{32#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Am'; AtomicNumber:95; MassNumber:232; Mass:232.04645{32#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Am'; AtomicNumber:95; MassNumber:233; Mass:233.04644{11#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Am'; AtomicNumber:95; MassNumber:234; Mass:234.04773{17#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Am'; AtomicNumber:95; MassNumber:235; Mass:235.047908{56} ; Abundance:0;  StandardWeight:0),
(Symbol:'Am'; AtomicNumber:95; MassNumber:236; Mass:236.04943{12#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Am'; AtomicNumber:95; MassNumber:237; Mass:237.049996{64#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Am'; AtomicNumber:95; MassNumber:238; Mass:238.051985{54} ; Abundance:0;  StandardWeight:0),
(Symbol:'Am'; AtomicNumber:95; MassNumber:239; Mass:239.0530247{26} ; Abundance:0;  StandardWeight:0),
(Symbol:'Am'; AtomicNumber:95; MassNumber:240; Mass:240.055300{15} ; Abundance:0;  StandardWeight:0),
(Symbol:'Am'; AtomicNumber:95; MassNumber:241; Mass:241.0568293{19} ; Abundance:0;  StandardWeight:0),
(Symbol:'Am'; AtomicNumber:95; MassNumber:242; Mass:242.0595494{19} ; Abundance:0;  StandardWeight:0),
(Symbol:'Am'; AtomicNumber:95; MassNumber:243; Mass:243.0613813{24} ; Abundance:0;  StandardWeight:0),
(Symbol:'Am'; AtomicNumber:95; MassNumber:244; Mass:244.0642851{22} ; Abundance:0;  StandardWeight:0),
(Symbol:'Am'; AtomicNumber:95; MassNumber:245; Mass:245.0664548{34} ; Abundance:0;  StandardWeight:0),
(Symbol:'Am'; AtomicNumber:95; MassNumber:246; Mass:246.069775{20#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Am'; AtomicNumber:95; MassNumber:247; Mass:247.07209{11#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Am'; AtomicNumber:95; MassNumber:248; Mass:248.07575{22#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Am'; AtomicNumber:95; MassNumber:249; Mass:249.07848{32#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Cm'; AtomicNumber:96; MassNumber:232; Mass:232.04982{22#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Cm'; AtomicNumber:96; MassNumber:233; Mass:233.050770{77} ; Abundance:0;  StandardWeight:0),
(Symbol:'Cm'; AtomicNumber:96; MassNumber:234; Mass:234.050160{20} ; Abundance:0;  StandardWeight:0),
(Symbol:'Cm'; AtomicNumber:96; MassNumber:235; Mass:235.05154{22#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Cm'; AtomicNumber:96; MassNumber:236; Mass:236.051374{20} ; Abundance:0;  StandardWeight:0),
(Symbol:'Cm'; AtomicNumber:96; MassNumber:237; Mass:237.052869{76} ; Abundance:0;  StandardWeight:0),
(Symbol:'Cm'; AtomicNumber:96; MassNumber:238; Mass:238.053081{13} ; Abundance:0;  StandardWeight:0),
(Symbol:'Cm'; AtomicNumber:96; MassNumber:239; Mass:239.054910{58} ; Abundance:0;  StandardWeight:0),
(Symbol:'Cm'; AtomicNumber:96; MassNumber:240; Mass:240.0555297{24} ; Abundance:0;  StandardWeight:0),
(Symbol:'Cm'; AtomicNumber:96; MassNumber:241; Mass:241.0576532{23} ; Abundance:0;  StandardWeight:0),
(Symbol:'Cm'; AtomicNumber:96; MassNumber:242; Mass:242.0588360{19} ; Abundance:0;  StandardWeight:0),
(Symbol:'Cm'; AtomicNumber:96; MassNumber:243; Mass:243.0613893{22} ; Abundance:0;  StandardWeight:0),
(Symbol:'Cm'; AtomicNumber:96; MassNumber:244; Mass:244.0627528{19} ; Abundance:0;  StandardWeight:0),
(Symbol:'Cm'; AtomicNumber:96; MassNumber:245; Mass:245.0654915{22} ; Abundance:0;  StandardWeight:0),
(Symbol:'Cm'; AtomicNumber:96; MassNumber:246; Mass:246.0672238{22} ; Abundance:0;  StandardWeight:0),
(Symbol:'Cm'; AtomicNumber:96; MassNumber:247; Mass:247.0703541{47} ; Abundance:0;  StandardWeight:0),
(Symbol:'Cm'; AtomicNumber:96; MassNumber:248; Mass:248.0723499{56} ; Abundance:0;  StandardWeight:0),
(Symbol:'Cm'; AtomicNumber:96; MassNumber:249; Mass:249.0759548{56} ; Abundance:0;  StandardWeight:0),
(Symbol:'Cm'; AtomicNumber:96; MassNumber:250; Mass:250.078358{12} ; Abundance:0;  StandardWeight:0),
(Symbol:'Cm'; AtomicNumber:96; MassNumber:251; Mass:251.082286{24} ; Abundance:0;  StandardWeight:0),
(Symbol:'Cm'; AtomicNumber:96; MassNumber:252; Mass:252.08487{32#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Bk'; AtomicNumber:97; MassNumber:234; Mass:234.05727{15#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Bk'; AtomicNumber:97; MassNumber:235; Mass:235.05658{43#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Bk'; AtomicNumber:97; MassNumber:236; Mass:236.05748{43#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Bk'; AtomicNumber:97; MassNumber:237; Mass:237.05710{24#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Bk'; AtomicNumber:97; MassNumber:238; Mass:238.05820{27#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Bk'; AtomicNumber:97; MassNumber:239; Mass:239.05824{22#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Bk'; AtomicNumber:97; MassNumber:240; Mass:240.05976{16#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Bk'; AtomicNumber:97; MassNumber:241; Mass:241.06016{22#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Bk'; AtomicNumber:97; MassNumber:242; Mass:242.06198{22#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Bk'; AtomicNumber:97; MassNumber:243; Mass:243.0630078{51} ; Abundance:0;  StandardWeight:0),
(Symbol:'Bk'; AtomicNumber:97; MassNumber:244; Mass:244.065181{16} ; Abundance:0;  StandardWeight:0),
(Symbol:'Bk'; AtomicNumber:97; MassNumber:245; Mass:245.0663618{24} ; Abundance:0;  StandardWeight:0),
(Symbol:'Bk'; AtomicNumber:97; MassNumber:246; Mass:246.068673{64} ; Abundance:0;  StandardWeight:0),
(Symbol:'Bk'; AtomicNumber:97; MassNumber:247; Mass:247.0703073{59} ; Abundance:0;  StandardWeight:0),
(Symbol:'Bk'; AtomicNumber:97; MassNumber:248; Mass:248.073088{76#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Bk'; AtomicNumber:97; MassNumber:249; Mass:249.0749877{27} ; Abundance:0;  StandardWeight:0),
(Symbol:'Bk'; AtomicNumber:97; MassNumber:250; Mass:250.0783167{42} ; Abundance:0;  StandardWeight:0),
(Symbol:'Bk'; AtomicNumber:97; MassNumber:251; Mass:251.080762{12} ; Abundance:0;  StandardWeight:0),
(Symbol:'Bk'; AtomicNumber:97; MassNumber:252; Mass:252.08431{22#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Bk'; AtomicNumber:97; MassNumber:253; Mass:253.08688{39#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Bk'; AtomicNumber:97; MassNumber:254; Mass:254.09060{32#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Cf'; AtomicNumber:98; MassNumber:237; Mass:237.062198{94} ; Abundance:0;  StandardWeight:0),
(Symbol:'Cf'; AtomicNumber:98; MassNumber:238; Mass:238.06149{32#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Cf'; AtomicNumber:98; MassNumber:239; Mass:239.06253{23#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Cf'; AtomicNumber:98; MassNumber:240; Mass:240.062256{20} ; Abundance:0;  StandardWeight:0),
(Symbol:'Cf'; AtomicNumber:98; MassNumber:241; Mass:241.06369{18#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Cf'; AtomicNumber:98; MassNumber:242; Mass:242.063754{14} ; Abundance:0;  StandardWeight:0),
(Symbol:'Cf'; AtomicNumber:98; MassNumber:243; Mass:243.06548{12#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Cf'; AtomicNumber:98; MassNumber:244; Mass:244.0660008{31} ; Abundance:0;  StandardWeight:0),
(Symbol:'Cf'; AtomicNumber:98; MassNumber:245; Mass:245.0680487{30} ; Abundance:0;  StandardWeight:0),
(Symbol:'Cf'; AtomicNumber:98; MassNumber:246; Mass:246.0688055{22} ; Abundance:0;  StandardWeight:0),
(Symbol:'Cf'; AtomicNumber:98; MassNumber:247; Mass:247.070965{16} ; Abundance:0;  StandardWeight:0),
(Symbol:'Cf'; AtomicNumber:98; MassNumber:248; Mass:248.0721851{57} ; Abundance:0;  StandardWeight:0),
(Symbol:'Cf'; AtomicNumber:98; MassNumber:249; Mass:249.0748539{23} ; Abundance:0;  StandardWeight:0),
(Symbol:'Cf'; AtomicNumber:98; MassNumber:250; Mass:250.0764062{22} ; Abundance:0;  StandardWeight:0),
(Symbol:'Cf'; AtomicNumber:98; MassNumber:251; Mass:251.0795886{48} ; Abundance:0;  StandardWeight:0),
(Symbol:'Cf'; AtomicNumber:98; MassNumber:252; Mass:252.0816272{56} ; Abundance:0;  StandardWeight:0),
(Symbol:'Cf'; AtomicNumber:98; MassNumber:253; Mass:253.0851345{67} ; Abundance:0;  StandardWeight:0),
(Symbol:'Cf'; AtomicNumber:98; MassNumber:254; Mass:254.087324{13} ; Abundance:0;  StandardWeight:0),
(Symbol:'Cf'; AtomicNumber:98; MassNumber:255; Mass:255.09105{22#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Cf'; AtomicNumber:98; MassNumber:256; Mass:256.09344{34#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Es'; AtomicNumber:99; MassNumber:239; Mass:239.06823{32#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Es'; AtomicNumber:99; MassNumber:240; Mass:240.06892{43#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Es'; AtomicNumber:99; MassNumber:241; Mass:241.06856{24#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Es'; AtomicNumber:99; MassNumber:242; Mass:242.06957{28#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Es'; AtomicNumber:99; MassNumber:243; Mass:243.06951{22#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Es'; AtomicNumber:99; MassNumber:244; Mass:244.07088{20#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Es'; AtomicNumber:99; MassNumber:245; Mass:245.07125{22#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Es'; AtomicNumber:99; MassNumber:246; Mass:246.07290{24#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Es'; AtomicNumber:99; MassNumber:247; Mass:247.073622{21} ; Abundance:0;  StandardWeight:0),
(Symbol:'Es'; AtomicNumber:99; MassNumber:248; Mass:248.075471{56#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Es'; AtomicNumber:99; MassNumber:249; Mass:249.076411{32#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Es'; AtomicNumber:99; MassNumber:250; Mass:250.07861{11#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Es'; AtomicNumber:99; MassNumber:251; Mass:251.0799936{67} ; Abundance:0;  StandardWeight:0),
(Symbol:'Es'; AtomicNumber:99; MassNumber:252; Mass:252.082980{54} ; Abundance:0;  StandardWeight:0),
(Symbol:'Es'; AtomicNumber:99; MassNumber:253; Mass:253.0848257{27} ; Abundance:0;  StandardWeight:0),
(Symbol:'Es'; AtomicNumber:99; MassNumber:254; Mass:254.0880222{45} ; Abundance:0;  StandardWeight:0),
(Symbol:'Es'; AtomicNumber:99; MassNumber:255; Mass:255.090275{12} ; Abundance:0;  StandardWeight:0),
(Symbol:'Es'; AtomicNumber:99; MassNumber:256; Mass:256.09360{11#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Es'; AtomicNumber:99; MassNumber:257; Mass:257.09598{44#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Es'; AtomicNumber:99; MassNumber:258; Mass:258.09952{32#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Fm'; AtomicNumber:100; MassNumber:241; Mass:241.07421{32#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Fm'; AtomicNumber:100; MassNumber:242; Mass:242.07343{43#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Fm'; AtomicNumber:100; MassNumber:243; Mass:243.07446{23#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Fm'; AtomicNumber:100; MassNumber:244; Mass:244.07404{22#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Fm'; AtomicNumber:100; MassNumber:245; Mass:245.07535{21#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Fm'; AtomicNumber:100; MassNumber:246; Mass:246.075350{17} ; Abundance:0;  StandardWeight:0),
(Symbol:'Fm'; AtomicNumber:100; MassNumber:247; Mass:247.07694{12#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Fm'; AtomicNumber:100; MassNumber:248; Mass:248.0771865{92} ; Abundance:0;  StandardWeight:0),
(Symbol:'Fm'; AtomicNumber:100; MassNumber:249; Mass:249.0789275{68} ; Abundance:0;  StandardWeight:0),
(Symbol:'Fm'; AtomicNumber:100; MassNumber:250; Mass:250.0795210{86} ; Abundance:0;  StandardWeight:0),
(Symbol:'Fm'; AtomicNumber:100; MassNumber:251; Mass:251.081540{16} ; Abundance:0;  StandardWeight:0),
(Symbol:'Fm'; AtomicNumber:100; MassNumber:252; Mass:252.0824671{61} ; Abundance:0;  StandardWeight:0),
(Symbol:'Fm'; AtomicNumber:100; MassNumber:253; Mass:253.0851846{37} ; Abundance:0;  StandardWeight:0),
(Symbol:'Fm'; AtomicNumber:100; MassNumber:254; Mass:254.0868544{30} ; Abundance:0;  StandardWeight:0),
(Symbol:'Fm'; AtomicNumber:100; MassNumber:255; Mass:255.0899640{52} ; Abundance:0;  StandardWeight:0),
(Symbol:'Fm'; AtomicNumber:100; MassNumber:256; Mass:256.0917745{78} ; Abundance:0;  StandardWeight:0),
(Symbol:'Fm'; AtomicNumber:100; MassNumber:257; Mass:257.0951061{69} ; Abundance:0;  StandardWeight:0),
(Symbol:'Fm'; AtomicNumber:100; MassNumber:258; Mass:258.09708{22#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Fm'; AtomicNumber:100; MassNumber:259; Mass:259.10060{30#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Fm'; AtomicNumber:100; MassNumber:260; Mass:260.10281{55#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Md'; AtomicNumber:101; MassNumber:245; Mass:245.08081{33#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Md'; AtomicNumber:101; MassNumber:246; Mass:246.08171{28#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Md'; AtomicNumber:101; MassNumber:247; Mass:247.08152{22#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Md'; AtomicNumber:101; MassNumber:248; Mass:248.08282{26#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Md'; AtomicNumber:101; MassNumber:249; Mass:249.08291{22#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Md'; AtomicNumber:101; MassNumber:250; Mass:250.08441{32#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Md'; AtomicNumber:101; MassNumber:251; Mass:251.084774{20} ; Abundance:0;  StandardWeight:0),
(Symbol:'Md'; AtomicNumber:101; MassNumber:252; Mass:252.08643{14#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Md'; AtomicNumber:101; MassNumber:253; Mass:253.087144{34#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Md'; AtomicNumber:101; MassNumber:254; Mass:254.08959{11#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Md'; AtomicNumber:101; MassNumber:255; Mass:255.0910841{73} ; Abundance:0;  StandardWeight:0),
(Symbol:'Md'; AtomicNumber:101; MassNumber:256; Mass:256.09389{13#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Md'; AtomicNumber:101; MassNumber:257; Mass:257.0955424{29} ; Abundance:0;  StandardWeight:0),
(Symbol:'Md'; AtomicNumber:101; MassNumber:258; Mass:258.0984315{50} ; Abundance:0;  StandardWeight:0),
(Symbol:'Md'; AtomicNumber:101; MassNumber:259; Mass:259.10051{22#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Md'; AtomicNumber:101; MassNumber:260; Mass:260.10365{34#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Md'; AtomicNumber:101; MassNumber:261; Mass:261.10583{62#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Md'; AtomicNumber:101; MassNumber:262; Mass:262.10910{45#} ; Abundance:0;  StandardWeight:0),
(Symbol:'No'; AtomicNumber:102; MassNumber:248; Mass:248.08655{24#} ; Abundance:0;  StandardWeight:0),
(Symbol:'No'; AtomicNumber:102; MassNumber:249; Mass:249.08780{30#} ; Abundance:0;  StandardWeight:0),
(Symbol:'No'; AtomicNumber:102; MassNumber:250; Mass:250.08756{22#} ; Abundance:0;  StandardWeight:0),
(Symbol:'No'; AtomicNumber:102; MassNumber:251; Mass:251.08894{12#} ; Abundance:0;  StandardWeight:0),
(Symbol:'No'; AtomicNumber:102; MassNumber:252; Mass:252.088967{10} ; Abundance:0;  StandardWeight:0),
(Symbol:'No'; AtomicNumber:102; MassNumber:253; Mass:253.0905641{75} ; Abundance:0;  StandardWeight:0),
(Symbol:'No'; AtomicNumber:102; MassNumber:254; Mass:254.090956{11} ; Abundance:0;  StandardWeight:0),
(Symbol:'No'; AtomicNumber:102; MassNumber:255; Mass:255.093191{16} ; Abundance:0;  StandardWeight:0),
(Symbol:'No'; AtomicNumber:102; MassNumber:256; Mass:256.0942829{84} ; Abundance:0;  StandardWeight:0),
(Symbol:'No'; AtomicNumber:102; MassNumber:257; Mass:257.0968878{74} ; Abundance:0;  StandardWeight:0),
(Symbol:'No'; AtomicNumber:102; MassNumber:258; Mass:258.09821{11#} ; Abundance:0;  StandardWeight:0),
(Symbol:'No'; AtomicNumber:102; MassNumber:259; Mass:259.10103{11#} ; Abundance:0;  StandardWeight:0),
(Symbol:'No'; AtomicNumber:102; MassNumber:260; Mass:260.10264{22#} ; Abundance:0;  StandardWeight:0),
(Symbol:'No'; AtomicNumber:102; MassNumber:261; Mass:261.10570{22#} ; Abundance:0;  StandardWeight:0),
(Symbol:'No'; AtomicNumber:102; MassNumber:262; Mass:262.10746{39#} ; Abundance:0;  StandardWeight:0),
(Symbol:'No'; AtomicNumber:102; MassNumber:263; Mass:263.11071{53#} ; Abundance:0;  StandardWeight:0),
(Symbol:'No'; AtomicNumber:102; MassNumber:264; Mass:264.11273{70#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Lr'; AtomicNumber:103; MassNumber:251; Mass:251.09418{32#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Lr'; AtomicNumber:103; MassNumber:252; Mass:252.09526{26#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Lr'; AtomicNumber:103; MassNumber:253; Mass:253.09509{22#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Lr'; AtomicNumber:103; MassNumber:254; Mass:254.09648{32#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Lr'; AtomicNumber:103; MassNumber:255; Mass:255.096562{19} ; Abundance:0;  StandardWeight:0),
(Symbol:'Lr'; AtomicNumber:103; MassNumber:256; Mass:256.098494{89} ; Abundance:0;  StandardWeight:0),
(Symbol:'Lr'; AtomicNumber:103; MassNumber:257; Mass:257.099418{47#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Lr'; AtomicNumber:103; MassNumber:258; Mass:258.10176{11#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Lr'; AtomicNumber:103; MassNumber:259; Mass:259.102902{76#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Lr'; AtomicNumber:103; MassNumber:260; Mass:260.10550{13#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Lr'; AtomicNumber:103; MassNumber:261; Mass:261.10688{22#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Lr'; AtomicNumber:103; MassNumber:262; Mass:262.10961{22#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Lr'; AtomicNumber:103; MassNumber:263; Mass:263.11136{30#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Lr'; AtomicNumber:103; MassNumber:264; Mass:264.11420{47#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Lr'; AtomicNumber:103; MassNumber:265; Mass:265.11619{65#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Lr'; AtomicNumber:103; MassNumber:266; Mass:266.11983{56#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Rf'; AtomicNumber:104; MassNumber:253; Mass:253.10044{44#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Rf'; AtomicNumber:104; MassNumber:254; Mass:254.10005{30#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Rf'; AtomicNumber:104; MassNumber:255; Mass:255.10127{12#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Rf'; AtomicNumber:104; MassNumber:256; Mass:256.101152{19} ; Abundance:0;  StandardWeight:0),
(Symbol:'Rf'; AtomicNumber:104; MassNumber:257; Mass:257.102918{12} ; Abundance:0;  StandardWeight:0),
(Symbol:'Rf'; AtomicNumber:104; MassNumber:258; Mass:258.103428{34} ; Abundance:0;  StandardWeight:0),
(Symbol:'Rf'; AtomicNumber:104; MassNumber:259; Mass:259.105596{78#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Rf'; AtomicNumber:104; MassNumber:260; Mass:260.10644{22#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Rf'; AtomicNumber:104; MassNumber:261; Mass:261.108773{54} ; Abundance:0;  StandardWeight:0),
(Symbol:'Rf'; AtomicNumber:104; MassNumber:262; Mass:262.10992{24#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Rf'; AtomicNumber:104; MassNumber:263; Mass:263.11249{20#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Rf'; AtomicNumber:104; MassNumber:264; Mass:264.11388{39#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Rf'; AtomicNumber:104; MassNumber:265; Mass:265.11668{39#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Rf'; AtomicNumber:104; MassNumber:266; Mass:266.11817{50#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Rf'; AtomicNumber:104; MassNumber:267; Mass:267.12179{62#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Rf'; AtomicNumber:104; MassNumber:268; Mass:268.12397{77#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Db'; AtomicNumber:105; MassNumber:255; Mass:255.10707{45#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Db'; AtomicNumber:105; MassNumber:256; Mass:256.10789{26#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Db'; AtomicNumber:105; MassNumber:257; Mass:257.10758{22#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Db'; AtomicNumber:105; MassNumber:258; Mass:258.10928{33#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Db'; AtomicNumber:105; MassNumber:259; Mass:259.109492{57} ; Abundance:0;  StandardWeight:0),
(Symbol:'Db'; AtomicNumber:105; MassNumber:260; Mass:260.11130{10#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Db'; AtomicNumber:105; MassNumber:261; Mass:261.11192{12#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Db'; AtomicNumber:105; MassNumber:262; Mass:262.11407{15#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Db'; AtomicNumber:105; MassNumber:263; Mass:263.11499{18#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Db'; AtomicNumber:105; MassNumber:264; Mass:264.11741{25#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Db'; AtomicNumber:105; MassNumber:265; Mass:265.11861{24#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Db'; AtomicNumber:105; MassNumber:266; Mass:266.12103{30#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Db'; AtomicNumber:105; MassNumber:267; Mass:267.12247{44#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Db'; AtomicNumber:105; MassNumber:268; Mass:268.12567{57#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Db'; AtomicNumber:105; MassNumber:269; Mass:269.12791{73#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Db'; AtomicNumber:105; MassNumber:270; Mass:270.13136{64#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Sg'; AtomicNumber:106; MassNumber:258; Mass:258.11298{44#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Sg'; AtomicNumber:106; MassNumber:259; Mass:259.11440{13#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Sg'; AtomicNumber:106; MassNumber:260; Mass:260.114384{22} ; Abundance:0;  StandardWeight:0),
(Symbol:'Sg'; AtomicNumber:106; MassNumber:261; Mass:261.115949{20} ; Abundance:0;  StandardWeight:0),
(Symbol:'Sg'; AtomicNumber:106; MassNumber:262; Mass:262.116337{38} ; Abundance:0;  StandardWeight:0),
(Symbol:'Sg'; AtomicNumber:106; MassNumber:263; Mass:263.11829{10#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Sg'; AtomicNumber:106; MassNumber:264; Mass:264.11893{30#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Sg'; AtomicNumber:106; MassNumber:265; Mass:265.12109{13#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Sg'; AtomicNumber:106; MassNumber:266; Mass:266.12198{26#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Sg'; AtomicNumber:106; MassNumber:267; Mass:267.12436{30#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Sg'; AtomicNumber:106; MassNumber:268; Mass:268.12539{50#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Sg'; AtomicNumber:106; MassNumber:269; Mass:269.12863{39#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Sg'; AtomicNumber:106; MassNumber:270; Mass:270.13043{60#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Sg'; AtomicNumber:106; MassNumber:271; Mass:271.13393{63#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Sg'; AtomicNumber:106; MassNumber:272; Mass:272.13589{83#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Sg'; AtomicNumber:106; MassNumber:273; Mass:273.13958{54#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Bh'; AtomicNumber:107; MassNumber:260; Mass:260.12166{26#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Bh'; AtomicNumber:107; MassNumber:261; Mass:261.12145{22#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Bh'; AtomicNumber:107; MassNumber:262; Mass:262.12297{33#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Bh'; AtomicNumber:107; MassNumber:263; Mass:263.12292{33#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Bh'; AtomicNumber:107; MassNumber:264; Mass:264.12459{19#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Bh'; AtomicNumber:107; MassNumber:265; Mass:265.12491{25#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Bh'; AtomicNumber:107; MassNumber:266; Mass:266.12679{18#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Bh'; AtomicNumber:107; MassNumber:267; Mass:267.12750{28#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Bh'; AtomicNumber:107; MassNumber:268; Mass:268.12969{41#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Bh'; AtomicNumber:107; MassNumber:269; Mass:269.13042{40#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Bh'; AtomicNumber:107; MassNumber:270; Mass:270.13336{31#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Bh'; AtomicNumber:107; MassNumber:271; Mass:271.13526{48#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Bh'; AtomicNumber:107; MassNumber:272; Mass:272.13826{58#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Bh'; AtomicNumber:107; MassNumber:273; Mass:273.14024{80#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Bh'; AtomicNumber:107; MassNumber:274; Mass:274.14355{65#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Bh'; AtomicNumber:107; MassNumber:275; Mass:275.14567{64#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Hs'; AtomicNumber:108; MassNumber:263; Mass:263.12852{14#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Hs'; AtomicNumber:108; MassNumber:264; Mass:264.128357{31} ; Abundance:0;  StandardWeight:0),
(Symbol:'Hs'; AtomicNumber:108; MassNumber:265; Mass:265.129793{26} ; Abundance:0;  StandardWeight:0),
(Symbol:'Hs'; AtomicNumber:108; MassNumber:266; Mass:266.130046{42} ; Abundance:0;  StandardWeight:0),
(Symbol:'Hs'; AtomicNumber:108; MassNumber:267; Mass:267.13167{10#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Hs'; AtomicNumber:108; MassNumber:268; Mass:268.13186{30#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Hs'; AtomicNumber:108; MassNumber:269; Mass:269.13375{13#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Hs'; AtomicNumber:108; MassNumber:270; Mass:270.13429{27#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Hs'; AtomicNumber:108; MassNumber:271; Mass:271.13717{32#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Hs'; AtomicNumber:108; MassNumber:272; Mass:272.13850{55#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Hs'; AtomicNumber:108; MassNumber:273; Mass:273.14168{40#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Hs'; AtomicNumber:108; MassNumber:274; Mass:274.14330{63#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Hs'; AtomicNumber:108; MassNumber:275; Mass:275.14667{63#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Hs'; AtomicNumber:108; MassNumber:276; Mass:276.14846{86#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Hs'; AtomicNumber:108; MassNumber:277; Mass:277.15190{58#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Mt'; AtomicNumber:109; MassNumber:265; Mass:265.13600{48#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Mt'; AtomicNumber:109; MassNumber:266; Mass:266.13737{33#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Mt'; AtomicNumber:109; MassNumber:267; Mass:267.13719{54#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Mt'; AtomicNumber:109; MassNumber:268; Mass:268.13865{25#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Mt'; AtomicNumber:109; MassNumber:269; Mass:269.13882{50#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Mt'; AtomicNumber:109; MassNumber:270; Mass:270.14033{18#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Mt'; AtomicNumber:109; MassNumber:271; Mass:271.14074{35#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Mt'; AtomicNumber:109; MassNumber:272; Mass:272.14341{52#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Mt'; AtomicNumber:109; MassNumber:273; Mass:273.14440{52#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Mt'; AtomicNumber:109; MassNumber:274; Mass:274.14724{38#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Mt'; AtomicNumber:109; MassNumber:275; Mass:275.14882{50#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Mt'; AtomicNumber:109; MassNumber:276; Mass:276.15159{59#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Mt'; AtomicNumber:109; MassNumber:277; Mass:277.15327{82#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Mt'; AtomicNumber:109; MassNumber:278; Mass:278.15631{68#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Mt'; AtomicNumber:109; MassNumber:279; Mass:279.15808{72#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Ds'; AtomicNumber:110; MassNumber:267; Mass:267.14377{15#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Ds'; AtomicNumber:110; MassNumber:268; Mass:268.14348{32#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Ds'; AtomicNumber:110; MassNumber:269; Mass:269.144752{34} ; Abundance:0;  StandardWeight:0),
(Symbol:'Ds'; AtomicNumber:110; MassNumber:270; Mass:270.144584{52} ; Abundance:0;  StandardWeight:0),
(Symbol:'Ds'; AtomicNumber:110; MassNumber:271; Mass:271.14595{10#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Ds'; AtomicNumber:110; MassNumber:272; Mass:272.14602{44#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Ds'; AtomicNumber:110; MassNumber:273; Mass:273.14856{14#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Ds'; AtomicNumber:110; MassNumber:274; Mass:274.14941{42#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Ds'; AtomicNumber:110; MassNumber:275; Mass:275.15203{45#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Ds'; AtomicNumber:110; MassNumber:276; Mass:276.15303{59#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Ds'; AtomicNumber:110; MassNumber:277; Mass:277.15591{41#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Ds'; AtomicNumber:110; MassNumber:278; Mass:278.15704{67#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Ds'; AtomicNumber:110; MassNumber:279; Mass:279.16010{64#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Ds'; AtomicNumber:110; MassNumber:280; Mass:280.16131{89#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Ds'; AtomicNumber:110; MassNumber:281; Mass:281.16451{59#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Rg'; AtomicNumber:111; MassNumber:272; Mass:272.15327{25#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Rg'; AtomicNumber:111; MassNumber:273; Mass:273.15313{56#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Rg'; AtomicNumber:111; MassNumber:274; Mass:274.15525{19#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Rg'; AtomicNumber:111; MassNumber:275; Mass:275.15594{56#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Rg'; AtomicNumber:111; MassNumber:276; Mass:276.15833{68#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Rg'; AtomicNumber:111; MassNumber:277; Mass:277.15907{61#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Rg'; AtomicNumber:111; MassNumber:278; Mass:278.16149{38#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Rg'; AtomicNumber:111; MassNumber:279; Mass:279.16272{51#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Rg'; AtomicNumber:111; MassNumber:280; Mass:280.16514{61#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Rg'; AtomicNumber:111; MassNumber:281; Mass:281.16636{89#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Rg'; AtomicNumber:111; MassNumber:282; Mass:282.16912{72#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Rg'; AtomicNumber:111; MassNumber:283; Mass:283.17054{79#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Cn'; AtomicNumber:112; MassNumber:276; Mass:276.16141{64#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Cn'; AtomicNumber:112; MassNumber:277; Mass:277.16364{15#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Cn'; AtomicNumber:112; MassNumber:278; Mass:278.16416{47#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Cn'; AtomicNumber:112; MassNumber:279; Mass:279.16654{50#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Cn'; AtomicNumber:112; MassNumber:280; Mass:280.16715{63#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Cn'; AtomicNumber:112; MassNumber:281; Mass:281.16975{42#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Cn'; AtomicNumber:112; MassNumber:282; Mass:282.17050{70#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Cn'; AtomicNumber:112; MassNumber:283; Mass:283.17327{65#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Cn'; AtomicNumber:112; MassNumber:284; Mass:284.17416{91#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Cn'; AtomicNumber:112; MassNumber:285; Mass:285.17712{60#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Uut'; AtomicNumber:113; MassNumber:278; Mass:278.17058{20#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Uut'; AtomicNumber:113; MassNumber:279; Mass:279.17095{75#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Uut'; AtomicNumber:113; MassNumber:280; Mass:280.17293{75#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Uut'; AtomicNumber:113; MassNumber:281; Mass:281.17348{75#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Uut'; AtomicNumber:113; MassNumber:282; Mass:282.17567{39#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Uut'; AtomicNumber:113; MassNumber:283; Mass:283.17657{52#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Uut'; AtomicNumber:113; MassNumber:284; Mass:284.17873{62#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Uut'; AtomicNumber:113; MassNumber:285; Mass:285.17973{89#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Uut'; AtomicNumber:113; MassNumber:286; Mass:286.18221{72#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Uut'; AtomicNumber:113; MassNumber:287; Mass:287.18339{81#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Fl'; AtomicNumber:114; MassNumber:285; Mass:285.18364{47#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Fl'; AtomicNumber:114; MassNumber:286; Mass:286.18423{71#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Fl'; AtomicNumber:114; MassNumber:287; Mass:287.18678{66#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Fl'; AtomicNumber:114; MassNumber:288; Mass:288.18757{91#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Fl'; AtomicNumber:114; MassNumber:289; Mass:289.19042{60#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Uup'; AtomicNumber:115; MassNumber:287; Mass:287.19070{52#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Uup'; AtomicNumber:115; MassNumber:288; Mass:288.19274{62#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Uup'; AtomicNumber:115; MassNumber:289; Mass:289.19363{89#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Uup'; AtomicNumber:115; MassNumber:290; Mass:290.19598{73#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Uup'; AtomicNumber:115; MassNumber:291; Mass:291.19707{88#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Lv'; AtomicNumber:116; MassNumber:289; Mass:289.19816{57#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Lv'; AtomicNumber:116; MassNumber:290; Mass:290.19864{71#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Lv'; AtomicNumber:116; MassNumber:291; Mass:291.20108{66#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Lv'; AtomicNumber:116; MassNumber:292; Mass:292.20174{91#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Lv'; AtomicNumber:116; MassNumber:293; Mass:293.20449{60#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Uus'; AtomicNumber:117; MassNumber:291; Mass:291.20553{68#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Uus'; AtomicNumber:117; MassNumber:292; Mass:292.20746{75#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Uus'; AtomicNumber:117; MassNumber:293; Mass:293.20824{89#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Uus'; AtomicNumber:117; MassNumber:294; Mass:294.21046{74#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Uuo'; AtomicNumber:118; MassNumber:293; Mass:293.21356{78#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Uuo'; AtomicNumber:118; MassNumber:294; Mass:294.21392{71#} ; Abundance:0;  StandardWeight:0),
(Symbol:'Uuo'; AtomicNumber:118; MassNumber:295; Mass:295.21624{69#} ; Abundance:0;  StandardWeight:0)
);
{$ENDIF}

{$IFDEF PREVIOUSVERSION}
const
Nuclei: array [0..348] of TNuclearInfo = (
(Symbol:'H'; AtomicNumber:1; MassNumber:1; Mass:1.007825; Abundance:99.9885; StandardWeight:1.00794),
(Symbol:'H'; AtomicNumber:1; MassNumber:2; Mass:2.014102; Abundance:0.0115; StandardWeight:1.00794),
(Symbol:'H'; AtomicNumber:1; MassNumber:3; Mass:3.016049; Abundance:0; StandardWeight:1.00794),
(Symbol:'D'; AtomicNumber:1; MassNumber:2; Mass:2.014102; Abundance:0.0115; StandardWeight:2.014102),
(Symbol:'T'; AtomicNumber:1; MassNumber:3; Mass:3.016049; Abundance:0; StandardWeight:3.016049),
(Symbol:'He'; AtomicNumber:2; MassNumber:3; Mass:3.016029; Abundance:0.000137; StandardWeight:4.002602),
(Symbol:'He'; AtomicNumber:2; MassNumber:4; Mass:4.002603; Abundance:99.99986; StandardWeight:4.002602),
(Symbol:'Li'; AtomicNumber:3; MassNumber:6; Mass:6.015122; Abundance:7.59; StandardWeight:6.941),
(Symbol:'Li'; AtomicNumber:3; MassNumber:7; Mass:7.016004; Abundance:92.41; StandardWeight:6.941),
(Symbol:'Be'; AtomicNumber:4; MassNumber:9; Mass:9.012182; Abundance:100; StandardWeight:9.012182),
(Symbol:'B'; AtomicNumber:5; MassNumber:10; Mass:10.01294; Abundance:19.9; StandardWeight:10.811),
(Symbol:'B'; AtomicNumber:5; MassNumber:11; Mass:11.00931; Abundance:80.1; StandardWeight:10.811),
(Symbol:'C'; AtomicNumber:6; MassNumber:12; Mass:12; Abundance:98.93; StandardWeight:12.0107),
(Symbol:'C'; AtomicNumber:6; MassNumber:13; Mass:13.00335; Abundance:1.07; StandardWeight:12.0107),
(Symbol:'C'; AtomicNumber:6; MassNumber:14; Mass:14.00324; Abundance:0; StandardWeight:12.0107),
(Symbol:'N'; AtomicNumber:7; MassNumber:14; Mass:14.00307; Abundance:99.632; StandardWeight:14.00674),
(Symbol:'N'; AtomicNumber:7; MassNumber:15; Mass:15.00011; Abundance:0.368; StandardWeight:14.00674),
(Symbol:'O'; AtomicNumber:8; MassNumber:16; Mass:15.99491; Abundance:99.757; StandardWeight:15.9994),
(Symbol:'O'; AtomicNumber:8; MassNumber:17; Mass:16.99913; Abundance:0.038; StandardWeight:15.9994),
(Symbol:'O'; AtomicNumber:8; MassNumber:18; Mass:17.99916; Abundance:0.205; StandardWeight:15.9994),
(Symbol:'F'; AtomicNumber:9; MassNumber:19; Mass:18.9984; Abundance:100; StandardWeight:18.9984),
(Symbol:'Ne'; AtomicNumber:10; MassNumber:20; Mass:19.99244; Abundance:90.48; StandardWeight:20.1797),
(Symbol:'Ne'; AtomicNumber:10; MassNumber:21; Mass:20.99385; Abundance:0.27; StandardWeight:20.1797),
(Symbol:'Ne'; AtomicNumber:10; MassNumber:22; Mass:21.99139; Abundance:9.25; StandardWeight:20.1797),
(Symbol:'Na'; AtomicNumber:11; MassNumber:23; Mass:22.98977; Abundance:100; StandardWeight:22.98977),
(Symbol:'Mg'; AtomicNumber:12; MassNumber:24; Mass:23.98504; Abundance:78.99; StandardWeight:24.305),
(Symbol:'Mg'; AtomicNumber:12; MassNumber:25; Mass:24.98584; Abundance:10; StandardWeight:24.305),
(Symbol:'Mg'; AtomicNumber:12; MassNumber:26; Mass:25.98259; Abundance:11.01; StandardWeight:24.305),
(Symbol:'Al'; AtomicNumber:13; MassNumber:27; Mass:26.98154; Abundance:100; StandardWeight:26.98154),
(Symbol:'Si'; AtomicNumber:14; MassNumber:28; Mass:27.97693; Abundance:92.2297; StandardWeight:28.0855),
(Symbol:'Si'; AtomicNumber:14; MassNumber:29; Mass:28.97649; Abundance:4.6832; StandardWeight:28.0855),
(Symbol:'Si'; AtomicNumber:14; MassNumber:30; Mass:29.97377; Abundance:3.0872; StandardWeight:28.0855),
(Symbol:'P'; AtomicNumber:15; MassNumber:31; Mass:30.97376; Abundance:100; StandardWeight:30.97376),
(Symbol:'S'; AtomicNumber:16; MassNumber:32; Mass:31.97207; Abundance:94.93; StandardWeight:32.066),
(Symbol:'S'; AtomicNumber:16; MassNumber:33; Mass:32.97146; Abundance:0.76; StandardWeight:32.066),
(Symbol:'S'; AtomicNumber:16; MassNumber:34; Mass:33.96787; Abundance:4.29; StandardWeight:32.066),
(Symbol:'S'; AtomicNumber:16; MassNumber:36; Mass:35.96708; Abundance:0.02; StandardWeight:32.066),
(Symbol:'Cl'; AtomicNumber:17; MassNumber:35; Mass:34.96885; Abundance:75.78; StandardWeight:35.4527),
(Symbol:'Cl'; AtomicNumber:17; MassNumber:37; Mass:36.9659; Abundance:24.22; StandardWeight:35.4527),
(Symbol:'Ar'; AtomicNumber:18; MassNumber:36; Mass:35.96755; Abundance:0.3365; StandardWeight:39.948),
(Symbol:'Ar'; AtomicNumber:18; MassNumber:38; Mass:37.96273; Abundance:0.0632; StandardWeight:39.948),
(Symbol:'Ar'; AtomicNumber:18; MassNumber:40; Mass:39.96238; Abundance:99.6003; StandardWeight:39.948),
(Symbol:'K'; AtomicNumber:19; MassNumber:39; Mass:38.96371; Abundance:93.2581; StandardWeight:39.0983),
(Symbol:'K'; AtomicNumber:19; MassNumber:40; Mass:39.964; Abundance:0.0117; StandardWeight:39.0983),
(Symbol:'K'; AtomicNumber:19; MassNumber:41; Mass:40.96183; Abundance:6.7302; StandardWeight:39.0983),
(Symbol:'Ca'; AtomicNumber:20; MassNumber:40; Mass:39.96259; Abundance:96.941; StandardWeight:40.078),
(Symbol:'Ca'; AtomicNumber:20; MassNumber:42; Mass:41.95862; Abundance:0.647; StandardWeight:40.078),
(Symbol:'Ca'; AtomicNumber:20; MassNumber:43; Mass:42.95877; Abundance:0.135; StandardWeight:40.078),
(Symbol:'Ca'; AtomicNumber:20; MassNumber:44; Mass:43.95548; Abundance:2.086; StandardWeight:40.078),
(Symbol:'Ca'; AtomicNumber:20; MassNumber:46; Mass:45.95369; Abundance:0.004; StandardWeight:40.078),
(Symbol:'Ca'; AtomicNumber:20; MassNumber:48; Mass:47.95253; Abundance:0.187; StandardWeight:40.078),
(Symbol:'Sc'; AtomicNumber:21; MassNumber:45; Mass:44.95591; Abundance:100; StandardWeight:44.95591),
(Symbol:'Ti'; AtomicNumber:22; MassNumber:46; Mass:45.95263; Abundance:8.25; StandardWeight:47.867),
(Symbol:'Ti'; AtomicNumber:22; MassNumber:47; Mass:46.95176; Abundance:7.44; StandardWeight:47.867),
(Symbol:'Ti'; AtomicNumber:22; MassNumber:48; Mass:47.94795; Abundance:73.72; StandardWeight:47.867),
(Symbol:'Ti'; AtomicNumber:22; MassNumber:49; Mass:48.94787; Abundance:5.41; StandardWeight:47.867),
(Symbol:'Ti'; AtomicNumber:22; MassNumber:50; Mass:49.94479; Abundance:5.18; StandardWeight:47.867),
(Symbol:'V'; AtomicNumber:23; MassNumber:50; Mass:49.94716; Abundance:0.25; StandardWeight:50.9415),
(Symbol:'V'; AtomicNumber:23; MassNumber:51; Mass:50.94396; Abundance:99.75; StandardWeight:50.9415),
(Symbol:'Cr'; AtomicNumber:24; MassNumber:50; Mass:49.94605; Abundance:4.345; StandardWeight:51.9961),
(Symbol:'Cr'; AtomicNumber:24; MassNumber:52; Mass:51.94051; Abundance:83.789; StandardWeight:51.9961),
(Symbol:'Cr'; AtomicNumber:24; MassNumber:53; Mass:52.94065; Abundance:9.501; StandardWeight:51.9961),
(Symbol:'Cr'; AtomicNumber:24; MassNumber:54; Mass:53.93888; Abundance:2.365; StandardWeight:51.9961),
(Symbol:'Mn'; AtomicNumber:25; MassNumber:55; Mass:54.93805; Abundance:100; StandardWeight:54.93805),
(Symbol:'Fe'; AtomicNumber:26; MassNumber:54; Mass:53.93961; Abundance:5.845; StandardWeight:55.845),
(Symbol:'Fe'; AtomicNumber:26; MassNumber:56; Mass:55.93494; Abundance:91.754; StandardWeight:55.845),
(Symbol:'Fe'; AtomicNumber:26; MassNumber:57; Mass:56.9354; Abundance:2.119; StandardWeight:55.845),
(Symbol:'Fe'; AtomicNumber:26; MassNumber:58; Mass:57.93328; Abundance:0.282; StandardWeight:55.845),
(Symbol:'Co'; AtomicNumber:27; MassNumber:59; Mass:58.9332; Abundance:100; StandardWeight:58.9332),
(Symbol:'Ni'; AtomicNumber:28; MassNumber:58; Mass:57.93535; Abundance:68.0769; StandardWeight:58.6934),
(Symbol:'Ni'; AtomicNumber:28; MassNumber:60; Mass:59.93079; Abundance:26.2231; StandardWeight:58.6934),
(Symbol:'Ni'; AtomicNumber:28; MassNumber:61; Mass:60.93106; Abundance:1.1399; StandardWeight:58.6934),
(Symbol:'Ni'; AtomicNumber:28; MassNumber:62; Mass:61.92835; Abundance:3.6345; StandardWeight:58.6934),
(Symbol:'Ni'; AtomicNumber:28; MassNumber:64; Mass:63.92797; Abundance:0.9256; StandardWeight:58.6934),
(Symbol:'Cu'; AtomicNumber:29; MassNumber:63; Mass:62.9296; Abundance:69.17; StandardWeight:63.546),
(Symbol:'Cu'; AtomicNumber:29; MassNumber:65; Mass:64.92779; Abundance:30.83; StandardWeight:63.546),
(Symbol:'Zn'; AtomicNumber:30; MassNumber:64; Mass:63.92915; Abundance:48.63; StandardWeight:65.39),
(Symbol:'Zn'; AtomicNumber:30; MassNumber:66; Mass:65.92604; Abundance:27.9; StandardWeight:65.39),
(Symbol:'Zn'; AtomicNumber:30; MassNumber:67; Mass:66.92713; Abundance:4.1; StandardWeight:65.39),
(Symbol:'Zn'; AtomicNumber:30; MassNumber:68; Mass:67.92485; Abundance:18.75; StandardWeight:65.39),
(Symbol:'Zn'; AtomicNumber:30; MassNumber:70; Mass:69.92533; Abundance:0.62; StandardWeight:65.39),
(Symbol:'Ga'; AtomicNumber:31; MassNumber:69; Mass:68.92558; Abundance:60.108; StandardWeight:69.723),
(Symbol:'Ga'; AtomicNumber:31; MassNumber:71; Mass:70.92471; Abundance:39.892; StandardWeight:69.723),
(Symbol:'Ge'; AtomicNumber:32; MassNumber:70; Mass:69.92425; Abundance:20.84; StandardWeight:72.61),
(Symbol:'Ge'; AtomicNumber:32; MassNumber:72; Mass:71.92208; Abundance:27.54; StandardWeight:72.61),
(Symbol:'Ge'; AtomicNumber:32; MassNumber:73; Mass:72.92346; Abundance:7.73; StandardWeight:72.61),
(Symbol:'Ge'; AtomicNumber:32; MassNumber:74; Mass:73.92118; Abundance:36.28; StandardWeight:72.61),
(Symbol:'Ge'; AtomicNumber:32; MassNumber:76; Mass:75.9214; Abundance:7.61; StandardWeight:72.61),
(Symbol:'As'; AtomicNumber:33; MassNumber:75; Mass:74.9216; Abundance:100; StandardWeight:74.9216),
(Symbol:'Se'; AtomicNumber:34; MassNumber:74; Mass:73.92248; Abundance:0.89; StandardWeight:78.96),
(Symbol:'Se'; AtomicNumber:34; MassNumber:76; Mass:75.91921; Abundance:9.37; StandardWeight:78.96),
(Symbol:'Se'; AtomicNumber:34; MassNumber:77; Mass:76.91991; Abundance:7.63; StandardWeight:78.96),
(Symbol:'Se'; AtomicNumber:34; MassNumber:78; Mass:77.91731; Abundance:23.77; StandardWeight:78.96),
(Symbol:'Se'; AtomicNumber:34; MassNumber:80; Mass:79.91652; Abundance:49.61; StandardWeight:78.96),
(Symbol:'Se'; AtomicNumber:34; MassNumber:82; Mass:81.9167; Abundance:8.73; StandardWeight:78.96),
(Symbol:'Br'; AtomicNumber:35; MassNumber:79; Mass:78.91834; Abundance:50.69; StandardWeight:79.904),
(Symbol:'Br'; AtomicNumber:35; MassNumber:81; Mass:80.91629; Abundance:49.31; StandardWeight:79.904),
(Symbol:'Kr'; AtomicNumber:36; MassNumber:78; Mass:77.92039; Abundance:0.35; StandardWeight:83.8),
(Symbol:'Kr'; AtomicNumber:36; MassNumber:80; Mass:79.91638; Abundance:2.28; StandardWeight:83.8),
(Symbol:'Kr'; AtomicNumber:36; MassNumber:82; Mass:81.91348; Abundance:11.58; StandardWeight:83.8),
(Symbol:'Kr'; AtomicNumber:36; MassNumber:83; Mass:82.91414; Abundance:11.49; StandardWeight:83.8),
(Symbol:'Kr'; AtomicNumber:36; MassNumber:84; Mass:83.91151; Abundance:57; StandardWeight:83.8),
(Symbol:'Kr'; AtomicNumber:36; MassNumber:86; Mass:85.91061; Abundance:17.3; StandardWeight:83.8),
(Symbol:'Rb'; AtomicNumber:37; MassNumber:85; Mass:84.91179; Abundance:72.17; StandardWeight:85.4678),
(Symbol:'Rb'; AtomicNumber:37; MassNumber:87; Mass:86.90918; Abundance:27.83; StandardWeight:85.4678),
(Symbol:'Sr'; AtomicNumber:38; MassNumber:84; Mass:83.91343; Abundance:0.56; StandardWeight:87.62),
(Symbol:'Sr'; AtomicNumber:38; MassNumber:86; Mass:85.90926; Abundance:9.86; StandardWeight:87.62),
(Symbol:'Sr'; AtomicNumber:38; MassNumber:87; Mass:86.90888; Abundance:7; StandardWeight:87.62),
(Symbol:'Sr'; AtomicNumber:38; MassNumber:88; Mass:87.90561; Abundance:82.58; StandardWeight:87.62),
(Symbol:'Y'; AtomicNumber:39; MassNumber:89; Mass:88.90585; Abundance:100; StandardWeight:88.90585),
(Symbol:'Zr'; AtomicNumber:40; MassNumber:90; Mass:89.9047; Abundance:51.45; StandardWeight:91.224),
(Symbol:'Zr'; AtomicNumber:40; MassNumber:91; Mass:90.90565; Abundance:11.22; StandardWeight:91.224),
(Symbol:'Zr'; AtomicNumber:40; MassNumber:92; Mass:91.90504; Abundance:17.15; StandardWeight:91.224),
(Symbol:'Zr'; AtomicNumber:40; MassNumber:94; Mass:93.90632; Abundance:17.38; StandardWeight:91.224),
(Symbol:'Zr'; AtomicNumber:40; MassNumber:96; Mass:95.90828; Abundance:2.8; StandardWeight:91.224),
(Symbol:'Nb'; AtomicNumber:41; MassNumber:93; Mass:92.90638; Abundance:100; StandardWeight:92.90638),
(Symbol:'Mo'; AtomicNumber:42; MassNumber:92; Mass:91.90681; Abundance:14.84; StandardWeight:95.94),
(Symbol:'Mo'; AtomicNumber:42; MassNumber:94; Mass:93.90509; Abundance:9.25; StandardWeight:95.94),
(Symbol:'Mo'; AtomicNumber:42; MassNumber:95; Mass:94.90584; Abundance:15.92; StandardWeight:95.94),
(Symbol:'Mo'; AtomicNumber:42; MassNumber:96; Mass:95.90468; Abundance:16.68; StandardWeight:95.94),
(Symbol:'Mo'; AtomicNumber:42; MassNumber:97; Mass:96.90602; Abundance:9.55; StandardWeight:95.94),
(Symbol:'Mo'; AtomicNumber:42; MassNumber:98; Mass:97.90541; Abundance:24.13; StandardWeight:95.94),
(Symbol:'Mo'; AtomicNumber:42; MassNumber:100; Mass:99.90748; Abundance:9.63; StandardWeight:95.94),
(Symbol:'Tc'; AtomicNumber:43; MassNumber:97; Mass:96.90636; Abundance:0; StandardWeight:98),
(Symbol:'Tc'; AtomicNumber:43; MassNumber:98; Mass:97.90722; Abundance:0; StandardWeight:98),
(Symbol:'Tc'; AtomicNumber:43; MassNumber:99; Mass:98.90625; Abundance:0; StandardWeight:98),
(Symbol:'Ru'; AtomicNumber:44; MassNumber:96; Mass:95.9076; Abundance:5.54; StandardWeight:101.07),
(Symbol:'Ru'; AtomicNumber:44; MassNumber:98; Mass:97.90529; Abundance:1.87; StandardWeight:101.07),
(Symbol:'Ru'; AtomicNumber:44; MassNumber:99; Mass:98.90594; Abundance:12.76; StandardWeight:101.07),
(Symbol:'Ru'; AtomicNumber:44; MassNumber:100; Mass:99.90422; Abundance:12.6; StandardWeight:101.07),
(Symbol:'Ru'; AtomicNumber:44; MassNumber:101; Mass:100.9056; Abundance:17.06; StandardWeight:101.07),
(Symbol:'Ru'; AtomicNumber:44; MassNumber:102; Mass:101.9043; Abundance:31.55; StandardWeight:101.07),
(Symbol:'Ru'; AtomicNumber:44; MassNumber:104; Mass:103.9054; Abundance:18.62; StandardWeight:101.07),
(Symbol:'Rh'; AtomicNumber:45; MassNumber:103; Mass:102.9055; Abundance:100; StandardWeight:102.9055),
(Symbol:'Pd'; AtomicNumber:46; MassNumber:102; Mass:101.9056; Abundance:1.02; StandardWeight:106.42),
(Symbol:'Pd'; AtomicNumber:46; MassNumber:104; Mass:103.904; Abundance:11.14; StandardWeight:106.42),
(Symbol:'Pd'; AtomicNumber:46; MassNumber:105; Mass:104.9051; Abundance:22.33; StandardWeight:106.42),
(Symbol:'Pd'; AtomicNumber:46; MassNumber:106; Mass:105.9035; Abundance:27.33; StandardWeight:106.42),
(Symbol:'Pd'; AtomicNumber:46; MassNumber:108; Mass:107.9039; Abundance:26.46; StandardWeight:106.42),
(Symbol:'Pd'; AtomicNumber:46; MassNumber:110; Mass:109.9052; Abundance:11.72; StandardWeight:106.42),
(Symbol:'Ag'; AtomicNumber:47; MassNumber:107; Mass:106.9051; Abundance:51.839; StandardWeight:107.8682),
(Symbol:'Ag'; AtomicNumber:47; MassNumber:109; Mass:108.9048; Abundance:48.161; StandardWeight:107.8682),
(Symbol:'Cd'; AtomicNumber:48; MassNumber:106; Mass:105.9065; Abundance:1.25; StandardWeight:112.411),
(Symbol:'Cd'; AtomicNumber:48; MassNumber:108; Mass:107.9042; Abundance:0.89; StandardWeight:112.411),
(Symbol:'Cd'; AtomicNumber:48; MassNumber:110; Mass:109.903; Abundance:12.49; StandardWeight:112.411),
(Symbol:'Cd'; AtomicNumber:48; MassNumber:111; Mass:110.9042; Abundance:12.8; StandardWeight:112.411),
(Symbol:'Cd'; AtomicNumber:48; MassNumber:112; Mass:111.9028; Abundance:24.13; StandardWeight:112.411),
(Symbol:'Cd'; AtomicNumber:48; MassNumber:113; Mass:112.9044; Abundance:12.22; StandardWeight:112.411),
(Symbol:'Cd'; AtomicNumber:48; MassNumber:114; Mass:113.9034; Abundance:28.73; StandardWeight:112.411),
(Symbol:'Cd'; AtomicNumber:48; MassNumber:116; Mass:115.9048; Abundance:7.49; StandardWeight:112.411),
(Symbol:'In'; AtomicNumber:49; MassNumber:113; Mass:112.9041; Abundance:4.29; StandardWeight:114.818),
(Symbol:'In'; AtomicNumber:49; MassNumber:115; Mass:114.9039; Abundance:95.71; StandardWeight:114.818),
(Symbol:'Sn'; AtomicNumber:50; MassNumber:112; Mass:111.9048; Abundance:0.97; StandardWeight:118.71),
(Symbol:'Sn'; AtomicNumber:50; MassNumber:114; Mass:113.9028; Abundance:0.66; StandardWeight:118.71),
(Symbol:'Sn'; AtomicNumber:50; MassNumber:115; Mass:114.9033; Abundance:0.34; StandardWeight:118.71),
(Symbol:'Sn'; AtomicNumber:50; MassNumber:116; Mass:115.9017; Abundance:14.54; StandardWeight:118.71),
(Symbol:'Sn'; AtomicNumber:50; MassNumber:117; Mass:116.903; Abundance:7.68; StandardWeight:118.71),
(Symbol:'Sn'; AtomicNumber:50; MassNumber:118; Mass:117.9016; Abundance:24.22; StandardWeight:118.71),
(Symbol:'Sn'; AtomicNumber:50; MassNumber:119; Mass:118.9033; Abundance:8.59; StandardWeight:118.71),
(Symbol:'Sn'; AtomicNumber:50; MassNumber:120; Mass:119.9022; Abundance:32.58; StandardWeight:118.71),
(Symbol:'Sn'; AtomicNumber:50; MassNumber:122; Mass:121.9034; Abundance:4.63; StandardWeight:118.71),
(Symbol:'Sn'; AtomicNumber:50; MassNumber:124; Mass:123.9053; Abundance:5.79; StandardWeight:118.71),
(Symbol:'Sb'; AtomicNumber:51; MassNumber:121; Mass:120.9038; Abundance:57.21; StandardWeight:121.76),
(Symbol:'Sb'; AtomicNumber:51; MassNumber:123; Mass:122.9042; Abundance:42.79; StandardWeight:121.76),
(Symbol:'Te'; AtomicNumber:52; MassNumber:120; Mass:119.904; Abundance:0.09; StandardWeight:127.6),
(Symbol:'Te'; AtomicNumber:52; MassNumber:122; Mass:121.903; Abundance:2.55; StandardWeight:127.6),
(Symbol:'Te'; AtomicNumber:52; MassNumber:123; Mass:122.9043; Abundance:0.89; StandardWeight:127.6),
(Symbol:'Te'; AtomicNumber:52; MassNumber:124; Mass:123.9028; Abundance:4.74; StandardWeight:127.6),
(Symbol:'Te'; AtomicNumber:52; MassNumber:125; Mass:124.9044; Abundance:7.07; StandardWeight:127.6),
(Symbol:'Te'; AtomicNumber:52; MassNumber:126; Mass:125.9033; Abundance:18.84; StandardWeight:127.6),
(Symbol:'Te'; AtomicNumber:52; MassNumber:128; Mass:127.9045; Abundance:31.74; StandardWeight:127.6),
(Symbol:'Te'; AtomicNumber:52; MassNumber:130; Mass:129.9062; Abundance:34.08; StandardWeight:127.6),
(Symbol:'I'; AtomicNumber:53; MassNumber:127; Mass:126.9045; Abundance:100; StandardWeight:126.9045),
(Symbol:'Xe'; AtomicNumber:54; MassNumber:124; Mass:123.9059; Abundance:0.09; StandardWeight:131.29),
(Symbol:'Xe'; AtomicNumber:54; MassNumber:126; Mass:125.9043; Abundance:0.09; StandardWeight:131.29),
(Symbol:'Xe'; AtomicNumber:54; MassNumber:128; Mass:127.9035; Abundance:1.92; StandardWeight:131.29),
(Symbol:'Xe'; AtomicNumber:54; MassNumber:129; Mass:128.9048; Abundance:26.44; StandardWeight:131.29),
(Symbol:'Xe'; AtomicNumber:54; MassNumber:130; Mass:129.9035; Abundance:4.08; StandardWeight:131.29),
(Symbol:'Xe'; AtomicNumber:54; MassNumber:131; Mass:130.9051; Abundance:21.18; StandardWeight:131.29),
(Symbol:'Xe'; AtomicNumber:54; MassNumber:132; Mass:131.9042; Abundance:26.89; StandardWeight:131.29),
(Symbol:'Xe'; AtomicNumber:54; MassNumber:134; Mass:133.9054; Abundance:10.44; StandardWeight:131.29),
(Symbol:'Xe'; AtomicNumber:54; MassNumber:136; Mass:135.9072; Abundance:8.87; StandardWeight:131.29),
(Symbol:'Cs'; AtomicNumber:55; MassNumber:133; Mass:132.9054; Abundance:100; StandardWeight:132.9055),
(Symbol:'Ba'; AtomicNumber:56; MassNumber:130; Mass:129.9063; Abundance:0.106; StandardWeight:137.327),
(Symbol:'Ba'; AtomicNumber:56; MassNumber:132; Mass:131.9051; Abundance:0.101; StandardWeight:137.327),
(Symbol:'Ba'; AtomicNumber:56; MassNumber:134; Mass:133.9045; Abundance:2.417; StandardWeight:137.327),
(Symbol:'Ba'; AtomicNumber:56; MassNumber:135; Mass:134.9057; Abundance:6.592; StandardWeight:137.327),
(Symbol:'Ba'; AtomicNumber:56; MassNumber:136; Mass:135.9046; Abundance:7.854; StandardWeight:137.327),
(Symbol:'Ba'; AtomicNumber:56; MassNumber:137; Mass:136.9058; Abundance:11.232; StandardWeight:137.327),
(Symbol:'Ba'; AtomicNumber:56; MassNumber:138; Mass:137.9052; Abundance:71.698; StandardWeight:137.327),
(Symbol:'La'; AtomicNumber:57; MassNumber:138; Mass:137.9071; Abundance:0.09; StandardWeight:138.9055),
(Symbol:'La'; AtomicNumber:57; MassNumber:139; Mass:138.9063; Abundance:99.91; StandardWeight:138.9055),
(Symbol:'Ce'; AtomicNumber:58; MassNumber:136; Mass:135.9071; Abundance:0.185; StandardWeight:140.116),
(Symbol:'Ce'; AtomicNumber:58; MassNumber:138; Mass:137.906; Abundance:0.251; StandardWeight:140.116),
(Symbol:'Ce'; AtomicNumber:58; MassNumber:140; Mass:139.9054; Abundance:88.45; StandardWeight:140.116),
(Symbol:'Ce'; AtomicNumber:58; MassNumber:142; Mass:141.9092; Abundance:11.114; StandardWeight:140.116),
(Symbol:'Pr'; AtomicNumber:59; MassNumber:141; Mass:140.9076; Abundance:100; StandardWeight:140.9076),
(Symbol:'Nd'; AtomicNumber:60; MassNumber:142; Mass:141.9077; Abundance:27.2; StandardWeight:144.24),
(Symbol:'Nd'; AtomicNumber:60; MassNumber:143; Mass:142.9098; Abundance:12.2; StandardWeight:144.24),
(Symbol:'Nd'; AtomicNumber:60; MassNumber:144; Mass:143.9101; Abundance:23.8; StandardWeight:144.24),
(Symbol:'Nd'; AtomicNumber:60; MassNumber:145; Mass:144.9126; Abundance:8.3; StandardWeight:144.24),
(Symbol:'Nd'; AtomicNumber:60; MassNumber:146; Mass:145.9131; Abundance:17.2; StandardWeight:144.24),
(Symbol:'Nd'; AtomicNumber:60; MassNumber:148; Mass:147.9169; Abundance:5.7; StandardWeight:144.24),
(Symbol:'Nd'; AtomicNumber:60; MassNumber:150; Mass:149.9209; Abundance:5.6; StandardWeight:144.24),
(Symbol:'Pm'; AtomicNumber:61; MassNumber:145; Mass:144.9127; Abundance:0; StandardWeight:145),
(Symbol:'Pm'; AtomicNumber:61; MassNumber:147; Mass:146.9151; Abundance:0; StandardWeight:145),
(Symbol:'Sm'; AtomicNumber:62; MassNumber:144; Mass:143.912; Abundance:3.07; StandardWeight:150.36),
(Symbol:'Sm'; AtomicNumber:62; MassNumber:147; Mass:146.9149; Abundance:14.99; StandardWeight:150.36),
(Symbol:'Sm'; AtomicNumber:62; MassNumber:148; Mass:147.9148; Abundance:11.24; StandardWeight:150.36),
(Symbol:'Sm'; AtomicNumber:62; MassNumber:149; Mass:148.9172; Abundance:13.82; StandardWeight:150.36),
(Symbol:'Sm'; AtomicNumber:62; MassNumber:150; Mass:149.9173; Abundance:7.38; StandardWeight:150.36),
(Symbol:'Sm'; AtomicNumber:62; MassNumber:152; Mass:151.9197; Abundance:26.75; StandardWeight:150.36),
(Symbol:'Sm'; AtomicNumber:62; MassNumber:154; Mass:153.9222; Abundance:22.75; StandardWeight:150.36),
(Symbol:'Eu'; AtomicNumber:63; MassNumber:151; Mass:150.9198; Abundance:47.81; StandardWeight:151.964),
(Symbol:'Eu'; AtomicNumber:63; MassNumber:153; Mass:152.9212; Abundance:52.19; StandardWeight:151.964),
(Symbol:'Gd'; AtomicNumber:64; MassNumber:152; Mass:151.9198; Abundance:0.2; StandardWeight:157.25),
(Symbol:'Gd'; AtomicNumber:64; MassNumber:154; Mass:153.9209; Abundance:2.18; StandardWeight:157.25),
(Symbol:'Gd'; AtomicNumber:64; MassNumber:155; Mass:154.9226; Abundance:14.8; StandardWeight:157.25),
(Symbol:'Gd'; AtomicNumber:64; MassNumber:156; Mass:155.9221; Abundance:20.47; StandardWeight:157.25),
(Symbol:'Gd'; AtomicNumber:64; MassNumber:157; Mass:156.924; Abundance:15.65; StandardWeight:157.25),
(Symbol:'Gd'; AtomicNumber:64; MassNumber:158; Mass:157.9241; Abundance:24.84; StandardWeight:157.25),
(Symbol:'Gd'; AtomicNumber:64; MassNumber:160; Mass:159.9271; Abundance:21.86; StandardWeight:157.25),
(Symbol:'Tb'; AtomicNumber:65; MassNumber:159; Mass:158.9253; Abundance:100; StandardWeight:158.9253),
(Symbol:'Dy'; AtomicNumber:66; MassNumber:156; Mass:155.9243; Abundance:0.06; StandardWeight:162.5),
(Symbol:'Dy'; AtomicNumber:66; MassNumber:158; Mass:157.9244; Abundance:0.1; StandardWeight:162.5),
(Symbol:'Dy'; AtomicNumber:66; MassNumber:160; Mass:159.9252; Abundance:2.34; StandardWeight:162.5),
(Symbol:'Dy'; AtomicNumber:66; MassNumber:161; Mass:160.9269; Abundance:18.91; StandardWeight:162.5),
(Symbol:'Dy'; AtomicNumber:66; MassNumber:162; Mass:161.9268; Abundance:25.51; StandardWeight:162.5),
(Symbol:'Dy'; AtomicNumber:66; MassNumber:163; Mass:162.9287; Abundance:24.9; StandardWeight:162.5),
(Symbol:'Dy'; AtomicNumber:66; MassNumber:164; Mass:163.9292; Abundance:28.18; StandardWeight:162.5),
(Symbol:'Ho'; AtomicNumber:67; MassNumber:165; Mass:164.9303; Abundance:100; StandardWeight:164.9303),
(Symbol:'Er'; AtomicNumber:68; MassNumber:162; Mass:161.9288; Abundance:0.14; StandardWeight:167.26),
(Symbol:'Er'; AtomicNumber:68; MassNumber:164; Mass:163.9292; Abundance:1.61; StandardWeight:167.26),
(Symbol:'Er'; AtomicNumber:68; MassNumber:166; Mass:165.9303; Abundance:33.61; StandardWeight:167.26),
(Symbol:'Er'; AtomicNumber:68; MassNumber:167; Mass:166.932; Abundance:22.93; StandardWeight:167.26),
(Symbol:'Er'; AtomicNumber:68; MassNumber:168; Mass:167.9324; Abundance:26.78; StandardWeight:167.26),
(Symbol:'Er'; AtomicNumber:68; MassNumber:170; Mass:169.9355; Abundance:14.93; StandardWeight:167.26),
(Symbol:'Tm'; AtomicNumber:69; MassNumber:169; Mass:168.9342; Abundance:100; StandardWeight:168.9342),
(Symbol:'Yb'; AtomicNumber:70; MassNumber:168; Mass:167.9339; Abundance:0.13; StandardWeight:173.04),
(Symbol:'Yb'; AtomicNumber:70; MassNumber:170; Mass:169.9348; Abundance:3.04; StandardWeight:173.04),
(Symbol:'Yb'; AtomicNumber:70; MassNumber:171; Mass:170.9363; Abundance:14.28; StandardWeight:173.04),
(Symbol:'Yb'; AtomicNumber:70; MassNumber:172; Mass:171.9364; Abundance:21.83; StandardWeight:173.04),
(Symbol:'Yb'; AtomicNumber:70; MassNumber:173; Mass:172.9382; Abundance:16.13; StandardWeight:173.04),
(Symbol:'Yb'; AtomicNumber:70; MassNumber:174; Mass:173.9389; Abundance:31.83; StandardWeight:173.04),
(Symbol:'Yb'; AtomicNumber:70; MassNumber:176; Mass:175.9426; Abundance:12.76; StandardWeight:173.04),
(Symbol:'Lu'; AtomicNumber:71; MassNumber:175; Mass:174.9408; Abundance:97.41; StandardWeight:174.967),
(Symbol:'Lu'; AtomicNumber:71; MassNumber:176; Mass:175.9427; Abundance:2.59; StandardWeight:174.967),
(Symbol:'Hf'; AtomicNumber:72; MassNumber:174; Mass:173.94; Abundance:0.16; StandardWeight:178.49),
(Symbol:'Hf'; AtomicNumber:72; MassNumber:176; Mass:175.9414; Abundance:5.26; StandardWeight:178.49),
(Symbol:'Hf'; AtomicNumber:72; MassNumber:177; Mass:176.9432; Abundance:18.6; StandardWeight:178.49),
(Symbol:'Hf'; AtomicNumber:72; MassNumber:178; Mass:177.9437; Abundance:27.28; StandardWeight:178.49),
(Symbol:'Hf'; AtomicNumber:72; MassNumber:179; Mass:178.9458; Abundance:13.62; StandardWeight:178.49),
(Symbol:'Hf'; AtomicNumber:72; MassNumber:180; Mass:179.9465; Abundance:35.08; StandardWeight:178.49),
(Symbol:'Ta'; AtomicNumber:73; MassNumber:180; Mass:179.9475; Abundance:0.012; StandardWeight:180.9479),
(Symbol:'Ta'; AtomicNumber:73; MassNumber:181; Mass:180.948; Abundance:99.988; StandardWeight:180.9479),
(Symbol:'W'; AtomicNumber:74; MassNumber:180; Mass:179.9467; Abundance:0.12; StandardWeight:183.84),
(Symbol:'W'; AtomicNumber:74; MassNumber:182; Mass:181.9482; Abundance:26.5; StandardWeight:183.84),
(Symbol:'W'; AtomicNumber:74; MassNumber:183; Mass:182.9502; Abundance:14.31; StandardWeight:183.84),
(Symbol:'W'; AtomicNumber:74; MassNumber:184; Mass:183.9509; Abundance:30.64; StandardWeight:183.84),
(Symbol:'W'; AtomicNumber:74; MassNumber:186; Mass:185.9544; Abundance:28.43; StandardWeight:183.84),
(Symbol:'Re'; AtomicNumber:75; MassNumber:185; Mass:184.953; Abundance:37.4; StandardWeight:186.207),
(Symbol:'Re'; AtomicNumber:75; MassNumber:187; Mass:186.9558; Abundance:62.6; StandardWeight:186.207),
(Symbol:'Os'; AtomicNumber:76; MassNumber:184; Mass:183.9525; Abundance:0.02; StandardWeight:190.23),
(Symbol:'Os'; AtomicNumber:76; MassNumber:186; Mass:185.9538; Abundance:1.59; StandardWeight:190.23),
(Symbol:'Os'; AtomicNumber:76; MassNumber:187; Mass:186.9557; Abundance:1.96; StandardWeight:190.23),
(Symbol:'Os'; AtomicNumber:76; MassNumber:188; Mass:187.9558; Abundance:13.24; StandardWeight:190.23),
(Symbol:'Os'; AtomicNumber:76; MassNumber:189; Mass:188.9581; Abundance:16.15; StandardWeight:190.23),
(Symbol:'Os'; AtomicNumber:76; MassNumber:190; Mass:189.9584; Abundance:26.26; StandardWeight:190.23),
(Symbol:'Os'; AtomicNumber:76; MassNumber:192; Mass:191.9615; Abundance:40.78; StandardWeight:190.23),
(Symbol:'Ir'; AtomicNumber:77; MassNumber:191; Mass:190.9606; Abundance:37.3; StandardWeight:192.217),
(Symbol:'Ir'; AtomicNumber:77; MassNumber:193; Mass:192.9629; Abundance:62.7; StandardWeight:192.217),
(Symbol:'Pt'; AtomicNumber:78; MassNumber:190; Mass:189.9599; Abundance:0.014; StandardWeight:195.078),
(Symbol:'Pt'; AtomicNumber:78; MassNumber:192; Mass:191.961; Abundance:0.782; StandardWeight:195.078),
(Symbol:'Pt'; AtomicNumber:78; MassNumber:194; Mass:193.9627; Abundance:32.967; StandardWeight:195.078),
(Symbol:'Pt'; AtomicNumber:78; MassNumber:195; Mass:194.9648; Abundance:33.832; StandardWeight:195.078),
(Symbol:'Pt'; AtomicNumber:78; MassNumber:196; Mass:195.9649; Abundance:25.242; StandardWeight:195.078),
(Symbol:'Pt'; AtomicNumber:78; MassNumber:198; Mass:197.9679; Abundance:7.163; StandardWeight:195.078),
(Symbol:'Au'; AtomicNumber:79; MassNumber:197; Mass:196.9666; Abundance:100; StandardWeight:196.9666),
(Symbol:'Hg'; AtomicNumber:80; MassNumber:196; Mass:195.9658; Abundance:0.15; StandardWeight:200.59),
(Symbol:'Hg'; AtomicNumber:80; MassNumber:198; Mass:197.9668; Abundance:9.97; StandardWeight:200.59),
(Symbol:'Hg'; AtomicNumber:80; MassNumber:199; Mass:198.9683; Abundance:16.87; StandardWeight:200.59),
(Symbol:'Hg'; AtomicNumber:80; MassNumber:200; Mass:199.9683; Abundance:23.1; StandardWeight:200.59),
(Symbol:'Hg'; AtomicNumber:80; MassNumber:201; Mass:200.9703; Abundance:13.18; StandardWeight:200.59),
(Symbol:'Hg'; AtomicNumber:80; MassNumber:202; Mass:201.9706; Abundance:29.86; StandardWeight:200.59),
(Symbol:'Hg'; AtomicNumber:80; MassNumber:204; Mass:203.9735; Abundance:6.87; StandardWeight:200.59),
(Symbol:'Tl'; AtomicNumber:81; MassNumber:203; Mass:202.9723; Abundance:29.524; StandardWeight:204.3833),
(Symbol:'Tl'; AtomicNumber:81; MassNumber:205; Mass:204.9744; Abundance:70.476; StandardWeight:204.3833),
(Symbol:'Pb'; AtomicNumber:82; MassNumber:204; Mass:203.973; Abundance:1.4; StandardWeight:207.2),
(Symbol:'Pb'; AtomicNumber:82; MassNumber:206; Mass:205.9744; Abundance:24.1; StandardWeight:207.2),
(Symbol:'Pb'; AtomicNumber:82; MassNumber:207; Mass:206.9759; Abundance:22.1; StandardWeight:207.2),
(Symbol:'Pb'; AtomicNumber:82; MassNumber:208; Mass:207.9766; Abundance:52.4; StandardWeight:207.2),
(Symbol:'Bi'; AtomicNumber:83; MassNumber:209; Mass:208.9804; Abundance:100; StandardWeight:208.9804),
(Symbol:'Po'; AtomicNumber:84; MassNumber:209; Mass:208.9824; Abundance:0; StandardWeight:209),
(Symbol:'Po'; AtomicNumber:84; MassNumber:210; Mass:209.9829; Abundance:0; StandardWeight:209),
(Symbol:'At'; AtomicNumber:85; MassNumber:210; Mass:209.9871; Abundance:0; StandardWeight:210),
(Symbol:'At'; AtomicNumber:85; MassNumber:211; Mass:210.9875; Abundance:0; StandardWeight:210),
(Symbol:'Rn'; AtomicNumber:86; MassNumber:211; Mass:210.9906; Abundance:0; StandardWeight:222),
(Symbol:'Rn'; AtomicNumber:86; MassNumber:220; Mass:220.0114; Abundance:0; StandardWeight:222),
(Symbol:'Rn'; AtomicNumber:86; MassNumber:222; Mass:222.0176; Abundance:0; StandardWeight:222),
(Symbol:'Fr'; AtomicNumber:87; MassNumber:223; Mass:223.0197; Abundance:0; StandardWeight:223),
(Symbol:'Ra'; AtomicNumber:88; MassNumber:223; Mass:223.0185; Abundance:0; StandardWeight:226),
(Symbol:'Ra'; AtomicNumber:88; MassNumber:224; Mass:224.0202; Abundance:0; StandardWeight:226),
(Symbol:'Ra'; AtomicNumber:88; MassNumber:226; Mass:226.0254; Abundance:0; StandardWeight:226),
(Symbol:'Ra'; AtomicNumber:88; MassNumber:228; Mass:228.0311; Abundance:0; StandardWeight:226),
(Symbol:'Ac'; AtomicNumber:89; MassNumber:227; Mass:227.0277; Abundance:0; StandardWeight:227),
(Symbol:'Th'; AtomicNumber:90; MassNumber:230; Mass:230.0331; Abundance:0; StandardWeight:232.0381),
(Symbol:'Th'; AtomicNumber:90; MassNumber:232; Mass:232.0381; Abundance:100; StandardWeight:232.0381),
(Symbol:'Pa'; AtomicNumber:91; MassNumber:231; Mass:231.0359; Abundance:100; StandardWeight:231.0359),
(Symbol:'U'; AtomicNumber:92; MassNumber:233; Mass:233.0396; Abundance:0; StandardWeight:238.0289),
(Symbol:'U'; AtomicNumber:92; MassNumber:234; Mass:234.0409; Abundance:0.0055; StandardWeight:238.0289),
(Symbol:'U'; AtomicNumber:92; MassNumber:235; Mass:235.0439; Abundance:0.72; StandardWeight:238.0289),
(Symbol:'U'; AtomicNumber:92; MassNumber:236; Mass:236.0456; Abundance:0; StandardWeight:238.0289),
(Symbol:'U'; AtomicNumber:92; MassNumber:238; Mass:238.0508; Abundance:99.2745; StandardWeight:238.0289),
(Symbol:'Np'; AtomicNumber:93; MassNumber:237; Mass:237.0482; Abundance:0; StandardWeight:237),
(Symbol:'Np'; AtomicNumber:93; MassNumber:239; Mass:239.0529; Abundance:0; StandardWeight:237),
(Symbol:'Pu'; AtomicNumber:94; MassNumber:238; Mass:238.0496; Abundance:0; StandardWeight:244),
(Symbol:'Pu'; AtomicNumber:94; MassNumber:239; Mass:239.0522; Abundance:0; StandardWeight:244),
(Symbol:'Pu'; AtomicNumber:94; MassNumber:240; Mass:240.0538; Abundance:0; StandardWeight:244),
(Symbol:'Pu'; AtomicNumber:94; MassNumber:241; Mass:241.0568; Abundance:0; StandardWeight:244),
(Symbol:'Pu'; AtomicNumber:94; MassNumber:242; Mass:242.0587; Abundance:0; StandardWeight:244),
(Symbol:'Pu'; AtomicNumber:94; MassNumber:244; Mass:244.0642; Abundance:0; StandardWeight:244),
(Symbol:'Am'; AtomicNumber:95; MassNumber:241; Mass:241.0568; Abundance:0; StandardWeight:243),
(Symbol:'Am'; AtomicNumber:95; MassNumber:243; Mass:243.0614; Abundance:0; StandardWeight:243),
(Symbol:'Cm'; AtomicNumber:96; MassNumber:243; Mass:243.0614; Abundance:0; StandardWeight:247),
(Symbol:'Cm'; AtomicNumber:96; MassNumber:244; Mass:244.0627; Abundance:0; StandardWeight:247),
(Symbol:'Cm'; AtomicNumber:96; MassNumber:245; Mass:245.0655; Abundance:0; StandardWeight:247),
(Symbol:'Cm'; AtomicNumber:96; MassNumber:246; Mass:246.0672; Abundance:0; StandardWeight:247),
(Symbol:'Cm'; AtomicNumber:96; MassNumber:247; Mass:247.0703; Abundance:0; StandardWeight:247),
(Symbol:'Cm'; AtomicNumber:96; MassNumber:248; Mass:248.0723; Abundance:0; StandardWeight:247),
(Symbol:'Bk'; AtomicNumber:97; MassNumber:247; Mass:247.0703; Abundance:0; StandardWeight:247),
(Symbol:'Bk'; AtomicNumber:97; MassNumber:249; Mass:249.075; Abundance:0; StandardWeight:247),
(Symbol:'Cf'; AtomicNumber:98; MassNumber:249; Mass:249.0748; Abundance:0; StandardWeight:251),
(Symbol:'Cf'; AtomicNumber:98; MassNumber:250; Mass:250.0764; Abundance:0; StandardWeight:251),
(Symbol:'Cf'; AtomicNumber:98; MassNumber:251; Mass:251.0796; Abundance:0; StandardWeight:251),
(Symbol:'Cf'; AtomicNumber:98; MassNumber:252; Mass:252.0816; Abundance:0; StandardWeight:251),
(Symbol:'Es'; AtomicNumber:99; MassNumber:252; Mass:252.083; Abundance:0; StandardWeight:252),
(Symbol:'Fm'; AtomicNumber:100; MassNumber:257; Mass:257.0951; Abundance:0; StandardWeight:257),
(Symbol:'Md'; AtomicNumber:101; MassNumber:256; Mass:256.094; Abundance:0; StandardWeight:258),
(Symbol:'Md'; AtomicNumber:101; MassNumber:258; Mass:258.0984; Abundance:0; StandardWeight:258),
(Symbol:'No'; AtomicNumber:102; MassNumber:259; Mass:259.101; Abundance:0; StandardWeight:259),
(Symbol:'Lr'; AtomicNumber:103; MassNumber:262; Mass:262.1097; Abundance:0; StandardWeight:262),
(Symbol:'Rf'; AtomicNumber:104; MassNumber:261; Mass:261.1087; Abundance:0; StandardWeight:261),
(Symbol:'Db'; AtomicNumber:105; MassNumber:262; Mass:262.1141; Abundance:0; StandardWeight:262),
(Symbol:'Sg'; AtomicNumber:106; MassNumber:266; Mass:266.1219; Abundance:0; StandardWeight:266),
(Symbol:'Bh'; AtomicNumber:107; MassNumber:264; Mass:264.1247; Abundance:0; StandardWeight:264),
(Symbol:'Hs'; AtomicNumber:108; MassNumber:269; Mass:269.1341; Abundance:0; StandardWeight:269),
(Symbol:'Mt'; AtomicNumber:109; MassNumber:268; Mass:268.1388; Abundance:0; StandardWeight:268),
(Symbol:'Uun'; AtomicNumber:110; MassNumber:271; Mass:271.1461; Abundance:0; StandardWeight:271),
(Symbol:'Uuu'; AtomicNumber:111; MassNumber:272; Mass:272.1535; Abundance:0; StandardWeight:272)
);
{$ENDIF}

const
  DummyNames: array [0..2] of string = ('Q', 'X', 'Du');

function TrimNucleusNumber(var Name: string):Integer;
var
  i: Integer;
begin
  { Trim trailing number and other junk from nucleus name }
  i := 0;
  while (i < Length(Name)) and (i < 3) do begin
    if not (Name[i+1] in ['A'..'Z', 'a'..'z']) then
      Break;
    Inc(i);
  end;
  if not TryStrToInt(Copy(Name,i+1,Length(Name)), Result) then
    Result := -1;
  SetLength(Name, i);
end;

function FindNucleus(ApproximateMass: Double; Name: string = ''; MinAbundance: Double = -1):TNuclearInfo;
var
  i: Integer;
begin
  TrimNucleusNumber(Name);
  if (Abs(ApproximateMass-2) < 0.5) and (Name = '') then
    Name := 'H';
  Result.Mass := -1;
  for i := 0 to High(Nuclei) do
    if (Abs(ApproximateMass-Nuclei[i].MassNumber) < 0.5) and (Nuclei[i].Abundance > MinAbundance) then
      if (Name = '') or (CompareText(Name,Nuclei[i].Symbol)=0) then begin
        if Result.Mass > 0 then begin
          { More than one match }
          if MinAbundance < 0 then
            { Try again, looking for long lived isotopes only }
            Result := FindNucleus(ApproximateMass, Name, 0)
          else
            Result.Mass := -2;
          Exit;
        end;
        Result := Nuclei[i];
      end;
end;

function FindExactMass(ApproximateMass: Double; Name: string; DummyOK: Boolean): Double;
var
  Nuc: TNuclearInfo;
begin
  if ApproximateMass <= 0 then
    Nuc := FindMostAbundantNucleus(Name, DummyOK)
  else
    Nuc := FindNucleus(ApproximateMass, Name);
  Result := Nuc.Mass;
end;

function FindMostAbundantNucleus(Name: string; DummyOK: Boolean): TNuclearInfo;
var
  i: Integer;
begin
  TrimNucleusNumber(Name);
  Result.Abundance := -1;
  for i := 0 to High(Nuclei) do
    if (CompareText(Name,Nuclei[i].Symbol)=0) and (Nuclei[i].Abundance > Result.Abundance) then
      Result := Nuclei[i];
  if (Result.Abundance < 0) and DummyOK then
    for i := 0 to High(DummyNames) do
      if (CompareText(Name,DummyNames[i])=0) then begin
        FillChar(Result, SizeOf(Result), 0);
        Result.Symbol := DummyNames[i];
        Exit;
      end;
  if Result.Abundance < 0 then
    Result.Mass := -1; { Not found }
end;

function FindMostAbundantNucleus(AtomicNumber: Integer): TNuclearInfo;
var
  i: Integer;
begin
  Result.Abundance := -1;
  for i := 0 to High(Nuclei) do
    if (AtomicNumber = Nuclei[i].AtomicNumber) and (Nuclei[i].Abundance > Result.Abundance) then
      Result := Nuclei[i];
  if Result.Abundance < 0 then
    Result.Mass := -1; { Not found }
end;

function StandardWeight(const Name: string): Double;
var
  Nuc: TNuclearInfo;
begin
  Nuc := FindMostAbundantNucleus(Name, False);
  if Nuc.Mass > 0 then
    Result := Nuc.StandardWeight
  else
    Result := -1;
end;

function TryFormulaToFloat(const s: string; out Value: Double):Boolean;
var
  p: Integer;
  ExactMass, OK: Boolean;
  Ch: Char;

  function NextC: Char;
  begin
    Inc(p);
    if p <= Length(s) then
      Result := s[p]
    else
      Result := #0;
    Ch := Result;
  end;

  function NextNumber:Integer;
  begin
    Result := Ord(Ch) - Ord('0');
    while NextC in ['0'..'9'] do
      Result := Result*10 + Ord(Ch) - Ord('0');
  end;

  function ElementList: Double;
  var
    M: Double;
    MassNo: Integer;
    Element: string;
    Exact: Boolean;
  begin
    Result := 0;
    while OK and not(Ch in [#0,')']) do begin
      if Ch = '(' then begin
        { e.g. (CH3) }
        NextC;
        M := ElementList();
        OK := OK and (Ch = ')');
        NextC;
      end else begin
        { Cl, ^Cl or ^35Cl }
        Exact := ExactMass;
        MassNo := 0;
        if Ch = '^' then begin
          Exact := True;
          if NextC in ['1'..'9'] then
            MassNo := NextNumber;
        end;
        if Ch in ['A'..'Z'] then begin
          Element := Ch;
          while NextC in ['a'..'z'] do
            Element := Element + Ch;
          if Exact then
            M := FindExactMass(MassNo, Element, False)
          else
            M := StandardWeight(Element);
          OK := OK and (M > 0);
        end else
          OK := False; { Bad element name }
      end;
      { Cl2 }
      if OK and (Ch in ['1'..'9']) then
        M := M * NextNumber;
      Result := Result + M;
    end;
  end;
begin
  Value := 0;
  p := 0;
  OK := NextC <> #0; { do not accept empty string }
  ExactMass := Pos('^', s) > 0;
  Value := ElementList;
  Result := OK;
end;

end.

Awk script for extracting data from ascii format listing:
BEGIN { CONVFMT = "%.15g" }
function valErr(v) {
  if ( match(v, /(.*)\((.*)\)/, mtch) ) {
    err = "{" mtch[2] "}";
    return mtch[1];
  } else {
    err = "";
    return v;
  }
}
/Atomic Number/ { AtomicNumber = $4 }
/Atomic Symbol/ { AtomicSymbol = $4 }
/Mass Number/ { MassNumber = $4 }
/Relative Atomic Mass/ { RelativeAtomicMass = valErr($5); MassErr = err; }
/Isotopic Composition/ { IsotopicComposition = valErr($4); CompErr = err; }
/Standard Atomic Weight/ { StandardAtomicWeight = $5 }
/Notes/ {
  Notes = $3;
  if (Notes != "") Notes = "{" Notes "}";
  AtNo[AtomicNumber]["Standard"][0] += RelativeAtomicMass * IsotopicComposition;
  if (AtNo[AtomicNumber]["Standard"][1] == "")
    AtNo[AtomicNumber]["Standard"][1] = StandardAtomicWeight
  else if ( AtNo[AtomicNumber]["Standard"][1] != StandardAtomicWeight )
    print "Error on" NR;
  AtNo[AtomicNumber][MassNumber] = "(Symbol:'" AtomicSymbol "'; AtomicNumber:" AtomicNumber "; MassNumber:" MassNumber ";" \
                                     " Mass:" RelativeAtomicMass MassErr " ; Abundance:" IsotopicComposition * 100 "; " Notes;
}
END {
  PROCINFO["sorted_in"] = "@ind_num_asc";
  for (AtomicNumber in AtNo) {
    Computed = AtNo[AtomicNumber]["Standard"][0];
    Standard = AtNo[AtomicNumber]["Standard"][1];
    delete AtNo[AtomicNumber]["Standard"];
    if ( match(Standard, /\[(.*),(.*)\]/, mtch) ) {
      if ( (Computed < mtch[1]+0) || (Computed > mtch[2]+0) )
        print "Error - " Standard, Computed;
      else
        Standard = " {" mtch[1] "-" mtch[2] "}";
      if ( match(mtch[1], /\.(.*)$/, mtch) ) {
        v = "%." length(mtch[1]) "f";
        Computed = sprintf(v, Computed);
      } else
        print "Error " mtch[1];
    } else if ( match(Standard, /\[(.*)\]/, mtch) ) {
      Computed = mtch[1];
      Standard = "";
    } else if ( match(Standard, /[0-9]*\([0-9]*\)/) ) {
      v = valErr(Standard) ;
      if ( Computed !~ v ".*" )
        print "Error:" v ":" Computed ":";
      Computed = v;
      Standard = err;
    } else if ( Standard == "" ) {
      Standard = "0";
      Computed = "";
    } else {
      print "Error with " Standard;
    }
    for (MassNumber in AtNo[AtomicNumber]) {
      v = AtNo[AtomicNumber][MassNumber] " StandardWeight:" Computed Standard "),";
      print v;
      if ( AtomicNumber == 1 && AtNo[AtomicNumber][MassNumber] !~ /'H'/ ) {
        sub(/'[A-Z]'/, "'H'", v);
        print v " Alternate ";
      }
    }
  }
}

